#! /bin/bash
# Generated automatically from options-setup.sh.in by configure.
# Script to build the predator options file

#functions
return_val="" #var used to return values from functions

#creates a directory ($1) if it doesn't exist
prompt_create_dir()  
{
   if [ -z "$1" ]; then
     echo "Parameter #1 is zero length.-"  # Also if no parameter is passed.
     return -1
   else
     if [ -d $1 ]; then
	#dir exists
	return 0
     else
	#perhaps we should make an actual prompt here
        echo "Creating directory $1"
	mkdir -p "$1"
	return $?
     fi
   fi
}

#get directory name input from user
#$1=default value, $2=name of option
get_directory_option()
{
if [ -z "$1" ]; then
  echo "-Parameter #1 is zero length.-"  # Also if no parameter is passed.
  exit -1
fi
value=$1
while [ "0" -ne "1" ]; do
    echo -ne "Enter $2 (absolute path)\n[$value]: "
    read value
    if [ -z "$value" ]; then 
	value=$1
    fi
    prompt_create_dir "$value" 
    if [ $? -eq 0 ]; then
	break
    else
	value=$1
    fi
done
return_val=$value
return 0
}


#get file name input from user, creates directories as needed
#i.e. file need not exist
#$1=default value, $2=name of option, $3=should file exist
get_file_option()
{
if [ -z "$1" ]; then
  echo "-Parameter #1 is zero length.-"  # Also if no parameter is passed.
  exit -1
fi
value=$1
while [ "0" -ne "1" ]; do
    echo -ne "Enter $2 (absolute path)\n[$value]: "
    read value
    if [ -z "$value" ]; then 
	value=$1
    fi
    if [ -z "$3" ]; then
	dir=`dirname $value`
	prompt_create_dir "$dir" 
	if [ $? -eq 0 ]; then
	    break
	fi
    else
	if [ -f $value ]; then
	    break
	else
	    echo "File could not be found. Please try again"
	fi
    fi
    value=$1
done
return_val=$value
return 0
}

#Reads an integer from a prompt
#$1=default value, $2=name of option
get_integer_option()
{
if [ -z "$1" ]; then
  echo "-Parameter #1 is zero length.-"  # Also if no parameter is passed.
  exit -1
fi
value=$1
while [ "0" -ne "1" ]; do
    echo -ne "Enter $2\n[$value]: "
    read value
    if [ -z "$value" ]; then 
	value=$1
    fi
    if [ $value -gt 0 ]; then
	break;
    fi
    value=$1
done
return_val=$value
return 0
}

#options file
#default values
options_file="/homes/mhammad/VDBMS1.1/Predator-2002-03-11/src/predator.options"
regression_file="/homes/mhammad/VDBMS1.1/Predator-2002-03-11/lair/suites/regressiontest.sh"
def_temp_dir="/homes/mhammad/VDBMS1.1/Predator-2002-03-11/tmp"
def_diskrw_file="/homes/mhammad/VDBMS1.1/shore-interim-2/installed/bin/diskrw"
def_defdev_size=100000
def_reldisk_size=100000
def_log_maxsize=50000

# main script
echo -e "\nThis script will build the options file for predator."
echo "The file is: $options_file"
echo -e "\nWhen prompted for a value a default value is provided in square brackets.\nHitting <Enter> will use the default value\n"

#  XXXTMPDATADIR
get_directory_option "$def_temp_dir" "XXXTMPDATADIR"
temp_dir=$return_val

#  XXXSHOREDEFDEVPATH
get_file_option "$temp_dir/defdisk" "XXXSHOREDEFDEVPATH"
defdev_file=$return_val

#  XXXSHOREDEFDEVSIZE
get_integer_option "$def_defdev_size" "XXXSHOREDEFDEVSIZE"
defdev_size=$return_val

#  XXXSHORERELDISKPATH
get_file_option "$temp_dir/reldisk" "XXXSHORERELDISKPATH"
reldisk_file=$return_val

#  XXXSHORERELDISKSIZE
get_integer_option "$def_reldisk_size" "XXXSHORERELDISKSIZE"
reldisk_size=$return_val

#  XXXSHOREDISKRW
if [ "linux" = "linux" ]; then
    get_file_option "$def_diskrw_file" "XXXSHOREDISKRW" "fileshouldexist"
    diskrw_file=$return_val
else
    diskrw_file="not_needed"
fi

#  XXXSHORELOGDIR
get_directory_option "$temp_dir/log" "XXXSHORELOGDIR"
log_dir=$return_val

#  XXXSHOREMAXLOGSIZE
get_integer_option "$def_log_maxsize" "XXXSHOREMAXLOGSIZE"
log_maxsize=$return_val

# write the options file
echo -e "\nWriting $options_file"

#if on windows convert paths back to windows paths
if [ "linux" = "cygwin" ]; then
    opt_temp_dir=`cygpath -w -a $temp_dir`
    opt_defdev_file=`cygpath -w -a $defdev_file`
    opt_reldisk_file=`cygpath -w -a $reldisk_file`
    opt_log_dir=`cygpath -w -a $log_dir` 
else
    opt_temp_dir=$temp_dir
    opt_defdev_file=$defdev_file
    opt_reldisk_file=$reldisk_file
    opt_log_dir=$log_dir
fi

cat << EOF > $options_file
XXXTMPDATADIR=$opt_temp_dir
XXXSHOREDEFDEVPATH=$opt_defdev_file
XXXSHOREDEFDEVSIZE=$defdev_size
XXXSHORERELDISKPATH=$opt_reldisk_file
XXXSHORERELDISKSIZE=$reldisk_size
XXXSHOREDISKRW=$diskrw_file
XXXSHORELOGDIR=$opt_log_dir
XXXSHOREMAXLOGSIZE=$log_maxsize
EOF

#write the regression test startup script
echo -e "\nWriting $regression_file"
cat << EOF > $regression_file 
#! /bin/bash
if [ -z "\$1" ]; then
    echo "No host specified. Trying \$HOST" 
    serverhost=\$HOST
else
    serverhost=\$1
fi
/homes/mhammad/VDBMS1.1/Predator-2002-03-11/lair/suites/runtest.pl $temp_dir \$serverhost
exit 0
EOF

chmod a+x $regression_file

exit 0

