#include <sys/types.h>
//#include <sys/uio.h>
#include <unistd.h>
#include <string.h>
#include <io.h>
#include "cgiClient.h"

Status FormInput::ArgStruct::AssignValue(const char *FormArg) 
{
  if(!FormArg){
   cerr << "Bad form argument : " << FormArg << endl;
   return PRD_BAD;
  }
 char *ValuePtr = strchr(FormArg, '=');
 if (!ValuePtr) {
   cerr << "Bad form argument : " << FormArg << endl;
   return PRD_BAD;
  }
 ArgName = new char[ValuePtr - FormArg + 1];
 strncpy(ArgName, FormArg, ValuePtr - FormArg);
 ArgName[ValuePtr - FormArg] = 0;

 ArgValue = new char[strlen(ValuePtr+1) + 1];
 strncpy(ArgValue, ValuePtr+1, strlen(ValuePtr+1));
 ArgValue[strlen(ValuePtr+1)] = 0;

 // Call decipher on the value parsed out of the input since it is encoded
 decipher(ArgValue);

 // cerr << "ArgName: " << ArgName << ", ArgValue: " << ArgValue << endl;
 return PRD_GOOD;
}


Status FormInput::parseFormArgs(const char *FArgs)
{
 // first make a copy of the args
  if(!FArgs){
    cerr << "Bad form argument : " << FArgs << endl;
    return PRD_BAD;
  }

 char *FormArgs = new char[strlen(FArgs) +1];
 strcpy(FormArgs, FArgs);
 FormArgs[strlen(FArgs)] = 0;

 // first find the number of arguments (separated by "&")
 int i;
 char *CurrentPtr = FormArgs;

 for (i = 0; ; i++) {
   if(!CurrentPtr) {
     cerr << "Bad form input" << endl;
     return PRD_BAD;
   }
   CurrentPtr = strchr(CurrentPtr, '&');
   if (CurrentPtr == NULL)
     break;
   CurrentPtr++;
  } // for

 numArgs = i+1;
 argArray = new  ArgStruct[numArgs];
 // now actually read each argument
 CurrentPtr = FormArgs;
 char *Arg = strtok(CurrentPtr, "&");
 if (!Arg) { 
   cerr << "No arguments!" << endl;
   return PRD_BAD;
  }

 if (argArray[0].AssignValue(Arg) != PRD_GOOD) {
   cerr << "Error assigning value 0" << endl;
   return PRD_BAD;
  }
 for (i = 1; i < numArgs; i++) {
   char *Arg = strtok(NULL, "&");
   if (argArray[i].AssignValue(Arg) != PRD_GOOD) {
     cerr << "Error assigning value " << i << endl;
     return PRD_BAD;
    }
  }
 delete [] FormArgs;
 return PRD_GOOD;
}


char FormInput::ArgStruct::hex2char(char *what) {
   register char digit;

   digit = (what[0] >= 'A' ? ((what[0] & 0xdf) - 'A')+10 : (what[0] - '0'));
   digit *= 16;
   digit += (what[1] >= 'A' ? ((what[1] & 0xdf) - 'A')+10 : (what[1] - '0'));
   return(digit);
}


void FormInput::ArgStruct::decipher(char *url) {
    register int i,j,k,z;

    // Change all plusses back to spaces
    for(k=0; url[k]; k++){
      if(url[k] == '+') url[k] = ' ' ;
    }
    for(i=0,j=0; url[j]; ++i,++j) {
        if((url[i] = url[j]) == '%') {
            url[i] = hex2char(&url[j+1]) ;
            j+= 2 ;
        }
    }
    // Change all double quotes to single quotes since they symbolize an 
	// end of a string to PREDATOR
	// TODO HACK!!  This must be changed in the server somehow
    for(z=0; url[z]; z++){
      if(url[z] == '\"') url[z] = '\'' ;
    }
    url[i] = '\0' ;
}


bool FormInput::FindFormArg(const char *ArgName,
			    char *& ArgValue)
{
 for (int i = 0; i < numArgs; i++) {
   if (!strcmp(argArray[i].ArgName, ArgName)) {
     ArgValue = argArray[i].ArgValue;
     return true;
    }
  }
 return false;
}


bool FormInput::FindNextFormArgAndValue(const char *& ArgName,
				char *& ArgValue, int& numLeft)
{
  if(CurrentArg >= numArgs){
    numLeft = 0;
    return false;
  }
  ArgName = argArray[CurrentArg].ArgName;
  ArgValue = argArray[CurrentArg++].ArgValue;
  numLeft = numArgs - CurrentArg;
  return true;
}


FormInput::FormInput(Status& rc, ostream* LogStream)
{
 // Init local vars
 rc = PRD_BAD;
 CurrentArg = 0;
 numArgs = 0;
 argArray = NULL;
 ServerName = NULL;
 RemoteAddr = NULL;
 LogStr = LogStream;
 
 // Get the length of the incoming data
 char *ContLengthStr = getenv("CONTENT_LENGTH");
 if (!ContLengthStr) {
   cerr << "Unable to find CONTENT_LENGTH" << endl;
   return;
  }

 int ContentLength = atoi(ContLengthStr);

 if (LogStr) (*LogStr) << "set CONTENT_LENGTH=" << ContentLength <<endl<<flush;

 // This will hold the incoming data
 char *FormArgs = new char[ContentLength + 10];

 // We had earlier problems with this.  We didn't take into account the fact that some
 // systems are slower in sending their data to the program, which may read the first
 // packet and not the remaining packets.  Now we force it to keep reading packets until
 // the amount of data read corresponds to the 'content length'
 int TotalBytesRead = 0;
 int BytesRead;
 while (BytesRead = read(0, FormArgs+TotalBytesRead, ContentLength - TotalBytesRead)) {
   TotalBytesRead += BytesRead;
 }
 FormArgs[TotalBytesRead] = 0; // null-terminate
 if (LogStr) (*LogStr) << "BytesRead=" << TotalBytesRead << "::" << FormArgs << endl<<flush;

 if (TotalBytesRead != ContentLength) {
   cerr << "Read " << TotalBytesRead << " instead of " << ContentLength << endl;
   delete [] FormArgs;
   return;
  }

 // Init the form args, and parse them into (name,value) pairs
 FormArgs[TotalBytesRead]=0;
 if (parseFormArgs(FormArgs) != PRD_GOOD) {
   cerr << "Error parsing form args" << endl;
   delete [] FormArgs;
   return;
  }
 delete [] FormArgs;

 // Get environment variables that may be useful for later use
 char *EnvValue;
 EnvValue = getenv("SERVER_NAME");
 if (EnvValue) {
   ServerName = new char[strlen(EnvValue)+1];
   strcpy(ServerName, EnvValue);
  }
 else {
   cerr << "No server name in environment";
   return;
  }
 if (LogStr) (*LogStr) << "set SERVER_NAME=" << ServerName <<endl<<flush;

 EnvValue = getenv("SERVER_PORT");
 if (EnvValue) {
   ServerPort = atoi(EnvValue);
  }
 else {
   cerr << "No server port in environment";
   return;
  }
 if (LogStr) (*LogStr) << "set SERVER_PORT=" << ServerPort <<endl<<flush;

 EnvValue = getenv("REMOTE_ADDR");
 if (EnvValue) {
   RemoteAddr = new char[strlen(EnvValue)+1];
   strcpy(RemoteAddr, EnvValue);
  }
 else {
   cerr << "No remote addr in environment";
   return;
 }
 if (LogStr) (*LogStr) << "set REMOTE_ADDR=" << RemoteAddr <<endl<<flush;
 
 rc = PRD_GOOD;
}


void SendEmail(const char* Recipient, char* Message, char* DocumentName)
{
  char Command[500];
  sprintf(Command, "/usr/ucb/mail -s \"%s\" %s < %s", Message, Recipient, DocumentName);
  if (system(Command)) {
    cerr << "Error in the emailing functionality" << endl;
  }
}
