#ifndef XXX_CGICLIENT_H
#define XXX_CGICLIENT_H

#include <iostream.h>
#include <strstream.h>
#include <fstream.h>
#include <stdlib.h>
#include <time.h>
#include "predconnect.h"

typedef enum {PRD_BAD = 0, PRD_GOOD = 1} Status;

/**
 ** FormInput Class takes a status enum type and returns PRD_GOOD on successful
 ** and PRD_BAD otherwise.
 ** Success is determined by satisfying BOTH of the following:
 **   - Reading in all of the input
 **   - Successfully parsing this data into recognizable (name,value) pairs
 ** Then the user may use this class to manipulate the form input.
 **/
class FormInput {
  class ArgStruct {
   public:
    char *ArgName;
    char *ArgValue;
    ArgStruct() { ArgName = NULL; ArgValue = NULL;}
    ~ArgStruct() {
      if (ArgName) delete [] ArgName;
      if (ArgValue) delete [] ArgValue;
     }
	/**
	** The way that form data is sent to this program is the following:
	** name=value&name2=value2&name3=value3&name....etc.
	** Here we take the input line and find the 'name', 'value' pair using
	** the equal sign as a deliminator. 
	**/
    Status AssignValue(const char *FormArg);
	/**
	** Reduce any %xx escape sequences to the characters they represent
	**/
    void decipher(char *FormArg);
	/**
	** Convert a two-char hex string into the char it represents
	**/
    char hex2char(char *FormArg);
   };
  int numArgs;
  int CurrentArg;
  ArgStruct* argArray;
  /**
  ** Handed the string of form arguments to decode for easier later access to
  ** each (name,value) pair.
  **/
  Status parseFormArgs(const char *FormArgs);

 public:
  // values of environment variables
  char *ServerName;
  int ServerPort;
  char *RemoteAddr;
  ostream* LogStr;

  FormInput(Status& rc, ostream* LogStream);
  ~FormInput() { 
    if (argArray && numArgs) delete [] argArray; 
    if (ServerName) delete [] ServerName;
    if (RemoteAddr) delete [] RemoteAddr;
   }
   /**
   ** Given the name of a specified form argument name, it returns the corresponding
   ** argument value and returns true, otherwise return false.
  **/
  bool FindFormArg(const char *ArgName,
		   char *& ArgValue);
   /**
   ** Keeps a running pointer to the current form argument, initialized to the 
   ** first form argument, and allows a 'stepping' behavior by the user.  Repeated
   ** calls to this function allows the user to scan through the inputs in order,
   ** allowing a relatively efficient way scan the inputs for a specific (name,value) pair
  **/
  bool FindNextFormArgAndValue(const char *& ArgName,
			       char *& ArgValue, int& numLeft);
 };

/**
 ** This class models the CGI side logic of a single 
 ** application form. It is a super-class.
 **/
class ApplForm {
  char applName[100];
 public:
  //// virtual root class for any execution data
  class ExecutionData {
   public:
    ExecutionData() {}
    virtual ~ExecutionData() {}
    virtual const char* ErrorString() {
      return "No Error";
     }
    virtual void WriteHTML(ostream& OStr) = 0;
   };
  ApplForm(const char* Name) 
   {
    strncpy(applName, Name, 100);
   }
  virtual ~ApplForm() {}
  const char* Name() { return applName;}
  //// semantic check of the form input
  virtual Status Verify(FormInput& FI, const char*& ErrorMessage)= 0;
  virtual Status Execute(PredConnection*, FormInput& FI, 
			 ExecutionData*& ExecData) = 0;
 };

/**
 ** Generic emailing function which should only be used if in UNIX environment
 **/ 
void SendEmail(const char* Recipient, char* Message, char* DocumentName);
#endif
