#include <fstream.h>
#include "cgiClient.h"
#include "sqlWindow.h"

const char* const ERR_FRM_BADARGS = "Bad form arguments";
const char* const ERR_FRM_MISSINGARGS = "Missing form entry";
const char* const ERR_FRM_BADREF = "Unknown referer for form";
const char* const ERR_FRM_NOAPP = "Unable to instantiate sql class";

ostream* LogStr = NULL;

/**
 ** Generic output routines that create html
 **/ 
void HandleInputError(const char* Message, const char* Referer) 
{
  cout << "Content-type: text/html\n\n";
  cout << "<HTML>";
  if(!Referer){
    cout << "<H2>There was a problem parsing the inputs with your browser's POST method<BR>Please try resubmitting your requests</H2>" << endl;
  }else{
    cout << Message << "<BR>" << endl << endl;
  }
  cout << "No change made to the database" << endl;
  cout << "</HTML>";
}

void HandleInputCorrect(const char* ApplName,
			FormInput& FI,
			ApplForm::ExecutionData* ExecData) 
{
  cout << "<p></p><p align=\"center\" color=\"#FF0000\">No More Results</p>" << "</HTML>";
}

void ShutDown(int ReturnCode)
{  
  exit(ReturnCode);
}

void main()
{ 
  LogStr = new ofstream("C:\\temp\\cgisqllog.txt");

  char* DBUserName =  getenv("XXXUSERNAME");
 if (!DBUserName) DBUserName = "user";

 char* DBUserPasswd = getenv("XXXPASSWORD");
 if (!DBUserPasswd) DBUserPasswd = "pass";

  ApplForm* SQLApplication = new sqlWindow();

  if (!SQLApplication) {
    HandleInputError(ERR_FRM_NOAPP,"SQL application");
    ShutDown(0);
  }

  // Initiate the socket
  XxxSocketStarter DummyStarter;
  
  // Step 1 : check and parse inputs
  Status rc;
  FormInput FI(rc, LogStr);
  char *Referer = NULL;
  
  if (rc != PRD_GOOD) {
    HandleInputError(ERR_FRM_BADARGS, Referer);
   ShutDown(0);
  }

  if (LogStr) (*LogStr) << "Recvd User:" << DBUserName << ", Passwd:" << DBUserPasswd << flush;

  char* DBServerName;
  if (!FI.FindFormArg("HostName", DBServerName)) {
   cerr << "Error finding HostName" << endl;
   HandleInputError(ERR_FRM_MISSINGARGS, DBServerName);
   ShutDown(0);
  }

  if (LogStr) (*LogStr) << "Host:" << DBServerName << flush;

  char* DBTempServerPort;
  if (!FI.FindFormArg("PortNumber", DBTempServerPort)) {
   HandleInputError("Unable to find PortNumber",DBTempServerPort);
   ShutDown(0);
  }

  const int DBServerPort = atoi(DBTempServerPort);

  if (LogStr) (*LogStr) << "Port:" << DBServerPort << flush;

  // Step 2 : depending on the source of the form, apply
  //          appropriate action
 bool Found = false;
 if (!FI.FindFormArg("Referer", Referer)) {
   cerr << "Error finding Referer" << endl;
   HandleInputError(ERR_FRM_MISSINGARGS, Referer);
   ShutDown(0);
  }

 if (LogStr) (*LogStr) << "Referer:" << Referer << endl << flush;

 if (strcmp(SQLApplication->Name(), Referer)) {
   cerr << "Error matching Referer to application name" << endl;
   HandleInputError(ERR_FRM_BADREF, Referer);
  ShutDown(0);
 }
 
  if (LogStr) (*LogStr) << "Verifying app ..." << flush;

 // now execute the application
 const char *ErrorMessage;
 if (SQLApplication->Verify(FI, ErrorMessage) != PRD_GOOD) {
   HandleInputError(ErrorMessage, Referer);
   ShutDown(0);     
  }

 if (LogStr) (*LogStr) << " .. OK" << endl << flush;

  if (LogStr) (*LogStr) << "Connecting to db server ..." << flush;

  // create a connection to the database *forcing* ascii and schemas printing
 PredConnection* Connection = 
   new PredConnection(DBServerName, DBServerPort,
		      DBUserName, DBUserPasswd, false, true);

 if(!Connection || !Connection->IsValid()) {
   HandleInputError("Unable to connect to the server at this time.", Referer);
   ShutDown(0);
 }

 if (LogStr) (*LogStr) << " .. OK" << endl << flush;

 ApplForm::ExecutionData* ExecData = NULL;
 if (SQLApplication->Execute(Connection, FI, ExecData) != PRD_GOOD) {
   if (ExecData) {
     HandleInputError(ExecData->ErrorString(), Referer);
     delete ExecData;
    }
   else 
     HandleInputError("Error executing application", Referer);
   delete Connection;
   ShutDown(0);
  }

 // generate the output html page
 HandleInputCorrect(SQLApplication->Name(), FI, ExecData);
 // remove any intermediate data
 if (ExecData) delete ExecData;

 // shutdown the connection to the database
 delete Connection;
 
 // cleanup
 delete SQLApplication;

 ShutDown(0);
}
