#include "cgiClient.h"
#include "sqlWindow.h"

extern ostream* LogStr;

const char* sqlWindow::ERR_SQL_NOQUERY = "Missing SQL query";

Status sqlWindow::Verify(FormInput& FI, const char*& ErrorMessage)
{
  // If a query has been entered, grab it and return ok, otherwise return not ok.
  if ((!FI.FindFormArg("SQLQuery", Query)) ||
      !sqlWindow::Query[0]) {
    // otherwise, maybe it is a canned query:
    char *QueryName;
    if ((!FI.FindFormArg("SQLQuery1", QueryName)) ||
 !QueryName[0]) {
 ErrorMessage = ERR_SQL_NOQUERY ;
 return PRD_BAD;
    }
    if ((!FI.FindFormArg(QueryName, Query)) ||
 !sqlWindow::Query[0]) {
 ErrorMessage = ERR_SQL_NOQUERY ;
 return PRD_BAD;
    }
    // now see if there are any parameters in the query
    for (int i = 0; ; i++) {
      // find the parameter <QueryName>Arg<i> in the Form
      char ParamName[50];
      sprintf(ParamName, "%sArg%d", QueryName, i);
      char* Param;
      if (!FI.FindFormArg(ParamName, Param)) {
 break;
      }
      if (LogStr) {
 (*LogStr) << "Query:" << Query << endl;
 (*LogStr) << "Param:" << Param << endl;
      }
      char ArgName[10];
      sprintf(ArgName, "%%Arg%d%%", i);
      if (LogStr) {
 (*LogStr) << "ArgName:" << ArgName << endl;
      }
      // substitute the parameter for the arg placeholder
      char* NewQuery = new char[strlen(Query) + 100];
      NewQuery[0] = 0;
      // find each occurrence of the string Arg<i> in the query
      // and replace it by Param
      char* SubStrLoc = Query;
      while (1) {
 char* NewSubStrLoc = strstr(SubStrLoc, ArgName);
 if (!NewSubStrLoc) break;
 strncat(NewQuery, SubStrLoc, (int)abs(SubStrLoc - NewSubStrLoc));
 strcat(NewQuery, Param);
 SubStrLoc = NewSubStrLoc + strlen(ArgName);
      }

      strcat(NewQuery, SubStrLoc);
      Query = NewQuery;
      if (LogStr) {
 (*LogStr) << "NewQuery:" << NewQuery << endl;
      }
    }
  }
  // try to see if there is any query prefix specified
 return PRD_GOOD;
}

Status sqlWindow::Execute(PredConnection* PredConn,
     FormInput& FI,
     ExecutionData*& ExecData)
{
  // make sure the connection exists
  if (!PredConn) return PRD_BAD;
  sqlWindowExecData* EData = new sqlWindowExecData;
  ExecData = EData;

  char Buf[8000];

  // Reset the buf for initialization purposes
  EData->resetBuf();

  // Construct and send off the query
  sprintf(Buf, "SQL: %s;", Query);

  if (!PredConn->SendRequest(Buf)) {
    cerr << "Error sending request" << endl;
    return PRD_BAD;
  }

  // To print a results page, the first line must be this for protocol's sake
  cout << "Content-type: text/html\n\n";

  // Now format the results page
  cout << "<HTML><BODY ";
  char* BackgroundImg = NULL;
  if ((FI.FindFormArg("BackgroundImage", BackgroundImg)) &&
      BackgroundImg) {
    cout << "background=\"" << BackgroundImg << "\" ";
  }
  char* BackgroundColor;
  if ((!FI.FindFormArg("BackgroundColor", BackgroundColor)) ||
      !BackgroundColor) {
    BackgroundColor =  "White";
  }
  cout << "BGCOLOR = \"" << BackgroundColor << "\" ";

  char* TextColor;
  if ((!FI.FindFormArg("TextColor", TextColor)) ||
      !TextColor) {
    TextColor =  "Black";
  }
  cout << "TEXT = \"" << TextColor << "\"> ";

  // cout << "<H2>Query sent to PREDATOR:</H2><BR>" << endl;
  char* ShowSQL;
  if (!FI.FindFormArg("DontShowSQL", ShowSQL)) {
    cout << "<CENTER><H3>" << Buf << "</H3></CENTER>" << endl;
  }
  char* CheetahLogo = NULL;
  if ((!FI.FindFormArg("CheetahLogo", CheetahLogo)) ||
      !CheetahLogo) {
    CheetahLogo =
"http://www.cs.cornell.edu/Info/Projects/PREDATOR/cheetah.gif";
  }
  // TEMP cout  << "<CENTER><IMG SRC=\"" << CheetahLogo << "\" WIDTH=\"20%\"> </CENTER>" << endl;
  cout.flush();

  // TEMP
  // PredConn->PrintReply(cout); 
  bool AtEnd = false;
  bool TimeOut = true;
  char AnsBuf[5000];
  while (!AtEnd) { 
	  cout.flush();
	  if (PredConn->CreateBufferScan(AnsBuf, TimeOut, AtEnd)) {
		  cout << AnsBuf ;
		  cout.flush();
	  } else {
		  AtEnd = true;
	  }

  }
  return PRD_GOOD;

 // Create the ReplyScan object to aid in interpretation of the results from the server
  ReplyScan* replyScan = PredConn->CreateReplyScan();
  cout << "reply scan created" << "<BR><BR>" << endl;
  cout.flush();


  // if an error is returned, print it out.  To do so correctly, you must replace all newlines with breaks
  if (replyScan->IsError()) {
    char errorChar;
    cout << "<BR><H2>Error returned:</H2><BR>" << endl;
    while (!replyScan->AtEOF()) {
      errorChar = replyScan->GetNextChar();
      // Format the error message since HTML doesn't recognize newline chars
      if (errorChar == '\n') {
 cout << "<BR>" << endl;
      } else {
 cout << errorChar;
      }
    }
	cout << "</TABLE></CENTER><BR></BODY>" << flush;
    cout.flush();
    return PRD_GOOD;
  }

  // if there is no error, loop on the result buffer making a nicely formatted table of results
  // until the result buffer is empty
  // cout << "<BR><CENTER><H3>Result:</H3></CENTER><BR>" << endl;
  cout << "no error <BR><BR>" << endl;

 // counter to help control printing of schemas
  int counter = 0;
  bool UnFormattedQuery = false;
  while (!replyScan->AtEOF()) {
	  // Philippe
	  if (!cout) {
		  delete PredConn;
	  }

    if (!replyScan->AnyResults()) {
      if (counter == 0) {
 UnFormattedQuery = true;
      }
      else {
 // replyScan->PrintSchema(cout);
      }
      break;
    }
    if (counter == 0) {
      // TEMP cout << "<CENTER><TABLE BORDER = 1 BGCOLOR=\"#COCOCO\">" << endl;
      // Prints out the schema of the response, if any schema is returned
      // TEMP replyScan->PrintSchema(cout);
		  cout << "PrintSchema" << "<BR><BR>" << endl;
		  cout.flush();

      cout.flush();
      counter++;
    } else {
      counter++;
      cout << "<TR>" ;
      // TEMP replyScan->GetNextLineFormatted(cout);
		  cout << "next line" << "<BR><BR>" << endl;
		  cout.flush();
	  replyScan->GetNextLine(cout);
      // TEMP cout << "</TR>" << endl;
	  cout.flush();
    }
  }

  char ListRelsChar;
  if (UnFormattedQuery) {
    while (replyScan->AtEOF() != 1) {
		  cout << "next char" << "<BR><BR>" << endl;
		  cout.flush();
      ListRelsChar = replyScan->GetNextChar();
      if (ListRelsChar == '\n') {
 cout << "<BR>" << endl<<flush;
      } else {
 cout << ListRelsChar<<flush;
      }
    }
  } else {
    // TEMP cout << "</TABLE></CENTER><BR></BODY>" << flush;
  }

  cout << "</HTML>";
  // flush the stream to make sure the data is printed to the browser
  cout.flush();


  // close the connection to the database
  return PRD_GOOD;
}
