package JDBCDriver;

import java.sql.*;
import PREDATOR.*;
import PREDATOR.ADT.*;

/**
 * A ResultSetMetaData object can be used to find out about the types 
 * and properties of the columns in a ResultSet.
 */

public class ResultSetMetaDataImpl implements java.sql.ResultSetMetaData {

	private ResultSet set;
	private QueryResult result;
	private Statement stmt;
	

	public ResultSetMetaDataImpl(java.sql.Statement stmt, QueryResult res, ResultSet rset) {

		this.stmt = stmt;
		result = res;
		set = rset;
	
	}
	
  /**
   * What's the number of columns in the ResultSet?
   *
   * @return the number
   */
  public int getColumnCount() throws SQLException {
		String name = getTableName(0);
		StringBuffer bufname = new StringBuffer(name);
		bufname.insert(0, '"');
		bufname.append('"');
		name = bufname.toString();
		String query = "SELECT arity FROM _STABLES WHERE name = " + name;
		
		ResultSet res = stmt.executeQuery(query);
		res.next();
		String str = res.getString(0);
		try {
			return (Integer.valueOf(str).intValue());
		} catch ( NumberFormatException e ) {
			throw new SQLException(" " + e);
		}
  }

  /**
   * Is the column automatically numbered, thus read-only?
   *
   * @param column the first column is 1, the second is 2, ...
   * @return true if so
   */
  public boolean isAutoIncrement(int column) throws SQLException {
    return false;
  }

  /**
   * Does a column's case matter?
   *
   * @param column the first column is 1, the second is 2, ...
   * @return true if so
   */
  public boolean isCaseSensitive(int column) throws SQLException {
    return false;
  }	

  /**
   * Can the column be used in a where clause?
   *
   * @param column the first column is 1, the second is 2, ...
   * @return true if so
   */
  public boolean isSearchable(int column) throws SQLException {
    return false;
  }

  /**
   * Is the column a cash value?
   *
   * @param column the first column is 1, the second is 2, ...
   * @return true if so
   */
  public boolean isCurrency(int column) throws SQLException {
    return false;
  }

  /**
   * Can you put a NULL in this column?		
   *
   * @param column the first column is 1, the second is 2, ...
   * @return columnNoNulls, columnNullable or columnNullableUnknown
   */
  public int isNullable(int column) throws SQLException {
    return columnNullableUnknown;
  }

  /**
   * Does not allow NULL values.
   */
  int columnNoNulls = 0;

  /**
   * Allows NULL values.
   */
  int columnNullable = 1;

  /**
   * Nullability unknown.
   */
  int columnNullableUnknown = 2;

  /**
   * Is the column a signed number?
   *
   * @param column the first column is 1, the second is 2, ...
   * @return true if so
   */
  public boolean isSigned(int column) throws SQLException {
    return false;
  }

  /**
   * What's the column's normal max width in chars?
   *
   * @param column the first column is 1, the second is 2, ...
   * @return max width
   */
  public int getColumnDisplaySize(int column) throws SQLException {
    return (set.getString(column)).length(); 
  }

  /**
   * What's the suggested column title for use in printouts and
   * displays?
   *
   * @param column the first column is 1, the second is 2, ...
   * @return true if so 
   */
  public String getColumnLabel(int column) throws SQLException {	
		RecordStreamResult res = (RecordStreamResult)result;
		Attribute ar = res.getSchema().getAttribute(column);
		return (getTableName(column) + "." + ar.getName()); 
  }

  /**
   * What's a column's name?
   *
   * @param column the first column is 1, the second is 2, ...
   * @return column name
   */
  public String getColumnName(int column) throws SQLException {
		RecordStreamResult res = (RecordStreamResult)result;
		Attribute ar = res.getSchema().getAttribute(column);
		return ar.getName();
  }

  /**
   * What's a column's table's schema?
   *
   * @param column the first column is 1, the second is 2, ...
   * @return schema name or "" if not applicable
   */
  public String getSchemaName(int column) throws SQLException {
    return null;
  }

  /**
   * What's a column's number of decimal digits?
   *
   * @param column the first column is 1, the second is 2, ...
   * @return precision
   */
  public int getPrecision(int column) throws SQLException {
    return -1;
  }

  /**
   * What's a column's number of digits to right of the decimal point?
   *
   * @param column the first column is 1, the second is 2, ...
   * @return scale
   */
  public int getScale(int column) throws SQLException {	
    return -1;
  }

  /**
   * What's a column's table name? 
   *
   * @return table name or "" if not applicable
   */
  public String getTableName(int column) throws SQLException {
		String colname = getColumnName(column);
		StringBuffer bufname = new StringBuffer(colname);
		bufname.insert(0, '"');
		bufname.append('"');
		colname = bufname.toString();
		String query = "SELECT relname FROM _SATTRS WHERE attrname = " + colname;
		
		ResultSet res = stmt.executeQuery(query);
		res.next();
		return res.getString(0); 
  }

  /**
   * What's a column's table's catalog name?
   *
   * @param column the first column is 1, the second is 2, ...
   * @return column name or "" if not applicable.
   */
  public String getCatalogName(int column) throws SQLException {
    return null;
  }

  /**
   * What's a column's SQL type?
   *
   * @param column the first column is 1, the second is 2, ...
   * @return SQL type
   * @see Types
   */
  public int getColumnType(int column) throws SQLException {
		RecordStreamResult res = (RecordStreamResult)result;
		Attribute ar = res.getSchema().getAttribute(column);
		ADT adt = ar.getTypeADT();
		String type = adt.getName();

		if (type.equals("int")) {
			return Types.INTEGER;
		} else if (type.equals("double")) {
			return Types.DOUBLE;
		} else if (type.equals("string")) {
			return Types.CHAR;
		}	else
			return Types.NULL; 
  }

  /**
   * What's a column's data source specific type name?
   *
   * @param column the first column is 1, the second is 2, ...
   * @return type name
   */
  public String getColumnTypeName(int column) throws SQLException {
		RecordStreamResult res = (RecordStreamResult)result;
		Attribute ar = res.getSchema().getAttribute(column);
		ADT adt = ar.getTypeADT();
		String type = adt.getName();

		if (type.equals("int")) {
			return "int";
		} else if (type.equals("double")) {
			return "double";
		} else if (type.equals("string")) {
			return "string";
		}	else
			return null; 
  }

  /**
   * Is a column definitely not writable?
   *
   * @param column the first column is 1, the second is 2, ...
   * @return true if so
   */
  public boolean isReadOnly(int column) throws SQLException {
    return false;
  }

  /**
   * Is it possible for a write on the column to succeed?
   *
   * @param column the first column is 1, the second is 2, ...
   * @return true if so
   */
  public boolean isWritable(int column) throws SQLException {
    return false;
  }

  /**
   * Will a write on the column definitely succeed?	
   *
   * @param column the first column is 1, the second is 2, ...
   * @return true if so
   */
  public boolean isDefinitelyWritable(int column) throws SQLException {
    return false;
  }

};

