package PREDATOR.ADT;

/**
 * This class represents meta-information which may be maintained for all
 * the objects of a certain attribute.  Frequently meta-info includes
 * data which specialize the objects in some way; for instance, selecting
 * some audio objects from a table might produce an attribute with meta-
 * info describing the format of the audio objects (e.g., <TT>.au</TT>
 * format).  This information is provided to all objects in all operations
 * so their behavior can be specialized when appropriate.  For a more 
 * detailed discussion of meta-information, refer to the 
 * <FONT COLOR="#CC0000"><FONT SIZE=+1>P</FONT>REDATOR</FONT> documentation.
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version 0.9
 * @see ADT#newMetaInfo(byte[])
 * @see PREDATOR.Attribute#getMetaInfo()
 */
public abstract class ADTMetaInfo {
  /**
   * The binary format of this object.
   */
  protected byte[] bytes;
  /**
   * Every data object is stored as a contiguous sequence of bytes at 
   * the server.  This function assigns those bytes to the object.
   * @param bytes The binary form of this object.
   */
  public void setBytes(byte[] bytes) 
  {
    this.bytes = bytes;
  }
  /**
   * Gets the binary form of this object.
   */
  public byte[] getBytes()
  {
    return bytes;
  }
  /**
   * Provides a textual version of this object suitable for user-display.
   * @return A textual version of this object
   */
  public abstract String toString();
}
