package PREDATOR.ADT;

import java.io.*;
import PREDATOR.NullValueException;

/**
 * Encapsulates all in-memory data regarding an boolean ADT object.
 */
public class BooleanObject extends DataObject {

  // temporary constructor
  public BooleanObject(boolean b)
  {
	// I set the length to six, in accord with the server
	// side SMaxObjectSize result (Tobias)
    byte[] tmp =  new byte[6];
    tmp[0]=tmp[1]=tmp[2]=tmp[3]=0;
    if(b)
      tmp[4]=1;
    else
      tmp[4]=0;
    setBytes(tmp);
  }
  public BooleanObject()
  {super();}

  /**
   * Maps the value to a Java boolean
   * @param MetaInfo The meta-info for this object
   * @return A boolean that represents the value of the object
   */
  public boolean getValue(ADTMetaInfo MetaInfo) 
       throws NullValueException 
  {
    if(is(NULL))
      throw new NullValueException();
    try {
      DataInputStream dataStream = 
        new DataInputStream(new ByteArrayInputStream(getBytes(), 4, 
                                                     getBytes().length - 4));
      byte val = dataStream.readByte();
      return (val > 0);
    } catch (IOException X) {
      throw new PREDATOR.InternalError(X);
    }
  }

  /**
   * Prints the value of the boolean
   * @param MetaInfo The meta-info for this object
   * @return A string holding the value of the boolean
   */
  public String toString(ADTMetaInfo MetaInfo) {
    String tmp = super.toString(MetaInfo);
    if (tmp != null)
      return (tmp);
    try {
      if (getValue(MetaInfo))
	return new String("true");
      else
	return new String("false");
    } catch (NullValueException X) {
      return("Null");
    }
  }
}

