package PREDATOR.ADT;

import java.net.*;

/**
 * The class extended by objects which can contact the server for
 * additional information and are also able to provide a binary representation
 * of this information so that the "full value" can be cached.
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version 0.9
 * @see CallbackObject
 */
public interface CachingObject extends CallbackObject {
  /**
   * Performs a callback to the server, using the given URL.  The URL
   * may be opened as a stream, if desired.  The contents of the URL
   * will be determined by what the server-side ADT writes out in the
   * <TT>ServiceCallback()<TT> function (in <TT>types.h</TT>).
   * @param CallbackURL The URL of the callback object
   * @param MetaInfo The meta-info for this object
   */
  public void callback(URL CallbackURL, ADTMetaInfo MetaInfo);
  /**
   * Determines if this object has a handle on the "full value", or if
   * a callback or deserialization is necessary.
   * @param MetaInfo The meta-info for this object
   * @return true if this object does not have its "full value" available
   */
  public boolean requiresCallback(ADTMetaInfo MetaInfo);
}
