package PREDATOR.ADT;

import java.io.*;

/**
 * Encapsulates all in-memory data regarding a complex-number ADT object.
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version 0.9
 * @see PREDATOR.ADT.ADT_int
 */
public class ComplexObject extends DataObject {
  /**
   * Prints the value of the complex number
   * @param MetaInfo The meta-info for this object
   * @return A string holding the value of the complex number
   */
  public String toString(ADTMetaInfo MetaInfo) {
    String tmp = super.toString(MetaInfo);
    if (tmp != null)
      return (tmp);
    try {
      DataInputStream dataStream = 
        new DataInputStream(new ByteArrayInputStream(getBytes(), 8, 
                                                     getBytes().length - 8));
      double RealPart = dataStream.readDouble();
      double ImagPart = dataStream.readDouble();
      if (ImagPart < 0.0)
	return Double.toString(RealPart) + "-" + Double.toString(-ImagPart) + "i";
      else
	return Double.toString(RealPart) + "+" + Double.toString(ImagPart) + "i";
    } catch (IOException X) {
      throw new PREDATOR.InternalError(X);
    }
  }
  public double getReal() {
    try {
      DataInputStream dataStream = 
        new DataInputStream(new ByteArrayInputStream(getBytes(), 8, 
                                                     getBytes().length - 8));
      return dataStream.readDouble();
    } catch (IOException X) {
      throw new PREDATOR.InternalError(X);
    }
  }
  public double getImag() {
    try {
      DataInputStream dataStream = 
        new DataInputStream(new ByteArrayInputStream(getBytes(), 8, 
                                                     getBytes().length - 8));
      dataStream.readDouble();
      return dataStream.readDouble();
    } catch (IOException X) {
      throw new PREDATOR.InternalError(X);
    }
  }
}

