package PREDATOR.ADT;

import PREDATOR.*;

/**
 * Extended by DataObject classes which can display themselves in some
 * fashion not captured by their toString() method.  For instance, an
 * image can display itself, or an audio can play itself.
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version 0.9
 * @see DataObject
 */
public interface DisplayObject {
  /**
   * Displays an object indepently.  Implementations of this function
   * should be non-blocking for best performance.  (For instance, an
   * image should be displayed in an independent frame and execution
   * should proceed while the frame is still open.)
   * @param MetaInfo The meta-info for this object
   * @exception RequiresCallbackException when a callback is required
   *                                      before a display can be made
   * @see CachingObject#callback(URL, ADTMetaInfo)
   */
  public void display(ADTMetaInfo MetaInfo) throws RequiresCallbackException;
}
