package PREDATOR.ADT;

/**
 * Encapsulates all in-memory data regarding an document ADT object.
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version 0.9
 * @see ADT_document
 */
public class DocumentObject
  extends DataObject implements CallbackObject, MIMETypeObject {
  /**
   * Prints the value of the document
   * @param MetaInfo The meta-info for this object
   * @return A string holding the value of the document
   */
  public String toString(ADTMetaInfo MetaInfo) {
	String tmp = super.toString(MetaInfo);
	if (tmp != null)
	  return (tmp);
    switch(((DocumentMetaInfo)MetaInfo).Format) {
    case DocumentMetaInfo.DOCUMENT_MSWORD:
      return "MS Word Document";
    case DocumentMetaInfo.DOCUMENT_HTML:
      return "Web Document";
    case DocumentMetaInfo.DOCUMENT_POSTSCRIPT:
      return "Postscript Document";
    default:
      return "Document";
    }
  }
  /**
   * Callback objects have the option of supplying a byte string
   * containing additional information about the callback.  For
   * instance, an image object might specify one of many stored
   * resolutions.
   * @return A binary message
   */
  public byte [] getMessage()
  {
    return null;
  }
  /**
   * Provides a filename extension suitable for files storing data of
   * this MIME type.  For instance, JPEG images will supply the extension
   * "jpg".
   * @return A suitable filename extension for files of this type
   */
  public String getFilenameExtension(ADTMetaInfo MetaInfo)
  {
    switch(((DocumentMetaInfo)MetaInfo).Format) {
    case DocumentMetaInfo.DOCUMENT_MSWORD:
      return "doc";
    case DocumentMetaInfo.DOCUMENT_HTML:
      return "htm";
    case DocumentMetaInfo.DOCUMENT_POSTSCRIPT:
      return "ps";
    default:
      return "txt";
    }
  }
}
