package PREDATOR.ADT;

import java.io.*;
import java.lang.reflect.*;
import PREDATOR.NullValueException;

/**
 * Encapsulates all in-memory data regarding a double ADT object.
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version 0.9
 * @see ADT_double
 */
public class DoubleObject extends DataObject implements JavaType 
{  
  private Double d;
	private boolean dataIsSet;
	private boolean byteIsSet;


	public DoubleObject( )
	{
		d = new Double( 0.0 );
		dataIsSet = false;
		byteIsSet = true;
	}

	public DoubleObject( double newD )
	{
		d = new Double( newD );
		dataIsSet = true;
		byteIsSet = false;
	}

	public DoubleObject( Double newD )
	{
		d = newD;
		dataIsSet = true;
		byteIsSet = false;
	}
	
	public boolean IsDataSet( )
	{
		return dataIsSet;
	}
	
	public boolean IsByteSet( )
	{
		return byteIsSet;
	}

	/**
	 * Reads bytes to set DoubleObject field unless bytes are not set, in which
	 * case, use the data field to set the bytes
	 * @return byte[] the data in a byte array
	 */
	public byte[] getBytes()
	{		
		byte[] bytes;

		try 
		{
			if (( !byteIsSet ) && ( dataIsSet ))
			{
				bytes = new byte[ 16 ];
				ByteArrayOutputStream byteStream = new ByteArrayOutputStream( );
				DataOutputStream ostream = new DataOutputStream( byteStream );
				ostream.writeDouble( d.doubleValue( ) );

				byte[] valueDbl = byteStream.toByteArray( );
				for ( int i = 0; i < 8; i++ )
					bytes[ i+8 ] = valueDbl[ i ];

				super.setBytes( bytes );
				byteIsSet = true;
			}
			else if (( byteIsSet ) && ( dataIsSet ))
			{
				bytes = new byte[ super.getBytes().length ];
				bytes = super.getBytes();				
			}
			else
			{
				bytes = new byte[ super.getBytes().length ];
				bytes = super.getBytes();	
				if(is(NULL))
					throw new NullValueException();
				DataInputStream dataStream = new DataInputStream(new ByteArrayInputStream(
					bytes, 8, bytes.length-8 ));
				d = new Double( dataStream.readDouble() );
			}

			dataIsSet = true;
		} 
		catch (Exception X) 
		{
			throw new PREDATOR.InternalError(X);
		}

		return bytes;
	}  
	
	/**
	 * Retrieves a copy of the double value
	 */
	public Object getValue( )
	{
		if ( !dataIsSet )
			getBytes();

		return d;
	}
	
	
	/**
	 * Sets the double value
	 */
	public void setValue( Object value )
	{
		try
		{
			d = (Double) value;
		  byteIsSet = false;
			dataIsSet = true;
		}
		catch (Exception X)
		{
			throw new PREDATOR.InternalError(X);
		}

		return;
	}

  /**
   * Prints the value of the double
   * @param MetaInfo The meta-info for this object
   * @return A string holding the value of the double
   */
  public String toString(ADTMetaInfo MetaInfo) 
  {
    String tmp = super.toString(MetaInfo);
    if (tmp != null)
      return (tmp);

		if ( !dataIsSet )
			getBytes();

    return d.toString();
  }
}
