package PREDATOR.ADT;

import java.io.*;
import java.util.*;
import java.awt.*;
import java.net.*;
import PREDATOR.NullValueException;


public class GenTimeSeriesObject extends TimeSeriesObject
{	
  protected int[] dataOfPeriod;

	public int max, min;

	// variables for the graphical interface
	protected int gwidth, gheight;				// the dimensions of the graph
	public int origX, origY;						// the offset on the window for the graph
	protected int topBorder;							// the thickness of the top border
	public int spacingX, spacingY;			// the scaling of values and days
	protected int scrollBorder;
	protected int dwidth, dheight;
	protected int wwidth, wheight;

	public int numDays;
	public int division = 1;

	public GenTimeSeriesObject( )
	{
		numDays = 0;
	}

	/**
	 */
	public byte[] getBytes()
	{		
		byte[] bytes;
		
		try 
		{			
			if (( !byteIsSet ) && ( dataIsSet ))
			{
				bytes = new byte[ numDays*4 + 8];
				for( int g = 0; g < (numDays*4 + 8); g++ )
					bytes[g] = 0;
				ByteArrayOutputStream baos = new ByteArrayOutputStream( numDays*4 );
				
				baos.write( numDays );
				for( int prd = 0; prd < numDays; prd++ )
					baos.write( dataOfPeriod[ prd ] );

				byte[] tmp = baos.toByteArray( );
				for ( int k = 0; k < tmp.length; k++ )
					bytes[ k+4 ] = tmp[ k ];

				super.setBytes( bytes );
				byteIsSet = true;
			}
			else if (( byteIsSet ) && ( dataIsSet ))
			{
				bytes = new byte[ super.getBytes().length ];
				bytes = super.getBytes();
			}
			else
			{
				bytes = new byte[ super.getBytes().length ];
				bytes = super.getBytes();
				
				if(is(NULL))
					throw new NullValueException();
				DataInputStream dataStream = new DataInputStream(new ByteArrayInputStream(
					bytes, 4, bytes.length-4));

				numDays = dataStream.readInt();

				dataOfPeriod = new int[numDays];

				max = 0; min = 10000;
				for ( int i = 0; i < numDays; i++ )
				{
					dataOfPeriod[i] = dataStream.readInt();
					if( dataOfPeriod[i] > max ) max = dataOfPeriod[i];
					if(( dataOfPeriod[i] < min ) && ( dataOfPeriod[i] != 0 ))
						min = dataOfPeriod[i];
				}
			}

			dataIsSet = true;
			setDisplayValues( );
		} 
		catch (Exception X) 
		{
			X.printStackTrace();
			throw new PREDATOR.InternalError(X);
		}

		return bytes;
	}


  /**
   * Prints the size of the time series data set
   * @param MetaInfo The meta-info for this object
   * @return A string holding the value of the time series data
   */
  public String toString(ADTMetaInfo MetaInfo) 
	{
    String tmp = super.toString(MetaInfo);
    if (tmp != null)
      return (tmp);

		if ( !dataIsSet )
			getBytes();

 
      return ( numDays + "-period time series." );
 
  }

	public void setDisplayValues( )
	{
		int numDigits1 = (int)Math.ceil( Math.log( (double) max ) / Math.log( 10.0 ) );
		int numDigits2 = (int)Math.ceil( Math.log( (double) Math.abs(min) ) / Math.log( 10.0 ) );
		int numDigits = Math.max( numDigits1, numDigits2 );
		origX = 20;
		origY = 30;
		topBorder = 30;
		scrollBorder = 30;
		spacingX = 8;
		spacingY = 4;
		gwidth = numDays * spacingX;

		int diff = max-min;
		if( diff < 0 )
			diff = 0;
		if( diff > 75 )
			division = diff / 75;
		gheight = ((diff)/division) * spacingY;
		wwidth = gwidth +2*origX + scrollBorder + 7*numDigits;
		wheight = (gheight == 0) ? 0 : gheight + 2*origY + topBorder + scrollBorder;
		dwidth = 600;
		dheight = Math.min( 400, wheight );
	}
	
	public ScrollPane display( )
	{		
		if ( !dataIsSet )
			getBytes();

		Canvas g;		
		g = new GenTSGraph( this, wwidth, wheight );

		GenTSScrollGraph sg = new GenTSScrollGraph( dwidth, dheight );
		sg.add( g );
		sg.setScrollPosition( dwidth, dheight );
		Point p = sg.getScrollPosition( );
		System.out.println( p );
												 
		return sg;
	}

	public Dimension getDimensions( )
	{
		if ( !dataIsSet )
			getBytes();

		return new Dimension( dwidth, dheight );
	}

  /**
   * Displays an object indepently.  Implementations of this function
   * should be non-blocking for best performance.  (For instance, an
   * image should be displayed in an independent frame and execution
   * should proceed while the frame is still open.)
   * @param MetaInfo The meta-info for this object
   */
  public void displayGraph( )
  {
		if ( !dataIsSet )
			getBytes();

		// Launch the thread
    ScrollPane sp = display();
      new Thread(new GenTSScrollWin( "Time Series", sp, dwidth, dheight )).run(); 
  }
  
	public int getDataOfPeriod( int period )
	{		
		if ( !dataIsSet )
			getBytes();
		
		if( period > dataOfPeriod.length )
			return 0;
		else
			return dataOfPeriod[ period ];
	}
	
  public IntegerObject getValue(IntegerObject Period)
  {
		int retInt;
		
		if ( !dataIsSet )
			getBytes();

	  int periodInt = ((Integer)Period.getValue()).intValue();
		if( periodInt > dataOfPeriod.length )
			retInt = 0;
		else retInt = dataOfPeriod[ periodInt ];
		
	  return new IntegerObject( retInt );
  }
  
	public IntegerObject getNumDays( )
	{
		if ( !dataIsSet )
			getBytes();

		return new IntegerObject( numDays );
	}
		
  public IntegerObject getNumItems( )
  {
		if ( !dataIsSet )
			getBytes();

		int count = 0;
		for( int i = 0; i < numDays; i++ )
			if( dataOfPeriod[i] != 0 ) count++;
		
	  return new IntegerObject( count );
  }
	
	public DoubleObject getMean( )
	{
		if ( !dataIsSet )
			getBytes();

		double total = 0.0;
		int count = 0;
		
		for( int i = 0; i < numDays; i++ )
			if( dataOfPeriod[i] != 0 )
			{
				total += (double)dataOfPeriod[i];
				count++;
			}
		
		if( count == 0 )
			return new DoubleObject( 0.0 );
		else
			return new DoubleObject( (double)(total / (double)count) );
	}
		
	public DoubleObject getStDev( )
	{
		if ( !dataIsSet )
			getBytes();
		
		double total = 0.0;
		int count = 0;
		for( int i = 0; i < numDays; i++ )
			if( dataOfPeriod[i] != 0 )
			{
				total += (double)dataOfPeriod[i];
				count++;
			}
		double mean;
		if( count == 0 )
			return new DoubleObject( 0.0 );
		else
			mean = (double)(total / (double)count);
		
		total = 0.0;
		for( int j = 0; j < numDays; j++ )
			if( dataOfPeriod[j] != 0 )
				total += Math.pow( (double)dataOfPeriod[j] - mean, 2.0 );
		
		double stdev = total * (1.0/(double)count);
		stdev = Math.pow( stdev, 0.5 );
		
		return new DoubleObject( stdev );
	}		
}



class GenTSScrollWin extends Frame implements Runnable 
{
	int pWidth, pHeight;
	
	public GenTSScrollWin( String title, ScrollPane sg, int prefWidth, int prefHeight )
	{
		super( title );

		add( sg );
		pWidth = prefWidth;
		pHeight = prefHeight;
		pack();
		setVisible( true );
	}
	
  public void run() 
	{
  }

  public boolean handleEvent(Event event) 
	{
    if (event.id == Event.WINDOW_DESTROY) 
		{
      dispose();	
      return true;
    }
    return false;
  }
		
	public Dimension getPreferredSize( )
	{
		return new Dimension( pWidth, pHeight );
	}
	
}

class GenTSScrollGraph extends ScrollPane
{
	int dwidth, dheight;

	public GenTSScrollGraph( int w, int h )
	{
		super( ScrollPane.SCROLLBARS_AS_NEEDED );
		
		dwidth = w;
		dheight = h;
	}

	public Dimension getPreferredSize( )
	{
		return new Dimension( dwidth - 50, dheight - 50 );
	}
}

class GenTSGraph extends Canvas
{

	GenTimeSeriesObject ts;
	int pX, pY;

  public GenTSGraph( GenTimeSeriesObject timeSeries, int dX, int dY ) 
	{
		ts = timeSeries;
		pX = dX;
		pY = dY;
    setBackground(Color.black);
  }
  
	public Dimension getPreferredSize( )
	{
		return new Dimension( pX, pY );
	}	

  private void Draw( Graphics g ) 
	{
		g.translate( 0, ts.topBorder );
		g.setColor(Color.yellow);
		for ( int i = 0; i < ts.numDays-1; i++ )
		{
			// draw the graph
			if(( ts.dataOfPeriod[i] != 0 ) && ( ts.dataOfPeriod[i+1] != 0 ))
				g.drawLine( ts.numDays*ts.spacingX+ts.origX - i*ts.spacingX, 
										((ts.max-ts.dataOfPeriod[i])/ts.division)*ts.spacingY + ts.origY, 
										ts.numDays*ts.spacingX+ts.origX - (i+1)*ts.spacingX, 
										((ts.max-ts.dataOfPeriod[i+1])/ts.division)*ts.spacingY + ts.origY );
			
			if( i == 0 ) 
			{
				if(( ts.dataOfPeriod[i] != 0 ) && ( ts.dataOfPeriod[i+1] == 0 ))
					g.drawOval( ts.numDays*ts.spacingX+ts.origX - i*ts.spacingX, 
										((ts.max-ts.dataOfPeriod[i])/ts.division)*ts.spacingY + ts.origY, 
										3, 3 );
			}
			else if(( ts.dataOfPeriod[i] != 0 ) && ( ts.dataOfPeriod[i+1] == 0 ) && ( ts.dataOfPeriod[i-1] == 0 ))
			{
				g.drawOval( ts.numDays*ts.spacingX+ts.origX - i*ts.spacingX, 
									((ts.max-ts.dataOfPeriod[i])/ts.division)*ts.spacingY + ts.origY, 
									3, 3 );
			}
			
			if( i == ts.numDays-2 )
			{
				if(( ts.dataOfPeriod[i] == 0 ) && ( ts.dataOfPeriod[i+1] != 0 ))
					g.drawOval( ts.numDays*ts.spacingX+ts.origX - (i+1)*ts.spacingX, 
										((ts.max-ts.dataOfPeriod[i+1])/ts.division)*ts.spacingY + ts.origY, 
										3, 3 );
			}
		}
		g.drawString( "Periods", ts.gwidth+ts.origX, ts.gheight+ts.origY+28 );
		g.drawString( "Value", ts.gwidth+ts.origX+ts.spacingX+10, 10 );
		// draw y-axis
		g.drawLine( ts.gwidth+ts.origX+ts.spacingX, ts.origY, ts.gwidth+ts.origX+ts.spacingX, ts.gheight+ts.origY );
		// draw x-axis
		g.drawLine( ts.origX, ts.gheight+ts.origY, ts.gwidth+ts.origX+ts.spacingX, ts.gheight+ts.origY );

		// draw x-axis markings
		for ( int j = 2*ts.spacingX; j < ts.gwidth+2*ts.spacingX; j += 2*ts.spacingX )
		{
			g.drawLine( j+ts.origX, ts.gheight+ts.origY-2, j+ts.origX, ts.gheight+ts.origY+1 );
			if ( (j-2*ts.spacingX) % (4*ts.spacingX) == 0 )
				g.drawString( (new Integer( (int)((ts.gwidth+ts.spacingX-j)/ts.spacingX))).toString(),
					j+ts.origX - 3, ts.gheight+ts.origY+12 );
		}

		// draw y-axis markings
		for ( int k = ts.gheight; k > 0; k -= 5*ts.spacingY )
		{
			g.drawLine( ts.gwidth+ts.origX+ts.spacingX-1, k+ts.origY, 
									ts.gwidth+ts.origX+ts.spacingX+2, k+ts.origY );
			if ( (ts.gheight-k) % (10*ts.spacingY) == 0 )
			{
				int val = (int)( ((ts.gheight - k)/(double)ts.spacingY) 
									+ (ts.min/ts.division) ) * ts.division;
				g.drawString( (new Integer( val )).toString(), ts.gwidth+ts.origX+ts.spacingX+10, k+ts.origY+5 );
			}
		}
	}
	
  public void paint( Graphics g )
	{
    Draw(g);
  }

}
