package PREDATOR.ADT;

import java.applet.*;
import java.io.*;
import java.awt.*;
import java.awt.image.*;
import java.net.*;
import PREDATOR.*;

/**
 * Encapsulates all in-memory data regarding an image ADT object.
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version 0.9
 * @see ADT_image
 */
public class ImageObject
  extends DataObject implements CallbackObject, MIMETypeObject 
{
  /**
   * The desired resolution number.
   */
  public int DesiredResolution = 2;
  /**
   * Sets the desired resolution of the object when a callback
   * is performed.  If the selected resolution does not exist, 
   * callbacks for this object will fail.
   * @param DesiredResolution Sets the desired resolution (0 = normal,
   *                          1 = first reduced resolution, 2 = 
   *                          second reduced resolution).
   */
  public void setDesiredResolution(int DesiredResolution)
  {
    this.DesiredResolution = DesiredResolution;
  }
  /**
   * Prints the value of the image
   * @param MetaInfo The meta-info for this object
   * @return A string holding the value of the image
   */
  public String toString(ADTMetaInfo MetaInfo) {
    String tmp = super.toString(MetaInfo);
    if (tmp != null)
      return (tmp);
    switch(((ImageMetaInfo)MetaInfo).Format) {
    case ImageMetaInfo.IMAGE_JPEG:
      return "JPEG image";
    case ImageMetaInfo.IMAGE_GIF:
      return "GIF image";
    default:
      return "image";
    }
  }
  /**
   * Callback objects have the option of supplying a byte string
   * containing additional information about the callback.  For
   * instance, an image object might specify one of many stored
   * resolutions.
   * @return A binary message
   */
  public byte [] getMessage()
  {
    try {
      ByteArrayOutputStream B = new ByteArrayOutputStream();
      DataOutputStream OStr = new DataOutputStream(B);
      OStr.writeInt(DesiredResolution);
      OStr.flush();
      return B.toByteArray();
    } catch (IOException X) {
      throw new PREDATOR.InternalError(X);
    }
  }
  /**
   * Provides a filename extension suitable for files storing data of
   * this MIME type.  For instance, JPEG images will supply the extension
   * "jpg".
   * @return A suitable filename extension for files of this type
   */
  public String getFilenameExtension(ADTMetaInfo MetaInfo)
  {
    switch(((ImageMetaInfo)MetaInfo).Format) {
    case ImageMetaInfo.IMAGE_JPEG:
      return "jpg";
    case ImageMetaInfo.IMAGE_GIF:
      return "gif";
    default:
      return "img";
    }
  }
}
