package PREDATOR.ADT;

import java.io.*;
import PREDATOR.NullValueException;

/**
 * Encapsulates all in-memory data regarding an integer ADT object.
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version 0.9
 * @see ADT_int
 */
public class IntegerObject extends DataObject implements JavaType 
{
	private Integer i;
	private boolean dataIsSet;
	private boolean byteIsSet;

	public IntegerObject( )
	{
		i = new Integer(0);
		dataIsSet = false;
		byteIsSet = true;
	}

	public IntegerObject( int newI )
	{
		i = new Integer( newI );
		dataIsSet = true;
		byteIsSet = false;
	}

	public IntegerObject( Integer newI )
	{
		i = newI;
		dataIsSet = true;
		byteIsSet = false;
	}
	
	public boolean IsDataSet( )
	{
		return dataIsSet;
	}
	
	public boolean IsByteSet( )
	{
		return byteIsSet;
	}

	/**
	 * Reads bytes to set IntegerObject field unless bytes are not set, in which
	 * case, use the data field to set the bytes
	 * @return byte[] the data in a byte array
	 */
	public byte[] getBytes()
	{		
		byte[] bytes;
		
		try 
		{
			if (( !byteIsSet ) && ( dataIsSet ))
			{
				bytes = new byte[ 8 ];
				for( int g = 0; g < 8; g++ )
					bytes[g] = 0;
				
				ByteArrayOutputStream baos = new ByteArrayOutputStream( 4 );
				DataOutputStream dos = new DataOutputStream(baos);
				dos.writeInt( i.intValue( ) );

				byte[] tmp = baos.toByteArray( );
				
				for ( int k = 0; k < 4; k++ )
					bytes[ k+4 ] = tmp[ k ];

				super.setBytes( bytes );
				byteIsSet = true;
			}
			else if (( byteIsSet ) && ( dataIsSet ))
			{
				bytes = new byte[ super.getBytes().length ];
				bytes = super.getBytes();				
			}
			else
			{
				bytes = new byte[ super.getBytes().length ];
				bytes = super.getBytes();	
				if(is(NULL))
					throw new NullValueException();
				DataInputStream dataStream = new DataInputStream(new ByteArrayInputStream(
					bytes, 4, bytes.length-4));
				i = new Integer( dataStream.readInt() );
			}

			dataIsSet = true;
		} 
		catch (Exception X) 
		{
			throw new PREDATOR.InternalError(X);
		}

		return bytes;
	} 


	/**
	 *
	 */
	public Object getValue( )
	{	
		if ( !dataIsSet )
			getBytes();

		return i;
	}

	
	/**
	 * Sets the double value
	 */
	public void setValue( Object value )
	{
		try
		{
			i = (Integer) value;
		  byteIsSet = false;
			dataIsSet = true;
		}
		catch (Exception X)
		{
			throw new PREDATOR.InternalError(X);
		}

		return;
	}

  /**
   * Prints the value of the integer
   * @param MetaInfo The meta-info for this object
   * @return A string holding the value of the integer
   */
  public String toString(ADTMetaInfo MetaInfo) 
	{
  	String tmp = super.toString(MetaInfo);
  	if (tmp != null)
  		return (tmp);

		if ( !dataIsSet )
			getBytes();

			return i.toString();
 }
}

