package PREDATOR;

import PREDATOR.ADT.*;

/**
 * An attribute of a schema.
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version 0.9
 * @see Schema
 */
public class Attribute {
  /// A handle on the ADT corresponding to the attribute's type
  ADT TypeADT;
  /// A handle on the meta-info for this attribute
  ADTMetaInfo MetaInfo;
  /// The name of the attribute
  String AttrName;
  /// The offset of the attribute (in bytes)
  int Offset;
  
  /**
   * Constructor.
   * @param DataType The ADT corresponding to the type of this attribute
   * @param MetaInfo The meta-info for this attribute
   * @param Name The name of the attribute
   */
  Attribute(ADT DataType, ADTMetaInfo MetaInfo, 
	    String Name, int Offset) {
    TypeADT = DataType;
    this.MetaInfo = MetaInfo;
    AttrName = Name;
    this.Offset = Offset;
  }
  /**
   * Gets the name of the attribute.
   * @return The name of the attribute.
   */
  public String getName()
  {
    return AttrName;
  }
  /**
   * Gets the type of the attribute.
   * @return the ADT associated with the type of this attribute.
   */
  public ADT getTypeADT()
  {
    return TypeADT;
  }
  /**
   * Gets the meta-info associated with this attribute.
   * @return this attribute's meta-info
   */
  public ADTMetaInfo getMetaInfo()
  {
    return MetaInfo;
  }
  /**
   * Gets the offset (in bytes) of this attribute
   * @return this attribute's offset
   */
  int getOffset()
  {
    return Offset;
  }
  /**
   * Creates a string representation of this attribute.
   */
  public String toString() {
    StringBuffer buffer = new StringBuffer(AttrName + ":" + TypeADT.getName());
    if (MetaInfo != null)
      buffer.append("[" + MetaInfo.toString() + "]");
    return (buffer.toString());
  }
}
