package PREDATOR;

import java.io.*;
import PREDATOR.ADT.*;

/**
 * Encapsulates the notion of a record.
 * @author Mark Paskin, paskin@cs.cornell.edu
 * @version 0.9
 */
class DataRecord {
  /// The binary data
  byte [] Data;
  /// The size (in bytes) of the record
  int Size;

  /**
   * Default Constructor.
   */
  DataRecord() {
    Data = null;
    Size = 0;
   }
  /**
   * Constructor.  Allocates the desired space for the record.
   * @param RecSize The number of bytes to allocate.
   */
  DataRecord(int RecSize) {
    Size = RecSize;
    Data = new byte[Size];
  }
  /**
   * Returns the binary contents of the attribute
   * specified.
   * @param Schema This record's schema
   * @param AttrNum The index of the attribute to extract
   * @return The bytes of the attribute
   */
  protected byte[] getAttrBytes(Schema schema, 
                                int AttrNum) 
  {
    try {
      int offset = 0, length = 0;
      // Get the offset of the attribute
      offset = schema.getAttribute(AttrNum).getOffset();
      if (schema.getAttribute(AttrNum).getTypeADT().
	  isVariableSize(schema.getAttribute(AttrNum).getMetaInfo())) {
	DataInputStream dataStream = 
	  new DataInputStream(new ByteArrayInputStream(Data, offset, 4));
	offset = dataStream.readInt();
	dataStream.close();
      }
      // Now get the offset of the next attribute to determine the length
      int nextOffset = Data.length, varOffset;
      for (int i = 0; i < schema.getArity(); i++) {
	if (schema.getAttribute(i).getOffset() > offset)
	  nextOffset = Math.min(nextOffset,
				schema.getAttribute(i).getOffset());
	if (schema.getAttribute(i).getTypeADT().
	    isVariableSize(schema.getAttribute(i).getMetaInfo())) {
	  DataInputStream dataStream =
	    new DataInputStream(new ByteArrayInputStream(Data, 
							 schema.
							 getAttribute(i).
							 Offset, 4));
	  varOffset = dataStream.readInt();
	  dataStream.close();
	  if (varOffset > offset)
	    nextOffset = Math.min(nextOffset, varOffset);
	}
      }
      length = nextOffset - offset;
      byte[] attr = new byte[length];
      System.arraycopy(Data, offset, attr, 0, length);
      return attr;
    } catch (IOException X) {
      throw new PREDATOR.InternalError(X);
    }
  }
  /**
   * Gets an array of the objects in this record
   * @param schema This record's schema
   * @param Subclasser The class which determines what subclass to use of
   *                   a given DataObject class
   * @exception BadSubclassSelectorException If the subclasser returns a
   *                                         null class or one which is not
   *                                         a subclass of the parameter class
   * @return An array of the objects in this record
   */
  DataObject[] getObjects(Schema schema, 
                          SubclassSelector Subclasser) 
    throws BadSubclassSelectorException
  {
    try {
      DataObject objects[] = new DataObject[schema.getArity()];
      for (int i = 0; i < schema.getArity(); i++) {
        ADTMetaInfo MetaInfo = schema.getAttribute(i).getMetaInfo();
        ADT AttrADT = schema.getAttribute(i).TypeADT;
        objects[i] = AttrADT.newObject(getAttrBytes(schema, i),
				       MetaInfo, Subclasser);
      }
      return objects;
    } catch (IllegalAccessException X) {
      throw new InternalError(X);
    } catch (InstantiationException X) {
      throw new InternalError(X);
    }
  }
  public void setObjects(DataObject[] objects)
  {
    Size = 0;
    for(int i=0; i<objects.length; i++)
      Size = Size + objects[i].getBytes().length;
    Data = new byte[Size];
    int endSoFar = 0;
    for(int i=0; i<objects.length; i++)
      {
	byte[] bytes = objects[i].getBytes();
	System.arraycopy(bytes,0,Data,endSoFar,bytes.length);
	endSoFar = endSoFar + bytes.length;
      }
  }
}
