package PREDATOR;

import java.io.*;
import java.net.*;
import java.util.*;
import PREDATOR.ADT.*;

public class JConnection extends Connection {
  protected final static String InitString = "NEW SESSION\n";
  protected final static String SERVER_SETUP_STRING = "Setup?";
  /**
   * The task socket connection; used for processing requests
   */
  protected SocketConnection task;

  /**
   * Opens a new connection to a 
   * <FONT COLOR="#CC0000"><FONT SIZE=+1>P</FONT>REDATOR</FONT> server.
   * @param Site The site to which the connection should be made
   * @exception CannotConnectException if the connection cannot be opened
   * @exception ConnectionAlreadyOpenException if the connection is already 
   *                                           open
   */
  public synchronized void open(ServerSite Site) 
    throws CannotConnectException, ConnectionAlreadyOpenException
  {
    if (isOpen())
      throw new ConnectionAlreadyOpenException();
    // Open the primary socket connection to the server
    primary = new SocketConnection();
    try { 
      primary.open(Site);
      // Send the string which informs the server to start a new session.
      DataOutputStream dStr = new DataOutputStream(primary.getOutputStream());
      dStr.writeBytes(InitString);
      dStr.flush();
      // Read back the server setup message
      String ServerSetUp;
      if (!(((ServerSetUp = primary.getInputStream().readLine()) != null) &&
	    (ServerSetUp.equals(SERVER_SETUP_STRING))))
	throw new CannotConnectException("Server setup failed.");
      // Send back client characteristics
      dStr.writeBytes("praveen password javaclient ");
      dStr.flush();
      // Read in the client ID
      clientId = primary.getInputStream().readInt();
      // Read in the ADT names and instantiate the ADT classes
      int numADTs = primary.getInputStream().readInt();
      ADTs = new ADT[numADTs];
      int nameSize, typeId;
      byte name[];
      for (int i = 0; i < numADTs; i++) {
	typeId = primary.getInputStream().readInt();
	nameSize = primary.getInputStream().readInt();
	name = new byte[nameSize];
	primary.getInputStream().readFully(name);
	try { 
	  String ADTName = new String(name, 0);
	  String ADTString = "PREDATOR.ADT.ADT_" + ADTName;
	  Class ADTClass = Class.forName(ADTString);
	  ADT ADTObject = (ADT)(ADTClass.newInstance());
	  ADTObject.setName(ADTName);
	  ADTObject.setTypeId(typeId);
	  ADTs[typeId] = ADTObject;
	} catch (ClassNotFoundException X) {
	  ADTs[typeId] = new NullADT();
	} catch (InstantiationException X) {
	  throw new PREDATOR.InternalError(X);
	} catch (IllegalAccessException X) {
	  throw new InternalError(X);
	}
      }
      // Read in the socket number of the secondary connection
      int secondaryPort = primary.getInputStream().readInt();
      // Open the secondary socket connection to the server
      secondary = new SocketConnection();
      secondary.open(new ServerSite(primary.getHostAddress(), secondaryPort));
      // Spawn the thread which asynchronously reads off the error stream
      errorBuffer = new ServerErrorBuffer();
      errorReader = new ErrorReader(secondary.getInputStream(), 
                                    secondary.getOutputStream(), 
				    errorBuffer);
      errorReader.setPriority(Thread.MAX_PRIORITY);
      errorReader.start();

      // Read in the socket number of the task connection
	  /*
      int taskPort = primary.getInputStream().readInt();
      // Open the task socket connection to the server
      task = new SocketConnection();
      task.open(new ServerSite(primary.getHostAddress(), taskPort));
	  */

    } catch (IOException X) {
      throw new CannotConnectException(X.getMessage());
    }
    site = Site;
  }
  /**
   * Closes the connection to the 
   * <FONT COLOR="#CC0000"><FONT SIZE=+1>P</FONT>REDATOR</FONT> server.
   * @exception NoOpenConnectionException if no connection is open
   * @exception UnterminatedQueryException if the previous query was not
   *                                       ended using <TT>endQuery()</TT>
   * @exception InterruptedException if this thread is interrupted while
   *                                 closing the connection
   * @see #endQuery()
   */
  public synchronized void close() 
    throws UnterminatedQueryException, NoOpenConnectionException, 
           InterruptedException
  {
    if (!isOpen())
      throw new NoOpenConnectionException();
    if (queryInProcess())
      throw new UnterminatedQueryException();

    /* TODO
      A Netscape bug makes this hang -- lets see if the bug is fixed.
      No, this isnt just a Netscape bug but something deeper.
      It crashes in VJ++ too but in a very strange place  
	  Binquan : It's due to the bug in "Thread.join()", so I changed
	  "errorReader.join()" to "errorReader.join(500)".
	*/
    errorReader.interrupt();
    errorReader.join(5000);

    errorReader.stop();

	try {
      primary.getOutputStream().writeInt(0);
      primary.getOutputStream().flush();
      primary.getInputStream().readLine();
      primary.close();
      secondary.close();
      //task.close();
    } catch (IOException X) {  }
    primary = null;
    site = null;
  }
  /**
   * Processes a query, returning its associated result object.  After this
   * function is called, any number of callbacks can be made for data 
   * objects (assuming the result was a <TT>RecordStreamResult</TT> object).
   * If the <TT>endQuery</TT> parameter is set to false, the query will not
   * be ended automatically after the results are read.  In this case,
   * the <TT>endQuery()</TT> function must be called before a new query can
   * be processed.  Note that while records are still being transferred
   * through a <TT>RecordStreamResult</TT>, the query has not ended.  Only
   * after the last record is read will the query be automatically ended.
   * @param Query The query to process
   * @param enwQuery If this parameter is false, the transaction will not
   *                 be automatically ended after the result is obtained;
   *                 otherwise, it will.
   * @return The result of the query; this can be either an UnformattedResult
   *         or a RecordStream result.
   * @exception NoOpenConnectionException if no connection is open
   * @exception UnterminatedQueryException if the previous query was not
   *                                       ended using <TT>endQuery()</TT>
   * @see #endQuery()
   * @see RecordStreamResult 
   * @see UnformattedResult
   */
  public synchronized QueryResult processQuery(Query Q, boolean endQuery) 
    throws NoOpenConnectionException, 
           UnterminatedQueryException
  {
    if (!isOpen())
      throw new NoOpenConnectionException();
    if (queryInProcess())
      throw new UnterminatedQueryException();
    queryInProcess = true;
    // Reset the error buffer
    errorBuffer.reset();
    try {
      // First send the number of bytes across
      primary.getOutputStream().writeInt(Q.toString().length());
      // Then send the actual data
      DataOutputStream dStr = new DataOutputStream(primary.getOutputStream());
      dStr.writeBytes(Q.toString());
      dStr.flush();
      // Now read the results
      ((FilterInputStream)primary.getInputStream()).mark(2);
      byte b;
      b = primary.getInputStream().readByte();
      if (b == FORMATTED_RESULTS_SENTINEL) {
	RecordStreamResult result = new RecordStreamResult(new Schema());
	// Read the transaction ID, size first
	xactId = new byte[primary.getInputStream().readInt()];
	primary.getInputStream().readFully(xactId);
	// Read the schema
	int schemaArity = primary.getInputStream().readInt();
	int attrTypeId, attrNameLen, attrOffset, metaInfoSize;
	ADT attrADT;
	byte attrName[], metaInfoBytes[];
	ADTMetaInfo MetaInfo = null;
	for (int i = 0; i < schemaArity; i++) {
	  // Read the type id
	  attrTypeId = primary.getInputStream().readInt();
	  attrADT = ADTs[attrTypeId];
	  // Read the attr name
	  attrNameLen = primary.getInputStream().readInt();
	  attrName = new byte[attrNameLen];
	  primary.getInputStream().readFully(attrName);
	  // Read the attr offset
	  attrOffset = primary.getInputStream().readInt();
	  // Read the ADT's meta-info (size first)
	  metaInfoSize = primary.getInputStream().readInt();
	  if (metaInfoSize > 0) {
	    metaInfoBytes = new byte[metaInfoSize];
	    primary.getInputStream().readFully(metaInfoBytes);
	    try {
	      MetaInfo = attrADT.newMetaInfo(metaInfoBytes);
	    } catch (Exception X) {
	      throw new InternalError(X);
	    }
	  } else {
	    MetaInfo = null;
	  }
	  Attribute attr = new Attribute(attrADT, MetaInfo, 
					 new String(attrName, 0), attrOffset);
	  result.getSchema().addAttribute(i, attr);
	}
	RecordReader reader = new RecordReader(this, result, endQuery);
	// The record reader will unlock the primary socket when it is finished
	reader.start();
	return result;
      } else {
	((FilterInputStream)primary.getInputStream()).reset();
	UnformattedResult result = readUnformattedResult();
	primary.unlock();
	if (endQuery) {
	  try {
	    endQuery();
	  } catch (Exception X) {
	    throw new InternalError(X);
	  }
	}
	return result;
      }
    } catch (IOException X) {
		System.out.println(X + "\ntostring :" + X.toString() + "\nmessage : " + X.getMessage());
      throw new InternalError(X);
    }
  }
}
