package PREDATOR;

import java.io.*;
import PREDATOR.ADT.*;

/**
 * Reads off formatted (record) results from the stream asynchronously, 
 * permitting the interface to perform other operations while results
 * are being read.
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version 0.9
 */
class RecordReader extends Thread {
  /**
   * The connection which spawned this thread
   */
  protected Connection conn;
  /**
   * The record stream result object into which the records are read
   */
  protected RecordStreamResult result;
  /**
   * true if the transaction query should be ended as soon as the stream 
   * is read completely.
   */
  protected boolean endQuery;
  /**
   * Default Constructor.
   * @param conn The connection object which spawned this thread
   * @param result The result in which the records are placed
   * @param endQuery true if the transaction query should be ended 
   *                 as soon as the stream is read completely.
   */
  RecordReader(Connection conn,
	       RecordStreamResult result,
	       boolean endQuery)
  {
    this.conn = conn;
    this.result = result;
    this.endQuery = endQuery;
  }
  /**
   * The thread's run method.  Reads records off the stream, inserting them
   * into the record stream.  Unlocks the socket when it is complete.
   */
  public void run() {
    DataRecord Rec;
    int recSize;
    DataInputStream IStr = conn.primary.getInputStream();
    try {
      //while (IStr.available() > 2) {
      while (true) {
	IStr.mark(3); // careful: need to backtrack
	if ((IStr.readByte() == (byte)'\n') &&
	    (IStr.readByte() == (byte)0)) {
	  IStr.reset(); // make sure to backtrack
	  break;
	}
	IStr.reset(); // make sure to backtrack
	// Read in the record length
	recSize = IStr.readInt();
	Rec = new DataRecord(recSize);
	IStr.readFully(Rec.Data);
	// Add the record to the result object
	result.putRecord(Rec);
      } 
      IStr.readByte(); // get rid of the \n from server
      IStr.readByte(); // get rid of the \0 from server
    } catch (IOException X) {
      throw new PREDATOR.InternalError(X);
    } catch (InterruptedException X) {
      throw new PREDATOR.InternalError(X);
    }
    result.setNoMoreInput();
    conn.primary.unlock();
    if (endQuery) {
      try {
	conn.endQuery();
      } catch (Exception X) {
	throw new InternalError(X);
      }
    }
  }
}
