package PREDATOR;

import java.io.*;

/**
 * A server error.
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version 0.9
 */
public class ServerError {
  /// The error message
  String message;
  /// The first (inner-most) error to occur
  ErrorStackNode stack;

  /**
   * Constructor.
   * @param message The general error message
   */
  ServerError(String message)
  {
    this.message = message;
  }
  /**
   * Pushes a stack entry onto the error stack.
   */
  void pushError (String filename, int line, 
		  String errorString, String detail)
  {
    if (stack == null)
      stack = new ErrorStackNode(filename, line, errorString, detail);
    else
      stack.push(new ErrorStackNode(filename, line, errorString, detail));
  }

  /**
   * Returns the message associated with the error.
   * @return The error message
   */
  public String getMessage()
  {
    return message;
  }
  /**
   * Prints a stack trace of the error.
   * @param OutStr The stream on which the error should be printed
   */
  public void printStackTrace(PrintStream OutStr)
  {
    printStackTrace(OutStr, 2);
  }
  /**
   * Prints a stack trace of the error at a specific error level
   * @param OutStr The stream on which the error should be printed
   */
  public void printStackTrace(PrintStream OutStr, int ErrorLevel)
  {
    if (stack != null)
      stack.print(OutStr, ErrorLevel);
  }
}

class ErrorStackNode {
  // The filename
  String filename;
  // The line number in that file
  int line;
  // The error string
  String errorString;
  // The detail associated with the error
  String detail;
  // A handle on the error which caused this error, or null if there was none
  ErrorStackNode cause;
  /**
   * Constructor.
   */
  ErrorStackNode (String filename, int line, 
		  String errorString, String detail)
  {
    this.filename = filename;
    this.line = line;
    this.errorString = errorString;
    this.detail = detail;
  }
  /**
   * Pushes a stack node.
   */
  void push(ErrorStackNode node)
  {
    if (cause == null)
      cause = node;
    else
      cause.push(node);
  }
  /**
   * Prints this error node and its cause to the stream.
   */
  void print(PrintStream OutStr, int ErrorLevel)
  {
    if (ErrorLevel > 0) {
      OutStr.print(errorString + ": " + detail);
      if (ErrorLevel > 1) {
	OutStr.print(":File " + filename + ", Line " + line );
      }
      OutStr.println("");
    }
    if (cause != null)
      cause.print(OutStr, ErrorLevel);
  }
}
