package PREDATOR;

import java.net.*;
import java.io.*;

/**
 * This class encapsulates a socket connection and the streams it holds.
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version 0.9
 */
class SocketConnection {
  /// The socket connection to the server
  protected Socket socket = null;
  /// The input stream to the socket
  protected DataInputStream inStr = null;
  /// The output stream to the socket
  protected DataOutputStream outStr = null;
  /// true if the socket has been "locked" by a user
  private boolean isLocked = false;
  /**
   * Opens the socket connection.
   * @param Site The site to connect to
   * @exception CannotConnectException if the connection fails
   */
  void open(ServerSite Site) throws IOException
  {
    socket = new Socket(Site.getHostAddress(), Site.getPortNumber());
    inStr = new DataInputStream(new BufferedInputStream(socket.
							getInputStream()));
    outStr = 
      new DataOutputStream(new BufferedOutputStream(socket.
						    getOutputStream()));
  }
  /**
   * Closes the socket connection.
   * @exception NoOpenConnectionException if no connection is open
   * @exception IOException if an error occurs closing the socket
   */
  void close() throws NoOpenConnectionException, IOException
  {
    if (socket == null)
      throw new NoOpenConnectionException();
    inStr.close();
    outStr.close();
    socket.close();
  }
  /**
   * Gets a handle on the input stream to the socket.
   * @return The input stream
   */
  DataInputStream getInputStream()
  {
    return inStr;
  }
  /**
   * Gets a handle on the output stream from the socket.
   * @return The output stream
   */
  DataOutputStream getOutputStream()
  {
    return outStr;
  }
  /**
   * Gets the internet address of the host.
   * @return The internet address of the host
   */
  InetAddress getHostAddress()
  {
    return socket.getInetAddress();
  }
  /**
   * Gets the port number of the host.
   * @return The port number of the host
   */
  int getPort()
  {
    return socket.getPort();
  }
  /**
   * Marks this socket as locked, and not to be used by another thread.
   * @return true if the lock is successful, false if it is already locked
   */
  synchronized boolean lock()
  {
    if (isLocked)
      return false;
    else {
      isLocked = true;
      return true;
    }
  }
  /**
   * Unlocks this socket.
   */
  synchronized void unlock() 
  {
    isLocked = false;
    notifyAll();
  }
}
