package JDBCGui;

import java.sql.*;
import java.awt.*;

/**
* This is a GUI which connects to a Predator server
* displays all available tables and runs queries.
* @version 1.0
* @author Emre Tezel
* @since JDK1.1
*/

public class PJGui extends Frame {

	private Button but1, but2, but3, but4, but5;
	private TextArea display;
	private Panel p1, p2;
	private Connection con;
	private Statement stmt;
	private MyDialog viewDialog;
	private String url;
	
	
	/** This is the main method. Notice that this is not an
	* applet but an application so there must be a main method.
	* The class itself extends java.awt.Frame. Inside the main
	* method I create a new instance of PJGui which is a frame 
	* and I show it
	* @param args[]   inputs from the user
	* @see java.awt.Frame
	*/
	public static void main(String args[]) {

		PJGui f = new PJGui();
		f.url = "jdbc:predator://" + args[0] + ":" + args[1];
		f.makeConnection();
		f.initializeGui();
		f.showTables();
		f.pack();
		f.show();

	}

	/** The constructor. First it sets the name of the Frame to "JGui"
	* then it opens a connection to the database
	* after that it initializes the layout of the GUI
	* then it calls the method showTables() to display the present tables
	* @see makeConnection();
	* @see initializeGui();
	* @see showTables();
	*/
	
	  public PJGui() {

		super ("JGui");
		
	}
	
	/**
	* This methos opens the actual connection to the database
	* Class.forName("Predator_drv") automatically loads and registers
	* the Predator driver into the DriverManager
	* then getConnection() method is called which makes DriverManager
	* to look at all the available drivers and use the one which can handle
	* the url
	* 
	*/
	  
	  public void makeConnection() {

		try {

			Class.forName("JDBCDriver.DriverImpl");
			con = DriverManager.getConnection(url, "", "");
		}
		catch (Exception e) {
			System.out.println(e.getMessage());
		}
	
	}

	
	/**
	* This method has tthe job to terminate the Frame
	* if needed
	*/
	
	public boolean handleEvent( Event e) {

		if ( e.id == Event.WINDOW_DESTROY ) {

			try{
				con.close();
			} catch (SQLException i) {
				System.out.println(i.getMessage());
			}
			hide();
			dispose();
			System.exit(0);
			return true;
		}

		return super.handleEvent(e);
	}
	
	
	/**
	* this method takes an SQL statement ( A String ) as a
	* parameter and displays all the query results.
	* @param s  the SQL Query String
	*/
	
	public void makeQuery(String s) {

		ResultSet rs = null;
		String label = "";
		int i = 0;
		try {
			rs = stmt.executeQuery(s);
			ResultSetMetaData meta = rs.getMetaData();
			while (true) {
				String name = meta.getColumnName(i);
				String type = meta.getColumnTypeName(i);
				label = label + "   " + name + ":" + type;
				i++;
			}	
		} catch (Exception e) {

			display.setText(label + "\n");
		}
		
		try {
			String temp = "";
			while (rs.next()) {
				String record = "";
				for ( int n=0; n < i; n++) {
					record = record + "   " + rs.getString(n);
				}
				if ( !temp.equals(record)) display.append(record + "\n");
				temp = record;
				
			}
			
			stmt.close();
			
		} catch (Exception e) {
		}

	}
	
	/**
	* this method takes an SQL statement ( A String ) as a
	* parameter, updates the table accordingly but does not display the results.
	* @param s  the SQL Update String
	*/
	
	public void updateTable(String s) {
		try {
			stmt.executeUpdate(s);
			stmt.close();
		} catch (Exception e) {
			System.out.println(e.getMessage());
		}
	}
	
	/** This method takes the name of a table as a parameter 
	* and displays all of his contents.
	* @param s  name of the table
	*/
	
	public void viewTable(String s){
		
		int i = 0;
		String label = "";
		ResultSet rs = null;
		try {
			String query = "SELECT * FROM " + s;
			rs = stmt.executeQuery(query);
			ResultSetMetaData meta = rs.getMetaData();
			while (true) {
				String name = meta.getColumnName(i);
				String type = meta.getColumnTypeName(i);
				label = label + "   " + name + ":" + type;
				i++;
			}
		} catch (Exception e) {

			display.setText(label + "\n");
		}

		try {
			String temp = "";
			while (rs.next()) {
				String record = "";
				for ( int n=0; n < i; n++) {
					record = record + "   " + rs.getString(n);
				}
				if ( !temp.equals(record)) display.append(record + "\n");
				temp = record;
			}
			
			stmt.close();
			
		} catch (Exception e) {
		}

	}



	/** 
	* This method is responsible of what to do
	* when a button is pressed. 
	* It brings out either a MyDialog, UpdateDialog or a QueryDialog.
	* @see MyDialog
	* @see QueryDialog
	* @see UpdateDialog
	* @param e
	* @param o
	*/
	
	public boolean action(Event e, Object o) {

		if ( e.target == but1 ) {

			viewDialog = new MyDialog(this);
		}
		else if ( e.target == but4 ) {
			showTables();
		}

		else if (e.target == but3) {

			QueryDialog Q = new QueryDialog(this);
		}

		else if (e.target == but2) {

			UpdateDialog Q = new UpdateDialog(this);
		}

		else if ( e.target == but5 ) {
			
			try{
				con.close();
			} catch (SQLException i) {
				System.out.println(i.getMessage());
			}
			hide();
			dispose();
			System.exit(0);
		
		}




		return true;
	}

	
	/**
	* This method display all of the present tables 
	* of the database.
	*/
	public void showTables() {
		
		try {
			display.setText("");
			stmt = con.createStatement();
			ResultSet rs = stmt.executeQuery("select * from _SATTRS");
			String temp = "";
			rs.next();
			while (rs.next()) {
				String name = rs.getString(0);
				if (name == null) break;
				if (!name.equals(temp)) {
					display.appendText(name + "\n");
				}
				temp = name;
			}
			
			stmt.close();	
		} catch ( Exception e ) {
			System.out.println(e.getMessage());
		}
		

	}
	
	
	public void initializeGui() {

		but1 = new Button("View Table");
		but2 = new Button("Update Table");
		but3 = new Button("Make Query");
		but4 = new Button("All Tables");
		but5 = new Button("Close");
		display = new TextArea(30,75);

		GridBagConstraints c = new GridBagConstraints();
		GridBagLayout GridBag = new GridBagLayout();
		setLayout(GridBag);
		c.gridheight = 5;
		c.gridwidth = 3;
		GridBag.setConstraints(display, c);
		add(display);
		c.gridheight = 1;
		c.gridwidth = GridBagConstraints.REMAINDER;
		c.gridx = GridBagConstraints.RELATIVE;
		GridBag.setConstraints(but1, c);
		add(but1);
		c.gridy = GridBagConstraints.RELATIVE;
		c.gridx = GridBagConstraints.RELATIVE;;
		GridBag.setConstraints(but2, c);
		GridBag.setConstraints(but3, c);
		GridBag.setConstraints(but4, c);
		GridBag.setConstraints(but5, c);
		add(but2);
		add(but3);
		add(but4);
		add(but5);

		resize(200,200);


		
	}
}






