package interfaces.JGUI;

import java.awt.*;
import PREDATOR.*;
import PREDATOR.ADT.*;
import jclass.table.*;
import jclass.util.*;

/**
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version 0.9
 */
public class JCAnswerPanel extends Panel implements JCSelectListener {
  /// The underlying JCTable
  protected JCTable DisplayTable;
  /// The current schema
  Schema schema;
  /// A handle on the applet
  JGUI gui;

  /**
   * Constructor.
   */
  public JCAnswerPanel(JGUI gui) {
    this.gui = gui;
    setLayout(new BorderLayout());
    DisplayTable = new JCTable();
    DisplayTable.setRepaint(true);
    DisplayTable.addSelectListener(this);	
    DisplayTable.setVisibleRows(15);
    DisplayTable.setVisibleColumns(4);
    DisplayTable.setSelectionPolicy(JCTblEnum.SELECT_SINGLE);
    DisplayTable.setHorizSBDisplay(JCTblEnum.SBDISPLAY_AS_NEEDED);
    DisplayTable.setVertSBDisplay(JCTblEnum.SBDISPLAY_AS_NEEDED);
    DisplayTable.setBackground(JCTblEnum.LABEL, JCTblEnum.ALL, Color.red);
    DisplayTable.setForeground(JCTblEnum.LABEL, JCTblEnum.ALL, Color.white);
    DisplayTable.setAlignment(JCTblEnum.LABEL, JCTblEnum.ALL, 
			      JCTblEnum.TOPCENTER);
    DisplayTable.setFont(JCTblEnum.LABEL, JCTblEnum.ALL, 
			 new Font("Helvetica", Font.PLAIN, 16));
    DisplayTable.setFont(JCTblEnum.ALL, JCTblEnum.ALL, 
			 new Font("Helvetica", Font.PLAIN, 14));
    DisplayTable.setFrameBorderType(JCTblEnum.BORDER_FRAME_IN);
    DisplayTable.setFrameShadowThickness(2);
    DisplayTable.setPixelWidth(JCTblEnum.ALL, 125);
    DisplayTable.setPixelHeight(JCTblEnum.ALL, JCTblEnum.VARIABLE);
    DisplayTable.setJumpScroll(JCTblEnum.ALL);
    add("Center", DisplayTable);
  }
	
  /**
   * Reads from the result and populates a table.  By default,
   * none of the cells can be selected.
   */
  public void showAnswers(RecordStreamResult recordStream) {
    schema = recordStream.getSchema();
    DisplayTable.setNumRows(0);
    DisplayTable.setNumColumns(schema.getArity());
    String[] TableColumnLabels = 
      new String[schema.getArity()];
    for (int i = 0; i < schema.getArity(); i++) {
      StringBuffer label = new StringBuffer(schema.getAttribute(i).getName());
      if (schema.getAttribute(i).getMetaInfo() != null)
	label.append("\n[" + schema.getAttribute(i).getMetaInfo() + "]");
      TableColumnLabels[i] = label.toString();
    }
    DisplayTable.setColumnLabels(TableColumnLabels);
    // Now read the answers into the table
    int Row = 0;
    DataObject record[];
    try {
      while (!recordStream.isEmpty()) {
	record = recordStream.getNextRecord(null);
	// Display record
	DisplayTable.addRow(JCTblEnum.MAXINT, null, null);
	DisplayTable.setEditable(Row, 0, false);
	DisplayTable.setCell(Row, 0, String.valueOf(Row));
	for (int Col = 0; Col < schema.getArity(); Col++) {
	  DisplayTable.setEditable(Row, Col, false);
	  DisplayTable.setCell(Row, Col, 
			       record[Col].
			       toString(schema.getAttribute(Col).
					getMetaInfo()));
	  DisplayTable.setUserdata(Row, Col, record[Col]);
	  if (record[Col] instanceof CachingObject) {
	    DisplayTable.setForeground(Row, Col, Color.red); 
	  }
	  if (record[Col] instanceof DisplayObject) {
	    DisplayTable.setForeground(Row, Col, Color.green); 
	  }
	  if ((record[Col] instanceof MIMETypeObject) && 
	      (record[Col] instanceof CallbackObject)) {
	    DisplayTable.setForeground(Row, Col, Color.blue); 
	  }
	}
	gui.showStatus("Processed record: " + Row++);
      }
    } catch (InterruptedException X) {
      // this should never happen
    } catch (BadSubclassSelectorException X) {
      // this should never happen
    }
  }
  /**
   * Performs the action necessary when a selection click is made.
   * Currently does nothing.
   * @param event The JCSelectEvent
   */
  public void selectBegin (JCSelectEvent event) {
    return;
  }
  /**
   * Performs the action necessary when a selection click is released.
   * @param event The JCSelectEvent
   */
  public void selectEnd (JCSelectEvent event) {
    if (event.SELECT != JCSelectEvent.SELECT)
      return;
    int Col = event.getColumn();
    int Row = event.getRow();
    DataObject Object = (DataObject)DisplayTable.getUserdata(Row, Col);
    if ((Object instanceof CachingObject) ||
	(Object instanceof DisplayObject) || 
	((Object instanceof MIMETypeObject) && 
	 (Object instanceof MIMETypeObject))) {
      ADT AttrADT = schema.getAttribute(Col).getTypeADT();
      ADTMetaInfo MetaInfo = schema.getAttribute(Col).getMetaInfo();
      new SelectionThread(Object, AttrADT, MetaInfo, gui).start();
    }
  }
}
