package interfaces.JGUI;

import java.applet.*;
import java.awt.*;
import java.util.*;
import PREDATOR.*;
import PREDATOR.ADT.*;

/**
 * This is a simple example of an applet client using PREDATOR's Java
 * Client Library.
 * @author Mark Paskin, paskin@cs.cornell.edu
 * @version 0.9
 */
public class JGUI extends Applet {
  /// The connection to PREDATOR
  Connection connection;
  /// The connection panel
  Panel connectionPanel;
  /// The button which makes the connection
  Button connectButton;
  /// The query panel
  Panel queryPanel;
  /// The text area in which the query is written
  TextArea queryTextArea;
  /// The button which submits a query
  Button submitButton;
  /// The button which clears the query text area
  Button clearButton;
  /// The button which quits the connection
  Button quitButton;
  /// The unformatted results panel
  Panel unformattedResultsPanel;
  /// The text area in which the unformatted results are written
  TextArea unformattedTextArea;
  /// The error panel
  Panel errorPanel;
  /// The formatted results panel
  Panel recordResultPanel;
  /// The JC answer panel
  JCAnswerPanel answerPanel;
  /// The button which ends the current query
  Button unformattedNewQueryButton;
  /// The button which ends the current query
  Button errorNewQueryButton;
  /// The button which ends the current query
  Button recordNewQueryButton;
  /// True if the applet is reading a query; false if displaying answers
  boolean queryMode = true;

  /**
   * Called by the browser when the applet is loaded.
   */ 
  public void init()
  {
    setLayout(new FlowLayout());
    setFont(new Font("Helvetica", Font.BOLD, 14));
    setBackground(Color.white);
    // Create the query panel
    queryPanel = new Panel();
    queryPanel.setLayout(new BorderLayout());
    queryPanel.add("North", new Label("Enter your query below."));
    queryPanel.add("Center", queryTextArea = new TextArea(15, 70));
    Panel buttonPanel = new Panel();
    buttonPanel.setLayout(new GridLayout(1, 3));
    buttonPanel.add(submitButton = new Button("Submit (F1)"));
    buttonPanel.add(clearButton = new Button("Clear (F2)"));
    buttonPanel.add(quitButton = new Button("Quit (F3)"));
    queryPanel.add("South", buttonPanel);
    queryPanel.validate();
    // Create the connection panel
    connectionPanel = new Panel();
    connectionPanel.setLayout(new FlowLayout());
    connectionPanel.add(connectButton = new Button("Reconnect"));
    connectionPanel.validate();
    // Create the unformatted results panel
    unformattedResultsPanel = new Panel();
    unformattedResultsPanel.setLayout(new BorderLayout());
    unformattedResultsPanel.add("North", new Label("Results:"));
    unformattedResultsPanel.add("Center", 
				unformattedTextArea = new TextArea(15, 70));
    Panel gridPanel = new Panel();
    gridPanel.setLayout(new GridLayout(1, 3));
    gridPanel.add(new Panel());
    gridPanel.add(unformattedNewQueryButton = new Button("New Query (F1)"));
    gridPanel.add(new Panel());
    unformattedResultsPanel.add("South", gridPanel);
    unformattedResultsPanel.validate();
    // Create the unformatted results panel
    errorPanel = new Panel();
    errorPanel.setLayout(new BorderLayout());
    gridPanel = new Panel();
    gridPanel.setLayout(new GridLayout(1, 3));
    gridPanel.add(new Panel());
    gridPanel.add(errorNewQueryButton = new Button("Continue (F1)"));
    gridPanel.add(new Panel());
    errorPanel.add("South", gridPanel);
    errorPanel.validate();
    // Create the formatted results panel
    recordResultPanel = new Panel();
    recordResultPanel.setLayout(new BorderLayout());
    recordResultPanel.add("North", new Label("Your query's results:"));
    recordResultPanel.add("Center", answerPanel = new JCAnswerPanel(this));
    gridPanel = new Panel();
    gridPanel.setLayout(new GridLayout(1, 3));
    gridPanel.add(new Panel());
    gridPanel.add(recordNewQueryButton = new Button("New Query (F1)"));
    gridPanel.add(new Panel());
    recordResultPanel.add("South", gridPanel);
    // Create the connection
    connection = new Connection();
  }
  /**
   * Called by the browser when the applet is destroyed.
   */ 
  public void destroy()
  {
    closeConnection();
  }
  /**
   * Called by the browser when the applet is started.
   */ 
  public void start()
  {
    openConnection();
    newQuery();
  }
  /**
   * Called by the browser when the applet is stopped.
   */ 
  public void stop()
  {
    closeConnection();
  }
  /**
   * The applet's action handler.
   */
  public boolean action(Event evt, Object what)
  {
    if (evt.target == connectButton) {
      start();
      return true;
    }
    if (evt.target == quitButton) {
      closeConnection();
      return true;
    }
    if (evt.target == submitButton) {
      processQuery();
      return true;
    }
    if (evt.target == clearButton) {
      queryTextArea.setText("");
      return true;
    }
    if ((evt.target == errorNewQueryButton) || 
	(evt.target == unformattedNewQueryButton) || 
	(evt.target == recordNewQueryButton)) {
      newQuery();
      return true;
    }
    return false;
  }
  /**
   * The applet's keypress handler.
   */
  public boolean keyDown(Event evt, int key)
  {
    if (key == Event.F1) {
      if (queryMode)
	processQuery();
      else
	newQuery();
      return true;
    }
    if ((key == Event.F2) && queryMode) {
      queryTextArea.setText("");      
      return true;
    }
    if ((key == Event.F3) && queryMode) {
      closeConnection();
      return true;
    }
    return false;
  }
  /**
   * Removes all elements, displaying a single message to the user.
   */
  protected void showMessage(String message)
  {
    removeAll();
    add(new Label(message));
    validate();
    showStatus(message);
  }
  /**
   * Closes the connection cleanly.
   */
  protected void closeConnection()
  {
    showMessage("Closing connection...");
    try {
      if ((connection != null) && (connection.isOpen())) {
	if (connection.queryInProcess())
	  connection.endQuery();
	connection.close();
      }
    } catch (InterruptedException X) {
      throw new 
	java.lang.InternalError("Interrupted while closing connection: " + X);
    } catch (NoOpenConnectionException X) {
      throw new java.lang.InternalError("The connection was not open: " + X);
    } catch (NoQueryRunningException X) {
      throw new java.lang.InternalError("There was no query running: " + X);
    } catch (UnterminatedQueryException X) {
      throw new 
	java.lang.InternalError("There was an unterminated query: " + X);
    }
    showMessage("Connection closed.");
    removeAll();
    add(connectionPanel);
    validate();
  }
  /**
   * Opens the connection to the server.
   */
  protected void openConnection()
  {
    // Get the port number embedded in the web page as a parameter
    String PortString = getParameter("PORT");
    if (PortString == null) {
      showMessage("Cannot start JGUI: PORT parameter not supplied");
      stop();
    }
    int Port = 0;
    try {
      Port = (Integer.valueOf(PortString)).intValue();
    } catch (java.lang.NumberFormatException X) {
      showMessage("Cannot start JGUI: PORT parameter was not an integer: " 
		  + PortString);
      stop();
    }
    try {
      ServerSite Site = new ServerSite(getCodeBase().getHost(), Port);
      connection.open(Site);
      showStatus("Connection established.");
    } catch (java.net.UnknownHostException X) {
      showMessage("Connection failed: Unknown Host");
      stop();
    } catch (CannotConnectException X) {
      showStatus("Connection failed.");
      stop();
    } catch (ConnectionAlreadyOpenException X) {
      throw new java.lang.InternalError("A connection was already open.");
    }
    if (connection.isError())
      processServerError(connection.getError());
  }
  /**
   * Displays the query panel.
   */
  public void newQuery()
  {
    queryMode = true;
    if (connection.queryInProcess())
      try {
	connection.endQuery();
      } catch (Exception X) {
	throw new java.lang.InternalError(X.toString());
      }
    removeAll();
    add(queryPanel);
    validate();
  }
  /**
   * Processes the query and displays the results.
   */
  public void processQuery()
  {
    String QueryText = queryTextArea.getText().trim();
    if (QueryText.length() < 1) {
      showStatus("Type an SQL query in the text box; then press \"Submit\".");
      return;
    }
    Query query;
    try {
      query = new Query("SQL", QueryText);
    } catch (BadQueryException X) {
      showStatus("The query is improperly formatted.  " + 
		 "Remember that semicolons are disallowed.");
      return;
    }
    queryMode = false;
    // Lock the display
    showMessage("Processing query...");
    if (connection.queryInProcess())
      try {
	connection.endQuery();
      } catch (Exception X) {
	throw new java.lang.InternalError(X.toString());
      }
    try {
      QueryResult Result = null;
      try {
	Result = connection.processQuery(query, false);
      } catch (UnterminatedQueryException X) {
	throw new 
	  java.lang.InternalError("The last query was improperly terminated.");
      }
      if (Result == null)
	processNoResponse();
      else if (Result instanceof UnformattedResult)
	processUnformatted((UnformattedResult)Result);
      else if (Result instanceof RecordStreamResult)
	processRecordStream((RecordStreamResult)Result);
      else
	throw new java.lang.InternalError("Unrecognized result type.");
    } catch (NoOpenConnectionException X) {
      throw new java.lang.InternalError("The connection is not open");
    }
    if (connection.isError())
      processServerError(connection.getError());
  }
  /**
   * Processes a query which has no result.
   */
  protected void processNoResponse()
  {
    showStatus("Command processed.");
    newQuery();
  }
  /**
   * Processes a server error.
   */
  protected void processServerError(ServerError Error)
  {
    removeAll();
    errorPanel.add("North", new Label("Server Error: " + Error.getMessage()));
    errorPanel.add("Center", 
		   new Label("Check the Java console for more information."));
    Date currentDate = new Date();
    System.err.println("A system occurred at " + currentDate + 
		       ".  Stack trace:");
    Error.printStackTrace(System.err);
    add(errorPanel);
    validate();
  }
  /**
   * Processes an unformatted result from a query.
   */
  protected void processUnformatted(UnformattedResult Unformatted)
  {
    removeAll();
    unformattedTextArea.setText(Unformatted.toString());
    add(unformattedResultsPanel);
    validate();
  }
  /**
   * Processes a RecordStream result from a query.
   */
  protected synchronized void processRecordStream(RecordStreamResult 
						  RecordStream)
  {
    removeAll();
    answerPanel.showAnswers(RecordStream);
    add(recordResultPanel);
    validate();
  }
}

