package interfaces.JGUI;

import java.net.*;
import PREDATOR.*;
import PREDATOR.ADT.*;

/**
 * Asynchronously performs the actions necessary when a JCAnswerPanel cell
 * is selected.  Current behavior:
 * <OL>
 *   <LI> If the DataObject inherits from <TT>CallbackObject</TT>
 *        the necessary callback is made.
 *   <LI> If the DataObject inherits from <TT>DisplayObject</TT>
 *        the display method is called.
 * </OL>
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version 0.9
 */
public class SelectionThread extends Thread {
  /// The DataObject associated with the selected cell
  DataObject Object;
  /// The ADT of the object
  ADT TypeADT;
  /// The object's meta-info
  ADTMetaInfo MetaInfo;
  /// A handle on the main applet
  JGUI gui;

  public SelectionThread (DataObject Object, ADT TypeADT, 
			  ADTMetaInfo MetaInfo, JGUI gui) {
    this.Object = Object;
    this.TypeADT = TypeADT;
    this.MetaInfo = MetaInfo;
    this.gui = gui;
  }

  public void run() {
    if ((Object instanceof CachingObject) && 
	(((CachingObject)(Object)).requiresCallback(MetaInfo))) {
      try {
	gui.showStatus("Transferring object from server...");
	URL callbackURL = 
	  gui.connection.getCallbackObjectURL(Object, TypeADT, MetaInfo);
	((CachingObject)(Object)).callback(callbackURL, MetaInfo);
	gui.showStatus("Transfer complete.");
      } catch (Exception X) {
	gui.showStatus("Transfer failed!  " + 
		       "Examine the standard output for details.");
	X.printStackTrace();
	return;
      }
    }
    try {
      if ((Object instanceof MIMETypeObject) && 
	  (Object instanceof CallbackObject)) {
	gui.showStatus("Displaying the object...");
	gui.getAppletContext().
	  showDocument(gui.connection.
		       getCallbackObjectURL(Object, TypeADT, MetaInfo),
		       "Requested Object");
	gui.showStatus("Display complete.");
      } else if (Object instanceof DisplayObject) {
	gui.showStatus("Displaying the object...");
	((DisplayObject)(Object)).display(MetaInfo);
	gui.showStatus("Display complete.");
      }
    } catch (Exception X) {
      // this should never happen
    }
  }
}
  
