#include "alias.h"
#include "expr.h"
#include "shoresm.h"
#include "filestream.h"
#include "typelist.h"
#include "xxx_thread.h"
#include "aggrexpr.h"
#include "optrules.h"

//---------------------------------------------------------------
XxxErrCode XxxAliasADT::Register()
{
  // Create serialized data what this aliasADT aliases to.
  int toAliasLen = strlen(AliasTo->TypeName);
  int metaLen, typeInfoSize;
  char* typeInfoData;
  if(MetaString){
    metaLen = strlen(MetaString);
    typeInfoSize = sizeof(int)*3+toAliasLen+metaLen;
    typeInfoData = new char[typeInfoSize];
    *((int *)typeInfoData) = TypeId;
    *((int *)(typeInfoData+sizeof(int))) = toAliasLen;
    memcpy(typeInfoData+2*sizeof(int), AliasTo->TypeName, toAliasLen);
    *((int *)(typeInfoData+2*sizeof(int)+toAliasLen)) = metaLen;
    memcpy(typeInfoData+3*sizeof(int)+toAliasLen, MetaString, metaLen);
  } else {
    metaLen = 0;
    typeInfoSize = sizeof(int)*3+toAliasLen;
    typeInfoData = new char[typeInfoSize];
    *((int *)typeInfoData) = TypeId;
    *((int *)(typeInfoData+sizeof(int))) = toAliasLen;
    memcpy(typeInfoData+2*sizeof(int), AliasTo->TypeName, toAliasLen);
    *((int *)(typeInfoData+2*sizeof(int)+toAliasLen)) = metaLen;
  }
	
  // Insert all the type infos into the default file, and store the 
  // handles for them into the "TypeList" file.
  TypeInfo* aliasInfo = new TypeInfo(TypeId, TypeName, "alias", 
				     typeInfoSize, typeInfoData);
	
  delete[] typeInfoData;
	
  void *aliasData = aliasInfo->Serialize();
  int aliasDataSize = aliasInfo->DataSize();
	
  delete aliasInfo;
	
  XxxShoreObjectId aliasDataId;
	
  XXX_CALL_RET(XxxShoreSM->InsertRecord(XxxShoreSM->defaultVolId, XxxShoreSM->typeFileId, 
					aliasData, aliasDataSize, aliasDataId), XXX_SHOREERR, 
	       "Cannot insert object into persistent file");
	
  delete[] aliasData;
	
  return XXX_OK;
}

//---------------------------------------------------------------
XxxErrCode XxxAliasADT::UnSerialize(const char* data, XxxBasicType& Id, char *&aliasToName, char *&MetaString){
  Id = *((int *)data);
  int aliasToLen = *((int *)(data + sizeof(int)));
  aliasToName = new char[aliasToLen + 1];
  memcpy(aliasToName, data+2*sizeof(int), aliasToLen);
  aliasToName[aliasToLen] = '\0';
	
  int metaLen = *((int *)(data+2*sizeof(int)+aliasToLen));
  if(metaLen){
    MetaString = new char[metaLen+1];
    memcpy(MetaString, data+3*sizeof(int)+aliasToLen, metaLen);
    MetaString[metaLen] = '\0';
  } else {
    MetaString = NULL;
  }
	
  return XXX_OK;
}

//---------------------------------------------------------------
XxxAliasADT::XxxAliasADT(XxxBasicType Id, const char *Name, const char *aliasToName, const char *MetaString, XxxErrCode& status) : 
  XxxADTClass(Id, Name, sizeof(int), XXX_TRUE, XXX_TRUE)
{
 XxxUseMutex AutoMutex(*XxxADTClass::Latch);
  int i;
  for(i = 0; i < XxxADTClass::NumADTs; i++) {
    if (XxxADTClass::ADTArray[i] != NULL){
      if (strcmp(aliasToName, XxxADTClass::ADTArray[i]->TypeName)==0)
	break;
    }
  }
	
  if (i == XxxADTClass::NumADTs) {
    XXX_NOTE_ERROR(XXX_BADTYPE,"Cannot find the ADT type to alias with");
    status = XXX_NOT_OK;
    return;
  }
	
  AliasToID = i;
  AliasTo = XxxADTClass::ADTArray[AliasToID];
	
  if (MetaString){
    this->MetaString = new char[strlen(MetaString)+1];
    strcpy(this->MetaString, MetaString);
    if (AliasTo->AddMetaInfo("dummy",XXX_FALSE,MetaString,MIHndl)!=XXX_OK){
      XXX_NOTE_ERROR(XXX_BADTYPE,"In aliasing metainfo");
      status = XXX_NOT_OK;
      return;
    }
  } else {
    this->MetaString = NULL;
  }
	
  //need to check if AliasTo is a basic type: we can't alias to a basic type
  /*if (AliasTo->IsBasicType) {
    XXX_NOTE_ERROR(XXX_BADADT, "Cannot create an alias to a basic type.");
    }*/
	
  //MaintainsCatalogs = (MetaString)?XXX_FALSE:AliasTo->MaintainsMeta();
  maintainsMetaInfo = AliasTo->MaintainsMetaInfo();
  StoreInField = AliasTo->StoreInField;
  status = XXX_OK;
}

XxxAliasADT::~XxxAliasADT(){
  if(MetaString != NULL){
    delete[] MetaString;
  }
}


//---------------------------------------------------------------
XxxErrCode XxxAliasADT::FuncTypeCheck(const char*FuncName,
				      const XxxValueExpression* Owner,
				      const XxxValueExprList *ArgList,
				      XxxBool IsAggr,
				      XxxBasicType& ReturnType,
				      XxxFuncParseInfo*& FuncParse) const
{

  int NumArgs=0;
  const XxxValueExprList *temp = ArgList;
  while(temp != NULL){
    NumArgs++;
    temp = ArgList->Next;
  }

  //need to check if the args is also the alias, change it if it is
  //for purposes of matching a function in the aliasto class
  XxxBool* isAlias = new XxxBool[NumArgs];
  int* aliasList = new int[NumArgs];

  int i;
  for(i=0; i<NumArgs; i++){
    if (XxxADTArray[((*ArgList)[i])->BasicType()]->IsAlias()==XXX_TRUE) {
      isAlias[i]= XXX_TRUE;
      aliasList[i] = ((*ArgList)[i])->BaseType;
      ((*ArgList)[i])->BaseType=((XxxAliasADT *)XxxADTArray[((*ArgList)[i])->BasicType()])->AliasToID;
    } else {
      isAlias[i] = XXX_FALSE;
    }
  }
	
  XxxErrCode returnCode = AliasTo->FuncTypeCheck(FuncName, Owner, ArgList, IsAggr, ReturnType,
						 FuncParse);
	
  //change back to original
  for(i=0; i<NumArgs; i++){
    if(isAlias[i] == XXX_TRUE){
      ((*ArgList)[i])->BaseType=aliasList[i];
    }
  }
	
  delete[] isAlias;
  delete[] aliasList;
  return returnCode;

}



//---------------------------------------------------------------
XxxADTMetaInfo* XxxAliasADT::GetMIInfo(const XxxADTMetaInfo* inMI) const
{
  if (MIHndl.Size==0) return (XxxADTMetaInfo*)inMI;
  XxxADTMetaInfo *MI;
  XxxBool found=XXX_FALSE;

  if(AliasTo->GetMetaInfo(AliasTo->TypeName,MIHndl,MI,found) != XXX_OK){
    XXX_NOTE_ERROR(XXX_BADHANDLE, "error in XxxStructADT::XxxStructADT");
    return NULL;
  }

  if (found==XXX_TRUE)
    return MI;
  else
    return (XxxADTMetaInfo*)inMI;
}
