/* This program will read a binary data file containing 5004 x 2168 (x,y) 
Land AVHRR Pathfinder subsetted Science CDROM gridded parameters in binary 
format and compute some simple statistics. It demonstrates how to access the
binary data files and convert the binary, scaled data to their geophysical
values.

Data wordsize is either 8-bit or 16-bit words. It has been tested
on a SGI 4D/340 VGX workstation running the IRIX 4.05.C  O/S
which uses "Big Endian" architecture.

Compile and link this program by typing

cc read_binary_sgi.c -o read_binary_sgi_c 

Run this program by typing the following command line. (Do not include
the quotation marks). 

avhrr2jpeg <input fn> # <output fn> <pixels / row> <rows / raster> 

Where # is parameter type dependent. 
           1 Channel_1 Reflectance
           2 Channel_2 Reflectance
	   3
           4 Channel_4 Blackbody brightness temperature
           5 Channel_5 Blackbody brightness temperature

Run this program by typing the following command line. (Do not include
For use with Predator, this program will exit with a value of 0
on success, and 1 on failure.

*/

#include "raster.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
extern "C" {
#include "jpeglib.h"
 }
//#define PIXELS 1600
//#define SCANS  2560

#define XXX_WRONG 2

struct RGBHeader {
  unsigned int Width;
  unsigned int Height;
  int Xbase;
  int Ybase;
 };
struct RGBPixel {
  unsigned char r;
  unsigned char g;
  unsigned char b;
 };
class RGB {
 public:
  RGBPixel* DataArray;
  RGBPixel** DataRows;  // array of pixel row pointers
  RGBHeader* Header;
  RGB(long x1, long y1, long x2, long y2);
  ~RGB();
 };

RGB::RGB(long x1, long y1, long x2, long y2)
{
 if (!((x2 >= x1) && (y2 >= y1))) {
   fprintf(stderr, "Bad RGB box!\n");
   exit(XXX_NOT_OK);
  }
 // allocate header
 Header = new RGBHeader();
 Header->Width = x2 - x1 + 1;
 Header->Height = y2 - y1 + 1;
 Header->Xbase = x1;
 Header->Ybase = y1;

 // allocate data area for image
 DataArray = new RGBPixel[Header->Height * Header->Width];

 // allocate row ptr structure and initialize
 DataRows = new (RGBPixel*)[Header->Height];

 int i;
 RGBPixel* p;
 for (i = 0, p = DataArray; i < Header->Height;
      i++, p += Header->Width) {
   *(DataRows + i) = p;
 }
}
RGB::~RGB()
{
 // reclaim row ptr array
 if (DataRows) delete [] DataRows;
 // reclaim actual data array
 if (DataArray) delete [] DataArray;;
 // reclaim header structure
 if (Header) delete Header;
}

int 
gains(int i, float *gain, long *offset)
{
 switch(i) {
  case 1:
   *gain = 0.002;
   *offset = 10;
   break; 
  case 2:
   *gain = 0.002;
   *offset = 10;
   break; 
  case 3: // I'm guessing on the gain and offset here...
  *gain = 0.005;
   *offset = -31990;
   break; 
  case 4:
   *gain = 0.005;
   *offset = -31990;
   break; 
  case 5:
   *gain = 0.005;
   *offset = -31990;
   break; 
  default:
   printf("Incorrect file name specified\n");
   printf("i is %d, %x\n", i, i);
   exit(XXX_NOT_OK);
   break;
  }
 return(1);
}

// Praveen :: scaling down the image for easy visibility
int 
avhrr2RGB(unsigned short* avhrr_data, struct RGB*& rgb_image, 
	  int pixels, int scans, float gain, long offset,
	  float minval, float maxval, float avgval) 
{
 int ScaleFactorX = 2;
 int ScaleFactorY = 2;
 rgb_image = new RGB(0,0,(pixels-1)/ScaleFactorX,(scans-1)/ScaleFactorY);
 
 // step thru rows
 for (int i =0; i < scans; i+=ScaleFactorY) {
   RGBPixel* ImRow = rgb_image->DataRows[i/ScaleFactorY];
   unsigned short* avhrrRow = avhrr_data+(i*pixels);
   // step thru pixels in each row
   for (int j = 0; j < pixels; 
	j+=ScaleFactorX, avhrrRow+=ScaleFactorX, ImRow++) {
     float Value = *avhrrRow;
     ImRow->g = 150;
     ImRow->b = 150;  
     if (Value > 9) 
       Value = (Value - offset)*gain;
     else
       Value = 0;
     // attempt to normalize
     Value = Value-minval; 
     if (Value < 0) {
       printf("this should not happen\n");
       Value = 0; // this should not happen, but in case it does
      }
     // scale value by the range MaxVal to MinVal
     Value = Value*255/(maxval-minval);

     ImRow->r = (unsigned char) (Value)%256;
    }    
  }// for all rows of the new image

 return(1);
}


int 
RGB2Jpeg(RGB* rgb_image, const char* jpeg_file, int pixels, int scans)
{
 // This struct contains the JPEG compression parameters 
 // and pointers to working space (which is allocated as
 // needed by the JPEG library).
 struct jpeg_compress_struct cinfo;
 struct jpeg_error_mgr jerr;
 
 // Step 1: allocate and initialize JPEG compression object 
 cinfo.err = jpeg_std_error(&jerr);
 jpeg_create_compress(&cinfo);

 // Step 2: specify SHORE data destination 
 FILE* fd = fopen(jpeg_file, "wb");
 fprintf(stderr,  "Writing image to: %s\n", jpeg_file);
 jpeg_stdio_dest(&cinfo, fd);

 // Step 3: set parameters for compression
 cinfo.image_width = pixels;
 cinfo.image_height = scans; 
 cinfo.input_components = 3;
 cinfo.in_color_space = JCS_RGB;
 jpeg_set_defaults(&cinfo);
 
 // Step 4: Start compressor 

 // TRUE ensures that a complete interchange-JPEG is written
 jpeg_start_compress(&cinfo, TRUE);

 //send all the data at once */
 jpeg_write_scanlines (&cinfo, 
		       (unsigned char **)(rgb_image->DataRows),
		       cinfo.image_height);

 // Step 6: Finish compression 

 jpeg_finish_compress(&cinfo);

 // Step 7: release JPEG compression object
 
 // This is an important step since it will release a 
 // good deal of memory.
 jpeg_destroy_compress(&cinfo);

 // write to a file 
 fclose(fd);
 // And we're done!
 return(1);
}


XxxErrCode 
XxxRasterADT::avhrrToJpeg(const char* FileName, int freq, 
			  const char* JpegName, int pixels, int scans) const
{

 unsigned short *data_s ;

 float gain;
 long offset;
 float minval, maxval, avgval;

 // read AVHRR file into memory
  FILE *fp1;
 int i,j,index; 
 long count=0;
 long skip_count=0;
 int file_type, result;
 unsigned char *data ;

 float temp_1;

 float min_temp_1 = 1.0e06 , max_temp_1 = 1.0e-06;
 double sum=0.0;
 int ipix_max = pixels;
 int iscan_max  = scans;

 if((fp1=fopen(FileName, "rb")) == NULL)
  {
   XXX_NOTE_ERROR(XXX_BADFILE, "Error: cannot open avhrr file of Raster Type");
   return XXX_NOT_OK;
  }

 result = freq;
 //
 if(result != 1 && result != 2 && result !=4 && result !=5 && result != 3)
  {
   XXX_NOTE_ERROR(XXX_BADVALUE, "Error freq specified");
   return XXX_NOT_OK;
  }
 else
  {
   file_type=result;
  }

 gains(file_type, &gain, &offset);
 //data_s = (unsigned short *)calloc(pixels, sizeof(unsigned short));
 data_s = (unsigned short *)calloc(iscan_max*ipix_max, sizeof(unsigned short));
 if (data_s == NULL) printf(" memory was not allocated correctly \n");
 for (j=0; j< iscan_max; j++)
  {
   unsigned short* data_ptr = data_s+j*ipix_max;
   fread(data_ptr, sizeof(unsigned short), ipix_max, fp1);
   for (i=0;i<ipix_max;i++)
    {
     if( (float) data_ptr[i] > 9  ) 
      { 
       temp_1 =  (float)data_ptr[i];
       temp_1 = (temp_1 - offset)*gain;
       if (min_temp_1 > temp_1) min_temp_1 = temp_1;
       if (max_temp_1 < temp_1) max_temp_1 = temp_1;
       sum = sum + temp_1;
       count++;
      } 
     else {
       skip_count++;
      }
    }      /*   End of for i loop */
  }           /*   End of for j loop  */

 minval = min_temp_1;
 maxval = max_temp_1;
 avgval = sum/count;

 printf(" stats for data set  %s  land pixels only \n",FileName);
 printf(" Number of pixel values = %ld, skips = %ld \n",count, skip_count);
 printf(" min/max of parameter value  %f %f \n",min_temp_1,max_temp_1);
 if(count > 0) printf(" avg parameter value %f \n", sum/count);
 fclose(fp1);


 RGB* rgb_image;

 // convert to RGB format
 if (!avhrr2RGB(data_s, rgb_image, pixels, scans, gain, offset, minval, maxval, avgval)) {
   fprintf(stderr, "Error converting avhrr data to RGB data\n");
   exit(XXX_NOT_OK);
  }
 delete data_s;
 
 // write to jpeg file
 if (!RGB2Jpeg(rgb_image, JpegName, rgb_image->Header->Width,rgb_image->Header->Height)) {
   fprintf(stderr, "Error writing RGB data to jpeg\n");
   exit(XXX_NOT_OK);
  }
 delete rgb_image;
 

 return XXX_OK;
}












