#include "gisadt.h"
#include "math.h"
#include "BoundingBox.h"

XxxBasicType XxxBoxADT::BoxTypeId = XXX_UNKNOWN;

// Constructor for BoxADT
XxxBoxADT::XxxBoxADT(XxxBasicType TypeId) 
 : XxxSmallFixedADT ( TypeId, "box",        
                      sizeof (BoxStruct))   
{
 BoxTypeId = TypeId;                         
 
 int PolygonId = XxxPolygonADT::PolygonTypeId;
 int PointId = XxxPointADT::PointTypeId;;
 // Register the methods for the box ADT
 NumMethods = 18; 
 MethArray = new FuncMethodInfo* [NumMethods]; 

 //Bjarke, sep. 14 2001: added '&' to all calls to FuncMethodInfo
 MethArray[0] = new FuncMethodInfo( "disjoint" , XXX_INT, 
  0 , 0 , (ExecuteFunc)(&PlanInfo::execDisjointp), 1  , PolygonId);

 MethArray[1] = new FuncMethodInfo( "disjoint" , XXX_INT, 
  0 , 0 , (ExecuteFunc)(&PlanInfo::execDisjointb), 1  , TypeId);

 MethArray[2] = new FuncMethodInfo( "meets" , XXX_INT, 
  0 , 0 , (ExecuteFunc)(&PlanInfo::execMeetsp), 1  , PolygonId);

 MethArray[3] = new FuncMethodInfo( "meets" , XXX_INT, 
  0 , 0 , (ExecuteFunc)(&PlanInfo::execMeetsb), 1  , TypeId);

 MethArray[4] = new FuncMethodInfo( "overlaps" , XXX_INT, 
  0 , 0 , (ExecuteFunc)(&PlanInfo::execOverlapsp), 1  , PolygonId);

 MethArray[5] = new FuncMethodInfo( "overlaps" , XXX_INT, 
  0 , 0 , (ExecuteFunc)(&PlanInfo::execOverlapsb), 1  , TypeId);

 MethArray[6] = new FuncMethodInfo( "covered_by" , XXX_INT, 
  0 , 0 , (ExecuteFunc)(&PlanInfo::execCovered_byp), 1  , PolygonId);

 MethArray[7] = new FuncMethodInfo( "covered_by" , XXX_INT, 
  0 , 0 , (ExecuteFunc)(&PlanInfo::execCovered_byb), 1  , TypeId);

 MethArray[8] = new FuncMethodInfo( "inside" , XXX_INT, 
  0 , 0 , (ExecuteFunc)(&PlanInfo::execInsidep), 1  , PolygonId);

 MethArray[9] = new FuncMethodInfo( "inside" , XXX_INT, 
  0 , 0 , (ExecuteFunc)(&PlanInfo::execInsideb), 1  , TypeId);

 MethArray[17] = new FuncMethodInfo( "equals" , XXX_INT, 
  0 , 0 , (ExecuteFunc)(&PlanInfo::execEqualsp), 1  , PolygonId);

 MethArray[10] = new FuncMethodInfo( "equals" , XXX_INT, 
  0 , 0 , (ExecuteFunc)(&PlanInfo::execEqualsb), 1  , TypeId);

 MethArray[11] = new FuncMethodInfo( "covers" , XXX_INT, 
  0 , 0 , (ExecuteFunc)(&PlanInfo::execCoversp), 1  , PolygonId);

 MethArray[12] = new FuncMethodInfo( "covers" , XXX_INT, 
  0 , 0 , (ExecuteFunc)(&PlanInfo::execCoversb), 1  , TypeId);

 MethArray[13] = new FuncMethodInfo( "covers" , XXX_INT, 
  0 , 0 , (ExecuteFunc)(&PlanInfo::execCoverspt), 1  , PointId);

 MethArray[14] = new FuncMethodInfo( "contains" , XXX_INT, 
  0 , 0 , (ExecuteFunc)(&PlanInfo::execContainsp), 1  , PolygonId);

 MethArray[15] = new FuncMethodInfo( "contains" , XXX_INT, 
  0 , 0 , (ExecuteFunc)(&PlanInfo::execContainsb), 1  , TypeId);

 MethArray[16] = new FuncMethodInfo( "contains" , XXX_INT, 
  0 , 0 , (ExecuteFunc)(&PlanInfo::execContainspt), 1  , PointId);

 }


XxxErrCode XxxBoxADT::ReadText(istream& InStream, 
			       XxxADTValue* ObjectRef, 
			       const XxxADTMetaInfo* MetaInfo) const 
{ 
 BoxStruct *BS = (BoxStruct *) ObjectRef;
 ASSERT((BS));
 BS->Init();
 XXX_BOX* box = &(BS->Box);
 
 InStream >> box->xl >> box->yl >> box->xh >> box->yh;
 if (InStream.bad()) {
   XXX_NOTE_ERROR(XXX_BADADT, "Error reading box from stream");
   return XXX_NOT_OK;
  }
 if ((box->xl > box->xh) ||
     (box->yl > box->yh)) {
   XXX_NOTE_ERROR(XXX_BADADT, "Invalid box read");
   return XXX_NOT_OK;
  }
 return XXX_OK; 
}


XxxErrCode XxxBoxADT::WriteText(ostream& OutStream, 
				const XxxADTValue* Object, 
				const XxxADTMetaInfo* MetaInfo) const 
{ 
  BoxStruct *BS = (BoxStruct *) Object;
  ASSERT((BS));
  XXX_BOX* box = &(BS->Box);
  
  OutStream << "[(" << box->xl << ", " << box->yl << "), ";
  OutStream << "(" << box->xh << ", " << box->yh << ")]";
  return XXX_OK; 
}

XxxBool XxxBoxADT::Equals(const XxxADTValue* Obj1, 
			  const XxxADTMetaInfo*,
			  const XxxADTValue* Obj2,
			  const XxxADTMetaInfo*) const
{
  XXX_BOX *b1 = &(((BoxStruct *) Obj1)->Box);
  XXX_BOX *b2 = &(((BoxStruct *) Obj2)->Box);

  if (box_equals(b1, b2)) return XXX_TRUE;
  else return XXX_FALSE;
}
  
 
XxxBool
XxxBoxADT::GetBoundingBox(const XxxADTValue*& Object,
			  XxxBoundingBox& Box) 
{
  XXX_BOX *box = &(((BoxStruct *)Object)->Box);
  int dim = 2;
  double bb_box[4];

  bb_box[0] = box->xl;
  bb_box[1] = box->yl;
  bb_box[2] = box->xh;
  bb_box[3] = box->yh;

  XxxBoundingBox tmp(dim, XXX_DOUBLE, (void *) bb_box);
  Box=tmp;
  return XXX_TRUE;
}

//AUTHOR: Fabian Camargo DATE: 3/5/97
//PURPOSE: Retrieve the compare method given a function name
//and a array of arguments for that function.
//NOTES: This is used for index scans and is index independent.

XxxErrCode
XxxBoxADT::GetCompOp(const char* FuncName,
		     XxxCompareOp& CompOp, XxxBool& Success) const
{
 Success = XXX_TRUE;

 // get compare method (e.g overlaps) from the function name

 if (strcmp(FuncName,"overlaps")==0) {
   CompOp = XXX_IDX_OVERLAPS;
  }
 else if (strcmp(FuncName,"inside")==0) {
   // inside goes to t_cover
   CompOp = XXX_IDX_INSIDE;
  }
 else if (strcmp(FuncName,"covers")==0) {
   // covers goes to t_inside
   CompOp = XXX_IDX_COVERS;
  }
 else if (strcmp(FuncName,"contains")==0) {
   CompOp = XXX_IDX_CONTAINS;
  }
 else if (strcmp(FuncName,"covered_by")==0) {
   CompOp = XXX_IDX_COVERED_BY;
  }
 else if (strcmp(FuncName,"disjoint")==0) {
   CompOp = XXX_IDX_DISJOINT;
  }
 else if (strcmp(FuncName,"meets")==0) {
   CompOp = XXX_IDX_MEETS;
  }
 else if (strcmp(FuncName,"equals")==0) {
   CompOp = XXX_IDX_EQUALS;
  }
 else {
   // Don't know how to get range from this function
   CompOp = XXX_IDX_INVALID;
   Success = XXX_FALSE;
  }
 return XXX_OK;
}

XxxErrCode XxxBoxADT::CopyHToN(const XxxADTValue* Object, 
			       XxxADTValue *ResultObject,
			       const XxxADTMetaInfo* MetaInfo) const
{
  XXX_BOX *box = &(((BoxStruct *)Object)->Box);
  XXX_BOX *Rbox = &(((BoxStruct *)ResultObject)->Box);
  ASSERT((box && Rbox));
  
  XxxNetworkValue(Rbox, *box);
  return XXX_OK;
}

XxxErrCode XxxBoxADT::CopyNToH(const XxxADTValue* Object, 
			       XxxADTValue *ResultObject,
			       const XxxADTMetaInfo* MetaInfo) const
{
  XXX_BOX *box = &(((BoxStruct *)Object)->Box);
  XXX_BOX *Rbox = &(((BoxStruct *)ResultObject)->Box);
  ASSERT((box && Rbox));
  
  XxxHostValue(Rbox, *box);
  return XXX_OK;
}

