#ifndef XXX_BOX_H
#define XXX_BOX_H
#include "genadt.h"
#include "optrules.h"
#include "geometry.h"
#include "adtfunc.h"

XXX_EXTERN_CLASS_DECL(XxxBoundingBox);

/**
 * XxxBoxADT:  
 * Basic ADT for handling any generic rectangle data.  Boxes are
 * stored as fixed size objects within a record and simply
 * contain the lower left and upper right points within a Cartesian
 * coordinate system.  Currently-supported methods are the same
 * as those supported by polygons: disjoint(), meets(), covers(),
 * contains(), covered_by(), inside(), overlaps(), and equals().
 * These methods were chosen via the 9-point intersection model
 * of spatial objects.
 */
class XxxBoxADT : public XxxSmallFixedADT {
  friend class XxxGISModule;
 public : 
  /// Box type ID
  static XxxBasicType BoxTypeId;
 protected : 

 class BoxStruct : public XxxADTValue { public:
   XXX_BOX Box;
 };
  /**
   * PlanInfo:
   * Plan structure for methods on boxes.
   */
  class PlanInfo : public XxxFuncPlanInfo {
    friend class XxxBoxADT;
    XxxErrCode execDisjointp(XxxValueEnv* Env, 
                             XxxADTValue *ReturnValue) const; 
    XxxErrCode execDisjointb(XxxValueEnv* Env, 
                             XxxADTValue *ReturnValue) const; 
    XxxErrCode execMeetsp(XxxValueEnv* Env, 
                             XxxADTValue *ReturnValue) const; 
    XxxErrCode execMeetsb(XxxValueEnv* Env, 
                             XxxADTValue *ReturnValue) const; 
    XxxErrCode execOverlapsp(XxxValueEnv* Env, 
                             XxxADTValue *ReturnValue) const; 
    XxxErrCode execOverlapsb(XxxValueEnv* Env, 
                             XxxADTValue *ReturnValue) const; 
    XxxErrCode execCovered_byp(XxxValueEnv* Env, 
                             XxxADTValue *ReturnValue) const; 
    XxxErrCode execCovered_byb(XxxValueEnv* Env, 
                             XxxADTValue *ReturnValue) const; 
    XxxErrCode execInsidep(XxxValueEnv* Env, 
                             XxxADTValue *ReturnValue) const; 
    XxxErrCode execInsideb(XxxValueEnv* Env, 
                             XxxADTValue *ReturnValue) const; 
    XxxErrCode execEqualsp(XxxValueEnv* Env, 
                             XxxADTValue *ReturnValue) const;
    XxxErrCode execEqualsb(XxxValueEnv* Env, 
                             XxxADTValue *ReturnValue) const; 
    XxxErrCode execCoversp(XxxValueEnv* Env, 
                             XxxADTValue *ReturnValue) const; 
    XxxErrCode execCoversb(XxxValueEnv* Env, 
                             XxxADTValue *ReturnValue) const; 
    XxxErrCode execCoverspt(XxxValueEnv* Env, 
                             XxxADTValue *ReturnValue) const; 
    XxxErrCode execContainsp(XxxValueEnv* Env, 
                             XxxADTValue *ReturnValue) const; 
    XxxErrCode execContainsb(XxxValueEnv* Env, 
                             XxxADTValue *ReturnValue) const; 
    XxxErrCode execContainspt(XxxValueEnv* Env, 
			      XxxADTValue *ReturnValue) const; 
   public :
    PlanInfo(const XxxFuncParseInfo* FuncParse, int ArgCount)
     : XxxFuncPlanInfo(FuncParse , ArgCount) {}
   };
  friend class PlanInfo;

  virtual XxxErrCode methOptimize(XxxFuncParseInfo* FuncParse,
      XxxValueExprPlan* OwnerPlan,
      XxxValExprPlanList* ArgPlans,
      XxxFuncPlanInfo*& FPlan) const{
    FPlan = new PlanInfo(FuncParse, (ArgPlans?ArgPlans->Count():0));
    return XXX_OK;
  }
 public:
  /// Constructor for XxxBoxADT.
  XxxBoxADT(XxxBasicType Id); 
  /// Destructor for XxxBoxADT.
  virtual ~XxxBoxADT(){} 
  /**
   * Display function for the txtclient protocol.
   *
   * @param OutStream Output stream to write to.
   * @param Object Buffer containing the BOX.
   * @param MetaInfo Meta info for box.
   * @return The status code.
   */
  virtual XxxErrCode WriteText(ostream& OutStream, 
			       const XxxADTValue* Object,
			       const XxxADTMetaInfo* MetaInfo) const;
  /**
   * Input function for loading boxes into tables.
   *
   * @param is Input stream from which to read the BOX.
   * @param ObjectRef Buffer in which to store the BOX.
   * @param MetaInfo Meta info for box.
   * @return The status code.
   */
  virtual XxxErrCode ReadText(istream& InStream, 
			      XxxADTValue* ObjectRef,
			      const XxxADTMetaInfo* MetaInfo) const;
  /// Equals method between boxes.
  virtual XxxBool  Equals(const XxxADTValue* Obj1, 
			  const XxxADTMetaInfo*,
			  const XxxADTValue* Obj2,
			  const XxxADTMetaInfo*) const;

  /**
   * Retrieves the bounding box of this spatial data type.  Used by
   * the indexing methods.
   */
  virtual XxxBool GetBoundingBox(const XxxADTValue *&Object,
				 XxxBoundingBox& Box);

  /**
   * Retrieves the compare method given a function name
   * and a array of arguments for that function.
   * This is used in index scans and is index independent.
   */
  virtual XxxErrCode GetCompOp(const char* FName, XxxCompareOp& CompOp, 
			       XxxBool& Success) const;

  virtual XxxErrCode CopyHToN(const XxxADTValue* SrcObject, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const;
  virtual XxxErrCode CopyNToH(const XxxADTValue* SrcObject, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const;
};

#endif    /*  XXX_Box_H */ 
