#include "bytearray.h"
#include "optrules.h"
#include "memmap.h"
#include "GetFile.h"


// Define static variables in XxxByteArrayADT 
XxxBasicType XxxByteArrayADT::ByteArrayTypeId = XXX_UNKNOWN; 
const char* const XxxByteArrayADT::BYTEARRAYFILENAME = "__BYTEARRAY_FILE";

// Constructor for ByteArrayADT
XxxByteArrayADT::XxxByteArrayADT(XxxBasicType TypeId) 
 : XxxLargeFixedADT (TypeId, "bytearray",        
                     sizeof ( ByteArrayStruct))
{
 ByteArrayTypeId = TypeId;                         
} 


XxxErrCode XxxByteArrayADT::makeMetaInfo(const MetaInfoHandle& MIHandle,
					 XxxADTMetaInfo*& MI) const 
{ 
  ByteArrayMetaInfo* HandleInfo = (ByteArrayMetaInfo* ) (MIHandle.Data); 
  MI = new ByteArrayMetaInfo(*HandleInfo); 
  return XXX_OK; 
 } 


XxxErrCode XxxByteArrayADT::parseMetaInfo(const char* MInfoStr,
			           MetaInfoHandle& MIHandle) const 
{ 

  MIHandle.Size = sizeof(ByteArrayMetaInfo); 
  ASSERT(MIHandle.Size <= XXX_MIHNDL_MAX); 
  ByteArrayMetaInfo* BMI = (ByteArrayMetaInfo*)(MIHandle.Data);
  BMI->Init();

  // Using the format of MInfoStr, get the metainfo
  // and save it in MIHandle
  istrstream InStr((char *)MInfoStr);
  
  InStr >> BMI->Size;
  if(InStr.bad()){
      XXX_NOTE_ERROR(XXX_BADADT, "Bad ByteArray meta Info");
      return XXX_NOT_OK;
  }
  return XXX_OK;
} 


XxxErrCode XxxByteArrayADT::
showDataFile(const char* FileName, 
	     ostream& OutStream,   
	     const XxxADTValue* Object,   
	     const XxxADTMetaInfo* MetaInfo) const 
{ 
 ByteArrayMetaInfo* MI = (ByteArrayMetaInfo *)MetaInfo; 

 // If your ADT support many different formats, you'll;
 // add codes for them. 
 char command[3 + XXX_DIRNAME_MAX]; 
 sprintf(command, "vi %s" , FileName); 
 if (system(command) ){ 
      XXX_NOTE_ERROR(XXX_BADADT, "Error playing ByteArrayADT "); 
      return XXX_NOT_OK; 
 }
 return XXX_OK;
}


XxxErrCode XxxByteArrayADT::ReadText(istream& InStream, 
				     XxxADTValue* ObjectRef, 
				     const XxxADTMetaInfo* MetaInfo) const 
{ 
 XXX_SM_CHECK ;
 XxxErrCode rc = XXX_OK; 

 char ByteArrayFileName[XXX_DIRNAME_MAX];

 InStream >> ByteArrayFileName ;

 if( InStream.bad()) {
   XXX_NOTE_ERROR(XXX_BADADT, "Cannot read ByteArray file"); 
   return XXX_NOT_OK ;
  }
 
 int URL= 0; 
 // determine file location 
 if(strncmp(ByteArrayFileName,"http://",7)==0){
    // load ByteArray file from URL
    URL = 1;
    char *TmpName = tempnam(XXX_THISENV().TmpDataDir, "xxx");
    if (XxxGetFile(ByteArrayFileName,TmpName) != XXX_OK) {
       XXX_NOTE_ERROR(XXX_BADADT, "Error calling XxxGetFile");
       return XXX_NOT_OK;
    }
    strcpy(ByteArrayFileName, TmpName); 
 }

 XxxShoreObjectId ByteArrayOID; 
 // insert a "record" of this object
 XxxLargeADTEnv 
   *ByteArrayEnv = (XxxLargeADTEnv*) XXX_THISADTENV(ByteArrayTypeId);
 // note: the constructor mmaps the file, while
 // the destructor unmaps the file
 XxxMemoryMap ArrayMap(ByteArrayFileName);
 void *MMapAddress = NULL;
 if (ArrayMap.GetHandle(MMapAddress) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_BADFILE, "Error in mmap");
   return XXX_NOT_OK;
 }

 if (XxxShoreSM ->InsertRecord(ByteArrayEnv->StorageVolume, 
			       ByteArrayEnv->StorageFile,
                               MMapAddress, ArrayMap.Size(), 
                               ByteArrayOID) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADFILE, "Unable to store bytearray in SHORE");
    return XXX_NOT_OK; 
  }

 ByteArrayStruct* ByteArrayS = (ByteArrayStruct *)ObjectRef; 
 ByteArrayS->ByteArrayId.VolId = ByteArrayEnv->StorageVolume;
 ByteArrayS->ByteArrayId.LocalId = ByteArrayOID; 
  
 // Add by user, to initialize the structure of {Name}Struct


 ByteArrayMetaInfo *MI = (ByteArrayMetaInfo*) (MetaInfo); 

 if(URL == 1){
    // Delete temporary File
    char command[3 + XXX_DIRNAME_MAX]; 
    sprintf(command, "rm %s", ByteArrayFileName);
    if(system(command)) {
       XXX_NOTE_ERROR(XXX_BADADT, "Error calling XxxGetFile");
       return XXX_NOT_OK; 
    } 
 } 

 return XXX_OK; 
}


XxxErrCode XxxByteArrayADT::ExecOIDRequest(ostream &OStr, char *RequestBuf) 
{ 

  // Write the ByteArray out to a file 
  ByteArrayStruct Buf; 
  memcpy(&Buf, RequestBuf, sizeof(ByteArrayStruct)); 

  char *TmpName = tempnam("tmpdata", "xxx"); 

  if(XxxShoreSM->WriteToOSFile(Buf.ByteArrayId.VolId, 
                               Buf.ByteArrayId.LocalId, 
                               TmpName) != XXX_OK) { 
     XXX_NOTE_ERROR(XXX_BADADT, "Error writing ByteArray to file"); 
     return XXX_NOT_OK; 
  } 

  // send back the tmp file name, bracketed by <> 
  OStr << "<" << TmpName << ">" << endl; 
  OStr.flush(); 
  return XXX_OK; 
}

XxxErrCode XxxByteArrayADT::
SerializeCopySize(const XxxADTMetaInfo* MetaInfo,
		  unsigned int& Size) const
{
  ByteArrayMetaInfo* BAMInfo = (ByteArrayMetaInfo*)MetaInfo;
  if (!BAMInfo) {
    XXX_NOTE_ERROR(XXX_BADADT, "No MInfo passed to byte array serialize");
    return XXX_NOT_OK;
  }
  // write the array size, followed by the actual array
  Size = BAMInfo->Size + sizeof(int); 
  return XXX_OK;
}

XxxErrCode XxxByteArrayADT::
SerializeCopy(const XxxADTValue* Object,
	      const XxxADTMetaInfo* MetaInfo,
	      char* MemLocation) const
{
  ByteArrayMetaInfo* BAMInfo = (ByteArrayMetaInfo*)MetaInfo;
  if (!BAMInfo) {
    XXX_NOTE_ERROR(XXX_BADADT, "No MInfo passed to byte array serialize");
    return XXX_NOT_OK;
  }
  int CopySize = BAMInfo->Size;
  // copy the size of the array
  memcpy(MemLocation, &CopySize, sizeof(int));
 // copy the actual array to the memory location
  ByteArrayStruct* ByteArrayS = (ByteArrayStruct *)Object; 
  if (XxxShoreSM->GetObjectPreAlloc(ByteArrayS->ByteArrayId.VolId,
				    ByteArrayS->ByteArrayId.LocalId, 
				    MemLocation + sizeof(int),
				    CopySize, 0, CopySize)) {
    XXX_NOTE_ERROR(XXX_BADADT, "Error reading bytearray from Shore");
    return XXX_NOT_OK;
  } 
  return XXX_OK;
}

XxxErrCode XxxByteArrayADT::
UnSerializeCopy(const char* MemLocation,
		const XxxADTMetaInfo* MetaInfo,
		XxxADTValue* Object) const
{
  XXX_NOTE_ERROR(XXX_BADADT, "Unserialize Copy not supported for bytearray");
  return XXX_NOT_OK;
}

XxxErrCode XxxByteArrayADT::CopyHToN(const XxxADTValue* Object, 
				     XxxADTValue* ResultObject,
				     const XxxADTMetaInfo* MetaInfo) const
{
  TypeCopy(Object, ResultObject, MetaInfo);
  return XXX_OK;
}

XxxErrCode XxxByteArrayADT::CopyNToH(const XxxADTValue* Object, 
				     XxxADTValue* ResultObject,
				     const XxxADTMetaInfo* MetaInfo) const
{
  TypeCopy(Object, ResultObject, MetaInfo);
  return XXX_OK;
}

 
