#ifndef XXX_BYTE_H
#define XXX_BYTE_H
#include "genadt.h"


/**
 * XxxByteArrayADT
 */
class XxxByteArrayADT : public XxxLargeFixedADT {
  public : 
  //// Type ID accessible without object
  static XxxBasicType ByteArrayTypeId;
  
  protected : 
  static const char* const BYTEARRAYFILENAME;
  
  ////  Structure used to represent ByteArrayADT within a record
  class ByteArrayStruct : public XxxADTValue {
    public : 
    XxxShoreLogicalOID ByteArrayId;     
   };
  
  //// meta Info for ByteArray ADT  
  class ByteArrayMetaInfo : public XxxADTMetaInfo {
    public : 
    long Size;
    void Init() {
      Size = 0;
     }
    ByteArrayMetaInfo (){Init() ;}
    virtual ~ByteArrayMetaInfo() {}
    
    virtual XxxErrCode Replicate(XxxADTMetaInfo*& NewMI) const{
      ByteArrayMetaInfo* NewMetaInfo = new ByteArrayMetaInfo(*this);
      NewMI = NewMetaInfo;
      return XXX_OK;
     }
    virtual void Print(ostream& OStr) const {
      OStr << "(\"" << Size << ")\"";
     }
    virtual XxxBool IsCompatible(const XxxADTMetaInfo* OtherMI) const {
      ByteArrayMetaInfo * ThatMI = (ByteArrayMetaInfo *)OtherMI;
      return ((Size == ThatMI->Size) ? XXX_TRUE : XXX_FALSE);
     }
    /**
     * Returns the number of bytes necessary to hold the serialized version
     * of this meta-info object.
     * @param Size (out) The required number of bytes
     * @return XXX_NOT_OK if the object cannot be serialized, XXX_OK otherwise
     */
    virtual XxxErrCode SerializeCopySize(int &Size) const {
      Size = sizeof(this->Size); return XXX_OK;
    }
    /**
     * Creates a binary version of this meta-info object.
     * @param Data An array of bytes of at least the size specified by the
     *             SerializeCopySize() function.
     * @return XXX_NOT_OK if the object cannot be serialized, XXX_OK otherwise
     */
    virtual XxxErrCode SerializeCopy(char *Data) const {
      memcpy(Data, (char *)(&Size), sizeof(this->Size));  return XXX_OK;
    }
  };
  
  virtual XxxErrCode makeMetaInfo(const MetaInfoHandle& MIHandle,
				  XxxADTMetaInfo*& MI) const;
  virtual XxxErrCode parseMetaInfo(const char* MetaInfoString,
				   MetaInfoHandle& MIHandle) const;
  
  // called from WriteText()
  virtual XxxErrCode showDataFile(const char* FileName,
				  ostream& OutStream, 
				  const XxxADTValue* Object,
				  const XxxADTMetaInfo* MetaInfo) const;
  
  virtual XxxErrCode getOid(const XxxADTValue*Object, 
			    XxxShoreVolumeId& VolId, 
			    XxxShoreObjectId& LocalId) const {
    XxxShoreLogicalOID* ByteArrayId = &(((ByteArrayStruct*)Object)->ByteArrayId);
    VolId = ByteArrayId->VolId;
    LocalId = ByteArrayId->LocalId;
    return XXX_OK;
   }

 public:
  XxxByteArrayADT(XxxBasicType Id);
  
  virtual XxxErrCode ReadText(istream& InStream, 
			      XxxADTValue* ObjectRef,
			      const XxxADTMetaInfo* MetaInfo) const;
  
  // This handler needed for OID requests! 
  virtual XxxErrCode ExecOIDRequest(ostream &OStr, char *RequestBuf);
  
  virtual XxxErrCode SerializeCopySize(const XxxADTMetaInfo* MetaInfo,
				       unsigned int& Size) const;
  virtual XxxErrCode SerializeCopy(const XxxADTValue* Object,
				   const XxxADTMetaInfo* MetaInfo,
				   char* MemLocation) const;
  virtual XxxErrCode UnSerializeCopy(const char* MemLocation,
				     const XxxADTMetaInfo* MetaInfo,
				     XxxADTValue* Object) const;
  virtual XxxErrCode CopyHToN(const XxxADTValue* SrcObject, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const;
  virtual XxxErrCode CopyNToH(const XxxADTValue* SrcObject, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const;
 };


#endif    /*  XXX_ByteArray_H */ 
