#ifndef XXX_CIRCLE_H
#define XXX_CIRCLE_H
#include "genadt.h"
#include "optrules.h"
#include "geometry.h"
#include "adtfunc.h"

/**
 * XxxCircleADT:  
 * Basic ADT for handling any generic circle data.  Circles are
 * stored as fixed size objects within a record and simply
 * contain the center point (x and y coordinate) and the radius.
 * The circle ADT does not support any methods yet.
 */
class XxxCircleADT : public XxxSmallFixedADT {
  friend class XxxGISModule;
 public : 
  /// Circle type ID.
  static XxxBasicType CircleTypeId;
 protected : 

 class CircleStruct : public XxxADTValue { public:
   XXX_CIRCLE Circle;
 };

  /// Note: There are no valid circle methods 
  virtual XxxErrCode methOptimize(XxxFuncParseInfo* FuncParse,
      XxxValueExprPlan* OwnerPlan,
      XxxValExprPlanList* ArgPlans,
      XxxFuncPlanInfo*& FPlan) const{
   return XXX_NOT_OK;
  }
  public:
  /// Destructor for XxxCircleADT.
  XxxCircleADT(XxxBasicType Id); 
  ///  Destructor for XxxCircleADT.
  virtual ~XxxCircleADT(){} 
  /**
   * Display function for the txtclient protocol.
   *
   * @param OutStream Output stream to write to.
   * @param Object Buffer containing the PolygonStruct.
   * @param MetaInfo Meta info for polygons.
   * @return The status code.
   */
  virtual XxxErrCode WriteText(ostream& OutStream, 
			       const XxxADTValue* Object,
			       const XxxADTMetaInfo* MetaInfo) const;
  /**
   * Input function for loading boxes into tables.
   *
   * @param is Input stream from which to read the PolygonStruct.
   * @param ObjectRef Buffer in which to store the PolygonStruct.
   * @param MetaInfo Meta info for polygons.
   * @return The status code.
   */
  virtual XxxErrCode ReadText(istream& InStream, 
			      XxxADTValue* ObjectRef,
			      const XxxADTMetaInfo* MetaInfo) const;
  /// Equals method between circles.
  virtual XxxBool  Equals(const XxxADTValue* Obj1,
			  const XxxADTMetaInfo*,
			  const XxxADTValue* Obj2,
			  const XxxADTMetaInfo*) const;

  /**
   * Retrieves the bounding box of this spatial data type.  Used by
   * the indexing methods.
   */
  virtual XxxBool GetBoundingBox(const XxxADTValue *&Object,
				 XxxBoundingBox& Box);

  virtual XxxErrCode CopyHToN(const XxxADTValue* SrcObject, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const;
  virtual XxxErrCode CopyNToH(const XxxADTValue* SrcObject, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const;
};

#endif    /*  XXX_Circle_H */ 
