#include <stdio.h>
#include <string.h>
#include "complex.h"
#include "funcexpr.h"
#include "aggrexpr.h"

//
// Complex Number ADT
//  

XxxBasicType XxxComplexADT::ComplexTypeId;

// OWNER: as  DATE: 2/20/97
// PURPOSE: constructor
// NOTES: RuleEngine is initialized 
XxxComplexADT::XxxComplexADT(XxxBasicType Id)
 : XxxSmallFixedADT(Id, "complex", sizeof(ComplexStruct))
{
  ComplexTypeId = Id;
  // Register methods for Complex
  NumMethods = 5;
  NumMethodSlots = NumMethods + XxxADTClass::NUM_SPARE_METHOD_SLOTS;
  MethArray = new FuncMethodInfo* [NumMethodSlots];

  //Bjarke, sep. 14 2001: added '&' to all calls to FuncMethodInfo
  MethArray[0] = new FuncMethodInfo("real",XXX_DOUBLE, 0,0,
				    (ExecuteFunc)(&PlanInfo::execReal),0);
  MethArray[1] = new FuncMethodInfo("imag",XXX_DOUBLE, 0,0,
				    (ExecuteFunc)(&PlanInfo::execImag),0);
  MethArray[2] = new FuncMethodInfo("plus",Id, 0,0,
				    (ExecuteFunc)(&PlanInfo::execPlus),1,Id);
  MethArray[3] = new FuncMethodInfo("minus",Id, 0,0,
				    (ExecuteFunc)(&PlanInfo::execMinus),1,Id);
  MethArray[4] = new FuncMethodInfo("invert",Id, 0,0,
				    (ExecuteFunc)(&PlanInfo::execInvert),0);

  // Register methods for aggregate function
  NumAggrMethods = 1;
  XxxFuncMethodInfo::ArgInfo EmptySignature(0);
  AggrMethArray = new AggrMethodInfo* [NumAggrMethods];
  //Bjarke, sep. 14 2001: added '&' to call to AggrMethodInfo
  AggrMethArray[0] = new AggrMethodInfo("csum", Id ,EmptySignature,
					0, 0,
					(AggrInitFunc)(&PlanInfo::csumInit), 
					(AggrAccumFunc)(&PlanInfo::csumAccum), 
					(AggrTermFunc)(&PlanInfo::csumTerm), 0);

  RuleEngine=new XxxOptRuleEngine;
  RuleEngine->AddRule(new XxxRule(csumInvertReorderCond,
				  csumInvertReorderAction));  
  //RuleEngine->AddRule(new XxxRule(plusCommuteCond, plusCommuteAct));
  //RuleEngine->AddRule(new XxxRule(plusMinusCond, plusMinusAct));
}

XxxComplexADT::~XxxComplexADT()
{
}

XxxErrCode XxxComplexADT::methOptimize(XxxFuncParseInfo* FuncParse,
				       XxxValueExprPlan* OwnerPlan,
				       XxxValExprPlanList* ArgPlans,
				       XxxFuncPlanInfo*& FPlan) const
{
  if(!MethArray[FuncParse->MethodIndex]->OptFunc) {
    FPlan = new PlanInfo(FuncParse, (ArgPlans?ArgPlans->Count():0));
    return XXX_OK;
  }
  else 
    return XxxADTClass::methOptimize(FuncParse, OwnerPlan,
				     ArgPlans, FPlan);
}


//-----------------------------------------------------------------------------
XxxErrCode XxxComplexADT::WriteText(ostream& OutStream, 
				    const XxxADTValue* Object,
				    const XxxADTMetaInfo* MetaInfo) const
{
 ComplexStruct *CVal = (ComplexStruct *)Object;
 OutStream << CVal->Real;
 if (CVal->Imag < 0.0)
   OutStream << "-" << -(CVal->Imag) << "i";
 else OutStream << "+" << CVal->Imag << "i";
 return XXX_OK;
}
//-----------------------------------------------------------------------------
XxxErrCode XxxComplexADT::ReadText(istream& InStream, 
				   XxxADTValue* ObjectRef,
				   const XxxADTMetaInfo* MetaInfo) const
{
  char c, i;
  ComplexStruct *CVal = (ComplexStruct *)ObjectRef;
  ASSERT(CVal);
  CVal->Init();
  /*
   * The proper syntax is "<real part>+<imaginary part>i". 
   * No whitespace is permitted.
   */
  // Read the first double
  double ValReal, ValImag;
  InStream >> CVal->Real >> c >> CVal->Imag >> i;
  if (InStream.bad() || (c != '+') || (i != 'i')) {
    XXX_NOTE_ERROR(XXX_BADPARSE, 
		   "The proper syntax for a complex number is <real part>+<imaginary part>i");
    return XXX_NOT_OK;
  }
  return XXX_OK;
}
//-----------------------------------------------------------------------------
XxxBool  XxxComplexADT::Equals(const XxxADTValue* Obj1, 
			       const XxxADTMetaInfo* MInfo1,
			       const XxxADTValue* Obj2,
			       const XxxADTMetaInfo* MInfo2) const
{
 ComplexStruct *CVal1 = (ComplexStruct *)Obj1;
 ComplexStruct *CVal2 = (ComplexStruct *)Obj2;
 
 ASSERT((CVal1 && CVal2));
 if ((CVal1->Real == CVal2->Real) && (CVal1->Imag == CVal2->Imag))
   return XXX_TRUE;
 else
   return XXX_FALSE;
}

// PURPOSE: if we have a selection of the type .csum(complexval).invert(), 
// see if we can change it to csum(complex.invert())
XxxErrCode 
XxxComplexADT::csumInvertReorderCond(XxxValueExpression *VExpr, 
				     XxxBool &Success, void *&RuleData) 
{
#ifdef XXX_DEBUG
  cout << "Trying to apply csumInvertReorderCond...." << endl;
#endif

  Success = XXX_FALSE;
  if (VExpr->ValType() != XxxValueExpression::FUNC_EXPR) {
    return XXX_OK;
  }
  else {
    // it is indeed a function expression
    XxxFuncExpression *FuncExpr = (XxxFuncExpression *)VExpr;
    
    XxxFuncParseInfo *FuncParse = FuncExpr->FuncParse;
    // TODO: yikes, remove this constant!!
    if (FuncParse->MethodIndex != 4) {
      return XXX_OK;
    }

    // the function is indeed 'INVERT': see if csum is its owner
    XxxValueExpression* Owner = FuncExpr->Owner;
    if ((Owner->ValType() != XxxValueExpression::FUNC_EXPR) ||
	!((XxxFuncExpression*)Owner)->IsAggr())
       return XXX_OK;

    XxxAggrExpression *AggrExpr = (XxxAggrExpression *)Owner;
    XxxFuncParseInfo *AggrParse = AggrExpr->FuncParse;

    // TODO: yikes, remove this constant!!
    if (AggrParse->MethodIndex != 0)
      return XXX_OK;

    Success = XXX_TRUE;
#ifdef XXX_DEBUG
  cout << "Successful!" << endl;
#endif
  }
  return XXX_OK;
}
//-----------------------------------------------------------------------------
// PURPOSE: change a selection of the type csum(complex).invert() to
// csum(complex.invert())
XxxErrCode XxxComplexADT::
csumInvertReorderAction(XxxValueExpression *InVExpr, void *RuleData, 
			XxxValueExpression *&OutVExpr) 
{
#ifdef XXX_DEBUG
  cout << "Applying csumInvertReorderAction...." << endl;
#endif
  XxxFuncExpression *FuncExpr = (XxxFuncExpression *) InVExpr;
  XxxAggrExpression *AggrExpr = (XxxAggrExpression *) FuncExpr->Owner;

  XxxValueExpression *AggrOwner = AggrExpr->Owner;

  // make the argument of the aggregate the argument of the function instead
  FuncExpr->Owner = AggrOwner;
  // make the function expression the argument of the aggregate
  AggrExpr->Owner = FuncExpr;
  // return the aggregate expression
  OutVExpr = AggrExpr;
#ifdef XXX_DEBUG
  cout << "Successful!" << endl;
#endif
  return XXX_OK;
}
//-----------------------------------------------------------------------------
XxxErrCode 
XxxComplexADT::plusCommuteCond(XxxValueExpression* VExpr,
			       XxxBool& Success,
			       void*& RuleData)
{
  if (VExpr->ValType()!=XxxValueExpression::FUNC_EXPR) {
     Success = XXX_FALSE;
     return XXX_OK;
  }
  XxxFuncExpression *FExpr = (XxxFuncExpression*) VExpr;

 Success = XXX_FALSE;
 XxxFuncParseInfo *ThisFuncParse = (FExpr->FuncParse);

 if (FExpr->ArgList &&
     (FExpr->ArgList->Count() == 1) && 
     (ThisFuncParse->MethodIndex == 2)) {
   Success = XXX_TRUE;  // condition fired
   RuleData = NULL;     // no extra rule data to pass
  }
 return XXX_OK;
}
//-----------------------------------------------------------------------------
XxxErrCode 
XxxComplexADT::plusCommuteAct(XxxValueExpression* InVExpr,
			      void* RuleData,
			      XxxValueExpression*& OutVExpr)
{
  // Since we know we can apply the rule, InVExpr must be a funcexpression
  XxxFuncExpression *InFExpr = (XxxFuncExpression*) InVExpr;

 XxxValueExprList* ArgList = InFExpr->ArgList;
 XxxValueExpression* Owner = InFExpr->Owner;

 OutVExpr = InFExpr;
 // see if we can commute the arguments
 InFExpr->Owner = ArgList->Expr;
 ArgList->Expr = Owner;

 cerr << "Commuted PLUS arguments" << endl;
 InFExpr->Print(cerr);

 return XXX_OK;
}
//-----------------------------------------------------------------------------
XxxErrCode 
XxxComplexADT::plusMinusCond(XxxValueExpression* VExpr,
			     XxxBool& Success,
			     void*& RuleData)
{
  if (VExpr->ValType()!=XxxValueExpression::FUNC_EXPR) {
    Success = XXX_FALSE;
    return XXX_OK;
  }
  XxxFuncExpression* FExpr=(XxxFuncExpression*) VExpr;

 Success = XXX_FALSE;
 XxxFuncParseInfo *ThisFuncParse = (FExpr->FuncParse);

 if (!(FExpr->ArgList &&
       (FExpr->ArgList->Count() == 1) && 
       (ThisFuncParse->MethodIndex == 2 ))) {
   return XXX_OK;
  }

 // now check that the owner is a MINUS operator
 XxxValueExpression* Owner = FExpr->Owner;
 if (!(Owner->ValType() == XxxValueExpression::FUNC_EXPR)) {
   return XXX_OK;
  }
 XxxFuncExpression* OwnerFuncExpr = (XxxFuncExpression*) Owner;
 if (OwnerFuncExpr->FuncParse &&
     (OwnerFuncExpr->FuncParse->TypeId == ThisFuncParse->TypeId)) {

   XxxFuncParseInfo* ChildParse = (OwnerFuncExpr->FuncParse);
   if (ChildParse->MethodIndex == 3) {
     Success = XXX_TRUE;  // condition fired
     RuleData = NULL;     // no extra rule data to pass
    }
  }
 return XXX_OK;
}
//-----------------------------------------------------------------------------
XxxErrCode 
XxxComplexADT::plusMinusAct(XxxValueExpression* InVExpr,
			    void* RuleData,
			    XxxValueExpression*& OutVExpr)
{
 // Since we know we can apply the rule, InVExpr must be a funcexpression
 XxxFuncExpression *InFExpr = (XxxFuncExpression*) InVExpr;

 XxxValueExprList* ArgList = InFExpr->ArgList;
 XxxValueExpression* Owner = InFExpr->Owner;

 OutVExpr = InFExpr;
 // see if we can commute the arguments
 InFExpr->Owner = ArgList->Expr;
 ArgList->Expr = Owner;

 cerr << "Commuted PLUS arguments" << endl;
 InFExpr->Print(cerr);

 return XXX_OK;
}
//-----------------------------------------------------------------------------
//OWNER: Fabian Camargo DATE: 1/29/97
//PURPOSE: Returns a bounding box, which can then be used for an rtree search
//NOTES: This is not really useful in itself, but it was used to test and
// debug the rtree code.

XxxBool
XxxComplexADT::GetBoundingBox(const XxxADTValue*& Object,
			      XxxBoundingBox& Box)
{
  double data[4];
  ComplexStruct *CVal = (ComplexStruct *)Object;

  if (CVal->Real < CVal->Imag)
    {
      data[0]=data[1]=CVal->Real;
      data[2]=data[3]=CVal->Imag;
    }
  else
    {
      data[0]=data[1]=CVal->Imag;
      data[2]=data[3]=CVal->Real;
    }
  XxxBoundingBox tmp(2,XXX_INT,data);
  Box = tmp; 
  return XXX_TRUE; 

}

XxxBool XxxComplexADT::CanCastTo(XxxBasicType TargetType, 
				 const XxxADTMetaInfo* SourceMetaInfo,
				 const XxxADTMetaInfo* TargetMetaInfo)
{
  if ((TargetType == XXX_INT) || (TargetType == XXX_DOUBLE))
    return XXX_TRUE;
  else
    return XXX_FALSE;
}

XxxErrCode XxxComplexADT::CastTo(XxxBasicType TargetType, 
				 const XxxADTMetaInfo* SourceMetaInfo,
				 const XxxADTMetaInfo* TargetMetaInfo,
				 XxxADTValue* OldObject, 
				 XxxADTValue* NewObject)
{
  NewObject->Init();
  if (TargetType == XXX_INT) {
    ((XxxCoreADTClass::IntValue *)NewObject)->Value = 
      (int)((ComplexStruct *)OldObject)->Real;
    return XXX_OK;
  } else if (TargetType == XXX_DOUBLE) {
    ((XxxCoreADTClass::DblValue *)NewObject)->Value = 
      ((ComplexStruct *)OldObject)->Real;
    return XXX_OK;
  } else {
    XXX_NOTE_ERROR(XXX_BADADT, "Cannot cast to a complex from supplied type");
    return XXX_NOT_OK;
  }
 
}

XxxBool XxxComplexADT::CanCastFrom(XxxBasicType SourceType, 
				   const XxxADTMetaInfo* SourceMetaInfo,
				   const XxxADTMetaInfo* TargetMetaInfo)
{
  if ((SourceType == XXX_INT) || (SourceType == XXX_DOUBLE))
    return XXX_TRUE;
  else
    return XXX_FALSE;
}

XxxErrCode XxxComplexADT::CastFrom(XxxBasicType SourceType, 
				   const XxxADTMetaInfo* SourceMetaInfo,
				   const XxxADTMetaInfo* TargetMetaInfo,
				   XxxADTValue* OldObject, 
				   XxxADTValue* NewObject)
{
  NewObject->Init();
  if (SourceType == XXX_INT) {
    ((ComplexStruct *)NewObject)->Real = 
      (double)((XxxCoreADTClass::IntValue *)OldObject)->Value;
    ((ComplexStruct *)NewObject)->Imag = 0.0;
    return XXX_OK;
  } else if (SourceType == XXX_DOUBLE) {
    ((ComplexStruct *)NewObject)->Real = 
      ((XxxCoreADTClass::DblValue *)OldObject)->Value;
    ((ComplexStruct *)NewObject)->Imag = 0.0;
    return XXX_OK;
  } else {
    XXX_NOTE_ERROR(XXX_BADADT, "Cannot cast to a complex from supplied type");
    return XXX_NOT_OK;
  }
}

XxxErrCode XxxComplexADT::CopyHToN(const XxxADTValue* Object, 
				   XxxADTValue *ResultObject,
				   const XxxADTMetaInfo* MetaInfo) const
{
  *ResultObject = *Object;
  ComplexStruct *CVal = (ComplexStruct *)Object;
  ComplexStruct *RCVal = (ComplexStruct *)ResultObject;
  XxxNetworkDouble(&(RCVal->Real), CVal->Real);
  XxxNetworkDouble(&(RCVal->Imag), CVal->Imag);
  return XXX_OK;
}

XxxErrCode XxxComplexADT::CopyNToH(const XxxADTValue* Object, 
				   XxxADTValue *ResultObject,
				   const XxxADTMetaInfo* MetaInfo) const
{
  *ResultObject = *Object;
  ComplexStruct *CVal = (ComplexStruct *)Object;
  ComplexStruct *RCVal = (ComplexStruct *)ResultObject;
  XxxHostDouble(&(RCVal->Real), CVal->Real);
  XxxHostDouble(&(RCVal->Imag), CVal->Imag);
  return XXX_OK;
}
