#ifndef XXX_COMPLEX_H
#define XXX_COMPLEX_H
#include "genadt.h"
#include "optrules.h"
#include "BoundingBox.h"
#include "adtfunc.h"

/**
 // this type represents complex numbers
 */
class XxxComplexADT : public XxxSmallFixedADT {
 public:
  static XxxBasicType ComplexTypeId;
 protected:
  //// 
  class ComplexStruct : public XxxADTValue { public:
    double Real;
    double Imag;
   };
  
  //// plan structure for methods on complex numbers
  class PlanInfo : public XxxFuncPlanInfo {
     friend class XxxComplexADT;
    protected :
    ///
    XxxErrCode execReal(XxxValueEnv* Env, XxxADTValue *ReturnValue) const;
    ///
    XxxErrCode execImag(XxxValueEnv* Env, XxxADTValue *ReturnValue) const;
    ///
    XxxErrCode execPlus(XxxValueEnv* Env, XxxADTValue *ReturnValue) const;
    ///
    XxxErrCode execMinus(XxxValueEnv* Env, XxxADTValue *ReturnValue) const;
    ///
    XxxErrCode execInvert(XxxValueEnv* Env, XxxADTValue *ReturnValue) const;
    /// Sum aggregate function
    XxxErrCode csumInit(XxxAggrOpHandle& Hndl) const;
    XxxErrCode csumAccum(XxxValueEnv* Env, XxxAggrOpHandle& Hndl) const;
    XxxErrCode csumTerm(XxxAggrOpHandle& Hndl, XxxADTValue* FinalVal) const;
    
    public :
     PlanInfo(const XxxFuncParseInfo* FuncParse, int argCount )
       :  XxxFuncPlanInfo(FuncParse, argCount){}
 };


  virtual XxxErrCode methOptimize(XxxFuncParseInfo* FuncParse,
				  XxxValueExprPlan* OwnerPlan,
				  XxxValExprPlanList* ArgPlans,
				  XxxFuncPlanInfo*& FPlan) const;
  /*
    todo rc moved and modified to complex.c {
    FPlan = new PlanInfo(FuncParse, (ArgPlans?ArgPlans->Count():0));
    return XXX_OK;
  }////////
  */

  // Simple rule to commute the arguments of a Plus operator.
  // Make sure that this rule doesnt get reapplied forever.
  // Currently disabled.
  static XxxErrCode plusCommuteCond(XxxValueExpression* VExpr,
				    XxxBool& Success,
				    void*& RuleData);
  static XxxErrCode plusCommuteAct(XxxValueExpression* InVExpr,
				   void* RuleData,
				   XxxValueExpression*& OutVExpr);

  // Rule to move Plus ahead of Minus
  // X.minus(Y).plus(Z) == X.plus(Z).minus(Y)
  static XxxErrCode plusMinusCond(XxxValueExpression* VExpr,
				  XxxBool& Success,
				  void*& RuleData);
  static XxxErrCode plusMinusAct(XxxValueExpression* InVExpr,
				 void* RuleData,
				 XxxValueExpression*& OutVExpr);
  static XxxErrCode csumInvertReorderCond(XxxValueExpression *VExpr, 
					  XxxBool &Success, 
					  void *&RuleData);
  // --> Anil Sachdeva 2/20/97
  static XxxErrCode csumInvertReorderAction(XxxValueExpression *InVExpr, 
					    void *RuleData, 
					    XxxValueExpression *&OutVExpr);
  // --> Anil Sachdeva 2/20/97
 public:
  XxxComplexADT(XxxBasicType Id);
  virtual ~XxxComplexADT();

  virtual XxxErrCode WriteText(ostream& OutStream, 
			       const XxxADTValue* Object,
			       const XxxADTMetaInfo* MetaInfo) const;
  virtual XxxErrCode ReadText(istream& InStream, 
			      XxxADTValue* ObjectRef,
			      const XxxADTMetaInfo* MetaInfo) const;
  virtual XxxBool  Equals(const XxxADTValue* Obj1, 
			  const XxxADTMetaInfo* MInfo1,
			  const XxxADTValue* Obj2,
			  const XxxADTMetaInfo* MInfo2) const;


  /**
   * Determines if this ADT can cast an object of another type into
   * its own type.
   * @param SourceType The type to cast from
   * @param SourceMetaInfo The meta info of the object to be casted
   * @param TargetMetaInfo The meta info of the new object to be created
   */
  virtual XxxBool CanCastFrom(XxxBasicType SourceType, 
			      const XxxADTMetaInfo* SourceMetaInfo,
			      const XxxADTMetaInfo* TargetMetaInfo);
  /**
   * Executes a cast from another type into this type.
   * @param CastType The type id of the old object
   * @param SourceMetaInfo The meta info of the object to be casted
   * @param TargetMetaInfo The meta info of the new object to be created
   * @param OldObject The object to be casted
   * @param NewObject The new object to be created
   */
  virtual XxxErrCode CastFrom(XxxBasicType SourceType, 
			      const XxxADTMetaInfo* SourceMetaInfo,
			      const XxxADTMetaInfo* TargetMetaInfo,
			      XxxADTValue* OldObject, 
			      XxxADTValue* NewObject);
  /**
   * Determines if this ADT can cast an object of its type into
   * another type.
   * @param TargetType The type to cast to
   * @param SourceMetaInfo The meta info of the object to be casted
   * @param TargetMetaInfo The meta info of the new object to be created
   */
  virtual XxxBool CanCastTo(XxxBasicType TargetType, 
			    const XxxADTMetaInfo* SourceMetaInfo,
			    const XxxADTMetaInfo* TargetMetaInfo);
  /**
   * Executes a cast from this type into another type.
   * @param CastType The type id of the new object
   * @param SourceMetaInfo The meta info of the object to be casted
   * @param TargetMetaInfo The meta info of the new object to be created
   * @param OldObject The object to be casted
   * @param NewObject The new object to be created
   */
  virtual XxxErrCode CastTo(XxxBasicType TargetType, 
			    const XxxADTMetaInfo* SourceMetaInfo,
			    const XxxADTMetaInfo* TargetMetaInfo,
			    XxxADTValue* OldObject, 
			    XxxADTValue* NewObject);

  // Function to return bounding box for indexing purposes  
  virtual XxxBool GetBoundingBox(const XxxADTValue*& Object,
				 XxxBoundingBox& Box);

  virtual XxxErrCode CopyHToN(const XxxADTValue* ObjectRef, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const;
  virtual XxxErrCode CopyNToH(const XxxADTValue* ObjectRef, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const;
};



#endif 
