#include "counter.h"
#include "types.h"
#include "shoresm.h"

XxxBasicType XxxCounterADT::CounterTypeId = XXX_UNKNOWN;

//
// Methods of XxxCounterADT::CounterMetaInfo
//

XxxErrCode XxxCounterADT::CounterMetaInfo::Replicate(XxxADTMetaInfo*& NewMI) const
{
  CounterMetaInfo *NewMetaInfo = new CounterMetaInfo();
  NewMetaInfo->nextValue = nextValue;
  NewMetaInfo->myId = myId;
  NewMI = NewMetaInfo;
  return XXX_OK;
}
  
void XxxCounterADT::CounterMetaInfo::Print(ostream& OStr) const 
{
  OStr << "[Next Value: " << nextValue << "]";
}

//
// Methods of XxxCounterADT
//
XxxCounterADT::XxxCounterADT(XxxBasicType Id) : 
  XxxSmallFixedADT(Id, "counter", 
	      sizeof(CounterStruct), XXX_TRUE) // Maintains catalog (meta) info
{ CounterTypeId = Id; }

XxxCounterADT::~XxxCounterADT() { }

void XxxCounterADT::Print(ostream& OStr, const XxxADTMetaInfo* MetaInfo) const 
{
  OStr << "counter";
  if (MetaInfo)
    ((XxxCounterADT::CounterMetaInfo *)MetaInfo)->Print(OStr);
}

XxxErrCode XxxCounterADT::WriteText(ostream& OutStream, 
				    const XxxADTValue* Object,
				    const XxxADTMetaInfo* MetaInfo) const
{
 int Val = ((CounterStruct *)Object)->value;
 OutStream << Val;
 return XXX_OK;
}

XxxErrCode XxxCounterADT::DefaultVal(XxxADTValue* ObjectRef, 
				     const XxxADTMetaInfo* MetaInfo) const
{ 
  CounterMetaInfo *CMI = (CounterMetaInfo *)MetaInfo;
  if (!CMI) {
    XXX_NOTE_ERROR(XXX_BADADT, "Invalid meta-info for counter default val");
    return XXX_NOT_OK;
   }
  ((CounterStruct *)ObjectRef)->value = 
    ((CounterMetaInfo *)MetaInfo)->nextValue;
  ((CounterMetaInfo *)MetaInfo)->nextValue++;
  // Update the persistent version
  if (XxxShoreSM->UpdateRecord(XxxShoreSM->DefaultVol(),
			       CMI->myId, 0, sizeof(CounterMetaInfo), 
			       (char *)MetaInfo) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot update persistent version of counter MI");
    return XXX_NOT_OK;
  }
  return XXX_OK; 
}

XxxErrCode XxxCounterADT::ReadText(istream& InStream, 
				   XxxADTValue* ObjectRef,
				   const XxxADTMetaInfo* MetaInfo) const
{
  char AssignedValue[100];
  AssignedValue[0] = 0;
  InStream >> AssignedValue;
	if (strlen(AssignedValue) == 0) {
     // Null string, therefore nothing to read; assign default value
	 if (DefaultVal(ObjectRef, MetaInfo) != XXX_OK) {
		XXX_NOTE_ERROR(XXX_BADADT, "Cannot assign default value to counter");
		return XXX_NOT_OK;
	}
   } else {
	   // This is the case when the user specifies a value
	   // to be entered, instead of incrementing the counter
	   CounterMetaInfo *CMI = (CounterMetaInfo *)MetaInfo;
	   if (!CMI) {
       XXX_NOTE_ERROR(XXX_BADADT, "Invalid meta-info for counter");
       return XXX_NOT_OK;
     }
     ((CounterStruct *)ObjectRef)->value = atoi(AssignedValue);
     if (XxxShoreSM->UpdateRecord(XxxShoreSM->DefaultVol(),
				  CMI->myId, 0, sizeof(CounterMetaInfo), 
				  (char *)MetaInfo) != XXX_OK) {
       XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot update persistent version of counter MI");
       return XXX_NOT_OK;
     }
   }
  return XXX_OK;
}

XxxErrCode XxxCounterADT::parseMetaInfo(const char* MetaInfoString,
					MetaInfoHandle& MIHandle) const 
{
 CounterMetaInfo CMI;
 XxxShoreObjectId RecordId;
 // Create the record (empty at first)
 if (XxxShoreSM->InsertRecord(XxxShoreSM->DefaultVol(),
			      XxxShoreSM->PersistentFileId(),
			      (char *)&CMI, sizeof(CounterMetaInfo), 
			      RecordId) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot store meta info in persistent file");
   return XXX_NOT_OK;
  }
 // The metainfo string can optionally specify the starting value
 if (MetaInfoString && strlen(MetaInfoString) > 0)
   CMI.nextValue = atoi(MetaInfoString);
 else
   CMI.nextValue = 0;
 // Now record the SHORE OID
 CMI.myId = RecordId;
 // Update the persistent version
 if (XxxShoreSM->UpdateRecord(XxxShoreSM->DefaultVol(),
			      CMI.myId, 0, sizeof(CounterMetaInfo), 
			      (char *)&CMI) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot update persistent version of counter MI");
   return XXX_NOT_OK;
  }
 assert(sizeof(RecordId) <= XXX_MIHNDL_MAX);
 MIHandle.Size = sizeof(XxxShoreObjectId);
 memcpy((char *)(&(MIHandle.Data)), (char *)&(CMI.myId),
	sizeof(XxxShoreObjectId));
 return XXX_OK;
}

XxxErrCode XxxCounterADT::makeMetaInfo(const MetaInfoHandle& MIHandle,
				      XxxADTMetaInfo* &MetaInfo) const 
{
 XxxShoreObjectId RecordId;
 memcpy((char *)&RecordId, (char *)(&(MIHandle.Data)), 
	sizeof(XxxShoreObjectId));
 char* TmpData = new char[sizeof(CounterMetaInfo)];
 if (XxxShoreSM->GetObjectPreAlloc(XxxShoreSM->DefaultVol(), RecordId,
				   TmpData, 
				   sizeof(CounterMetaInfo), 0,
				   sizeof(CounterMetaInfo)) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot get meta info from persistent file");
   return XXX_NOT_OK;
  }
 // TODO:L2 : all over the ADT code, there is dangerous code
 // that copies ADTMetaInfo objects from stored versions to 
 // in-memory versions -- this presents a major problem.
 // Instead, only the data component of the MetaInfo should be
 // copied back and forth.
 MetaInfo = new CounterMetaInfo(*((CounterMetaInfo*)TmpData));
 delete [] TmpData;
 return XXX_OK;
}

XxxBool XxxCounterADT::CanCastTo(XxxBasicType TargetType, 
				 const XxxADTMetaInfo* SourceMetaInfo,
				 const XxxADTMetaInfo* TargetMetaInfo) 
{
  return (TargetType == XXX_INT);
}

XxxErrCode XxxCounterADT::CastTo(XxxBasicType TargetType, 
				 const XxxADTMetaInfo* SourceMetaInfo,
				 const XxxADTMetaInfo* TargetMetaInfo,
				 XxxADTValue* OldObject, 
				 XxxADTValue* NewObject)
{
  if (TargetType != XXX_INT) {
    XXX_NOTE_ERROR(XXX_BADADT, "Cannot cast a counter to anything but an integer");
    return XXX_NOT_OK;
  }
  ((XxxCoreADTClass::IntValue *)NewObject)->Value = ((CounterStruct *)OldObject)->value;
  return XXX_OK;
}

XxxErrCode XxxCounterADT::CopyHToN(const XxxADTValue* Object, 
				   XxxADTValue *ResultObject,
				   const XxxADTMetaInfo* MetaInfo) const
{
 // copy the header fields
 *ResultObject = *Object;

 int Val = ((CounterStruct *)Object)->value;
 unsigned long LTemp = htonl(Val);
 ((CounterStruct*)ResultObject)->value = LTemp;
 return XXX_OK;
}

XxxErrCode XxxCounterADT::CopyNToH(const XxxADTValue* Object, 
				   XxxADTValue *ResultObject,
				   const XxxADTMetaInfo* MetaInfo) const
{
 // copy the header fields
 *ResultObject = *Object;

 int Val = ((CounterStruct *)Object)->value;
 unsigned long LTemp = ntohl(Val);
 ((CounterStruct*)ResultObject)->value = LTemp;
 return XXX_OK;
}
