#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include <iomanip.h>
#include <sys/types.h>
#ifdef __GNUG__
#pragma implementation
#endif
#include "prd-config.h"
#include "date.h"

int JulianMDYCalADT::MDYOrdDomId = -1;
const char JulianCalADT::NowString[] = "now";
const int JulianCalADT::SEQ_WEEK = 8;
const int JulianCalADT::SEQ_MONTH = 32;
const int JulianCalADT::SEQ_YEAR = 13*32;
const int JulianCalADT::SEQ_CENTURY = 13*32*100;
const int JulianCalADT::StartDay = 1;
const int JulianCalADT::StartWeek = 1;
const int JulianCalADT::StartMonth = 1;
const int JulianCalADT::StartYear = 0;

const int JulianCalADT::YearsPerCent = 100;
// assuming that it is not a century divisible by 4!
const int JulianCalADT::DaysPerCent  = 76*365+24*366;
  
// some constants useful for approximate calculations
const float JulianCalADT::DaysPerMonth = 30.5;
const float JulianCalADT::DaysPerYear  = 365.25;

// const values of interest for hour/minute/sec calcs
const int JulianCalADT::SEQ_SECOND = 60;
const int JulianCalADT::SEQ_MINUTE = 60;
const int JulianCalADT::SEQ_HOUR = 24;

const int JulianCalADT::StartSecond = 0;
const int JulianCalADT::EndSecond = 59;
const int JulianCalADT::StartMinute = 0;
const int JulianCalADT::EndMinute = 59;
const int JulianCalADT::StartHour = 0;
const int JulianCalADT::EndHour = 23;
const int JulianCalADT::MonthDays[13] = {0,31,28,31,30,31,30,31,31,30,31,30,31};
const int JulianCalADT::AccumMonthDays[13] = {0,31,59,90,120,151,181,212,243,273,304,334,365};

const int 
JulianCalADT::LeapYears[100] = { 0, 0, 0, 0, 1, 0, 0, 0, 1, 0,
                                 0, 0, 1, 0, 0, 0, 1, 0, 0, 0,
                                 1, 0, 0, 0, 1, 0, 0, 0, 1, 0,
                                 0, 0, 1, 0, 0, 0, 1, 0, 0, 0,
                                 1, 0, 0, 0, 1, 0, 0, 0, 1, 0,
                                 0, 0, 1, 0, 0, 0, 1, 0, 0, 0,
                                 1, 0, 0, 0, 1, 0, 0, 0, 1, 0,
                                 0, 0, 1, 0, 0, 0, 1, 0, 0, 0,
                                 1, 0, 0, 0, 1, 0, 0, 0, 1, 0,
                                 0, 0, 1, 0, 0, 0, 1, 0, 0, 0};
const int 
JulianCalADT::AccumLeapYears[100] =  { 0, 0, 0, 0, 1, 1, 1, 1, 2, 2,
                                       2, 2, 3, 3, 3, 3, 4, 4, 4, 4,
                                       5, 5, 5, 5, 6, 6, 6, 6, 7, 7,
                                       7, 7, 8, 8, 8, 8, 9, 9, 9, 9,
                                       10,10,10,10,11,11,11,11,12,12,
                                       12,12,13,13,13,13,14,14,14,14,
                                       15,15,15,15,16,16,16,16,17,17,
                                       17,17,18,18,18,18,19,19,19,19,
                                       20,20,20,20,21,21,21,21,22,22,
                                       22,22,23,23,23,23,24,24,24,24};

const int JulianCalADT::DateMetaInfo::DefaultCent = 19;
const int JulianCalADT::DateMetaInfo::DefaultYear = 0;
const int JulianCalADT::DateMetaInfo::DefaultMonth = 1;
const int JulianCalADT::DateMetaInfo::DefaultDay = 1;
const int JulianCalADT::DateMetaInfo::DefaultHr = 0;
const int JulianCalADT::DateMetaInfo::DefaultMin = 0;
const int JulianCalADT::DateMetaInfo::DefaultSec = 0;

// The commented out one is right. Put the other in just to check
// that the user-defined format is propagated everywhere
const char JulianCalADT::DateFormat::DefaultSpec[20] = "M/D/CYe"; // "CY-M-Dwh:m:se";

/********************************************************************************
 * JulianCalADT
 *******************************************************************************/

XxxBool JulianCalADT::DateFormat::CentSpecified() 
{
 XxxBool found = XXX_FALSE;
 char* FormatString = Spec;
 while (!found && *FormatString && (*FormatString != 'e')) {
   if (*FormatString == 'C')
     found = XXX_TRUE;
   FormatString++;
  }
 return found;
}
//------------------------------------------------------------------------------
XxxBool JulianCalADT::DateMetaInfo::ValidDateFormat(const char*FormatString)
{
 if (strlen (FormatString) > DateFormat::SpecSize()) 
   return XXX_FALSE;
 
 while (*FormatString != 'e') {
   if (!(*FormatString)) {
     // should end with an 'e'
     return XXX_FALSE;
    }
   // take care of separator characters
   if (ispunct(*FormatString) || isspace(*FormatString)) {
     FormatString++;
     continue;
    }

   switch (*FormatString) {
    case 'C':
    case 'Y':
    case 'M':
    case 'D':
    case 'h':
    case 'm':
    case 's':
    case 'w': // whitespace
    case 't': // tab
     break;
    default:
     // unknown separator character
     return XXX_FALSE;
    }
   FormatString++;
  }

 return XXX_TRUE;
}

// called by XxxADTClass::AddMetaInfo
XxxErrCode JulianCalADT::parseMetaInfo(const char* MetaInfoString,
				       MetaInfoHandle& MIHandle) const 
{
 MIHandle.Size = sizeof(DateMetaInfo);
 ASSERT(MIHandle.Size <= (unsigned int)XXX_MIHNDL_MAX);

 if (!strcmp(MetaInfoString, "")) {
   DateMetaInfo* DMI = (DateMetaInfo*) (MIHandle.Data);
   DMI->Format.ResetDefault();
   return XXX_OK;
  } 
 else if (DateMetaInfo::ValidDateFormat(MetaInfoString)) {
   DateMetaInfo* DMI = (DateMetaInfo*) (MIHandle.Data);
   strcpy(DMI->Format.Spec, MetaInfoString);
   return XXX_OK;
  } 
 else {
   char buf[100];
   sprintf(buf, "Invalid Date format: \"%s\"", MetaInfoString);
   XXX_NOTE_ERROR(XXX_BADADT, buf);
   return XXX_NOT_OK;
  }
}
//------------------------------------------------------------------------------
// call by XxxADTClass::GetMetaInfo

XxxErrCode JulianCalADT::makeMetaInfo(const MetaInfoHandle& MIHandle,
				      XxxADTMetaInfo*& MI) const {
					DateMetaInfo* HandleInfo = (DateMetaInfo*) (MIHandle.Data);
					MI = new DateMetaInfo(*HandleInfo);
					return XXX_OK;
				       }


//-------------------------------------------------------------- DatePlanInfo
void JulianCalADT::DatePlanInfo::
NextPos(const JulianCalADT::DateValue* Pos, int HowMany, 
	const XxxADTMetaInfo* MetaInfo, 
	JulianCalADT::DateValue* Next) const
{
 if (!HowMany) {
   Next->Value = Pos->Value;
   return;
  }

 if (HowMany < 0) {
   PrevPos(Pos, -HowMany, MetaInfo, Next);
   return;
  }
 for (int i=0; i < HowMany; i++ ){
   if (i == 0)
     NextPos(Pos, MetaInfo, Next);
   else
     NextPos(Next, MetaInfo, Next);
  }
}

void JulianCalADT::DatePlanInfo::
PrevPos(const JulianCalADT::DateValue* Pos, int HowMany, 
	const XxxADTMetaInfo* MetaInfo, 
	JulianCalADT::DateValue* Prev) const
{
 if (!HowMany) {
   Prev->Value = Pos->Value;
   return;
  }

 if (HowMany < 0) {
   NextPos(Pos, -HowMany, MetaInfo, Prev);
   return;
  }
 for (int i=0; i < HowMany; i++ ){
   if (i == 0)
     PrevPos(Pos, MetaInfo, Prev);
   else
     PrevPos(Prev, MetaInfo, Prev);
  }
}

/********************************************************************************
 * JulianMDYCalADT Implements (MONTH, DAY, YEAR)
 *******************************************************************************/

JulianMDYCalADT::JulianMDYCalADT(XxxBasicType Id)
: JulianDayCalADT(Id, "pdate")
{ 
 // init stats
 MDYOrdDomId = Id;
 int MaxCentury = 99;
 int MaxYear = endYear(MaxCentury);
 int MaxMonth = endMonth(MaxYear);
 int MaxDay = JulianDayCalADT::MonthDays[MaxMonth];
 maxPos = makeIntPos(MaxDay,MaxMonth,MaxYear,MaxCentury);
 minPos = makeIntPos(StartDay,StartMonth,StartYear,0);

 // registration of functions
 NumMethods = 11;
 MethArray = new FuncMethodInfo* [NumMethods];
 //Bjarke, sep. 14 2001: added '&' to all calls to FuncMethodInfo
 MethArray[0] = new FuncMethodInfo("getMonth", XXX_INT, (MetaInfoGen) OwnerMIGen, 0,
				   (ExecuteFunc)(&MDYPlanInfo::execMonth), 0);
 MethArray[1] = new FuncMethodInfo("getDay", XXX_INT, (MetaInfoGen) OwnerMIGen, 0,
				   (ExecuteFunc)(&MDYPlanInfo::execDay), 0);
 MethArray[2] = new FuncMethodInfo("getYear", XXX_INT, (MetaInfoGen) OwnerMIGen, 0,
				   (ExecuteFunc)(&MDYPlanInfo::execYear), 0);
 MethArray[3] = new FuncMethodInfo("getCentyear", XXX_INT, (MetaInfoGen) OwnerMIGen, 0,
				   (ExecuteFunc)(&MDYPlanInfo::execCentYear), 0);
 MethArray[4] = new FuncMethodInfo("getCent", XXX_INT, (MetaInfoGen) OwnerMIGen, 0,
				   (ExecuteFunc)(&MDYPlanInfo::execCent), 0);
 MethArray[5] = new FuncMethodInfo("nextYear", Id, (MetaInfoGen) OwnerMIGen, 0,
				   (ExecuteFunc)(&MDYPlanInfo::execNextYear), 1, XXX_INT);
 MethArray[6] = new FuncMethodInfo("prevYear", Id, (MetaInfoGen) OwnerMIGen, 0,
				   (ExecuteFunc)(&MDYPlanInfo::execPrevYear), 1, XXX_INT);
 MethArray[7] = new FuncMethodInfo("nextMonth", Id, (MetaInfoGen) OwnerMIGen, 0,
				   (ExecuteFunc)(&MDYPlanInfo::execNextMonth), 1, XXX_INT);
 MethArray[8] = new FuncMethodInfo("prevMonth", Id, (MetaInfoGen) OwnerMIGen, 0,
				   (ExecuteFunc)(&MDYPlanInfo::execPrevMonth), 1, XXX_INT);
 MethArray[9] = new FuncMethodInfo("next", Id, (MetaInfoGen) OwnerMIGen, 0,
				   (ExecuteFunc)(&MDYPlanInfo::execNextPos), 1, XXX_INT);
 MethArray[10] = new FuncMethodInfo("prev", Id, (MetaInfoGen) OwnerMIGen, 0,
				   (ExecuteFunc)(&MDYPlanInfo::execPrevPos), 1, XXX_INT);

 // no aggr
 NumAggrMethods = 0;

 // no rules to apply
 // RuleEngine = new XxxOptRuleEngine;
}
//------------------------------------------------------------------------------
XxxErrCode JulianMDYCalADT::
WriteText(ostream& OutStream, 
	  const XxxADTValue* Pos, 
	  const XxxADTMetaInfo* MInfo) const 
{
 DateMetaInfo* MetaData = (DateMetaInfo*)MInfo;
 int Century, Year, Month, Day;
 splitIntPos(((DateValue*)Pos)->Value, 
	     Day, Month, Year, Century);
 const char* FormatChar;
 if (MetaData)
   FormatChar = MetaData->Format.Spec;
 else
   FormatChar = DateFormat::DefaultSpec;
 OutStream.fill('0');

 while (*FormatChar && *FormatChar != 'e') {
   switch (*FormatChar) {
    case 'w':
     OutStream << ' ';
     break;
    case 'C':
     OutStream << setw(2) << Century << setw(0);
     break;
    case 'Y':
     OutStream << setw(2) << Year << setw(0);
     break;
    case 'M':
     OutStream << setw(2) << Month << setw(0);
     break;
    case 'D':
     OutStream << setw(2) << Day << setw(0);
     break;
    case 'h':
    case 'm':
    case 's':
     // ignore these three for MDY domain
     break;
    default:
     // separator character: write it out ...
     OutStream << *FormatChar;
     break;
    }
   FormatChar++;
  }
 OutStream.fill(' ');

 return XXX_OK;
}
//------------------------------------------------------------------------------
XxxErrCode JulianMDYCalADT::ReadText(istream& InStream, 
				     XxxADTValue* PosData,
				     const XxxADTMetaInfo* MInfo) const 
{
 DateMetaInfo* MetaData = (DateMetaInfo*)MInfo;
 int Century, Year, Month, Day;
 char TempChar;
 XxxBool GetCurDate = XXX_FALSE;
 struct tm CurCalTime;

 InStream >> ws;
 TempChar = InStream.peek();
 if (TempChar == JulianMDYCalADT::NowString[0]) {
   char InitInput[100];
   InStream.width(99);
   InStream >> InitInput;
   if (!strcmp(InitInput, JulianMDYCalADT::NowString)) {
     time_t CurTime;
     if (time(&CurTime) == -1)
       return XXX_NOT_OK;
     else {
       struct tm* tmpCurCalTime = localtime(&CurTime);
       CurCalTime = *tmpCurCalTime;
       GetCurDate = XXX_TRUE;
      }
    }
  }

 const char* FormatChar;
 if (MetaData) {
   FormatChar = MetaData->Format.Spec;
   Century = MetaData->DefaultCent;
   Year = MetaData->DefaultYear;
   Month = MetaData->DefaultMonth;
   Day = MetaData->DefaultDay;
  }
 else {
   FormatChar = DateFormat::DefaultSpec;
  }
 while (*FormatChar && *FormatChar != 'e') {
   switch (*FormatChar) {
    case 'C':
     // special case!
     if (GetCurDate)
       Century = CurCalTime.tm_year + 1900;
     else
       InStream >> Century;
     if (*(FormatChar+1) == 'Y') {
       Year = Century%100;
       Century = Century/100;
       FormatChar++;
      }
     break;
    case 'Y':
     if (GetCurDate)
       Year = CurCalTime.tm_year + 1900 - Century*100;
     else 
       InStream >> Year;
     break;
    case 'M':
     if (GetCurDate)
       Month = CurCalTime.tm_mon + 1;
     else {
       InStream >> Month;
       // allow 0n for month n.
       if (Month == 0)
         InStream >> Month;
     }
     break;
    case 'D':
     if (GetCurDate)
       Day = CurCalTime.tm_mday;
     else {
       InStream >> Day;
       if (Day == 0)
         InStream >> Day;
     }
     break;
    case 'w':
    case 'h':
    case 'm':
    case 's':
     // ignore these three for MDY domain
     break;
    default:
     // separator character
     if (GetCurDate == XXX_FALSE) {
       InStream >> TempChar;
       if (TempChar != *FormatChar) {
	 XXX_NOTE_ERROR(XXX_BADPARSE, "Error reading date");
	 return XXX_NOT_OK;
	}
      }
     break;
    }
   FormatChar++;
  } // while

 if (InStream.bad()) return XXX_NOT_OK;

 if (Year > endYear(Century)) {
   XXX_NOTE_ERROR(XXX_BADPARSE, "Read invalid year");
   return XXX_NOT_OK;
  }
 if (Month < StartMonth || Month > endYear(Century)) {
   XXX_NOTE_ERROR(XXX_BADPARSE, "Read invalid month");
   return XXX_NOT_OK;
  }
 if (Day < StartDay ||
     Day > (MonthDays[Month]+(Month==2 && isLeapYear(Year, Century))) ) {
   XXX_NOTE_ERROR(XXX_BADPARSE, "Read invalid day");
   return XXX_NOT_OK;
  }
 
 ((DateValue *)PosData)->Value = makeIntPos(Day, Month, Year, Century);

 if (((DateValue*)PosData)->Value >
     ((JulianMDYCalADT*)(XxxADTArray[JulianMDYCalADT::MDYOrdDomId]))->maxPos) {
   XXX_NOTE_ERROR(XXX_BADPARSE, "Read date is too large");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

XxxErrCode JulianMDYCalADT::CopyHToN(const XxxADTValue* Pos, 
				     XxxADTValue *RPos,
				     const XxxADTMetaInfo* MetaInfo) const
{
  *((DateValue*)RPos) = *((DateValue*)Pos);
  XxxNetworkInt(&(((DateValue*)RPos)->Value), ((DateValue*)Pos)->Value);
  return XXX_OK;
}

XxxErrCode JulianMDYCalADT::CopyNToH(const XxxADTValue* Pos, 
				     XxxADTValue *RPos,
				     const XxxADTMetaInfo* MetaInfo) const
{
  *((DateValue*)RPos) = *((DateValue*)Pos);
  XxxHostInt(&(((DateValue*)RPos)->Value), ((DateValue*)Pos)->Value);
  return XXX_OK;
}


//---------------------------------------------------------------- MDYPlanInfo
// TODO: error check for invalid date
void JulianMDYCalADT::MDYPlanInfo::
NextYear(const JulianCalADT::DateValue* Pos, int HowMany, 
	 const XxxADTMetaInfo* MetaInfo, 
	 JulianCalADT::DateValue* NextPos) const
{
 int Century, Year, Month, Day;

 if (!HowMany) {
   NextPos->Value = Pos->Value;
   return;
  }
 if (HowMany < 0) {
   PrevYear(Pos, -HowMany, MetaInfo, NextPos);
   return;
  }

 splitIntPos(Pos->Value, Day, Month, Year, Century);
 for (int i = 1; i <= HowMany; i++) {
     nextYear(Year, Century);
  }
 NextPos->Value = makeIntPos(Day, Month, Year, Century);
}

void JulianMDYCalADT::MDYPlanInfo::
PrevYear(const JulianCalADT::DateValue* Pos, int HowMany, 
	 const XxxADTMetaInfo* MetaInfo, 
	 JulianCalADT::DateValue* PrevPos) const
{
 int Century, Year, Month, Day;
 if (!HowMany) {
   PrevPos->Value = Pos->Value;
   return;
  }
 if (HowMany < 0) {
   NextYear(Pos, -HowMany, MetaInfo, PrevPos);
   return;
  }

 splitIntPos(Pos->Value, Day, Month, Year, Century);
 for (int i = 1; i <= HowMany; i++ ){
    prevYear(Year, Century);
  }
 PrevPos->Value = makeIntPos(Day, Month, Year, Century);
}

void JulianMDYCalADT::MDYPlanInfo::
NextMonth(const JulianCalADT::DateValue* Pos, int HowMany, 
	  const XxxADTMetaInfo* MetaInfo, 
	  JulianCalADT::DateValue* NextPos) const
{
 int Century, Year, Month, Day;

 if (!HowMany) {
   NextPos->Value = Pos->Value;
   return;
  }
 if (HowMany < 0) {
   PrevMonth(Pos, -HowMany, MetaInfo, NextPos);
   return;
  }

 splitIntPos(Pos->Value, Day, Month, Year, Century);
 for (int i = 1; i <= HowMany; i++) {
     nextMonth(Month, Year, Century);
  }
 NextPos->Value = makeIntPos(Day, Month, Year, Century);
}

void JulianMDYCalADT::MDYPlanInfo::
PrevMonth(const JulianCalADT::DateValue* Pos, int HowMany, 
	  const XxxADTMetaInfo* MetaInfo, 
	  JulianCalADT::DateValue* PrevPos) const
{
 int Century, Year, Month, Day;
 if (!HowMany) {
   PrevPos->Value = Pos->Value;
   return;
  }
 if (HowMany < 0) {
   NextMonth(Pos, -HowMany, MetaInfo, PrevPos);
   return;
  }

 splitIntPos(Pos->Value, Day, Month, Year, Century);
 for (int i = 1; i <= HowMany; i++ ){
    prevMonth(Month, Year, Century);
  }
 PrevPos->Value = makeIntPos(Day, Month, Year, Century);
}

//----------------------------------------------------------------

void JulianMDYCalADT::MDYPlanInfo::
NextPos(const JulianCalADT::DateValue* Pos, int HowMany, 
	const XxxADTMetaInfo* MetaInfo, 
	JulianCalADT::DateValue* NextPos) const
{
  // ((JulianDayCalADT::DatePlanInfo *)this)->NextPos(Pos, HowMany, MetaInfo, NextPos);
 this->DatePlanInfo::NextPos(Pos, HowMany, MetaInfo, NextPos);
}

void JulianMDYCalADT::MDYPlanInfo::
PrevPos(const JulianCalADT::DateValue* Pos, int HowMany, 
	const XxxADTMetaInfo* MetaInfo, 
	JulianCalADT::DateValue* PrevPos) const
{
  // ((JulianDayCalADT::DatePlanInfo *)this)->PrevPos(Pos, HowMany, MetaInfo, PrevPos);
 this->DatePlanInfo::PrevPos(Pos, HowMany, MetaInfo, PrevPos);
}

//------------------------------------------------------------------------------
void JulianMDYCalADT::MDYPlanInfo::
NextPos(const JulianCalADT::DateValue* Pos, 
	const XxxADTMetaInfo* MetaInfo, 
	JulianCalADT::DateValue* Next) const
{
 int Century, Year, Month, Day;
 splitIntPos(Pos->Value, Day, Month, Year, Century);

 nextDay(Day, Month, Year, Century);
 Next->Value = makeIntPos(Day,Month,Year,Century);
}
//------------------------------------------------------------------------------
void JulianMDYCalADT::MDYPlanInfo::
PrevPos(const JulianCalADT::DateValue* Pos, 
	const XxxADTMetaInfo* MetaInfo, 
	JulianCalADT::DateValue* Prev) const
{
 int IntPos = Pos->Value;

 // TODO should find less messy way of getting info from ADT owning this plan
 if (IntPos >
     ((JulianMDYCalADT*)(XxxADTArray[JulianMDYCalADT::MDYOrdDomId]))->minPos) {
   int Century, Year, Month, Day;
   splitIntPos(IntPos, Day, Month, Year, Century);

   prevDay(Day, Month, Year, Century);
   Prev->Value = makeIntPos(Day,Month,Year,Century);
  } 
 else
   Prev->Value = IntPos;
}
//------------------------------------------------------------------------------
int JulianMDYCalADT::MDYPlanInfo::
NumPositions(const JulianCalADT::DateValue* Pos1, 
	     const JulianCalADT::DateValue* Pos2) const
{
 int IntPos1 = Pos1->Value;

 int Century1, Year1, Month1, Day1;
 splitIntPos(IntPos1, Day1, Month1, Year1, Century1);

 int IntPos2 = Pos2->Value;

 int Century2, Year2, Month2, Day2;
 splitIntPos(IntPos2, Day2, Month2, Year2, Century2);

 return numDays(Day1, Month1, Year1, Century1,
		Day2, Month2, Year2, Century2);
}

//------------------------------------------------------------------------------
// ADT methods

XxxErrCode JulianMDYCalADT::MDYPlanInfo::
execMonth(XxxValueEnv* Env, XxxADTValue* ReturnValue) const 
{
 JulianCalADT::DateValue* SrcDate = (DateValue*) (Env->OwnerValPtr);
 ASSERT((Env->NumArgs == 0));
 int Day, Month, Yr, Cent;

 splitIntPos(SrcDate->Value, Day, Month, Yr, Cent);
 ((JulianCalADT::DateValue*)ReturnValue)->Value = Month;
 return XXX_OK;
}
//------------------------------------------------------------------------------
XxxErrCode JulianMDYCalADT::MDYPlanInfo::
execDay(XxxValueEnv* Env, XxxADTValue* ReturnValue) const 
{
 JulianCalADT::DateValue* SrcDate = (DateValue*) (Env->OwnerValPtr);
 ASSERT((Env->NumArgs == 0));
 int Day, Month, Yr, Cent;

 splitIntPos(SrcDate->Value, Day, Month, Yr, Cent);
 ((JulianCalADT::DateValue*)ReturnValue)->Value = Day;
 return XXX_OK;
}
//------------------------------------------------------------------------------
XxxErrCode JulianMDYCalADT::MDYPlanInfo::
execYear(XxxValueEnv* Env, XxxADTValue* ReturnValue) const 
{
 JulianCalADT::DateValue* SrcDate = (DateValue*) (Env->OwnerValPtr);
 DateMetaInfo* DMI = (DateMetaInfo*) (Env->OwnerCatInfo);
 ASSERT(DMI && (Env->NumArgs == 0));
 int Day, Month, Yr, Cent;

 splitIntPos(SrcDate->Value, Day, Month, Yr, Cent);
 if (DMI->Format.CentSpecified())
   ((JulianCalADT::DateValue*)ReturnValue)->Value = Cent*100 + Yr;
 else
   ((JulianCalADT::DateValue*)ReturnValue)->Value = Yr;
 return XXX_OK;
}
//------------------------------------------------------------------------------
XxxErrCode JulianMDYCalADT::MDYPlanInfo::
execCentYear(XxxValueEnv* Env, XxxADTValue* ReturnValue) const 
{
 JulianCalADT::DateValue* SrcDate = (DateValue*) (Env->OwnerValPtr);
 ASSERT((Env->NumArgs == 0));
 int Day, Month, Yr, Cent;

 splitIntPos(SrcDate->Value, Day, Month, Yr, Cent);
 ((JulianCalADT::DateValue*)ReturnValue)->Value = Cent*100 + Yr;
 return XXX_OK;
}
//------------------------------------------------------------------------------
XxxErrCode JulianMDYCalADT::MDYPlanInfo::
execCent(XxxValueEnv* Env, XxxADTValue* ReturnValue) const 
{
 JulianCalADT::DateValue* SrcDate = (DateValue*) (Env->OwnerValPtr);
 ASSERT((Env->NumArgs == 0));
 int Day, Month, Yr, Cent;

 splitIntPos(SrcDate->Value, Day, Month, Yr, Cent);
 ((JulianCalADT::DateValue*)ReturnValue)->Value = Cent;
 return XXX_OK;
}

//----------------------------------------------------------------------
XxxErrCode JulianMDYCalADT::MDYPlanInfo::
execNextYear(XxxValueEnv* Env, XxxADTValue* ReturnValue) const 
{
 JulianCalADT::DateValue* StartDate = (JulianCalADT::DateValue*) (Env->OwnerValPtr);
 ASSERT((Env->NumArgs == 1));
 int HowMany = ((XxxCoreADTClass::IntValue*)(Env->ArgValsPtrs[0]))->Value;

 NextYear(StartDate, HowMany, Env->OwnerCatInfo, 
	     (JulianCalADT::DateValue *)ReturnValue);
 return XXX_OK;
}

XxxErrCode JulianMDYCalADT::MDYPlanInfo::
execPrevYear(XxxValueEnv* Env, XxxADTValue* ReturnValue) const 
{
 JulianCalADT::DateValue* StartDate = 
   (JulianCalADT::DateValue*) (Env->OwnerValPtr);
 ASSERT((Env->NumArgs == 1));
 int HowMany = ((XxxCoreADTClass::IntValue*)(Env->ArgValsPtrs[0]))->Value;

 PrevYear(StartDate, HowMany, Env->OwnerCatInfo, 
	     (JulianCalADT::DateValue*)ReturnValue);
 return XXX_OK;
}

XxxErrCode JulianMDYCalADT::MDYPlanInfo::
execNextMonth(XxxValueEnv* Env, XxxADTValue* ReturnValue) const 
{
 JulianCalADT::DateValue* StartDate = (JulianCalADT::DateValue*) (Env->OwnerValPtr);
 ASSERT((Env->NumArgs == 1));
 int HowMany = ((XxxCoreADTClass::IntValue*)(Env->ArgValsPtrs[0]))->Value;

 NextMonth(StartDate, HowMany, Env->OwnerCatInfo, 
	     (JulianCalADT::DateValue *)ReturnValue);
 return XXX_OK;
}

XxxErrCode JulianMDYCalADT::MDYPlanInfo::
execPrevMonth(XxxValueEnv* Env, XxxADTValue* ReturnValue) const 
{
 JulianCalADT::DateValue* StartDate = 
   (JulianCalADT::DateValue*) (Env->OwnerValPtr);
 ASSERT((Env->NumArgs == 1));
 int HowMany = ((XxxCoreADTClass::IntValue*)(Env->ArgValsPtrs[0]))->Value;

 PrevMonth(StartDate, HowMany, Env->OwnerCatInfo, 
	     (JulianCalADT::DateValue*)ReturnValue);
 return XXX_OK;
}

XxxErrCode JulianMDYCalADT::MDYPlanInfo::
execNextPos(XxxValueEnv* Env, XxxADTValue* ReturnValue) const 
{
 JulianCalADT::DateValue* StartDate = (JulianCalADT::DateValue*) (Env->OwnerValPtr);
 ASSERT((Env->NumArgs == 1));
 int HowMany = ((XxxCoreADTClass::IntValue*)(Env->ArgValsPtrs[0]))->Value;

 NextPos(StartDate, HowMany, Env->OwnerCatInfo, 
	     (JulianCalADT::DateValue *)ReturnValue);
 return XXX_OK;
}

XxxErrCode JulianMDYCalADT::MDYPlanInfo::
execPrevPos(XxxValueEnv* Env, XxxADTValue* ReturnValue) const 
{
 JulianCalADT::DateValue* StartDate = 
   (JulianCalADT::DateValue*) (Env->OwnerValPtr);
 ASSERT((Env->NumArgs == 1));
 int HowMany = ((XxxCoreADTClass::IntValue*)(Env->ArgValsPtrs[0]))->Value;

 PrevPos(StartDate, HowMany, Env->OwnerCatInfo, 
	     (JulianCalADT::DateValue*)ReturnValue);
 return XXX_OK;
}
