#include "document.h"
#include "iomanip.h"
#include "docutils.h"

XxxErrCode 
XxxDocumentADT::PlanInfo::execPrint(XxxValueEnv* Env, 
				    XxxADTValue *ReturnValue) const
{
 // Get the Document structure to print.
 DocumentStruct* DS = (DocumentStruct *)(Env->OwnerValPtr);
  
 char *TmpName = tempnam(XXX_THISENV().TmpDataDir, "xxx");
 
 if (XxxShoreSM->WriteToOSFile(DS->DocumentId.VolId, 
			       DS->DocumentId.LocalId,
			       TmpName) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_BADADT, "Error writing document file to file");
   return XXX_NOT_OK;
  }
  
 char Command[500];
 sprintf(Command, "lpr %s", TmpName);
 if (system(Command)) {
   char buf[200];
   sprintf(buf, "Error printing document file %s", TmpName);
   XXX_NOTE_ERROR(XXX_BADADT, buf);
   return XXX_NOT_OK;
  }
 ((XxxCoreADTClass::IntValue *)ReturnValue)->Value = 1; // successful print
 return XXX_OK;
}


XxxErrCode 
XxxDocumentADT::PlanInfo::execEmail(XxxValueEnv* Env, 
				    XxxADTValue *ReturnValue) const
{
 // Get the Document structure to attach to the email.
 DocumentStruct* DS = (DocumentStruct *)(Env->OwnerValPtr);
 // Get the recipient of this email.
 char* recipient = (char *)(((XxxCoreADTClass::StrValue *)(Env->ArgValsPtrs[0]))->Value);
 // Get the subject line of this email.
 char* subject = (char *)(((XxxCoreADTClass::StrValue *)(Env->ArgValsPtrs[1]))->Value);

 char *TmpName = tempnam(XXX_THISENV().TmpDataDir, "xxx");
  
 if (XxxShoreSM->WriteToOSFile(DS->DocumentId.VolId, 
			       DS->DocumentId.LocalId,
			       TmpName) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_BADADT, "Error writing document file to file");
   return XXX_NOT_OK;
  }
  
 char Command[500];
 sprintf(Command, "mail -s \"%s\" %s < %s", subject, recipient, TmpName);
 if (system(Command)) {
   char buf[200];
   sprintf(buf, "Error emailing document file %s", TmpName);
   XXX_NOTE_ERROR(XXX_BADADT, buf);
   return XXX_NOT_OK;
  }
 ((XxxCoreADTClass::IntValue *)ReturnValue)->Value = 1; // successful email
 return XXX_OK;
}


//// A much harder search.  This returns the number of times the
//// string exists in the document as it's own WORD.  So this is
//// more of a word search, not string search
XxxErrCode 
XxxDocumentADT::PlanInfo::execExists(XxxValueEnv* Env,
									 XxxADTValue *ReturnValue) const
{
    ASSERT((Env->NumArgs == 1));
    ASSERT((Env->OwnerValPtr ));
 // Get the Document structure to grep for the input string.
 DocumentStruct* DS = (DocumentStruct *)(Env->OwnerValPtr);
 // Get the string to search for.
 char *searchString = (char *)(((XxxCoreADTClass::StrValue *)(Env->ArgValsPtrs[0]))->Value);

 char *TmpName = tempnam(XXX_THISENV().TmpDataDir, "xxx");
  
 if (XxxShoreSM->WriteToOSFile(DS->DocumentId.VolId, 
			       DS->DocumentId.LocalId,
			       TmpName) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_BADADT, "Error writing document file to file");
   return XXX_NOT_OK;
  }
 
 int numberOfOccurrences = 0;
 ifstream iFile(TmpName, ios::in);
 //Bjarke, aug. 3.: unsigned int used as type for bufsize
 const unsigned int bufsize = 1024;
 char buf[bufsize];

 while(iFile >> setw(bufsize) >> buf){
   if(!strcmp(searchString,buf)){
     numberOfOccurrences++;
   }
 }
 
 ((XxxCoreADTClass::IntValue *)ReturnValue)->Value = numberOfOccurrences; // Not a successful search
 return XXX_OK;
}

/**
 * This is another implementation of string search.
 * Troubles:
 * 1. only handle ASCII
 * 2. try to load the WHOLE object at once into memory - not feasible for large object
 * @see doc_utils.h for auxillary functions
 * @author Wai-Kwong Sam LEE
 * @version
 */

XxxErrCode 
XxxDocumentADT::PlanInfo::execSearch(XxxValueEnv* Env,
									 XxxADTValue *ReturnValue) const
{
	ASSERT((Env->NumArgs == 1));
    ASSERT((Env->OwnerValPtr ));
 // Get the Document structure to grep for the input string.
 DocumentStruct* DS = (DocumentStruct *)(Env->OwnerValPtr);
 // Get the string to search for.
 char* SearchString = (char*)(((XxxCoreADTClass::StrValue *)(Env->ArgValsPtrs[0]))->Value);

 char* Doc;
 unsigned int DocSize;
 const int CheckCase = 0;

 if (XxxShoreSM->GetAllocateObject(DS->DocumentId.VolId, 
			       DS->DocumentId.LocalId,
			       Doc,
			       DocSize) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_BADADT, "Error reading document file from storage");
   return XXX_NOT_OK;
  }

 int numberOfOccurrences = 0;
 switch(DS->DocType) {
 case ASCII:
   if (!CheckCase) {
     tolowerstr(SearchString);
     tolowerstr(Doc);
   }
   numberOfOccurrences = XxxStrFind(Doc, SearchString);
   break;
 case MSWORD:
 case HTML:
 case POSTSCRIPT:
 case ANY:
 default:
   cerr << "ERROR: document.search() - unsupported document type\n";
 } // switch(DS.DocType);
 
 
((XxxCoreADTClass::IntValue *)ReturnValue)->Value = numberOfOccurrences;

 delete []Doc;

 return XXX_OK;
}
