#include "document.h"
#include "optrules.h"
#include "syscmd.h"
#include "memmap.h"
#include "GetFile.h"

/*****
Here's the info on Document EADT.

 Each document schema has :
    *) a format
    *) a flag specifying whether input data is "inlin" or not
        (by default, this is FALSE)
    
 Document input format: File-Name/URL name or inline text
****/

// define the static variables
XxxBasicType  XxxDocumentADT::DocumentTypeId = XXX_UNKNOWN;
const char* const XxxDocumentADT::ASCII_FORMAT = "ascii";
const char* const XxxDocumentADT::MSWORD_FORMAT = "msword";
const char* const XxxDocumentADT::HTML_FORMAT = "html";
const char* const XxxDocumentADT::POSTSCRIPT_FORMAT = "ps";
const char* const XxxDocumentADT::ANY_FORMAT = "any";

const char* const XxxDocumentADT::DocumentMetaInfo::XXX_FORMAT = "format";
const char* const XxxDocumentADT::DocumentMetaInfo::XXX_INLINE = "inline";

//
// Document Data Type
//
XxxDocumentADT::XxxDocumentADT(XxxBasicType TypeId, 
			       const char *MyName,
			       int ObjSize)
: XxxLargeFixedADT(TypeId, MyName, ObjSize) 
{
 DocumentTypeId = TypeId;

 NumMethods = 4;
 MethArray = new FuncMethodInfo*[NumMethods];
 //Bjarke, sep. 14 2001: added '&' to all calls to FuncMethodInfo
 MethArray[0] = new FuncMethodInfo("print", XXX_INT, 
			       (MetaInfoGen)(OwnerMIGen), 0, (ExecuteFunc)(&PlanInfo::execPrint), 0);
 MethArray[1] = new FuncMethodInfo("email", XXX_INT, 
				(MetaInfoGen)(OwnerMIGen), 0, (ExecuteFunc)(&PlanInfo::execEmail), 
				    2, XXX_STRING, XXX_STRING);
 MethArray[2] = new FuncMethodInfo("exists", XXX_INT, 
			       (MetaInfoGen)(OwnerMIGen), 0, (ExecuteFunc)(&PlanInfo::execExists),
				   1, XXX_STRING);
 MethArray[3] = new FuncMethodInfo("search", XXX_INT, 
				(MetaInfoGen)(OwnerMIGen), 0, (ExecuteFunc)(&PlanInfo::execSearch), 
				    1, XXX_STRING);
}

XxxDocumentADT::~XxxDocumentADT()
{
}

/**
 * Makes the Document Meta Info (The Format)
 */
XxxErrCode XxxDocumentADT::makeMetaInfo(const MetaInfoHandle& MIHandle,
					XxxADTMetaInfo*& MI) const
{
 DocumentMetaInfo* HandleInfo = (DocumentMetaInfo*)(MIHandle.Data);
 MI = new DocumentMetaInfo(*HandleInfo);
 return XXX_OK;
}

/**
 * Parses the Document Meta Info (Checks The Format)
 */
XxxErrCode XxxDocumentADT::parseMetaInfo(const char* MetaInfoString,
					 MetaInfoHandle& MIHandle) const
{
 MIHandle.Size = sizeof(DocumentMetaInfo);
 ASSERT((MIHandle.Size <= (unsigned int)XXX_MIHNDL_MAX));
 DocumentMetaInfo* DMI = (DocumentMetaInfo*) (MIHandle.Data);
 DMI->Init();
 XxxADTMetaInfo::Itemizer Itemizer(",");
 XxxADTMetaInfo::StringTokenizer Tokenizer;
 XxxSetCursor SCsr;

 Itemizer.InitScan(SCsr, MetaInfoString);
 while (1) {
   char* Entry;
   Itemizer.NextItem(SCsr, Entry);
   if (!SCsr.IsValid()) 
     break;
   else {
     XxxSetCursor SCsr1;
     char* MIEntry;
     Tokenizer.InitScan(SCsr1, Entry);
     Tokenizer.NextItem(SCsr1, MIEntry);
     if (SCsr1.IsValid()) {
       if (!strcasecmp(MIEntry, DocumentMetaInfo::XXX_FORMAT)) {
	 XXX_CALL_RET(parseFormat(Tokenizer, SCsr1, DMI),
		      XXX_BADADT, "Error parsing format");
	}
       else if (!strcasecmp(MIEntry, DocumentMetaInfo::XXX_INLINE)) {
	 XXX_CALL_RET(parseInline(Tokenizer, SCsr1, DMI),
		      XXX_BADADT, "Error parsing inline format");
	}
       else {
	 XXX_NOTE_ERROR(XXX_BADADT, "Unknown document metainfo item");
	 return XXX_NOT_OK;
	}
      }
     Tokenizer.CloseScan(SCsr1);
    }
  }//while (1)
 Itemizer.CloseScan(SCsr);
 
 return XXX_OK;
}//parseMetaInfo

XxxErrCode XxxDocumentADT::
parseFormat(XxxADTMetaInfo::StringTokenizer& Tokenizer,
	    XxxSetCursor& SCsr1,
	    DocumentMetaInfo* DMI) const
{
 char* Format;
 Tokenizer.NextItem(SCsr1, Format);
 if (SCsr1.IsValid()) {
   if (!strcasecmp(Format, ASCII_FORMAT)) 
    { DMI->Format = ASCII; }
   else if (!strcasecmp(Format, MSWORD_FORMAT)) 
    { DMI->Format = MSWORD; }
   else if (!strcasecmp(Format, HTML_FORMAT)) 
    { DMI->Format = HTML; }
   else if (!strcasecmp(Format, POSTSCRIPT_FORMAT)) 
    { DMI->Format = POSTSCRIPT; }
   else if (!strcasecmp(Format, ANY_FORMAT)) 
    { DMI->Format = ANY; }
   else {
     XXX_NOTE_ERROR(XXX_BADADT, "Unknown document format");
     return XXX_NOT_OK;
    }//else (wrong format specified)
  } // if (correct format is found)
 else {
   XXX_NOTE_ERROR(XXX_BADADT, "No format specified for document");
   return XXX_NOT_OK;
 }
 return XXX_OK;
}

XxxErrCode XxxDocumentADT::
parseInline(XxxADTMetaInfo::StringTokenizer& Tokenizer,
	    XxxSetCursor& SCsr1,
	    DocumentMetaInfo* DMI) const
{
 char* Inline;
 Tokenizer.NextItem(SCsr1, Inline);
 if (SCsr1.IsValid()) {
   if (!strcasecmp(Inline, "yes")) 
    { DMI->InputInline = XXX_TRUE; }
   else if (!strcasecmp(Inline, "no")) 
    { DMI->InputInline = XXX_FALSE; }
   else {
     XXX_NOTE_ERROR(XXX_BADADT, "Unknown document inline format");
     return XXX_NOT_OK;
    } 
  } 
 else {
   XXX_NOTE_ERROR(XXX_BADADT, "No inline format specified for document");
   return XXX_NOT_OK;
 }
 return XXX_OK;
}

//// called by WriteText to show the document file 
XxxErrCode XxxDocumentADT::showDataFile(const char* FileName,
					ostream& OutStream, 
					const XxxADTValue* Object,
					const XxxADTMetaInfo* MetaInfo) const
{
 DocumentMetaInfo* DMI = (DocumentMetaInfo*)MetaInfo;
 char Command[20 + 2*XXX_DIRNAME_MAX];
 DocumentType DocFormat = DMI->Format;
 DocumentStruct* DS = (DocumentStruct*)Object;
 if (DocFormat == ANY) {
   // get the format from the object
   DocFormat = DS->DocType; 
  }
 else {
   if (DocFormat != DS->DocType) {
     XXX_NOTE_ERROR(XXX_BADADT, "Format in document doesnt match format of meta-info");
     return XXX_NOT_OK;
    }
  }
 switch (DocFormat) {
  case POSTSCRIPT:
   sprintf(Command, "ghostview %s; rm -f %s", FileName, FileName);
   break;
  case ASCII:
  case MSWORD:
  case HTML:
  case ANY:
   sprintf(Command, "emacs %s; rm -f %s", FileName, FileName);
   break;
  }
 if (system(Command)) {
   XXX_NOTE_ERROR(XXX_BADADT, "Error displaying document");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

/**
 * Used to read in the document file name and its corresponding data
 */
XxxErrCode XxxDocumentADT::ReadText(istream& InStream, 
				    XxxADTValue* ObjectRef,
				    const XxxADTMetaInfo* MetaInfo) const
{
 XXX_SM_CHECK;
 XxxErrCode rc = XXX_OK;

 char DocumentFileName[XXX_DIRNAME_MAX];
 XxxBool TmpFileCreated = XXX_FALSE;

 DocumentMetaInfo* DMI = (DocumentMetaInfo*) (MetaInfo);
 if (!DMI->InputInline) {
   InStream >> DocumentFileName;
   if (InStream.bad()) {
     XXX_NOTE_ERROR(XXX_BADADT, "Cannot read document file name");
     return XXX_NOT_OK;
    }
   
   // determine file location
   if (strncmp(DocumentFileName,"http://",7)==0){
     // load document file from URL
     char *TmpName = tempnam(XXX_THISENV().TmpDataDir, "xxx");
     if (XxxGetFile(DocumentFileName,TmpName) != XXX_OK) {
       XXX_NOTE_ERROR(XXX_BADADT, "Error calling XxxGetFile for document");
       return XXX_NOT_OK;
      }
     strcpy(DocumentFileName,TmpName);
     TmpFileCreated = XXX_TRUE;
    } // if it is a URL
  } // if the input is not inline
 else {
   // write the inline input to a tmp file & assign its name to DocumentFileName
   char *TmpName = tempnam(XXX_THISENV().TmpDataDir, "xxx");
   ofstream OStream(TmpName);
   if (OStream.bad()) {
     XXX_NOTE_ERROR(XXX_BADADT, "Error creating temp file");
     return XXX_NOT_OK;
    }
   char Buf[100];  int BufCount;
   while (InStream.good()) {
     InStream.read(Buf, 100);
     BufCount = InStream.gcount();
     OStream.write(Buf, BufCount);
     if (OStream.bad()) {
       XXX_NOTE_ERROR(XXX_BADADT, "Error writing to temp file");
       return XXX_NOT_OK;
      }
    }
   strcpy(DocumentFileName,TmpName);
   TmpFileCreated = XXX_TRUE;
  } // else if inline input

 // open the document file
 XxxShoreObjectId DocumentOID;
 // insert a "record" of this object
 XxxLargeADTEnv 
   *DocumentEnv = (XxxLargeADTEnv*) XXX_THISADTENV(DocumentTypeId);
 // note: the constructor mmaps the file, while
 // the destructor unmaps the file
 XxxMemoryMap DocMap(DocumentFileName);
 void *MMapAddress = NULL;
 if (DocMap.GetHandle(MMapAddress) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_BADFILE, "Error in mmap");
   return XXX_NOT_OK;
 }

 if (XxxShoreSM->InsertRecord(DocumentEnv->StorageVolume, DocumentEnv->StorageFile,
			      MMapAddress,
			      DocMap.Size(),
			      DocumentOID) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_BADFILE, "Unable to store document file in SHORE");
   return XXX_NOT_OK;
  }
 //Fill in all fields necessary for accurate document 
 //data file (found in document structure)
 DocumentStruct* DS = (DocumentStruct*)ObjectRef;
 DS->DocumentId.VolId = DocumentEnv->StorageVolume;
 DS->DocumentId.LocalId = DocumentOID;
 DS->DocSize = DocMap.Size();
 DS->DocType = DMI->Format;
 // TODO:L2 Add code here to check the format to make sure
 // that the document does really belong to the right format

 // cerr << "DocSize " << DS->DocSize << " IDs " << DS->DocumentId.VolId << " " << DS->DocumentId.LocalId << endl;

  // Free the locks the memory map had on the temp file
 if (DocMap.close() != XXX_OK) {
	 // XxxMemoryMap takes care of all the error reporting
	 return XXX_NOT_OK;
 }

 if (TmpFileCreated) {
   // delete temporary file
   if (XxxSystemCmd::DeleteFile((const char*)DocumentFileName) < 0) {
     XXX_NOTE_ERROR(XXX_BADADT, "Error calling remove for document");
     return XXX_NOT_OK;
   }
  } // if (TmpFileCreated)
 return XXX_OK;
}//ReadText

const char * XxxDocumentADT::GetContentType(void *ResultHandle) const
{
  DocumentStruct *DSBuf = (DocumentStruct *)ResultHandle;
  if (!DSBuf) return "text/plain"; // should not happen

  if (DSBuf->DocType == ASCII)
    return "text/plain";
  if (DSBuf->DocType == HTML)
    return "text/html";
  if (DSBuf->DocType == MSWORD)
    return "application/msword";
  if (DSBuf->DocType == POSTSCRIPT)
    return "application/postscript";
  // Fail-safe
  return "text/plain";
}


XxxErrCode XxxDocumentADT::CopyHToN(const XxxADTValue* Object, 
				    XxxADTValue *ResultObject,
				    const XxxADTMetaInfo* MetaInfo) const
{
  // copy the header fields
  *ResultObject = *Object;

  DocumentStruct *DSBuf = (DocumentStruct *)Object;
  DocumentStruct *RDSBuf = (DocumentStruct *)ResultObject;
  ASSERT ((DSBuf && RDSBuf));
  return RDSBuf->CopyHToN(DSBuf, (DocumentMetaInfo*)MetaInfo);
}

XxxErrCode XxxDocumentADT::CopyNToH(const XxxADTValue* Object, 
				    XxxADTValue *ResultObject,
				    const XxxADTMetaInfo* MetaInfo) const
{
  // copy the header fields
  *ResultObject = *Object;

  DocumentStruct *DSBuf = (DocumentStruct *)Object;
  DocumentStruct *RDSBuf = (DocumentStruct *)ResultObject;
  ASSERT ((DSBuf && RDSBuf));
  return RDSBuf->CopyNToH(DSBuf, (DocumentMetaInfo*)MetaInfo);
}

