#ifndef XXX_DOCUMENT_H
#define XXX_DOCUMENT_H
#include "genadt.h"
#include "shoresm.h"
#include "adtfunc.h"

/**
 *  Document EADT -> See document.C for 
 *  more info.
 *
 */
class XxxDocumentADT : public XxxLargeFixedADT {
 public:
  // type id accessible without object
  static XxxBasicType DocumentTypeId;
  // local typedefs
  typedef enum {ASCII,
	        MSWORD,
		HTML,
		POSTSCRIPT,
		ANY,
	        } DocumentType;
 protected:

  // local consts
  static const char* const ASCII_FORMAT;
  static const char* const MSWORD_FORMAT;
  static const char* const HTML_FORMAT;
  static const char* const POSTSCRIPT_FORMAT;
  static const char* const ANY_FORMAT;

  //// The Meta Info for the Document EADT 
  class DocumentMetaInfo : public XxxADTMetaInfo { 
   public:
    static const char* const XXX_FORMAT;
    static const char* const XXX_INLINE;
    DocumentType Format;
    XxxBool InputInline; // is the input document "inline", or is it
                         // specified as a filename/URL?
    void Init() {
      Format = ANY;
      InputInline = XXX_FALSE;
    }
    DocumentMetaInfo () { Init(); }
    virtual ~DocumentMetaInfo() {}
    virtual XxxErrCode Replicate(XxxADTMetaInfo*& NewMI) const {
      DocumentMetaInfo* NewMetaInfo = new DocumentMetaInfo(*this);
      NewMI = NewMetaInfo;
      return XXX_OK;
    }
    virtual XxxBool IsCompatible(const XxxADTMetaInfo* OtherMI) const {
      if ((((DocumentMetaInfo*)OtherMI)->Format == ANY) || 
	  (Format == ((DocumentMetaInfo*)OtherMI)->Format))
	return XXX_TRUE;
      else return XXX_FALSE;
    }
    virtual void Print(ostream& OStr) const {
      OStr << "(" << Format << ", " 
	   << (InputInline ? "Inline" : "Not Inline") << ")" ;
    }
    /**
     * Returns the number of bytes necessary to hold the serialized version
     * of this meta-info object.
     * @param Size (out) The required number of bytes
     * @return XXX_NOT_OK if the object cannot be serialized, XXX_OK otherwise
     */
    virtual XxxErrCode SerializeCopySize(int &Size) const {
      Size = sizeof(int) + sizeof(InputInline); return XXX_OK;
    }
    /**
     * Creates a binary version of this meta-info object.
     * @param Data An array of bytes of at least the size specified by the
     *             SerializeCopySize() function.
     * @return XXX_NOT_OK if the object cannot be serialized, XXX_OK otherwise
     */
    virtual XxxErrCode SerializeCopy(char *Data) const {
      int intFormat = (int)Format;
      memcpy(Data, (char *)(&intFormat), sizeof(intFormat));
      memcpy((char *)(Data + sizeof(intFormat)), (char *)(&InputInline), 
	     sizeof(InputInline));
      return XXX_OK;
    }
   };  
  friend class DocumentMetaInfo;

  //  Structure used to represent a Document file within a record
  //  Be careful that the size does not cross XXX_MIHNDL_MAX
  class DocumentStruct : public XxxADTValue {
   public: 
    XxxShoreLogicalOID  DocumentId;
    int DocSize;
    DocumentType DocType;
    XxxErrCode CopyHToN(const DocumentStruct* DS,
			const DocumentMetaInfo* MetaInfo) {
      XxxNetworkInt((int *)&(DocSize), DS->DocSize);
      XxxNetworkInt((int *)&(DocType), DS->DocType);
      memcpy((char *)&(DocumentId), (char *)&(DS->DocumentId),
	     sizeof(XxxShoreLogicalOID));
      return XXX_OK;
    }
    XxxErrCode CopyNToH(const DocumentStruct* DS,
			const DocumentMetaInfo* MetaInfo) {
      XxxHostInt((int *)&(DocSize), DS->DocSize);
      XxxHostInt((int *)&(DocType), DS->DocType);
      memcpy((char *)&(DocumentId), (char *)&(DS->DocumentId),
	     sizeof(XxxShoreLogicalOID));
      return XXX_OK;
    }
  };

  class PlanInfo : public XxxFuncPlanInfo {
    friend class XxxDocumentADT;
    XxxErrCode execPrint(XxxValueEnv* Env,
			 XxxADTValue *ReturnValue) const;
    XxxErrCode execEmail(XxxValueEnv* Env,
			 XxxADTValue *ReturnValue) const;
    XxxErrCode execExists(XxxValueEnv* Env,
			 XxxADTValue *ReturnValue) const;
    XxxErrCode execSearch(XxxValueEnv* Env,
			 XxxADTValue *ReturnValue) const;
  public:
    PlanInfo(const XxxFuncParseInfo* FuncParse, int ArgCount) 
      : XxxFuncPlanInfo(FuncParse, ArgCount)  {}
  };
  
 
  virtual XxxErrCode makeMetaInfo(const MetaInfoHandle& MIHandle,
				  XxxADTMetaInfo*& MI) const;
  virtual XxxErrCode parseMetaInfo(const char* MetaInfoString,
				   MetaInfoHandle& MIHandle) const;
  /// helper function for parseMetaInfo
  XxxErrCode parseFormat(XxxADTMetaInfo::StringTokenizer& Tokenizer,
			 XxxSetCursor& SCsr1,
			 DocumentMetaInfo* DMI) const;
  /// helper function for parseMetaInfo
  XxxErrCode parseInline(XxxADTMetaInfo::StringTokenizer& Tokenizer,
			 XxxSetCursor& SCsr1,
			 DocumentMetaInfo* DMI) const;

  static XxxADTMetaInfo* OwnerMIGen(XxxValueExpression* Owner,
				 XxxValueExprList *ArgList) {
    DocumentMetaInfo* OwnerMetaInfo = (DocumentMetaInfo* )(Owner->GenerateCatInfo());
    if (OwnerMetaInfo) {
      DocumentMetaInfo* ReturnMetaInfo =new DocumentMetaInfo(*OwnerMetaInfo);
      return ReturnMetaInfo;
     }
    else return NULL;
   }


  // called from WriteText()
  virtual XxxErrCode showDataFile(const char* FileName,
				  ostream& OutStream, 
				  const XxxADTValue* Object,
				  const XxxADTMetaInfo* MetaInfo) const;
  virtual XxxErrCode getOid(const XxxADTValue* Object, 
			    XxxShoreVolumeId& VolId, 
			    XxxShoreObjectId& LocalId) const {
    XxxShoreLogicalOID* DocumentId = &(((DocumentStruct*)Object)->DocumentId);
    VolId = DocumentId->VolId;
    LocalId = DocumentId->LocalId;
    return XXX_OK;
   }

public:
  XxxDocumentADT(XxxBasicType TypeId, const char *MyName = "document",
		 int ObjSize = sizeof(DocumentStruct));
  virtual ~XxxDocumentADT();

  virtual XxxErrCode ReadText(istream& InStream, 
			      XxxADTValue* SrcObject,
			      const XxxADTMetaInfo* MetaInfo) const;
  /**
   * This function will only be called on ADTs which support callbacks.
   * Any data type which would like to take advantage of browser capability
   * to handle some data specially, for instance, displaying images, must
   * return a valid Content Type string as described in RFC 1590.
   */
  virtual const char * GetContentType(void *ResultHandle) const;

  virtual XxxErrCode CopyHToN(const XxxADTValue* SrcObject, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const;
  virtual XxxErrCode CopyNToH(const XxxADTValue* SrcObject, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const;
};

#endif /* XXX_DOCUMENT_H */
