#include <string.h>
#include <iostream.h>

/**
 * Auxillary prefix computing function,
 * called by Kunth-Morris-Pratt algorithm String matching function.
 */
void
XxxKMPComputePrefixFunction(const char * Pattern, 
			    int* Pi)
{
	 int m = strlen(Pattern);
	 Pi[0] = 0;
	 int k = 0;
	 for (int q = 1; q < m; q++) {
		 while ((k > 0) && Pattern[k] != Pattern[q]) {
			 k = Pi[k];
		 }
		 if (Pattern[k] == Pattern[q]) {
			 k++;
		 }
		 Pi[q] = k;
	 }

} // XxxKMPComputePrefixFunction(char* Pattern, int* Pi)


/**
 * Sub String match function
 * Kunth-Morris-Pratt algorithm Implementation
 * see CLR(algo text book) Ch. 33.4, P.869 - 875
 */
int
XxxKMPStrFind(const char * text, 
	      const char * Pattern)
{

   int numberOfOccurences = 0;

   int n = strlen(text);
   int m = strlen(Pattern);
   int* Pi = new int[m];
   XxxKMPComputePrefixFunction(Pattern, Pi);

 #ifdef DEBUG
   for (int k = 0; k < m; k++) {
	   cout << Pi[k] << " ";
   }
   cout << endl;
 #endif

   int q = 0;
   for (int i = 0; i < n; i++) {
	   while ((q > 0) && (Pattern[q] != text[i])) {
		   q = Pi[q-1];
	   }
	   if (Pattern[q] == text[i]) {
		   q++;
	   }
	   if (q == m-1) {
		   numberOfOccurences++;
 #ifdef DEBUG
		   cout << "The String: " << text + i - m << endl;
 #endif
		   q = Pi[q-1];
	   }
   }

   return numberOfOccurences;
} // XxxKMPStrFind(const char * String, const char * SearchString)


/**
 * String search routine - make use of C library func strstr
 */
int
XxxCStrFind(const char * String, 
	    const char * SearchString)
{
  int numberOfOccurences = 0;
  const char* SubString = String;
  while((SubString = strstr(SubString, SearchString)) != NULL) {
    numberOfOccurences++;
    // move to the next possible place(the dumb way by ++) 
    // and break if reached the end.
    SubString++;
    if (SubString[0] == '\0') {
      break;
    }
  }

  return numberOfOccurences;

} // XxxCStrFind(const char * String, const char * SearchString)

/**
 * Naive String search routines
 */
int
XxxNaiveStrFind(const char * String, 
		const char * SearchString)
{
  
  int numberOfOccurences = 0;
  int StringSize = strlen(String);
  int SearchStringSize = strlen(SearchString);
  int offset = 0;
  
  for(offset = 0; offset + SearchStringSize <= StringSize; offset++) {
    if (strncmp(String + offset, SearchString, SearchStringSize) == 0) {
      numberOfOccurences++;
    }
  }

  return numberOfOccurences;
} // XxxNaiveStrFind(const char * String, const char * SearchString)