#include "enum.h"
#include "typelist.h"
//---------------------------------------------------------------
XxxErrCode XxxEnumADT::Register()
{
  char* typeInfoData = new char[serialSize];
  int size;

  *((int *)typeInfoData) = numEnumerations;

  char * curPtr = typeInfoData + sizeof(int);
  for(int i=0; i<numEnumerations; i++){
	size = strlen(enumerationNames[i]);
	*((int *)(curPtr)) = size;
	curPtr += sizeof(int);
	memcpy(curPtr, enumerationNames[i], size);
	curPtr += size;
  }

  // Serialize the information needed to recreate the Enum type into "enumData", and insert the data as a record
  // in the typelist file
  TypeInfo* enumInfo = new TypeInfo(TypeId, TypeName, "enum", serialSize, typeInfoData);

  delete[] typeInfoData;

  void *enumData = enumInfo->Serialize();
  int enumDataSize = enumInfo->DataSize();

  delete enumInfo;

  XxxShoreObjectId enumDataId;

  if(XxxShoreSM->InsertRecord(XxxShoreSM->defaultVolId, XxxShoreSM->typeFileId, enumData, enumDataSize, enumDataId) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot insert object into persistent file");
     return XXX_NOT_OK;
  }

  delete[] enumData;

  return XXX_OK;
}

//---------------------------------------------------------------
XxxErrCode XxxEnumADT::UnSerialize(const char* typeInfoData, int &enumNo, char** &enumNames){
	int size;

	enumNo = *((int *)typeInfoData);
	
	if(enumNo == 0){
		enumNames = NULL;
		return XXX_OK;
	}
	
	enumNames = new char*[enumNo];

	const char * curPtr = typeInfoData + sizeof(int);
	for(int i=0; i<enumNo; i++){
		size = *((int *)(curPtr));
		curPtr += sizeof(int);
		enumNames[i] = new char[size+1];
		memcpy(enumNames[i], curPtr, size);
		enumNames[i][size] = '\0';
		curPtr += size;
	}
	return XXX_OK;
}

//---------------------------------------------------------------
XxxErrCode XxxEnumADT::GetEnumList(RelDataSrc::KeyList *enumerations, int &enumNo, char** &enumNames, int &serialSize)
{
	  int i=0, size;
	  enumNo=0;
	  enumNames=NULL;
	  serialSize = 0;
	  RelDataSrc::KeyList *enumListPtr = enumerations, *copyPtr;

	  copyPtr = enumListPtr;
	  while (enumListPtr!=NULL) {
		  enumNo++;
		  enumListPtr = enumListPtr->Next;
	  }
	
	  serialSize = sizeof(int);
	  if(enumNo == 0){
		  return XXX_OK;
	  }

	  enumNames = new char*[enumNo];

	  enumListPtr = copyPtr;
	  while (enumListPtr!=NULL) {
		  i++;
		  size = strlen(enumListPtr->AttrName);
		  serialSize += sizeof(int) + size;
		  enumNames[enumNo - i]= new char[size+1];
		  strcpy(enumNames[enumNo-i], enumListPtr->AttrName);
		  enumNames[enumNo-i][size] = '\0';
		  enumListPtr = enumListPtr->Next;
	  }

	  return XXX_OK;
}

//---------------------------------------------------------------
XxxEnumADT::XxxEnumADT(XxxBasicType Id, char *enumTypeName, int enumNo, char** enumNames, int serialSize) :
  XxxSmallFixedADT(Id, enumTypeName, sizeof(EnumStruct))
{
	numEnumerations = enumNo;
	// Shallow copy, will be deleted in destructor
	enumerationNames = enumNames;

	this->serialSize = serialSize;
}

//---------------------------------------------------------------
XxxEnumADT::~XxxEnumADT(){
	if(enumerationNames){
		for(int i=0; i<numEnumerations; i++){
			delete[] enumerationNames[i];
		}
		delete[] enumerationNames;
	}
}

//---------------------------------------------------------------
//given raw byte, return the proper enumeration name.
XxxErrCode XxxEnumADT::WriteText(ostream& OutStream, const XxxADTValue* Object,
			       const XxxADTMetaInfo* MetaInfo) const
{
	OutStream << enumerationNames[((EnumStruct*)Object)->Val];
	return XXX_OK;
}

//---------------------------------------------------------------
//given enumeration name, return enumeration value
XxxErrCode XxxEnumADT::ReadText(istream& InStream, XxxADTValue* ObjectRef,
			      const XxxADTMetaInfo* MetaInfo) const
{
  char readIn[100];
  int i;
  EnumStruct *enumVal = (EnumStruct *)ObjectRef;
  InStream >> readIn;
  for(i=0;i<numEnumerations;i++){
	  if(strcmp(enumerationNames[i],readIn)==0){
		  enumVal->Val = i;
		  return XXX_OK;
	  }
  }
  return XXX_NOT_OK;
}

//---------------------------------------------------------------
//compares if 2 enum vals are equal
XxxBool  XxxEnumADT::Equals(const XxxADTValue* Obj1, const XxxADTMetaInfo* MI1, 
			  const XxxADTValue* Obj2, const XxxADTMetaInfo* MI2)  const 

{
	if ( ((EnumStruct *)Obj1)->Val == ((EnumStruct *)Obj2)->Val )
		return XXX_TRUE;
	else
		return XXX_FALSE;
}

//---------------------------------------------------------------
//checks if casting is possible
XxxBool XxxEnumADT::CastCheck(XxxBasicType CastType, const XxxADTMetaInfo* MetaInfo)
{
	return (CastType==XXX_INT);
}
  
//---------------------------------------------------------------
//casts from enum type to another type (integer)
XxxErrCode XxxEnumADT::CastExec(XxxBasicType CastType, 
			      const XxxADTMetaInfo* MetaInfo,
			      XxxADTValue* OldObject, 
			      XxxADTValue* NewObject)
{
	((EnumStruct*)NewObject)->Val = ((EnumStruct *)OldObject)->Val;
	return XXX_OK;
}

//---------------------------------------------------------------
XxxErrCode XxxEnumADT::CopyHToN(const XxxADTValue* Object, 
				   XxxADTValue *ResultObject,
				   const XxxADTMetaInfo* MetaInfo) const
{
  EnumStruct *En = (EnumStruct*)Object;
  EnumStruct *REn = (EnumStruct*)ResultObject;
  int LTemp = (int)htonl((unsigned long)(En->Val)); //int is 32 bits;
  REn->Val = LTemp;
  return XXX_OK;
}

//---------------------------------------------------------------
XxxErrCode XxxEnumADT::CopyNToH(const XxxADTValue* Object, 
				   XxxADTValue *ResultObject,
				   const XxxADTMetaInfo* MetaInfo) const
{
  EnumStruct *En = (EnumStruct*)Object;
  EnumStruct *REn = (EnumStruct*)ResultObject;
  int LTemp = (int)ntohl((unsigned long)(En->Val)); //int is 32 bits;
  REn->Val = LTemp;
  return XXX_OK;
}
