#ifndef XXX_ENUM_H
#define XXX_ENUM_H

#include "genadt.h"
#include "reldata.h"

/** 
 * XxxEnumADT is a data type which allows the user
 * to dynamically create his/her own enumerated type and 
 * use it in the database. The enumeration is represented
 * internally by int. Each creation of an enumerated type
 * will result in an instantiation of this class
 * @author Jyhtar Chong
 * @version January 28, 1998
 * @see
 */
class XxxEnumADT : public XxxSmallFixedADT {
 protected:
  class EnumStruct : public XxxADTValue {
	public:
      int Val;
  };
  // The number of enumeration attributes
  int numEnumerations;
  // The name of the enumeration attributes
  char **enumerationNames;
  // Size of the serialized data
  int serialSize;
 public:
  XxxEnumADT(XxxBasicType Id, char *enumTypeName, int enumNo, char** enumNames, int serialSize);

  virtual ~XxxEnumADT();

  // Register this enumeration type with server system
  virtual XxxErrCode Register();

  static XxxErrCode UnSerialize(const char* typeInfoData, int &enumNo, char** &enumNames);

  static XxxErrCode GetEnumList(RelDataSrc::KeyList *enumerations, int &enumNo, char** &enumNames, int &serialSize);
  //given raw byte, return the proper enumeration name.
  virtual XxxErrCode WriteText(ostream& OutStream, const XxxADTValue* Object,
			       const XxxADTMetaInfo* MetaInfo) const;
  
  //given enumeration name, return enumeration value
  virtual XxxErrCode ReadText(istream& InStream, XxxADTValue* ObjectRef,
			      const XxxADTMetaInfo* MetaInfo) const;
  //compares if 2 enum vals are equal
  virtual XxxBool  Equals(const XxxADTValue* Obj1, const XxxADTMetaInfo* MI1, 
			  const XxxADTValue* Obj2, const XxxADTMetaInfo* MI2)  const;

  //checks if casting is possible
  virtual XxxBool CastCheck(XxxBasicType CastType, 
			    const XxxADTMetaInfo* MetaInfo);
				
  //casts from enum type to another type (integer)
  virtual XxxErrCode CastExec(XxxBasicType CastType, 
			      const XxxADTMetaInfo* MetaInfo,
			      XxxADTValue* OldObject, 
			      XxxADTValue* NewObject);

  virtual XxxErrCode CopyHToN(const XxxADTValue* Object, 
				   XxxADTValue *ResultObject,
				   const XxxADTMetaInfo* MetaInfo) const;

  virtual XxxErrCode CopyNToH(const XxxADTValue* Object, 
				   XxxADTValue *ResultObject,
				   const XxxADTMetaInfo* MetaInfo) const;

};

#endif
