#include <stdio.h>
#include <string.h>
#include "client.h"
#include "genadt.h"
#include "funcexpr.h"
#include "httprequest.h"
#include "suspcomp.h"

/*****************************************************************************
 * Generic Large ADT
 *****************************************************************************/

const char* const XxxLargeADT::XxxLargeADTEnv::DISPLAY_TYPE="display_type";

//-----------------------------------------------------------------------------
XxxErrCode XxxLargeADT::WriteText(ostream& OutStream, 
				  const XxxADTValue* Object,
				  const XxxADTMetaInfo* MetaInfo) const
{
 XXX_SM_CHECK;
 XxxLargeADTEnv *Env = (XxxLargeADTEnv*)XXX_THISADTENV(TypeId);

 if (Env->DisplayType == XxxLargeADTEnv::NO_DISPLAY)
   return XXX_OK;
 else if (Env->DisplayType == XxxLargeADTEnv::CALLBACK_URL) {
   if (!SupportsCallback()) {
     OutStream << "Callback unsupported";
     return XXX_OK;
   }
   XxxShoreXactId ClientXactId;
   if (XxxShoreSM->XactToId(XXX_CLIENTINFO()->ThisThread->xct(),
			    ClientXactId) != XXX_OK) {
     XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot get client's transaction ID");
     return XXX_NOT_OK;
   }
   if (XxxHTTPRequest::WriteCallbackURL(XXX_CLIENTINFO()->UserId,
					ClientXactId, TypeId,
					Object, MetaInfo, 0, NULL,
					OutStream) != XXX_OK) {
     XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot create callback URL");
     return XXX_NOT_OK;
   }
 } else {
   XxxShoreVolumeId VolId;
   XxxShoreObjectId LocalId;
   if (getOid(Object, VolId, LocalId) != XXX_OK) {
     XXX_NOTE_ERROR(XXX_BADADT, "Bad Shore OID");
     return XXX_NOT_OK;
   }
   // write the data out to a file
   char TmpName[L_tmpnam];
   // PRD: CHG2
   tmpnam(TmpName);
   if (XxxShoreSM->WriteToOSFile(VolId, LocalId, TmpName)!= XXX_OK) {
     XXX_NOTE_ERROR(XXX_BADADT, "Error writing data to file");
     return XXX_NOT_OK;
   }
   if (Env->DisplayType == XxxLargeADTEnv::FILESYSTEM_DOWNLOAD) {
     OutStream << TmpName;
   } else if (Env->DisplayType == XxxLargeADTEnv::SERVERSIDE_DISPLAY) {
     if (showDataFile(TmpName, OutStream, Object, 
		      (XxxADTMetaInfo*)MetaInfo)!= XXX_OK) {
       XXX_NOTE_ERROR(XXX_BADADT, "Error displaying data file");
       return XXX_NOT_OK;
     }
     OutStream << "Displayed at Server";
     // Note that it is unsafe to delete the file, since the display could
     // be a background process which doesn't access the file immediately.
   } else if (Env->DisplayType == XxxLargeADTEnv::STREAM_CONTENTS) {
     ifstream inFile (TmpName,ios::in);
     char ch;
     OutStream.put('\"');
     while ( OutStream && inFile.get(ch) ) {
       // Praveen: this is a HACK! avoid putting in tabs
       if (ch == '\t') {
	 OutStream.put(' ');OutStream.put(' ');OutStream.put(' ');OutStream.put(' ');
	}
       else if (ch == '\"') {
		   OutStream.put('\'');
	   } else 
		   OutStream.put(ch);
	 }
     OutStream.put('\"');
   }
 }

 return XXX_OK;
}

XxxErrCode XxxLargeADT::CreateEnv( XxxADTEnv* &Env ) const
{
 Env = new XxxLargeADTEnv(XXX_THISENV().StorageVolume,
			  XXX_THISENV().StorageFile,
			  XXX_THISENV().TempFileId);
 return XXX_OK;
}

XxxErrCode XxxLargeADT::DeleteEnv( XxxADTEnv* Env ) const
{
 if (!Env) return XXX_OK;
 XxxLargeADTEnv* ADTEnv = (XxxLargeADTEnv*)Env;
 delete ADTEnv;
 return XXX_OK;
}


//-----------------------------------------------------------------------------
void XxxLargeADT::
XxxLargeADTEnv::Print(ostream& OStr) const
{
  XxxADTEnv::Print(OStr);
  OStr << DISPLAY_TYPE 
       << ": The method of display (";
  OStr << (int)(NO_DISPLAY) << " = none, ";
  OStr << (int)(CALLBACK_URL) << " = callback via URL, ";
  OStr << (int)(FILESYSTEM_DOWNLOAD) << " = filesystem download, ";
  OStr << (int)(STREAM_CONTENTS) << " = stream contents, ";
  OStr << (int)(SERVERSIDE_DISPLAY) << " = server-side display): ";
  OStr << (int)(DisplayType) << endl;
}

XxxErrCode XxxLargeADT::
XxxLargeADTEnv::SetValue(const char* ValName, void *NewVal)
{
 if (!strcmp(ValName, DISPLAY_TYPE)) {
   DisplayType = (XxxLargeADTDisplayType)((int)(*((char*)NewVal)-'0'));
   return XXX_OK;
  }
 return XxxADTEnv::SetValue(ValName, NewVal);
}

XxxErrCode XxxLargeADT::GetResultSize(const XxxADTValue *ObjectRef,
				      const XxxADTMetaInfo *MetaInfo,
				      const char *Message,
				      const int MessageSize,
				      int &ResultSize) const
{
  XxxShoreVolumeId VolId;
  XxxShoreObjectId LocalId;
  if (getOid(ObjectRef, VolId, LocalId) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADADT, "Bad Shore OID");
    return XXX_NOT_OK;
  }
  if (XxxShoreSM->GetObjectSize(VolId, LocalId,
				ResultSize) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADADT, "Cannot determine size of object");
    return XXX_NOT_OK;
  }
  return XXX_OK;
}

XxxErrCode XxxLargeADT::ServiceCallback(const XxxADTValue *ObjectRef,
					const XxxADTMetaInfo *MetaInfo,
					const char *Message,
					const int MessageSize,
					ostream &OStr) const 
{
  XxxShoreVolumeId VolId;
  XxxShoreObjectId LocalId;
  if (getOid(ObjectRef, VolId, LocalId) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADADT, "Bad Shore OID");
    return XXX_NOT_OK;
  }
  if (XxxShoreSM->WriteToStream(VolId, LocalId, OStr) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADADT, "Error writing object to stream");
    return XXX_NOT_OK;
  }
  OStr.flush();
  return XXX_OK;
}
