#ifndef XXX_GENADT_H
#define XXX_GENADT_H
#include <iostream.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "types.h"
#include "xxx_thread.h"  // for using XXX_NOTE_ERROR
#include "error.h"       // for using error codes
#include "generic.h"     // used to define generic Parse and Plan
#include "values.h"      // expr lists for arguments
#include "funcexpr.h"
#include "shoresm.h"
#include "optrules.h"
//SMostafa
#include "stream-mgr.h"
//EMostafa


/**
 GenADT defines the following  classes hierarchy:


  XxxADTClass: 
  --- XxxSmallADT:      Generic Small ADT
          --- XxxSmallFixedADT:
          --- XxxSmallVarADT:        
  --- XxxLargeADT:      Generic Large ADT
          --- XxxLargeFixedADT:
          --- XxxLargeVarADT:
   

-------------------------------------------------------------------------
  Name                       Size        Fixed       
------------------------------------------------------------------------
 XxxSmallFixedADT           Small       Yes          
 XxxSmallVarADT             Small       No           
 XxxLargeFixedADT           Large       Yes           
 XxxLargeVarADT             Large       No            
**/


/**
 *  Generic Small ADTs which are stored within the record
 */
class XxxSmallADT : public XxxADTClass {
public:
  XxxSmallADT(XxxBasicType Id, const char* Name, int ObjSize, XxxBool InField, XxxBool MaintainsCat) :
  XxxADTClass(Id, Name, ObjSize, InField, MaintainsCat) {}
  /**
   * Are the values "small" i.e., stored totally within the record?
   */
  virtual XxxBool IsSmall(const XxxADTMetaInfo*) const 
   { return XXX_TRUE; }
  //******************************************************************
  //SMostafa
  // This function must be implemented be each stream data type to 
  // indicate that the data type is stream
  //******************************************************************
  virtual XxxBool IsStream() const
    { return XXX_FALSE; }
  virtual XxxErrCode InitStream(XxxADTValue* ObjectRef)
    //				const XxxADTMetaInfo* MetaInfo)
    { return XXX_OK; }
  virtual XxxErrCode ReadStream(XxxADTValue* ObjectRef,
				//const XxxADTMetaInfo* MetaInfo,
				int *retVal,
				struct timeval *TS)
    { *retVal = STREAM_NO_VALUE; return XXX_OK; }
  virtual XxxErrCode CloseStream(XxxADTValue* ObjectRef)
    //				 const XxxADTMetaInfo* MetaInfo)
    { return XXX_OK; }
  //EMostafa
};

/**
 *  Small (stored within the record) fixed size ADTs
 */
class XxxSmallFixedADT : public XxxSmallADT {
public:
  XxxSmallFixedADT(XxxBasicType Id, const char* Name, int ObjSize,
		   XxxBool MaintainsCat = XXX_FALSE) :
  XxxSmallADT(Id, Name, ObjSize, XXX_TRUE, MaintainsCat) {}
};


/**
 *  Small (stored within the record) variable size ADTs
 */
class XxxSmallVarADT : public XxxSmallADT {
public:
  XxxSmallVarADT(XxxBasicType Id, const char* Name, 
		 int MaxSize,  XxxBool MaintainsCat = XXX_FALSE) :
  XxxSmallADT(Id, Name, MaxSize, XXX_FALSE, MaintainsCat) {}
};

/**
 * Generic large ADT.
 */
class XxxLargeADT : public XxxADTClass {
 protected:
  // called from WriteText()
  virtual XxxErrCode showDataFile(const char* FileName,
				  ostream& OutStream, 
				  const XxxADTValue* Object,
				  const XxxADTMetaInfo* MetaInfo) const = 0;

  // called from WriteText()
  virtual XxxErrCode getOid(const XxxADTValue*Object, 
			    XxxShoreVolumeId& VolId, 
			    XxxShoreObjectId& LocalId) const = 0;

 public:
  XxxLargeADT(XxxBasicType TypeId, const char* TypeName, int MaxSize ,
	      XxxBool InField, XxxBool MaintainsCat)  
       :    XxxADTClass(TypeId, TypeName, MaxSize, InField, MaintainsCat){}

  /**
   * Are the values "small" i.e., stored totally within the record?
   */
  virtual XxxBool IsSmall(const XxxADTMetaInfo*) const 
   { return XXX_FALSE; }
  //******************************************************************
  //SMostafa
  // This function must be implemented be each stream data type to 
  // indicate that the data type is stream
  //******************************************************************
  virtual XxxBool IsStream() const
  { return XXX_FALSE; }
  
  virtual XxxErrCode InitStream(XxxADTValue* ObjectRef)
    //				const XxxADTMetaInfo* MetaInfo)
    { return XXX_OK; }
  virtual XxxErrCode ReadStream(XxxADTValue* ObjectRef,
				//const XxxADTMetaInfo* MetaInfo,
			        int *retVal,
				struct timeval *TS)
    { *retVal = STREAM_NO_VALUE; return XXX_OK; }
  virtual XxxErrCode CloseStream(XxxADTValue* ObjectRef)
    //				 const XxxADTMetaInfo* MetaInfo)
    { return XXX_OK; }
  //EMostafa
  /**
   * Must return XXX_TRUE for ADTs which wish to implement callbacks.
   */
  virtual XxxBool SupportsCallback() const
  { return XXX_TRUE; }
  /**
   * This function will only be called on ADTs which support callbacks.
   * Given a handle on the result, this function must
   * determine the size (in bytes) of the response it will send in a 
   * call to the ServiceCallback() function.
   * @param ObjectRef The object for which a callback is being made
   * @param MetaInfo The meta info for that object
   * @param Message A binary message that was sent with the callback
   *                (or NULL if none was sent)
   * @param MessageSize The size of the Message array
   * @param ResultSize (out) The size of the object to be sent
   */
  virtual XxxErrCode GetResultSize(const XxxADTValue *ObjectRef,
				   const XxxADTMetaInfo *MetaInfo,
				   const char *Message,
				   const int MessageSize,
				   int &ResultSize) const;
  /**
   * This function will only be called on ADTs which support callbacks.
   * This is the sister-function to the Callback function in the Java
   * Client Library.  It communicates with the JCL and transfers the 
   * "full value" of a data object over the stream provided.
   * @param RequestBuffer The request message sent from the JCL
   * @param BufferSize The buffer's size
   * @param OStr The output stream to the JCL data type
   */
  virtual XxxErrCode ServiceCallback(const XxxADTValue *ObjectRef,
				     const XxxADTMetaInfo *MetaInfo,
				     const char *Message,
				     const int MessageSize,
				     ostream &OStr) const;

  virtual XxxErrCode WriteText(ostream& OutStream, 
			       const XxxADTValue* Object,
			       const XxxADTMetaInfo* MetaInfo) const;
  /**
   * Creates an ADT environment.
   */
  virtual XxxErrCode CreateEnv( XxxADTEnv* &Env ) const;
  /**
   * Reclaims an ADT environment.
   */
  virtual XxxErrCode DeleteEnv( XxxADTEnv* Env ) const;
  /**
   * Environment variable class
   */
  class XxxLargeADTEnv: public XxxADTEnv {
  public:
    /// Determines the display type
    enum XxxLargeADTDisplayTypeEnum {
      NO_DISPLAY,           // Do not display the item at all
      CALLBACK_URL,         // Write out a URL which can be used for callbacks
      FILESYSTEM_DOWNLOAD,  // Download to a local file and write its name
	  STREAM_CONTENTS,		// Download to a local file and stream its contents
      SERVERSIDE_DISPLAY    // Display the item at the server (for debugging)
    };
    /// A typedef for easy reference
    typedef enum XxxLargeADTDisplayTypeEnum XxxLargeADTDisplayType;
    /// The command referencing the next variable
    static const char* const DISPLAY_TYPE;
    /// The desired display type
    XxxLargeADTDisplayType DisplayType;


    /// The volume on which the large ADTs store information
    XxxShoreVolumeId StorageVolume;
    /// A file on the default volume which is persistent
    XxxShoreObjectId StorageFile;
    /// A file on the default volume which is cleared after all Xacts end
    XxxShoreObjectId TemporaryFile;

    XxxLargeADTEnv(XxxShoreVolumeId VolId, XxxShoreObjectId FileId,
		   XxxShoreObjectId TempFileId) :
      XxxADTEnv(), DisplayType(CALLBACK_URL), StorageVolume(VolId), 
      StorageFile(FileId), TemporaryFile(TempFileId)  { }
    virtual ~XxxLargeADTEnv() {}
    virtual void Print(ostream& OStr) const;
    virtual XxxErrCode SetValue(const char* ValName,
				void *NewVal);
  }; // class XxxLargeADTEnv

 };

/**
 * Large (stored outside record) fixed size ADT
 */
class XxxLargeFixedADT : public XxxLargeADT {
 public:
  XxxLargeFixedADT(XxxBasicType TypeId, const char* TypeName,  
		   int MaxTypeSize, XxxBool MaintainsCat = XXX_TRUE)
    : XxxLargeADT(TypeId, TypeName, MaxTypeSize, XXX_TRUE, MaintainsCat){}
  //SMostafa
  virtual XxxBool IsStream() const
    { return XXX_FALSE; }
  virtual XxxErrCode InitStream(XxxADTValue* ObjectRef)
    //				const XxxADTMetaInfo* MetaInfo)
    { return XXX_OK; }
  virtual XxxErrCode ReadStream(XxxADTValue* ObjectRef,
				//const XxxADTMetaInfo* MetaInfo,
				int *retVal,
				struct timeval *TS)
    { *retVal = STREAM_NO_VALUE; return XXX_OK; }
  virtual XxxErrCode CloseStream(XxxADTValue* ObjectRef)
    //				 const XxxADTMetaInfo* MetaInfo)
    { return XXX_OK; }
  //EMostafa
};
/**
 * Large (stored outside record) variable size ADT
 */
class XxxLargeVarADT : public XxxLargeADT {
 public:
  XxxLargeVarADT(XxxBasicType TypeId, const char* TypeName,  
		 int MaxTypeSize, XxxBool MaintainsCat = XXX_FALSE)
    : XxxLargeADT(TypeId, TypeName, MaxTypeSize, XXX_FALSE, MaintainsCat) {}
//SMostafa
  virtual XxxBool IsStream() const
    { return XXX_FALSE; }
  virtual XxxErrCode InitStream(XxxADTValue* ObjectRef)
    //				const XxxADTMetaInfo* MetaInfo)
    { return XXX_OK; }
  virtual XxxErrCode ReadStream(XxxADTValue* ObjectRef,
				//const XxxADTMetaInfo* MetaInfo,
				int *retVal,
				struct timeval *TS)
    { *retVal = STREAM_NO_VALUE; return XXX_OK; }
  virtual XxxErrCode CloseStream(XxxADTValue* ObjectRef)
    //				 const XxxADTMetaInfo* MetaInfo)
    { return XXX_OK; }
  //EMostafa
 };

#endif /* XXX_GENADT_H */
