#include "gentimeseries.h"
#include "optrules.h"
#include "memmap.h"
//#include "GetFile.h"


// Define static variables in XxxGenTimeSeriesADT 
XxxBasicType XxxGenTimeSeriesADT::GenTimeSeriesTypeId = XXX_UNKNOWN; 

// Constructor for GenTimeSeriesADT
XxxGenTimeSeriesADT::XxxGenTimeSeriesADT(XxxBasicType TypeId) 
 : XxxSmallVarADT (TypeId, "genTimeSeries", sizeof ( GenTimeSeriesStruct ))
{
  GenTimeSeriesTypeId = TypeId;

  NumMethods = 0;
  NumMethodSlots = NumMethods + XxxADTClass::NUM_SPARE_METHOD_SLOTS;
  MethArray = new FuncMethodInfo* [NumMethodSlots];

	maintainsMetaInfo = XXX_TRUE;
} 

XxxErrCode XxxGenTimeSeriesADT::makeMetaInfo(const MetaInfoHandle& MIHandle,
					 XxxADTMetaInfo*& MI) const 
{ 
  GenTimeSeriesMetaInfo* HandleInfo = (GenTimeSeriesMetaInfo* ) (MIHandle.Data); 
  MI = new GenTimeSeriesMetaInfo(*HandleInfo); 
  return XXX_OK; 
 } 


XxxErrCode XxxGenTimeSeriesADT::parseMetaInfo(const char* MInfoStr,
			           MetaInfoHandle& MIHandle) const 
{ 

  MIHandle.Size = sizeof(GenTimeSeriesMetaInfo); 
  //Bjarke, sep. 17 2001: added cast
  ASSERT(MIHandle.Size <= (unsigned int)XXX_MIHNDL_MAX); 
  GenTimeSeriesMetaInfo* TMI = (GenTimeSeriesMetaInfo*)(MIHandle.Data);
  TMI->Init();

  // Using the format of MInfoStr, get the metainfo
  // and save it in MIHandle
  istrstream InStr((char *)MInfoStr);
  
  InStr >> TMI->numDays;

  if( InStr.bad() )
	{
    XXX_NOTE_ERROR(XXX_BADADT, "Bad GenTimeSeries Meta Info");
    return XXX_NOT_OK;
  }
  return XXX_OK;
} 

XxxErrCode XxxGenTimeSeriesADT::WriteText(ostream& OutStream, 
				    const XxxADTValue* Object,
				    const XxxADTMetaInfo* MetaInfo) const
{
 GenTimeSeriesStruct *TSVal = (GenTimeSeriesStruct *)Object;
 //OutStream << "\"";
 //OutStream.write( TSVal->TimeData, 5000 );
 //OutStream << "\"";
 //OutStream.flush();

 return XXX_OK;
}

XxxErrCode XxxGenTimeSeriesADT::ReadText(istream& InStream, 
				   XxxADTValue* ObjectRef,
				   const XxxADTMetaInfo* MetaInfo) const
{
  GenTimeSeriesStruct *TSVal = (GenTimeSeriesStruct *)ObjectRef;
  ASSERT(TSVal);
	int MaxStrLen = 5000;
	char *data = new char[MaxStrLen];
	memset( data, 0, MaxStrLen );

  char CurrentChar = 0;
	char Delimiter = 0;
  int j = 0;

	while( Delimiter != '"' )
	{
	  InStream.get( Delimiter );
	  if (InStream.fail())
		return XXX_NOT_OK;
	}

  InStream.get(CurrentChar);
  if (InStream.fail())
    return XXX_NOT_OK;
  
  while( !InStream.eof() && (CurrentChar != Delimiter) && 
	 (CurrentChar != '\n') && (j<MaxStrLen-1) ) 
	{
		if( CurrentChar == ' ' ) 
		{
			*(data + j++) = CurrentChar;
			InStream.get(CurrentChar);
			while( !InStream.eof() && (CurrentChar != Delimiter) && 
				(CurrentChar != '\n') && (j<MaxStrLen-1) && (CurrentChar == ' ')) 
				InStream.get(CurrentChar);
		}
		else
		{
			*(data + j++) = CurrentChar;
			InStream.get(CurrentChar);
		}

    if (InStream.fail()) 
      return XXX_NOT_OK;
  }

  GenTimeSeriesMetaInfo* TSMInfo = (GenTimeSeriesMetaInfo*)MetaInfo;
	TSVal->numDays = TSMInfo->numDays;

	int k = 0, i = 0;
	for( i = 0; i < TSMInfo->numDays; i++ )
	{
		char *buf = new char[7];
		memset( buf, 0, 7 );
		int ptr = 0;
	  while( (k < j) && (*(data+k) != ' ') )
		{
			buf[ptr++] = *(data+k++);
		}
		k++;

		if( buf[0] == '@' )
			TSVal->dataOfPeriod[i] = 0;
		else TSVal->dataOfPeriod[i] = (int)(atof( buf ) * 1000);
		//cout << "dataOfPeriod[" << i << "] = " << TSVal->dataOfPeriod[i] << "\n";
	}

  if (InStream.bad()) 
	{
    XXX_NOTE_ERROR(XXX_BADPARSE, 
			"The proper syntax for a GenTimeSeries data is to enclose the data around quotes.");
    return XXX_NOT_OK;
  }
  return XXX_OK;
}


int XxxGenTimeSeriesADT::TypeSize( const XxxADTValue*, const XxxADTMetaInfo* MetaInfo ) const
{
  GenTimeSeriesMetaInfo* TSMInfo = (GenTimeSeriesMetaInfo*)MetaInfo;
  if (!TSMInfo) {
    XXX_NOTE_ERROR(XXX_BADADT, "No MInfo passed to time series serialize");
    return XXX_NOT_OK;
  }
  
  return TSMInfo->numDays * sizeof(int) + sizeof(int) + 4; 
}


int XxxGenTimeSeriesADT::TypeCopy( const XxxADTValue* SrcObject,
				XxxADTValue *ResultObject, const XxxADTMetaInfo* MetaInfo ) const
{
	int Size = TypeSize( SrcObject, MetaInfo );
	memcpy( (char*)ResultObject, (char *)SrcObject, Size );
	return Size;
}


XxxErrCode XxxGenTimeSeriesADT::
SerializeCopySize(const XxxADTMetaInfo* MetaInfo, unsigned int& Size) const
{
  GenTimeSeriesMetaInfo* TSMInfo = (GenTimeSeriesMetaInfo*)MetaInfo;
  if (!TSMInfo) {
    XXX_NOTE_ERROR(XXX_BADADT, "No MInfo passed to time series serialize");
    return XXX_NOT_OK;
  }
  
  Size = TSMInfo->numDays * sizeof(int) + sizeof(int) + 4; 

  return XXX_OK;
}

XxxErrCode XxxGenTimeSeriesADT::
SerializeCopy(const XxxADTValue* Object, const XxxADTMetaInfo* MetaInfo,
	      char* MemLocation) const
{
  GenTimeSeriesMetaInfo* TSMInfo = (GenTimeSeriesMetaInfo*)MetaInfo;
  if (!TSMInfo) 
	{
    XXX_NOTE_ERROR(XXX_BADADT, "No MInfo passed to GenTimeSeries serialize");
    return XXX_NOT_OK;
  }

  int CopySize = TSMInfo->numDays * sizeof(int) + sizeof(int);

  // copy the size of the array
  memcpy(MemLocation, &CopySize, sizeof(int));

	// copy the actual array to the memory location
  GenTimeSeriesStruct* GenTimeSeriesS = (GenTimeSeriesStruct *)Object; 

  memcpy(MemLocation + sizeof(int), (char *)GenTimeSeriesS, CopySize);  
  return XXX_OK;
}


XxxErrCode XxxGenTimeSeriesADT::
UnSerializeCopy(const char* MemLocation,
		const XxxADTMetaInfo* MetaInfo,
		XxxADTValue* Object) const
{
  XXX_NOTE_ERROR(XXX_BADADT, "Unserialize Copy not supported for GenTimeSeries");
  return XXX_NOT_OK;
}


XxxErrCode XxxGenTimeSeriesADT::CopyHToN(const XxxADTValue* Object, 
				     XxxADTValue* ResultObject,
				     const XxxADTMetaInfo* MetaInfo) const
{
  GenTimeSeriesMetaInfo* TMI = (GenTimeSeriesMetaInfo*)(MetaInfo);
  *ResultObject = *Object;
  GenTimeSeriesStruct *TSVal  = (GenTimeSeriesStruct *)Object;
  GenTimeSeriesStruct *RTSVal = (GenTimeSeriesStruct *)ResultObject;

	XxxNetworkInt(&(RTSVal->numDays), TSVal->numDays);
	for( int i = 0; i < TMI->numDays; i++ )
		XxxNetworkInt(&(RTSVal->dataOfPeriod[i]), TSVal->dataOfPeriod[i]);

  return XXX_OK;
}

XxxErrCode XxxGenTimeSeriesADT::CopyNToH(const XxxADTValue* Object, 
				     XxxADTValue* ResultObject,
				     const XxxADTMetaInfo* MetaInfo) const
{
  GenTimeSeriesMetaInfo* TMI = (GenTimeSeriesMetaInfo*)(MetaInfo);
  *ResultObject = *Object;
  GenTimeSeriesStruct *TSVal  = (GenTimeSeriesStruct *)Object;
  GenTimeSeriesStruct *RTSVal = (GenTimeSeriesStruct *)ResultObject;

	XxxHostInt(&(RTSVal->numDays), TSVal->numDays);
	for( int i = 0; i < TMI->numDays; i++ )
		XxxHostInt(&(RTSVal->dataOfPeriod[i]), TSVal->dataOfPeriod[i]);

  return XXX_OK;
}

 
