/*
 * geometry.h
 */

#ifndef GEOMETRY_H
#define GEOMETRY_H

#include <math.h>
#include <stdlib.h>

#define EPSILON     (double) 0.000001

#define iszero(A)               (fabs(A) <= EPSILON)
#define isequal(A,B)            (fabs((A) - (B)) <= EPSILON)
#define islt(A,B)               ((B) - (A) > EPSILON)
#define isle(A,B)               ((A) - (B) <= EPSILON)
#define isgt(A,B)               ((A) - (B) > EPSILON)
#define isge(A,B)               ((B) - (A) <= EPSILON)

#define HYPOT(A, B)             sqrt((A) * (A) + (B) * (B))

#define box_contains(b1,b2)     (box_in_box((b2),(b1)))
#define box_ccontains(b1,b2)    (box_iin_box((b2),(b1)))
#define box_contained(b1,b2)    (box_in_box((b1),(b2)))
#define box_ccontained(b1,b2)   (box_iin_box((b1),(b2)))

#define GEOM_OK 1;
#define GEOM_NOT_OK -1;

#include "types.h"

// Points:  (x, y)

class XxxPointStruct  { public:
  double x;  // x-coordinate
  double y;  // y-coordinate
};
typedef XxxPointStruct XXX_POINT;

// Circles: ((x,y), r)

class XxxCircleStruct { public:
  XXX_POINT center;
  double radius;
};
typedef XxxCircleStruct XXX_CIRCLE;

// Rectangles:

class XxxRectangleStruct  { public:
  double xl, yl;  // lower-left point
  double xh, yh;  // upper-right point
};
typedef XxxRectangleStruct XXX_BOX;

// Line Segments:

typedef struct {
  XXX_POINT p[2];
  double m;
} XxxLSegStruct, XXX_LSEG;

// Paths:

typedef struct {
  long length;
  short closed;
  long npts;
  XXX_POINT p[1];
} XxxPathStruct, XXX_PATH;

// Lines:

typedef struct {
  double A, B, C;
  double m;
} XxxLineStruct, XXX_LINE;

// Polygons:  ((X[], Y[]), n, mbr)

class XxxPolygonStruct { public:
  double *pts;  // pts are like (x[0], x[1],  .. x[npts-1], y[0], y[1], .. y[npts-1]))
  long npts;  // number of vertices in polygon
  XXX_BOX mbr;       // minimum bounding rectangle for polygon
};
typedef XxxPolygonStruct XXX_POLYGON;

/*
 * Functions:
 */


XXX_CIRCLE *circle_make(XXX_POINT *loc, double radius);
XXX_BOX *circle_MBR(XXX_CIRCLE *c);
int circle_equals(XXX_CIRCLE *c1, XXX_CIRCLE *c2);
XXX_BOX *box_make(double xl, double yl, double xh, double yh);
XXX_BOX *box_make(XXX_POINT *ll, XXX_POINT *ur);
int box_in_box(XXX_BOX *b1, XXX_BOX *b2);
int box_iin_box(XXX_BOX *b1, XXX_BOX *b2);
int box_equals(XXX_BOX *b1, XXX_BOX *b2);
XXX_POLYGON *box_2_poly(XXX_BOX *box);
int box_overlap(XXX_BOX *b1, XXX_BOX *b2);
int box_ooverlap(XXX_BOX *b1, XXX_BOX *b2);
XXX_POINT *point_make(double x, double y);
int point_equals(XXX_POINT *p1, XXX_POINT *p2);
int point_equals(XXX_POINT *p1, XXX_POINT p2);
XXX_BOX *point_2_box(XXX_POINT *p);
double point_dist(XXX_POINT *p1, XXX_POINT *p2);
int point_in_box(XXX_POINT *p, XXX_BOX *b);
int point_iin_box(XXX_POINT *p, XXX_BOX *b);
int point_on_box(XXX_POINT *p, XXX_BOX *b);
XXX_POLYGON *point_2_poly(XXX_POINT *pt);
int point_in_poly(XXX_POINT *pt, XXX_POLYGON *poly);
int point_iin_poly(XXX_POINT *pt, XXX_POLYGON *poly);
int point_on_poly(XXX_POINT *pt, XXX_POLYGON *poly);
void lseg_smake(XXX_LSEG *lseg, XXX_POINT *p1, XXX_POINT *p2);
XXX_LSEG *lseg_make(XXX_POINT *p1, XXX_POINT *p2);
int lseg_intersect(XXX_LSEG *l1, XXX_LSEG *l2);
int lseg_parallel(XXX_LSEG *l1, XXX_LSEG *l2);
XXX_POINT *lseg_intercept(XXX_LSEG *l1, XXX_LSEG *l2);
XXX_POLYGON *poly_make(long npts, double *xpts, double *ypts);
int poly_MBR(XXX_POLYGON *p);
int poly_MBR(XXX_POLYGON *p, XXX_BOX *b);
int poly_2_chars(XXX_POLYGON *p, char *s);
int poly_2_chars(XXX_POLYGON *p, const char *s);
int chars_2_poly(XXX_POLYGON *p, char *s);
int chars_2_poly(XXX_POLYGON *p, const char *s);
int poly_free(XXX_POLYGON *p);
int poly_equals(XXX_POLYGON *p1, XXX_POLYGON *p2);
int poly_in_circle(XXX_POLYGON *p, XXX_CIRCLE *c);
int poly_iin_circle(XXX_POLYGON *p, XXX_CIRCLE *c);
int poly_overlap(XXX_POLYGON *p1, XXX_POLYGON *p2);
int poly_ooverlap(XXX_POLYGON *p1, XXX_POLYGON *p2);
int poly_in_poly(XXX_POLYGON *p1, XXX_POLYGON *p2);
int poly_iin_poly(XXX_POLYGON *p1, XXX_POLYGON *p2);
double poly_size(XXX_POLYGON *p1);

/*
 * point_equals --
 *
 * Return 1 if the two XXX_POINTs are equal; 0 otherwise.
 */

inline int point_equals(XXX_POINT *p1, XXX_POINT *p2) { 
  return (isequal(p1->x, p2->x) && isequal(p1->y, p2->y));
}

/* Overloaded function */
inline int point_equals(XXX_POINT *p1, XXX_POINT p2) {
  return (isequal(p1->x, p2.x) && isequal(p1->y, p2.y));
}

/*
 * point_in_poly --
 *
 * Return 1 if the given XXX_POINT lies within the
 * XXX_POLYGON; otherwise 0.  This a relaxed check.
 * I.e. the point can lie on the boundary of the
 * polygon.
 */

inline int point_in_poly(XXX_POINT *pt, XXX_POLYGON *poly) {
  XXX_BOX b;
  int nI, nJ;                        /* Counters through the PATH points */
  int c = 0;                         /* Innocent until guilty            */
  int p = 0;
  double *xp, *yp;                   /* Pointers to the PATH points      */
  XXX_POINT * pAt;
  
  b.xl = b.xh = pt->x;
  b.yl = b.yh = pt->y;
  /*
   * If it's a candidate for being inside the
   * polygon.
   */
  if (box_in_box(&b, &poly->mbr)) {    
    xp = (double *) poly->pts;
    yp = (double *) (poly->pts + poly->npts);
    
    for (nI=0, nJ=poly->npts-1;((nI<poly->npts)&&!p); nJ = nI++) { /* Around points */
      if (((isle(xp[nI],pt->x)&&islt(pt->x,xp[nJ])) || ((isle(xp[nJ],pt->x)&&islt(pt->x,xp[nI]))))
	  &&((islt(pt->y,yp[nJ])&&islt(pt->y,yp[nI])) 
	     || (islt(pt->y,((yp[nJ]-yp[nI]) * (pt->x-xp[nI]) / (xp[nJ]-xp[nI])+yp[nI])))))
	c=!c;
      if (!p) {
	pAt = point_make(xp[nI], yp[nI]);
	p = point_equals(pt, pAt);
	delete pAt;
      }
      /* p=((!p)?(point_equals(pt, make_point( xp[nI], yp[nI] ))):(p)); */
    }
  }
  return(c||p);
}


/*
 * point_iin_poly --
 *
 * Return 1 if the XXX_POINT lies strictly within (does
 * not lie on the boundary) the given XXX_POLYGON; otherwise
 * 0.
 */

inline int point_iin_poly(XXX_POINT *pt, XXX_POLYGON *poly) {
/*

  Borrowed from polyfuncs.c at the S2K site at UW-Madison.

  Modified by Paul Brown, February, 1994, to get the point_in_poly
  better. Check the box boundary to exclude all by the likely cases,
  and then perform the complete point_in_bound_path algorhythm. 0(n)

  Note that we are excluding points on the perimiter. Only points deemed
  to be in and not on the perim (such points sometimes return an
  'in' status).

  For a detailed explaination of this algorhythm, refer to    

  "Point in Polygon Strategies", Eric Haines,
  Graphics Gems IV, ed. Paul Heckbert, Academic Press, p. 23-45.

*/

  XXX_BOX b;
  int nI, nJ;                       /* Counters through the PATH points */
  int c = 0;                        /* Innocent until guilty            */
  int p = 0;
  double *xp, *yp;                  /* Pointers to the PATH points      */
  
  b.xl = b.xh = pt->x;
  b.yl = b.yh = pt->y;
  /*
   * If it's a candidate for being inside the
   * polygon.
   */
  if (box_iin_box(&b, &poly->mbr))  {  
    xp = (double *) poly->pts;
    yp = (double *) (poly->pts + poly->npts);
    
    for (nI=0, nJ=poly->npts-1; ((nI<poly->npts)&&!p); nJ = nI++) {/* Around points */
      if ((((isle(xp[nI],pt->x)) && (islt(pt->x,xp[nJ]))) || 
	   ((isle(xp[nJ],pt->x)) && (islt(pt->x,xp[nI]))))
	  &&( ( (islt(pt->y,yp[nJ])) && (islt(pt->y,yp[nI])) ) 
	      || (islt(pt->y, ( (yp[nJ]-yp[nI]) * (pt->x-xp[nI]) / (xp[nJ]-xp[nI])+yp[nI]) )) ))
	c=!c;
    }
  }
  return(c&&!p);                                     /* If in and not on a   */
}                                                    /* perimiter point.     */



#endif // GEOMETRY_H
