#include "gisadt.h"
#include "math.h"
#include "BoundingBox.h"

XxxBasicType XxxPointADT::PointTypeId = XXX_UNKNOWN;

// Constructor for PointADT
XxxPointADT::XxxPointADT(XxxBasicType TypeId) 
 : XxxSmallFixedADT ( TypeId, "point",        
                      sizeof (PointStruct))
{
 PointTypeId = TypeId;                         
 int BoxId     = XxxBoxADT::BoxTypeId;
 int CircleId  = XxxCircleADT::CircleTypeId;
 int PolygonId = XxxPolygonADT::PolygonTypeId;
 // Register the methods for the point ADT
 NumMethods = 9; 
 MethArray = new FuncMethodInfo* [NumMethods]; 
//Bjarke, sep. 14 2001: added '&' to all calls to FuncMethodInfo
 MethArray[0] = new FuncMethodInfo( "x" , XXX_DOUBLE, 
  0 , 0 , (ExecuteFunc)(&PlanInfo::execX), 0 );
 MethArray[1] = new FuncMethodInfo( "y" , XXX_DOUBLE, 
  0 , 0 , (ExecuteFunc)(&PlanInfo::execY), 0 );
 MethArray[2] = new FuncMethodInfo( "make_box" , BoxId, 
  0 , 0 , (ExecuteFunc)(&PlanInfo::execMake_box), 1  , XXX_DOUBLE);
 MethArray[3] = new FuncMethodInfo( "make_circle" , CircleId, 
  0 , 0 , (ExecuteFunc)(&PlanInfo::execMake_circle), 1  , XXX_DOUBLE);
 MethArray[4] = new FuncMethodInfo( "inside" , XXX_INT, 
  0 , 0 , (ExecuteFunc)(&PlanInfo::execInsidep), 1  , PolygonId);
 MethArray[5] = new FuncMethodInfo( "inside" , XXX_INT, 
  0 , 0 , (ExecuteFunc)(&PlanInfo::execInsideb), 1  , BoxId);
 MethArray[6] = new FuncMethodInfo( "outside" , XXX_INT, 
  0 , 0 , (ExecuteFunc)(&PlanInfo::execOutsidep), 1 , PolygonId);
 MethArray[7] = new FuncMethodInfo( "outside" , XXX_INT, 
  0 , 0 , (ExecuteFunc)(&PlanInfo::execOutsideb), 1  , BoxId);
 MethArray[8] = new FuncMethodInfo( "equals" , XXX_INT, 
  0 , 0 , (ExecuteFunc)(&PlanInfo::execEquals), 1  , TypeId);

 // Register the methods for the point ADT
 NumAggrMethods = 0; 
 AggrMethArray = new AggrMethodInfo* [NumAggrMethods]; 
}



XxxErrCode XxxPointADT::ReadText(istream& InStream, 
				 XxxADTValue* ObjectRef, 
				 const XxxADTMetaInfo* MetaInfo) const 
{ 
/*
  long x, y;
  PointStruct p;
 */

  PointStruct *PS = (PointStruct *) ObjectRef;
  ASSERT((PS));
  XXX_POINT *point = &(PS->Point);
  
  InStream >> point->x >> point->y;

/*
  p.x = x;
  p.y = y;
  memcpy(ObjectRef, (char *)&p, sizeof(PointStruct));
 */

  return XXX_OK;

}


XxxErrCode XxxPointADT::WriteText(ostream& OutStream, 
				  const XxxADTValue* Object, 
				  const XxxADTMetaInfo* MetaInfo) const 
{ 
  PointStruct *PS = (PointStruct *) Object;
  ASSERT((PS));
  XXX_POINT *point = &(PS->Point);

  OutStream << "(" << point->x << ", " << point->y << ")";
  return XXX_OK; 
}


XxxBool  XxxPointADT::Equals(const XxxADTValue* Obj1, 
			     const XxxADTMetaInfo* MInfo1,
			     const XxxADTValue* Obj2,
			     const XxxADTMetaInfo* MInfo2) const
{
  PointStruct *PS1 = (PointStruct *) Obj1;
  ASSERT((PS1));
  XXX_POINT *p1 = &(PS1->Point);
  PointStruct *PS2 = (PointStruct *) Obj2;
  ASSERT((PS2));
  XXX_POINT *p2 = &(PS2->Point);

  if (point_equals(p1, p2)) return XXX_TRUE;
  else return XXX_FALSE;
}
  
// Index related functions.

//AUTHOR: Fabian Camargo 
//PURPOSE: Returns a bounding box for the point.
//NOTES: none


XxxBool 
XxxPointADT::GetBoundingBox(const XxxADTValue*& Object, XxxBoundingBox& Box) 
{
  PointStruct *PS = (PointStruct *) Object;
  ASSERT((PS));
  XXX_POINT *point = &(PS->Point);

  int dim = 2;
  double bb_box[4];
  
  bb_box[0] = point->x;
  bb_box[1] = point->y;
  bb_box[2] = point->x;
  bb_box[3] = point->y;
  
  XxxBoundingBox tmp(dim, XXX_DOUBLE, (void *) bb_box);
  Box = tmp;
  return XXX_TRUE;
}


//AUTHOR: Fabian Camargo DATE: 4/16/97
//PURPOSE: Retrieve the compare method for a given function
//and an array of arguments for that function.
//NOTES: This is index independent.

XxxErrCode
XxxPointADT::GetCompOp(const char* FuncName,
		       XxxCompareOp& CompOp, XxxBool& Success) const
{
 Success = XXX_TRUE;

 // get compare method (e.g overlaps) from the function name
 if (strcmp(FuncName,"inside")==0) {
   CompOp = XXX_IDX_INSIDE;
  }
 else if (strcmp(FuncName,"outside")==0) {
   CompOp = XXX_IDX_DISJOINT;
  }
 else if (strcmp(FuncName,"equals")==0) {
   CompOp = XXX_IDX_EQUALS;
  }
 else {
   CompOp = XXX_IDX_INVALID;
   Success = XXX_FALSE;
  }
 return XXX_OK;
}

XxxErrCode XxxPointADT::CopyHToN(const XxxADTValue* Object, 
				 XxxADTValue *ResultObject,
				 const XxxADTMetaInfo* MetaInfo) const
{
  XXX_POINT *point = &(((PointStruct *) Object)->Point);
  XXX_POINT *Rpoint = &(((PointStruct *) ResultObject)->Point);
  XxxNetworkValue(Rpoint, *point);
  return XXX_OK;
}

XxxErrCode XxxPointADT::CopyNToH(const XxxADTValue* Object, 
				 XxxADTValue *ResultObject,
				 const XxxADTMetaInfo* MetaInfo) const
{
  XXX_POINT *point = &(((PointStruct *) Object)->Point);
  XXX_POINT *Rpoint = &(((PointStruct *) ResultObject)->Point);
  XxxHostValue(Rpoint, *point);
  return XXX_OK;
}

