#ifndef XXX_POINT_H
#define XXX_POINT_H
#include "genadt.h"
#include "optrules.h"
#include "geometry.h"
#include "adtfunc.h"

XXX_EXTERN_CLASS_DECL(XxxBoundingBox);

/**
 * XxxPointADT:  
 * Basic ADT for handling any generic point data.  Points are
 * stored as fixed size objects within a record and simply
 * contain the x and y coordinates.  Currently-supported methods
 * include make_box(), make_circle(), inside(), outside(), x(), 
 * equals(), and y().
 *
 * @see geometry.h
 * @author Dave Koster dkoster@cs.cornell.edu
 * @version 6/17/97
 */
class XxxPointADT : public XxxSmallFixedADT {
  friend class XxxGISModule;
  public : 
  /// Point Type ID
    static XxxBasicType PointTypeId;
  protected : 

  class PointStruct: public XxxADTValue { public:
    XXX_POINT Point;
  };

  /// Plan structure for function over Point 
    class PlanInfo : public XxxFuncPlanInfo {
      friend class XxxPointADT;
      XxxErrCode execX(XxxValueEnv* Env, 
		       XxxADTValue *ReturnValue) const; 
      XxxErrCode execY(XxxValueEnv* Env, 
		       XxxADTValue *ReturnValue) const; 
      XxxErrCode execMake_box(XxxValueEnv* Env, 
			      XxxADTValue *ReturnValue) const; 
      XxxErrCode execMake_circle(XxxValueEnv* Env, 
				 XxxADTValue *ReturnValue) const; 
      XxxErrCode execInsidep(XxxValueEnv* Env, 
                             XxxADTValue *ReturnValue) const; 
      XxxErrCode execInsideb(XxxValueEnv* Env, 
                             XxxADTValue *ReturnValue) const; 
      XxxErrCode execOutsidep(XxxValueEnv* Env, 
			      XxxADTValue *ReturnValue) const; 
      XxxErrCode execOutsideb(XxxValueEnv* Env, 
			      XxxADTValue *ReturnValue) const; 
      XxxErrCode execEquals(XxxValueEnv* Env, 
			    XxxADTValue *ReturnValue) const; 
      public :
      PlanInfo(const XxxFuncParseInfo* FuncParse, int ArgCount)
	: XxxFuncPlanInfo(FuncParse , ArgCount) {}
     };
  
  virtual XxxErrCode methOptimize(XxxFuncParseInfo* FuncParse,
				  XxxValueExprPlan* OwnerPlan,
				  XxxValExprPlanList* ArgPlans,
				  XxxFuncPlanInfo*& FPlan) const{
				    FPlan = new PlanInfo(FuncParse, (ArgPlans?ArgPlans->Count():0));
				    return XXX_OK;
				   }

 public:
  /// Constructor for XxxPointADT
    XxxPointADT(XxxBasicType Id); 
  /// Destructor for XxxPointADT
    virtual ~XxxPointADT() {}
  /**
   * Display function for the txtclient protocol.
   *
   * @param OutStream Output stream to write to.
   * @param Object Buffer containing the PointStruct.
   * @param MetaInfo Meta info for points.
   * @return The status code.
   */
  virtual XxxErrCode WriteText(ostream& OutStream, 
			       const XxxADTValue* Object,
			       const XxxADTMetaInfo* MetaInfo) const;
  /**
   * Input function for loading points into tables.
   *
   * @param is Input stream from which to read the PointStruct.
   * @param ObjectRef Buffer in which to store the PointStruct.
   * @param MetaInfo Meta info for points.
   * @return The status code.
   */
  virtual XxxErrCode ReadText(istream& InStream, 
			      XxxADTValue* ObjectRef,
			      const XxxADTMetaInfo* MetaInfo) const;
  /// Equals method between points.
    
    virtual XxxBool  Equals(const XxxADTValue* Obj1, const XxxADTMetaInfo*,
			    const XxxADTValue* Obj2, const XxxADTMetaInfo*) const;
  // Methods for allowing indexes to be used with the point type.
  /**
   * Retrieves the bounding box of this spatial data type.  Used by
   * the indexing methods.
   */
  virtual XxxBool GetBoundingBox(const XxxADTValue*& Object,
				 XxxBoundingBox& Box);
  /**
   * Retrieves the compare method given a function name
   * and a array of arguments for that function.
   * This is used in index scans and is index independent.
   */
  virtual XxxErrCode GetCompOp(const char* FName, XxxCompareOp& CompOp, 
			       XxxBool& Success) const;

  virtual XxxErrCode CopyHToN(const XxxADTValue* SrcObject, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const;
  virtual XxxErrCode CopyNToH(const XxxADTValue* SrcObject, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const;
 };

#endif    /*  XXX_Point_H */ 
