/*
 * polygon.C
 */

#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include "BoundingBox.h"
#include "gisadt.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "memmap.h"

// static variables:

XxxBasicType XxxPolygonADT::PolygonTypeId = XXX_UNKNOWN;

XxxPolygonADT::XxxPolygonADT(XxxBasicType TypeId) :
  XxxLargeFixedADT(TypeId,
		   "polygon",
		   sizeof(PolygonStruct))
{
    PolygonTypeId = TypeId;
    
    NumMethods = 22;
    MethArray = new FuncMethodInfo*[NumMethods];
  
    MethArray[0] = new FuncMethodInfo("disjoint", XXX_INT, 0, 0,
				      (ExecuteFunc) &(PlanInfo::execDisjoint),
				      1, XxxBoxADT::BoxTypeId);
    MethArray[1] = new FuncMethodInfo("disjoint", XXX_INT, 0, 0,
				      (ExecuteFunc) &(PlanInfo::execDisjoint),
				      1, TypeId);
    MethArray[2] = new FuncMethodInfo("meets", XXX_INT, 0, 0,
				      (ExecuteFunc) &(PlanInfo::execMeets),
				      1, XxxBoxADT::BoxTypeId);
    MethArray[3] = new FuncMethodInfo("meets", XXX_INT, 0, 0,
				      (ExecuteFunc) &(PlanInfo::execMeets),
				      1, TypeId);
    MethArray[4] = new FuncMethodInfo("covers", XXX_INT, 0, 0,
				      (ExecuteFunc) &(PlanInfo::execCovers),
				      1, XxxPointADT::PointTypeId);
    MethArray[5] = new FuncMethodInfo("covers", XXX_INT, 0, 0,
				      (ExecuteFunc) &(PlanInfo::execCovers),
				      1, XxxBoxADT::BoxTypeId);
    MethArray[6] = new FuncMethodInfo("covers", XXX_INT, 0, 0,
				      (ExecuteFunc) &(PlanInfo::execCovers),
				      1, TypeId);
    MethArray[7] = new FuncMethodInfo("contains", XXX_BOOLEAN, 0, 0,
				      (ExecuteFunc) &(PlanInfo::execContains),
				      1, XxxPointADT::PointTypeId);
    MethArray[8] = new FuncMethodInfo("contains", XXX_BOOLEAN, 0, 0,
				      (ExecuteFunc) &(PlanInfo::execContains),
				      1, XxxBoxADT::BoxTypeId);
    MethArray[9] = new FuncMethodInfo("contains", XXX_BOOLEAN, 0, 0,
				      (ExecuteFunc) &(PlanInfo::execContains),
				      1, TypeId);
    MethArray[10] = new FuncMethodInfo("overlaps", XXX_INT, 0, 0,
				       (ExecuteFunc) &(PlanInfo::execOverlaps),
				       1, XxxBoxADT::BoxTypeId);
    MethArray[11] = new FuncMethodInfo("overlaps", XXX_INT, 0, 0,
				       (ExecuteFunc) &(PlanInfo::execOverlaps),
				       1, TypeId);
    MethArray[12] = new FuncMethodInfo("inside", XXX_INT, 0, 0,
				       (ExecuteFunc) &(PlanInfo::execInside),
				       1, XxxCircleADT::CircleTypeId);
    MethArray[13] = new FuncMethodInfo("inside", XXX_INT, 0, 0,
				       (ExecuteFunc) &(PlanInfo::execInside),
				       1, XxxBoxADT::BoxTypeId);
    MethArray[14] = new FuncMethodInfo("inside", XXX_INT, 0, 0,
				       (ExecuteFunc) &(PlanInfo::execInside),
				       1, TypeId);
    MethArray[15] = new FuncMethodInfo("covered_by", XXX_INT, 0, 0,
				       (ExecuteFunc) &(PlanInfo::execCoveredBy),
				       1, XxxCircleADT::CircleTypeId);
    MethArray[16] = new FuncMethodInfo("covered_by", XXX_INT, 0, 0,
				       (ExecuteFunc) &(PlanInfo::execCoveredBy),
				       1, XxxBoxADT::BoxTypeId);
    MethArray[17] = new FuncMethodInfo("covered_by", XXX_INT, 0, 0,
				       (ExecuteFunc) &(PlanInfo::execCoveredBy),
				       1, TypeId);
    MethArray[18] = new FuncMethodInfo("equals", XXX_INT, 0, 0,
				       (ExecuteFunc) &(PlanInfo::execEquals),
				       1, XxxBoxADT::BoxTypeId);
    MethArray[19] = new FuncMethodInfo("equals", XXX_INT, 0, 0,
				       (ExecuteFunc) &(PlanInfo::execEquals),
				       1, TypeId);
    MethArray[20] = new FuncMethodInfo("bounding_box", XxxBoxADT::BoxTypeId, 0, 0,
				       (ExecuteFunc) &(PlanInfo::execBoundingBox),
				       0);
    MethArray[21] = new FuncMethodInfo("size", XXX_DOUBLE, 0, 0,
				       (ExecuteFunc) &(PlanInfo::execSize),
				       0); 
}


XxxErrCode XxxPolygonADT::ReadText(istream& InStream, 
				   XxxADTValue* ObjectRef, 
				   const XxxADTMetaInfo* MetaInfo) const
{
  XXX_SM_CHECK;
  XxxErrCode rc = XXX_OK;
  
  char polygonName[XXX_DIRNAME_MAX] = "\0";
  char numPts[12] = "\0";
  char xl[20] = "\0";
  char yl[20] = "\0";
  char xh[20] = "\0";
  char yh[20] = "\0";

  InStream >> numPts >> xl >> yl >> xh >> yh >> polygonName;

  if (!InStream.good()) {
    XXX_NOTE_ERROR(XXX_BADADT, "Cannot read polygon data");
    return XXX_NOT_OK;
  }

  PolygonStruct* PS = (PolygonStruct*)ObjectRef;
  PS->polygon.npts = atoi(numPts);

  XxxShoreObjectId PolygonOID;
  XxxLargeADTEnv *PolygonEnv = (XxxLargeADTEnv*) XXX_THISADTENV(TypeId);
  // note: the constructor mmaps the file, while
  // the destructor unmaps the file
  XxxMemoryMap PolyMap(polygonName);
  void *MMapAddress = NULL;
  XXX_CALL_RET(PolyMap.GetHandle(MMapAddress), 
	       XXX_BADFILE, "Error in mmap");

  // the data format for polygon binary data is big-endian
#ifdef XXX_LITTLEENDIAN
  double *NewData = new double[PS->polygon.npts * 2];
  double *OldData = (double *)MMapAddress;
  for (int i = 0; i < PS->polygon.npts * 2; i++) {
    XxxHostDouble(NewData+i, OldData[i]);
  }
  MMapAddress = (void *)NewData;
#endif

  // insert a "record" of this object
  XXX_CALL_RET(XxxShoreSM->InsertRecord(PolygonEnv->StorageVolume, 
					PolygonEnv->StorageFile,
					MMapAddress,
					PolyMap.Size(),
					PolygonOID),
	       XXX_BADFILE, "Unable to store in SHORE");
  PS->PolygonId.VolId = PolygonEnv->StorageVolume;
  PS->PolygonId.LocalId = PolygonOID;
  
  PS->MemPolygon = NULL;
  
  PolygonMetaInfo* PMI = (PolygonMetaInfo*) (MetaInfo);
  
  sscanf(xl, "%lf", &(PS->polygon.mbr.xl));
  sscanf(yl, "%lf", &(PS->polygon.mbr.yl));
  sscanf(xh, "%lf", &(PS->polygon.mbr.xh));
  sscanf(yh, "%lf", &(PS->polygon.mbr.yh));

  //cout << PS->polygon.mbr.xl << PS->polygon.mbr.yl << PS->polygon.mbr.xh << PS->polygon.mbr.yh << endl;

  // Now compute the mbr of the polygon!

#ifdef XXX_LITTLEENDIAN
  if (NewData) delete [] NewData;
#endif

  return XXX_OK;
}


XxxErrCode XxxPolygonADT::WriteText(ostream& OutStream, 
				    const XxxADTValue* ObjectRef, 
				    const XxxADTMetaInfo* MetaInfo) const
{
  PolygonStruct PS;
  memcpy((char *)&PS, ObjectRef, sizeof(PolygonStruct));

  InMemPolygon* InMemPolygon1 = NULL;
  if (XxxPolygonADT::PlanInfo::getMemPolygon(&PS, InMemPolygon1) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADFUNC, "Error retrieving polygon!");
    return XXX_NOT_OK;
  }

  double *points = InMemPolygon1->polygon.pts;

  OutStream << " npts: " << PS.polygon.npts;
  OutStream << " mbr:  (" <<  PS.polygon.mbr.xl << "," << PS.polygon.mbr.yl << "),("
	    << PS.polygon.mbr.xh << "," << PS.polygon.mbr.yh << ")";

  OutStream << "[";
  for (int i = 0; i < PS.polygon.npts; i++) {
    OutStream << "(" << points[i] << ", " << points[i+PS.polygon.npts] << ")";
  }
  OutStream << "]";

  delete InMemPolygon1;

  return XXX_OK;
}


XxxErrCode
XxxPolygonADT::parseMetaInfo(const char* MetaInfoString,
                            MetaInfoHandle& MIHandle) const
{
 MIHandle.Size = sizeof(PolygonMetaInfo);
//Bjarke, sep. 17 2001: added cast
 ASSERT((MIHandle.Size <= (unsigned int)XXX_MIHNDL_MAX));

 istrstream InStr((char *)MetaInfoString);
 int AvgSizeKB;
 InStr >> AvgSizeKB;
 if (InStr.bad()) {
   XXX_NOTE_ERROR(XXX_BADADT, "Bad Polygon meta info");
   return XXX_NOT_OK;
  }
 PolygonMetaInfo* PMI = (PolygonMetaInfo*) (MIHandle.Data);
 PMI->AvgPolygonSize = AvgSizeKB;
 return XXX_OK;
}


XxxErrCode
XxxPolygonADT::makeMetaInfo(const MetaInfoHandle& MIHandle,
			    XxxADTMetaInfo* &MetaInfo) const
{
 PolygonMetaInfo* HandleInfo = (PolygonMetaInfo*)(MIHandle.Data);
 MetaInfo = new PolygonMetaInfo(*HandleInfo);
 return XXX_OK;
}

// Index related methods.

XxxBool 
XxxPolygonADT::GetBoundingBox(const XxxADTValue*& Object, 
			      XxxBoundingBox& Box) 
{
  PolygonStruct ps;
  int dim = 2;
  double bb_box[4];

  memcpy((char *)&ps, Object, sizeof(PolygonStruct));

  bb_box[0] = ps.polygon.mbr.xl;
  bb_box[1] = ps.polygon.mbr.yl;
  bb_box[2] = ps.polygon.mbr.xh;
  bb_box[3] = ps.polygon.mbr.yh;

  XxxBoundingBox tmp(dim, XXX_DOUBLE, (void *) bb_box);
  Box = tmp;
  return XXX_TRUE;
}

//AUTHOR: Fabian Camargo DATE: 4/16/97
//PURPOSE: Retrieve the compare method given a function name
//and a array of arguments for that function.
//NOTES: This is used in index scans and is index independent.

XxxErrCode 
XxxPolygonADT::GetCompOp(const char* FuncName,
			 XxxCompareOp& CompOp, XxxBool& Success) const
{
 Success = XXX_TRUE;
 
 // get compare method (e.g overlaps) from the function name
 
 if (strcmp(FuncName,"overlaps")==0) {
   CompOp = XXX_IDX_OVERLAPS;
  }
 else if (strcmp(FuncName,"inside")==0) {
   // inside goes to t_cover
   CompOp = XXX_IDX_INSIDE;
  }
 else if (strcmp(FuncName,"covers")==0) {
   // covers goes to t_inside
   CompOp = XXX_IDX_COVERS;
  }
 else if (strcmp(FuncName,"contains")==0) {
   CompOp = XXX_IDX_CONTAINS;
  }
 else if (strcmp(FuncName,"covered_by")==0) {
   CompOp = XXX_IDX_COVERED_BY;
  }
 else if (strcmp(FuncName,"disjoint")==0) {
   CompOp = XXX_IDX_DISJOINT;
  }
 else if (strcmp(FuncName,"meets")==0) {
   CompOp = XXX_IDX_MEETS;
  }
 else if (strcmp(FuncName,"equals")==0) {
   CompOp = XXX_IDX_EQUALS;
  }
 else {
   // Don't know how to get range from this function
   CompOp = XXX_IDX_INVALID;
   Success = XXX_FALSE;
  }
 return XXX_OK;
}

XxxErrCode XxxPolygonADT::CopyHToN(const XxxADTValue* Object, 
				   XxxADTValue *ResultObject,
				   const XxxADTMetaInfo* MetaInfo) const
{
  PolygonStruct* PS = (PolygonStruct*)Object;
  PolygonStruct* RPS = (PolygonStruct*)ResultObject;
  ASSERT ((RPS && PS));
  return RPS->CopyHToN(PS, (PolygonMetaInfo*)MetaInfo);
}

XxxErrCode XxxPolygonADT::CopyNToH(const XxxADTValue* Object, 
				   XxxADTValue *ResultObject,
				   const XxxADTMetaInfo* MetaInfo) const
{
  PolygonStruct* PS = (PolygonStruct*)Object;
  PolygonStruct* RPS = (PolygonStruct*)ResultObject;
  ASSERT ((RPS && PS));
  return RPS->CopyNToH(PS, (PolygonMetaInfo*)MetaInfo);
}


XxxErrCode XxxPolygonADT::PolygonStruct::
CopyHToN(const PolygonStruct* PS,
	 const PolygonMetaInfo* MetaInfo) 
{
  XxxNetworkValue(&(polygon), PS->polygon);
  memcpy((char *)&(PolygonId), (char *)&(PS->PolygonId),
	 sizeof(XxxShoreLogicalOID));
  MemPolygon = 0;
  return XXX_OK;
}

XxxErrCode XxxPolygonADT::PolygonStruct::
CopyNToH(const PolygonStruct* PS,
	 const PolygonMetaInfo* MetaInfo) 
{
  XxxHostValue(&(polygon), PS->polygon);
  memcpy((char *)&(PolygonId), (char *)&(PS->PolygonId),
	 sizeof(XxxShoreLogicalOID));
  MemPolygon = 0;
  return XXX_OK;
}
