/*
 * polygonmethods.C
 */

#include <stdio.h>
#include "polygon.h"
#include "point.h"
#include "box.h"
#include "circle.h"
#include "geometry.h"
#include "gisadt.h"
#include "BoundingBox.h"

XxxErrCode 
XxxPolygonADT::PlanInfo::execDisjoint(XxxValueEnv* Env, XxxADTValue *ReturnValue) const
{

  PolygonStruct *PS1 = (PolygonStruct *)(Env->OwnerValPtr);
  InMemPolygon* InMemPolygon1 = NULL;
  if (getMemPolygon(PS1, InMemPolygon1) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADFUNC, "Error retrieving polygon!");
    return XXX_NOT_OK;
  }
  XXX_POLYGON *poly = &(InMemPolygon1->polygon);

  // Argument is a polygon
  if (Env->ArgVals.Types[0] == TypeId) {
    PolygonStruct PS2;
    memcpy((char *)&PS2, (char *)(Env->ArgValsPtrs[0]), sizeof(PolygonStruct));
    InMemPolygon* InMemPolygon2 = NULL;
    if (getMemPolygon(&PS2, InMemPolygon2) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADFUNC, "Error retrieving polygon!");
      return XXX_NOT_OK;
    }
    ((XxxCoreADTClass::IntValue*)ReturnValue)->Value = (!poly_overlap(poly, &(InMemPolygon2->polygon)));
    delete InMemPolygon2;
  }
  // Argument is a box
  else if (Env->ArgVals.Types[0] == XxxBoxADT::BoxTypeId) {
    XxxGISModule::BoxStruct *BS = (XxxGISModule::BoxStruct*)Env->ArgValsPtrs[0];
    XXX_BOX* b = &(BS->Box);
    XXX_POLYGON *poly2;
    poly2 = box_2_poly(b);
    ((XxxCoreADTClass::IntValue*)ReturnValue)->Value = (!(poly_overlap(poly, poly2)));
    poly_free(poly2);
  }
  delete InMemPolygon1;
  return XXX_OK;
}


XxxErrCode 
XxxPolygonADT::PlanInfo::execMeets(XxxValueEnv* Env, XxxADTValue *ReturnValue) const
{

  PolygonStruct *PS1 = (PolygonStruct *)(Env->OwnerValPtr);
  InMemPolygon* InMemPolygon1 = NULL;
  if (getMemPolygon(PS1, InMemPolygon1) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADFUNC, "Error retrieving polygon!");
    return XXX_NOT_OK;
  }
  XXX_POLYGON *poly = &(InMemPolygon1->polygon);

  // Argument is a polygon
  if (Env->ArgVals.Types[0] == TypeId) {
    PolygonStruct PS2;
    memcpy((char *)&PS2, (char *)(Env->ArgValsPtrs[0]), sizeof(PolygonStruct));
    InMemPolygon* InMemPolygon2 = NULL;
    if (getMemPolygon(&PS2, InMemPolygon2) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADFUNC, "Error retrieving polygon!");
      return XXX_NOT_OK;
    }
    ((XxxCoreADTClass::IntValue*)ReturnValue)->Value = ((poly_overlap(poly, &(InMemPolygon2->polygon))) && !(poly_ooverlap(poly, &(InMemPolygon2->polygon))));
    delete InMemPolygon2;
  }
  // Argument is a box
  else if (Env->ArgVals.Types[0] == XxxBoxADT::BoxTypeId) {
    XxxGISModule::BoxStruct *BS = (XxxGISModule::BoxStruct*)Env->ArgValsPtrs[0];
    XXX_BOX* b = &(BS->Box);

    XXX_POLYGON *poly2;
    poly2 = box_2_poly(b);
    ((XxxCoreADTClass::IntValue*)ReturnValue)->Value = ((poly_overlap(poly, poly2)) && !(poly_ooverlap(poly, poly2)));
    poly_free(poly2);
  }
  delete InMemPolygon1;
  return XXX_OK;
}


XxxErrCode
XxxPolygonADT::PlanInfo::execCovers(XxxValueEnv* Env, XxxADTValue *ReturnValue) const
{

  PolygonStruct *PS1 = (PolygonStruct *)(Env->OwnerValPtr);
  InMemPolygon* InMemPolygon1 = NULL;
  if (getMemPolygon(PS1, InMemPolygon1) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADFUNC, "Error retrieving polygon!");
    return XXX_NOT_OK;
  }
  XXX_POLYGON *poly = &(InMemPolygon1->polygon);

  // Argument is a polygon
  if (Env->ArgVals.Types[0] == TypeId) {
    PolygonStruct PS2;
    memcpy((char *)&PS2, (char *)(Env->ArgValsPtrs[0]), sizeof(PolygonStruct));
    InMemPolygon* InMemPolygon2 = NULL;
    if (getMemPolygon(&PS2, InMemPolygon2) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADFUNC, "Error retrieving polygon!");
      return XXX_NOT_OK;
    }
    ((XxxCoreADTClass::IntValue*)ReturnValue)->Value = (poly_in_poly(&(InMemPolygon2->polygon), poly));
    delete InMemPolygon2;
  }
  // Argument is a box
  else if (Env->ArgVals.Types[0] == XxxBoxADT::BoxTypeId) {
    XxxGISModule::BoxStruct *BS = (XxxGISModule::BoxStruct*)Env->ArgValsPtrs[0];
    XXX_BOX* b = &(BS->Box);
    XXX_POLYGON *poly2;
    poly2 = box_2_poly(b);
    ((XxxCoreADTClass::IntValue*)ReturnValue)->Value = (poly_in_poly(poly2, poly));
    poly_free(poly2);
  }
  // Argument is a point
  else if (Env->ArgVals.Types[0] == XxxPointADT::PointTypeId) {
    XxxGISModule::PointStruct* PS = (XxxGISModule::PointStruct*)(Env->ArgValsPtrs[0]);
    XXX_POINT* p = &(PS->Point);
    ((XxxCoreADTClass::IntValue*)ReturnValue)->Value = (point_in_poly(p, poly));
  }
  delete InMemPolygon1;
  return XXX_OK;
}


XxxErrCode
XxxPolygonADT::PlanInfo::execContains(XxxValueEnv* Env, XxxADTValue *ReturnValue) const
{

  PolygonStruct *PS1 = (PolygonStruct *)(Env->OwnerValPtr);
  InMemPolygon* InMemPolygon1 = NULL;
  if (getMemPolygon(PS1, InMemPolygon1) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADFUNC, "Error retrieving polygon!");
    return XXX_NOT_OK;
  }
  XXX_POLYGON *poly = &(InMemPolygon1->polygon);

  // Argument is a polygon
  if (Env->ArgVals.Types[0] == TypeId) {
    PolygonStruct PS2;
    memcpy((char *)&PS2, (char *)(Env->ArgValsPtrs[0]), sizeof(PolygonStruct));
    InMemPolygon* InMemPolygon2 = NULL;
    if (getMemPolygon(&PS2, InMemPolygon2) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADFUNC, "Error retrieving polygon!");
      return XXX_NOT_OK;
    }
    ((XxxCoreADTClass::BoolValue*)ReturnValue)->Value = 
      (poly_iin_poly(&(InMemPolygon2->polygon), poly))?XXX_TRUE:XXX_FALSE;
    delete InMemPolygon2;
  }
  // Argument is a box
  else if (Env->ArgVals.Types[0] == XxxBoxADT::BoxTypeId) {
    XxxGISModule::BoxStruct *BS = (XxxGISModule::BoxStruct*)Env->ArgValsPtrs[0];
    XXX_BOX* b = &(BS->Box);

    XXX_POLYGON *poly2;
    poly2 = box_2_poly(b);
    ((XxxCoreADTClass::IntValue*)ReturnValue)->Value = 
      (poly_iin_poly(poly2, poly))?XXX_TRUE:XXX_FALSE;
    poly_free(poly2);
  }
  // Argument is a point
  else if (Env->ArgVals.Types[0] == XxxPointADT::PointTypeId) {
    XxxGISModule::PointStruct* PS = (XxxGISModule::PointStruct*)(Env->ArgValsPtrs[0]);
    XXX_POINT* p = &(PS->Point);
    ((XxxCoreADTClass::IntValue*)ReturnValue)->Value = 
      (point_iin_poly(p, poly))?XXX_TRUE:XXX_FALSE;
  }
  delete InMemPolygon1;
  return XXX_OK;
}


XxxErrCode 
XxxPolygonADT::PlanInfo::execOverlaps(XxxValueEnv* Env, XxxADTValue *ReturnValue) const
{

  PolygonStruct *PS1 = (PolygonStruct *)(Env->OwnerValPtr);
  InMemPolygon* InMemPolygon1 = NULL;
  if (getMemPolygon(PS1, InMemPolygon1) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADFUNC, "Error retrieving polygon!");
    return XXX_NOT_OK;
  }
  XXX_POLYGON *poly = &(InMemPolygon1->polygon);

  // Argument is a polygon
  if (Env->ArgVals.Types[0] == TypeId) {
    PolygonStruct PS2;
    memcpy((char *)&PS2, (char *)(Env->ArgValsPtrs[0]), sizeof(PolygonStruct));
    InMemPolygon* InMemPolygon2 = NULL;
    if (getMemPolygon(&PS2, InMemPolygon2) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADFUNC, "Error retrieving polygon!");
      return XXX_NOT_OK;
    }
    ((XxxCoreADTClass::IntValue*)ReturnValue)->Value = (poly_overlap(poly, &(InMemPolygon2->polygon)));
    delete InMemPolygon2;
  }
  // Argument is a box
  else if (Env->ArgVals.Types[0] == XxxBoxADT::BoxTypeId) {
    XxxGISModule::BoxStruct *BS = (XxxGISModule::BoxStruct*)Env->ArgValsPtrs[0];
    XXX_BOX* b = &(BS->Box);

    XXX_POLYGON *poly2;
    poly2 = box_2_poly(b);
    ((XxxCoreADTClass::IntValue*)ReturnValue)->Value = (poly_overlap(poly, poly2));
    poly_free(poly2);
  } 
  delete InMemPolygon1;
  return XXX_OK;
}


XxxErrCode 
XxxPolygonADT::PlanInfo::execInside(XxxValueEnv* Env, XxxADTValue *ReturnValue) const
{

  PolygonStruct *PS1 = (PolygonStruct *)(Env->OwnerValPtr);
  InMemPolygon* InMemPolygon1 = NULL;
  if (getMemPolygon(PS1, InMemPolygon1) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADFUNC, "Error retrieving polygon!");
    return XXX_NOT_OK;
  }
  XXX_POLYGON *poly = &(InMemPolygon1->polygon);

  // Argument is a polygon
  if (Env->ArgVals.Types[0] == TypeId) {
    PolygonStruct PS2;
    memcpy((char *)&PS2, (char *)(Env->ArgValsPtrs[0]), sizeof(PolygonStruct));
    InMemPolygon* InMemPolygon2 = NULL;
    if (getMemPolygon(&PS2, InMemPolygon2) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADFUNC, "Error retrieving polygon!");
      return XXX_NOT_OK;
    }
    ((XxxCoreADTClass::IntValue*)ReturnValue)->Value = (poly_iin_poly(poly, &(InMemPolygon2->polygon)));
    delete InMemPolygon2;
  }
  // Argument is a box
  else if (Env->ArgVals.Types[0] == XxxBoxADT::BoxTypeId) {
    XxxGISModule::BoxStruct *BS = (XxxGISModule::BoxStruct*)Env->ArgValsPtrs[0];
    XXX_BOX* b = &(BS->Box);

    XXX_POLYGON *poly2;
    poly2 = box_2_poly(b);
    ((XxxCoreADTClass::IntValue*)ReturnValue)->Value = (poly_iin_poly(poly, poly2));
    poly_free(poly2);
  } 
  // Argument is a circle
  else if (Env->ArgVals.Types[0] == XxxCircleADT::CircleTypeId) {
    XxxGISModule::CircleStruct *CS = (XxxGISModule::CircleStruct*)Env->ArgValsPtrs[0];
    XXX_CIRCLE* c = &(CS->Circle);
    ((XxxCoreADTClass::IntValue*)ReturnValue)->Value = (poly_iin_circle(poly, c));
  }
  delete InMemPolygon1;
  return XXX_OK;
}


XxxErrCode
XxxPolygonADT::PlanInfo::execCoveredBy(XxxValueEnv* Env, XxxADTValue *ReturnValue) const
{

  PolygonStruct *PS1 = (PolygonStruct *)(Env->OwnerValPtr);
  InMemPolygon* InMemPolygon1 = NULL;
  if (getMemPolygon(PS1, InMemPolygon1) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADFUNC, "Error retrieving polygon!");
    return XXX_NOT_OK;
  }
  XXX_POLYGON *poly = &(InMemPolygon1->polygon);

  // Argument is a polygon
  if (Env->ArgVals.Types[0] == TypeId) {
    PolygonStruct PS2;
    memcpy((char *)&PS2, (char *)(Env->ArgValsPtrs[0]), sizeof(PolygonStruct));
    InMemPolygon* InMemPolygon2 = NULL;
    if (getMemPolygon(&PS2, InMemPolygon2) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADFUNC, "Error retrieving polygon!");
      return XXX_NOT_OK;
    }
    ((XxxCoreADTClass::IntValue*)ReturnValue)->Value = (poly_in_poly(poly, &(InMemPolygon2->polygon)));
    delete InMemPolygon2;
  }
  // Argument is a box
  else if (Env->ArgVals.Types[0] == XxxBoxADT::BoxTypeId) {
    XxxGISModule::BoxStruct *BS = (XxxGISModule::BoxStruct*)Env->ArgValsPtrs[0];
    XXX_BOX* b = &(BS->Box);

    XXX_POLYGON *poly2;
    poly2 = box_2_poly(b);
    ((XxxCoreADTClass::IntValue*)ReturnValue)->Value = (poly_in_poly(poly, poly2));
    poly_free(poly2);
  } 
  // Argument is a circle
  else if (Env->ArgVals.Types[0] == XxxCircleADT::CircleTypeId) {
    XxxGISModule::CircleStruct *CS = (XxxGISModule::CircleStruct*)Env->ArgValsPtrs[0];
    XXX_CIRCLE* c = &(CS->Circle);
    ((XxxCoreADTClass::IntValue*)ReturnValue)->Value = (poly_in_circle(poly, c));
  }
  delete InMemPolygon1;
  return XXX_OK;
}


XxxErrCode 
XxxPolygonADT::PlanInfo::execEquals(XxxValueEnv* Env, XxxADTValue *ReturnValue) const
{

  PolygonStruct *PS1 = (PolygonStruct *)(Env->OwnerValPtr);
  InMemPolygon* InMemPolygon1 = NULL;
  if (getMemPolygon(PS1, InMemPolygon1) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADFUNC, "Error retrieving polygon!");
    return XXX_NOT_OK;
  }
  XXX_POLYGON *poly = &(InMemPolygon1->polygon);

  // Argument is a polygon
  if (Env->ArgVals.Types[0] == TypeId) {
    PolygonStruct PS2;
    memcpy((char *)&PS2, (char *)(Env->ArgValsPtrs[0]), sizeof(PolygonStruct));
    InMemPolygon* InMemPolygon2 = NULL;
    if (getMemPolygon(&PS2, InMemPolygon2) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADFUNC, "Error retrieving polygon!");
      return XXX_NOT_OK;
    }
    ((XxxCoreADTClass::IntValue*)ReturnValue)->Value = (poly_equals(poly, &(InMemPolygon2->polygon)));
    delete InMemPolygon2;
  }
  // Argument is a box
  else if (Env->ArgVals.Types[0] == XxxBoxADT::BoxTypeId) {
    XxxGISModule::BoxStruct *BS = (XxxGISModule::BoxStruct*)Env->ArgValsPtrs[0];
    XXX_BOX* b = &(BS->Box);

    XXX_POLYGON *poly2;
    poly2 = box_2_poly(b);
    ((XxxCoreADTClass::IntValue*)ReturnValue)->Value = (poly_equals(poly, poly2));
    poly_free(poly2);
  } 
  delete InMemPolygon1;
  return XXX_OK;
}


XxxErrCode
XxxPolygonADT::PlanInfo::execBoundingBox(XxxValueEnv* Env, XxxADTValue *ReturnValue) const
{

  PolygonStruct *PS1 = (PolygonStruct *)(Env->OwnerValPtr);
  InMemPolygon* InMemPolygon1 = NULL;
  if (getMemPolygon(PS1, InMemPolygon1) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADFUNC, "Error retrieving polygon!");
    return XXX_NOT_OK;
  }
  XXX_POLYGON *poly = &(InMemPolygon1->polygon);

  XXX_BOX *b;
  b = &(poly->mbr);
  XxxGISModule::BoxStruct *BS = (XxxGISModule::BoxStruct *)ReturnValue;

  memcpy((char*)&(BS->Box), (char *)b, sizeof(XXX_BOX));
  delete InMemPolygon1;
  return XXX_OK;
}


XxxErrCode XxxPolygonADT::
PlanInfo::execSize(XxxValueEnv* Env, XxxADTValue *ReturnValue) const
{
  PolygonStruct *PS1 = (PolygonStruct *)(Env->OwnerValPtr);
  InMemPolygon* InMemPolygon1 = NULL;
  XXX_CALL_RET(getMemPolygon(PS1, InMemPolygon1),
	       XXX_BADFUNC, "Error retrieving polygon!");

  XXX_POLYGON *poly = &(InMemPolygon1->polygon);
  XXX_BOX *b = &(poly->mbr);

  ((XxxCoreADTClass::DblValue*)ReturnValue)->Value = ((b->xh-b->xl) * (b->yh-b->yl));

  delete InMemPolygon1;
  return XXX_OK;
}


XxxErrCode XxxPolygonADT::
PlanInfo::getMemPolygon(PolygonStruct *PS, InMemPolygon*& Polygon) 
{
  if (PS->MemPolygon) {
    Polygon = PS->MemPolygon;
    return XXX_OK;
  }
  else {
    XxxShoreLogicalOID *PolygonId = &(PS->PolygonId);
    int PolygonSize = PS->polygon.npts * sizeof(double) * 2;
    Polygon = new InMemPolygon(PS->polygon);
    XXX_CALL_RET(XxxShoreSM->
		 GetObjectPreAlloc(PolygonId->VolId, PolygonId->LocalId,
				   (char *)(Polygon->polygon.pts),
				   PolygonSize, 0, PolygonSize),
		 XXX_BADADT, "Error reading polygon");
  }
  return XXX_OK;
}


XxxErrCode XxxPolygonADT::
PlanInfo::putMemPolygon(InMemPolygon *Polygon,
			PolygonStruct *PS,
			XxxShoreVolumeId &VolId,
			XxxShoreObjectId &FileId)
{
  PS->polygon = Polygon->polygon;

  int DataSize = Polygon->polygon.npts * sizeof(double) * 2;
  if (XxxShoreSM->InsertRecord(VolId, FileId,
			       (char *)(Polygon->polygon.pts),
			       DataSize,
			       PS->PolygonId.LocalId) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADFILE, "Unable to store in SHORE");
    return XXX_NOT_OK;
  }
  PS->PolygonId.VolId = VolId;
  PS->MemPolygon = NULL;
  delete Polygon;
  return XXX_OK;
}


XxxErrCode XxxPolygonADT::
PlanInfo::reclaimMemPolygon(PolygonStruct* PS)
{
 if (PS->MemPolygon) {
   delete PS->MemPolygon;
   PS->MemPolygon = NULL;
  }
 return XXX_OK;
}
