/*
 * rastermethods.C
 */

#include "raster.h"
#include "gisadt.h"

//-----------------------------------------------------------------------------
// execClip takes llx, lly, urx, ury as inputs in box form
XxxErrCode 
XxxRasterADT::PlanInfo::execClip(XxxValueEnv* Env,
				 XxxADTValue* ReturnValue) const
{

  RasterStruct *RS = (RasterStruct *)(Env->OwnerValPtr);
  RasterStruct* newRS = (RasterStruct*)(ReturnValue);

  // the args for the boxs
  XXX_BOX newBox;
  XxxGISModule::BoxStruct* BS = (XxxGISModule::BoxStruct*)(Env->ArgValsPtrs[0]);
  memcpy((char *)&newBox, (char *)&(BS->Box) , sizeof(XXX_BOX));

  int newLLX = (int)newBox.xl;
  int newLLY = (int)newBox.yl;
  int newURX = (int)newBox.xh;  
  int newURY = (int)newBox.yh;  

  // boundary in real coordinates of current raster:
  int oldLLX = (int)RS->geoArea.xl;
  int oldLLY = (int)RS->geoArea.yl;
  int oldURX = (int)RS->geoArea.xh;
  int oldURY = (int)RS->geoArea.yh;

  /*****
  // sanity check on clip box-- smaller or same size as current raster
  if ((oldLLX > newLLX) || (oldLLY > newLLY) 
      || (oldURX < newURX) ||  (oldURY < newURY)){
    XXX_NOTE_ERROR(XXX_BADFUNC, "Clip box out of bounds");
    return XXX_NOT_OK;
  }

  if ((newLLX > newURX) || (newLLY > newURY)) {
    XXX_NOTE_ERROR(XXX_BADFUNC, "Clip box empty");
    return XXX_NOT_OK;
  }
  ****/
  // sanity check on clip box-- smaller or same size as current raster
  if (oldLLX > newLLX) {
    newLLX = oldLLX;
    newBox.xl = newLLX;
   }
  if (oldLLY > newLLY) {
    newLLY = oldLLY;
    newBox.yl = newLLY;
   }
  if (oldURX < newURX) {
    newURX = oldURX;
    newBox.xh = newURX;
   }
  if (oldURY < newURY) {
    newURY = oldURY;
    newBox.yh = newURY;
   }
  if (newLLX > newURX) {
    int tmp = newLLX;
    newLLX = newURX;
    newURX = tmp;
    newBox.xl = newLLX;
    newBox.xh = newURX;
  }
  if (newLLY > newURY) {
    int tmp = newLLY;
    newLLY = newURY;
    newURY = tmp;
    newBox.yl = newLLY;
    newBox.yh = newURY;
  }


  InMemRaster* InMemRaster1 = NULL;
  // get raster
  // OLD: if (!getenv("XXXNORASTERCLIP")) {
  if (!XXX_THISENV().Options->getopt("XXXNORASTERCLIP")) {
    if (getMemRaster(RS, decompressOwner, InMemRaster1, &newBox) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADFUNC, "Error retrieving raster");
      return XXX_NOT_OK;
     }
   }
  else {
    if (getMemRaster(RS, decompressOwner, InMemRaster1) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADFUNC, "Error retrieving raster");
      return XXX_NOT_OK;
     }
   }

  // boundary in real coordinates of retrieved raster:
  oldLLX = (int)InMemRaster1->GeoArea.xl;
  oldLLY = (int)InMemRaster1->GeoArea.yl;
  oldURX = (int)InMemRaster1->GeoArea.xh;
  oldURY = (int)InMemRaster1->GeoArea.yh;
  // now determine the indices into the old raster.
  double xDist = floor(abs(oldLLX - oldURX) / InMemRaster1->Samples);
  unsigned int jStart = (int) floor(abs(oldLLX -  newLLX) / xDist);
  unsigned int jDone =  (int) floor(abs(oldLLX -  newURX) / xDist);
  
  double yDist = floor(abs(oldLLY - oldURY) / InMemRaster1->Lines);
  unsigned int iStart = (int) floor(abs(oldLLY -  newLLY) / yDist);
  unsigned int iDone =  (int) floor(abs(oldLLY -  newURY) / yDist);

  // Because of rounding errors we need to check the boundaries

  if (jDone-jStart >= InMemRaster1->Samples)
    jDone = jStart + InMemRaster1->Samples -1;

  if (iDone-iStart >= InMemRaster1->Lines)
    iDone = iStart + InMemRaster1->Lines - 1;

  // copy old data into new raster

  unsigned int NumLines = iDone-iStart+1;
  unsigned int NumSamples = jDone-jStart+1;

  XXX_BOX NewGeoArea;
  // boundary in real coordinates of current raster:
  NewGeoArea.xl = newLLX;
  NewGeoArea.yl = newLLY;
  NewGeoArea.xh = newURX;
  NewGeoArea.yh = newURY;

  InMemRaster* NewRaster = new InMemRaster(NumLines, NumSamples,
					   InMemRaster1->Freq, NewGeoArea,
					   InMemRaster1->resolution);

  unsigned int iCount = 0; 
  unsigned int i;

  for (i = iStart; i <= iDone; i++, iCount++){
    /***
    memcpy((char *)(((unsigned short *)(NewRaster->RasData))+ (iCount*NumSamples)),
	   (char *)(((unsigned short *)(InMemRaster1->RasData))+(i*InMemRaster1->Samples + jStart)),
	   NumSamples*sizeof(unsigned short));
    ***/
    int SUS = sizeof(unsigned short);
    memcpy(((char *)(NewRaster->RasData)) + (iCount*NumSamples*SUS),
	   ((char *)(InMemRaster1->RasData))+(i*InMemRaster1->Samples + jStart)*SUS,
	   NumSamples*SUS);
  }

  // store the newRS in shore:
  XxxLargeADTEnv 
    *RasterEnv = (XxxLargeADTEnv*) XXX_THISADTENV(RasterTypeId);
  if (putMemRaster(NewRaster, compressOutput, newRS, 
 		   RasterEnv->StorageVolume, 
 		   RasterEnv->TemporaryFile)) {
    XXX_NOTE_ERROR(XXX_BADFUNC, "Error writing final image");
    return XXX_NOT_OK;
  }

  delete InMemRaster1;
 
  return XXX_OK;
}

//-----------------------------------------------------------------------------
// Deres lowers the resolution of a raster by a given reduction factor s
// 		for every s scan lines, discard s - 1 of them
// 		for every s samples, discard s - 1 of them
XxxErrCode 
XxxRasterADT::PlanInfo::execDeres(XxxValueEnv* Env,
				  XxxADTValue *ReturnValue) const
{

  RasterStruct *RS = (RasterStruct *)(Env->OwnerValPtr);
  RasterStruct* newRS = (RasterStruct*)(ReturnValue);
  // function parameter
  int Reduction = ((XxxCoreADTClass::IntValue *)(Env->ArgValsPtrs[0]))->Value;
  // TODO:L2 dont we need to check if this value makes sense?
  // Also, i think we could allow double values instead with fractional
  // deresolution too -- it may be slightly trickier, but possible: Praveen
  
  // load raster into memory
  InMemRaster* InMemRaster1 = NULL;
  // get raster
  if (getMemRaster(RS, decompressOwner, InMemRaster1) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADFUNC, "Error retrieving raster");
    return XXX_NOT_OK;
  }
  
  // copy old data into new raster
  unsigned int NumLines = InMemRaster1->Lines/Reduction;
  unsigned int NumSamples = InMemRaster1->Samples/Reduction;
  InMemRaster* NewRaster = new InMemRaster(NumLines, NumSamples,
					   RS->freq, RS->geoArea,
					   Reduction);

  unsigned int i, j;
  unsigned int iCount = 0;
  unsigned int jCount = 0;
  for (i = 0; i < InMemRaster1->Lines; i++) {
    if (i%Reduction == 0) {
      jCount=0;
      for (j = 0; j < InMemRaster1->Samples; j++) {
	if (j%Reduction == 0) {
	  NewRaster->SetVal(iCount, jCount, InMemRaster1->GetVal(i, j));
	  jCount++;
	  if (jCount >= NumSamples) break;
	}
      }
      iCount++;
      if (iCount >= NumLines) break;
    }
  }
	
  // store the newRS in shore:
  XxxLargeADTEnv 
    *RasterEnv = (XxxLargeADTEnv*) XXX_THISADTENV(RasterTypeId);
  if (putMemRaster(NewRaster, compressOutput, newRS, 
 		   RasterEnv->StorageVolume, 
 		   RasterEnv->TemporaryFile)) {
      XXX_NOTE_ERROR(XXX_BADFUNC, "Error writing final image");
      return XXX_NOT_OK;
     }

  delete InMemRaster1;
 
  return XXX_OK;
}

/*
 * Methods for Raster Aggregate functions:
 * Init, Accum, Drop, Terminate, Reclaim
 */

XxxErrCode XxxRasterADT::
PlanInfo::ravgInit(XxxAggrOpHandle &Handle) const
{
  RasterAvgInfo *RAI = new RasterAvgInfo();
  Handle.State = RAI;
  return XXX_OK;
}

XxxErrCode XxxRasterADT::
PlanInfo::ravgAccum(XxxValueEnv* Env, XxxAggrOpHandle& Hndl) const
{
  RasterStruct *RS = (RasterStruct*)(Hndl.AccumVal);
  RasterStruct *newRS = (RasterStruct*)(Env->OwnerValPtr);
  RasterAvgInfo *RAI = (RasterAvgInfo *)(Hndl.State);
    
  if (Hndl.Count == 0) {
    RAI->Raster = new InMemRaster(newRS->lines, newRS->samples,
				  newRS->freq, newRS->geoArea,
				  newRS->resolution);
  }
  else {
    // TODO:L2 Check this eventually
    /*** Assume that the dimension is correct
      if (RS->lines != newRS->lines || RS->samples != newRS->samples || RS->geoArea.xl != newRS->geoArea.xl || RS->geoArea.yl != newRS->geoArea.yl || RS->geoArea.xh != newRS->geoArea.xh || RS->geoArea.yh != newRS->geoArea.yh) {
      XXX_NOTE_ERROR(XXX_BADFUNC, "raster incompatible for average");
      return XXX_NOT_OK;
      }
      ***/
  }
  InMemRaster* NewRaster;
  XXX_CALL_RET(getMemRaster(newRS, decompressOwner, NewRaster),
	       XXX_BADFUNC, "RAVG::Accum: Error retrieving new raster");
  
  for (int i = 0; i < newRS->lines; i++){
    for (int j = 0; j < newRS->samples; j++){
      // Since we're using short's. We might have problems with
      // overflow.
      float tmp2 = ((float)(RAI->Raster->GetVal(i,j)) * Hndl.Count) + 
	(float)(NewRaster->GetVal(i, j));
      RAI->Raster->SetVal(i, j, (unsigned short)(tmp2/(Hndl.Count+1)));
    }
  }
  delete NewRaster;
  return XXX_OK;
}

XxxErrCode XxxRasterADT::
PlanInfo::ravgTerm(XxxAggrOpHandle& Hndl, XxxADTValue* FinalVal) const
{
  RasterStruct *RS = (RasterStruct*)(Hndl.AccumVal);
  RasterStruct *FinalRS = (RasterStruct*)(FinalVal);
  RasterAvgInfo *RAI = (RasterAvgInfo *)(Hndl.State);
  
    // write the answer
  XxxLargeADTEnv *RasterEnv = 
    (XxxLargeADTEnv*) XXX_THISADTENV(RasterTypeId);
  if (putMemRaster(RAI->Raster, compressOutput, FinalRS,
		   RasterEnv->StorageVolume, 
		   RasterEnv->TemporaryFile) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADFUNC, "Error writing final image");
    return XXX_NOT_OK;
  }
  // make sure to nullify it, because the RAI destructor will delete
  // it, and putMemRaster also does it
  RAI->Raster = NULL;
  return XXX_OK;
}

//---------------------------------------------------------------------------
XxxErrCode XxxRasterADT::
PlanInfo::ranyInit(XxxAggrOpHandle &Handle) const
{
  return XXX_OK;
}

XxxErrCode XxxRasterADT::
PlanInfo::ranyAccum(XxxValueEnv* Env, XxxAggrOpHandle& Hndl) const
{
  RasterStruct *RS = (RasterStruct*)(Hndl.AccumVal);
  RasterStruct *newRS = (RasterStruct*)(Env->OwnerValPtr);
  if(Hndl.Count == 0) {
    // we only want to get the image the first time around
    InMemRaster* NewRaster;
    XXX_CALL_RET(getMemRaster(newRS, decompressOwner, NewRaster),
		 XXX_BADFUNC, "RANY::Accum: Error retrieving new raster");
    
    // allocate some memory for the new raster
    RS->MemRaster = new InMemRaster(newRS->lines, newRS->samples,
				    newRS->freq, newRS->geoArea);
    // copy the new (first) raster into the accumval
    for (int i=0; i < newRS->lines; i++) { 
      for (int j=0; j < newRS->samples; j++) {
	RS->MemRaster->SetVal(i, j, NewRaster->GetVal(i, j));
      }
    }
    delete NewRaster;
  } // if Hndl.Count == 0
  return XXX_OK;
}

XxxErrCode XxxRasterADT::
PlanInfo::ranyTerm(XxxAggrOpHandle& Hndl, XxxADTValue* FinalVal) const
{
  RasterStruct *RS = (RasterStruct*)(Hndl.AccumVal);
  RasterStruct *FinalRS = (RasterStruct*)(FinalVal);
  // write the answer
  XxxLargeADTEnv *RasterEnv = 
    (XxxLargeADTEnv*) XXX_THISADTENV(RasterTypeId);
  XXX_CALL_RET(putMemRaster(RS->MemRaster, compressOutput, FinalRS, 
			    RasterEnv->StorageVolume, 
			    RasterEnv->TemporaryFile),
	       XXX_BADFUNC, "Error writing final image");
  return XXX_OK;
}

//-----------------------------------------------------------------------------
// the returned raster is a superset of the desired box
XxxErrCode XxxRasterADT::
PlanInfo::getMemRaster(RasterStruct* RS,
		       XxxBool ToDecompress,
		       InMemRaster*& Raster,
		       const XXX_BOX* DesiredBox) const
{
 if (RS->MemRaster) {
   if (ToDecompress) {
     XXX_NOTE_ERROR(XXX_BADADT, "being asked to decompress main-mem raster");
     return XXX_NOT_OK;
    }
   Raster = RS->MemRaster;
   return XXX_OK;
 }
 else {
   if (!ToDecompress) {
     XXX_NOTE_ERROR(XXX_BADADT, "not being asked to decompress disk raster");
     return XXX_NOT_OK;
    }
   XxxShoreLogicalOID* RasterId = &(RS->RasterId);
   unsigned int Lines, Samples, RasSize;
   int Freq;
   int RasStart;
   int Resolution;
   XXX_BOX BoundingBox = RS->geoArea;
   if (!DesiredBox) {
     Lines  = RS->lines;
     Samples = RS->samples;
     RasStart = 0;
    }
   else {
     double yDist = floor(abs((int)RS->geoArea.yl - (int)RS->geoArea.yh) / RS->lines);
     int iStart = (int) floor(abs((int)RS->geoArea.yl -  (int)DesiredBox->yl) / yDist);
     int iDone = (int) floor(abs((int)RS->geoArea.yl -  (int)DesiredBox->yh) / yDist);
     if (iDone >= RS->lines)
       iDone = RS->lines-1;
     Lines = iDone-iStart+1;
     BoundingBox.yl = DesiredBox->yl;
     BoundingBox.yh = DesiredBox->yh;
     Samples = RS->samples; // this doesnt change
     RasStart = iStart*RS->samples * sizeof(unsigned short);
    }
   Freq = RS->freq;
   Resolution = RS->resolution;
   RasSize = Lines * Samples * sizeof(unsigned short);
   Raster = new InMemRaster(Lines, Samples, Freq, BoundingBox, Resolution);
   
   XXX_CALL_RET(XxxShoreSM->GetObjectPreAlloc(RasterId->VolId,
					      RasterId->LocalId, 
					      (char *)(Raster->RasData),
					      RasSize, RasStart, RasSize),
		XXX_BADADT, "Error reading raster from Shore");
  }
 return XXX_OK;
}

//---------------------------------------------------------------------------------
XxxErrCode XxxRasterADT::
PlanInfo::putMemRaster(InMemRaster* Raster, 
		       XxxBool ToCompress,
		       RasterStruct* RS,
		       XxxShoreVolumeId &VolId,
		       XxxShoreObjectId &FileId) const
{
 RS->freq = Raster->Freq;
 RS->samples = Raster->Samples;
 RS->lines  = Raster->Lines;
 RS->geoArea = Raster->GeoArea;
 RS->resolution = Raster->resolution;
 if (ToCompress) {
   int DataSize = Raster->Lines * Raster->Samples * sizeof(unsigned short);
   if (XxxShoreSM->InsertRecord(VolId, FileId,
				(char *)Raster->RasData,
				DataSize,
				RS->RasterId.LocalId) != XXX_OK) {
     XXX_NOTE_ERROR(XXX_BADFILE, "Unable to store in SHORE");
     return XXX_NOT_OK;
    }
   RS->RasterId.VolId = VolId;

   RS->MemRaster = NULL;
   delete Raster;
  }
 else {
   RS->MemRaster = Raster;
  }
 return XXX_OK;
}

//------------------------------------------------------------------------------
XxxErrCode XxxRasterADT::
PlanInfo::reclaimMemRaster(RasterStruct* RS,
			   XxxBool ToDecompress) const
{
 if (RS->MemRaster) {
   if (ToDecompress) {
     XXX_NOTE_ERROR(XXX_BADADT, "being asked to decompress main-mem raster");
     return XXX_NOT_OK;
    }
   delete RS->MemRaster;
   RS->MemRaster = NULL;
  }
 return XXX_OK;
}

