#include "struct.h"
#include "typelist.h"
#include "shoresm.h"
#include "releng.h"
#include "FlexLexer.h"
#include "adtfunc.h"

// define the static variables
XxxBasicType  XxxStructADT::StructTypeId = XXX_UNKNOWN;
//------------------------------------------------------------------------------

XxxErrCode 
StructPlanInfo::Evaluate(XxxValueEnv* ValueEnv,
						 XxxFuncMethodInfo* ThisMethInfo,
					     XxxADTValue *ReturnValue) const
{

    XxxStructADT::StructMetaInfo *OwnerCatInfo = 
		(XxxStructADT::StructMetaInfo *)ValueEnv->OwnerCatInfo;
	XxxRecordSchema *StructSchema = OwnerCatInfo->StructSchema;

	// Create record from serialized data
	unsigned int size = *(unsigned int*)(ValueEnv->OwnerValPtr);
	XxxRecord *Rec = new XxxRecord(size, (char *)(ValueEnv->OwnerValPtr)+sizeof(unsigned int));

	// Get the data from the column 'col' of this struct object
	int col = MethodIndex;
	XxxBasicType BType = StructSchema->Attrs[col].Type;
	XxxADTMetaInfo *MI = StructSchema->Attrs[col].TypeCatInfo;
	
	//Bjarke, sep. 13 2001: added const qualifier to variable
	const char *data;
	XXX_CALL_RET(Rec->GetField(col,StructSchema,data), XXX_BADRECORD,
		"Cannot get field of a record");

	XxxADTArray[BType]->TypeCopy((XxxADTValue*)data,ReturnValue,MI);
	return XXX_OK;
}

//------------------------------------------------------------------------------
// XxxStructADT::StructMetaInfo
//------------------------------------------------------------------------------
XxxStructADT::
StructMetaInfo::StructMetaInfo (XxxGenDataSrc::AttrTypeList *structAttr,
								XxxADTClass::MetaInfoHandle &MIHandle, 
								XxxErrCode& status)
{
	char **attrNames;
	XxxBasicType *attrTypes;
	XxxADTClass::MetaInfoHandle *attrMIHndl;
	
	// The string that contains the serialized data of the meta attribute list
	// The maximum size is set to 4000.
	char *metaRecord = new char[4000];
	// curPtr is the position in "metaRecord" where the next data field should be put.
	char *curPtr;

	numAttrs=0;
	attrNames=NULL;
	XxxGenDataSrc::AttrTypeList *attrListPtr = structAttr, *copyPtr;
	
	// Get the number of member fields for this struct ADT
	copyPtr = attrListPtr;
	while (attrListPtr!=NULL) {
	  numAttrs++;
	  attrListPtr = attrListPtr->Next;
	}
	
	*((int *)metaRecord) = numAttrs;
	curPtr = metaRecord + sizeof(int);
	
	// Allocate memory for the name, type, meta handle of each struct fields
	attrNames = new char*[numAttrs];
	attrTypes = new int[numAttrs];
	attrMIHndl = new XxxADTClass::MetaInfoHandle[numAttrs];
	
	attrListPtr = copyPtr;
	StructSchema = new XxxRecordSchema(numAttrs);
	for (int i=0; i<numAttrs; i++) {
	  int len;
	  
	  len = strlen(attrListPtr->Name);
	  attrNames[i] = new char[ len + 1];
	  strcpy(attrNames[i], attrListPtr->Name);
	  
	  *((int *)curPtr) = len;
	  curPtr += sizeof(int);
	  
	  memcpy(curPtr, attrListPtr->Name, len);
	  curPtr += len;
	  
	  attrTypes[i]=attrListPtr->Type;
	  *((int *)curPtr) = attrListPtr->Type;
	  curPtr += sizeof(int);
	  
	  attrMIHndl[i].Size = (attrListPtr->MIHndl).Size;
	  *((int *)curPtr) = (attrListPtr->MIHndl).Size;
	  curPtr += sizeof(int);
	  
	  if (attrMIHndl[i].Size > 0) {
	    memcpy(attrMIHndl[i].Data, (attrListPtr->MIHndl).Data, (attrListPtr->MIHndl).Size);
	    memcpy(curPtr, (attrListPtr->MIHndl).Data, (attrListPtr->MIHndl).Size);
	    curPtr += (attrListPtr->MIHndl).Size;
	    
	    XxxADTMetaInfo *MI;
	    XxxBool found=XXX_FALSE;
	    if (XxxADTArray[attrTypes[i]]->GetMetaInfo(attrNames[i],attrMIHndl[i],MI,found)!=XXX_OK){
	      cerr<<"error in XxxStructADT::XxxStructADT";
	      status = XXX_NOT_OK;
	      return;
	    }
	    
	    if (found==XXX_TRUE){
	      StructSchema->Attrs[i].Assign(attrNames[i], attrTypes[i], MI);
	    } else {
	      StructSchema->Attrs[i].Assign(attrNames[i], attrTypes[i], NULL);
	    }
	  } else {
	    StructSchema->Attrs[i].Assign(attrNames[i], attrTypes[i], NULL);
	  }
	  attrListPtr = attrListPtr->Next;
	}

	// Insert the metaRecord into shore, and store the handle in the MIHandle
	int metaRecLen = curPtr - metaRecord;
	XxxShoreObjectId metaRecordId;
	
	if (XxxShoreSM->InsertRecord(XxxShoreSM->defaultVolId,
				     XxxShoreSM->persistentFileId,
				     metaRecord, metaRecLen, metaRecordId) != XXX_OK) {
	  XXX_NOTE_ERROR(XXX_BADFILE, "Unable to store in SHORE");
	  status = XXX_NOT_OK;
	  return;
	}
	
	// Store the size of the actual meta record
	MIHandle.Size = metaRecLen;
	*((XxxShoreObjectId *)MIHandle.Data) = metaRecordId;
	
	delete [] attrTypes;
	//Bjarke, aug. 3: i declared variable in for stmt.
	//Bjarke, nov. 30: i renamed to ii, because of problems with VC++
	for (int ii = 0; ii < numAttrs; ii++) {
		delete [] attrNames[ii];
	}
	delete [] attrNames;
	delete [] attrMIHndl;
	delete [] metaRecord;

	StructSchema->ComputeStats();
	status = XXX_OK;
}

//------------------------------------------------------------------------------
XxxStructADT::StructMetaInfo::~StructMetaInfo() {
	if(StructSchema) {
		delete StructSchema;
	} else {
		XXX_NOTE_ERROR(XXX_BADADT, "StructSchema should not be NULL");
	}
}

//------------------------------------------------------------------------------
void XxxStructADT::StructMetaInfo::Print(ostream& OStr) const {
	OStr <<"(";
	for(int i=0; i<StructSchema->RecArity; i++){
		OStr << StructSchema->Attrs[i].Name;
		OStr << " "<<XxxADTArray[StructSchema->Attrs[i].Type]->TypeName;
		if(StructSchema->Attrs[i].TypeCatInfo){
			OStr << "(";
			StructSchema->Attrs[i].TypeCatInfo->Print(OStr);
			OStr << ")";
		}
		if(i!=(StructSchema->RecArity-1)){
			OStr << ",";
		}
	}
	OStr <<")";
}

//------------------------------------------------------------------------------
XxxBool XxxStructADT::StructMetaInfo::IsCompatible(const XxxADTMetaInfo* OtherMI) const {
	StructMetaInfo *mInfo = ((StructMetaInfo *)OtherMI);
	if(numAttrs != mInfo->numAttrs){
		return XXX_FALSE;
	}

	XxxRecordSchema *otherSchema = mInfo->StructSchema;
	// Compare the Record Schemas, and this could become a method of XxxRecordSchema
	if(StructSchema->RecArity != otherSchema->RecArity) {
		return XXX_FALSE;
	}

	for(int i=0; i<StructSchema->RecArity; i++){
		if(StructSchema->Attrs[i].Type != otherSchema->Attrs[i].Type){
			return XXX_FALSE;
		}

		if(strcmp(StructSchema->Attrs[i].Name, otherSchema->Attrs[i].Name)){
			return XXX_FALSE;
		}

		if(StructSchema->Attrs[i].TypeCatInfo && !StructSchema->Attrs[i].TypeCatInfo->IsCompatible(otherSchema->Attrs[i].TypeCatInfo)){
			return XXX_FALSE;
		}
	}

	return XXX_TRUE;
}

//------------------------------------------------------------------------------
XxxErrCode 
XxxStructADT::StructMetaInfo::SerializeCopySize(int &Size) const {
	Size = 0;

	// First add the size of integer, for RecArity
	Size += sizeof(int);

	for (int i=0; i<numAttrs; i++) {
		int len;

		len = strlen(StructSchema->Attrs[i].Name);

		// sizeof(int) for the length of Name, len for the Name chars
		Size += (sizeof(int) + len);

		// sizeof(int) for the typeID
		Size += sizeof(int);

		if(StructSchema->Attrs[i].TypeCatInfo){
			StructSchema->Attrs[i].TypeCatInfo->SerializeCopySize(len);
			// sizeof(int) for the length of Metainfo, len for the serialized meta
			// data
			Size += sizeof(int)+len;
		} else {
			Size += sizeof(int);
		}
	}

	return XXX_OK;
}

//------------------------------------------------------------------------------
XxxErrCode 
XxxStructADT::StructMetaInfo::SerializeCopy(char *Data) const {
	// curPtr is the position in "Data" where the next data field should be put.
	char *curPtr;

	*((int *)Data) = numAttrs;
	curPtr = Data + sizeof(int);

	for (int i=0; i<numAttrs; i++) {
		int len;

		len = strlen(StructSchema->Attrs[i].Name);

		*((int *)curPtr) = len;
		curPtr += sizeof(int);

		memcpy(curPtr, StructSchema->Attrs[i].Name, len);
		curPtr += len;

		*((int *)curPtr) = StructSchema->Attrs[i].Type;
		curPtr += sizeof(int);

		if(StructSchema->Attrs[i].TypeCatInfo){
			StructSchema->Attrs[i].TypeCatInfo->SerializeCopySize(len);
			*((int *)curPtr) = len;
			curPtr += sizeof(int);

			StructSchema->Attrs[i].TypeCatInfo->SerializeCopy(curPtr);
			curPtr += len;
		} else {
			*((int *)curPtr) = 0;
			curPtr += sizeof(int);
		}
	}
	return XXX_OK;
}


//------------------------------------------------------------------------------
// XxxStructADT
//------------------------------------------------------------------------------
XxxStructADT::XxxStructADT(XxxBasicType Id): 
XxxSmallVarADT(Id, "struct", 1, XXX_TRUE)
{
	StructTypeId=Id;
	
	NumMethods = 1;
	MethArray = new XxxFuncMethodInfo*[NumMethods];
	MethArray[0] = new XxxFuncMethodInfo("dummy", XXX_INT, 0,0,
		0, 0);

}
//------------------------------------------------------------------------------
void XxxStructADT::Print(ostream& OStr, const XxxADTMetaInfo* MetaInfo) const
{
  XxxStructADT::StructMetaInfo* metaInfo = (XxxStructADT::StructMetaInfo *)MetaInfo;

  if (metaInfo) metaInfo->Print(OStr);
  else OStr << "struct";
}
//------------------------------------------------------------------------------
XxxErrCode 
XxxStructADT::parseMetaString(const char* MetaInfoString,
							XxxGenDataSrc::AttrTypeList*& structAttr) const
{
 istrstream InStream ((char *)MetaInfoString, strlen(MetaInfoString));
 XxxRelationEngine::parseStruct PDS (&InStream, NULL);
 
 XxxRelationEngine* relEngine = (XxxRelationEngine *)XxxDBEngineArray[RelationEngineId];
 int ParseReturn = relEngine->sqlparse(PDS);
 if (PDS.LexerPtr) {
   delete ((yyFlexLexer*)(PDS.LexerPtr));
   PDS.LexerPtr = NULL;
  }
 if (ParseReturn) {
   REL_NOTE_ERROR (REL_BADQUERY, "Error in indexed expression");
   return XXX_NOT_OK;
 }

 structAttr = (XxxGenDataSrc::AttrTypeList *)PDS.OutArg;
 return XXX_OK;
}

//------------------------------------------------------------------------------
XxxErrCode 
XxxStructADT::parseMetaInfo(const char* MetaInfoString, MetaInfoHandle& MIHandle) const
{
	XxxGenDataSrc::AttrTypeList* structAttr;
	int len;
	if(!MetaInfoString || (len= strlen(MetaInfoString)) == 0){
		XXX_NOTE_ERROR(XXX_BADPARSE, "Meta String should not be null");
		return XXX_NOT_OK;
	}
	char* parseStr = new char[len+3];
	parseStr[0] = '{';
	parseStr[len+1] = '}';
	parseStr[len+2] = '\0';
	memcpy(parseStr+1, MetaInfoString, len);

	// In order to use parseMetaString function, wrap the meta string with {}
	XXX_CALL_RET(parseMetaString(parseStr, structAttr), XXX_BADPARSE, 
		"Meta string is invalid");

	XxxErrCode status;
	StructMetaInfo* MI = new StructMetaInfo(structAttr, MIHandle, status);
	
	if(status != XXX_OK){
		XXX_NOTE_ERROR(XXX_BADVALUE, "Meta info class is invalid");
		return XXX_NOT_OK;
	}

	delete[] parseStr;
	return XXX_OK;
}
//------------------------------------------------------------------------------
XxxErrCode XxxStructADT::makeMetaInfo(const MetaInfoHandle& MIHandle,
		             XxxADTMetaInfo*& structMI) const
{
  int metaRecLen = MIHandle.Size;
  XxxShoreObjectId metaRecordId = *((XxxShoreObjectId *)MIHandle.Data);
  
  char *metaRecord = new char[metaRecLen];
  XXX_CALL_RET(XxxShoreSM->GetObjectPreAlloc(XxxShoreSM->defaultVolId,
					     metaRecordId, metaRecord, metaRecLen, 0, metaRecLen) ,
	       XXX_SHOREERR, "Unable to store in SHORE");
  
  // curPtr is the position in "metaRecord" where the next data field should be put.
  char *curPtr;
  
  int numAttrs = *((int *)metaRecord);
  curPtr = metaRecord + sizeof(int);
  
  // Allocate memory for the name, type, meta handle of each struct fields
  char **attrNames = new char*[numAttrs];
  int *attrTypes = new int[numAttrs];
  XxxADTClass::MetaInfoHandle *attrMIHndl = new XxxADTClass::MetaInfoHandle[numAttrs];
  
  XxxRecordSchema *StructSchema = new XxxRecordSchema(numAttrs);
  for (int i=0; i<numAttrs; i++) {
    int len;
    
    len = *((int *)curPtr);
    curPtr += sizeof(int);
    
    attrNames[i] = new char[ len + 1];
    memcpy(attrNames[i], curPtr, len);
    attrNames[i][len] = '\0';
    curPtr += len;
    
    attrTypes[i] = *((int *)curPtr);
    curPtr += sizeof(int);
    
    attrMIHndl[i].Size = *((int *)curPtr);
    curPtr += sizeof(int);
    
    if (attrMIHndl[i].Size > 0) {
      memcpy(attrMIHndl[i].Data, curPtr, attrMIHndl[i].Size);
      curPtr += attrMIHndl[i].Size;
      
      XxxADTMetaInfo *MI;
      XxxBool found=XXX_FALSE;
      XXX_CALL_RET(XxxADTArray[attrTypes[i]]->GetMetaInfo(attrNames[i],attrMIHndl[i],MI,found),
		   XXX_BADHANDLE, "error in XxxStructADT::XxxStructADT");
      
      if (found==XXX_TRUE){
	StructSchema->Attrs[i].Assign(attrNames[i], attrTypes[i], MI);
      } else {
	StructSchema->Attrs[i].Assign(attrNames[i], attrTypes[i], NULL);
      }
    } else {
      StructSchema->Attrs[i].Assign(attrNames[i], attrTypes[i], NULL);
    }
  }
  delete [] attrTypes;
  //Bjarke, aug. 3: i declared variable in for stmt.
  //Bjarke, feb 1 2002: I changed name to j to accomodate VC++
  for (int j=0; j< numAttrs; j++){
    delete [] attrNames[j];
  }
  delete [] attrNames;
  delete [] attrMIHndl;
  delete [] metaRecord;
  
  StructSchema->ComputeStats();
  
  structMI = new StructMetaInfo(numAttrs, StructSchema);
  
  delete StructSchema;
  return XXX_OK;
}

//------------------------------------------------------------------------------
int XxxStructADT::MaxObjectSize(const XxxADTMetaInfo* MetaInfo) const
{
	XxxStructADT::StructMetaInfo* metaInfo = (XxxStructADT::StructMetaInfo *)MetaInfo;
	XxxRecordSchema *StructSchema = metaInfo->StructSchema;
	int sum=0, len;
	for(int i=0;i<metaInfo->numAttrs;i++){
		XxxADTMetaInfo *MI;
		MI = StructSchema->Attrs[i].TypeCatInfo;
		len=XxxADTArray[StructSchema->Attrs[i].Type]->MaxObjectSize(MI);
		if(!XxxADTArray[StructSchema->Attrs[i].Type]->StoreInField) {
			len+=sizeof(int);
		}
		XXX_WORD_ALIGN(len);
		sum+=len;
	}
	return (sum+sizeof(unsigned int)+sizeof(XxxADTValue)); 
}
//------------------------------------------------------------------------------
XxxErrCode XxxStructADT::WriteText(ostream& OutStream, const XxxADTValue* Object,
								   const XxxADTMetaInfo* MetaInfo) const
{
	XxxStructADT::StructMetaInfo* metaInfo = (XxxStructADT::StructMetaInfo *)MetaInfo;
	XxxRecordSchema *StructSchema = metaInfo->StructSchema;

	// Get the length of the record
	unsigned int size = *(unsigned int*)Object;
	XxxRecord *Rec = new XxxRecord(size,(char*)Object+sizeof(unsigned int));
	OutStream << '(';
	
	for (int i=0;i<metaInfo->numAttrs;i++){
		XxxADTMetaInfo *MI;
		MI = StructSchema->Attrs[i].TypeCatInfo;
		//Bjarke, sep. 13 2001: added const qualifier to variable
		const char *data;
		XXX_CALL_RET(Rec->GetField(i,StructSchema,data), XXX_BADRECORD,
			"Cannot get a field of a record");
		
		XxxADTClass *AttrClass = XxxADTArray[StructSchema->Attrs[i].Type];
		if (XxxCoreADTClass::IsCoreADT(StructSchema->Attrs[i].Type)) {
			XXX_CALL_RET(XxxCoreADTClass::SWriteText(StructSchema->Attrs[i].Type, 
				OutStream, (XxxADTValue*)data, MI), XXX_BADFILE, "Cannot write to file");
		}
		else {
			XXX_CALL_RET(XxxADTArray[StructSchema->Attrs[i].Type]->WriteText(OutStream,
				(XxxADTValue*)data,	MI), XXX_BADFILE, "Cannot write to file");
		}
		if(i<(metaInfo->numAttrs-1))
			OutStream<<',';
	}
	OutStream<<')';
	
	delete Rec;
	return XXX_OK;
}
//------------------------------------------------------------------------------
XxxErrCode XxxStructADT::ReadText(istream& InStream, 
				  XxxADTValue* ObjectRef,
				  const XxxADTMetaInfo* MetaInfo) const
{
  //not called directly. processing done in XxxStructCastExpression::StructCastPlan::Evaluate()
  return XXX_OK;
}
//------------------------------------------------------------------------------
XxxErrCode XxxStructADT::FuncTypeCheck(const char*FuncName,
				       const XxxValueExpression* Owner,
				       const XxxValueExprList *ArgList,
				       XxxBool IsAggr,
				       XxxBasicType& ReturnType,
				       XxxFuncParseInfo*& FuncParse) const
{
  XxxStructADT::StructMetaInfo* metaInfo = (XxxStructADT::StructMetaInfo *)(Owner->GenerateCatInfo());
  XxxRecordSchema *StructSchema = metaInfo->StructSchema;
  
  // Check whether the method name is one of the attribute name in the struct schema
  //Bjarke, aug. 3: i declared outside for stmt.
  int i;
  for (i=0;i<metaInfo->numAttrs;i++){
    if (strcmp(FuncName,StructSchema->Attrs[i].Name)==0)
      break;
  }
  if (i==metaInfo->numAttrs){
    XXX_NOTE_ERROR(XXX_BADEXPR, "No such attribute in struct");
    return XXX_NOT_OK;
  }
  
  ReturnType= StructSchema->Attrs[i].Type;
  FuncParse = new XxxFuncParseInfo(FuncName, i, NULL, TypeId, ReturnType, 
				   StructSchema->Attrs[i].TypeCatInfo);
  
  return XXX_OK;
}
//------------------------------------------------------------------------------
XxxErrCode XxxStructADT::FuncOptimize(XxxFuncParseInfo* FParse,
									  XxxValueExprPlan* OwnerPlan,
									  XxxValExprPlanList* ArgPlans,
									  XxxFuncPlanInfo*& FuncPlan) const
{ 
	XxxFuncParseInfo* FuncParse = (XxxFuncParseInfo*) FParse;
	XxxFuncPlanInfo* ThisPlanInfo;
	XXX_CALL_RET(methOptimize(FuncParse, OwnerPlan, ArgPlans, ThisPlanInfo) ,
		XXX_BADADT, "Error optimizing method");
	FuncPlan = ThisPlanInfo;
	return XXX_OK; 
}
//------------------------------------------------------------------------------
XxxErrCode XxxStructADT::CopyHToN(const XxxADTValue* Object, 
								  XxxADTValue *ResultObject,
								  const XxxADTMetaInfo* MetaInfo) const
{
	unsigned int size = *(unsigned int*)(((StructStruct*)Object)->Val);
	memcpy((char*)ResultObject,(char*)Object,size+sizeof(unsigned int));
	return XXX_OK;
}
//------------------------------------------------------------------------------
XxxErrCode XxxStructADT::CopyNToH(const XxxADTValue* Object, 
								  XxxADTValue *ResultObject,
								  const XxxADTMetaInfo* MetaInfo) const
{
	unsigned int size = *(unsigned int*)(((StructStruct*)Object)->Val);
	memcpy((char*)ResultObject,(char*)Object,size+sizeof(unsigned int));
	return XXX_OK;
}
