/** odbc\driver\handle.cpp
*	Author: zhiyuan chen, zhychen@cs.cornell.edu / Philippe Bonnet bonnet@cs.cornell.edu
*   Purpose: All implementation of the four kinds of handlers in ODBC driver
*	Date: 4/24/98
*/


#include <ctype.h>
#include <sys/types.h>
#include <sys/timeb.h>

/*#ifndef _WINSOCK2API_
#define _WINSOCK2API_
#define _WINSOCKAPI_   // Prevent inclusion of winsock.h in windows.h 
#endif
*/

//#include <windows.h>

#include "socket.h"
#include "handle.h"

//#include "Statement.h"
#include <ctype.h>
#include <iostream.h>
#include <string.h>


#ifdef XXX_WIN32
#include <strstrea.h>
#else
#include <strstream.h>
#endif

extern int dump(char * message); // for debugging. Dump useful information to c:\temp\log.txt.

extern unsigned long binatoi(char buf[4]); // a function convert int in network order to host order.

#include "odbcinst.h"
ConnectionHandle::ConnectionHandle()
{	DummyStarter = NULL;
predConnection = NULL;
count = 0;
sprintf(version,"00.00.0001");
errorGot = 0;
hMutex = CreateMutex( NULL, FALSE, NULL ); 
}

ConnectionHandle::~ConnectionHandle()
{	Env->count--;
dump("delete predCOnnection");
delete predConnection;
}

SQLRETURN ConnectionHandle::FreeConnectionHandle()
{	if (count > 0) // still some statement handle associated with it, error
return SQL_ERROR;
else {
	Env->count--;
	dump("delete connection handle");
	delete this;
	return SQL_SUCCESS;
}
}

void ConnectionHandle::PrintADTs()
{	int i;
for ( i = 0; i < predConnection->numADTs; i++)
{	cout << "ADT " << i << " Type" << predConnection->ADTs[i] << endl;
}
}

/** Suppose ServerName is the DSN name?
*/
int	SQL_API ConnectionHandle::Connect(SQLCHAR *		ServerName,
									  SQLSMALLINT	NameLength1,
									  SQLCHAR *		UserName,
									  SQLSMALLINT	NameLength2,
									  SQLCHAR *		Authentication,
									  SQLSMALLINT	NameLength3)
{	if ( (DummyStarter = new XxxSocketStarter()) == NULL )
return SQL_ERROR;
char serverName[50];
bool success;
dump("Enter COnnetionHandle,Connect");
strcpy(serverName,(char *)ServerName);
dump("Enter COnnect");
if (predConnection != NULL)
delete predConnection;
predConnection = NULL;
char port[64];
SQLGetPrivateProfileString((char *)ServerName,"Port",
						   "5090", port, 64,"ODBC.INI");
predConnection = 
new PredConnection((const char * )serverName, atoi(port),
				   (const char * )UserName, (const char * )Authentication,success,
				   TRUE, TRUE);
if (predConnection != NULL && success == TRUE) {
	PrintADTs();
	dump("Successful connected to server");
	return SQL_SUCCESS;
}
else 
return SQL_ERROR;
}

StatementHandle::StatementHandle(ConnectionHandle *	ptr)
{	WaitForSingleObject( ptr->hMutex, INFINITE );
ptr->count++;
connectionHandle = ptr; 
ReleaseMutex( ptr->hMutex );
columns = NULL;
schema = NULL;
record = NULL;
errorInfo = NULL;
special = 0;
result = 0;
errorGot = 0;
}

void StatementHandle::Free()
{ WaitForSingleObject( connectionHandle->hMutex, INFINITE );
connectionHandle->count--;
ReleaseMutex( connectionHandle->hMutex );

GlobalUnlock(GlobalHandle(App_Row_Desc));
GlobalFree(GlobalHandle(App_Row_Desc));
GlobalUnlock(GlobalHandle(App_Param_Desc));
GlobalFree(GlobalHandle(App_Param_Desc));
GlobalUnlock(GlobalHandle(Imp_Row_Desc));
GlobalFree(GlobalHandle(Imp_Row_Desc));
GlobalUnlock(GlobalHandle(Imp_Param_Desc));
GlobalFree(GlobalHandle(Imp_Param_Desc));
if ( columns) delete columns;
if ( errorInfo) delete errorInfo;
if (schema && schema->RecArity > 0) delete schema;
if ( record) delete record;
if (result) delete result;
}

StatementHandle::~StatementHandle()
{ // decrease ref counts of connection handle, 
  /*	if (columns)
		delete columns;
		if ( errorInfo)
		delete errorInfo;
		if ( schema)
		delete schema;
		if ( record)
		delete record;
		
	*/
	dump("Free it");
}




// send a request to server, called in SQLEXECDIRECT
SQLRETURN StatementHandle::SendRequest(SQLCHAR *		StatementText)
{	char * text;
char * ptr;
int len;
errorGot = 0;
// add 6 in length since we need SQL: and ; 
text = (char *)malloc(sizeof(char)* ((strlen((char *) (StatementText))+6)));
// filter out all tab, space, whitespace
for ( ptr = (char *)StatementText; *ptr != 0; ptr++) {
	if (*ptr != ' ' && *ptr != '\t' && *ptr != '\n')
		break;
}
sprintf(text,"SQL:%s",ptr);	
buf = new char[strlen(text)+1];
len = strlen(text) - 4;
strncpy(buf,ptr,len);
buf[len] = 0;
if (schema != NULL && schema->RecArity > 0 ) {
	delete schema;
	schema = NULL; // reclaim schema to make sure we read it when first fetch
}
WaitForSingleObject( connectionHandle->hMutex, INFINITE ); // must lock the connection handle until no result
this->connectionHandle->predConnection->SendRequest(text);
preProcessed = 0;
// decide the type of request
if((ptr = strstr(buf,"list")) && ptr == buf ) // result in lines
special = 1;
else if ((ptr = strstr(buf,"select")) && ptr == buf) // result in tables
special = 0;
else special = 2; // no result
done = 0;
return SQL_SUCCESS;
}

// columns is a link list to store the result comlumns binding
Columns::Columns()
{ 
}

Columns::~Columns()
{	
}

Columns::Columns(SQLUSMALLINT	ColumnNumber,SQLSMALLINT		TargetType,	SQLPOINTER		TargetValuePtr,SQLINTEGER		BufferLength,SQLINTEGER *		StrLen_or_IndPtr)
{	columnNumber = ColumnNumber;
targetType = TargetType;
targetValuePtr = TargetValuePtr;
bufferLength = BufferLength;
strLen_or_IndPtr = StrLen_or_IndPtr;
next = NULL;
}

SQLRETURN StatementHandle::AddColumns(SQLUSMALLINT	ColumnNumber,SQLSMALLINT		TargetType,	SQLPOINTER		TargetValuePtr,SQLINTEGER		BufferLength,SQLINTEGER *		StrLen_or_IndPtr)
{	// add a column
	Columns *cur, *next, *temp;
	
	if (columns == NULL) { // add a column to null
		if ( TargetValuePtr == NULL)
			return SQL_ERROR;
		else { // add the column
			columns = new Columns(ColumnNumber,TargetType,TargetValuePtr,BufferLength,StrLen_or_IndPtr);
			return SQL_SUCCESS;
		}
	}
	for ( cur = columns, next = cur; next != NULL && next->columnNumber < ColumnNumber; ) {
		if (cur == next) // first time
			next = cur->next;
		else {// otherwise, both cur and next march
			cur = next;
			if (cur != NULL)
				next = cur -> next;
		}
	}
	if ( next == NULL) { // add in tail
		if (TargetValuePtr == NULL)
			return SQL_SUCCESS;
		else {
			next = new Columns(ColumnNumber,TargetType,TargetValuePtr,BufferLength,StrLen_or_IndPtr);
			if ( cur != NULL)
				cur->next = next;
			return SQL_SUCCESS;
		}
	}
	else if ( cur == next) {// add in head
		if (TargetValuePtr == NULL)
			return SQL_SUCCESS;
		else {
			next = new Columns(ColumnNumber,TargetType,	TargetValuePtr,BufferLength,StrLen_or_IndPtr);
			if ( ColumnNumber != cur->columnNumber) // if not found, simply add
				next->next = cur;
			columns = next;
			return SQL_SUCCESS;
		}
	}
	else { // add in middle, after cur and before next
		if (next->columnNumber == ColumnNumber) { // if column found
			if (TargetValuePtr == NULL) {// unbind this column
				cur->next = next->next;
				next->next = NULL;
				delete next;
				return SQL_SUCCESS;
			}
			else { // column exists, update
				temp = next->next;
				next = new Columns(ColumnNumber,TargetType,TargetValuePtr,BufferLength,StrLen_or_IndPtr);
				next->next = temp;
				cur->next = next;
				return SQL_SUCCESS;
			}
		} 
		else { // when column not found
			if (TargetValuePtr == NULL)
				return SQL_SUCCESS;
			else {// add the column before next and after cur
				cur->next = new Columns(ColumnNumber,TargetType,TargetValuePtr,BufferLength,StrLen_or_IndPtr);
				cur->next->next = next;
				return SQL_SUCCESS;
			}
		}
	}
}

/**
* SQLBindCol Bind column for a statement handle
*/
SQLRETURN StatementHandle::SQLBindCol(SQLUSMALLINT	ColumnNumber,SQLSMALLINT	TargetType,SQLPOINTER	TargetValuePtr,SQLINTEGER	BufferLength,SQLINTEGER *	StrLen_or_IndPtr)
{	
	return AddColumns(ColumnNumber,TargetType,TargetValuePtr,BufferLength,StrLen_or_IndPtr);
}





/** 
* SQLFetch
* Fetch one row of data, and binding columns to buffers application provides
*/
SQLRETURN StatementHandle::SQLFetch()
{ 
	// First, if schema unread, read schema first
	SQLRETURN rc;
	if (schema == NULL) {
		rc = GetSchema();
		if (rc != SQL_SUCCESS)
			return SQL_ERROR;
	}
	// Next read next row of data into a XXXRecord
	if (special == 2) // no result
		return SQL_SUCCESS;
	dump("Get Record");
	rc = GetRecord();
	// no result, now may release lock on connnection handle
	
	if ( rc != SQL_SUCCESS) {
		ReleaseMutex( connectionHandle->hMutex );
		return rc;
	}
	dump("Get fields");
	rc = GetFields();
	return rc;
}

/** GetRecord
*	Read the record from socket and store it in buffer
*/
SQLRETURN StatementHandle::GetRecord()
{  
	char tempBuf[10];
	char Buf[1024];
	char info[256];
	int BytesRead;

	// socket init
	XxxSocketConnection* socket, * errorsocket;
	socket = connectionHandle->predConnection->socket;
	errorsocket = connectionHandle->predConnection->socket2;
	
	if (!socket) 
	{
		cerr << "Invalid connection" << endl;
		return SQL_ERROR;
	}
	
	// clear former record
	if (record != NULL)
		delete record;
	
	record = new Record;
	int SocketDesc = socket->socketDesc;


	// select for the result socket and error socket, 
	fd_set FAR  tempSockets;
	
	tempSockets.fd_count = 2;
	tempSockets.fd_array[0] = socket->socketDesc;
	tempSockets.fd_array[1] = errorsocket->socketDesc;
	struct timeval nowait = {0,0};
	
	if (select( 0, &tempSockets, NULL,NULL, &nowait) != SOCKET_ERROR) 
	{
		// some information returned, polling to see which one is ok
		preProcessed = 1;
		tempSockets.fd_count = 1;
		tempSockets.fd_array[0] = errorsocket->socketDesc;
		if (select (0,&tempSockets, NULL,NULL, &nowait ) == 1) 
		{ 
			if ( errorInfo != NULL)
				delete errorInfo;
			errorInfo = new XxxErrInfo();
			if (ReadErrors()==XXX_OK)
			{
				// Error messages
				dump("Error message exists!");
				
				// Now printout errors 
				//errorInfo->Print(,errorMessage);
				// close the request
				int SocketDesc = socket->socketDesc;
				if ((BytesRead = ::recv(SocketDesc, tempBuf, sizeof (unsigned long), 0)) <= 0) 
				{
					cerr << "Bad Read" << endl;
					schema = 0;
					return 0;
				}
				// judge whether it is end of records, meet \n (10) 0
				if (tempBuf[0]== 10 && tempBuf[1] == 0) 
				{
					// send back END , receive OK
					sprintf(tempBuf,"END");
					if (::send(SocketDesc, tempBuf, 3, 0) <= 0) 
					{
						cerr << "Error writing End" << endl;
						return SQL_ERROR;
					}
					// receive OK
					if ((BytesRead = ::recv(SocketDesc, tempBuf, 2*sizeof (unsigned char), 0)) <= 0) 
					{
						cerr << "Bad Read" << endl;
						return SQL_ERROR;
					}
					if (tempBuf[0]!= 'O' || tempBuf[1]!='K') 
					{
						cerr << "Bad END" << endl;
						return SQL_ERROR;
					}
					return SQL_ERROR;
				}
			} else 
			{
				dump("no error");
				
				// send end
				int SocketDesc = errorsocket->socketDesc;
				sprintf(tempBuf,"END");
				// end no error
				if (::send(SocketDesc, tempBuf, 3, 0) <= 0) 
				{
					dump( "Error writing End");
					return SQL_ERROR;
				}
				dump("sent END - socket");
				
			} 	
		}
		
		tempSockets.fd_count = 1;
		tempSockets.fd_array[0] = socket->socketDesc;
		if (select (0,&tempSockets, NULL,NULL, NULL) == 1) 
		{ 
			int readSize;
			if (special == 1)
			{
				readSize = 1024;
			}
			else
			{
				readSize = sizeof (unsigned long);
			}

			SocketDesc = socket->socketDesc;
			if ((BytesRead = ::recv(SocketDesc, Buf, readSize, 0)) <= 0) 
			{
				dump("bad read");
				return SQL_ERROR;
			}
			
			// judge whether it is end of records, meet \n (10) 0
			if (Buf[0]== 10 && Buf[1] == 0) 
			{
				dump("close request found");
				// end no data
				sprintf(tempBuf,"END");
				if (::send(SocketDesc, tempBuf, 3, 0) <= 0) 
				{
					dump( "Error writing End");
					return SQL_ERROR;
				}
				dump("sent END");
				// receive OK
				if ((BytesRead = ::recv(SocketDesc, tempBuf, 2*sizeof (unsigned char), 0)) <= 0) 
				{
					dump("Bad Read");
					return SQL_ERROR;
				}
				if (tempBuf[0]!= 'O' || tempBuf[1]!='K') 
				{
					dump("Bad END");
					return SQL_ERROR;
				}
				dump("SQL_NO_DATA");
				return SQL_NO_DATA;
			}
			
			if (special == 1)
			{
				record->DataArea = new char[BytesRead+1];
				Buf[BytesRead] = '\0';
				strcpy(record->DataArea, Buf);
				//ptr += strlen(record->DataArea)+1;
				record->size = strlen(record->DataArea);
			}
			else 
			{ // special == 0
				record->size = binatoi(Buf);
				// read full record
				record->DataArea = new char[record->size+1];
				if ((BytesRead = ::recv(SocketDesc, record->DataArea, record->size, 0)) <= 0) 
				{
					cerr << "Bad Read" << endl;
					return 0;
				}
			}
			return SQL_SUCCESS;
			
		} 
		
	}
	return SQL_ERROR;
}

/** TypeName
*  Return the typename of a specified type ID
*/
char * StatementHandle::TypeName(int TypeId)
{ 
	return connectionHandle->predConnection->ADTs[TypeId];
}

/** VarSize
* To tell whether the field is stored in field
*/
int StatementHandle::VarSize(int TypeId)
{ char * typeName;
typeName = connectionHandle->predConnection->ADTs[TypeId];
if ( strcmp(typeName, "string") == 0) // varsize
return 1;
else
return 0;
}

// Read network order double
double GetNetworkDouble(char *Value)
{	  // Reverse all 8 bytes
	char NValue[8];
	for (int i = 0; i < sizeof(double); i++)
		((char *)NValue)[i] = ((char *)Value)[sizeof(double) - 1 - i];
	return *((double *) NValue);
}

/** GetFields
*  Read Fields from result
*  and do type changes if the bounded c type is not the same as result type.
*  Currently, only support convert from basic types to string.
*/
SQLRETURN StatementHandle::GetFields()
{ 
	// First, get the field's offset
	Columns *cur;
	char *ptr, *temp,tmp[20];
	int i,j;
	int varsize = 0;
	int start, nextoffset, offset;
	// for each arity
	ptr = record->DataArea;
	if ( !ptr ) {
		
		return SQL_NO_DATA;
		
	}
	
	char * typeName;
	// extract when special == 1
	if ( special == 1) {
		cur = columns;
		strcpy((char *)cur->targetValuePtr,record->DataArea);
		*(cur->strLen_or_IndPtr) = record->size;
		return SQL_SUCCESS;
	}
	for ( i = 0, cur = columns; i < schema->RecArity && cur != NULL;i++ ) {
		// march the data area
		ptr = record->DataArea + schema->RecordOffsets[i];
		if (cur->columnNumber == i+1) { // need to get out this field
			// judge whether it is of variable size
			
			if(VarSize(schema->Attrs[i].Type)) {
				offset = ntohl(*((int *)ptr));
				typeName = TypeName(schema->Attrs[i].Type);
				if (strcmp(typeName,"string"))
					return SQL_ERROR;
				// calculate next offset
				nextoffset = record->size;
				for ( j = i; j < schema->RecArity; j++) {
					if (VarSize(schema->Attrs[j].Type) )
						nextoffset = ntohl(*((int *)(record->DataArea+schema->RecordOffsets[j])));
					else 
						nextoffset = schema->RecordOffsets[j];
					if (nextoffset > offset)
						break; // this is the next offset place
				}
				if (nextoffset <= offset) nextoffset = record->size;
				ptr = record->DataArea + offset;
				start = 0;
				for ( j = offset; j < nextoffset; j++) {// get string
					if ( *ptr != 0 && start == 0) {// begin now
						start = 1;
						temp = ptr;
						strcpy((char *)cur->targetValuePtr,temp);
						*(cur->strLen_or_IndPtr) = strlen(temp);
						break;
					}
					else
						ptr++;
				}
			} // of var size field
			else {// for fixed field, only treat int , double, boolean now
				// read the last 4 bytes for int and last 8 bytes for double
				typeName = TypeName(schema->Attrs[i].Type);
				if (strcmp(typeName,"int") == 0) { // int
					if ( i < schema->RecArity -1 )
						nextoffset = schema->RecordOffsets[i+1];
					else nextoffset = schema->RecordOffsets[i] + 8; // default value
					offset = nextoffset - 4;
					int tmpint;
					tmpint = ntohl(*((int *)(record->DataArea+offset)));
					// check if the result type if of SQL_INTEGER or string.
					switch (cur->targetType) {
					case SQL_C_CHAR:
						sprintf(tmp,"%d",tmpint);
						strcpy((char *)cur->targetValuePtr,tmp);
						*(cur->strLen_or_IndPtr) = strlen(tmp);
						break;
					default:					
						*(( int *)(cur->targetValuePtr)) = tmpint;
						*(cur->strLen_or_IndPtr) = 4;
					}
					
				}
				if (strcmp(typeName,"double") == 0) { // double
					if ( i < schema->RecArity -1 )
						nextoffset = schema->RecordOffsets[i+1];
					else nextoffset = schema->RecordOffsets[i] + 16; // default value, not quite sure
					offset = nextoffset - 8;
					// convert double from network order
					temp = record->DataArea+offset;
					double tmpdouble;
					tmpdouble = GetNetworkDouble(temp);
					// check if the result type if of SQL_INTEGER or string.
					switch (cur->targetType) {
					case SQL_C_CHAR:
						sprintf(tmp,"%f",tmpdouble);
						strcpy((char *)cur->targetValuePtr,tmp);
						*(cur->strLen_or_IndPtr) = strlen(tmp);
						break;
					default:					
						*((double *) cur->targetValuePtr) = GetNetworkDouble(temp);
						*(cur->strLen_or_IndPtr) = 8;
					}
					
				}
				if (!strcmp(typeName,"boolean")) { //boolean
					// not known whether supported or not
					return SQL_ERROR;
				}
			}
			cur = cur->next;
		}
	}
	return SQL_SUCCESS;
}

/** 
* ReadErrors
*  Read Errors from the socket, store in statement handle
*/
SQLRETURN StatementHandle::ReadErrors()
{  
	char tempBuf[10];
	int length = 0;
	int BytesRead = 0;
	int SocketDesc = 0;
	int lineNumber = 0;
	
	dump("Read Errors");
	
	XxxSocketConnection * socket;
	socket = connectionHandle->predConnection->socket2;
	SocketDesc = socket->socketDesc;
	
	// read error message size
	BytesRead = ::recv(SocketDesc, tempBuf, sizeof (unsigned long), 0);
	if (BytesRead <= 0) {
		cerr << "Bad Read" << endl;
		return 0;
	}
	// read error message
	length = binatoi(tempBuf);
	if (length == 0)
	{
		// no error
		return XXX_NOT_OK;
	}
	else 
	{
		if ((BytesRead = ::recv(SocketDesc, errorMessage, length, 0)) <= 0) 
		{
			cerr << "Bad Read" << endl;
			return SQL_ERROR;
		}
		errorMessage[length] = 0;
		if (errorInfo != NULL)  // already have error messages
			errorInfo->Clear();
		else errorInfo = new XxxErrInfo();
		// Now read errors one by one until there is no errors returned.
		char fileName[64];
		char errorString[XXX_ERRMESSAGE_MAX];
		char message[XXX_ERRMESSAGE_MAX];
		
		while (1) 
		{
			// read filename size.
			if ((BytesRead = ::recv(SocketDesc, tempBuf, sizeof (unsigned long), 0)) <= 0) 
			{
				cerr << "Bad Read" << endl;
				return 0;
			}	
			
			length = binatoi(tempBuf);
			if (!length)  // finish when zero size fileName
				break;
			if ((BytesRead = ::recv(SocketDesc, fileName, length, 0)) <= 0) 
			{
				cerr << "Bad Read" << endl;
				return SQL_ERROR;
			}
			fileName[length] = 0;
			// read line number
			if ((BytesRead = ::recv(SocketDesc, tempBuf, sizeof (unsigned long), 0)) <= 0) 
			{
				cerr << "Bad Read" << endl;
				return 0;
			}
			// read linenumber
			lineNumber = binatoi(tempBuf);
			// read error string, size first
			if ((BytesRead = ::recv(SocketDesc, tempBuf, sizeof (unsigned long), 0)) <= 0) 
			{
				cerr << "Bad Read" << endl;
				return 0;
			}
			
			length = binatoi(tempBuf);
			if ((BytesRead = ::recv(SocketDesc, errorString, length, 0)) <= 0) 
			{
				cerr << "Bad Read" << endl;
				return SQL_ERROR;
			}
			errorString[length] = 0;
			// read message
			if ((BytesRead = ::recv(SocketDesc, tempBuf, sizeof (unsigned long), 0)) <= 0) 
			{
				cerr << "Bad Read" << endl;
				return 0;
			}
			
			length = binatoi(tempBuf);
			if ((BytesRead = ::recv(SocketDesc, message, length, 0)) <= 0) 
			{
				cerr << "Bad Read" << endl;
				return SQL_ERROR;
			}
			message[length] = 0;
			// push into stack
			errorInfo->Note(0,message,0,fileName,lineNumber);
		}
		
		// write back an END
		sprintf(tempBuf,"END");
		if (::send(SocketDesc, tempBuf, 3, 0) <= 0) 
		{
			cerr << "Error writing End" << endl;
			return SQL_ERROR;
		}
		/* // receive OK
		if ((BytesRead = ::recv(SocketDesc, tempBuf, 2*sizeof (unsigned char), 0)) <= 0) {
		cerr << "Bad Read" << endl;
		return SQL_ERROR;
		}
		if (tempBuf[0]!= 'O' || tempBuf[1]!='K') {
		cerr << "Bad END" << endl;
		return SQL_ERROR;
	}*/
		return XXX_OK;
	}
}

/** GetSchema
*	Read the schema from socket, store at statement handle
*/
SQLRETURN StatementHandle::GetSchema()
{ 
	char tempBuf[10], Buf[500];
	int length, i;
	int BytesRead;
	
	XxxSocketConnection* socket, * errorsocket;
	socket = connectionHandle->predConnection->socket;
	errorsocket = connectionHandle->predConnection->socket2;
	dump("Enter GetSchema");
	
	if (!socket || !errorsocket) 
	{
		cerr << "Invalid connection" << endl;
		dump("Invalid connection");
		return SQL_ERROR;
	}
	
	// select for the result socket and error socket, 
	fd_set FAR  tempSockets;
	
	tempSockets.fd_count = 2;
	tempSockets.fd_array[0] = socket->socketDesc;
	tempSockets.fd_array[1] = errorsocket->socketDesc;
	
	if (select( 0, &tempSockets, NULL,NULL, NULL) != SOCKET_ERROR) 
	{
		// some information returned, polling to see which one is ok
		preProcessed = 1;
		tempSockets.fd_count = 1;
		tempSockets.fd_array[0] = errorsocket->socketDesc;
		const struct timeval nowait = {0,0};
		if (select (0,&tempSockets, NULL,NULL, &nowait ) == 1) 
		{ 
			if ( errorInfo != NULL)
				delete errorInfo;
			errorInfo = new XxxErrInfo();
			if (ReadErrors()==XXX_OK)
			{
				// Error messages
				dump("Error message exists!");
				
				// Now printout errors 
				//errorInfo->Print(,errorMessage);
				// close the request
				int SocketDesc = socket->socketDesc;
				if ((BytesRead = ::recv(SocketDesc, tempBuf, sizeof (unsigned long), 0)) <= 0) 
				{
					cerr << "Bad Read" << endl;
					schema = 0;
					return 0;
				}
				// judge whether it is end of records, meet \n (10) 0
				if (tempBuf[0]== 10 && tempBuf[1] == 0) 
				{
					// send back END , receive OK
					sprintf(tempBuf,"END");
					if (::send(SocketDesc, tempBuf, 3, 0) <= 0) 
					{
						cerr << "Error writing End" << endl;
						return SQL_ERROR;
					}
					// receive OK
					if ((BytesRead = ::recv(SocketDesc, tempBuf, 2*sizeof (unsigned char), 0)) <= 0) 
					{
						cerr << "Bad Read" << endl;
						return SQL_ERROR;
					}
					if (tempBuf[0]!= 'O' || tempBuf[1]!='K') 
					{
						cerr << "Bad END" << endl;
						return SQL_ERROR;
					}
					return SQL_ERROR;
				}
			}
		}
		
		// no error!
		dump("No error message");	 
		
		schema = new RecordSchema();
		
		int SocketDesc = socket->socketDesc;
		if (special == 1) 
		{ // list functions, return lines
			dump("list commands");
			schema->RecArity = 1;
			schema->Attrs = new Attribute[schema->RecArity];
			schema->RecordOffsets = new int[schema->RecArity];
			schema->Attrs[0].Name = new char[2];
			strcpy(schema->Attrs[0].Name,"t");
			schema->Attrs[0].Type = 2; // string
			schema->Attrs[0].TypeCatInfo = NULL;
			return SQL_SUCCESS;
		}
		if ( special == 2) {// no result
			dump("no result");
			// reach end of sending
			// END - write no error
			SocketDesc = errorsocket->socketDesc;
			sprintf(tempBuf,"END");
			// end no error
			if (::send(SocketDesc, tempBuf, 3, 0) <= 0) 
			{
				dump( "Error writing End");
				return SQL_ERROR;
			}
			dump("sent END - socket");
			
			// END - Close request
			SocketDesc = socket->socketDesc;
			if ((BytesRead = ::recv(SocketDesc, tempBuf, sizeof (unsigned long), 0)) <= 0) 
			{
				dump( "Bad Read");
				schema = 0;
				return 0;
			}
			
			// judge whether it is end of records, meet \n (10) 0
			if (tempBuf[0]== 10 && tempBuf[1] == 0) 
			{
				// end no data
				sprintf(tempBuf,"END");
				if (::send(SocketDesc, tempBuf, 3, 0) <= 0) 
				{
					dump( "Error writing End");
					return SQL_ERROR;
				}
				dump("sent END");
				// receive OK
				if ((BytesRead = ::recv(SocketDesc, tempBuf, 2*sizeof (unsigned char), 0)) <= 0) 
				{
					dump("Bad Read");
					return SQL_ERROR;
				}
				if (tempBuf[0]!= 'O' || tempBuf[1]!='K') 
				{
					dump("Bad END");
					return SQL_ERROR;
				}
				dump("SQL_NO_DATA");
				return SQL_NO_DATA;
			}
			
		}	
		
		// read 27, formatted sentinel
		if ((BytesRead = ::recv(SocketDesc, tempBuf, sizeof (unsigned char), 0)) <= 0) 
		{
			cerr << "Bad Read" << endl;
			return SQL_ERROR;
		}
		dump("got sentinel");
		// Get transactionID length, name
		if ((BytesRead = ::recv(SocketDesc, tempBuf, sizeof (unsigned long), 0)) <= 0) 
		{
			cerr << "Bad Read" << endl;
			return SQL_ERROR;
		}
		dump("got length");
		length = binatoi(tempBuf);
		if ((BytesRead = ::recv(SocketDesc, Buf,length , 0)) <= 0) 
		{
			cerr << "Bad Read" << endl;
			return SQL_ERROR;
		}
		dump("got TID");
		// Get Schema Arity
		if ((BytesRead = ::recv(SocketDesc, tempBuf, sizeof (unsigned long), 0)) <= 0) 
		{
			cerr << "Bad Read" << endl;
			return SQL_ERROR;
		}
		schema->RecArity = binatoi(tempBuf);
		// Get each meta infor
		dump("Read RecArity");
		schema->Attrs = new Attribute[schema->RecArity];
		schema->RecordOffsets = new int[schema->RecArity];
		for (i = 0; i < schema->RecArity;i++) 
		{
			// get TypeID
			if ((BytesRead = ::recv(SocketDesc, tempBuf, sizeof (unsigned long), 0)) <= 0) 
			{
				cerr << "Bad Read" << endl;
				return SQL_ERROR;
			}
			schema->Attrs[i].Type = binatoi(tempBuf);
			// get AttrNameLength
			if ((BytesRead = ::recv(SocketDesc, tempBuf, sizeof (unsigned long), 0)) <= 0) {
				cerr << "Bad Read" << endl;
				return SQL_ERROR;
			}
			length = binatoi(tempBuf);
			schema->Attrs[i].Name = new char[length+1];
			
			if ((BytesRead = ::recv(SocketDesc, Buf, length, 0)) <= 0) 
			{
				cerr << "Bad Read" << endl;
				return SQL_ERROR;
			}
			Buf[length] = 0;
			strcpy(schema->Attrs[i].Name,Buf);
			// Get AttrOffset.
			if ((BytesRead = ::recv(SocketDesc, tempBuf, sizeof (unsigned long), 0)) <= 0) 
			{
				cerr << "Bad Read" << endl;
				return SQL_ERROR;
			}
			schema->RecordOffsets[i] = binatoi(tempBuf);
			// Get Meta info
			if ((BytesRead = ::recv(SocketDesc, tempBuf, sizeof (unsigned long), 0)) <= 0) 
			{
				cerr << "Bad Read" << endl;
				return SQL_ERROR;
			}
			length = binatoi(tempBuf);
			
			char info[200];
			sprintf(info,"meta info length is %d",length);
			dump(info);
			if ( length != 0) 
			{ //  meta info
				schema->Attrs[i].TypeCatInfo = new char[length+1];	 
				if ((BytesRead = ::recv(SocketDesc, Buf, length, 0)) <= 0) 
				{
					cerr << "Bad Read" << endl;
					return SQL_ERROR;
				}
				Buf[length] = 0;
				strcpy(schema->Attrs[i].TypeCatInfo,Buf);
			}
			else 
			{
				schema->Attrs[i].TypeCatInfo = NULL;
			}
		}
		
		return SQL_SUCCESS;
	}
	dump("socket error");
	return SQL_ERROR;
}
