#include "predconnect.h"

//
// Methods of PredConnection
//
extern int dump(char *);
const char *PredConnection::initString = "NEW SESSION\n";

PredConnection::PredConnection(const char* HostName, int PortNumber,
							   const char* UserName, const char* Password, bool& success,
							   bool Binary, bool GetSchemas)
{ 
	success = FALSE;
	init();
	binaryMode = Binary;
	getSchemas = GetSchemas;
	portNumber = PortNumber;
	hostName = new char[strlen(HostName)+1];
	strcpy(hostName, HostName);
	userName = new char[strlen(UserName)+1];
	strcpy(userName, UserName);
	password = new char[strlen(Password)+1];
	strcpy(password, Password);
	
	socket = new XxxSocketConnection(hostName, portNumber);
	if (!socket->Valid()) {
		cerr << "Did not make connection to server" << endl;
		dump("Socket init fail!");
		return;
	}
	dump("Now begin Setup Protocol");
	// set up protocol
	if (!setupProtocol()) {
		cerr << "Could not set up protocol" << endl;
		dump("Setup Protocol Fail!");
		return;
	}
	valid = true;
	success = TRUE;
}

PredConnection::~PredConnection()
{
	// expect the server to close out the protocol with a \0
	if (valid) {
		if (!terminateProtocol()) {
			cerr << "No clean shutdown from server" << endl;
		}
	}
	reclaim();
}

unsigned long binatoi(char buf[4])
{	
	unsigned char tempbuf[4];
	int i;
	for ( i = 0; i < 4; i++)
		tempbuf[i] = buf[i];
	return (((tempbuf[0]*256+tempbuf[1])*256+tempbuf[2])*256+tempbuf[3]);
}



int PredConnection::setupProtocol()
{
	if (!socket) {
		cerr << "Invalid connection" << endl;
		return 0;
	}
	int SocketDesc = socket->socketDesc;
	// Send the string which requests a new client session
	::send(SocketDesc, initString, strlen(initString),0);
	
	char Buf[500];
	char tempBuf[4];
	int BytesRead;
	int clientID;
	
	if ((BytesRead = ::recv(SocketDesc, Buf, 500, 0)) <= 0) {
		dump("Bad Read");
		cerr << "Bad Read" << endl;
		return 0;
	}
	if (strncmp(Buf, "Setup?\n", BytesRead)) {
		cerr << "Unexpected protocol string from server:" << Buf << endl;
		cerr << "Bytes Read: " << BytesRead << endl;
		return 0;
	}
	// now send back a string with initial info
	sprintf (Buf, "%s %s %s ", userName, password, // passwd
		"javaclient"); // protocolname
	::send(SocketDesc, Buf, strlen(Buf)+1, 0);
	// get client ID
	if ((BytesRead = ::recv(SocketDesc, tempBuf, sizeof (unsigned long), 0)) <= 0) {
		cerr << "Bad Read" << endl;
		return 0;
	}
	
	clientID = binatoi(tempBuf);
	
	if ((BytesRead = ::recv(SocketDesc, tempBuf, sizeof (unsigned long), 0)) <= 0) {
		cerr << "Bad Read" << endl;
		return 0;
	}
	
	int i, length;
	numADTs = binatoi(tempBuf);
	if ( numADTs > 0)
		ADTs = (char **)malloc(numADTs * sizeof(char *));
	for ( i = 0; i < numADTs; i++) {
		// read typeID, just is i in java client protocol
		if ((BytesRead = ::recv(SocketDesc, tempBuf, sizeof (unsigned long), 0)) <= 0) {
			cerr << "Bad Read" << endl;
			return 0;
		}
		// read length of typename
		if ((BytesRead = ::recv(SocketDesc, tempBuf, sizeof (unsigned long), 0)) <= 0) {
			cerr << "Bad Read" << endl;
			return 0;
		}
		length = binatoi(tempBuf);
		// read  typename
		if ((BytesRead = ::recv(SocketDesc, Buf, length * sizeof (char), 0)) <= 0) {
			cerr << "Bad Read" << endl;
			return 0;
		}
		Buf[length] = '\0';
		ADTs[i] = (char *) malloc((length+1)*sizeof(char));
		strncpy(ADTs[i],Buf,length+1);
	}
	
	// read secondary socket port number
	if ((BytesRead = ::recv(SocketDesc, tempBuf, sizeof (unsigned long), 0)) <= 0) {
		cerr << "Bad Read" << endl;
		return 0;
	}
	int	portNumber2 = binatoi(tempBuf);
	// new socke2
	socket2 = new XxxSocketConnection(hostName, portNumber2);
	if (!socket2->Valid()) {
		cerr << "Did not make connection to server" << endl;
		return 0;
	}
	
	
	return 1;
}

bool PredConnection::requestTerminated(const char*Buf, int BytesRead) const
{
	if (!valid) {
		cerr << "Invalid connection" << endl;
		return true;
	}
	// answers terminated by \0
	if (!Buf[BytesRead-1]) { 
		return true;
	}
	return false;
}

int PredConnection::terminateProtocol()
{
	if (!valid) {
		cerr << "Invalid connection" << endl;
		return 0;
	}
	int SocketDesc = socket->socketDesc;
	char Buf[100];
	int BytesToRead = 0; // signal EOF
	::send(SocketDesc, (char*)&BytesToRead, sizeof(BytesToRead), 0);
	
	if (::recv(SocketDesc, Buf, 100, 0) < 0) {
		return 0;
	}
	return 1;
}

int PredConnection::SendRequest(const char* Buf)
{
	if (!valid) {
		cerr << "Invalid connection" << endl;
		return 0;
	}
	int SocketDesc = socket->socketDesc;
	// first send the length
	int BufLength = strlen(Buf)+1;
	int NetworkBufLength = htonl(BufLength);
	//cerr << "Send Request Length" << NetworkBufLength << endl;
	
	if (::send(SocketDesc, (char*)&NetworkBufLength, sizeof(int), 0) != sizeof(int)) {
		cerr << "Error writing data length" << endl;
		valid = false;
		return 0;
	}
	// then send the data
	//cerr << "Send data:" << Buf << endl;
	if (::send(SocketDesc, Buf, strlen(Buf)+1, 0) != (int)(strlen(Buf)+1)) {
		cerr << "Error writing data" << endl;
		valid = false;
		return 0;
	}
	
	return 1;
}