/** The stub of all ODBC API fuctions, those in set up DLL is in setup.c
*  Author: zhiyuan chen zhychen@cs.cornell.eud
*  Tricky things: ODBC is a damn, you have to put in all functions even if you
*  haven't implement them. It is always the case the driver manager will call some
*  functions you don't know.
*
*	Refer to Readme.doc for implemented fuctions.
*
*	Warning: Even for implemented functions, some features are not fully supported.
*
*	The export of these functions is in testdll.def.
*
*/

#include <ctype.h>
#include <sys/types.h>
#include <sys/timeb.h>

#ifndef _WINSOCK2API_
#define _WINSOCK2API_
#define _WINSOCKAPI_   // Prevent inclusion of winsock.h in windows.h 
#endif


#include <windows.h>
#ifndef HANDLE_H
#include "handle.h"
#endif
FILE *fp;

#include "sql.h"
#include "sqlext.h"
#include "odbcinst.h"

int dump(char * message) {
	fp = fopen("c:\\temp\\log.txt","a");
	fprintf(fp,"\n%s",message);
	fclose(fp);
	return 0;
}

/** SQLAllocHandle
*  Allocate handles
*/
SQLRETURN    SQL_API SQLAllocHandle(
									SQLSMALLINT			HandleType,
									SQLHANDLE 			InputHandle,
									SQLHANDLE *			OutputHandlePtr)
{	
	char temp[100];
	HGLOBAL hdbc;
	Conn *hd;
	Env *henv;
	Stmt *hstmt;
	strcpy(temp,"\n begin allocate");
	dump(temp);
	switch (HandleType) 
	{	case SQL_HANDLE_DBC:
	dump(" allocate dbc handle");
	// global allocate
	hdbc = GlobalAlloc (GMEM_MOVEABLE | GMEM_ZEROINIT, sizeof (Conn));
	if (!hdbc || (hd = (Conn *)GlobalLock (hdbc)) == SQL_NULL_HDBC) {
        GlobalFree (hdbc);      //      Free it if lock fails
		dump("\n Can not allocate globla memory for connnection handle");
        return SQL_ERROR;
	} 
	// Now allocate our local connectionhandler
	hd->conn = new ConnectionHandle();
	if ( hd->conn == NULL) {
		dump("\n wuwu, allocate null connection handle");
		*OutputHandlePtr = NULL;
		return SQL_ERROR;
	}
	dump("\n allocate local connection handle success");
	
	if (InputHandle == NULL) {
		dump("\n Null environment handle!");
		return SQL_ERROR;			
	}
	henv = (Env *)InputHandle;
	hd->conn->Env = henv->env; 
	hd->conn->Env->count++;
	dump("\n success set env handle memember");
	sprintf(temp,"current Env count is %d",hd->conn->Env->count);
	dump(temp);
	*OutputHandlePtr = hd;
	return SQL_SUCCESS;
	
	case SQL_HANDLE_STMT: // allocate statement handle
		dump("\n Begin Allocate Statement Handle");
		if (InputHandle == NULL) {
			dump("\n Null environment handle!");
			return SQL_ERROR;			
		}
		hd = (Conn *)InputHandle; // get input handle
		hdbc = GlobalAlloc (GMEM_MOVEABLE | GMEM_ZEROINIT, sizeof (Stmt));
		sprintf(temp,"Stmt is %d", sizeof(Stmt));
		dump(temp);
		if (!hdbc || (hstmt = (Stmt *)GlobalLock (hdbc)) == SQL_NULL_HDBC) {
			GlobalFree (hdbc);      //      Free it if lock fails
			dump("\n Can not allocate globla memory for stat handle");
			return SQL_ERROR;
		}
		hstmt->stmt = new StatementHandle(hd->conn);
		sprintf(temp,"Host name is %s", hstmt->stmt->connectionHandle->predConnection->hostName);
		dump(temp);
		if ( hstmt->stmt == NULL) {
			dump("Can not allocate local statement handle");
			*OutputHandlePtr = NULL;
			return SQL_ERROR;
		}
		// allocate four implicit desc handlers 
		if ( AllocDesc((HSTMT)hstmt,(SQLHDESC *)&hstmt->stmt->App_Row_Desc, (SQLHDESC *)&hstmt->stmt->App_Param_Desc,
			(SQLHDESC *)&hstmt->stmt->Imp_Row_Desc, (SQLHDESC *)&hstmt->stmt->Imp_Param_Desc) != SQL_SUCCESS)
			return SQL_ERROR;
		*OutputHandlePtr = hstmt;
		sprintf(temp, "stmt is %d", *OutputHandlePtr);
		dump(temp);
		dump("Success allocate stat handle");
		return SQL_SUCCESS;
	case SQL_HANDLE_ENV: 
		dump("\n allocate Env handle");
		hdbc = GlobalAlloc (GMEM_MOVEABLE | GMEM_ZEROINIT, sizeof (Env));
		sprintf(temp,"size is %d",sizeof(Env));
		dump(temp);
		if (!hdbc || (henv = (Env *)GlobalLock (hdbc)) == SQL_NULL_HDBC) {
			GlobalFree (hdbc);      //      Free it if lock fails
			dump("\n Can not allocate globla memory for stat handle");
			return SQL_ERROR;
		}	
		henv->env = new EnvHandle();
		if (henv->env == NULL) {
			dump("\n Can not allocate local env handle!");
			return SQL_ERROR;
		}
		
		sprintf(temp,"Current count is %d",henv->env->count);
		*OutputHandlePtr = henv;
		dump(temp);
		return SQL_SUCCESS;
	case SQL_HANDLE_DESC :
		dump("Allocate Desc handle!");
		dump("Not supported yet!");
		return SQL_ERROR;
	default: dump("Try to allocate unrecognized Handle");
		return  SQL_INVALID_HANDLE;
	}
}

/** Allocate implicit four desc handlers for a statement
*	We just store all information in statement handle and desc handle is just cheating
*/
SQLRETURN SQL_API AllocDesc(HSTMT InputHandle,SQLHDESC * App_Row_Desc, SQLHDESC * App_Param_Desc,
							SQLHDESC * Imp_Row_Desc, SQLHDESC * Imp_Param_Desc)
{
	HGLOBAL hd;
	Desc * desc;
	Stmt * stmt;
	dump("Begin allocate description handlers ");
	stmt = (Stmt *) InputHandle;
	
	hd = GlobalAlloc (GMEM_MOVEABLE | GMEM_ZEROINIT, sizeof (Desc));
	if (!hd || ( desc = (Desc *)GlobalLock (hd)) == SQL_NULL_HDBC) {
		GlobalFree (hd);      //      Free it if lock fails
		dump("\n Can not allocate globla memory for Desc handle");
		return SQL_ERROR;
	} 
	desc->desc = new DescHandle(stmt->stmt);
	(*App_Row_Desc) = desc;
	
	hd = GlobalAlloc (GMEM_MOVEABLE | GMEM_ZEROINIT, sizeof (Desc));
	if (!hd || ( desc = (Desc *)GlobalLock (hd)) == SQL_NULL_HDBC) {
		GlobalFree (hd);      //      Free it if lock fails
		dump("\n Can not allocate globla memory for Desc handle");
		return SQL_ERROR;
	} 
	desc->desc = new DescHandle(stmt->stmt);
	(*App_Param_Desc) = desc;
	
	hd = GlobalAlloc (GMEM_MOVEABLE | GMEM_ZEROINIT, sizeof (Desc));
	if (!hd || ( desc = (Desc *)GlobalLock (hd)) == SQL_NULL_HDBC) {
		GlobalFree (hd);      //      Free it if lock fails
		dump("\n Can not allocate globla memory for Desc handle");
		return SQL_ERROR;
	} 
	desc->desc = new DescHandle(stmt->stmt);
	(*Imp_Row_Desc) = desc;
	
	hd = GlobalAlloc (GMEM_MOVEABLE | GMEM_ZEROINIT, sizeof (Desc));
	if (!hd || ( desc = (Desc *)GlobalLock (hd)) == SQL_NULL_HDBC) {
		GlobalFree (hd);      //      Free it if lock fails
		dump("\n Can not allocate globla memory for Desc handle");
		return SQL_ERROR;
	} 
	desc->desc = new DescHandle(stmt->stmt);
	(*Imp_Param_Desc) = desc;
	return SQL_SUCCESS;
}


SQLRETURN SQL_API SQLAllocStmt(
							   HDBC      InputHandle,
							   HSTMT  *OutputHandlePtr)
{
	dump("Enter SQLAllocStmt");
	char temp[100];
	HGLOBAL hdbc;
	Conn *hd;
	Stmt *hstmt;
	
	strcpy(temp,"\n begin allocate");
	dump(temp);
	dump("\n Begin Allocate Statement Handle");
	if (InputHandle == NULL) {
		dump("\n Null environment handle!");
		return SQL_ERROR;			
	}
	hd = (Conn *)InputHandle; // get input handle
	hdbc = GlobalAlloc (GMEM_MOVEABLE | GMEM_ZEROINIT, sizeof (Stmt));
	if (!hdbc || (hstmt = (Stmt *)GlobalLock (hdbc)) == SQL_NULL_HDBC) {
		GlobalFree (hdbc);      //      Free it if lock fails
		dump("\n Can not allocate globla memory for stat handle");
		return SQL_ERROR;
	}
	hstmt->stmt = new StatementHandle(hd->conn);
	if ( hstmt->stmt == NULL) {
		dump("Can not allocate local statement handle");
		*OutputHandlePtr = NULL;
		return SQL_ERROR;
	}
	*OutputHandlePtr = hstmt;
	sprintf(temp,"Alloc stmt handle is %d",hstmt);
	dump(temp);
	dump("Success allocate stat handle");
	return SQL_SUCCESS;
}


/** SQLSetEnvAttr
*	Necessary because always needs to set ODBC version after
*	allocating the environment handle and before allocating 
*	connection handle
*/
SQLRETURN   SQL_API SQLSetEnvAttr(
								  SQLHENV			environmentHandle,
								  SQLINTEGER	Attribute,
								  SQLPOINTER	ValuePtr,
								  SQLINTEGER	StringLength)
{	
	char temp[100];
	Env * henv;
	strcpy(temp,"\n Begin call sql set env attr");
	dump(temp);
	sprintf(temp," \n Set Attr %d",Attribute);
	dump(temp);
	henv = (Env *) environmentHandle;
	switch (Attribute) {
	case SQL_ATTR_ODBC_VERSION:// henv->env->version = (unsigned long)ValuePtr;
		sprintf(temp," Version Number is %d", henv->env->version);
		dump(temp);
		break;
	default: dump("Unsupported Attribute. Sorry.");
		return SQL_ERROR;
	}
	
	return SQL_SUCCESS;
}

/**
*  SQLExecDirect execute a query directly.
*/
SQLRETURN   SQL_API SQLExecDirect(
								  SQLHSTMT		statementHandle,
								  SQLCHAR *		StatementText,
								  SQLINTEGER	TextLength)
{	
	Stmt * hstmt;
	dump("Enter SQLExecDirect");
	if (statementHandle == NULL) {
		dump("\n Error! Null statement handle in sqlexecdirect");
		return SQL_ERROR;
	}
	hstmt = (Stmt *) statementHandle;
	return (hstmt->stmt->SendRequest( StatementText ));
}

/** 
* SQLBindCol Bind a column for result. StrLen or IndPtr is the real length
*/
SQLRETURN   SQL_API SQLBindCol(
							   SQLHSTMT			statementHandle,
							   SQLUSMALLINT	ColumnNumber,
							   SQLSMALLINT		TargetType,
							   SQLPOINTER		TargetValuePtr,
							   SQLINTEGER		BufferLength,
							   SQLINTEGER *		StrLen_or_IndPtr)
{	
	Stmt * hstmt;
	if (statementHandle == NULL) {
		dump("\n Error! Null statement handle in sqlexecdirect");
		return SQL_ERROR;
	}
	hstmt = (Stmt *) statementHandle;
	return (hstmt->stmt->SQLBindCol(ColumnNumber,TargetType,TargetValuePtr,BufferLength,StrLen_or_IndPtr));
	
}

/** 
* SQLFetch Fetch result, one row at a time
*/
SQLRETURN   SQL_API SQLFetch(
							 SQLHSTMT	statementHandle)
{	
	Stmt * hstmt;
	if (statementHandle == NULL) {
		dump("\n Error! Null statement handle in sqlexecdirect");
		return SQL_ERROR;
	}
	hstmt = (Stmt *) statementHandle;
	return hstmt->stmt->SQLFetch();
}

/**
* SQLFreeStmt
*/ 
SQLRETURN   SQL_API SQLFreeStmt(
								SQLHSTMT			statementHandle,
								SQLUSMALLINT	Option)
{ 
	Stmt * hstmt;
	dump("Enter SQLFreeStmt");
	if (statementHandle == NULL) {
		dump("\n Null input statementHandle");
		return SQL_ERROR;
	}
	hstmt = (Stmt *)statementHandle;
	switch (Option) {
	case SQL_CLOSE: // close the cursor, do nothing
		dump("Close cursor");
		break;
	case SQL_UNBIND: 
		dump("Unbind");
		hstmt->stmt->connectionHandle->count--;
		break;
	}
	return SQL_SUCCESS;
}

/** 
* strncpy2 a copy function, which copy src to des, and not more than length
*	Add 0 at end
*/
void SQL_API strncpy2(char * des, char * src, int length)
{
	if ( length >= strlen(src)) {
		strcpy(des,src);
		*(des+strlen(src)) = 0;
	}
	else {
		strncpy(des,src,length);
		*(des+length) = 0;
	}
}

/** SQLConnect Connect to the data source
*/	
SQLRETURN     SQL_API SQLConnect(
								 SQLHDBC			connectionHandle,
								 SQLCHAR *		ServerName,
								 SQLSMALLINT	NameLength1,
								 SQLCHAR *		UserName,
								 SQLSMALLINT	NameLength2,
								 SQLCHAR *		Authentication,
								 SQLSMALLINT	NameLength3)
{	
	Conn *conn;
	char info[555];
	char server[200],user[200], passwd[200];
	dump("Now begin SQLConnect");
	sprintf(info,"Name length1 is %d", NameLength1);
	dump(info);
	strncpy2(server,(char *)ServerName,NameLength1);
	strncpy2(user,(char *)UserName,NameLength2);
	strncpy2(passwd,(char *)Authentication,NameLength3);
	
	sprintf(info,"ServerName is %s, User Name is %s, Autthentication is %s", server,user,passwd);
	dump(info);
	if ( connectionHandle == NULL) {
		dump("\n Null handler");
		return SQL_ERROR;
	}
	conn = (Conn *) connectionHandle;
	return conn->conn->Connect((SQLCHAR *)server,NameLength1,(SQLCHAR *)user,NameLength2,(SQLCHAR *)passwd,NameLength3);
	
}

/**
* SQLDisconnect
*/ 
SQLRETURN   SQL_API SQLDisconnect(SQLHDBC	connectionHandle)
{	
	Conn *conn;
	dump("Now begin SQLConnect");
	if ( connectionHandle == NULL) {
		dump("\n Null handler");
		return SQL_ERROR;
	}
	conn = (Conn *) connectionHandle;
	return conn->conn->FreeConnectionHandle();
}

/** 
* SQLFreeHandle Free handles
*/
SQLRETURN   SQL_API SQLFreeHandle(SQLSMALLINT	HandleType,	SQLHANDLE		Handle)
{	
	Conn *conn;
	Env *henv;
	Stmt * hstmt;
	
	dump("\n Enter SQLFREE");
	switch (HandleType) {
	case SQL_HANDLE_ENV:  
		dump("Free env handle!");
		henv = (Env *)Handle;
		if ( henv->env->Free() != SQL_SUCCESS) {
			dump("\n Error free local env");
			return SQL_ERROR;
		}
		GlobalUnlock(GlobalHandle(henv));
		GlobalFree(GlobalHandle(henv));
		return SQL_SUCCESS;
	case SQL_HANDLE_DBC: 
		dump("Free connection handle");
		conn = (Conn *)Handle;
		if (conn == NULL) {
			dump("\n Null input handle!");
			return SQL_ERROR;
		}
		GlobalUnlock(GlobalHandle(conn));
		GlobalFree(GlobalHandle(conn));
		return SQL_SUCCESS;
	case SQL_HANDLE_STMT: dump("Free Statement Handle");  
		if (Handle == NULL) {
			dump("\n Null input statementHandle");
			return SQL_ERROR;
		}
		hstmt = (Stmt *)Handle;				
		hstmt->stmt->Free();
		//  delete hstmt->stmt;
		GlobalUnlock(GlobalHandle(hstmt));
		GlobalFree(GlobalHandle(hstmt));
		break;
	case SQL_HANDLE_DESC: 
		dump("free desc handle");
		delete ((DescHandle *)Handle);
		break;
	default: return SQL_ERROR;
	}
	return SQL_SUCCESS;
}
									
/**
*
 helper function for SQLDriverConnect (below)
 */
void GetDSN(char * input, char * output)
{   
	dump("GetDSN");
	char *ptr,*ptr2;
	int len;
	for ( ptr = input; *ptr != 0; ptr++) {
		if( ((*ptr) == 'D') && (*(ptr+1)='S') && (*(ptr+2) = 'N') && ((*(ptr+3)) == '='))
			break;
	}
	if (! *ptr ) 
		*output = 0;
	else {
		ptr+=4;
		for ( len = 0,ptr2 = ptr; *ptr2 != 0; ptr2++, len++)
			if ( *ptr2 == ';')
				break;
			strncpy(output,ptr,len);
			*(output+len) = 0;
	}
	char info[64];
	sprintf(info,"DSN=%s",output);
	dump(info);
	
}
									

/**
* SQLDriverConnect
*/
SQLRETURN   SQL_API SQLDriverConnect(
									 SQLHDBC				ConnectionHandle,
									 SQLHWND			WindowHandle,
									 SQLCHAR *			InConnectionString,
									 SQLSMALLINT		StringLength1,
									 SQLCHAR *			OutConnectionString,
									 SQLSMALLINT		BufferLength,
									 SQLSMALLINT *	StringLength2Ptr,
									 SQLUSMALLINT	DriverCompletion)
{	
	dump("\n Driver Connect!");
	char info[256];
	char *ptr;
	char dsnName[64];
	char host[64];
	sprintf(info,"%s, Completions:%d",InConnectionString,DriverCompletion);
	dump(info);
	if ( ! StringLength2Ptr)
		StringLength2Ptr = new short;
	StringLength2Ptr = 0;
	// find the severname from onbc.ini
	// First get the DSN
	GetDSN((char *)InConnectionString,dsnName);
	if (!dsnName[0]) {
		dump("No DSN!");
		return SQL_ERROR;
	}
	
	
	SQLGetPrivateProfileString(dsnName,"Servername",
		"localhost", host, 64,"ODBC.INI");
	dump(host);
	return	SQLConnect(ConnectionHandle,(unsigned char *)host,64,(unsigned char *)"abahd",64,(unsigned char *)"feh",64);
	
}

									

/** 
* Not supported yet
*/
SQLRETURN   SQL_API SQLBrowseConnect(
									 SQLHDBC				ConnectionHandle,
									 SQLCHAR *			InConnectionString,
									 SQLSMALLINT		StringLength1,
									 SQLCHAR *			OutConnectionString,
									 SQLSMALLINT		BufferLength,
									 SQLSMALLINT *	StringLength2Ptr)
{ 
	Conn *conn;
	char temp[50];
	SQLCHAR 		ServerName[100];
	SQLSMALLINT	NameLength1;
	SQLCHAR 		UserName[100];
	SQLSMALLINT	NameLength2;
	SQLCHAR 		Authentication[100];
	SQLSMALLINT	NameLength3;
	strcpy((char *)ServerName,"oslo.cs.cornell.edu");
	NameLength1 = strlen((char *) ServerName);
	strcpy((char *)UserName,"zhiyuan");
	NameLength2 = strlen((char *)UserName);
	strcpy((char *)Authentication,"aaa");
	NameLength3 = strlen((char *)Authentication);
	strcpy((char *)OutConnectionString,"DSN=Predator Datasource;HOST=olso.cs.cornell.edu;UID=zhiyuan;PWD=aaa");
	*StringLength2Ptr = strlen((char *)OutConnectionString);
	dump("\n SQLBrowseconect");
	sprintf(temp,"\nOutputString Length is %d",*StringLength2Ptr);
	dump(temp);
	// current implementation, simply run connection to default data source, Predator Datasource
	if ( ConnectionHandle == NULL) {
		dump("\n Null handler");
		return SQL_INVALID_HANDLE;
	}
	conn = (Conn *) ConnectionHandle;
	conn->conn->Connect(ServerName,NameLength1,UserName,NameLength2,Authentication,NameLength3);
	return SQL_SUCCESS;
	
}
									

/**
* string helper functions
*/
char *strncpy_null(char *dst, const char *src, size_t len)
{
	unsigned int i;
	
	if (NULL != dst) {
		for(i = 0; src[i] && i < len - 1; i++) {
												dst[i] = src[i];
		}
		
		if(len > 0) {
												dst[i] = '\0';
		}
	}
	return dst;
}

void strnullcpy(char * des, char * source)
{ 
	char * ptr1, *ptr2;
	for ( ptr1 = source,ptr2 = des; *ptr1 != 0; ptr1++,ptr2++)
		*ptr2 = *ptr1;
	*ptr2 = 0;
	dump(des);
}

char *strmake(char *dst, char *src, int length)
{
	while (length--)
		if (! (*dst++ = *src++))
		  return dst-1;
		*dst=0;
		return dst;
}


/** 
* SQLGetInfo
*  Return necessary information for driver manager
*/
SQLRETURN SQL_API SQLGetInfo(SQLHDBC hdbc, SQLUSMALLINT fInfoType, SQLPOINTER rgbInfoValue,
							 SQLSMALLINT cbInfoValueMax, SQLSMALLINT FAR *pcbInfoValue)
{
	Conn * conn = (Conn *) hdbc;
	char dummy2[16];
	SWORD dummy;
	char info[40];
	sprintf(info,"fInfoType: %d",fInfoType);
	
	if (!pcbInfoValue)
		pcbInfoValue=&dummy;
	if (!rgbInfoValue)
	{						/* This is because of VB5 */
		rgbInfoValue=dummy2;
		cbInfoValueMax=sizeof(dummy2)-1;
	}
	switch (fInfoType) {
	case SQL_MAX_USER_NAME_LEN:
		
		*(UNALIGNED SWORD FAR *) rgbInfoValue = 16;
		*pcbInfoValue = sizeof(SWORD);
		break;
	case SQL_DRIVER_ODBC_VER:
		*pcbInfoValue=(SWORD) (strmake((char*) rgbInfoValue,"03.00",
												cbInfoValueMax)
												- (char*) rgbInfoValue);
		break;
	case SQL_ACTIVE_CONNECTIONS:
	case SQL_ACTIVE_STATEMENTS:
		*((SWORD*) rgbInfoValue)=0;			/* No limits */
		*pcbInfoValue = sizeof(SWORD);
		break;
#ifdef SQL_ASYNC_MODE
	case SQL_ASYNC_MODE:
		*((SQLUINTEGER*) rgbInfoValue)= SQL_AM_NONE;
		*pcbInfoValue = sizeof(SQLUINTEGER);
		break;
#endif
		
	case SQL_CREATE_TABLE:
		*((SQLUINTEGER*) rgbInfoValue)= SQL_CT_CREATE_TABLE | SQL_CT_COLUMN_DEFAULT;
		*pcbInfoValue = sizeof(SQLUINTEGER);
		break;
		
	case SQL_DATA_SOURCE_NAME:
		*pcbInfoValue=(SWORD) (strmake((char*) rgbInfoValue,"PredatorDataSource",
												cbInfoValueMax)
												- (char*) rgbInfoValue);
		break;
	case SQL_DATABASE_NAME:
		*pcbInfoValue=(SWORD) (strmake((char*) rgbInfoValue,"default",
												cbInfoValueMax)
												- (char*) rgbInfoValue);
		break;
	case SQL_DRIVER_NAME:
		*pcbInfoValue=(SWORD) (strmake((char*) rgbInfoValue,"testdll.dll",
												cbInfoValueMax)
												- (char*) rgbInfoValue);
		break;
	case SQL_DRIVER_VER:
	case SQL_DBMS_VER:
		*pcbInfoValue=(SWORD) (strmake((char*) rgbInfoValue,"03.00.0000",
												cbInfoValueMax)
												- (char*) rgbInfoValue);
		break;
	case SQL_FETCH_DIRECTION:
		*((long*) rgbInfoValue)=SQL_FD_FETCH_NEXT | SQL_FD_FETCH_FIRST |
												SQL_FD_FETCH_LAST | SQL_FD_FETCH_PREV | SQL_FD_FETCH_ABSOLUTE |
												SQL_FD_FETCH_RELATIVE;
		*pcbInfoValue=sizeof(long);
		break;
	case SQL_ODBC_VER:
		*pcbInfoValue=(SWORD) (strmake((char*) rgbInfoValue,"03.00.0000",
												cbInfoValueMax)
												- (char*) rgbInfoValue);
		*pcbInfoValue=sizeof(SWORD);
		break;
	case SQL_ODBC_SAG_CLI_CONFORMANCE:
		*((SWORD*) rgbInfoValue)=SQL_OSCC_COMPLIANT;
		*pcbInfoValue=sizeof(SWORD);
		break;
	case SQL_SERVER_NAME:
		*pcbInfoValue=(SWORD) (strmake((char*) rgbInfoValue,conn->conn->predConnection->hostName,
												cbInfoValueMax)
												- (char*) rgbInfoValue);
		break;
	case SQL_SEARCH_PATTERN_ESCAPE:
		*pcbInfoValue=(SWORD) (strmake((char*) rgbInfoValue,"\\",cbInfoValueMax)
												- (char*) rgbInfoValue);
		break;
	case SQL_ODBC_SQL_CONFORMANCE:
	case SQL_ODBC_API_CONFORMANCE:
		*((SWORD*) rgbInfoValue)=SQL_OAC_LEVEL1;	/* Lie to make things work */
		*pcbInfoValue=sizeof(SWORD);
		break;
	case SQL_DBMS_NAME:
		*pcbInfoValue=(SWORD) (strmake((char*) rgbInfoValue,"Predator",cbInfoValueMax)
												- (char*) rgbInfoValue);
		break;
	case SQL_ROW_UPDATES:
	case SQL_ACCESSIBLE_PROCEDURES:
	case SQL_PROCEDURES:
	case SQL_EXPRESSIONS_IN_ORDERBY:
	case SQL_NEED_LONG_DATA_LEN:
	case SQL_DATA_SOURCE_READ_ONLY:
	case SQL_ODBC_SQL_OPT_IEF:
	case SQL_LIKE_ESCAPE_CLAUSE:
	case SQL_ORDER_BY_COLUMNS_IN_SELECT:
		*pcbInfoValue=(SWORD) (strmake((char*) rgbInfoValue,"N",cbInfoValueMax)
												- (char*) rgbInfoValue);
		break;
	case SQL_OUTER_JOINS:
		*pcbInfoValue=(SWORD) (strmake((char*) rgbInfoValue,"N",cbInfoValueMax)
												- (char*) rgbInfoValue);
		break;
#ifdef SQL_OJ_CAPABILITIES
	case SQL_OJ_CAPABILITIES:
		*((SQLUINTEGER*) rgbInfoValue)= SQL_OJ_LEFT | SQL_OJ_NESTED | SQL_OJ_NOT_ORDERED |
												SQL_OJ_INNER | SQL_OJ_ALL_COMPARISON_OPS;
		*pcbInfoValue=sizeof(SQLUINTEGER);
		break;
#endif
	case SQL_COLUMN_ALIAS:
	case SQL_ACCESSIBLE_TABLES:
	case SQL_MULT_RESULT_SETS:
	case SQL_MULTIPLE_ACTIVE_TXN:
	case SQL_MAX_ROW_SIZE_INCLUDES_LONG:
		*pcbInfoValue=(SWORD) (strmake((char*) rgbInfoValue,"Y",cbInfoValueMax)
												- (char*) rgbInfoValue);
		break;
#ifdef SQL_MAX_IDENTIFIER_LEN
	case SQL_MAX_IDENTIFIER_LEN:
		*((SQLUSMALLINT*) rgbInfoValue)=100;
		break;
#endif
	case SQL_CONCAT_NULL_BEHAVIOR:
		*((SWORD*) rgbInfoValue)=SQL_CB_NULL;
		*pcbInfoValue=sizeof(SWORD);
		break;
	case SQL_CURSOR_COMMIT_BEHAVIOR:		/* Lie */
	case SQL_CURSOR_ROLLBACK_BEHAVIOR:
		*((SWORD*) rgbInfoValue)=SQL_CB_PRESERVE;
		*pcbInfoValue=sizeof(SWORD);
		break;
#ifdef SQL_CURSOR_SENSITIVITY
	case SQL_CURSOR_SENSITIVITY:
		*((SQLUINTEGER*) rgbInfoValue)=SQL_UNSPECIFIED;
		break;
#endif
	case SQL_DEFAULT_TXN_ISOLATION:
		*((long*) rgbInfoValue)= SQL_TXN_READ_UNCOMMITTED;
		*pcbInfoValue=sizeof(long);
		break;
#ifdef SQL_DESCRIBE_PARAMETER
	case SQL_DESCRIBE_PARAMETER:
		*pcbInfoValue=(SWORD) (strmake((char*) rgbInfoValue,"N",cbInfoValueMax)
												- (char*) rgbInfoValue);
		break;
#endif
	case SQL_TXN_ISOLATION_OPTION:
		*((long*) rgbInfoValue)= SQL_TXN_READ_UNCOMMITTED;
		*pcbInfoValue=sizeof(long);
		break;
	case SQL_IDENTIFIER_CASE:
		*((SWORD*) rgbInfoValue)=SQL_IC_MIXED;
		*pcbInfoValue=sizeof(SWORD);
		break;
	case SQL_IDENTIFIER_QUOTE_CHAR:
		*pcbInfoValue=(SWORD) (strmake((char*) rgbInfoValue," ",cbInfoValueMax)
												- (char*) rgbInfoValue);
		break;
	case SQL_MAX_COLUMN_NAME_LEN:
		*((SWORD*) rgbInfoValue)=100;
		*pcbInfoValue=sizeof(SWORD);
		break;
	case SQL_MAX_CURSOR_NAME_LEN:
	case SQL_MAX_OWNER_NAME_LEN:
	case SQL_MAX_PROCEDURE_NAME_LEN:
	case SQL_MAX_QUALIFIER_NAME_LEN:
		*((SWORD*) rgbInfoValue)=0;
		*pcbInfoValue=sizeof(SWORD);
		break;
	case SQL_MAX_TABLE_NAME_LEN:
		*((SWORD*) rgbInfoValue)=100;
		break;
	case SQL_OWNER_TERM:
		*pcbInfoValue=(SWORD) (strmake((char*) rgbInfoValue,"owner",cbInfoValueMax)
												- (char*) rgbInfoValue);
		break;
	case SQL_PROCEDURE_TERM:
		*pcbInfoValue=(SWORD) (strmake((char*) rgbInfoValue,"procedure",
												cbInfoValueMax) - (char*) rgbInfoValue);
		break;
	case SQL_QUALIFIER_NAME_SEPARATOR:		/* what should this be */
		*pcbInfoValue=(SWORD) (strmake((char*) rgbInfoValue,",",cbInfoValueMax)
												- (char*) rgbInfoValue);
		break;
	case SQL_QUALIFIER_TERM:
		*pcbInfoValue=(SWORD) (strmake((char*) rgbInfoValue,"database",cbInfoValueMax)
												- (char*) rgbInfoValue);
		break;
	case SQL_SCROLL_CONCURRENCY:
	case SQL_SCROLL_OPTIONS:
		*((long*) rgbInfoValue)=0L;
		*pcbInfoValue=sizeof(long);
		break;
	case SQL_TABLE_TERM:
		*pcbInfoValue=(SWORD) (strmake((char*) rgbInfoValue,"table",cbInfoValueMax)
												- (char*) rgbInfoValue);
		break;
	case SQL_TXN_CAPABLE:
		*((SWORD*) rgbInfoValue)=SQL_TC_NONE;
		*pcbInfoValue=sizeof(SWORD);
		break;
	case SQL_USER_NAME:
		*pcbInfoValue=(SWORD) (strmake((char*) rgbInfoValue,"user",cbInfoValueMax)
												- (char*) rgbInfoValue);
		break;
	case SQL_CONVERT_FUNCTIONS:
		*((long*) rgbInfoValue)=0L;
		*pcbInfoValue=sizeof(long);
		break;
	case SQL_SYSTEM_FUNCTIONS:
		*((long*) rgbInfoValue)= SQL_FN_SYS_DBNAME | SQL_FN_SYS_IFNULL | SQL_FN_SYS_USERNAME;
		*pcbInfoValue=sizeof(long);
		break;
	case SQL_NUMERIC_FUNCTIONS:
		*((long*) rgbInfoValue)=SQL_FN_NUM_ABS | SQL_FN_NUM_ACOS |
												SQL_FN_NUM_ASIN | SQL_FN_NUM_ATAN | SQL_FN_NUM_ATAN2 |
												SQL_FN_NUM_CEILING | SQL_FN_NUM_COS | SQL_FN_NUM_COT | SQL_FN_NUM_EXP |
												SQL_FN_NUM_FLOOR | SQL_FN_NUM_LOG | SQL_FN_NUM_MOD | SQL_FN_NUM_SIGN |
												SQL_FN_NUM_SIN | SQL_FN_NUM_SQRT | SQL_FN_NUM_TAN | SQL_FN_NUM_PI |
												SQL_FN_NUM_RAND | SQL_FN_NUM_DEGREES | SQL_FN_NUM_LOG10 |
												SQL_FN_NUM_POWER | SQL_FN_NUM_RADIANS | SQL_FN_NUM_ROUND |
												SQL_FN_NUM_TRUNCATE;
		*pcbInfoValue=sizeof(long);
		break;
	case SQL_STRING_FUNCTIONS:
		*((long*) rgbInfoValue)=SQL_FN_STR_CONCAT | SQL_FN_STR_INSERT |
												SQL_FN_STR_LEFT | SQL_FN_STR_LTRIM | SQL_FN_STR_LENGTH |
												SQL_FN_STR_LOCATE | SQL_FN_STR_LCASE | SQL_FN_STR_REPEAT |
												SQL_FN_STR_REPLACE | SQL_FN_STR_RIGHT | SQL_FN_STR_RTRIM |
												SQL_FN_STR_SUBSTRING | SQL_FN_STR_UCASE | SQL_FN_STR_ASCII |
												SQL_FN_STR_CHAR | SQL_FN_STR_LOCATE_2 | SQL_FN_STR_SOUNDEX |
												SQL_FN_STR_SPACE;
		*pcbInfoValue=sizeof(long);
		break;
	case SQL_TIMEDATE_FUNCTIONS:
		*((long*) rgbInfoValue)= SQL_FN_TD_NOW | SQL_FN_TD_CURDATE |
												SQL_FN_TD_DAYOFMONTH | SQL_FN_TD_DAYOFWEEK | SQL_FN_TD_DAYOFYEAR |
												SQL_FN_TD_MONTH | SQL_FN_TD_QUARTER | SQL_FN_TD_WEEK | SQL_FN_TD_YEAR |
												SQL_FN_TD_CURTIME | SQL_FN_TD_HOUR | SQL_FN_TD_MINUTE |
												SQL_FN_TD_SECOND | SQL_FN_TD_DAYNAME | SQL_FN_TD_MONTHNAME;
		*pcbInfoValue=sizeof(long);
		break;
	case SQL_CONVERT_BIGINT:
	case SQL_CONVERT_BIT:
	case SQL_CONVERT_CHAR:
	case SQL_CONVERT_DATE:
	case SQL_CONVERT_DECIMAL:
	case SQL_CONVERT_DOUBLE:
	case SQL_CONVERT_FLOAT:
	case SQL_CONVERT_INTEGER:
	case SQL_CONVERT_LONGVARCHAR:
	case SQL_CONVERT_NUMERIC:
	case SQL_CONVERT_REAL:
	case SQL_CONVERT_SMALLINT:
	case SQL_CONVERT_TIME:
	case SQL_CONVERT_TIMESTAMP:
	case SQL_CONVERT_TINYINT:
	case SQL_CONVERT_VARCHAR:
		/* lie that we can handle anything */
		*((long*) rgbInfoValue)=SQL_CVT_CHAR | SQL_CVT_NUMERIC | SQL_CVT_DECIMAL |
												SQL_CVT_INTEGER | SQL_CVT_SMALLINT | SQL_CVT_FLOAT | SQL_CVT_REAL |
												SQL_CVT_DOUBLE | SQL_CVT_VARCHAR | SQL_CVT_LONGVARCHAR |
												SQL_CVT_BIT | SQL_CVT_TINYINT | SQL_CVT_BIGINT |
												SQL_CVT_DATE | SQL_CVT_TIME | SQL_CVT_TIMESTAMP;
		*pcbInfoValue=sizeof(long);
		break;
	case SQL_CONVERT_BINARY:
	case SQL_CONVERT_VARBINARY:
	case SQL_CONVERT_LONGVARBINARY:
	case SQL_POSITIONED_STATEMENTS:
	case SQL_LOCK_TYPES:
	case SQL_BOOKMARK_PERSISTENCE:
	case SQL_OWNER_USAGE:
	case SQL_QUALIFIER_USAGE:
	case SQL_SUBQUERIES:
	case SQL_UNION:
	case SQL_TIMEDATE_ADD_INTERVALS:
	case SQL_TIMEDATE_DIFF_INTERVALS:
		*((long*) rgbInfoValue)=0L;
		*pcbInfoValue=sizeof(long);
		break;
	case SQL_QUOTED_IDENTIFIER_CASE:
		*((SQLUSMALLINT*) rgbInfoValue)=SQL_IC_MIXED;
		break;
	case SQL_POS_OPERATIONS:
		*((long*) rgbInfoValue)=SQL_POS_POSITION;
		*pcbInfoValue=sizeof(long);
		break;
	case SQL_ALTER_TABLE:
		*((long*) rgbInfoValue)=SQL_AT_ADD_COLUMN | SQL_AT_DROP_COLUMN;
		*pcbInfoValue=sizeof(long);
		break;
	case SQL_CORRELATION_NAME:
		*((SWORD*) rgbInfoValue)=SQL_CN_DIFFERENT;
		*pcbInfoValue=sizeof(SWORD);
		break;
	case SQL_NON_NULLABLE_COLUMNS:
		*((SWORD*) rgbInfoValue)=SQL_NNC_NON_NULL;
		*pcbInfoValue=sizeof(SWORD);
		break;
	case SQL_NULL_COLLATION:
		*((SWORD*) rgbInfoValue)=SQL_NC_START;
		*pcbInfoValue=sizeof(SWORD);
		break;
	case SQL_MAX_COLUMNS_IN_GROUP_BY:
	case SQL_MAX_COLUMNS_IN_ORDER_BY:
	case SQL_MAX_COLUMNS_IN_SELECT:
	case SQL_MAX_COLUMNS_IN_TABLE:
	case SQL_MAX_ROW_SIZE:
		*((SWORD*) rgbInfoValue)=0;
		*pcbInfoValue=sizeof(SWORD);
		break;
	case SQL_MAX_TABLES_IN_SELECT:
		*((SWORD*) rgbInfoValue)=32;
		*pcbInfoValue=sizeof(SWORD);
		break;
	case SQL_MAX_BINARY_LITERAL_LEN:
	case SQL_MAX_CHAR_LITERAL_LEN:
		*((long*) rgbInfoValue)=0;
		*pcbInfoValue=sizeof(long);
		break;
	case SQL_MAX_COLUMNS_IN_INDEX:
		*((SWORD*) rgbInfoValue)=16;
		*pcbInfoValue=sizeof(SWORD);
		break;
	case SQL_MAX_INDEX_SIZE:
		*((long*) rgbInfoValue)=120;
		*pcbInfoValue=sizeof(long);
		break;
	case SQL_MAX_STATEMENT_LEN:
		*((long*) rgbInfoValue)=1000;
		*pcbInfoValue=sizeof(long);
		break;
	case SQL_QUALIFIER_LOCATION:
		*((SWORD*) rgbInfoValue)=SQL_QL_START;
		*pcbInfoValue=sizeof(SWORD);
		break;
	case SQL_GETDATA_EXTENSIONS:
		*((long*) rgbInfoValue)= SQL_GD_ANY_COLUMN | SQL_GD_ANY_ORDER | SQL_GD_BOUND;
		*pcbInfoValue=sizeof(long);
		break;
	case SQL_STATIC_SENSITIVITY:
		// *((long*) rgbInfoValue)=SQL_SS_DELETIONS | SQL_SS_UPDATES;
		*((long*) rgbInfoValue)=0;
		*pcbInfoValue=sizeof(long);
		break;
	case SQL_FILE_USAGE:
		*((SWORD*) rgbInfoValue)=SQL_FILE_NOT_SUPPORTED;
		*pcbInfoValue=sizeof(SWORD);
		break;
	case SQL_GROUP_BY:
		*((SWORD*) rgbInfoValue)=SQL_GB_NO_RELATION;
		*pcbInfoValue=sizeof(SWORD);
		break;
	case SQL_KEYWORDS:
		*pcbInfoValue=(SWORD) (strmake((char*) rgbInfoValue,
												"UNIQUE,ZEROFILL,UNSIGNED,BIGINT,BLOB,TINYBLOB,MEDIMUMBLOB,LONGBLOB,MEDIUMINT,PROCEDURE,SHOW,LIMIT,DEFAULT,TABLES,REGEXP,RLIKE,KEYS,TINTTEXT,MEDIUMTEXT",
												cbInfoValueMax)
												- (char*) rgbInfoValue);
		break;
	case SQL_SPECIAL_CHARACTERS:
		{
			*pcbInfoValue=(SWORD) (strmake((char*) rgbInfoValue,
								"",
								cbInfoValueMax)
								- (char*) rgbInfoValue);
			break;
		}
	default:
		{
			char buff[80];
			sprintf(buff,"Unsupported option: %d to SQLGetInfo",fInfoType);
			dump(buff);
			return(SQL_ERROR);
		}
  }
  return SQL_SUCCESS;
}




/** 
* Set the corresponding function to be supported or not
*  Assume the corresponding bit is 0 at this time
*/
void SetBitVector( SQLUSMALLINT * p, bool supported, SQLUSMALLINT id)
{	
	char info[20];
	if ( supported)
		p[id/16] = *(((UWORD*) (p)) + ((id) >> 4)) | (1 << ((id) & 0x000F));
	sprintf(info,"Set Bit %d to 1", id/4);
	dump(info);
	sprintf(info," It is %d", p[id/16]);
	dump(info);
	
}

/** 
* SQLGetFunctions
*	Return the functions supported
*/
SQLRETURN   SQL_API SQLGetFunctions(
									SQLHDBC				ConnectionHandle,
									SQLUSMALLINT	FunctionId,
									SQLUSMALLINT *	pfExists)
{	
	char info[50];
	dump("\n SQLGetFuncs");
	sprintf(info,"FuctionId %d", FunctionId);
	dump(info);
	switch ( FunctionId) {
	case SQL_API_ODBC3_ALL_FUNCTIONS: // return a SMALLINT array of SQL_API_ODBC3_ALL_FUNCTIONS_SIZE
		// It is a 4,000 bit vector, each bit stand for the # of that function, 1 is supported,0not
		// Damn! Who figured out such a stupid thing?
		memset(pfExists, 0, sizeof(SQLSMALLINT)*SQL_API_ODBC3_ALL_FUNCTIONS_SIZE);
		
		// ODBC core functions
		
		SetBitVector(pfExists,TRUE,SQL_API_SQLALLOCCONNECT);
		SetBitVector(pfExists,TRUE,SQL_API_SQLALLOCENV);
		SetBitVector(pfExists,TRUE,SQL_API_SQLALLOCSTMT);
		SetBitVector(pfExists,TRUE,SQL_API_SQLBINDCOL);
		SetBitVector(pfExists,TRUE,SQL_API_SQLCONNECT);
		SetBitVector(pfExists,TRUE,SQL_API_SQLDISCONNECT);
		SetBitVector(pfExists,TRUE,SQL_API_SQLALLOCHANDLE);
		SetBitVector(pfExists,TRUE,SQL_API_SQLFREEHANDLE);
		
		//       pfExists[SQL_API_SQLERROR]            = TRUE;
		SetBitVector(pfExists,TRUE,SQL_API_SQLEXECDIRECT);
		
		//       pfExists[SQL_API_SQLEXECUTE]          = TRUE;
		SetBitVector(pfExists,TRUE,SQL_API_SQLFETCH);
		
		//pfExists[SQL_API_SQLFREECONNECT]      = TRUE;
		//pfExists[SQL_API_SQLFREEENV]          = TRUE;
		SetBitVector(pfExists,TRUE,SQL_API_SQLFREESTMT);
		SetBitVector(pfExists,TRUE,SQL_API_SQLGETDIAGREC);
		SetBitVector(pfExists,TRUE,SQL_API_SQLGETSTMTATTR);
		SetBitVector(pfExists,TRUE,SQL_API_SQLSETSTMTATTR);
		SetBitVector(pfExists,TRUE,SQL_API_SQLNUMRESULTCOLS);
		//       pfExists[SQL_API_SQLGETCURSORNAME]    = FALSE;
		
		//   pfExists[SQL_API_SQLPREPARE]          = TRUE;  // complete?
		// pfExists[SQL_API_SQLROWCOUNT]         = TRUE;
		//pfExists[SQL_API_SQLSETCURSORNAME]    = FALSE;
		// pfExists[SQL_API_SQLSETPARAM]         = FALSE;
		// pfExists[SQL_API_SQLTRANSACT]         = TRUE;
		
		// ODBC level 1 functions
		//pfExists[SQL_API_SQLBINDPARAMETER]    = TRUE;
		//pfExists[SQL_API_SQLCOLUMNS]          = TRUE;
		SetBitVector(pfExists,TRUE,SQL_API_SQLDRIVERCONNECT);
		
		//      pfExists[SQL_API_SQLGETCONNECTOPTION] = TRUE;  // partial
		//    pfExists[SQL_API_SQLGETDATA]          = TRUE;
		SetBitVector(pfExists,TRUE,SQL_API_SQLGETFUNCTIONS);
		//   pfExists[SQL_API_SQLGETFUNCTIONS]     = TRUE;  // sadly,  I still
		// had to think about
		// this one
		SetBitVector(pfExists,TRUE,SQL_API_SQLGETINFO);
		SetBitVector(pfExists,TRUE,SQL_API_SQLDESCRIBECOL);
		// pfExists[SQL_API_SQLGETSTMTOPTION]    = TRUE;  // very partial
		// pfExists[SQL_API_SQLGETTYPEINFO]      = TRUE;
		// pfExists[SQL_API_SQLPARAMDATA]        = FALSE;
		//pfExists[SQL_API_SQLPUTDATA]          = FALSE;
		//pfExists[SQL_API_SQLSETCONNECTOPTION] = TRUE;  // partial
		// pfExists[SQL_API_SQLSETSTMTOPTION]    = TRUE;
		// pfExists[SQL_API_SQLSPECIALCOLUMNS]   = TRUE;
		// pfExists[SQL_API_SQLSTATISTICS]       = TRUE;
		// pfExists[SQL_API_SQLTABLES]           = TRUE;
		
		// ODBC level 2 functions
		SetBitVector(pfExists,TRUE,SQL_API_SQLBROWSECONNECT);
		
		//pfExists[SQL_API_SQLCOLUMNPRIVILEGES] = FALSE;
		SetBitVector(pfExists,TRUE,SQL_API_SQLDATASOURCES);
		//pfExists[SQL_API_SQLDATASOURCES]      = FALSE;  // only implemented by DM
		// pfExists[SQL_API_SQLDESCRIBEPARAM]    = FALSE;
		SetBitVector(pfExists,TRUE,SQL_API_SQLDRIVERS);
		// pfExists[SQL_API_SQLDRIVERS]          = FALSE;
		// pfExists[SQL_API_SQLEXTENDEDFETCH]    = TRUE;
		// pfExists[SQL_API_SQLFOREIGNKEYS]      = FALSE;
		//pfExists[SQL_API_SQLMORERESULTS]      = TRUE;
		//pfExists[SQL_API_SQLNATIVESQL]        = TRUE;
		// pfExists[SQL_API_SQLNUMPARAMS]        = TRUE;
		// pfExists[SQL_API_SQLPARAMOPTIONS]     = FALSE;
		/*pfExists[SQL_API_SQLPRIMARYKEYS]      = FALSE;
		pfExists[SQL_API_SQLPROCEDURECOLUMNS] = FALSE;
		pfExists[SQL_API_SQLPROCEDURES]       = FALSE;
		pfExists[SQL_API_SQLSETPOS]           = FALSE;
		pfExists[SQL_API_SQLSETSCROLLOPTIONS] = FALSE;
		pfExists[SQL_API_SQLTABLEPRIVILEGES]  = FALSE;*/
		break;
	case SQL_API_SQLALLOCCONNECT:     *pfExists = TRUE; break;
	case SQL_API_SQLALLOCENV:         *pfExists = TRUE; break;
	case SQL_API_SQLALLOCSTMT:        *pfExists = TRUE; break;
	case SQL_API_SQLALLOCHANDLE:        *pfExists = TRUE; break;
	case SQL_API_SQLFREEHANDLE:        *pfExists = TRUE; break;
	case SQL_API_SQLBINDCOL:          *pfExists = TRUE; break;  // partial
	case SQL_API_SQLCANCEL:           *pfExists = FALSE; break;
	case SQL_API_SQLCOLATTRIBUTES:    *pfExists = FALSE; break;
	case SQL_API_SQLCONNECT:          *pfExists = TRUE; break;
	case SQL_API_SQLDESCRIBECOL:      *pfExists = TRUE; break;  // partial
	case SQL_API_SQLDISCONNECT:       *pfExists = TRUE; break;
	case SQL_API_SQLERROR:            *pfExists = FALSE; break;
	case SQL_API_SQLEXECDIRECT:       *pfExists = TRUE; break;
	case SQL_API_SQLEXECUTE:          *pfExists = FALSE; break;
	case SQL_API_SQLFETCH:            *pfExists = TRUE; break;
	case SQL_API_SQLFREECONNECT:      *pfExists = FALSE; break;
	case SQL_API_SQLFREEENV:          *pfExists = TRUE; break;
	case SQL_API_SQLFREESTMT:         *pfExists = TRUE; break;
	case SQL_API_SQLGETCURSORNAME:    *pfExists = FALSE; break;
	case SQL_API_SQLNUMRESULTCOLS:    *pfExists = TRUE; break;
	case SQL_API_SQLPREPARE:          *pfExists = FALSE; break;
	case SQL_API_SQLROWCOUNT:         *pfExists = FALSE; break;
	case SQL_API_SQLSETCURSORNAME:    *pfExists = FALSE; break;
	case SQL_API_SQLSETPARAM:         *pfExists = FALSE; break;
	case SQL_API_SQLTRANSACT:         *pfExists = FALSE; break;
		
		// ODBC level 1 functions
	case SQL_API_SQLBINDPARAMETER:    *pfExists = FALSE; break;
	case SQL_API_SQLCOLUMNS:          *pfExists = FALSE; break;
	case SQL_API_SQLDRIVERCONNECT:    *pfExists = TRUE; break;
	case SQL_API_SQLGETCONNECTOPTION: *pfExists = FALSE; break;  // partial
	case SQL_API_SQLGETDATA:          *pfExists = FALSE; break;
	case SQL_API_SQLGETFUNCTIONS:     *pfExists = TRUE; break;
	case SQL_API_SQLGETINFO:          *pfExists = TRUE; break;
	case SQL_API_SQLGETSTMTOPTION:    *pfExists = FALSE; break;  // very partial
	case SQL_API_SQLGETTYPEINFO:      *pfExists = FALSE; break;
	case SQL_API_SQLPARAMDATA:        *pfExists = FALSE; break;
	case SQL_API_SQLPUTDATA:          *pfExists = FALSE; break;
	case SQL_API_SQLSETCONNECTOPTION: *pfExists = FALSE; break;  // partial
	case SQL_API_SQLSETSTMTOPTION:    *pfExists = FALSE; break;
	case SQL_API_SQLSPECIALCOLUMNS:   *pfExists = FALSE; break;
	case SQL_API_SQLSTATISTICS:       *pfExists = FALSE; break;
	case SQL_API_SQLTABLES:           *pfExists = FALSE; break;
		
		// ODBC level 2 functions
	case SQL_API_SQLBROWSECONNECT:    *pfExists = FALSE; break;
	case SQL_API_SQLCOLUMNPRIVILEGES: *pfExists = FALSE; break;
	case SQL_API_SQLDATASOURCES:      *pfExists = FALSE; break;  // only implemented by DM
	case SQL_API_SQLDESCRIBEPARAM:    *pfExists = FALSE; break;
	case SQL_API_SQLDRIVERS:          *pfExists = FALSE; break;
	case SQL_API_SQLEXTENDEDFETCH:    *pfExists = FALSE; break;
	case SQL_API_SQLFOREIGNKEYS:      *pfExists = FALSE; break;
	case SQL_API_SQLMORERESULTS:      *pfExists = FALSE; break;
	case SQL_API_SQLNATIVESQL:        *pfExists = FALSE; break;
	case SQL_API_SQLNUMPARAMS:        *pfExists = FALSE; break;
	case SQL_API_SQLPARAMOPTIONS:     *pfExists = FALSE; break;
	case SQL_API_SQLPRIMARYKEYS:      *pfExists = FALSE; break;
	case SQL_API_SQLPROCEDURECOLUMNS: *pfExists = FALSE; break;
	case SQL_API_SQLPROCEDURES:       *pfExists = FALSE; break;
	case SQL_API_SQLSETPOS:           *pfExists = FALSE; break;
	case SQL_API_SQLSETSCROLLOPTIONS: *pfExists = FALSE; break;
	case SQL_API_SQLTABLEPRIVILEGES:  *pfExists = FALSE; break;
	case SQL_API_SQLGETDIAGREC:		  *pfExists = TRUE; break;
	case SQL_API_SQLGETSTMTATTR:*pfExists = TRUE; break;
	case SQL_API_SQLSETSTMTATTR:*pfExists = TRUE; break;
	default: return SQL_ERROR;
	}

	dump("Return Success");
	return SQL_SUCCESS;
			
}

/**
* SQLSetConnectAttr
*/
SQLRETURN   SQL_API SQLSetConnectAttr(
									  SQLHDBC			ConnectionHandle,
									  SQLINTEGER	Attribute,
									  SQLPOINTER	ValuePtr,
									  SQLINTEGER	StringLength)
{
	dump("\n SQLSetConnectAttr");
	return SQL_ERROR;
}


/**
* SQLGetConnectAttr
*/
SQLRETURN   SQL_API SQLGetConnectAttr(
									  SQLHDBC			ConnectionHandle,
									  SQLINTEGER	Attribute,
									  SQLPOINTER	ValuePtr,
									  SQLINTEGER	BufferLength,
									  SQLINTEGER *	StringLengthPtr)
{
	dump("\n SQLGetConnectAttr");
	return SQL_ERROR;
}


/**
* SQLGetEnvAttr
*/
SQLRETURN   SQL_API SQLGetEnvAttr(
								  SQLHENV			EnvironmentHandle,
								  SQLINTEGER	Attribute,
								  SQLPOINTER	ValuePtr,
								  SQLINTEGER	BufferLength,
								  SQLINTEGER *	StringLengthPtr)
{
	dump("\n SQLGetEnvAttr");
	return SQL_ERROR;
}

/** 
* SQLGetDiagRec
*	Return error message
*	Application calls this function until returning SQL_NO_DATA
*/
SQLRETURN SQL_API SQLGetDiagRec(
								SQLSMALLINT		HandleType,
								SQLHANDLE			Handle,
								SQLSMALLINT		RecNumber,
								SQLCHAR *			Sqlstate,
								SQLINTEGER *		NativeErrorPtr,
								SQLCHAR *			MessageText,
								SQLSMALLINT		BufferLength,
								SQLSMALLINT *	TextLengthPtr)
{	
	char info[50];
	dump(" Enter SQLGetDiagRec");
	sprintf(info,"Handle type %d",HandleType);
	dump(info);
	if(!Sqlstate)
		return SQL_ERROR;
	strcpy((char *)Sqlstate,"01000");
	if(!NativeErrorPtr) {
		NativeErrorPtr = new long;
	}
	*NativeErrorPtr = 0;
	switch (HandleType) {
	case SQL_HANDLE_DBC:
		dump("Connection Hanle");
		Conn *conn;
		conn = (Conn *)Handle;
		if ( conn->conn->errorGot == 1) {
			conn->conn->errorGot = 0;
			return SQL_NO_DATA;
		}
		if (!MessageText) 
			MessageText = (unsigned char *)new char[strlen("Connection error")+1];
		strcpy((char *)MessageText,"Connection error");
		*TextLengthPtr = strlen("Connection error");
		conn->conn->errorGot = 1;
		return SQL_SUCCESS;
	case SQL_HANDLE_STMT:
		dump("Statement Hanlde");
		Stmt * stmt;
		stmt = (Stmt *)Handle;
		if ( stmt->stmt->errorGot == 1) {
			stmt->stmt->errorGot = 0;
			dump("Erro info got!");
			return SQL_NO_DATA;
		}
		
		int length;
		if ( stmt->stmt->errorInfo) {
			stmt->stmt->errorInfo->Size(length);
			sprintf(info,"size %d,bufferleng %d",length,BufferLength);
			dump(info);
			if (!MessageText) 
				MessageText = (unsigned char *)new char[length];
			else if ( BufferLength < length)
				return SQL_NO_DATA;
			stmt->stmt->errorInfo->Print((char*)MessageText,"Statement Error");
			*TextLengthPtr = length;
			stmt->stmt->errorGot = 1;
			return SQL_SUCCESS;
		}
		else 
			return SQL_NO_DATA;
	case SQL_HANDLE_ENV: 
	case SQL_HANDLE_DESC :
		dump("Env Handle or Desc Handle, unimplemented yet");
		return SQL_NO_DATA;
	default: return SQL_ERROR;
	}
}

/** 
* SQLGetStmtAttr
*	Get statement attributes
*	Only implement returning desc handler
*/
SQLRETURN SQL_API SQLGetStmtAttr(
								 SQLHSTMT		StatementHandle,
								 SQLINTEGER	Attribute,
								 SQLPOINTER	ValuePtr,
								 SQLINTEGER	BufferLength,
								 SQLINTEGER *	StringLengthPtr)
{	
	char temp[60];
	Stmt * stmt;
	stmt = (Stmt *)StatementHandle;
	
	dump(" Enter SQLGetStmtAttr");
	sprintf(temp,"Get Attribute %d", Attribute);
	dump(temp);
	switch (Attribute) {
	case SQL_ATTR_APP_PARAM_DESC:
		dump(" Get ATTR_APP_PARAM_DESC");
		*StringLengthPtr = 4;
		*(SQLHDESC *)ValuePtr = (SQLHDESC)stmt->stmt->App_Param_Desc;
		break;
	case SQL_ATTR_APP_ROW_DESC: 
		dump("Get Attr_App_Row_Desc");
		*StringLengthPtr = 4;
		*(SQLHDESC *)ValuePtr = (SQLHDESC)stmt->stmt->App_Row_Desc;
		break;
	case SQL_ATTR_IMP_ROW_DESC:
		dump("Get Attr_Imp_Row_Desc");
		*StringLengthPtr = 4;
		*(SQLHDESC *)ValuePtr = (SQLHDESC)stmt->stmt->Imp_Row_Desc;
		break;
	case SQL_ATTR_IMP_PARAM_DESC:
		dump("Get Attr_Imp_Param_Desc");
		*StringLengthPtr = 4;
		*(SQLHDESC *)ValuePtr = (SQLHDESC)stmt->stmt->Imp_Param_Desc;
		break;
	default: dump("Sorry, current not supported attributes");
		return SQL_ERROR;
	}
	return SQL_SUCCESS;
}


/**
* SQLSetStmtAttr
*/
SQLRETURN SQL_API SQLSetStmtAttr(
								 SQLHSTMT		StatementHandle,
								 SQLINTEGER	Attribute,
								 SQLPOINTER	ValuePtr,
								 SQLINTEGER	StringLength)
{
	dump("Enter SetStmtAttr");
	return SQL_SUCCESS;
}


/**
* SQLDataSources
*/
SQLRETURN SQL_API SQLDataSources(
								 SQLHENV				EnvironmentHandle,
								 SQLUSMALLINT	Direction,
								 SQLCHAR *			ServerName,
								 SQLSMALLINT		BufferLength1,
								 SQLSMALLINT *	NameLength1Ptr,
								 SQLCHAR *			Description,
								 SQLSMALLINT		BufferLength2,
								 SQLSMALLINT *	NameLength2Ptr)
{
	dump("Enter SQLDataSources");
	return SQL_ERROR;
}


/**
* SQLGetTypeInfo
*/
SQLRETURN SQL_API SQLGetTypeInfo(
								 SQLHSTMT		StatementHandle,
								 SQLSMALLINT	DataType)
{
	dump("Enter SQLGETTYPEINFO");
	return SQL_ERROR;
}


/**
* SQLGetDescField
*/
SQLRETURN SQL_API SQLGetDescField(
								  SQLHDESC		DescriptorHandle,
								  SQLSMALLINT	RecNumber,
								  SQLSMALLINT	FieldIdentifier,
								  SQLPOINTER	ValuePtr,
								  SQLINTEGER	BufferLength,
								  SQLINTEGER *	StringLengthPtr)
{
	dump("Enter GetDescField");
	return SQL_ERROR;
}


/**
* SQLGetDescRec
*/
SQLRETURN SQL_API SQLGetDescRec(
								SQLHDESC			DescriptorHandle,
								SQLSMALLINT		RecNumber,
								SQLCHAR *			Name,
								SQLSMALLINT		BufferLength,
								SQLSMALLINT *	StringLengthPtr,
								SQLSMALLINT *	TypePtr,
								SQLSMALLINT *	SubTypePtr, 
								SQLINTEGER *		LengthPtr,
								SQLSMALLINT *	PrecisionPtr,
								SQLSMALLINT *	ScalePtr,
								SQLSMALLINT *	NullablePtr)
{
	dump("Enter GetDescRec");
	return SQL_ERROR;
}


/**
* SQLSetDescField
*/
SQLRETURN SQL_API SQLSetDescField(
								  SQLHDESC		DescriptorHandle,
								  SQLSMALLINT	RecNumber,
								  SQLSMALLINT	FieldIdentifier,
								  SQLPOINTER	ValuePtr,
								  SQLINTEGER	BufferLength)
{
	dump("Enter SetDescField");
	return SQL_ERROR;
}


/**
* SQLSetDescRec
*/
SQLRETURN SQL_API SQLSetDescRec(
								SQLHDESC			DescriptorHandle,
								SQLSMALLINT		RecNumber,
								SQLSMALLINT		Type,
								SQLSMALLINT		SubType,
								SQLINTEGER		Length,
								SQLSMALLINT		Precision,
								SQLSMALLINT		Scale,
								SQLPOINTER		DataPtr,
								SQLINTEGER *		StringLengthPtr,
								SQLINTEGER *		IndicatorPtr)
{
	dump("Enter SetDescRec");
	return SQL_ERROR;
}


/**
* SQLPrepare
*/
SQLRETURN SQL_API SQLPrepare(
							 SQLHSTMT		StatementHandle,
							 SQLCHAR *		StatementText,
							 SQLINTEGER	TextLength)
{
	dump("Enter SQLPrepare");
	return SQL_ERROR;
}


/**
* SQLBindParameter
*/
SQLRETURN SQL_API SQLBindParameter(
								   SQLHSTMT			StatementHandle,
								   SQLUSMALLINT	ParameterNumber,
								   SQLSMALLINT		InputOutputType,
								   SQLSMALLINT		ValueType,
								   SQLSMALLINT		ParameterType,
								   SQLUINTEGER		ColumnSize,
								   SQLSMALLINT		DecimalDigits,
								   SQLPOINTER		ParameterValuePtr,
								   SQLINTEGER		BufferLength,
								   SQLINTEGER *		StrLen_or_IndPtr)
{
	dump("Enter BindParameter");
	return SQL_ERROR;
}


/**
* SQLGetCursorName
*/
SQLRETURN SQL_API SQLGetCursorName(
								   SQLHSTMT			StatementHandle,
								   SQLCHAR *			CursorName,
								   SQLSMALLINT		BufferLength,
								   SQLSMALLINT *	NameLengthPtr)
{
	dump("Enter GetCursorName");
	return SQL_ERROR;
}


/**
* SQLSetCursorName
*/
SQLRETURN SQL_API SQLSetCursorName(
								   SQLHSTMT		StatementHandle,
								   SQLCHAR *		CursorName,
								   SQLSMALLINT	NameLength)
{
	dump("Enter SQLSetCUrsorName");
	return SQL_ERROR;
}


/**
* SQLExecute
*/
SQLRETURN SQL_API SQLExecute(
							 SQLHSTMT	StatementHandle)
{	
	dump("Enter SQLExecute");
	return SQL_ERROR;
}


/**
* SQLNativeSql
*/
SQLRETURN SQL_API SQLNativeSql(
							   SQLHDBC			ConnectionHandle,
							   SQLCHAR *		InStatementText,
							   SQLINTEGER	TextLength1,
							   SQLCHAR *		OutStatementText,
							   SQLINTEGER	BufferLength,
							   SQLINTEGER *	TextLength2Ptr)
{
	dump("Enter Nativesql");
	return SQL_ERROR;
}


/**
* SQLDescribeParam
*/
SQLRETURN SQL_API SQLDescribeParam(
								   SQLHSTMT			StatementHandle,
								   SQLUSMALLINT	ParameterNumber,
								   SQLSMALLINT *	DataTypePtr,
								   SQLUINTEGER *	ParameterSizePtr,
								   SQLSMALLINT *	DecimalDigitsPtr,
								   SQLSMALLINT *	NullablePtr)
{
	dump("DescribeParam");
	return SQL_ERROR;
}


/**
* SQLNumParams
*/
SQLRETURN SQL_API SQLNumParams(
							   SQLHSTMT			StatementHandle,
							   SQLSMALLINT *	ParameterCountPtr)
{
	dump("Enter SQlNumparams");
	return SQL_ERROR;
}


/**
* SQLParamData
*/
SQLRETURN SQL_API SQLParamData(
							   SQLHSTMT		StatementHandle,
							   SQLPOINTER *	ValuePtrPtr)
{
	dump("Enter SQLParamData");
	return SQL_ERROR;
}


/**
* SQLPutData
*/
SQLRETURN SQL_API SQLPutData(
							 SQLHSTMT		StatementHandle,
							 SQLPOINTER	DataPtr,
							 SQLINTEGER	StrLen_or_Ind)
{
	dump("Enter PutData");
	return SQL_ERROR;
}


/** 
* SQLRowCount
*/
SQLRETURN SQL_API SQLRowCount(
							  SQLHSTMT		StatementHandle,
							  SQLINTEGER *	RowCountPtr)
{
	dump("Enter RowCOunt");
	return SQL_ERROR;
}

/** 
* SQLNumResultCols
*	Return number of result columns
*	Get The schema of SQLNumResultCols and set the preprocess tag turn on
*/
SQLRETURN SQL_API SQLNumResultCols(
								   SQLHSTMT			statementHandle,
								   SQLSMALLINT *	ColumnCountPtr)
{
	dump("Enter SQLNumResultCols");
	Stmt * htmt;
	StatementHandle * hstmt;
	char info[200];
	SQLRETURN rc;
	
	if (statementHandle == NULL) {
		dump("\n Error! Null statement handle in sqlexecdirect");
		return SQL_ERROR;
	}
	htmt = (Stmt *) statementHandle;
	hstmt = htmt->stmt;
	// check whether preprocessed, if yes, 
	if ( hstmt->preProcessed ) {
		dump("already preProcessed ");
		if ( !hstmt->schema) { // Null schema, error occurs
			*ColumnCountPtr = 0;
			return SQL_SUCCESS;
		}
		sprintf(info,"NumCols is %d",hstmt->schema->RecArity);
		dump(info);
		*ColumnCountPtr = hstmt->schema->RecArity;
		return SQL_SUCCESS;
	}
	// Get Schema
	if ((rc = hstmt->GetSchema()) != SQL_SUCCESS) {
		if ( rc == SQL_NO_DATA) {// no result
			dump("No data");
			*ColumnCountPtr = 0;
			return SQL_SUCCESS;
		}
		dump("Error GetSchema");
		return SQL_ERROR;
	}
	hstmt->preProcessed = 1;
	sprintf(info,"NumCols is %d",hstmt->schema->RecArity);
	dump(info);
	if ( !hstmt->schema)
		*ColumnCountPtr = 0;
	else
		*ColumnCountPtr = hstmt->schema->RecArity;
	return SQL_SUCCESS;
	
}

/** 
* SQLDescribeCol
*	Not fully debugged
*/
SQLRETURN SQL_API SQLDescribeCol(
								 SQLHSTMT			statementHandle,
								 SQLSMALLINT		ColumnNumber,
								 SQLCHAR *			ColumnName,
								 SQLSMALLINT		BufferLength,
								 SQLSMALLINT *	nameLengthPtr,
								 SQLSMALLINT *	DataTypePtr,
								 SQLUINTEGER *		ColumnSizePtr,
								 SQLSMALLINT *	DecimalDigitsPtr,
								 SQLSMALLINT *	NullablePtr)
{
	dump("Enter DescribeCol");
	
	Stmt * stmt;
	StatementHandle * hstmt;
	
	if (statementHandle == NULL) {
		dump("\n Error! Null statement handle in sqlexecdirect");
		return SQL_ERROR;
	}
	stmt = (Stmt *) statementHandle;
	hstmt = stmt->stmt;
	
	// check whether preprocessed, if yes, 
	if ( hstmt->preProcessed ) {
		if ( !hstmt->schema) { // Null schema, error occurs
			return SQL_ERROR;
		}
	}
	else {
		// Get Schema
		if (hstmt->GetSchema() != SQL_SUCCESS)
			return SQL_ERROR;
		hstmt->preProcessed = 1;
		if ( !hstmt->schema) { // Null schema, error occurs
			return SQL_ERROR;
		}
	}
	// check if column number fall in range
	if ( ColumnNumber < 0 || ColumnNumber > hstmt->schema->RecArity)
		return SQL_ERROR;
	// Now describe columnNuma, with NameLengthPtr
	// copy the column name into
	if ( ColumnName)
		strncpy2((char *)ColumnName,hstmt->schema->Attrs[ColumnNumber-1].Name,BufferLength);
	// set NameLength
	*nameLengthPtr = strlen((char *)ColumnName);
	// get Typ
	int typeId = hstmt->schema->Attrs[ColumnNumber-1].Type;
	
	char * typeName = hstmt->connectionHandle->predConnection->ADTs[typeId];
	while (1) {
		if ( !strcmp(typeName,"int")) {
			*DataTypePtr = SQL_INTEGER;
			*ColumnSizePtr = sizeof(int);
			*DecimalDigitsPtr = 0;
			*NullablePtr = SQL_NO_NULLS;
			break;
		}
		if ( !strcmp(typeName,"string")) {
			*DataTypePtr = SQL_VARCHAR;
			*ColumnSizePtr = 0;
			*DecimalDigitsPtr = 0;
			*NullablePtr = SQL_NO_NULLS;
			break;
		}
		if ( !strcmp(typeName,"double")) {
			*DataTypePtr = SQL_DOUBLE;
			*ColumnSizePtr = sizeof(double);
			*DecimalDigitsPtr = 0;
			*NullablePtr = SQL_NO_NULLS;
			break;
		}
		if ( !strcmp(typeName,"boolean")) {
			*DataTypePtr = SQL_CHAR;
			*ColumnSizePtr = sizeof(char);
			*DecimalDigitsPtr = 0;
			*NullablePtr = SQL_NO_NULLS;
			break;
		}
		return SQL_ERROR;
	}
	return SQL_SUCCESS;
}


/**
* SQLColAttribute
*/ 
SQLRETURN SQL_API SQLColAttribute (
								   SQLHSTMT				StatementHandle,
								   SQLUSMALLINT		ColumnNumber,
								   SQLUSMALLINT		FieldIdentifier,
								   SQLPOINTER			CharacterAttributePtr,
								   SQLSMALLINT			BufferLength,
								   SQLSMALLINT *		StringLengthPtr,
								   SQLPOINTER			NumericAttributePtr)
{
	dump("Enter ColAttribute");
	return SQL_ERROR;
}


/**
* SQLFetchScroll
*/
SQLRETURN SQL_API SQLFetchScroll(
								 SQLHSTMT		StatementHandle,
								 SQLSMALLINT	FetchOrientation,
								 SQLINTEGER	FetchOffset)
{
	dump("Enter SQLFetchScroll");
	return SQL_ERROR;
}


/**
* SQLGetData
*/
SQLRETURN SQL_API SQLGetData(
							 SQLHSTMT			StatementHandle,
							 SQLUSMALLINT	ColumnNumber,
							 SQLSMALLINT		TargetType,
							 SQLPOINTER		TargetValuePtr,
							 SQLINTEGER		BufferLength,
							 SQLINTEGER *		StrLen_or_IndPtr)
{
	dump("Enter SQLGetData");
	return SQL_ERROR;
}


/**
* SQLSetPos
*/
SQLRETURN SQL_API SQLSetPos(
							SQLHSTMT			StatementHandle,
							SQLUSMALLINT	RowNumber,
							SQLUSMALLINT	Operation,
							SQLUSMALLINT	LockType)
{
	dump("Enter SetPos");
	return SQL_ERROR;
}


/**
* SQLBulkOperations
*/
SQLRETURN SQL_API SQLBulkOperations(
									SQLHSTMT			StatementHandle,
									SQLUSMALLINT	Operation)
{
	dump("Enter BulkOperations");
	return SQL_ERROR;
}


/**
* SQLMoreResults
*/
SQLRETURN SQL_API SQLMoreResults(
								 SQLHSTMT	StatementHandle)
{
	dump("Enter MoreResults");
	return SQL_ERROR;
}


/**
* SQLGetDiagField
*/
SQLRETURN SQL_API SQLGetDiagField(
								  SQLSMALLINT		HandleType,
								  SQLHANDLE			Handle,
								  SQLSMALLINT		RecNumber,
								  SQLSMALLINT		DiagIdentifier,
								  SQLPOINTER		DiagInfoPtr,
								  SQLSMALLINT		BufferLength,
								  SQLSMALLINT *	StringLengthPtr)
{
	dump("Enter GetDiagField");
	return SQL_ERROR;
}


/**
* SQLColumnPrivileges
*/
SQLRETURN SQL_API SQLColumnPrivileges(
									  SQLHSTMT		StatementHandle,
									  SQLCHAR *		CatalogName,
									  SQLSMALLINT	NameLength1,
									  SQLCHAR *		SchemaName,
									  SQLSMALLINT	NameLength2,
									  SQLCHAR *		TableName,
									  SQLSMALLINT	NameLength3,
									  SQLCHAR *		ColumnName,
									  SQLSMALLINT	NameLength4)
{
	dump("Enter ColumnPrivileges");
	return SQL_ERROR;
}


/**
* SQLColumns
*/
SQLRETURN SQL_API SQLColumns(
							 SQLHSTMT		StatementHandle,
							 SQLCHAR *		CatalogName,
							 SQLSMALLINT	NameLength1,
							 SQLCHAR *		SchemaName,
							 SQLSMALLINT	NameLength2,
							 SQLCHAR *		TableName,
							 SQLSMALLINT	NameLength3,
							 SQLCHAR *		ColumnName,
							 SQLSMALLINT	NameLength4)
{
	dump("Enter SQLColums");
	return SQL_ERROR;
}


/**
* SQLForeignKeys
*/
SQLRETURN SQL_API SQLForeignKeys(
								 SQLHSTMT		StatementHandle,
								 SQLCHAR *		PKCatalogName,
								 SQLSMALLINT	NameLength1,
								 SQLCHAR *		PKSchemaName,
								 SQLSMALLINT	NameLength2,
								 SQLCHAR *		PKTableName,
								 SQLSMALLINT	NameLength3,
								 SQLCHAR *		FKCatalogName,
								 SQLSMALLINT	NameLength4,
								 SQLCHAR *		FKSchemaName,
								 SQLSMALLINT	NameLength5,
								 SQLCHAR *		FKTableName,
								 SQLSMALLINT	NameLength6)
{
	dump("Enter SQLForeignKeys");
	return SQL_ERROR;
}


/**
* SQLPrimaryKeys
*/
SQLRETURN SQL_API SQLPrimaryKeys(
								 SQLHSTMT		StatementHandle,
								 SQLCHAR *		CatalogName,
								 SQLSMALLINT	NameLength1,
								 SQLCHAR *		SchemaName,
								 SQLSMALLINT	NameLength2,
								 SQLCHAR *		TableName,
								 SQLSMALLINT	NameLength3)
{
	dump("Enter PrimaryKeys");
	return SQL_ERROR;
}



/**
* SQLProcedureColumns
*/
SQLRETURN SQL_API SQLProcedureColumns(
									  SQLHSTMT		StatementHandle,
									  SQLCHAR *		CatalogName,
									  SQLSMALLINT	NameLength1,
									  SQLCHAR *		SchemaName,
									  SQLSMALLINT	NameLength2,
									  SQLCHAR *		ProcName,
									  SQLSMALLINT	NameLength3,
									  SQLCHAR *		ColumnName,
									  SQLSMALLINT	NameLength4)
{
	dump("Enter ProcedureColums");
	return SQL_ERROR;
}


/**
* SQLProcedures
*/
SQLRETURN SQL_API SQLProcedures(
								SQLHSTMT		StatementHandle,
								SQLCHAR *		CatalogName,
								SQLSMALLINT	NameLength1,
								SQLCHAR *		SchemaName,
								SQLSMALLINT	NameLength2,
								SQLCHAR *		ProcName,
								SQLSMALLINT	NameLength3)
{
	dump("Enter SQLProcedures");
	return SQL_ERROR;
}


/**
* SQLSpecialColumns
*/
SQLRETURN SQL_API SQLSpecialColumns(
									SQLHSTMT			StatementHandle,
									SQLSMALLINT		IdentifierType,
									SQLCHAR *			CatalogName,
									SQLSMALLINT		NameLength1,
									SQLCHAR *			SchemaName,
									SQLSMALLINT		NameLength2,
									SQLCHAR *			TableName,
									SQLSMALLINT		NameLength3,
									SQLSMALLINT		Scope,
									SQLSMALLINT		Nullable)
{
	dump("Enter SpecialColumns");
	return SQL_ERROR;
}


/**
*  SQLStatistics(
*/ 
SQLRETURN SQL_API SQLStatistics(
								SQLHSTMT			StatementHandle,
								SQLCHAR *			CatalogName,
								SQLSMALLINT		NameLength1,
								SQLCHAR *			SchemaName,
								SQLSMALLINT		NameLength2,
								SQLCHAR *			TableName,
								SQLSMALLINT		NameLength3,
								SQLUSMALLINT	Unique,
								SQLUSMALLINT	Reserved)
{
	dump("Enter Statistics");
	return SQL_ERROR;
}


/**
* SQLTablePrivileges
*/
SQLRETURN SQL_API SQLTablePrivileges(
									 SQLHSTMT		StatementHandle,
									 SQLCHAR *		CatalogName,
									 SQLSMALLINT	NameLength1,
									 SQLCHAR *		SchemaName,
									 SQLSMALLINT	NameLength2,
									 SQLCHAR *		TableName,
									 SQLSMALLINT	NameLength3)
{
	dump("Enter TablePrivileges");
	return SQL_ERROR;
}


/**
* SQLTables
*/
SQLRETURN SQL_API SQLTables(
							SQLHSTMT		StatementHandle,
							SQLCHAR *		CatalogName,
							SQLSMALLINT	NameLength1,
							SQLCHAR *		SchemaName,
							SQLSMALLINT	NameLength2,
							SQLCHAR *		TableName,
							SQLSMALLINT	NameLength3,
							SQLCHAR *		TableType,
							SQLSMALLINT	NameLength4)
{
	dump("Enter Tables");
	return SQL_ERROR;
}


/**
* SQLCloseCursor
*/
SQLRETURN SQL_API SQLCloseCursor(
								 SQLHSTMT	StatementHandle)
{
	dump("Enter CloseCursor");
	return SQL_SUCCESS;
}


/**
* SQLCancel
*/
SQLRETURN SQL_API SQLCancel(
							SQLHSTMT	StatementHandle)
{
	dump("Enter SQLCancel");
	return SQL_ERROR;
}



/**
* SQLEndTran
*/
SQLRETURN SQL_API SQLEndTran(
							 SQLSMALLINT	HandleType,
							 SQLHANDLE		Handle,
							 SQLSMALLINT	CompletionType)
{
	dump("Enter EndTran");
	return SQL_ERROR;
}












