#include "odbcclient.h"
#include <windows.h>
#include "sql.h"
#include "sqlext.h"


#ifdef XXX_WIN32
#define strcasecmp(X, Y) _stricmp(X, Y)
#endif
#define MaxCols 40
#define MaxDataLen 4096

const char* const XxxTxtClientEnv::DefaultPrompt = "Q> ";
const int XxxTxtClientEnv::DefaultPort = 6005;
const int XxxTxtClientEnv::DefaultTiming = 1;
const int XxxTxtClientEnv::DefaultVerbose = 1;
const bool XxxTxtClientEnv::DefaultPrintSchemas = true;
const bool XxxTxtClientEnv::DefaultPrintBinary = false;

void DisplayError(SQLRETURN nResult, SWORD fHandleType, SQLHANDLE handle)
{
	unsigned char    szErrState[10];    // SQL Error State string
	unsigned char    szErrText[4096];    // SQL Error Text string, be sure to leave it large enough, otherwise,
	// no error message will be returned.
	// formatted Error text Buffer
	SWORD    wErrMsgLen;             // Error message length
	long   dwErrCode;              // Native Error code
	//int  iSize;                      // Display Error Text size
	SQLRETURN  nErrResult;             // Return Code from SQLGetDiagRec
	SWORD sMsgNum = 1;
	
	do
	{
		// continue to bring messageboxes till all errors are displayed.
		// more than one message box may be reqd. as err text has fixed
		// string size.
		
		// call SQLGetDiagRec function with proper ODBC handles, repeatedly until
		// function returns SQL_NO_DATA. Concatenate all error strings
		// in the display buffer and display all results.
		while ((nErrResult = SQLGetDiagRec(fHandleType, handle, sMsgNum++,
			szErrState, &dwErrCode, szErrText,
			4095, &wErrMsgLen))
			!= SQL_NO_DATA) {
			
			
			if(nErrResult == SQL_ERROR || nErrResult == SQL_INVALID_HANDLE)
				break;
			cout << szErrState << szErrText << endl;
		}
		
		// display proper ERROR or WARNING message with proper title
		
		
	}
	while (!(nErrResult == SQL_NO_DATA || nErrResult == SQL_ERROR || nErrResult == SQL_INVALID_HANDLE));
}

// Returns 1 if the string is terminated by
// a semicolon (possibly followed by whitespace).
// Otherwise returns 0.
inline int SemiColonTerminated(const char* Buf)
{
	// if this is a comment line, dont count it as a terminated line
	if (Buf[0] == '#') return 0; 
	if (Buf[0] == '%') return 0; 
	
	int BufLen = strlen(Buf);
	for (int i = BufLen-1; i >=0; i--) {
		if (Buf[i] == ';') return 1;
		if (!isspace(Buf[i])) return 0;
	}
	return 0;
}

#ifndef XXX_READLINE
char* readline(char *Prompt)
{
	static char Buf[XXX_INPUT_LINEBUFFER_MAX];
	
	cout << Prompt;
	
	cin.getline(Buf,XXX_INPUT_LINEBUFFER_MAX);
	
	if (!cin.good()) return NULL;
	
	return Buf;
}
#endif

int IStreamList::GetCommand(char* Buf, int BufLength, int& CommandLength)
{
	while (1) {
		while (Next && !Next->Stream->good()) {
			IStreamList* Tmp = Next->Next;
			Next->Next = NULL;
			delete Next;
			Next = Tmp;
			Count--;
		}
		if (Empty() || !Next) return 0;
		
		//if (Next->Stream->get(Buf, BufLength, ';') && (Next->Stream->good()) &&
		if (Next->Stream->get(Buf, BufLength, '\n') && (!Next->Stream->bad()) &&
			(CommandLength = Next->Stream->gcount()))
		{
			// return this command
			// get the ; out of the istream
			//Next->Stream->get();
			//Buf[CommandLength++] = ';';
			//Buf[CommandLength] = 0;
			return 1;
		}
		// else try again!!
	}
}

//
// this is supposed to return a List of IStreams that contains all the strings
// in this particular user request
//
IStreamList* XxxTxtClientEnv::GetInput(int &Status)
{
	char *Buf1;
	IStreamList *InputList = new IStreamList();
	while (1) {
		if (!(Buf1 = readline(Prompt))) {
			Status = XXX_CL_QUIT;
			delete InputList;
			return NULL;
		}
		
#ifdef XXX_READLINE
		add_history(Buf1);
#endif
		
		if (!strncmp(Buf1, XXX_CL_QUITSTR, strlen(XXX_CL_QUITSTR))) {
			Status = XXX_CL_QUIT;
			delete InputList;
			return NULL;
		}
		
		
		if (!strncmp(Buf1, XXX_CL_CONSULTSTR, strlen(XXX_CL_CONSULTSTR))) {
			// get rid of trailing semicolon
			if (Buf1[strlen(Buf1)-1] == ';') Buf1[strlen(Buf1)-1] = 0;
			// account for starting white space
			char *FileName = Buf1+strlen(XXX_CL_CONSULTSTR) + 1;
			ifstream* InputFile = new ifstream(FileName);
			if (!InputFile || !InputFile->good()) {
				cout << "txtclient::Bad input file name " << FileName << endl;
				if (InputFile) delete InputFile;
			}
			else {
				InputList->Add(InputFile);
			} 
		}
		else if (!strncmp(Buf1, XXX_CL_SHELLSTR, strlen(XXX_CL_SHELLSTR))) {
			// get rid of trailing semicolon
			if (Buf1[strlen(Buf1)-1] == ';') Buf1[strlen(Buf1)-1] = 0;
			// account for starting white space
			char *SysCommand = Buf1+strlen(XXX_CL_SHELLSTR) + 1;
			system(SysCommand);
		}
		
		else if (!strncmp(Buf1, XXX_CL_LABELSTR, strlen(XXX_CL_LABELSTR))) {
			// get rid of trailing semicolon
			if (Buf1[strlen(Buf1)-1] == ';') Buf1[strlen(Buf1)-1] = 0;
			// account for starting white space
			char *TLabel = Buf1+strlen(XXX_CL_LABELSTR) + 1;
			strcpy(TimerLabel, TLabel);
		}
		
		else if (!strncmp(Buf1, XXX_CL_LANGSTR, strlen(XXX_CL_LANGSTR))) {
			// get rid of trailing semicolon
			if (Buf1[strlen(Buf1)-1] == ';') Buf1[strlen(Buf1)-1] = 0;
			else Buf1[strlen(Buf1)] = 0;
			char *NewLang = Buf1+strlen(XXX_CL_LANGSTR) + 1;
			// account for starting white space
			while (*NewLang && isspace(*NewLang)) NewLang++;
			if (*NewLang) {
				strcpy(CurrentLang, NewLang);
				//strcat(CurrentLang, ":");
				cout << "txtclient::Current Language set to:" << CurrentLang << endl;
			}
			else {
				cout << "txtclient::Current Language:" << CurrentLang << endl;
			}
		}
		
		else if (!strncmp(Buf1, XXX_CL_DEFSTR, strlen(XXX_CL_DEFSTR))) {
			PrintDefaults(cout);
		}
		
		else if (!strncmp(Buf1, XXX_CL_OUTFSTR, strlen(XXX_CL_OUTFSTR))) {
			cout << "txtclient::Havent implemented o/p redirection yet" << endl;
		}
		
		else if (!strncmp(Buf1, XXX_CL_TIMERFSTR, strlen(XXX_CL_TIMERFSTR))) {
			// get rid of trailing semicolon
			if (Buf1[strlen(Buf1)-1] == ';') Buf1[strlen(Buf1)-1] = 0;
			// account for starting white space
			char *TimerFileName = Buf1+strlen(XXX_CL_TIMERFSTR) + 1;
			TimerStr = new ofstream(TimerFileName);
		}
		
		else if (!strncmp(Buf1, XXX_CL_HELPSTR, strlen(XXX_CL_HELPSTR))) {
			PrintCommandsHelp(cout);
		}
		
		else if (!strncmp(Buf1, XXX_CL_TOGGLESTR, strlen(XXX_CL_TOGGLESTR))) {
			// account for starting white space
			char *SysCommand = Buf1+strlen(XXX_CL_SHELLSTR) + 1;
			while (*SysCommand && isspace(*SysCommand)) SysCommand++;
			if (*SysCommand) {
				if (!strcasecmp(SysCommand, "timing"))
					TimingOn = !TimingOn;
				else if (!strcasecmp(SysCommand, "schema"))
					PrintSchemas = !PrintSchemas;
				else if (!strcasecmp(SysCommand, "binary"))
					PrintBinary = !PrintBinary;
			}
		}
		else {
			// have to reallocate buffer, since gnu readline tends to
			// reuse allocated memory
			// This is probably a memory leak!
			char* NewBuf = new char[strlen(Buf1)+1];
			memcpy(NewBuf, Buf1, strlen(Buf1)+1);
			InputList->Add(new istrstream(NewBuf));
			// new addition
			if (SemiColonTerminated(NewBuf)) {
				// here's the primary return point
				Status = XXX_CL_OK;
				return InputList;
			}
		} // if not a special command keyword
  } // while (1)
}

/****
-h <hostname> : server hostname
-p <prompt>   : client prompt string
-s <portnum>  : server socket port
-u <username> : client's user's name
-r            : schemas printed with tables
-t            : timing on
****/
void XxxTxtClientEnv::PrintFlagsHelp(ostream& OStr) const
{
	OStr << " PREDATOR Textual Client" << endl;
	OStr << "    -h <server host>" << endl;
	OStr << "    -u <user name>" << endl;
	OStr << "    -l <top language>" << endl;
	OStr << "    -p <prompt>" << endl;
	OStr << "    -s <server port>" << endl;
	OStr << "    -t <0/1> timing off/on" << endl;
	OStr << "    -r <0/1> schema printing off/on" << endl;
	OStr << "    -b <0/1> binary printing off/on" << endl;
}

void XxxTxtClientEnv::PrintCommandsHelp(ostream& OStr) const
{
	OStr << " XXX Textual Client Commands" << endl;
	OStr << "    " << XXX_CL_CONSULTSTR << " <filename>" << endl;
	OStr << "    " << XXX_CL_DEFSTR << " : show default settings" << endl;
	OStr << "    " << XXX_CL_HELPSTR << " : list of commands" << endl;
	OStr << "    " << XXX_CL_LANGSTR << " <language name>  : set the top-level language" << endl;
	OStr << "    " << XXX_CL_LABELSTR << " <timer label>  : set the label for timer info" << endl;
	OStr << "    " << XXX_CL_OUTFSTR << " <outfile name>  : set the answer o/p file" << endl;
	OStr << "    " << XXX_CL_QUITSTR << " (or CTRL-D)  : exit system" << endl;
	OStr << "    " << XXX_CL_SHELLSTR << " <shell command>" << endl;
	OStr << "    " << XXX_CL_TIMERFSTR << " <timerfile name>  : set the timer o/p file" << endl;
	OStr << "    " << XXX_CL_TOGGLESTR << " <timing/schema printing>" << endl;
}

void XxxTxtClientEnv::PrintDefaults(ostream& OStr) const
{
	OStr << " XXX Textual Client Defaults" << endl;
	OStr << "    Server Host  : " << HostName << endl;
	OStr << "    Server Port  : " << ServerPort << endl;
	OStr << "    User Name    : " << UserName << endl;
	OStr << "    Top Language : " << CurrentLang << endl;
	OStr << "    User Prompt  : " << Prompt << endl;
	OStr << "    Output File  : " << "cout" << endl;
	OStr << "    Timing       : " << (TimingOn?"on":"off") << endl;
	OStr << "    TimerLabel   : " << TimerLabel << endl;
	OStr << "    Schemas are  : " << (PrintSchemas?"printed":"not printed") << endl;
	OStr << "    Printing  : " << (PrintBinary?"binary":"ASCII") << endl;
}

void XxxTxtClientEnv::ReadCommandLine(int argc, char**argv)
{
	for (int i = 1; i < argc; i++){
		char *flag = argv[i];
		if (flag[0] != '-') {
			cout <<  "txtclient::Bad command line flag " << argv[i] << endl;
			continue;
		}
		switch (flag[1]) {
		case 'h' :
			if (i+1 < argc)
				strcpy(HostName, argv[++i]);
			else 
				cout << "txtclient::Bad command line flag " <<  flag << endl;
			break;
			
		case 'u' :
			if (i+1 < argc)
				strcpy(UserName, argv[++i]);
			else 
				cout << "txtclient::Bad command line flag " <<  flag << endl;
			break;
			
		case 'l' :
			if (i+1 < argc)
				strcpy(CurrentLang, argv[++i]);
			else 
				cout << "txtclient::Bad command line flag " << flag << endl;
			break;
			
		case 'p' :
			if (i+1 < argc)
				strcpy(Prompt, argv[++i]);
			else 
				cout << "txtclient::Bad command line flag " <<  flag << endl;
			break;
			
		case 's' :
			if (i+1 < argc)
				ServerPort = atoi(argv[++i]);
			else 
				cout << "txtclient::Bad command line flag " << flag << endl;
			break;
			
		case 't' :
			if (i+1 < argc)
				TimingOn = atoi(argv[++i]);
			else 
				cout << "txtclient::Bad command line flag " << flag << endl;
			break;
			
		case 'r' :
			if (i+1 < argc)
				PrintSchemas = (atoi(argv[++i])?true:false);
			else 
				cout << "txtclient::Bad command line flag " << flag << endl;
			break;
			
		case 'b' :
			if (i+1 < argc)
				PrintBinary = (atoi(argv[++i]) ? true : false);
			else 
				cout << "txtclient::Bad command line flag " << flag << endl;
			break;
		case 'd' :
			if ( i+1 < argc)
				strcpy(DataSource,argv[++i]);
			else
				cout << "odbcclient: Bad Command line flag" << flag << endl;
			break;
		default :
			PrintFlagsHelp(cout);
			exit(EXIT_FAILURE);
		}
	}
	
}

void XxxTxtClientEnv::TimerStart()
{
#ifdef XXX_WIN32
	Time = CTime::GetCurrentTime();
#else
	struct timezone Temp;
	gettimeofday(&Time, &Temp);
#endif
}

void XxxTxtClientEnv::TimerDisplay(ostream& OStr)
{
	OStr << TimerLabel << " ";
#ifdef XXX_WIN32
	CTime NewTime = CTime::GetCurrentTime();
	
	CTimeSpan TSpan = NewTime - Time;
	OStr << TSpan.GetTotalSeconds();
#else
	struct timeval NewTime;
	struct timezone Temp;
	gettimeofday(&NewTime, &Temp);
	
	OStr << (NewTime.tv_sec-Time.tv_sec)+(NewTime.tv_usec-Time.tv_usec)/1000000.0;
#endif
	OStr << " secs" << endl;
}


main(int argc, char**argv)
{
	char results[MaxCols][MaxDataLen];
	long lenResults[MaxCols];
	cout << "Usage: odbcclient -d datasourcename -s portnumber -l SQL" << endl;
	XxxTxtClientEnv* Env = new XxxTxtClientEnv(argc, argv);
	// cout << "providing : binary " << (int)Env->PrintBinary ;
	SQLHENV hdenv;
	HDBC hdbc;
	HSTMT hsmt;
	SQLRETURN retcode;
	unsigned char sqlstate[100], msgText[300];
	char datasource[100];
	char user[100];
	char authentication[100];
	
	unsigned char instr[100];
	SQLSMALLINT len1,len2, *len3;
	len3 = new short;
	long nativeErrorPtr;
	short textLengthPtr;
	if (!Env->DataSource) {
		cout << "No data source specified!" << endl;
		exit(2);
	}
	
	strcpy(datasource,Env->DataSource);
	strcpy(user,"zhiyuan");
	strcpy(authentication,"aabc");
	retcode = SQLAllocHandle(SQL_HANDLE_ENV, SQL_NULL_HANDLE, &hdenv);
	if ( retcode != SQL_SUCCESS)
		exit(0);
	retcode = SQLSetEnvAttr(hdenv, SQL_ATTR_ODBC_VERSION,(SQLPOINTER) SQL_OV_ODBC3, SQL_IS_INTEGER);
	if ( retcode != SQL_SUCCESS)
		exit(0);
	retcode = SQLAllocHandle(SQL_HANDLE_DBC, hdenv, &hdbc);
	if ( retcode != SQL_SUCCESS)
		exit(0);
	strcpy((char *)instr,"DRIVER=Dr. DeeBee Sample Driver 32;DSN=Dr.;UID=praveen;PWD=aaaa");
	len1 = strlen((char *)instr)+1;
	len2 = 200;
	//	retcode = SQLDriverConnect(hdbc,NULL,(SQLCHAR *)instr,SQL_NTS,NULL,0,NULL,SQL_DRIVER_COMPLETE);
	//	retcode = SQLBrowseConnect(hdbc,(SQLCHAR *)instr,SQL_NTS,NULL,0,NULL);
	//	if ( retcode != SQL_SUCCESS)
	//		exit(0);
	retcode = SQLConnect(hdbc, (SQLCHAR *)datasource , strlen(datasource), (SQLCHAR *)user, strlen(user), (SQLCHAR *)authentication, strlen(authentication));
	   if (retcode != SQL_SUCCESS ) {
		   
		   retcode = SQLGetDiagRec(SQL_HANDLE_DBC,hdbc,1,	sqlstate,&nativeErrorPtr,msgText,300,&textLengthPtr);
		   
		   
	   }
	   //	cout << "Shutting down connection ... " ;
	   
	   
	   IStreamList* InputList;
	   int length;
	   char Buf[XXX_INPUT_BUFFER_MAX];
	   while (1) {
		   Buf[0]= 0;
		   int Status;
		   InputList = Env->GetInput(Status);
		   if (Status == XXX_CL_OK) {
			   int BytesToWrite;
			   char Buf1[XXX_INPUT_BUFFER_MAX];
			   Buf1[0] = 0;
			   //    strcpy(Buf1, Env->CurrentLang);
			   //  strcat(Buf1, ":");
			   
			   while (InputList->GetCommand(Buf, XXX_INPUT_BUFFER_MAX, 
				   BytesToWrite)) {
				   // check that there is no overflow
				   if ((length = strlen(Buf1) + strlen(Buf) + 2) >= XXX_INPUT_BUFFER_MAX) {
					   cout << "txtclient::Overflowing txtclient input buffer!" << endl;
					   exit(EXIT_FAILURE);
				   }
				   strcat(Buf1, Buf);
				   // tack on an extra white space instead of the newline
				   // strcat(Buf1, " ");
				   strcat(Buf1, "\n");
			   }
			   retcode = SQLAllocHandle(SQL_HANDLE_STMT, hdbc, &hsmt);
			   if (retcode != SQL_SUCCESS ) {
				   
				   retcode = SQLGetDiagRec(SQL_HANDLE_DBC,hdbc,1,	sqlstate,&nativeErrorPtr,msgText,100,&textLengthPtr);
				   
			   }
			   if (	retcode = SQLExecDirect(hsmt,(SQLCHAR *)Buf1,(SQLINTEGER)(length))) {
				   cout << "txtclient::Error sending request to server" << endl;
				   exit(EXIT_FAILURE);
			   }
			   delete InputList;
		   }
		   else { // XXX_CL_QUIT
			   if (InputList) delete InputList;
			   break;
		   }
		   
		   if (Env->TimingOn) Env->TimerStart();
		   
		   // print result:
		   short nCols;
		   // retriev number of columns
		   // If error, usually means sqlexec meets an error. Check command
		   retcode = SQLNumResultCols(hsmt, &nCols);
		   if (retcode == SQL_NO_DATA) 
		   {
		   }
		   else if (retcode == SQL_SUCCESS) 
		   {
			   if ( nCols == 0 || nCols > MaxCols)
			   {
				   // Close the open result set.
				   SQLCloseCursor(hsmt);
				   
			   }
			   else 
			   {
				   // SQLBindColumn, however, just bind all of them to char
				   for(int nCount=0; nCount<nCols; nCount++) 
				   {
					   retcode = SQLBindCol(hsmt, (nCount+1), SQL_C_CHAR, results[nCount], MaxDataLen, &lenResults[nCount]);
					   if (retcode != SQL_SUCCESS) 
					   {
						   cout << "txtclient::Error binding columns of results" << endl;
						   exit(EXIT_FAILURE);				
					   }	
				   }
				   // Now getting result lines and keeping printing them out
				   for(int nRows = 0; (retcode = SQLFetch(hsmt))==SQL_SUCCESS || retcode == SQL_SUCCESS_WITH_INFO;) 
				   {
					   if (retcode != SQL_SUCCESS)
						   DisplayError(retcode, SQL_HANDLE_STMT, hsmt);
					   for(nCount=0; nCount<nCols; nCount++) 
					   {
						   // check if the column is a null value?
						   cout << results[nCount] << '\t';
					   }
					   cout << endl;
				   }
				   
			   } // else
			   
		   } 
		   else 
		   {
			   // retcode == SQL_ERROR
			   cout << "txtclient::Error in executing, check commands!" << endl;
			   DisplayError(retcode, SQL_HANDLE_STMT, hsmt);
			   SQLFreeHandle(SQL_HANDLE_STMT,hsmt);
			   continue;
		   }
		   
		   if (Env->TimingOn)
			   Env->TimerDisplay(*(Env->TimerStr));
		   SQLFreeHandle(SQL_HANDLE_STMT,hsmt);
	   } // while (1)
	   
	   cout << "Shutting down connection ... " ;
	   
	   SQLDisconnect(hdbc);
	   SQLFreeHandle(SQL_HANDLE_ENV,hdenv);
	   cout <<  "goodbye" << endl;
	   delete Env;
	   
	   return 0;
}




