#ifdef __GNUG__
#pragma implementation
#endif

#include "xxx_thread.h"
#include "reladt.h"
#include "relopt.h"
#include "relquery.h"
#include "relcat.h"
#include "releng.h"
#include "reldata.h"
#include "relexec.h"
#include "commands.h"
#include "relation.h"

// static field
const char* const XxxRelADT::RelADTEnv::RELSTATS = "RelStatsPtr";

XxxBasicType RelationTypeId;
// The constructor must do all the setup necessary to make this 
// type usable. This includes setting up catalogs.
XxxRelADT::XxxRelADT(XxxBasicType TypeIden) : 
XxxADTClass(TypeIden, "rel", 40, // need a better way to get the max size
	    XXX_FALSE, // stored in place in a record
	    XXX_TRUE)  // maintains catalogs
{ 
 // it is important to initialize this global here!!
 // a lot of the relation  code uses it as an index into the
 // the XxxADT array to get a handle on the relation type object.
 RelationTypeId = TypeIden;
 DataEngines = NULL;
 QPEngines = NULL;
}

XxxRelADT::~XxxRelADT() 
{
}

int XxxRelADT::TypeSize(const XxxADTValue* Object, 
			const XxxADTMetaInfo* MetaInfo) const
{
#ifdef XXX_STATS
 RELSTATS()->TypeSize++;
#endif
 return DataEngines->GetHandleSize((const RelHandle*)Object);
}

int XxxRelADT::TypeCopy(const XxxADTValue* Object, XxxADTValue *ResultObject,
			const XxxADTMetaInfo* MetaInfo) const
{
#ifdef XXX_STATS
 RELSTATS()->TypeCopy++;
#endif
 return DataEngines->CopyHandle((const RelHandle*)Object,
				(RelHandle*)ResultObject,
				(RelCatalogInfo*)MetaInfo);
}
// need to modify this to check for storable implementation
int XxxRelADT::TypeStoreCopy(const XxxADTValue* Object, 
			     XxxADTValue *ResultObject,
			     const XxxADTMetaInfo* MetaInfo) const
{
#ifdef XXX_STATS
 RELSTATS()->TypeStoreCopy++;
#endif
 return DataEngines->CopyHandle((const RelHandle*)Object,
				(RelHandle*)ResultObject,
				(RelCatalogInfo*)MetaInfo);
}

void XxxRelADT::TypeDestroy(XxxADTValue *Object, const XxxADTMetaInfo* MetaInfo) const
{
#ifdef XXX_STATS
 RELSTATS()->TypeDestroy++;
#endif
 DataEngines->RemoveHandle((const RelHandle*)Object,
			   (RelCatalogInfo*)MetaInfo);
}

XxxErrCode XxxRelADT::WriteText(ostream& OutStream, 
				const XxxADTValue* Object,
				const XxxADTMetaInfo* MetaInfo) const
{
#ifdef XXX_STATS
 RELSTATS()->WriteText++;
#endif
 OutStream << " { " << endl;
 if (!Object || !MetaInfo) {
   REL_NOTE_ERROR(REL_BADEXEC, 
		  "RelADT:WriteText called without Relation or Schema");
   return XXX_NOT_OK;
  }
 XxxRelation* Rel;
 XxxRecord* RelRecord;
 RelHandle* Hndl = (RelHandle*)Object;
 RelCatalogInfo* CatInfo = (RelCatalogInfo*) MetaInfo;
 // get the relation
 if (DataEngines->GetRelation(Hndl, Rel, CatInfo->ImplInfo()) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, 
		  "RelADT:WriteText called with bad relation handle");
   return XXX_NOT_OK;
  }
 Rel->CatalogInfo = CatInfo;

 // open a cursor on the relation
 XxxSetCursor RCsr;
 if (Rel->InitCursor(RCsr) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, "Unable to open cursor");
   return XXX_NOT_OK;
  }

 // write each tuple 
 while ((Rel->NextItem(RCsr, RelRecord) == XXX_OK) && 
	(RCsr.IsValid() == XXX_TRUE)) {
   RelRecord->Print(OutStream, CatInfo->RecordSchema());
   // OutStream << endl;
  }

 // close the cursor
 if (Rel->CloseCursor(RCsr) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, "Unable to open cursor");
   return XXX_NOT_OK;
  }

 OutStream << " } ";
 return XXX_OK;
}

XxxErrCode XxxRelADT::ReadText(istream& InStream, 
			       XxxADTValue* Object,
			       const XxxADTMetaInfo* MetaInfo) const
{
#ifdef XXX_STATS
 RELSTATS()->ReadText++;
#endif
 if (!MetaInfo) {
   REL_NOTE_ERROR(REL_BADCAT, "Empty catalog info for ReadText");
   return XXX_NOT_OK;
  }
 // call function loadRelation 
 RelHandle *Handle;
 if (DataEngines->LoadRelation(NULL, InStream, 
			       (RelCatalogInfo*)MetaInfo,
			       Handle, XXX_TRUE) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADTXTFILE, "RelADT:ReadText failed");
   return XXX_NOT_OK;
  }
 // copy Handle into Object
 memcpy((char*)Object, (char *)(Handle), TypeSize(Handle, MetaInfo));
 return XXX_OK;
}
  
XxxErrCode XxxRelADT::AddMetaInfo(const char* Name, const XxxBool Named,
				  const char* MetaInfoString,
				  MetaInfoHandle& MIHandle)
{
#ifdef XXX_STATS
 RELSTATS()->AddMetaInfo++;
#endif
 XxxErrCode rc = XXX_OK;

 // first parse the string
 if (!QPEngines || !DataEngines) return XXX_NOT_OK;

 // For now, just store the name in MIHandle.
 // Later,change this to get a handle from AddMetaInfo and
 // store that handle in MIHandle
 strncpy(MIHandle.Data, Name, XXX_MIHNDL_MAX);
 MIHandle.Data[XXX_MIHNDL_MAX-1] = 0;
 MIHandle.Size = strlen(Name)+1;
 ASSERT((MIHandle.Size > 1)); // because it contains the name
 if (MIHandle.Size > (unsigned int)XXX_MIHNDL_MAX) {
   MIHandle.Size = XXX_MIHNDL_MAX;
   MIHandle.Data[XXX_MIHNDL_MAX-1] = 0;
  }

 XxxGenDataSrc::AttrTypeList* Attrs;
 XxxGenDataSrc::KeyList* Keys;
 char* DevName;
 char *RelImplType;
 RelSrcTypes SrcType;
 if (QPEngines->ParseMetaInfo(MetaInfoString, Attrs, Keys, 
			      (int*)(&SrcType), DevName, 
			      RelImplType) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADCAT, "Unable to parse catalog info");
   return XXX_NOT_OK;
  }
 rc = DataEngines->AddCatInfo(Name, Named, Attrs, Keys, SrcType,
			      DevName, RelImplType);
 if (DevName) delete DevName;
 if (RelImplType) delete RelImplType;
 // what about Attrs and Keys
 if (Attrs) delete Attrs;
 if (Keys) delete Keys;

 return rc;
}

//Bjarke, aug. 8: const added to char* parameter
XxxErrCode XxxRelADT::GetMetaInfo(const char* ObjName, // ignore this (defunct)
				  const MetaInfoHandle& MIHandle,
				  XxxADTMetaInfo* &CatInfo,
				  XxxBool& Found) const
{
#ifdef XXX_STATS
 RELSTATS()->GetMetaInfo++;
#endif
 ASSERT((MIHandle.Size > 1)); // because it contains the name
 const char* ObjectName = MIHandle.Data;

 RelCatalogInfo* TmpCatInfo = new RelCatalogInfo(ObjectName);

 if (DataEngines->GetCatInfo(ObjectName, *TmpCatInfo, Found) != XXX_OK) {
   delete TmpCatInfo;
   Found = XXX_FALSE;
   return XXX_NOT_OK;
  }
 if (Found == XXX_TRUE)
   CatInfo = TmpCatInfo;
 else
   CatInfo = NULL;

 return XXX_OK;
}

XxxErrCode XxxRelADT::RegisterQPEngine(XxxRelationEngine* QPEng)
{
 QPEngines = QPEng;
 return XXX_OK;
}

XxxErrCode XxxRelADT::RegisterDataEngine(RelDataEngine* DEng)
{
 DataEngines = DEng;
 return XXX_OK;
}

XxxBool XxxRelADT::CanCastTo(XxxBasicType CastType, 
			     const XxxADTMetaInfo* SourceMetaInfo,
			     const XxxADTMetaInfo* TargetMetaInfo)
{
#ifdef XXX_STATS
 RELSTATS()->CastCheck++;
#endif
 if (!SourceMetaInfo) {
   XXX_NOTE_ERROR(XXX_BADEXEC, "MetaInfo should not be NULL here !");
   return XXX_FALSE;
  }
 RelCatalogInfo* CatInfo = (RelCatalogInfo*)SourceMetaInfo;
 XxxRecordSchema* Schema = CatInfo->RecordSchema();
 if (Schema->Arity() != 1) 
   return XXX_FALSE;

 return (Schema->GetAttr(0)->Type == CastType);
}

XxxErrCode XxxRelADT::CastTo(XxxBasicType CastType, 
			     const XxxADTMetaInfo* SourceMetaInfo,
			     const XxxADTMetaInfo* TargetMetaInfo,
			     XxxADTValue* OldObject, 
			     XxxADTValue* NewObject)
{
#ifdef XXX_STATS
 RELSTATS()->CastExec++;
#endif
 if (!SourceMetaInfo) {
   XXX_NOTE_ERROR(XXX_BADEXEC, "MetaInfo should not be NULL here !");
   return XXX_NOT_OK;
  }
 RelCatalogInfo* CatInfo = (RelCatalogInfo*)SourceMetaInfo;
 XxxRecordSchema* Schema = CatInfo->RecordSchema();
 if (Schema->Arity() != 1) {
   XXX_NOTE_ERROR(XXX_BADEXEC, "Arity should be 1");
   return XXX_NOT_OK;
  }

 // Step 1: get a relation out of the OldObject
 XxxRelation* Rel;
 if (DataEngines->GetRelation((RelHandle*) OldObject, 
			      Rel, CatInfo->ImplInfo()) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, "Handle does not correspond to any relation");
   return XXX_NOT_OK;
  }
 Rel->CatalogInfo = CatInfo;

 // Step 2: open a cursor on the relation
 XxxSetCursor RelCsr;
 if (Rel->InitCursor(RelCsr) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, "Unable to open cursor");
   return XXX_NOT_OK;
  }

 // Step 3: get a record from the relation
 XxxRecord* RelRecord;
 if ((Rel->NextItem(RelCsr, RelRecord) != XXX_OK) ||
     (RelCsr.IsValid() != XXX_TRUE)) {
   REL_NOTE_ERROR(REL_BADEXEC, "Unable to get record from relation");
   return XXX_NOT_OK;
  }

 // Step 4: extract the first field from the record into NewObject
 //Bjarke, sep. 13 2001: added const qualifier to variable
 const char *Fld0Ptr;
 RelRecord->GetField(0, Schema, Fld0Ptr);
 XxxADTValue* Fld0Value = (XxxADTValue*)Fld0Ptr;
 if (XxxCoreADTClass::IsCoreADT(CastType)) {
   XxxCoreADTClass::
     STypeCopy(CastType, Fld0Value, NewObject,
	       Schema->GetAttr(0)->TypeCatInfo);
 }
 else {
   // reaches here if it is an ADT
   XxxADTClass* AttrClass = XxxADTArray[CastType];
   AttrClass->TypeCopy(Fld0Value, NewObject,
			 Schema->GetAttr(0)->TypeCatInfo);
 }

 // Step 5: check that there are no more records in the relation
 XxxRecord* RelRecord2;
 if ((Rel->NextItem(RelCsr, RelRecord2) != XXX_OK) ||
     (RelCsr.IsValid() == XXX_TRUE)) {
   REL_NOTE_ERROR(REL_BADEXEC, "Record not unique");
   return XXX_NOT_OK;
  }

 // Step 6: close the cursor on the relation
 if (Rel->CloseCursor(RelCsr) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, "Error closing relation cursor");
   return XXX_NOT_OK;
  }

 // Step 7: delete the record
 // this is necessary, because the consumer is responsible for memory
 delete RelRecord;
 // Step 8: delete the relation
 delete Rel;

 return XXX_OK;
}

void XxxRelADT::RelStats::Print(ostream& OStr) const
{
 OStr << "   RelADT : Statistics: " << endl;
 OStr << "\t TypeSize : " << TypeSize << endl;
 OStr << "\t TypeCopy : " << TypeCopy << endl;
 OStr << "\t TypeStoreCopy : " << TypeStoreCopy << endl;
 OStr << "\t TypeDestroy : " << TypeDestroy << endl;
 OStr << "\t WriteText : " << WriteText << endl;
 OStr << "\t ReadText : " << ReadText << endl;
 OStr << "\t GetMetaInfo : " << GetMetaInfo << endl;
 OStr << "\t AddMetaInfo : " << AddMetaInfo << endl;
 OStr << "\t CastCheck : " << CastCheck << endl;
 OStr << "\t CastExec : " << CastExec << endl;
}

//
// RelADTEnv methods and RelADT methods relating to environment
//

XxxRelADT::
RelADTEnv::RelADTEnv() : XxxADTEnv()
{
  RelStatsPtr=new RelStats();
}

XxxRelADT::
RelADTEnv::~RelADTEnv() 
{
  if (RelStatsPtr)
    delete RelStatsPtr;
}

XxxErrCode 
XxxRelADT::CreateEnv(XxxADTEnv* &Env) const
{
  Env=new RelADTEnv;
  return (Env==NULL)?XXX_NOT_OK:XXX_OK;
}

XxxErrCode XxxRelADT::
RelADTEnv::SetValue(const char* ValName, void* NewVal)
{
  // we don't let user set RelStatsPtr variable
  return XxxADTEnv::SetValue(ValName,NewVal);
}

void XxxRelADT::
RelADTEnv::Print(ostream& OStr) const
{
  // print the parent's variables.
  XxxADTEnv::Print(OStr); 
}
