#ifndef XXX_RELADT_H
#define XXX_RELADT_H

#include "xxx_thread.h"
#include "env.h"
#include "types.h"

XXX_EXTERN_CLASS_DECL(XxxRelationEngine);
XXX_EXTERN_CLASS_DECL(RelDataEngine);
// global initialized in XxxRelADT constructor
extern XxxBasicType RelationTypeId; 

class XxxRelADT : public XxxADTClass {
 protected:
  class RelStats {
   public:
    int NumQPEngines;
    int NumDataEngines;

    // stats on invocations of each method
    int TypeSize;
    int TypeCopy;
    int TypeStoreCopy;
    int TypeDestroy;
    int WriteText;
    int ReadText;
    int GetMetaInfo;
    int AddMetaInfo;
    int CastCheck;
    int CastExec;
    int Extra[5];

    void Reset() {
      TypeSize = TypeCopy = TypeStoreCopy = TypeDestroy = 0;
      WriteText = ReadText = GetMetaInfo = 0;
      AddMetaInfo = 0;
      CastCheck = CastExec = 0;
      Extra[0] = Extra[1] = Extra[2] = Extra[3] = Extra[4] = 0;
     }
    void Print(ostream& OStr) const;
    RelStats() { Reset(); }
   };

  inline static RelStats* RELSTATS()
    { return ((RelADTEnv *)(XXX_THISADTENV(RelationTypeId)))->RelStatsPtr; }

  // environment structure for RelADT
  class RelADTEnv : public XxxADTEnv {
  public:
    // store pointer to relation statistics object in this environment.
    static const char* const RELSTATS;
    
    RelStats* RelStatsPtr;

    RelADTEnv();
    virtual ~RelADTEnv();
    virtual void Print(ostream& OStr) const;
    virtual XxxErrCode SetValue(const char* ValName, void* NewVal);
  };
 public:
  RelDataEngine* DataEngines; // generalize to a collection
  XxxRelationEngine* QPEngines; // generalize to a collection

  XxxRelADT(XxxBasicType);
  virtual ~XxxRelADT();

  virtual int TypeSize(const XxxADTValue* Object, const XxxADTMetaInfo* MetaInfo) const;
  virtual int TypeCopy(const XxxADTValue* SrcObject, 
		       XxxADTValue *ResultObject,
		       const XxxADTMetaInfo* MetaInfo) const;
  virtual int TypeStoreCopy(const XxxADTValue* SrcObject, 
			    XxxADTValue *ResultObject,
			    const XxxADTMetaInfo* MetaInfo) const;
  virtual void TypeDestroy(XxxADTValue *Object, const XxxADTMetaInfo* MetaInfo) const;

  virtual XxxErrCode WriteText(ostream& OutStream, 
			       const XxxADTValue* Object,
			       const XxxADTMetaInfo* MetaInfo) const;
  virtual XxxErrCode ReadText(istream& InStream, 
			      XxxADTValue* ObjectRef,
			      const XxxADTMetaInfo* MetaInfo) const;
  
  virtual XxxErrCode AddMetaInfo(const char* Name, 
				 const XxxBool Named,
				 const char* MetaInfoString,
				 MetaInfoHandle& MIHandle);
//Bjarke, aug. 8: const added to char* parameter
  virtual XxxErrCode GetMetaInfo(const char* ObjName, 
				 const MetaInfoHandle& MIHandle,
				 XxxADTMetaInfo* &MetaInfo, 
				 XxxBool& Found) const;

  /**
   * Determines if this ADT can cast an object of its type into
   * another type.
   * @param TargetType The type to cast to
   * @param SourceMetaInfo The meta info of the object to be casted
   * @param TargetMetaInfo The meta info of the new object to be created
   */
  virtual XxxBool CanCastTo(XxxBasicType TargetType, 
			    const XxxADTMetaInfo* SourceMetaInfo,
			    const XxxADTMetaInfo* TargetMetaInfo);
  /**
   * Executes a cast from this type into another type.
   * @param CastType The type id of the new object
   * @param SourceMetaInfo The meta info of the object to be casted
   * @param TargetMetaInfo The meta info of the new object to be created
   * @param OldObject The object to be casted
   * @param NewObject The new object to be created
   */
  virtual XxxErrCode CastTo(XxxBasicType TargetType, 
			    const XxxADTMetaInfo* SourceMetaInfo,
			    const XxxADTMetaInfo* TargetMetaInfo,
			    XxxADTValue* OldObject, 
			    XxxADTValue* NewObject);

  XxxErrCode RegisterQPEngine(XxxRelationEngine* QPEng);
  XxxErrCode RegisterDataEngine(RelDataEngine* DEng);

  virtual XxxErrCode CreateEnv( XxxADTEnv* &Env ) const; //-->sandman 3/29/97
  // no need for a DeleteEnv; the parent class DeleteEnv is fine.

  virtual void PrintStats(ostream& OStr) const
   { RELSTATS()->Print(OStr); }
  virtual void ResetStats() 
   { RELSTATS()->Reset(); }

  virtual XxxErrCode CopyHToN(const XxxADTValue* SrcObject, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const {
    // TODO: does this need to reverse bytes within RelHandle?
    TypeCopy(SrcObject, ResultObject, MetaInfo);
    return XXX_OK;
  }
  virtual XxxErrCode CopyNToH(const XxxADTValue* SrcObject, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const {
    // TODO: does this need to reverse bytes within RelHandle?
    TypeCopy(SrcObject, ResultObject, MetaInfo);
    return XXX_OK;
  }
 };

#endif  /* XXX_RELADT_H */
