#ifdef __GNUG__
#pragma implementation
#pragma implementation "commands.h"
#endif

#include "xxx_thread.h"
#include "reladt.h"
#include "relopt.h"
#include "relquery.h"
#include "relcat.h"
#include "reldata.h"
#include "relexec.h"
#include "commands.h"
#include "releng.h"
#include "relenv.h"
#include "shoresm.h"
#include "client.h"

#define yyFlexLexer sqlFlexLexer
#include <FlexLexer.h>

int RelationEngineId;
const int XxxRelationEngine::NumErrorConds = 16;
const char *const XxxRelationEngine::ErrorTable[] = {
  "Missing relation",             // 0 REL_MSSNGREL
  "Bad relation device",          // 1 REL_BADDEV
  "Duplicate view definition",    // 2 REL_DUPVIEW
  "Incorrect query definition",   // 3 REL_BADQUERY
  "Bad query optimization",       // 4 REL_BADOPT
  "Bad execution plan",           // 5 REL_BADPLAN
  "Bad query execution",          // 6 REL_BADEXEC
  "Bad query operator",           // 7 REL_BADOP
  "Missing catalog entry",        // 8 REL_MSSNGCAT
  "Bad catalog entry",            // 9 REL_BADCAT
  "Bad text file",                // 10 REL_BADTXTFILE
  "Bad SHORE relation",           // 11 REL_BADSHOREREL
  "Bad Table",                    // 12 REL_BADSHOREREL
  "Bad Index",                    // 13 REL_BADSHOREREL
  "Bad Data Source",              // 14 REL_BADSRC
  "Bad Cursor",                   // 15 REL_BADCURSOR
  "No view definition"            // 16 REL_NOVIEW
   };

XxxRelationEngine::XxxRelationEngine(const int Id) : 
XxxDBEngine(Id, "__Rel", 1)
{
 // Important : do this rightaway!
 RelationEngineId = Id;

 strncpy(LanguageNames[0], "SQL", XXX_LANGNAME_MAX);
 LanguageNames[0][XXX_LANGNAME_MAX-1] = 0;
 // create all of the engines
 Optimizer = new RelOptEngine(this);
 ExecEng = new RelExecEngine(this);
 CommandEng = new RelCommandEngine(this);
 QueryEng = new RelQueryEngine(this);
 DataEng = new RelDataEngine(this);

 // look at the startup files for Relations
 // this should specify where catalog files will be found, etc
 if ((establishRegistry() != XXX_OK) || (XXX_IS_ERROR())) {
   XXX_PRINT_ERROR(*(XXX_THISENV().ErrorStr), "Cannot establish registry");
   XXX_CLEAR_ERROR();
  }

 // now register this engine with whichever types it serves
 XxxRelADT* RelADT = (XxxRelADT*)(XxxADTArray[RelationTypeId]);
 if (RelADT->RegisterQPEngine(this) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_BADENG, "Unable to register relation QP engine");
   return;
  }
 if (RelADT->RegisterDataEngine(DataEng) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_BADENG, "Unable to register relation QP engine");
   return;
  }
}

XxxRelationEngine::~XxxRelationEngine()
{
 delete Optimizer;
 delete ExecEng;
 delete CommandEng;
 delete QueryEng;
 delete DataEng;
}

XxxErrCode XxxRelationEngine::establishRegistry()
{
  // Get the environment variables which give the size and location of the 
  // relation data disk
  // OLD: char *RelDiskPath = getenv("XXXSHORERELDISKPATH");
  char *RelDiskPath = XXX_THISENV().Options->getopt("XXXSHORERELDISKPATH");
  if (!RelDiskPath) {
    XXX_NOTE_ERROR(XXX_BADEXEC, 
		   "XXXSHORERELDISKPATH not specified: check environment");
    return XXX_NOT_OK;
  }
  // OLD: char *RelDiskSizeString = getenv("XXXSHORERELDISKSIZE");
  char *RelDiskSizeString = XXX_THISENV().Options->getopt("XXXSHORERELDISKSIZE");
  if (!RelDiskSizeString) {
    XXX_NOTE_ERROR(XXX_BADEXEC, 
		   "XXXSHORERELDISKSIZE not specified: check environment");
    return XXX_NOT_OK;
  }
  int RelDiskSize = atoi(RelDiskSizeString);
  // First check to see if the file exists already
  XxxBool Found = XXX_FALSE;
  ifstream InFStream;
  InFStream.open(RelDiskPath, ios::nocreate);
  Found = (InFStream.good() ? XXX_TRUE : XXX_FALSE);
  InFStream.close();
  /*
   * For now we limit the relation disk to a SHORE disk; but the capability
   * exists for other formats.
   */
  if (!Found) {
    // Creating relational disk
    cerr << "Creating relational data disk: " << RelDiskPath 
	 << "(" << RelDiskSize << " KB)" << endl;
    if (DataEng->CreateDevice(RelDiskPath, RelDiskSize, 
			      REL_SHORE_SRC) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot create relation data disk");
      return XXX_NOT_OK;
    }
  }
  // Mounting relational disk
  cerr << "Mounting relational data disk: " << RelDiskPath << endl;
  if (DataEng->MountDevice(RelDiskPath, REL_SHORE_SRC) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot mount relation data disk");
    return XXX_NOT_OK;
  }
  return XXX_OK;
}

XxxErrCode XxxRelationEngine::ExprParse(char* ExprString,
					const XxxArgEnv* Env,
					XxxBasicType& ReturnType,
					XxxGenericParse*& ParseStructure)
{
 XxxBool Success = XXX_FALSE;
 ParseStructure = NULL;
 ReturnType = XXX_UNKNOWN;

 // TODO:L3 Its silly to step thru the languages. the chosen language
 // should be passed in (since it is already determined in client.C
 for (int j = 0; j < NumLanguages; j++) {
   if (!strncasecmp(LanguageNames[j], ExprString, strlen(LanguageNames[j]))) {
     Success = XXX_TRUE;
     istrstream InStream(ExprString+strlen(LanguageNames[j])+1);
     parseStruct PDS(&InStream, NULL);
     if (sqlparse(PDS) || 
	 (XXX_IS_ERROR())) {
       if (PDS.LexerPtr) {
	 delete ((yyFlexLexer*)(PDS.LexerPtr));
	 PDS.LexerPtr = NULL;
       }
       REL_NOTE_ERROR(REL_BADQUERY, "Error parsing query");
       Success = XXX_FALSE; 
       break;
     } else { 
       if (PDS.LexerPtr) {
	 delete ((yyFlexLexer*)(PDS.LexerPtr));
	 PDS.LexerPtr = NULL;
       }
     }
     // If ViewExpr is non-NULL, do the type checking, etc
     RelViewClass *ViewExpr;
     if ((ViewExpr = ((RelViewClass *)(PDS.OutArg)))) {
       // Enters this only if this is a nested language expression
       RelQueryClass* Query = NULL;
       if (QueryEng->CheckQuery(ViewExpr, Query, Env) != XXX_OK) {
	 REL_NOTE_ERROR(REL_BADQUERY, "Error checking query");
	 delete ViewExpr;
	 Success = XXX_FALSE; 
	 break;
	}
       else {
	 ParseStructure = Query;
	 ReturnType = Query->GenerateType();
	} // else there were no errors
      } // if this is a nested expression
    } // if this is the right language
  } // for all languages supported by this engine

 return (Success ? XXX_OK : XXX_NOT_OK);
}

XxxErrCode XxxRelationEngine::ExprOptimize(XxxGenericParse* ParseStructure,
        XxxGenericPlan*& PlanStructure)
{
 RelQueryPlan* Plan = NULL;
 RelQueryClass* Query = (RelQueryClass*)ParseStructure;
 if (!Query) {
   REL_NOTE_ERROR(REL_BADOPT, "Null query plan");
   return XXX_NOT_OK;
  }
 RelSpecParameters Params(XXX_FALSE); // specify that answers should 
                                      // not be printed, but stored instead
 if (Optimizer->OptimizeQuery(Query, &Params, Plan) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error optimizing query");
   return XXX_NOT_OK;
  }
 PlanStructure = Plan;
 return XXX_OK;
}

XxxErrCode XxxRelationEngine::
ExprEvaluate(XxxGenericPlan* PlanStructure,
	     const XxxValueEnv* ArgEnv,
	     XxxADTValue *ReturnValue)
{
 // this should evaluate the query and return the appropriate type object
 RelExecParameters Params(ArgEnv);
 return (ExecEng->EvalQuery((RelQueryPlan *)PlanStructure, 
			    &Params, ReturnValue));
} 

const char* XxxRelationEngine::ErrorString(XxxErrCond Condition) const
{
 if ((Condition >= NumErrorConds) || (Condition < 0))
   return "???";
 else return ErrorTable[Condition];
}

XxxErrCode 
XxxRelationEngine::ParseMetaInfo(const char* MetaInfoString,
     XxxGenDataSrc::AttrTypeList*& Attrs,
     XxxGenDataSrc::KeyList*& Keys,
     int* SrcType,
     char*& DeviceName,
     char*& RelImplType)
{
 istrstream InStream((char *)MetaInfoString);
 RelEngCatParse CatInfo;
 CatInfo.SrcType = SrcType;
 parseStruct PDS(&InStream, (void*)(&CatInfo));

 // hack: normally, sqlparse treats the third argument as
 // an _out_ parameter. in this case though, it is being
 // treated as an _in_ parameter.
 int ParseReturn = sqlparse(PDS);
 if (PDS.LexerPtr) { 
   delete ((yyFlexLexer*)(PDS.LexerPtr));
   PDS.LexerPtr = NULL;
  }
 if (ParseReturn || (XXX_IS_ERROR())){
   REL_NOTE_ERROR(REL_BADCAT, "Error parsing catalog info");
   return XXX_NOT_OK;
 } 
 Attrs = CatInfo.Attrs;
 Keys = CatInfo.Keys;
 DeviceName = CatInfo.DevName;
 RelImplType = CatInfo.ImplType;
 // src type is implicitly set in the parser by 
 // dereferncing the ptr

 return XXX_OK;
}
         
XxxErrCode XxxRelationEngine::CreateEnv( XxxEngineEnv* &Env ) const
{
 Env = new XxxRelEnv();
 return XXX_OK;
}
XxxErrCode XxxRelationEngine::DeleteEnv( XxxEngineEnv* Env ) const
{
 if (!Env)
   return XXX_NOT_OK;
 XxxRelEnv *REnv = (XxxRelEnv *) Env;
 delete REnv;
 return XXX_OK;
}

void XxxRelationEngine::PrintStats(ostream& OStr) const
{
 OStr << "Statistics for Relation Engine" << endl;
}

//
// Takes in a value expression. Parse it and return it as XxxValueExpression
// This is used mainly when constructing function index.
//
XxxErrCode 
XxxRelationEngine::ParseValueExpression(const char* IndexString,
					XxxValueExpression *&Expr)
{
 istrstream InStream ((char *)IndexString, strlen(IndexString));
 parseStruct PDS (&InStream, NULL);

 int ParseReturn = sqlparse(PDS);
 if (PDS.LexerPtr) {
   delete ((yyFlexLexer*)(PDS.LexerPtr));
   PDS.LexerPtr = NULL;
  }
 if (ParseReturn) {
   REL_NOTE_ERROR (REL_BADQUERY, "Error in indexed expression");
   return XXX_NOT_OK;
 }
 Expr = (XxxValueExpression *)PDS.OutArg;
 return XXX_OK;
}

//
// Takes in a boolean value expression. Parse it and return it as XxxBooleanExpression
// This is used mainly when constructing function index.
//
XxxErrCode 
XxxRelationEngine::ParseBoolExpression(const char* String,
				       XxxBooleanExpression *&Expr)
{
  XxxValueExpression* VExpr;
  if (ParseValueExpression(String, VExpr) != XXX_OK)
    return XXX_NOT_OK;
  if (VExpr->ValType() != XxxValueExpression::BOOL_EXPR) {
   REL_NOTE_ERROR (REL_BADQUERY, "Not a boolean expression");
   return XXX_NOT_OK;
  }
  Expr = (XxxBooleanExpression*)VExpr;
  return XXX_OK;
}

XxxErrCode XxxRelationEngine::execQuery(RelViewClass* QueryView,
					RelSpecParameters& Params) const
{
  RelQueryClass* QueryQC = NULL;
  RelQueryPlan* PlanQP = NULL;
  RelExecParameters ExecParams;
  XxxBool PushDiagStack = XXX_FALSE;
  XxxErrCode rc = XXX_OK;
  
  // If the client is a Java client, open the file which will hold the
  // information regarding query processing
  XxxRelEnv *GEnv = XXX_THISRELENV();
  if (GEnv->ShowQPInfo) {
    char fullpath[100];
    strcpy(fullpath, XXX_THISENV().TmpDataDir);
    char file[100];
    sprintf(file, "/query_proc_info.client%d.txt", XXX_CLIENTINFO()->UserId); 
    strcat(fullpath, file);
    GEnv->QPInfoFile = new ofstream(fullpath);
    // put the QPInfoFile onto the DiagnosticStream stack
    XXX_THISENV().DiagnosticStream.PushStack(GEnv->QPInfoFile);
    PushDiagStack = XXX_TRUE;
  }
  
  if (QueryEng->CheckQuery(QueryView, QueryQC) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, "Check Query failed");
    rc =  XXX_NOT_OK;
  }
  else {
    // check that the types match for INSERT INTO statement
    if (Params.NamedRel) {
      // get hold of the catalog info
      XxxBool MIFound = XXX_FALSE;
      RelCatalogInfo* CatInfo = new RelCatalogInfo(Params.RelName);
      REL_CALL_FN(DataEng->GetCatInfo(Params.RelName, *CatInfo, 
				      MIFound),
		  REL_BADTABLE, "Cannot find catalog for relation");
      if (MIFound) {
	// check that CatInfo matches QueryQC
	XxxBool Matches;
	XxxRecordSchema* QuerySchema = ((RelCatalogInfo*)(QueryQC->GenerateCatInfo()))->RecordSchema();
	REL_CALL_FN(CatInfo->RecordSchema()->Matches(*QuerySchema, Matches),
		    REL_BADTABLE, "INSERT schemas do not match");
	delete CatInfo;
	if (!Matches) {
	  REL_NOTE_ERROR (REL_BADTABLE, "INSERT schemas do not match");
	  return XXX_NOT_OK;
	}
      } // if the target table already exists
      // Check for halloween problem: make sure target table is not in query.
      // TODO: this does not check for halloween problem thru a view or
      // through tables in nested subqueries
      RelViewClass::FromInfo* FromList = QueryView->FromList;
      for (int j = 0; j < FromList->Arity; j++) {
	RelInfo* TableInfo = FromList->RelList[j];
	if (TableInfo && !TableInfo->Nested &&
	    !strcmp(((SimpleRelInfo*)TableInfo)->RelName, Params.RelName)) {
	  REL_NOTE_ERROR (REL_BADQUERY, "INSERT might have halloween problem");
	  return XXX_NOT_OK;
	}
      } // for each table in the query
    } // if it is an INSERT INTO query
    //Begin Moustafa
    XxxBool Er = XXX_IS_ERROR();
    //End Moustafa
    if (Optimizer->OptimizeQuery(QueryQC, &Params, PlanQP) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADOPT, "Optimize Query failed");
      rc = XXX_NOT_OK;
    }
    else {
      //Begin Moustafa
      Er = XXX_IS_ERROR();
      // Add window information from window class to RelExecParameters
      if( QueryView->WindowValue!=NULL)
	QueryView->WindowValue->GetWindowValue(&(ExecParams.TheWindowSec), &(ExecParams.TheWindowMicroSec));
      //End Moustafa
      if (ExecEng->EvalQuery(PlanQP, &ExecParams, NULL) != XXX_OK) {
	REL_NOTE_ERROR(REL_BADEXEC, "Eval Query failed");
	rc = XXX_NOT_OK;
      }
    }
  }
  // successful, now delete structures
  if (QueryQC) delete QueryQC;
  if (PlanQP) delete PlanQP;
  
  // Close the file stream
  if (GEnv->QPInfoFile) {
    GEnv->QPInfoFile->flush();
    delete GEnv->QPInfoFile;
    GEnv->QPInfoFile = NULL;
  }
  if (PushDiagStack) {
    // make sure to pop the DiagnosticStream stack
    ASSERT((!XXX_THISENV().DiagnosticStream.IsEmpty()));
    ostream* TmpOstr;
    XXX_THISENV().DiagnosticStream.PopStack(TmpOstr);
  }
  return rc;
}

//
// XxxRelEnv
//
// define static fields

// use sort-merge join when appropriate
const char* const XxxRelEnv::SORT_MERGE_JOIN = "sort_merge_join";
const char XxxRelEnv::SORT_MERGE_ENABLED = 1;

// hash-join flag
const char* const XxxRelEnv::HASH_JOIN = "hash_join";
const char XxxRelEnv::HASH_JOIN_ENABLED = 0;

// show query results
const char* const XxxRelEnv::SHOW_QUERY_RESULTS = "show_query_results";
const bool XxxRelEnv::SHOW_Q_RESULTS = true;

// show query processing information
const char* const XxxRelEnv::SHOW_QUERY_PROC_INFO = "show_query_proc_info";
const char XxxRelEnv::SHOW_QP_INFO = 1;

// perform query rewrite
const char* const XxxRelEnv::REWRITE = "rewrite";
const char XxxRelEnv::REL_OPT_REWRITE  = 1;

// push selections
const char* const XxxRelEnv::PUSHSEL = "pushsel";
const char XxxRelEnv::REL_OPT_PUSHSEL  = 1;

// push projections
const char* const XxxRelEnv::PUSHPROJ = "pushproj";
const char XxxRelEnv::REL_OPT_PUSHPROJ  = 1;

const char* const XxxRelEnv::JOINORDER = "joinorder";
const char XxxRelEnv::REL_OPT_JOINORDER = XxxRelEnv::KBZ_JO;

const char* const XxxRelEnv::TRY_CROSS_PRODUCT = "crossproduct";
const char XxxRelEnv::REL_OPT_CROSSPRODUCT = 0;

// default NL join method
const char* const XxxRelEnv::NLJOINMETH = "nljoin";
const char XxxRelEnv::REL_OPT_NLJOINMETH = XxxRelEnv::PAGE_NL;

//Begin Moustafa
// default window join method
// user may choose between HBEW_JOIN, NLBEW_JOIN, HFEW_JOIN, NLFEW_JOIN
const char* const XxxRelEnv::WJOINMETH = "wjoin";
const char XxxRelEnv::REL_OPT_WJOINMETH = XxxRelEnv::HBEW_JOIN;
//End Moustafa

// pipeline joins by default
const char* const XxxRelEnv::PIPELINE = "pipeline";
const char XxxRelEnv::REL_OPT_PIPELINE  = 1;

// materialize join results by default
const char* const XxxRelEnv::MATERIALIZE = "materialize";
const char XxxRelEnv::REL_OPT_MATERIALIZE  = 0;
 
// flags for data loading
const char* const XxxRelEnv::DELIMITED = "delimited";
const char XxxRelEnv::REL_OPT_DELIMITED  = 0;
const char* const XxxRelEnv::DELIMITER = "delimiter";
const char XxxRelEnv::REL_OPT_DELIMITER  = '\t';

// flags for Selection Ordering
const char *const XxxRelEnv::SELECTIONORDERING = "selection_ordering";
const char XxxRelEnv::REL_OPT_SELECTIONORDERING = SELECTIVITY_PO;

// flags for client side UDF batch size (concurrency factor)
const char *const XxxRelEnv::UDFBATCHSIZE = "udf_batch_size";
const int XxxRelEnv::REL_OPT_UDFBATCHSIZE = 0;

// flags for client side UDF execution ( sets REL_OPT_UDF_EXEC bitwise)
const char *const XxxRelEnv::SET_UDF_EXEC = "set_udf_exec";
const char *const XxxRelEnv::UNSET_UDF_EXEC = "unset_udf_exec";
const int XxxRelEnv::REL_OPT_UDF_EXEC = 0;

// flags for expensive predicates
const char *const XxxRelEnv::EXPENSIVE_PRED = "expensive_pred";
const enum XxxRelEnv::ExpensivePredStrategy XxxRelEnv::REL_OPT_EXPPRED = PULLUP_EP;

XxxRelEnv::XxxRelEnv()
{
 ShowQueryResults = SHOW_Q_RESULTS;

 SortMergeEnabled = SORT_MERGE_ENABLED;

 ShowQPInfo = SHOW_QP_INFO;
 QPInfoFile = NULL;

 RewriteQuery = REL_OPT_REWRITE ;

 PushSelection = REL_OPT_PUSHSEL;
 PushProjection = REL_OPT_PUSHPROJ;

 TryCrossProduct = REL_OPT_CROSSPRODUCT;

 JoinOrder = REL_OPT_JOINORDER;
 // Use a getenv to dynamically set this!
 // that is a very useful hack to avoid having to
 // recompile before running testscripts
 // OLD: char* OptChoice = getenv("XXX_REL_OPT_JOINORDER");
 char* OptChoice = XXX_THISENV().Options->getopt("XXX_REL_OPT_JOINORDER");
 if (OptChoice) {
   int NewJoinOrder = atoi(OptChoice);
   if ((NewJoinOrder >= 0) && (NewJoinOrder < 5))
     JoinOrder = NewJoinOrder;
  }

 NLJoinMethod = REL_OPT_NLJOINMETH ;
 // Use a getenv to dynamically set this!
 // that is a very useful hack to avoid having to
 // recompile before running testscripts
 // OLD: OptChoice = getenv("XXXNLJOINVALUE");
 OptChoice = XXX_THISENV().Options->getopt("XXXNLJOINVALUE");
 if (OptChoice) {
   int NewNLJoinMethod = atoi(OptChoice);
   if ((NewNLJoinMethod >= 0) && (NewNLJoinMethod < 5))
     NLJoinMethod = NewNLJoinMethod;
  }

 //Begin Mousatafa
 WJoinMethod = REL_OPT_WJOINMETH ;
 // Use a getenv to dynamically set this!
 // that is a very useful hack to avoid having to
 // recompile before running testscripts
 OptChoice = XXX_THISENV().Options->getopt("XXXWJOINVALUE");
 if (OptChoice) {
   int NewWJoinMethod = atoi(OptChoice);
   if ((NewWJoinMethod >= 0) && (NewWJoinMethod < 5))
     WJoinMethod = NewWJoinMethod;
  }
 //End Moustafa

 PipelineJoins = REL_OPT_PIPELINE;
 MaterializeTemps = REL_OPT_MATERIALIZE;

 Delimited = REL_OPT_DELIMITED ;
 Delimiter = REL_OPT_DELIMITER ;

 SelectionOrdering = REL_OPT_SELECTIONORDERING;
 UDFBatchSize = REL_OPT_UDFBATCHSIZE;
 UDFExec = REL_OPT_UDF_EXEC;
 
 ExpPredStrategy = REL_OPT_EXPPRED;

 CheapPredMax = 49.99;  // TODO: need to find a better way to specify this

 XxxBool BVal = XXX_FALSE;
 /**** Too much junk printed to screen -- by default turn all flags off
#ifdef XXX_DEBUG
 BVal = XXX_TRUE;
#endif
***/
 DebugQrw = DebugOpt = DebugExec = DebugStorage = DebugStats = DebugCompile = BVal;

 for (int i = 0; i < 10; i++)
   MoreFlags[i] = 1;
}

XxxRelEnv::~XxxRelEnv()
{
}

void XxxRelEnv::Print(ostream& OStr) const
{
 // print out the current settings
 OStr << SORT_MERGE_JOIN << " : Should Sort-Merge join be considered as a join alternative? " <<
   (SortMergeEnabled?"yes":"no") << endl;
 OStr << HASH_JOIN << " : Should symmetric hash join be considered as a join alternative? " <<
   (HashJoinEnabled?"yes":"no") << endl;
 OStr << SHOW_QUERY_RESULTS << " : should query results be output? " <<
   (ShowQueryResults?"yes":"no") << endl;
 OStr << SHOW_QUERY_PROC_INFO << " : should query processing information be shown? " <<
   (ShowQPInfo?"yes":"no") << endl;
 OStr << REWRITE << " : should query rewrite be used? " << 
   (RewriteQuery?"yes":"no") << endl;
 OStr << PUSHSEL << " : should selections be pushed down? " << 
   (PushSelection?"yes":"no") << endl;
 OStr << PUSHPROJ << " : should projections be pushed down? " << 
   (PushProjection?"yes":"no") << endl;
 OStr << JOINORDER << " : which Join Order optimizer to use? "
      << "(0:NAIVE_JO/1:SEMINAIVE_JO/2:GREEDY_JO/3:KBZ_JO/4:DYNPROG_JO)? " 
      << ((int)(JoinOrder) == 0?  "NAIVE_JO": 
	  ((int)(JoinOrder) == 1?  "SEMINAIVE_JO": 
	   ((int)(JoinOrder) == 2?  "GREEDY_JO": 
	    ((int)(JoinOrder) == 3?  "KBZ_JO": "DYNPROG_JO"))))  << endl;

 OStr << NLJOINMETH << " : which NL Join method to use? " << 
   (NLJoinMethod?"page-at-a-time":"tuple-at-a-time") << endl;
 //Begin Moustafa
 OStr << WJOINMETH << " : which Window Join method to use? " 
      << "(0:HBEW_JOIN/1:NLBEW_JOIN/2:HFEW_JOIN/3:NLFEW_JOIN)? " 
      << (WJoinMethod == 0?  "HBEW_JOIN": 
	  (WJoinMethod == 1? "NLBEW_JOIN" : 
	   (WJoinMethod == 2? "HFEW_JOIN" : "NLFEW_JOIN"))) << endl;
 //End Moustafa

 OStr << PIPELINE << " : should joins be pipelined? " << 
   (PipelineJoins?"yes":"no") << endl;
 OStr << MATERIALIZE << " : materialize intermediate results? " << 
   (MaterializeTemps?"yes":"no") << endl;

 OStr << DELIMITED << " : is there a delimiter for load data? " << 
   (Delimited?"yes":"no") << endl;
 OStr << DELIMITER << " : what is the delimiter for load data? "<< 
   Delimiter << endl;

 OStr << SELECTIONORDERING << " : how are the selections to be ordered "
      << "(0:Selectivity/1:Cost/2:Rank)? " 
      << (SelectionOrdering == 0?  "Selectivity": 
	  (SelectionOrdering == 1? "Cost" : "Rank")) << endl;

 // mmaasha
 /*
   OStr << UDFBATCHSIZE << " : what is the batch size for client side UDFs? "<< OStr << SET_UDF_EXEC << " : what are the flags for client UDF execution? "<<
   OStr << UNSET_UDF_EXEC << " : what are the flags for client UDF execution? "<<
   Delimiter << endl;
   */
}

XxxErrCode XxxRelEnv::SetValue(const char* ValName,
			       char NewVal)
{
 if (!strcmp(ValName, REWRITE)) {
   RewriteQuery = NewVal;
  } 
 else if (!strcmp(ValName, SORT_MERGE_JOIN)) {
   SortMergeEnabled = NewVal;
 }
 else if (!strcmp(ValName, HASH_JOIN)) {
   HashJoinEnabled = NewVal;
 }
 else if (!strcmp(ValName, SHOW_QUERY_RESULTS)) {
   ShowQueryResults = (NewVal?true:false);
 }
 else if (!strcmp(ValName, SHOW_QUERY_PROC_INFO)) {
   ShowQPInfo = NewVal;
 }
 else if (!strcmp(ValName, REWRITE)) {
   RewriteQuery = NewVal;
  }
 else if (!strcmp(ValName, PUSHSEL)) {
   PushSelection = NewVal;
  }
 else if (!strcmp(ValName, EXPENSIVE_PRED)) {
   ExpPredStrategy = (enum XxxRelEnv::ExpensivePredStrategy)NewVal;
  }
 else if (!strcmp(ValName, PUSHPROJ)) {
   PushProjection = NewVal;
  }
 else if (!strcmp(ValName, JOINORDER)) {
   JoinOrder = NewVal;
  }
 else if (!strcmp(ValName, NLJOINMETH)) {
   NLJoinMethod = NewVal;
  }
 //Begin Moustafa
 else if (!strcmp(ValName, WJOINMETH)) {
   WJoinMethod = NewVal;
 }
 //End Moustafa
 else if (!strcmp(ValName, PIPELINE)) {
   PipelineJoins = NewVal;
  }
 else if (!strcmp(ValName, MATERIALIZE)) {
   MaterializeTemps = NewVal;
  }
 else if (!strcmp(ValName, DELIMITED)) {
   Delimited = NewVal;
  }
 else if (!strcmp(ValName, DELIMITER)) {
   Delimiter = NewVal;
  }
 else if (!strcmp(ValName, SELECTIONORDERING)) {
   SelectionOrdering = NewVal;
  }
 else if (!strcmp(ValName, UDFBATCHSIZE)) {
   if (NewVal > 0)
     UDFBatchSize = NewVal;
   else 
     UDFBatchSize = NewVal * (-100);
	cerr << "UDFBatchSize set to " << UDFBatchSize << "\n";
  }
 else if (!strcmp(ValName, SET_UDF_EXEC)) {
   UDFExec = UDFExec | (1<<NewVal);
   cerr << "UDFExec flags set to " << UDFExec <<"\n";
   }
 else if (!strcmp(ValName, UNSET_UDF_EXEC)) {
   UDFExec = UDFExec && !(1<<NewVal);
   cerr << "UDFExec flags set to " << UDFExec <<"\n";
   }
 else if (!strcmp(ValName, "debug_qrw")) {
   DebugQrw = (NewVal ? XXX_TRUE : XXX_FALSE);
  }
 else if (!strcmp(ValName, "debug_opt")) {
   DebugOpt = (NewVal ? XXX_TRUE : XXX_FALSE);
  }
 else if (!strcmp(ValName, "debug_exec")) {
   DebugExec = (NewVal ? XXX_TRUE : XXX_FALSE);
  }
 else if (!strcmp(ValName, "debug_store")) {
   DebugStorage = (NewVal ? XXX_TRUE : XXX_FALSE);
  }
 else if (!strcmp(ValName, "debug_stats")) {
   DebugStats = (NewVal ? XXX_TRUE : XXX_FALSE);
  }
 else if (!strcmp(ValName, "debug_comp")) {
   DebugCompile = (NewVal ? XXX_TRUE : XXX_FALSE);
  }
 else {
   return XXX_NOT_OK;
  }
 return XXX_OK;
}



