#ifndef XXX_RELENV_H
#define XXX_RELENV_H

#include <fstream.h>
#include "env.h"
#include "error.h"
#include "releng.h"

// some consts are defined in relconfig.h. Define the rest here (for
// ease of recompilation)
// Define a XxxRelEnv Class
XXX_EXTERN_CLASS_DECL(XxxArgEnv); // look in utils/values.h
class XxxRelEnv : public XxxEngineEnv {
 public:
  // Maintain a stack of nested environments for correlations values
  XxxStack<XxxArgEnv *> NestedState;

  // Maintain a stack of table names (used by the parser)
  XxxStack<char *> NestedTableName;

  /******** FLAGS FOR JOIN OPTIMIZER ********/
  // push projections
  static const char* const PUSHPROJ;
  static const char REL_OPT_PUSHPROJ;
  char PushProjection;

  enum JoinEnumeration {
    NAIVE_JO = 0,
    SEMINAIVE_JO = 1,
    GREEDY_JO = 2,
    KBZ_JO = 3,
    DYNPROG_JO = 4
  };
  static const char* const JOINORDER;
  static const char REL_OPT_JOINORDER;
  char JoinOrder;

  // should cross-products be considered
  static const char* const TRY_CROSS_PRODUCT;
  static const char REL_OPT_CROSSPRODUCT;
  char TryCrossProduct;

  // pipeline joins by default
  static const char* const PIPELINE;
  static const char REL_OPT_PIPELINE;
  char PipelineJoins;

  // materialize join results by default
  static const char* const MATERIALIZE;
  static const char REL_OPT_MATERIALIZE;
  char MaterializeTemps;

  /***** FLAGS FOR JOIN METHODS ****/
  // use sort-merge join when appropriate
  static const char* const SORT_MERGE_JOIN;
  static const char SORT_MERGE_ENABLED;
  char SortMergeEnabled;

  // added by Wee Hyong, Tok
  // tokweehy@comp.nus.edu.sg
  // use symmetric hash join when set
  static const char* const HASH_JOIN;
  static const char HASH_JOIN_ENABLED;
  char HashJoinEnabled;

  enum NLJoinMethod {
    TUPLE_NL = 0,
    PAGE_NL = 1,
  };
  // default NL join method
  static const char* const NLJOINMETH;
  static const char REL_OPT_NLJOINMETH;
  char NLJoinMethod;


  //Begin Moustafa
  enum WJoinMethod {
    HBEW_JOIN = 0,
    NLBEW_JOIN = 1,
    HFEW_JOIN = 2,
    NLFEW_JOIN = 3,
  };
  // default W join method
  static const char* const WJOINMETH;
  static const char REL_OPT_WJOINMETH;
  char WJoinMethod;
  //End Moustafa


  /***** FLAGS FOR PREDICATES *****/
  // push selections
  static const char* const PUSHSEL;
  static const char REL_OPT_PUSHSEL;
  char PushSelection;

  enum PredOrdering {
    SELECTIVITY_PO = 0,
    COST_PO = 1,
    RANK_PO = 2,
  };
  // flags for Selection Ordering
  static const char *const SELECTIONORDERING;
  static const char REL_OPT_SELECTIONORDERING;
  char SelectionOrdering;

  /* flags for client site UDF batch size (concurrency factor, the number
  of tuples that are in the pipeline sender-client-receiver */
  static const char *const UDFBATCHSIZE;
  static const int REL_OPT_UDFBATCHSIZE;
  int UDFBatchSize;
  /* bitfield for various UDF execution options:
  bit 0 : reads and writes from the local file system, instead of the network
  bit 1 : uses separate thread for the sender
  bit 2 : do a client-site join instead of the usual semi-join
  */
  static const char *const SET_UDF_EXEC;
  static const char *const UNSET_UDF_EXEC;
  static const int REL_OPT_UDF_EXEC;
  int UDFExec;

  enum ExpensivePredStrategy {
    IGNORE_EP = 0,
    PULLUP_EP = 1,  // predicate pullup
    MIGRATE_EP = 2, // predicate migration -- not supported
  };
  static const char* const EXPENSIVE_PRED;
  static const enum ExpensivePredStrategy REL_OPT_EXPPRED;
  enum ExpensivePredStrategy ExpPredStrategy;

  double CheapPredMax; // max cost for a non-expensive predicate

  /***** FLAGS FOR GENERIC QUERY PROCESSING *****/

  // show query results
  static const char* const SHOW_QUERY_RESULTS;
  static const bool SHOW_Q_RESULTS;
  bool ShowQueryResults;

  // show query processing information
  static const char* const SHOW_QUERY_PROC_INFO;
  static const char SHOW_QP_INFO;
  char ShowQPInfo;
  ofstream *QPInfoFile;

  // perform query rewrite
  static const char* const REWRITE;
  static const char REL_OPT_REWRITE;
  char RewriteQuery;

  /***** FLAGS for data loading *****/
  static const char* const DELIMITED;
  static const char REL_OPT_DELIMITED;
  char Delimited;

  static const char* const DELIMITER;
  static const char REL_OPT_DELIMITER;
  char Delimiter;

  /***** debugging flags *****/
  XxxBool DebugQrw;
  XxxBool DebugOpt;
  XxxBool DebugExec;
  XxxBool DebugStats;
  XxxBool DebugStorage;
  XxxBool DebugCompile;

  // use this temporarily if you want to avoid changing header file
  int MoreFlags[10];

  XxxRelEnv();
  ~XxxRelEnv();
  virtual void Print(ostream& OStr) const;
  virtual XxxErrCode SetValue(const char* ValName,
			      char NewVal);
		      
 };

// Get the current ADT environment
inline XxxRelEnv* XXX_THISRELENV()
{ return (XxxRelEnv*)(XXX_THISENGENV(RelationEngineId)); }

#endif  /* XXX_RELENV_H */
