#include <stdio.h>
#include <stdlib.h>

#ifdef __GNUG__
#pragma implementation
#endif

#include "relops.h"
#include "relcat.h"
#include "relquery.h"

// OWNER: as  DATE: 02/24/97
// PURPOSE: sets the DistinctCond for the Query Rewrite Engine
// NOTES: added to facilitate query rewrite
RelOperator::RelOperator(XxxBool Distinct) {
  if(Distinct == XXX_FALSE) 
    DistinctCond = QR_PRESERVE;
  else
    DistinctCond = QR_ENFORCE;
}

// OWNER: as  DATE: 02/24/97
// PURPOSE: sets the DistinctCond for the Query Rewrite Engine
// NOTES: added to facilitate query rewrite
// default constructor
RelOperator::RelOperator() {
  DistinctCond = QR_PRESERVE;
}

RelSPJOp::RelSPJOp(XxxBooleanExpression* Cond, 
		   XxxValueExprList* ExprList,
		   XxxBool Distinct, XxxErrCode &RC) 
  : RelOperator(Distinct)
{
  this->Distinct = Distinct;
  Condition = Cond;
  // convert the expr to CNF
  XxxBooleanExpression::ConvertToCNF(Condition);
  TargetList = ExprList;
  RC = XXX_OK;
}

RelSPJOp::~RelSPJOp()
{
 if (Condition) delete Condition;
 if (TargetList) TargetList->Destroy();
}

RelImplInfo*
RelSPJOp::GenerateImplInfo(int NumChildren, RelQueryNode** Children)
{
 if (!NumChildren)
   return NULL;
 
 return new RelImplInfo(*(Children[0]->CatalogInfo->ImplInfo()), XXX_FALSE);
 // return Children[0]->CatalogInfo->ImplInfo()->Replicate(XXX_FALSE);
}

XxxRecordSchema* 
RelSPJOp::GenerateRecordSchema(int NumChildren, RelQueryNode** Children,
			       const XxxArgEnv* Env)
{
  if (!NumChildren) return NULL;
  /*
   * Replace all "TableName.*" unknown value plans with their lists of
   * variables.
   */
  XxxValueExprList *Tmp = TargetList;
  char NodeName[XXX_NODE_NAME_MAX];
  while (Tmp) {
    if (Tmp->Expr->ValType() != XxxValueExpression::COLUMN_VAR) {
      Tmp = Tmp->Next;
      continue;
    }
    if (strcmp(((XxxUnknownValue *)(Tmp->Expr))->Name, "*")) {
      Tmp = Tmp->Next;
      continue;
    }
    strcpy(NodeName, ((XxxUnknownValue *)(Tmp->Expr))->XxxName);
    // This is a wildcard expression; first remove it
    if (Tmp == TargetList) {
      TargetList = Tmp->Next;
    } else {
      XxxValueExprList *Parent;
      for (Parent = TargetList; 
	   Parent->Next != Tmp;
	   Parent = Parent->Next);
      Parent->Next = Tmp->Next;
    }
    Tmp->Next = NULL;
    Tmp->Destroy();
    // Now find the child which matched the node name of the parent
    int SrcChild, AttrIndex;
    for (SrcChild = 0; 
	 ((SrcChild < NumChildren) && 
	  !(Children[SrcChild]->MatchNodeName(NodeName))); SrcChild++);
    if (SrcChild >= NumChildren) {
      // The specified input name was not found
      char buf[100];
      sprintf(buf, "%s is not a specified input", NodeName);
      REL_NOTE_ERROR(REL_BADQUERY, buf);
      return NULL;
    }
    // Add all the attributes of that child to the targetlist
    XxxRecordSchema* ChildSchema = 
      Children[SrcChild]->CatalogInfo->RecordSchema();
    for (AttrIndex = 0; AttrIndex < ChildSchema->Arity(); AttrIndex++) {
      const char* AttrName = ((*ChildSchema)[AttrIndex]).Name;
      XxxValueExpression* NewExpr = 
	new XxxUnknownValue(AttrName, NodeName);
      ((XxxUnknownValue *)NewExpr)->SrcAttrIndex = AttrIndex;
      ((XxxUnknownValue *)NewExpr)->SrcChild = SrcChild;
      ((XxxUnknownValue *)NewExpr)->BaseType = 
	((*ChildSchema)[AttrIndex]).Type;
      if (TargetList)
	TargetList = TargetList->AddExpr(NewExpr);
      else
	TargetList = new XxxValueExprList(NewExpr);
    }
    // Now start over
    Tmp = TargetList;
  }

  // resolve all column names referenced by this op
  if (ResolveReferences(NumChildren, Children, Env) 
      != XXX_OK) {
    char buf[100];
    sprintf(buf, "resolve references of %d failed!\n", (int)(Type()));
    REL_NOTE_ERROR(REL_BADQUERY, buf);
    return NULL;
  }
  
  if (TypeCheck() != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, "Type Error");
    return NULL;
  }
  
  int NumAttrs;
  NumAttrs = TargetList->Count();
  
  XxxRecordSchema *ReturnSchema = new XxxRecordSchema(NumAttrs);
  
  // fill in the attributes
  XxxValueExprList* EList = TargetList;
  int i;
  for (i = 0; i < NumAttrs; i++, EList = EList->Next) {
    if (EList->Expr->ValType() == XxxValueExpression::COLUMN_VAR) { 
      // simple expression R.X
      XxxUnknownValue* VarExpr = (XxxUnknownValue *)(EList->Expr);
      if ((VarExpr->SrcChild < 0) || (VarExpr->SrcChild >= NumChildren)) {
	XXX_NOTE_ERROR(XXX_INTERNAL, "Error in SrcChild of ColumnExpression");
	return NULL;
      }
      int k = VarExpr->SrcAttrIndex;
      XxxRecordSchema* ChildSchema = 
	Children[VarExpr->SrcChild]->CatalogInfo->RecordSchema();
      
      if (VarExpr->ExprName) {
	(*ReturnSchema)[i].Assign(VarExpr->ExprName,
				  ChildSchema->GetAttr(k)->Type,
				  VarExpr->GenerateCatInfo());
      }
      else {
	(*ReturnSchema)[i].Assign(*(ChildSchema->GetAttr(k)));
      }
      (*ReturnSchema)[i].AttrSource = VarExpr->SrcChild;
      (*ReturnSchema)[i].AttrIndex = k;
    }
    else { // complex expression or constant
      if (EList->Expr->ExprName) {
	(*ReturnSchema)[i].Assign(EList->Expr->ExprName,
				  EList->Expr->BasicType(),
				  EList->Expr->GenerateCatInfo());
      }
      else {
	char TmpBuf[10]; sprintf(TmpBuf, "#%d", i);
	(*ReturnSchema)[i].Assign(TmpBuf, EList->Expr->BasicType(),
				  EList->Expr->GenerateCatInfo());
      }
      (*ReturnSchema)[i].AttrSource = -1;
    }
  }
  
  // have to do this before returning
  ReturnSchema->ComputeStats();
  return ReturnSchema;
}

RelStatsInfo* 
RelSPJOp::GenerateStatsInfo(int NumChildren, RelQueryNode** Children)
{
 if (!NumChildren)
   return NULL;

 // determine the selectivity information for the predicates
 XxxBooleanExpression::SelectivityStatsInfo* SelectivityStats = 
   new XxxBooleanExpression::SelectivityStatsInfo[NumChildren];
 int AttrCount = 0;
 for (int i = 0; i < NumChildren; i++) {
   RelCatalogInfo* CatInfo = Children[i]->CatalogInfo;
   SelectivityStats[i].RecSchema = CatInfo->RecordSchema();
   SelectivityStats[i].AttrStats = 
     ((CatInfo->StatsInfo())?(CatInfo->StatsInfo()->AttrStats): 0);
   AttrCount += CatInfo->RecordSchema()->Arity();
  }

 RelStatsInfo* StatsInfo = new RelStatsInfo(AttrCount);
 if (NumChildren == 1) {
   if (Condition) 
     StatsInfo->Cardinality 
       = (int)(StatsInfo->Cardinality * 
	       Condition->Selectivity(SelectivityStats, NumChildren));
  }
 else {
   for (int i = 1; i < NumChildren; i++ ) {
     StatsInfo->Cardinality *= 
       Children[i]->CatalogInfo->StatsInfo()->Cardinality;
    }
   if (Condition)
     StatsInfo->Cardinality 
       = (int)(StatsInfo->Cardinality 
	       * Condition->Selectivity(SelectivityStats, NumChildren));
  }
 delete [] SelectivityStats;
 return StatsInfo;
}

void RelSPJOp::Print(ostream& OStr) const 
{
 OStr << "RelSPJOp";
}

void RelSPJOp::NodePrint(ostream &OStr, int curBlockId, int Level, 
			 int numChildren, int *childrenIds) const
{
  OStr << "NODE Select Project Join" << endl;
  OStr << "ID " << curBlockId << endl;
  OStr << "LEVEL " << Level << endl;
  if (numChildren > 0) {
    OStr << "CHILDREN ";
    for (int i = 0; i < numChildren; i++) {
      OStr << childrenIds[i] << " ";
    }
    OStr << endl;
  }
  OStr << "COLOR " << 0x00FF00 << endl;
  OStr << "INFO Distinct: " << ((Distinct == XXX_TRUE) ? "Yes" : "No") << endl;
  OStr << "INFO Projection: ";
  if (TargetList)
    TargetList->Print(OStr);
  else
    OStr << "*";
  OStr << endl;
  OStr << "INFO Selection: ";
  if (Condition) {
    Condition->Print(OStr);
    OStr << endl;
  } else {
    OStr << "none" << endl;
  }
  OStr << "ENDNODE" << endl;
}

XxxErrCode 
RelSPJOp::ResolveReferences(int NumChildren,
			    RelQueryNode** Children,
			    const XxxArgEnv* Env)
{
 for (int i = 0; i < NumChildren; i++) {
   if (Condition && (Condition->ResolveVars(Children[i], i) != XXX_OK)) {
     REL_NOTE_ERROR(REL_BADQUERY, "Error resolving condition references");
     return XXX_NOT_OK;
    }
   if (TargetList && (TargetList->ResolveVars(Children[i], i) != XXX_OK)) {
     REL_NOTE_ERROR(REL_BADQUERY, "Error resolving targetlist references");
     return XXX_NOT_OK;
    }
  } // for

 if (Env) {
   if (Condition && (Condition->ResolveVars(Env) != XXX_OK)) {
     REL_NOTE_ERROR(REL_BADQUERY, 
		    "Unable to resolve refs of SPJ condition using host env");
     return XXX_NOT_OK;
    }
   if (TargetList && (TargetList->ResolveVars(Env) != XXX_OK)) {
     REL_NOTE_ERROR(REL_BADQUERY, 
		    "Unable to resolve refs of SPJ targetlist using host env");
     return XXX_NOT_OK;
    }
  } // if Env

 return XXX_OK;
}

XxxErrCode RelSPJOp::GetDependencies(XxxBitMap& Dependency,
				     int CorrelationHt) const
{
  if (Condition) {
    XXX_CALL_RET(Condition->GetDependencies(Dependency, CorrelationHt),
		 REL_BADQUERY, "Error getting correlation dependencies");
  }
  if (TargetList) {
    XXX_CALL_RET(TargetList->GetDependencies(Dependency, CorrelationHt),
		 REL_BADQUERY, "Error getting correlation dependencies");
  }
  return XXX_OK;
}

XxxErrCode 
RelSPJOp::TypeCheck()
{
 if (Condition && (Condition->TypeCheck() != XXX_OK)) {
   REL_NOTE_ERROR(REL_BADQUERY, "Condition Type Check failed for SPJ op");
   return XXX_NOT_OK;
  }
 if (TargetList && (TargetList->TypeCheck() != XXX_OK)) {
   REL_NOTE_ERROR(REL_BADQUERY, "TargetList Type Check failed for SPJ op");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

//
// Methods of RelGrpByOp
//

RelGrpByOp::RelGrpByOp(XxxValueExprList* TList,
		       XxxValueExprList* AList,
		       XxxValueExprList* GList, 
		       XxxErrCode &RC) 
: TargetList(TList), GroupList(GList), AggrList(AList)
{
  RC = XXX_OK;

  if (!AggrList) {
    REL_NOTE_ERROR(REL_BADQUERY, "No aggregates on which to group");
    RC = XXX_NOT_OK;
  }
/*
  for (int i = 0; i < AList->Count(); i++)
    if (!(((*AList)[i]->ValType() == XxxValueExpression::FUNC_EXPR) ||
           ((*AList)[i]->ValType() == XxxValueExpression::CONST_EXPR))) {
      XXX_NOTE_ERROR(XXX_INCOMPL, 
		     "Cannot currently aggregate complex non-functional "
		     "expressions");
      RC = XXX_NOT_OK;
      return;
    }
*/
}

// OWNER: as  DATE: 2/20/97
// PURPOSE: destructor
RelGrpByOp::~RelGrpByOp() {
  if (TargetList) TargetList->Destroy();
  if (GroupList) GroupList->Destroy();
  if (AggrList) AggrList->Destroy();
}

RelImplInfo*
RelGrpByOp::GenerateImplInfo(int NumChildren, RelQueryNode** Children)
{
 if (!NumChildren)
   return NULL;
 
 return new RelImplInfo(*(Children[0]->CatalogInfo->ImplInfo()), XXX_FALSE);
 // return Children[0]->CatalogInfo->ImplInfo()->Replicate(XXX_FALSE);

}
RelStatsInfo* 
RelGrpByOp::GenerateStatsInfo(int NumChildren, RelQueryNode** Children)
{
 if (NumChildren != 1) {
   XXX_NOTE_ERROR(XXX_INTERNAL, 
		  "Relational Aggregate should have one input");
   return NULL;
  }

 RelCatalogInfo* ChildCatInfo = Children[0]->CatalogInfo;
 RelStatsInfo* StatsInfo = 
   new RelStatsInfo(ChildCatInfo->RecordSchema()->Arity());
 StatsInfo->Cardinality = 1;  
 // TODO:LOpt this must be changed now that multiple columns are implemented
 return StatsInfo;
}

void RelGrpByOp::Print(ostream& OStr) const 
{
 OStr << "RelGBYOp";
}

void RelGrpByOp::NodePrint(ostream &OStr, int curBlockId, int Level, 
			   int numChildren, int *childrenIds) const
{
  OStr << "NODE Grouping / Aggregation" << endl;
  OStr << "ID " << curBlockId << endl;
  OStr << "LEVEL " << Level << endl;
  if (numChildren > 0) {
    OStr << "CHILDREN ";
    for (int i = 0; i < numChildren; i++) {
      OStr << childrenIds[i] << " ";
    }
    OStr << endl;
  }
  OStr << "COLOR " << 0x0000FF << endl;
  OStr << "INFO Projection: ";
  if (TargetList)
    TargetList->Print(OStr);
  else
    OStr << "*";
  OStr << endl;
  OStr << "INFO Groupings: ";
  if (GroupList)
    GroupList->Print(OStr);
  else
    OStr << "none";
  OStr << endl;
  OStr << "INFO Aggregations: ";
  if (AggrList)
    AggrList->Print(OStr);
  else
    OStr << "none";
  OStr << endl;
  OStr << "ENDNODE" << endl;
}

XxxErrCode 
RelGrpByOp::TypeCheck() {
  if (TargetList && (TargetList->TypeCheck() != XXX_OK)) {
    REL_NOTE_ERROR(REL_BADQUERY, "TargetList Type Check failed for GBY op");
    return XXX_NOT_OK;
  }
  if (AggrList && (AggrList->TypeCheck() != XXX_OK)) {
    REL_NOTE_ERROR(REL_BADQUERY, "AggrList Type Check failed for GBY op");
    return XXX_NOT_OK;
  }
  if (GroupList && (GroupList->TypeCheck() != XXX_OK)) {
    REL_NOTE_ERROR(REL_BADQUERY, "GroupList Type Check failed for GBY op");
    return XXX_NOT_OK;
  }
  return XXX_OK;
}

XxxErrCode 
RelGrpByOp::ResolveReferences(int NumChildren,
			      RelQueryNode** Children,
			      const XxxArgEnv* Env) 
 {  
  if (TargetList && TargetList->ResolveVars(Children[0], 0) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, 
		   "Unable to resolve vars of TargetList");
    return XXX_NOT_OK;
  }
  if (GroupList && GroupList->ResolveVars(Children[0], 0) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, 
		   "Unable to resolve vars of GroupList");
    return XXX_NOT_OK;
  }
  if (AggrList && AggrList->ResolveVars(Children[0], 0) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, 
		   "Unable to resolve vars of AggrList");
    return XXX_NOT_OK;
  }
  if (Env) {
   if (TargetList && TargetList->ResolveVars(Env) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, 
     "Unable to resolve refs of TargetList using host env");
    return XXX_NOT_OK;
    }
   if (GroupList && GroupList->ResolveVars(Env) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, 
     "Unable to resolve refs of GroupList using host env");
    return XXX_NOT_OK;
    }
   if (AggrList && AggrList->ResolveVars(Env) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, 
     "Unable to resolve refs of AggrList using host env");
    return XXX_NOT_OK;
    }
   }

  return XXX_OK;
}

XxxRecordSchema* 
RelGrpByOp::GenerateRecordSchema(int NumChildren, RelQueryNode** Children,
				 const XxxArgEnv* Env)
{
  if (!NumChildren) 
    return NULL;
  
  // resolve all column names referenced by this op
  if (ResolveReferences(NumChildren, Children, Env) 
      != XXX_OK) {
    char buf[100];
    sprintf(buf, "resolve references of %d failed!\n", (int)(Type()));
    REL_NOTE_ERROR(REL_BADQUERY, buf);
    return NULL;
  }

  if (TypeCheck() != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, "Type Error");
    return NULL;
  }

  int NumAttrs;
  NumAttrs = TargetList->Count();

  XxxRecordSchema *ReturnSchema = new XxxRecordSchema(NumAttrs);

  // fill in the attributes
  XxxValueExprList* EList = TargetList;
  for (int i = 0; i < NumAttrs; i++, EList = EList->Next) {
    if (EList->Expr->ValType() == XxxValueExpression::COLUMN_VAR) { 
      // simple expression R.X
      XxxUnknownValue* VarExpr = (XxxUnknownValue *)(EList->Expr);
      if ((VarExpr->SrcChild < 0) || (VarExpr->SrcChild >= NumChildren)) {
	XXX_NOTE_ERROR(XXX_INTERNAL, "Error in SrcChild of ColumnExpression");
	return NULL;
      }
      int k = VarExpr->SrcAttrIndex;
      XxxRecordSchema* ChildSchema = 
	Children[VarExpr->SrcChild]->CatalogInfo->RecordSchema();

      if (VarExpr->ExprName) {
	(*ReturnSchema)[i].Assign(VarExpr->ExprName,
				  ChildSchema->GetAttr(k)->Type,
				  VarExpr->GenerateCatInfo());
      }
      else {
	(*ReturnSchema)[i].Assign(*(ChildSchema->GetAttr(k)));
      }
      (*ReturnSchema)[i].AttrSource = VarExpr->SrcChild;
      (*ReturnSchema)[i].AttrIndex = k;
    }
    else { // complex expression or constant
      if (EList->Expr->ExprName) {
	(*ReturnSchema)[i].Assign(EList->Expr->ExprName,
				  EList->Expr->BasicType(),
				  EList->Expr->GenerateCatInfo());
      }
      else {
	char TmpBuf[10]; sprintf(TmpBuf, "#%d", i);
	(*ReturnSchema)[i].Assign(TmpBuf, EList->Expr->BasicType(),
				  EList->Expr->GenerateCatInfo());
      }
      (*ReturnSchema)[i].AttrSource = -1;
    }
  }
  
  // have to do this before returning
  ReturnSchema->ComputeStats();
  return ReturnSchema;
}

XxxErrCode RelGrpByOp::GetDependencies(XxxBitMap& Dependency,
				       int CorrelationHt) const
{
  if (TargetList) {
    XXX_CALL_RET(TargetList->GetDependencies(Dependency, CorrelationHt),
		 REL_BADQUERY, "Error getting correlation dependencies");
  }
  if (GroupList) {
    XXX_CALL_RET(GroupList->GetDependencies(Dependency, CorrelationHt),
		 REL_BADQUERY, "Error getting correlation dependencies");
  }
  if (AggrList) {
    XXX_CALL_RET(AggrList->GetDependencies(Dependency, CorrelationHt),
		 REL_BADQUERY, "Error getting correlation dependencies");
  }
  return XXX_OK;
}

//
// Methods of RelOrderByOp
//
void RelOrderByOp::Print(ostream& OStr) const 
{
 OStr << "RelOBYOp";
}

void RelOrderByOp::NodePrint(ostream &OStr, int curBlockId, int Level, 
			     int numChildren, int *childrenIds) const
{
  OStr << "NODE Ordering" << endl;
  OStr << "ID " << curBlockId << endl;
  OStr << "LEVEL " << Level << endl;
  if (numChildren > 0) {
    OStr << "CHILDREN ";
    for (int i = 0; i < numChildren; i++) {
      OStr << childrenIds[i] << " ";
    }
    OStr << endl;
  }
  OStr << "COLOR " << 0x00CCCC << endl;
  // print TargetList
  OStr << "INFO Ordering attributes: ";
  if (TargetList)
    TargetList->Print(OStr);
  else
    OStr << "none";
  OStr << endl;
  OStr << "ENDNODE" << endl;
}

RelOrderByOp::RelOrderByOp (XxxValueExprList *TList,
			    XxxBool *Asc, XxxErrCode &RC)
: TargetList(TList), Ascending(Asc)
{
  if (!Ascending) {
    REL_NOTE_ERROR(REL_BADOP, "Ascending not supplied");
    RC = XXX_NOT_OK;
    return;
  }
  for (int i = 0; i < TList->Count(); i++)
    if ((*TList)[i]->ValType() != XxxValueExpression::COLUMN_VAR) {
      XXX_NOTE_ERROR(XXX_INCOMPL, 
		     "Cannot currently order by complex expressions");
      RC = XXX_NOT_OK;
      return;
    }

  RC = XXX_OK;
  return;
}

RelOrderByOp::~RelOrderByOp()
{
 if (TargetList) TargetList->Destroy();
 if (Ascending) delete Ascending;
}

RelImplInfo*
RelOrderByOp::GenerateImplInfo(int NumChildren, RelQueryNode** Children)
{
 if (!NumChildren)
   return NULL;
 
 return new RelImplInfo(*(Children[0]->CatalogInfo->ImplInfo()), XXX_FALSE);
 // return Children[0]->CatalogInfo->ImplInfo()->Replicate(XXX_FALSE);

}

RelStatsInfo* 
RelOrderByOp::GenerateStatsInfo(int NumChildren, RelQueryNode** Children) 
{
  if (NumChildren != 1) {
    XXX_NOTE_ERROR(XXX_INTERNAL, 
		   "orderby should have one input");
    return NULL;
  }
  
  RelStatsInfo* StatsInfo;
  RelCatalogInfo* ChildCatInfo = Children[0]->CatalogInfo;
  StatsInfo = ChildCatInfo->StatsInfo()->Replicate(ChildCatInfo->RecordSchema());
  return StatsInfo;
}

XxxRecordSchema* 
RelOrderByOp::GenerateRecordSchema(int NumChildren, RelQueryNode** Children,
				   const XxxArgEnv* Env) 
{
  if (NumChildren != 1) {
    REL_NOTE_ERROR(REL_BADQUERY, "Order By should have only one input");
    return NULL;
  }
  // resolve all column names referenced by this op
  if (ResolveReferences(NumChildren, Children, Env) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, "resolve references of OrderByOp failed");
    return NULL;
  }
  if (TypeCheck() != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, "Type Error");
    return NULL;
  }
  XxxRecordSchema* ReturnSchema = 
    new XxxRecordSchema(*(Children[0]->CatalogInfo->RecordSchema()));
  for (int i = 0; i < ReturnSchema->Arity(); i++) {
    (*ReturnSchema)[i].AttrSource = 0;
  }
  ReturnSchema->ComputeStats();
  return ReturnSchema;
}

// ResolveReferences
XxxErrCode 
RelOrderByOp::ResolveReferences(int NumChildren,
				RelQueryNode** Children,
				const XxxArgEnv* Env) 
{
  if (TargetList && TargetList->ResolveVars(Children[0], 0) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, 
		   "Unable to resolve vars of TargetList");
    return XXX_NOT_OK;
  }
  if (Env) {
    if (TargetList && TargetList->ResolveVars(Env) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADQUERY, 
		     "Unable to resolve refs of TargetList using host env");
      return XXX_NOT_OK;
    }
  }
  return XXX_OK;
}

XxxErrCode 
RelOrderByOp::TypeCheck() 
{
  if (!TargetList) {
    REL_NOTE_ERROR(REL_BADEXEC, "No TargetList supplied");
    return XXX_NOT_OK;
  }
  if (TargetList->TypeCheck() != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, 
		   "TargetList Type Check failed for OrderBy op");
    return XXX_NOT_OK;
  }
  return XXX_OK;
}

XxxErrCode RelOrderByOp::GetDependencies(XxxBitMap& Dependency,
					 int CorrelationHt) const
{
  if (TargetList) {
    XXX_CALL_RET(TargetList->GetDependencies(Dependency, CorrelationHt),
		 REL_BADQUERY, "Error getting correlation dependencies");
  }
  return XXX_OK;
}


//begin Ihab
//methods for KRJOP
void RelKRJOp::Print(ostream& OStr) const 
{
 OStr << "RelKRJOp";
}

void RelKRJOp::NodePrint(ostream &OStr, int curBlockId, int Level, 
			 int numChildren, int *childrenIds) const
{
  OStr << "NODE Multiple NN" << endl;
  OStr << "ID " << curBlockId << endl;
  OStr << "LEVEL " << Level << endl;
  if (numChildren > 0) {
    OStr << "CHILDREN ";
    for (int i = 0; i < numChildren; i++) {
      OStr << childrenIds[i] << " ";
    }
    OStr << endl;
  }
  OStr << "COLOR " << 0x00FF00 << endl;
  OStr << "INFO Distinct: " << ((Distinct == XXX_TRUE) ? "Yes" : "No") << endl;
  OStr << "INFO Projection: ";
  //if (TargetList)
  //  TargetList->Print(OStr);
  //else
    OStr << "*";
  OStr << endl;
  OStr << "INFO Selection: ";
  if (Condition) {
    Condition->Print(OStr);
    OStr << endl;
  } else {
    OStr << "none" << endl;
  }
  OStr << "ENDNODE" << endl;
}


XxxErrCode 
RelKRJOp::ResolveReferences(int NumChildren,
			   RelQueryNode** Children,
			   const XxxArgEnv* Env)
{
  for (int i = 0; i < NumChildren; i++) {
    if (Condition && (Condition->ResolveVars((RelQueryNode*)Children[i], i) != XXX_OK)) {
      REL_NOTE_ERROR(REL_BADQUERY, "Error resolving condition references");
      return XXX_NOT_OK;
    }
    if (TargetList && (TargetList->ResolveVars(Children[i], i) != XXX_OK)) {
      REL_NOTE_ERROR(REL_BADQUERY, "Error resolving targetlist references");
      return XXX_NOT_OK;
    }
    if (Ordering && (Ordering->ResolveVars(Children[i], i) != XXX_OK)) {
      REL_NOTE_ERROR(REL_BADQUERY, "Error resolving ranking list references");
      return XXX_NOT_OK;
    }
  } // for
  
 if (Env) {
   if (Condition && (Condition->ResolveVars(Env) != XXX_OK)) {
     REL_NOTE_ERROR(REL_BADQUERY, 
		    "Unable to resolve refs of KRJ condition using host env");
     return XXX_NOT_OK;
    }
   if (TargetList && (TargetList->ResolveVars(Env) != XXX_OK)) {
     REL_NOTE_ERROR(REL_BADQUERY, 
		    "Unable to resolve refs of KRJ targetlist using host env");
     return XXX_NOT_OK;
   }
   if (Ordering && (Ordering->ResolveVars(Env) != XXX_OK)) {
     REL_NOTE_ERROR(REL_BADQUERY, "Error resolving KRJ ranking list references");
     return XXX_NOT_OK;
   }
   
 } // if Env
 
 return XXX_OK;
}

XxxErrCode RelKRJOp::GetDependencies(XxxBitMap& Dependency,
				    int CorrelationHt) const
{
  if (Condition) {
    XXX_CALL_RET(Condition->GetDependencies(Dependency, CorrelationHt),
		 REL_BADQUERY, "Error getting correlation dependencies");
  }
  if (TargetList) {
    XXX_CALL_RET(TargetList->GetDependencies(Dependency, CorrelationHt),
		 REL_BADQUERY, "Error getting correlation dependencies");
  }

  if (Ordering) {
    XXX_CALL_RET(Ordering->GetDependencies(Dependency, CorrelationHt),
		 REL_BADQUERY, "Error getting correlation dependencies");
  }
  return XXX_OK;
}

XxxErrCode RelKRJOp::TypeCheck()
{
  if (Condition && (Condition->TypeCheck() != XXX_OK)) {
    REL_NOTE_ERROR(REL_BADQUERY, "Condition Type Check failed for KRJ op");
    return XXX_NOT_OK;
  }
  if (TargetList && (TargetList->TypeCheck() != XXX_OK)) {
    REL_NOTE_ERROR(REL_BADQUERY, "TargetList Type Check failed for KRJ op");
    return XXX_NOT_OK;
  }

  if (Ordering && (Ordering->TypeCheck() != XXX_OK)) {
    REL_NOTE_ERROR(REL_BADQUERY, "TargetList Type Check failed for KRJ op");
    return XXX_NOT_OK;
  } 
  return XXX_OK;
}

XxxRecordSchema* 
RelKRJOp::GenerateRecordSchema(int NumChildren, RelQueryNode** Children,const XxxArgEnv* Env)
{
  if (!NumChildren) return NULL;
  /*
   * Replace all "TableName.*" unknown value plans with their lists of
   * variables.
   */
  XxxValueExprList *Tmp = TargetList;
  char NodeName[XXX_NODE_NAME_MAX];
  while (Tmp) {
    if (Tmp->Expr->ValType() != XxxValueExpression::COLUMN_VAR) {
      Tmp = Tmp->Next;
      continue;
    }
    if (strcmp(((XxxUnknownValue *)(Tmp->Expr))->Name, "*")) {
      Tmp = Tmp->Next;
      continue;
    }
    strcpy(NodeName, ((XxxUnknownValue *)(Tmp->Expr))->XxxName);
    // This is a wildcard expression; first remove it
    if (Tmp == TargetList) {
      TargetList = Tmp->Next;
    } else {
      XxxValueExprList *Parent;
      for (Parent = TargetList; 
	   Parent->Next != Tmp;
	   Parent = Parent->Next);
      Parent->Next = Tmp->Next;
    }
    Tmp->Next = NULL;
    Tmp->Destroy();
    // Now find the child which matched the node name of the parent
    int SrcChild, AttrIndex;
    for (SrcChild = 0; 
	 ((SrcChild < NumChildren) && 
	  !(Children[SrcChild]->MatchNodeName(NodeName))); SrcChild++);
    if (SrcChild >= NumChildren) {
      // The specified input name was not found
      char buf[100];
      sprintf(buf, "%s is not a specified input", NodeName);
      REL_NOTE_ERROR(REL_BADQUERY, buf);
      return NULL;
    }
    // Add all the attributes of that child to the targetlist
    XxxRecordSchema* ChildSchema = 
      Children[SrcChild]->CatalogInfo->RecordSchema();
    for (AttrIndex = 0; AttrIndex < ChildSchema->Arity(); AttrIndex++) {
      const char* AttrName = ((*ChildSchema)[AttrIndex]).Name;
      XxxValueExpression* NewExpr = 
	new XxxUnknownValue(AttrName, NodeName);
      ((XxxUnknownValue *)NewExpr)->SrcAttrIndex = AttrIndex;
      ((XxxUnknownValue *)NewExpr)->SrcChild = SrcChild;
      ((XxxUnknownValue *)NewExpr)->BaseType = 	((*ChildSchema)[AttrIndex]).Type;
      if (TargetList)
	TargetList = TargetList->AddExpr(NewExpr);
      else
	TargetList = new XxxValueExprList(NewExpr);
    }
    // Now start over
       Tmp = TargetList;
  }

  // resolve all column names referenced by this op
  if (ResolveReferences(NumChildren, Children, Env) 
      != XXX_OK) {
    char buf[100];
    sprintf(buf, "resolve references of %d failed!\n", (int)(Type()));
    REL_NOTE_ERROR(REL_BADQUERY, buf);
    return NULL;
  }
  
  if (TypeCheck() != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, "Type Error");
    return NULL;
  }
  
  int NumAttrs;
  NumAttrs = TargetList->Count();
  
  XxxRecordSchema *ReturnSchema = new XxxRecordSchema(NumAttrs); 
  
  // fill in the attributes
  XxxValueExprList* EList = TargetList;
  int i;
  for (i = 0; i < NumAttrs; i++, EList = EList->Next) {
    if (EList->Expr->ValType() == XxxValueExpression::COLUMN_VAR ) {// not for WG and BG 
      // simple expression R.X
      XxxUnknownValue* VarExpr = (XxxUnknownValue *)(EList->Expr);
      if ((VarExpr->SrcChild < 0) || (VarExpr->SrcChild >= NumChildren)) {
	XXX_NOTE_ERROR(XXX_INTERNAL, "Error in SrcChild of ColumnExpression");
	return NULL;
      }
      int k = VarExpr->SrcAttrIndex;
      XxxRecordSchema* ChildSchema = 
	Children[VarExpr->SrcChild]->CatalogInfo->RecordSchema();
      
      if (VarExpr->ExprName) {
	(*ReturnSchema)[i].Assign(VarExpr->ExprName,
				  ChildSchema->GetAttr(k)->Type,
				  VarExpr->GenerateCatInfo());
      }
      else {
	(*ReturnSchema)[i].Assign(*(ChildSchema->GetAttr(k)));
      }
      (*ReturnSchema)[i].AttrSource = VarExpr->SrcChild;
      (*ReturnSchema)[i].AttrIndex = k;
    }
    else { // complex expression or constant
      if (EList->Expr->ExprName) {
	(*ReturnSchema)[i].Assign(EList->Expr->ExprName,
				  EList->Expr->BasicType(),
				  EList->Expr->GenerateCatInfo());
      }
      else {
	char TmpBuf[10]; sprintf(TmpBuf, "#%d", i);
	(*ReturnSchema)[i].Assign(TmpBuf, EList->Expr->BasicType(),
				  EList->Expr->GenerateCatInfo());
      }
      (*ReturnSchema)[i].AttrSource = -1;
    }
  }
  // have to do this before returning
  ReturnSchema->ComputeStats();
  return ReturnSchema;
}






//
// Methods of RelStopAfterOp
//
void RelStopAfterOp::Print(ostream& OStr) const 
{
 OStr << "RelStopAfterOp";
}

void RelStopAfterOp::NodePrint(ostream &OStr, int curBlockId, int Level, 
			     int numChildren, int *childrenIds) const
{
  OStr << "NODE Ordering" << endl;
  OStr << "ID " << curBlockId << endl;
  OStr << "LEVEL " << Level << endl;
  if (numChildren > 0) {
    OStr << "CHILDREN ";
    for (int i = 0; i < numChildren; i++) {
      OStr << childrenIds[i] << " ";
    }
    OStr << endl;
  }
  OStr << "COLOR " << 0x00CCCC << endl;
  // print K
  OStr << "INFO Stopping attributes: ";
  OStr << K;
  OStr << endl;
  OStr << "ENDNODE" << endl;
}

RelStopAfterOp::RelStopAfterOp (int k, XxxErrCode &RC)
: K(k)
{
  RC = XXX_OK;
  return;
}

RelStopAfterOp::~RelStopAfterOp()
{
}

RelImplInfo*
RelStopAfterOp::GenerateImplInfo(int NumChildren, RelQueryNode** Children)
{
  if (!NumChildren)
    return NULL;
  return new RelImplInfo(*(Children[0]->CatalogInfo->ImplInfo()), XXX_FALSE);
  
}

RelStatsInfo* 
RelStopAfterOp::GenerateStatsInfo(int NumChildren, RelQueryNode** Children) 
{
  if (NumChildren != 1) {
    XXX_NOTE_ERROR(XXX_INTERNAL, "Stop after should have one input");
    return NULL;
  }
  RelStatsInfo* StatsInfo;
  RelCatalogInfo* ChildCatInfo = Children[0]->CatalogInfo;
  StatsInfo = ChildCatInfo->StatsInfo()->Replicate(ChildCatInfo->RecordSchema());
  return StatsInfo;
}

XxxRecordSchema* 
RelStopAfterOp::GenerateRecordSchema(int NumChildren, RelQueryNode** Children,
				   const XxxArgEnv* Env) 
{
  if (NumChildren != 1) {
    REL_NOTE_ERROR(REL_BADQUERY, "Stop After should have only one input");
    return NULL;
  }
  
  XxxRecordSchema* ReturnSchema = 
    new XxxRecordSchema(*(Children[0]->CatalogInfo->RecordSchema()));
  
  ReturnSchema->ComputeStats();
  return ReturnSchema;
}

// ResolveReferences
XxxErrCode 
RelStopAfterOp::ResolveReferences(int NumChildren,
				  RelQueryNode** Children,
				  const XxxArgEnv* Env) 
{
  return XXX_OK;
}

XxxErrCode 
RelStopAfterOp::TypeCheck() 
{
  return XXX_OK;
}

XxxErrCode RelStopAfterOp::GetDependencies(XxxBitMap& Dependency,
					 int CorrelationHt) const
{
  return XXX_OK;
}



//end Ihab
