#ifndef REL_OPs_H
#define REL_OPs_H

#include <stdio.h>
#include <string.h>
#include "relconfig.h"
#include "relerror.h"
#include "expr.h"
#include "values.h"

XXX_EXTERN_CLASS_DECL(RelQueryNode);
XXX_EXTERN_CLASS_DECL(XxxQueryNameStruct);

XXX_EXTERN_CLASS_DECL(RelRecordSchema);
XXX_EXTERN_CLASS_DECL(RelStatsInfo);
XXX_EXTERN_CLASS_DECL(RelImplInfo);

typedef enum {
  REL_SPJ,
//Begin Ihab
  REL_KRJ,  
  REL_STOP,
//End Ihab
  REL_GBY,
  REL_OBY
} RelOpTypes;

typedef enum {
  QR_ENFORCE,
  QR_PRESERVE,
  QR_PERMIT
} RewriteDupCond;  // --> Anil 2/24/97: Added to facilitate Query Rewrite

/**
 *
 */
class RelOperator {
 public:
  // necessary for the Query Rewrite Engine.
  // indicates what this operator needs to do with duplicates.
  // this member is different from the Distinct Member of the descendants
  // in that they merely tell whether the output table has distinct values
  RewriteDupCond DistinctCond; // --> Anil 2/24/97: Added for Query Rewrite

  RelOperator(XxxBool Distinct);
  RelOperator();
  virtual ~RelOperator() {}
  virtual RelOpTypes Type() = 0;
  virtual void Print(ostream& OStr)  const = 0;
  virtual void NodePrint(ostream &OStr, int curBlockId, int Level, 
			 int numChildren, int *childrenIds) const = 0;
  virtual XxxErrCode ResolveReferences(int NumChildren,
				       RelQueryNode** Children,
				       const XxxArgEnv* Env = NULL) 
   { return XXX_OK; };

  virtual XxxErrCode TypeCheck() {return XXX_OK;};

  // functions used to propagate info up the graph
  virtual XxxRecordSchema* GenerateRecordSchema
    (int NumChildren, RelQueryNode** Children, const XxxArgEnv* Env = NULL)= 0;
  virtual RelStatsInfo* GenerateStatsInfo
    (int NumChildren, RelQueryNode** Children)=0;
  virtual RelImplInfo* GenerateImplInfo
    (int NumChildren, RelQueryNode** Children)=0;
  //// Used to get correlation dependencies
  virtual XxxErrCode GetDependencies(XxxBitMap& Dependendency,
				     int CorrelationHt) const = 0;
 };

/**
 *
 */
class RelSPJOp : public RelOperator {
 public:
  XxxBooleanExpression* Condition;
  XxxValueExprList* TargetList;
  XxxBool Distinct;

  RelSPJOp(XxxBooleanExpression* Cond, XxxValueExprList* ExprList, 
	   XxxBool Distinct, XxxErrCode &RC);
  virtual ~RelSPJOp();
  virtual RelOpTypes Type() { return REL_SPJ; }
  virtual void Print(ostream& OStr)  const ;
  virtual void NodePrint(ostream &OStr, int curBlockId, int Level, 
			 int numChildren, int *childrenIds) const;
  virtual XxxErrCode ResolveReferences(int NumChildren,
				       RelQueryNode** Children,
				       const XxxArgEnv* Env = NULL);
  virtual XxxErrCode TypeCheck();
  // functions used to propagate info up the graph
  virtual XxxRecordSchema* GenerateRecordSchema
    (int NumChildren, RelQueryNode** Children, const XxxArgEnv* Env = NULL);
  virtual RelStatsInfo* GenerateStatsInfo
    (int NumChildren, RelQueryNode** Children);
  virtual RelImplInfo* GenerateImplInfo
    (int NumChildren, RelQueryNode** Children);
  //// Used to get correlation dependencies
  virtual XxxErrCode GetDependencies(XxxBitMap& Dependendency,
				     int CorrelationHt) const;
 };


//Begin Ihab
class RelKRJOp : public RelSPJOp {
 public:
  XxxValueExprList * Ordering; 
  
  
  RelKRJOp(XxxBooleanExpression* Cond, XxxValueExprList* ExprList, 
	  XxxBool Distinct, XxxValueExprList * ordering, XxxErrCode &RC):
    RelSPJOp(Cond,ExprList,Distinct,RC){Ordering = ordering;}
  virtual ~RelKRJOp(){}
  virtual RelOpTypes Type() { return REL_KRJ; } 
  virtual void Print(ostream& OStr)  const ;
  virtual void NodePrint(ostream &OStr, int curBlockId, int Level, 
			 int numChildren, int *childrenIds) const;
  virtual XxxErrCode ResolveReferences(int NumChildren,  RelQueryNode** Children,
				       const XxxArgEnv* Env = NULL);
  virtual XxxErrCode TypeCheck(); 
  
  virtual XxxErrCode GetDependencies(XxxBitMap& Dependendency,int CorrelationHt) const; 
  virtual XxxRecordSchema* GenerateRecordSchema
    (int NumChildren, RelQueryNode** Children, const XxxArgEnv* Env = NULL);
};

class RelStopAfterOp : public RelOperator {
 public:
  int K;
  RelStopAfterOp(int k, XxxErrCode &RC);
  virtual ~RelStopAfterOp();
  virtual RelOpTypes Type() { return REL_STOP; }
  virtual void Print(ostream& OStr)  const ;
  virtual void NodePrint(ostream &OStr, int curBlockId, int Level, int numChildren, int *childrenIds) const;
  virtual XxxErrCode ResolveReferences(int NumChildren, RelQueryNode** Children, const XxxArgEnv* Env = NULL);
  virtual XxxErrCode TypeCheck();
  // functions used to propagate info up the graph
  virtual XxxRecordSchema* GenerateRecordSchema
    (int NumChildren, RelQueryNode** Children, const XxxArgEnv* Env = NULL);
  virtual RelStatsInfo* GenerateStatsInfo
    (int NumChildren, RelQueryNode** Children);
  virtual RelImplInfo* GenerateImplInfo
    (int NumChildren, RelQueryNode** Children);
  //// Used to get correlation dependencies
  virtual XxxErrCode GetDependencies(XxxBitMap& Dependendency,
				     int CorrelationHt) const;
};






//End Ihab


/**
 *
 */
class RelGrpByOp : public RelOperator {
 public:
  XxxValueExprList *TargetList;
  XxxValueExprList *GroupList;
  XxxValueExprList *AggrList;
  XxxBool AllAggs;

  RelGrpByOp(XxxValueExprList *TList,
	     XxxValueExprList *AList,
	     XxxValueExprList *GList,
	     XxxErrCode &RC);
  virtual ~RelGrpByOp();
  virtual RelOpTypes Type() { return REL_GBY; }
  virtual void Print(ostream& OStr)  const ;
  virtual void NodePrint(ostream &OStr, int curBlockId, int Level, 
			 int numChildren, int *childrenIds) const;
  virtual XxxErrCode ResolveReferences(int NumChildren,
				       RelQueryNode** Children,
				       const XxxArgEnv* Env = NULL);
  virtual XxxErrCode TypeCheck();
  // functions used to propagate info up the graph
  virtual XxxRecordSchema* GenerateRecordSchema
    (int NumChildren, RelQueryNode** Children, const XxxArgEnv* Env = NULL);
  virtual RelStatsInfo* GenerateStatsInfo
    (int NumChildren, RelQueryNode** Children);
  virtual RelImplInfo* GenerateImplInfo
    (int NumChildren, RelQueryNode** Children);
  //// Used to get correlation dependencies
  virtual XxxErrCode GetDependencies(XxxBitMap& Dependendency,
				     int CorrelationHt) const;
 };

/**
 *
 */
class RelOrderByOp : public RelOperator {
 public:
  XxxValueExprList *TargetList;
  // Ascending has an element for each ValueExpr in 
  // TargetList specifying its ordering
  // (the first element corresponds to the first in the Targetist)
  XxxBool *Ascending;
  
  RelOrderByOp(XxxValueExprList *TList, XxxBool *Asc, XxxErrCode &RC);
  virtual ~RelOrderByOp();
  virtual RelOpTypes Type() { return REL_OBY; }
  virtual void Print(ostream& OStr)  const ;
  virtual void NodePrint(ostream &OStr, int curBlockId, int Level, 
			 int numChildren, int *childrenIds) const;
  virtual XxxErrCode ResolveReferences(int NumChildren,
				       RelQueryNode** Children,
				       const XxxArgEnv* Env = NULL);
  virtual XxxErrCode TypeCheck();
  // functions used to propagate info up the graph
  virtual XxxRecordSchema* GenerateRecordSchema
    (int NumChildren, RelQueryNode** Children, const XxxArgEnv* Env = NULL);
  virtual RelStatsInfo* GenerateStatsInfo
    (int NumChildren, RelQueryNode** Children);
  virtual RelImplInfo* GenerateImplInfo
    (int NumChildren, RelQueryNode** Children);
  //// Used to get correlation dependencies
  virtual XxxErrCode GetDependencies(XxxBitMap& Dependendency,
				     int CorrelationHt) const;
};

#endif /* REL_OPERATOR_H */

