#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fstream.h>

#ifdef __GNUG__
#pragma implementation
#endif

#include "prd-config.h"
#include "relquery.h"
#include "reldata.h"
#include "releng.h"
#include "relenv.h"
#include "funcexpr.h"
#include "relrewrite.h"
#include "client.h"
#include "types.h"
#include "viewlist.h"
#include "shoresm.h"


//
// Methods of RelViewClass
//
RelViewClass::RelViewClass(const char* Name) : XxxViewClass(Name) 
{
  FromList = NULL;
  TargetList = NULL;
  WhereClause = NULL;
  GroupByList = NULL;
  HavingClause = NULL;
  OrderByList = NULL;
  StopValue = 0; //ihab: initialize stop value
  //Begin Moustafa: Window predicate
  WindowValue = NULL;
  //End Moustafa: Window predicate
}

RelViewClass::~RelViewClass() 
{
  if (FromList) {
    for (int i = 0; i < FromList->Arity; i++) {
      RelInfo *CurRel = (*(FromList))[i];
      delete CurRel;
    }
    delete FromList;
  }
  if (TargetList) {
    if (TargetList->TargetList)
      TargetList->TargetList->Destroy();
    delete TargetList;
  }
  if(WhereClause)
    delete WhereClause;
  if (GroupByList) {
    GroupByList->GroupList->Destroy();
    delete GroupByList;
  }
  if (OrderByList) {
    OrderByList->TargetList->Destroy();
    delete OrderByList;
  }
  if(HavingClause)
    delete HavingClause;
}

void RelViewClass::Print(ostream& OStr) const
{
  OStr << ViewName << " (";
  if (TargetList) { TargetList->Print(OStr);  OStr <<  "   "; }
  if (FromList) FromList->Print(OStr);
  OStr << "   ";
  if (WhereClause) {
    OStr << " WHERE ";
    WhereClause->Print(OStr);
    OStr << "   ";
  }
  if (GroupByList) { 
    OStr << " GROUP BY ";
    GroupByList->Print(OStr);  OStr << "   "; 
  }
  if (HavingClause) { 
    OStr << " HAVING ";
    HavingClause->Print(OStr);  OStr << "  "; 
  }
  if (OrderByList) { OrderByList->Print(OStr);  OStr << "  "; }
	
  OStr << " )" << endl;
}

void RelViewClass::PrintWithoutName(ostream& OStr) const
{
  if (TargetList) { TargetList->Print(OStr);  OStr <<  "   "; }
  if (FromList) FromList->Print(OStr);
  OStr << "   ";
  if (WhereClause) {
    OStr << " WHERE ";
    WhereClause->Print(OStr);
    OStr << "   ";
  }
  if (GroupByList) { 
    OStr << " GROUP BY ";
    GroupByList->Print(OStr);  OStr << "   "; 
  }
  if (HavingClause) { 
    OStr << " HAVING ";
    HavingClause->Print(OStr);  OStr << "   "; 
  }
  if (OrderByList) { OrderByList->Print(OStr);  OStr << "   "; }
  OStr << '\0';
	
}

void RelViewClass::AddFrom(FromInfo* FInfo)
{
  FromList = FInfo;
}
void RelViewClass::FromInfo::AddRelInfo(RelInfo* RInfo)
{
  int OldArity = Arity++;
  RelInfo** NewList = new RelInfo*[Arity];
  for (int i = 0; i < OldArity; i++)
    NewList[i] = RelList[i];
  NewList[OldArity] = RInfo;
  delete [] RelList;
  RelList = NewList;
}

void RelViewClass::AddSelection(XxxBooleanExpression *WherePred)
{
  WhereClause = WherePred;
}
void RelViewClass::AddProjection(ProjInfo* PInfo)
{
  TargetList = PInfo;
}
void RelViewClass::AddGroupBy(GroupByInfo *GrpByInfo)
{
  GroupByList = GrpByInfo;
}
void RelViewClass::AddHaving(XxxBooleanExpression* HavingPred)
{
  HavingClause = HavingPred;
}
void RelViewClass::AddOrderBy(OrderInfo *OrdByCols)
{
  OrderByList = OrdByCols;
}
//ihab: add stop after info
void RelViewClass::AddStopAfter(StopInfo *stopinfo)
{
  StopValue = stopinfo;
}

//Begin Moustafa: Window predicate
void RelViewClass::AddWindow(WindowInfo *windowinfo)
{
  WindowValue = windowinfo;
}
//End Moustafa: Window predicate
void SimpleRelInfo::Print(ostream& OStr) const
{
  OStr <<  RelName << " AS " << AliasName;
}
void NestedRelInfo::Print(ostream& OStr) const
{
  View->Print(OStr);
  OStr << " AS " << AliasName;
}
void RelViewClass::FromInfo::Print(ostream& OStr) const
{
  OStr << "FROM ";
  for (int i = 0; i < Arity; i++) {
    RelList[i]->Print(OStr);
    if (i < Arity-1)
      OStr << ", ";
  }
}
void RelViewClass::ProjInfo::Print(ostream& OStr) const
{
  OStr << "SELECT ";
  if (TargetList) TargetList->Print(OStr);
  else OStr << "* ";
}
void RelViewClass::OrderInfo::Print(ostream& OStr) const
{
  OStr <<  "ORDER BY ";
  XxxValueExprList *Tmp = TargetList;
  int Col = 0;
  while (Tmp) {
    Tmp->Expr->Print(OStr);
    if (Tmp->Expr->ExprName)
      OStr << " AS " << Tmp->Expr->ExprName;
    OStr << (Ascending[Col++] == XXX_TRUE ? " ASC" : " DESC");
    Tmp = Tmp->Next;
    if (Tmp) 
      OStr << ", ";
  }
}
void RelViewClass::GroupByInfo::Print(ostream& OStr) const
{
  XxxValueExprList *Tmp = GroupList;
  int Col = 0;
  while (Tmp) {
    Tmp->Expr->Print(OStr);
    if (Tmp->Expr->ExprName)
      OStr << " AS " << Tmp->Expr->ExprName;
    Tmp = Tmp->Next;
    if (Tmp) 
      OStr << ", ";
  }
}

//ihab: print stop after info
void RelViewClass::StopInfo::Print(ostream& OStr) const
{
  OStr << " STOP AFTER  " << K;
}

//Begin Moustafa
void RelViewClass::WindowInfo::Print(ostream& OStr) const
{
  OStr << " WINDOW  " << IntervalSeconds << " " << IntervalMicroSecs;
}
//End Moustafa

//
// Methods of RelQueryNode
//
RelQueryNode::RelQueryNode(RelOperator* Oper,
			   int ChildArity, RelQueryNode** Children,
			   int NumNames, char** NodeNames) 
  : XxxQueryNode(ChildArity, (XxxQueryNode**)Children, NumNames, NodeNames)
{
  Op = Oper;
  CatalogInfo = NULL;
}

RelQueryNode::~RelQueryNode() 
{
  if (Op) delete Op;
  if (CatalogInfo) { 
    CatalogInfo->Destroy();
    delete CatalogInfo; 
    CatalogInfo = NULL; 
  }
}

// OWNER: as  DATE: 02/24/97
// PURPOSE: adds a bunch of children to the end of this node's children list
// NOTES: added to facilitate query rewrite
XxxErrCode RelQueryNode::AddChildren(RelQueryNode **NodeList, int Quantity) {
  if (Quantity==0)
    return XXX_OK;
	
  XxxQueryNode **NewChildList=new XxxQueryNode*[arity+Quantity];
  for (int i=0;i<arity;i++) 
    NewChildList[i]=children[i];
  for (int j=0;j<Quantity;j++) {
    NewChildList[arity+j]=NodeList[j];
    NewChildList[arity+j]->SetChildNumber(arity+j);
    NewChildList[arity+j]->SetParent(this);
  }
  arity+=Quantity;
  if (children!=NULL)
    delete[] children;
  children=NewChildList;
  return XXX_OK;
}

void RelQueryNode::PrintThis(ostream& OStr) const 
{
  if (Op) {
    Op->Print(OStr);
  }
  else {
    OStr << nameInQuery->Name;
  }
  if (refCnt > 1) {
    OStr << ":{" << refCnt << "}";
  }
}

void RelQueryNode::NodePrint(ostream &OStr, int curBlockId, int Level, 
			     int numChildren, int *childrenIds) const
{
  // Assumption: this is a base table
  OStr << "NODE " << nameInQuery->Name << endl;
  OStr << "ID " << curBlockId << endl;
  OStr << "LEVEL " << Level << endl;
  OStr << "COLOR " << 0xFF0000 << endl;
  OStr << "INFO ";
  nameInQuery->Print(OStr);
  OStr << endl;
  // print the schema
  OStr << "INFO Schema: ";
  if (CatalogInfo)
    for (int i = 0; i < CatalogInfo->recSchema->Arity(); i++) {
      OStr << CatalogInfo->recSchema->Attrs[i].Name << ':';
      OStr << XxxADTArray[CatalogInfo->recSchema->Attrs[i].Type]->TypeName;
      if (i < CatalogInfo->recSchema->Arity() - 1)
        OStr << ", ";
    }
  OStr << endl;
  OStr << "ENDNODE" << endl;
}

// OWNER: map  DATE: 02/27/97
// PURPOSE: Print block info to file (usually for query tree GUI).
// NOTES: none
void RelQueryNode::TreePrint(ostream& OStr, int &curBlockId, 
			     int Level) const
{
  // print children first
  int *childrenIds = new int[arity];
  for (int i = 0; i < arity; i++) {
    children[i]->TreePrint(OStr, curBlockId, Level + 1);
    childrenIds[i] = curBlockId++;
  }
  // print this node
  if (Op)
    Op->NodePrint(OStr, curBlockId, Level, arity, childrenIds);
  else
    NodePrint(OStr, curBlockId, Level, arity, childrenIds);
  delete []childrenIds;
}

XxxErrCode RelQueryNode::PropagateMetaInfoUp(const XxxArgEnv* Env)
{
  // if this node has been reached already
  // this is important, because the query may be a dag!
  if (CatalogInfo) return XXX_OK;
	
  // if the relation is a host variable (say $1), get the desired catalog
  // info from the appropriate host environment variable
  if (Env && !Op && (nameInQuery->Name[0] == '$')) {
    int Index = atoi(nameInQuery->Name+1);
    if (Index < 0) {
      char buf[100];
      sprintf(buf, "Bad relation name: %s", nameInQuery->Name);
      REL_NOTE_ERROR(REL_BADQUERY, buf);
      return XXX_NOT_OK;
    }
    else {
      XxxValueExpression* HostExpr = (*Env)[Index-1];
      RelCatalogInfo* Tmp;
      if (HostExpr && (HostExpr->BasicType() == RelationTypeId) &&
	  (Tmp = (RelCatalogInfo *)(HostExpr->GenerateCatInfo())) ) {
	CatalogInfo = Tmp->Replicate1();
	return XXX_OK;
      }
      else {
	char buf[100];
	sprintf(buf, "Bad type for relation :%s", nameInQuery->Name);
	REL_NOTE_ERROR(REL_BADQUERY, buf);
	return XXX_NOT_OK;
      }
    }
  }
	
  // Put the current state on the stack for nested queries.
  // Note: Praveen:: nested queries are currently not dealt with
  // in a uniform manner. Nested views/table expressions from the
  // FROM clause appear as XxxQueryNode "Children". They are
  // type-checked before the current node. But nestd subqueries within
  // other expressions (eg. in the WHERE clause), are only
  // type-checked when the embedding expression is type-checked.
  // For both categories, we need to ensure that the state of the
  // current node is available on the stack to resolve correlation
  // values. So push it on the stack now, and remove it at the end
  // of this method.
  XxxArgEnv CurrentArgEnv(0, NULL);
  CurrentArgEnv.NPEnv.Init(Arity(), (XxxQueryNode**)children, Env);
  XXX_THISRELENV()->NestedState.PushStack(&CurrentArgEnv);
  XxxArgEnv* DummyArgEnv;
  
  // recursively call on children
  for (int i = 0; i < Arity(); i++) {
    if (children[i]->PropagateMetaInfoUp(Env) != XXX_OK) {
      XXX_THISRELENV()->NestedState.PopStack(DummyArgEnv);
      return XXX_NOT_OK;
    }
  }
	
  // at this stage, all the children have successfully computed their
  // meta-information
  CatalogInfo = new RelCatalogInfo(nameInQuery->Name);
	
  if (Op) { // derived relation
    RelQueryNode** Children = (RelQueryNode **)(children);
    CatalogInfo->recSchema = 
      Op->GenerateRecordSchema(Arity(), Children, Env);
    CatalogInfo->implInfo = Op->GenerateImplInfo(Arity(), Children);
    CatalogInfo->statsInfo =
      Op->GenerateStatsInfo(Arity(), Children);
  }
  else { // base relation
    XxxBool Found = XXX_FALSE;
    if ((((XxxRelationEngine*)(XxxDBEngineArray[RelationEngineId]))->DataEng->
	 GetCatInfo(nameInQuery->Name, *CatalogInfo, Found) != XXX_OK) ||
	(Found != XXX_TRUE)) {
      REL_NOTE_ERROR(REL_MSSNGCAT, nameInQuery->Name);
    }
  }
	
  // remove the state from the stack
  XXX_THISRELENV()->NestedState.PopStack(DummyArgEnv);

  if (CatalogInfo->Valid() != XXX_TRUE) {
    char buf[100];
    sprintf(buf, "Schema for %s not found", nameInQuery->Name);
    delete CatalogInfo;
    CatalogInfo = NULL;
    REL_NOTE_ERROR(REL_BADQUERY, buf);
    return XXX_NOT_OK;
  }
  return XXX_OK;
}

XxxErrCode RelQueryNode::PropagateMetaInfoDown()
{
  return XXX_OK;
}
void RelQueryClass::Print(ostream& OStr) const
{
  *(XXX_THISENV().ErrorStr) << "RelQueryClass::Print: Who called me!!!" << endl;
}

XxxGenericParse*
RelQueryClass::Replicate() const {
  REL_NOTE_ERROR(REL_BADSRC,"Attempted to replicate a RelQueryClass."
		 "\nNote: RelQueryNode *Root is NOT replicated!");
  return new RelQueryClass(*this);
}

//
// Methods of RelQueryEngine
//
RelQueryEngine::RelQueryEngine(XxxRelationEngine* SrcEng) 
{ 
  RelEng = SrcEng; 
}

XxxErrCode RelQueryEngine::StoreView(const char* ViewName, RelViewClass* View)
{
  strcpy(View->ViewName, ViewName);
  if (XXX_THISRELENV()->DebugCompile) {
    cerr << "Received view " << ViewName << endl;
    View->Print(cerr);
  }
	
  // Put in some perfunctory checks to check the view definition.
  // For now, the only check is whether the view is cyclically defined
  RelViewClass::FromInfo* FromList = View->FromList;
  for (int j = 0; j < FromList->Arity; j++) {
    RelInfo* TableInfo = FromList->RelList[j];
    if (TableInfo && !TableInfo->Nested &&
	!strcmp(((SimpleRelInfo*)TableInfo)->RelName, ViewName)) {
      REL_NOTE_ERROR (REL_BADQUERY, "Cyclic view definition");
      return XXX_NOT_OK;
    }
  } // for each table in the view
  // TODO: this still doesnt check for a larger cycle thru another view

  char vInfo[1000];
  ostrstream OStr(vInfo,1000);
  View->PrintWithoutName(OStr);
  
  RelDataEngine* DataEng = (RelDataEngine* )(((XxxRelationEngine *)XxxDBEngineArray[RelationEngineId])->DataEng);
  XXX_CALL_RET(DataEng->CreateNewView(ViewName, vInfo),
	       XXX_BADQUERY, "Cannot create view");
  return XXX_OK;
}

XxxErrCode RelQueryEngine::DropView(const char* ViewName)
{
  if (XXX_THISRELENV()->DebugCompile) {
    cerr << "Deleting view " << ViewName << endl;
  }
  RelDataEngine* DataEng = (RelDataEngine* )(((XxxRelationEngine *)XxxDBEngineArray[RelationEngineId])->DataEng);
  XXX_CALL_RET(DataEng->DeleteView(ViewName), 
	       XXX_BADQUERY, "Cannot drop view");
  return XXX_OK;
}

XxxErrCode RelQueryEngine::RetrieveView(const char* ViewName, 
					RelViewClass*& View,
					XxxBool& Found) const
{
  RelDataEngine* DataEng = (RelDataEngine* )(((XxxRelationEngine *)XxxDBEngineArray[RelationEngineId])->DataEng);
  XXX_CALL_RET(DataEng->RetrieveView(ViewName, View, Found), 
	       XXX_BADQUERY, "Cannot retrieve view");
  return XXX_OK;
}

XxxErrCode RelQueryEngine::PrintViews(ostream& OStr) const
{
  RelDataEngine* DataEng = (RelDataEngine* )(((XxxRelationEngine *)XxxDBEngineArray[RelationEngineId])->DataEng);
  XXX_CALL_RET(DataEng->PrintViews(OStr), 
	       XXX_BADQUERY, "Cannot retrieve view");
  return XXX_OK;
}

//
// this function takes a view (that possibly references other queries) and 
// produces a complete query dag with query operators, catalog information
// and query rewrites performed
//  Here are the steps of the function:
//      1) convert View into a operator dag Q
//      2) propagate meta-info up the graph
//      3) propagate meta-info down the graph
//      4) perform rewrites
//      5) return Q
//
XxxErrCode RelQueryEngine::CheckQuery(RelViewClass* View, 
				      RelQueryClass*& Query,
				      const XxxArgEnv* Env) const
{
  //      1) convert View into a operator dag Q
  RelQueryNode* RootNode;
   if (ViewToOpDag(View, RootNode) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, "Unable to expand view");
    return XXX_NOT_OK;
  }
	
  //      2) convert operator dag into type-checked query class
  if (CheckOpDag(RootNode, Env, Query) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, "Error type-checking query");
    return XXX_NOT_OK;
  }
	
  return XXX_OK;
}

XxxErrCode RelQueryEngine::CheckOpDag(RelQueryNode* RootNode,
				      const XxxArgEnv* Env,
				      RelQueryClass*& Query) const
{
  ASSERT((RootNode));
  XxxRelEnv *GEnv = (XxxRelEnv*)(XXX_THISENGENV(RelationEngineId));
	
  //      1) propagate meta-info up the graph 
  //   bottom-up traversal implemented using recursive calls
  if (RootNode->PropagateMetaInfoUp(Env) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, "Unable to propagate meta-info up");
    return XXX_NOT_OK;
  }
	
  //      2) propagate meta-info down the graph
  if (RootNode->PropagateMetaInfoDown() != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, "Unable to propagate meta-info down");
    return XXX_NOT_OK;
  }
	
  //      3) Print the block info to file before rewrite and draw 
  // tree if a file stream has been provided
  if (GEnv->QPInfoFile) {
    *(GEnv->QPInfoFile) << "SEQ Compilation" << endl;
    *(GEnv->QPInfoFile) << "TREE Tree before Rewrite" << endl;
    int blockId = 0;
    //ivan@3.14.99
    if (GEnv->RewriteQuery) {
      RootNode->TreePrint(*(GEnv->QPInfoFile), blockId, 0);
    *(GEnv->QPInfoFile) << "ENDTREE\nENDSEQ" << endl;
    } else {
      RootNode->TreePrint(*(GEnv->QPInfoFile), blockId, 0);
      *(GEnv->QPInfoFile) << "ENDTREE\nENDSEQ" << endl;
    }

  }
	
  //      4) perform rewrites
  if (GEnv->RewriteQuery) {
    RelRewriteEngine *Rewriter = new RelRewriteEngine;
    RelQueryNode *OutNode;
    if (Rewriter->Rewrite(RootNode, OutNode)==XXX_NOT_OK) {
      delete Rewriter;
      return XXX_NOT_OK;
    }
    // Print the block info to file after rewrite and draw tree if 
    // a file stream has been provided.
    if (GEnv->QPInfoFile) {
      *(GEnv->QPInfoFile) << "SEQ FinalTree\nTREE Final Tree (Input to Optimizer)" << endl;
      int blockId = 0;
      RootNode->TreePrint(*(GEnv->QPInfoFile), blockId, 0);
      *(GEnv->QPInfoFile) << "ENDTREE\nENDSEQ" << endl;
    }
    // all's well
    delete Rewriter;
    RootNode = OutNode;
  }
	
  //      5) return Q
  if (XXX_THISRELENV()->DebugCompile) {
    cerr << "Query : "; RootNode->Print(cerr); cerr << endl;
  }
  Query = new RelQueryClass(RootNode);
  return XXX_OK;
}

XxxErrCode RelQueryEngine::ViewToOpDag(RelViewClass *View,
				       RelQueryNode *&Root) const
{
  RelOpDagBuilder *Builder = new RelOpDagBuilder(this);
  if (Builder->BuildQueryDag(View, Root) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, "Cannot build query graph");
    return XXX_NOT_OK;
  }
  delete Builder;
  return XXX_OK;
}

RelQueryEngine::~RelQueryEngine()
{
}

//
// Methods of RelOpDagBuilder
//

RelOpDagBuilder::RelOpDagBuilder(const RelQueryEngine *Owner) :
  Engine(Owner),
  UserProjection(NULL),
  InitProjection(NULL),
  GroupByProjection(NULL),
  Groupings(NULL),
  Aggregates(NULL),
  FinalSPJNeeded(XXX_FALSE),
  FinalProjection(NULL),
  DistinctSelection(XXX_FALSE),
  Selection(NULL),
  SelectionExprsRef(NULL),
  SourceRelations(NULL),
  Arity(0),
  GroupSelection(NULL),
  GrpSelExprsRef(NULL),
  Ordering(NULL),
  Ascending(NULL), 
 //Ihab begin
  StopValue(0),
  KRJOp (XXX_FALSE)
  //Ihab
{ }

RelOpDagBuilder::~RelOpDagBuilder()
{ 
  if (UserProjection)
    UserProjection->Destroy();
  if (SelectionExprsRef)
    SelectionExprsRef->DestroyListOnly();
  if (GrpSelExprsRef)
    GrpSelExprsRef->DestroyListOnly();
}

XxxErrCode 
RelOpDagBuilder::BuildQueryDag(RelViewClass *View,
			       RelQueryNode *&Root)
{
  // Extract the view information
  if (GetViewInfo(View) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADEXEC, 
		   "Cannot get necessary information from query");
    return XXX_NOT_OK;
  }
  // Reorganize it
  if (ReorganizeInfo() != XXX_OK) {
    REL_NOTE_ERROR(REL_BADEXEC, "Cannot reorganize query information");
    return XXX_NOT_OK;
  }
  // Perform all semantic checking
  if (CheckSemantics() != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, "Query is semantically incorrect");
    return XXX_NOT_OK;
  }
  // Add the implied projections
  if (AddImpliedProjections() != XXX_OK) {
    REL_NOTE_ERROR(REL_BADEXEC, "Cannot add additional projections");
    return XXX_NOT_OK;
  }
  if (XXX_THISRELENV()->DebugCompile) {
    // Print it out
    PrintQueryDag(cerr);
  }
  // Build the query DAG from the bottom up
  RelQueryNode **FromNodes = new RelQueryNode*[Arity];
  if (BuildFromNodes(FromNodes) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, "Cannot build one or more FROM nodes");
    return XXX_NOT_OK;
  }

   //begin Ihab
  if(KRJOp){
    if (AddInitialKRJNode(Root, FromNodes) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADQUERY, "Cannot build initial KEY RJ node");
      return XXX_NOT_OK;
    }
  }
  else{ // no order by in key RJ passed already to the operator
     //end Ihab
     if (AddInitialSPJNode(Root, FromNodes) != XXX_OK) {
	REL_NOTE_ERROR(REL_BADQUERY, "Cannot build initial SPJ node");
	return XXX_NOT_OK;
     }
     
     if (Ordering) {
	if (AddOrderByNode(Root) != XXX_OK) {
	   REL_NOTE_ERROR(REL_BADQUERY, "Cannot build order by node");
	   return XXX_NOT_OK;
	}    
     }
     
  }
  
  
  delete [] FromNodes;
  if (Aggregates) {
     if (AddGroupByNode(Root) != XXX_OK) {
	REL_NOTE_ERROR(REL_BADQUERY, "Cannot build group by node");
	return XXX_NOT_OK;
     }    
  }
  if (FinalSPJNeeded == XXX_TRUE) {
    if (AddFinalSPJNode(Root) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADQUERY, "Cannot build final SPJ node");
      return XXX_NOT_OK;
    }
  }
  //ihab: add stop after node
  if (AddStopAfterNode(Root) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, "Cannot build StopAfter node");
    return XXX_NOT_OK;
  }
  
/*  if (AddInitialSPJNode(Root, FromNodes) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADQUERY, "Cannot build initial SPJ node");
    return XXX_NOT_OK;
  }
  delete [] FromNodes;
  if (Aggregates) {
    if (AddGroupByNode(Root) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADQUERY, "Cannot build group by node");
      return XXX_NOT_OK;
    }    
  }
  if (Ordering) {
    if (AddOrderByNode(Root) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADQUERY, "Cannot build order by node");
      return XXX_NOT_OK;
    }    
  }
  if (FinalSPJNeeded == XXX_TRUE) {
    if (AddFinalSPJNode(Root) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADQUERY, "Cannot build final SPJ node");
      return XXX_NOT_OK;
    }
  }*/
  return XXX_OK;
}

XxxErrCode 
RelOpDagBuilder::GetViewInfo(const RelViewClass *View)
{
  /* Here's the info we need to acquire (and the order in which we do it):
   * 
   * SourceRelations:  The source relations/views (FROM clause)
   * Arity:            The number of inputs (source relations)
   * DistinctSelection:Is the selection distinct?
   * Selection:        The selection condition (WHERE clause)
   * InitProjection:   All projected expressions.
   * Ordering:         The ordering expressions
   * Ascending:        A parallel array specifying the ordering of those exprs.
   * Groupings:        All group-by expressions
   * GroupSelection:   The HAVING condition
   *
   * All information retrieved from the view is replicated to simplify
   * memory management.  Some initial manipulation is performed on the 
   * supplied projection list, such as removing replacing *.* expressions
   * and resolving Alias.* expressions early.  This is solely to simplify
   * later coding.
	*/
	
  if (!View) {
    REL_NOTE_ERROR(REL_BADEXEC, "NULL view passed in");
    return XXX_NOT_OK;
  }
  // Get the source relations
  if (View->FromList) {
    SourceRelations = View->FromList->RelList;
    Arity = View->FromList->Arity;
	 //begin Ihab
    if(View->StopValue)
      StopValue = View->StopValue->K;
    else 
      StopValue = 0;
    KRJOp = View->FromList->KRJOp;
    //End Ihab
  } else {
    REL_NOTE_ERROR(REL_BADEXEC, "No source relation information provided");
    return XXX_NOT_OK;
  }
  // Get the distinct info
  DistinctSelection = View->TargetList->Distinct;
  //Begin Moustafa: Window predicate
  
  if(View->WindowValue)
    WindowValue = View->WindowValue;
  else
    WindowValue = NULL;    // No window attached


  //End Moustafa: Window predicate

  // Get the selection
  if (View->WhereClause)
    Selection = (XxxBooleanExpression *)View->WhereClause->Replicate();
  // Get the initial projection list
  if (!((View->TargetList) && (View->TargetList->TargetList))) {
    REL_NOTE_ERROR(REL_BADEXEC, "No projection provided");
    return XXX_NOT_OK;
  }
  /*
	* If a complete-wildcard expression ("*.*") appears, it must be replaced
	* with a list of "Alias.*" expressions to simplify post-projection.
	*/
  XxxValueExprList *Tmp = View->TargetList->TargetList;
  XxxUnknownValue Star("*", "*");
  while (Tmp) {
    if (Tmp->Expr->Match(&Star)) {
      // Remove the expression
      if (Tmp == View->TargetList->TargetList) {
	View->TargetList->TargetList = Tmp->Next;
      } else {
	XxxValueExprList *Parent;
	for (Parent = View->TargetList->TargetList; 
	     Parent->Next != Tmp;
	     Parent = Parent->Next);
	Parent->Next = Tmp->Next;
      }
      Tmp->Next = NULL;
      Tmp->Destroy();
      // Now add a "Alias.*" expression for each relational input
      for (int i = 0; i < Arity; i++) {
	XxxValueExpression *Expr = (XxxValueExpression *)
	  (new XxxUnknownValue("*", SourceRelations[i]->AliasName));
	if (View->TargetList->TargetList)
	  View->TargetList->TargetList->AddExpr(Expr);
	else
	  View->TargetList->TargetList = new XxxValueExprList(Expr);
      }
      // Now start over
      Tmp = View->TargetList->TargetList;
      continue;
    }
    Tmp = Tmp->Next;
  }
  // Do a complete replication
  UserProjection = View->TargetList->TargetList->Replicate();
  // Get the ordering information
  if (View->OrderByList) {
    if (View->OrderByList->TargetList)
      Ordering = View->OrderByList->TargetList->Replicate();
    Ascending = View->OrderByList->Ascending;
  }
  // Get the grouping information
  if (View->GroupByList) {
    Groupings = View->GroupByList->GroupList->Replicate();
  }
  // Get the having clause information
  if (View->HavingClause) {
    GroupSelection = (XxxBooleanExpression *)View->HavingClause->Replicate();
  }
	
  return XXX_OK;
}

XxxErrCode 
RelOpDagBuilder::ReorganizeInfo() {
  /*
   * Get a list of all of the value expressions referenced in the
   * selection condition and group selection condition.
	*/
  if (Selection) {
    SelectionExprsRef = Selection->ValExprsReferenced();
  }
  if (GroupSelection) {
    GrpSelExprsRef = GroupSelection->ValExprsReferenced();
  }
  /*
	* Gather all expressions which reference an aggregate.  In a correct
	* query, these can exist only in the projection, group selection,
	* and ordering clauses.
	*/
  XxxValueExprList *Lists[3] = { UserProjection, 
				 GrpSelExprsRef,
				 Ordering };
  XxxValueExprList *Tmp = NULL;
  int i;
  for (i = 0; i < 3; i++) {
    Tmp = Lists[i];
    while (Tmp) {
      if (Tmp->Expr->ReferencesAggregate()) {
	if (Aggregates) {
	  if (!Aggregates->Contains(Tmp->Expr))
	    Aggregates->AddExpr((XxxValueExpression *)Tmp->Expr->Replicate());
	} else
	  Aggregates = 
	    new XxxValueExprList((XxxValueExpression *)
				 Tmp->Expr->Replicate());
      }
      Tmp = Tmp->Next;
    }
  }
  return XXX_OK;
}

XxxErrCode 
RelOpDagBuilder::CheckSemantics()
{
  /*
   * For now we explicitly disallow the use of aggregate expressions in
   * the HAVING clause.
   */
  if (GroupSelection && GrpSelExprsRef->ReferencesAggregate()) {
    XXX_NOTE_ERROR(XXX_INCOMPL, "HAVING referencing aggregates is currently not supported");
    return XXX_NOT_OK;
  }
  /*
   * There must be at least one relational input to the query.
   */
  if (Arity < 1) {
    REL_NOTE_ERROR(REL_BADQUERY, "No source relations/views provided.");
    return XXX_NOT_OK;
  }
  /*
   * All of the aliases referenced in the query must be distinct to insure
   * the expected results.
   */
  int i, j;
  for (i = 0; i < Arity; i++)
    for (j = i + 1; j < Arity; j++)
      if (!strcmp(SourceRelations[i]->AliasName, 
		  SourceRelations[j]->AliasName)) {
	char buf[200];
	sprintf(buf, "The alias \"%s\" is defined twice in the query", 
		SourceRelations[i]->AliasName);
	REL_NOTE_ERROR(REL_BADQUERY, buf);
	return XXX_NOT_OK;
      }
  /*
   * A group selection (HAVING clause) cannot exist unless a grouping
   * clause exists.
   */
  if (!Groupings && GroupSelection) {
    REL_NOTE_ERROR(REL_BADQUERY, "HAVING supplied with no GROUP BY clause");
    return XXX_NOT_OK;
  }
  /*
   * Two expressions, if both are named, cannot share the same name.
   */
  XxxValueExprList *Tmp = UserProjection, *Tmp2;
  while (Tmp) {
    Tmp2 = Tmp->Next;
    while (Tmp2) {
      if (Tmp->Expr->ExprName && Tmp2->Expr->ExprName &&
	  !strcmp(Tmp->Expr->ExprName, Tmp2->Expr->ExprName)) {
	ostrstream ss;
	Tmp->Expr->Print(ss);
	ss << " and ";
	Tmp2->Expr->Print(ss);
	ss << " cannot both be named ";
	ss << Tmp->Expr->ExprName << '\0';
	REL_NOTE_ERROR(REL_BADQUERY, ss.str());
	return XXX_NOT_OK;
      } 
      Tmp2 = Tmp2->Next;
    }
    Tmp = Tmp->Next;
  }
  /*
   * The selection (WHERE) clause cannot reference aggregates.
   */
  if (SelectionExprsRef && SelectionExprsRef->ReferencesAggregate()) {
    REL_NOTE_ERROR(REL_BADQUERY, 
		   "The WHERE clause cannot reference an aggregate.");
    return XXX_NOT_OK;
  }
  /*
   * If a grouping exists, aggregates must be referenced.
   */
  if (Groupings && !Aggregates) {
    REL_NOTE_ERROR(REL_BADQUERY, 
		   "Cannot perform GROUP BY if no aggregates are supplied");
    return XXX_NOT_OK;
  }
  /*
   * If aggregates exist, all projected items must be aggregates,
   * grouping, or constant expressions. This is a limitation ---
   * actually, the expressions must all be one of constants,
   * aggregates, or expressions that use only the grouping expressions.
   */
  if (Aggregates) {
    Tmp = UserProjection;
    while (Tmp) {
      if (!(Aggregates->Contains(Tmp->Expr) ||
	    (Groupings && Groupings->Contains(Tmp->Expr)) ||
            (Tmp->Expr->ValType() == XxxValueExpression::CONST_EXPR))) {
	ostrstream ss;
	ss << "Cannot SELECT ";
	Tmp->Expr->Print(ss);
	ss << " because it is not an aggregate or a grouping term" << '\0';
	REL_NOTE_ERROR(REL_BADQUERY, ss.str());
	return XXX_NOT_OK;
      }
      Tmp = Tmp->Next;
    }
  }
  /*
   * The same expression cannot appear twice in the grouping list or
   * in the ordering list.
   */
  Tmp = Ordering;
  while (Tmp && Tmp->Next) {
    if (Tmp->Next->Contains(Tmp->Expr)) {
      ostrstream ss;
      ss << "Cannot ORDER BY ";
      Tmp->Expr->Print(ss);
      ss << " twice" << '\0';
      REL_NOTE_ERROR(REL_BADQUERY, ss.str());
      return XXX_NOT_OK;
    }
    Tmp = Tmp->Next;
  }
  Tmp = Groupings;
  while (Tmp && Tmp->Next) {
    if (Tmp->Next->Contains(Tmp->Expr)) {
      ostrstream ss;
      ss << "Cannot GROUP BY ";
      Tmp->Expr->Print(ss);
      ss << " twice" << '\0';
      REL_NOTE_ERROR(REL_BADQUERY, ss.str());
      return XXX_NOT_OK;
    }
    Tmp = Tmp->Next;
  }
  /*
   * If there are aggregates present, the expressions in the ordering 
   * list must be aggregates, grouping, or constant terms.
   */
  /****
  if (Ordering && Aggregates) {
    Tmp = Ordering;
    while (Tmp) {
      if (!(Aggregates->Contains(Tmp->Expr) ||
	    (Groupings && Groupings->Contains(Tmp->Expr)) ||
            (Tmp->Expr->ValType() == XxxValueExpression::CONST_EXPR))) {
	ostrstream ss;
	ss << "Cannot ORDER BY ";
	Tmp->Expr->Print(ss);
	ss << " because it is not an attribute of a group" << '\0';
	REL_NOTE_ERROR(REL_BADQUERY, ss.str());
	return XXX_NOT_OK;
      }
      Tmp = Tmp->Next;
    }
  }
  ****/
  //Begin Moustafa: Window predicte
  // window currently is of syntax 
  // WINDOW YYYYMMDD.HHMMSSZZZZZZ
  if(WindowValue!=NULL)
    if (!WindowValue->IsValid()) {
      REL_NOTE_ERROR(REL_BADQUERY, 
		     " Invalid window specification: Syntax WINDOW YYYYMMDD.HHMMSSZZZZZZ");
      return XXX_NOT_OK;
    }
  
  //End Moustafa:

  return XXX_OK;
}

/**
 * Does Aggregates contain expression named the same as Expr?
 */
const char RelOpDagBuilder::AggContains(XxxValueExpression *Expr)
{
  XxxValueExprList *Tmp = Aggregates;
  while (Tmp) {
    if (Tmp->Expr->ExprName &&
		(strcmp(Tmp->Expr->ExprName, ((XxxUnknownValue *)Expr)->Name) == 0))
      return XXX_TRUE;
    Tmp = Tmp->Next;
  }
  return XXX_FALSE;
}

XxxErrCode RelOpDagBuilder::AddImpliedProjections() 
{
  /*
   * We start out by creating a preliminary initial projection list;
   * this list will consist of all items the query projects, minus
   * aggregates: they cannot be computed by the initial SPJ node.
	*/
  XxxValueExprList *Tmp = UserProjection;
  while (Tmp) {
    if (!(Tmp->Expr->ReferencesAggregate())) {
      if (InitProjection)
	InitProjection->AddExpr((XxxValueExpression *)
				(Tmp->Expr->Replicate()));
      else
	InitProjection = new XxxValueExprList((XxxValueExpression *)
					      (Tmp->Expr->Replicate()));
    }
    Tmp = Tmp->Next;
  }
  /*
	* Here we must create a projection list for the group by node.
	* It should be exactly what the user requested, plus any expressions
	* referenced in the group selection (HAVING clause), since those
	* expressions will later be selected upon by the final SPJ node,
	* and any expressions referenced by the Ordering list.
	*/
  if (Aggregates) {
    GroupByProjection = UserProjection->Replicate();
    XxxValueExprList *Tmp = GrpSelExprsRef;
    while (Tmp) {
      if ((Tmp->Expr->ValType() != XxxValueExpression::CONST_EXPR) &&
	  !GroupByProjection->Contains(Tmp->Expr))
	GroupByProjection->AddExpr((XxxValueExpression *)
				   (Tmp->Expr->Replicate()));
      Tmp = Tmp->Next;
    }
    Tmp = Ordering;
    while (Tmp) {
      if (!GroupByProjection->Contains(Tmp->Expr)
           && !AggContains(Tmp->Expr))
	GroupByProjection->AddExpr((XxxValueExpression *)
				   (Tmp->Expr->Replicate()));
      Tmp = Tmp->Next;
    }
  }
  /*
	* The projection lists must be reorganized in the following way:
	* attributes which are legally referenced in the aggregates, grouping terms,
	* ordering terms, and group selection need not be explicitly projected.  
	* Those attributes that are not explicitly projected must be added 
	* to the initial projection list.
	*/
  int TempVarNum = 0;
  XxxValueExprList *Lists[4] = { Ordering, Groupings,
				 Aggregates, GrpSelExprsRef };
  XxxValueExprList *ListCursor = NULL;
  XxxValueExprList *UnknownValsRef = NULL;
  int i;
  for (i = 0; i < 4; i++) { // for each list
    ListCursor = Lists[i];
    while (ListCursor) {
      /*
       * Each ListCursor->Expr is an expression which can reference many
       * column expressions.  We must get a list of all such references
       * and insure they exist in the initial projection, adding them if
       * they do not.
       */
      Tmp = UnknownValsRef = ListCursor->Expr->ValExprsReferenced();
      if (UnknownValsRef) 
	Tmp = UnknownValsRef = 
	  UnknownValsRef->TypeFilter(XxxValueExpression::COLUMN_VAR);
      while (Tmp) {
	/*
	 * If the expression references a column and the projection list
	 * includes it, ignore it.
	 */
        XxxUnknownValue Star("*", ((XxxUnknownValue *)Tmp->Expr)->XxxName);
        int cond = (i == 0 || i == 3) ?  AggContains(Tmp->Expr) : 0;

        if (!(InitProjection &&
              (InitProjection->Contains(Tmp->Expr) ||
              InitProjection->Contains(&Star))) && !cond) {
	  XxxUnknownValue *Expr = 
	    new XxxUnknownValue(*(XxxUnknownValue *)(Tmp->Expr));
	  // Add the expression to the projection list
	  if (InitProjection)
	    InitProjection->AddExpr(Expr);
	  else
	    InitProjection = new XxxValueExprList(Expr);
	  /** Praveen: not necessary -- in fact, this 
	    can cause errors if two tables both have fields
	    of the same name!
	    // Now replace all references to the expression in the list with
	    // column references.
	    XxxUnknownValue ColRef(Expr->Name);
	    Lists[i]->Replace(Tmp->Expr, &ColRef);
	    ****/
	}
	Tmp = Tmp->Next;
      } // while (Tmp)
      if (UnknownValsRef)
	UnknownValsRef->Destroy();
      ListCursor = ListCursor->Next;
    } // while (ListCursor)
  } // for each list
  /*
	  * In addition, the ordering and grouping lists can make reference to
	  * complex expressions that they are unable to compute themselves.  In this
	  * case, they must be preprojected in.
	*/
  for (i = 0; i < 2; i++) { // for each list
    ListCursor = Lists[i];
    while (ListCursor) {
      /*
       * Ignore column expressions; they were dealt with above.
       */
      if (ListCursor->Expr->ValType() == XxxValueExpression::COLUMN_VAR) {
	ListCursor = ListCursor->Next;
	continue;
      }
      /*
       * For the Ordering List only:
       * If the expression references an aggregate, it is being computed
       * by the group-by node.  In this case, name the expression in the
       * aggregate list and change this reference to point to that expression.
       */
      if (ListCursor->Expr->ReferencesAggregate()) {
	char buf[XXX_ATTR_NAME_MAX];
	ostrstream ss(buf, XXX_ATTR_NAME_MAX, ios::out);
	ss << TempVarNum++ << ":";
	ListCursor->Expr->Print(ss);
	ss << '\0' << flush;
	buf[XXX_ATTR_NAME_MAX - 1] = '\0';
	ListCursor->Expr->ExprName = new char[strlen(buf) + 1];
	strcpy(ListCursor->Expr->ExprName, buf);
	GroupByProjection->Replace(ListCursor->Expr, ListCursor->Expr);
	delete ListCursor->Expr;
	ListCursor->Expr = new XxxUnknownValue(buf);
	ListCursor = ListCursor->Next;
	continue;
      }
      /*
       * Now finally check to see if the item needs to be pre-projected.
       */
      if (!InitProjection ||
	  !InitProjection->Contains(ListCursor->Expr)) {
	// Add the expression to the projection list
	if (InitProjection)
	  InitProjection->AddExpr((XxxValueExpression *)
				  ListCursor->Expr->Replicate());
	else
	  InitProjection =
	    new XxxValueExprList((XxxValueExpression *)
				 ListCursor->Expr->Replicate());
      }
      // Name all instances of the expression in the initial projection
      char buf[XXX_ATTR_NAME_MAX];
      ostrstream ss(buf, XXX_ATTR_NAME_MAX, ios::out);
      ss << TempVarNum++ << ":";
      ListCursor->Expr->Print(ss);
      ss << '\0' << flush;
      buf[XXX_ATTR_NAME_MAX - 1] = '\0';
      ListCursor->Expr->ExprName = new char[strlen(buf) + 1];
      strcpy(ListCursor->Expr->ExprName, buf);
      if (InitProjection)
	InitProjection->Replace(ListCursor->Expr, ListCursor->Expr);
      // Replace the expression in the list with a column expression
      delete ListCursor->Expr;
      ListCursor->Expr = new XxxUnknownValue(buf);
      ListCursor = ListCursor->Next;
    }
  }
  /*
	* If the user did not specify a wildcard expression ("*"), and if
	* any additional projections have been named or a having clause
	* exists, create a final projection which eliminates all additional
	* projections.  We can take the first <n> elements of the incoming
	* projection (where n = the number of projections requested), since
	* we only added to the end of the requested projection.
	*/
  XxxUnknownValue Star("*");
  if (!(UserProjection->Contains(&Star)) && 
      ((GroupSelection 
	|| 
	(!Aggregates && !(InitProjection->IsSubsetOf(UserProjection) &&
			  UserProjection->IsSubsetOf(InitProjection))) 
	||
	(Aggregates && !(GroupByProjection->IsSubsetOf(UserProjection) &&
			 UserProjection->IsSubsetOf(GroupByProjection)))))) {
    FinalSPJNeeded = XXX_TRUE;
    int i;
    for (i = 0; i < UserProjection->Count(); i++) {
      char AttrName[XXX_ATTR_NAME_MAX];
      ostrstream ss(AttrName, XXX_ATTR_NAME_MAX, ios::out);
      (*UserProjection)[i]->Print(ss);
      ss << '\0';
      // Create the expression as a simple reference
      XxxValueExpression *Expr = 
	new XxxUnknownValue(AttrName, NULL, XXX_UNKNOWN, i, 0);
      if (FinalProjection)
	FinalProjection->AddExpr(Expr);
      else
	FinalProjection = new XxxValueExprList(Expr);
    }
  }
  return XXX_OK;
}

void RelOpDagBuilder::PrintQueryDag(ostream &OStr) const {
  OStr << "--------------------------------------------------------" << endl;
  OStr << "Query DAG construction: " << endl;
  OStr << " Initial SPJ node:" << endl;
  if (DistinctSelection)
    OStr << "  Distinct " << endl;
  if (Selection) {
    OStr << "  Initial selection: ";
    Selection->Print(OStr);
    OStr << endl;
  }
  if (InitProjection) {
    OStr << "  Initial projection: ";
    InitProjection->Print(OStr);
    OStr << endl;
  }
  if (Aggregates) {
    OStr << " GroupBy node:" << endl;
    OStr << "  Projection: ";
    if (GroupByProjection)
      GroupByProjection->Print(OStr);
    else
      OStr << "None";
    OStr << endl;
    OStr << "  Aggregates referenced: ";
    if (Aggregates)
      Aggregates->Print(OStr);
    else
      OStr << "None";
    OStr << endl;
    OStr << "  Grouping terms: ";
    if (Groupings)
      Groupings->Print(OStr);
    else
      OStr << "none";
    OStr << endl;
  }
  if (Ordering) {
    OStr << " OrderBy Node:" << endl;
    OStr << "  Ordering terms: ";
    if (Ordering)
      Ordering->Print(OStr);
    else
      OStr << "None";
    OStr << endl;
  }
  if (FinalSPJNeeded) {
    OStr << " Final SPJ Node:" << endl;
    if (FinalProjection) {
      OStr << "  Final projection: ";
      FinalProjection->Print(OStr);
      OStr << endl;
    }
    if (GroupSelection) {
      OStr << "  Final selection: ";
      GroupSelection->Print(OStr);
      OStr << endl;
    }
  }
  OStr << "--------------------------------------------------------" << endl;
}

XxxErrCode 
RelOpDagBuilder::BuildFromNodes(RelQueryNode **&FromNodes) 
{
  /* Make a base node for each SimpleRelInfo in the From list
   * and an SPJ node above them.  For the base nodes, make common 
   * subexpressions out of duplicates and expand out views if defined. 
   * Also, for offset nodes, create offset operators.
   */
  for (int i = 0; i < Arity; i++) {
    RelInfo *CurRel = SourceRelations[i];
    if (CurRel->Nested) {
      /*
       * If the input is a nested query, create a query DAG for that query.
       */
      RelViewClass* NestedView = ((NestedRelInfo *)(CurRel))->View;
      if (Engine->ViewToOpDag(NestedView, FromNodes[i]) != XXX_OK) {
	REL_NOTE_ERROR(REL_BADQUERY, "Cannot form query dag for from node");
	return XXX_NOT_OK;
      }
      FromNodes[i]->AddName(CurRel->AliasName);
    } else {
      /*
       * There are three possibilities:
       * 1) the relation name is the same as that of a previous relation
       *    in the from list (not currently handled--see below)
       * 2) the relation name is the name of a registered view
       * 3) the relation name is that of a base relation
       */
      FromNodes[i] = NULL;
      char* SiName = ((SimpleRelInfo *)(CurRel))->RelName;
      // DO NOT CREATE CSEs FOR RELATIONS !! 09/95 - Praveen
      if (0) {
	// We can do beter than this !! : TODO:L3
	// consider if the same relation is shared across view levels
	// code this by also maintaining a list of nodes seen so far
	for (int j = 0; j < i; j++) {
	  if (!SourceRelations[j]->Nested) {
	    char* SjName =
	      ((SimpleRelInfo *)(SourceRelations[j]))->RelName;
	    if (!strcmp(SiName, SjName)) {
	      // case 1
	      FromNodes[i] = FromNodes[j];
	      FromNodes[j]->IncrRef();
	      FromNodes[i]->AddName(CurRel->AliasName);
	      break;
	    }
	  }
	} // for loop
      } // if (0)
			
      if (!FromNodes[i]) { // not case 1
	RelViewClass* ThisView;
	XxxBool Found = XXX_FALSE;
	REL_CALL_FN(Engine->RetrieveView(SiName, ThisView, Found),
		    REL_BADQUERY, "Error checking view");
	if (Found) {
	  /*
	   * The input is a reference to an existing view.
	   */
	  if (Engine->ViewToOpDag(ThisView, FromNodes[i]) != XXX_OK) {
	    REL_NOTE_ERROR(REL_BADQUERY, 
			   "Cannot form query dag for from node");
	    return XXX_NOT_OK;
	  }
	  FromNodes[i]->AddName(CurRel->AliasName);
	} else {
	  /*
	   * This is a base relation reference.  Note that the ordering
	   * of node names is important.  The relation name must be the
	   * first name.
	   */
	  char *NodeNames[2];
	  NodeNames[0] = ((SimpleRelInfo *)(CurRel))->RelName;
	  NodeNames[1] = CurRel->AliasName;
	  FromNodes[i] = new RelQueryNode(NULL,0,NULL,2,NodeNames);
	}
      } // if ! case 1
    } // if not nested view
  }// for
  return XXX_OK;
}

XxxErrCode 
RelOpDagBuilder::AddInitialSPJNode(RelQueryNode *&SPJNode,
				   RelQueryNode **FromNodes)
{
  /*
   * In a very special case, we do not need an initial SPJ; in this
   * case, eliminate it.
	*/
  XxxUnknownValue TableStar("*", SourceRelations[0]->AliasName);
  if (!DistinctSelection && !Selection && (Arity == 1) &&
      InitProjection &&
      (InitProjection->Count() == 1) && 
      (InitProjection->Expr->Match(&TableStar))) {
    SPJNode = FromNodes[0];
    return XXX_OK;
  }
    
  XxxErrCode Status = XXX_OK;
  // there must be an InitProjection: check here
  ASSERT ((InitProjection));
  RelSPJOp *SPJOp = new RelSPJOp(Selection, InitProjection, 
				 DistinctSelection, Status);
  if (Status != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOP, "Cannot instantiate SPJ operator");
    return XXX_NOT_OK;
  }
  SPJNode = new RelQueryNode(SPJOp, Arity, FromNodes);
  // set parent pointers for all the children
  int i;
  for (i = 0; i < Arity; i++) {
    FromNodes[i]->SetParent(SPJNode);
  }
  return XXX_OK;
}

//begin Ihab
XxxErrCode 
RelOpDagBuilder::AddInitialKRJNode(RelQueryNode *&KRJNode,RelQueryNode **FromNodes)
{
  /*
   * In a very special case, we do not need an initial KRJ; in this
   * case, eliminate it.
	*/
  XxxUnknownValue TableStar("*", SourceRelations[0]->AliasName);
  if (!DistinctSelection && !Selection && (Arity == 1) &&
      InitProjection &&
      (InitProjection->Count() == 1) && 
      (InitProjection->Expr->Match(&TableStar))) {
    KRJNode = FromNodes[0];
    return XXX_OK;
  }
    
  XxxErrCode Status = XXX_OK;
  // there must be an InitProjection: check here
  ASSERT ((InitProjection));
  RelKRJOp *KRJOp = new RelKRJOp(Selection, InitProjection, DistinctSelection, Ordering, Status);
  if (Status != XXX_OK || XXX_IS_ERROR()) {//ihab
    REL_NOTE_ERROR(REL_BADOP, "Cannot instantiate Key RJ operator");
    return XXX_NOT_OK;
  }
  KRJNode = new RelQueryNode(KRJOp, Arity, FromNodes);
  // set parent pointers for all the children
  int i;
  for (i = 0; i < Arity; i++) {
    FromNodes[i]->SetParent(KRJNode);
  }
  return XXX_OK;
}

//end Ihab

XxxErrCode 
RelOpDagBuilder::AddGroupByNode(RelQueryNode *&Root)
{
  XxxErrCode Status = XXX_OK;
  RelGrpByOp *GrpByOp = new RelGrpByOp(GroupByProjection, Aggregates, 
				       Groupings, Status);
  if (Status != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOP, "Cannot instantiate group-by operator");
    return XXX_NOT_OK;
  }
  RelQueryNode *Child = Root;
  Root = new RelQueryNode(GrpByOp, 1, &Root);
  // set the parent pointer
  Child->SetParent(Root);
	
  return XXX_OK;
}

XxxErrCode 
RelOpDagBuilder::AddOrderByNode(RelQueryNode *&Root)
{
  XxxErrCode Status = XXX_OK;
  RelOrderByOp *OrderByOp = 
    new RelOrderByOp(Ordering, Ascending, Status);
  if (Status != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOP, "Cannot instantiate order-by operator");
    return XXX_NOT_OK;
  }
  RelQueryNode *Child = Root;
  Root = new RelQueryNode(OrderByOp, 1, &Root);
  // set the parent pointer
  Child->SetParent(Root);
	
  return XXX_OK;
}

//ihab:actual add stop after node o top of the dag
XxxErrCode 
RelOpDagBuilder::AddStopAfterNode(RelQueryNode *&Root)
{
  
  XxxErrCode Status = XXX_OK;
  if(StopValue){
    RelStopAfterOp *StopAfterOp =  new RelStopAfterOp(StopValue,Status);
    if (Status != XXX_OK) {
      REL_NOTE_ERROR(REL_BADOP, "Cannot instantiate Stop-after operator");
      return XXX_NOT_OK;
    }
    RelQueryNode *Child = Root;
    Root = new RelQueryNode(StopAfterOp, 1, &Root);
    // set the parent pointer
    Child->SetParent(Root);
  }
  return XXX_OK;
 
}

XxxErrCode 
RelOpDagBuilder::AddFinalSPJNode(RelQueryNode *&Root)
{
  XxxErrCode Status = XXX_OK;  
  RelSPJOp *SPJOp = new RelSPJOp(GroupSelection, FinalProjection, 
				 XXX_FALSE, Status);
  if (Status != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOP, "Cannot instantiate SPJ operator");
    return XXX_NOT_OK;
  }
  RelQueryNode *Child = Root;
  Root = new RelQueryNode(SPJOp, 1, &Root);
  // set the parent pointer
  Child->SetParent(Root);
	
  return XXX_OK;
}

//begin Ihab
XxxErrCode 
RelOpDagBuilder::AddFinalKRJNode(RelQueryNode *&Root)
{
  XxxErrCode Status = XXX_OK;  
  RelSPJOp *KRJOp = new RelKRJOp(GroupSelection, FinalProjection,   XXX_FALSE, Ordering,Status);
  if (Status != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOP, "Cannot instantiate Key RJ operator");
    return XXX_NOT_OK;
  }
  RelQueryNode *Child = Root;
  Root = new RelQueryNode(KRJOp, 1, &Root);
  // set the parent pointer
  Child->SetParent(Root);
	
  return XXX_OK;
}
