%{
#include <stdio.h>
#include <stdlib.h>
#include "xxx_thread.h"   // basic thread stuff : to get current thread
#include "expr.h"         // various kinds of expressions
#include "aggrexpr.h"
#include "caseexpr.h"
#include "castexpr.h"
#include "funcexpr.h"
#include "inlistexpr.h"
#include "nullexpr.h"
#include "valbool.h"
#include "values.h"
#include "reladt.h"       // includes types.h
#include "relerror.h"     // error messages
#include "releng.h"
#include "reldata.h"
#include "relexpr.h"
#include "relenv.h"

#include "commands.h"     // commands supported by the relational server
#include "relquery.h"     // relational query tree
#include "relopt.h"       // relational optimizer and plans
#include "relexec.h"      // relational query execution

#include "alias.h"
#include "enum.h"
#include "struct.h"
#include "structcastexpr.h"
#include "genfunc.h"

/* for HACK */
#include "stringtypes.h"

#ifdef XXX_WIN32
// Bison for Win32 output checks for "MSDOS", so we need to define it
#ifndef MSDOS
#define MSDOS
#endif
//Bjarke, namespace std workaround for VC++ ... argh!:
#define std 
#endif

#if (XXX_BISON <= 122)
/** For Bison Version 1.22, use yyparse(), for 1.24 use yyparse(X) **/
#define yyparse() XxxRelationEngine::sqlparse(parseStruct& PDS) const
#else
#define yyparse(X) XxxRelationEngine::sqlparse(parseStruct& PDS) const
#endif

#if (XXX_BISON >= 125)
/** We want a more detailed error report that indicates the location of the error **/
#define YYERROR_VERBOSE 
#else
/** This seems to cause an unexplained _compile_ error in Bison
Version 1.24! So too bad -- no meaningful error messages ***/
#endif

struct FullType {
  XxxBasicType Type;
  char* MetaInfo;
 };

typedef union {
  XxxBool boolVal;
  int  intNum;
  double doubleNum;
  char* string;
  XxxBasicType type;
  struct FullType* fulltype;
  struct StructType* structtype;
  XxxConstValue* constVal;
  XxxRelCommandInfo* command;
  XxxArithOpType valueOp;
  XxxRelOpType relOp;
  RelUnarySetOpType  setUnRelOp;
  XxxBooleanExpression* boolExprVal;
  XxxValueExpression* valueExprVal;
  XxxValueExpression** valueExprs;
  XxxValueExprList* valueExprList;
  RelInfo* relInfo;
  RelViewClass::ProjInfo* projInfo;
  RelViewClass::FromInfo* fromInfo;
  RelViewClass::GroupByInfo* groupByInfo;
  RelViewClass::OrderInfo* orderInfo;
  RelViewClass::WindowInfo* windowInfo;		
  RelViewClass::StopInfo* stopInfo;	 // ihab: stop after info

  RelViewClass* view;
  RelDataSrc::AttrTypeList* attrTypeList;
  RelDataSrc::KeyList*   keyList;
  RelDataEngine::UpdateItemList* updateList;
} stackType;

#define YYSTYPE stackType

#include "sql.lex.C"

/** We want to display the location of the error **/
/***
#define yyerror(X) { \
 if ((yyFlexLexer *)PDS.LexerPtr != NULL) { \
   char buf[500]; \
   const char *CurText; \
   sprintf(buf, "Parsing error %s at position: %s\n", X, ((CurText=((yyFlexLexer *)PDS.LexerPtr)->YYText()) ? CurText : "NULL"));\
   REL_NOTE_ERROR(REL_BADQUERY, buf); \
  } \
 }
 ***/
#define yyerror(X) { \
 if ((yyFlexLexer *)PDS.LexerPtr != NULL) { \
   char buf[500]; \
   const char *CurText; \
   sprintf(buf, "Parsing error %s\n", X);\
   REL_NOTE_ERROR(REL_BADQUERY, buf); \
  } \
 }

#undef yylex
#if (XXX_FLEX <= 247)
/** For older Flex versions, replace "->yylex(X)" by "->sqllex(X)" **/
#define yylex(X) ((yyFlexLexer *)(PDS.LexerPtr?PDS.LexerPtr:(PDS.LexerPtr=(void*)(new yyFlexLexer(PDS.IStr)))))->sqllex((void*)(X))
#else
#define yylex(X) ((yyFlexLexer *)(PDS.LexerPtr?PDS.LexerPtr:(PDS.LexerPtr=(void*)(new yyFlexLexer(PDS.IStr)))))->yylex((void*)(X))
#endif

#undef yytext
#define yytext ((yyFlexLexer *)PDS.LexerPtr)->YYText()

/*
 * Return TypeId for specified TypeName.
 */
//Bjarke, aug. 9: const addded to char* parameter
static int GetTypeId(const char *TypeName) {
  int TypeId;
  XxxUseMutex *AutoMutex = new XxxUseMutex(*XxxADTClass::Latch);

  for (TypeId = 0; TypeId < XxxADTClass::NumADTs; TypeId++)
    if (!strcasecmp(XxxADTClass::ADTArray[TypeId]->TypeName, TypeName)) {
      break;
    }
  if (TypeId == XxxADTClass::NumADTs)
    TypeId = XXX_UNKNOWN;
  delete AutoMutex;
  return TypeId;
}

/*
 * Convert a string into a real MetaInfo.
 */
//Bjarke, aug. 9: const addded to char* parameter
static XxxErrCode ToMetaInfo(int TypeId,
                             const char *MetaString,
                             XxxADTMetaInfo** MetaInfo,
  			     XxxADTClass::MetaInfoHandle& MIHandle) {
  XxxBool Found = XXX_FALSE;

  if (MetaString) {
    /**
     * Note: XXX_FIXEDSTRING does not work, so using XXX_STRING.
     */
    if (TypeId == XXX_STRING) {
       return XXX_OK;
    }
    else if (TypeId == XXX_UNKNOWN ||
      XxxADTArray[TypeId]->AddMetaInfo("dummy",
  			             XXX_FALSE, 
  			             MetaString,
  			             MIHandle) != XXX_OK) {
      return XXX_NOT_OK;
    }
    if ((XxxADTArray[TypeId]->GetMetaInfo("dummy",
  				        MIHandle,
  				        *MetaInfo,
  				        Found) != XXX_OK)
        || (Found != XXX_TRUE)) {
      return XXX_NOT_OK;
    }
  }
  return XXX_OK;
}

%}
%pure_parser

%type <boolVal> unique clustered asc_desc null_info 
%type <intNum> XXX_TOK_INTEGER col_number idx_attr opt_src
%type <doubleNum> XXX_TOK_DOUBLE
%type <string> XXX_TOK_STRING XXX_TOK_QUOTESTRING alias opt_alias relation_name opt_relation_name agg_op function_name indx_name indx_type db_name opt_dev opt_type file_name any_string opt_meta_info alias_type_name alias_to_type enum_type_name struct_type_name
%type <type> data_type
%type <fulltype> full_data_type check_full_data_type
/*%type <structtype> struct_data_type */
%type <command> opt_args
%type <constVal> const s_const q_const trim_char
%type <relOp> rel_op
%type <setUnRelOp> set_unrel_op

%type <view> query_block paren_query_block
%type <projInfo>  select_clause
%type <relInfo> relinfo 
%type <fromInfo> from_clause relation_list
%type <orderInfo> order_by_clause order_by_list
%type <windowInfo> window_clause
%type <stopInfo> stop_after_clause //ihab: stop after 
%type <groupByInfo> group_by_clause group_by_list
%type <boolExprVal> where_clause having_clause boolean_predicate rel_predicate
%type <valueExprVal> star_expr projection_expr column_expr named_column_expr col_name col_position date_expr date func_expr trim_spec value_expr alias_ref case_expr case_default
%type <valueExprList> projection_list column_expr_list opt_column_expr_list value_list
%type <valueExprs> in_list case_item_list case_item
%type <attrTypeList> col_data_type_list struct_col_data_type_list
%type <keyList> col_list opt_key
%type <updateList> update_list 

%token XXX_TOK_WINDOW
%token XXX_TOK_STRING XXX_TOK_QUOTESTRING XXX_TOK_INTEGER XXX_TOK_DOUBLE
%token XXX_TOK_SEMICOLON XXX_TOK_L_PAREN XXX_TOK_R_PAREN XXX_TOK_L_BRACE XXX_TOK_R_BRACE XXX_TOK_L_CBRACE XXX_TOK_R_CBRACE XXX_TOK_COMMA XXX_TOK_DOT XXX_TOK_DOLLAR XXX_TOK_POUND
%token XXX_TOK_VIEW XXX_TOK_WITH XXX_TOK_SELECT XXX_TOK_FROM XXX_TOK_WHERE XXX_TOK_GROUP XXX_TOK_BY XXX_TOK_HAVING XXX_TOK_ORDER XXX_TOK_AS
%token XXX_TOK_CREATE XXX_TOK_DROP XXX_TOK_TABLE XXX_TOK_INDEX XXX_TOK_DATABASE XXX_TOK_OPEN XXX_TOK_CLOSE XXX_TOK_ON XXX_TOK_KEY XXX_TOK_LOAD XXX_TOK_NULL XXX_TOK_RUNSTATS XXX_TOK_DUMPTABLE
%token XXX_TOK_SOURCE XXX_TOK_DEVICE XXX_TOK_IMPLTYPE
%token XXX_TOK_GT XXX_TOK_LT XXX_TOK_EQ XXX_TOK_AND XXX_TOK_OR XXX_TOK_NOT XXX_TOK_MINUS XXX_TOK_PLUS XXX_TOK_DIV XXX_TOK_STAR XXX_TOK_GEQ XXX_TOK_LEQ XXX_TOK_NEQ XXX_TOK_LIKE XXX_TOK_CONCAT
%token XXX_TOK_ASC XXX_TOK_DESC XXX_TOK_ALL XXX_TOK_ANY XXX_TOK_UNIQUE XXX_TOK_CLUSTERED XXX_TOK_DISTINCT XXX_TOK_EXISTS XXX_TOK_BETWEEN XXX_TOK_IN XXX_TOK_IS
%token XXX_TOK_INSERT XXX_TOK_INTO XXX_TOK_DELETE XXX_TOK_VALUES XXX_TOK_UPDATE XXX_TOK_SET 
%token XXX_TOK_ALIAS XXX_TOK_TYPE XXX_TOK_ENUM XXX_TOK_STRUCT

%token XXX_TOK_COUNT
%token XXX_TOK_CASE XXX_TOK_WHEN XXX_TOK_THEN XXX_TOK_ELSE XXX_TOK_END
%token XXX_TOK_CAST %token XXX_TOK_CHAR XXX_TOK_VARYING XXX_TOK_VARCHAR
%token XXX_TOK_DATE XXX_TOK_INTERVAL XXX_TOK_YEAR XXX_TOK_MONTH
%token XXX_TOK_DAY XXX_TOK_HOUR XXX_TOK_MINUTE XXX_TOK_SECOND
%token XXX_TOK_ABSOLUTE XXX_TOK_UPPER XXX_TOK_LOWER
%token XXX_TOK_TRIM XXX_TOK_LEADING XXX_TOK_TRAILING XXX_TOK_BOTH
%token XXX_TOK_KRJ  // ihab
%token XXX_TOK_STOP XXX_TOK_AFTER //ihab

%left XXX_TOK_AND XXX_TOK_OR
%left XXX_TOK_PLUS XXX_TOK_MINUS
%left XXX_TOK_STAR XXX_TOK_DIV
%left XXX_TOK_NOT

%%
input :
       input_block_list 
         { 
	   // IF THIS IS THE TOP LEVEL PARSER
	   PDS.OutArg = NULL; 
	 }
        | query_block 
        {
	   // IF THIS IS A NESTED PARSER
	   PDS.OutArg = (void *)($1);
	}
        | column_expr
        {
           // For parsing index expression
           PDS.OutArg = (void *) ($1);
        }
/***
	| boolean_predicate
	{
	   // For parsing boolean expression
	   PDS.OutArg = (void *) ($1);
	}
	***/
       // The '{' and '}' must be provided to parse the 
       // meta info string of struct ADT correctly, and in the
       // same time, to avoid conflict with the rule of
       // input_block_list, such as "list rels;"
       | XXX_TOK_L_CBRACE struct_col_data_type_list XXX_TOK_R_CBRACE
       {
	 // IF THIS IS An Attribute List
	 PDS.OutArg = (void *)($2);
       }
       | input_schema
         {
         }
;
input_block_list: input_block_list input_block
         {
	 }
        | input_block
         {
	 }
;
/* An SQL command can be a query, update or utility */
input_block:	  
                  query  XXX_TOK_SEMICOLON 
                | utility XXX_TOK_SEMICOLON 
		| update XXX_TOK_SEMICOLON
                | XXX_TOK_SEMICOLON
    	        | error XXX_TOK_SEMICOLON 
                  { }
                   input_block {}
		;

query:
             XXX_TOK_CREATE XXX_TOK_VIEW XXX_TOK_STRING XXX_TOK_AS paren_query_block
             {
	      if (QueryEng->StoreView($3, $5) != XXX_OK) {
		REL_NOTE_ERROR(REL_DUPVIEW, $3);
		if ($5) delete $5;
		if ($3) delete [] $3;
		YYERROR;
	       }
              if ($3) delete [] $3;
	     }
           | query_block
             {
	      // specify that answers should be printed
	      RelSpecParameters Params(XXX_TRUE); 
              if (execQuery($1, Params) != XXX_OK) {
		REL_NOTE_ERROR(REL_BADQUERY, "Error executing query");
		if ($1) delete $1;
		YYERROR;
	       }
              if ($1) delete $1;
	     }
;
//ihab:stop after
query_block:	select_clause from_clause where_clause order_by_clause stop_after_clause 
                {
		 $$ = new RelViewClass("");
		 $$->AddFrom($2);
		 $$->AddSelection($3);
		 $$->AddProjection($1);
		 $$->AddOrderBy($4);
		 $$->AddStopAfter($5);
                }
                |	
select_clause from_clause where_clause order_by_clause window_clause
                {
		 $$ = new RelViewClass("");
		 $$->AddFrom($2);
		 $$->AddSelection($3);
		 $$->AddProjection($1);
		 $$->AddOrderBy($4);
  		 $$->AddWindow($5);	
                }
                |
                select_clause from_clause where_clause 
			group_by_clause having_clause order_by_clause
                {
		 $$ = new RelViewClass("");
		 $$->AddFrom($2);
		 $$->AddSelection($3);
		 $$->AddProjection($1);
		 $$->AddGroupBy($4);
		 $$->AddHaving($5);
		 $$->AddOrderBy($6);
                }
;
paren_query_block: XXX_TOK_L_PAREN query_block XXX_TOK_R_PAREN
                {  $$ = $2; }

;

window_clause: 
		{$$=NULL;}	
		| XXX_TOK_WINDOW XXX_TOK_INTEGER XXX_TOK_COMMA XXX_TOK_INTEGER XXX_TOK_COMMA XXX_TOK_INTEGER XXX_TOK_COMMA XXX_TOK_INTEGER
                {
 		 RelViewClass::WindowInfo *Tmp = new RelViewClass::WindowInfo();
		 Tmp->IntervalHours = $2;
		 Tmp->IntervalMinutes = $4;
		 Tmp->IntervalSeconds = $6;
		 Tmp->IntervalMicroSecs = $8;
		 $$ = Tmp;
		}
;

//ihab :stop after
stop_after_clause: 
		{$$=NULL;}
		| XXX_TOK_STOP XXX_TOK_AFTER XXX_TOK_INTEGER 
                {
		RelViewClass::StopInfo *Tmp = new RelViewClass::StopInfo();
		Tmp->K = $3;
		$$ = Tmp;
		} 	
;

select_clause:
		XXX_TOK_SELECT unique projection_list
                {
		 RelViewClass::ProjInfo *Tmp= new RelViewClass::ProjInfo();
		 Tmp->Distinct = $2;
		 Tmp->TargetList = $3;
		 $$ = Tmp;
                }
;
projection_list: projection_list XXX_TOK_COMMA projection_expr 
               {
		$$ = $1->AddExpr($3);
	       }
               | projection_expr { $$ = new XxxValueExprList($1); } 
;
projection_expr: 
               named_column_expr { $$ = $1; }
               |
               star_expr { $$ = $1; }
;
named_column_expr: column_expr XXX_TOK_AS XXX_TOK_STRING
               {
		$1->ExprName = new char[strlen($3)+1];
		strcpy($1->ExprName, $3);
		if ($3) delete [] $3;
		$$ = $1;
               }
            | column_expr
               {
		$1->ExprName = NULL;
		$$ = $1;
               }
;
star_expr:
		XXX_TOK_STAR                    /* e.g. select * from */
                {
		  $$ = new XxxUnknownValue("*", "*");
		}
                |
		XXX_TOK_STRING XXX_TOK_DOT XXX_TOK_STAR         /* e.g. Employees.*   */
                {
	         $$ = new XxxUnknownValue("*", $1);
	         if ($1) delete [] $1;
		}
;
opt_column_expr_list: column_expr_list { $$ = $1 ;}
          |  { $$ = NULL; }
;
column_expr_list: column_expr_list XXX_TOK_COMMA column_expr 
               {
		$$ = $1->AddExpr($3);
	       }
            | column_expr { $$ = new XxxValueExprList($1); } 
;
column_expr: col_name { $$ = $1;}
           | col_position { $$ = $1;}
           | alias_ref { $$ = $1;}
           | q_const { $$ = $1;} 
           | XXX_TOK_DOLLAR XXX_TOK_INTEGER 
             {
	       $$ = new XxxArgValue($2);
	     }
           | XXX_TOK_MINUS column_expr
             {
               XxxExpression* Tmp[1];
               Tmp[0] = $2;
               $$ = new XxxArithExpression(XXX_UMINUS_OP, 1, Tmp);
             }
           | column_expr XXX_TOK_PLUS column_expr
             {
	       XxxExpression* Tmp[2];
	       Tmp[0] = $1; Tmp[1] = $3;
	       $$ = new XxxArithExpression(XXX_PLUS_OP, 2, Tmp);
             }
           | column_expr XXX_TOK_MINUS column_expr
             {
	       XxxExpression* Tmp[2];
	       Tmp[0] = $1; Tmp[1] = $3;
	       $$ = new XxxArithExpression(XXX_MINUS_OP, 2, Tmp);
             }
           | column_expr XXX_TOK_STAR column_expr
             {
	       XxxExpression* Tmp[2];
	       Tmp[0] = $1; Tmp[1] = $3;
	       $$ = new XxxArithExpression(XXX_TIMES_OP, 2, Tmp);
             }
           | column_expr XXX_TOK_DIV column_expr
             {
	       XxxExpression* Tmp[2];
	       Tmp[0] = $1; Tmp[1] = $3;
	       $$ = new XxxArithExpression(XXX_DIV_OP, 2, Tmp);
             }
           | XXX_TOK_L_PAREN column_expr XXX_TOK_R_PAREN
             { $$ = $2; }
           | agg_op XXX_TOK_L_PAREN column_expr_list XXX_TOK_R_PAREN 
             {
	       $$ = new XxxAggrExpression($1, $3);
	     }
           | XXX_TOK_COUNT XXX_TOK_L_PAREN column_expr_list XXX_TOK_R_PAREN 
             {
	       $$ = new XxxAggrExpression("count", $3);
	     }
           | XXX_TOK_COUNT XXX_TOK_L_PAREN XXX_TOK_STAR XXX_TOK_R_PAREN
             {
  	       XxxValueExprList *t;
	       t = new XxxValueExprList(new XxxUnknownValue(0, ""));
	       $$ = new XxxAggrExpression("count", t);
	     }
           | column_expr XXX_TOK_CONCAT column_expr
             {
               XxxValueExprList *args;
               args = new XxxValueExprList($1);
               args->AddExpr($3);
               $$ = new XxxFuncExpression("$concat", args);
             }
           | case_expr { $$ = $1; }
/* Extensible function expression */
           | func_expr
             {
	       $$ = $1;
             }
/* ADT function expression */
           | column_expr XXX_TOK_DOT function_name XXX_TOK_L_PAREN opt_column_expr_list XXX_TOK_R_PAREN
             {
	      $$ = new XxxFuncExpression($3, $5, $1);
             }
           | date { $$ = $1; }
           | date_expr { $$ = $1; }
           | XXX_TOK_CAST XXX_TOK_L_PAREN column_expr XXX_TOK_AS full_data_type XXX_TOK_R_PAREN
             {
	       XxxADTMetaInfo* MetaInfo = NULL;
               XxxADTClass::MetaInfoHandle MIHandle;
               if (ToMetaInfo($5->Type, $5->MetaInfo, &MetaInfo, MIHandle) != XXX_OK) {
                 XXX_NOTE_ERROR(XXX_BADTYPE, "ToMetaInfo failed");
                 YYERROR;
               }
	       $$ = new XxxCastValueExpression($5->Type, MetaInfo, $3);
	       // TODO:L2 also remove the MetaInfoHandle from the data type
	       if ($5->MetaInfo) delete $5->MetaInfo;
	       if ($5) delete $5;
             }
           | XXX_TOK_L_BRACE full_data_type XXX_TOK_R_BRACE column_expr
             {
	       XxxADTMetaInfo* MetaInfo = NULL;
               XxxADTClass::MetaInfoHandle MIHandle;
               if (ToMetaInfo($2->Type, $2->MetaInfo, &MetaInfo, MIHandle) != XXX_OK) {
                 XXX_NOTE_ERROR(XXX_BADTYPE, "ToMetaInfo failed");
                 YYERROR;
               }
	       $$ = new XxxCastValueExpression($2->Type, MetaInfo, $4);
	       if ($2->MetaInfo) delete $2->MetaInfo;
	       if ($2) delete $2;
             }
           | XXX_TOK_L_BRACE full_data_type XXX_TOK_R_BRACE XXX_TOK_L_PAREN column_expr_list XXX_TOK_R_PAREN
             {
	       XxxADTMetaInfo* MetaInfo = NULL;
               XxxADTClass::MetaInfoHandle MIHandle;
               if (ToMetaInfo($2->Type, $2->MetaInfo, &MetaInfo, MIHandle) != XXX_OK) {
                 XXX_NOTE_ERROR(XXX_BADTYPE, "ToMetaInfo failed");
                 YYERROR;
               }
	       if ($2->MetaInfo) delete $2->MetaInfo;
	       $$ = new XxxStructCastExpression($2->Type, ((XxxStructADT::StructMetaInfo *)MetaInfo), $5);
	       if ($2) delete $2;
             }
/*** wierd syntax for now: [[ bool_pred ]] ***/
           | XXX_TOK_L_BRACE XXX_TOK_L_BRACE boolean_predicate XXX_TOK_R_BRACE XXX_TOK_R_BRACE 
             {
	      $$ = $3;
             }

 /*          | XXX_TOK_L_BRACE struct_data_type XXX_TOK_R_BRACE XXX_TOK_L_PAREN column_expr_list XXX_TOK_R_PAREN
			{
			XxxADTClass::MetaInfoHandle MIHandle;
			XxxErrCode status;
		    if ($2->AttrsList) {
				XxxStructADT::StructMetaInfo *structMetaInfo = new XxxStructADT::StructMetaInfo($2->AttrsList, MIHandle, status);
				if(status != XXX_OK){
					YYERROR;
				}
				//used by the struct data type
				$$ = new XxxStructCastExpression($2->Type, structMetaInfo, $5);

			} else {
				XXX_NOTE_ERROR(XXX_BADTYPE, "struct ADT must have meta string with it");
				YYERROR;
			}

			if ($2) delete $2;
			}
*/
;
agg_op: XXX_TOK_STRING
        { // underspecified for extensibility.
          $$ = $1;
        }
;

case_expr: XXX_TOK_CASE case_item_list case_default XXX_TOK_END
         {
           int n = (int)($2[0]);

           $2[n] = $3;
           $$ = new XxxCaseExpression(n, &($2[1]));
         }
case_item_list: case_item_list XXX_TOK_COMMA case_item
                {
                  int n = (int)($1[0]);
                  $1[n++] = $3[0];
                  $1[n++] = $3[1];
                  $1[0] = (XxxValueExpression *)n;
                  $$ = $1;
	        }
              | case_item
                {
                  // List used to store all arguments to case expression
                  // The first element used for counting number of
                  // filled elements.
                  // TODO: check for #arg > 100.
	          XxxExpression* Args[100];
                  Args[0] = (XxxValueExpression *)3;
                  Args[1] = $1[0];
                  Args[2] = $1[1];
                  $$ = Args;
                }
;
case_item: XXX_TOK_WHEN boolean_predicate XXX_TOK_THEN column_expr
           {
	     XxxExpression* Tmp[2];
             Tmp[0] = $2;
             Tmp[1] = $4;
             $$ = Tmp;
           }
         | XXX_TOK_WHEN boolean_predicate XXX_TOK_THEN XXX_TOK_NULL
           {
	     XxxExpression* Tmp[2];
             Tmp[0] = $2;
             Tmp[1] = new XxxConstIntValue(0); // TODO: this is wrong
             $$ = Tmp;
           }
;
case_default: XXX_TOK_ELSE column_expr { $$ = $2; }
        | XXX_TOK_ELSE XXX_TOK_NULL
          {
            $$ = new XxxConstIntValue(0);
          }
        | { $$ = new XxxConstIntValue(0); }

;

date_expr: column_expr XXX_TOK_PLUS XXX_TOK_INTERVAL column_expr XXX_TOK_YEAR
          {
            XxxValueExprList *arglist = new XxxValueExprList($4);
            $$ = new XxxFuncExpression("nextYear", arglist, $1);
          }
         | column_expr XXX_TOK_PLUS XXX_TOK_INTERVAL column_expr XXX_TOK_MONTH
          {
            XxxValueExprList *arglist = new XxxValueExprList($4);
            $$ = new XxxFuncExpression("nextMonth", arglist, $1);
          }
         | column_expr XXX_TOK_PLUS XXX_TOK_INTERVAL column_expr XXX_TOK_DAY
          {
            XxxValueExprList *arglist = new XxxValueExprList($4);
            $$ = new XxxFuncExpression("next", arglist, $1);
          }
         | column_expr XXX_TOK_MINUS XXX_TOK_INTERVAL column_expr XXX_TOK_YEAR
          {
            XxxValueExprList *arglist = new XxxValueExprList($4);
            $$ = new XxxFuncExpression("prevYear", arglist, $1);
          }
         | column_expr XXX_TOK_MINUS XXX_TOK_INTERVAL column_expr XXX_TOK_MONTH
          {
            XxxValueExprList *arglist = new XxxValueExprList($4);
            $$ = new XxxFuncExpression("prevMonth", arglist, $1);
          }
         | column_expr XXX_TOK_MINUS XXX_TOK_INTERVAL column_expr XXX_TOK_DAY
          {
            XxxValueExprList *arglist = new XxxValueExprList($4);
            $$ = new XxxFuncExpression("prev", arglist, $1);
          }
;
date: XXX_TOK_DATE XXX_TOK_QUOTESTRING
    { // e.g., DATE '1998-12-01'
	       XxxADTMetaInfo* MetaInfo = NULL;
      XxxADTClass::MetaInfoHandle MIHandle;
      int TypeId = GetTypeId("Pdate");
      if (TypeId == XXX_UNKNOWN) {
         XXX_NOTE_ERROR(XXX_BADTYPE, "Pdate");
         YYERROR;
      }
      if (ToMetaInfo(TypeId, "CY-M-De", &MetaInfo, MIHandle) != XXX_OK) {
        XXX_NOTE_ERROR(XXX_BADTYPE, "ToMetaInfo failed");
        YYERROR;
      }
      XxxExpression *t = new XxxConstStringValue($2);
      $$ = new XxxCastValueExpression(TypeId, MetaInfo, t);
    }

func_expr:   XXX_TOK_DOLLAR function_name XXX_TOK_L_PAREN opt_column_expr_list XXX_TOK_R_PAREN
             {
               char s[100];
               s[0] = '$';
               strcpy(s + 1, $2);
	       $$ = new XxxFuncExpression(s, $4);
             }
           | XXX_TOK_ABSOLUTE XXX_TOK_L_PAREN column_expr_list XXX_TOK_R_PAREN
             {
               $$ = new XxxFuncExpression("$abs", $3);
             }
           | XXX_TOK_UPPER XXX_TOK_L_PAREN column_expr_list XXX_TOK_R_PAREN
             {
	       $$ = new XxxFuncExpression("$ucase", $3);
             }
           | XXX_TOK_LOWER XXX_TOK_L_PAREN column_expr_list XXX_TOK_R_PAREN
             {
	       $$ = new XxxFuncExpression("$lcase", $3);
             }
           | XXX_TOK_TRIM XXX_TOK_L_PAREN trim_spec XXX_TOK_R_PAREN
             {
               $$ = $3;
             }
;
trim_spec: XXX_TOK_LEADING trim_char XXX_TOK_FROM column_expr_list
           {
             XxxConstIntValue *loc = new XxxConstIntValue(XXX_TRIM_LEADING);
             $4->AddExpr(loc);
             $$ = new XxxFuncExpression("$trim", $4->AddExpr($2));
           }
         | XXX_TOK_TRAILING trim_char XXX_TOK_FROM column_expr_list
           {
             XxxConstIntValue *loc = new XxxConstIntValue(XXX_TRIM_TRAILING);
             $4->AddExpr(loc);
             $$ = new XxxFuncExpression("$trim", $4->AddExpr($2));
           }
         | XXX_TOK_BOTH trim_char XXX_TOK_FROM column_expr_list
           {
             XxxConstIntValue *loc = new XxxConstIntValue(XXX_TRIM_BOTH);
             $4->AddExpr(loc);
             $$ = new XxxFuncExpression("$trim", $4->AddExpr($2));
           }
         | XXX_TOK_QUOTESTRING XXX_TOK_FROM column_expr_list
           {
             XxxConstIntValue *loc = new XxxConstIntValue(XXX_TRIM_BOTH);
             XxxExpression *s = new XxxConstStringValue($1);
             $3->AddExpr(loc);
             $$ = new XxxFuncExpression("$trim", $3->AddExpr(s));
           }
         | column_expr_list
           {
             XxxConstIntValue *loc = new XxxConstIntValue(XXX_TRIM_BOTH);
	     XxxExpression *s = new XxxConstStringValue(" ");
             $1->AddExpr(loc);
             $$ = new XxxFuncExpression("$trim", $1->AddExpr(s));
           }
;
trim_char: XXX_TOK_QUOTESTRING
           {
             $$ = new XxxConstStringValue($1);
           }
         | /* NULL */
           {
             $$ = new XxxConstStringValue(" ");
           }
;
function_name: XXX_TOK_STRING { $$ = $1; }
;

from_clause:	XXX_TOK_FROM relation_list { $2->KRJOp = XXX_FALSE;$$ = $2;}
				| XXX_TOK_KRJ relation_list	{ $2->KRJOp = XXX_TRUE; $$ = $2;}
;
relation_list:	 relation_list XXX_TOK_COMMA relinfo
                 {
		  $1->AddRelInfo($3);
		  $$ = $1;
		 }
		| relinfo
                 {
		   $$ = new RelViewClass::FromInfo($1);
		 }
		;
relinfo:        relation_name opt_alias
                 {
		  $$ = new SimpleRelInfo($1, ($2?$2:$1));
                  if ($1) delete [] $1;
                  if ($2) delete [] $2;
		 }
/********
            |   func_expr alias
                 {
		  // $$ = new SimpleRelInfo($1, $2);
		  $$ = NULL; // TODO:L2
                  if ($2) delete [] $2;
		 }
*******/
            |   paren_query_block alias
                 {
		  $$ = new NestedRelInfo($1, $2);
                  if ($2) delete [] $2;
		 }
;
opt_relation_name: relation_name { $$ = $1;}
            |  { $$ = NULL; }
;
relation_name:	  XXX_TOK_STRING { $$ = $1;}
            |
                  XXX_TOK_DOLLAR XXX_TOK_INTEGER 
                  { 
		   $$ = new char[10];
		   sprintf($$, "$%d", $2);
		  }
;
opt_alias :  alias {$$ = $1;}
         | { $$ = NULL; }
;
alias :    XXX_TOK_STRING { $$ = $1;}
         |  XXX_TOK_AS XXX_TOK_STRING { $$ = $2; }
;

where_clause:	XXX_TOK_WHERE boolean_predicate
                  { $$ = $2;}
         |        { $$ = NULL;}
;

group_by_clause:	XXX_TOK_GROUP XXX_TOK_BY group_by_list
                 {
		   $$ = $3;
		 }
;

group_by_list:	 group_by_list XXX_TOK_COMMA named_column_expr
                 {
                  ($1->ColArity)++;
		  $1->GroupList->AddExpr($3);
		 }
                | named_column_expr
                 {
		  RelViewClass::GroupByInfo *Tmp = 
                    new RelViewClass::GroupByInfo();
                  Tmp->ColArity = 1;
                  Tmp->GroupList = new XxxValueExprList($1);
                  $$ = Tmp;
                 }
                ;

having_clause:  { $$ = NULL;}
                | XXX_TOK_HAVING boolean_predicate
                 {
		  $$ = $2;
		 }
                ;

order_by_clause:  { $$ = NULL;}
		| XXX_TOK_ORDER XXX_TOK_BY order_by_list
                  { $$ = $3;}
		;
order_by_list:	 order_by_list XXX_TOK_COMMA column_expr asc_desc
                 {
                  ($1->ColArity)++;
		  $1->TargetList->AddExpr($3);
                  // this adds the new asc_desc XxxBool
                  XxxBool *NewAscending = new XxxBool[$1->ColArity];
                  for (int i = 0; i < $1->ColArity - 1; i++) {
                    NewAscending[i] = $1->Ascending[i];
		  }
                  NewAscending[$1->ColArity - 1] = $4;
                  if ($1->Ascending) delete $1->Ascending;
                  $1->Ascending = NewAscending;
                  $$ = $1;
		 }
                | column_expr asc_desc
                 {
		  RelViewClass::OrderInfo *Tmp = 
                    new RelViewClass::OrderInfo();
                  Tmp->ColArity = 1;
                  Tmp->TargetList = new XxxValueExprList($1);
                  Tmp->Ascending = new XxxBool($2);
                  $$ = Tmp;
                 }
                ;
asc_desc:	  { $$ = XXX_TRUE;}
		| XXX_TOK_ASC { $$ = XXX_TRUE;}
		| XXX_TOK_DESC { $$ = XXX_FALSE;}
		;

boolean_predicate:
		  boolean_predicate XXX_TOK_AND boolean_predicate
                  {
		   XxxExpression* Tmp[2];
		   Tmp[0] = $1; Tmp[1] = $3;
		   $$ = new XxxBoolBooleanExpression(XXX_AND_OP, 2, Tmp);
		  }
		| boolean_predicate XXX_TOK_OR  boolean_predicate
                  {
		   XxxExpression* Tmp[2];
		   Tmp[0] = $1; Tmp[1] = $3;
		   $$ = new XxxBoolBooleanExpression(XXX_OR_OP, 2, Tmp);
		  }
		| XXX_TOK_NOT boolean_predicate
                  {
		   XxxExpression* Tmp[1];
		   Tmp[0] = $2;
		   $$ = new XxxBoolBooleanExpression(XXX_NOT_OP, 1, Tmp);
		  }
                | XXX_TOK_L_PAREN boolean_predicate XXX_TOK_R_PAREN
                  {$$ = $2;}
                | rel_predicate
                  {$$ = $1;}
                | set_unrel_op paren_query_block
                  { 
		    if ($1 == REL_EXISTS_OP) {
		      $$ = new RelExistsExpression(this, $2);
		      // do not delete $2 (it is used by the expression)
		    }
		    else if ($1 == REL_UNIQUE_OP) {
		      $$ = new RelUniqueExpression(this, $2);
		      // do not delete $2 (it is used by the expression)
		    }
		    else {
		      $$ = NULL;
		      YYERROR;
		    }
		  }
/**  value expression of type XXX_BOOLEAN **/
                | column_expr
                  {
		   $$ = new XxxValBooleanExpression($1);
		  }
		;

rel_predicate:   
                 column_expr rel_op column_expr 
                  { 
		   XxxExpression* Tmp[2];
		   Tmp[0] = $1; Tmp[1] = $3;
		   $$ = new XxxRelBooleanExpression($2, 2, Tmp); 
		  }
                | column_expr rel_op XXX_TOK_ANY paren_query_block 
                  {
		    $$ = new RelAnyExpression(this, $4, $2, $1);
		  }
                | column_expr rel_op XXX_TOK_ALL paren_query_block 
                  {
		    // treat > ALL as NOT (<= ANY)
		    XxxExpression* Tmp[1];
		    Tmp[0] = new RelAnyExpression(this, $4, RelOpNeg($2), $1);
		    $$ = new XxxBoolBooleanExpression(XXX_NOT_OP, 1, Tmp);
		   }
                | column_expr XXX_TOK_IN paren_query_block 
                  {
		    $$ = new RelInExpression(this, $3, $1);
		  }
                | column_expr XXX_TOK_IN XXX_TOK_L_PAREN in_list XXX_TOK_R_PAREN
                  {
                    int n = (int)($4[0]);
                    $4[0] = $1;
                    $$ = new XxxInListExpression(n, $4);
                  }
                | column_expr XXX_TOK_NOT XXX_TOK_IN paren_query_block 
                  {
		    XxxExpression* Tmp[1];
		    Tmp[0] = new RelInExpression(this, $4, $1);
		    $$ = new XxxBoolBooleanExpression(XXX_NOT_OP, 1, Tmp);
		  }
                | column_expr XXX_TOK_NOT XXX_TOK_IN XXX_TOK_L_PAREN in_list XXX_TOK_R_PAREN
                  {
		    XxxExpression* Tmp[1];
                    int n = (int)($5[0]);
                    $5[0] = $1;
                    Tmp[0] = new XxxInListExpression(n, $5);
		    $$ = new XxxBoolBooleanExpression(XXX_NOT_OP, 1, Tmp);
                  }
		| column_expr XXX_TOK_IS XXX_TOK_NULL
                  { 
		   $$ = new XxxIsNullExpression($1); 
		  }		
		| column_expr XXX_TOK_IS XXX_TOK_NOT XXX_TOK_NULL
                  { 
		    XxxExpression* Tmp[1];
		    Tmp[0] = new XxxIsNullExpression($1);
		    $$ = new XxxBoolBooleanExpression(XXX_NOT_OP, 1, Tmp);
		  }
                | column_expr XXX_TOK_LIKE column_expr
                  {
                    XxxValueExprList *a;
                    XxxExpression *b;
                    a = new XxxValueExprList($1); 
                    b = new XxxFuncExpression("$strmatch", a->AddExpr($3));
		    $$ = new XxxValBooleanExpression(b);
                  }
                | column_expr XXX_TOK_NOT XXX_TOK_LIKE column_expr
                  {
                    XxxValueExprList *a;
                    XxxExpression* b;
                    a = new XxxValueExprList($1); 
                    b = new XxxFuncExpression("$notstrmatch", a->AddExpr($4));
		    $$ = new XxxValBooleanExpression(b);
                  }
                | column_expr XXX_TOK_BETWEEN column_expr XXX_TOK_AND column_expr 
                  {
                   //  $1 >= $3 AND $1 <= $5
		   XxxExpression *T[2], *L[2], *R[2];
                   L[0] = $1;
                   L[1] = $3;
		   T[0] = new XxxRelBooleanExpression(XXX_GEQ_OP, 2, L);
                   R[0] = $1->Replicate();
                   R[1] = $5;
		   T[1] = new XxxRelBooleanExpression(XXX_LEQ_OP, 2, R);
		   $$ = new XxxBoolBooleanExpression(XXX_AND_OP, 2, T);
		  }
		| column_expr XXX_TOK_NOT XXX_TOK_BETWEEN column_expr XXX_TOK_AND column_expr
                  {
                   //  NOT ($1 >= $3 AND $1 <= $5)
		   XxxExpression *T[2], *L[2], *R[2];
                   L[0] = $1;
                   L[1] = $4;
		   T[0] = new XxxRelBooleanExpression(XXX_LT_OP, 2, L);
                   R[0] = $1->Replicate();
                   R[1] = $6;
		   T[1] = new XxxRelBooleanExpression(XXX_GT_OP, 2, R);
		   $$ = new XxxBoolBooleanExpression(XXX_OR_OP, 2, T);
		  }
/***
		| column_expr NOT MATCH sub_query
		| column_expr MATCH sub_query
****/
		;

rel_op: 
        XXX_TOK_GT { $$ = XXX_GT_OP;}
      | XXX_TOK_LT { $$ = XXX_LT_OP;}
      | XXX_TOK_EQ { $$ = XXX_EQ_OP;}
      | XXX_TOK_LEQ {$$ = XXX_LEQ_OP;}
      | XXX_TOK_GEQ { $$ = XXX_GEQ_OP;}
      | XXX_TOK_NEQ { $$ = XXX_NEQ_OP;}
;

set_unrel_op: XXX_TOK_EXISTS { $$ = REL_EXISTS_OP;}
          |   XXX_TOK_UNIQUE { $$ = REL_UNIQUE_OP;}
;


in_list : in_list XXX_TOK_COMMA column_expr
          {
            int n = (int)($1[0]);
            $1[n++] = $3;
            $1[0] = (XxxValueExpression *)n;
            $$ = $1;
	  }
        | column_expr
          {
            // List used to store all items in list.  The first
            // element used for counting number of filled elements.
            // TODO: check for size bigger than 100.
	    XxxExpression* Args[100];
            Args[0] = (XxxValueExpression *)2;
            Args[1] = $1;
            $$ = Args;
          }
;
const : s_const
         { $$ = $1;}
      |
        q_const
         { $$ = $1;}
;
s_const : XXX_TOK_STRING
         { $$ = new XxxConstStringValue($1);
           if ($1) delete [] $1;
         }
;
q_const : XXX_TOK_INTEGER 
         { $$ = new XxxConstIntValue($1) ;} 
      | XXX_TOK_DOUBLE 
         { $$ = new XxxConstDoubleValue($1);} 
      | XXX_TOK_QUOTESTRING 
         { $$ = new XxxConstStringValue($1);
           if ($1) delete [] $1;
         }
;
any_string : XXX_TOK_STRING { $$ = $1;}
      |      XXX_TOK_QUOTESTRING { $$ = $1;}
;

col_name:       XXX_TOK_STRING XXX_TOK_DOT XXX_TOK_STRING 	/* e.g.   S.deposit   */
                {
	         $$ = new XxxUnknownValue($3, $1);
	         if ($1) delete [] $1;
	         if ($3) delete [] $3;
		} 
                ;
col_position:   XXX_TOK_STRING XXX_TOK_DOT XXX_TOK_POUND XXX_TOK_INTEGER /* e.g.   S.#1   */
                {
	         $$ = new XxxUnknownValue($4, $1);
	         if ($1) delete [] $1;
		}
                ;
alias_ref:	XXX_TOK_STRING
                {
	         $$ = new XxxUnknownValue($1);
	         if ($1) delete [] $1;
		} 
		;

col_number:	XXX_TOK_DOLLAR XXX_TOK_INTEGER { $$ = $2; }
		;

/* Return value is XXX_TRUE if Distinct, XXX_FALSE if not Distinct */
unique:		XXX_TOK_ALL { $$ = XXX_FALSE; }
		| XXX_TOK_DISTINCT { $$ = XXX_TRUE;}
		| XXX_TOK_UNIQUE { $$ = XXX_TRUE;}
		| { $$ = XXX_FALSE;}
		;

clustered:	XXX_TOK_CLUSTERED { $$ = XXX_TRUE;}
		| { $$ = XXX_FALSE;}
		;

/********

set_op:		  INTERSECT
		| MINUS
		| UNION all
		;

query:		| L_PAREN query R_PAREN set_op L_PAREN query R_PAREN 
			set_order_by_clause
                {
		 // (select ...) UNION (select ...) ORDER BY 3  
		 fprintf(stderr, "Not implemented yet\n");
                }
set_order_by_clause: 
		| ORDER BY set_order_by_list
		;

set_order_by_list: INTEGER
		| set_order_by_list COMMA INTEGER
		;
******/

/* utilities create and drop indices, relations (tables), and databases */
/* also to create new types, such as alias, enum and struct */
utility:
 	     XXX_TOK_CREATE unique indx_type XXX_TOK_INDEX indx_name XXX_TOK_ON relation_name
		  XXX_TOK_L_PAREN column_expr XXX_TOK_R_PAREN clustered
                 {
		  if (DataEng->CreateNewIndex($7, $5, $9, $2, $11, $3) != XXX_OK) {
		    REL_NOTE_ERROR(REL_BADINDEX, $5);
		    YYERROR;
		   }
		  if ($3) delete [] $3;
		  if ($5) delete [] $5;
		  if ($7) delete [] $7;
                 }
		| XXX_TOK_DROP XXX_TOK_INDEX indx_name
                 {
                  if (DataEng->DeleteIndex($3) != XXX_OK) {
		    REL_NOTE_ERROR(REL_BADINDEX, $3);
		    YYERROR;
		  }
                 }
		| XXX_TOK_CREATE XXX_TOK_TABLE relation_name 
                 {  
		  XXX_THISRELENV()->NestedTableName.PushStack($3);
		 }
                  XXX_TOK_L_PAREN col_data_type_list XXX_TOK_R_PAREN
			opt_key opt_src opt_dev opt_type
                 {
		  RelHandle* TmpHandle;
		  if (DataEng->CreateNewRelation(TmpHandle,$3,(RelSrcTypes)$9,
						 $10, $11, $6, $8, 
						 XXX_TRUE) != XXX_OK) {
		    REL_NOTE_ERROR(REL_BADTABLE, $3);
		    if ($6) delete $6;
		    if ($8) delete $8; 
		    YYERROR;
		   }
                  // TODO:L1 nobody reclaims $6 and $8
		  if ($3) delete [] $3;
		  if ($10) delete [] ($10);
		  if ($11) delete [] ($11);
                  // make sure to pop the stack
		  ASSERT((!XXX_THISRELENV()->NestedTableName.IsEmpty()));
		  char *TmpVal;
		  XXX_THISRELENV()->NestedTableName.PopStack(TmpVal);
                 }
		| XXX_TOK_LOAD XXX_TOK_TABLE relation_name XXX_TOK_FROM file_name
                 {
		  if (DataEng->LoadRelation($3, $5) != XXX_OK) {
		    REL_NOTE_ERROR(REL_BADTABLE, $3);
		    YYERROR;
		   }
		  if ($3) delete [] $3;
		  if ($5) delete [] $5;
                 }
		| XXX_TOK_DROP XXX_TOK_TABLE relation_name
                 {
		  if (DataEng->DeleteRelation($3) != XXX_OK) {
		    REL_NOTE_ERROR(REL_BADTABLE, $3);
		    YYERROR;
		  }
                  if ($3) delete [] $3;
                 }
                | XXX_TOK_DROP XXX_TOK_VIEW XXX_TOK_STRING
                 {
		   if (QueryEng->DropView($3) != XXX_OK) {
		     REL_NOTE_ERROR(REL_DUPVIEW, $3);
		     if ($3) delete [] $3;
		     YYERROR;
		   }
		   if ($3) delete [] $3;
		 }
		| XXX_TOK_RUNSTATS opt_relation_name
                 {
		  if (DataEng->RunStats($2) != XXX_OK) {
		    REL_NOTE_ERROR(REL_BADTABLE, ($2?$2:""));
		    YYERROR;
		  }
                  if ($2) delete [] $2;
                 }
		| XXX_TOK_DUMPTABLE opt_relation_name
                 {
		   /****
		  if (DataEng->DumpTable($2) != XXX_OK) {
		    REL_NOTE_ERROR(REL_BADTABLE, ($2?$2:""));
		    YYERROR;
		  }
		  ****/
                  if ($2) delete [] $2;
                 }
		| XXX_TOK_CREATE XXX_TOK_DATABASE db_name
                 {
		  fprintf(stderr, "Not implemented yet!");
                 }
		| XXX_TOK_DROP XXX_TOK_DATABASE db_name
                 {
		  fprintf(stderr, "Not implemented yet!");
                 }
		| XXX_TOK_OPEN XXX_TOK_DATABASE db_name
                 {
		  fprintf(stderr, "Not implemented yet!");
                 }
		| XXX_TOK_CLOSE XXX_TOK_DATABASE db_name
                 {
		  fprintf(stderr, "Not implemented yet!");
                 }
                | 
                XXX_TOK_STRING opt_args
                 {
		  $2->Name = $1; // important to assign name
		  CommandEng->ProcessCommand($2);
		  if ($2) delete $2;
                 }
		| XXX_TOK_CREATE XXX_TOK_TYPE alias_type_name XXX_TOK_ALIAS alias_to_type
		 {
		   XxxUseMutex *AutoMutex=new XxxUseMutex(*XxxADTClass::Latch);
		   if (XxxADTClass::CheckName(XxxADTClass::NumADTs, $3) == XXX_OK) {
		     XxxErrCode status;
		     new XxxAliasADT(XxxADTClass::NumADTs, $3, $5, NULL, status);
		     if(status != XXX_OK) {
		       YYERROR;
		     }
		     XxxADTClass::ADTArray[XxxADTClass::NumADTs]->Register();
		     XxxADTClass::NumADTs++;
		   }
		   if ($3) delete [] $3; if ($5) delete [] $5;
		   delete AutoMutex;
		 }
		| XXX_TOK_CREATE XXX_TOK_TYPE alias_type_name XXX_TOK_ALIAS alias_to_type XXX_TOK_L_PAREN XXX_TOK_QUOTESTRING XXX_TOK_R_PAREN
		 {
		   XxxUseMutex *AutoMutex=new XxxUseMutex(*XxxADTClass::Latch);
		   if (XxxADTClass::CheckName(XxxADTClass::NumADTs, $3) == XXX_OK) {
		       XxxErrCode status;
		       new XxxAliasADT(XxxADTClass::NumADTs, $3, $5, $7, status);
		       if (status != XXX_OK) {
			 YYERROR;
		       }
		       XxxADTClass::ADTArray[XxxADTClass::NumADTs]->Register();
		       XxxADTClass::NumADTs++;
		     }
		   if ($3) delete [] $3; if ($5) delete [] $5;
		   delete AutoMutex;
		 }
		 | XXX_TOK_CREATE XXX_TOK_TYPE enum_type_name XXX_TOK_ENUM XXX_TOK_L_PAREN col_list XXX_TOK_R_PAREN
		    {
		      XxxUseMutex *AutoMutex = new XxxUseMutex(*XxxADTClass::Latch);
		      if (XxxADTClass::CheckName(XxxADTClass::NumADTs, $3) == XXX_OK){
			int numEnum;
			char** enumNames;
			int serialSize;
			  
			if (XxxEnumADT::GetEnumList($6, numEnum, enumNames, serialSize) != XXX_OK) {
			  XXX_NOTE_ERROR(XXX_BADTYPE, "Cannot get enum list");
			  YYERROR;
			}
			  
			new XxxEnumADT(XxxADTClass::NumADTs, $3, numEnum, enumNames, serialSize);
			XxxADTClass::ADTArray[XxxADTClass::NumADTs]->Register();
			XxxADTClass::NumADTs++;
			if ($3) delete [] $3;
			if ($6) {
			  RelDataSrc::KeyList *tmp1 = $6, *tmp2;
			  while(tmp1->Next!=NULL){
			    tmp2 = tmp1->Next;
			    delete tmp1;
			    tmp1 = tmp2;
			  }
			  delete tmp1;
			}
		      }
		      delete AutoMutex;
		    }
	/**	 | XXX_TOK_CREATE XXX_TOK_TYPE struct_type_name XXX_TOK_STRUCT XXX_TOK_L_PAREN struct_col_data_type_list XXX_TOK_R_PAREN
		   {
		   XxxUseMutex *AutoMutex=new XxxUseMutex(*XxxADTClass::Latch);
		   new XxxStructADT(XxxADTClass::NumADTs++, $3, $6);
		   if ($3) delete [] $3;
		   if ($6) delete $6;
		   delete AutoMutex;
		   }
			**/

;
input_schema: XXX_TOK_L_PAREN col_data_type_list XXX_TOK_R_PAREN opt_key opt_src opt_dev opt_type
                 {
		  RelEngCatParse* CatInfo = (RelEngCatParse *)(PDS.InArg);
		  CatInfo->Attrs = $2;
		  CatInfo->Keys = $4;
		  *(CatInfo->SrcType) = (int)($5);
		  CatInfo->DevName = $6;
		  CatInfo->ImplType = $7;
                 }
;
indx_name:	  XXX_TOK_STRING { $$ = $1;}
;
indx_type:        XXX_TOK_STRING { $$ = $1;}
              | {  $$ = new char[2]; strcpy($$, "B") ;}
;
idx_attr:         XXX_TOK_INTEGER { $$ = $1;}
;
db_name:	  XXX_TOK_STRING { $$ = $1;}
;
alias_type_name:	XXX_TOK_STRING { $$ = $1; }
;
alias_to_type:		XXX_TOK_STRING { $$ = $1; }
;
enum_type_name:		XXX_TOK_STRING { $$ = $1; }
;
struct_type_name:	XXX_TOK_STRING { $$ = $1; }
;
file_name:	  any_string { $$ = $1;}
;
opt_key:     XXX_TOK_KEY XXX_TOK_L_PAREN col_list XXX_TOK_R_PAREN
              { $$ = $3; }
            | { $$ = NULL;}
;
opt_src:     XXX_TOK_SOURCE XXX_TOK_STRING 
              {
	       $$ = DataEng->FindSrcType($2);
	       if ($2) delete [] $2;
	      }
            |
              { 
	       $$ = DataEng->DefaultSrcType();
	      }
;
opt_dev:     XXX_TOK_DEVICE XXX_TOK_STRING { $$ = $2; }
            | { $$ = NULL; }
;
opt_type:    XXX_TOK_IMPLTYPE XXX_TOK_STRING  { $$ = $2; }
            | { $$ = NULL; }
;
col_data_type_list: XXX_TOK_STRING check_full_data_type null_info XXX_TOK_COMMA col_data_type_list
                   {
 		    XxxADTClass::MetaInfoHandle MIHandle;
		    if ($2->MetaInfo) {
		      // construct composite name
		      char CompositeName[XXX_ATTR_NAME_MAX+REL_RELNAME_MAX];
		      ASSERT((!XXX_THISRELENV()->NestedTableName.IsEmpty()));
		      sprintf(CompositeName, "%s.%s", 
			      XXX_THISRELENV()->NestedTableName.TopStack(), 
			      $1);
		      if ($2->Type == XXX_UNKNOWN ||
			  XxxADTArray[$2->Type]->AddMetaInfo(CompositeName, 
						       XXX_FALSE, 
						       $2->MetaInfo,
						       MIHandle) != XXX_OK)
		       {
			XXX_NOTE_ERROR(XXX_BADTYPE, $1);
			YYERROR;
		       }
		      if ($2->MetaInfo) delete [] $2->MetaInfo; 
		     } else if(XxxADTArray[$2->Type]->IsAlias()&&XxxADTArray[$2->Type]->MaintainsMetaInfo() ){
				$2->MetaInfo = new char[strlen(((XxxAliasADT *)XxxADTArray[$2->Type])->MetaString) + 1];
				strcpy($2->MetaInfo, ((XxxAliasADT *)XxxADTArray[$2->Type])->MetaString);
			    if (XxxADTArray[$2->Type]->AddMetaInfo("dummy", XXX_FALSE, $2->MetaInfo, MIHandle) != XXX_OK) {
					XXX_NOTE_ERROR(XXX_BADTYPE, $1);
					YYERROR;
		        }
			 }
		    if (!($5->Contains($1))) {
		      $$ = $5->Add($1, $2->Type, $3, MIHandle);
		      if ($1) delete [] $1;
		      if ($2) delete $2;
		    } else {
		      char buf[200];
		      sprintf (buf, "%s appears twice in the schema", $1);
		      if ($1) delete [] $1;
		      if ($2) delete $2;
		      REL_NOTE_ERROR(REL_BADQUERY, buf);
		      YYERROR;
		    }
		   }
	/*	|   XXX_TOK_STRING struct_data_type null_info XXX_TOK_COMMA col_data_type_list
		{
		    XxxADTClass::MetaInfoHandle MIHandle;
			XxxErrCode status;

		    if ($2->AttrsList) {
				XxxStructADT::StructMetaInfo *structMetaInfo = new XxxStructADT::StructMetaInfo($2->AttrsList, MIHandle, status);
				if(status != XXX_OK){
					YYERROR;
				}
			} else {
				XXX_NOTE_ERROR(XXX_BADTYPE, "struct ADT must have meta string with it");
				YYERROR;
			}		
			
			$$ = new RelDataSrc::AttrTypeList($1, XxxStructADT::StructTypeId, $3, MIHandle);
		    if ($1) delete [] $1;
                    if ($2) delete $2;
		    if (!($5->Contains($1))) {
		      $$ = $5->Add($1, $2->Type, $3, MIHandle);
		      if ($1) delete [] $1;
		      if ($2) delete $2;
		    } else {
		      char buf[200];
		      sprintf (buf, "%s appears twice in the schema", $1);
		      if ($1) delete [] $1;
		      if ($2) delete $2;
		      REL_NOTE_ERROR(REL_BADQUERY, buf);
		      YYERROR;
		    }

		}
		|   XXX_TOK_STRING struct_data_type null_info 
		{
		    XxxADTClass::MetaInfoHandle MIHandle;
			XxxErrCode status;
		    if ($2->AttrsList) {
				XxxStructADT::StructMetaInfo *structMetaInfo = new XxxStructADT::StructMetaInfo($2->AttrsList, MIHandle, status);
				if(status != XXX_OK){
					YYERROR;
				}
			} else {
				XXX_NOTE_ERROR(XXX_BADTYPE, "struct ADT must have meta string with it");
				YYERROR;
			}
			$$ = new RelDataSrc::AttrTypeList($1, XxxStructADT::StructTypeId, $3, MIHandle);
		    if ($1) delete [] $1;
                    if ($2) delete $2;

		}
	*/	|   XXX_TOK_STRING check_full_data_type null_info 
                   {
 		    XxxADTClass::MetaInfoHandle MIHandle;
		    if ($2->MetaInfo) {
		      // construct composite name
		      char CompositeName[XXX_ATTR_NAME_MAX+REL_RELNAME_MAX];
		      ASSERT((!XXX_THISRELENV()->NestedTableName.IsEmpty()));
		      sprintf(CompositeName, "%s.%s", 
			      XXX_THISRELENV()->NestedTableName.TopStack(), 
			      $1);
		      if ($2->Type == XXX_UNKNOWN ||
			  XxxADTArray[$2->Type]->AddMetaInfo(CompositeName, 
						       XXX_FALSE, 
						       $2->MetaInfo,
						       MIHandle) != XXX_OK) {
			XXX_NOTE_ERROR(XXX_BADTYPE, $1);
			YYERROR;
		       }
		      if ($2->MetaInfo) delete [] $2->MetaInfo;
		     } 	else if(XxxADTArray[$2->Type]->IsAlias()&&XxxADTArray[$2->Type]->MaintainsMetaInfo() ){
				$2->MetaInfo = new char[strlen(((XxxAliasADT *)XxxADTArray[$2->Type])->MetaString) + 1];
				strcpy($2->MetaInfo, ((XxxAliasADT *)XxxADTArray[$2->Type])->MetaString);
			    if (XxxADTArray[$2->Type]->AddMetaInfo("dummy", XXX_FALSE, $2->MetaInfo, MIHandle) != XXX_OK) {
					XXX_NOTE_ERROR(XXX_BADTYPE, $1);
					YYERROR;
		        }

			 }
			$$ = new RelDataSrc::AttrTypeList($1, $2->Type, $3, MIHandle);
		    if ($1) delete [] $1;
                    if ($2) delete $2;
		   }
		;
		
//used for struct data type, very similar to col_data_type_list
struct_col_data_type_list: XXX_TOK_STRING check_full_data_type null_info XXX_TOK_COMMA struct_col_data_type_list
                   {
 		    XxxADTClass::MetaInfoHandle MIHandle;
		    if ($2->MetaInfo) {
		      // construct composite name
		      char CompositeName[XXX_ATTR_NAME_MAX+REL_RELNAME_MAX];
		      sprintf(CompositeName, "%s", $1);
		      if ($2->Type == XXX_UNKNOWN ||
			  XxxADTArray[$2->Type]->AddMetaInfo(CompositeName, 
						       XXX_FALSE, 
						       $2->MetaInfo,
						       MIHandle) != XXX_OK)
		       {
			XXX_NOTE_ERROR(XXX_BADTYPE, $1);
			YYERROR;
		       }
		      if ($2->MetaInfo) delete [] $2->MetaInfo; 
		     } else if(XxxADTArray[$2->Type]->IsAlias()&&XxxADTArray[$2->Type]->MaintainsMetaInfo() ){
				$2->MetaInfo = new char[strlen(((XxxAliasADT *)XxxADTArray[$2->Type])->MetaString) + 1];
				strcpy($2->MetaInfo, ((XxxAliasADT *)XxxADTArray[$2->Type])->MetaString);
			    if (XxxADTArray[$2->Type]->AddMetaInfo("dummy", XXX_FALSE, $2->MetaInfo, MIHandle) != XXX_OK) {
					XXX_NOTE_ERROR(XXX_BADTYPE, $1);
					YYERROR;
		        }
			 }
		    if (!($5->Contains($1))) {
		      $$ = $5->Add($1, $2->Type, $3, MIHandle);
		      if ($1) delete [] $1;
		      if ($2) delete $2;
		    } else {
		      char buf[200];
		      sprintf (buf, "%s appears twice in the schema", $1);
		      if ($1) delete [] $1;
		      if ($2) delete $2;
		      REL_NOTE_ERROR(REL_BADQUERY, buf);
		      YYERROR;
		    }
		   }
		|   XXX_TOK_STRING check_full_data_type null_info 
                   {
 		    XxxADTClass::MetaInfoHandle MIHandle;
		    if ($2->MetaInfo) {
		      // construct composite name
		      char CompositeName[XXX_ATTR_NAME_MAX+REL_RELNAME_MAX];
		      sprintf(CompositeName, "%s", $1);
		      if ($2->Type == XXX_UNKNOWN ||
			  XxxADTArray[$2->Type]->AddMetaInfo(CompositeName, 
						       XXX_FALSE, 
						       $2->MetaInfo,
						       MIHandle) != XXX_OK) {
			XXX_NOTE_ERROR(XXX_BADTYPE, $1);
			YYERROR;
		       }
		      if ($2->MetaInfo) delete [] $2->MetaInfo;
		     } else if(XxxADTArray[$2->Type]->IsAlias()&&XxxADTArray[$2->Type]->MaintainsMetaInfo() ){
				$2->MetaInfo = new char[strlen(((XxxAliasADT *)XxxADTArray[$2->Type])->MetaString) + 1];
				strcpy($2->MetaInfo, ((XxxAliasADT *)XxxADTArray[$2->Type])->MetaString);
			    if (XxxADTArray[$2->Type]->AddMetaInfo("dummy", XXX_FALSE, $2->MetaInfo, MIHandle) != XXX_OK) {
					XXX_NOTE_ERROR(XXX_BADTYPE, $1);
					YYERROR;
		        }
			 }

		    $$ = new XxxGenDataSrc::AttrTypeList($1, $2->Type, $3, MIHandle);
		    if ($1) delete [] $1;
                    if ($2) delete $2;
		   }
		;
		
null_info:  XXX_TOK_NOT XXX_TOK_NULL  { $$ = XXX_TRUE;}
         |  { $$= XXX_FALSE;}
;
check_full_data_type: full_data_type 
            {
	     if ((!$1->MetaInfo) && XxxADTArray[$1->Type]->SpecifiedMetaInfo()) {
	       // if there is no meta-info, check that this is ok
	       char buf[100];
	       sprintf(buf, "Need meta-info for type %s", 
		       XxxADTArray[$1->Type]->TypeName);
	       XXX_NOTE_ERROR(XXX_BADTYPE, buf);
               if ($1) delete $1;
	       YYERROR;
	      }
              $$ = $1;
 	    }
full_data_type: data_type opt_meta_info
                {
	         $$ = new struct FullType();
	         $$->Type = $1;
	         $$->MetaInfo = $2;
 	        }
;
/**
 * Note: XXX_FIXEDSTRING does not work, so using XXX_STRING.
 */
data_type: XXX_TOK_CHAR { $$ = XXX_STRING; }
         | XXX_TOK_CHAR XXX_TOK_VARYING { $$ = XXX_STRING; }
         | XXX_TOK_VARCHAR { $$ = XXX_STRING; }
         | XXX_TOK_STRING 
           {
             $$ = GetTypeId($1);
             if ($$ == XXX_UNKNOWN) {
               XXX_NOTE_ERROR(XXX_BADTYPE, $1);
               YYERROR;
             }
	     if ($1) delete [] $1;
	   }
;
/*
struct_data_type: XXX_TOK_STRUCT XXX_TOK_L_PAREN struct_col_data_type_list XXX_TOK_R_PAREN
		{
	     $$ = new struct StructType();
	     $$->Type = XxxStructADT::StructTypeId;
	     $$->AttrsList = $3;
		}
;
*/
opt_meta_info : 
           XXX_TOK_L_PAREN XXX_TOK_QUOTESTRING XXX_TOK_R_PAREN
           { $$ = $2; }
         | XXX_TOK_L_BRACE XXX_TOK_QUOTESTRING XXX_TOK_R_BRACE
           { $$ = $2; }
         | XXX_TOK_L_PAREN XXX_TOK_INTEGER XXX_TOK_R_PAREN
           { // need this for char(FIXED_SIZE)
             $$ = new char[10];
	     sprintf($$, "%d", $2);
           }
         | XXX_TOK_L_BRACE XXX_TOK_INTEGER XXX_TOK_R_BRACE
           { // need this for char[FIXED_SIZE]
             $$ = new char[10];
	     sprintf($$, "%d", $2);
           }
         | { $$ = NULL; }
;

col_list:      col_list XXX_TOK_COMMA XXX_TOK_STRING 
            {
	     $$ = $1->Add($3);
	     if ($3) delete [] $3;
            }
         |     XXX_TOK_STRING
            {
	     $$ = new RelDataSrc::KeyList($1);
	     if ($1) delete [] $1;
            }
;

/* the update stmt can delete/insert tuples from/into relations and change values of attributes */
update:		  delete_stmt
		| insert_stmt
 		| update_stmt
		;

delete_stmt:	  XXX_TOK_DELETE XXX_TOK_FROM relation_name where_clause
                 {
                  if (DataEng->DeleteTuples($3, $4) != XXX_OK) {
		    REL_NOTE_ERROR(REL_BADTABLE, $3);
		    YYERROR;
		  }
		  if ($3) delete [] $3;
                  if ($4) delete $4;
                 }
		;

insert_stmt:	  XXX_TOK_INSERT XXX_TOK_INTO relation_name XXX_TOK_VALUES XXX_TOK_L_PAREN value_list XXX_TOK_R_PAREN
                 {
		  if (DataEng->InsertTuples($3, $6) != XXX_OK) {
		    REL_NOTE_ERROR(REL_BADTABLE, $3);
		    YYERROR;
		   }
		  if ($3) delete [] $3;
		  if ($6) delete $6;
                 }
		| XXX_TOK_INSERT XXX_TOK_INTO relation_name query_block 
                 {
		  // specify that answers should be printed
		  RelSpecParameters Params($3, 
					   XXX_TRUE, // named
					   XXX_TRUE, // create
					   XXX_TRUE); // append
		  if (execQuery($4, Params) != XXX_OK) {
		    REL_NOTE_ERROR(REL_BADQUERY, "Error executing query");
		    YYERROR;
		   }
		  if ($3) delete [] $3;
		  if ($4) delete $4;
                 }
		;

value_list:	 value_list XXX_TOK_COMMA value_expr 
                 {
		  $$ = $1->AddExpr($3);
		 }
                |  value_expr { $$ = new XxxValueExprList($1); }
		;

value_expr:	column_expr 
                 { 
		  if ($1->IsConstant())
		    $$ = $1;
		  else { 
		    if ($1) delete $1;
		    REL_NOTE_ERROR(REL_BADQUERY, 
				   "Should have only constants in value list");
		    YYERROR;
		   }
		 }

update_stmt:	  XXX_TOK_UPDATE relation_name XXX_TOK_SET update_list where_clause	
                 {
		  if (DataEng->UpdateTuples($2, $5, $4) != XXX_OK) {
		    REL_NOTE_ERROR(REL_BADTABLE, $2);
		    YYERROR;
		   }
		  if ($2) delete [] $2;
		  if ($4) delete $4;
		  if ($5) delete $5;
                 }
		;

update_list:	update_list XXX_TOK_COMMA XXX_TOK_STRING XXX_TOK_EQ column_expr 
                {
		  $$ = $1->Add($3, $5);
		  if ($3) delete [] $3;
		}
		| XXX_TOK_STRING XXX_TOK_EQ column_expr 
                {
		  $$ = new RelDataEngine::UpdateItemList($1, $3);
		  if ($1) delete [] $1;
		}
		;

opt_args : opt_args const
           { if ($1->ArgCount < 4)
	      $$ = $1->AddArg($2);
  	     else {
	       REL_NOTE_ERROR(REL_BADQUERY, 
			      "Too many arguments for command");
	       YYERROR;
	     }
	   }
         | { $$ = new XxxRelCommandInfo(); }
;

%%

