// Ihab
// The Main KRJN algorith encapsulated in the KRJN planoperator (physical)
#ifdef __GNUG__
#pragma implementation
#endif

#include "relplanop.h"
#include "relops.h"
#include "reldata.h"
#include "releng.h"
#include "planKRJ.h"
//------------------------------------------------------
double ComputeScore  (double  LeftVal, double  RightVal)
{ 
  double score = RightVal+LeftVal;
  return score; 
}

//------------------------------------------------------

// Queue Methods

XxxBool RelKRJPlanOp::Queue::isEqual(XxxRecord *Rec, int AttrIndex,
				   XxxRecord *QRec, int QAttrIndex, 
				   XxxRecordSchema *Schema, XxxRecordSchema* QSchema) const
{
  XxxBool result;
  ASSERT(Rec && QRec);
  ASSERT((*Schema)[AttrIndex].Type ==  (*QSchema)[QAttrIndex].Type);
  XxxBasicType TypeId = (*Schema)[AttrIndex].Type;
  XxxADTMetaInfo* MInfo = (*Schema)[AttrIndex].TypeCatInfo;
  XxxADTMetaInfo* QMInfo = (*QSchema)[QAttrIndex].TypeCatInfo;
  const char *Field, *QField;
  Rec->GetField(AttrIndex, Schema, Field);
  int oid1 = ((XxxCoreADTClass::IntValue*)Field)->Value ;
  QRec->GetField(QAttrIndex, QSchema, QField);
  int oid2 = ((XxxCoreADTClass::IntValue*)QField)->Value ;
  if (XxxCoreADTClass::IsCoreADT(TypeId)) {
    result = XxxCoreADTClass::SEquals(TypeId, (XxxADTValue*)Field, MInfo, (XxxADTValue*)QField, QMInfo);
  }
  else {
    result = XxxADTArray[TypeId]->Equals((XxxADTValue*)Field, MInfo, (XxxADTValue*)QField, QMInfo);
  }
  return result;
}
//-----------------------------------
RelKRJPlanOp::Queue::Queue(){
 Front = Rear = NULL;
 Size = 0;
 MaxB = 0.0;
}
//------------------------------
RelKRJPlanOp::Queue::~Queue(){
 RelKRJPlanOp::Tuple *q, *p, *s;
 p= Front;
  while(p!=NULL ) {
    q= p->Prev;
    delete p->LeftRec;
    delete p->RightRec;
    delete p;
    p=q;
  }
}

//-------------------------------
XxxBool RelKRJPlanOp::Queue::IsEmpty(){
  return Rear==NULL;
}
//------------------------------
void RelKRJPlanOp::Queue::AddTuple(RelKRJPlanOp::Tuple *t,int k){
  XxxRecord * leftRec = NULL,*rightRec=NULL;
  if(t->LeftRec) leftRec= new XxxRecord(*t->LeftRec);
  if(t->RightRec) rightRec= new XxxRecord(*t->RightRec);
  RelKRJPlanOp::Tuple * tuple = new RelKRJPlanOp::Tuple(leftRec,rightRec, t->Grade, t->WG,t->BG,t->Source);
 
  RelKRJPlanOp::Tuple *q, *p, *s;
  p= Front;
  q=NULL;
  
  if(p==NULL){ //empty
    Rear=Front=tuple;
    Size++;
    return;
  }
  int i=0;
  while(p!=NULL && p->WG >=t->WG) {
    if( t->WG == p->WG)
      if (t->BG > p->BG ) break;
    q=p ; 
    p = p->Prev;
    i++;
  }
  
  if(q==NULL)    Front=tuple;
  else           q->Prev=tuple;
  
  tuple->Next=q;
  tuple->Prev=p;
  
  if(p==NULL)    Rear=tuple;
  else p->Next=tuple;

  // update maxB for elements not in top k
  int j;
  s=tuple;
  if(i<=k){ // element inserted in top k
    for(j=i;j<=k && s!=NULL;j++) s=s->Prev;
    if(j>k && s) if(MaxB < s->BG) MaxB = s->BG;
  }else
    if(MaxB < t->BG) MaxB = t->BG;
  
  
  Size++;
  return;
}
//------------------------------
void RelKRJPlanOp::Queue::updateMaxB (int k){
  RelKRJPlanOp::Tuple *p;
  p=Front;
  int j;

  MaxB=0;
  for(j=0;j<=k && p!=NULL;j++) p=p->Prev;
  if(j>k) {
    while(p!=NULL){
      if(MaxB < p->BG) MaxB = p->BG;
      p=p->Prev;
    }
  }
}
//---------------------------------------
void  RelKRJPlanOp::Queue::updateRanks( double LeftTh, double RightTh){
 RelKRJPlanOp::Tuple *p;
 p=Front;
 MaxB=0;
 
 if (p!=NULL) p=p->Prev;  // advance one 
 
 while(p!=NULL){	
   if(p->Source==1){
     p->BG = ComputeScore(p->Grade,RightTh);
   }
   else if (p->Source==2){
     p->BG = ComputeScore(p->Grade,LeftTh);
   }
   if(MaxB < p->BG) MaxB = p->BG;
   p=p->Prev;
 }
}
//-----------------------------
RelKRJPlanOp::Tuple* RelKRJPlanOp::Queue::GetNext (){
  RelKRJPlanOp::Tuple* result;
  
  if (IsEmpty()) return NULL;
  Size--;
  result = Front;
  RelKRJPlanOp::Tuple * p = Front;
  Front = Front->Prev;
  if(Front) Front->Next = NULL;
  else Rear = NULL;
  return result;
}
//------------------------------
RelKRJPlanOp::Tuple* RelKRJPlanOp::Queue::GetTuple(XxxRecord* Rec,
						 XxxRecordSchema *Schema, 
						 int RecSource,
						 int AttrIndex, 
						 XxxDerivedRelation::CursorInfo* Handle,
						 XxxBooleanExpressionPlan** ResidPredPlans,
						 int remove){
  RelKRJPlanOp::Tuple *q, *p , *s;  
  RelKRJPlanOp::Tuple* result=NULL;
  XxxBool Basic_Ok = XXX_FALSE;
  
  p=Rear;
  q=NULL;  
  if(p==NULL)  return NULL;
  while(p!=NULL ) {
    if( RecSource==2 && p->Source==1 ){
      if(isEqual(Rec,AttrIndex, p->LeftRec, LeftAttrIndex, Schema, LeftSchema))
	{
	  Basic_Ok = XXX_TRUE;
	  Handle->ChildRecs[RightGradeRel] = Rec;
	  Handle->ChildRecs[LeftGradeRel]  = p->LeftRec; 
	}
      else Basic_Ok = XXX_FALSE;
    }
      else 
      if( RecSource==1 && p->Source==2 ){
	if( isEqual(Rec,AttrIndex, p->RightRec, RightAttrIndex, Schema, RightSchema)){	  
	  Basic_Ok = XXX_TRUE;
	  Handle->ChildRecs[RightGradeRel] = p->RightRec;
	  Handle->ChildRecs[LeftGradeRel]  = Rec;
	}     
	else Basic_Ok = XXX_FALSE;
      }
      
    if (Basic_Ok){  
      XxxCoreADTClass::BoolValue Satisfied(XXX_TRUE);
      if (ResidPredPlans[1] && (ResidPredPlans[1]->Evaluate((const XxxRecord **)Handle->ChildRecs, 
							    (const XxxRecordSchema **)Handle->ChildSchema, 
							    &Satisfied,
							    (Handle->Params) ? Handle->Params->Params : 0) != XXX_OK)) {
	XXX_NOTE_ERROR(XXX_BADEXPR, "Error evaluating residual expression");
	return NULL;
      }
      
      if(Satisfied.Value) 
	break;
      Basic_Ok = XXX_FALSE;
    }
    p = p->Next;
  }
  if(p==NULL)  return NULL;
  if(remove){
    s= p->Next;
    q= p->Prev;
    if(q==NULL)  Rear = p->Next;
    else q->Next = p->Next;
    
    if(s==NULL) Front = p->Prev;
    else s->Prev = q;
    
    result= p;
    Size--;
  }
  return result;
  
}
//----------------------------
RelKRJPlanOp::Tuple* RelKRJPlanOp::Queue::ElementAt(int k){
  return Front; //.. for now k=0
}

//-----------------------------
RelKRJPlanOp::KRJJoinInfo::KRJJoinInfo(int RightRelIndex, int LeftRelIndex,
				    int RightAttrIndex, int LeftAttrIndex,
				    int LeftWGradeIndex, int RightWGradeIndex,
				    int LeftBGradeIndex, int RightBGradeIndex,
				    int LeftGradeRel, int RightGradeRel):
  Valid(XXX_TRUE), LeftRelIndex(LeftRelIndex), RightRelIndex(RightRelIndex),
  LeftAttrIndex(LeftAttrIndex), RightAttrIndex(RightAttrIndex), 
  LeftGradeRel(LeftGradeRel), RightGradeRel(RightGradeRel),
  LeftWGradeIndex(LeftWGradeIndex), RightWGradeIndex(RightWGradeIndex),
  LeftBGradeIndex(LeftBGradeIndex), RightBGradeIndex(RightBGradeIndex)
 { 
   Top = 0; 
   Valid = XXX_TRUE;
   eof = XXX_FALSE;
   depth = 0;
   K = -1;
   LeftTh = RightTh = 0.0;
   Buffer = new RelKRJPlanOp::Queue();
} 

RelKRJPlanOp::KRJJoinInfo::~KRJJoinInfo(){
  delete Buffer;
}


//-------------------------------
//
// Methods of RelKRJPlanOp



RelKRJPlanOp::RelKRJPlanOp(int NumChildren, 
			 RelPlanOp* Children[],
			 RelCatalogInfo* CatInfo,
			 XxxValExprPlanList* ExprPlanList,
			 XxxValExprPlanList* OrderPlanList,
			 XxxBooleanExpressionPlan** SelPredPlans,
			 XxxBooleanExpressionPlan** JoinPredPlans,
			 XxxBooleanExpressionPlan** ResidPredPlans,
			 XxxBooleanExpressionPlan* InitPlan,
			 int top)
  : RelSPJPlanOp(NumChildren, Children, CatInfo, ExprPlanList, 
		 SelPredPlans, JoinPredPlans, ResidPredPlans, InitPlan),
    OrderPlanList(OrderPlanList), Top(top) 
{
  
   ASSERT((NumChildren == 2));

}
//------------------------------------
XxxErrCode RelKRJPlanOp::getNextItem(XxxRelation* *RelationArray,
				    XxxDerivedRelation::CursorInfo* Handle,
				    int Index, XxxBool& Valid) const 
{
  if (RelationArray[Index]->
      NextItem(Handle->ChildCursorArray[Index], 
	       Handle->ChildRecs[Index]) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADEXEC, "Error getting NextItem from relation");
    return XXX_NOT_OK;
  }
  
   Valid = Handle->ChildCursorArray[Index].IsValid();
  return XXX_OK;
}
//-----------------------------------------------------
XxxErrCode 
RelKRJPlanOp::parseJoinPlan(XxxDerivedRelation::CursorInfo* Handle,
			   int &LeftAttrIndex, int &LeftRelIndex,
			   int &RightAttrIndex, int &RightRelIndex,
			   int &LeftWGradeIndex, int &RightWGradeIndex,
			   int &LeftBGradeIndex, int &RightBGradeIndex,
			   int &LeftGradeRel,  int &RightGradeRel)const
{
  XxxBooleanExpressionPlan* JoinPlan = JoinPredPlans[1]; 
  XxxValExprPlanList *OList = OrderPlanList;

  ASSERT (JoinPlan);
  XxxRelBooleanPlan* RJoinPlan = (XxxRelBooleanPlan *)JoinPlan;
  ASSERT((JoinPlan->BoolType == XXX_REL_BOOL_EXPR) && (RJoinPlan->OperType == XXX_EQ_OP));
  XxxValueExprPlan *LeftPlan = RJoinPlan->SubPlans[0];
  XxxValueExprPlan *RightPlan = RJoinPlan->SubPlans[1];
  ASSERT ((RightPlan->ValType() == 
	   XxxValueExpression::COLUMN_VAR)&&(LeftPlan->ValType() == XxxValueExpression::COLUMN_VAR));
  ASSERT (RightPlan->BasicType() == LeftPlan->BasicType());
  // Extract the indices in the RelationArray of the inputs
  LeftRelIndex = ((XxxUnknownValue::UnknownValuePlan *) (LeftPlan))->SrcChild;
  RightRelIndex = ((XxxUnknownValue::UnknownValuePlan *)(RightPlan))->SrcChild;
  LeftGradeRel = LeftRelIndex;
  RightGradeRel = RightRelIndex;
  

  // Extract the indices of the attributes in the records
  
  LeftAttrIndex = ((XxxUnknownValue::UnknownValuePlan *)(LeftPlan))->SrcAttrIndex;
  RightAttrIndex = ((XxxUnknownValue::UnknownValuePlan *)(RightPlan))->SrcAttrIndex;
  
  if(OList){
    LeftBGradeIndex = LeftWGradeIndex = ((XxxUnknownValue::UnknownValuePlan *) (OList->Plan))->SrcAttrIndex;
    OList = OList->Next;
  }
  if(OList){ //bottom plan
    RightBGradeIndex = RightWGradeIndex = ((XxxUnknownValue::UnknownValuePlan *) (OList->Plan))->SrcAttrIndex;
  }
  else {    //temporary plan
    RightWGradeIndex = Handle->ChildSchema[RightRelIndex]->Arity()-2;
    RightBGradeIndex = Handle->ChildSchema[RightRelIndex]->Arity()-1;
  }
  return XXX_OK; 
}
//----------------------------------------------------
XxxErrCode RelKRJPlanOp::createJoinInfo(XxxDerivedRelation::CursorInfo* Handle,
				       KRJJoinInfo *&JoinInfo) const
{
  // Get the locations of the join columns
  int LeftAttrIndex, RightAttrIndex, LeftRelIndex, RightRelIndex;
  int LeftWGradeIndex, RightWGradeIndex;
  int LeftBGradeIndex, RightBGradeIndex;
  int LeftGradeRel, RightGradeRel;

  if (parseJoinPlan(Handle, LeftAttrIndex, LeftRelIndex,
		    RightAttrIndex, RightRelIndex,
		    LeftWGradeIndex, RightWGradeIndex,
		    LeftBGradeIndex, RightBGradeIndex,
		    LeftGradeRel, RightGradeRel) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADEXEC,
		   "Attempt to separate join condition failed");
    return XXX_NOT_OK;
  }

  JoinInfo = new KRJJoinInfo(RightRelIndex, LeftRelIndex,
			    RightAttrIndex, LeftAttrIndex,
			    LeftWGradeIndex, RightWGradeIndex,
			    LeftBGradeIndex, RightBGradeIndex,
			    LeftGradeRel, RightGradeRel);
  CreateDummyRec( Handle->ChildSchema[LeftGradeRel] , JoinInfo->DummyLeftRec);
  CreateDummyRec( Handle->ChildSchema[RightGradeRel], JoinInfo->DummyRightRec);

  return XXX_OK;
}

//----------------------------------------------------
XxxBool RelKRJPlanOp::isEqual( KRJJoinInfo* JoinInfo,
			      XxxDerivedRelation::CursorInfo* Handle,
			      XxxBooleanExpressionPlan** ResidPredPlans) const
{
  XxxBasicType TypeId = (*Handle->ChildSchema[JoinInfo->LeftRelIndex])[JoinInfo->LeftAttrIndex].Type;
  XxxADTMetaInfo* LMInfo = (*Handle->ChildSchema[JoinInfo->LeftRelIndex])[JoinInfo->LeftAttrIndex].TypeCatInfo;
  XxxADTMetaInfo* RMInfo = (*Handle->ChildSchema[JoinInfo->RightRelIndex])[JoinInfo->RightAttrIndex].TypeCatInfo;
  const char *LField, *RField;
  Handle->ChildRecs[JoinInfo->LeftRelIndex]->GetField(JoinInfo->LeftAttrIndex, 
						      Handle->ChildSchema[JoinInfo->LeftRelIndex], 
						      LField);

  Handle->ChildRecs[JoinInfo->RightRelIndex]->GetField(JoinInfo->RightAttrIndex, 
						      Handle->ChildSchema[JoinInfo->RightRelIndex], 
						      RField);
  XxxBool result;
  if (XxxCoreADTClass::IsCoreADT(TypeId)) {
    result = XxxCoreADTClass::SEquals(TypeId, (XxxADTValue*)LField, LMInfo,(XxxADTValue*)RField, RMInfo);
  }
  else {
    result = XxxADTArray[TypeId]->Equals((XxxADTValue*)LField, LMInfo, (XxxADTValue*)RField, RMInfo);
  }
  XxxCoreADTClass::BoolValue Satisfied(XXX_TRUE);
  if (ResidPredPlans[1] && (ResidPredPlans[1]->Evaluate((const XxxRecord **)Handle->ChildRecs, 
							(const XxxRecordSchema **)Handle->ChildSchema, 
							&Satisfied,
							(Handle->Params) ? Handle->Params->Params : 0) != XXX_OK)) {
    XXX_NOTE_ERROR(XXX_BADEXPR, "Error evaluating residual expression");
    return XXX_FALSE;
  }
  
  result = result && Satisfied.Value ;
    
  return result;

}

//------------------------------------------------------
double GetGrade  (XxxRecord *R, int Index, XxxRecordSchema *S) 
{ 
  ASSERT(R);
  const char *Field;
  R->GetField(Index, S, Field);
  double score = ((XxxCoreADTClass::DblValue*)Field)->Value ;
  return score; 
}
//------------------------------------------------------
void  GetIDS  (XxxRecord *Rec, int AttrIndex,XxxRecordSchema *Schema, int & x, int & y, int & z) 
{ 
  ASSERT(Rec);
  const char *Field1, *Field2, *Field3;
  Rec->GetField(AttrIndex, Schema, Field1);
  x = ((XxxCoreADTClass::IntValue*)Field1)->Value ;
  Rec->GetField(AttrIndex+1, Schema, Field2);
  y = ((XxxCoreADTClass::IntValue*)Field2)->Value ;
  Rec->GetField(AttrIndex+2, Schema, Field3);
  z = ((XxxCoreADTClass::IntValue*)Field3)->Value ;
}
//--------------------------------------------------------
XxxErrCode RelKRJPlanOp::InitHandle(XxxDerivedRelation::CursorInfo* Handle,
			XxxRelation* *RelationArray) const
{
 if (Handle->ArrayLen != 2) {
   REL_NOTE_ERROR(REL_BADOP, "KRJ JOIN Need  two Inputs");
   return XXX_NOT_OK;
  }

 // Create the operator handle
 KRJJoinInfo* JoinInfo;
 if (createJoinInfo(Handle, JoinInfo) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, "Cannot create KRJ join info");
   return XXX_NOT_OK;
 } 
 Handle->OperatorHandle = (void *)JoinInfo;
 for (int i = 0; i < Handle->ArrayLen; i++) {
   XxxBooleanExpressionPlan* PredPlan =   (SelPredPlans?SelPredPlans[i]:0);
   if (RelationArray[i]->
       InitCursor1(Handle->ChildCursorArray[i],
		   PredPlan, 
		   NULL,
		   NULL,
		   Handle->Params, Handle->ArrayLen, 0) != XXX_OK) {
     char buf[100];
     sprintf (buf, "Unable to open cursor on input %d", i);
     XXX_NOTE_ERROR(XXX_BADCURSOR, buf);
     return XXX_NOT_OK;
   }
 }  
 
 JoinInfo->Buffer->LeftSchema= Handle->ChildSchema[JoinInfo->LeftRelIndex];
 JoinInfo->Buffer->RightSchema= Handle->ChildSchema[JoinInfo->RightRelIndex];
 JoinInfo->Buffer->LeftAttrIndex =JoinInfo->LeftAttrIndex;
 JoinInfo->Buffer->RightAttrIndex =JoinInfo->RightAttrIndex;
 JoinInfo->Buffer->LeftGradeRel=JoinInfo->LeftGradeRel;
 JoinInfo->Buffer->RightGradeRel =JoinInfo->RightGradeRel;
 

 return XXX_OK;
}
//---------------------------------------------------
XxxErrCode RelKRJPlanOp::CreateDummyRec(  const XxxRecordSchema* Schema,  
					 XxxRecord*& ReturnRec) const
{
  XxxValuesArray  ValArray(Schema);
  ReturnRec = new XxxRecord (ValArray, Schema);
  return XXX_OK;
}

//-------------------------------------------------------------------
// close all open cursors
XxxErrCode 
RelKRJPlanOp::CloseHandle(XxxDerivedRelation::CursorInfo* Handle,
			  XxxRelation* *RelationArray) const
{
  // close cursors on all the inputs
  for (int i = 0; i < Handle->ArrayLen; i++) {
    // NOTE: if this has already been closed, set the cursor state to 0!!
    if (Handle->ChildCursorArray[i].State &&
	RelationArray[i]->
	CloseCursor(Handle->ChildCursorArray[i]) != XXX_OK) {
      char buf[100];
      sprintf(buf, "Unable to close sequence cursor %d\n", i);
      XXX_NOTE_ERROR(XXX_BADCURSOR, buf);
      return XXX_NOT_OK;
    }
  }
  delete ((KRJJoinInfo *)(Handle->OperatorHandle));
  return XXX_OK;
}
//-------------------------------------------------------------------
XxxErrCode RelKRJPlanOp::createRecord(XxxDerivedRelation::CursorInfo* Handle,
				     XxxRelation* *RelationArray,
				     const XxxRecordSchema* Schema,
				     XxxRecord*& ReturnRec,
				     double WG,
				     double BG) const
{
 // a successful instantiation of the inputs has happened. now make a record
  if (ExprPlanList) {
    ASSERT((Schema->Arity() == ExprPlanList->Count())); 
    
    XXX_CALL_RET(ExprPlanList->
		 Evaluate((const XxxRecord **)Handle->ChildRecs, 
			  (const XxxRecordSchema **)Handle->ChildSchema,
			  Handle->RecValuesArray.Values, 
			  Handle->Params?Handle->Params->Params:0), 
		 XXX_BADVALUE, "Error evaluating value list");
    
    // now make a new record out of the RecValuesArray
    ReturnRec = new XxxRecord(Handle->RecValuesArray, Schema); 
    if(Top<1){
      const char *WField, *BField;
      ReturnRec->GetField(Schema->Arity()-2,Schema, WField);
      ReturnRec->GetField(Schema->Arity()-1,Schema, BField);
      ((XxxCoreADTClass::DblValue*)BField)->Value = BG;
      ((XxxCoreADTClass::DblValue*)WField)->Value = WG;
    }
    return XXX_OK;
  }
  return XXX_NOT_OK;
}
//------------------------------------------------------------------
double MaxVal(double v1, double v2){
  if(v1>=v2) return v1; else return v2;
}
//-------------------------------------------------------------------
XxxErrCode RelKRJPlanOp::GetNextRecord(XxxSetCursor& SCsr,
				      XxxRelation* *RelationArray,
				      const XxxRecordSchema* Schema,
				      XxxRecord*& ReturnRec) const
{


// do a simple pipelined RJOIN 

   XxxDerivedRelation::CursorInfo* Handle = (XxxDerivedRelation::CursorInfo*)(SCsr.State);
    if (!Handle) {
      REL_NOTE_ERROR(REL_BADEXEC, "NULL Exec handle");
      return XXX_NOT_OK;
    }
  
    KRJJoinInfo* JoinInfo = ((KRJJoinInfo *)(Handle->OperatorHandle));
    Queue* & Buffer = JoinInfo->Buffer;

 
    XxxBool LeftScanValid = XXX_TRUE;
    XxxBool RightScanValid = XXX_TRUE;
    XxxRecord *LastLeftRec  = Handle->ChildRecs[JoinInfo->LeftRelIndex];
    XxxRecord *LastRightRec = Handle->ChildRecs[JoinInfo->RightRelIndex];
    XxxBool eof = XXX_FALSE;
    Tuple * q;
  

    // check if u can satisfy the request without
    // going deeper
    q= Buffer->ElementAt(0);
    if(q){
      if(q->BG== Buffer->MaxB) Buffer->updateMaxB(JoinInfo->K);
    
      double Mkd = q->WG;
      double BRd = MaxVal(ComputeScore(JoinInfo->RightTh, JoinInfo->LeftTh),Buffer->MaxB);
      if(BRd <= Mkd) { // beware of rounding errors in case of equal...TODO
        Buffer->GetNext();
        if(q->Source==1)     
  	q->RightRec= JoinInfo->DummyRightRec;
        if(q->Source==2)     
  	q->LeftRec = q->RightRec; 
//JoinInfo->DummyLeftRec; // to be fixed .. assuming same source schemas 
      
        Handle->ChildRecs[JoinInfo->LeftRelIndex]=q->LeftRec;
        Handle->ChildRecs[JoinInfo->RightRelIndex]=q->RightRec;  
	XxxErrCode  status = createRecord(Handle, RelationArray, Schema, ReturnRec,q->WG, q->BG);
        return status; 
      }
    }
  
    if (JoinInfo->eof)
      {
        JoinInfo->Valid = XXX_FALSE;   
        SCsr.AtEnd = XXX_TRUE;
        ReturnRec = 0;
        return XXX_OK;
      }
  
    
    double LeftWGrade, LeftBGrade ,RightWGrade,RightBGrade;
    LeftWGrade = LeftBGrade = RightWGrade = RightBGrade = 0.0;

    while ((LeftScanValid == XXX_TRUE) && (LeftScanValid == XXX_TRUE))    {
    
// for relations, the consumer is reponsible for reclaiming memory
      if (reclaimCurrent(Handle, 0) != XXX_OK) {
        REL_NOTE_ERROR(REL_BADEXEC, "Error reclaiming old memory");
        return XXX_NOT_OK;
      }
//for relations, the consumer is reponsible for reclaiming memory
      if (reclaimCurrent(Handle, 1) != XXX_OK) {
        REL_NOTE_ERROR(REL_BADEXEC, "Error reclaiming old memory");
        return XXX_NOT_OK;
      }
      if (getNextItem(RelationArray, Handle,JoinInfo->LeftRelIndex, LeftScanValid) != XXX_OK) {
        REL_NOTE_ERROR(REL_BADEXEC, "Cannot get next record from input");
        return XXX_NOT_OK;
      } 
      LastLeftRec = Handle->ChildRecs[JoinInfo->LeftRelIndex];
      // Get Grades  
      
      if(LeftScanValid){
	 // LeftWGrade = LeftBGrade = 100000 - JoinInfo->depth;
	 LeftWGrade = GetGrade(LastLeftRec,
			       JoinInfo->LeftWGradeIndex, 
			       Handle->ChildSchema[JoinInfo->LeftGradeRel]);
	 LeftBGrade = GetGrade(LastLeftRec,
			       JoinInfo->LeftBGradeIndex, 
			       Handle->ChildSchema[JoinInfo->LeftGradeRel]);
	 
	 JoinInfo->LeftTh = LeftBGrade;
      }
      
      
      if (getNextItem(RelationArray, Handle, JoinInfo->RightRelIndex,RightScanValid) != XXX_OK) {
	 REL_NOTE_ERROR(REL_BADEXEC, "Cannot get next record from input");
	 return XXX_NOT_OK;
      }
      LastRightRec = Handle->ChildRecs[JoinInfo->RightRelIndex];
      //GetGrades
      if(RightScanValid){
	 //RightWGrade = RightBGrade = 100000 - JoinInfo->depth;
	 RightWGrade = GetGrade(LastRightRec,
  				JoinInfo->RightWGradeIndex, 
  				Handle->ChildSchema[JoinInfo->RightGradeRel]);
	 
  	 RightBGrade = GetGrade(LastRightRec,
  				JoinInfo->RightBGradeIndex, 
  				Handle->ChildSchema[JoinInfo->RightGradeRel]);
	 JoinInfo->RightTh = RightBGrade;
      }
      
      // for now they must end together 
      // this should be fixed
      if(!LeftScanValid || !RightScanValid){
	 JoinInfo->Valid = XXX_FALSE;   
	 SCsr.AtEnd = XXX_TRUE;
	 ReturnRec = 0;
	 return XXX_OK;
      }
      
      else
      {
	XxxBool l = XXX_TRUE;
  	XxxBool r = XXX_TRUE;
  	JoinInfo->depth ++;
	
  	if((q=Buffer->GetTuple(LastLeftRec,
  			       Handle->ChildSchema[JoinInfo->LeftRelIndex],
  			       1,
  			       JoinInfo->LeftAttrIndex,
  			       Handle,
  			       ResidPredPlans,
  			       1))){
//  	  // left rec found in buffer
//  	  //  ASSERT (q->Source == 2);
  	  l = XXX_FALSE;
  	  q->WG= ComputeScore(q->WG,LeftWGrade);
  	  q->BG= ComputeScore(q->Grade,LeftBGrade);
  	  q->LeftRec = LastLeftRec;
  	  q->Source =0;
  	  Buffer->AddTuple(q,JoinInfo->K);
  	}
  	if((q=Buffer->GetTuple(LastRightRec,
  			       Handle->ChildSchema[JoinInfo->RightRelIndex],
  			       2,
  			       JoinInfo->RightAttrIndex,
  			       Handle,
  			       ResidPredPlans,
  			       1))){
  	  // right rec found in buffer	  
  	  ASSERT (q->Source == 1);
  	  r = XXX_FALSE;
  	  q->WG= ComputeScore(q->WG,RightWGrade);
  	  q->BG= ComputeScore(q->Grade,RightBGrade);	
  	  q->RightRec= LastRightRec;
  	  q->Source =0; // final
  	  Buffer->AddTuple(q,JoinInfo->K);
  	}
  	Handle->ChildRecs[JoinInfo->LeftRelIndex]  = LastLeftRec;
  	Handle->ChildRecs[JoinInfo->RightRelIndex] = LastRightRec;
	
  	if(l==XXX_TRUE && isEqual(JoinInfo,Handle, ResidPredPlans) == XXX_TRUE)
  	  {
  	    double newWgrade = ComputeScore(LeftWGrade,RightBGrade);	
  	    double newBgrade = ComputeScore(LeftWGrade,RightBGrade);
	    
  	    q=new Tuple(LastLeftRec, LastRightRec,newWgrade,newWgrade,newBgrade,0); // has the final schema
  	    Buffer->AddTuple(q,JoinInfo->K);
  	  }
  	else{
  	  if (l==XXX_TRUE){
  	    double worstgrade = ComputeScore(LeftWGrade,0);
  	    double bestgrade  = ComputeScore(LeftBGrade,JoinInfo->RightTh);
  	    q=new Tuple(LastLeftRec,(XxxRecord*)NULL,LeftWGrade,worstgrade,bestgrade,1); // has left schema
  	    Buffer->AddTuple(q,JoinInfo->K);
  	  }
  	  if( r== XXX_TRUE){
	     double worstgrade = ComputeScore(RightWGrade,0);
	     double bestgrade  = ComputeScore(RightBGrade,JoinInfo->LeftTh);
	     q=new Tuple((XxxRecord*)NULL, LastRightRec,RightWGrade,worstgrade,bestgrade,2); //has right schema
	     Buffer->AddTuple(q,JoinInfo->K);
  	  }
  	}
      }// else of leftscanvalid and rightscanvalid 
      
// Begin Queue Processing
  
      Buffer->updateRanks(JoinInfo->LeftTh, JoinInfo->RightTh);//        //Stopping Criteria
      
      
      
      q= Buffer->ElementAt(0);
      if(q){
	 double Mkd = q->WG;
	 double BRd = MaxVal(ComputeScore(JoinInfo->RightTh, JoinInfo->LeftTh), Buffer->MaxB);
	 if(BRd <= Mkd) break;
      }
        
    }// while
    
    if(Buffer->IsEmpty()) // Buffer is empty
       JoinInfo->eof = XXX_TRUE;
    
    if (JoinInfo->eof)
    {
       JoinInfo->Valid = XXX_FALSE;   
       SCsr.AtEnd = XXX_TRUE;
       ReturnRec = 0;
       return XXX_OK;
    }
    else if(!LeftScanValid || !RightScanValid){ 
       JoinInfo->eof = XXX_TRUE;   
    }
    // before reporting th grade should be adjusted to the WG and the orig_b to BG ----TODO
    Buffer->GetNext();
    
    if(q->Source==1)     
       q->RightRec= JoinInfo->DummyRightRec;
    if(q->Source==2)     
       q->LeftRec = q->RightRec; 
     //JoinInfo->DummyLeftRec; // to be fixed.. for missing records assume same schema..
    Handle->ChildRecs[JoinInfo->LeftRelIndex]=q->LeftRec;
    Handle->ChildRecs[JoinInfo->RightRelIndex]=q->RightRec;  
    XxxErrCode  status = createRecord(Handle, RelationArray, Schema, ReturnRec,q->WG,q->BG);
    
    return status;
    
}

//-------------------------------------------------------------------------
XxxErrCode RelKRJPlanOp::reclaimCurrent(XxxDerivedRelation::CursorInfo* Handle, int NLPos) const 
{
 if (Handle->ChildRecs[0]) {
   delete Handle->ChildRecs[0];
   Handle->ChildRecs[0] = 0;
 }
 if (Handle->ChildRecs[1]) {
   delete Handle->ChildRecs[1];
   Handle->ChildRecs[1] = 0;
 }
 return XXX_OK;
}
//----------------------------------------------------------------------------

XxxErrCode RelKRJPlanOp::openChildScan(XxxDerivedRelation::CursorInfo* Handle,
	      XxxRelation* *RelationArray,
	      int NLPos) const 
{
 // Extract local predicates
 XxxBooleanExpressionPlan* InnerPredPlan = 
   (SelPredPlans?SelPredPlans[NLPos]:0);

 if (JoinPredPlans && JoinPredPlans[NLPos]) {
   if (XxxBooleanExpressionPlan::MakeConjunction
       (InnerPredPlan, JoinPredPlans[NLPos]) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
     return XXX_NOT_OK;
   }
 }
 if (ResidPredPlans && ResidPredPlans[NLPos]) {
   if (XxxBooleanExpressionPlan::MakeConjunction
       (InnerPredPlan, ResidPredPlans[NLPos]) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
     return XXX_NOT_OK;
   }
 }
 
 if (RelationArray[NLPos]->
     InitCursor1(Handle->ChildCursorArray[NLPos],
		 InnerPredPlan,
		 (const XxxRecord **)Handle->ChildRecs,
		 (const XxxRecordSchema **)Handle->ChildSchema,
		 Handle->Params, Handle->ArrayLen,
		 NLPos) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, "Error opening relation scan");
   return XXX_NOT_OK;
  } // if
 return XXX_OK;
}
//-------------------------------------------------------------------------------
XxxErrCode RelKRJPlanOp::closeChildScan(XxxDerivedRelation::CursorInfo* Handle,
	       XxxRelation* *RelationArray,
	       int NLPos) const 
{
 return RelationArray[NLPos]->CloseCursor(Handle->ChildCursorArray[NLPos]);
}
//------------------------------------------------------------------------------
void RelKRJPlanOp::PrintNodeInfo(ostream &OStr) const 
{
  RelSPJPlanOp::PrintNodeInfo(OStr);
  // now print any additional stuff
}
//----------------------------------------------------------------------------
