#ifdef __GNUG__
#pragma implementation
#endif

#include "xxx_thread.h"
#include "relconfig.h"
#include "types.h"
#include "values.h"
#include "commands.h"
#include "releng.h"
#include "relenv.h"
#include "relcat.h"
#include "reldata.h"
#include "relquery.h"
#include "shoresm.h"
#include "defeng.h"

extern XxxErrCode XxxListAlias(ostream& OStr);

RelCommandEngine::RelCommandEngine(XxxRelationEngine* SrcEng) : 
commandTable(REL_COMMANDS_MAX)
{
 RelEng = SrcEng;
 // populate table of commands
 XxxRelCommand DummyCmd;
 char listcmdname[20], setcmdname[20];
 strcpy(listcmdname,XXX_DEF_LISTADTENV);
 strcpy(setcmdname,XXX_DEF_SETADTENV);
 //Bjarke, sep. 13 2001: added '&RelCommandEngine::' to all calls to AddItem below
 commandTable.AddItem(DummyCmd.Assign((char *)(listcmdname), 
				      (XxxRelCommand::CommandFunc)(&RelCommandEngine::ListADTEnvFnc)));
 commandTable.AddItem(DummyCmd.Assign((char *)(setcmdname),
				      (XxxRelCommand::CommandFunc)(&RelCommandEngine::SetADTEnvFnc)));
 commandTable.AddItem(DummyCmd.Assign((char *)("list"), 
				      (XxxRelCommand::CommandFunc)(&RelCommandEngine::ListUtilityFnc)));
 commandTable.AddItem(DummyCmd.Assign((char *)("help"), 
				      (XxxRelCommand::CommandFunc)(&RelCommandEngine::HelpFnc)));
 // disable this: we dont want people adding/removing devices in
 // the current release.
 // commandTable.AddItem(DummyCmd.Assign("dev", &RelCommandEngine::DeviceFnc));
 commandTable.AddItem(DummyCmd.Assign((char *)("trans"),
				      (XxxRelCommand::CommandFunc)(&RelCommandEngine::TransFnc)));
 commandTable.AddItem(DummyCmd.Assign((char *)("setenv"), 
				      (XxxRelCommand::CommandFunc)(&RelCommandEngine::OptionFnc)));
 //Begin: Moustafa
 // adds stream_stop command from the client side
 // a similar command exist at the server side as well however 
 // it needs client number
  commandTable.AddItem(DummyCmd.Assign((char *)"stream_stop", 
				       (XxxRelCommand::CommandFunc)(&StreamStopFnc)));
 //End: Moustafa
}

RelCommandEngine::~RelCommandEngine()
{
}

const int XxxRelCommand::COMMAND_NAME_MAX = 30;

XxxErrCode RelCommandEngine::HelpCommand(char *CommandName, ostream& OStr) const
{
 XxxErrCode rc;
 // retrieve associated function from table
 XxxRelCommand UserCommand;
 strncpy(UserCommand.commandName, CommandName, 
	 XxxRelCommand::COMMAND_NAME_MAX);
 UserCommand.commandName[XxxRelCommand::COMMAND_NAME_MAX-1] = 0;
 if (commandTable.FindTableItem(UserCommand, UserCommand) != XXX_TRUE) {
   XXX_NOTE_ERROR(XXX_BADCMD, CommandName);
   rc = XXX_NOT_OK;
  }
 else {
   // print help for current command
   rc = UserCommand.ShowHelp(OStr);
  }
 return rc;
}

static int printCommandItem(XxxRelCommand& TableEntry, void *arg)
{
 TableEntry.Print(*((ostream *)arg));
 return 1;
}
void RelCommandEngine::PrintCommands(ostream& OStr)
{
 OStr << "\tREL:: List of Commands..." << endl;
 commandTable.ApplyAll(printCommandItem ,(void *)(&OStr));
}

XxxErrCode RelCommandEngine::ProcessCommand(XxxRelCommandInfo* CommandInfo)
{
 // retrieve associated function from table
 XxxRelCommand UserCommand;
 strncpy(UserCommand.commandName, CommandInfo->Name, 
	 XxxRelCommand::COMMAND_NAME_MAX);
 UserCommand.commandName[XxxRelCommand::COMMAND_NAME_MAX-1] = 0;
 if (commandTable.FindTableItem(UserCommand, UserCommand) != XXX_TRUE) {
   XXX_NOTE_ERROR(XXX_BADCMD, CommandInfo->Name);
   return XXX_NOT_OK;
  }
 else {
   // execute function
   XxxRelCommand::CommandFunc MemberFunc = UserCommand.commandFunc;
  return (this->*MemberFunc)(CommandInfo);
  }
}

XxxRelCommandInfo::~XxxRelCommandInfo()
{
 if (Name) delete [] Name;
 for (int i = 0; i < ArgCount; i++)
   delete Args[i];
}

void XxxRelCommand::Print(ostream& OStr) const
{
 OStr << "Command " << commandName << endl;
}

//
// Here are the relation utility commands
//

XxxErrCode RelCommandEngine::HelpFnc(XxxRelCommandInfo* CommandArgs)
{
 if (!CommandArgs) {
   XXX_NOTE_ERROR(XXX_BADCMD, "");
   return XXX_NOT_OK;
  }

 if (CommandArgs->ArgCount < 1) {
   PrintCommands(*(XXX_THISENV().OutputStr));
  }
 else {
   char *HelpTopic = ((XxxConstStringValue *)((*CommandArgs)[0]))->c;
   HelpCommand(HelpTopic, *(XXX_THISENV().OutputStr));
  }
 return XXX_OK;
}


//
//  LIST utility
//   Takes several variants:
//     LIST CATS
//     LIST VIEWS
//     LIST RELS
//     LIST INDXS
//
XxxErrCode RelCommandEngine::ListUtilityFnc(XxxRelCommandInfo* CommandArgs)
{
 if (!CommandArgs || (CommandArgs->ArgCount < 1) || 
     ((*CommandArgs)[0]->BasicType() != XXX_STRING) ) {
   XXX_NOTE_ERROR(XXX_BADCMD, "Need 1 string argument to List");
   return XXX_NOT_OK;
  }
 char *ListCommand = ((XxxConstStringValue *)((*CommandArgs)[0]))->c;

 if (!strcmp(ListCommand, "cats")) {
   RelEng->DataEng->PrintCats(*(XXX_THISENV().OutputStr));
  }
 else if (!strcmp(ListCommand, "schemas")) {
   RelEng->DataEng->PrintSchemas(*(XXX_THISENV().OutputStr));
  }
 else if (!strcmp(ListCommand, "views")) {
   RelEng->QueryEng->PrintViews(*(XXX_THISENV().OutputStr));
  }
 else if (!strcmp(ListCommand, "rels")) {
   RelEng->DataEng->PrintRels(*(XXX_THISENV().OutputStr));
  }
 else if (!strcmp(ListCommand, "env")) {
   ((XxxRelEnv*)XXX_THISENGENV(RelationEngineId))->
     Print(*(XXX_THISENV().OutputStr));
 }
 else if (!strcmp(ListCommand, "type_alias")) {
	 if(XxxListAlias(*(XXX_THISENV().OutputStr)) != XXX_OK){
	     XXX_NOTE_ERROR(XXX_BADADT, "List alias failed");
 		 return XXX_NOT_OK;
	 }
 }
 else {
   XXX_NOTE_ERROR(XXX_BADCMD, "Unknown argument to List");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

XxxErrCode RelCommandEngine::ListADTEnvFnc(XxxRelCommandInfo* CommandArgs)
{
  if (!CommandArgs || (CommandArgs->ArgCount <1) ||
      ((*CommandArgs)[0]->BasicType() != XXX_STRING) ) {
    char buf[100];
    sprintf(buf,"Need 1 string argument to %s",CommandArgs->Name);
    XXX_NOTE_ERROR(XXX_BADCMD, buf);
    return XXX_NOT_OK;
  }
  char *AdtName=((XxxConstStringValue*) ((*CommandArgs)[0]))->c;
  char cmd[XXX_TEXTLINE_MAX];
  sprintf(cmd,"XXX:%s %s;",XXX_DEF_LISTADTENV,AdtName);
  int dummy;
  XxxGenericParse* dummy2;
  return ((XxxDefaultEngine*)XxxDBEngineArray[XxxDefaultEngine::DefEngId])->
    ExprParse(cmd,NULL,dummy,dummy2);
}

XxxErrCode RelCommandEngine::SetADTEnvFnc(XxxRelCommandInfo* CommandArgs)
{
  if (!CommandArgs || (CommandArgs->ArgCount <3) ||
      (*CommandArgs)[0]->BasicType() != XXX_STRING ||
      (*CommandArgs)[1]->BasicType() != XXX_STRING) {
    char buf[100];
    sprintf(buf,"Missing arguments or arguments of wrong type.");
    XXX_NOTE_ERROR(XXX_BADCMD, buf);
    return XXX_NOT_OK;
  }
  char cmd[XXX_TEXTSTRING_MAX+XXX_ADT_NAME_MAX+XXX_ATTR_NAME_MAX+20];
  sprintf(cmd,"XXX:%s %s %s ",XXX_DEF_SETADTENV,
	  ((XxxConstStringValue*)((*CommandArgs)[0]))->c,
	  ((XxxConstStringValue*)((*CommandArgs)[1]))->c);

  // now comes the tough part.  The remaining "arguments" are the value we
  // want to place in the environment variable.  So, we have to construct
  // the value string from the remaining arguments.
  char NewValue[XXX_TEXTSTRING_MAX]="";
  char temp[XXX_TEXTSTRING_MAX];
  for (int i=2; i<CommandArgs->ArgCount;i++) {
    *temp=0; // initialize to 0 length string
    switch ((*CommandArgs)[i]->BasicType()) {
    case XXX_INT:
      sprintf(temp,"%d ",((XxxConstIntValue*)((*CommandArgs)[i]))->i);
      break;
    case XXX_DOUBLE:
      sprintf(temp,"%f ",((XxxConstDoubleValue*)((*CommandArgs)[i]))->d);
      break;
    case XXX_STRING:
      sprintf(temp,"%s ",((XxxConstStringValue*)((*CommandArgs)[i]))->c);
      break;
    }
    if (*temp==0) {
      // we've hit an ADT object that can't be expressed in a string!
      char buf[100];
      strcpy(buf,"NewValue is not an int, string, or double!");
      XXX_NOTE_ERROR(XXX_BADCMD, buf);
      return XXX_NOT_OK;
    }
    strcat(NewValue,temp);
  }
  // append this "NewValue" to our command
  strcat(cmd, NewValue);
  int dummy;
  XxxGenericParse *dummy2;
  return ((XxxDefaultEngine*)XxxDBEngineArray[XxxDefaultEngine::DefEngId])->
    ExprParse(cmd,NULL,dummy,dummy2);
}

//
//  DEV manipulations : disabled for now
//   Takes several variants:
//     DEV Build <dev-name> [<dev-size> <dev-impl>]
//     DEV Destroy <dev-name> [<dev-impl>]
//     DEV Mount <dev-name> [<dev-impl>]
//     DEV UnMount <dev-name> [<dev-impl>]
//     DEV List [<dev-impl>]
//
XxxErrCode RelCommandEngine::DeviceFnc(XxxRelCommandInfo* CommandArgs)
{
 if (!CommandArgs || (CommandArgs->ArgCount < 1) || 
     ((*CommandArgs)[0]->BasicType() != XXX_STRING) ) {
   XXX_NOTE_ERROR(XXX_BADCMD, "Need 1 string argument to DEV");
   return XXX_NOT_OK;
  }
 char *DevCommand = ((XxxConstStringValue *)((*CommandArgs)[0]))->c;

 if (!XXX_THISTHREAD()->xct() && !strcasecmp(DevCommand, "build")) {
   if ((CommandArgs->ArgCount < 2) || 
       ((*CommandArgs)[1]->BasicType() != XXX_STRING)) {
     XXX_NOTE_ERROR(XXX_BADCMD,"Need dev-name string argument for DEV BUILD");
     return XXX_NOT_OK;
    }
   char *DevName = ((XxxConstStringValue *)((*CommandArgs)[1]))->c;
   if ((CommandArgs->ArgCount == 4) && 
       ((*CommandArgs)[2]->BasicType() == XXX_INT) &&
       ((*CommandArgs)[3]->BasicType() == XXX_STRING)) {
     // determine the type of the data src
     RelSrcTypes SrcType = RelEng->DataEng->FindSrcType
       (((XxxConstStringValue *)((*CommandArgs)[3]))->c);
     if (SrcType == REL_INVALID_SRC) {
       XXX_NOTE_ERROR(XXX_BADCMD, "Bad Src Type argument for DEV CREATE");
       return XXX_NOT_OK;
      }
     return RelEng->DataEng->CreateDevice
       (DevName, ((XxxConstIntValue *)((*CommandArgs)[2]))->i, SrcType);
    }
   else 
     return RelEng->DataEng->CreateDevice(DevName, 
					  RelEng->DataEng->DefaultDevSize(),
					  RelEng->DataEng->DefaultSrcType());
  }
 else if (!XXX_THISTHREAD()->xct() && !strcasecmp(DevCommand, "destroy")) {
   if ((CommandArgs->ArgCount < 2) || 
       ((*CommandArgs)[1]->BasicType() != XXX_STRING)) {
     XXX_NOTE_ERROR(XXX_BADCMD,"Need dev-name string argument for DEV CREATE");
     return XXX_NOT_OK;
    }
   char *DevName = ((XxxConstStringValue *)((*CommandArgs)[1]))->c;
   if ((CommandArgs->ArgCount == 3) && 
       ((*CommandArgs)[2]->BasicType() == XXX_STRING)) {
     // determine the type of the data src
     RelSrcTypes SrcType = RelEng->DataEng->FindSrcType
       (((XxxConstStringValue *)((*CommandArgs)[2]))->c);
     if (SrcType == REL_INVALID_SRC) {
       XXX_NOTE_ERROR(XXX_BADCMD, "Bad Src Type argument for DEV DESTROY");
       return XXX_NOT_OK;
      }
     return RelEng->DataEng->DestroyDevice(DevName, SrcType);
    }
   else 
     return RelEng->DataEng->DestroyDevice(DevName, 
					   RelEng->DataEng->DefaultSrcType());
  }
 else if (!XXX_THISTHREAD()->xct() && !strcasecmp(DevCommand, "mount")) {
   if ((CommandArgs->ArgCount < 2) || 
       ((*CommandArgs)[1]->BasicType() != XXX_STRING)) {
     XXX_NOTE_ERROR(XXX_BADCMD,"Need dev-name string argument for DEV MOUNT");
     return XXX_NOT_OK;
    }
   char *DevName = ((XxxConstStringValue *)((*CommandArgs)[1]))->c;
   if ((CommandArgs->ArgCount == 3) && 
       ((*CommandArgs)[2]->BasicType() == XXX_STRING)) {
     // determine the type of the data src
     RelSrcTypes SrcType = RelEng->DataEng->FindSrcType
       (((XxxConstStringValue *)((*CommandArgs)[2]))->c);
     if (SrcType == REL_INVALID_SRC) {
       XXX_NOTE_ERROR(XXX_BADCMD, "Bad Src Type argument for DEV MOUNT");
       return XXX_NOT_OK;
      }
     return RelEng->DataEng->MountDevice(DevName, SrcType);
    }
   else 
     return RelEng->DataEng->MountDevice(DevName, 
					 RelEng->DataEng->DefaultSrcType());
  }
 else if (!XXX_THISTHREAD()->xct() && !strcasecmp(DevCommand, "unmount")) {
   if ((CommandArgs->ArgCount < 2) || 
       ((*CommandArgs)[1]->BasicType() != XXX_STRING)) {
     XXX_NOTE_ERROR(XXX_BADCMD,
		    "Need dev-name string argument for DEV UNMOUNT");
     return XXX_NOT_OK;
    }
   char *DevName = ((XxxConstStringValue *)((*CommandArgs)[1]))->c;
   if ((CommandArgs->ArgCount == 3) && 
       ((*CommandArgs)[2]->BasicType() == XXX_STRING)) {
     // determine the type of the data src
     RelSrcTypes SrcType = RelEng->DataEng->FindSrcType
       (((XxxConstStringValue *)((*CommandArgs)[2]))->c);
     if (SrcType == REL_INVALID_SRC) {
       XXX_NOTE_ERROR(XXX_BADCMD, "Bad Src Type argument for DEV UNMOUNT");
       return XXX_NOT_OK;
      }
     return RelEng->DataEng->UnMountDevice(DevName, SrcType);
    }
   else 
     return RelEng->DataEng->UnMountDevice(DevName, 
					   RelEng->DataEng->DefaultSrcType());
  }
 else if (!strcasecmp(DevCommand, "list")) {
   RelEng->DataEng->PrintDevs(*(XXX_THISENV().OutputStr));
  }
 else {
   XXX_NOTE_ERROR(XXX_BADCMD, "Unknown argument to DEV");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

XxxErrCode RelCommandEngine::TransFnc(XxxRelCommandInfo* CommandArgs)
{
 if (!CommandArgs || (CommandArgs->ArgCount != 1) || 
     ((*CommandArgs)[0]->BasicType() != XXX_STRING) ) {
   XXX_NOTE_ERROR(XXX_BADCMD, "Need 1 string argument to TRANS");
   return XXX_NOT_OK;
  }
 char *TransCommand = ((XxxConstStringValue *)((*CommandArgs)[0]))->c;

 if (!strcmp(TransCommand, "commit")) {
   if (!XXX_THISENV().MultiStmtXact) {
     XXX_NOTE_ERROR(XXX_BADCMD, "Can only be used for multi-stmt xact");
     return XXX_NOT_OK;
    }
   if (XXX_IS_ERROR()) {
     XXX_NOTE_ERROR(XXX_BADCMD, "Cannot commit: error has occurred;");
     return XXX_NOT_OK;
    }
   if (!XxxShoreSM || (XxxCommitTrans() != XXX_OK)) {
     XXX_NOTE_ERROR(XXX_BADCMD, "Unable to commit transaction");
     return XXX_NOT_OK;
    }
  }
 else if (!strcmp(TransCommand, "abort")) {
   if (!XXX_THISENV().MultiStmtXact) {
     XXX_NOTE_ERROR(XXX_BADCMD, "Can only be used for multi-stmt xact");
     return XXX_NOT_OK;
    }
   if (!XxxShoreSM || (XxxAbortTrans() != XXX_OK)) {
     XXX_NOTE_ERROR(XXX_BADCMD, "Unable to abort transaction");
     return XXX_NOT_OK;
    }
  }
 else if (!strcmp(TransCommand, "end")) {
   if (!XXX_THISENV().MultiStmtXact) {
     XXX_NOTE_ERROR(XXX_BADCMD, "Can only be used for multi-stmt xact");
     return XXX_NOT_OK;
    }
   // abort if an error occurred, otherwise commit
   XxxBool ToAbort = XXX_FALSE;
   if (XXX_IS_ERROR()) {
     ToAbort = XXX_TRUE;
    }
   if (ToAbort &&
       (!XxxShoreSM || (XxxAbortTrans() != XXX_OK))) {
     XXX_NOTE_ERROR(XXX_BADCMD, "Unable to abort transaction");
     return XXX_NOT_OK;
    }
   if (!ToAbort &&
       (!XxxShoreSM || (XxxCommitTrans() != XXX_OK))) {
     XXX_NOTE_ERROR(XXX_BADCMD, "Unable to commit transaction");
     return XXX_NOT_OK;
    }
  }
 else if (!strcmp(TransCommand, "begin")) {
   if (XXX_THISENV().MultiStmtXact) {
     XXX_NOTE_ERROR(XXX_BADCMD, "Another multi-stmt xact exists");
     return XXX_NOT_OK;
    }
   // commit any existing xact
   if (!XxxShoreSM || (XxxCommitTrans() != XXX_OK)) {
     XXX_NOTE_ERROR(XXX_BADCMD, "Unable to commit existing xact");
     return XXX_NOT_OK;
    }
   XXX_THISENV().MultiStmtXact = XXX_TRUE;
   if (!XxxShoreSM || (XxxBeginTrans() != XXX_OK)) {
     XXX_NOTE_ERROR(XXX_BADCMD, "Unable to begin multi-stmt xact");
     return XXX_NOT_OK;
    }
  }
 else {
   XXX_NOTE_ERROR(XXX_BADCMD, "Unknown argument to TRANS");
   return XXX_NOT_OK;
  }

 // Now start up a local xact for the rest of this request.
 // This is needed for the abort and commit commands.
 if (!XXX_THISTHREAD()->xct()) {
   XXX_THISENV().MultiStmtXact = XXX_FALSE;
   if (!XxxShoreSM || (XxxBeginTrans() != XXX_OK)) {
     XXX_NOTE_ERROR(XXX_BADCMD, "Unable to begin transaction");
     return XXX_NOT_OK;
    }
  }
 return XXX_OK;
 
}

XxxErrCode RelCommandEngine::OptionFnc(XxxRelCommandInfo* CommandArgs)
{
 if (!CommandArgs || (CommandArgs->ArgCount < 2) || 
     ((*CommandArgs)[0]->BasicType() != XXX_STRING) ||
     ((*CommandArgs)[1]->BasicType() != XXX_INT) ) {
   XXX_NOTE_ERROR(XXX_BADCMD, "Need one string and one int argument to Set");
   return XXX_NOT_OK;
  }
 char *Option = ((XxxConstStringValue *)((*CommandArgs)[0]))->c;
 int OptValue = ((XxxConstIntValue *)((*CommandArgs)[1]))->i;
 XxxRelEnv* REnv = ((XxxRelEnv*)(XXX_THISENGENV(RelationEngineId)));

 if (REnv &&
     REnv->SetValue(Option, (char)OptValue) == XXX_OK) {
   return XXX_OK;
 }
 // else try setting it in XXX_THISENV()
 if (XXX_THISENV().SetValue(Option, (char)OptValue) != XXX_OK) {
   XXX_NOTE_ERROR(REL_BADEXEC, "Bad Set command");
   return XXX_NOT_OK;
 }
 return XXX_OK;
}

//Begin: Moustafa
// stream_stop client command handler
// Stops streams flow to current query and hence results in query stop

XxxErrCode RelCommandEngine::StreamStopFnc(XxxRelCommandInfo* CommandArgs){

  ((XxxRequestThread* )(XXX_THISTHREAD()))->StopQuery = 1;
  return XXX_OK;
}

//End: Moustafa

XxxErrCode XxxRelCommand::ShowHelp(ostream& OStr) const
{
 if (!strcmp(commandName, "set")) {
   OStr << "\tset <option-name> <option-value>" << endl;
   XxxRelEnv* REnv = ((XxxRelEnv*)(XXX_THISENGENV(RelationEngineId)));
   REnv->Print(OStr);
  }
 else if (!strcmp(commandName, XXX_DEF_LISTADTENV)) {
   OStr << "\t" << XXX_DEF_LISTADTENV << " <adt-name>" << endl;
  }
 else if (!strcmp(commandName, XXX_DEF_SETADTENV)) {
   OStr << "\t" << XXX_DEF_SETADTENV << " <adt-name> <option-name> <option-value>" << endl;
  }
 else if (!strcmp(commandName, "list")) {
   OStr << "\tlist <item-name>" << endl;
   OStr << "list cats" << endl;
   OStr << "list indxs" << endl;
   OStr << "list rels" << endl;
   OStr << "list views" << endl;
  }
 else if (!strcmp(commandName, "trans")) {
   OStr << "\ttrans <action>" << endl;
   OStr << "trans begin" << endl;
   OStr << "trans commit" << endl;
   OStr << "trans abort" << endl;
  }
 else {
   OStr << "No detailed help available on " << commandName << endl;
   XXX_NOTE_ERROR(XXX_INCOMPL, "You are beyond help!\n");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

