#ifndef XXX_RELCOMMANDS_H
#define XXX_RELCOMMANDS_H
#include <string.h>
#include "templates.h"
#include "reladt.h"
#include "adtfunc.h"

XXX_EXTERN_CLASS_DECL(XxxConstValue);
XXX_EXTERN_CLASS_DECL(RelCommandEngine);

class XxxRelCommandInfo {
 public:
  char *Name;
  XxxConstValue *Args[5];
  int ArgCount;
  XxxRelCommandInfo() {ArgCount = 0; Name = NULL;}
  ~XxxRelCommandInfo();
  XxxRelCommandInfo* AddArg(XxxConstValue *Argument) {
    Args[ArgCount++] = Argument;
    return this;
   }
  XxxConstValue* operator [] (int index) const
    { return Args[index]; }
  XxxErrCode Process();
 };


class XxxRelCommand {  
 public:
  typedef XxxErrCode (RelCommandEngine::*CommandFunc)(XxxRelCommandInfo *);
  static const int COMMAND_NAME_MAX; // = 30
  char commandName[30];
  CommandFunc commandFunc;
  
  XxxRelCommand() {};
  XxxRelCommand& Assign(char *Name, CommandFunc Func)
   {
    commandFunc = Func;
    strncpy(commandName, Name, COMMAND_NAME_MAX);
    commandName[COMMAND_NAME_MAX-1] = 0;
    return (*this);
   };
  void Print(ostream& OStr) const;
  XxxErrCode ShowHelp(ostream& OStr) const;
  
  // following two needed by the template Table class
  int Hash() const { return (commandName[0] + commandName[1]);}
  int operator==(const XxxRelCommand& Command) const
   { return !strcmp(commandName, Command.commandName);}
 };

class RelCommandEngine {
  XxxTable<XxxRelCommand> commandTable;
  void PrintCommands(ostream& OStr);
  XxxErrCode HelpCommand(char *CommandName, ostream& OStr) const;
  XxxRelationEngine* RelEng;

  XxxErrCode HelpFnc(XxxRelCommandInfo*);
  XxxErrCode ListUtilityFnc(XxxRelCommandInfo*);
  XxxErrCode DeviceFnc(XxxRelCommandInfo*);
  XxxErrCode TransFnc(XxxRelCommandInfo*);
  XxxErrCode OptionFnc(XxxRelCommandInfo*);
  XxxErrCode ListADTEnvFnc(XxxRelCommandInfo*);
  XxxErrCode SetADTEnvFnc(XxxRelCommandInfo*);
  XxxErrCode IncrementalFunctionReportFnc(XxxRelCommandInfo*);
  XxxErrCode FunctionReportFnc(XxxRelCommandInfo*);
  XxxErrCode AddClientUDFFnc(XxxRelCommandInfo*);
  XxxErrCode AddServerUDFFnc(XxxRelCommandInfo*);
  //Begin: Moustafa
  // adds stream_stop function
  XxxErrCode StreamStopFnc(XxxRelCommandInfo*);
  //End: Moustafa
  // used by Add...UDFFnc
  XxxErrCode determineUDFTypes(XxxRelCommandInfo* CommandArgs, 
                         XxxFuncMethodInfo::ArgInfo &Signature,
                         XxxBasicType &ResultType);

 public:  
  RelCommandEngine(XxxRelationEngine* SrcEng);
  virtual ~RelCommandEngine();
  XxxErrCode ProcessCommand(XxxRelCommandInfo* CommandInfo);

  // used by AddClientUDFFnc for UDF upload
  static int UploadPortNumber;

 };

#endif
