#ifdef __GNUG__
#pragma implementation
#endif

#include "relplanop.h"

//
// Methods of RelDistinctPlanOp
//

RelDistinctPlanOp::RelDistinctPlanOp(RelPlanOp* Child,
				     RelCatalogInfo* CatInfo)
: RelDerivedPlanOp(1, &Child, CatInfo)
{
}

RelDistinctPlanOp::~RelDistinctPlanOp() 
{ }

void RelDistinctPlanOp::NodePrint (ostream &OStr, int curBlockId, int Level, 
				   int numChildren, int *childrenIds) const {
  OStr << "NODE Eliminate Duplicates" << endl;
  OStr << "ID " << curBlockId << endl;
  OStr << "LEVEL " << Level << endl;
  if (numChildren > 0) {
    OStr << "CHILDREN ";
    for (int i = 0; i < numChildren; i++) {
      OStr << childrenIds[i] << " ";
    }
    OStr << endl;
  }
  OStr << "COLOR " << 0x66C66C << endl;
  OStr << "ENDNODE" << endl;
}

XxxErrCode 
RelDistinctPlanOp::InitHandle(XxxDerivedRelation::CursorInfo* Handle,
			      XxxRelation* *RelationArray) const
{
  if (Handle->ArrayLen != 1) {
    REL_NOTE_ERROR(REL_BADEXEC, "DistinctPlanOp needs exactly one input");
    return XXX_NOT_OK;
  }
  // Open a scan on the input
  if (RelationArray[0]->InitCursor1(Handle->ChildCursorArray[0],
				    Handle->Filter, NULL, NULL,
				    Handle->Params) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADCURSOR, 
		   "Unable to open sorted input relation cursor");
    return XXX_NOT_OK; 
  }
  Handle->OperatorHandle = NULL;
  return XXX_OK;
}

XxxErrCode 
RelDistinctPlanOp::CloseHandle(XxxDerivedRelation::CursorInfo* Handle,
			       XxxRelation* *RelationArray) const
{
  XxxRecord *lastReturnedRec = (XxxRecord *)Handle->OperatorHandle;
  if (lastReturnedRec)
    delete lastReturnedRec;
  // Close the scan on the input relation
  if (Handle->ChildCursorArray[0].State &&
      RelationArray[0]->CloseCursor(Handle->ChildCursorArray[0]) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADCURSOR, "Unable to close cursor");
    return XXX_NOT_OK;
  }
  return XXX_OK;
}

XxxErrCode RelDistinctPlanOp::GetNextRecord(XxxSetCursor& SCsr,
					    XxxRelation* *RelationArray,
					    const XxxRecordSchema* Schema,
					    XxxRecord*& ReturnRec) const
{
  XxxDerivedRelation::CursorInfo* Handle = 
    (XxxDerivedRelation::CursorInfo*)(SCsr.State);
  if (!Handle) {
    REL_NOTE_ERROR(REL_BADEXEC, "NULL Exec handle");
    return XXX_NOT_OK;
  }
  XxxRecord *&lastReturnedRec = (XxxRecord *&)Handle->OperatorHandle;
  if (RelationArray[0]->NextItem(Handle->ChildCursorArray[0],
				 Handle->ChildRecs[0]) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADEXEC, "Error getting NextItem from relation");
    return XXX_NOT_OK;
  }
  while (Handle->ChildCursorArray[0].IsValid() == XXX_TRUE) {
    if (!lastReturnedRec ||
	!lastReturnedRec->Equals(Handle->ChildRecs[0],
				 CatalogInfo->recSchema))
      break;
    if (RelationArray[0]->NextItem(Handle->ChildCursorArray[0],
				   Handle->ChildRecs[0]) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADEXEC, "Error getting NextItem from relation");
      return XXX_NOT_OK;
    }
  }
  if (Handle->ChildCursorArray[0].IsValid()) {
    ReturnRec = Handle->ChildRecs[0];
    delete lastReturnedRec;
    lastReturnedRec = new XxxRecord(*ReturnRec);
    return XXX_OK;
  } else {
    // The scan is complete
    SCsr.AtEnd = XXX_TRUE;
    ReturnRec = 0;
    return XXX_OK;
  }
}

void RelDistinctPlanOp::PrintThis(ostream& OStr) const
{
  OStr << PrintString();
}

