#ifdef __GNUG__
#pragma implementation
#endif

#include "relplanop.h"

//
// Methods of RelFilterPlanOp
//

RelFilterPlanOp::RelFilterPlanOp(RelPlanOp* Child,
				 RelCatalogInfo* CatInfo,
				 XxxBooleanExpressionPlan* SPredPlan)
: RelDerivedPlanOp(1, &Child, CatInfo)
{
 SelPredPlan = SPredPlan;
}

RelFilterPlanOp::~RelFilterPlanOp() 
{ }

void RelFilterPlanOp::NodePrint (ostream &OStr, int curBlockId, int Level, 
				   int numChildren, int *childrenIds) const {
  OStr << "NODE Filter" << endl;
  OStr << "ID " << curBlockId << endl;
  OStr << "LEVEL " << Level << endl;
  if (numChildren > 0) {
    OStr << "CHILDREN ";
    for (int i = 0; i < numChildren; i++) {
      OStr << childrenIds[i] << " ";
    }
    OStr << endl;
  }
  OStr << "COLOR " << 0x76C667 << endl;
  OStr << "ENDNODE" << endl;
}

XxxErrCode 
RelFilterPlanOp::InitHandle(XxxDerivedRelation::CursorInfo* Handle,
			    XxxRelation* *RelationArray) const
{
  if (Handle->ArrayLen != 1) {
    REL_NOTE_ERROR(REL_BADEXEC, "FilterPlanOp needs exactly one input");
    return XXX_NOT_OK;
  }
  // Open a scan on the input: do not pass any conditions in
  if (RelationArray[0]->InitCursor1(Handle->ChildCursorArray[0],
				    NULL, NULL, NULL,
				    Handle->Params) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADCURSOR, 
		   "Unable to open sorted input relation cursor");
    return XXX_NOT_OK; 
  }
  Handle->OperatorHandle = NULL;
  return XXX_OK;
}

XxxErrCode 
RelFilterPlanOp::CloseHandle(XxxDerivedRelation::CursorInfo* Handle,
			       XxxRelation* *RelationArray) const
{
  // Close the scan on the input relation
  if (Handle->ChildCursorArray[0].State &&
      RelationArray[0]->CloseCursor(Handle->ChildCursorArray[0]) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADCURSOR, "Unable to close cursor");
    return XXX_NOT_OK;
  }
  return XXX_OK;
}

XxxErrCode RelFilterPlanOp::GetNextRecord(XxxSetCursor& SCsr,
					  XxxRelation* *RelationArray,
					  const XxxRecordSchema* Schema,
					  XxxRecord*& ReturnRec) const
{
  XxxDerivedRelation::CursorInfo* Handle = 
    (XxxDerivedRelation::CursorInfo*)(SCsr.State);
  if (!Handle) {
    REL_NOTE_ERROR(REL_BADEXEC, "NULL Exec handle");
    return XXX_NOT_OK;
  }

  while (1) {
    // get the next record
    if (RelationArray[0]->NextItem(Handle->ChildCursorArray[0],
				   Handle->ChildRecs[0]) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADEXEC, "Error getting NextItem from relation");
      return XXX_NOT_OK;
    }
    if (!Handle->ChildCursorArray[0].IsValid()) {
      // The scan is complete
      SCsr.AtEnd = XXX_TRUE;
      ReturnRec = 0;
      return XXX_OK;
    }
    // note that if this is the inner of a NL join, ReturnRec is
    // actually a reference to Handle->SiblingRecs[NLJoinPos]
    ReturnRec = Handle->ChildRecs[0];
    
    if (!SelPredPlan) {
      break; // out of while (1) loop
    }
    // now apply any conditions that exist
    XxxCoreADTClass::BoolValue CondResult(XXX_TRUE);
    if (!Handle->SiblingRecs || !Handle->SiblingSchema) {
      // in this case, simply evaluate the filter assuming a single input
      if (SelPredPlan->Evaluate(ReturnRec, 
				CatalogInfo->RecordSchema(),
				&CondResult, 
				(Handle->Params?Handle->Params->Params:0))
	  != XXX_OK) {
	XXX_NOTE_ERROR(XXX_BADEXPR, "Cannot evaluate filter expression");
	return XXX_NOT_OK;
      }
    } // if there is no "join environment"
    else {
      if (SelPredPlan->Evaluate(Handle->SiblingRecs,
				Handle->SiblingSchema,
				&CondResult,
				(Handle->Params?Handle->Params->Params:0))
	  != XXX_OK) {
	XXX_NOTE_ERROR(XXX_BADEXPR, "Cannot evaluate filter expression");
	return XXX_NOT_OK;
      }
    }
    if (CondResult.Value == XXX_TRUE) {
      // break out of while loop and return valid record
      break;
    }
    else {
      delete ReturnRec;
      // reset the ReturnRecord: this is important!!
      ReturnRec = NULL;
      // continue in the loop
    }
  }
  SCsr.NumberRead++;
  return XXX_OK;
}

void RelFilterPlanOp::PrintThis(ostream& OStr) const
{
  OStr << PrintString();
}

