#ifndef REL_GBYPLANOP_H
#define REL_GBYPLANOP_H

#include "relplanop.h"
#include "aggrexpr.h"

/**
 * Plan Op for aggregates and grouping
 */
class RelGrpByPlanOp : public RelDerivedPlanOp {
 protected:
  XxxValExprPlanList* TargetExprPlanList;
  XxxValExprPlanList* AggrExprPlanList;
  XxxValExprPlanList* GroupExprPlanList;

  class ScanInfo { public:
    XxxRecord *LastRec;
    int NumAggrExprs;
    XxxAggrOpHandle **OpHandles;
    ScanInfo(XxxValExprPlanList* AList, XxxErrCode& rc);
    ~ScanInfo();
  };
  friend class ScanInfo;

  static XxxFuncExpression::AggrPlan *getAggrPlan(XxxValueExprPlan *Plan);

  XxxErrCode aggregatePostProcess(const XxxFuncExpression::FuncPlan* CurPlan, 
				  const XxxRecord **Rec, 
				  const XxxRecordSchema **RecSchema,
				  XxxADTValue *AggValue, int AggSize,
				  XxxADTValue *ReturnValue) const; 

  XxxBool sameGroup (const XxxRecord *LastRec, 
		     const XxxRecord *NextRec,
		     const XxxRecordSchema *Schema,
		     const XxxValueEnv *Env,
		     XxxValExprPlanList *TargetList) const;//--> paskin 1/24/97
  XxxErrCode 
    aggregateRecord (XxxRecord *Record,
		     XxxValExprPlanList *AggrExprPlanList,
		     XxxDerivedRelation::CursorInfo* Handle,
		     XxxAggrOpHandle** OpHandles) const;//--> paskin 1/24/97

  XxxErrCode generateRecValues(XxxAggrOpHandle** OpHandles,
			       const XxxRecordSchema* Schema, 
			       XxxRecord *LastRec,
			       const XxxRecordSchema* LastRecSchema, 
			       XxxValuesArray& ValArray) const;
  XxxErrCode generateAggrValue(XxxValueExprPlan* CurPlan, 
			       XxxFuncExpression::AggrPlan *AggrPlan,
			       XxxAggrOpHandle* OpHandle, 
			       const XxxRecordSchema* Schema, 
			       XxxRecord *LastRec,
			       const XxxRecordSchema* LastRecSchema,
			       XxxADTValue* Result) const;
  XxxErrCode generateScalarValue(XxxValueExprPlan* CurPlan, 
				 int CurColIndex, 
				 const XxxRecordSchema* Schema, 
				 XxxRecord *LastRec,
				 const XxxRecordSchema* LastRecSchema,
				 XxxADTValue* Result) const;
  XxxErrCode initAggrExprs(XxxAggrOpHandle** OpHandles) const;

 public:
  virtual void PrintThis(ostream& OStr)  const;
  virtual const char *PrintString()  const { return "GRPBY" ;}
  virtual XxxErrCode InitHandle(XxxDerivedRelation::CursorInfo* Handle,
				XxxRelation* *RelationArray) const;
  virtual XxxErrCode CloseHandle(XxxDerivedRelation::CursorInfo* Handle,
				 XxxRelation* *RelationArray) const;
  virtual XxxErrCode GetNextRecord(XxxSetCursor& SCsr,
				   XxxRelation* *RelationArray,
				   const XxxRecordSchema* Schema,
				   XxxRecord*& ReturnRec) const;
  RelGrpByPlanOp(RelPlanOp* Child,
		 XxxValExprPlanList* TargetList,
		 XxxValExprPlanList* AggrList,
		 XxxValExprPlanList* GroupList,
		 RelCatalogInfo* CatInfo);
  virtual ~RelGrpByPlanOp() ;
  virtual void NodePrint (ostream &OStr, int curBlockId, int Level, 
			  int numChildren, int *childrenIds) const;
 };

#endif
