#ifdef __GNUG__
#pragma implementation
#endif

#include "relops.h"
#include "reldata.h"
#include "releng.h"
#include "planspj.h"
#include "funcexpr.h"
#include "shoresm.h"
#include "hfewplanop.h"

/*
  FEW-join (Hash Impl)
*/


/*
  Methods of RelHFEWPlanOp
*/

/*
  Methods of the buckets in the join buffers
*/


void RelHFEWPlanOp::LargeBucket::AdvanceBucket(Group *eGroup, int* cnt, int w){
  
  // in this method delete expired tuples from top of
  // the bucket and add them to eGroup.
  struct timezone Temp1; 
  struct timeval ts1;
  gettimeofday(&ts1, &Temp1);
  chain->RemoveExpired(ts1, eGroup, cnt, w);
}




XxxErrCode RelHFEWPlanOp::LargeBucket::insertRec(XxxRecord *Rec){

  if(size >  MAX_HASH_SIZE)
    ASSERT(0);

  if (chain==NULL)
    // first time insertion
    chain = new BucketChain;
  ASSERT(chain);
  // check if hash table is overflow
  if (!BucketOverflow()){
    size++;
    if(OverallMaxSize < size)
      OverallMaxSize = size;
    struct timeval ts;
    Entry* E = new Entry(Rec, 0, ts);
    ASSERT(E!=NULL);
    chain->AddEntry(E);
  }
  else{
    // Must swap this tuple.
    BucketChain sGroup(0);
    Entry *SwapEntry;
    // add the new record to the swap group, 
    // with an updated time stamp
    struct timeval ts;
    Entry* E = new Entry(Rec, 0, ts);
    ASSERT(E!=NULL);
    E->swapped = 1;
    sGroup.AddEntry(E);
    
    TheSwapFile *sfile = SwapFile;
    if(sfile==NULL){
      sfile = new TheSwapFile();
    }
    
    // For each record in swap group, insert it into the swap file
    char *ExtraData=NULL;
    int SizeOfExtraData;

    sGroup.InitScan();
    SwapEntry = sGroup.GetNext();
    cerr << " Swapping these vlaues" << endl;
    while (sGroup.ScanIsValid()) {
      // Insert the bucket entry into the swap file
      // 1) append the other fields to the record
      // 
      cerr << SwapEntry->value << endl;
      SwapEntry->SwapTimeStamp = ts;
      SwapEntry->ConvertEntryToRawData(ExtraData, &SizeOfExtraData);
      ASSERT(ExtraData!=NULL);
      //-----------------------    
      // structure is :
      // DataSize + Data + ExtraData  
      char *DataArea = new char[sizeof(int)+SwapEntry->Rec->Size()+SizeOfExtraData];
      
      int recSize = SwapEntry->Rec->Size();
      int totalSize = recSize+SizeOfExtraData;
      memcpy(DataArea, (char *)&recSize,  sizeof(int));
      memcpy(DataArea+sizeof(int), SwapEntry->Rec->DataArea, recSize);
      memcpy(DataArea+sizeof(int)+recSize, ExtraData, SizeOfExtraData);
      
      sfile->InsertRecord(DataArea, totalSize);
      delete DataArea;

      delete ExtraData;
      ExtraData = NULL;
      SwapEntry = sGroup.GetNext();
    };
    // delete the swap group
    sGroup.FlushAllRecords(XXX_TRUE);
  }
  return XXX_OK;
}

// Retreive "cnt" tuples from the swap disk
XxxErrCode RelHFEWPlanOp::LargeBucket::FloatTuples(int cnt, BucketChain *fGroup){

  int currCnt=0;

  SwapFile->InitScan();

  while(currCnt<cnt){

    // extract the total data first
    //    char* DataArea = new char[TotalSize];
    int TotalSize = SwapFile->RecordSize;
    char* DataArea = new char[SwapFile->RecordSize];
    if(SwapFile->DeleteTopRecord(DataArea)==0)
      break;

    // Extract the hash table entry
    
    Entry *E = new Entry(DataArea, TotalSize);
    delete DataArea;
    E->swapped = 1; //????
    fGroup->AddEntry(E);
    currCnt++;
  }

  SwapFile->CloseScan();

  return XXX_OK;
}


// Delete the swap file
XxxErrCode RelHFEWPlanOp::LargeBucket::DeleteSwapFile(const XxxRecordSchema* RecSchema){

  // First scan the file and reclaim the records:
  // this step should be removed when you store large objects on disk
  if(SwapFile!=NULL){  
    delete SwapFile;
  }
  return XXX_OK;
}

XxxRecord* RelHFEWPlanOp::OutWindowLargeBucket::GetOldestRec(XxxRelation *TheRelation, 
							   XxxDerivedRelation::CursorInfo* Handle, 
							   int Index, XxxBool& Valid){ 

  if((chain==NULL)||(chain->IsEmpty())){
    
    XxxRecord* TmpRec = NULL;
    
    if (TheRelation->NextItem(Handle->ChildCursorArray[Index], 
			      Handle->ChildRecs[Index]) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADEXEC, "Error getting NextItem from relation");
      return NULL;
    }
    
    Valid = Handle->ChildCursorArray[Index].IsValid();
    if(Valid == XXX_TRUE){
      TmpRec = Handle->ChildRecs[Index];
      if(TmpRec!=NULL){
	insertRec(TmpRec);
      }
    }    
    return TmpRec;
  }
  else{
    Entry* E = chain->GetTop();
    return E->Rec;
  }
}
/*
  Methods of RelHFEWPlanOp
*/
/*****************************************
 *
 *		 Methods of the hash table
 *
 ******************************************/
/// The methods of the Shore Swap file

int RelHFEWPlanOp::HashTable::TheShoreSwapFile::GetNext(char* data){
  
  bool AtEnd;
  XxxErrCode rc = XXX_OK;
  w_rc_t RetVal;
  pin_i* Pin = 0;

  XXX_CALL((XXX_SHORE_RETFN(ScanIter->next(Pin, 0, AtEnd), RetVal)),
	   XXX_SHOREERR, "Couldn't scan the swapped hash table", rc);
  if(rc==XXX_NOT_OK) return 0;
  
  if (AtEnd) return 0;
  
  if (!XxxShoreSM->PinValid(Pin, SH)) {
    XXX_NOTE_ERROR(XXX_SHOREERR, "Invalid pin at swap hash table");
    return XXX_NOT_OK;
  }
  
  int TotalSize = Pin->body_size();

  ASSERT(TotalSize == RecordSize);

  if (XxxShoreSM->GetObjectPreAlloc(Pin, data, TotalSize, 0, TotalSize) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_SHOREERR, "Error getting pinned record");
    Pin->unpin();
    return XXX_NOT_OK;
  }

  // delete record from the swap file
  /* // Not here
     XxxShoreObjectId* ObjId = new XxxShoreObjectId(XxxShoreSM->PinObjectId(Pin));
     if (XxxShoreSM->DeleteRecord(XxxShoreSM->DefaultVol(), *ObjId) != XXX_OK) {
     REL_NOTE_ERROR (REL_BADTABLE, "Unable to delete record");
     return XXX_NOT_OK;
     }
     delete ObjId;
  */
  //--------------
  Pin->unpin();
  return TotalSize;
}

int RelHFEWPlanOp::HashTable::
TheShoreSwapFile::DeleteTopRecord(char* data){
  
  bool AtEnd;
  XxxErrCode rc = XXX_OK;
  w_rc_t RetVal;
  pin_i* Pin = 0;
  

  if(ScanIter->next(Pin, 0, AtEnd) != XXX_OK) {
    cerr << "Error in pinning next record from seap file" << endl;
    return 0;
  }
  
  if (AtEnd) return 0;
  
  if (!XxxShoreSM->PinValid(Pin, SH)) {
    cerr << "Invalid pin at swap hash table" << endl;
    return 0;
  }
  
  int TotalSize = Pin->body_size();

  ASSERT(TotalSize == RecordSize);

  if (XxxShoreSM->GetObjectPreAlloc(Pin, data, TotalSize, 0, TotalSize) != XXX_OK) {
    cerr << "Error getting pinned record" << endl;
    Pin->unpin();
    return 0;
  }

  // delete record from the swap file
  XxxShoreObjectId* ObjId = new XxxShoreObjectId(XxxShoreSM->PinObjectId(Pin));
  Pin->unpin();
  if (XxxShoreSM->DeleteRecord(XxxShoreSM->DefaultVol(), *ObjId) != XXX_OK) {
    cerr << "Unable to delete record" << endl;
    delete ObjId;
    return 0;
  }

  delete ObjId;

  return TotalSize;
}

//-----------------------------------------------------------------
/// The methods of the Swap file

int RelHFEWPlanOp::HashTable::TheSwapFile::GetNext(char* data){
  // read a record
  int retVal = 0;
  if(!feof(SFile)){
    retVal = fread(data, sizeof(char), RecordSize, SFile);
    // if valid return the record and set the offset to next
    // record
    if(retVal==0) return 0;
    ASSERT(retVal == RecordSize);
  };
  return retVal;
}

int RelHFEWPlanOp::HashTable::TheSwapFile::DeleteTopRecord(char* data){
  
  int retVal = 0;
  retVal = GetNext(data);
  if(!retVal)
    FileOffset = FileOffset + RecordSize;
  
  return retVal;

  // only set a position of next record
  // read a record
  /*
    int retVal = 0;
    if(!feof(SFile)){
    retVal = fread(data, sizeof(char), RecordSize, SFile);
    // if valid return the record and set the offset to next
    // record
    if(retVal==0) return 0;
    ASSERT(retVal == RecordSize);
    FileOffset = FileOffset + RecordSize;
    };
    return retVal;
  */
}

//----------------------------------------------------------------------------------
XxxErrCode RelHFEWPlanOp::HashTable::insertRec(XxxRecord *Rec, int v, const XxxRecordSchema* RecSchema){
  if(hashSize >  maxSize)
    ASSERT(0);

  // check if hash table is overflow
  if (!TableOverflow()){
    hashSize++;
    if(OverallMaxHashSize < hashSize)
      OverallMaxHashSize = hashSize;
    // append time stamp for each record in the hash table
    struct timezone Temp; 
    struct timeval ts;
    gettimeofday(&ts, &Temp); 
    Entry* E = new Entry(Rec, v, ts);
    ASSERT(E!=NULL);

    // First, get a free slot from the Free List
    HashDirectoryEntry& FreeList = TableDir[bucketCnt+1];
    int position = FreeList.BucketHead;
    FreeList.BucketHead = TableEntries[position].NextHashEntry;
    FreeList.size--;

    if(FreeList.BucketHead == -1)
      FreeList.BucketTail = -1;


    TableEntries[position].NextHashEntry=-1;
    TableEntries[position].PrevHashEntry=-1;
    TableEntries[position].NextBucketEntry=-1;
    TableEntries[position].PrevBucketEntry=-1;

    // Second, append it to the Hash List
    HashDirectoryEntry& HashList = TableDir[bucketCnt];
    if(HashList.BucketTail == -1){
      // Add First Entry
      HashList.BucketHead = position;
      HashList.BucketTail = position;
    }
    else{
      TableEntries[HashList.BucketTail].NextHashEntry=position;
      TableEntries[position].PrevHashEntry=HashList.BucketTail;
      HashList.BucketTail = position;
    }
    HashList.size++;

    // Third, append it to the bucket
    HashDirectoryEntry& TheBucket = TableDir[hash(v)];
    if(TheBucket.BucketTail == -1){
      // Add First Entry
      TheBucket.BucketHead = position;
      TheBucket.BucketTail = position;
    }
    else{
      TableEntries[TheBucket.BucketTail].NextBucketEntry=position;
      TableEntries[position].PrevBucketEntry=TheBucket.BucketTail;
      TheBucket.BucketTail = position;
    }
    TheBucket.size++;
    //------------------------------------
    TableEntries[position].BucketEntry = E;
  }
  else{
    cerr << " Overflow.............";
    ASSERT(XXX_FALSE);
  }// else, swap the tuple
  return XXX_OK;
}
/************************************************************************
 * RangeDelete: deletes all entries from Head up to and included the to entry
 * add them to eGroup
 ***********************************************************************/
void RelHFEWPlanOp::HashTable::RangeDelete(int to, int &DeletedCnt, const XxxRecordSchema* RecSchema){

  int From	= TableDir[bucketCnt].BucketHead;
  int curr = From;
  while(curr!=to){
    // Add to expired group
    int BucketNumber = hash(TableEntries[curr].BucketEntry->value);

    TableEntries[curr].BucketEntry->Rec->Reclaim(RecSchema);
      //eGroup->AddRecord(TableEntries[curr].BucketEntry->Rec);
    delete TableEntries[curr].BucketEntry;
    DeletedCnt++;

    // delete entry from the bucket list first
    HashDirectoryEntry& TheBucket = TableDir[BucketNumber];

    if(TableEntries[curr].NextBucketEntry==-1){
      // Deleteing Tail of a bucket
      TheBucket.BucketTail = TableEntries[curr].PrevBucketEntry;
      if(TheBucket.BucketTail!=-1)
	TableEntries[TheBucket.BucketTail].NextBucketEntry = -1;
      else
	// a single entry in the bucket
	TheBucket.BucketHead = -1;
    }
    else if(TableEntries[curr].PrevBucketEntry==-1){
      // Deleteing Head of a bucket
      TheBucket.BucketHead = TableEntries[curr].NextBucketEntry;
      if(TheBucket.BucketHead!=-1)
	TableEntries[TheBucket.BucketHead].PrevBucketEntry = -1;
      else
	// a single entry in the bucket
	TheBucket.BucketTail = -1;
    }
    else{
      // Deleting intermediate entry in the bucket
      int prev = TableEntries[curr].PrevBucketEntry;
      int next = TableEntries[curr].NextBucketEntry;
      TableEntries[prev].NextBucketEntry = next;
      TableEntries[next].PrevBucketEntry = prev;
    }
	
    TableEntries[curr].PrevBucketEntry=-1;
    TableEntries[curr].NextBucketEntry=-1;
    TheBucket.size--;
    //----------------------------------------------------------
    int NextCurrent = TableEntries[curr].NextHashEntry;
    //----------------------------------------------------------
    // Delete the entry from the hash list
    HashDirectoryEntry& HashList = TableDir[bucketCnt];
	
    if(TableEntries[curr].NextHashEntry==-1){
      // Deleteing Tail of a hash list
      HashList.BucketTail = TableEntries[curr].PrevHashEntry;
      if(HashList.BucketTail!=-1)
	TableEntries[HashList.BucketTail].NextHashEntry = -1;
      else
	// a single entry in the bucket
	HashList.BucketHead = -1;
    }
    else if(TableEntries[curr].PrevHashEntry==-1){
      // Deleteing Head of a bucket
      HashList.BucketHead = TableEntries[curr].NextHashEntry;
      if(HashList.BucketHead!=-1)
	TableEntries[HashList.BucketHead].PrevHashEntry = -1;
      else
	// a single entry in the bucket
	HashList.BucketTail = -1;
    }
    else{
      // Deleting intermediate entry in the bucket
      int prev = TableEntries[curr].PrevHashEntry;
      int next = TableEntries[curr].NextHashEntry;
      TableEntries[prev].NextHashEntry = next;
      TableEntries[next].PrevHashEntry = prev;
    }
	
    TableEntries[curr].PrevHashEntry=-1;
    TableEntries[curr].NextHashEntry=-1;
    HashList.size--;
    //----------------------------------------------------------
    // Add the entry to the free list
    HashDirectoryEntry& FreeList = TableDir[bucketCnt+1];
    if(FreeList.BucketHead==-1){
      // First entry in the free list
      FreeList.BucketHead = curr;
      FreeList.BucketTail = curr;
    }
    else{
      ASSERT(FreeList.BucketTail!=-1);
      // adding entry to the tail
      TableEntries[FreeList.BucketTail].NextHashEntry = curr;
      TableEntries[curr].PrevHashEntry = FreeList.BucketTail;
      FreeList.BucketTail = curr;
    }
    FreeList.size++;
    curr = NextCurrent;
  }// while

  if(curr==to){
    // delete current item
    // Add to expired group
    int BucketNumber = hash(TableEntries[curr].BucketEntry->value);

    TableEntries[curr].BucketEntry->Rec->Reclaim(RecSchema);
      //eGroup->AddRecord(TableEntries[curr].BucketEntry->Rec);
    delete TableEntries[curr].BucketEntry;
    DeletedCnt++;

    // delete entry from the bucket list first
    HashDirectoryEntry& TheBucket = TableDir[BucketNumber];

    if(TableEntries[curr].NextBucketEntry==-1){
      // Deleteing Tail of a bucket
      TheBucket.BucketTail = TableEntries[curr].PrevBucketEntry;
      if(TheBucket.BucketTail!=-1)
	TableEntries[TheBucket.BucketTail].NextBucketEntry = -1;
      else
	// a single entry in the bucket
	TheBucket.BucketHead = -1;
    }
    else if(TableEntries[curr].PrevBucketEntry==-1){
      // Deleteing Head of a bucket
      TheBucket.BucketHead = TableEntries[curr].NextBucketEntry;
      if(TheBucket.BucketHead!=-1)
	TableEntries[TheBucket.BucketHead].PrevBucketEntry = -1;
      else
	// a single entry in the bucket
	TheBucket.BucketTail = -1;
    }
    else{
      // Deleting intermediate entry in the bucket
      int prev = TableEntries[curr].PrevBucketEntry;
      int next = TableEntries[curr].NextBucketEntry;
      TableEntries[prev].NextBucketEntry = next;
      TableEntries[next].PrevBucketEntry = prev;
    }
	
    TableEntries[curr].PrevBucketEntry=-1;
    TableEntries[curr].NextBucketEntry=-1;
    TheBucket.size--;
    //----------------------------------------------------------
    int NextCurrent = TableEntries[curr].NextHashEntry;
    //----------------------------------------------------------
    // Delete the entry from the hash list
    HashDirectoryEntry& HashList = TableDir[bucketCnt];
	
    if(TableEntries[curr].NextHashEntry==-1){
      // Deleteing Tail of a hash list
      HashList.BucketTail = TableEntries[curr].PrevHashEntry;
      if(HashList.BucketTail!=-1)
	TableEntries[HashList.BucketTail].NextHashEntry = -1;
      else
	// a single entry in the bucket
	HashList.BucketHead = -1;
    }
    else if(TableEntries[curr].PrevHashEntry==-1){
      // Deleteing Head of a bucket
      HashList.BucketHead = TableEntries[curr].NextHashEntry;
      if(HashList.BucketHead!=-1)
	TableEntries[HashList.BucketHead].PrevHashEntry = -1;
      else
	// a single entry in the bucket
	HashList.BucketTail = -1;
    }
    else{
      // Deleting intermediate entry in the bucket
      int prev = TableEntries[curr].PrevHashEntry;
      int next = TableEntries[curr].NextHashEntry;
      TableEntries[prev].NextHashEntry = next;
      TableEntries[next].PrevHashEntry = prev;
    }
	
    TableEntries[curr].PrevHashEntry=-1;
    TableEntries[curr].NextHashEntry=-1;
    HashList.size--;
    //----------------------------------------------------------
    // Add the entry to the free list
    HashDirectoryEntry& FreeList = TableDir[bucketCnt+1];
    if(FreeList.BucketHead==-1){
      // First entry in the free list
      FreeList.BucketHead = curr;
      FreeList.BucketTail = curr;
    }
    else{
      ASSERT(FreeList.BucketTail!=-1);
      // adding entry to the tail
      TableEntries[FreeList.BucketTail].NextHashEntry = curr;
      TableEntries[curr].PrevHashEntry = FreeList.BucketTail;
      FreeList.BucketTail = curr;
    }
    FreeList.size++;
    curr = NextCurrent;
  }// delete last item
}


/* 
   Methods of the hash table
*/

//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX
// Searches for a matching tuples in a right hash table
void RelHFEWPlanOp::HashTable::searchRecInRightHT_2(int v, TheTimeStampVector* tupleV,
						  Group* ResG, Group* eGroup,
						  unsigned int window, TheTimeStampVector* LeftTSVec,
						  const XxxRecordSchema* RecSchema){
  return;
  /*
  
  int DeletedCnt=0;
  
  HashDirectoryEntry& TheBucket = TableDir[hash(v)];
  
  if(TheBucket.BucketHead==-1) // bucket is empty
    return; // No matching tuples
  
  else{// Bucket is not empty
    // Second Alternative: online refresh and refill
    // 1) Scan the memory part of the bucket
    // 2) Scan the disk part of the bucket
    
    // Iterate over the Bucket
    int retV;
    struct timeval sts;
    struct timeval ets;
    
    sts.tv_sec = tupleV->maxTS.tv_sec - window;
    sts.tv_usec = tupleV->maxTS.tv_usec;
    
    ets.tv_sec = tupleV->minTS.tv_sec + window;
    ets.tv_usec = tupleV->minTS.tv_usec;

    // Now check if tuples are within the left interval
    int CurrentEntry = TheBucket.BucketHead;

    while (CurrentEntry!=-1) {
      Entry  *E = TableEntries[CurrentEntry].BucketEntry;
      // get time stamp of right record
      struct timeval ts;

      if(E->Rec->TSVec->IsToLeftByWindowOf(LeftTSVec, window)){
	// Remove entry from the table ?????
	RangeDelete(CurrentEntry, eGroup, DeletedCnt);
	hashSize = hashSize - DeletedCnt;
	DeletedCnt = 0;
      }
      else{
	ts.tv_sec = E->Rec->TSVec->TheVec[0].tv_sec;
	ts.tv_usec = E->Rec->TSVec->TheVec[0].tv_usec;
	
	if(!TimeStampsGreater(sts,ts) &&
	   !TimeStampsGreater(ts,ets)){
	  if (E->value==v)
	    ResG->AddRecord(E->Rec);
	}
	else{
	  // tuples in memory are out of the left window
	  return;
	}
      }

      CurrentEntry = TableEntries[CurrentEntry].NextBucketEntry;
    } // CurrentEntry!=-1

    //---------------------------------
    // 
    // Check tuples in disk if any
    //---------------------------------
    int EmptySlots = maxSize - hashSize;

    TheSwapFile *SwapFile = TheBucket.SwapFile;
    //TheShoreSwapFile *SwapFile = table[hash(v)]->SwapFile;
	
    XxxBool DeleteRec = XXX_TRUE;

    if (SwapFile!=NULL){
      SwapFile->InitScan();
      while(1){
	int TotalSize = SwapFile->RecordSize;
	char* DataArea = new char[TotalSize];
	Entry *E = NULL;

	DeleteRec = XXX_TRUE;

	if(EmptySlots){
 	  int returnSize = 0;
	  // reinsert the tuple from disk into memory
	  if((returnSize=SwapFile->DeleteTopRecord(DataArea))==0){
	    free(DataArea);
	    break;
	  }
	  ASSERT(returnSize==TotalSize);
	  E = new Entry(DataArea, TotalSize);
	  free(DataArea);
	  
	  if(E->Rec->TSVec->IsToLeftByWindowOf(LeftTSVec, window));
	  else{ // Insert the floatted tuple in the corresponding hash table
	    insertRec(E->Rec, E->value, RecSchema);
	    DeleteRec = XXX_FALSE;
	  }
	}
	else{
	  if(SwapFile->GetNext(DataArea)==0){
	    free(DataArea);
	    break;
	  }
	  E = new Entry(DataArea, TotalSize);
	  free(DataArea);
	}
	
	// get time stamp of right record
	struct timeval ts;

	ts.tv_sec = E->Rec->TSVec->TheVec[0].tv_sec;
	ts.tv_usec = E->Rec->TSVec->TheVec[0].tv_usec;


	if(!TimeStampsGreater(sts,ts) &&
	   !TimeStampsGreater(ts,ets)){
	  if (E->value==v){
	    ResG->AddRecord(E->Rec);
	    DeleteRec = XXX_FALSE;
	  }
	}
	else{ // tuples in disk are out of the left window
	  if(DeleteRec) E->Rec->Reclaim(RecSchema);
	  delete E;
	  break;
	}

	if(DeleteRec)
	  E->Rec->Reclaim(RecSchema);

	delete E;
      } // while(1)
      SwapFile->CloseScan();
    } // 
  } // Bucket is not empty
  */
}
//XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX

//xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
// Searches for a matching tuples in a left hash table
void RelHFEWPlanOp::HashTable::searchRecInLeftHT_2(int v,  struct timeval* tupleTS, Group* ResG,
						 Group *eGroup, unsigned int window,
						 TheTimeStampVector* RightTSVec,
						 const XxxRecordSchema* RecSchema){
  return;
  /*
  int DeletedCnt=0;

  struct timeval rts;
  rts.tv_sec = tupleTS->tv_sec;
  rts.tv_usec = tupleTS->tv_usec;
  
  HashDirectoryEntry& TheBucket = TableDir[hash(v)];
  
  if(TheBucket.BucketHead==-1) // bucket is empty
    return; // No matching tuples

  else{// Bucket is not empty
    // Second Alternative: online refresh and refill
    // 1) Scan the memory part of the bucket
    // 2) Scan the disk part of the bucket

    // Iterate over the group
    int retV;

    struct timeval sts;
    struct timeval ets;
	
    // Now check if tuples are within the left interval
    int CurrentEntry = TheBucket.BucketHead;
	
    while (CurrentEntry!=-1) {

      Entry  *E = TableEntries[CurrentEntry].BucketEntry;

      if(E->Rec->TSVec->IsToLeftByWindowOf(RightTSVec, window)){
	// Remove entry from the table ?????
	RangeDelete(CurrentEntry, eGroup, DeletedCnt);
	hashSize = hashSize - DeletedCnt;
	DeletedCnt = 0;
      }
      else{
	sts.tv_sec = E->Rec->TSVec->maxTS.tv_sec - window;
	sts.tv_usec = E->Rec->TSVec->maxTS.tv_usec;
	
	ets.tv_sec = E->Rec->TSVec->minTS.tv_sec + window;
	ets.tv_usec = E->Rec->TSVec->minTS.tv_usec;
	
	if(!TimeStampsGreater(sts, rts) &&
	   !TimeStampsGreater(rts, ets) &&
	   (E->value==v))
	  ResG->AddRecord(E->Rec);
      }    
      CurrentEntry = TableEntries[CurrentEntry].NextBucketEntry;
    }

    //---------------------------------
    // 
    // Check tuples in disk if any
    //---------------------------------
    int EmptySlots = maxSize - hashSize;

    TheSwapFile *SwapFile = TheBucket.SwapFile;
    //TheShoreSwapFile *SwapFile = table[hash(v)]->SwapFile;

    XxxBool DeleteRec = XXX_TRUE;

    if (SwapFile!=NULL){
      // check swapped tuples as well
      SwapFile->InitScan();
      while(1){
	int TotalSize = SwapFile->RecordSize;
	char* DataArea = new char[TotalSize];
	Entry *E = NULL;
	XxxBool DeleteRec = XXX_TRUE;

	if(EmptySlots){
	  int returnSize = 0;
	  // reinsert the tuple from disk into memory
	  if((returnSize=SwapFile->DeleteTopRecord(DataArea))==0){
	    free(DataArea);
	    break;
	  }
	  ASSERT(returnSize==TotalSize);
	  E = new Entry(DataArea, TotalSize);
	  free(DataArea);
	  
	  if(E->Rec->TSVec->IsToLeftByWindowOf(RightTSVec, window));
	  else{ // Insert the floatted tuple in the corresponding hash table
	    insertRec(E->Rec, E->value, RecSchema);
	    DeleteRec = XXX_FALSE;
	  }	
	}
	else{
	  if(SwapFile->GetNext(DataArea)==0){
	    free(DataArea);
	    break;
	  }
	  E = new Entry(DataArea, TotalSize);
	  free(DataArea);
	}

	// compare the tuple
	// 
	sts.tv_sec = E->Rec->TSVec->maxTS.tv_sec - window;
	sts.tv_usec = E->Rec->TSVec->maxTS.tv_usec;
	
	ets.tv_sec = E->Rec->TSVec->minTS.tv_sec + window;
	ets.tv_usec = E->Rec->TSVec->minTS.tv_usec;

	// get time stamp of right record
	if(!TimeStampsGreater(sts, rts) &&
	   !TimeStampsGreater(rts, ets) &&
	   (E->value==v)){
	  ResG->AddRecord(E->Rec);
	  DeleteRec = XXX_FALSE;
	}
	else if(DeleteRec)
	  E->Rec->Reclaim(RecSchema);
	
	delete E;
      }// while(1)
      SwapFile->CloseScan();
    }// (SwapFile!=NULL)
  }// Bucket is not empty
  */
}

/************************************************************************
 * DeleteOldest: deletes oldest tuple from 
 ***********************************************************************/
void RelHFEWPlanOp::HashTable::DeleteOldest(const XxxRecordSchema* RecSchema){

  int curr = TableDir[bucketCnt].BucketHead;
  // delete current item (the oldest)
  // Add to expired group
  int BucketNumber = hash(TableEntries[curr].BucketEntry->value);
  
  TableEntries[curr].BucketEntry->Rec->Reclaim(RecSchema);
  delete TableEntries[curr].BucketEntry;
  
  // delete entry from the bucket list first
  HashDirectoryEntry& TheBucket = TableDir[BucketNumber];
  
  if(TableEntries[curr].NextBucketEntry==-1){
    // Deleteing Tail of a bucket
    TheBucket.BucketTail = TableEntries[curr].PrevBucketEntry;
    if(TheBucket.BucketTail!=-1)
      TableEntries[TheBucket.BucketTail].NextBucketEntry = -1;
    else
      // a single entry in the bucket
      TheBucket.BucketHead = -1;
  }
  else if(TableEntries[curr].PrevBucketEntry==-1){
    // Deleteing Head of a bucket
    TheBucket.BucketHead = TableEntries[curr].NextBucketEntry;
    if(TheBucket.BucketHead!=-1)
      TableEntries[TheBucket.BucketHead].PrevBucketEntry = -1;
    else
      // a single entry in the bucket
      TheBucket.BucketTail = -1;
  }
  else{
    // Deleting intermediate entry in the bucket
    int prev = TableEntries[curr].PrevBucketEntry;
    int next = TableEntries[curr].NextBucketEntry;
    TableEntries[prev].NextBucketEntry = next;
    TableEntries[next].PrevBucketEntry = prev;
  }
  
  TableEntries[curr].PrevBucketEntry=-1;
  TableEntries[curr].NextBucketEntry=-1;
  TheBucket.size--;
  //----------------------------------------------------------
  int NextCurrent = TableEntries[curr].NextHashEntry;
  //----------------------------------------------------------
  // Delete the entry from the hash list
  HashDirectoryEntry& HashList = TableDir[bucketCnt];
  
  if(TableEntries[curr].NextHashEntry==-1){
    // Deleteing Tail of a hash list
    HashList.BucketTail = TableEntries[curr].PrevHashEntry;
    if(HashList.BucketTail!=-1)
      TableEntries[HashList.BucketTail].NextHashEntry = -1;
    else
      // a single entry in the bucket
      HashList.BucketHead = -1;
  }
  else if(TableEntries[curr].PrevHashEntry==-1){
    // Deleteing Head of a bucket
    HashList.BucketHead = TableEntries[curr].NextHashEntry;
    if(HashList.BucketHead!=-1)
      TableEntries[HashList.BucketHead].PrevHashEntry = -1;
    else
      // a single entry in the bucket
      HashList.BucketTail = -1;
  }
  else{
    // Deleting intermediate entry in the bucket
    int prev = TableEntries[curr].PrevHashEntry;
    int next = TableEntries[curr].NextHashEntry;
    TableEntries[prev].NextHashEntry = next;
    TableEntries[next].PrevHashEntry = prev;
  }
  
  TableEntries[curr].PrevHashEntry=-1;
  TableEntries[curr].NextHashEntry=-1;
  HashList.size--;
  //----------------------------------------------------------
  // Add the entry to the free list
  HashDirectoryEntry& FreeList = TableDir[bucketCnt+1];
  if(FreeList.BucketHead==-1){
    // First entry in the free list
    FreeList.BucketHead = curr;
    FreeList.BucketTail = curr;
  }
  else{
    ASSERT(FreeList.BucketTail!=-1);
    // adding entry to the tail
    TableEntries[FreeList.BucketTail].NextHashEntry = curr;
    TableEntries[curr].PrevHashEntry = FreeList.BucketTail;
    FreeList.BucketTail = curr;
  }
  FreeList.size++;
  curr = NextCurrent;
}

void RelHFEWPlanOp::JoinInWindowBuffers(XxxDerivedRelation::CursorInfo* Handle, int v) const{

  HFEWJoinInfo* JoinInfo = ((HFEWJoinInfo *)(Handle->OperatorHandle));  
  JoinInfo->inner = (JoinInfo->Mode2_CurrentStream+1) % JoinInfo->StreamsCnt;  
  RecursiveNestedLoop(Handle, JoinInfo->inner, v);
}


// Searches for a matching tuples
void RelHFEWPlanOp::RecursiveNestedLoop(XxxDerivedRelation::CursorInfo* Handle, 
				      int CurrStrm, int v) const{
  
  HFEWJoinInfo* JoinInfo = ((HFEWJoinInfo *)(Handle->OperatorHandle));	  
  if(CurrStrm == JoinInfo->SkippedStream){ 
    // Completed a loop over all streams				  
    XxxCoreADTClass::BoolValue Satisfied(XXX_FALSE);
    // To re-use old code, temporarily use the Handle; get its current records
    XxxRecord *TempRec;
    for (int i=0;i<JoinInfo->StreamsCnt;i++){
      TempRec = Handle->ChildRecs[i];
      Handle->ChildRecs[i] = JoinInfo->CandidateForJoin[i];
      JoinInfo->CandidateForJoin[i] = TempRec;;
    }
    if (checkFinal(Handle, Satisfied) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADEXEC, "Error checking residual plan");
      return;
    }
    if (Satisfied.Value){
      XxxRecord *RetRec;
      createRecord(Handle, JoinInfo->RelationArray, JoinInfo->Schema, RetRec);
      JoinInfo->OutputGroup.AddRecord(RetRec);	  
    }
    // return handle to its original values
    for (int j=0;j<JoinInfo->StreamsCnt;j++){
      TempRec = Handle->ChildRecs[j];
      Handle->ChildRecs[j] = JoinInfo->CandidateForJoin[j];
      JoinInfo->CandidateForJoin[j] = TempRec;;
    }
    //	  cerr << FieldValue << endl;
  }
  else{
    HashTable* hTable = JoinInfo->InWindowBuffer[CurrStrm];
    HashTable::HashDirectoryEntry& TheBucket = hTable->TableDir[hTable->hash(v)];
    if(TheBucket.BucketHead==-1) // bucket is empty
      return; // No matching tuples

    int CurrentEntry = TheBucket.BucketTail;
    while (CurrentEntry!=-1) {
      Entry  *E = hTable->TableEntries[CurrentEntry].BucketEntry;
      XxxRecord *tempRec = E->Rec;
      if(E->value==v){
	JoinInfo->CandidateForJoin[CurrStrm] = tempRec;
	int inner = (CurrStrm+1) % JoinInfo->StreamsCnt;
	RecursiveNestedLoop(Handle, inner, v);
      }
      CurrentEntry = hTable->TableEntries[CurrentEntry].PrevBucketEntry;
    }//(CurrentEntry!=-1)
  }//InWindowBuffer[CurrStrm]->chain!=NULL
}

//-----------------------------------------------------
void RelHFEWPlanOp::PrintNodeInfo(ostream &OStr) const{

  RelSPJPlanOp::PrintNodeInfo(OStr);
  // now print any additional stuff
}

RelHFEWPlanOp::RelHFEWPlanOp(int NumChildren, 
			 RelPlanOp* Children[],
			 RelCatalogInfo* CatInfo, 
			 XxxValExprPlanList* ExprPlanList,
			 XxxBooleanExpressionPlan** SelPredPlans, 
			 XxxBooleanExpressionPlan** JoinPredPlans,
			 XxxBooleanExpressionPlan** ResidPredPlans, 
			 XxxBooleanExpressionPlan* InitPlan,
			 double LWSize, double RWSize): 
  RelSPJPlanOp(NumChildren, Children, CatInfo, ExprPlanList, 
	       SelPredPlans, JoinPredPlans, ResidPredPlans, InitPlan){
  
  LeftWindowSize = LWSize;   // The window size for left buffer
  RightWindowSize = RWSize;  // The window size for right buffer
}

XxxErrCode RelHFEWPlanOp::checkFinal(XxxDerivedRelation::CursorInfo* Handle,
				   XxxCoreADTClass::BoolValue& Satisfied) const 
{
  Satisfied.Value = XXX_TRUE;
  XxxBooleanExpressionPlan* PredPlan = NULL;
  for (int i = 0; i < Handle->ArrayLen; i++) {
    if (JoinPredPlans && JoinPredPlans[i])
      if (XxxBooleanExpressionPlan::MakeConjunction(PredPlan, JoinPredPlans[i]) != XXX_OK) {
	REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
	return XXX_NOT_OK;
      }
    if (ResidPredPlans && ResidPredPlans[i])
      if (XxxBooleanExpressionPlan::MakeConjunction(PredPlan, ResidPredPlans[i]) != XXX_OK) {
	REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
	return XXX_NOT_OK;
      }
  }
  if (PredPlan)
    return PredPlan->Evaluate((const XxxRecord **)Handle->ChildRecs, 
			      (const XxxRecordSchema**)Handle->ChildSchema, 
			      &Satisfied, Handle->Params?Handle->Params->Params:0);
  return XXX_OK;
}

XxxErrCode RelHFEWPlanOp::getNextItem(XxxRelation *TheRelation,
				    XxxDerivedRelation::CursorInfo* Handle,
				    int Index, XxxBool& Valid) const
{
  if (TheRelation->NextItem(Handle->ChildCursorArray[Index],
			    Handle->ChildRecs[Index]) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADEXEC, "Error getting NextItem from relation");
    return XXX_NOT_OK;
  }
  Valid = Handle->ChildCursorArray[Index].IsValid();
  return XXX_OK;
}


XxxErrCode RelHFEWPlanOp::joinGroups(XxxDerivedRelation::CursorInfo* Handle,
				   XxxRelation* *RelationArray,
				   const XxxRecordSchema* Schema,
				   HFEWJoinInfo *JoinInfo) const
{  
  return XXX_OK;
}

XxxErrCode RelHFEWPlanOp::NLJoinTuples(XxxDerivedRelation::CursorInfo* Handle,
				     XxxRelation* *RelationArray,
				     const XxxRecordSchema* Schema,			  
				     HFEWJoinInfo *JoinInfo,
				     XxxRecord *LeftRec,
				     XxxRecord *RightRec) const
{  
  return XXX_OK;
}

XxxErrCode RelHFEWPlanOp::parseJoinPlan(XxxDerivedRelation::CursorInfo* Handle, XxxValueExprPlan** AttrFuncP) const

{ 
  XxxBool success=XXX_FALSE;
  /* We consider multi-way hash window join only if :
   * 1) the join condition involves the streaming attribute, otherwise another type
   * of window join should be used. Why? becuase if using hash join on 
   * a fixed (non-streaming attribute) the number of collisions within
   * a bucket will be huge and O(1) of hashing will not be possible
   * 2) A join condition contains at least a single equality RelBooleanPlan
   * such as " S.col.fn() = R.col.fn() = Z.col.fn() = W.col.fn() 
   * (i.e. eqaulity over the same attribute), 
   * where fn() returns an integer value".  
   */

  for(int i=1;i<Handle->ArrayLen;i++){

    XxxBooleanExpressionPlan* JoinPlan = JoinPredPlans[i];
    ASSERT (JoinPlan);
    if(JoinPlan->BoolType == XXX_REL_BOOL_EXPR){
      XxxRelBooleanPlan* RJoinPlan = (XxxRelBooleanPlan *)JoinPlan;
      if(RJoinPlan->OperType == XXX_EQ_OP){
	// Check to make sure the plans are of the form 
	//	s.col.fn() = b.col.fn()
	XxxValueExprPlan *LeftPlan = RJoinPlan->SubPlans[0];
	XxxValueExprPlan *RightPlan = RJoinPlan->SubPlans[1];
	
	if ((RightPlan->ValType() == XxxValueExpression::FUNC_EXPR) &&
	    (LeftPlan->ValType() == XxxValueExpression::FUNC_EXPR)){
	  if((LeftPlan->BasicType() == XXX_INT)&&
	     (RightPlan->BasicType() == XXX_INT)){
	    XxxFuncExpression::FuncPlan* LeftFuncPlan = (XxxFuncExpression::FuncPlan*)LeftPlan;
	    XxxFuncExpression::FuncPlan* RightFuncPlan = (XxxFuncExpression::FuncPlan*)RightPlan;
	    // mak sure the owners are of type COLUMN_VAR
	    if((LeftFuncPlan->OwnerPlan->ValType()==XxxValueExpression::COLUMN_VAR) &&
	       (RightFuncPlan->OwnerPlan->ValType()==XxxValueExpression::COLUMN_VAR)){
	      // TODO: the column must be of stream type???
	      
	      // Extract the indices in the RelationArray of the inputs
	      
	      //	      LeftRelIndex[i] = ((XxxUnknownValue::UnknownValuePlan *)(LeftFuncPlan->OwnerPlan))->SrcChild;
	      //	      RightRelIndex[i] = ((XxxUnknownValue::UnknownValuePlan *)(RightFuncPlan->OwnerPlan))->SrcChild;
	      
	      //	      LeftAttrIndex[i] = ((XxxUnknownValue::UnknownValuePlan *)(LeftFuncPlan->OwnerPlan))->SrcAttrIndex;
	      //	      RightAttrIndex[i] = ((XxxUnknownValue::UnknownValuePlan *)(RightFuncPlan->OwnerPlan))->SrcAttrIndex;
	      
	      
	      // Extract the function plans of the attributes in the records
	      
	      //	      AttrFuncP[i] = LeftFuncPlan;
	      AttrFuncP[i-1] = RightFuncPlan;
	      if(i==(Handle->ArrayLen-1))
		AttrFuncP[i] = LeftFuncPlan;
	      success = XXX_TRUE;
	    }
	  }
	}
      }// if operator is XXX_EQ_OP
    }
  }
  /*
    else{
    if (((XxxBoolBooleanPlan*)JoinPlan)->OpType != XXX_AND_OP)
    return XXX_OK; // failure
    for (int i = 0; i < JoinPlan->Arity(); i++) {
    XxxBoolBooleanPlan* BoolJoinPlan = (XxxBoolBooleanPlan*)JoinPlan;
    XxxBooleanExpressionPlan* SubPlan = (XxxBooleanExpressionPlan*)(BoolJoinPlan->SubPlans[i]);
    XxxRelBooleanPlan* RSubPlan = (XxxRelBooleanPlan *)SubPlan;
    // only accept anded predicates
    ASSERT(SubPlan->BoolType == XXX_REL_BOOL_EXPR);
    if(RSubPlan->OperType == XXX_EQ_OP){
    // Check to make sure the plans are simple column references
    XxxValueExprPlan *LeftPlan = RSubPlan->SubPlans[0];
    XxxValueExprPlan *RightPlan = RSubPlan->SubPlans[1];
    
    if ((RightPlan->ValType() == XxxValueExpression::FUNC_EXPR) &&
    (LeftPlan->ValType() == XxxValueExpression::FUNC_EXPR)){
    if((LeftPlan->BasicType() == XXX_INT)&&(RightPlan->BasicType() == XXX_INT)){
    XxxFuncExpression::FuncPlan* LeftFuncPlan = (XxxFuncExpression::FuncPlan*)LeftPlan;
    XxxFuncExpression::FuncPlan* RightFuncPlan = (XxxFuncExpression::FuncPlan*)RightPlan;
    // mak sure the owners are of type COLUMN_VAR
    if((LeftFuncPlan->OwnerPlan->ValType()==XxxValueExpression::COLUMN_VAR) &&
    (RightFuncPlan->OwnerPlan->ValType()==XxxValueExpression::COLUMN_VAR)){
    
    // Extract the indices in the RelationArray of the inputs
    //
    //	      LeftRelIndex = ((XxxUnknownValue::UnknownValuePlan *)(LeftFuncPlan->OwnerPlan))->SrcChild;
    //	      RightRelIndex = ((XxxUnknownValue::UnknownValuePlan *)(RightFuncPlan->OwnerPlan))->SrcChild;
    //
    //	      LeftAttrIndex = ((XxxUnknownValue::UnknownValuePlan *)(LeftFuncPlan->OwnerPlan))->SrcAttrIndex;
    //	      RightAttrIndex = ((XxxUnknownValue::UnknownValuePlan *)(RightFuncPlan->OwnerPlan))->SrcAttrIndex;
    //
    // Extract the function plans of the attributes in the records
    
    //	      LeftAttrFuncP = LeftFuncPlan;
    //	      RightAttrFuncP = RightFuncPlan;
    
    success = XXX_TRUE;
    break;
    }
    }
    }
    }// if operator is XXX_EQ_OP
    }// for arity
    }
  */
  
  //  ASSERT(success);
  return XXX_OK;
}

XxxErrCode
RelHFEWPlanOp::createJoinInfo(XxxDerivedRelation::CursorInfo* Handle,
			    HFEWJoinInfo *&JoinInfo) const
{
  // Get the locations of the join columns

  XxxValueExprPlan* AttrFuncP[MAX_JOINED_STREAMS];

  if (parseJoinPlan(Handle, AttrFuncP) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADEXEC, "Attempt to separate join condition failed");
    return XXX_NOT_OK;
  }

  // Make the input groups
  JoinInfo = new HFEWJoinInfo(Handle->Params->TheWindowSec,
			    Handle->Params->TheWindowMicroSec,
			    Handle->ArrayLen,
			    AttrFuncP);

  return XXX_OK;
}

XxxErrCode RelHFEWPlanOp::GetFieldValue(const XxxRecord* Rec[], 
				      XxxRecord* NewRec, int* v, int Index,
				      XxxValueExprPlan* AttrFuncPlan,
				      const XxxRecordSchema* Schema[]) const
{

  ASSERT(Rec);
  // GetFieldValue works on tuples on Handle->ChildRecs only, so
  // must prpeare for the calling by moving the target record on Handle->ChildRecs[x]
  XxxRecord *SwapRec = Rec[Index];
  Rec[Index] = NewRec;
  XxxADTValue *Result = new XxxCoreADTClass::IntValue();
  XxxFuncExpression::FuncPlan* theFuncPlan= 
    (XxxFuncExpression::FuncPlan*)AttrFuncPlan;

  if(theFuncPlan->Evaluate(Rec, Schema, Result, NULL) != XXX_OK) {
    char buf[100];
    sprintf(buf, "Error evaluating list expression %d", 0);
    XXX_NOTE_ERROR(XXX_BADEXEC, buf);
    Rec[Index] = SwapRec;
    return XXX_NOT_OK;
  }

  *v =	((XxxCoreADTClass::IntValue *)Result)->Value;  
  delete Result;
  //  ASSERT((*Schema)[AttrIndex].Type == XXX_INT);
  /*
    XxxBasicType TypeId = (*Schema)[AttrIndex].Type;
    XxxADTMetaInfo* MInfo = (*Schema)[AttrIndex].TypeCatInfo;
    char *tField;
  
    if(Rec->GetField(AttrIndex, Schema, tField)!=XXX_OK){
    return XXX_NOT_OK;
    }
    *v =	((XxxCoreADTClass::IntValue *)tField)->Value;
    */
  Rec[Index] = SwapRec;
  return XXX_OK;
}

XxxErrCode 
RelHFEWPlanOp::InitHandle(XxxDerivedRelation::CursorInfo* Handle,
			XxxRelation* *RelationArray) const 
{
  if (Handle->ArrayLen < 2) {
    REL_NOTE_ERROR(REL_BADOP, "Merge-Window join works on 2 or more inputs only");
    return XXX_NOT_OK;
  }
  // Create the operator handle
  HFEWJoinInfo* JoinInfo;
  if (createJoinInfo(Handle, JoinInfo) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADEXEC, "Cannot create Merge-window join info");
    return XXX_NOT_OK;
  }
  Handle->OperatorHandle = (void *)JoinInfo;
  
  XxxCoreADTClass::BoolValue Satisfied(XXX_TRUE);
  if (InitPlan) {
    if (InitPlan->Evaluate((const XxxRecord **)Handle->ChildRecs, 
			   (const XxxRecordSchema **)Handle->ChildSchema, 
			   &Satisfied, (Handle->Params) ? Handle->Params->Params : 0)!= XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADEXPR, "Error evaluating expression!");
      return XXX_NOT_OK;
    }
  }
  if (!Satisfied.Value) {
    JoinInfo->Valid = XXX_FALSE;
    return XXX_OK;
  }
  // Initialize cursors for both inputs
  for (int i = 0; i < Handle->ArrayLen; i++) {
    XxxBooleanExpressionPlan* PredPlan = (SelPredPlans?SelPredPlans[i]:0);
    if (RelationArray[i]->InitCursor1(Handle->ChildCursorArray[i], PredPlan, NULL, NULL,
				      Handle->Params, Handle->ArrayLen, 0) != XXX_OK) {
      char buf[100];
      sprintf (buf, "Unable to open cursor on input %d", i);
      XXX_NOTE_ERROR(XXX_BADCURSOR, buf);
      return XXX_NOT_OK;
    }	
  }
  // ---- Initialization
  JoinInfo->RelationArray = RelationArray;
  JoinInfo->StreamsCnt = Handle->ArrayLen;
  JoinInfo->CurrentPeriod.SPeriod.tv_sec=0; JoinInfo->CurrentPeriod.SPeriod.tv_usec=0;
  JoinInfo->CurrentPeriod.EPeriod.tv_sec=0; JoinInfo->CurrentPeriod.EPeriod.tv_usec=0;
  //Initialize InWindowBuffers and OutWindowBuffers
  JoinInfo->IsMode1 = 1;
  JoinInfo->CurrentlyIn = 0;
  JoinInfo->StartingStream = 0;
  JoinInfo->CurrentStream = 0;
  JoinInfo->Mode2_CurrentStream = 0;  
  JoinInfo->IsMode1 = 1;
  JoinInfo->inner=0;
  JoinInfo->SkippedStream=0;
  JoinInfo->NewTupleAddedToInBuffers=0;
  JoinInfo->EndOfMode2Flag = 0;

  JoinInfo->InWindowBuffer = (HashTable **) malloc(sizeof(HashTable *)*JoinInfo->StreamsCnt);
  //  JoinInfo->InWindowBuffer = (InWindowLargeBucket **) malloc(sizeof(InWindowLargeBucket *)*JoinInfo->StreamsCnt);
  JoinInfo->OutWindowBuffer = (OutWindowLargeBucket **) malloc(sizeof(OutWindowLargeBucket *)*JoinInfo->StreamsCnt);
  for (int k=0; k < JoinInfo->StreamsCnt; k++){
    JoinInfo->InWindowBuffer[k] = new HashTable(MAX_HASH_SIZE, MAX_BUCKET_CNT);
    //    JoinInfo->InWindowBuffer[k] = new InWindowLargeBucket;
    JoinInfo->OutWindowBuffer[k] = new OutWindowLargeBucket;
  }
  return XXX_OK;
}

XxxErrCode RelHFEWPlanOp::CloseHandle(XxxDerivedRelation::CursorInfo* Handle,
				    XxxRelation* *RelationArray) const
{ 

  // Close cursors on all the inputs
  for (int i = 0; i < Handle->ArrayLen; i++) {
    if (Handle->ChildCursorArray[i].State && RelationArray[i]->CloseCursor(Handle->ChildCursorArray[i]) != XXX_OK) {
      char buf[100];
      sprintf(buf, "Unable to close sequence cursor %d", i);
      XXX_NOTE_ERROR(XXX_BADCURSOR, buf);
      return XXX_NOT_OK;
    }
  }
  
  //TODO: you must 
  // delete the tables and the non-empty groups
  HFEWJoinInfo* JoinInfo = ((HFEWJoinInfo *)(Handle->OperatorHandle));  
  
  JoinInfo->OutputGroup.FlushAllRecords(XXX_TRUE,(const XxxRecordSchema *)Handle->RecordSchema);  
  JoinInfo->ExpiredGroup.FlushAllRecords(XXX_TRUE);
  JoinInfo->FloatGroup.FlushAllRecords(XXX_TRUE);


  // TODO: delete the swap files
  //  JoinInfo->LeftHashTable.DeleteSwapFile((const XxxRecordSchema *)Handle->ChildSchema[0]);
  //  JoinInfo->LeftHashTable.DeleteSwapFile((const XxxRecordSchema *)Handle->ChildSchema[1]);

  delete ((HFEWJoinInfo *)(Handle->OperatorHandle));
  
  return XXX_OK;
}

// TODO: Looks like the input records never get reclaimed properly!!
XxxErrCode 
RelHFEWPlanOp::GetNextRecord(XxxSetCursor& SCsr, XxxRelation* *RelationArray,
			   const XxxRecordSchema* Schema, XxxRecord*& ReturnRec) const 

{
  // Get the HFEWJoinInfo from the operator handle
  XxxDerivedRelation::CursorInfo* Handle = (XxxDerivedRelation::CursorInfo*)(SCsr.State);
  if (!Handle) {
    REL_NOTE_ERROR(REL_BADEXEC, "NULL Exec handle");
    return XXX_NOT_OK;
  }  
  HFEWJoinInfo* JoinInfo = ((HFEWJoinInfo *)(Handle->OperatorHandle));

  JoinInfo->RelationArray = RelationArray;
  JoinInfo->Schema = (XxxRecordSchema *)Schema;

  Group & OutputGroup = JoinInfo->OutputGroup;  
  Group & ExpiredGroup = JoinInfo->ExpiredGroup;
  BucketChain & FloatGroup = JoinInfo->FloatGroup; 

  // If there is a valid scan on the OutputGroup, return its next tuple
  if (OutputGroup.ScanIsValid() == XXX_TRUE) {
    ReturnRec = OutputGroup.GetNext();
    if (OutputGroup.ScanIsValid() == XXX_TRUE)
      return XXX_OK;
  }
  // Empty the output group; all of its records have been returned
  OutputGroup.FlushAllRecords(XXX_FALSE);
  // If there are no further records to return and a scan has finished, return
  if (!(JoinInfo->Valid)) {
    SCsr.AtEnd = XXX_TRUE;
    ReturnRec = 0;
    return XXX_OK;
  }
  
  XxxBool LeftScanValid = JoinInfo->LeftScanValid, 
    RightScanValid = JoinInfo->RightScanValid;

  int IsNewRecord=0;	
  // check which mode you are in
  while(1){
    if(JoinInfo->IsMode1){
      // In Mode 1
      int i;
      JoinInfo->CurrentStream = JoinInfo->StartingStream;
      JoinInfo->CurrentlyIn = 0;      
      JoinInfo->CurrentPeriod.SPeriod.tv_sec=0; 
      JoinInfo->CurrentPeriod.SPeriod.tv_usec=0;
      JoinInfo->CurrentPeriod.EPeriod.tv_sec=0; 
      JoinInfo->CurrentPeriod.EPeriod.tv_usec=0;
      while(1){
	i = JoinInfo->CurrentStream;
	JoinInfo->StartRec[i] = 
	  JoinInfo->InWindowBuffer[i]->GetOldestRec(this,JoinInfo->OutWindowBuffer[i], 
						    JoinInfo->RelationArray[i], Handle, 
						    i, JoinInfo->StreamScanValid[i], 
						    JoinInfo->IsMode1, 
						    JoinInfo->NewTupleAddedToInBuffers,
						    IsNewRecord);
	if(JoinInfo->StartRec[i]!=NULL){
	  if(TimeStampsGreater(JoinInfo->CurrentPeriod.SPeriod, 
			       JoinInfo->StartRec[i]->TSVec->TheVec[0])){
	    // Oldest record has time stamp older than the period, 
	    // Delete this tuple from its hashTable, skip and get next from this stream
	    HashTable* hTable = JoinInfo->InWindowBuffer[i];
	    hTable->DeleteOldest((const XxxRecordSchema *)Handle->ChildSchema[i]);
	    hTable->hashSize = hTable->hashSize - 1;
	    //x Already deleted in DeleteOldest; 
	    //xxxxxxx JoinInfo->StartRec[i]->Reclaim((const XxxRecordSchema *)Handle->ChildSchema[i]);
	    JoinInfo->StartRec[i] = NULL;
	  }
	  else {
	    if(TimeStampsGreater(JoinInfo->StartRec[i]->TSVec->TheVec[0], 
				 JoinInfo->CurrentPeriod.EPeriod)){
	      // Oldest record has time stamp newer than the left of the period, 
	      // drop current period and start with a new period centered at this record
	      JoinInfo->StartingStream = JoinInfo->CurrentStream;
	      JoinInfo->CurrentPeriod.SPeriod.tv_sec = 
		JoinInfo->StartRec[JoinInfo->CurrentStream]->TSVec->TheVec[0].tv_sec - JoinInfo->TheWindow;
	      JoinInfo->CurrentPeriod.SPeriod.tv_usec = 
		JoinInfo->StartRec[JoinInfo->CurrentStream]->TSVec->TheVec[0].tv_usec;
	      JoinInfo->CurrentPeriod.EPeriod.tv_sec = 
		JoinInfo->StartRec[JoinInfo->CurrentStream]->TSVec->TheVec[0].tv_sec + JoinInfo->TheWindow;
	      JoinInfo->CurrentPeriod.EPeriod.tv_usec = 
		JoinInfo->StartRec[JoinInfo->CurrentStream]->TSVec->TheVec[0].tv_usec;
	      JoinInfo->CurrentlyIn = 1 << JoinInfo->CurrentStream;

	      JoinInfo->CurrentStream = (JoinInfo->CurrentStream+1) % JoinInfo->StreamsCnt;
	      while((JoinInfo->CurrentlyIn&(1 << JoinInfo->CurrentStream)))
		JoinInfo->CurrentStream = (JoinInfo->CurrentStream+1) % JoinInfo->StreamsCnt;
	    }
	    else{
	      // Oldest record within the period
	      // update the period
	      struct timeval SNewPeriod, ENewPeriod;
	      SNewPeriod.tv_sec  = JoinInfo->StartRec[JoinInfo->CurrentStream]->TSVec->TheVec[0].tv_sec - JoinInfo->TheWindow;
	      SNewPeriod.tv_usec = JoinInfo->StartRec[JoinInfo->CurrentStream]->TSVec->TheVec[0].tv_usec;
	      ENewPeriod.tv_sec  = JoinInfo->StartRec[JoinInfo->CurrentStream]->TSVec->TheVec[0].tv_sec + JoinInfo->TheWindow;
	      ENewPeriod.tv_usec = JoinInfo->StartRec[JoinInfo->CurrentStream]->TSVec->TheVec[0].tv_usec;
	      if(TimeStampsGreater(SNewPeriod, JoinInfo->CurrentPeriod.SPeriod)){
		JoinInfo->CurrentPeriod.SPeriod.tv_sec = SNewPeriod.tv_sec; 
		JoinInfo->CurrentPeriod.SPeriod.tv_usec = SNewPeriod.tv_usec; 
	      }
	      if(TimeStampsGreater(JoinInfo->CurrentPeriod.EPeriod, ENewPeriod)){
		JoinInfo->CurrentPeriod.EPeriod.tv_sec = ENewPeriod.tv_sec; 
		JoinInfo->CurrentPeriod.EPeriod.tv_usec = ENewPeriod.tv_usec; 
	      }
	      // if this stream is counted before then skip
	      if(!(JoinInfo->CurrentlyIn&(1 << JoinInfo->CurrentStream)))
		JoinInfo->CurrentlyIn = JoinInfo->CurrentlyIn + (1 << JoinInfo->CurrentStream);
	      if(JoinInfo->CurrentlyIn==((1<<JoinInfo->StreamsCnt) - 1)){
		JoinInfo->IsMode1 = 0;
		JoinInfo->EndOfMode2Flag = 0;
		break;
	      }
	      else{
		JoinInfo->CurrentStream = (JoinInfo->CurrentStream+1) % JoinInfo->StreamsCnt;
		while((JoinInfo->CurrentlyIn&(1 << JoinInfo->CurrentStream)))
		  JoinInfo->CurrentStream = (JoinInfo->CurrentStream+1) % JoinInfo->StreamsCnt;
	      }
	    }
	  }// Oldest tuple not older tha the period
	}//StartRec!=NULL
	else{
	  JoinInfo->CurrentStream = (JoinInfo->CurrentStream+1) % JoinInfo->StreamsCnt;
	  while((JoinInfo->CurrentlyIn&(1 << JoinInfo->CurrentStream)))
	    JoinInfo->CurrentStream = (JoinInfo->CurrentStream+1) % JoinInfo->StreamsCnt;
	}
      }//while(1)
    }// Mode 1
    
    if(!JoinInfo->IsMode1){ // Mode 2
      // at this point 
      if(JoinInfo->NewTupleAddedToInBuffers){
	int j = JoinInfo->StreamWithNewTuple;
	JoinInfo->Mode2_CurrentStream = j;
	JoinInfo->SkippedStream = j;
	HashTable* hTable = JoinInfo->InWindowBuffer[j];
	
	int curr = hTable->TableDir[hTable->bucketCnt].BucketHead;
	ASSERT(curr!=-1);
	int hValue = hTable->TableEntries[curr].BucketEntry->value;
	JoinInfo->CandidateForJoin[j] = hTable->TableEntries[curr].BucketEntry->Rec;

	// do a nested loop on all the cuurent tuples in the buffers
	JoinInWindowBuffers(Handle, hValue);
	JoinInfo->Mode2_CurrentStream=0;
	JoinInfo->NewTupleAddedToInBuffers = 0;
	if(!JoinInfo->OutputGroup.IsEmpty())
	  // to start next time at a new stream 
	  break; // get out of the while loop
      }

      int j;
      while(1){
	j = JoinInfo->Mode2_CurrentStream;
	JoinInfo->LastRec[j] = 
	  JoinInfo->OutWindowBuffer[j]->GetOldestRec(JoinInfo->RelationArray[j],
						     Handle,
						     j,
						     //(j > 1) ? 1:j, 
						     JoinInfo->StreamScanValid[j]); 
	if(JoinInfo->LastRec[j]!=NULL){
	  if(TimeStampsGreater(JoinInfo->LastRec[j]->TSVec->TheVec[0], 
			       JoinInfo->CurrentPeriod.EPeriod)){
	    // new record has time stamp newer than the period, 
	    //JoinInfo->Mode2_CurrentStream =  (JoinInfo->Mode2_CurrentStream+1) % JoinInfo->StreamsCnt;
	    JoinInfo->EndOfMode2Flag = JoinInfo->EndOfMode2Flag | (1<<j);
	    if(JoinInfo->EndOfMode2Flag==((1<<JoinInfo->StreamsCnt)-1)){
	      // all new tuples are newer than the period, return to Mode 1 
	      JoinInfo->IsMode1 = 1;
	      // delete oldest tuple in InWindowBuffer
	      unsigned int minStream=0;
	      for (int k=1;k<JoinInfo->StreamsCnt;k++){
		if(TimeStampsGreater(JoinInfo->StartRec[minStream]->TSVec->TheVec[0], 
				     JoinInfo->StartRec[k]->TSVec->TheVec[0]))
		  minStream = k;
	      }
	      // delete oldest tuple from its hashTable
	      HashTable* hTable = JoinInfo->InWindowBuffer[minStream];
	      hTable->DeleteOldest((const XxxRecordSchema *)Handle->ChildSchema[minStream]);
	      hTable->hashSize = hTable->hashSize - 1;
	      JoinInfo->StartingStream = minStream;
	      break;
	    }
	  }
	  else{
	    // new record is in the period, It can not be older
	    // insert the new tuple in its hash table
	    JoinInfo->OutWindowBuffer[j]->DeleteTopRec(XXX_FALSE, NULL);
	    int FieldValue;
	    GetFieldValue((const XxxRecord **)Handle->ChildRecs, JoinInfo->LastRec[j], &FieldValue, 
			  j, JoinInfo->AttrFuncP[j], (const XxxRecordSchema **)Handle->ChildSchema);
	    JoinInfo->InWindowBuffer[j]->insertRec(JoinInfo->LastRec[j], FieldValue, 
						   (const XxxRecordSchema *)Handle->ChildSchema[j]);
	    // perform the join to get the result
	    JoinInfo->SkippedStream = j;
	    JoinInfo->CandidateForJoin[j] = JoinInfo->LastRec[j];
	    JoinInWindowBuffers(Handle, FieldValue);
	    //JoinInWindowBuffers(Handle, JoinInfo->LastRec[j]);
	    if(!JoinInfo->OutputGroup.IsEmpty()){
	      // to start next time at a new stream 
	      JoinInfo->Mode2_CurrentStream =  (JoinInfo->Mode2_CurrentStream+1) % JoinInfo->StreamsCnt;
	      while((JoinInfo->EndOfMode2Flag&(1 << JoinInfo->Mode2_CurrentStream)))
		JoinInfo->Mode2_CurrentStream =  (JoinInfo->Mode2_CurrentStream+1) % JoinInfo->StreamsCnt;
	      break;
	    }
	  }// else
	}// LastRec!=NULL
	JoinInfo->Mode2_CurrentStream =  (JoinInfo->Mode2_CurrentStream+1) % JoinInfo->StreamsCnt;
	while((JoinInfo->EndOfMode2Flag&(1 << JoinInfo->Mode2_CurrentStream)))
	  JoinInfo->Mode2_CurrentStream =  (JoinInfo->Mode2_CurrentStream+1) % JoinInfo->StreamsCnt;
      }//while(1)
    }// ! IsMode1
    if(!JoinInfo->OutputGroup.IsEmpty())
      break;
  }//while(1)  
  

  JoinInfo->Valid = XXX_FALSE;
  for (int i=0; i<JoinInfo->StreamsCnt; i++)
    if(JoinInfo->StreamScanValid[i]){
      JoinInfo->Valid = XXX_TRUE;
      break;
    }      

  // If the join was successful, and the OutputGroup has been repopulated.
  // Return its first tuple.
  if (!OutputGroup.IsEmpty()) {
    OutputGroup.InitScan();
    ReturnRec = OutputGroup.GetNext();
    ASSERT (OutputGroup.ScanIsValid());
    return XXX_OK;
  }

  // Otherwise, one of the inputs' scans is complete and the join need not
  // proceed.  Return end-of-scan.
  SCsr.AtEnd = XXX_TRUE;
  ReturnRec = 0;
  return XXX_OK;
}
