#ifdef __GNUG__
#pragma implementation
#endif

#include "relops.h"
#include "reldata.h"
#include "releng.h"
#include "planspj.h"
#include "funcexpr.h"
#include "shoresm.h"
#include "nlbewplanop.h"

/*
  THE MULTI-WAY NL WINDOW JOIN
*/


/*
  Methods of RelNLBEWPlanOp
*/

/*
  Methods of the buckets in the join buffers
*/


void RelNLBEWPlanOp::LargeBucket::AdvanceBucket(Group *eGroup, int* cnt, int w){
  
  // in this method delete expired tuples from top of
  // the bucket and add them to eGroup.
  struct timezone Temp1; 
  struct timeval ts1;
  gettimeofday(&ts1, &Temp1);
  chain->RemoveExpired(ts1, eGroup, cnt, w);
}




XxxErrCode RelNLBEWPlanOp::LargeBucket::insertRec(XxxRecord *Rec){

  if(size >  MAX_HASH_SIZE)
    ASSERT(0);

  if (chain==NULL)
    // first time insertion
    chain = new BucketChain;
  ASSERT(chain);
  // check if hash table is overflow
  if (!BucketOverflow()){
    size++;
    if(OverallMaxSize < size)
      OverallMaxSize = size;
    struct timeval ts;
    Entry* E = new Entry(Rec, 0, ts);
    ASSERT(E!=NULL);
    chain->AddEntry(E);
  }
  else{
    // Must swap this tuple.
    BucketChain sGroup(0);
    Entry *SwapEntry;
    // add the new record to the swap group, 
    // with an updated time stamp
    struct timeval ts;
    Entry* E = new Entry(Rec, 0, ts);
    ASSERT(E!=NULL);
    E->swapped = 1;
    sGroup.AddEntry(E);
    
    TheSwapFile *sfile = SwapFile;
    if(sfile==NULL){
      sfile = new TheSwapFile();
    }
    
    // For each record in swap group, insert it into the swap file
    char *ExtraData=NULL;
    int SizeOfExtraData;

    sGroup.InitScan();
    SwapEntry = sGroup.GetNext();
    cerr << " Swapping these vlaues" << endl;
    while (sGroup.ScanIsValid()) {
      // Insert the bucket entry into the swap file
      // 1) append the other fields to the record
      // 
      cerr << SwapEntry->value << endl;
      SwapEntry->SwapTimeStamp = ts;
      SwapEntry->ConvertEntryToRawData(ExtraData, &SizeOfExtraData);
      ASSERT(ExtraData!=NULL);
      //-----------------------    
      // structure is :
      // DataSize + Data + ExtraData  
      char *DataArea = new char[sizeof(int)+SwapEntry->Rec->Size()+SizeOfExtraData];
      
      int recSize = SwapEntry->Rec->Size();
      int totalSize = recSize+SizeOfExtraData;
      memcpy(DataArea, (char *)&recSize,  sizeof(int));
      memcpy(DataArea+sizeof(int), SwapEntry->Rec->DataArea, recSize);
      memcpy(DataArea+sizeof(int)+recSize, ExtraData, SizeOfExtraData);
      
      sfile->InsertRecord(DataArea, totalSize);
      delete DataArea;

      delete ExtraData;
      ExtraData = NULL;
      SwapEntry = sGroup.GetNext();
    };
    // delete the swap group
    sGroup.FlushAllRecords(XXX_TRUE);
  }
  return XXX_OK;
}

// Retreive "cnt" tuples from the swap disk
XxxErrCode RelNLBEWPlanOp::LargeBucket::FloatTuples(int cnt, BucketChain *fGroup){

  int currCnt=0;

  SwapFile->InitScan();

  while(currCnt<cnt){

    // extract the total data first
    //    char* DataArea = new char[TotalSize];
    int TotalSize = SwapFile->RecordSize;
    char* DataArea = new char[SwapFile->RecordSize];
    if(SwapFile->DeleteTopRecord(DataArea)==0)
      break;

    // Extract the hash table entry
    
    Entry *E = new Entry(DataArea, TotalSize);
    delete DataArea;
    E->swapped = 1; //????
    fGroup->AddEntry(E);
    currCnt++;
  }

  SwapFile->CloseScan();

  return XXX_OK;
}


// Delete the swap file
XxxErrCode RelNLBEWPlanOp::LargeBucket::DeleteSwapFile(const XxxRecordSchema* RecSchema){

  // First scan the file and reclaim the records:
  // this step should be removed when you store large objects on disk
  if(SwapFile!=NULL){  
    delete SwapFile;
  }
  return XXX_OK;
}

void RelNLBEWPlanOp::UpdateMaxTimeStamps(TheTimeStampVector* TSVec, 
				       struct timeval ts, int currS) const{

  TSVec->TheVec[currS].tv_sec = ts.tv_sec;
  TSVec->TheVec[currS].tv_usec = ts.tv_usec;
  // recalculate the MinMaxTimeStamps
  TSVec->minTS.tv_sec = TSVec->TheVec[currS].tv_sec; 
  TSVec->minTS.tv_usec = TSVec->TheVec[currS].tv_usec; 
  
  int VecSize = TSVec->VecSize;
  for(int i=0; i < VecSize;i++){
    if(TimeStampsGreater(TSVec->minTS, TSVec->TheVec[i])){
      TSVec->minTS.tv_sec = TSVec->TheVec[i].tv_sec; 
      TSVec->minTS.tv_usec = TSVec->TheVec[i].tv_usec; 
    }
  }
}

void RelNLBEWPlanOp::JoinInWindowBuffers(XxxDerivedRelation::CursorInfo* Handle) const{

  NLBEWJoinInfo* JoinInfo = ((NLBEWJoinInfo *)(Handle->OperatorHandle));  
  JoinInfo->inner = (JoinInfo->CurrentStream+1) % JoinInfo->StreamsCnt;  
  RecursiveNestedLoop(Handle, JoinInfo->inner);

}


// Searches for a matching tuples
void RelNLBEWPlanOp::RecursiveNestedLoop(XxxDerivedRelation::CursorInfo* Handle, 
				       int CurrStrm) const{
  
  NLBEWJoinInfo* JoinInfo = ((NLBEWJoinInfo *)(Handle->OperatorHandle));	  
  if(CurrStrm == JoinInfo->SkippedStream){ 
    // Completed a loop over all streams				  
    XxxCoreADTClass::BoolValue Satisfied(XXX_FALSE);
    // To re-use old code, temporarily use the Handle; get its current records
    XxxRecord *TempRec;
    for (int i=0;i<JoinInfo->StreamsCnt;i++){
      TempRec = Handle->ChildRecs[i];
      Handle->ChildRecs[i] = JoinInfo->CandidateForJoin[i];
      JoinInfo->CandidateForJoin[i] = TempRec;;
    }
    if (checkFinal(Handle, Satisfied) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADEXEC, "Error checking residual plan");
      return;
    }
    if (Satisfied.Value){
      XxxRecord *RetRec;
      createRecord(Handle, JoinInfo->RelationArray, JoinInfo->Schema, RetRec);
      JoinInfo->OutputGroup.AddRecord(RetRec);	  
    }
    // return handle to its original values
    for (int j=0;j<JoinInfo->StreamsCnt;j++){
      TempRec = Handle->ChildRecs[j];
      Handle->ChildRecs[j] = JoinInfo->CandidateForJoin[j];
      JoinInfo->CandidateForJoin[j] = TempRec;;
    }
    //	  cerr << FieldValue << endl;
  }
  else{
    if(JoinInfo->InWindowBuffer[CurrStrm]->chain==NULL)
      return;
    JoinInfo->InWindowBuffer[CurrStrm]->InitBackwardScan();
    XxxRecord *tempRec = JoinInfo->InWindowBuffer[CurrStrm]->GetPrevious();
    while(JoinInfo->InWindowBuffer[CurrStrm]->ScanIsValid()){			  
      // Check if time stamp of tempRec lies within the current period
      if(TimeStampsGreater(JoinInfo->CurrentPeriod.SPeriod, 
			   tempRec->TSVec->TheVec[0])){
	// tempRec is older than than period, 
	// since the scan is from newest to oldest, this tuple and all that succeed it are also
	// older than the period
	// check if the old tuples should be dropped out of the window buffers
	//??????
	struct timeval tempTS;
	tempTS.tv_sec = tempRec->TSVec->TheVec[0].tv_sec+JoinInfo->TheWindow;
	tempTS.tv_usec = tempRec->TSVec->TheVec[0].tv_usec;
	
	if(TimeStampsGreater(JoinInfo->LastTuplesTimeStamps->minTS, tempTS)){
	  // drop this tuple and all the tuples which are elder than it in this stream
	  //	  cerr << "Before: "<< CurrStrm << ": " << JoinInfo->InWindowBuffer[CurrStrm]->size << endl;
	  JoinInfo->InWindowBuffer[CurrStrm]->DropFromHeadToTuple(tempRec,(const XxxRecordSchema *)Handle->ChildSchema[CurrStrm]);
	  //	  cerr << "After: "<< CurrStrm << ": " << JoinInfo->InWindowBuffer[CurrStrm]->size<<endl;
	  return;
	}
      } 
      else if(TimeStampsGreater(tempRec->TSVec->TheVec[0], 
				JoinInfo->CurrentPeriod.EPeriod)){
	// tempRec is newer than the period, skip this tuple and look for the next tuple
	// in this stream
	;
      } 
      else{
	// new record within the period
	// Update the current period
	struct timeval SNewPeriod, ENewPeriod;
	SNewPeriod.tv_sec  = tempRec->TSVec->TheVec[0].tv_sec - JoinInfo->TheWindow;
	SNewPeriod.tv_usec = tempRec->TSVec->TheVec[0].tv_usec;
	ENewPeriod.tv_sec  = tempRec->TSVec->TheVec[0].tv_sec + JoinInfo->TheWindow;
	ENewPeriod.tv_usec = tempRec->TSVec->TheVec[0].tv_usec;
	if(TimeStampsGreater(SNewPeriod, JoinInfo->CurrentPeriod.SPeriod)){
	  JoinInfo->CurrentPeriod.SPeriod.tv_sec = SNewPeriod.tv_sec; 
	  JoinInfo->CurrentPeriod.SPeriod.tv_usec = SNewPeriod.tv_usec; 
	}
	if(TimeStampsGreater(JoinInfo->CurrentPeriod.EPeriod, ENewPeriod)){
	  JoinInfo->CurrentPeriod.EPeriod.tv_sec = ENewPeriod.tv_sec; 
	  JoinInfo->CurrentPeriod.EPeriod.tv_usec = ENewPeriod.tv_usec; 
	}
	JoinInfo->CandidateForJoin[CurrStrm] = tempRec;
	/*int ChildNum = 1;if(CurrStrm==0){ChildNum = 0;}//???  */
	int inner = (CurrStrm+1) % JoinInfo->StreamsCnt;
	RecursiveNestedLoop(Handle, inner);
      } 
      tempRec = JoinInfo->InWindowBuffer[CurrStrm]->GetPrevious();
    }// while(JoinInfo->InWindowBuffer[CurrStrm]->ScanIsValid())
  }//InWindowBuffer[CurrStrm]->chain!=NULL
}
//xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
void RelNLBEWPlanOp::PrintNodeInfo(ostream &OStr) const{

  RelSPJPlanOp::PrintNodeInfo(OStr);
  // now print any additional stuff
}

RelNLBEWPlanOp::RelNLBEWPlanOp(int NumChildren, 
			 RelPlanOp* Children[],
			 RelCatalogInfo* CatInfo, 
			 XxxValExprPlanList* ExprPlanList,
			 XxxBooleanExpressionPlan** SelPredPlans, 
			 XxxBooleanExpressionPlan** JoinPredPlans,
			 XxxBooleanExpressionPlan** ResidPredPlans, 
			 XxxBooleanExpressionPlan* InitPlan,
			 double LWSize, double RWSize): 
  RelSPJPlanOp(NumChildren, Children, CatInfo, ExprPlanList, 
	       SelPredPlans, JoinPredPlans, ResidPredPlans, InitPlan){
  
  LeftWindowSize = LWSize;   // The window size for left buffer
  RightWindowSize = RWSize;  // The window size for right buffer
}

XxxErrCode RelNLBEWPlanOp::checkFinal(XxxDerivedRelation::CursorInfo* Handle,
				   XxxCoreADTClass::BoolValue& Satisfied) const 
{
  Satisfied.Value = XXX_TRUE;
  XxxBooleanExpressionPlan* PredPlan = NULL;
  for (int i = 0; i < Handle->ArrayLen; i++) {
    if (JoinPredPlans && JoinPredPlans[i])
      if (XxxBooleanExpressionPlan::MakeConjunction(PredPlan, JoinPredPlans[i]) != XXX_OK) {
	REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
	return XXX_NOT_OK;
      }
    if (ResidPredPlans && ResidPredPlans[i])
      if (XxxBooleanExpressionPlan::MakeConjunction(PredPlan, ResidPredPlans[i]) != XXX_OK) {
	REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
	return XXX_NOT_OK;
      }
  }
  if (PredPlan)
    return PredPlan->Evaluate((const XxxRecord **)Handle->ChildRecs, 
			      (const XxxRecordSchema**)Handle->ChildSchema, 
			      &Satisfied, Handle->Params?Handle->Params->Params:0);
  return XXX_OK;
}

XxxErrCode RelNLBEWPlanOp::getNextItem(XxxRelation *TheRelation,
				    XxxDerivedRelation::CursorInfo* Handle,
				    int Index, XxxBool& Valid) const
{
  if (TheRelation->NextItem(Handle->ChildCursorArray[Index],
			    Handle->ChildRecs[Index]) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADEXEC, "Error getting NextItem from relation");
    return XXX_NOT_OK;
  }
  Valid = Handle->ChildCursorArray[Index].IsValid();
  return XXX_OK;
}


XxxErrCode RelNLBEWPlanOp::joinGroups(XxxDerivedRelation::CursorInfo* Handle,
				   XxxRelation* *RelationArray,
				   const XxxRecordSchema* Schema,
				   NLBEWJoinInfo *JoinInfo) const
{  
  return XXX_OK;
}

XxxErrCode RelNLBEWPlanOp::NLJoinTuples(XxxDerivedRelation::CursorInfo* Handle,
				     XxxRelation* *RelationArray,
				     const XxxRecordSchema* Schema,			  
				     NLBEWJoinInfo *JoinInfo,
				     XxxRecord *LeftRec,
				     XxxRecord *RightRec) const
{  
  return XXX_OK;
}



XxxErrCode RelNLBEWPlanOp::parseJoinPlan(XxxDerivedRelation::CursorInfo* Handle,
				      XxxValueExprPlan* &LeftAttrFuncP, 
				      int &LeftAttrIndex,
				      int &LeftRelIndex,
				      XxxValueExprPlan* &RightAttrFuncP, 
				      int &RightAttrIndex,
				      int &RightRelIndex) const

{ 
  return XXX_OK;
}


XxxErrCode 
RelNLBEWPlanOp::createJoinInfo(XxxDerivedRelation::CursorInfo* Handle,
			    NLBEWJoinInfo *&JoinInfo) const
{
  // Get the locations of the join columns
  XxxValueExprPlan *LeftAttrFuncP, *RightAttrFuncP;
  int LeftRelIndex, RightRelIndex;
  int LeftAttrIndex, RightAttrIndex;

  if (parseJoinPlan(Handle, LeftAttrFuncP, LeftAttrIndex, LeftRelIndex,
		    RightAttrFuncP, RightAttrIndex, RightRelIndex) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADEXEC,
		   "Attempt to separate join condition failed");
    return XXX_NOT_OK;
  }

  // Make the input groups
  JoinInfo = new NLBEWJoinInfo(RightRelIndex, LeftRelIndex, RightAttrIndex, 
			    LeftAttrIndex, RightAttrFuncP, LeftAttrFuncP,
			    Handle->Params->TheWindowSec,
			    Handle->Params->TheWindowMicroSec);

  return XXX_OK;
}

XxxErrCode 
RelNLBEWPlanOp::InitHandle(XxxDerivedRelation::CursorInfo* Handle,
			XxxRelation* *RelationArray) const 
{
  if (Handle->ArrayLen < 2) {
    REL_NOTE_ERROR(REL_BADOP, "Merge-Window join works on 2 or more inputs only");
    return XXX_NOT_OK;
  }
  // Create the operator handle
  NLBEWJoinInfo* JoinInfo;
  if (createJoinInfo(Handle, JoinInfo) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADEXEC, "Cannot create Merge-window join info");
    return XXX_NOT_OK;
  }
  Handle->OperatorHandle = (void *)JoinInfo;
  
  XxxCoreADTClass::BoolValue Satisfied(XXX_TRUE);
  if (InitPlan) {
    if (InitPlan->Evaluate((const XxxRecord **)Handle->ChildRecs, 
			   (const XxxRecordSchema **)Handle->ChildSchema, 
			   &Satisfied, (Handle->Params) ? Handle->Params->Params : 0)!= XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADEXPR, "Error evaluating expression!");
      return XXX_NOT_OK;
    }
  }
  if (!Satisfied.Value) {
    JoinInfo->Valid = XXX_FALSE;
    return XXX_OK;
  }
  // Initialize cursors for both inputs
  for (int i = 0; i < Handle->ArrayLen; i++) {
    XxxBooleanExpressionPlan* PredPlan = (SelPredPlans?SelPredPlans[i]:0);
    if (RelationArray[i]->InitCursor1(Handle->ChildCursorArray[i], PredPlan, NULL, NULL,
				      Handle->Params, Handle->ArrayLen, 0) != XXX_OK) {
      char buf[100];
      sprintf (buf, "Unable to open cursor on input %d", i);
      XXX_NOTE_ERROR(XXX_BADCURSOR, buf);
      return XXX_NOT_OK;
    }	
  }
  // ---- Initialization
  JoinInfo->RelationArray = RelationArray;
  JoinInfo->StreamsCnt = Handle->ArrayLen;
  JoinInfo->CurrentPeriod.SPeriod.tv_sec=0; JoinInfo->CurrentPeriod.SPeriod.tv_usec=0;
  JoinInfo->CurrentPeriod.EPeriod.tv_sec=0; JoinInfo->CurrentPeriod.EPeriod.tv_usec=0;
  JoinInfo->LastTuplesTimeStamps = new TheTimeStampVector(JoinInfo->StreamsCnt);
  //Initialize InWindowBuffers and OutWindowBuffers
  JoinInfo->IsMode1 = 1;
  JoinInfo->CurrentlyIn = 0;
  JoinInfo->StartingStream = 0;
  JoinInfo->CurrentStream = 0;
  JoinInfo->Mode2_CurrentStream = 0;  
  JoinInfo->IsMode1 = 1;
  JoinInfo->inner=0;
  JoinInfo->SkippedStream=0;
  JoinInfo->NewTupleAddedToInBuffers=0;
  JoinInfo->EndOfMode2Flag = 0;

  JoinInfo->InWindowBuffer = (InWindowLargeBucket **) malloc(sizeof(InWindowLargeBucket *)*JoinInfo->StreamsCnt);
  for (int k=0; k < JoinInfo->StreamsCnt; k++){
    JoinInfo->InWindowBuffer[k] = new InWindowLargeBucket;
  }
  return XXX_OK;
}

XxxErrCode RelNLBEWPlanOp::CloseHandle(XxxDerivedRelation::CursorInfo* Handle,
				    XxxRelation* *RelationArray) const
{ 

  // Close cursors on all the inputs
  for (int i = 0; i < Handle->ArrayLen; i++) {
    if (Handle->ChildCursorArray[i].State && RelationArray[i]->CloseCursor(Handle->ChildCursorArray[i]) != XXX_OK) {
      char buf[100];
      sprintf(buf, "Unable to close sequence cursor %d", i);
      XXX_NOTE_ERROR(XXX_BADCURSOR, buf);
      return XXX_NOT_OK;
    }
  }
  
  //TODO: you must 
  // delete the tables and the non-empty groups
  NLBEWJoinInfo* JoinInfo = ((NLBEWJoinInfo *)(Handle->OperatorHandle));  
  
  JoinInfo->OutputGroup.FlushAllRecords(XXX_TRUE,(const XxxRecordSchema *)Handle->RecordSchema);  
  JoinInfo->ExpiredGroup.FlushAllRecords(XXX_TRUE);
  JoinInfo->FloatGroup.FlushAllRecords(XXX_TRUE);


  // TODO: delete the swap files
  //  JoinInfo->LeftHashTable.DeleteSwapFile((const XxxRecordSchema *)Handle->ChildSchema[0]);
  //  JoinInfo->LeftHashTable.DeleteSwapFile((const XxxRecordSchema *)Handle->ChildSchema[1]);

  delete ((NLBEWJoinInfo *)(Handle->OperatorHandle));
  
  return XXX_OK;
}

// TODO: Looks like the input records never get reclaimed properly!!
XxxErrCode 
RelNLBEWPlanOp::GetNextRecord(XxxSetCursor& SCsr, XxxRelation* *RelationArray,
			   const XxxRecordSchema* Schema, XxxRecord*& ReturnRec) const 

{
  // Get the NLBEWJoinInfo from the operator handle
  XxxDerivedRelation::CursorInfo* Handle = (XxxDerivedRelation::CursorInfo*)(SCsr.State);
  if (!Handle) {
    REL_NOTE_ERROR(REL_BADEXEC, "NULL Exec handle");
    return XXX_NOT_OK;
  }  
  NLBEWJoinInfo* JoinInfo = ((NLBEWJoinInfo *)(Handle->OperatorHandle));

  JoinInfo->RelationArray = RelationArray;
  JoinInfo->Schema = (XxxRecordSchema *)Schema;

  Group & OutputGroup = JoinInfo->OutputGroup;  
  Group & ExpiredGroup = JoinInfo->ExpiredGroup;
  BucketChain & FloatGroup = JoinInfo->FloatGroup; 

  // If there is a valid scan on the OutputGroup, return its next tuple
  if (OutputGroup.ScanIsValid() == XXX_TRUE) {
    ReturnRec = OutputGroup.GetNext();
    if (OutputGroup.ScanIsValid() == XXX_TRUE)
      return XXX_OK;
  }
  // Empty the output group; all of its records have been returned
  OutputGroup.FlushAllRecords(XXX_FALSE);
  // If there are no further records to return and a scan has finished, return
  if (!(JoinInfo->Valid)) {
    SCsr.AtEnd = XXX_TRUE;
    ReturnRec = 0;
    return XXX_OK;
  }
  
  XxxBool LeftScanValid = JoinInfo->LeftScanValid, 
    RightScanValid = JoinInfo->RightScanValid;
  ///--------------------------------------------------------------------
  JoinInfo->CurrentStream = JoinInfo->StartingStream;
  JoinInfo->CurrentlyIn = 0;      
  JoinInfo->CurrentPeriod.SPeriod.tv_sec=0; 
  JoinInfo->CurrentPeriod.SPeriod.tv_usec=0;
  JoinInfo->CurrentPeriod.EPeriod.tv_sec=0; 
  JoinInfo->CurrentPeriod.EPeriod.tv_usec=0;

  while(1){
    // Seek a new tuple in any of the streams
    int i = JoinInfo->CurrentStream;
    XxxRecord* tempRec = JoinInfo->InWindowBuffer[i]->GetNewTopRec(JoinInfo->RelationArray[i], 
								   Handle, i, 
								   JoinInfo->StreamScanValid[i]);
    
    if(tempRec!=NULL){
      JoinInfo->Cnt[i]++;
      // Found new tuple in one of the streams
      // 1) Build a period centered at this tuple
      JoinInfo->StartingStream = JoinInfo->CurrentStream;
      JoinInfo->CurrentPeriod.SPeriod.tv_sec = tempRec->TSVec->TheVec[0].tv_sec - JoinInfo->TheWindow;
      JoinInfo->CurrentPeriod.SPeriod.tv_usec = tempRec->TSVec->TheVec[0].tv_usec;
      JoinInfo->CurrentPeriod.EPeriod.tv_sec = tempRec->TSVec->TheVec[0].tv_sec + JoinInfo->TheWindow;
      JoinInfo->CurrentPeriod.EPeriod.tv_usec = tempRec->TSVec->TheVec[0].tv_usec;
      //2) Update the MinMaxTimeStamps of the tuples (used to drop old tuples)
      UpdateMaxTimeStamps(JoinInfo->LastTuplesTimeStamps, 
			  tempRec->TSVec->TheVec[0], JoinInfo->CurrentStream);
      //3) Iterate on the rest of the streams
      JoinInfo->SkippedStream = JoinInfo->CurrentStream;
      JoinInfo->CandidateForJoin[JoinInfo->SkippedStream] = tempRec;
      JoinInWindowBuffers(Handle);
      if(!JoinInfo->OutputGroup.IsEmpty()){
	// to start next time at a new stream 
	JoinInfo->StartingStream = (JoinInfo->CurrentStream+1) % JoinInfo->StreamsCnt;
	break;
      }
    }
    JoinInfo->CurrentStream = (JoinInfo->CurrentStream+1) % JoinInfo->StreamsCnt;
  }


  JoinInfo->Valid = XXX_FALSE;
  for (int i=0; i<JoinInfo->StreamsCnt; i++)
    if(JoinInfo->StreamScanValid[i]){
      JoinInfo->Valid = XXX_TRUE;
      break;
    }      
  
  // If the join was successful, and the OutputGroup has been repopulated.
  // Return its first tuple.
  if (!OutputGroup.IsEmpty()) {
    OutputGroup.InitScan();
    ReturnRec = OutputGroup.GetNext();
    ASSERT (OutputGroup.ScanIsValid());
    return XXX_OK;
  }

  // Otherwise, one of the inputs' scans is complete and the join need not
  // proceed.  Return end-of-scan.
  SCsr.AtEnd = XXX_TRUE;
  ReturnRec = 0;
  return XXX_OK;
}
