#ifdef __GNUG__
#pragma implementation
#endif

#include "relops.h"
#include "reldata.h"
#include "releng.h"
#include "planspj.h"
#include "funcexpr.h"
#include "shoresm.h"
#include "nlfewplanop.h"

/*
  FEW-join (Nested Loop Impl)
*/


/*
  Methods of RelNLFEWPlanOp
*/

/*
  Methods of the buckets in the join buffers
*/


void RelNLFEWPlanOp::LargeBucket::AdvanceBucket(Group *eGroup, int* cnt, int w){
  
  // in this method delete expired tuples from top of
  // the bucket and add them to eGroup.
  struct timezone Temp1; 
  struct timeval ts1;
  gettimeofday(&ts1, &Temp1);
  chain->RemoveExpired(ts1, eGroup, cnt, w);
}




XxxErrCode RelNLFEWPlanOp::LargeBucket::insertRec(XxxRecord *Rec){

  if(size >  MAX_HASH_SIZE)
    ASSERT(0);

  if (chain==NULL)
    // first time insertion
    chain = new BucketChain;
  ASSERT(chain);
  // check if hash table is overflow
  if (!BucketOverflow()){
    size++;
    if(OverallMaxSize < size)
      OverallMaxSize = size;
    struct timeval ts;
    Entry* E = new Entry(Rec, 0, ts);
    ASSERT(E!=NULL);
    chain->AddEntry(E);
  }
  else{
    // Must swap this tuple.
    BucketChain sGroup(0);
    Entry *SwapEntry;
    // add the new record to the swap group, 
    // with an updated time stamp
    struct timeval ts;
    Entry* E = new Entry(Rec, 0, ts);
    ASSERT(E!=NULL);
    E->swapped = 1;
    sGroup.AddEntry(E);
    
    TheSwapFile *sfile = SwapFile;
    if(sfile==NULL){
      sfile = new TheSwapFile();
    }
    
    // For each record in swap group, insert it into the swap file
    char *ExtraData=NULL;
    int SizeOfExtraData;

    sGroup.InitScan();
    SwapEntry = sGroup.GetNext();
    cerr << " Swapping these vlaues" << endl;
    while (sGroup.ScanIsValid()) {
      // Insert the bucket entry into the swap file
      // 1) append the other fields to the record
      // 
      cerr << SwapEntry->value << endl;
      SwapEntry->SwapTimeStamp = ts;
      SwapEntry->ConvertEntryToRawData(ExtraData, &SizeOfExtraData);
      ASSERT(ExtraData!=NULL);
      //-----------------------    
      // structure is :
      // DataSize + Data + ExtraData  
      char *DataArea = new char[sizeof(int)+SwapEntry->Rec->Size()+SizeOfExtraData];
      
      int recSize = SwapEntry->Rec->Size();
      int totalSize = recSize+SizeOfExtraData;
      memcpy(DataArea, (char *)&recSize,  sizeof(int));
      memcpy(DataArea+sizeof(int), SwapEntry->Rec->DataArea, recSize);
      memcpy(DataArea+sizeof(int)+recSize, ExtraData, SizeOfExtraData);
      
      sfile->InsertRecord(DataArea, totalSize);
      delete DataArea;

      delete ExtraData;
      ExtraData = NULL;
      SwapEntry = sGroup.GetNext();
    };
    // delete the swap group
    sGroup.FlushAllRecords(XXX_TRUE);
  }
  return XXX_OK;
}

// Retreive "cnt" tuples from the swap disk
XxxErrCode RelNLFEWPlanOp::LargeBucket::FloatTuples(int cnt, BucketChain *fGroup){

  int currCnt=0;

  SwapFile->InitScan();

  while(currCnt<cnt){

    // extract the total data first
    //    char* DataArea = new char[TotalSize];
    int TotalSize = SwapFile->RecordSize;
    char* DataArea = new char[SwapFile->RecordSize];
    if(SwapFile->DeleteTopRecord(DataArea)==0)
      break;

    // Extract the hash table entry
    
    Entry *E = new Entry(DataArea, TotalSize);
    delete DataArea;
    E->swapped = 1; //????
    fGroup->AddEntry(E);
    currCnt++;
  }

  SwapFile->CloseScan();

  return XXX_OK;
}


// Delete the swap file
XxxErrCode RelNLFEWPlanOp::LargeBucket::DeleteSwapFile(const XxxRecordSchema* RecSchema){

  // First scan the file and reclaim the records:
  // this step should be removed when you store large objects on disk
  if(SwapFile!=NULL){  
    delete SwapFile;
  }
  return XXX_OK;
}

XxxRecord* RelNLFEWPlanOp::
OutWindowLargeBucket::GetTopRec(XxxRelation *TheRelation, 
				XxxDerivedRelation::CursorInfo* Handle, 
				int Index, XxxBool& Valid){ 

  if((chain==NULL)||(chain->IsEmpty())){
    
    XxxRecord* TmpRec = NULL;
    
    if (TheRelation->NextItem(Handle->ChildCursorArray[Index], 
			      Handle->ChildRecs[Index]) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADEXEC, "Error getting NextItem from relation");
      return NULL;
    }
    
    Valid = Handle->ChildCursorArray[Index].IsValid();
    if(Valid == XXX_TRUE){
      TmpRec = Handle->ChildRecs[Index];
      if(TmpRec!=NULL){
	insertRec(TmpRec);
	NLFEWJoinInfo* JoinInfo = ((NLFEWJoinInfo *)(Handle->OperatorHandle));
      }
    }    
    return TmpRec;
  }
  else{
    Entry* E = chain->GetTop();
    return E->Rec;
  }
};


void RelNLFEWPlanOp::JoinInWindowBuffers(XxxDerivedRelation::CursorInfo* Handle) const{
  
  NLFEWJoinInfo* JoinInfo = ((NLFEWJoinInfo *)(Handle->OperatorHandle));

  JoinInfo->Mode2_CurrentStream = 0;
  JoinInfo->SkippedStream = JoinInfo->Mode2_CurrentStream;

  JoinInfo->InWindowBuffer[JoinInfo->Mode2_CurrentStream]->InitBackwardScan();	  
  JoinInfo->CandidateForJoin[JoinInfo->Mode2_CurrentStream] = 
    JoinInfo->InWindowBuffer[JoinInfo->Mode2_CurrentStream]->GetPrevious();
  while(JoinInfo->InWindowBuffer[JoinInfo->Mode2_CurrentStream]->ScanIsValid()){
    JoinInfo->inner = (JoinInfo->Mode2_CurrentStream+1) % JoinInfo->StreamsCnt;  
    RecursiveNestedLoop(Handle, JoinInfo->inner); // inner must be >=1
    JoinInfo->CandidateForJoin[JoinInfo->Mode2_CurrentStream] = 
      JoinInfo->InWindowBuffer[JoinInfo->Mode2_CurrentStream]->GetPrevious();		  
  }

}


void RelNLFEWPlanOp::JoinInWindowBuffers(XxxDerivedRelation::CursorInfo* Handle, XxxRecord *TmpRec) const{
	
  NLFEWJoinInfo* JoinInfo = ((NLFEWJoinInfo *)(Handle->OperatorHandle));  
  JoinInfo->inner = (JoinInfo->Mode2_CurrentStream+1) % JoinInfo->StreamsCnt;  
  RecursiveNestedLoop(Handle, JoinInfo->inner); // inner must be >=1  
}


// Searches for a matching tuples
void RelNLFEWPlanOp::RecursiveNestedLoop(XxxDerivedRelation::CursorInfo* Handle, int CurrStrm) const{

  NLFEWJoinInfo* JoinInfo = ((NLFEWJoinInfo *)(Handle->OperatorHandle));	  
  if(CurrStrm!=JoinInfo->SkippedStream){
    JoinInfo->InWindowBuffer[CurrStrm]->InitBackwardScan();
    JoinInfo->CandidateForJoin[CurrStrm] = JoinInfo->InWindowBuffer[CurrStrm]->GetPrevious();
    while(JoinInfo->InWindowBuffer[CurrStrm]->ScanIsValid()){	
      int inner = (CurrStrm+1) % JoinInfo->StreamsCnt;
      RecursiveNestedLoop(Handle, inner);
      JoinInfo->CandidateForJoin[CurrStrm] = JoinInfo->InWindowBuffer[CurrStrm]->GetPrevious();
    }
  }
  else{ // we consider last tuple in this relation
    // reached end of streams and all join				  
    XxxCoreADTClass::BoolValue Satisfied(XXX_FALSE);
    // To re-use old code, temporarily use the Handle; get its current records
    XxxRecord *TempRec;
    for (int i=0;i<JoinInfo->StreamsCnt;i++){
      TempRec = Handle->ChildRecs[i];
      Handle->ChildRecs[i] = JoinInfo->CandidateForJoin[i];
      JoinInfo->CandidateForJoin[i] = TempRec;;
    }
    if (checkFinal(Handle, Satisfied) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADEXEC, "Error checking residual plan");
      return ;
    }
    if (Satisfied.Value){
      XxxRecord *RetRec;
      createRecord(Handle, JoinInfo->RelationArray, JoinInfo->Schema, RetRec);
      JoinInfo->OutputGroup.AddRecord(RetRec);	  
    }      
    // return handle to its original values
    for (int j=0;j<JoinInfo->StreamsCnt;j++){
      TempRec = Handle->ChildRecs[j];
      Handle->ChildRecs[j] = JoinInfo->CandidateForJoin[j];
      JoinInfo->CandidateForJoin[j] = TempRec;;
    }
  }
}

//-----------------------------------------------------
void RelNLFEWPlanOp::PrintNodeInfo(ostream &OStr) const{

  RelSPJPlanOp::PrintNodeInfo(OStr);
  // now print any additional stuff
}

RelNLFEWPlanOp::RelNLFEWPlanOp(int NumChildren, 
			 RelPlanOp* Children[],
			 RelCatalogInfo* CatInfo, 
			 XxxValExprPlanList* ExprPlanList,
			 XxxBooleanExpressionPlan** SelPredPlans, 
			 XxxBooleanExpressionPlan** JoinPredPlans,
			 XxxBooleanExpressionPlan** ResidPredPlans, 
			 XxxBooleanExpressionPlan* InitPlan,
			 double LWSize, double RWSize): 
  RelSPJPlanOp(NumChildren, Children, CatInfo, ExprPlanList, 
	       SelPredPlans, JoinPredPlans, ResidPredPlans, InitPlan){
  
  LeftWindowSize = LWSize;   // The window size for left buffer
  RightWindowSize = RWSize;  // The window size for right buffer
}

XxxErrCode RelNLFEWPlanOp::checkFinal(XxxDerivedRelation::CursorInfo* Handle,
				   XxxCoreADTClass::BoolValue& Satisfied) const 
{
  Satisfied.Value = XXX_TRUE;
  XxxBooleanExpressionPlan* PredPlan = NULL;
  for (int i = 0; i < Handle->ArrayLen; i++) {
    if (JoinPredPlans && JoinPredPlans[i])
      if (XxxBooleanExpressionPlan::MakeConjunction(PredPlan, JoinPredPlans[i]) != XXX_OK) {
	REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
	return XXX_NOT_OK;
      }
    if (ResidPredPlans && ResidPredPlans[i])
      if (XxxBooleanExpressionPlan::MakeConjunction(PredPlan, ResidPredPlans[i]) != XXX_OK) {
	REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
	return XXX_NOT_OK;
      }
  }
  if (PredPlan)
    return PredPlan->Evaluate((const XxxRecord **)Handle->ChildRecs, 
			      (const XxxRecordSchema**)Handle->ChildSchema, 
			      &Satisfied, Handle->Params?Handle->Params->Params:0);
  return XXX_OK;
}

XxxErrCode RelNLFEWPlanOp::getNextItem(XxxRelation *TheRelation,
				    XxxDerivedRelation::CursorInfo* Handle,
				    int Index, XxxBool& Valid) const
{
  if (TheRelation->NextItem(Handle->ChildCursorArray[Index],
			    Handle->ChildRecs[Index]) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADEXEC, "Error getting NextItem from relation");
    return XXX_NOT_OK;
  }
  Valid = Handle->ChildCursorArray[Index].IsValid();
  return XXX_OK;
}


XxxErrCode RelNLFEWPlanOp::joinGroups(XxxDerivedRelation::CursorInfo* Handle,
				   XxxRelation* *RelationArray,
				   const XxxRecordSchema* Schema,
				   NLFEWJoinInfo *JoinInfo) const
{  
  return XXX_OK;
}

XxxErrCode RelNLFEWPlanOp::NLJoinTuples(XxxDerivedRelation::CursorInfo* Handle,
				     XxxRelation* *RelationArray,
				     const XxxRecordSchema* Schema,			  
				     NLFEWJoinInfo *JoinInfo,
				     XxxRecord *LeftRec,
				     XxxRecord *RightRec) const
{  
  return XXX_OK;
}



XxxErrCode RelNLFEWPlanOp::parseJoinPlan(XxxDerivedRelation::CursorInfo* Handle,
				      XxxValueExprPlan* &LeftAttrFuncP, 
				      int &LeftAttrIndex,
				      int &LeftRelIndex,
				      XxxValueExprPlan* &RightAttrFuncP, 
				      int &RightAttrIndex,
				      int &RightRelIndex) const

{ 
  return XXX_OK;
}


XxxErrCode 
RelNLFEWPlanOp::createJoinInfo(XxxDerivedRelation::CursorInfo* Handle,
			    NLFEWJoinInfo *&JoinInfo) const
{
  // Get the locations of the join columns
  XxxValueExprPlan *LeftAttrFuncP, *RightAttrFuncP;
  int LeftRelIndex, RightRelIndex;
  int LeftAttrIndex, RightAttrIndex;

  if (parseJoinPlan(Handle, LeftAttrFuncP, LeftAttrIndex, LeftRelIndex,
		    RightAttrFuncP, RightAttrIndex, RightRelIndex) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADEXEC,
		   "Attempt to separate join condition failed");
    return XXX_NOT_OK;
  }

  // Make the input groups
  JoinInfo = new NLFEWJoinInfo(RightRelIndex, LeftRelIndex, RightAttrIndex, 
			    LeftAttrIndex, RightAttrFuncP, LeftAttrFuncP,
			    Handle->Params->TheWindowSec,
			    Handle->Params->TheWindowMicroSec);

  return XXX_OK;
}

XxxErrCode 
RelNLFEWPlanOp::InitHandle(XxxDerivedRelation::CursorInfo* Handle,
			XxxRelation* *RelationArray) const 
{
  if (Handle->ArrayLen < 2) {
    REL_NOTE_ERROR(REL_BADOP, "Merge-Window join works on 2 or more inputs only");
    return XXX_NOT_OK;
  }
  // Create the operator handle
  NLFEWJoinInfo* JoinInfo;
  if (createJoinInfo(Handle, JoinInfo) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADEXEC, "Cannot create Merge-window join info");
    return XXX_NOT_OK;
  }
  Handle->OperatorHandle = (void *)JoinInfo;
  
  XxxCoreADTClass::BoolValue Satisfied(XXX_TRUE);
  if (InitPlan) {
    if (InitPlan->Evaluate((const XxxRecord **)Handle->ChildRecs, 
			   (const XxxRecordSchema **)Handle->ChildSchema, 
			   &Satisfied, (Handle->Params) ? Handle->Params->Params : 0)!= XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADEXPR, "Error evaluating expression!");
      return XXX_NOT_OK;
    }
  }
  if (!Satisfied.Value) {
    JoinInfo->Valid = XXX_FALSE;
    return XXX_OK;
  }
  // Initialize cursors for both inputs
  for (int i = 0; i < Handle->ArrayLen; i++) {
    XxxBooleanExpressionPlan* PredPlan = (SelPredPlans?SelPredPlans[i]:0);
    if (RelationArray[i]->InitCursor1(Handle->ChildCursorArray[i], PredPlan, NULL, NULL,
				      Handle->Params, Handle->ArrayLen, 0) != XXX_OK) {
      char buf[100];
      sprintf (buf, "Unable to open cursor on input %d", i);
      XXX_NOTE_ERROR(XXX_BADCURSOR, buf);
      return XXX_NOT_OK;
    }	
  }
  // ---- Initialization
  JoinInfo->RelationArray = RelationArray;
  JoinInfo->StreamsCnt = Handle->ArrayLen;
  JoinInfo->CurrentPeriod.SPeriod.tv_sec=0; JoinInfo->CurrentPeriod.SPeriod.tv_usec=0;
  JoinInfo->CurrentPeriod.EPeriod.tv_sec=0; JoinInfo->CurrentPeriod.EPeriod.tv_usec=0;
  
  //Initialize InWindowBuffers and OutWindowBuffers
  JoinInfo->IsMode1 = 1;
  JoinInfo->CurrentlyIn = 0;
  JoinInfo->StartingStream = 0;
  JoinInfo->CurrentStream = 0;
  JoinInfo->Mode2_CurrentStream = 0;  
  JoinInfo->IsMode1 = 1;
  JoinInfo->inner=0;
  JoinInfo->SkippedStream=0;
  JoinInfo->NewTupleAddedToInBuffers=0;
  JoinInfo->EndOfMode2Flag = 0;

  JoinInfo->InWindowBuffer = (InWindowLargeBucket **) malloc(sizeof(InWindowLargeBucket *)*JoinInfo->StreamsCnt);
  JoinInfo->OutWindowBuffer = (OutWindowLargeBucket **) malloc(sizeof(OutWindowLargeBucket *)*JoinInfo->StreamsCnt);
  for (int k=0; k < JoinInfo->StreamsCnt; k++){
    JoinInfo->InWindowBuffer[k] = new InWindowLargeBucket;
    JoinInfo->OutWindowBuffer[k] = new OutWindowLargeBucket;
  }
  return XXX_OK;
}

XxxErrCode RelNLFEWPlanOp::CloseHandle(XxxDerivedRelation::CursorInfo* Handle,
				    XxxRelation* *RelationArray) const
{ 

  // Close cursors on all the inputs
  for (int i = 0; i < Handle->ArrayLen; i++) {
    if (Handle->ChildCursorArray[i].State && RelationArray[i]->CloseCursor(Handle->ChildCursorArray[i]) != XXX_OK) {
      char buf[100];
      sprintf(buf, "Unable to close sequence cursor %d", i);
      XXX_NOTE_ERROR(XXX_BADCURSOR, buf);
      return XXX_NOT_OK;
    }
  }
  
  //TODO: you must 
  // delete the tables and the non-empty groups
  NLFEWJoinInfo* JoinInfo = ((NLFEWJoinInfo *)(Handle->OperatorHandle));  
  
  JoinInfo->OutputGroup.FlushAllRecords(XXX_TRUE,(const XxxRecordSchema *)Handle->RecordSchema);  
  JoinInfo->ExpiredGroup.FlushAllRecords(XXX_TRUE);
  JoinInfo->FloatGroup.FlushAllRecords(XXX_TRUE);


  // TODO: delete the swap files
  //  JoinInfo->LeftHashTable.DeleteSwapFile((const XxxRecordSchema *)Handle->ChildSchema[0]);
  //  JoinInfo->LeftHashTable.DeleteSwapFile((const XxxRecordSchema *)Handle->ChildSchema[1]);

  delete ((NLFEWJoinInfo *)(Handle->OperatorHandle));
  
  return XXX_OK;
}

// TODO: Looks like the input records never get reclaimed properly!!
XxxErrCode 
RelNLFEWPlanOp::GetNextRecord(XxxSetCursor& SCsr, XxxRelation* *RelationArray,
			   const XxxRecordSchema* Schema, XxxRecord*& ReturnRec) const{
  // Get the NLFEWJoinInfo from the operator handle
  XxxDerivedRelation::CursorInfo* Handle = 
    (XxxDerivedRelation::CursorInfo*)(SCsr.State);
  if (!Handle) {
    REL_NOTE_ERROR(REL_BADEXEC, "NULL Exec handle");
    return XXX_NOT_OK;
  }  
  NLFEWJoinInfo* JoinInfo = ((NLFEWJoinInfo *)(Handle->OperatorHandle));

  JoinInfo->RelationArray = RelationArray;
  JoinInfo->Schema = (XxxRecordSchema *)Schema;

  Group & OutputGroup = JoinInfo->OutputGroup;  
  Group & ExpiredGroup = JoinInfo->ExpiredGroup;
  BucketChain & FloatGroup = JoinInfo->FloatGroup; 

  // If there is a valid scan on the OutputGroup, return its next tuple
  if (OutputGroup.ScanIsValid() == XXX_TRUE) {
    ReturnRec = OutputGroup.GetNext();
    if (OutputGroup.ScanIsValid() == XXX_TRUE)
      return XXX_OK;
  }
  // Empty the output group; all of its records have been returned
  OutputGroup.FlushAllRecords(XXX_FALSE);
  // If there are no further records to return and a scan has finished, return
  if (!(JoinInfo->Valid)) {
    SCsr.AtEnd = XXX_TRUE;
    ReturnRec = 0;
    return XXX_OK;
  }
  
  XxxBool LeftScanValid = JoinInfo->LeftScanValid, 
    RightScanValid = JoinInfo->RightScanValid;

  int IsNewRecord=0;	
 
  // check which mode you are in
  while(1){
    if(JoinInfo->IsMode1){
      // In Mode 1
      //      cerr << " ================================  Mode 1 " << endl;
      int i;
      JoinInfo->CurrentStream = JoinInfo->StartingStream;
      JoinInfo->CurrentlyIn = 0;      
      JoinInfo->CurrentPeriod.SPeriod.tv_sec=0; 
      JoinInfo->CurrentPeriod.SPeriod.tv_usec=0;
      JoinInfo->CurrentPeriod.EPeriod.tv_sec=0; 
      JoinInfo->CurrentPeriod.EPeriod.tv_usec=0;
      while(1){
	i = JoinInfo->CurrentStream;
	//cerr << i << endl;
	JoinInfo->StartRec[i] = 
	  JoinInfo->InWindowBuffer[i]->GetTopRec(JoinInfo->OutWindowBuffer[i], 
						 JoinInfo->RelationArray[i], Handle, 
						 i, JoinInfo->StreamScanValid[i], 
						 JoinInfo->IsMode1, 
						 JoinInfo->NewTupleAddedToInBuffers,
						 IsNewRecord); 
	if(JoinInfo->StartRec[i]!=NULL){
	  //cerr << i << endl;
	  if(TimeStampsGreater(JoinInfo->CurrentPeriod.SPeriod, JoinInfo->StartRec[i]->TSVec->TheVec[0])){
	    // new record is older than the period, 
	    // skip this tuple and get next from this stream
	    JoinInfo->InWindowBuffer[i]->DeleteTopRec(XXX_TRUE, (const XxxRecordSchema *)Handle->ChildSchema[i]);
	  }
	  else if(TimeStampsGreater(JoinInfo->StartRec[i]->TSVec->TheVec[0], 
				    JoinInfo->CurrentPeriod.EPeriod)){
	    // new record is newer than the period
	    // skip all current records and start with period
	    // centered at this record
	    //	    for(int k=StartingStream; k<i; k=(k+1)% StreamsCnt)
	    //	      InWindowBuffer[i]->DeleteTopRec(XXX_TRUE);
	    JoinInfo->StartingStream = JoinInfo->CurrentStream;
	    JoinInfo->CurrentPeriod.SPeriod.tv_sec = JoinInfo->StartRec[i]->TSVec->TheVec[0].tv_sec - JoinInfo->TheWindow;
	    JoinInfo->CurrentPeriod.SPeriod.tv_usec = JoinInfo->StartRec[i]->TSVec->TheVec[0].tv_usec;
	    JoinInfo->CurrentPeriod.EPeriod.tv_sec = JoinInfo->StartRec[i]->TSVec->TheVec[0].tv_sec + JoinInfo->TheWindow;
	    JoinInfo->CurrentPeriod.EPeriod.tv_usec = JoinInfo->StartRec[i]->TSVec->TheVec[0].tv_usec;
	    JoinInfo->CurrentlyIn = 1 << JoinInfo->CurrentStream;
	    JoinInfo->CurrentStream = (JoinInfo->CurrentStream+1) % JoinInfo->StreamsCnt;
	  }
	  else{
	    // new record within the period
	    // update the period
	    struct timeval SNewPeriod, ENewPeriod;
	    SNewPeriod.tv_sec  = JoinInfo->StartRec[JoinInfo->CurrentStream]->TSVec->TheVec[0].tv_sec - JoinInfo->TheWindow;
	    SNewPeriod.tv_usec = JoinInfo->StartRec[JoinInfo->CurrentStream]->TSVec->TheVec[0].tv_usec;
	    ENewPeriod.tv_sec  = JoinInfo->StartRec[JoinInfo->CurrentStream]->TSVec->TheVec[0].tv_sec + JoinInfo->TheWindow;
	    ENewPeriod.tv_usec = JoinInfo->StartRec[JoinInfo->CurrentStream]->TSVec->TheVec[0].tv_usec;
	    if(TimeStampsGreater(SNewPeriod, JoinInfo->CurrentPeriod.SPeriod)){
	      JoinInfo->CurrentPeriod.SPeriod.tv_sec = SNewPeriod.tv_sec; 
	      JoinInfo->CurrentPeriod.SPeriod.tv_usec = SNewPeriod.tv_usec; 
	    }
	    if(TimeStampsGreater(JoinInfo->CurrentPeriod.EPeriod, ENewPeriod)){
	      JoinInfo->CurrentPeriod.EPeriod.tv_sec = ENewPeriod.tv_sec; 
	      JoinInfo->CurrentPeriod.EPeriod.tv_usec = ENewPeriod.tv_usec; 
	    }
	    // if this stream is counted before then skip
	    if(!(JoinInfo->CurrentlyIn&(1 << JoinInfo->CurrentStream)))
	      JoinInfo->CurrentlyIn = JoinInfo->CurrentlyIn + (1 << JoinInfo->CurrentStream);
	    if(JoinInfo->CurrentlyIn==((1<<JoinInfo->StreamsCnt) - 1)){
	      JoinInfo->IsMode1 = 0;
	      JoinInfo->EndOfMode2Flag = 0;
	      break;
	      // end of mode 1
	    }
	    else{
	      JoinInfo->CurrentStream = (JoinInfo->CurrentStream+1) % JoinInfo->StreamsCnt;
	    }
	  }
	}//StartRec!=NULL
	else
	  JoinInfo->CurrentStream = (JoinInfo->CurrentStream+1) % JoinInfo->StreamsCnt;
      }//while(1)
    }// Mode 1
    
    
    if(!JoinInfo->IsMode1){ // Mode 2
      // at this point 
      //      cerr << " >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>  Mode 2 (a)" << endl;
      if(JoinInfo->NewTupleAddedToInBuffers){
	// If mode 1 results in new tuple in the buffers
	// do a nested loop on all the cuurent tuples in the buffers
	//	cerr << " here before " << endl;
	JoinInWindowBuffers(Handle);
	JoinInfo->Mode2_CurrentStream=0;
	JoinInfo->NewTupleAddedToInBuffers = 0;
	//	cerr << " here After " << endl;
      }
      if(JoinInfo->OutputGroup.IsEmpty()){
	int j;
	//cerr << " >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>  Mode 2 (b)" << endl;
	while(1){
	  j = JoinInfo->Mode2_CurrentStream;
	  struct timezone Temp; 
	  //	  struct timeval StartTime;
	  //	  gettimeofday(&StartTime, &Temp);
	  JoinInfo->LastRec[j] = 
	    JoinInfo->OutWindowBuffer[j]->GetTopRec(JoinInfo->RelationArray[j],
						    Handle, j, JoinInfo->StreamScanValid[j]); 
	  if(JoinInfo->LastRec[j]!=NULL){
	    
	    if(TimeStampsGreater(JoinInfo->LastRec[j]->TSVec->TheVec[0], 
				 JoinInfo->CurrentPeriod.EPeriod)){
	      // new record newer than the period
	      //cerr << "............................. - 2nrn " << endl;
	      JoinInfo->EndOfMode2Flag = JoinInfo->EndOfMode2Flag | (1<<j);
	      if(JoinInfo->EndOfMode2Flag==((1<<JoinInfo->StreamsCnt)-1)){
		JoinInfo->IsMode1 = 1;
		// delete oldest tuple in InWindowBuffer
		unsigned int minStream=0;
		for (int k=1;k<JoinInfo->StreamsCnt;k++){
		  if(TimeStampsGreater(JoinInfo->StartRec[minStream]->TSVec->TheVec[0], JoinInfo->StartRec[k]->TSVec->TheVec[0]))
		    minStream = k;
		}
		JoinInfo->InWindowBuffer[minStream]->DeleteTopRec(XXX_TRUE, (const XxxRecordSchema *)Handle->ChildSchema[minStream]);
		JoinInfo->StartingStream = minStream;
		break;
	      }
	      JoinInfo->Mode2_CurrentStream =  (JoinInfo->Mode2_CurrentStream+1) % JoinInfo->StreamsCnt;
	      while((JoinInfo->EndOfMode2Flag&(1 << JoinInfo->Mode2_CurrentStream)))
		JoinInfo->Mode2_CurrentStream =  (JoinInfo->Mode2_CurrentStream+1) % JoinInfo->StreamsCnt;
	    }
	    else{	      
	      //cerr << "----------------------------- - 2nri " << endl;
	      // new record is in period, It can't be to left
	      // perform nested loop to get result
	      JoinInfo->SkippedStream = j;
	      JoinInfo->CandidateForJoin[j] = JoinInfo->LastRec[j];
	      //	      cerr << " Before " << endl;
	      JoinInWindowBuffers(Handle, JoinInfo->LastRec[j]);			  
	      //	      cerr << " After " << endl;
	      JoinInfo->OutWindowBuffer[j]->DeleteTopRec(XXX_FALSE, NULL);
	      JoinInfo->InWindowBuffer[j]->insertRec(JoinInfo->LastRec[j]);
	      JoinInfo->Mode2_CurrentStream =  (JoinInfo->Mode2_CurrentStream+1) % JoinInfo->StreamsCnt;
	      while((JoinInfo->EndOfMode2Flag&(1 << JoinInfo->Mode2_CurrentStream)))
		JoinInfo->Mode2_CurrentStream =  (JoinInfo->Mode2_CurrentStream+1) % JoinInfo->StreamsCnt;
	      if(!JoinInfo->OutputGroup.IsEmpty())
		break;
	    }
	  }// LastRec!=NULL
	  else{
	    //cerr << "----------------------------- - 2nr-null " << j <<endl;
	    JoinInfo->Mode2_CurrentStream =  (JoinInfo->Mode2_CurrentStream+1) % JoinInfo->StreamsCnt;
	    while((JoinInfo->EndOfMode2Flag&(1 << JoinInfo->Mode2_CurrentStream)))
	      JoinInfo->Mode2_CurrentStream =  (JoinInfo->Mode2_CurrentStream+1) % JoinInfo->StreamsCnt;
	  }
	  //CurrentStream = (CurrentStream+1) % StreamsCnt;
	}//while(1)
      }//  if(JoinInfo->OutputGroup.IsEmpty())
    }// ! IsMode1
    if(!JoinInfo->OutputGroup.IsEmpty())
      break;
  }//while(1)  
  

  JoinInfo->Valid = XXX_FALSE;
  for (int i=0; i<JoinInfo->StreamsCnt; i++)
    if(JoinInfo->StreamScanValid[i]){
      JoinInfo->Valid = XXX_TRUE;
      break;
    }      

  // If the join was successful, and the OutputGroup has been repopulated.
  // Return its first tuple.
  if (!OutputGroup.IsEmpty()) {
    OutputGroup.InitScan();
    ReturnRec = OutputGroup.GetNext();
    ASSERT (OutputGroup.ScanIsValid());
    //OutputGroup.FlushAllRecords(XXX_TRUE, (const XxxRecordSchema *)Handle->RecordSchema); // <------------- Added
    //    ReturnRec = 0; // <---------------- Added
    return XXX_OK;
  }

  // Otherwise, one of the inputs' scans is complete and the join need not
  // proceed.  Return end-of-scan.
  SCsr.AtEnd = XXX_TRUE;
  ReturnRec = 0;
  return XXX_OK;
}
