#include <stdio.h>
#include <stdlib.h>
#include <iostream.h>
#include "xxx_thread.h"
#include "reladt.h"
#include "relerror.h"
#include "relation.h"
#include "relplanop.h"
#include "relcat.h"


//
// Methods for Derived Relations
//
XxxDerivedRelation::XxxDerivedRelation(RelDerivedPlanOp* Oper,
				       RelCatalogInfo* CatInfo,
				       const XxxValueEnv* Params)
: XxxRelation(CatInfo)
{
 planOp = Oper;
 inputArity = Oper->Arity();
 relInputs = new XxxRelation*[inputArity];
 for (int i = 0; i< inputArity; i++) {
   if (Oper->GetChild(i)->MakeRelation(Params, relInputs[i]) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADPLAN, "Unable to make relation");
     relInputs[i] = NULL;
    }
  }
}

XxxDerivedRelation::~XxxDerivedRelation()
{
 for (int i = 0; i< inputArity; i++) 
   delete relInputs[i];
 delete [] relInputs;
}

XxxErrCode XxxDerivedRelation::
InitCursor1(XxxSetCursor& SCsr,
	    XxxBooleanExpressionPlan* Filter,
	    const XxxRecord* Rec[],
	    const XxxRecordSchema* Schema[],
	    const RelExecParameters* Parameters,
	    int   SibArity,
	    int   CurrJoinPos) const
{
 CursorInfo* Handle = 
   new CursorInfo(CatalogInfo->RecordSchema(), Filter, 
		  XXX_FALSE, // initially, set ExactMatch == FALSE
		  inputArity, Parameters, Rec, Schema, CurrJoinPos);

 for (int i = 0; i < inputArity; i++) {
   // collect the child schema
   Handle->ChildSchema[i] = relInputs[i]->CatalogInfo->RecordSchema();
  }

 // this could modify Handle->ExactMatch
 if (planOp->InitHandle(Handle, relInputs) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_BADCURSOR, "Unable to initialize cursor");
   return XXX_NOT_OK;
  }

 SCsr.State = Handle;
 SCsr.AtEnd = XXX_FALSE;
 SCsr.NumberRead = 0;
 return XXX_OK;
}

XxxErrCode XxxDerivedRelation::CloseCursor(XxxSetCursor& SCsr) const
{
 CursorInfo* Handle = (CursorInfo *)SCsr.State;

 if (Handle) {
   if (planOp->CloseHandle(Handle, relInputs) != XXX_OK) {
     XXX_NOTE_ERROR(XXX_BADCURSOR, "Unable to close cursor");
     return XXX_NOT_OK;
    }
   delete Handle;
  }

 SCsr.AtEnd = XXX_TRUE;
 SCsr.State = NULL;
 return XXX_OK;
}

XxxErrCode XxxDerivedRelation::NextItem(XxxSetCursor& SCsr, 
					XxxRecord* &ReturnRec) const
{
 if (!SCsr.State) {
   XXX_NOTE_ERROR(XXX_BADCURSOR, "no state in cursor");
   SCsr.AtEnd = XXX_TRUE;
   return XXX_NOT_OK;
  }
 if (SCsr.IsValid() != XXX_TRUE)
   return XXX_OK;

 // use planop to determine how to get the next item
 CursorInfo* Handle = (CursorInfo *)SCsr.State;

 while (1) {
   if (planOp->GetNextRecord(SCsr, relInputs,
			     CatalogInfo->RecordSchema(),
			     ReturnRec) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADEXEC, "Unable to generate next record");
     return XXX_NOT_OK;
    }

   // break from the loop if there are no more answers
   if (SCsr.IsValid() != XXX_TRUE)
     return XXX_OK;

   // if it is an exact match cursor, break out of loop
   if (Handle->ExactMatch || !Handle->Filter) break;

   // otherwise the Filter condition needs to be checked again
   XxxCoreADTClass::BoolValue CondResult(XXX_TRUE);
   if (!Handle->SiblingRecs || !Handle->SiblingSchema) {
     // in this case, simply evaluate the filter assuming a single input
     if (Handle->Filter->Evaluate(ReturnRec, 
				  CatalogInfo->RecordSchema(),
				  &CondResult, 
				  (Handle->Params?Handle->Params->Params:0))
	 != XXX_OK) {
       XXX_NOTE_ERROR(XXX_BADEXPR, "Cannot evaluate filter expression");
       return XXX_NOT_OK;
      }
    } // if there is no "join environment"
    else {
      // first put the current record into the Handle->SiblingRecs array
      // Handle->SiblingRecs[Handle->CurrJoinPos] = ReturnRec;
      // Unnecessary: happens automatically (see how ReturnRec is passed in)

      // then evaluate the Filter expression
      if (Handle->Filter->Evaluate(Handle->SiblingRecs,
				   Handle->SiblingSchema,
				   &CondResult,
				   (Handle->Params?Handle->Params->Params:0))
	  != XXX_OK) {
	XXX_NOTE_ERROR(XXX_BADEXPR, "Cannot evaluate filter expression");
	return XXX_NOT_OK;
       }
     }

   if (CondResult.Value == XXX_TRUE) {
     // break out of while loop and return valid record
     break;
    }
   else {
     delete ReturnRec;
     // reset the ReturnRecord: this is important!!
     ReturnRec = NULL;
     // continue in the loop
    }
  } // while forever

 SCsr.NumberRead++;
 return XXX_OK;
}

XxxDerivedRelation::CursorInfo::
CursorInfo(XxxRecordSchema* Schema,
	   XxxBooleanExpressionPlan* Expr,
	   XxxBool Exact,
	   int InputArity,
	   const RelExecParameters* Parameters,
	   const XxxRecord** SibRecs,
	   const XxxRecordSchema** SibSchema,
	   int CurrentJoinPos) 
  : CommonCursorInfo(Schema, Expr, Exact, Parameters, 
		     SibRecs, SibSchema, CurrentJoinPos),
  RecValuesArray(Schema)
{
 ArrayLen = InputArity;

 ChildCursorArray = new XxxSetCursor[ArrayLen];
 ChildRecs = new XxxRecord*[ArrayLen];
 ChildSchema = new XxxRecordSchema*[ArrayLen];
 for (int i = 0; i < ArrayLen; i++) {
   // This is important to set to NULL initially.
   // If it is non-NULL, it means that it points to data that
   // will need to be reclaimed.
   ChildRecs[i] = NULL;
   // these will get assigned by NextItem
   // so there is no need to assign them here
   ChildSchema[i] = NULL;
  }
}

XxxDerivedRelation::CursorInfo::~CursorInfo()
{
 if (ChildCursorArray) delete [] ChildCursorArray;
 if (ChildRecs) delete [] ChildRecs;
 if (ChildSchema) delete [] ChildSchema;
}
