#include <iostream.h>

#include "reladt.h"
#include "relerror.h"
#include "relopt.h"
#include "relation.h"
#include "reldata.h"
#include "releng.h"
#include "relenv.h"
#include "relexec.h"
#include "relplanop.h"
#include "client.h"
#include "protocol.h"

XxxErrCode RelExecEngine::EvalQuery(RelQueryPlan* Plan, 
				    const RelExecParameters* Params,
				    XxxADTValue *ReturnValue)
{
 if (!Plan || Plan->IsValid() != XXX_TRUE) {
   REL_NOTE_ERROR(REL_BADPLAN, "Invalid relational plan");
   return XXX_NOT_OK;
  }
 
 if (XXX_THISRELENV()->DebugExec) {
   cerr << "Evaluating : ";
   Plan->Print(cerr);
   cerr << endl;
 }

 RelPlanExecutor* PlanExec;

 // first translate query plan to an "executable" structure
 if (Plan->GenerateExecPlan(Params, PlanExec) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, "Unable to interpret query plan");
   return XXX_NOT_OK;
  }

 // If a query-processing information file was provided, append a 
 // representation of the plan tree.
// XxxRelEnv *GEnv = (XxxRelEnv*)(XXX_THISENGENV(RelationEngineId));
 ///if (GEnv->QPInfoFile) {
 //  *(GEnv->QPInfoFile) << "SEQ Execution" << endl;
 //  *(GEnv->QPInfoFile) << "TREE Execution Plan" << endl;
 //  PlanExec->TreePrint(*(GEnv->QPInfoFile));
 //  *(GEnv->QPInfoFile) << "ENDTREE\nENDSEQ" << endl;
 //}

 // now "interpret" the executable
 if (PlanExec->Execute(Params, ReturnValue) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, "Failure executing query plan");
   return XXX_NOT_OK;
  }

 // If a query-processing information file was provided, append a 
 // representation of the plan tree.
 XxxRelEnv *GEnv = (XxxRelEnv*)(XXX_THISENGENV(RelationEngineId));
 if (GEnv->QPInfoFile) {
   *(GEnv->QPInfoFile) << "SEQ Execution" << endl;
   *(GEnv->QPInfoFile) << "TREE Final Plan" << endl;
   PlanExec->TreePrint(*(GEnv->QPInfoFile));
   *(GEnv->QPInfoFile) << "ENDTREE\nENDSEQ" << endl;
 }

 // reclaim the execution plan
 delete PlanExec;

 return XXX_OK;
}

XxxErrCode RelExecEngine::GetEvalCursor(RelQueryPlan* Plan, 
					const RelExecParameters* Params,
					RelEvalCursor*& EvalCursor)
{
 if (!Plan || Plan->IsValid() != XXX_TRUE) {
   REL_NOTE_ERROR(REL_BADPLAN, "Invalid relational plan");
   return XXX_NOT_OK;
  }
 
 if (XXX_THISRELENV()->DebugExec) {
   cerr << "Creating plan cursor on : ";
   Plan->Print(cerr);
   cerr << endl;
 }


 // translate query plan to an "executable" cursor
 if (Plan->GenerateEvalCursor(Params, EvalCursor) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, "Unable to interpret query plan");
   return XXX_NOT_OK;
  }

 return XXX_OK;
}

void RelPlanExecutor::TreePrint (ofstream &QPInfoFile) {
  int curPlanId = 0;
  planOp->TreePrint(QPInfoFile, curPlanId, 0);
}


XxxErrCode 
RelPlanExecutor::Execute(const RelExecParameters* Params, 
			 XxxADTValue* ReturnValue)
{
 XxxRecordSchema* AnswerSchema = planOp->CatalogInfo->RecordSchema();
 XxxRelation *QueryRel;
 void *ExtraData;

 if ((planOp->MakeRelation((Params?Params->Params:0), QueryRel) != XXX_OK) ||
    (!QueryRel) || (XXX_IS_ERROR())) {
   REL_NOTE_ERROR(REL_BADPLAN, "Unable to make relation from plan");
   return XXX_NOT_OK;
  }

 if (initExtraData(ExtraData) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADPLAN, "Unable to perform setup for query execution");
   return XXX_NOT_OK;
 }

 XxxSetCursor RelCsr;  XxxRecord* RelRecord;

 if (QueryRel->InitCursor1(RelCsr, NULL, NULL, NULL, Params) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, "Unable to open cursor");
   return XXX_NOT_OK;
  }
 while ((QueryRel->NextItem(RelCsr, RelRecord) == XXX_OK) && 
	(RelCsr.IsValid() == XXX_TRUE)) {
   //Begin Moustafa:
   // since pipline may have NULL tuples , NULL tupes should not be processed
   if (RelRecord!=NULL){
     processRecord(RelRecord, AnswerSchema, ExtraData);


     // The consumer is reponsible for reclaiming memory.
     // First recover any ADT objects in the record
     RelRecord->Reclaim(AnswerSchema);
   }
   //End Moustafa
   //Begin Moustafa
   // verify that the client did not stop the query
   //   if(((XxxRequestThread* )XXX_THISTHREAD())->IsQueryRequestStop()){
   //     ((XxxRequestThread* )XXX_THISTHREAD())->StopQuery=1;
   //     break;
   //   }
   //End Moustafa
  } // while

 if (QueryRel->CloseCursor(RelCsr) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, "Error closing relation cursor");
   return XXX_NOT_OK;
  }

 // reclaim the relation
 delete QueryRel;

 // remove extra data
 termExtraData(ExtraData, ReturnValue);
 //Begin: Moustafa
 ((XxxRequestThread* )XXX_THISTHREAD())->StopQuery = 0;
 //End: Moustafa
 return XXX_OK;
}

RelPlanExecutor::~RelPlanExecutor()
{ 
  if (planOp) delete planOp; 
}

XxxErrCode RelExecEngine::
PrintPlanExecutor::initExtraData(void *&ED)
{
 struct ExtraData* ExtraD = new ExtraData;
 ExtraD->OutputStr = XXX_THISENV().OutputStr;
 ED = (void *)(ExtraD);

 if (XXX_CLIENTINFO()->Protocol->
     PrefixResults(XXX_THISENV().OutputStr,
		   planOp->CatalogInfo->RecordSchema()) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, "Error writing prefix for results");
   return XXX_NOT_OK;
  }

 return XXX_OK;
}

XxxErrCode RelExecEngine::
PrintPlanExecutor::processRecord(XxxRecord* RelRecord,
				 XxxRecordSchema* Schema,
				 void *ED)
{
 XxxRelEnv *REnv = (XxxRelEnv*)(XXX_THISENGENV(RelationEngineId));
 if (!REnv->ShowQueryResults) {
   return XXX_OK;
  }

 struct ExtraData* ExtraD = (struct ExtraData*)ED;
 if (XXX_CLIENTINFO()->Protocol->WriteRecord(ExtraD->OutputStr,
					     Schema, RelRecord) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, "Error writing result record");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

XxxErrCode RelExecEngine::
PrintPlanExecutor::termExtraData(void *ED, XxxADTValue* Result)
{
 struct ExtraData* ExtraD = (struct ExtraData*)ED;
 delete ExtraD;
 return XXX_OK;
}

XxxErrCode RelExecEngine::
StorePlanExecutor::initExtraData(void *&ED)
{

 struct ExtraData* ExtraD = new ExtraData;
 ExtraD->AnswerRel = NULL;
 ExtraD->RelEng = (XxxRelationEngine*)XxxDBEngineArray[RelationEngineId];
 RelDataEngine* RelDataEng = ExtraD->RelEng->DataEng;
 RelCatalogInfo* CatInfo; 

 if (ToAppend) {    // look for the relation
   // get hold of the catalog info
   XxxBool MIFound = XXX_FALSE;
   // reclaimed in termExtraData
   CatInfo = new RelCatalogInfo(RelName);

   if (RelDataEng->GetCatInfo(RelName, *CatInfo, MIFound) != XXX_OK) {
     REL_NOTE_ERROR (REL_BADTABLE, "Cannot find catalog for relation");
     return XXX_NOT_OK;
    }
   if (MIFound == XXX_TRUE) {
     if (RelDataEng->GetRelation(RelName, ExtraD->Hndl) != XXX_TRUE) {
       REL_NOTE_ERROR (REL_BADTABLE, "Cannot find relation to load");
       return XXX_NOT_OK;
      }
     REL_CALL_FN(RelDataEng->GetRelation(ExtraD->Hndl, ExtraD->AnswerRel, 
					 CatInfo->ImplInfo()),
		 REL_BADTABLE, "Cannot find relation to load");

     ExtraD->AnswerRel->CatalogInfo = CatInfo;
     // here's the place to change the store property of the relation
     REL_CALL_FN(ExtraD->AnswerRel->PrepareForLoad(),
		 REL_BADTABLE, "Not a loadable table");
    } // if the relation exists already
  }

 if (!ExtraD->AnswerRel && ToCreate) {
   // WHO RECOVERS THIS RELATION?? PS 5/95 : TODO:L1
   XxxErrCode rc;
   if (RelName) {
     RelCatalogInfo* ThisCatInfo = planOp->CatalogInfo;
     XxxRecordSchema* RecSchema = ThisCatInfo->RecordSchema();
     const char* KeyVector = RecSchema->GetKeyVec();
     RelDataSrc::KeyList* KList = NULL;
     RelDataSrc::AttrTypeList* AList = NULL;

     // TODO:L2 this should reflect the contents of Attr->TypeCatInfo
     XxxADTClass::MetaInfoHandle MIHndl;
     // have to go backward to get correct order
     for (int i = RecSchema->Arity()-1; i>=0; i--) {
       // first handle attr info
       XxxAttribute* Attr = RecSchema->GetAttr(i);
       if (!AList) {
	 // TODO:L2 NULL value info should be stored with Attr
	 AList = new RelDataSrc::AttrTypeList(Attr->Name, Attr->Type,
					      XXX_TRUE, MIHndl);
	} else {
	  AList = AList->Add(Attr->Name, Attr->Type,
			     XXX_TRUE, MIHndl);
	 }
       // now handle key info
       if (KeyVector[i]) {
	 if (!KList) {
	   KList = new RelDataSrc::KeyList(Attr->Name);
	  }
	 else {
	   KList = KList->Add(Attr->Name);
	  }
	} // key info
      }

     rc = RelDataEng->CreateNewRelation(ExtraD->Hndl, RelName,
					ThisCatInfo->ImplInfo(),
					AList, KList, 
					XXX_TRUE);
     // TODO:L1 these do not really reclaim recursively!
     delete AList;
     delete KList;
    }
   else {
     REL_CALL_FN(RelDataEng->CreateNewRelation(ExtraD->Hndl, NULL, 
					       NULL, //planOp->CatalogInfo->ImplInfo(),
					       NULL, NULL, 
					       XXX_FALSE),
		 REL_BADTABLE, "Unable to create answer relation");
    }

   REL_CALL_FN(RelDataEng->GetRelation(ExtraD->Hndl, ExtraD->AnswerRel,
				       NULL /*planOp->CatalogInfo->ImplInfo()*/),
	       REL_BADTABLE, "Unable to create relation from handle");
   // reclaimed in termExtraData
   CatInfo = planOp->CatalogInfo->Replicate1(XXX_FALSE /* CopyIndexes */);
   // make sure to remove the implinfo because it is not valid
   delete CatInfo->implInfo;
   CatInfo->implInfo = NULL;
   ExtraD->AnswerRel->CatalogInfo = CatInfo;
  }  // if we are to create the AnswerRel

 if (ExtraD->AnswerRel) {
   ED = (void *)(ExtraD);
   return XXX_OK;
  }
 else {
   REL_NOTE_ERROR(REL_BADEXEC, "Unable to create answer relation");
   return XXX_NOT_OK;
  }
}

XxxErrCode RelExecEngine::
StorePlanExecutor::processRecord(XxxRecord* RelRecord,
				 XxxRecordSchema* Schema,
				 void *ED)
{
 struct ExtraData* ExtraD = (struct ExtraData*)ED;
 if (ExtraD->AnswerRel->AddItem(RelRecord) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, "Unable to store answer record");
   return XXX_NOT_OK;
  }
 // TODO:L1 add something to repeatedly commit the transaction so
 // that log space doesnt overflow
 return XXX_OK;
}

XxxErrCode RelExecEngine::
StorePlanExecutor::termExtraData(void *ED, XxxADTValue* Result)
{
 struct ExtraData* ExtraD = (struct ExtraData*)ED;

 // there are two uses of StorePlanExecutor
 //  1) for INSERT INTO: in this case, Result is NULL
 //  2) for nested queries: in this case, Result is non-NULL
 if (Result) {
   // copy the handle into the Return Value
   XxxADTArray[RelationTypeId]->TypeCopy(ExtraD->Hndl, 
					 Result,
					 ExtraD->AnswerRel->CatalogInfo);
  }

 delete ExtraD->AnswerRel->CatalogInfo;
 delete ExtraD->AnswerRel;
 delete ExtraD->Hndl;
 delete ExtraD;

 return XXX_OK;
}

//
// Methods of RelEvalCursor
//

RelEvalCursor::RelEvalCursor(RelPlanOp* PlanOperator,
			     const RelExecParameters* Params)
{
  Initialized = XXX_FALSE;
  PlanOp = PlanOperator; 
  AnswerSchema = PlanOp->CatalogInfo->RecordSchema();

  if ((PlanOp->MakeRelation((Params?Params->Params:0), 
			    QueryRel) != XXX_OK) ||
      (!QueryRel) || (XXX_IS_ERROR())) {
    REL_NOTE_ERROR(REL_BADPLAN, "Unable to make relation from plan");
    return;
  }

  RelRecord = NULL;
  if (QueryRel->InitCursor1(RelCsr, NULL, NULL, 
			    NULL, Params) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, "Unable to open cursor");
   return;
  }

  Initialized = XXX_TRUE;
}

RelEvalCursor::~RelEvalCursor()
{ 
  // reclaim any old record
  if (RelRecord) {
    // The consumer is reponsible for reclaiming memory.
    // First recover any ADT objects in the record
    RelRecord->Reclaim(AnswerSchema);
    RelRecord = NULL;
  }

  // close the scan on the relation
  if (QueryRel->CloseCursor(RelCsr) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADEXEC, "Error closing relation cursor");
  }
  
  // reclaim the relation
  delete QueryRel;
  
  // reclaim the plan operator
  if (PlanOp) delete PlanOp;
}

XxxErrCode 
RelEvalCursor::GetNext(XxxRecord*& NextRecord,
		       XxxBool& Valid)
{
  if (!Initialized) {
    REL_NOTE_ERROR(REL_BADEXEC, "Uninitialized EvalCursor");
    return XXX_NOT_OK;
  }
  if (RelRecord) {
    // The consumer is reponsible for reclaiming memory.
    // First recover any ADT objects in the record
    RelRecord->Reclaim(AnswerSchema);
    RelRecord = NULL;
  }

  if (QueryRel->NextItem(RelCsr, RelRecord) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADEXEC, "Error getting next record");
    return XXX_NOT_OK;
  }

  if (RelCsr.IsValid() == XXX_TRUE) {
    NextRecord = RelRecord;
    Valid = XXX_TRUE;
    return XXX_OK;
  }
  else {
    NextRecord= NULL;
    Valid = XXX_FALSE;
    return XXX_OK;
  }
}
