#ifndef XXX_RELEXEC_H
#define XXX_RELEXEC_H

#include "reladt.h"
#include "values.h"

XXX_EXTERN_CLASS_DECL(RelQueryPlan);
XXX_EXTERN_CLASS_DECL(RelPlanOp);

/**
 *
 */
class RelExecParameters {
 public:
  XxxBool FirstTuple;
  const XxxValueEnv* Params;
  //Begin Moustafa: add window as executin parametes
  //TODO: multiple window?
  unsigned int TheWindowSec;  // the seconds part
  unsigned int TheWindowMicroSec; // the microseconds part    
  //End Moustafa
  RelExecParameters(const XxxValueEnv* Parameters) 
    : Params(Parameters) { FirstTuple = XXX_FALSE;}
  RelExecParameters() : Params(NULL) { }
  ~RelExecParameters(){ }
 };

XXX_EXTERN_CLASS_DECL(RelHandle);

/**
 * This class represents the top-level "driver" of a query
 * execution plan.
 * Note: This should be inside RelExecEngine but g++ dies
 * with an internal error.
 */
class RelPlanExecutor { 
 protected:
  RelPlanOp* planOp;
  virtual XxxErrCode initExtraData(void *& ED) = 0;
  virtual XxxErrCode processRecord(XxxRecord* RelRecord,
				   XxxRecordSchema* Schema,
				   void * ED) = 0;
  virtual XxxErrCode termExtraData(void * ED, XxxADTValue *Result) = 0;
 public:
  RelPlanExecutor(RelPlanOp* PlanOperator) 
   { planOp = PlanOperator; }
  virtual ~RelPlanExecutor();
  virtual XxxErrCode Execute(const RelExecParameters* Params, 
			     XxxADTValue* ReturnValue);
  void TreePrint (ofstream &QPInfoFile);
 };

/**
 * This class is used instead of RelPlanExecutor when an iterator
 * over the query is needed.
 */
class RelEvalCursor { 
 protected:
  RelPlanOp* PlanOp;
  XxxRelation *QueryRel;
  XxxSetCursor RelCsr;  
  XxxRecord* RelRecord;

  XxxBool Initialized;
 public:
  XxxRecordSchema* AnswerSchema;
  RelEvalCursor(RelPlanOp* PlanOperator,
		const RelExecParameters* Params);

  virtual ~RelEvalCursor();

  // the record returned by this is "owned" by this object,
  // and should not be deleted outside the methods of this class
  virtual XxxErrCode GetNext(XxxRecord*& NextRecord,
			     XxxBool& Valid);
 };

/**
 *
 */
class RelExecEngine {
  XxxRelationEngine* RelEng;

 public:  
  //// 
  class StorePlanExecutor : public RelPlanExecutor { 
   protected:
    struct ExtraData {
      RelHandle* Hndl;
      XxxRelation* AnswerRel;
      XxxRelationEngine* RelEng;
     };
    virtual XxxErrCode initExtraData(void *& ED);
    virtual XxxErrCode processRecord(XxxRecord* RelRecord,
				     XxxRecordSchema* Schema,
				     void * ED);
    virtual XxxErrCode termExtraData(void * ED, XxxADTValue *Result);

   public:
    char RelName[XXX_NODE_NAME_MAX];
    XxxBool ToCreate;
    XxxBool ToAppend;
    StorePlanExecutor(RelPlanOp* PlanOperator,
		      const char* RName,
		      XxxBool Create,
		      XxxBool Append)
      :RelPlanExecutor(PlanOperator) {
	if (RName) {
	  strncpy(RelName, RName, XXX_NODE_NAME_MAX);
	  RelName[XXX_NODE_NAME_MAX-1] = 0;
	 }
	else
	  strcpy(RelName, "ERROR");
	ToCreate = Create;
	ToAppend = Append;
       };
   };

  //// 
  class PrintPlanExecutor : public RelPlanExecutor { 
   protected:
    struct ExtraData {
      ostream* OutputStr;
     };
    virtual XxxErrCode initExtraData(void *& ED);
    virtual XxxErrCode processRecord(XxxRecord* RelRecord,
				     XxxRecordSchema* Schema,
				     void * ED);
    virtual XxxErrCode termExtraData(void * ED, XxxADTValue *Result);

   public:
    PrintPlanExecutor(RelPlanOp* PlanOperator)
      :RelPlanExecutor(PlanOperator) {}
   };


  RelExecEngine(XxxRelationEngine* SrcEng) { RelEng = SrcEng; }
  XxxErrCode EvalQuery(RelQueryPlan* Plan, 
		       const RelExecParameters* Params, XxxADTValue *Result);
  XxxErrCode GetEvalCursor(RelQueryPlan* Plan, 
			   const RelExecParameters* Params, 
			   RelEvalCursor*& EvalCursor);
 };

#endif  /* XXX_RELEXEC_H */
