#ifdef __GNUG__
#pragma implementation
#endif

#include "relplanop.h"
#include "relops.h"
#include "reldata.h"
#include "releng.h"

//
// Methods of RelPlanOp
//

// PURPOSE: used in printing the planOp tree
void RelPlanOp::NodePrint (ostream &OStr, int curBlockId, int Level, 
			   int numChildren, int *childrenIds) const {
}

RelPlanOp::RelPlanOp(int NumChildren, RelPlanOp* Children[],
		     RelCatalogInfo* CatInfo)
{
 childPlanArrayLen = NumChildren;
 childPlanArray = new RelPlanOp*[NumChildren];

 for (int i = 0; i < NumChildren; i++)
   childPlanArray[i] = Children[i];

 // deep copy of the catalog information
 CatalogInfo = CatInfo->Replicate1();
 // the subclasses of RelPlanOp can assign the SortInfo appropriately
}

RelPlanOp::~RelPlanOp()
{
 // delete all the children plans
 for (int i = 0; i < childPlanArrayLen; i++)
   delete childPlanArray[i];

 delete [] childPlanArray ;

 // reclaim CatalogInfo
 if (CatalogInfo) {
   CatalogInfo->Destroy();
   delete CatalogInfo;
   CatalogInfo = NULL;
  }
}

void RelPlanOp::TreePrint (ofstream &OStr, int &curPlanId, int level) 
const {
  int *childrenIds = new int[Arity()];
  for (int i = 0; i < Arity(); i++) {
    GetChild(i)->TreePrint(OStr, curPlanId, level + 1);
    childrenIds[i] = curPlanId++;
  }
  NodePrint(OStr, curPlanId, level, Arity(), childrenIds);
  delete [] childrenIds;
}

void RelPlanOp::Print(ostream& OStr) const
{
 PrintThis(OStr);
 OStr << "(";
 for (int i = 0; i < Arity(); i++) {
   childPlanArray[i]->Print(OStr);
   if (i < Arity()-1) {
     OStr << ", ";
    }
  }
 OStr << ")";
}

//
// Methods of RelBasePlanOp
//

// PURPOSE: used in printing the planOp tree
void RelBasePlanOp::NodePrint (ostream &OStr, int curBlockId, int Level, 
			       int numChildren, int *childrenIds) const {
  OStr << "NODE " << relName << endl;
  OStr << "ID " << curBlockId << endl;
  OStr << "LEVEL " << Level << endl;
  if (numChildren > 0) {
    OStr << "CHILDREN ";
    for (int i = 0; i < numChildren; i++) {
      OStr << childrenIds[i] << " ";
    }
    OStr << endl;
  }
  OStr << "COLOR " << 0xFF0000 << endl;
  OStr << "ENDNODE" << endl;
}

XxxErrCode RelBasePlanOp::MakeRelation(const XxxValueEnv* Params,
				       XxxRelation*& ReturnRel)
{
 XxxRelationEngine* RelEng = 
   (XxxRelationEngine*)XxxDBEngineArray[RelationEngineId];

 // if the relation is an environment parameter, get the appropriate handle
 if (Params && (relName[0] == '$')) {
   int Index = atoi(relName+1);
   if ((Index < 0) || (Index > Params->NumArgs)){
     char buf[100];
     sprintf(buf, "Bad relation name: %s", relName);
     REL_NOTE_ERROR(REL_BADQUERY, buf);
     return XXX_NOT_OK;
    }
   // check that the type of the parameter is "relation"
   if (Params->ArgVals.Types[Index] != RelationTypeId) {
     char buf[100];
     sprintf(buf, "Bad type for dynamic relation: %s", relName);
     REL_NOTE_ERROR(REL_BADEXEC, buf);
     return XXX_NOT_OK;
    }
   RelHandle* Handle = (RelHandle *)(Params->ArgValsPtrs[Index]);
   // Get a shell for the relation from the data engine
   if (RelEng->DataEng->GetRelation(Handle,
				    ReturnRel, 
				    CatalogInfo->ImplInfo()) != XXX_OK) {
     char buf[100];
     sprintf(buf, "Unable to get relation %s from env handle", relName);
     REL_NOTE_ERROR(REL_BADEXEC, buf);
     return XXX_NOT_OK;
    }
  }
 else {
   // Else use look up the relation name in the catalogs.
   // Get a shell for the relation from the data engine
   if (RelEng->DataEng->GetRelation(relName,
				    ReturnRel, 
				    CatalogInfo->ImplInfo()) != XXX_TRUE) {
     char buf[100];
     sprintf(buf, "Unable to get relation %s from data engine", relName);
     REL_NOTE_ERROR(REL_BADEXEC, buf);
     return XXX_NOT_OK;
    }
  }
 // fill in the shell
 ReturnRel->CatalogInfo = CatalogInfo;
 return XXX_OK;
}

// PURPOSE: used in printing the planOp tree
void RelFScanPlanOp::NodePrint (ostream &OStr, int curBlockId, int Level, 
				int numChildren, int *childrenIds) const {
  OStr << "NODE File scan of " << relName << endl;
  OStr << "ID " << curBlockId << endl;
  OStr << "LEVEL " << Level << endl;
  if (numChildren > 0) {
    OStr << "CHILDREN ";
    for (int i = 0; i < numChildren; i++) {
      OStr << childrenIds[i] << " ";
    }
    OStr << endl;
  }
  OStr << "COLOR " << 0xFF0000 << endl;
  OStr << "INFO Selection: ";
  if (Selection) {
    Selection->Print(OStr);
  } else {
    OStr << "none";
  }
  OStr << endl;
  OStr << "ENDNODE" << endl;
}

XxxBool RelFScanPlanOp::IsFullFileScan() const
{
  RelImplInfo *ImplInfo = CatalogInfo->ImplInfo();
  /***
  if ((ImplInfo->RelType == REL_STORED) &&
      (ImplInfo->SourceIndex == REL_SHORE_SRC) &&
      (((RelShoreDataSrc::RelShoreImplInfo *)ImplInfo)->ImplType == 
       RelShoreDataSrc::REL_SHORE_FILE_IMPL)) {
    return (Selection ? XXX_FALSE : XXX_TRUE);
  }
  return XXX_FALSE;
  ***/
  return (Selection ? XXX_FALSE : XXX_TRUE);
}

XxxErrCode 
RelFScanPlanOp::MakeRelation(const XxxValueEnv* Params,
			     XxxRelation*& ReturnRel)
{
 if (RelBasePlanOp::MakeRelation(Params, ReturnRel) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, "Error making relation");
   return XXX_NOT_OK;
  }
 // now specify that the relation should use a filescan
 XxxStoredRelation* NewRel = (XxxStoredRelation*)ReturnRel;

 NewRel->AccMode = XxxStoredRelation::FILESCAN;
 NewRel->AccInfo = this;
 return XXX_OK;
}

//Begin: Moustafa
// Adding the fucntions for RelSScanPlanOp
// PURPOSE: used in printing the planOp tree
void RelSScanPlanOp::NodePrint (ostream &OStr, int curBlockId, int Level, 
				int numChildren, int *childrenIds) const {
  OStr << "NODE Stream scan of " << relName << endl;
  OStr << "ID " << curBlockId << endl;
  OStr << "LEVEL " << Level << endl;
  if (numChildren > 0) {
    OStr << "CHILDREN ";
    for (int i = 0; i < numChildren; i++) {
      OStr << childrenIds[i] << " ";
    }
    OStr << endl;
  }
  OStr << "COLOR " << 0xFF0000 << endl;
  OStr << "INFO Selection: ";
  if (Selection) {
    Selection->Print(OStr);
  } else {
    OStr << "none";
  }
  OStr << endl;
  OStr << "ENDNODE" << endl;
}

XxxErrCode 
RelSScanPlanOp::MakeRelation(const XxxValueEnv* Params,
			     XxxRelation*& ReturnRel)
{
 if (RelBasePlanOp::MakeRelation(Params, ReturnRel) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, "Error making relation");
   return XXX_NOT_OK;
  }
 // now specify that the relation should use a streamscan
 XxxStoredRelation* NewRel = (XxxStoredRelation*)ReturnRel;

 NewRel->AccMode = XxxStoredRelation::SSCAN;
 NewRel->AccInfo = this;
 return XXX_OK;
}
//End: Moustafa

// PURPOSE: used in printing the planOp tree
void RelIScanPlanOp::NodePrint (ostream &OStr, int curBlockId, int Level, 
				int numChildren, int *childrenIds) const {
  OStr << "NODE Index scan of " << relName << endl;
  OStr << "ID " << curBlockId << endl;
  OStr << "LEVEL " << Level << endl;
  if (numChildren > 0) {
    OStr << "CHILDREN ";
    for (int i = 0; i < numChildren; i++) {
      OStr << childrenIds[i] << " ";
    }
    OStr << endl;
  }
  OStr << "COLOR " << 0xFF0000 << endl;
  OStr << "INFO Index name: " << IndxName << endl;
  OStr << "INFO Index Plan: ";
  if (IndxPlan) {
    IndxPlan->Print(OStr);
  } else {
    OStr << "none";
  }
  OStr << endl;
  OStr << "INFO Residual Plan: ";
  if (ResidPlan) {
    ResidPlan->Print(OStr);
  } else {
    OStr << "none";
  }
  OStr << endl;
  OStr << "ENDNODE" << endl;
}

XxxErrCode 
RelIScanPlanOp::MakeRelation(const XxxValueEnv* Params,
			     XxxRelation*& ReturnRel)
{
 if (RelBasePlanOp::MakeRelation(Params, ReturnRel) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, "Error making relation");
   return XXX_NOT_OK;
  }
 // now specify that the relation should use an indexscan
 XxxStoredRelation* NewRel = (XxxStoredRelation*)ReturnRel;

 NewRel->AccMode = XxxStoredRelation::INDEXED;
 NewRel->AccInfo = this;
 return XXX_OK;
}

// PURPOSE: used in printing the planOp tree
void RelTmpScanPlanOp::NodePrint (ostream &OStr, int curBlockId, int Level, 
				  int numChildren, int *childrenIds) const {
  OStr << "NODE Scan of temp relation" << endl;
  OStr << "ID " << curBlockId << endl;
  OStr << "LEVEL " << Level << endl;
  if (numChildren > 0) {
    OStr << "CHILDREN ";
    for (int i = 0; i < numChildren; i++) {
      OStr << childrenIds[i] << " ";
    }
    OStr << endl;
  }
  OStr << "COLOR " << 0xFF0000 << endl;
  OStr << "ENDNODE" << endl;
}

XxxErrCode 
RelTmpScanPlanOp::MakeRelation(const XxxValueEnv* Params,
			       XxxRelation*& ReturnRel)
{
 REL_NOTE_ERROR(REL_BADEXEC, "Tmp scan not yet implemented");
 return XXX_NOT_OK;
}

//
// Methods of RelDerivedPlanOp
//

// PURPOSE: used in printing the planOp tree
void RelDerivedPlanOp::NodePrint (ostream &OStr, int curBlockId, int Level, 
				  int numChildren, int *childrenIds) const {
}

XxxErrCode RelDerivedPlanOp::MakeRelation(const XxxValueEnv* Params,
					  XxxRelation*& ReturnRel)
{
 ReturnRel = new XxxDerivedRelation(this, CatalogInfo, Params);
 return XXX_OK;
}

