#ifdef __GNUG__
#pragma implementation
#endif

#include <sys/time.h>
#include "relplanop.h"
#include "relops.h"
#include "reldata.h"
#include "relenv.h"
#include "prd-sort.h"
#include "releng.h"
#include "shorerel.h"
#include "shoresm.h"

//
// Methods of RelSortPlanOp
//
RelSortPlanOp::RelSortPlanOp(RelPlanOp* Child,
			     const XxxSortInfo &SortInfo,
			     XxxBool Distinct,
			     XxxBool UseTempFile,
			     int MaxRunSize,
			     RelCatalogInfo* CatInfo)
: RelDerivedPlanOp(1, &Child, CatInfo), 
  sortInfo(SortInfo), distinct(Distinct), 
  useTempFile(UseTempFile), maxRunSize(MaxRunSize)
{
}

RelSortPlanOp::~RelSortPlanOp() 
{ }

void RelSortPlanOp::NodePrint (ostream &OStr, int curBlockId, int Level, 
			       int numChildren, int *childrenIds) const {
  OStr << "NODE Sort" << endl;
  OStr << "ID " << curBlockId << endl;
  OStr << "LEVEL " << Level << endl;
  if (numChildren > 0) {
    OStr << "CHILDREN ";
    for (int i = 0; i < numChildren; i++) {
      OStr << childrenIds[i] << " ";
    }
    OStr << endl;
  }
  OStr << "COLOR " << 0x00CCCC << endl;
  // print TargetList
  OStr << "INFO Sorting attributes: ";
  for (unsigned int i = 0; i < sortInfo.NumOrderedAttrs(); i++) {
    OStr << "(Attr " << sortInfo[i].AttrIndex << ", " << 
      (sortInfo[i].Asc ? "ascending)" : "descending)");
  }
  OStr << endl;
  OStr << "ENDNODE" << endl;
}

XxxErrCode 
RelSortPlanOp::InitHandle(XxxDerivedRelation::CursorInfo* Handle,
			  XxxRelation* *RelationArray) const
{
  XXX_SM_CHECK;

  if (Handle->ArrayLen != 1) {
    REL_NOTE_ERROR(REL_BADEXEC, "SortPlanOp needs exactly one input");
    return XXX_NOT_OK;
  }
  struct SortHandleInfo *HandleInfo = new struct SortHandleInfo();
  Handle->OperatorHandle = HandleInfo;
  HandleInfo->TempFile = new XxxShorePhysicalId;
  XxxComparisonData CompData;
  CompData.Schema = Handle->ChildSchema[0];
  CompData.SortInfo = &sortInfo;

  if (!useTempFile) {
#ifdef XXX_SHORE_V2
	// Create a temporary file to hold the results
	// sort_file has changed and the output file must now be created by the caller
    XXX_CALL_RET(XxxShoreSM->CreateTempFile(XxxShoreSM->DefaultVol(),
					    *(HandleInfo->TempFile)),
		 XXX_SHOREERR, "Cannot create temporary file");
#endif

    // Convert the logical id to a physical one
    XxxShorePhysicalId FileId;
    XXX_CALL_RET(XxxShoreSM->LogicalToPhysical(((XxxShoreRelation *)
						RelationArray[0])->ShoreOID(),
					       FileId),
		 XXX_SHOREERR, "Cannot get physical id of file");
    // Sort the file, producing a new temporary file
    XXX_CALL_RET(XxxShoreSM->SortPhysicalFile(FileId, 
					      ((XxxShoreRelation *) RelationArray[0])->ShoreOID().VolId, 
					      *(HandleInfo->TempFile), 
					      &XxxCompareFunc, &CompData, 
					      distinct, maxRunSize),
		 XXX_SHOREERR, "Cannot sort existing SHORE file");
  }  // if no temp file
  else {
   int NumRecs = 0;

#ifndef XXX_WIN32
    struct timeval Time[2];
    struct timezone TZ[2];
    if (XXX_THISRELENV()->DebugExec) {
      gettimeofday(&Time[0], &TZ[0]);
    }
#endif
    // Create a temporary file to hold the stream's contents
    XXX_CALL_RET(XxxShoreSM->CreateTempFile(XxxShoreSM->DefaultVol(),
					    *(HandleInfo->TempFile)),
		 XXX_SHOREERR, "Cannot create temporary file");
    // Open a scan
    XXX_CALL_RET(RelationArray[0]->InitCursor1(Handle->ChildCursorArray[0],
					       Handle->Filter, NULL, NULL,
					       Handle->Params),
		 XXX_BADCURSOR, "Unable to open sorted input relation cursor");

    REL_CALL_FN(RelationArray[0]->NextItem(Handle->ChildCursorArray[0],
					   Handle->ChildRecs[0]),
		REL_BADEXEC, "Error getting NextItem from relation");

    // Populate the temporary file
    XxxShoreObjectId RecordId;
    while (Handle->ChildCursorArray[0].IsValid() == XXX_TRUE) {
      NumRecs++;
      REL_CALL_FN(XxxShoreSM->InsertTempRecord(*(HandleInfo->TempFile),
					       Handle->ChildRecs[0]),
		  REL_BADEXEC, "Error inserting record into temp file");
      delete Handle->ChildRecs[0];
      REL_CALL_FN(RelationArray[0]->NextItem(Handle->ChildCursorArray[0],
					     Handle->ChildRecs[0]),
		  REL_BADEXEC, "Error getting NextItem from relation");
    }
    // Close the scan
    if (Handle->ChildCursorArray[0].State &&
	RelationArray[0]->
	CloseCursor(Handle->ChildCursorArray[0]) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADCURSOR, "Unable to close cursor");
      return XXX_NOT_OK;
    }
#ifndef XXX_WIN32
    if (XXX_THISRELENV()->DebugExec) {
      gettimeofday(&Time[1], &TZ[1]);
      double ElapsedTime = (Time[1].tv_sec - Time[0].tv_sec) +
	(Time[1].tv_usec - Time[0].tv_usec)/1000000.0;
      cerr << "Materialized " << NumRecs << " records for sort in " << ElapsedTime 
	   << " seconds." << endl;
    }
#endif

#ifdef XXX_SHORE_V2
	// Create a temporary file to hold the results
	// sort_file has changed and the output file must now be created by the caller
	XxxShorePhysicalId *TempFile = new XxxShorePhysicalId;
    XXX_CALL_RET(XxxShoreSM->CreateTempFile(XxxShoreSM->DefaultVol(),
					    *TempFile),
		 XXX_SHOREERR, "Cannot create temporary file");

    // Sort the temporary file
    if (XxxShoreSM->SortPhysicalFile(*(HandleInfo->TempFile),
				     XxxShoreSM->DefaultVol(), 
				     *TempFile, 
				     &XxxCompareFunc, &CompData, 
				     distinct, maxRunSize) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot sort existing SHORE file");
      return XXX_NOT_OK;
	}

	HandleInfo->TempFile = TempFile;

#else
    // Sort the temporary file
    if (XxxShoreSM->SortPhysicalFile(*(HandleInfo->TempFile),
				     XxxShoreSM->DefaultVol(), 
				     *(HandleInfo->TempFile), 
				     &XxxCompareFunc, &CompData, 
				     distinct, maxRunSize) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot sort existing SHORE file");
      return XXX_NOT_OK;
    }
#endif
  }
  // Open a scan on the temporary file
  HandleInfo->ScanIter = new scan_file_i(*(HandleInfo->TempFile),
					 ss_m::t_cc_file, true);
  XxxErrCode rc = XXX_OK;
  w_rc_t RetVal;
  XXX_CALL((XXX_SHORE_RETFN(HandleInfo->ScanIter->error_code(), RetVal)),
	   XXX_SHOREERR, "Couldn't init scan on temporary file", rc);
  XXX_BADRC_RCRETURN(rc);

  return XXX_OK;
}

XxxErrCode 
RelSortPlanOp::CloseHandle(XxxDerivedRelation::CursorInfo* Handle,
			   XxxRelation* *RelationArray) const
{
  struct SortHandleInfo *HandleInfo = 
    (struct SortHandleInfo *)Handle->OperatorHandle;
  if (HandleInfo && HandleInfo->ScanIter) {
    HandleInfo->ScanIter->finish();
    delete HandleInfo->ScanIter;
  }
  if (HandleInfo && HandleInfo->TempFile) {
    XXX_CALL_RET(XxxShoreSM->DeleteTempFile(*(HandleInfo->TempFile)),
		 XXX_SHOREERR, "Cannot delete temporary file");
    delete HandleInfo->TempFile;
  }
  if (HandleInfo)
    delete HandleInfo;
  return XXX_OK;
}

XxxErrCode RelSortPlanOp::GetNextRecord(XxxSetCursor& SCsr,
					XxxRelation* *RelationArray,
					const XxxRecordSchema* Schema,
					XxxRecord*& ReturnRec) const
{
  XxxDerivedRelation::CursorInfo* Handle = 
    (XxxDerivedRelation::CursorInfo*)(SCsr.State);
  if (!Handle) {
    REL_NOTE_ERROR(REL_BADEXEC, "NULL Exec handle");
    return XXX_NOT_OK;
  }
  struct SortHandleInfo *HandleInfo = 
    (struct SortHandleInfo *)Handle->OperatorHandle;

  XxxErrCode rc;
  pin_i* Pin = 0;
  bool AtEnd;
  w_rc_t RetVal;
  XXX_CALL((XXX_SHORE_RETFN(HandleInfo->ScanIter->next(Pin, 0, AtEnd), 
			    RetVal)),
	   XXX_SHOREERR, "Couldn't scan file", rc);
  XXX_BADRC_RCRETURN(rc);
  
  if (AtEnd) {
    // Scan is complete
    SCsr.AtEnd = XXX_TRUE;
    ReturnRec = NULL;
    return XXX_OK;
  }
  if (!XxxShoreSM->PinValid(Pin, SH)) {
    XXX_NOTE_ERROR(XXX_SHOREERR, "Invalid pin");
    return XXX_NOT_OK;
  }
  
  int TotalSize = Pin->body_size();
  ReturnRec = new XxxRecord(TotalSize);
  if (XxxShoreSM->GetObjectPreAlloc(Pin, (char *)(ReturnRec->DataArea),
				    TotalSize, 0, TotalSize) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_SHOREERR, "Error getting pinned record");
    Pin->unpin();
    return XXX_NOT_OK;
  }
  Pin->unpin();

  return XXX_OK;
}

void RelSortPlanOp::PrintThis(ostream& OStr) const
{
  OStr << PrintString();
  OStr << "[";
  for (unsigned int i = 0; i < sortInfo.NumOrderedAttrs(); i++) {
    OStr << "(Attr " << sortInfo[i].AttrIndex << ", " << 
      (sortInfo[i].Asc ? "ascending)" : "descending)");
  }
  OStr << "]";
}

