#ifdef __GNUG__
#pragma implementation
#endif

#include "relplanop.h"
#include "relops.h"
#include "reldata.h"
#include "releng.h"
#include "planspj.h"
#include "shoresm.h"
//
// Methods of RelSPJPlanOp
//
RelSPJPlanOp::RelSPJPlanOp(int NumChildren, 
			   RelPlanOp* Children[],
			   RelCatalogInfo* CInfo,
			   XxxValExprPlanList* EPlanList,
			   XxxBooleanExpressionPlan** SPredPlans,
			   XxxBooleanExpressionPlan** JPredPlans,
			   XxxBooleanExpressionPlan** RPredPlans,
			   XxxBooleanExpressionPlan* IPlan)
: RelDerivedPlanOp(NumChildren, Children, CInfo)
{
 ExprPlanList = EPlanList;
 SelPredPlans = SPredPlans;
 JoinPredPlans = JPredPlans;
 ResidPredPlans = RPredPlans;
 InitPlan = IPlan;
}

RelSPJPlanOp::~RelSPJPlanOp()
{
 // reclaim the pred arrays allocated in the GeneratePlanOp method of
 // the optimizer
 if (SelPredPlans) {
	for (int i = 0; i < Arity(); i++) {
		if (SelPredPlans[i]) delete SelPredPlans[i];
	}
	 delete [] SelPredPlans;
 }
 if (JoinPredPlans) {
	for (int i = 0; i < Arity(); i++) {
		if (JoinPredPlans[i]) delete JoinPredPlans[i];
	}	 
	delete [] JoinPredPlans;
 }
 if (ResidPredPlans) {
	for (int i = 0; i < Arity(); i++) {
		if (ResidPredPlans[i]) delete ResidPredPlans[i];
	}
	delete [] ResidPredPlans;
 }
}

// OWNER: map DATE: 04/26/97
// PURPOSE: used in printing the planOp tree
// NOTES: none
void RelSPJPlanOp::NodePrint (ostream &OStr, int curBlockId, int Level, 
			      int numChildren, int *childrenIds) const {
  OStr << "NODE " << PrintString() << " SPJ" << endl;
  OStr << "ID " << curBlockId << endl;
  OStr << "LEVEL " << Level << endl;
  if (numChildren > 0) {
    OStr << "CHILDREN ";
    for (int i = 0; i < numChildren; i++) {
      OStr << childrenIds[i] << " ";
    }
    OStr << endl;
  }
  OStr << "COLOR " << 0x00FF00 << endl;
  PrintNodeInfo(OStr);
  OStr << "ENDNODE" << endl;
}

XxxErrCode RelSPJPlanOp::createRecord(XxxDerivedRelation::CursorInfo* Handle,
				      XxxRelation* *RelationArray,
				      const XxxRecordSchema* Schema,
				      XxxRecord*& ReturnRec) const
{
 // a successful instantiation of the inputs has happened. now make a record
 if (ExprPlanList) {
   ASSERT((Schema->Arity() == ExprPlanList->Count())); 
   // get the record values into the ValArray
   XXX_CALL_RET(ExprPlanList->
		Evaluate((const XxxRecord **)Handle->ChildRecs, 
			 (const XxxRecordSchema **)Handle->ChildSchema,
			 Handle->RecValuesArray.Values, 
			 Handle->Params?Handle->Params->Params:0), 
		XXX_BADVALUE, "Error evaluating value list");

   // now make a new record out of the RecValuesArray
   ReturnRec = new XxxRecord(Handle->RecValuesArray, Schema);\
   // Begin: Moustafa
   // must now delete the temp data that is created in evaluate
   for (int i=0; i < Schema->Arity(); i++) {
     if (XxxCoreADTClass::IsCoreADT((*Schema)[i].Type));
     else {
       // reaches here if it is an ADT
       XxxADTClass* AttrClass = XxxADTArray[(*Schema)[i].Type];
       AttrClass->TypeDestroy((XxxADTValue*)(Handle->RecValuesArray.Values[i]),
			      (*Schema)[i].TypeCatInfo);
     }
   }
   // End: Moustafa
 }
 else {
   // a NULL ExprPlanList means there is no projection to apply
   // (so just concat all the records)
   // This is usually used for simple selection operations
   // over temporary results.
   XxxRecord::RecSchemaList RSList;
   int ListSize = 0;
   for (int i = 0; i < Arity(); i++) {
     RSList.Add(Handle->ChildRecs[i], Handle->ChildSchema[i]);
     ListSize += Handle->ChildRecs[i]->Size();
   }
   // concat the records
   ReturnRec = XxxRecord::Concat(RSList, ListSize);
 }

 return XXX_OK;
}

void RelSPJPlanOp::PrintNodeInfo(ostream &OStr) const 
{
 if (InitPlan) {
   OStr << "INFO Initial Plan:";
   InitPlan->Print(OStr);
   OStr << ", Cost: " << InitPlan->GetPlanUsage().FullCost();
   OStr << endl;
  }
 int i;
 if (SelPredPlans) {
   OStr << "INFO Selection Predicate Plans: ";
   for (i = 0; i < Arity(); i++) {
	 if (i==0) 
       OStr << "(" << i << ":" ;
	 else 
       OStr << ",(" << i << ":" ;
     if (SelPredPlans[i]) {
       SelPredPlans[i]->Print(OStr);
       OStr << ", Cost: " << SelPredPlans[i]->GetPlanUsage().FullCost();
     }
     else 
       OStr << "_";
     OStr << ")";
    }
   OStr << endl;
  }
 if (JoinPredPlans) {
   OStr << "INFO Join Predicate Plans: ";
   for (i = 0; i < Arity(); i++) {
	 if (i==0) 
       OStr << "(" << i << ":" ;
	 else 
       OStr << ",(" << i << ":" ;
     if (JoinPredPlans[i]) {
       JoinPredPlans[i]->Print(OStr);
       OStr << ", Cost: " << JoinPredPlans[i]->GetPlanUsage().FullCost();
     }
     else 
       OStr << "_";
     OStr << ")";
    }
   OStr << endl;
  }
 if (ResidPredPlans) {
   OStr << "INFO Residual Predicate Plans: ";
   for (i = 0; i < Arity(); i++) {
	 if (i==0) 
       OStr << "(" << i << ":" ;
	 else 
       OStr << ",(" << i << ":" ;
     if (ResidPredPlans[i]) {
       ResidPredPlans[i]->Print(OStr);
       OStr << ", Cost: " << ResidPredPlans[i]->GetPlanUsage().FullCost();
     }
     else 
       OStr << "_";
     OStr << ")";
    }
   OStr << endl;
  }
 if (ExprPlanList) {
   OStr << "INFO Target List: ";
   ExprPlanList->Print(OStr);
   OStr << endl;
  }
}

void RelSPJPlanOp::PrintThis(ostream& OStr) const
{
 int i;
 OStr << PrintString();
 if (InitPlan) {
   OStr << "[Initial Plan:";
   InitPlan->Print(OStr);
   OStr << "]";
  }
 if (SelPredPlans) {
   OStr << "[Selection Predicate Plans: ";
   for (i = 0; i < Arity(); i++) {
     OStr << "(" << i << ":" ;
     if (SelPredPlans[i])
       SelPredPlans[i]->Print(OStr);
     else 
       OStr << "_";
     OStr << ")";
    }
   OStr << "]";
  }
 if (JoinPredPlans) {
   OStr << "[Join Predicate Plans: ";
   for (i = 0; i < Arity(); i++) {
     OStr << "(" << i << ":" ;
     if (JoinPredPlans[i])
       JoinPredPlans[i]->Print(OStr);
     else 
       OStr << "_";
     OStr << ")";
    }
   OStr << "]";
  }
 if (ResidPredPlans) {
   OStr << "[ Residual Predicate Plans: ";
   for (i = 0; i < Arity(); i++) {
     OStr << "(" << i << ":" ;
     if (ResidPredPlans[i])
       ResidPredPlans[i]->Print(OStr);
     else 
       OStr << "_";
     OStr << ")";
    }
   OStr << "]";
  }
 if (ExprPlanList) {
   OStr << "[Target List: ";
   ExprPlanList->Print(OStr);
   OStr << "]";
  }
}

//
// Methods of RelNLPlanOp
//
XxxErrCode 
RelNLPlanOp::InitHandle(XxxDerivedRelation::CursorInfo* Handle,
			XxxRelation* *RelationArray) const
{
 if (Handle->ArrayLen < 1) {
   REL_NOTE_ERROR(REL_BADOP, "Need at least one input");
   return XXX_NOT_OK;
  }

 // do an InitCursor on just the outer relation in the NL join
 XxxBooleanExpressionPlan* PredPlan =
   (SelPredPlans?SelPredPlans[0]:0);
 if (JoinPredPlans && JoinPredPlans[0]) 
   if (XxxBooleanExpressionPlan::MakeConjunction
       (PredPlan, JoinPredPlans[0]) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
     return XXX_NOT_OK;
    }
 if (ResidPredPlans && ResidPredPlans[0]) 
   if (XxxBooleanExpressionPlan::MakeConjunction
       (PredPlan, ResidPredPlans[0]) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
     return XXX_NOT_OK;
    }

 if (RelationArray[0]->
     InitCursor1(Handle->ChildCursorArray[0],
		 PredPlan,
		 NULL,
		 NULL,
		 Handle->Params,
		 Handle->ArrayLen, 0) != XXX_OK){
   XXX_NOTE_ERROR(XXX_BADCURSOR, "Unable to open outer relation cursor");
   return XXX_NOT_OK;
  }

 // create the operator handle 
 NLJoinInfo* JoinInfo = createJoinInfo(); 
 Handle->OperatorHandle = (void *)JoinInfo;

 // check if initial predicates if any are satisfied
 XxxCoreADTClass::BoolValue Satisfied(XXX_TRUE);
 if (InitPlan) {
   if (InitPlan->Evaluate((const XxxRecord **)Handle->ChildRecs, 
			  (const XxxRecordSchema **)Handle->ChildSchema, 
			  &Satisfied, 
			  (Handle->Params) ? Handle->Params->Params : 0) 
       != XXX_OK) {
     XXX_NOTE_ERROR(XXX_BADEXPR, "Error evaluating init expression!");
     return XXX_NOT_OK;
    }
  } // if InitPlan
 if (!Satisfied.Value) {
   JoinInfo->Valid = XXX_FALSE;
  }

 return XXX_OK;
}

// close all open cursors
XxxErrCode 
RelNLPlanOp::CloseHandle(XxxDerivedRelation::CursorInfo* Handle,
			  XxxRelation* *RelationArray) const
{
 // close cursors on all the inputs
 for (int i = 0; i < Handle->ArrayLen; i++) {
   // NOTE: if this has already been closed, set the cursor state to 0!!
   if (Handle->ChildCursorArray[i].State &&
       RelationArray[i]->
       CloseCursor(Handle->ChildCursorArray[i]) != XXX_OK) {
     char buf[100];
     sprintf(buf, "Unable to close sequence cursor %d\n", i);
     XXX_NOTE_ERROR(XXX_BADCURSOR, buf);
     return XXX_NOT_OK;
    }
  }

 delete ((NLJoinInfo *)(Handle->OperatorHandle));
 return XXX_OK;
}

XxxErrCode RelNLPlanOp::GetNextRecord(XxxSetCursor& SCsr,
				      XxxRelation* *RelationArray,
				      const XxxRecordSchema* Schema,
				      XxxRecord*& ReturnRec) const
{
 // do a simple pipelined nested loops join 
 XxxDerivedRelation::CursorInfo* Handle = 
   (XxxDerivedRelation::CursorInfo*)(SCsr.State);
 if (!Handle) {
   REL_NOTE_ERROR(REL_BADEXEC, "NULL Exec handle");
   return XXX_NOT_OK;
  }

 NLJoinInfo* JoinInfo = ((NLJoinInfo *)(Handle->OperatorHandle));
 if (!JoinInfo->Valid) {
   // usually because of initpred failing
   SCsr.AtEnd = XXX_TRUE;
   ReturnRec = 0;
   return XXX_OK;
  }

 // make this a reference so that any changes to NLPos get reflected in the src
 int& NLPos = JoinInfo->NLPos;

 while (1) {
   // for relations, the consumer is reponsible for reclaiming memory
   //Begin: Moustafa
   //   if (reclaimCurrent(Handle, NLPos) != XXX_OK) {
   if (reclaimCurrent(Handle, NLPos, Handle->ChildSchema[NLPos]) != XXX_OK) {
     //End: Moustafa
     REL_NOTE_ERROR(REL_BADEXEC, "Error reclaiming old memory");
     return XXX_NOT_OK;
   }

   XxxBool Valid = XXX_FALSE;
   if (getNextItem(RelationArray, Handle, NLPos, Valid) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADEXEC, "Error getting NextItem from relation");
     return XXX_NOT_OK;
    }

   if (Valid == XXX_FALSE) {
     // close the cursor on the relation
     if (closeChildScan(Handle, RelationArray, NLPos) != XXX_OK) {
       REL_NOTE_ERROR(REL_BADEXEC, "Error closing relation scan");
       return XXX_NOT_OK;
      }
     if (--NLPos < 0) {
       // nested loops join over
       SCsr.AtEnd = XXX_TRUE;
       ReturnRec = 0;
       return XXX_OK;
      }
    } // if reached eof for relation NLPos
   else { 
     if (++NLPos >= Handle->ArrayLen) {
       // decrement, so that the nxt invocation finds the right NLPos
       NLPos--;	
       XxxCoreADTClass::BoolValue Satisfied(XXX_FALSE);
       if (checkFinal(Handle, Satisfied) != XXX_OK) {
	 REL_NOTE_ERROR(REL_BADEXEC, "Error checking residual plan");
	 return XXX_NOT_OK;
	}
       if (Satisfied.Value) break;
      } // if NLPos >= ArrayLen
     else {
       // Need to re-open a cursor on the relation
       if (openChildScan(Handle, RelationArray, NLPos) != XXX_OK) {
	 REL_NOTE_ERROR(REL_BADEXEC, "Error opening relation scan");
	 return XXX_NOT_OK;
	}
      } // // else if !(NLPos >= ArrayLen)
    } // else if !(reached eof for NLPos)
  } // while

 return createRecord(Handle, RelationArray, Schema, ReturnRec);
}

//
// Methods of RelTNLPlanOp (Tuple-at-a-time NL Join)
//
void RelTNLPlanOp::PrintNodeInfo(ostream &OStr) const 
{
  RelSPJPlanOp::PrintNodeInfo(OStr);
  // now print any additional stuff
}

XxxErrCode RelTNLPlanOp::
reclaimCurrent(XxxDerivedRelation::CursorInfo* Handle,
	       int NLPos) const 
{
 if (Handle->ChildRecs[NLPos]) {
   delete Handle->ChildRecs[NLPos];
   Handle->ChildRecs[NLPos] = 0;
  }
 return XXX_OK;
}

//Begin: Moustafa
// reclaiming a record must be through a call to Reclaim
// not just deleting the record......
XxxErrCode RelTNLPlanOp::
reclaimCurrent(XxxDerivedRelation::CursorInfo* Handle,
	       int NLPos,
	       const XxxRecordSchema* RecSchema) const 
{
  if (Handle->ChildRecs[NLPos]) {
    //Begin: Moustafa:
    // reclaiming a record must be through a call to Reclaim
    // not just deleting the record......
    Handle->ChildRecs[NLPos]->Reclaim(Handle->ChildSchema[NLPos]);
    Handle->ChildRecs[NLPos] = 0;
    //End: Moustafa
  }
  return XXX_OK;
}
//End: Moustafa

XxxErrCode RelTNLPlanOp::
getNextItem(XxxRelation* *RelationArray,
	    XxxDerivedRelation::CursorInfo* Handle,
	    int NLPos, XxxBool& Valid) const 
{

 if (RelationArray[NLPos]->
     NextItem(Handle->ChildCursorArray[NLPos],
	      Handle->ChildRecs[NLPos]) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, "Error getting NextItem from relation");
   return XXX_NOT_OK;
  }
 Valid = Handle->ChildCursorArray[NLPos].IsValid();
 return XXX_OK;
}

XxxErrCode RelTNLPlanOp::
openChildScan(XxxDerivedRelation::CursorInfo* Handle,
	      XxxRelation* *RelationArray,
	      int NLPos) const 
{
 // Extract local predicates
 XxxBooleanExpressionPlan* InnerPredPlan = 
   (SelPredPlans?SelPredPlans[NLPos]:0);
 if (JoinPredPlans && JoinPredPlans[NLPos]) {
   if (XxxBooleanExpressionPlan::MakeConjunction
       (InnerPredPlan, JoinPredPlans[NLPos]) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
     return XXX_NOT_OK;
   }
 }
 if (ResidPredPlans && ResidPredPlans[NLPos]) {
   if (XxxBooleanExpressionPlan::MakeConjunction
       (InnerPredPlan, ResidPredPlans[NLPos]) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
     return XXX_NOT_OK;
   }
 }
 
 if (RelationArray[NLPos]->
     InitCursor1(Handle->ChildCursorArray[NLPos],
		 InnerPredPlan,
		 (const XxxRecord **)Handle->ChildRecs,
		 (const XxxRecordSchema **)Handle->ChildSchema,
		 Handle->Params, Handle->ArrayLen,
		 NLPos) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, "Error opening relation scan");
   return XXX_NOT_OK;
  } // if
 return XXX_OK;
}
XxxErrCode RelTNLPlanOp::
closeChildScan(XxxDerivedRelation::CursorInfo* Handle,
	       XxxRelation* *RelationArray,
	       int NLPos) const 
{
 return RelationArray[NLPos]->CloseCursor(Handle->ChildCursorArray[NLPos]);
}

//
// Methods of RelPNLPlanOp (Page-at-a-time NL Join)
//
// change this to depend on the tuple size and page size
const int RelPNLPlanOp::DEFAULT_PNL_BLOCKCARD = 200;

void RelPNLPlanOp::PrintNodeInfo(ostream &OStr) const 
{
  RelSPJPlanOp::PrintNodeInfo(OStr);
  // now print any additional stuff
}

RelPNLPlanOp::
BlockRelTuples::BlockRelTuples(int Size)
{
 BlockSize = Size;
 NumRecords = 0;
 CurrentPos = 0;
 Records = new XxxRecord*[BlockSize];
 for (int i =0; i < BlockSize; i++) {
   Records[i] = NULL;
  }
}
RelPNLPlanOp::
BlockRelTuples::~BlockRelTuples()
{
 for (int i =0; i < BlockSize; i++) {
   if (Records[i]) delete Records[i];
  }
 delete [] Records;
}

RelPNLPlanOp::
PNLJoinInfo::PNLJoinInfo(int NumChildren, unsigned int BlockCard) 
{
 Arity = NumChildren;
 BlockCardinality = BlockCard;
 RelBlocks = new BlockRelTuples*[Arity];
 for (int i = 0; i < Arity; i++) {
   RelBlocks[i] = new BlockRelTuples(BlockCardinality);
  }
}

RelPNLPlanOp::
PNLJoinInfo::~PNLJoinInfo() 
{
 for (int i = 0; i < Arity; i++) {
   if (RelBlocks[i]) delete RelBlocks[i];
  }
 delete [] RelBlocks;
}
XxxErrCode RelPNLPlanOp::
openChildScan(XxxDerivedRelation::CursorInfo* Handle,
	      XxxRelation* *RelationArray,
	      int NLPos) const
{
 // Extract local predicates
 XxxBooleanExpressionPlan* InnerPredPlan = 
   (SelPredPlans?SelPredPlans[NLPos]:0);
 
 if (RelationArray[NLPos]->
     InitCursor1(Handle->ChildCursorArray[NLPos],
		 InnerPredPlan,
		 (const XxxRecord **)Handle->ChildRecs,
		 (const XxxRecordSchema **)Handle->ChildSchema,
		 Handle->Params, Handle->ArrayLen,
		 NLPos) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, "Error opening relation scan");
   return XXX_NOT_OK;
  } // if
 return XXX_OK;
}
XxxErrCode RelPNLPlanOp::
closeChildScan(XxxDerivedRelation::CursorInfo* Handle,
	       XxxRelation* *RelationArray,
	       int NLPos) const 
{
 return RelTNLPlanOp::closeChildScan(Handle, RelationArray,
				     NLPos);
}

XxxErrCode RelPNLPlanOp::
checkFinal(XxxDerivedRelation::CursorInfo* Handle,
	   XxxCoreADTClass::BoolValue& Satisfied) const 
{
 Satisfied.Value = XXX_TRUE;
 XxxBooleanExpressionPlan* PredPlan = NULL;
 for (int i = 0; i < Handle->ArrayLen; i++) {
   if (JoinPredPlans && JoinPredPlans[i])
     if (XxxBooleanExpressionPlan::MakeConjunction
	 (PredPlan, JoinPredPlans[i]) != XXX_OK) {
       REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
       return XXX_NOT_OK;
      }
   if (ResidPredPlans && ResidPredPlans[i])
     if (XxxBooleanExpressionPlan::MakeConjunction
	 (PredPlan, ResidPredPlans[i]) != XXX_OK) {
       REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
       return XXX_NOT_OK;
      }
  }

 if (PredPlan)
   return PredPlan->Evaluate((const XxxRecord **)Handle->ChildRecs, 
			     (const XxxRecordSchema**)Handle->ChildSchema, 
			     &Satisfied, 
			     Handle->Params?Handle->Params->Params:0);
 return XXX_OK;
}

XxxErrCode RelPNLPlanOp::
getNextItem(XxxRelation* *RelationArray,
	    XxxDerivedRelation::CursorInfo* Handle,
	    int NLPos, XxxBool& Valid) const 
{
 PNLJoinInfo* JoinInfo = (PNLJoinInfo*)(Handle->OperatorHandle);
 BlockRelTuples* TupleBlock = JoinInfo->RelBlocks[NLPos];

 // Step 1: if needed, read in a chunk of records
 if (TupleBlock->NumRecords == 0) {
   if (Handle->ChildCursorArray[NLPos].IsValid()) {
     while (TupleBlock->NumRecords < TupleBlock->BlockSize) {
       // read an input record
       // if (RelationArray[NLPos]->
       // NextItem(Handle->ChildCursorArray[NLPos],
       // TupleBlock->Records[TupleBlock->NumRecords]) 
       if (RelationArray[NLPos]->
	   NextItem(Handle->ChildCursorArray[NLPos],
		    Handle->ChildRecs[NLPos])
	   != XXX_OK) {
	 REL_NOTE_ERROR(REL_BADEXEC, "Error in NextItem");
	 return XXX_NOT_OK;
	}
       // break out of loop if done with input
       if (!Handle->ChildCursorArray[NLPos].IsValid()) 
	 break;

       // need to read it first into Handle->ChildRecs[NLPos]
       // then copy it here! The reason is that any predicates
       // which may have been pushed down into the scan
       // expect to get values from Handle->ChildRecs[NLPos] 
       TupleBlock->Records[TupleBlock->NumRecords] =
	 Handle->ChildRecs[NLPos];
       TupleBlock->NumRecords++;
      } // while loop
    } // if the scan on the input is still valid
   if (TupleBlock->NumRecords == 0) {
     Valid = XXX_FALSE;
     return XXX_OK;
    }
   TupleBlock->CurrentPos = 0;
   // reset all tuple blocks before NLPos
   for (int j = 0; j < NLPos; j++) {
     Handle->ChildRecs[j] = 
       JoinInfo->RelBlocks[j]->Records[0];
     JoinInfo->RelBlocks[j]->CurrentPos = 1;
    }
  } // if new records have to be read

 // Step 2: transfer a record to Handle->ChildRecs[NLPos]
 // At this stage, we are guaranteed that there is at least
 // one new record to provide
 Handle->ChildRecs[NLPos] = 
   TupleBlock->Records[TupleBlock->CurrentPos++];

 Valid = XXX_TRUE;
 return XXX_OK;
}
XxxErrCode RelPNLPlanOp::
reclaimCurrent(XxxDerivedRelation::CursorInfo* Handle,
	       int NLPos) const 
{
 PNLJoinInfo* JoinInfo = (PNLJoinInfo*)(Handle->OperatorHandle);
 BlockRelTuples* TupleBlock = JoinInfo->RelBlocks[NLPos];
 // if used up current bunch of records, delete them
 if (TupleBlock->NumRecords &&
     (TupleBlock->CurrentPos == TupleBlock->NumRecords)) {
   // A decision needs to be made: should these tuples be replaced?
   // Remember that this is a page-at-a-time NL join,
   // but the driving loop in GetNextRecord imagines it
   // is tuple-at-a-time. So, here, instead of deleting
   // the records, check to see if one of the earlier
   // relations has more records its block ...
   for (int TmpNLPos = NLPos-1; TmpNLPos >= 0; TmpNLPos--) {
     BlockRelTuples* TmpTupleBlock = JoinInfo->RelBlocks[TmpNLPos];
     if (TmpTupleBlock->CurrentPos < TmpTupleBlock->NumRecords) {
       // found an earlier block : increment its current record
       Handle->ChildRecs[TmpNLPos] = 
	 TmpTupleBlock->Records[TmpTupleBlock->CurrentPos++];
       // reset all tuple blocks after TmpNLPos
       for (int j = TmpNLPos+1; j <= NLPos; j++) {
	 Handle->ChildRecs[j] = 
	   JoinInfo->RelBlocks[j]->Records[0];
	 // BUG FIX?? JoinInfo->RelBlocks[j]->CurrentPos = 1;
	 JoinInfo->RelBlocks[j]->CurrentPos = 0;
	}
       // return from this function
       return XXX_OK;
      } // if a half-processed block was found
    } // for all tuple blocks before NLPos

   // if it reaches here, it means that the entire contents
   // of all blocks before this have been joined already
   for (int i = 0; i < TupleBlock->NumRecords; i++) {
     delete TupleBlock->Records[i];
     TupleBlock->Records[i] = NULL;
    } // for
   TupleBlock->NumRecords = 0;
   TupleBlock->CurrentPos = 0;
  } // if
 return XXX_OK;
}

//Begin: Moustafa
// reclaiming a record must be through a call to Reclaim
// not just deleting the record......
XxxErrCode RelPNLPlanOp::
reclaimCurrent(XxxDerivedRelation::CursorInfo* Handle,
	       int NLPos,
	       const XxxRecordSchema* RecSchema) const 
{
  PNLJoinInfo* JoinInfo = (PNLJoinInfo*)(Handle->OperatorHandle);
  BlockRelTuples* TupleBlock = JoinInfo->RelBlocks[NLPos];
  // if used up current bunch of records, delete them
  if (TupleBlock->NumRecords &&
      (TupleBlock->CurrentPos == TupleBlock->NumRecords)) {
    // A decision needs to be made: should these tuples be replaced?
    // Remember that this is a page-at-a-time NL join,
    // but the driving loop in GetNextRecord imagines it
    // is tuple-at-a-time. So, here, instead of deleting
    // the records, check to see if one of the earlier
    // relations has more records its block ...
    for (int TmpNLPos = NLPos-1; TmpNLPos >= 0; TmpNLPos--) {
      BlockRelTuples* TmpTupleBlock = JoinInfo->RelBlocks[TmpNLPos];
      if (TmpTupleBlock->CurrentPos < TmpTupleBlock->NumRecords) {
	// found an earlier block : increment its current record
	Handle->ChildRecs[TmpNLPos] = 
	  TmpTupleBlock->Records[TmpTupleBlock->CurrentPos++];
	// reset all tuple blocks after TmpNLPos
	for (int j = TmpNLPos+1; j <= NLPos; j++) {
	  Handle->ChildRecs[j] = 
	    JoinInfo->RelBlocks[j]->Records[0];
	  // BUG FIX?? JoinInfo->RelBlocks[j]->CurrentPos = 1;
	  JoinInfo->RelBlocks[j]->CurrentPos = 0;
	}
	// return from this function
	return XXX_OK;
      } // if a half-processed block was found
    } // for all tuple blocks before NLPos

    // if it reaches here, it means that the entire contents
    // of all blocks before this have been joined already
    for (int i = 0; i < TupleBlock->NumRecords; i++) {
      //SMostafa:
      // reclaiming a record must be through a call to Reclaim
      // not just deleting the record......
      TupleBlock->Records[i]->Reclaim(RecSchema);
      TupleBlock->Records[i] = NULL;
      //EMostafa
    } // for
    TupleBlock->NumRecords = 0;
    TupleBlock->CurrentPos = 0;
  } // if
  return XXX_OK;
}
//End: Moustafa

//
// Methods of RelSMPlanOp
//

void RelSMPlanOp::PrintNodeInfo(ostream &OStr) const 
{
  RelSPJPlanOp::PrintNodeInfo(OStr);
  // now print any additional stuff
}

RelSMPlanOp::RelSMPlanOp(int NumChildren, 
			 RelPlanOp* Children[],
			 RelCatalogInfo* CatInfo,
			 XxxValExprPlanList* ExprPlanList,
			 XxxBooleanExpressionPlan** SelPredPlans,
			 XxxBooleanExpressionPlan** JoinPredPlans,
			 XxxBooleanExpressionPlan** ResidPredPlans,
			 XxxBooleanExpressionPlan* InitPlan)
: RelSPJPlanOp(NumChildren, Children, CatInfo, ExprPlanList, 
	       SelPredPlans, JoinPredPlans, ResidPredPlans, InitPlan)
{
  /*
   * Note: This SPJ must have its input sorted, and as a result,
   *       does not generate complex expressions.  The join plan
   *       must be an equality of two column references; more complex
   *       expressions are disallowed, because if they were generated
   *       here, there could be no guarantee that they were sorted.
   */
  ASSERT((NumChildren == 2));
}

XxxErrCode 
RelSMPlanOp::getNextItem(XxxRelation* *RelationArray,
			 XxxDerivedRelation::CursorInfo* Handle,
			 int Index, XxxBool& Valid) const
{
  if (RelationArray[Index]->
      NextItem(Handle->ChildCursorArray[Index],
	       Handle->ChildRecs[Index]) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADEXEC, "Error getting NextItem from relation");
    return XXX_NOT_OK;
  }
  Valid = Handle->ChildCursorArray[Index].IsValid();
  return XXX_OK;
}

XxxErrCode 
RelSMPlanOp::joinGroups(XxxDerivedRelation::CursorInfo* Handle,
			XxxRelation* *RelationArray,
			const XxxRecordSchema* Schema,
			SMJoinInfo *JoinInfo) const
{
  Group & LeftGroup = JoinInfo->LeftGroup;
  Group & RightGroup = JoinInfo->RightGroup;
  Group & OutputGroup = JoinInfo->OutputGroup;
  // Empty out the output group
  if (!(OutputGroup.IsEmpty()))
    OutputGroup.FlushAllRecords(XXX_FALSE);
  // To re-use old code, temporarily use the Handle; get its current records
  XxxRecord *TempRightRec = Handle->ChildRecs[JoinInfo->RightRelIndex];
  XxxRecord *TempLeftRec = Handle->ChildRecs[JoinInfo->LeftRelIndex];
  // Open a scan on the left (outer) group
  LeftGroup.InitScan();
  XxxRecord *LeftRec = LeftGroup.GetNext();
  XxxRecord *RightRec, *ReturnRec;
  while (LeftGroup.ScanIsValid()) {
    // Open a scan on the right (inner) group
    RightGroup.InitScan();
    RightRec = RightGroup.GetNext();
    while (RightGroup.ScanIsValid()) {
      Handle->ChildRecs[JoinInfo->LeftRelIndex] = LeftRec;
      Handle->ChildRecs[JoinInfo->RightRelIndex] = RightRec;
      // Check any residual predicates
      XxxCoreADTClass::BoolValue Satisfied(XXX_TRUE);
      if (ResidPredPlans[1] &&
	  (ResidPredPlans[1]->
	   Evaluate((const XxxRecord **)Handle->ChildRecs, 
		    (const XxxRecordSchema **)Handle->ChildSchema, 
		    &Satisfied,
		    (Handle->Params) ? Handle->Params->Params : 0) != XXX_OK)) {
	XXX_NOTE_ERROR(XXX_BADEXPR, "Error evaluating residual expression");
	return XXX_NOT_OK;
       }
      if (Satisfied.Value) {
	if (createRecord(Handle, RelationArray, Schema, ReturnRec) != XXX_OK) {
	  XXX_NOTE_ERROR(XXX_BADEXPR, "Error creating record");
	  return XXX_NOT_OK;
	 }
	OutputGroup.AddRecord(ReturnRec);
       } // if Satisfied
      RightRec = RightGroup.GetNext();
    } // inner while loop
    RightGroup.CloseScan();
    LeftRec = LeftGroup.GetNext();
  } // outer while loop
  LeftGroup.CloseScan();

  // Empty out the input groups
  LeftGroup.FlushAllRecords(XXX_TRUE);
  RightGroup.FlushAllRecords(XXX_TRUE);

  // Restore the handle to its previous state
  Handle->ChildRecs[JoinInfo->RightRelIndex] = TempRightRec;
  Handle->ChildRecs[JoinInfo->LeftRelIndex] = TempLeftRec;
  
  return XXX_OK;
}

XxxErrCode 
RelSMPlanOp::parseJoinPlan(XxxDerivedRelation::CursorInfo* Handle,
			   int &LeftAttrIndex, int &LeftRelIndex,
			   int &RightAttrIndex, int &RightRelIndex) const
{
  /* Check that the join plan passed in (which contains the join
   * condition) consists of a single equality RelBooleanPlan
   * such as "s.id = b.id".  Extract the XxxValueExpression on each side
   * of the equality.
   */
  XxxBooleanExpressionPlan* JoinPlan = JoinPredPlans[1];
  ASSERT (JoinPlan);
  XxxRelBooleanPlan* RJoinPlan = (XxxRelBooleanPlan *)JoinPlan;
  ASSERT((JoinPlan->BoolType == XXX_REL_BOOL_EXPR) &&
	 (RJoinPlan->OperType == XXX_EQ_OP));
  // Check to make sure the plans are simple column references
  XxxValueExprPlan *LeftPlan = RJoinPlan->SubPlans[0];
  XxxValueExprPlan *RightPlan = RJoinPlan->SubPlans[1];
  ASSERT ((RightPlan->ValType() == XxxValueExpression::COLUMN_VAR) &&
	  (LeftPlan->ValType() == XxxValueExpression::COLUMN_VAR));
  ASSERT (RightPlan->BasicType() == LeftPlan->BasicType());
  // Extract the indices in the RelationArray of the inputs
  LeftRelIndex = ((XxxUnknownValue::UnknownValuePlan *)
		  (LeftPlan))->SrcChild;
  RightRelIndex = ((XxxUnknownValue::UnknownValuePlan *)
		   (RightPlan))->SrcChild;
  // Extract the indices of the attributes in the records
  LeftAttrIndex = ((XxxUnknownValue::UnknownValuePlan *)
		   (LeftPlan))->SrcAttrIndex;
  RightAttrIndex = ((XxxUnknownValue::UnknownValuePlan *)
		    (RightPlan))->SrcAttrIndex;
  return XXX_OK;
}

XxxErrCode 
RelSMPlanOp::createJoinInfo(XxxDerivedRelation::CursorInfo* Handle,
			    SMJoinInfo *&JoinInfo) const
{
  // Get the locations of the join columns
  int LeftAttrIndex, RightAttrIndex, LeftRelIndex, RightRelIndex;
  if (parseJoinPlan(Handle, LeftAttrIndex, LeftRelIndex,
		    RightAttrIndex, RightRelIndex) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADEXEC,
		   "Attempt to separate join condition failed");
    return XXX_NOT_OK;
  }

  // Make the input groups
  JoinInfo = new SMJoinInfo(RightRelIndex, LeftRelIndex,
			    RightAttrIndex, LeftAttrIndex);

  return XXX_OK;
}

XxxBool 
RelSMPlanOp::isEqual(XxxRecord *LeftRec, int LeftAttrIndex,
		     XxxRecordSchema *LeftSchema,
		     XxxRecord *RightRec, int RightAttrIndex,
		     XxxRecordSchema *RightSchema) const
{
  ASSERT(LeftRec && RightRec);
  ASSERT((*LeftSchema)[LeftAttrIndex].Type == 
	 (*RightSchema)[RightAttrIndex].Type);
  XxxBasicType TypeId = (*LeftSchema)[LeftAttrIndex].Type;
  XxxADTMetaInfo* LMInfo = (*LeftSchema)[LeftAttrIndex].TypeCatInfo;
  XxxADTMetaInfo* RMInfo = (*RightSchema)[RightAttrIndex].TypeCatInfo;
  //Bjarke, sep. 13 2001: added const qualifier to variable
  const char *LtField, *RtField;
  LeftRec->GetField(LeftAttrIndex, LeftSchema, LtField);
  RightRec->GetField(RightAttrIndex, RightSchema, RtField);
  if (XxxCoreADTClass::IsCoreADT(TypeId)) {
    return XxxCoreADTClass::SEquals(TypeId, (XxxADTValue*)LtField, LMInfo,
				    (XxxADTValue*)RtField, RMInfo);
  }
  else {
    return XxxADTArray[TypeId]->Equals((XxxADTValue*)LtField, LMInfo, 
				       (XxxADTValue*)RtField, RMInfo);
   }
}

XxxBool 
RelSMPlanOp::isLessThan(XxxRecord *LeftRec, int LeftAttrIndex,
			XxxRecordSchema *LeftSchema,
			XxxRecord *RightRec, int RightAttrIndex,
			XxxRecordSchema *RightSchema) const
{
  ASSERT(LeftRec && RightRec);
  ASSERT((*LeftSchema)[LeftAttrIndex].Type == 
	 (*RightSchema)[RightAttrIndex].Type);
  XxxBasicType TypeId = (*LeftSchema)[LeftAttrIndex].Type;
  XxxADTMetaInfo* LMInfo = (*LeftSchema)[LeftAttrIndex].TypeCatInfo;
  XxxADTMetaInfo* RMInfo = (*RightSchema)[RightAttrIndex].TypeCatInfo;
  //Bjarke, sep. 13 2001: added const qualifier to variable
  const char *LtField, *RtField;
  LeftRec->GetField(LeftAttrIndex, LeftSchema, LtField);
  RightRec->GetField(RightAttrIndex, RightSchema, RtField);
  if (XxxCoreADTClass::IsCoreADT(TypeId)) {
    return XxxCoreADTClass::SLessThan(TypeId, (XxxADTValue*)LtField, LMInfo, 
				      (XxxADTValue*)RtField, RMInfo);
  }
  else {
    return XxxADTArray[TypeId]->LessThan((XxxADTValue*)LtField, LMInfo, 
					 (XxxADTValue*)RtField, RMInfo);
   }
}
// TODO:L2 there may be currently one problem with SMJoin.
// If there are any selection conditions on the inputs,
// they do not get pushed properly into the pre-sort
// scan of the inputs, because the inputs are treated 
// as DerivedRelations (product of a sortplanop).
// Need to change DerivedRelation code to also consider
// the case where predicates can be performed by the 
// operator directly.

XxxErrCode 
RelSMPlanOp::InitHandle(XxxDerivedRelation::CursorInfo* Handle,
			XxxRelation* *RelationArray) const 
{
  if (Handle->ArrayLen != 2) {
    REL_NOTE_ERROR(REL_BADOP, "Sort-merge join works on 2 inputs only");
    return XXX_NOT_OK;
  }
  // Create the operator handle
  SMJoinInfo* JoinInfo;
  if (createJoinInfo(Handle, JoinInfo) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADEXEC, "Cannot create sort-merge join info");
    return XXX_NOT_OK;
  }
  Handle->OperatorHandle = (void *)JoinInfo;
  // Before doing anything, check the Initial plan; if it fails, we
  // want to stop right here before we sort the inputs.
  XxxCoreADTClass::BoolValue Satisfied(XXX_TRUE);
  if (InitPlan) {
    if (InitPlan->Evaluate((const XxxRecord **)Handle->ChildRecs, 
			   (const XxxRecordSchema **)Handle->ChildSchema, 
			   &Satisfied, 
			   (Handle->Params) ? Handle->Params->Params : 0) 
	!= XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADEXPR, "Error evaluating expression!");
      return XXX_NOT_OK;
    }
  }
  if (!Satisfied.Value) {
    JoinInfo->Valid = XXX_FALSE;
    return XXX_OK;
  }
  // Initialize cursors for both inputs
  for (int i = 0; i < Handle->ArrayLen; i++) {
    XxxBooleanExpressionPlan* PredPlan =
      (SelPredPlans?SelPredPlans[i]:0);
    /***  handle JoinPredPlan as part of SMJoin, and
          ResidPredPlan as part of post-processing of join
    if (JoinPredPlans && JoinPredPlans[i])
      PredPlan = XxxBooleanExpressionPlan::
                 MakeConjunction(PredPlan, JoinPredPlans[i]);
    if (ResidPredPlans && ResidPredPlans[i]) 
      PredPlan = XxxBooleanExpressionPlan::
                 MakeConjunction(PredPlan, ResidPredPlans[i]);
    ***/

    // CurrJoinPos should be 0 because there is no join context 
    // for a sort-merge join -- praveen. 
    if (RelationArray[i]->
	InitCursor1(Handle->ChildCursorArray[i],
		    PredPlan, 
		    NULL,
		    NULL,
		    Handle->Params, Handle->ArrayLen, 0) != XXX_OK) {
      char buf[100];
      sprintf (buf, "Unable to open cursor on input %d", i);
      XXX_NOTE_ERROR(XXX_BADCURSOR, buf);
      return XXX_NOT_OK;
    }
  }
  // Read in the first record of each of the two inputs
  XxxBool ScanValid = XXX_TRUE;
  if (getNextItem(RelationArray, Handle, 
		  JoinInfo->LeftRelIndex, ScanValid) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADEXEC, "Cannot get next record from input");
    return XXX_NOT_OK;
  }
  if (!ScanValid) {
    JoinInfo->Valid = XXX_FALSE;
    return XXX_OK;
  }
  if (getNextItem(RelationArray, Handle, 
		  JoinInfo->RightRelIndex, ScanValid) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADEXEC, "Cannot get next record from input");
    return XXX_NOT_OK;
  }
  if (!ScanValid) {
    JoinInfo->Valid = XXX_FALSE;
    return XXX_OK;
  }
  return XXX_OK;
}

XxxErrCode 
RelSMPlanOp::CloseHandle(XxxDerivedRelation::CursorInfo* Handle,
			 XxxRelation* *RelationArray) const
{
  // Close cursors on all the inputs
  for (int i = 0; i < Handle->ArrayLen; i++) {
    if (Handle->ChildCursorArray[i].State && RelationArray[i]->
	CloseCursor(Handle->ChildCursorArray[i]) != XXX_OK) {
      char buf[100];
      sprintf(buf, "Unable to close sequence cursor %d", i);
      XXX_NOTE_ERROR(XXX_BADCURSOR, buf);
      return XXX_NOT_OK;
    }
  }
  delete ((SMJoinInfo *)(Handle->OperatorHandle));
  return XXX_OK;
}

// TODO: Looks like the input records never get reclaimed properly!!
XxxErrCode 
RelSMPlanOp::GetNextRecord(XxxSetCursor& SCsr,
			   XxxRelation* *RelationArray,
			   const XxxRecordSchema* Schema,
			   XxxRecord*& ReturnRec) const 
{
  /*
   * Description: The notion of a "group", or collection of records
   * with the same critical attribute (the one being joined upon)
   * defines how Sort-merge works.  The left and right inputs are 
   * scanned until two records are found which satisfy the join's
   * equality condition.  Then, while the incoming streams belong to
   * the same group, they are scanned into group objects.  When the
   * scanning of the groups is complete, the two input groups are
   * joined in a cross-product fashion to produce an output group,
   * which holds the result tuples from the join.
   *
   * Below is a large "while" loop; in pseudocode, this is what it does:
   *
   *   while (both inputs are valid) {
   *     while (left scan is valid and left_tuple < right_tuple)
   *       advance left scan
   *     while (left scan is valid,
   *            right scan is valid and right_tuple < left_tuple)
   *       advance right scan
   *     if (left_tuple == right_tuple) {
   *       scan in both groups and join them
   *     }
   *   }
   *
   * The tricky part about this code is how the iterator interface
   * is preserved for the plan operator.  When the output group has
   * a valid scan upon it, the next tuple to be returned is the next
   * tuple from that scan.  If and when that scan fails, a new output
   * group must be produced, a scan must be initialized upon it, and 
   * its first tuple must be returned.  The operator produces no more
   * tuples when the output group is empty, AND the JoinInfo->Valid,
   * which indicates whether the scans continue to be valid, is XXX_FALSE.
   *
   * Another subtlety is that look-ahead records are maintained for 
   * each of the two inputs.  When we are scanning in groups, we know
   * the group is completely scanned in when we scan a record which doesn't
   * belong in it; but by this time the record has been read off of the
   * input and must be maintained.  Thus when we start to read a new group,
   * we start with the records in the Handle->ChildRecs[] array, which
   * holds the look-aheads.
   */

  // Get the SMJoinInfo from the operator handle
  XxxDerivedRelation::CursorInfo* Handle = 
    (XxxDerivedRelation::CursorInfo*)(SCsr.State);
  if (!Handle) {
    REL_NOTE_ERROR(REL_BADEXEC, "NULL Exec handle");
    return XXX_NOT_OK;
  }  
  SMJoinInfo* JoinInfo = ((SMJoinInfo *)(Handle->OperatorHandle));
  Group & LeftGroup = JoinInfo->LeftGroup;
  Group & RightGroup = JoinInfo->RightGroup;
  Group & OutputGroup = JoinInfo->OutputGroup;
  // If there is a valid scan on the OutputGroup, return its next tuple
  if (OutputGroup.ScanIsValid() == XXX_TRUE) {
    ReturnRec = OutputGroup.GetNext();
    if (OutputGroup.ScanIsValid() == XXX_TRUE)
      return XXX_OK;
  }
  // Empty the output group; all of its records have been returned
  OutputGroup.FlushAllRecords(XXX_FALSE);
  // If there are no further records to return and a scan has finished, return
  if (!(JoinInfo->Valid)) {
    SCsr.AtEnd = XXX_TRUE;
    ReturnRec = 0;
    return XXX_OK;
  }
  /* Populate the next output group.  Keep advancing the scan in both
   * input relations until a set is found which satisfies the join condition.
   * When one is found, populate the InputGroup corresponding to each input
   * as long as the incoming tuples belongs in the group.  Note that if this
   * is not the first time this function is called, Handle->ChildRecs[] has
   * the "look-ahead" records that were retrieved last time but were found
   * not to belong to the last group.
   */
  XxxBool LeftScanValid = XXX_TRUE, RightScanValid = XXX_TRUE;
  XxxRecord *LastLeftRec, *LastRightRec = Handle->ChildRecs[JoinInfo->RightRelIndex];
  while ((LeftScanValid == XXX_TRUE) && (RightScanValid == XXX_TRUE)) {
    // Advance the left input
    while ((LeftScanValid == XXX_TRUE) &&
	   (isLessThan(Handle->ChildRecs[JoinInfo->LeftRelIndex], 
		       LeftGroup.AttrIndex(),
		       Handle->ChildSchema[JoinInfo->LeftRelIndex],
		       LastRightRec, RightGroup.AttrIndex(),
		       Handle->ChildSchema[JoinInfo->RightRelIndex]) == XXX_TRUE)) {
      if (getNextItem(RelationArray, Handle, 
		      JoinInfo->LeftRelIndex, LeftScanValid) != XXX_OK) {
	REL_NOTE_ERROR(REL_BADEXEC, "Cannot get next record from input");
	return XXX_NOT_OK;
      }
    } // while advancing left scan
    LastLeftRec = Handle->ChildRecs[JoinInfo->LeftRelIndex];
    // Advance the right input
    while ((LeftScanValid == XXX_TRUE) && (RightScanValid == XXX_TRUE) &&
	   (isLessThan(Handle->ChildRecs[JoinInfo->RightRelIndex], 
		       RightGroup.AttrIndex(),
		       Handle->ChildSchema[JoinInfo->RightRelIndex],
		       LastLeftRec, LeftGroup.AttrIndex(),
		       Handle->ChildSchema[JoinInfo->LeftRelIndex]) == XXX_TRUE)) {
      if (getNextItem(RelationArray, Handle, JoinInfo->RightRelIndex, 
		      RightScanValid) != XXX_OK) {
	REL_NOTE_ERROR(REL_BADEXEC, "Cannot get next record from input");
	return XXX_NOT_OK;
      }
    } // while advancing right scan
    LastRightRec = Handle->ChildRecs[JoinInfo->RightRelIndex];
    // If a match has been found, scan each group in and compute the
    // output group
    if ((LeftScanValid == XXX_TRUE) && (RightScanValid == XXX_TRUE) &&
	(isEqual(LastLeftRec, LeftGroup.AttrIndex(),
		Handle->ChildSchema[JoinInfo->LeftRelIndex],
		LastRightRec, RightGroup.AttrIndex(),
		Handle->ChildSchema[JoinInfo->RightRelIndex]) == XXX_TRUE)) {
      // Scan in the left group
      while ((LeftScanValid == XXX_TRUE) &&
	     (isEqual(LastLeftRec, LeftGroup.AttrIndex(),
		      Handle->ChildSchema[JoinInfo->LeftRelIndex],
		      Handle->ChildRecs[JoinInfo->LeftRelIndex], 
		      LeftGroup.AttrIndex(),
		      Handle->ChildSchema[JoinInfo->LeftRelIndex]) == XXX_TRUE)) {
	LeftGroup.AddRecord(Handle->ChildRecs[JoinInfo->LeftRelIndex]);
	if (getNextItem(RelationArray, Handle, JoinInfo->LeftRelIndex, 
			LeftScanValid) != XXX_OK){
	  REL_NOTE_ERROR(REL_BADEXEC, "Cannot get next record from input");
	  return XXX_NOT_OK;
	}
      } // while scanning in left group
      // Scan in the right group
      while ((RightScanValid == XXX_TRUE) &&
	     (isEqual(LastRightRec, RightGroup.AttrIndex(),
		      Handle->ChildSchema[JoinInfo->RightRelIndex],
		      Handle->ChildRecs[JoinInfo->RightRelIndex], 
		      RightGroup.AttrIndex(),
		      Handle->ChildSchema[JoinInfo->RightRelIndex]) == XXX_TRUE)) {
	RightGroup.AddRecord(Handle->ChildRecs[JoinInfo->RightRelIndex]);
	if (getNextItem(RelationArray, Handle, JoinInfo->RightRelIndex, 
			RightScanValid)!=XXX_OK){
	  REL_NOTE_ERROR(REL_BADEXEC, "Cannot get next record from input");
	  return XXX_NOT_OK;
	}
      } // while scanning in right group
      // Now both input groups are non-empty; compute the output group
      if (joinGroups(Handle, RelationArray, Schema, JoinInfo) != XXX_OK) {
	REL_NOTE_ERROR(REL_BADEXEC, "Cannot compute output group");
	return XXX_NOT_OK;
      }
      // The OutputGroup may be empty if the residual preds were false
      if (!OutputGroup.IsEmpty()) break;
      else // we need to continue : make sure LastRightRec is correct
        LastRightRec = Handle->ChildRecs[JoinInfo->RightRelIndex];
    } // if a match was found
  } // outermost while loop

  // If either scan is complete, note it in the JoinInfo
  if (!LeftScanValid || !RightScanValid)
    JoinInfo->Valid = XXX_FALSE;
  // If the join was successful, a the OutputGroup has been repopulated.
  // Return its first tuple.
  if (!OutputGroup.IsEmpty()) {
    OutputGroup.InitScan();
    ReturnRec = OutputGroup.GetNext();
    ASSERT (OutputGroup.ScanIsValid());
    return XXX_OK;
  }
  // Otherwise, one of the inputs' scans is complete and the join need not
  // proceed.  Return end-of-scan.
  SCsr.AtEnd = XXX_TRUE;
  ReturnRec = 0;
  return XXX_OK;
}

//
// Methods of RelHashJoinPlanOp
//

void RelHashJoinPlanOp::PrintNodeInfo(ostream &OStr) const 
{
  RelSPJPlanOp::PrintNodeInfo(OStr);
  // now print any additional stuff
}

RelHashJoinPlanOp::RelHashJoinPlanOp(int NumChildren, 
								RelPlanOp* Children[],
								RelCatalogInfo* CatInfo,
								XxxValExprPlanList* ExprPlanList,
								XxxBooleanExpressionPlan** SelPredPlans,
								XxxBooleanExpressionPlan** JoinPredPlans,
								XxxBooleanExpressionPlan** ResidPredPlans,
								XxxBooleanExpressionPlan* InitPlan)
		: RelSPJPlanOp(NumChildren, Children, CatInfo, ExprPlanList, 
	       SelPredPlans, JoinPredPlans, ResidPredPlans, InitPlan)
{
  /*
   *       must be an equality of two column references; more complex
   *       expressions are disallowed
   */
  ASSERT((NumChildren == 2));
}

XxxErrCode 
RelHashJoinPlanOp::getNextItem(XxxRelation* *RelationArray,
			 XxxDerivedRelation::CursorInfo* Handle,
			 int Index, XxxBool& Valid) const
{
	
	if (RelationArray[Index]->
		NextItem(Handle->ChildCursorArray[Index],
		Handle->ChildRecs[Index] ) != XXX_OK) {
		REL_NOTE_ERROR(REL_BADEXEC, "Error getting NextItem from relation");
		return XXX_NOT_OK;
	}
	Valid = Handle->ChildCursorArray[Index].IsValid();
	

	return XXX_OK;
}

// -----------------------------------------------------------------------------------------------
// newly added code for XxxRecordBuffer and PopulateThread
// used by the threaded symmetric hash join implementation 
//
// @author	tokweehy@comp.nus.edu.sg
// @version	13 June 2000
// -----------------------------------------------------------------------------------------------
RelHashJoinPlanOp::XxxRecordBuffer::XxxRecordBuffer(int bufIndex) 
{
	bIndex = bufIndex;	// used for debugging purpose to indicate which buffer is being used

	for (int i =0; i < XXX_MAX_SHAREDBUFFER_RECORDS; i++) {
		// Init all XxxRecords to be NULL
		Table[i] = NULL;
		State[i] = XXX_FALSE;
	}

	// Cond var are created
	CondFull  = new XxxCondVar("f1"); 
	CondEmpty = new XxxCondVar("e1");
	Mutex = new XxxBaseMutex();

	// used to indicate the begin/end of the RecordBuffer
	First = 0;
	Last = 0;

	this->moreRecs = XXX_TRUE;

	timeoutCount = 0;
	itemCount = 0;
	isValid = XXX_FALSE;
}

RelHashJoinPlanOp::XxxRecordBuffer::~XxxRecordBuffer() {
   // perform housekeeping before Record Buffer is destroyed
	for (int i =0; i < XXX_MAX_SHAREDBUFFER_RECORDS; i++) 
		delete Table[i];
}

XxxBool RelHashJoinPlanOp::XxxRecordBuffer::IsFull() { return (((Last - First)%XXX_MAX_SHAREDBUFFER_RECORDS) == XXX_MAX_SHAREDBUFFER_RECORDS); }

XxxBool RelHashJoinPlanOp::XxxRecordBuffer::IsEmpty() { return ((Last-First) == 0);	}

// Add Item to RecordBuffer, used by Add to add item
// Assume proper synchornization mechanism is already set up by Add
XxxErrCode 
RelHashJoinPlanOp::XxxRecordBuffer::AddItem(XxxRecord *Item) {
	   if ( IsFull()) {
		   // should never reach here coz buffer should not
		   // be full at all.
		   return XXX_NOT_OK;
	   }

	   Table[Last] = Item;
	   State[Last] = XXX_TRUE;
	
	   Last = (Last + 1 ) % XXX_MAX_SHAREDBUFFER_RECORDS;
	   return XXX_OK;
}

// Add Item to Record Buffer, with synchornization mechanisms in place		
XxxErrCode 
RelHashJoinPlanOp::XxxRecordBuffer::Add(XxxRecord *Item) {
	// Ensure that buffer is not full before adding item
	// If buffer is full, wait conditionally till signal given
	// cout << "sb " << bIndex << " add" << endl;   

	while(1) {
		if (IsFull()) {
			// Buffer is full, conditional wait
			cout << "Buffer is full!" << endl;
			CondFull->Wait();
		}

		// Acquire mutex before adding to buffer
		W_COERCE(Mutex->acquire());
			

		ASSERT(Item);
		this->moreRecs = XXX_TRUE;

		if (AddItem(Item) != XXX_OK) {
			XXX_NOTE_ERROR(XXX_BADEXEC, "Unable to add item in record buffer");
			return XXX_NOT_OK;
		}

		// Mutex no longer needed. Release it
		Mutex->release();

		// signal to indicate that it is no longer empty
		// to allow items waiting on an empty buffer to proceed
		CondEmpty->Signal();

		//if ( this->moreRecs == XXX_TRUE) 
		//XxxBaseThread::me()->yield();
		return XXX_OK;
	}

	// should never reach here!
	return XXX_NOT_OK;
}


XxxErrCode 
RelHashJoinPlanOp::XxxRecordBuffer::Pop() {

	if (IsEmpty()) {
		return XXX_NOT_OK;
	}
	curRec = Table[First];
	State[First] = XXX_FALSE;


	while (State[First] == XXX_FALSE && !IsEmpty()) {
		First = (First+1)%XXX_MAX_SHAREDBUFFER_RECORDS;
	}

	return XXX_OK;
	
}

void RelHashJoinPlanOp::XxxRecordBuffer::signal() {
	// set flag to indicate populatethread has terminated
	this->moreRecs = XXX_FALSE;
	CondEmpty->Signal();
	CondFull->Signal();	
}

XxxErrCode 
RelHashJoinPlanOp::XxxRecordBuffer::Get() {
	//	cout << "sb " << bIndex << " get" << endl;

	while (1) {
		//cout << "sb " << bIndex << " Get: Acquiring mutex " << endl;
		//cout << "sb " << bIndex << " Get: Mutex acquired! " << endl;

		isValid = XXX_FALSE;
		
		if ( IsEmpty() ) {
		   	//cout << "sb Get: Empty buffer!!! " << timeoutCount << endl;
		   	//cout << "conditional wait! " << timeoutCount << endl;
			if ( this->moreRecs )
				CondEmpty->Wait();
		}

		if ( moreRecs == XXX_FALSE && IsEmpty() ) {
			// populat thread has terminated
			//cout << "sb " << bIndex << "No more recs, returning false " << endl;

			//(this->Mutex)->release();	
			isValid = XXX_FALSE;
			return XXX_NOT_OK;
		}
		else {
			//cout << "sb " << bIndex << " No longer empty?? Grab mutex!!" << endl;
			W_COERCE((this->Mutex)->acquire());
		
			if (Pop() != XXX_OK) {
				XXX_NOTE_ERROR(XXX_BADEXEC, "Unable to get item from request buffer");
				cout << "sb Get: Unable to pop item off sb ?? " << endl;
						
				(this->Mutex)->release();
				return XXX_NOT_OK;
			}
			else {
				CondFull->Signal();

				(this->Mutex)->release();
				//cout << "sb " << bIndex << " Get: loop: Mutex released with item! (total :" << itemCount << ")" << endl;
				isValid = XXX_TRUE;
				return XXX_OK;
			}
		}
	  
		// iterate here until there is a value of timeout occurs
		(this->Mutex)->release();	
		//cout << "sb Get: Mutex released! " << endl;
		
		/*
		if ( timeoutCount > 10 ) {
				(this->Mutex)->release();	
				isValid = XXX_FALSE;
				return XXX_NOT_OK;				
		}
		*/
	} // while


}

XxxBool 
RelHashJoinPlanOp::XxxRecordBuffer::IsValid() { return isValid; }

// get the item from the shared buffer specified by Index
XxxErrCode 
RelHashJoinPlanOp::getNextItemSb(HashJoinInfo* JoinInfo, 
							   int Index, XxxBool& Valid) const
{
	if ( JoinInfo->sb[Index]->Get() != XXX_NOT_OK ) {
		Valid = JoinInfo->sb[Index]->IsValid();
		return XXX_OK;
	}
	Valid = XXX_FALSE;
	return XXX_NOT_OK;
}



void RelHashJoinPlanOp::PopulateThread::run()
{
   XxxBeginTrans();

   // init the cursor for the specific relation
   if ( populateRelationArray[populateIndex]->InitCursor1(childCursorArray[populateIndex],
				PredPlan, NULL, NULL, parameters, 
				ArrayLen, 0 ) != XXX_OK ) 
			{
				char buf[100];
				sprintf(buf, "Unable to open cursor on input %d", populateIndex);
				XXX_NOTE_ERROR(XXX_BADCURSOR, buf);
				ShutDown();
			}

   // loop will end when there is no more items, and ShutDown is initiated
   while ( !Shutdown->isSet()) {
	   if ( populateRelationArray[populateIndex]->NextItem(childCursorArray[populateIndex],CurRec) != XXX_OK ) {
			   REL_NOTE_ERROR(REL_BADEXEC, "Error getting NextItem from relation");
			   Buffer->signal();		
			   ShutDown();
	   }
	   else {
		   if (childCursorArray[populateIndex].IsValid()) {
			   ASSERT(CurRec);
			   Buffer->Add(CurRec);

#ifdef XXX_SHORE_V2
			   this->yield();
#endif
			   //XxxBaseThread::me()->yield();
		   }
		   else {
			   // indicate scan is not valid	
			   Buffer->signal();
			   ShutDown();			   			   	
		   } // else
	   } // else
   } // while


   // no more items, close cursor
   if (childCursorArray[populateIndex].State && populateRelationArray[populateIndex]->
	CloseCursor(childCursorArray[populateIndex]) != XXX_OK) {
      char buf[100];
      sprintf(buf, "Unable to close sequence cursor %d", populateIndex);
      XXX_NOTE_ERROR(XXX_BADCURSOR, buf);
	  cout << buf << endl;
   }

   XxxCommitTrans();

   // end the thread's execution
   this->end();
}

XxxErrCode 
RelHashJoinPlanOp::parseJoinPlan(XxxDerivedRelation::CursorInfo* Handle,
			   int &LeftAttrIndex, int &LeftRelIndex,
			   int &RightAttrIndex, int &RightRelIndex) const
{
  /* Check that the join plan passed in (which contains the join
   * condition) consists of a single equality RelBooleanPlan
   * such as "s.id = b.id".  Extract the XxxValueExpression on each side
   * of the equality.
   */
  XxxBooleanExpressionPlan* JoinPlan = JoinPredPlans[1];
  ASSERT (JoinPlan);
  XxxRelBooleanPlan* RJoinPlan = (XxxRelBooleanPlan *)JoinPlan;
  ASSERT((JoinPlan->BoolType == XXX_REL_BOOL_EXPR) &&
	 (RJoinPlan->OperType == XXX_EQ_OP));
  
  // Check to make sure the plans are simple column references
  XxxValueExprPlan *LeftPlan = RJoinPlan->SubPlans[0];
  XxxValueExprPlan *RightPlan = RJoinPlan->SubPlans[1];
  ASSERT ((RightPlan->ValType() == XxxValueExpression::COLUMN_VAR) &&
	  (LeftPlan->ValType() == XxxValueExpression::COLUMN_VAR));
  ASSERT (RightPlan->BasicType() == LeftPlan->BasicType());
  
  // Extract the indices in the RelationArray of the inputs
  LeftRelIndex = ((XxxUnknownValue::UnknownValuePlan *)
		  (LeftPlan))->SrcChild;
  RightRelIndex = ((XxxUnknownValue::UnknownValuePlan *)
		   (RightPlan))->SrcChild;
  
  // Extract the indices of the attributes in the records
  LeftAttrIndex = ((XxxUnknownValue::UnknownValuePlan *)
		   (LeftPlan))->SrcAttrIndex;
  RightAttrIndex = ((XxxUnknownValue::UnknownValuePlan *)
		    (RightPlan))->SrcAttrIndex;
  return XXX_OK;
}

XxxErrCode 
RelHashJoinPlanOp::createJoinInfo(XxxDerivedRelation::CursorInfo* Handle,
			    HashJoinInfo *&JoinInfo) const
{
  // Get the locations of the join columns
  int LeftAttrIndex, RightAttrIndex, LeftRelIndex, RightRelIndex;

  if (parseJoinPlan(Handle, LeftAttrIndex, LeftRelIndex,
		    RightAttrIndex, RightRelIndex) != XXX_OK) {
	REL_NOTE_ERROR(REL_BADEXEC,
		   "Attempt to separate join condition failed");
    return XXX_NOT_OK;
  }

  // Make the input groups
  JoinInfo = new HashJoinInfo(RightRelIndex, LeftRelIndex,
			    RightAttrIndex, LeftAttrIndex);

  return XXX_OK;
}

XxxBool 
RelHashJoinPlanOp::isEqual(XxxRecord *LeftRec, int LeftAttrIndex,
		     XxxRecordSchema *LeftSchema,
		     XxxRecord *RightRec, int RightAttrIndex,
		     XxxRecordSchema *RightSchema) const
{
  ASSERT(LeftRec && RightRec);
  ASSERT((*LeftSchema)[LeftAttrIndex].Type == 
	 (*RightSchema)[RightAttrIndex].Type);
  XxxBasicType TypeId = (*LeftSchema)[LeftAttrIndex].Type;
  XxxADTMetaInfo* LMInfo = (*LeftSchema)[LeftAttrIndex].TypeCatInfo;
  XxxADTMetaInfo* RMInfo = (*RightSchema)[RightAttrIndex].TypeCatInfo;
  //Bjarke, sep. 13 2001: added const qualifier to variable
  const char *LtField, *RtField;
  LeftRec->GetField(LeftAttrIndex, LeftSchema, LtField);
  RightRec->GetField(RightAttrIndex, RightSchema, RtField);

  if (XxxCoreADTClass::IsCoreADT(TypeId)) {
    return XxxCoreADTClass::SEquals(TypeId, (XxxADTValue*)LtField, LMInfo,
				    (XxxADTValue*)RtField, RMInfo);
  }
  else {
    return XxxADTArray[TypeId]->Equals((XxxADTValue*)LtField, LMInfo, 
				       (XxxADTValue*)RtField, RMInfo);
   }
}


XxxErrCode 
RelHashJoinPlanOp::InitHandle(XxxDerivedRelation::CursorInfo* Handle,
			XxxRelation* *RelationArray) const 
{
  if (Handle->ArrayLen != 2 ) {
	REL_NOTE_ERROR(REL_BADOP, "Hash join works on 2 inputs only");
	return XXX_NOT_OK;
  }

  // Create the hash join operator
  HashJoinInfo *JoinInfo;
  if ( createJoinInfo(Handle,JoinInfo) != XXX_OK ) {
	REL_NOTE_ERROR(REL_BADEXEC, "Cannot create hash join info");
	return XXX_NOT_OK;
  }
   
  Handle->OperatorHandle = (void *)JoinInfo;

  // Need to check initial plan, if it fails, do not need to proceed
  XxxCoreADTClass:: BoolValue Satisfied(XXX_TRUE);
  if ( InitPlan ) {
	if ( InitPlan->Evaluate((const XxxRecord **)Handle->ChildRecs,
			(const XxxRecordSchema **) Handle->ChildSchema,
			&Satisfied, (Handle->Params) ? Handle->Params->Params : 0 )
			!= XXX_OK ) {
		XXX_NOTE_ERROR(XXX_BADEXPR, "Error evaluating expression!");
		return XXX_NOT_OK;
	}
  }

  if ( !Satisfied.Value ) {
	JoinInfo->Valid = XXX_FALSE;
	return XXX_OK;
  }

  // Initialize cursors for both inputs
  for (int i = 0; i < Handle->ArrayLen; i++) {
	  XxxBooleanExpressionPlan* PredPlan = ( SelPredPlans?SelPredPlans[i]:0);
	 /***  handle JoinPredPlan as part of Hash Join, and
          ResidPredPlan as part of post-processing of join
		if (JoinPredPlans && JoinPredPlans[i])
			PredPlan = XxxBooleanExpressionPlan::
                 MakeConjunction(PredPlan, JoinPredPlans[i]);
				 
		if (ResidPredPlans && ResidPredPlans[i]) 
			PredPlan = XxxBooleanExpressionPlan::
                 MakeConjunction(PredPlan, ResidPredPlans[i]);
    ***/

	/* depreceated! Used by non-threaded symmetric hash join
	if ( RelationArray[i]->InitCursor1(Handle->ChildCursorArray[i],
			PredPlan, NULL, NULL, Handle->Params, 
			Handle->ArrayLen, 0 ) != XXX_OK ) 
	{
		char buf[100];
		sprintf(buf, "Unable to open cursor on input %d", i);
		XXX_NOTE_ERROR(XXX_BADCURSOR, buf);
		return XXX_NOT_OK;
	}
  */
	
	  // start the  populate thread here
	  // @author tokweehy@comp.nus.edu.sg
	  // @version 7 Jun 2000

	  JoinInfo->PopulThread[i] = new PopulateThread(RelationArray,Handle->ChildCursorArray,i,JoinInfo->sb[i],PredPlan,Handle->Params,Handle->ArrayLen);


  }

  
  return XXX_OK;
} // Init Handle

XxxErrCode 
RelHashJoinPlanOp::CloseHandle(XxxDerivedRelation::CursorInfo* Handle,
			 XxxRelation* *RelationArray) const
{
  // Close cursors on all the inputs
	/*
  for (int i = 0; i < Handle->ArrayLen; i++) {
    if (Handle->ChildCursorArray[i].State && RelationArray[i]->
	CloseCursor(Handle->ChildCursorArray[i]) != XXX_OK) {
      char buf[100];
      sprintf(buf, "Unable to close sequence cursor %d", i);
      XXX_NOTE_ERROR(XXX_BADCURSOR, buf);
      return XXX_NOT_OK;
    }
  }
  */
	HashJoinInfo* JoinInfo = ((HashJoinInfo *) (Handle->OperatorHandle));

	for (int i = 0; i < Handle->ArrayLen; i++) {
		delete JoinInfo->sb[i];
	}
  
  delete ((HashJoinInfo *)(Handle->OperatorHandle));
  return XXX_OK;
}


XxxErrCode RelHashJoinPlanOp::
checkFinal(XxxDerivedRelation::CursorInfo* Handle,
	   XxxCoreADTClass::BoolValue& Satisfied) const 
{
 Satisfied.Value = XXX_TRUE;
 XxxBooleanExpressionPlan* PredPlan = NULL;
 for (int i = 0; i < Handle->ArrayLen; i++) {
   if (JoinPredPlans && JoinPredPlans[i])
     if (XxxBooleanExpressionPlan::MakeConjunction
	 (PredPlan, JoinPredPlans[i]) != XXX_OK) {
       REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
       return XXX_NOT_OK;
      }
   if (ResidPredPlans && ResidPredPlans[i])
     if (XxxBooleanExpressionPlan::MakeConjunction
	 (PredPlan, ResidPredPlans[i]) != XXX_OK) {
       REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
       return XXX_NOT_OK;
      }
  }

 if (PredPlan)
   return PredPlan->Evaluate((const XxxRecord **)Handle->ChildRecs, 
			     (const XxxRecordSchema**)Handle->ChildSchema, 
			     &Satisfied, 
			     Handle->Params?Handle->Params->Params:0);
 return XXX_OK;
}

XxxErrCode 
RelHashJoinPlanOp::GetNextRecord(XxxSetCursor& SCsr,
			   XxxRelation* *RelationArray,
			   const XxxRecordSchema* Schema,
			   XxxRecord*& ReturnRec) const 
{
	/*
	 * Description : a hash table is defined for both inputs
	 * 
	 * Assume relation R and S
	 * Assuming both the hash table will fit into memory
	 * 
	 * while ( R has tuple && S has tuple ) {
     *     tr = Read a tuple from R
     *     ts = Read a tuple from S
	 *     Hash tr into hash table HTR using hash-function h
	 *     Hash ts into hash table HTS using hash-function s
	 *
	 *     Use tr to probe ts for a +ve match
	 *	   If +ve match, output
	 * }
	 * 
	 * while loop ended because one of the operands has no more tuples.
	 * Suppose R has no more tuple
     *
	 * while ( S has tuple ) {
	 *	 ts = Read a tuple from S
	 *   Use ts to probe HTR
	 *   If a +ve match, output
	 * }
	 *
	 */
	
	// Get the HashJoinInfo from the operator handle
	XxxDerivedRelation::CursorInfo* Handle = (XxxDerivedRelation::CursorInfo*) (SCsr.State);

	if ( !Handle ) {
		REL_NOTE_ERROR(REL_BADEXEC, "Null Exec Handle");
		return XXX_NOT_OK;
	}

	HashJoinInfo* JoinInfo = ((HashJoinInfo *) (Handle->OperatorHandle));
	HashTable & LeftHash = JoinInfo->LeftHashT;
	HashTable & RightHash = JoinInfo->RightHashT;
	Store & OutputStore = JoinInfo->OutputStore;
	
	XxxRecord *ResultRec;
	XxxRecord *TempRightRec = Handle->ChildRecs[JoinInfo->RightRelIndex];
	XxxRecord *TempLeftRec  = Handle->ChildRecs[JoinInfo->LeftRelIndex];

		
	// if there is a valid scan on the output
	if ( OutputStore.ScanIsValid() == XXX_TRUE ) {
		ReturnRec = OutputStore.GetNext();
		if ( OutputStore.ScanIsValid() == XXX_TRUE ) 
			return XXX_OK;
	}

	// Empty the OutputStore if all of its records have been returned
	OutputStore.FlushAllRecords(XXX_FALSE);

	// If there are no further records to return and a scan has finished, return
	if ( !(JoinInfo->Valid)) {
		SCsr.AtEnd = XXX_TRUE;
		ReturnRec = 0;
		return XXX_OK;
	}

	/* 
	   Populate the next output store. Keep probing until
	   a positive match occur and all the elements in the bucket
	   has been checked for a match 
	*/
	XxxBool ValidLeftProbe = XXX_FALSE, ValidRightProbe = XXX_FALSE;
	XxxBool LeftScanValid = XXX_TRUE,  RightScanValid = XXX_TRUE;
	XxxRecord *LastLeftRec, *LastRightRec,  *CurLeftRec, *CurRightRec;
	
	CurLeftRec = NULL;
	CurRightRec = NULL;

	// When both inputs are valid, start building in-memory hash
	// for both relations
	while ( (LeftScanValid == XXX_TRUE) || (RightScanValid == XXX_TRUE)) {
		// get an item from left input

		
	  if ( LeftScanValid ) {
		/*if (getNextItem(RelationArray, Handle, JoinInfo->LeftRelIndex, LeftScanValid)
			!= XXX_OK ) {
			REL_NOTE_ERROR(REL_BADEXEC, "Cannot get next record from input");
			return XXX_NOT_OK;
		}
		if ( !LeftScanValid ) break;
		*/

		 // !!! tok : commented out 7 Jun 2000 to test out threads
		//getNextItem(RelationArray, Handle, JoinInfo->LeftRelIndex, LeftScanValid);

//W_COERCE(JoinInfo->populateMutex->acquire());			
getNextItemSb(JoinInfo, JoinInfo->LeftRelIndex, LeftScanValid);
//JoinInfo->populateMutex->release();

		
		if ( LeftScanValid ) {
			// Hash it into left hash table
			// !!! tok: commented out to test out thread
			//CurLeftRec = Handle->ChildRecs[JoinInfo->LeftRelIndex];
			CurLeftRec = JoinInfo->sb[JoinInfo->LeftRelIndex]->curRec;

			ASSERT(CurLeftRec);
			//cout << "Left hash ";
			LeftHash.Insert(CurLeftRec, Handle->ChildSchema[JoinInfo->LeftRelIndex]);
		}
		else
			CurLeftRec = NULL;
	  } 
		

	  if ( RightScanValid ) {
		// get an item from right input
		/*
		if (getNextItem(RelationArray, Handle, JoinInfo->RightRelIndex, RightScanValid)
			!= XXX_OK ) {
			REL_NOTE_ERROR(REL_BADEXEC, "Cannot get next record from input");
			return XXX_NOT_OK;
		}

		if ( !RightScanValid ) break;
		*/
		// !!! tok : commented out 7 Jun 2000 to test out threads
		//getNextItem(RelationArray, Handle, JoinInfo->RightRelIndex, RightScanValid);

//W_COERCE(JoinInfo->populateMutex->acquire());			
getNextItemSb(JoinInfo, JoinInfo->RightRelIndex, RightScanValid);
//JoinInfo->populateMutex->release();


		if ( RightScanValid ) {
			// !!! tok: commented out to test out thread
			//CurRightRec = Handle->ChildRecs[JoinInfo->RightRelIndex];
			CurRightRec = JoinInfo->sb[JoinInfo->RightRelIndex]->curRec;
			ASSERT(CurRightRec);
			// Hash it into right hash table
			//cout << "RIght hash ";
			RightHash.Insert(CurRightRec, Handle->ChildSchema[JoinInfo->RightRelIndex]);
		}
		else
			CurRightRec = NULL;
	  }

		// Flush the outputstore
	  OutputStore.FlushAllRecords(XXX_FALSE);
	  if ( LeftScanValid ) {	
		// Now use the child tuple from left relation to probe right relation
		// cout << "leftscanvalid: probing right hash with curleftrec " << endl;
		if ( CurLeftRec && ( RightHash.probe(CurLeftRec, 
			Handle->ChildSchema[JoinInfo->LeftRelIndex]) == XXX_TRUE) ) {
			ValidLeftProbe = XXX_TRUE;

			// possible matching tuple
			// start performing match
		
			while ( RightHash.ScanIsValid() ) {
				LastRightRec = RightHash.GetNext();

				if ( LastRightRec != NULL
					&& ( isEqual(CurLeftRec, LeftHash.AttrIndex(),
						Handle->ChildSchema[JoinInfo->LeftRelIndex],
						LastRightRec, RightHash.AttrIndex(),
						Handle->ChildSchema[JoinInfo->RightRelIndex]) == XXX_TRUE ) ) {
					
					Handle->ChildRecs[JoinInfo->LeftRelIndex] = CurLeftRec;
					Handle->ChildRecs[JoinInfo->RightRelIndex] = LastRightRec;

					XxxCoreADTClass::BoolValue Satisfied(XXX_TRUE);
					
					// Join condition is equal
					// Check any residual predicates
					if (ResidPredPlans[1] &&
						ResidPredPlans[1]->Evaluate( (const XxxRecord **) Handle->ChildRecs,
						(const XxxRecordSchema **) Handle->ChildSchema,
						&Satisfied,	(Handle->Params) ? Handle->Params->Params : 0) != XXX_OK ) {
						XXX_NOTE_ERROR(XXX_BADEXPR, "Error evaluating residual expression");
						return XXX_NOT_OK;
					}

					if  (Satisfied.Value) {
						if (createRecord(Handle,RelationArray, Schema, ResultRec) != XXX_OK ) {
							XXX_NOTE_ERROR(XXX_BADEXPR, "Error Creating Record ");
							return XXX_NOT_OK;
						}
						OutputStore.AddRecord(ResultRec);
					}
				} // 
			} // while
		} // if valid right probe
	  }
	  
	  if ( RightScanValid ) {
		// cout << "rightscanvalid: probing left hash with currightrec " << endl;
		if ( CurRightRec && ( LeftHash.probe(CurRightRec, 
			Handle->ChildSchema[JoinInfo->RightRelIndex]) == XXX_TRUE) ) {
			
			// If left probe wasn't valid, try using the child tuple from right 
			// relation to probe left hash
			ValidRightProbe = XXX_TRUE;

			// possible matching tuple
			// start performing match
			
			while ( LeftHash.ScanIsValid() ) {
				LastLeftRec = LeftHash.GetNext();

				if ( (LastLeftRec != NULL) && (LastLeftRec != CurLeftRec ) 
					&& ( isEqual(CurRightRec, RightHash.AttrIndex(),
						Handle->ChildSchema[JoinInfo->RightRelIndex],
						LastLeftRec, LeftHash.AttrIndex(),
						Handle->ChildSchema[JoinInfo->LeftRelIndex]) == XXX_TRUE ) ) {
					
					Handle->ChildRecs[JoinInfo->LeftRelIndex] = LastLeftRec;
					Handle->ChildRecs[JoinInfo->RightRelIndex] = CurRightRec;

					XxxCoreADTClass::BoolValue Satisfied(XXX_TRUE);
					
					// Join condition is equal
					// Check any residual predicates
					if (ResidPredPlans[1] &&
						ResidPredPlans[1]->Evaluate( (const XxxRecord **) Handle->ChildRecs,
						(const XxxRecordSchema **) Handle->ChildSchema,
						&Satisfied,	(Handle->Params) ? Handle->Params->Params : 0) != XXX_OK ) {
						XXX_NOTE_ERROR(XXX_BADEXPR, "Error evaluating residual expression");
						return XXX_NOT_OK;
					}

					if  (Satisfied.Value) {
						if (createRecord(Handle,RelationArray, Schema, ResultRec) != XXX_OK ) {
							XXX_NOTE_ERROR(XXX_BADEXPR, "Error Creating Record ");
							return XXX_NOT_OK;
						}
						OutputStore.AddRecord(ResultRec);
					}
				} // 
			} // while
		} // if valid left probe
	  }
		// Restore the handle to its previous state
		// Handle->ChildRecs[JoinInfo->RightRelIndex] = TempRightRec;
		// Handle->ChildRecs[JoinInfo->LeftRelIndex] = TempLeftRec;
		
		if ( ValidLeftProbe == XXX_TRUE || ValidRightProbe == XXX_TRUE ) {
			// Either a left probe or right probe is valid
			// If OutputStore is populated, break out of loop
			if ( !OutputStore.IsEmpty() ) break;
		}
	} // both scan is valid

	// Restore the handle to its previous state
	Handle->ChildRecs[JoinInfo->RightRelIndex] = TempRightRec;
	Handle->ChildRecs[JoinInfo->LeftRelIndex] = TempLeftRec;
	
	// If both scan are complete, note it in the JoinInfo
	if (!LeftScanValid && !RightScanValid)
		JoinInfo->Valid = XXX_FALSE;

	// If the join was successful, the OutputGroup has been repopulated.
	// Return its first tuple.
	if (!OutputStore.IsEmpty()) {
		OutputStore.InitScan();
		ReturnRec = OutputStore.GetNext();
		ASSERT (OutputStore.ScanIsValid());
		return XXX_OK;
	}
	
	// Otherwise, one of the inputs' scans is complete and the join need not
	// proceed.  Return end-of-scan.
	SCsr.AtEnd = XXX_TRUE;
	ReturnRec = 0;
	return XXX_OK;
}
