#ifdef __GNUG__
#pragma implementation
#endif

#include "relplanop.h"

//
// Methods of RelStopAfterOp
//

RelStopAfterPlanOp::RelStopAfterPlanOp(RelPlanOp* Child,int k,
				       RelCatalogInfo* CatInfo)
  : RelDerivedPlanOp(1, &Child, CatInfo), K(k)
{
}

RelStopAfterPlanOp::~RelStopAfterPlanOp() 
{ }

void RelStopAfterPlanOp::NodePrint (ostream &OStr, int curBlockId, int Level, 
				   int numChildren, int *childrenIds) const {
  OStr << "NODE Stop After" << endl;
  OStr << "ID " << curBlockId << endl;
  OStr << "LEVEL " << Level << endl;
  if (numChildren > 0) {
    OStr << "CHILDREN ";
    for (int i = 0; i < numChildren; i++) {
      OStr << childrenIds[i] << " ";
    }
    OStr << endl;
  }
  OStr << "COLOR " << 0x66C66C << endl;
  OStr << "ENDNODE" << endl;
}

XxxErrCode 
RelStopAfterPlanOp::InitHandle(XxxDerivedRelation::CursorInfo* Handle,
			       XxxRelation* *RelationArray) const
{
  if (Handle->ArrayLen != 1) {
    REL_NOTE_ERROR(REL_BADEXEC, "StopAfterPlanOp needs exactly one input");
    return XXX_NOT_OK;
  }
  // Open a scan on the input
  if (RelationArray[0]->InitCursor1(Handle->ChildCursorArray[0],
				    Handle->Filter, NULL, NULL,
				    Handle->Params) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADCURSOR, 
		   "Unable to open sorted input relation cursor");
    return XXX_NOT_OK; 
  }
  Handle->OperatorHandle = NULL;
  return XXX_OK;
}

XxxErrCode 
RelStopAfterPlanOp::CloseHandle(XxxDerivedRelation::CursorInfo* Handle,
			       XxxRelation* *RelationArray) const
{
   // Close the scan on the input relation
  if (Handle->ChildCursorArray[0].State &&
      RelationArray[0]->CloseCursor(Handle->ChildCursorArray[0]) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADCURSOR, "Unable to close cursor");
    return XXX_NOT_OK;
  }
  return XXX_OK;
}

XxxErrCode RelStopAfterPlanOp::GetNextRecord(XxxSetCursor& SCsr,
					    XxxRelation* *RelationArray,
					    const XxxRecordSchema* Schema,
					    XxxRecord*& ReturnRec) const
{
  static int counter=0;
  
  XxxDerivedRelation::CursorInfo* Handle =    (XxxDerivedRelation::CursorInfo*)(SCsr.State);
  if (!Handle) {
    REL_NOTE_ERROR(REL_BADEXEC, "NULL Exec handle");
    return XXX_NOT_OK;
  }
 
  if (RelationArray[0]->NextItem(Handle->ChildCursorArray[0],
				 Handle->ChildRecs[0]) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADEXEC, "Error getting NextItem from relation");
    return XXX_NOT_OK;
  }
  
  
  if (Handle->ChildCursorArray[0].IsValid()) {
    ReturnRec = Handle->ChildRecs[0];
    if(counter>K-1){
      counter =0;
      ReturnRec = 0;
      SCsr.AtEnd = XXX_TRUE;
    }
    else counter++;
    return XXX_OK;
  } else {
    // The scan is complete
    SCsr.AtEnd = XXX_TRUE;
    ReturnRec = 0;
    counter = 0;;
    return XXX_OK;
  }

}

void RelStopAfterPlanOp::PrintThis(ostream& OStr) const
{
  OStr << PrintString();
}

