#ifndef XXX_COSTSPJOPT_H
#define XXX_COSTSPJOPT_H
    
//
// Defines a superclass of all cost-based optimizers.
// Important sub-classes are :
//      a) Greedy optimizer
//      b) Dyn prog optimizer
//
// Both use the plan structures defined in this class
//
    
#include "relenv.h"
#include "seminaivespj.h"
 
     
/**
 *  RelSPJCostBasedOpt is a sub-class of SemiNaiveOpt because
 * it differs only by specifying additional classes.
 * Importantly, it uses the same format to specify the final plan.
 */
class RelSPJCostBasedOpt : public RelSPJSemiNaiveOpt {
public:
  RelSPJCostBasedOpt(XxxErrCode& RC,
    		     XxxValueExprList* TargetList,
    		     XxxBooleanExpression* Condition,
    		     XxxBool Unique,
    		     RelSPJBlockPlan* SPJPlan,
    		     RelOptEngine::OptQueryEnv* QEnv) :
    RelSPJSemiNaiveOpt(RC, TargetList, Condition, Unique, SPJPlan, QEnv) { 
      XxxBool IgnoreExpPreds = XXX_FALSE;
      if (XXX_THISRELENV()->ExpPredStrategy == XxxRelEnv::PULLUP_EP)
	IgnoreExpPreds = XXX_TRUE;
      double MaxCheapPredCost = XXX_THISRELENV()->CheapPredMax;
      analyzeJoinGraph(IgnoreExpPreds, MaxCheapPredCost);
  }
 
  virtual ~RelSPJCostBasedOpt() {}
  virtual XxxErrCode GeneratePlanOp(RelOptEngine::PlanObject* Plan,
    				    RelPlanOp*& PlanOp) const;
    
protected:
  // define physical plans
  //@Include: physplan.h
#include "physplan.h"
    
  // define logical plans
  //@Include: logplan.h
#include "logplan.h"
    
public:
  //// Statistics for plan table
  class PlanTableStats { 
  public:
    int PlansGenerated;
    int PlansAdded;
    int PlansPruned;
    PlanTableStats() { 
      PlansGenerated = PlansAdded = PlansPruned = 0; }
  };
  //// Virtual class representing a collection of physical plans.
  //// Each kind of optimizer can define its own version of a PlanTable
  //// with its own kind of pruning.
  class PlanTable { public:
    int Joinplannow;	// added by zhiyuan
    int AddedPlanPruned;
    int Minplannow;
 
    PlanTableStats Stats;
    PlanTable() {}
    virtual ~PlanTable() {}
    virtual XxxErrCode AddPlan(const XxxDBEnv& Env,
    			       PhysicalPlan* NewPlan,
    			       int& NumPlansAdded,
    			       int& NumPlansPruned) = 0;
    virtual void Reset() {}
  };
  // Simple plan collection structure used to simply keep
  // the cheapest plan
  class MinPlanTable : public PlanTable {
    PhysicalPlan* minPPlan;
  public:
   

    MinPlanTable() { Joinplannow = 0; AddedPlanPruned = 0; Minplannow = 0; minPPlan = NULL;}
    virtual ~MinPlanTable() { if (minPPlan) delete minPPlan; }
    virtual XxxErrCode AddPlan(const XxxDBEnv& Env,
    			       PhysicalPlan* NewPPlan,
    			       int& NumPlansAdded,
    			       int& NumPlansPruned) {
      XxxRelEnv *GEnv;
      // if debug, dump all plans created here and the winning one

      GEnv = (XxxRelEnv*)(XXX_THISENGENV(RelationEngineId));
      AddedPlanPruned = 0;

      if (!minPPlan) {
    	minPPlan = NewPPlan;
    	NumPlansAdded = 1;
    	NumPlansPruned = 0;
	if(Joinplannow) {
	  AddedPlanPruned = 0;
	  minPPlan->PrintTree(*(GEnv->QPInfoFile));
	}
      }
      else {
    	PhysicalPlan* Victim;
    	if (minPPlan->Usage.FullCost() > NewPPlan->Usage.FullCost()) {
    	  Victim = minPPlan;
    	  minPPlan = NewPPlan;
    	  NumPlansAdded = 1;
    	  NumPlansPruned = 1;
	  if(Joinplannow) {
	    AddedPlanPruned = 0;
	    NewPPlan->PrintTree(*(GEnv->QPInfoFile));
	  }
    	}
    	else {
 	  if(Joinplannow) {
	    AddedPlanPruned = 1;
	    NewPPlan->PrintTree(*(GEnv->QPInfoFile));
          }
    	  Victim = NewPPlan;
    	  NumPlansAdded = 0;
    	  NumPlansPruned = 1;
    	}
    	if (Victim) delete Victim; // the if() is a sanity check
      }

      return XXX_OK;
    }
    
    XxxErrCode GetMinPlan(PhysicalPlan*& PhysPlan) const { 
    
      if (!minPPlan) {
    	REL_NOTE_ERROR(REL_BADOPT, "No min plan available");
    	return XXX_NOT_OK;
      }
      PhysPlan = minPPlan; 
      if(Joinplannow) {
	XxxRelEnv *GEnv = (XxxRelEnv*)(XXX_THISENGENV(RelationEngineId));
	if (GEnv->QPInfoFile) {
  
	  minPPlan->PrintTree(*(GEnv->QPInfoFile));
  
	}
      }
      return XXX_OK;
    }
    virtual void Reset() { minPPlan = NULL; }
  };
    
protected:
  //// Can the join graph be connected by a linear chain? 
  //// If not, cross products must be considered for left-deep trees
  XxxBool linearJoinGraph;
    
  // this should assign the value of linearJoinGraph
  void analyzeJoinGraph(XxxBool AvoidExpensivePreds, 
  			double MaxCheapPredCost);

    
  // Put all physical plans for one logical table access into table
  XxxErrCode generateRelPlans(const XxxDBEnv& Env,
    			      const RelOptEngine::LogProperties& RelLProps,
    			      unsigned char RelId,
    			      PlanTable& PlanTable);
  // Get best physical plan for one logical table
  XxxErrCode generateBestRelPlan(const XxxDBEnv& Env,
    				 const RelOptEngine::LogProperties& RelLProps,
    				 unsigned char RelId,
    				 PhysicalPlan*& RelPPlan);
      
  // Enumerate all physical plans for a join with some PPlan in OuterLPlan
  // and the inner table specified by InnerLPlan
  XxxErrCode generateJoinPlans(const XxxDBEnv& Env,
    			       const LogicalPlan* OuterLPlan,
    			       const LogicalPlan* InnerLPlan,
    			       const RelCatalogInfo* InnerCatInfo,
    			       PlanTable& InnerPlanTab,
    			       const RelOptEngine::LogProperties& JoinLProps,
    			       const XxxBitMap& NewPreds,
    			       PlanTable& PlanTable);
  // Enumerate all physical plan for a join with OuterPPlan and the inner
  // table specified by InnerLProps
  XxxErrCode generateJoinPlans(const XxxDBEnv& Env,
    			       const PhysicalPlan* OuterPPlan,
    			       const RelOptEngine::LogProperties& InnerLProps,
    			       const RelCatalogInfo* InnerCatInfo,
    			       PlanTable& InnerPlanTab,
    			       const RelOptEngine::LogProperties& JoinLProps,
    			       const XxxBitMap& NewPreds,
    			       PlanTable& PlanTab);
  // Get best physical plan for a join with OuterPPlan and the inner
  // table specified by InnerLProps
  XxxErrCode generateBestJoinPlan(const XxxDBEnv& Env,
    				  const PhysicalPlan* OuterPPlan,
    				  const RelOptEngine::LogProperties& InnerLProps,
    				  const RelCatalogInfo* InnerCatInfo,
    				  const RelOptEngine::LogProperties& JoinLProps,
    				  const XxxBitMap& JoinPreds,
    				  PhysicalPlan*& JoinPPlan);
    
  // make a specific join plan (method specified)
  XxxErrCode createJoinPlan(const XxxDBEnv& Env,
    			    const RelOptEngine::LogProperties& JoinLProps,
    			    const PhysicalPlan* OuterPPlan,
    			    const RelOptEngine::LogProperties& InnerLProps,
    			    const RelStatsInfo* InnerStatsInfo,
    			    PlanTable& InnerPlanTab,
    			    const XxxBitMap& NewPreds,
    			    RelOptEngine::JoinMethod JMeth,
    			    XxxBool& Success,
    			    JoinPhysPlan*& JPlan);
    
  /*******
    // all physical plans for joins involving two physical plans 
    XxxErrCode generateJoinPlans(const XxxDBEnv& Env,
    const PhysicalPlan* PPlan0, // outer
    const PhysicalPlan* PPlan1, // inner
    const RelCatalogInfo* InnerCatInfo,
    const RelOptEngine::LogProperties& JoinLProps,
    const XxxBitMap& NewPreds,
    PlanTable& PlanTable);
    // one specific physical plan for a join
    XxxErrCode createJoinPlan(const RelOptEngine::LogProperties& JoinLProps,
    const PhysicalPlan* OuterPlan,
    const PhysicalPlan* InnerPlan,
    const RelStatsInfo* InnerStatsInfo,
    const XxxBitMap& NewPreds,
    RelOptEngine::JoinMethod JMeth,
    XxxBool& Success,
    JoinPhysPlan*& JPlan);
    ********/
};    
    
#endif
    
    
