#include <iostream.h>

#include "relerror.h"
#include "relquery.h"
#include "dynspj.h"
#include "relplanop.h"
#include "relenv.h"

//
// Methods of RelSPJDynOptimizer
//
/**
 * Optimize the query. the description of the query is in
 * the fields of the RelSPJOptimizer superclass.
 * the system parameters are in the Env field
 */
XxxErrCode RelSPJDynOptimizer::Optimize()
{
  if (XXX_THISRELENV()->DebugOpt) {
    PrintOptInfo(cerr);
  }
 // TODO:L2 Identify which of these predicates are "expensive"
 // and so should be applied like joins

 // These counts will help define XxxBitMaps for Rels and Preds.
 int RelCount = BlockPlan->ChildArrayLen;
 int PredCount = NumPreds; 

 // define a "plan space" to hold the results of planning.
 // this will be reclaimed at the end of this function,
 // so copy out any final plans before that
 PlanSpaceClass PlanSpace(RelCount, this);

 if (PlanSpace.GeneratePlans(Env->OptSpecs->DBEnv) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Unable to generate plans");
   return XXX_NOT_OK;
  }

 // get one or more plans out of the plan space
 int NumPlans = 0;
 XxxSetCursor Csr;
 if (PlanSpace.OpenPPlanScan(Csr) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Unable to open scan of plans");
   return XXX_NOT_OK;
  }
 while (1) {
   PhysicalPlan* Plan;
   if (PlanSpace.GetNextPPlan(Csr, Plan) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error getting next plan");
     return XXX_NOT_OK;
    }

   if (!Csr.IsValid()) break;
   // otherwise, add this plan to the block
   BlockPlan->AddPlan(Plan);

   if (XXX_THISRELENV()->DebugOpt) {
     cerr << "DynOpt creating:: ";
     if (Plan) Plan->Print(cerr);
     cerr << endl;
   }
   NumPlans++;
  }
 if (PlanSpace.ClosePPlanScan(Csr) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Unable to close scan of plans");
   return XXX_NOT_OK;
  }

  if (XXX_THISRELENV()->DebugOpt) {
    cerr << "Done with DYNOPT: NumPlans = " << NumPlans << endl;
  }
 return XXX_OK;
}



