#ifndef XXX_DYNSPJ_H
#define XXX_DYNSPJ_H

#include "costspjopt.h"

/**
 * The Dynamic programming optimizer has the following structure:
   It has a PlanSpaceClass object. This is an array of
   PlanTableClass objects. Each PlanTableClass holds
   LogicalPlans for joins of k relations (k is fixed for
   each PlanTableClass, and different across PlanTableClasses).
   Each LogicalPlan corresponds to a logical partial join
   result, and leads to a number of PhysicalPlans.

   DynOptimizer is a sub-class of RelSPJCostBasedOpt because
   it differs only in the Optimize() and GeneratePlanOp() methods. 
   Importantly, it uses the same format to specify the final plan.
 */
class RelSPJDynOptimizer : public RelSPJCostBasedOpt {
 public:
  RelSPJDynOptimizer(XxxErrCode& RC,
		     XxxValueExprList* TargetList,
		     XxxBooleanExpression* Condition,
		     XxxBool Unique,
		     RelSPJBlockPlan* SPJPlan,
		     RelOptEngine::OptQueryEnv* QEnv) :
  RelSPJCostBasedOpt(RC, TargetList, Condition, Unique, SPJPlan, QEnv)
   {}
  virtual ~RelSPJDynOptimizer() {}
  virtual XxxErrCode Optimize();

 protected:
  // define the plan space and plan table classes
  //@Include: planspace.h
  #include "planspace.h"
  friend class PlanSpaceClass;

 };


#endif
