#include <iostream.h>
#include <math.h>
    
#include "relerror.h"
#include "relquery.h"
#include "relplanop.h"
#include "relenv.h"
#include "mykbz.h"
    
//
// Methods of RelSPJMyKBZOpt
//
/**
 * Optimize the query. the description of the query is in
 * the fields of the RelSPJOptimizer superclass.
 * the system parameters are in the Env field.
 * This method should create a PhysicalPlan object and
 * call BlockPlan->Add() with it as a parameter.
 */
XxxErrCode RelSPJMyKBZOpt::Optimize()
{
  XxxRelEnv *GEnv;
  
  if (XXX_THISRELENV()->DebugOpt) {
    cerr << "Opimizing using MyKBZ optimizer" << endl;
    PrintOptInfo(cerr);
  }
    
  XxxErrCode rc = XXX_OK;
  PhysicalPlan* PhysPlan = NULL;
  // These counts will help define XxxBitMaps for Rels and Preds.
  int RelCount = BlockPlan->ChildArrayLen;
  int PredCount = NumPreds; 
  if (RelCount <= 0) {
    REL_NOTE_ERROR(REL_BADOPT, "Asked to optimize SPJ node with no tables");
    return XXX_NOT_OK;
  }
    
  RelOptEngine::Cost MinCost;
  MinCost.SetMax();
  // dump physical plans to qpinfoffile
  if(XXX_THISRELENV() ) {
    GEnv = (XxxRelEnv*)(XXX_THISENGENV(RelationEngineId));
  }
  // find best plan for each possible outmost relation
  for (int i = 0; i < RelCount; i++) {
    PhysicalPlan* TempPlan;
    
    if (getLinearPlan(i, RelCount, TempPlan) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADOPT, "Error determining KBZ join order");
      return XXX_NOT_OK;
    }
    if (XXX_THISRELENV() && XXX_THISRELENV()->DebugOpt) {
      TempPlan->Print(cerr);
    
    }
    // If the new cost is less than the current cost, choose the new plan.
    if (TempPlan->Usage.FullCost() < MinCost.FullCost()) {
      PhysPlan = TempPlan;
      MinCost = TempPlan->Usage;
      if (XXX_THISRELENV() && XXX_THISRELENV()->DebugOpt) {
	cerr << "Plan is chosen as lowest cost" << endl;
      }
    }
    else {
      if (XXX_THISRELENV() && XXX_THISRELENV()->DebugOpt) {
      }
      // throw away this plan
      delete TempPlan;
    }
  } // for each starting relation
    
  // Now add the plan to the block
  if ((MinCost.FullCost() == DBL_MAX) || (!PhysPlan)) {
    REL_NOTE_ERROR(REL_BADOPT, "Error in determining KBZ plan");
    return XXX_NOT_OK;
  }
  if (PhysPlan) {
    BlockPlan->AddPlan(PhysPlan);
    if (XXX_THISRELENV()->DebugOpt) {
      cerr << "MyKBZ: Generated Plan: ";
      PhysPlan->Print(cerr);
      // also print out the physplan to qpinfo
      cerr << endl;
    }
    if (GEnv->QPInfoFile) {
      *(GEnv->QPInfoFile) << "SEQ Optimizing Block "<< (int)BlockPlan->BlockId  << " Final" << endl;
      PhysPlan->PrintTree(*(GEnv->QPInfoFile));
      *(GEnv->QPInfoFile) << "ENDSEQ" << endl;
    }
  }
  return XXX_OK;
}
    
XxxErrCode RelSPJMyKBZOpt::getLinearPlan(int StartRel, int NumRels,
    					 PhysicalPlan*& PhysPlan)
{ 
  XxxRelEnv *GEnv;
  if (XXX_THISRELENV() && XXX_THISRELENV()->DebugOpt) {
    cerr << "Starting with relation:" << StartRel << " .. ";
  }
  //dump query plan
  if(XXX_THISRELENV()) {
    GEnv = (XxxRelEnv*)(XXX_THISENGENV(RelationEngineId));
    
  }
  PhysicalPlan* OuterPPlan = NULL;
    
  XxxBool IgnoreExpPreds = XXX_FALSE;
  if (XXX_THISRELENV()->ExpPredStrategy == XxxRelEnv::PULLUP_EP)
    IgnoreExpPreds = XXX_TRUE;
  double MaxCheapPredCost = XXX_THISRELENV()->CheapPredMax;
    
  RelOptEngine::Cost OuterCost;
  OuterCost.SetMin();
    
  RelOptEngine::LogProperties OuterLogProps(NumRels, NumPreds);
  OuterLogProps.Card = 1;
  OuterLogProps.TupleSize = 0;
  OuterLogProps.Preds.ClearAll();
  OuterLogProps.Rels.ClearAll();
    
  if (SetRelLProps(StartRel, 
		   ((NumRels > 1)?IgnoreExpPreds:XXX_FALSE), // input
		   MaxCheapPredCost, // input
		   OuterLogProps) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, "Error creating logprops");
    return XXX_NOT_OK;
  }
  // dump the best plan for outer most relation
  if(XXX_THISRELENV()) {
    GEnv = (XxxRelEnv*)(XXX_THISENGENV(RelationEngineId));
    *(GEnv->QPInfoFile) << "SEQ Optimizing Block "<< (int)BlockPlan->BlockId << " w. Relation " << StartRel << " as Outermost" << endl;
  }
  // create the single-rel plan for StartRel
  if (generateBestRelPlan(OptSpecs()->DBEnv, OuterLogProps,
			  StartRel, OuterPPlan) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, "Error generating phys rel plan");
    return XXX_NOT_OK;
  }
    
  // choose the best plan for remaining relations
  int i;
  for (i=1; i < NumRels; i++) {
    double MinCost = DBL_MAX;
    
    PhysicalPlan* ChosenPPlan = NULL;
    RelOptEngine::LogProperties CurrLogProps(NumRels, NumPreds);
    RelOptEngine::LogProperties RelLogProps(NumRels, NumPreds);
    
    // try each of the possible relations and choose the best one
    for (int j=0; j<NumRels; j++) {
      if (OuterPPlan->LogProps.Rels.GetBit(j)) continue;
    
      // the plan for the current join being considered
      PhysicalPlan* CurrentPPlan;
      // determine plan if relation j was used in the next spot
      if (SetRelLProps(j, IgnoreExpPreds, 
		       MaxCheapPredCost, RelLogProps) != XXX_OK) {
	REL_NOTE_ERROR(REL_BADOPT, "Error creating logprops");
	return XXX_NOT_OK;
      }
      const RelCatalogInfo* InnerCatInfo = ChildCat(j);
    
      // first check that the two plans are joinable
      XxxBool Joinable = XXX_FALSE;
      XxxBitMap NewPreds(NumPreds);
      RelOptEngine::LogProperties JoinLogProps(NumRels, NumPreds);
      if (JoinsWith(OuterPPlan->LogProps, // input
		    RelLogProps, // input
		    ((i<NumRels-1)?IgnoreExpPreds:XXX_FALSE), // input
		    MaxCheapPredCost, // input
		    JoinLogProps, // output
		    NewPreds, // output
		    Joinable) != XXX_OK) {
	REL_NOTE_ERROR(REL_BADOPT, "Error checking joinable status");
	return XXX_NOT_OK;
      }
      // what if a cross-product is needed ?? uh-oh: check for it
      if (!Joinable && !XXX_THISRELENV()->TryCrossProduct && linearJoinGraph) continue;
    
    
      // create the join plans 
      if (generateBestJoinPlan(OptSpecs()->DBEnv, OuterPPlan, 
			       RelLogProps, InnerCatInfo,
			       JoinLogProps, NewPreds, CurrentPPlan)
	  != XXX_OK) {
	REL_NOTE_ERROR(REL_BADOPT, "Error generating join plan");
	return XXX_NOT_OK;
      }
    
      if (CurrentPPlan->Cost() < MinCost) {
	MinCost = CurrentPPlan->Cost();
	if (ChosenPPlan) delete ChosenPPlan;
	ChosenPPlan = CurrentPPlan;
    
      }
      else
	delete CurrentPPlan;
 
    } // loop over all the children for a particular spot
    
    if (MinCost == DBL_MAX) {
      REL_NOTE_ERROR(REL_BADOPT, "Error in finding next join child");
      return XXX_NOT_OK;
    }
    OuterPPlan = ChosenPPlan;
  } // loop over all the spots
    
  // make sure that all bits are set in RelsUsed and PredsUsed
  if (OuterPPlan->LogProps.Preds.NumBitsSet() != NumPreds) {
    REL_NOTE_ERROR(REL_BADOPT, "All the predicates were not used.");
    return XXX_NOT_OK;
  }      
  if (OuterPPlan->LogProps.Rels.NumBitsSet() != NumRels) {
    REL_NOTE_ERROR(REL_BADOPT, "All the relations were not used.");
    return XXX_NOT_OK;
  }      
  if (XXX_THISRELENV() && XXX_THISRELENV()->DebugOpt) {
    cerr << "Plan chosen by MyKBZ optimizer" << endl;
    OuterPPlan->Print(cerr);
    // dump winning plan   	
     
  }
  PhysPlan = OuterPPlan;
  if (GEnv->QPInfoFile ) {
    
    *(GEnv->QPInfoFile) << "ENDSEQ" << endl;
    
    *(GEnv->QPInfoFile) << "SEQ Optimizing Block "<< (int)BlockPlan->BlockId  << " w. Relation " << StartRel << " as Outermost (Best)" << endl;
    PhysPlan->PrintTree(*(GEnv->QPInfoFile));
    *(GEnv->QPInfoFile) << "ENDSEQ" << endl;
  }
  return XXX_OK;
}
