#ifndef XXX_KBZSPJ_H
#define XXX_KBZSPJ_H

/**
 *  This is my simplified version of the KBZ algorithm -- actually, it
 *  isn't really based on the "rank" idea of KBZ. The main idea it
 *  uses is to run a greedy algorithm trying each relation as the
 *  outermost relation of a left-deep tree.
 */
#include "costspjopt.h"

/**
 * Varies from the naive optimizer in that it uses a greedy algorithm
 * based on intermediate tables cardinality to determine the join order.
 * Also creates the plan in the same format as generated by the
 * dyn. prog optimizer (i.e. in the format expected by RelSPJCostBasedOpt)
 */
class RelSPJMyKBZOpt : public RelSPJCostBasedOpt {
 protected:
  //// Estimate the properties and cost of joining with CurrentRel
  XxxErrCode includeNewRel(int CurrentRel, // index of rel to join in
			   int CurrentRelPos, // position in join order
			   const RelOptEngine::Cost& InputCost,
			   const RelOptEngine::LogProperties& InputLProps,
			   const RelOptEngine::PhysProperties& InputPProps,
			   RelOptEngine::Cost& OutputCost,
			   RelOptEngine::LogProperties& OutputLProps,
			   RelOptEngine::PhysProperties& OutputPProps,
			   RelOptEngine::JoinMethod& JM) const;

  XxxErrCode getLinearPlan(int StartRel, int NumRels,
			   PhysicalPlan*& PhysPlan);

 public:
  RelSPJMyKBZOpt(XxxErrCode& RC,
		 XxxValueExprList* TargetList,
		 XxxBooleanExpression* Condition,
		 XxxBool Unique,
		 RelSPJBlockPlan* SPJPlan,
		 RelOptEngine::OptQueryEnv* QEnv) :
  RelSPJCostBasedOpt(RC, TargetList, Condition, Unique, SPJPlan, QEnv) 
  {  alwaysTrySM = XXX_FALSE; }
  virtual ~RelSPJMyKBZOpt() {}

  virtual XxxErrCode Optimize();
 };


#endif
