//Ihab
//the basic naive optimizer (multiway physical operator)

#include <iostream.h>

#include "relerror.h"
#include "relquery.h"
#include "naiveKRJ.h"
#include "relplanop.h"
#include "relenv.h"
#include "funcexpr.h"
#include "adtfunc.h"

//
// Methods of RelKRJNaiveOptimizer
//
RelKRJNaiveOptimizer::
RelKRJNaiveOptimizer(XxxErrCode& RC,
		    XxxValueExprList* TargetList,
		    XxxBooleanExpression* Condition,
		    XxxValueExprList* Ordering,
		    XxxBool Unique,
		    RelKRJBlockPlan* KRJPlan,
		    RelOptEngine::OptQueryEnv* QEnv)
: RelKRJOptimizer(RC, TargetList, Condition, Ordering, Unique, KRJPlan, QEnv)
{
 if (RC != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error creating RelKRJOptimizer");
   return;
  }
 // the parent class constructor has already split the WHERE
 // clause predicates into conjunctive boolean factors, and
 // determined the dependencies of these factors.

 // What remains is to construct a query plan from this.
 
 RC = XXX_OK;

}

RelKRJNaiveOptimizer::~RelKRJNaiveOptimizer()
{
}

XxxErrCode RelKRJNaiveOptimizer::Optimize()
{
 XxxErrCode rc = XXX_OK;
 KRJNaivePlan* Plan = new KRJNaivePlan(rc, BlockPlan->HighestPlanId++,
				     BlockPlan->BlockId,
				     BlockPlan->ChildArrayLen,
				     this);
 if (rc != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error creating naive plan");
   return XXX_NOT_OK;
  }
 
 if (Plan->ArraySize) {
   int i;
   //
   // Since this is an ultra-naive optimizer, the join order
   // is determined independent of any of the predicates.
   // A slightly less naive optimizer would use the predicate
   // selectivities as well. In that case, the PredDependencyInfo
   // would be useful.
   // 
   Plan->JoinOrder = new int[Plan->ArraySize];
   // given N relations, there are (N-1) joins
   if (Plan->ArraySize > 1) 
     Plan->JoinMethods =  new RelOptEngine::JoinMethod[Plan->ArraySize-1];
   // At this stage, any join order manipulation should be done.
   if (determineJoinStrategy(Plan->ArraySize, 
			     Plan->JoinOrder,
			     Plan->JoinMethods) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error generating join strategy");
     return XXX_NOT_OK;
    }

   // use pred dependencies to determine placement
   // allocate the SelPreds, JoinPreds, and ResidPreds arrays
   Plan->SelPreds = new XxxBitMap[Plan->ArraySize];
   Plan->JoinPreds = new XxxBitMap[Plan->ArraySize];
   Plan->ResidPreds = new XxxBitMap[Plan->ArraySize];
   for (i = 0; i < Plan->ArraySize; i++) {
     Plan->SelPreds[i].ChangeSize(NumPreds);
     Plan->JoinPreds[i].ChangeSize(NumPreds);
     Plan->ResidPreds[i].ChangeSize(NumPreds);
    }

   // I have Changed JoinOrder -> InverseJoinOrder and am computing 
   // InverseJoinOrder here.
   // InverseJoinorder is as the name implies, the inverse
   // of the JoinOrder mapping. JoinOrder[i] is the index of
   // the child that is at the i'th position in the join.
   // InverseJoinOrder[i] is the position of the i'th child.
   int *InverseJoinOrder = new int[Plan->ArraySize];
   for (i=0; i < Plan->ArraySize; i++) {
     InverseJoinOrder[Plan->JoinOrder[i]] = i;
   }

   // for each of the boolean factors, find a place for it.
   // factor in the join order here!!
   for (int j = 0; j < NumPreds; j++) {
     int DepMax = -1;
     int DepCount = 0;
     // note: dont have to check Arity()+1, because if it is only 
     // dependent on the Env, then it should be applied as an InitPred
     for (i =0; i < Plan->ArraySize; i++) {
       if (PredDependencyInfo[j][i] == XXX_TRUE) {
	 DepCount++;
	 if (InverseJoinOrder[i] > DepMax)
	   DepMax = InverseJoinOrder[i];
	}
      } // for each child
     if (DepMax < 0) {
       // trivial selection of the form 3 = 2+1 or 3 = $1
       Plan->InitPreds.SetBit(j);
      }
     else if (DepCount == 1) {
       Plan->SelPreds[DepMax].SetBit(j);
      }
     else {
       Plan->JoinPreds[DepMax].SetBit(j);
      }
    } // for each boolean factor

    delete[] InverseJoinOrder;
  } // if (Plan->ArraySize)

 BlockPlan->AddPlan(Plan);
 if (XXX_THISRELENV()->DebugOpt) {
   cerr << "SN: Generating Plan: ";
   Plan->Print(cerr);
   cerr << endl;
 }
 return XXX_OK;
}

// For the naive optimizer, we do not worry about join order,
// and join method. The join order is right to left, and
// the join method is NL join. For now, lets not even worry
// about indexing...
XxxErrCode RelKRJNaiveOptimizer::
determineJoinStrategy(int NumChildren, 
		      int* JoinOrder,
		      RelOptEngine::JoinMethod* JoinMethods) const
{
 XxxRelEnv* REnv =((XxxRelEnv*)(XXX_THISENGENV(RelationEngineId)));
 // For now: use the reverse join order as the user spec
 for (int i=0; i < NumChildren; i++) {
   //JoinOrder[i] = i;  // simple join order
   JoinOrder[i] =  NumChildren-i-1; // reverse join order
   if (i > 0) {
     JoinMethods[i-1] = RelOptEngine::RJOIN; 
   } // if i > 0
 } // for each input
 return XXX_OK;
}

RelKRJNaiveOptimizer::
KRJNaivePlan::KRJNaivePlan(XxxErrCode& RC,
			   int PlanIdent,
			   unsigned char BlockIdent,
			   int NumChildren,
			   RelKRJNaiveOptimizer* Opt)
: RelOptEngine::PlanObject(PlanIdent, BlockIdent),
  InitPreds(Opt->NumPreds)
{
 ArraySize = NumChildren;
 JoinOrder = NULL;
 JoinMethods = NULL;

 RC = XXX_OK;
}

//-----------------------------------------Till here no executable plan generated-------------

/**
 *  Generate a RelPlanOp for this block
 */
XxxErrCode RelKRJNaiveOptimizer::
GeneratePlanOp(RelOptEngine::PlanObject* Plan,   RelPlanOp*& PlanOp) const
{
 KRJNaivePlan* KRJPlan = (KRJNaivePlan*)(Plan);
 int NumChildren = KRJPlan->ArraySize;
 
 ostream* OSt = XXX_THISENV().OutputStr; 
 
 // Step 1: Generate plan ops for all the child query blocks.
 RelPlanOp** ChildPlanOps = new RelPlanOp*[NumChildren];
 // InputArity keeps track of the arity of each input relation
 // before join reordering. This could be needed for Remapping
 int *InputArity = new int[NumChildren];
 if (getChildPlans(KRJPlan, ChildPlanOps, InputArity) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error getting child plans");
   return XXX_NOT_OK;
  }

 // Step 2: Create the preds to be applied during the joins
 RelKRJOptimizer::PredPlanStruct PredPlans;
 if (createPreds(KRJPlan, PredPlans) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error creating preds");
   return XXX_NOT_OK;
  }

 // Step 3: Construct a remapping
 XxxExprPlan::VarRemapping Remap(NumChildren);
 if (constructRemap(KRJPlan, Remap, NumChildren, InputArity, XXX_FALSE) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error constructing remap");
   return XXX_NOT_OK;
  }
 // Step 4: modify all the expressions using the VarRemapping
 if (exprRemap(Remap, PredPlans, NumChildren, ExprPlanList,OrderPlanList) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error applying remap");
   return XXX_NOT_OK;
  }

 // Step 5: generate a plan operator for the 

 if (makePlanOp(ChildPlanOps, NumChildren, PredPlans, ExprPlanList,OrderPlanList,
		BlockPlan->CatalogInfo,
		((NumChildren>1)?KRJPlan->JoinMethods[0]:RelOptEngine::RJOIN),
		Distinct,
		PlanOp) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error making plan operator");
   return XXX_NOT_OK;
  }

 delete [] InputArity;
 delete [] ChildPlanOps;
 // Do not reclaim the pred arrays, because the plan ops use them.
 // Change this eventually : TODO:L2
 // PredPlans.Reclaim();

 //KRJPlan->Print(*OSt); //ihab
 return XXX_OK;
}

//----------------------------------------

XxxErrCode RelKRJNaiveOptimizer::
getChildPlans(KRJNaivePlan* KRJPlan, RelPlanOp** ChildPlanOps, int* InputArity) const
{
 int NumChildren = KRJPlan->ArraySize;
 // no need for exec parameters for children
 RelExecParameters* Params = NULL; 

 // Since this is a naive optimizer, we do not have
 // planIds for these child plans. so instead, get the
 // best plans from each. Change this later.
 // here is the place to permute the ChildPlanOps
 // based on the join order
 for (int i = 0; i < NumChildren; i++) {
   if (BlockPlan->GetChild(KRJPlan->JoinOrder[i])->GeneratePlanOp(Params, ChildPlanOps[i]) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "No planop for child of KRJ");
     return XXX_NOT_OK;
    }
   // fill in the InputArity array
   InputArity[KRJPlan->JoinOrder[i]] = ChildPlanOps[i]->CatalogInfo->RecordSchema()->Arity();
  }
 return XXX_OK;
}
//----------------------------------------
XxxErrCode RelKRJNaiveOptimizer::
getInitPlan(KRJNaivePlan* KRJPlan,
	    XxxBooleanExpressionPlan*& InitPlan) const
{
 int i;
 for (i=0; i < KRJPlan->InitPreds.ArraySize; i++) {
   if (KRJPlan->InitPreds.GetBit(i)) {
     if (XxxBooleanExpressionPlan::MakeConjunction
	 (InitPlan, PredArray[i]) != XXX_OK) {
       REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
       return XXX_NOT_OK;
      }
    }
  }
 return XXX_OK;
}
//------------------------------------------
XxxErrCode RelKRJNaiveOptimizer::
getJoinPlans(KRJNaivePlan* KRJPlan,
	     RelKRJOptimizer::PredPlanStruct& PredPlans,
	     int NumChildren,
	     int JoinOffset,
	     int SkipTemporary) const
{
 int i, j;
 // This space is reclaimed in the destructor of the planop
 PredPlans.SelPredPlans = new XxxBooleanExpressionPlan*[NumChildren];
 PredPlans.JoinPredPlans = new XxxBooleanExpressionPlan*[NumChildren];
 PredPlans.ResidPredPlans = new XxxBooleanExpressionPlan*[NumChildren];
 for (j=0; j < NumChildren; j++) {
   PredPlans.SelPredPlans[j] = NULL;
   PredPlans.JoinPredPlans[j] = NULL;
   PredPlans.ResidPredPlans[j] = NULL;

   // no preds to be applied to intermediate results
   if (SkipTemporary && (JoinOffset > 0) && (j == 0)) continue; 

   for (i=0; i < KRJPlan->SelPreds[JoinOffset+j].ArraySize; i++) {
     if (KRJPlan->SelPreds[JoinOffset+j].GetBit(i)) {
       if (XxxBooleanExpressionPlan::MakeConjunction
	   (PredPlans.SelPredPlans[j], PredArray[i]) != XXX_OK) {
	 REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
	 return XXX_NOT_OK;
	}
      }
    } // for each SelPred
   for (i=0; i < KRJPlan->JoinPreds[JoinOffset+j].ArraySize; i++) {
     if (KRJPlan->JoinPreds[JoinOffset+j].GetBit(i)) {
       if (XxxBooleanExpressionPlan::MakeConjunction
	   (PredPlans.JoinPredPlans[j], PredArray[i]) != XXX_OK) {
	 REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
	 return XXX_NOT_OK;
	}
      }
    }// for each JoinPred
   for (i=0; i < KRJPlan->ResidPreds[JoinOffset+j].ArraySize; i++) {
     if (KRJPlan->ResidPreds[JoinOffset+j].GetBit(i)) {
       if (XxxBooleanExpressionPlan::MakeConjunction
	   (PredPlans.ResidPredPlans[j], PredArray[i]) != XXX_OK) {
	 REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
	 return XXX_NOT_OK;
	}
      }
    } // for each ResidPred
  } // for each child
 return XXX_OK;
}
//-----------------------------------------
/**
 * Create the preds to be applied during the joins
 */
XxxErrCode RelKRJNaiveOptimizer::
createPreds(KRJNaivePlan* KRJPlan, RelKRJOptimizer::PredPlanStruct& PredPlans) const
{
 int NumChildren = KRJPlan->ArraySize;

 // Map from pred bitmaps to actual expression plans.
 PredPlans.InitPlan = NULL;
 if (getInitPlan(KRJPlan, PredPlans.InitPlan) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error in getInitPlan");
   return XXX_NOT_OK;
  }

 int Offset = 0;
 if (getJoinPlans(KRJPlan, PredPlans, NumChildren, Offset,
		  XXX_FALSE) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error in getJoinPlan");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}
//----------------------------
XxxErrCode RelKRJNaiveOptimizer::
constructRemap(KRJNaivePlan* KRJPlan,
	       XxxExprPlan::VarRemapping& Remap,
	       int NewArity,
	       int *InputArity,
	       XxxBool MapAttrs) const
{
 int NumChildren = KRJPlan->ArraySize;
 int i, j;

 // InverseJoinorder is as the name implies, the inverse
 // of the JoinOrder mapping. JoinOrder[i] is the index of
 // the child that is at the i'th position in the join.
 // InverseJoinOrder[i] is the position of the i'th child.
 int *InverseJoinOrder = new int[NumChildren];
 for (i=0; i < NumChildren; i++) {
   InverseJoinOrder[KRJPlan->JoinOrder[i]] = i;
  }

 // modify the expression plans to reflect the join order
 Remap.NewArity = NewArity;
 for (i=0; i < NumChildren; i++) {
   Remap.SrcMappingArray[i].NewSrc = InverseJoinOrder[i];
   if (MapAttrs) {
     Remap.SrcMappingArray[i].NewAttrMapping = new int[InputArity[i]]; 
     // initially, leave the attribute mapping alone
     for (j = 0; j < InputArity[i]; j++) {
       Remap.SrcMappingArray[i].NewAttrMapping[j] = j;
      }
    }
   else {
     // leave the attr mapping as NULL
     Remap.SrcMappingArray[i].NewAttrMapping = NULL;
    }
  }

 delete [] InverseJoinOrder;
 return XXX_OK;
}

XxxErrCode RelKRJNaiveOptimizer::
exprRemap(XxxExprPlan::VarRemapping& Remap,
	  RelKRJOptimizer::PredPlanStruct& PredPlans,
	  int NumChildren,
	  XxxValExprPlanList* ExprPlanList,
	  XxxValExprPlanList* OrderPlanList ) const
{
 int i;

 for (i = 0; i < NumChildren; i++) {
   if (PredPlans.SelPredPlans[i] &&
       PredPlans.SelPredPlans[i]->RedirectDependencies(Remap) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error redirecting sel pred");
     return XXX_NOT_OK;
    }
   if (PredPlans.JoinPredPlans[i] &&
       PredPlans.JoinPredPlans[i]->RedirectDependencies(Remap) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error redirecting join pred");
     return XXX_NOT_OK;
    }
   if (PredPlans.ResidPredPlans[i] &&
       PredPlans.ResidPredPlans[i]->RedirectDependencies(Remap) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error redirecting resid pred");
     return XXX_NOT_OK;
    }
  }
 if (ExprPlanList &&
     ExprPlanList->RedirectDependencies(Remap) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error redirecting target");
   return XXX_NOT_OK;
  }
 
 if (OrderPlanList &&
     OrderPlanList->RedirectDependencies(Remap) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error redirecting Ordering");
   return XXX_NOT_OK;
 }
 return XXX_OK;
}

//---------------------------------------------------------------------------

XxxErrCode RelKRJNaiveOptimizer::
makePlanOp(RelPlanOp** ChildPlanOps,
	   int NumChildren,
	   RelKRJOptimizer::PredPlanStruct& PredPlans,
	   XxxValExprPlanList* ExprPlanList,
	   XxxValExprPlanList* OrderPlanList,
	   RelCatalogInfo* CatInfo,
	   RelOptEngine::JoinMethod JoinMethod,
	   XxxBool ApplyDistinct,
	   RelPlanOp*& PlanOp,
	   int top) const
{
 
/*   PlanOp = new RelTNLPlanOp(NumChildren, ChildPlanOps,
				    CatInfo,
				    ExprPlanList, 
				    PredPlans.SelPredPlans,
				    PredPlans.JoinPredPlans,
				    PredPlans.ResidPredPlans,
				    PredPlans.InitPlan);
	
	
*/
  PlanOp = new RelKRJPlanOp(NumChildren, ChildPlanOps,
			   CatInfo,
			   ExprPlanList, 
			   OrderPlanList,
			   PredPlans.SelPredPlans,
			   PredPlans.JoinPredPlans,
			   PredPlans.ResidPredPlans,
			   PredPlans.InitPlan,
			   top);
  

 if ((NumChildren == 1) && (ChildPlanOps[0]->CatalogInfo->ImplInfo()->SortInfo.NumSortOrders)) {
   RelImplInfo* ImplInfo = ChildPlanOps[0]->CatalogInfo->ImplInfo();
   XxxBool* UsefulSortOrder = new XxxBool[ImplInfo->SortInfo.NumSortOrders];
   int* UsefulSortAttrs = new int[ImplInfo->SortInfo.NumSortOrders];
   unsigned int i;
   for (i = 0; i < ImplInfo->SortInfo.NumSortOrders; i++) {
     UsefulSortOrder[i] = XXX_FALSE;
     UsefulSortAttrs[i] = 0;
    }
   int NumUsefulOrders = 0;
   // Any sort orders in the input may carry over to the output.
   // Check each sort order in the input
   for (i = 0; i < ImplInfo->SortInfo.NumSortOrders; i++) {
     if ((UsefulSortAttrs[i] =
	  ImplInfo->SortInfo.SortInfoArray[i].CommonAttrs(ExprPlanList))) {
       UsefulSortOrder[i] = XXX_TRUE;
       NumUsefulOrders++;
      }
    } // for each input sort order
   if (NumUsefulOrders) {
     PlanOp->CatalogInfo->ImplInfo()->SortInfo.Init(NumUsefulOrders);
     int j = 0;
     for (i = 0; i < ImplInfo->SortInfo.NumSortOrders; i++) {
       if (UsefulSortOrder[i]) {
	 XxxSortInfo OutSInfo(UsefulSortAttrs[i]);
	 if (ImplInfo->SortInfo.SortInfoArray[i].
	     FillTarget(ExprPlanList, OutSInfo) != XXX_OK) {
	   REL_NOTE_ERROR(REL_BADEXEC, "Error making sort info");
	   return XXX_NOT_OK;
	  }
	 PlanOp->CatalogInfo->ImplInfo()->SortInfo[j].Init(OutSInfo);
	 j++;
	}
      } // for each input sort order
    } // if there are useful join orders
   delete [] UsefulSortOrder;
   delete [] UsefulSortAttrs;
  } // if there should be an output sort order assigned

 return XXX_OK;
  }

//----------------------------------------------------------------------------------
// Determines if a given join can be computed using Rank join
XxxErrCode
RelKRJNaiveOptimizer::isKRJJoinable(XxxBooleanExpressionPlan* JoinPlan,
				  XxxRelBooleanPlan*& KRJPlan,
				  XxxBooleanExpressionPlan*& ResidualPlan,
				  XxxBool& Success) const
{
  // TODO:L2 there could be more than one SM join plan!!
  Success = XXX_FALSE;
  
  XxxRelEnv *REnv = XXX_THISRELENV();
  if (!REnv->SortMergeEnabled) {
    return XXX_OK;
  }
  
  // Make sure the join condition is of the form "COLUMN_VAR = COLUMN_VAR"
 if (!JoinPlan)
   return XXX_OK;

 if (JoinPlan->BoolType == XXX_REL_BOOL_EXPR) {
   if (isKRJJoinable((XxxRelBooleanPlan*)JoinPlan, Success) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error checking relboolean plan");
     return XXX_NOT_OK;
   }
   KRJPlan = (XxxRelBooleanPlan*)JoinPlan;
   ResidualPlan = NULL;
   return XXX_OK;
  } // if it is a RelBoolean plan
 else { // it is a BoolBoolean plan
   if (((XxxBoolBooleanPlan*)JoinPlan)->OpType != XXX_AND_OP)
     return XXX_OK; // failure
   for (int i = 0; i < JoinPlan->Arity(); i++) {
     XxxBoolBooleanPlan* BoolJoinPlan  =      (XxxBoolBooleanPlan*)JoinPlan;
     XxxBooleanExpressionPlan* SubPlan =      (XxxBooleanExpressionPlan*)(BoolJoinPlan->SubPlans[i]);
     if (isKRJJoinable(SubPlan, KRJPlan,   ResidualPlan, Success) != XXX_OK) {
       REL_NOTE_ERROR(REL_BADOPT, "Error checking sub plan");
       return XXX_NOT_OK;
      }
     if (Success) {
       // tack on all the other plans to the Residual plan
       for (int j = 0; j < JoinPlan->Arity(); j++) {
	 if (i == j) continue;
	 XxxBooleanExpressionPlan* TmpPlan =   (XxxBooleanExpressionPlan*)(BoolJoinPlan->SubPlans[j]);
	 if (XxxBooleanExpressionPlan::MakeConjunction(ResidualPlan, TmpPlan) != XXX_OK) {
	   REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
	   return XXX_NOT_OK;
	 }
       }
       break; // break out of the loop is successful
     } // if (Success)
   } // for each child
 } // if it is a BoolBoolean
 return XXX_OK;
}

// Determines if a given join can be computed using Rank join
XxxErrCode
RelKRJNaiveOptimizer::isKRJJoinable(XxxRelBooleanPlan* JoinPlan,
				  XxxBool& Success) const
{
  Success = XXX_FALSE;
  
  if (JoinPlan->OperType != XXX_EQ_OP)
    return XXX_OK;
  XxxValueExprPlan *LeftPlan = JoinPlan->SubPlans[0];
  XxxValueExprPlan *RightPlan = JoinPlan->SubPlans[1];
  if ((RightPlan->ValType() != XxxValueExpression::COLUMN_VAR) ||
      (LeftPlan->ValType() != XxxValueExpression::COLUMN_VAR))
    return XXX_OK;
  // Insure the type is linearly ordered
  if (!(XxxADTArray[RightPlan->BasicType()]->OrderedType(RightPlan->GenerateCatInfo())))
    return XXX_OK;
  
  Success = XXX_TRUE;
  return XXX_OK;
}


//-----------------------------Prints ---------------------------------------------
void RelKRJNaiveOptimizer::printPlan(ostream& OStr, int NumRels,
				     int* JoinOrder,
				     RelOptEngine::JoinMethod* JoinMethods,
				     const RelOptEngine::Cost& Cost) const
{
  for (int i = 0; i < NumRels; i++) {
      switch (JoinMethods[i-1]) {
      case RelOptEngine::RJOIN:
	OStr << " RJOIN "; break;
      default:
	OStr << " BAD "; break;
      } // switch
      OStr << JoinOrder[i];
  } // for each child
  
  OStr << endl;
  Cost.Print(OStr);
}
void RelKRJNaiveOptimizer::
PrintPlans(const RelOptEngine::PlanObject* Plan,
	   ostream& OStr) const
{
 KRJNaivePlan* KRJPlan = (KRJNaivePlan*)(Plan);
 OStr << " <PROJ EXPR: ";
 ExprPlanList->Print(OStr);
 OStr << " >\n" << "SEL EXPR: ";
 for (int i = 0; i < NumPreds; i++) {
   PredArray[i]->Print(OStr);
   if (i != NumPreds-1) OStr << ", ";
  }
 OStr << " >\n" ;
 KRJPlan->Print(OStr);
 OStr << " >\n" ;
}

RelKRJNaiveOptimizer::
KRJNaivePlan::~KRJNaivePlan()
{
 if (JoinOrder) delete [] JoinOrder;
 if (JoinMethods) delete [] JoinMethods;

 if (SelPreds) delete [] SelPreds;
 if (JoinPreds) delete [] JoinPreds;
 if (ResidPreds) delete [] ResidPreds;
}

void RelKRJNaiveOptimizer::
KRJNaivePlan::Print(ostream& OStr) const
{
 if (JoinOrder) {
   OStr << " [KRJ Join Order : ";
   for (int i =0; i < ArraySize; i++) {
     if (i > 0)
       OStr << RelOptEngine::JoinMethodNames[JoinMethods[i-1]];
     else OStr << " ";
     OStr << JoinOrder[i];
    }
   OStr << "] ";
  }
 if (InitPreds.NumBitsSet()) {
   OStr << "[INIT:";
   InitPreds.Print(OStr);
   OStr << "]";
  }
 if (SelPreds) {
   OStr << "[SelPredPlans:";
   for (int i = 0; i < ArraySize; i++) {
     if (i>0) OStr << ",";
     if (SelPreds[i].NumBitsSet())
       SelPreds[i].Print(OStr);
     else 
       OStr << "_";
    }
   OStr << "]";
  }
 if (JoinPreds) {
   OStr << "[JoinPredPlans:";
   for (int i = 0; i < ArraySize; i++) {
     if (i>0) OStr << ",";
     if (JoinPreds[i].NumBitsSet())
       JoinPreds[i].Print(OStr);
     else 
       OStr << "_";
    }
   OStr << "]";
  }
 if (ResidPreds) {
   OStr << "[ResidPredPlans:";
   for (int i = 0; i < ArraySize; i++) {
     if (ResidPreds[i].NumBitsSet())
       ResidPreds[i].Print(OStr);
     else 
       OStr << "_";
    }
   OStr << "]";
  }
}


