/*
Ihab
 new file for the basic KRJ optimizer (multiway)
*/
#ifndef XXX_NAIVEKRJ_H
#define XXX_NAIVEKRJ_H

#include "planKRJ.h"

/** 
 * The most naive KRJ optimizer.
 */
class RelKRJNaiveOptimizer : public RelKRJOptimizer {
 public:
  /**
   // This plan maintains the order in which the relations
   // are to be joined and the predicates to be applied at
   // each stage.
   */
  class KRJNaivePlan : public RelOptEngine::PlanObject { public:

    int ArraySize;
    // join ordering effected by using one level of indirection
    int *JoinOrder;
    // join method array: size = 1-NumChildren
    RelOptEngine::JoinMethod *JoinMethods;

    // The size of each array == NumChildren.
    // array of selection predicate bitmaps
    XxxBitMap *SelPreds;
    // array of join predicate bitmaps: these are important 
    // for SM joins (and hash-joins when we implement them) 
    // -- for NL joins, this combines with the Residual preds
    XxxBitMap *JoinPreds;
    // Residual predicates that need to be applied after each join.
    XxxBitMap *ResidPreds;

    // Initial predicates that should be applied before any tables 
    // are accessed. Usually, this should be empty unless there 
    // are preds of the form 3 = 2 + 1
    // XxxBooleanExpressionPlan* InitPlan;
    XxxBitMap InitPreds;

    KRJNaivePlan(XxxErrCode& RC,
		 int PlanIdent,
		 unsigned char BlockIdent,
		 int NumChildren,
		 RelKRJNaiveOptimizer* Opt);
    virtual ~KRJNaivePlan();
    virtual void Print(ostream& OStr) const;
   };

  RelKRJNaiveOptimizer(XxxErrCode& RC,
		      XxxValueExprList* TargetList,
		      XxxBooleanExpression* Condition,
		      XxxValueExprList* Ordering,
		      XxxBool Unique,
		      RelKRJBlockPlan* KRJPlan,
		      RelOptEngine::OptQueryEnv* QEnv);
  virtual ~RelKRJNaiveOptimizer();

  virtual XxxErrCode Optimize();
  virtual XxxErrCode GeneratePlanOp(RelOptEngine::PlanObject* Plan,
				    RelPlanOp*& PlanOp) const;
  virtual void PrintPlans(const RelOptEngine::PlanObject* Plan,
			  ostream& OStr) const;
 // inside this complex plan, is there one that can be used for Rank join?
  XxxErrCode isKRJJoinable (XxxBooleanExpressionPlan* JoinPlan,
			   XxxRelBooleanPlan*& KRJPlan,
			   XxxBooleanExpressionPlan*& ResidualPlan,
			   XxxBool& Success) const;
  
  // is this a simple plan that can be used for a Rank join?
  XxxErrCode isKRJJoinable (XxxRelBooleanPlan* JoinPlan,
			   XxxBool& Success) const;
  
  // This virtual method simply provides some permutation
  // of the tables as a join order.
  // Therefore, one can subclass this and simply redefine
  // this one method to have another naive optimizer
  // that uses a different join order heuristic.
  virtual XxxErrCode determineJoinStrategy
    (int NumChildren, int* JoinOrder,
     RelOptEngine::JoinMethod* JoinMethods) const;

 protected:
  //// Get plans for children of this block
  XxxErrCode getChildPlans(KRJNaivePlan* KRJPlan,
			   RelPlanOp** ChildPlanOps,
			   int* InputArity) const;
  //// Create the preds to be applied during the joins
  XxxErrCode createPreds(KRJNaivePlan* KRJPlan,
			 RelKRJOptimizer::PredPlanStruct& PredPlans) const;
  //// Get the InitPlan -- used by createPreds
  XxxErrCode getInitPlan(KRJNaivePlan* KRJPlan,
			 XxxBooleanExpressionPlan*& InitPlan) const;
  //// Get the preds for the joins -- used by createPreds
  XxxErrCode getJoinPlans(KRJNaivePlan* KRJPlan,
			  RelKRJOptimizer::PredPlanStruct& PredPlans,
			  int NumChildren,
			  int JoinOffset,
			  int SkipTemporary) const;

  //// Construct a remapping
  XxxErrCode constructRemap(KRJNaivePlan* KRJPlan,
			    XxxExprPlan::VarRemapping& Remap,
			    int NewArity,
			    int *InputArity,
			    XxxBool MapAttrs) const;
  //// Modify all the expressions using the VarRemapping
  XxxErrCode exprRemap(XxxExprPlan::VarRemapping& Remap,
		       RelKRJOptimizer::PredPlanStruct& PredPlans,
		       int NumChildren,
		       XxxValExprPlanList* ExprPlanList,
		       XxxValExprPlanList* OrderPlanList) const;
  

  XxxErrCode makePlanOp(RelPlanOp** ChildPlanOps,
			int NumChildren,
			RelKRJOptimizer::PredPlanStruct& PredPlans,
			XxxValExprPlanList* ExprPlanList,
			XxxValExprPlanList* OrderPlanList,
			RelCatalogInfo* CatInfo,
			RelOptEngine::JoinMethod JoinMethod,
			XxxBool ApplyDistinct,
			RelPlanOp*& PlanOp,int top =0) const;

  void printPlan(ostream& OStr, int NumRels,
		 int* JoinOrder,
		 RelOptEngine::JoinMethod* JoinMethods,
		 const RelOptEngine::Cost& Cost) const;
 };

#endif // XXX_NAIVEKRJ_H
