#include <iostream.h>

#include "relerror.h"
#include "relquery.h"
#include "naivespj.h"
#include "relplanop.h"
#include "relenv.h"
#include "funcexpr.h"
#include "adtfunc.h"
#include "valbool.h"
//Begin Moustafa
#include "hfewplanop.h"
#include "nlfewplanop.h"
#include "hbewplanop.h"
#include "nlbewplanop.h"
//End Moustafa
#include "ClientSiteOperator.h"
#include "JavaPlans.h"

#include "deviceplanop.h"


XxxBool IsJavaExpr();
XxxErrCode makeJavaExpr(int NumChildren, RelPlanOp** ChildPlanOp, RelSPJOptimizer::PredPlanStruct& PredPlans);

//
// Methods of RelSPJNaiveOptimizer
//
RelSPJNaiveOptimizer::
RelSPJNaiveOptimizer(XxxErrCode& RC,
		     XxxValueExprList* TargetList,
		     XxxBooleanExpression* Condition,
		     XxxBool Unique,
		     RelSPJBlockPlan* SPJPlan,
		     RelOptEngine::OptQueryEnv* QEnv)
: RelSPJOptimizer(RC, TargetList, Condition, Unique, SPJPlan, QEnv)
{
 if (RC != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error creating RelSPJOptimizer");
   return;
  }
 // the parent class constructor has already split the WHERE
 // clause predicates into conjunctive boolean factors, and
 // determined the dependencies of these factors.

 // What remains is to construct a query plan from this.
 
 RC = XXX_OK;

}

RelSPJNaiveOptimizer::~RelSPJNaiveOptimizer()
{
}

XxxErrCode RelSPJNaiveOptimizer::Optimize()
{
 XxxErrCode rc = XXX_OK;
 SPJNaivePlan* Plan = new SPJNaivePlan(rc, BlockPlan->HighestPlanId++,
				       BlockPlan->BlockId,
				       BlockPlan->ChildArrayLen,
				       this);
 if (rc != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error creating naive plan");
   return XXX_NOT_OK;
  }

 if (Plan->ArraySize) {
   int i;
   //
   // Since this is an ultra-naive optimizer, the join order
   // is determined independent of any of the predicates.
   // A slightly less naive optimizer would use the predicate
   // selectivities as well. In that case, the PredDependencyInfo
   // would be useful.
   // 
   Plan->JoinOrder = new int[Plan->ArraySize];
   // given N relations, there are (N-1) joins
   if (Plan->ArraySize > 1) 
     Plan->JoinMethods = 
       new RelOptEngine::JoinMethod[Plan->ArraySize-1];
   //Begin Moustafa
   // Add HasStreamInput parameter
   // At this stage, any join order manipulation should be done.
   //   if (determineJoinStrategy(Plan->ArraySize, 
   //			     Plan->JoinOrder,
   //			     Plan->JoinMethods) != XXX_OK) {

   if (determineJoinStrategy(Plan->ArraySize, 
			     Plan->JoinOrder,
			     Plan->JoinMethods, 
			     BlockPlan->HasStreamInput) != XXX_OK) {
     //End Moustafa
     REL_NOTE_ERROR(REL_BADOPT, "Error generating join strategy");
     return XXX_NOT_OK;
    }

   // use pred dependencies to determine placement
   // allocate the SelPreds, JoinPreds, and ResidPreds arrays
   Plan->SelPreds = new XxxBitMap[Plan->ArraySize];
   Plan->JoinPreds = new XxxBitMap[Plan->ArraySize];
   Plan->ResidPreds = new XxxBitMap[Plan->ArraySize];
   for (i = 0; i < Plan->ArraySize; i++) {
     Plan->SelPreds[i].ChangeSize(NumPreds);
     Plan->JoinPreds[i].ChangeSize(NumPreds);
     Plan->ResidPreds[i].ChangeSize(NumPreds);
    }

   // I have Changed JoinOrder -> InverseJoinOrder and am computing 
   // InverseJoinOrder here.
   // InverseJoinorder is as the name implies, the inverse
   // of the JoinOrder mapping. JoinOrder[i] is the index of
   // the child that is at the i'th position in the join.
   // InverseJoinOrder[i] is the position of the i'th child.
   int *InverseJoinOrder = new int[Plan->ArraySize];
   for (i=0; i < Plan->ArraySize; i++) {
     InverseJoinOrder[Plan->JoinOrder[i]] = i;
   }

   // for each of the boolean factors, find a place for it.
   // factor in the join order here!!
   for (int j = 0; j < NumPreds; j++) {
     int DepMax = -1;
     int DepCount = 0;
     // note: dont have to check Arity()+1, because if it is only 
     // dependent on the Env, then it should be applied as an InitPred
     for (i =0; i < Plan->ArraySize; i++) {
       if (PredDependencyInfo[j][i] == XXX_TRUE) {
	 DepCount++;
	 if (InverseJoinOrder[i] > DepMax)
	   DepMax = InverseJoinOrder[i];
	}
      } // for each child
     if (DepMax < 0) {
       // trivial selection of the form 3 = 2+1 or 3 = $1
       Plan->InitPreds.SetBit(j);
      }
     else if (DepCount == 1) {
       Plan->SelPreds[DepMax].SetBit(j);
      }
     else {
       Plan->JoinPreds[DepMax].SetBit(j);
      }
    } // for each boolean factor

    delete[] InverseJoinOrder;
  } // if (Plan->ArraySize)

 BlockPlan->AddPlan(Plan);
 if (XXX_THISRELENV()->DebugOpt) {
   cerr << "SN: Generating Plan: ";
   Plan->Print(cerr);
   cerr << endl;
 }
 return XXX_OK;
}

// For the naive optimizer, we do not worry about join order,
// and join method. The join order is right to left, and
// the join method is NL join. For now, lets not even worry
// about indexing...
// Begin Moustafa
//XxxErrCode RelSPJNaiveOptimizer::
//determineJoinStrategy(int NumChildren, 
//		      int* JoinOrder,
//		      RelOptEngine::JoinMethod* JoinMethods) const

XxxErrCode RelSPJNaiveOptimizer::
determineJoinStrategy(int NumChildren, 
		      int* JoinOrder,
		      RelOptEngine::JoinMethod* JoinMethods, 
		      XxxBool HasStreamInput) const
// End Moustafa
{
 XxxRelEnv* REnv =((XxxRelEnv*)(XXX_THISENGENV(RelationEngineId)));
 // For now: use the reverse join order as the user spec
 for (int i=0; i < NumChildren; i++) {
   // JoinOrder[i] = i;  // simple join order
   JoinOrder[i] = NumChildren-i-1; // reverse join order
   if (i > 0) {
     if(HasStreamInput)
       switch (REnv->WJoinMethod) {
       case 0:
	 JoinMethods[i-1] = RelOptEngine::HBEWINDOW; 
	 break;
       case 1:
	 JoinMethods[i-1] = RelOptEngine::NLBEWINDOW; 
	 break;
       case 2:
	 JoinMethods[i-1] = RelOptEngine::HFEWINDOW; 
	 break;
       case 3:
	 JoinMethods[i-1] = RelOptEngine::NLFEWINDOW; 
	 break;
       default:
	 REL_NOTE_ERROR(REL_BADOPT, "Unknown window join method");
	 return XXX_NOT_OK;
       }
       //JoinMethods[i-1] = RelOptEngine::HFEWINDOW;
       //JoinMethods[i-1] = RelOptEngine::NLFEWINDOW; 
       //JoinMethods[i-1] = RelOptEngine::HBEWINDOW; 
       //JoinMethods[i-1] = RelOptEngine::NLBEWINDOW; 
     else
       switch (REnv->NLJoinMethod) {
       case 0:
	 JoinMethods[i-1] = RelOptEngine::TNLOOPS; 
	 break;
       case 1:
	 JoinMethods[i-1] = RelOptEngine::PNLOOPS; 
	 break;
       default:
	 REL_NOTE_ERROR(REL_BADOPT, "Unknown join method");
	 return XXX_NOT_OK;
       } // switch
   } // if i > 0
 } // for each input
 return XXX_OK;
}

RelSPJNaiveOptimizer::
SPJNaivePlan::SPJNaivePlan(XxxErrCode& RC,
			   int PlanIdent,
			   unsigned char BlockIdent,
			   int NumChildren,
			   RelSPJNaiveOptimizer* Opt)
: RelOptEngine::PlanObject(PlanIdent, BlockIdent),
  InitPreds(Opt->NumPreds)
{
 ArraySize = NumChildren;
 JoinOrder = NULL;
 JoinMethods = NULL;

 RC = XXX_OK;
}

/**
 *  Generate a RelPlanOp for this block
 */
XxxErrCode RelSPJNaiveOptimizer::
GeneratePlanOp(RelOptEngine::PlanObject* Plan,
	       RelPlanOp*& PlanOp) const
{
 SPJNaivePlan* SPJPlan = (SPJNaivePlan*)(Plan);
 int NumChildren = SPJPlan->ArraySize;

 // Step 1: Generate plan ops for all the child query blocks.
 RelPlanOp** ChildPlanOps = new RelPlanOp*[NumChildren];
 // InputArity keeps track of the arity of each input relation
 // before join reordering. This could be needed for Remapping
 int *InputArity = new int[NumChildren];
 if (getChildPlans(SPJPlan, ChildPlanOps, InputArity) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error getting child plans");
   return XXX_NOT_OK;
  }

 // Step 2: Create the preds to be applied during the joins
 RelSPJOptimizer::PredPlanStruct PredPlans;
 if (createPreds(SPJPlan, PredPlans) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error creating preds");
   return XXX_NOT_OK;
  }

 // Step 3: Construct a remapping
 XxxExprPlan::VarRemapping Remap(NumChildren);
 if (constructRemap(SPJPlan, Remap, NumChildren, InputArity,
		    XXX_FALSE) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error constructing remap");
   return XXX_NOT_OK;
  }

 // Step 4: modify all the expressions using the VarRemapping
 if (exprRemap(Remap, PredPlans, NumChildren, ExprPlanList) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error applying remap");
   return XXX_NOT_OK;
  }

 // Step 5: generate a plan operator for the SPJ
 if (makePlanOp(ChildPlanOps, NumChildren, PredPlans, ExprPlanList,
		BlockPlan->CatalogInfo,
		((NumChildren>1)?SPJPlan->JoinMethods[0]:RelOptEngine::TNLOOPS),
		Distinct,
		PlanOp) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error making plan operator");
   return XXX_NOT_OK;
  }

 delete [] InputArity;
 delete [] ChildPlanOps;
 // Do not reclaim the pred arrays, because the plan ops use them.
 // Change this eventually : TODO:L2
 // PredPlans.Reclaim();
 return XXX_OK;
}

void RelSPJNaiveOptimizer::
PrintPlans(const RelOptEngine::PlanObject* Plan,
	   ostream& OStr) const
{
 SPJNaivePlan* SPJPlan = (SPJNaivePlan*)(Plan);
 OStr << " <PROJ EXPR: ";
 ExprPlanList->Print(OStr);
 OStr << " >\n" << "SEL EXPR: ";
 for (int i = 0; i < NumPreds; i++) {
   PredArray[i]->Print(OStr);
   if (i != NumPreds-1) OStr << ", ";
  }
 OStr << " >\n" ;
 SPJPlan->Print(OStr);
 OStr << " >\n" ;
}

RelSPJNaiveOptimizer::
SPJNaivePlan::~SPJNaivePlan()
{
 if (JoinOrder) delete [] JoinOrder;
 if (JoinMethods) delete [] JoinMethods;

 if (SelPreds) delete [] SelPreds;
 if (JoinPreds) delete [] JoinPreds;
 if (ResidPreds) delete [] ResidPreds;
}

void RelSPJNaiveOptimizer::
SPJNaivePlan::Print(ostream& OStr) const
{
 if (JoinOrder) {
   OStr << " [Join Order : ";
   for (int i =0; i < ArraySize; i++) {
     if (i > 0)
       OStr << RelOptEngine::JoinMethodNames[JoinMethods[i-1]];
     else OStr << " ";
     OStr << JoinOrder[i];
    }
   OStr << "] ";
  }
 if (InitPreds.NumBitsSet()) {
   OStr << "[INIT:";
   InitPreds.Print(OStr);
   OStr << "]";
  }
 if (SelPreds) {
   OStr << "[SelPredPlans:";
   for (int i = 0; i < ArraySize; i++) {
     if (i>0) OStr << ",";
     if (SelPreds[i].NumBitsSet())
       SelPreds[i].Print(OStr);
     else 
       OStr << "_";
    }
   OStr << "]";
  }
 if (JoinPreds) {
   OStr << "[JoinPredPlans:";
   for (int i = 0; i < ArraySize; i++) {
     if (i>0) OStr << ",";
     if (JoinPreds[i].NumBitsSet())
       JoinPreds[i].Print(OStr);
     else 
       OStr << "_";
    }
   OStr << "]";
  }
 if (ResidPreds) {
   OStr << "[ResidPredPlans:";
   for (int i = 0; i < ArraySize; i++) {
     if (ResidPreds[i].NumBitsSet())
       ResidPreds[i].Print(OStr);
     else 
       OStr << "_";
    }
   OStr << "]";
  }
}
// Determines if a given join can be computed using sort-merge join
XxxErrCode
RelSPJNaiveOptimizer::isSMJoinable(XxxBooleanExpressionPlan* JoinPlan,
				   XxxRelBooleanPlan*& SMPlan,
				   XxxBooleanExpressionPlan*& ResidualPlan,
				   XxxBool& Success) const
{
 // TODO:L2 there could be more than one SM join plan!!
 Success = XXX_FALSE;

 XxxRelEnv *REnv = XXX_THISRELENV();
 if (!REnv->SortMergeEnabled) {
   return XXX_OK;
  }

 // Make sure the join condition is of the form "COLUMN_VAR = COLUMN_VAR"
 if (!JoinPlan)
   return XXX_OK;

 if (JoinPlan->BoolType == XXX_REL_BOOL_EXPR) {
   if (isSMJoinable((XxxRelBooleanPlan*)JoinPlan, 
		    Success) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error checking relboolean plan");
     return XXX_NOT_OK;
    }
   SMPlan = (XxxRelBooleanPlan*)JoinPlan;
   ResidualPlan = NULL;
   return XXX_OK;
  } // if it is a RelBoolean plan
 else { // it is a BoolBoolean plan
   if (((XxxBoolBooleanPlan*)JoinPlan)->OpType != XXX_AND_OP)
     return XXX_OK; // failure
   for (int i = 0; i < JoinPlan->Arity(); i++) {
     XxxBoolBooleanPlan* BoolJoinPlan = 
       (XxxBoolBooleanPlan*)JoinPlan;
     XxxBooleanExpressionPlan* SubPlan = 
       (XxxBooleanExpressionPlan*)(BoolJoinPlan->SubPlans[i]);
     if (isSMJoinable(SubPlan, SMPlan, 
		      ResidualPlan, Success) != XXX_OK) {
       REL_NOTE_ERROR(REL_BADOPT, "Error checking sub plan");
       return XXX_NOT_OK;
      }
     if (Success) {
       // tack on all the other plans to the Residual plan
       for (int j = 0; j < JoinPlan->Arity(); j++) {
	 if (i == j) continue;
	 XxxBooleanExpressionPlan* TmpPlan =
	   (XxxBooleanExpressionPlan*)(BoolJoinPlan->SubPlans[j]);
	 if (XxxBooleanExpressionPlan::MakeConjunction
	     (ResidualPlan, TmpPlan) != XXX_OK) {
	   REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
	   return XXX_NOT_OK;
	  }
	}
       break; // break out of the loop is successful
      } // if (Success)
    } // for each child
  } // if it is a BoolBoolean
 return XXX_OK;
}

// Determines if a given join can be computed using sort-merge join
XxxErrCode
RelSPJNaiveOptimizer::isSMJoinable(XxxRelBooleanPlan* JoinPlan,
				   XxxBool& Success) const
{
 Success = XXX_FALSE;

 if (JoinPlan->OperType != XXX_EQ_OP)
   return XXX_OK;
 XxxValueExprPlan *LeftPlan = JoinPlan->SubPlans[0];
 XxxValueExprPlan *RightPlan = JoinPlan->SubPlans[1];
 if ((RightPlan->ValType() != XxxValueExpression::COLUMN_VAR) ||
     (LeftPlan->ValType() != XxxValueExpression::COLUMN_VAR))
   return XXX_OK;
 // Insure the type is linearly ordered
 if (!(XxxADTArray[RightPlan->BasicType()]->OrderedType(RightPlan->GenerateCatInfo())))
   return XXX_OK;

 Success = XXX_TRUE;
 return XXX_OK;
}

XxxErrCode 
RelSPJNaiveOptimizer::CreateSMPlanOp(RelPlanOp*& PlanOp,
				     int NumChildren,
				     RelPlanOp* Children[],
				     RelCatalogInfo* CatInfo,
				     XxxValExprPlanList* ExprPlanList,
				     XxxBooleanExpressionPlan** SelPredPlans,
				     XxxBooleanExpressionPlan** JoinPredPlans,
				     XxxBooleanExpressionPlan** ResidPredPlans,
				     XxxBooleanExpressionPlan* InitPlan,
				     const SMPlanGlue& Glue)
{
 ASSERT(NumChildren == 2);
 
 RelPlanOp *RightSortPlanOp = Children[Glue.RightRelIndex];
 RelPlanOp *LeftSortPlanOp = Children[Glue.LeftRelIndex];
 // Create the sort planops
 if (Glue.SortRight) {
   RightSortPlanOp = 
     new RelSortPlanOp(Children[Glue.RightRelIndex], Glue.RightSInfo, 
		       XXX_FALSE, Glue.MaterializeRtInput, Glue.RightRunSize,
		       Children[Glue.RightRelIndex]->CatalogInfo);
  }
 if (Glue.SortLeft) {
   LeftSortPlanOp = 
     new RelSortPlanOp(Children[Glue.LeftRelIndex], Glue.LeftSInfo, 
		       XXX_FALSE, Glue.MaterializeLtInput, Glue.LeftRunSize,
		       Children[Glue.LeftRelIndex]->CatalogInfo);
  }
 Children[Glue.LeftRelIndex] = LeftSortPlanOp;
 Children[Glue.RightRelIndex] = RightSortPlanOp;
 
 PlanOp = new RelSMPlanOp(2, Children,
			  CatInfo,
			  ExprPlanList, 
			  SelPredPlans,
			  JoinPredPlans,
			  ResidPredPlans,
			  InitPlan);
 // make sure to specify the SortInfo in the result
 int RSortAttrs = 0;
 int LSortAttrs = 0;
 int NumSortOrders = 0;

 // !tok22
 if (RSortAttrs == Glue.RightSInfo.CommonAttrs(ExprPlanList, 
					      Glue.RightRelIndex)) {
   NumSortOrders++;
  }
 if (LSortAttrs == Glue.LeftSInfo.CommonAttrs(ExprPlanList, 
					     Glue.LeftRelIndex)) {
   NumSortOrders++;
  }
 if (NumSortOrders) {
   PlanOp->CatalogInfo->ImplInfo()->SortInfo.Init(NumSortOrders);
   if (RSortAttrs) {
     XxxSortInfo OutSInfo(RSortAttrs);
     if (Glue.RightSInfo.FillTarget(ExprPlanList, OutSInfo,
				    Glue.RightRelIndex) != XXX_OK) {
       REL_NOTE_ERROR(REL_BADEXEC, "Error making sort info");
       return XXX_NOT_OK;
      }
     PlanOp->CatalogInfo->ImplInfo()->SortInfo[0].Init(OutSInfo);
    } // if (RSortAttrs)
   if (LSortAttrs) {
     XxxSortInfo OutSInfo(LSortAttrs);
     if (Glue.LeftSInfo.FillTarget(ExprPlanList, OutSInfo,
				   Glue.LeftRelIndex) != XXX_OK) {
       REL_NOTE_ERROR(REL_BADEXEC, "Error making sort info");
       return XXX_NOT_OK;
      }
     PlanOp->CatalogInfo->ImplInfo()->SortInfo[NumSortOrders-1].Init(OutSInfo);
    } // if (LSortAttrs)
  } // if NumSortOrders

 return XXX_OK;
}

XxxErrCode RelSPJNaiveOptimizer::
getChildPlans(SPJNaivePlan* SPJPlan,
	      RelPlanOp** ChildPlanOps,
	      int* InputArity) const
{
 int NumChildren = SPJPlan->ArraySize;
 // no need for exec parameters for children
 RelExecParameters* Params = NULL; 

 // Since this is a naive optimizer, we do not have
 // planIds for these child plans. so instead, get the
 // best plans from each. Change this later.

 // here is the place to permute the ChildPlanOps
 // based on the join order
 for (int i = 0; i < NumChildren; i++) {
   if (BlockPlan->GetChild(SPJPlan->JoinOrder[i])->
       GeneratePlanOp(Params, ChildPlanOps[i]) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "No planop for child of spj");
     return XXX_NOT_OK;
    }
   // fill in the InputArity array
   InputArity[SPJPlan->JoinOrder[i]] = 
     ChildPlanOps[i]->CatalogInfo->RecordSchema()->Arity();
  }
 return XXX_OK;
}

XxxErrCode RelSPJNaiveOptimizer::
getInitPlan(SPJNaivePlan* SPJPlan,
	    XxxBooleanExpressionPlan*& InitPlan) const
{
 int i;
 for (i=0; i < SPJPlan->InitPreds.ArraySize; i++) {
   if (SPJPlan->InitPreds.GetBit(i)) {
     if (XxxBooleanExpressionPlan::MakeConjunction
	 (InitPlan, PredArray[i]) != XXX_OK) {
       REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
       return XXX_NOT_OK;
      }
    }
  }
 return XXX_OK;
}

XxxErrCode RelSPJNaiveOptimizer::
getJoinPlans(SPJNaivePlan* SPJPlan,
	     RelSPJOptimizer::PredPlanStruct& PredPlans,
	     int NumChildren,
	     int JoinOffset,
	     int SkipTemporary) const
{
 int i, j;
 // This space is reclaimed in the destructor of the planop
 PredPlans.SelPredPlans = new XxxBooleanExpressionPlan*[NumChildren];
 PredPlans.JoinPredPlans = new XxxBooleanExpressionPlan*[NumChildren];
 PredPlans.ResidPredPlans = new XxxBooleanExpressionPlan*[NumChildren];
 for (j=0; j < NumChildren; j++) {
   PredPlans.SelPredPlans[j] = NULL;
   PredPlans.JoinPredPlans[j] = NULL;
   PredPlans.ResidPredPlans[j] = NULL;

   // no preds to be applied to intermediate results
   if (SkipTemporary && (JoinOffset > 0) && (j == 0)) continue; 

   for (i=0; i < SPJPlan->SelPreds[JoinOffset+j].ArraySize; i++) {
     if (SPJPlan->SelPreds[JoinOffset+j].GetBit(i)) {
       if (XxxBooleanExpressionPlan::MakeConjunction
	   (PredPlans.SelPredPlans[j], PredArray[i]) != XXX_OK) {
	 REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
	 return XXX_NOT_OK;
	}
      }
    } // for each SelPred
   for (i=0; i < SPJPlan->JoinPreds[JoinOffset+j].ArraySize; i++) {
     if (SPJPlan->JoinPreds[JoinOffset+j].GetBit(i)) {
       if (XxxBooleanExpressionPlan::MakeConjunction
	   (PredPlans.JoinPredPlans[j], PredArray[i]) != XXX_OK) {
	 REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
	 return XXX_NOT_OK;
	}
      }
    }// for each JoinPred
   for (i=0; i < SPJPlan->ResidPreds[JoinOffset+j].ArraySize; i++) {
     if (SPJPlan->ResidPreds[JoinOffset+j].GetBit(i)) {
       if (XxxBooleanExpressionPlan::MakeConjunction
	   (PredPlans.ResidPredPlans[j], PredArray[i]) != XXX_OK) {
	 REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
	 return XXX_NOT_OK;
	}
      }
    } // for each ResidPred
  } // for each child
 return XXX_OK;
}

/**
 * Create the preds to be applied during the joins
 */
XxxErrCode RelSPJNaiveOptimizer::
createPreds(SPJNaivePlan* SPJPlan,
	    RelSPJOptimizer::PredPlanStruct& PredPlans) const
{
 int NumChildren = SPJPlan->ArraySize;

 // Map from pred bitmaps to actual expression plans.
 PredPlans.InitPlan = NULL;
 if (getInitPlan(SPJPlan, PredPlans.InitPlan) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error in getInitPlan");
   return XXX_NOT_OK;
  }

 int Offset = 0;
 if (getJoinPlans(SPJPlan, PredPlans, NumChildren, Offset,
		  XXX_FALSE) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error in getJoinPlan");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

XxxErrCode RelSPJNaiveOptimizer::
constructRemap(SPJNaivePlan* SPJPlan,
	       XxxExprPlan::VarRemapping& Remap,
	       int NewArity,
	       int *InputArity,
	       XxxBool MapAttrs) const
{
 int NumChildren = SPJPlan->ArraySize;
 int i, j;

 // InverseJoinorder is as the name implies, the inverse
 // of the JoinOrder mapping. JoinOrder[i] is the index of
 // the child that is at the i'th position in the join.
 // InverseJoinOrder[i] is the position of the i'th child.
 int *InverseJoinOrder = new int[NumChildren];
 for (i=0; i < NumChildren; i++) {
   InverseJoinOrder[SPJPlan->JoinOrder[i]] = i;
  }

 // modify the expression plans to reflect the join order
 Remap.NewArity = NewArity;
 for (i=0; i < NumChildren; i++) {
   Remap.SrcMappingArray[i].NewSrc = InverseJoinOrder[i];
   if (MapAttrs) {
     Remap.SrcMappingArray[i].NewAttrMapping = new int[InputArity[i]]; 
     // initially, leave the attribute mapping alone
     for (j = 0; j < InputArity[i]; j++) {
       Remap.SrcMappingArray[i].NewAttrMapping[j] = j;
      }
    }
   else {
     // leave the attr mapping as NULL
     Remap.SrcMappingArray[i].NewAttrMapping = NULL;
    }
  }

 delete [] InverseJoinOrder;
 return XXX_OK;
}

XxxErrCode RelSPJNaiveOptimizer::
exprRemap(XxxExprPlan::VarRemapping& Remap,
	  RelSPJOptimizer::PredPlanStruct& PredPlans,
	  int NumChildren,
	  XxxValExprPlanList* ExprPlanList) const
{
 int i;

 for (i = 0; i < NumChildren; i++) {
   if (PredPlans.SelPredPlans[i] &&
       PredPlans.SelPredPlans[i]->RedirectDependencies(Remap) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error redirecting sel pred");
     return XXX_NOT_OK;
    }
   if (PredPlans.JoinPredPlans[i] &&
       PredPlans.JoinPredPlans[i]->RedirectDependencies(Remap) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error redirecting join pred");
     return XXX_NOT_OK;
    }
   if (PredPlans.ResidPredPlans[i] &&
       PredPlans.ResidPredPlans[i]->RedirectDependencies(Remap) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error redirecting resid pred");
     return XXX_NOT_OK;
    }
  }
 if (ExprPlanList &&
     ExprPlanList->RedirectDependencies(Remap) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error redirecting target");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

void RelSPJNaiveOptimizer::MakeSMGlue(int NumChildren,
				      const RelPlanOp** Children,
				      const XxxRelBooleanPlan *JoinPlan, 
				      SMPlanGlue& Glue)
{
 const XxxUnknownValue::UnknownValuePlan *LeftPlan = 
   (const XxxUnknownValue::UnknownValuePlan*)(JoinPlan->SubPlans[0]);
 Glue.LeftSInfo[0].AttrIndex = LeftPlan->SrcAttrIndex;
 Glue.LeftSInfo[0].Asc = XXX_TRUE;
 Glue.LeftRelIndex = LeftPlan->SrcChild;
 const RelPlanOp *LeftPlanOp = Children[Glue.LeftRelIndex];
 // the optimizer decides if the sorting is necessary
 if (!LeftPlanOp->CatalogInfo->ImplInfo()->SortInfo.Matches(Glue.LeftSInfo)) {
   Glue.SortLeft = XXX_TRUE;
  /*
   * Ordinarily, we create a temporary SHORE file and dump the records
   * into it--then sort it--and return the results in a scan.  However,
   * it is very likely that the input to this operator is a SHORE file
   * relation.  We can save the overhead of creating a file by sorting 
   * the original file, if the input relation is created by a filescan
   * and there is no filter condition specified for the filescan.
   */
   Glue.MaterializeLtInput = !LeftPlanOp->IsFullFileScan();
   Glue.LeftRunSize = XXX_THISENV().DBEnv.BufPoolSize/2;
  }

 const XxxUnknownValue::UnknownValuePlan *RightPlan = 
   (const XxxUnknownValue::UnknownValuePlan*)(JoinPlan->SubPlans[1]);
 Glue.RightSInfo[0].AttrIndex = RightPlan->SrcAttrIndex;
 Glue.RightSInfo[0].Asc = XXX_TRUE;
 Glue.RightRelIndex = RightPlan->SrcChild;
 const RelPlanOp *RightPlanOp = Children[Glue.RightRelIndex];
 // the optimizer decides if the sorting is necessary
 if (!RightPlanOp->CatalogInfo->ImplInfo()->SortInfo.Matches(Glue.RightSInfo)) {
   Glue.SortRight = XXX_TRUE;
   Glue.MaterializeRtInput = !RightPlanOp->IsFullFileScan();
   Glue.RightRunSize = XXX_THISENV().DBEnv.BufPoolSize/2;
  }
}


/**
 * For the moment we consider that device UDFs are applied in the context
 * of specific query blocks where (a) all predicates are empty and (b)
 * the device UDFs are applied in the projection list.
 *
 * We should modify the optimizer (so that device UDFs can be applied in 
 * the predicates of a query block involving several relations)? This implies
 * finding the device UDFs, creating two RelOp nodes out of one query block 
 * (with their respective projection list and predicate lists),
 * setting the dependencies so that the two nodes are correctly connected.
 * 
 * Maybe first look at the rewrite rules to create specific query blocks
 * whenever device UDFs are applied.
 *
 * @param NumChildren the number of children of this query block
 * @param PredPlans a structure containing all predicates in this query block
 * for the moment none of them should contain a device UDF
 * @param ExprPlanList the projection list that may contain Device UDFs
 * @param CatInfo 
 * @author Praveen
 */
static bool IsDeviceUDFNode(int NumChildren,
			    RelSPJOptimizer::PredPlanStruct& PredPlans,
			    XxxValExprPlanList* ExprPlanList,
			    RelCatalogInfo* CatInfo)
{
  if (NumChildren != 1) {
    return false;
  }
  // for now, we expect special preds to be in the SelPredPlans condition list
  // and all other preds to be empty
  if (PredPlans.InitPlan || PredPlans.JoinPredPlans[0] || PredPlans.ResidPredPlans[0] ) {
    return false;
  }
  XxxBool UDFPlan;
  RelDevicePlanOp::IsDeviceUDFPlan(ExprPlanList, PredPlans, UDFPlan);
    
  if (UDFPlan) {
	  return true;
  }	else {
	  return false;
  }
}


XxxErrCode RelSPJNaiveOptimizer::
makePlanOp(RelPlanOp** ChildPlanOps,
	   int NumChildren,
	   RelSPJOptimizer::PredPlanStruct& PredPlans,
	   XxxValExprPlanList* ExprPlanList,
	   RelCatalogInfo* CatInfo,
	   RelOptEngine::JoinMethod JoinMethod,
	   XxxBool ApplyDistinct,
	   RelPlanOp*& PlanOp) const
{
  // added by Tobias for client site UDFs
  XxxBool applicable = XXX_FALSE;
  if(clientSiteExecApplicable(NumChildren,
			      PredPlans,
			      ExprPlanList,
			      CatInfo,
			      applicable ) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, 
		   "Could not determine if client site executable.");
    return XXX_NOT_OK;
  }
  if(applicable) 
    {
      int BatchSize;
      if (XXX_THISRELENV() -> UDFBatchSize )
	BatchSize = XXX_THISRELENV() -> UDFBatchSize;
      else 
	{
	  // OLD: char *BSString = getenv("XXX_BATCH_SIZE");
	  char *BSString = XXX_THISENV().Options->getopt("XXX_BATCH_SIZE");
	  if (BSString)
	    BatchSize = atoi(BSString);
	  else
	    BatchSize = 5;
	}
      PlanOp = new RelClientSitePlanOp( ChildPlanOps,
					NumChildren,
					PredPlans,
					ExprPlanList,
					CatInfo,
					BatchSize );
      if ( ! ((RelClientSitePlanOp*)PlanOp) -> wasInitOK () ) 
	{
	  REL_NOTE_ERROR(REL_BADOPT, 
			 "Cannot construct client site plan operator");
	  return XXX_NOT_OK;
	}
    }
  else if (IsDeviceUDFNode(NumChildren, PredPlans, ExprPlanList, CatInfo)) 
    {
      // HOW TO DETECT DUPLICATE ELIMINATION?
      XXX_CALL_RET(RelDevicePlanOp::MakePlanOp(ChildPlanOps[0],
					       ExprPlanList,
					       PredPlans,
					       ApplyDistinct,
					       CatInfo,
					       PlanOp),
		   XXX_BADOPT, "Error making device operator");
    } // if it is a device UDF
  else 
    {
      switch (JoinMethod) 
	{
	case RelOptEngine::TNLOOPS:
	case RelOptEngine::NDXLOOPS: // put this in here too: run-time index choice
	  // Philippe: creation of a java stream operator for SelPredPlans expressions in
	  // a nested loop. Will be replaced by a tighter integration in the optimizer.
	  if (IsJavaExpr()) 
	    {
	      XXX_CALL_RET(makeJavaExpr(NumChildren, ChildPlanOps, PredPlans),
			   XXX_BADEXEC, "Problem inserting a java node in the expression plan");
	    }
	  PlanOp = new RelTNLPlanOp(NumChildren, ChildPlanOps,
				    CatInfo,
				    ExprPlanList, 
				    PredPlans.SelPredPlans,
				    PredPlans.JoinPredPlans,
				    PredPlans.ResidPredPlans,
				    PredPlans.InitPlan);
	  break;
	case RelOptEngine::PNLOOPS: 
	  PlanOp = new RelPNLPlanOp(NumChildren, ChildPlanOps,
				    CatInfo,
				    ExprPlanList, 
				    PredPlans.SelPredPlans,
				    PredPlans.JoinPredPlans,
				    PredPlans.ResidPredPlans,
				    PredPlans.InitPlan,
				    RelPNLPlanOp::DEFAULT_PNL_BLOCKCARD);
	  break;
	  //Begin: Moustafa
	case RelOptEngine::HBEWINDOW: 
	  PlanOp = new RelHBEWPlanOp(NumChildren, ChildPlanOps,
				     CatInfo,
				     ExprPlanList, 
				     PredPlans.SelPredPlans,
				     PredPlans.JoinPredPlans,
				     PredPlans.ResidPredPlans,
				     PredPlans.InitPlan, 0, 0);
	  break;
	case RelOptEngine::NLBEWINDOW: 
	  PlanOp = new RelNLBEWPlanOp(NumChildren, 
				      ChildPlanOps,	
				      CatInfo, 
				      ExprPlanList, 
				      PredPlans.SelPredPlans, 
				      PredPlans.JoinPredPlans, 
				      PredPlans.ResidPredPlans, 
				      PredPlans.InitPlan, 
				      0, 0);
	  break;
	  
	case RelOptEngine::HFEWINDOW: 
	  PlanOp = new RelHFEWPlanOp(NumChildren, 
				     ChildPlanOps,
				     CatInfo, 
				     ExprPlanList,
				     PredPlans.SelPredPlans,
				     PredPlans.JoinPredPlans,
				     PredPlans.ResidPredPlans, 
				     PredPlans.InitPlan,
				     0, 0);
	  break;
	case RelOptEngine::NLFEWINDOW: 
	  PlanOp = new RelNLFEWPlanOp(NumChildren, 
				      ChildPlanOps,
				      CatInfo, 
				      ExprPlanList,
				      PredPlans.SelPredPlans,
				      PredPlans.JoinPredPlans,
				      PredPlans.ResidPredPlans, 
				      PredPlans.InitPlan,
				      0, 0);
	  break;
	  
	  //End Moustafa
	case RelOptEngine::SORTMERGE: 
	  {
	    // make the SMGlue information
	    SMPlanGlue Glue(1);
	    MakeSMGlue(NumChildren, (const RelPlanOp**)ChildPlanOps,
		       (XxxRelBooleanPlan *)(PredPlans.JoinPredPlans[1]), Glue);
	    if (CreateSMPlanOp(PlanOp, NumChildren, ChildPlanOps,
			       CatInfo,	
			       ExprPlanList, 	
			       PredPlans.SelPredPlans,	
			       PredPlans.JoinPredPlans,
			       PredPlans.ResidPredPlans,
			       PredPlans.InitPlan, Glue) != XXX_OK) 
	      {
		REL_NOTE_ERROR(REL_BADOPT, "Cannot make SM plan operator");
		return XXX_NOT_OK;
	      }
	    break;
	  }
	case RelOptEngine::HASHJOIN: {
		PlanOp = new RelHashJoinPlanOp(NumChildren, ChildPlanOps,
			     CatInfo,
			     ExprPlanList, 
			     PredPlans.SelPredPlans,
			     PredPlans.JoinPredPlans,
			     PredPlans.ResidPredPlans,
			     PredPlans.InitPlan);
	  break;
	}	
	default:
	  // Should never get here
	  REL_NOTE_ERROR(REL_BADOPT, "Invalid join method");
	  return XXX_NOT_OK;
	}
    } // end of else, if not client site execution applicable 
  if ((NumChildren == 1) &&
      (ChildPlanOps[0]->CatalogInfo->ImplInfo()->SortInfo.NumSortOrders)) 
    {
      RelImplInfo* ImplInfo = ChildPlanOps[0]->CatalogInfo->ImplInfo();
      XxxBool* UsefulSortOrder = new XxxBool[ImplInfo->SortInfo.NumSortOrders];
      int* UsefulSortAttrs = new int[ImplInfo->SortInfo.NumSortOrders];
   unsigned int i;
   for (i = 0; i < ImplInfo->SortInfo.NumSortOrders; i++) 
     {
       UsefulSortOrder[i] = XXX_FALSE;
       UsefulSortAttrs[i] = 0;
     }
   int NumUsefulOrders = 0;
   // Any sort orders in the input may carry over to the output.
   // Check each sort order in the input
   for (i = 0; i < ImplInfo->SortInfo.NumSortOrders; i++) 
     {
       if ((UsefulSortAttrs[i] =
	    ImplInfo->SortInfo.SortInfoArray[i].CommonAttrs(ExprPlanList))) 
	 {
	   UsefulSortOrder[i] = XXX_TRUE;
	   NumUsefulOrders++;
	 }
     } // for each input sort order
    if (NumUsefulOrders) 
      {
	PlanOp->CatalogInfo->ImplInfo()->SortInfo.Init(NumUsefulOrders);
	int j = 0;
	for (i = 0; i < ImplInfo->SortInfo.NumSortOrders; i++) 
	  {
	    if (UsefulSortOrder[i]) 
	      {
		XxxSortInfo OutSInfo(UsefulSortAttrs[i]);
		if (ImplInfo->SortInfo.SortInfoArray[i].
		    FillTarget(ExprPlanList, OutSInfo) != XXX_OK) 
		  {
		    REL_NOTE_ERROR(REL_BADEXEC, "Error making sort info");
		    return XXX_NOT_OK;
		  }
		PlanOp->CatalogInfo->ImplInfo()->SortInfo[j].Init(OutSInfo);
		j++;
	      }
	  } // for each input sort order
      } // if there are useful join orders
    delete [] UsefulSortOrder;
    delete [] UsefulSortAttrs;
    } // if there should be an output sort order assigned
  
  // If the selection is distinct, add a distinct planop on top.
  // Note that the distinct requires a sort.
  if (ApplyDistinct) 
    {
      // Create the sort planop below
      XxxSortInfo SInfo(BlockPlan->CatalogInfo->recSchema->Arity());
      int i;
      for (i = 0; i < BlockPlan->CatalogInfo->recSchema->Arity(); i++) 
	{
	  SInfo[i].AttrIndex = i;
	  SInfo[i].Asc = XXX_TRUE;
	}
      // This takes some explaining: if the input is already sorted in
      // the necessary order, of course, just add a Distinct planop.
      // However, if the input is not sorted, add a sort operator.
      // The sort operator takes a "distinct" flag which allows 
      // duplicates to be eliminated early in the sort. So in this
      // case, combine the two planops.
      /**** For some reason, there is a bug in the sort that removes
      *** non-equal values!!!
      **** TODO: find out why and fix it!! likewise in costspjopt.C
      if (!PlanOp->CatalogInfo->ImplInfo()->SortInfo.Matches(SInfo)) 
      {
      // Create the sort + distinct planop 
      PlanOp = new RelSortPlanOp(PlanOp, SInfo, XXX_TRUE, XXX_TRUE,
      XXX_THISENV().DBEnv.BufPoolSize/2,
      BlockPlan->CatalogInfo);
      }
      else 
      {
      // Create the distinct planop 
      PlanOp = new RelDistinctPlanOp(PlanOp, BlockPlan->CatalogInfo);
      }
      *****/
      
      if (!PlanOp->CatalogInfo->ImplInfo()->SortInfo.Matches(SInfo)) {
	PlanOp = new RelSortPlanOp(PlanOp, SInfo, XXX_FALSE, XXX_TRUE,
				   XXX_THISENV().DBEnv.BufPoolSize/2,
				   BlockPlan->CatalogInfo);
      }
      PlanOp = new RelDistinctPlanOp(PlanOp, BlockPlan->CatalogInfo);
      
      
      // make sure to specify the SortInfo in the result
      PlanOp->CatalogInfo->ImplInfo()->SortInfo.Init(1);
      PlanOp->CatalogInfo->ImplInfo()->SortInfo.SortInfoArray[0].Init(SInfo);
    }
  
  return XXX_OK;
}


// determines if this block can be executed on the client side
XxxErrCode RelSPJNaiveOptimizer::
clientSiteExecApplicable ( int NumChildren,
			   RelSPJOptimizer::PredPlanStruct& PredPlans,
			   XxxValExprPlanList* ExprPlanList,
			   RelCatalogInfo* CatInfo,
			   XxxBool &Result) const
{
  Result = XXX_TRUE;
  XxxBool AtLeastOne = XXX_FALSE;
  if ( NumChildren != 1 ) {
    Result = XXX_FALSE;
    return XXX_OK;
  }
  
  /* PredPlanStruct (planspj.h:37) works, according to Praveen, 
     like this: 
     SelPredPlans holds the predicates to be applied to each input
     JoinPredPlans holds the predicates for each join
     InitPlan is any constant predicate
     ResidPredPlans holds predicates applied after each join...
     */
  if ( PredPlans . InitPlan ||
       PredPlans . JoinPredPlans[0] ||
       PredPlans . ResidPredPlans[0] ) {
    Result = XXX_FALSE; // none of this should exist
    return XXX_OK;
  }
  if ( PredPlans . SelPredPlans[0] ) {
    // only one child, so we only have to look at SelPredPlans[0]

      // if we do a client-site join, we have to allow pushable 
      // predicates
      if(XXX_THISRELENV() -> UDFExec & 4) {

          const XxxValueExpression *SourceExpression =
            PredPlans . SelPredPlans[0] -> SourceExpr();
          ASSERT (( SourceExpression -> ValType() == 
                    XxxValueExpression :: BOOL_EXPR ));
          if ( ((XxxBooleanExpression *) SourceExpression) 
              -> BoolType == XXX_VAL_BOOL_EXPR ) {
              clientSiteExecApplicable(SourceExpression ->
                                       ChildExpr(0), Result,
                                       AtLeastOne);
              if ( ! Result ) {
                if ( AtLeastOne) {
                    REL_NOTE_ERROR(REL_BADOPT, "Client-site operations cannot be mixed with server operations.");
                    return XXX_NOT_OK;
                }
                return XXX_OK;
              }
           } else {
              ASSERT (((XxxBooleanExpression *) SourceExpression) 
                  -> BoolType == XXX_REL_BOOL_EXPR );
              XxxRelOpType PredicateOpType = 
                  ((XxxRelBooleanExpression *) SourceExpression) 
                  -> OpType;
              ASSERT ((PredicateOpType < XXX_NEQ_OP));
                // only gt,lt,geq,leq,eq, and not neq, like, other
              // Now check both children
              clientSiteExecApplicable(SourceExpression ->
                                       ChildExpr(0), Result,
                                       AtLeastOne);
              if ( ! Result ) {
                if ( AtLeastOne ) {
                    REL_NOTE_ERROR(REL_BADOPT, "Client-site operations cannot be mixed with server operations.");
                    return XXX_NOT_OK;
                }
                return XXX_OK;
              }
              clientSiteExecApplicable(SourceExpression ->
                                       ChildExpr(1), Result,
                                       AtLeastOne);
              if ( ! Result ) {
                if ( AtLeastOne ) {
                    REL_NOTE_ERROR(REL_BADOPT, "Client-site operations cannot be mixed with server operations.");
                    return XXX_NOT_OK;
                }
                return XXX_OK;
              }
           }
      } else { // XXX_THISRELENV() -> UDFExec & 4

    clientSiteExecApplicable 
      ( PredPlans . SelPredPlans[0] -> SourceExpr() ,
	Result, AtLeastOne );
    if ( ! Result ) {
        if ( AtLeastOne ) {
        REL_NOTE_ERROR(REL_BADOPT, "Client-site operations cannot be mixed with server operations.");
        return XXX_NOT_OK;
      }
      return XXX_OK;
    }
  } } 
  XxxValExprPlanList *tmp = ExprPlanList;
  while (tmp) {
    clientSiteExecApplicable 
      ( tmp -> Plan -> SourceExpr() ,
	Result , AtLeastOne );
    if ( ! Result ) {
      if ( AtLeastOne ) {
        REL_NOTE_ERROR(REL_BADOPT, "Client-site operations cannot be mixed with server operations.");
        return XXX_NOT_OK;
      }
      return XXX_OK;
    }
    tmp = tmp -> Next;
  }
  if ( ! AtLeastOne )
    Result = false;
  else if ( ! Result ) {
     REL_NOTE_ERROR(REL_BADOPT, "Client-site operations cannot be mixed with server operations.");
     return XXX_NOT_OK;

  }
  return XXX_OK;
}

XxxErrCode RelSPJNaiveOptimizer::
clientSiteExecApplicable ( const XxxValueExpression* Expression,
			   XxxBool &Result ,
			   XxxBool &AtLeastOne ) const
{
  Result = XXX_TRUE;
  //TODO: check if there is a better way of fixing this
  if( ! Expression ) {
    Result = XXX_FALSE;
    return XXX_OK;
  }

  XxxValueExprList* Expressions = 
    Expression -> SubExprsReferenced();
  XxxValueExprList* ListWeHaveToDelete = Expressions;

  while ( Expressions && Result ) {
    // go through the list of subexpressions 
    // no need to go on if one is not client site executable
    XxxValueExpression *expr =
      Expressions -> Expr;
    // here we have to decide what expressions are executed on
    // the client, see list in utils/expr.h:140
    if ( expr -> ValType() == XxxValueExpression :: FUNC_EXPR ) {
      Result =
	    XxxADTArray[((XxxFuncExpression*)expr) -> OwnerType] ->
	    GetMethByIndex(((XxxFuncExpression*)expr) -> 
		         FuncParse -> MethodIndex ) ->
	             IsClientSiteExecutable () ;
      if ( ! Result ) 
	    break;
      else
        AtLeastOne = XXX_FALSE; 
    }
    else if ( expr -> ValType() == 
	    XxxValueExpression :: CONST_EXPR )  ; // do nothing
	else if ( expr -> ValType() == 
		XxxValueExpression :: COLUMN_VAR )  ; // do nothing
    else if ( expr -> ValType() ==
	       XxxValueExpression :: BOOL_EXPR ) {
       if ( ((XxxBooleanExpression *) expr) -> BoolType
	    != XXX_VAL_BOOL_EXPR ) {
	     Result = XXX_FALSE;
	     break;
       } }
    else {
      Result = XXX_FALSE; 
      break;
    }       
    Expressions = Expressions -> Next;
  }
  if ( ListWeHaveToDelete )
    ListWeHaveToDelete -> Destroy();    
  return XXX_OK;
}
  
void RelSPJNaiveOptimizer::printPlan(ostream& OStr, int NumRels,
				     int* JoinOrder,
				     RelOptEngine::JoinMethod* JoinMethods,
				     const RelOptEngine::Cost& Cost) const
{
  for (int i = 0; i < NumRels; i++) {
    if (i > 0) {
      switch (JoinMethods[i-1]) {
      case RelOptEngine::SORTMERGE:
	OStr << " SM "; break;
      case RelOptEngine::PNLOOPS:
	OStr << " PNL "; break;
      case RelOptEngine::TNLOOPS:
	OStr << " TNL "; break;
      case RelOptEngine::NDXLOOPS:
	OStr << " INL "; break;
      default:
	OStr << " BAD "; break;
      } // switch
    } // if
    OStr << JoinOrder[i];
  } // for each child
  
  OStr << endl;
  Cost.Print(OStr);
}


/**
* Inserts a JavaExprPlan that wraps every PredPlans.SelPredPlans.
* There is thus only one input schema for each expression and 
* the result type is a boolean.
* The target is a remote java exec engine.
*
*/
XxxErrCode
makeJavaExpr(int NumChildren, RelPlanOp** ChildPlanOp, RelSPJOptimizer::PredPlanStruct& PredPlans)
{
	// initialize java target: remote java exec engine
	XxxJavaTarget *JavaTarget = new XxxJavaTarget();
	JavaTarget->setTargetRemote();

	for (int i=0; i<NumChildren; i++) {
		XxxValueExprPlan* BoolExpr[1];

		if (!PredPlans.SelPredPlans[i]) {
			// no condition for child i
		} else {

			if (PredPlans.SelPredPlans[i]->BoolType == XXX_BOOL_BOOL_EXPR) {
				BoolExpr[0]	= (XxxValueExprPlan*) ((XxxBoolBooleanPlan*) PredPlans.SelPredPlans[i])->Replicate();
			} else if (PredPlans.SelPredPlans[i]->BoolType == XXX_REL_BOOL_EXPR) {
				BoolExpr[0]	= (XxxValueExprPlan*) ((XxxRelBooleanPlan*) PredPlans.SelPredPlans[i])->Replicate();
			} else if (PredPlans.SelPredPlans[i]->BoolType == XXX_VAL_BOOL_EXPR) {
				BoolExpr[0]	= (XxxValueExprPlan*) ((XxxValBooleanPlan*) PredPlans.SelPredPlans[i])->Replicate();
			}

			XxxRecordSchema *BoolSchema[1];
			BoolSchema[0] = ChildPlanOp[i]->CatalogInfo->RecordSchema();

			PredPlans.SelPredPlans[i] = new XxxValBooleanPlan(new XxxValBooleanExpression(new XxxJavaExpression()), 
															  new XxxJavaExprPlan(JavaTarget, BoolExpr, BoolSchema,1,
																				  XXX_BOOLEAN, NULL));
		}
	}

	return XXX_OK;
}

XxxBool IsJavaExpr()
{
	// old: char *JavaExpr = getenv("XXX_JAVAEXPR");
	char *JavaExpr = XXX_THISENV().Options->getopt("XXX_JAVAEXPR");
	if (JavaExpr)
		return XXX_TRUE;
	else 
		return XXX_FALSE;
}
