#ifndef XXX_NAIVESPJ_H
#define XXX_NAIVESPJ_H

#include "planspj.h"

/** 
 * The most naive SPJ optimizer.
 */
class RelSPJNaiveOptimizer : public RelSPJOptimizer {
 public:
  /**
   // This plan maintains the order in which the relations
   // are to be joined and the predicates to be applied at
   // each stage.
   */
  class SPJNaivePlan : public RelOptEngine::PlanObject { public:

    int ArraySize;
    // join ordering effected by using one level of indirection
    int *JoinOrder;
    // join method array: size = 1-NumChildren
    RelOptEngine::JoinMethod *JoinMethods;

    // The size of each array == NumChildren.
    // array of selection predicate bitmaps
    XxxBitMap *SelPreds;
    // array of join predicate bitmaps: these are important 
    // for SM joins (and hash-joins when we implement them) 
    // -- for NL joins, this combines with the Residual preds
    XxxBitMap *JoinPreds;
    // Residual predicates that need to be applied after each join.
    XxxBitMap *ResidPreds;

    // Initial predicates that should be applied before any tables 
    // are accessed. Usually, this should be empty unless there 
    // are preds of the form 3 = 2 + 1
    // XxxBooleanExpressionPlan* InitPlan;
    XxxBitMap InitPreds;

    SPJNaivePlan(XxxErrCode& RC,
		 int PlanIdent,
		 unsigned char BlockIdent,
		 int NumChildren,
		 RelSPJNaiveOptimizer* Opt);
    virtual ~SPJNaivePlan();
    virtual void Print(ostream& OStr) const;
   };

  RelSPJNaiveOptimizer(XxxErrCode& RC,
		       XxxValueExprList* TargetList,
		       XxxBooleanExpression* Condition,
		       XxxBool Unique,
		       RelSPJBlockPlan* SPJPlan,
		       RelOptEngine::OptQueryEnv* QEnv);
  virtual ~RelSPJNaiveOptimizer();

  virtual XxxErrCode Optimize();
  virtual XxxErrCode GeneratePlanOp(RelOptEngine::PlanObject* Plan,
				    RelPlanOp*& PlanOp) const;
  virtual void PrintPlans(const RelOptEngine::PlanObject* Plan,
			  ostream& OStr) const;
  /**
   * Determines the glue information to decide what sorts are needed
   * on the inputs for a SM join
  **/
  static void MakeSMGlue(int NumChildren,
			 const RelPlanOp** Children,
			 const XxxRelBooleanPlan *JoinPlan, 
			 SMPlanGlue& Glue);
  /**
   * Creates a Sort-merge Join planOp, placing sort planOps below
   * if incoming tuple streams are not already sorted.
   */
  static XxxErrCode CreateSMPlanOp(RelPlanOp*& PlanOp, 
				   int NumChildren,
				   RelPlanOp* Children[],
				   RelCatalogInfo* CatInfo,
				   XxxValExprPlanList* ExprPlanList,
				   XxxBooleanExpressionPlan** SelPredPlans,
				   XxxBooleanExpressionPlan** JoinPredPlans,
				   XxxBooleanExpressionPlan** ResidPredPlans,
				   XxxBooleanExpressionPlan* InitPlan,
				   const SMPlanGlue& Glue);
 protected:
  // inside this complex plan, is there one that can be used for sort-merge?
  XxxErrCode isSMJoinable (XxxBooleanExpressionPlan* JoinPlan,
			   XxxRelBooleanPlan*& SMPlan,
			   XxxBooleanExpressionPlan*& ResidualPlan,
			   XxxBool& Success) const;

  // is this a simple plan that can be used for a sort-merge join?
  XxxErrCode isSMJoinable (XxxRelBooleanPlan* JoinPlan,
			   XxxBool& Success) const;

  // This virtual method simply provides some permutation
  // of the tables as a join order.
  // Therefore, one can subclass this and simply redefine
  // this one method to have another naive optimizer
  // that uses a different join order heuristic.
  //Begin Moustafa
  // Add HasStreamInput parameter
  //  virtual XxxErrCode determineJoinStrategy
  //    (int NumChildren, int* JoinOrder,
  //     RelOptEngine::JoinMethod* JoinMethods) const;

  virtual XxxErrCode determineJoinStrategy
    (int NumChildren, int* JoinOrder,
     RelOptEngine::JoinMethod* JoinMethods, XxxBool HasStreamInput) const;
  //End Moustafa

 protected:
  //// Get plans for children of this block
  XxxErrCode getChildPlans(SPJNaivePlan* SPJPlan,
			   RelPlanOp** ChildPlanOps,
			   int* InputArity) const;
  //// Create the preds to be applied during the joins
  XxxErrCode createPreds(SPJNaivePlan* SPJPlan,
			 RelSPJOptimizer::PredPlanStruct& PredPlans) const;
  //// Get the InitPlan -- used by createPreds
  XxxErrCode getInitPlan(SPJNaivePlan* SPJPlan,
			 XxxBooleanExpressionPlan*& InitPlan) const;
  //// Get the preds for the joins -- used by createPreds
  XxxErrCode getJoinPlans(SPJNaivePlan* SPJPlan,
			  RelSPJOptimizer::PredPlanStruct& PredPlans,
			  int NumChildren,
			  int JoinOffset,
			  int SkipTemporary) const;

  //// Construct a remapping
  XxxErrCode constructRemap(SPJNaivePlan* SPJPlan,
			    XxxExprPlan::VarRemapping& Remap,
			    int NewArity,
			    int *InputArity,
			    XxxBool MapAttrs) const;
  //// Modify all the expressions using the VarRemapping
  XxxErrCode exprRemap(XxxExprPlan::VarRemapping& Remap,
		       RelSPJOptimizer::PredPlanStruct& PredPlans,
		       int NumChildren,
		       XxxValExprPlanList* ExprPlanList) const;
  

  XxxErrCode makePlanOp(RelPlanOp** ChildPlanOps,
			int NumChildren,
			RelSPJOptimizer::PredPlanStruct& PredPlans,
			XxxValExprPlanList* ExprPlanList,
			RelCatalogInfo* CatInfo,
			RelOptEngine::JoinMethod JoinMethod,
			XxxBool ApplyDistinct,
			RelPlanOp*& PlanOp) const;

  // used by MakePlanOp to decide if client side operator 
  // should be generated (Tobias)
  XxxErrCode clientSiteExecApplicable 
  ( int NumChildren,
    RelSPJOptimizer::PredPlanStruct& PredPlans,
    XxxValExprPlanList* ExprPlanList,
    RelCatalogInfo* CatInfo,
    XxxBool &Result) const;
  
  XxxErrCode clientSiteExecApplicable 
  ( const XxxValueExpression* Expression,
    XxxBool &Result,
    XxxBool &AtLeastOne ) const;

  void printPlan(ostream& OStr, int NumRels,
		 int* JoinOrder,
		 RelOptEngine::JoinMethod* JoinMethods,
		 const RelOptEngine::Cost& Cost) const;
 };

#endif // XXX_NAIVESPJ_H
