//
// TODO: fix all the cost formulas
//
    
#include <iostream.h>
    
#include "relerror.h"
#include "relquery.h"
#include "dynspj.h"
#include "relplanop.h"
    
//
// Methods of RelSPJCostBasedOpt::JoinPhysPlan
//
RelSPJCostBasedOpt::
JoinPhysPlan::JoinPhysPlan(int PlanIdent,
    			   unsigned char BlockIdent,
    			   const RelOptEngine::LogProperties& LProps,
				   const RelOptEngine::PhysProperties& PProps,
    			   const PhysicalPlan* Plan0,
    			   const PhysicalPlan* Plan1,
    			   const XxxBitMap& NPreds,
    			   RelOptEngine::JoinMethod JMeth)
  : PhysicalPlan (PlanIdent, BlockIdent, LProps,PProps, NPreds),
    JoinPreds(NPreds), ResidPreds(NPreds.ArraySize)
{
  // initially, assume JoinPreds == NewPreds, and ResidPreds is null
  OuterPlan = Plan0;
  InnerPlan = Plan1;
  Method = JMeth;
}
    
XxxErrCode RelSPJCostBasedOpt::
JoinPhysPlan::getCatInfo(const RelSPJOptimizer* Opt,
    			 RelPlanOp** TmpPlanOps,
    			 XxxBool TopLevel,
    			 RelCatalogInfo*& CatInfo) const
{
  if (TopLevel) {
    CatInfo = Opt->BlockPlan->CatalogInfo;
  }
  else {
    CatInfo = new RelCatalogInfo();
    // create the schema
    const XxxRecordSchema* RecSchema[2];
    RecSchema[0] = TmpPlanOps[0]->CatalogInfo->RecordSchema();
    RecSchema[1] = TmpPlanOps[1]->CatalogInfo->RecordSchema();
    CatInfo->recSchema = new XxxRecordSchema(2, RecSchema);
    CatInfo->implInfo = new RelImplInfo();
    CatInfo->statsInfo = new RelStatsInfo();
  }
  return XXX_OK;
}
XxxErrCode RelSPJCostBasedOpt::
JoinPhysPlan::getProjInfo(const RelSPJOptimizer* Opt,
    			  RelPlanOp** TmpPlanOps,
    			  XxxBool TopLevel,
    			  const XxxExprPlan::VarRemapping& Remap,
    			  XxxValExprPlanList*& TargetList) const
{
  if (TopLevel) {
    TargetList = Opt->ExprPlanList;
    // modify the mapping in the targetlist
    if (TargetList &&
	TargetList->RedirectDependencies(Remap) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADOPT, "Error redirecting target");
      return XXX_NOT_OK;
    }
  }
  else {
    //  new targetlist is the concatenation of the record schemas
    TargetList = NULL;
    // Two steps: first create an XxxValueExprList, then 
    // optimize it to get an XxxValExprPlanList
       
    XxxValueExprList *TmpExprList = NULL;
    for (int ci =0; ci < 2; ci++) {
      const XxxRecordSchema* ChildSchema = 
	TmpPlanOps[ci]->CatalogInfo->RecordSchema();
      for (int j = 0; j < ChildSchema->Arity(); j++) {
	XxxValueExpression* NewExpr = 
	  new XxxUnknownValue(j, 
			      ChildSchema->GetAttr(j)->Type,
			      ChildSchema->GetAttr(j)->TypeCatInfo);
	((XxxUnknownValue *)NewExpr)->SrcChild = ci;
	if (!TmpExprList)
	  TmpExprList = new XxxValueExprList(NewExpr);
	else 
	  TmpExprList = TmpExprList->AddExpr(NewExpr);
      } // for each attribute
    } // for both children
    // now optimize TmpExprList to create plan list
    if (TmpExprList && 
	(TmpExprList->Optimize(TargetList) != XXX_OK)) {
      REL_NOTE_ERROR(REL_BADOPT, "Error optimizing tmp expr list");
      return XXX_NOT_OK;
    }
  } // if not TopLevel
  return XXX_OK;
}
XxxErrCode RelSPJCostBasedOpt::
JoinPhysPlan::GeneratePlanOp(const RelSPJOptimizer* Opt,
    			     XxxBool TopLevel,
    			     XxxExprPlan::VarRemapping& Remap,
    			     int InitRemapSrcNum, // ignored
    			     int FinalRemapSrcNum,
    			     RelPlanOp*& PlanOp) const
{
  RelPlanOp* TmpPlanOps[2];
  // consider the difference between sort-merge and NL
  // For sort-merge, both outer and inner need to be treated
  // as 0 for SelPreds, but not for join preds...
  int PreMapPosition; // remapping before the relation is accessed
  int PostMapPosition; // remapping after the relation is accessed
  getRemap(0, PreMapPosition, PostMapPosition);
  if (OuterPlan->GeneratePlanOp(Opt, XXX_FALSE, Remap, 
				PreMapPosition, PostMapPosition, 
				TmpPlanOps[0]) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, "Outer of TNL join");
    return XXX_NOT_OK;
  }
  getRemap(1, PreMapPosition, PostMapPosition);
  if (InnerPlan->GeneratePlanOp(Opt, XXX_FALSE, Remap, 
				PreMapPosition, PostMapPosition, 
				TmpPlanOps[1]) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, "Inner of TNL join");
    return XXX_NOT_OK;
  }
    
  XxxBooleanExpressionPlan** SelPredPlans, **JoinPredPlans, **ResidPredPlans;
  SelPredPlans = new XxxBooleanExpressionPlan*[2];
  JoinPredPlans = new XxxBooleanExpressionPlan*[2];
  ResidPredPlans = new XxxBooleanExpressionPlan*[2];
    
  // for each input
  for (int j=0; j < 2; j++) {
    SelPredPlans[j] = NULL;
    JoinPredPlans[j] = NULL;
    ResidPredPlans[j] = NULL;
    
    // no preds to be applied to intermediate results
    if (j == 0) continue;
    
    for (int k=0; k < JoinPreds.ArraySize; k++) {
      // currently, ignore SelPredPlans, because any selection predicates
      // should have been already absorbed in the children plan ops.
      if (JoinPreds.GetBit(k)) {
	if (XxxBooleanExpressionPlan::MakeConjunction
	    (JoinPredPlans[j], Opt->PredArray[k]) != XXX_OK) {
	  REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
	  return XXX_NOT_OK;
    	}
      } // if
      else if (ResidPreds.GetBit(k)) {
	if (XxxBooleanExpressionPlan::MakeConjunction
	    (ResidPredPlans[j], Opt->PredArray[k]) != XXX_OK) {
	  REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
	  return XXX_NOT_OK;
    	}
      } // if
    } // for each predicate
    if (JoinPredPlans[j] &&
	JoinPredPlans[j]->RedirectDependencies(Remap) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADOPT, "Error redirecting join pred");
      return XXX_NOT_OK;
    }
    if (ResidPredPlans[j] &&
	ResidPredPlans[j]->RedirectDependencies(Remap) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADOPT, "Error redirecting resid pred");
      return XXX_NOT_OK;
    }   
  } // for each input
    
  RelCatalogInfo* CatInfo;
  XxxValExprPlanList* TargetList;
    
  if (getCatInfo(Opt, TmpPlanOps, TopLevel, CatInfo) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, "Error getting CatInfo");
    return XXX_NOT_OK;
  }
  if (getProjInfo(Opt, TmpPlanOps, TopLevel, Remap, TargetList) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, "Error getting CatInfo");
    return XXX_NOT_OK;
  }
    
  PlanOp = makePlanOp(2, TmpPlanOps, CatInfo,
		      TargetList, SelPredPlans,
		      JoinPredPlans, ResidPredPlans,
		      NULL);
    
  // now adjust the Remap to account for this plan
  for (int k=0; k < Remap.OldArity; k++) {
    int SrcNum = Remap.SrcMappingArray[k].NewSrc;
    if ((SrcNum == 0) || (SrcNum == 1)) {
      Remap.SrcMappingArray[k].NewSrc = FinalRemapSrcNum;
      if (SrcNum == 1) {
	int Child1Arity = 
	  TmpPlanOps[1]->CatalogInfo->RecordSchema()->Arity();
	int Child0Arity = 
	  TmpPlanOps[0]->CatalogInfo->RecordSchema()->Arity();
	for (int j = 0; j < Child1Arity; j++)
	  Remap.SrcMappingArray[k].NewAttrMapping[j] += Child0Arity;
      } // if SrcNum == i+1
    } // if (SrcNum == 0 or 1)
  } // for k < Remap.OldArity
    		     
  return XXX_OK;
}
    
void RelSPJCostBasedOpt::
JoinPhysPlan::Print(ostream& OStr) const
{
  OStr << "Join: (" ;
  OStr << "[" << RelOptEngine::JoinMethodNames[Method] << "]";
  OStr << " JPreds:";
  JoinPreds.Print(OStr);
  OStr << " RPreds:";
  ResidPreds.Print(OStr);
  OStr << " Usage: ";
  Usage.Print(OStr);
  OStr << "[";
  OuterPlan->Print(OStr);
  OStr << " --> ";
  InnerPlan->Print(OStr);
  OStr << "]";
  printThis(OStr);
}
    
void RelSPJCostBasedOpt::
PhysicalPlan::TreePrint(ofstream &OStr, int &curPlanId, int level) const
{
}
  
// print out a relation physplan
void RelSPJCostBasedOpt::
RelIScanPlan::TreePrint(ofstream &OStr, int &curPlanId, int level) const
{
  OStr << "NODE " << " Index Scan of Rel" << (int)RelId << " (" << Usage.FullCost() << ")" << endl;
  OStr << "ID " << curPlanId << endl;
  OStr << "LEVEL " << level << endl;
  OStr << "COLOR " << 0xFF0000 << endl;
//  OStr << "INFO Preds:";
//  NewPreds.Print(OStr);
//  OStr << " Cost: " << Usage.FullCost() << endl;
  Usage.Print(OStr);
  OStr << "ENDNODE" << endl;
  curPlanId++;
}
  
void RelSPJCostBasedOpt::
RelFScanPlan::TreePrint(ofstream &OStr, int &curPlanId, int level) const
{
  OStr << "NODE " << " File Scan of Rel" << (int)RelId << " (" << Usage.FullCost() << ")" << endl;
  OStr << "ID " << curPlanId << endl;
  OStr << "LEVEL " << level << endl;
  OStr << "COLOR " << 0xFF0000 << endl;
//  OStr << "INFO Preds:";
//  NewPreds.Print(OStr);
//  OStr << " Cost: " << Usage.FullCost() << endl;
  Usage.Print(OStr);
  OStr << "ENDNODE" << endl;
  curPlanId++;
}

//Begin:Moustafa
// To print node information for Stream Scan 
void RelSPJCostBasedOpt::
RelSScanPlan::TreePrint(ofstream &OStr, int &curPlanId, int level) const
{
  OStr << "NODE " << " Stream Scan of Rel" << (int)RelId << " (" << Usage.FullCost() << ")" << endl;
  OStr << "ID " << curPlanId << endl;
  OStr << "LEVEL " << level << endl;
  OStr << "COLOR " << 0xFF0000 << endl;
//  OStr << "INFO Preds:";
//  NewPreds.Print(OStr);
//  OStr << " Cost: " << Usage.FullCost() << endl;
  Usage.Print(OStr);
  OStr << "ENDNODE" << endl;
  curPlanId++;
}

//End:Moustafa

// Print out a subtree, with assigned cur plan Id, level
void RelSPJCostBasedOpt::JoinPhysPlan::
TreePrint(ofstream &OStr, int &curPlanId, int level) const
{
  int PlanId0, PlanId1;
  PlanId0 = curPlanId +1;
  OuterPlan->TreePrint(OStr, PlanId0, level+1); // PlanId0 is now next available Id
  PlanId1 = PlanId0;
  InnerPlan->TreePrint(OStr, PlanId1, level+1); // PlanId1 is now next available Id
  // Print current node
  OStr << "NODE " << RelOptEngine::JoinMethodNames[Method] << " SPJ (" << Usage.FullCost() << ")" << endl;
  OStr << "ID " << curPlanId << endl;
  OStr << "LEVEL " << level << endl;
  OStr << "CHILDREN " << curPlanId + 1 << " " << PlanId0 << endl; 
  OStr << "COLOR " << 0x00FF00 << endl;
//  OStr << "INFO JPreds:";
//  JoinPreds.Print(OStr);
//  OStr << endl;
//  OStr << "INFO RPreds:";
//  ResidPreds.Print(OStr);
//  OStr << endl;
//  OStr << "INFO Usage: ";
  Usage.Print(OStr);
//  OStr << endl;
//  OStr << "INFO Cost: " << Usage.FullCost();
//  OStr << endl;
  OStr << "INFO ";
  printThis(OStr);
  OStr << endl;
  OStr << "ENDNODE" << endl;
  curPlanId = PlanId1;
}
    
void RelSPJCostBasedOpt::PhysicalPlan::PrintTree(ofstream& OStr) const
{
}

void RelSPJCostBasedOpt::RelPhysPlan::
TreePrint(ofstream &OStr, int &curPlanId, int level) const
{
}

// Add by zhiyuan chen, print out physical scan in tree model
void RelSPJCostBasedOpt::JoinPhysPlan::PrintTree(ofstream& OStr) const
{
  int curPlanId = 0, level = 0;
  OStr << "TREE Plan " << curPlanId << " Level " << level << endl;
  TreePrint(OStr, curPlanId, level);
  OStr << "ENDTREE" << endl;
}

//
// Methods of RelSPJCostBasedOpt::TNLJoinPlan
//
void RelSPJCostBasedOpt::
TNLJoinPlan::planStats(const XxxDBEnv& Env,
    		       const RelStatsInfo* InnerStatsInfo)
{
  int TableCard;
  int AvgTupleSize;
  if (InnerStatsInfo) {
    TableCard = InnerStatsInfo->Cardinality;
    AvgTupleSize = InnerStatsInfo->AvgTupleSize;
  }
  else {
    TableCard = RelStatsInfo::DefaultTableCard;
    AvgTupleSize = RelStatsInfo::DefaultTupleSize;
  }
    
  CostTNLJoin(OuterPlan->LogProps,
    	      OuterPlan->PhysProps,
    	      OuterPlan->Usage,
    	      TableCard, AvgTupleSize,
    	      Env, Usage);
}
RelPlanOp* RelSPJCostBasedOpt::
TNLJoinPlan::makePlanOp(int NumChildren, 
    			RelPlanOp* Children[],
    			RelCatalogInfo* CatInfo,
    			XxxValExprPlanList* ExprPlanList,
    			XxxBooleanExpressionPlan** SelPredPlans,
    			XxxBooleanExpressionPlan** JoinPredPlans,
    			XxxBooleanExpressionPlan** ResidPredPlans,
    			XxxBooleanExpressionPlan* InitPlan) const
{
  return new RelTNLPlanOp(NumChildren, Children,
			  CatInfo, ExprPlanList,
			  SelPredPlans,
			  JoinPredPlans,
			  ResidPredPlans, 
			  InitPlan);
}
    
//
// Methods of RelSPJCostBasedOpt::PNLJoinPlan
//
void RelSPJCostBasedOpt::
PNLJoinPlan::planStats(const XxxDBEnv& Env,
    		       const RelStatsInfo* InnerStatsInfo)
{
  int TableCard;
  int AvgTupleSize;
  if (InnerStatsInfo) {
    TableCard = InnerStatsInfo->Cardinality;
    AvgTupleSize = InnerStatsInfo->AvgTupleSize;
  }
  else {
    TableCard = RelStatsInfo::DefaultTableCard;
    AvgTupleSize = RelStatsInfo::DefaultTupleSize;
  }
    
  CostPNLJoin(OuterPlan->LogProps,
	      OuterPlan->PhysProps,
	      OuterPlan->Usage,
	      TableCard, AvgTupleSize,
	      Env, Usage, TuplesPerBlock);
    
}
void RelSPJCostBasedOpt::
PNLJoinPlan::printThis(ostream& OStr) const
{
  OStr << "Block Size = " << TuplesPerBlock << "tuples";
}

RelPlanOp* RelSPJCostBasedOpt::
PNLJoinPlan::makePlanOp(int NumChildren, 
    			RelPlanOp* Children[],
    			RelCatalogInfo* CatInfo,
    			XxxValExprPlanList* ExprPlanList,
    			XxxBooleanExpressionPlan** SelPredPlans,
    			XxxBooleanExpressionPlan** JoinPredPlans,
    			XxxBooleanExpressionPlan** ResidPredPlans,
    			XxxBooleanExpressionPlan* InitPlan) const
{
  return new RelPNLPlanOp(NumChildren, Children,
			  CatInfo, ExprPlanList,
			  SelPredPlans,
			  JoinPredPlans,
			  ResidPredPlans, 
			  InitPlan,
			  TuplesPerBlock);
}
//
// Methods of RelSPJCostBasedOpt::INLJoinPlan
//
void RelSPJCostBasedOpt::
INLJoinPlan::planStats(const XxxDBEnv& Env,
    		       const RelStatsInfo* InnerStatsInfo)
{
  int TableCard;
  int AvgTupleSize;
  if (InnerStatsInfo) {
    TableCard = InnerStatsInfo->Cardinality;
    AvgTupleSize = InnerStatsInfo->AvgTupleSize;
  }
  else {
    TableCard = RelStatsInfo::DefaultTableCard;
    AvgTupleSize = RelStatsInfo::DefaultTupleSize;
  }
    
  RelIScanPlan* IndxPlan = (RelIScanPlan*)InnerPlan;
  // TODO
  CostINLJoin(OuterPlan->LogProps,
	      OuterPlan->PhysProps,
	      OuterPlan->Usage,
	      TableCard, AvgTupleSize,
	      /* IndxInfo */ NULL, IndxPlan->MatchedInfo,
	      // IsEqualityPred, Selectivity,
	      XXX_TRUE, 0.1,
	      Env, Usage);
}
void RelSPJCostBasedOpt::
INLJoinPlan::printThis(ostream& OStr) const
{
  // TODO:LOpt
}
    
//
// Methods of RelSPJCostBasedOpt::SMJoinPlan
//
void RelSPJCostBasedOpt::
SMJoinPlan::planStats(const XxxDBEnv& Env,
    		      const RelStatsInfo* InnerStatsInfo)
{
  int TableCard;
  int AvgTupleSize;
  if (InnerStatsInfo) {
    TableCard = InnerStatsInfo->Cardinality;
    AvgTupleSize = InnerStatsInfo->AvgTupleSize;
  }
  else {
    TableCard = RelStatsInfo::DefaultTableCard;
    AvgTupleSize = RelStatsInfo::DefaultTupleSize;
  }
  
  CostSMJoin(OuterPlan->LogProps, OuterPlan->PhysProps,
	     OuterPlan->Usage, ToSortPlan0,
	     TableCard, AvgTupleSize, ToSortPlan1,
	     Env, Usage, SortRunLength0, SortRunLength1);
}
    
void RelSPJCostBasedOpt::
SMJoinPlan::printThis(ostream& OStr) const
{
 
  OStr << "INFO Sort0:" << (ToSortPlan0?"yes":"no");
  OStr << " Sort1:" << (ToSortPlan0?"yes":"no");
}
    
RelPlanOp* RelSPJCostBasedOpt::
SMJoinPlan::makePlanOp(int NumChildren, 
    		       RelPlanOp* Children[],
    		       RelCatalogInfo* CatInfo,
    		       XxxValExprPlanList* ExprPlanList,
    		       XxxBooleanExpressionPlan** SelPredPlans,
    		       XxxBooleanExpressionPlan** JoinPredPlans,
    		       XxxBooleanExpressionPlan** ResidPredPlans,
    		       XxxBooleanExpressionPlan* InitPlan) const
{
  RelPlanOp* PlanOp;
  // for now, do a run-time dynamic check of the sort order.
    
  // make the SMGlue information (used in sort order determination)
  RelSPJCostBasedOpt::SMPlanGlue Glue(1);

  RelSPJCostBasedOpt::MakeSMGlue(NumChildren, (const RelPlanOp**)Children,
    				 (XxxRelBooleanPlan *)(JoinPredPlans[1]), Glue);
  if (RelSPJCostBasedOpt::CreateSMPlanOp(PlanOp, NumChildren, Children,
    					 CatInfo,	
    					 ExprPlanList, 	
    					 SelPredPlans,	
    					 JoinPredPlans,
    					 ResidPredPlans,
    					 InitPlan, Glue) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, "Cannot make SM plan operator");
    return NULL;
  }
  return PlanOp;
}
    
//
// Methods of RelSPJCostBasedOpt::RelPhysPlan
//
void RelSPJCostBasedOpt::
RelPhysPlan::Print(ostream& OStr) const
{
  OStr << "Rel : " << (int)RelId;
  OStr << " Preds:";
  NewPreds.Print(OStr);
  OStr << " Cost: " << Usage.FullCost();
}
    
//
// Methods of RelSPJCostBasedOpt::RelFScanPlan
//
RelSPJCostBasedOpt::
RelFScanPlan::RelFScanPlan(int PlanIdent,
    			   unsigned char BlockIdent,
    			   const RelOptEngine::LogProperties& LProps,
    			   unsigned char RelationId,
    			   const XxxBitMap& NPreds,
    			   const XxxDBEnv& Env,
    			   const RelStatsInfo* StatsInfo) 
  : RelPhysPlan(PlanIdent, BlockIdent, LProps, 
    		RelationId, NPreds, RelOptEngine::FILESCAN)
{
  // make sure to specify the plan usage
  Usage.Reset();
  int TableCard;
  int AvgTupleSize;
  if (StatsInfo) {
    TableCard = StatsInfo->Cardinality;
    AvgTupleSize = StatsInfo->AvgTupleSize;
  }
  else {
    TableCard = RelStatsInfo::DefaultTableCard;
    AvgTupleSize = RelStatsInfo::DefaultTupleSize;
  }
  CostFileScan(TableCard, AvgTupleSize, XXX_TRUE, Env, Usage);
}
    
void RelSPJCostBasedOpt::
RelFScanPlan::Print(ostream& OStr) const
{
  OStr << "FSCAN:: ";
  RelPhysPlan::Print(OStr);
}
    
XxxErrCode RelSPJCostBasedOpt::
RelFScanPlan::GeneratePlanOp(const RelSPJOptimizer* Opt,
    			     XxxBool TopLevel,
    			     XxxExprPlan::VarRemapping& Remap,
    			     int InitRemapSrcNum,
    			     int FinalRemapSrcNum,
    			     RelPlanOp*& PlanOp) const
{
  // step 1: get the planop from the child
  RelExecParameters* Params = NULL; // no need for exec parameters for children
  RelBlockPlan* ChildBlockPlan = Opt->BlockPlan->GetChild(RelId);
  RelPlanOp* ChildPlanOp;
  if (ChildBlockPlan->GeneratePlanOp(Params, ChildPlanOp) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, "No planop for RelPhysPlan");
    return XXX_NOT_OK;
  }
    
  // step 2: modify the Remap
  Remap.SrcMappingArray[RelId].NewSrc = InitRemapSrcNum;
    
  // step 3: get and remap the selection predicates
  XxxBooleanExpressionPlan* SelPredPlan = NULL;
  for (int k=0; k < NewPreds.ArraySize; k++) {
    if (NewPreds.GetBit(k)) {
      if (XxxBooleanExpressionPlan::MakeConjunction
	  (SelPredPlan, Opt->PredArray[k]) != XXX_OK) {
	REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
	return XXX_NOT_OK;
      }
    } // if
  } // for each predicate
  if (SelPredPlan &&
      SelPredPlan->RedirectDependencies(Remap) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, "Error redirecting sel pred");
    return XXX_NOT_OK;
  }
    
  // add the selection predicates
  if (ChildBlockPlan->Arity() == 0) {
    // it is a base plan op
    RelBasePlanOp* BasePlan = (RelBasePlanOp*)ChildPlanOp;
     //Begin: Moustafa
    //  Make sure that, a decision is not made to make the plan stream scan. (Mar. 2003)
    if(!BasePlan->IsStreamScan()){
      //End: Moustafa
      // create a new planop over that relation and throw
      // away the old one
      RelFScanPlanOp* NewPlanOp = 
	new RelFScanPlanOp(BasePlan->RelName(), 
			   BasePlan->CatalogInfo,
			   SelPredPlan);
      delete BasePlan;
      PlanOp = NewPlanOp;
      //Begin: Moustafa 
      // see previous comment (Mar. 2003)
    }
    else
      PlanOp = ChildPlanOp;
    //End: Moustafa
  }
  else {
    // Note: TODO:LOpt : if the child is actually a subquery, a
    // decision should be made: should the subquery be reevaluated
    // each time or should a temp be created: for now, TmpPlanOps
    // have not been considered ....
    // This decision (Tmp or not) should be set as a field of the
    // RelFScanPlan during optimization
    /*** 
      PlanOp = new RelTmpScanPlanOp(ChildPlanOp->CatalogInfo,
      ChildPlanOp,
      SelPredPlan);
      ****/
    if (SelPredPlan) {
      /*********
	//  create a plan that applies the selection-cond
	XxxBooleanExpressionPlan** SelPredPlans = new XxxBooleanExpressionPlan*[1];
	SelPredPlans[0] = SelPredPlan;
	PlanOp = new RelTNLPlanOp(1, &ChildPlanOp,
	ChildPlanOp->CatalogInfo,
	NULL, // empty exprplanlist -- no projection
	SelPredPlans, NULL, NULL, NULL);
	********/
      PlanOp = new RelFilterPlanOp(ChildPlanOp, 
				   ChildPlanOp->CatalogInfo,
				   SelPredPlan);
    }
    else
      PlanOp = ChildPlanOp;
  }
    
  // step 5: adjust the Remap
  Remap.SrcMappingArray[RelId].NewSrc = FinalRemapSrcNum;
    
  return XXX_OK;
}
    
//
// Methods of RelSPJCostBasedOpt::RelIScanPlan
//
RelSPJCostBasedOpt::
RelIScanPlan::RelIScanPlan(int PlanIdent,
    			   unsigned char BlockIdent,
    			   const RelOptEngine::LogProperties& LProps,
    			   unsigned char RelationId,
    			   const XxxBitMap& NPreds,
    			   const char* IName,
    			   const XxxBitMap& IPreds,
    			   const XxxBitMap& RelDeps,
    			   const XxxMatchInfo* MInfo,
    			   const XxxDBEnv& Env,
    			   const RelStatsInfo* StatsInfo)
  : RelPhysPlan(PlanIdent, BlockIdent, LProps, 
    		RelationId, NPreds, RelOptEngine::NDXSCAN),
    IndxPreds(IPreds), RelDependencies(RelDeps), MatchedInfo(MInfo)
{
  strncpy(IndexName, IName, REL_RELNAME_MAX); 
  IndexName[REL_RELNAME_MAX-1] = 0;
    
  Usage.Reset();
    
  int TableCard;
  int AvgTupleSize;
  if (StatsInfo) {
    TableCard = StatsInfo->Cardinality;
    AvgTupleSize = StatsInfo->AvgTupleSize;
  }
  else {
    TableCard = RelStatsInfo::DefaultTableCard;
    AvgTupleSize = RelStatsInfo::DefaultTupleSize;
  }
  // TODO: put in true values for IndxInfo, etc
  CostIndexScan(TableCard, AvgTupleSize, 
		NULL, MatchedInfo,
		XXX_TRUE, 0.1,
		Env, Usage);
}
    
XxxErrCode RelSPJCostBasedOpt::
RelIScanPlan::GeneratePlanOp(const RelSPJOptimizer* Opt,
    			     XxxBool TopLevel,
    			     XxxExprPlan::VarRemapping& Remap,
    			     int InitRemapSrcNum,
    			     int FinalRemapSrcNum,
    			     RelPlanOp*& PlanOp) const
{
  // step 1: get the planop from the child
  RelExecParameters* Params = NULL; // no need for exec parameters for children
  RelBlockPlan* ChildBlockPlan = Opt->BlockPlan->GetChild(RelId);
  // indexed plans can only be over stored relations
  if (ChildBlockPlan->Arity() != 0) {
    REL_NOTE_ERROR(REL_BADEXEC, "Index on non-stored rel??");
    return XXX_NOT_OK;
  }
    
  RelPlanOp* ChildPlanOp;
  if (ChildBlockPlan->GeneratePlanOp(Params, ChildPlanOp) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, "No planop for RelPhysPlan");
    return XXX_NOT_OK;
  }
    
  // step 2: modify the Remap
  // now adjust the Remap to account for this plan
  Remap.SrcMappingArray[RelId].NewSrc = InitRemapSrcNum;
    
     // step 3: get and remap the index and residual predicates
  XxxBooleanExpressionPlan* IndxPredPlan = NULL;
  for (int k=0; k < IndxPreds.ArraySize; k++) {
    if (IndxPreds.GetBit(k)) {
      if (XxxBooleanExpressionPlan::MakeConjunction
	  (IndxPredPlan, Opt->PredArray[k]) != XXX_OK) {
	REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
	return XXX_NOT_OK;
      }
    } // if
  } // for each predicate
  if (IndxPredPlan &&
      IndxPredPlan->RedirectDependencies(Remap) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, "Error redirecting indx pred");
    return XXX_NOT_OK;
  }
    
  XxxBooleanExpressionPlan* ResidPredPlan = NULL;
  for (int l=0; l < NewPreds.ArraySize; l++) {
    if (NewPreds.GetBit(l) && !IndxPreds.GetBit(l)) {
      if (XxxBooleanExpressionPlan::MakeConjunction
	  (ResidPredPlan, Opt->PredArray[l]) != XXX_OK) {
	REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
	return XXX_NOT_OK;
      }
    } // if
  } // for each predicate
  if (ResidPredPlan &&
      ResidPredPlan->RedirectDependencies(Remap) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, "Error redirecting resid pred");
    return XXX_NOT_OK;
  }
    
    
  // it is a base plan op
  RelBasePlanOp* BasePlan = (RelBasePlanOp*)ChildPlanOp;
  //Begin: Moustafa
  //  Make sure that, a decision is not made to make the plan stream scan.

  if(!BasePlan->IsStreamScan()){
    //End: Moustafa
    // create a new planop over that relation and throw
    // away the old one
    RelIScanPlanOp* NewPlanOp = 
      new RelIScanPlanOp(BasePlan->RelName(), 
			 BasePlan->CatalogInfo,
			 IndxPredPlan, ResidPredPlan,
			 MatchedInfo, IndexName);
    delete BasePlan;
    PlanOp = NewPlanOp;
    //Begin: Moustafa 
    // see previous comment (Mar. 2003)
  }
  else
    PlanOp = ChildPlanOp;
  //End: Moustafa

  // step 5: modify the Remap
  Remap.SrcMappingArray[RelId].NewSrc = FinalRemapSrcNum;
  
  return XXX_OK;
}

void RelSPJCostBasedOpt::
RelIScanPlan::Print(ostream& OStr) const
{
  OStr << "ISCAN:: ";
  RelPhysPlan::Print(OStr);
}
    
//-------------------------------------------------------------------
//-------------------------------------------------------------------
//Begin: Moustafa
//
// Methods of RelSPJCostBasedOpt::RelSScanPlan
//
RelSPJCostBasedOpt::
RelSScanPlan::RelSScanPlan(int PlanIdent,
    			   unsigned char BlockIdent,
    			   const RelOptEngine::LogProperties& LProps,
    			   unsigned char RelationId,
    			   const XxxBitMap& NPreds,
    			   const XxxDBEnv& Env,
    			   const RelStatsInfo* StatsInfo) 
  : RelPhysPlan(PlanIdent, BlockIdent, LProps, 
    		RelationId, NPreds, RelOptEngine::SSCAN)
{
  // make sure to specify the plan usage
  Usage.Reset();
  int TableCard;
  int AvgTupleSize;
  if (StatsInfo) {
    TableCard = StatsInfo->Cardinality;
    AvgTupleSize = StatsInfo->AvgTupleSize;
  }
  else {
    TableCard = RelStatsInfo::DefaultTableCard;
    AvgTupleSize = RelStatsInfo::DefaultTupleSize;
  }
  CostStreamScan(TableCard, AvgTupleSize, XXX_TRUE, Env, Usage);
}
    
void RelSPJCostBasedOpt::
RelSScanPlan::Print(ostream& OStr) const
{
  OStr << "SSCAN:: ";
  RelPhysPlan::Print(OStr);
}
    
XxxErrCode RelSPJCostBasedOpt::
RelSScanPlan::GeneratePlanOp(const RelSPJOptimizer* Opt,
    			     XxxBool TopLevel,
    			     XxxExprPlan::VarRemapping& Remap,
    			     int InitRemapSrcNum,
    			     int FinalRemapSrcNum,
    			     RelPlanOp*& PlanOp) const
{
  // step 1: get the planop from the child
  RelExecParameters* Params = NULL; // no need for exec parameters for children
  RelBlockPlan* ChildBlockPlan = Opt->BlockPlan->GetChild(RelId);
  RelPlanOp* ChildPlanOp;
  if (ChildBlockPlan->GeneratePlanOp(Params, ChildPlanOp) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, "No planop for RelPhysPlan");
    return XXX_NOT_OK;
  }
    
  // step 2: modify the Remap
  Remap.SrcMappingArray[RelId].NewSrc = InitRemapSrcNum;
    
  // step 3: get and remap the selection predicates
  XxxBooleanExpressionPlan* SelPredPlan = NULL;
  for (int k=0; k < NewPreds.ArraySize; k++) {
    if (NewPreds.GetBit(k)) {
      if (XxxBooleanExpressionPlan::MakeConjunction
	  (SelPredPlan, Opt->PredArray[k]) != XXX_OK) {
	REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
	return XXX_NOT_OK;
      }
    } // if
  } // for each predicate
  if (SelPredPlan &&
      SelPredPlan->RedirectDependencies(Remap) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, "Error redirecting sel pred");
    return XXX_NOT_OK;
  }
    
  // add the selection predicates
  ASSERT(ChildBlockPlan->Arity() == 0); // the stream scan must be on a relation not a subquery
  
  // it is a base plan op
  RelBasePlanOp* BasePlan = (RelBasePlanOp*)ChildPlanOp;

  RelSScanPlanOp* NewPlanOp = 
    new RelSScanPlanOp(BasePlan->RelName(), 
		       BasePlan->CatalogInfo,
		       SelPredPlan);
  delete BasePlan;
  PlanOp = NewPlanOp;
    
  // step 5: adjust the Remap
  Remap.SrcMappingArray[RelId].NewSrc = FinalRemapSrcNum;
  
  return XXX_OK;
}
//End: Moustafa
