// Ihab
// New file to implement the logical KRJ plan op
//
#include <iostream.h>
#include <math.h>

#include "relerror.h"
#include "relquery.h"
#include "planKRJ.h"
#include "seminaiveKRJ.h"
#include "relplanop.h"
#include "relenv.h"

XxxErrCode RelOptEngine::planKRJBlock(RelQueryNode* Node,
				     unsigned char BlockIdent,
				     OptQueryEnv  *Env,
				     RelBlockPlan*& Plan)
{
  ASSERT((Node->Arity() > 0));
  RelBlockPlan** ChildPlans = new RelBlockPlan*[Node->Arity()];

  // first find plans for all the children: they should have
  // already been computed by the depth-first postfix travesal
  for (int i = 0; i < Node->Arity(); i++) {
    RelQueryNode* ChildNode = (RelQueryNode*)(Node->GetChild(i));
    XxxBool ChildFound = XXX_FALSE;
    if ((findChildBlockPlan(ChildNode, Env, 
			    ChildFound, ChildPlans[i]) != XXX_OK) ||
	(ChildFound != XXX_TRUE)) {
      REL_NOTE_ERROR(REL_BADOPT, "Unable to find plan for child of SPJ");
      delete [] ChildPlans;
      return XXX_NOT_OK;
    }
  } // for all children
  XxxErrCode rc = XXX_OK;
  RelKRJOptimizer::RelKRJBlockPlan* KRJPlan = 
    new RelKRJOptimizer::RelKRJBlockPlan(rc, BlockIdent,
				      Node->Op, Node->Arity(), 
				      ChildPlans, Node->CatalogInfo,
				      Env);
  delete [] ChildPlans;
  
  Plan = KRJPlan;
  
  if (rc != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, "Error optimizing SPJ block");
  }
  return rc;
}
//------------------------------------------------------
//
// Methods of RelKRJBlockPlan
//
RelKRJOptimizer::
RelKRJBlockPlan::RelKRJBlockPlan(XxxErrCode& RC,
			       unsigned char BlockIdent,
			       RelOperator* Op, 
			       int NumChildren,
			       RelBlockPlan* Children[],
			       RelCatalogInfo* CatInfo,
			       RelOptEngine::OptQueryEnv* Env)
  : RelBlockPlan(RC, BlockIdent, NumChildren, Children, CatInfo, Env)
{
 Opt = NULL;
 if (RC != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOP, "Error in RelBlockPlan constructor");
   return;
 }
 XxxBool Er = XXX_IS_ERROR();
 // Here we determine the type of optimization we use
 // for now we use the seminaiive optimizer
 // enough to form a default pipeline of physical operators
 // the join order is as specified in the query
 Opt = new RelKRJSemiNaiveOpt(RC,
			     ((RelKRJOp*)Op)->TargetList,
			     ((RelKRJOp*)Op)->Condition,
			     ((RelKRJOp*)Op)->Ordering, // new for ranking
			     ((RelKRJOp*)Op)->Distinct,
			     this, Env);
 Er = XXX_IS_ERROR();
 if (!Opt->IsValid() || (RC != XXX_OK)) {
   REL_NOTE_ERROR(REL_BADOP, "Invalid optimizer");
   if (Opt) delete Opt;
   Opt = NULL;
   return;
 }
 
 if ((RC = Opt->Optimize()) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOP, "Error during optimization");
   if (Opt) delete Opt;
   Opt = NULL;
   return;
 }
 // do not delete Opt here: it carries the state of the optimizer
 RC = XXX_OK;
}
//--------------------------------------------------------------
RelKRJOptimizer::
RelKRJBlockPlan::~RelKRJBlockPlan()
{
 if (Opt) delete Opt;
}
//----------------------------------------------------------------
void RelKRJOptimizer::
RelKRJBlockPlan::Print(ostream& OStr) const
{
 OStr << "(KRJ PLAN";
 if (Opt && Opt->IsValid()) {
   // Opt->PrintPlans(OStr);
 }
 else {
   OStr << "-- Invalid";
  }
 OStr << ")";
 
 RelBlockPlan::Print(OStr);
}
//---------------------------------------------------------------
XxxErrCode RelKRJOptimizer::
RelKRJBlockPlan::GeneratePlanOp(int PlanId,
				RelPlanOp*& PlanOp) const
{
 ASSERT((PlanId >= 0));

 RelOptEngine::PlanObject* ChosenPlan;
 if (GetPlan(PlanId, ChosenPlan) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Unable to find chosen plan");
   return XXX_NOT_OK;
  }
 if (Opt->GeneratePlanOp(ChosenPlan, PlanOp) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Unable to find chosen plan");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}
//---------------------------------------------------------------
XxxErrCode RelKRJOptimizer::
RelKRJBlockPlan::RedirectDependencies(const XxxValueExprPlan::VarRemapping& Remap,
				      int CorrelationHt)
{
  REL_CALL_FN(redirectDependencies(Remap, CorrelationHt),
	      REL_BADOPT, "Unable to remap plan");

  REL_CALL_FN(Opt->ExprPlanList->RedirectDependencies(Remap, CorrelationHt),
	      REL_BADOPT, "Unable to remap target list");
  if(Opt->OrderPlanList)
    REL_CALL_FN(Opt->OrderPlanList->RedirectDependencies(Remap, CorrelationHt),
		REL_BADOPT, "Unable to remap ranking list");
  for (int i = 0; i < Opt->NumPreds; i++) {
    REL_CALL_FN(Opt->PredArray[i]->
		RedirectDependencies(Remap, CorrelationHt),
		REL_BADOPT, "Unable to remap target list");
  }
  return XXX_OK;
}
//----------------------------------------------------------------------
//
// Methods of abstract RelKRJOptimizer class
//
RelKRJOptimizer::
RelKRJOptimizer(XxxErrCode& RC,
		XxxValueExprList* TargetList,
		XxxBooleanExpression* Condition,
	        XxxValueExprList * Ordering,
		XxxBool Unique,
		RelKRJBlockPlan* KRJPlan,
		RelOptEngine::OptQueryEnv* QEnv)
{
  
  XxxBool Er = XXX_IS_ERROR();
  
  isValid = XXX_FALSE;
  BlockPlan = KRJPlan;
  Env = QEnv;
  Distinct = Unique;
  
  NumPreds = 0;
  PredArray = NULL;
  PredSelectivity = NULL;
  
  ExprPlanList = NULL;
  OrderPlanList = NULL;
  // should fill ExprPlanList and PredArray;
  
  // first make sure that the target list is fine
  if (TargetList && (TargetList->Optimize(ExprPlanList) != XXX_OK)) {
    REL_NOTE_ERROR(REL_BADOPT, "Unable to optimize target list");
    RC = XXX_NOT_OK;
    return;
  }
  Er = XXX_IS_ERROR();
  
  // fill the ranking plan list from the ordering information to that block
  if (Ordering && (Ordering->Optimize(OrderPlanList) != XXX_OK)) {
    REL_NOTE_ERROR(REL_BADOP, "Unable to optimize ranking plans");
    RC = XXX_NOT_OK;
    return;
  }
  //XXX_CLEAR_ERROR(); // need fix  Ihab
  Er = XXX_IS_ERROR();

  if (Condition) {
    // Now try to break up Condition into conjunctive
    // boolean factors
    int NumFactors;
    XxxBooleanExpression** Factors;
    if (XxxBooleanExpression::GenerateBooleanFactors(Condition, NumFactors,
						     Factors) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADOPT, "Unable to split conditions");
      RC = XXX_NOT_OK;
      return;
    }
    
    // now keep track of the dependencies for each boolean factor,
    // and generate a plan for each
    NumPreds = NumFactors;
    PredArray = new XxxBooleanExpressionPlan*[NumPreds];
    PredDependencyInfo = new XxxBitMap[NumPreds];
    int RelArity = KRJPlan->ChildArrayLen;
    int i;
    for (i =0; i < NumPreds; i++) {
      XxxExprPlan* TmpPlan = NULL;
      if (Factors[i]->Optimize(TmpPlan) != XXX_OK) {
	REL_NOTE_ERROR(REL_BADOPT, "Error optimizing boolean factor");
	RC = XXX_NOT_OK;
	return;
      }
      PredArray[i] = (XxxBooleanExpressionPlan*)TmpPlan;
      
      // Get predicate dependencies.
      // The actual dependencies also include one bit
      // for "env" values. So we need to get rid of this bit.
      XxxBitMap TmpPredDep(RelArity+1);
      if (Factors[i]->GetDependencies(TmpPredDep, 0) != XXX_OK) {
	REL_NOTE_ERROR(XXX_BADEXPR, "Unable to get expr dependencies");
	RC = XXX_NOT_OK;
	return;
      }
      PredDependencyInfo[i].ChangeSize(RelArity);
      for (unsigned char j = 0; j < RelArity; j++)  {
	if (TmpPredDep.GetBit(j+1))
	  PredDependencyInfo[i].SetBit(j);
      }
      
    } // for each boolean factor
    
    // determine the selectivity information for the predicates
    XxxBooleanExpression::SelectivityStatsInfo* SelectivityStats = 
      new XxxBooleanExpression::SelectivityStatsInfo[BlockPlan->Arity()];
    for (i = 0; i < BlockPlan->Arity(); i++) {
      RelCatalogInfo* CatInfo = BlockPlan->GetChild(i)->CatalogInfo;
      SelectivityStats[i].RecSchema = CatInfo->RecordSchema();
      SelectivityStats[i].AttrStats = 
	((CatInfo->StatsInfo())?(CatInfo->StatsInfo()->AttrStats): 0);
    }
    
    PredSelectivity = new double[NumPreds];
    for (i =0; i < NumPreds; i++) {
      PredSelectivity[i] = Factors[i]->Selectivity(SelectivityStats, 
						   BlockPlan->Arity());
      PredArray[i]->Selectivity = PredSelectivity[i];
    } 
    
   delete [] SelectivityStats;
  } // if Condition
  
  isValid = XXX_TRUE;
  RC = XXX_OK;
}
//------------------------------------------------------------------------
RelKRJOptimizer::~RelKRJOptimizer()
{
 // Delete all the structures used in the optimizer.
 // Any plans that use these structures are already
 // deleted at this stage.
 if (ExprPlanList) ExprPlanList->Destroy();
 //if (OrderPlanList) OrderPlanList->Destroy();

 if (NumPreds) {
   if (PredArray) {
     for (int i = 0; i < NumPreds; i++) {
       /******* Seems to cause a crash : am i reclaiming this elsewhere too?
	 TODO:L1
       if (PredArray[i])
	 delete PredArray[i];
	 *******/
      }
     delete [] PredArray;
    }
   // The destructor of XxxBitMap gets called on each element.
   if (PredDependencyInfo) {
     delete [] PredDependencyInfo;
    }
   if (PredSelectivity) {
     delete [] PredSelectivity;
    }
  } // if NumPreds
}

void RelKRJOptimizer::PrintOptInfo(ostream& OStr) const
{
 OStr << "Num Rels == " << BlockPlan->Arity() << endl;
 int i;
 for (i = 0; i < BlockPlan->Arity(); i++) {
   OStr << "\t " << i << " : ";
   OStr << ChildCat(i)->RelationName() << endl;
  }
 OStr << "Num Preds == " << NumPreds << endl;
 for (i = 0; i < NumPreds; i++) {
   OStr << "\t " << i << " : ";
   PredArray[i]->Print(OStr);
   OStr << "  dep:";
   PredDependencyInfo[i].Print(OStr);
   OStr << " sel:" << PredSelectivity[i] << endl;
  }
 OStr << " Targetlist: ";
 if (Distinct)
   OStr << " DISTINCT: ";
 ExprPlanList->Print(OStr);
 OStr << endl;
}

//-----------------------------------------------no Idea yet------------------------

XxxErrCode RelKRJOptimizer::
JoinsWith(const RelOptEngine::LogProperties& LProps0,
	  const RelOptEngine::LogProperties& LProps1,
	  XxxBool AvoidExpensivePreds, 
	  double MaxCheapPredCost, // definition of "expensive"
	  RelOptEngine::LogProperties& JoinProps,
	  XxxBitMap& NewPreds,
	  XxxBool& Joinable) const
{
 Joinable = XXX_FALSE;

 ASSERT((LProps0.Rels.ArraySize == LProps1.Rels.ArraySize));
 ASSERT((LProps0.Preds.ArraySize == LProps1.Preds.ArraySize));

 // first make sure that the rels in LProps0 and LProps1 differ.
 XxxBitMap OverlapRels(LProps0.Rels.ArraySize); 
 LProps0.Rels.And(LProps1.Rels, OverlapRels);
 if (OverlapRels.NumBitsSet() > 0) {
   return XXX_OK;
  }
 // likewise for the preds
 XxxBitMap OverlapPreds(LProps0.Preds.ArraySize); 
 LProps0.Preds.And(LProps1.Preds, OverlapPreds);
 if (OverlapPreds.NumBitsSet() > 0) {
   return XXX_OK;
  }

 // set the Rels bits in JoinProps
 JoinProps.Rels.ClearAll();
 JoinProps.Rels.Or(LProps0.Rels, JoinProps.Rels);
 JoinProps.Rels.Or(LProps1.Rels, JoinProps.Rels);

 // set the Preds bits in JoinProps
 JoinProps.Preds.ClearAll();
 JoinProps.Preds.Or(LProps0.Preds, JoinProps.Preds);
 JoinProps.Preds.Or(LProps1.Preds, JoinProps.Preds);

 JoinProps.Card = LProps0.Card * LProps1.Card;
 // TODO:L2 this should take projections into account
 JoinProps.TupleSize = LProps0.TupleSize + LProps1.TupleSize;

 // look for unused predicates that will act as join preds
 XxxBool JoinPredFound = XXX_FALSE;
 XxxBool CandidatePredFound = XXX_FALSE;
 for (int i = 0; i < LProps0.Preds.ArraySize; i++) {
   if (JoinProps.Preds.GetBit(i)) continue;

   // check here is this is an expensive predicate
   if (AvoidExpensivePreds && 
       PredArray[i]->GetPlanUsage().FullCost() > MaxCheapPredCost) {
     continue;
   }

   CandidatePredFound = XXX_TRUE;
   // otherwise, this predicate may be a candidate. check further.
   // check if the dependency of the predicate
   // is fully covered by JoinProps.Rels.
   // Reuse OverlapRels for this purpose
   OverlapRels.ClearAll();
   JoinProps.Rels.And(PredDependencyInfo[i], OverlapRels);
   if (OverlapRels.Equals(PredDependencyInfo[i])) {
     // this is a join predicate
     JoinPredFound = XXX_TRUE;
     JoinProps.Preds.SetBit(i);
     NewPreds.SetBit(i);
     double PredSel = PredSelectivity[i];
     JoinProps.Card = (int)(((double)JoinProps.Card) * PredSel);
    }
  }

 if (JoinPredFound) Joinable = XXX_TRUE;

 // otherwise make sure cross products are used if absolutely needed
 if (!CandidatePredFound) Joinable = XXX_TRUE;

 return XXX_OK;
}

XxxErrCode RelKRJOptimizer::
SetRelLProps(unsigned char RelId,
	     XxxBool AvoidExpensivePreds, 
	     double MaxCheapPredCost, // definition of "expensive"
	     RelOptEngine::LogProperties& RelLProps) const
{
 RelLProps.Rels.ClearAll();
 RelLProps.Rels.SetBit(RelId);

 const RelCatalogInfo* CatInfo = ChildCat(RelId);
 if (!CatInfo) {
   REL_NOTE_ERROR(REL_BADOPT, "Invalid cat info for child");
   return XXX_NOT_OK;
 }
 RelStatsInfo* StatsInfo = CatInfo->StatsInfo();
 if (StatsInfo) {
   RelLProps.Card = StatsInfo->Cardinality;
   RelLProps.TupleSize = StatsInfo->AvgTupleSize;
 } 
 else {
   RelLProps.Card = RelStatsInfo::DefaultTableCard;
   RelLProps.TupleSize = RelStatsInfo::DefaultTupleSize;
 }

 // determine what preds if any should be applied ...
 RelLProps.Preds.ClearAll();
 for (int j = 0; j < RelLProps.Preds.ArraySize; j++) {
   if ((PredDependencyInfo[j].NumBitsSet() == 1) &&
       (PredDependencyInfo[j].GetBit(RelId))) {

     // check here is this is an expensive predicate
     if (AvoidExpensivePreds && 
	 PredArray[j]->GetPlanUsage().FullCost() > MaxCheapPredCost) {
       continue;
     }

     RelLProps.Preds.SetBit(j);
     double PredSel = PredSelectivity[j];
     RelLProps.Card = (int)(((double)RelLProps.Card) * PredSel);
   }
 }
 
 return XXX_OK;
}



