/*
Ihab
a new file to define the KRJ logical operator (block)
*/
#ifndef XXX_PLANKRJ_H
#define XXX_PLANKRJ_H

#include "relopt.h"
#include "misc.h"
#include "values.h"

/**
 * This is the root class of all KRJ optimizers
 */
class RelKRJOptimizer {
 public:
  XXX_EXTERN_CLASS_DECL(PlanOptions);
  /**
   *
   */
  class RelKRJBlockPlan : public RelBlockPlan { 
   public:
    RelKRJOptimizer* Opt; // this is the "optimizer" object
    virtual void Print(ostream& OStr) const;
    RelKRJBlockPlan(XxxErrCode& RC,
		    unsigned char BlockIdent,
		    RelOperator*, 
		    int NumChildren,
		    RelBlockPlan* Children[],
		    RelCatalogInfo* CatInfo,
		    RelOptEngine::OptQueryEnv* Env);
    virtual ~RelKRJBlockPlan();
    virtual XxxErrCode GeneratePlanOp(int PlanId, RelPlanOp*& PlanOp) const;
    // This method replaces the dependencies of all correlation
    // variables (i.e XxxUnknownVars) within this plan.
    virtual XxxErrCode RedirectDependencies(const XxxValueExprPlan::VarRemapping& Remap,
					    int CorrelationHt = 0);
   };

  struct PredPlanStruct {
    XxxBooleanExpressionPlan* InitPlan;
    XxxBooleanExpressionPlan** SelPredPlans;
    XxxBooleanExpressionPlan** JoinPredPlans;
    XxxBooleanExpressionPlan** ResidPredPlans;
    PredPlanStruct() {
      InitPlan = NULL;
      SelPredPlans = NULL;
      JoinPredPlans = NULL;
      ResidPredPlans = NULL;
     }
    ~PredPlanStruct() {}
    void Reclaim() {
      if (SelPredPlans) delete [] SelPredPlans;
      if (JoinPredPlans) delete [] JoinPredPlans;
      if (ResidPredPlans) delete [] ResidPredPlans;
     }
   };
 //// dummy superclass that indicates glue information for a plan
  class PlanGlue {
   public:
    virtual ~PlanGlue() {}
   };
  //// Glue information for sort merge join
  class SMPlanGlue : public PlanGlue {
   public:
    int LeftRelIndex, RightRelIndex;
    XxxBool SortLeft, SortRight;
    XxxSortInfo LeftSInfo, RightSInfo;
    //// indicate whether the inputs need a temp to sort them
    XxxBool MaterializeLtInput, MaterializeRtInput;
    //// indicate max run sizes for sorting
    int LeftRunSize, RightRunSize;
    SMPlanGlue(int NumAttrs) 
      : LeftSInfo(NumAttrs), RightSInfo(NumAttrs) {}
    virtual ~SMPlanGlue() {}
   };

  
  XxxBool isValid;
 public:
  RelOptEngine::OptQueryEnv *Env;
  RelKRJBlockPlan* BlockPlan;
  XxxValExprPlanList* ExprPlanList;  // list of target plans
  XxxValExprPlanList* OrderPlanList; // list of ranking plans // new
  XxxBool Distinct;
  int NumPreds;                      // conjunctive boolean predicates
  XxxBooleanExpressionPlan** PredArray; // array of pred plans
  XxxBitMap* PredDependencyInfo;        // array of predicate dependencies 
  double* PredSelectivity;              // array of predicate selectivities
  
  const RelCatalogInfo* ChildCat(int i) const {
    if (i >= BlockPlan->Arity()) return NULL;
    else return BlockPlan->GetChild(i)->CatalogInfo;
  }
  
  RelKRJOptimizer(XxxErrCode& RC,
		 XxxValueExprList* TargetList,
		 XxxBooleanExpression* Condition,
		 XxxValueExprList * Ordering,   //ranking expression
		 XxxBool Unique,
		 RelKRJBlockPlan* KRJPlan,
		 RelOptEngine::OptQueryEnv* QEnv);
  virtual ~RelKRJOptimizer();

  const RelSpecParameters* OptSpecs() const { return Env->OptSpecs;}
    
  virtual XxxErrCode Optimize() = 0;
  // dump the contents of the RelSPJOptimizer
  virtual void PrintOptInfo(ostream& OStr) const;

  virtual XxxErrCode GeneratePlanOp(RelOptEngine::PlanObject* Plan,
				    RelPlanOp*& PlanOp) const = 0;

  XxxBool IsValid() { return isValid;}
  virtual void PrintPlans(const RelOptEngine::PlanObject* Plan,
			  ostream& OStr) const = 0;
  // Do the two sub-plans represented by LProps0 and LProps1 join?
  // What are the resulting JoinProps? What NewPreds are applied?
  virtual XxxErrCode 
    JoinsWith(const RelOptEngine::LogProperties& LProps0,
	      const RelOptEngine::LogProperties& LProps1,
	      XxxBool AvoidExpensivePreds, 
	      double MaxCheapPredCost, // definition of "expensive"
	      RelOptEngine::LogProperties& JoinProps,
	      XxxBitMap& NewPreds,
	      XxxBool& Joinable) const;
  // For a table scan, what are the resulting logical props?
  virtual XxxErrCode 
    SetRelLProps(unsigned char RelId,
		 XxxBool AvoidExpensivePreds, 
		 double MaxCheapPredCost, // definition of "expensive"
		 RelOptEngine::LogProperties& RelLProps) const;
  
};

#endif
