#include <iostream.h>

#include "planbase.h"
#include "relplanop.h"
#include "relerror.h"
#include "relquery.h"
//Begin: Moustafa
#include "../../newtypes/genadt.h"
//End: Moustafa

XxxErrCode RelOptEngine::planBaseBlock(RelQueryNode* Node,
				       unsigned char BlockIdent,
				       OptQueryEnv  *Env,
				       RelBlockPlan*& Plan)
{
 XxxErrCode rc = XXX_OK;

 Plan = new RelBaseBlockPlan(rc,
			     BlockIdent,
			     Node->BasicName(),
			     Node->CatalogInfo,
			     Env);

 return rc;
}

RelBaseBlockPlan::RelBaseBlockPlan(XxxErrCode& RC,
				   unsigned char BlockIdent,
				   const char* Name,
				   RelCatalogInfo* CatInfo,
				   RelOptEngine::OptQueryEnv* Env)
: RelBlockPlan(RC, BlockIdent,
	       0 /* NumChildren */, NULL /* Children */, CatInfo, Env)
{
 strncpy(RelName, Name, REL_RELNAME_MAX);
 RelName[REL_RELNAME_MAX-1] = 0;

 //SMostafa
 // At this point a decision is made by the query compiler
 // to mark the query block as of streaming input
 
 // check all relaion attributes 
 
 for(int i=0; i< CatInfo->recSchema->RecArity; i++){
   XxxADTClass* attrADT = XxxADTArray[(CatInfo->recSchema->Attrs+i)->Type];
   
   if (!XxxCoreADTClass::IsCoreADT((CatInfo->recSchema->Attrs+i)->Type)){
     if(attrADT->IsSmall(NULL))
       HasStreamInput = ((XxxSmallADT*)attrADT)->IsStream();
     else
       HasStreamInput = ((XxxLargeADT*)attrADT)->IsStream();
     if (HasStreamInput)
       break;
   }
 }
 //EMostafa
 // factor in the RelSpecParameters in Env->OptSpecs

 // Initially, create a simple Scan plan over the table.
 // Index scans will only be used if selections or projections
 // are to be pushed down. These plans will be created dynamically
 // based on requests from a higher SPJ block.

 RelAccessPlan* ScanPlan = new RelAccessPlan(HighestPlanId++,
					     BlockId);
 // add the plan to this RelBlockPlan
 AddPlan(ScanPlan);

 RC = XXX_OK;
}

RelBaseBlockPlan::~RelBaseBlockPlan()
{
}

void RelBaseBlockPlan::Print(ostream& OStr) const
{
 OStr << RelName;

 RelBlockPlan::Print(OStr);
}

// this will generate a RelBasePlanOp
XxxErrCode 
RelBaseBlockPlan::GeneratePlanOp(int PlanId,
				 RelPlanOp*& PlanOp) const
{
 ASSERT((PlanId >= 0));
 RelOptEngine::PlanObject* ChosenPlan;
 if (GetPlan(PlanId, ChosenPlan) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Unable to find chosen plan");
   return XXX_NOT_OK;
  }
 
 RelAccessPlan* ChosenAccessPath = 
   (RelAccessPlan*)(ChosenPlan);
 if (ChosenAccessPath->GeneratePlanOp(RelName,
				      CatalogInfo, 
				      PlanOp) != XXX_OK){
   REL_NOTE_ERROR(REL_BADOPT, "Unable to generate base plan operator");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

XxxErrCode RelBaseBlockPlan::
RedirectDependencies(const XxxValueExprPlan::VarRemapping& Remap,
		     int CorrelationHt)
{
  REL_CALL_FN(redirectDependencies(Remap, CorrelationHt),
	      REL_BADOPT, "Unable to remap plan");
  return XXX_OK;
}

void RelBaseBlockPlan::
RelAccessPlan::Print(ostream& OStr) const
{
 OStr << "(BaseRel)";
}

XxxErrCode RelBaseBlockPlan::
RelAccessPlan::GeneratePlanOp(const char* RelName,
			      RelCatalogInfo* CatInfo,
			      RelPlanOp*& PlanOp) const
{

  //SMostafa: At this point a decision is made by the query plan generator
  //          to use stream scan operator if at least one of the attributes
  //          is a subclass of Stream Class ADT
  
  
  // check all relaion attributes 
  
  XxxBool IsStrm = XXX_FALSE;

  for(int i=0; i< CatInfo->recSchema->RecArity; i++){
    XxxADTClass* attrADT = XxxADTArray[(CatInfo->recSchema->Attrs+i)->Type];
    
    if (!XxxCoreADTClass::IsCoreADT((CatInfo->recSchema->Attrs+i)->Type)){
      if(attrADT->IsSmall(NULL))
	IsStrm = ((XxxSmallADT*)attrADT)->IsStream();
      else
	IsStrm = ((XxxLargeADT*)attrADT)->IsStream();
      if (IsStrm){
	PlanOp = new RelSScanPlanOp(RelName, CatInfo, NULL);
	return XXX_OK;
      }
    }
  }
  
  //EMostafa
  PlanOp = new RelBasePlanOp(RelName, CatInfo);
  return XXX_OK;
}

