#include <iostream.h>

#include "relerror.h"
#include "relquery.h"
#include "plangby.h"
#include "gbyplanop.h"
#include "prd-sort.h"

XxxErrCode RelOptEngine::planAggrBlock(RelQueryNode* Node,
				       unsigned char BlockIdent,
				       OptQueryEnv  *Env,
				       RelBlockPlan*& Plan)
{
  ASSERT((Node->Arity() == 1));
  
  RelGrpByBlockPlan* GBYPlan = NULL;
  RelBlockPlan* ChildBlockPlan = NULL;
  
  RelQueryNode* ChildNode = (RelQueryNode*)(Node->GetChild(0));
  XxxBool ChildFound = XXX_FALSE;
  if ((findChildBlockPlan(ChildNode, Env, 
			  ChildFound, ChildBlockPlan) != XXX_OK) ||
      (ChildFound != XXX_TRUE))  {
    REL_NOTE_ERROR(REL_BADOPT, "Unable to find plan for child of GBY");
    return XXX_NOT_OK;
  }
  
  XxxErrCode rc = XXX_OK;
  GBYPlan = new RelGrpByBlockPlan(rc, BlockIdent, Node->Op, 
				  ChildBlockPlan, Node->CatalogInfo, Env);
  Plan = GBYPlan;
  return rc;
}

// OWNER: as  DATE: 2/20/97
// PURPOSE: constructor
RelGrpByBlockPlan::RelGrpByBlockPlan(XxxErrCode& RC,
				     unsigned char BlockIdent,
				     RelOperator* Op, 
				     RelBlockPlan* Child,
				     RelCatalogInfo* CatInfo,
				     RelOptEngine::OptQueryEnv* Env)
: RelBlockPlan(RC, BlockIdent, 1, &Child, CatInfo, Env),
  TargetExprPlanList(NULL), AggrExprPlanList(NULL), 
  GroupExprPlanList(NULL) 
{

  XxxValueExprList *TargetExprList = ((RelGrpByOp *)Op)->TargetList;
  if (TargetExprList && 
      TargetExprList->Optimize(TargetExprPlanList) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOP, "Unable to optimize target list");
    RC = XXX_NOT_OK;
    return;
  }
  XxxValueExprList *AggrExprList = ((RelGrpByOp *)Op)->AggrList;
  if (AggrExprList && 
      AggrExprList->Optimize(AggrExprPlanList) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOP, "Unable to optimize aggregate list");
    RC = XXX_NOT_OK;
    return;
  }
  XxxValueExprList *GroupExprList = ((RelGrpByOp *)Op)->GroupList;
  if (GroupExprList && 
      GroupExprList->Optimize(GroupExprPlanList) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOP, "Unable to optimize grouping list");
    RC = XXX_NOT_OK;
    return;
  }
  // TODO:L_OPT
  // Step thru all the plans of the lower query block and find
  // those plans that match
  // Step 1 : Create a PhysProp with the groupby columns
  //          specifying the interesting order.
  // For now, since we do not support grouping columns, leave this
  // blank (NOT TRUE ANY MORE)
  RelOptEngine::PhysProperties DesiredProps, *ChildProps;
  XxxBool PlanFound = XXX_FALSE;
  int ChildPlanId;
  XxxGenericPlan::PlanUsage ChildUsage;
  
  if ((Child->GetBestPlan(DesiredProps, ChildProps, ChildPlanId,
			  ChildUsage, PlanFound) != XXX_OK) ||
      (PlanFound != XXX_TRUE)) {
    REL_NOTE_ERROR(REL_BADOPT, "Cannot get child plan");
    RC = XXX_NOT_OK;
    return;
  }
  
  //
  // Step 2:  Create a PlanObject using this child plan
  //
  GbyPlan* ThisPlan = new GbyPlan(HighestPlanId++,
				  BlockId,
				  ChildPlanId,
				  ChildUsage);
  
  // add the plan to this RelBlockPlan
  AddPlan(ThisPlan);
  
  RC = XXX_OK;
}

// OWNER: as  DATE: 2/20/97
// PURPOSE: destructor
// NOTES: Only the TargetExprPlanList owns the expression plans
RelGrpByBlockPlan::~RelGrpByBlockPlan()
{
  // TargetExprPlanList is the owner list, i.e. it owns the exprplans
  if (TargetExprPlanList) {
    TargetExprPlanList->Destroy();
  }
  if (AggrExprPlanList) {
    AggrExprPlanList->DestroyListOnly();
  }
  if (GroupExprPlanList) {
    GroupExprPlanList->DestroyListOnly();
  }
}

RelGrpByBlockPlan::
GbyPlan::GbyPlan(int PlanIdent,
		 unsigned char BlockIdent,
		 int ChildPlan,
		 XxxGenericPlan::PlanUsage& ChildUsage)
      :RelOptEngine::PlanObject(PlanIdent, BlockIdent)
{
 ChildPlanId = ChildPlan;
 Usage.Add(ChildUsage);
 // TODO:L_OPT
 //  Modify Usage to reflect the current plan
}

void RelGrpByBlockPlan::
GbyPlan::Print(ostream& OStr) const
{
 OStr << "[GBY Plan]";
}

void RelGrpByBlockPlan::Print(ostream& OStr) const
{
  OStr << "GRPBY PLAN ";
  if (TargetExprPlanList) {
    OStr << "[TARGETS: ";
    TargetExprPlanList->Print(OStr);
    OStr << "]";
  }
  if (AggrExprPlanList) {
    OStr << "[AGGRS: ";
    AggrExprPlanList->Print(OStr);
    OStr << "]";
  }
  if (GroupExprPlanList) {
    OStr << "[GROUP: ";
    GroupExprPlanList->Print(OStr);
    OStr << "]";
  }
  RelBlockPlan::Print(OStr);
}

XxxErrCode 
RelGrpByBlockPlan::GeneratePlanOp(int PlanId,
				  RelPlanOp*& PlanOp) const
{
 ASSERT((PlanId >= 0));
 
 RelOptEngine::PlanObject* ChosenPlan;
 if (GetPlan(PlanId, ChosenPlan) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Unable to find chosen plan");
   return XXX_NOT_OK;
  }
 // currently, the GbyPlan contains no info, so isnt used
 GbyPlan* ThePlan = (GbyPlan*)(ChosenPlan);

 RelExecParameters* Params = NULL; 
 RelPlanOp* ChildPlanOp;
 // first generate a planop for its child
 if (GetChild(0)->GeneratePlanOp(Params, 
				 ChildPlanOp) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADEXEC, 
		  "Cannot generate planop for child of aggr");
   return XXX_NOT_OK;
  }

 XxxSortInfo SInfo;
 if (GroupExprPlanList != NULL) {
   // if there are non-aggregate expressions, they are grouping terms,
   // and the input to the RelGrpByPlanOp must be sorted on them.
   
   // form the SortInfo
   SInfo.Init(GroupExprPlanList->Count());
   XxxValExprPlanList *GList = GroupExprPlanList;
   int i = 0;
   while (GList) {
     ASSERT(GList->Plan->ValType() == XxxValueExpression::COLUMN_VAR);
     SInfo[i].AttrIndex = 
       ((XxxUnknownValue::UnknownValuePlan *)(GList->Plan))->SrcAttrIndex;
     SInfo[i].Asc = XXX_TRUE;
     GList = GList->Next;  i++;
    }
   
   if (!ChildPlanOp->CatalogInfo->ImplInfo()->SortInfo.Matches(SInfo)) {
     // Create the sort plan op
     RelPlanOp *SortPlanOp = 
       new RelSortPlanOp(ChildPlanOp, SInfo, XXX_FALSE, XXX_TRUE, 
			 XXX_THISENV().DBEnv.BufPoolSize/2,
			 ChildPlanOp->CatalogInfo);
     ChildPlanOp = SortPlanOp;
    }
  }
 // now generate a plan operator for the group by
 PlanOp = new RelGrpByPlanOp(ChildPlanOp, 
			     TargetExprPlanList,
			     AggrExprPlanList,
			     GroupExprPlanList,
			     CatalogInfo);

 if (GroupExprPlanList != NULL) {
   // Make sure to specify the SortInfo in the result
   // First check how many attributes of the Grouping list are in the TargetList.
   unsigned int OutputSortedAttrs = SInfo.CommonAttrs(TargetExprPlanList);
   
   // If there is some sort order in the output
   if (OutputSortedAttrs) {
     ASSERT (OutputSortedAttrs <= SInfo.NumOrderedAttrs());
     // TODO:L2: if the same attribute is repeated in the output, there may
     // be more than one output sort order : for now, we ignore this possibility
     XxxSortInfo OutSInfo(OutputSortedAttrs);
     if (SInfo.FillTarget(TargetExprPlanList, OutSInfo) != XXX_OK) {
       REL_NOTE_ERROR(REL_BADEXEC, "Cannot generate sort info for aggr");
       return XXX_NOT_OK;
      }
     
     PlanOp->CatalogInfo->ImplInfo()->SortInfo.Init(1);
     PlanOp->CatalogInfo->ImplInfo()->SortInfo.SortInfoArray[0].Init(OutSInfo);
    }
  }
 return XXX_OK;
}

XxxErrCode RelGrpByBlockPlan::
RedirectDependencies(const XxxValueExprPlan::VarRemapping& Remap,
		     int CorrelationHt)
{
  REL_CALL_FN(redirectDependencies(Remap, CorrelationHt),
	      REL_BADOPT, "Unable to remap plan");

  if (TargetExprPlanList)
    REL_CALL_FN(TargetExprPlanList->RedirectDependencies(Remap, CorrelationHt),
	      REL_BADOPT, "Unable to remap target list");
  if (AggrExprPlanList)
    REL_CALL_FN(AggrExprPlanList->RedirectDependencies(Remap, CorrelationHt),
	      REL_BADOPT, "Unable to remap aggr list");
  if (GroupExprPlanList)
    REL_CALL_FN(GroupExprPlanList->RedirectDependencies(Remap, CorrelationHt),
	      REL_BADOPT, "Unable to remap group list");
  return XXX_OK;
}
