#include <iostream.h>

#include "relerror.h"
#include "relquery.h"
#include "planoby.h"
#include "relplanop.h"
#include "prd-sort.h"

XxxErrCode RelOptEngine::planOrderBlock(RelQueryNode* Node,
					unsigned char BlockIdent,
					OptQueryEnv  *Env,
					RelBlockPlan *& Plan)
{
  ASSERT((Node->Arity() == 1));

  RelOrderByBlockPlan* OBYPlan = NULL;
  RelBlockPlan* ChildBlockPlan = NULL;
  
  RelQueryNode* ChildNode = (RelQueryNode*)(Node->GetChild(0));
  XxxBool ChildFound = XXX_FALSE;
  if ((findChildBlockPlan(ChildNode, Env, 
			  ChildFound, ChildBlockPlan) != XXX_OK) ||
      (ChildFound != XXX_TRUE))  {
    REL_NOTE_ERROR(REL_BADOPT, "Unable to find plan for child of OBY");
    return XXX_NOT_OK;
  }
  
  XxxErrCode rc = XXX_OK;
  OBYPlan = new RelOrderByBlockPlan(rc, BlockIdent, Node->Op, 
				    ChildBlockPlan, Node->CatalogInfo, Env);
  Plan = OBYPlan;
  return rc;
}

RelOrderByBlockPlan::RelOrderByBlockPlan(XxxErrCode& RC,
					 unsigned char BlockIdent,
					 RelOperator* Op, 
					 RelBlockPlan* Child,
					 RelCatalogInfo* CatInfo,
					 RelOptEngine::OptQueryEnv* Env)
: RelBlockPlan(RC, BlockIdent, 1, &Child, CatInfo, Env)
{
  XxxValueExprList* TargetExprList = ((RelOrderByOp *)Op)->TargetList;
  if (TargetExprList &&
      TargetExprList->Optimize(TargetExprPlanList) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOP, "Unable to optimize order by target list");
    RC = XXX_NOT_OK;
    return;
  }
  if (!(Ascending = ((RelOrderByOp *)Op)->Ascending)) {
    REL_NOTE_ERROR(REL_BADOPT, "No Ascending info provided");
    RC = XXX_NOT_OK;
    return;
  }

  // @author:  Wee Hyong
  // @version: 20 June 2000
  // added creation of new SortOrder based on TargetExprPlanList
  XxxSortInfo SInfo(TargetExprPlanList->Count());
  XxxValExprPlanList *TList = TargetExprPlanList;

  for (unsigned int i = 0; (i < SInfo.NumOrderedAttrs()) && TList; i++) {
    SInfo[i].AttrIndex = ((XxxUnknownValue::UnknownValuePlan *)
			  (TList->Plan))->SrcAttrIndex;
    // sort merge ascending flag is always true
    //  SInfo[i].Asc = XXX_TRUE;
    //Bjarke, aug. 22: The ascending flag must be passed
    SInfo[i].Asc = Ascending[i];
   
    TList = TList->Next;
  }

  // Step 1 : Create a PhysProp with the groupby columns
  //          specifying the interesting order.
  RelOptEngine::PhysProperties DesiredProps(SInfo), *ChildProps;
  XxxBool PlanFound = XXX_FALSE;
  int ChildPlanId;
  XxxGenericPlan::PlanUsage ChildUsage;
  
  
  if ((Child->GetBestPlan(DesiredProps, ChildProps, ChildPlanId,
			  ChildUsage, PlanFound) != XXX_OK) ||
      (PlanFound != XXX_TRUE)) {
    REL_NOTE_ERROR(REL_BADOPT, "Cannot get child plan");
    RC = XXX_NOT_OK;
    return;
  }
 
  //
  // Step 2:  Create a PlanObject using this child plan
  // Check for intersting order
  //
  //if ( ChildProps->SortOrder.interestingMatches(DesiredProps.SortOrder) ) {
  if ( ChildProps->SortOrder.NumOrderedAttrs() > 0 &&
	  DesiredProps.SortOrder.interestingMatches(ChildProps->SortOrder) ) {
	// no need for order block anymore since the orderby plan is not added
	// instead, add the ChildPlan instead it is not added
	Env->BlockCount--;	
	RC = XXX_OK;
  }
  else {
	ObyPlan* ThisPlan = new ObyPlan(HighestPlanId++,
				  BlockId,
				  ChildPlanId,
				  ChildUsage);
	// add the plan to this RelBlockPlan
	AddPlan(ThisPlan);
	RC = XXX_OK;
  }
}

RelOrderByBlockPlan::~RelOrderByBlockPlan()
{
  if (TargetExprPlanList) {
    TargetExprPlanList->Destroy();
  }
}

RelOrderByBlockPlan::
ObyPlan::ObyPlan(int PlanIdent,
		 unsigned char BlockIdent,
		 int ChildPlan,
		 XxxGenericPlan::PlanUsage& ChildUsage)
:RelOptEngine::PlanObject(PlanIdent, BlockIdent)
{
 ChildPlanId = ChildPlan;
 Usage.Add(ChildUsage);
 // TODO:L_OPT
 //  Modify Usage to reflect the current plan
}

void RelOrderByBlockPlan::
ObyPlan::Print(ostream& OStr) const
{
  OStr << "[OBY Plan]";
}

void RelOrderByBlockPlan::Print(ostream& OStr) const
{
 OStr << "ORDERBY PLAN";
  if (TargetExprPlanList) {
    OStr << "[";
    const XxxValExprPlanList* Tmp = TargetExprPlanList;
    XxxBool *Asc = Ascending; 
    int i = 0;
    while (Tmp) {
      Tmp->Plan->Print(OStr);
      if (Asc[i++] == XXX_TRUE)
	OStr << " ASC";
      else
	OStr << " DESC";
      Tmp = Tmp->Next;
      if (Tmp) 
	OStr << ", ";
    }
    OStr << "]";
  }
}

XxxErrCode 
RelOrderByBlockPlan::GeneratePlanOp(int PlanId,
				    RelPlanOp*& PlanOp) const
{
  ASSERT((PlanId >= 0));
  
  RelOptEngine::PlanObject* ChosenPlan;
  if (GetPlan(PlanId, ChosenPlan) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, "Unable to find chosen plan");
    return XXX_NOT_OK;
  }
  // currently, the ObyPlan contains no info, so isn't used
  ObyPlan* ThePlan = (ObyPlan*)(ChosenPlan);
  
  RelExecParameters* Params = NULL; 
  RelPlanOp* ChildPlanOp;
  // first generate a planop for its child
  REL_CALL_FN(GetChild(0)->GeneratePlanOp(Params, ChildPlanOp),
	      REL_BADEXEC, "Cannot generate planop for child of order by");

  // Generate the sort information
  XxxSortInfo SInfo(TargetExprPlanList->Count());
  XxxValExprPlanList *TList = TargetExprPlanList;
  for (unsigned int i = 0; (i < SInfo.NumOrderedAttrs()) && TList; i++) {
    SInfo[i].AttrIndex = ((XxxUnknownValue::UnknownValuePlan *)
			  (TList->Plan))->SrcAttrIndex;
    SInfo[i].Asc = Ascending[i];
    TList = TList->Next;
  }
  PlanOp = new RelSortPlanOp(ChildPlanOp, SInfo, XXX_FALSE, XXX_TRUE, 
			     XXX_THISENV().DBEnv.BufPoolSize/2,
			     ChildPlanOp->CatalogInfo);
  // make sure to specify the SortInfo in the result
  PlanOp->CatalogInfo->ImplInfo()->SortInfo.Init(1);
  PlanOp->CatalogInfo->ImplInfo()->SortInfo.SortInfoArray[0].Init(SInfo);

  return XXX_OK;
}

XxxErrCode RelOrderByBlockPlan::
RedirectDependencies(const XxxValueExprPlan::VarRemapping& Remap,
		     int CorrelationHt)
{
  REL_CALL_FN(redirectDependencies(Remap, CorrelationHt),
	      REL_BADOPT, "Unable to remap plan");

  REL_CALL_FN(TargetExprPlanList->RedirectDependencies(Remap, CorrelationHt),
	      REL_BADOPT, "Unable to remap target list");
  return XXX_OK;
}
