#include <iostream.h>

#include "relerror.h"
#include "relquery.h"
#include "relenv.h"
#include "dynspj.h"
#include "relplanop.h"

//
// Methods of RelSPJDynOptimizer::PlanSpaceClass
//
RelSPJDynOptimizer::
PlanSpaceClass::PlanSpaceClass(unsigned char NumRels,
			       RelSPJDynOptimizer* DynOpt)
{
 Opt = DynOpt;
 RelCount = NumRels;
 SubPlanArray = new DynPlanTable[RelCount];
 for (int i = 0; i < RelCount; i++) {
   SubPlanArray[i].Init(i+1, this);
  }
}

RelSPJDynOptimizer::
PlanSpaceClass::~PlanSpaceClass()
{
 if (RelCount && SubPlanArray) delete [] SubPlanArray;
}

XxxErrCode RelSPJDynOptimizer::
PlanSpaceClass::GeneratePlans(const XxxDBEnv& DBEnv)
{
 // standard system R optimizer: go bottom up building
 // larger and larger plans
 for (int i = 0; i < RelCount; i++) {
   if (generateSubPlans(DBEnv, 
			SubPlanArray[i]) != XXX_OK) {
     char buf[50];
     sprintf(buf, "Error generating subplans of length %d", i);
     REL_NOTE_ERROR(REL_BADOP, buf);
     return XXX_NOT_OK;
    }
  }

 return XXX_OK;
}

XxxErrCode RelSPJDynOptimizer::
PlanSpaceClass::generateSubPlans(const XxxDBEnv& Env,
				 DynPlanTable& PlanTab)
{
 if (PlanTab.NumRels > 1) {
   // plan joins
   return generateJoinPlans(Env, PlanTab);
  }
 else {
   return generateRelPlans(Env, PlanTab);
  }
}

/*********
XxxErrCode RelSPJDynOptimizer::
PlanSpaceClass::generateJoinPlans(const XxxDBEnv& Env,
				  DynPlanTable& PlanTab)
{
  if (XXX_THISRELENV()->DebugOpt) {
    cerr << "Generating join plans of length " << PlanTab.NumRels << endl;
  }
 // for each plan of size PlanTab.NumRels-1,
 XxxSetCursor Csr;
 RelOptEngine::LogProperties DummyLogProps; // dummy. invalid log properties
 if (SubPlanArray[PlanTab.NumRels-2].
     OpenLPlanScan(DummyLogProps, Csr) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error opening scan ");
   return XXX_NOT_OK;
  }
 while (1) {
   LogicalPlan* LPlan;
   if (SubPlanArray[PlanTab.NumRels-2].
       GetNextLPlan(Csr, LPlan) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, 
		    "Error getting next logical plan");
     return XXX_NOT_OK;
    }
   if (!Csr.IsValid() || !LPlan) break;

   // for each plan of size 1
   XxxSetCursor Csr1;
   RelOptEngine::LogProperties DummyLogProps1; // dummy. invalid log properties
   if (SubPlanArray[0].OpenLPlanScan(DummyLogProps1,
				     Csr1) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error opening scan ");
     return XXX_NOT_OK;
    }
   while (1) {
     LogicalPlan* LPlan1;
     if (SubPlanArray[0].GetNextLPlan(Csr1, 
				      LPlan1) != XXX_OK) {
       REL_NOTE_ERROR(REL_BADOPT, "Error getting next plan option");
       return XXX_NOT_OK;
      }
     if (!Csr1.IsValid() || !LPlan1) break;

     // first check that the two plans are joinable
     XxxBool Joinable = XXX_FALSE;
     XxxBitMap NewPreds(Opt->NumPreds);
     RelOptEngine::LogProperties 
       JoinLogProps(Opt->BlockPlan->ChildArrayLen,
		    Opt->NumPreds);
     if (Opt->JoinsWith(LPlan->LogProps, // input
			LPlan1->LogProps, // input
			JoinLogProps, // output
			NewPreds, // output
			Joinable) != XXX_OK) {
       REL_NOTE_ERROR(REL_BADOPT, "Error checking joinable status");
       return XXX_NOT_OK;
      }
     if (!Joinable) continue;

     if (Opt->generateJoinPlans(Env, LPlan, 
				LPlan1, JoinLogProps,
				NewPreds, PlanTab) != XXX_OK) {
       REL_NOTE_ERROR(REL_BADOPT, "Error generating join plans");
       return XXX_NOT_OK;
      }
    } // while (1)
   if (SubPlanArray[0].CloseLPlanScan(Csr1) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error closing scan ");
     return XXX_NOT_OK;
    }
  } // inner loop
 if (SubPlanArray[PlanTab.NumRels-2].CloseLPlanScan(Csr) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error closing scan ");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}
********/

XxxErrCode RelSPJDynOptimizer::
PlanSpaceClass::generateJoinPlans(const XxxDBEnv& Env,
				  DynPlanTable& PlanTab)
{
  if (XXX_THISRELENV()->DebugOpt) {
    cerr << "Generating join plans of length " << PlanTab.NumRels << endl;
  }
 XxxBool IgnoreExpPreds = XXX_FALSE;
 if (XXX_THISRELENV()->ExpPredStrategy == XxxRelEnv::PULLUP_EP)
   IgnoreExpPreds = XXX_TRUE;
 double MaxCheapPredCost = XXX_THISRELENV()->CheapPredMax;

 // for each plan of size PlanTab.NumRels-1,
 XxxSetCursor Csr;
 RelOptEngine::LogProperties DummyLogProps; // dummy. invalid log properties
 if (SubPlanArray[PlanTab.NumRels-2].
     OpenLPlanScan(DummyLogProps, Csr) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error opening scan ");
   return XXX_NOT_OK;
  }
 while (1) {
   LogicalPlan* LPlan;
   if (SubPlanArray[PlanTab.NumRels-2].
       GetNextLPlan(Csr, LPlan) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, 
		    "Error getting next logical plan");
     return XXX_NOT_OK;
    }
   if (!Csr.IsValid() || !LPlan) break;

   // for each plan of size 1
   XxxSetCursor Csr1;
   RelOptEngine::LogProperties DummyLogProps1; // dummy. invalid log properties
   if (SubPlanArray[0].OpenLPlanScan(DummyLogProps1,
				     Csr1) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error opening scan ");
     return XXX_NOT_OK;
    }
   while (1) {
     LogicalPlan* LPlan1;
     if (SubPlanArray[0].GetNextLPlan(Csr1, 
				      LPlan1) != XXX_OK) {
       REL_NOTE_ERROR(REL_BADOPT, "Error getting next plan option");
       return XXX_NOT_OK;
      }
     if (!Csr1.IsValid() || !LPlan1) break;

     // first check that the two plans are joinable
     XxxBool Joinable = XXX_FALSE;
     XxxBitMap NewPreds(Opt->NumPreds);
     RelOptEngine::LogProperties 
       JoinLogProps(Opt->BlockPlan->ChildArrayLen,
		    Opt->NumPreds);
     if (Opt->JoinsWith(LPlan->LogProps, // input
			LPlan1->LogProps, // input
			IgnoreExpPreds, // input
			MaxCheapPredCost, // input
			JoinLogProps, // output
			NewPreds, // output
			Joinable) != XXX_OK) {
       REL_NOTE_ERROR(REL_BADOPT, "Error checking joinable status");
       return XXX_NOT_OK;
      }
     // what if a cross-product is needed ?? uh-oh: check for it
     if (!Joinable && !XXX_THISRELENV()->TryCrossProduct && Opt->linearJoinGraph) continue;

     // find out which rel is the Inner relation
     unsigned char InnerRelId;
     for (InnerRelId = 0; InnerRelId < LPlan1->LogProps.Rels.ArraySize; InnerRelId++) {
       if (LPlan1->LogProps.Rels.GetBit(InnerRelId)) break;
     }
     const RelCatalogInfo* InnerCatInfo = Opt->ChildCat(InnerRelId);
     if (Opt->generateJoinPlans(Env, LPlan, LPlan1, 
				InnerCatInfo, SubPlanArray[0],
				JoinLogProps, NewPreds, PlanTab) != XXX_OK) {
       REL_NOTE_ERROR(REL_BADOPT, "Error generating join plans");
       return XXX_NOT_OK;
      }
    } // while (1)
   if (SubPlanArray[0].CloseLPlanScan(Csr1) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error closing scan ");
     return XXX_NOT_OK;
    }
  } // inner loop
 if (SubPlanArray[PlanTab.NumRels-2].CloseLPlanScan(Csr) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error closing scan ");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}


XxxErrCode RelSPJDynOptimizer::
PlanSpaceClass::generateRelPlans(const XxxDBEnv& Env,
				 DynPlanTable& PlanTab)
{
  if (XXX_THISRELENV()->DebugOpt) {
    cerr << "Generating rel plans " << endl;
  }
 XxxBool IgnoreExpPreds = XXX_FALSE;
 if (XXX_THISRELENV()->ExpPredStrategy == XxxRelEnv::PULLUP_EP)
   IgnoreExpPreds = XXX_TRUE;
 double MaxCheapPredCost = XXX_THISRELENV()->CheapPredMax;

 // plan single table access
 // plans for individual relation access methods
 // for each relation
 for (unsigned char i = 0; i < RelCount; i++) {
   RelOptEngine::LogProperties LogProps(RelCount, Opt->NumPreds);
   // fill in the fact that this plan deals with relation i
   if (Opt->SetRelLProps(i, 
			 ((RelCount > 1)?IgnoreExpPreds:XXX_FALSE),
			 MaxCheapPredCost, 
			 LogProps) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error creating logprops");
     return XXX_NOT_OK;
    }
   if (Opt->generateRelPlans(Env, LogProps, i, PlanTab)
       != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error generating phys rel plan");
     return XXX_NOT_OK;
    }
  }
 return XXX_OK;
}

// scan over plans options
XxxErrCode RelSPJDynOptimizer::
PlanSpaceClass::OpenPPlanScan(XxxSetCursor& Csr) const
{
 Csr.NumberRead = 0;
 Csr.AtEnd = XXX_FALSE;
 planScanHandle* ScanHandle = 
   new planScanHandle(RelCount, Opt->NumPreds);
 ScanHandle->LSpec.Rels.SetAll();
 ScanHandle->LSpec.Preds.SetAll();

 Csr.State = ScanHandle;
 if (SubPlanArray[RelCount-1].
     OpenPPlanScan(ScanHandle->LSpec,
		   ScanHandle->Csr) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error opening scan on plantable");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}
XxxErrCode RelSPJDynOptimizer::
PlanSpaceClass::GetNextPPlan(XxxSetCursor& Csr,
			     PhysicalPlan*& Plan) const
{
 planScanHandle* ScanHandle = (planScanHandle*)(Csr.State);
 if (SubPlanArray[RelCount-1].
     GetNextPPlan(ScanHandle->Csr, Plan) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error getting next from plantable");
   return XXX_NOT_OK;
  }
 if (!ScanHandle->Csr.IsValid()) {
   Csr.AtEnd = XXX_TRUE;
   return XXX_OK;
  }
 Csr.NumberRead++;
 return XXX_OK;
}
XxxErrCode RelSPJDynOptimizer::
PlanSpaceClass::ClosePPlanScan(XxxSetCursor& Csr) const
{
 planScanHandle* ScanHandle = (planScanHandle*)(Csr.State);
 if (SubPlanArray[RelCount-1].
     ClosePPlanScan(ScanHandle->Csr) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error closing scan on plantable");
   return XXX_NOT_OK;
  }
 delete ScanHandle;
 return XXX_OK;
}

