//
// Note: these classes are nested within RelSPJDynOptimizer (see dynspj.h)
//
// Classes defined:
//        -- PlanTable : a collection of logical plans of the same size
//        -- PlanSpace : a collection of PlanTables, one for each size
//
// Here, "size" means the number of relations joined.
//
#ifndef XXX_PLANSPACE_H
#define XXX_PLANSPACE_H

  XXX_EXTERN_CLASS_DECL(PlanSpaceClass);
  /**
   * Holds subplans of a particular "size" 
   * (size == number of tables joined)
   */
  class DynPlanTable : public RelSPJCostBasedOpt::PlanTable { 
   public:
    ////
    struct LPlanScanHandle : public XxxSetCursor::CursorState {
      const RelOptEngine::LogProperties& LogProps;
      int ScanIndex;
      LPlanScanHandle(const RelOptEngine::LogProperties& LProps) :
      LogProps(LProps) { 
	ScanIndex = 0; 
       }
     };
    ////
    struct PPlanScanHandle : public XxxSetCursor::CursorState {
      XxxSetCursor LPlanCsr; // for scan over LogicalPlan
      LogicalPlan* CurrLPlan; // current LogicalPlan

      RelOptEngine::PlanObjectList* CurrPlanList;
      //PhysicalPlan* CurrPPlan; // current phys plan 
      PPlanScanHandle() {  
	CurrLPlan = NULL;
	CurrPlanList = NULL;}
      ~PPlanScanHandle() {}
      XxxErrCode Init(const DynPlanTable* PlanTab, 
		      const RelOptEngine::LogProperties& LProps);
      XxxErrCode ShutDown(const DynPlanTable* PlanTab);
     };

    int NumRels;
    int PlanArraySize;
    int EndArray; // last entry in array
    // which plan space does this plan table belong to?
    const PlanSpaceClass* PlanSpace; 

    LogicalPlan** PlanArray;

    DynPlanTable();
    virtual ~DynPlanTable();

    void Init(int Size, const PlanSpaceClass* PSpace);
    virtual XxxErrCode AddPlan(const XxxDBEnv& Env,
			       PhysicalPlan* NewPlan,
			       int& NumPlansAdded,
			       int& NumPlansPruned);

    // scan over plan options
    XxxErrCode OpenLPlanScan(const RelOptEngine::LogProperties& LSpec,
			       XxxSetCursor& Csr) const;
    XxxErrCode GetNextLPlan(XxxSetCursor& Csr,
			    LogicalPlan*& LPlan) const;
    XxxErrCode CloseLPlanScan(XxxSetCursor& Csr) const;

    // scan over plans objects
    XxxErrCode OpenPPlanScan(const RelOptEngine::LogProperties& LSpec,
			     XxxSetCursor& Csr) const;
    XxxErrCode GetNextPPlan(XxxSetCursor& Csr,
			    PhysicalPlan*& Plan) const;
    XxxErrCode ClosePPlanScan(XxxSetCursor& Csr) const;

    XxxErrCode PrintStats(ostream& OStr) const;
    XxxErrCode Print(ostream& OStr) const;
   };

  /**
   * Holds subplans of all sizes.
   */
  class PlanSpaceClass { 
   public:
    unsigned char RelCount;
    RelSPJDynOptimizer* Opt;

    DynPlanTable* SubPlanArray; // array of PlanTables, 

    DynPlanTable& GetPlanTable(unsigned int PlanSize) {
      return SubPlanArray[PlanSize-1];
     }
    PlanSpaceClass(unsigned char NumRels,
		   RelSPJDynOptimizer* DynOpt);
    ~PlanSpaceClass();
    XxxErrCode GeneratePlans(const XxxDBEnv& Env);
    void Print(ostream& OStr) const;
    // XxxErrCode PrintStats(ostream& OStr) const;
    // scan over plan objects: called after opt is complete
    XxxErrCode OpenPPlanScan(XxxSetCursor& Csr) const;
    XxxErrCode GetNextPPlan(XxxSetCursor& Csr,
			    PhysicalPlan*& Plan) const;
    XxxErrCode ClosePPlanScan(XxxSetCursor& Csr) const;

   private:
    struct planScanHandle : public XxxSetCursor::CursorState {
      RelOptEngine::LogProperties LSpec;
      XxxSetCursor Csr;
      planScanHandle(int NumRels = 0, int NumPreds = 0) 
	: LSpec(NumRels, NumPreds) {}
     };

    // all plans for a particular PlanTable
    XxxErrCode generateSubPlans(const XxxDBEnv& Env,
				DynPlanTable& PlanTable);
    
    // all plans of k joins
    XxxErrCode generateJoinPlans(const XxxDBEnv& Env,
				 DynPlanTable& PlanTable);

    // all plans involving base relation access
    XxxErrCode generateRelPlans(const XxxDBEnv& Env,
				DynPlanTable& PlanTable);
   };
#endif
