//
// General TODO: it would be good to replace the hardcoded join
// algorithms by an extensible table of join algorithms
//
#include <iostream.h>
#include <math.h>

#include "relerror.h"
#include "relquery.h"
#include "planspj.h"
#include "naivespj.h"
#include "seminaivespj.h"
#include "greedyspj.h"
#include "mykbz.h"
#include "dynspj.h"
#include "relplanop.h"
#include "relenv.h"

XxxErrCode RelOptEngine::planSPJBlock(RelQueryNode* Node,
				      unsigned char BlockIdent,
				      OptQueryEnv  *Env,
				      RelBlockPlan*& Plan)
{
 ASSERT((Node->Arity() > 0));

 RelBlockPlan** ChildPlans = new RelBlockPlan*[Node->Arity()];

 // first find plans for all the children: they should have
 // already been computed by the depth-first postfix travesal
 for (int i = 0; i < Node->Arity(); i++) {
   RelQueryNode* ChildNode = (RelQueryNode*)(Node->GetChild(i));
   XxxBool ChildFound = XXX_FALSE;
   if ((findChildBlockPlan(ChildNode, Env, 
			   ChildFound, ChildPlans[i]) != XXX_OK) ||
       (ChildFound != XXX_TRUE)) {
     REL_NOTE_ERROR(REL_BADOPT, "Unable to find plan for child of SPJ");
     delete [] ChildPlans;
     return XXX_NOT_OK;
    }
  } // for all children

 XxxErrCode rc = XXX_OK;
 RelSPJOptimizer::RelSPJBlockPlan* SPJPlan = 
   new RelSPJOptimizer::RelSPJBlockPlan(rc, BlockIdent,
					Node->Op, Node->Arity(), 
					ChildPlans, Node->CatalogInfo,
					Env);
 delete [] ChildPlans;
 
 Plan = SPJPlan;

 if (rc != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Error optimizing SPJ block");
  }
 return rc;
}

//
// Methods of RelSPJBlockPlan
//
RelSPJOptimizer::
RelSPJBlockPlan::RelSPJBlockPlan(XxxErrCode& RC,
				 unsigned char BlockIdent,
				 RelOperator* Op, 
				 int NumChildren,
				 RelBlockPlan* Children[],
				 RelCatalogInfo* CatInfo,
				 RelOptEngine::OptQueryEnv* Env)
: RelBlockPlan(RC, BlockIdent, NumChildren, Children, CatInfo, Env)
{
 Opt = NULL;

 if (RC != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOP, "Error in RelBlockPlan constructor");
   return;
  }

 if (NumChildren < 2) {
   // for single table queries, no need for fancy join optimizer
   Opt = new RelSPJNaiveOptimizer(RC,
				  ((RelSPJOp*)Op)->TargetList,
				  ((RelSPJOp*)Op)->Condition,
				  ((RelSPJOp*)Op)->Distinct,
				  this, Env);
  }
 else {
   RC = XXX_NOT_OK;

   XxxRelEnv* REnv =((XxxRelEnv*)(XXX_THISENGENV(RelationEngineId)));
   switch (REnv->JoinOrder) {
    case 0:
     Opt = new RelSPJNaiveOptimizer(RC,
				    ((RelSPJOp*)Op)->TargetList,
				    ((RelSPJOp*)Op)->Condition,
				    ((RelSPJOp*)Op)->Distinct,
				    this, Env);
     break;
    case 1:
     Opt = new RelSPJSemiNaiveOpt(RC,
				  ((RelSPJOp*)Op)->TargetList,
				  ((RelSPJOp*)Op)->Condition,
				  ((RelSPJOp*)Op)->Distinct,
				  this, Env);
     break;
    case 2:
     Opt = new RelSPJGreedyOpt(RC,
			       ((RelSPJOp*)Op)->TargetList,
			       ((RelSPJOp*)Op)->Condition,
			       ((RelSPJOp*)Op)->Distinct,
			       this, Env);
     break;
    case 3:
     Opt = new RelSPJMyKBZOpt(RC,
			      ((RelSPJOp*)Op)->TargetList,
			      ((RelSPJOp*)Op)->Condition,
			      ((RelSPJOp*)Op)->Distinct,
			      this, Env);
     break;
    case 4:
     Opt = new RelSPJDynOptimizer(RC,
				  ((RelSPJOp*)Op)->TargetList,
				  ((RelSPJOp*)Op)->Condition,
				  ((RelSPJOp*)Op)->Distinct,
				  this, Env);
     break;
    default:
     // SHOULD NEVER GET HERE!!
     ASSERT((1 == 0));
    }
  }

 if (!Opt->IsValid() || (RC != XXX_OK)) {
   REL_NOTE_ERROR(REL_BADOP, "Invalid optimizer");
   if (Opt) delete Opt;
   Opt = NULL;
   return;
  }

 if ((RC = Opt->Optimize()) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOP, "Error during optimization");
   if (Opt) delete Opt;
   Opt = NULL;
   return;
  }
 // do not delete Opt here: it carries the state of the optimizer

 RC = XXX_OK;
}

RelSPJOptimizer::
RelSPJBlockPlan::~RelSPJBlockPlan()
{
 if (Opt) delete Opt;
}

void RelSPJOptimizer::
RelSPJBlockPlan::Print(ostream& OStr) const
{
 OStr << "(SPJ PLAN";
 if (Opt && Opt->IsValid()) {
   // Opt->PrintPlans(OStr);
  }
 else {
   OStr << "-- Invalid";
  }
 OStr << ")";

 RelBlockPlan::Print(OStr);
}

XxxErrCode RelSPJOptimizer::
RelSPJBlockPlan::GeneratePlanOp(int PlanId,
				RelPlanOp*& PlanOp) const
{
 ASSERT((PlanId >= 0));

 RelOptEngine::PlanObject* ChosenPlan;
 if (GetPlan(PlanId, ChosenPlan) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Unable to find chosen plan");
   return XXX_NOT_OK;
  }
 if (Opt->GeneratePlanOp(ChosenPlan, PlanOp) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADOPT, "Unable to find chosen plan");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

XxxErrCode RelSPJOptimizer::
RelSPJBlockPlan::RedirectDependencies(const XxxValueExprPlan::VarRemapping& Remap,
				      int CorrelationHt)
{
  REL_CALL_FN(redirectDependencies(Remap, CorrelationHt),
	      REL_BADOPT, "Unable to remap plan");

  REL_CALL_FN(Opt->ExprPlanList->RedirectDependencies(Remap, CorrelationHt),
	      REL_BADOPT, "Unable to remap target list");

  for (int i = 0; i < Opt->NumPreds; i++) {
    REL_CALL_FN(Opt->PredArray[i]->
		RedirectDependencies(Remap, CorrelationHt),
		REL_BADOPT, "Unable to remap target list");
  }
  return XXX_OK;
}

//
// Methods of abstract RelSPJOptimizer class
//
RelSPJOptimizer::
RelSPJOptimizer(XxxErrCode& RC,
		XxxValueExprList* TargetList,
		XxxBooleanExpression* Condition,
		XxxBool Unique,
		RelSPJBlockPlan* SPJPlan,
		RelOptEngine::OptQueryEnv* QEnv)
{
 isValid = XXX_FALSE;
 BlockPlan = SPJPlan;
 Env = QEnv;
 Distinct = Unique;

 NumPreds = 0;
 PredArray = NULL;
 PredSelectivity = NULL;

 ExprPlanList = NULL;

 // should fill ExprPlanList and PredArray;

 // first make sure that the target list is fine
 if (TargetList && (TargetList->Optimize(ExprPlanList) != XXX_OK)) {
   REL_NOTE_ERROR(REL_BADOPT, "Unable to optimize target list");
   RC = XXX_NOT_OK;
   return;
  }

 if (Condition) {
   // Now try to break up Condition into conjunctive
   // boolean factors
   int NumFactors;
   XxxBooleanExpression** Factors;
   if (XxxBooleanExpression::
       GenerateBooleanFactors(Condition, NumFactors,
			      Factors) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Unable to split conditions");
     RC = XXX_NOT_OK;
     return;
    }

   // now keep track of the dependencies for each boolean factor,
   // and generate a plan for each
   NumPreds = NumFactors;
   PredArray = new XxxBooleanExpressionPlan*[NumPreds];
   PredDependencyInfo = new XxxBitMap[NumPreds];
   int RelArity = SPJPlan->ChildArrayLen;
   int i;
   for (i =0; i < NumPreds; i++) {
     XxxExprPlan* TmpPlan = NULL;
     if (Factors[i]->Optimize(TmpPlan) != XXX_OK) {
       REL_NOTE_ERROR(REL_BADOPT, "Error optimizing boolean factor");
       RC = XXX_NOT_OK;
       return;
      }
     PredArray[i] = (XxxBooleanExpressionPlan*)TmpPlan;

     // Get predicate dependencies.
     // The actual dependencies also include one bit
     // for "env" values. So we need to get rid of this bit.
     XxxBitMap TmpPredDep(RelArity+1);
     if (Factors[i]->GetDependencies(TmpPredDep, 0) != XXX_OK) {
       REL_NOTE_ERROR(XXX_BADEXPR, "Unable to get expr dependencies");
       RC = XXX_NOT_OK;
       return;
      }
     PredDependencyInfo[i].ChangeSize(RelArity);
     for (unsigned char j = 0; j < RelArity; j++)  {
       if (TmpPredDep.GetBit(j+1))
	 PredDependencyInfo[i].SetBit(j);
      }

    } // for each boolean factor

   // determine the selectivity information for the predicates
   XxxBooleanExpression::SelectivityStatsInfo* SelectivityStats = 
     new XxxBooleanExpression::SelectivityStatsInfo[BlockPlan->Arity()];
   for (i = 0; i < BlockPlan->Arity(); i++) {
     RelCatalogInfo* CatInfo = BlockPlan->GetChild(i)->CatalogInfo;
     SelectivityStats[i].RecSchema = CatInfo->RecordSchema();
     SelectivityStats[i].AttrStats = 
       ((CatInfo->StatsInfo())?(CatInfo->StatsInfo()->AttrStats): 0);
    }
   
   PredSelectivity = new double[NumPreds];
   for (i =0; i < NumPreds; i++) {
     PredSelectivity[i] = Factors[i]->Selectivity(SelectivityStats, 
						  BlockPlan->Arity());
     PredArray[i]->Selectivity = PredSelectivity[i];
    } 
   
   delete [] SelectivityStats;
  } // if Condition

 isValid = XXX_TRUE;
 RC = XXX_OK;
}

RelSPJOptimizer::~RelSPJOptimizer()
{
 // Delete all the structures used in the optimizer.
 // Any plans that use these structures are already
 // deleted at this stage.
 if (ExprPlanList) ExprPlanList->Destroy();

 if (NumPreds) {
   if (PredArray) {
     for (int i = 0; i < NumPreds; i++) {
       /******* Seems to cause a crash : am i reclaiming this elsewhere too?
	 TODO:L1
       if (PredArray[i])
	 delete PredArray[i];
	 *******/
      }
     delete [] PredArray;
    }
   // The destructor of XxxBitMap gets called on each element.
   if (PredDependencyInfo) {
     delete [] PredDependencyInfo;
    }
   if (PredSelectivity) {
     delete [] PredSelectivity;
    }
  } // if NumPreds
}

void RelSPJOptimizer::PrintOptInfo(ostream& OStr) const
{
 OStr << "Num Rels == " << BlockPlan->Arity() << endl;
 int i;
 for (i = 0; i < BlockPlan->Arity(); i++) {
   OStr << "\t " << i << " : ";
   OStr << ChildCat(i)->RelationName() << endl;
  }
 OStr << "Num Preds == " << NumPreds << endl;
 for (i = 0; i < NumPreds; i++) {
   OStr << "\t " << i << " : ";
   PredArray[i]->Print(OStr);
   OStr << "  dep:";
   PredDependencyInfo[i].Print(OStr);
   OStr << " sel:" << PredSelectivity[i] << endl;
  }
 OStr << " Targetlist: ";
 if (Distinct)
   OStr << " DISTINCT: ";
 ExprPlanList->Print(OStr);
 OStr << endl;
}

XxxErrCode RelSPJOptimizer::
JoinsWith(const RelOptEngine::LogProperties& LProps0,
	  const RelOptEngine::LogProperties& LProps1,
	  XxxBool AvoidExpensivePreds, 
	  double MaxCheapPredCost, // definition of "expensive"
	  RelOptEngine::LogProperties& JoinProps,
	  XxxBitMap& NewPreds,
	  XxxBool& Joinable) const
{
 Joinable = XXX_FALSE;

 ASSERT((LProps0.Rels.ArraySize == LProps1.Rels.ArraySize));
 ASSERT((LProps0.Preds.ArraySize == LProps1.Preds.ArraySize));

 // first make sure that the rels in LProps0 and LProps1 differ.
 XxxBitMap OverlapRels(LProps0.Rels.ArraySize); 
 LProps0.Rels.And(LProps1.Rels, OverlapRels);
 if (OverlapRels.NumBitsSet() > 0) {
   return XXX_OK;
  }
 // likewise for the preds
 XxxBitMap OverlapPreds(LProps0.Preds.ArraySize); 
 LProps0.Preds.And(LProps1.Preds, OverlapPreds);
 if (OverlapPreds.NumBitsSet() > 0) {
   return XXX_OK;
  }

 // set the Rels bits in JoinProps
 JoinProps.Rels.ClearAll();
 JoinProps.Rels.Or(LProps0.Rels, JoinProps.Rels);
 JoinProps.Rels.Or(LProps1.Rels, JoinProps.Rels);

 // set the Preds bits in JoinProps
 JoinProps.Preds.ClearAll();
 JoinProps.Preds.Or(LProps0.Preds, JoinProps.Preds);
 JoinProps.Preds.Or(LProps1.Preds, JoinProps.Preds);

 JoinProps.Card = LProps0.Card * LProps1.Card;
 // TODO:L2 this should take projections into account
 JoinProps.TupleSize = LProps0.TupleSize + LProps1.TupleSize;

 // look for unused predicates that will act as join preds
 XxxBool JoinPredFound = XXX_FALSE;
 XxxBool CandidatePredFound = XXX_FALSE;
 for (int i = 0; i < LProps0.Preds.ArraySize; i++) {
   if (JoinProps.Preds.GetBit(i)) continue;

   // check here is this is an expensive predicate
   if (AvoidExpensivePreds && 
       PredArray[i]->GetPlanUsage().FullCost() > MaxCheapPredCost) {
     continue;
   }

   CandidatePredFound = XXX_TRUE;
   // otherwise, this predicate may be a candidate. check further.
   // check if the dependency of the predicate
   // is fully covered by JoinProps.Rels.
   // Reuse OverlapRels for this purpose
   OverlapRels.ClearAll();
   JoinProps.Rels.And(PredDependencyInfo[i], OverlapRels);
   if (OverlapRels.Equals(PredDependencyInfo[i])) {
     // this is a join predicate
     JoinPredFound = XXX_TRUE;
     JoinProps.Preds.SetBit(i);
     NewPreds.SetBit(i);
     double PredSel = PredSelectivity[i];
     JoinProps.Card = (int)(((double)JoinProps.Card) * PredSel);
    }
  }

 if (JoinPredFound) Joinable = XXX_TRUE;

 // otherwise make sure cross products are used if absolutely needed
 if (!CandidatePredFound) Joinable = XXX_TRUE;

 return XXX_OK;
}

XxxErrCode RelSPJOptimizer::
SetRelLProps(unsigned char RelId,
	     XxxBool AvoidExpensivePreds, 
	     double MaxCheapPredCost, // definition of "expensive"
	     RelOptEngine::LogProperties& RelLProps) const
{
 RelLProps.Rels.ClearAll();
 RelLProps.Rels.SetBit(RelId);

 const RelCatalogInfo* CatInfo = ChildCat(RelId);
 if (!CatInfo) {
   REL_NOTE_ERROR(REL_BADOPT, "Invalid cat info for child");
   return XXX_NOT_OK;
 }
 RelStatsInfo* StatsInfo = CatInfo->StatsInfo();
 if (StatsInfo) {
   RelLProps.Card = StatsInfo->Cardinality;
   RelLProps.TupleSize = StatsInfo->AvgTupleSize;
 } 
 else {
   RelLProps.Card = RelStatsInfo::DefaultTableCard;
   RelLProps.TupleSize = RelStatsInfo::DefaultTupleSize;
 }

 // determine what preds if any should be applied ...
 RelLProps.Preds.ClearAll();
 for (int j = 0; j < RelLProps.Preds.ArraySize; j++) {
   if ((PredDependencyInfo[j].NumBitsSet() == 1) &&
       (PredDependencyInfo[j].GetBit(RelId))) {

     // check here is this is an expensive predicate
     if (AvoidExpensivePreds && 
	 PredArray[j]->GetPlanUsage().FullCost() > MaxCheapPredCost) {
       continue;
     }

     RelLProps.Preds.SetBit(j);
     double PredSel = PredSelectivity[j];
     RelLProps.Card = (int)(((double)RelLProps.Card) * PredSel);
   }
 }
 
 return XXX_OK;
}

//
// This is a NL join which will be pipelined with the next join.
// But we dont know how many joins lie ahead of this one.
// So here's what we do: use half the memory available and
// leave the rest for any other joins ahead in the pipeline.
void RelSPJOptimizer::
CostPNLJoin(const RelOptEngine::LogProperties& OuterLProps,
	    const RelOptEngine::PhysProperties& OuterPProps,
	    const RelOptEngine::Cost& OuterCost,
	    int TableCard, int AvgTupleSize,
	    const XxxDBEnv& DBEnv,
	    RelOptEngine::Cost& OutputCost,
	    int TuplesPerBlock)
{
  OutputCost.SetMin();

  int NumInvocs = OuterLProps.Card / TuplesPerBlock;
  if (NumInvocs == 0) NumInvocs = 1;
  RelOptEngine::Cost TempCost;
  CostFileScan(TableCard, AvgTupleSize, XXX_TRUE, DBEnv, TempCost);
  OutputCost.Add(TempCost, NumInvocs, NumInvocs);

  // add "in-memory" cost of join
  OutputCost.FullUsage.Proc += TableCard * TuplesPerBlock * 100;
  // add Outer cost
  OutputCost.Add(OuterCost);
}

void RelSPJOptimizer::
CostTNLJoin(const RelOptEngine::LogProperties& OuterLProps,
	    const RelOptEngine::PhysProperties& OuterPProps,
	    const RelOptEngine::Cost& OuterCost,
	    int TableCard, int AvgTupleSize,
	    const XxxDBEnv& DBEnv,
	    RelOptEngine::Cost& OutputCost)
{
  OutputCost.SetMin();
  int NumInvocs = OuterLProps.Card;
  RelOptEngine::Cost TempCost;
  // the first file scan must pay the full price
  CostFileScan(TableCard, AvgTupleSize, XXX_TRUE, DBEnv, TempCost);
  OutputCost.Add(TempCost);

  if (NumInvocs > 1) {
    int BufPagesAvailable = DBEnv.BufPoolSize/DBEnv.MPL - OuterCost.FullUsage.Buf.PagesInUse;
    int NumPagesInner = (int)(((double)TableCard * (AvgTupleSize + DBEnv.RecordOverhead)) / DBEnv.PageSize);
    int NumPagesUsed;
    // buffer usage: use half of available buffers
    if (NumPagesInner < BufPagesAvailable/2) {
      NumPagesUsed = NumPagesInner;
    }
    else {
      // isn't accurate, since we use LRU! 
      NumPagesUsed = BufPagesAvailable/2;
    }

    OutputCost.FullUsage.Buf.PagesUsed += NumPagesUsed;
    OutputCost.FullUsage.Buf.PagesInUse += NumPagesUsed;

    double InnerScanFraction = 1.0 - ((double)NumPagesUsed)/NumPagesInner;
    // add the cost for the extra scans of the inner
    OutputCost.Add(TempCost, (NumInvocs-1)*InnerScanFraction, 
		   (NumInvocs-1), 0.0, 0.0);
  } // if (NumInvocs > 1)

  // add Outer cost
  OutputCost.Add(OuterCost);
}

void RelSPJOptimizer::
CostINLJoin(const RelOptEngine::LogProperties& OuterLProps,
	    const RelOptEngine::PhysProperties& OuterPProps,
	    const RelOptEngine::Cost& OuterCost,
	    int TableCard, int AvgTupleSize,
	    const RelImplInfo::IndexImplInfo* IndxInfo,
	    const XxxMatchInfo* MatchedInfo,
	    XxxBool IsEqualityPred,
	    double Selectivity,
	    const XxxDBEnv& DBEnv,
	    RelOptEngine::Cost& OutputCost)
{
  OutputCost.SetMin();
  int NumInvocs = OuterLProps.Card;

  RelOptEngine::Cost TempCost;
  CostIndexScan(TableCard, AvgTupleSize, IndxInfo, MatchedInfo, 
		IsEqualityPred, Selectivity, DBEnv, TempCost);
  OutputCost.Add(TempCost, NumInvocs, NumInvocs);
  // add Outer cost
  OutputCost.Add(OuterCost);
}

void RelSPJOptimizer::
CostSMJoin(const RelOptEngine::LogProperties& OuterLProps,
	   const RelOptEngine::PhysProperties& OuterPProps,
	   const RelOptEngine::Cost& OuterCost,
	   XxxBool ToSortPlan0,
	   int TableCard, int AvgTupleSize,
	   XxxBool ToSortPlan1,
	   const XxxDBEnv& DBEnv,
	   RelOptEngine::Cost& OutputCost,
	   int& SortRunLengthOuter,
	   int& SortRunLengthInner)
{
  OutputCost.SetMin();
  // cost of sorting each relation + cost of merging
  RelOptEngine::Cost TempCost, TempCost1;

  if (ToSortPlan0) {
    CostSort(OuterLProps.Card, OuterLProps.TupleSize, OuterCost, DBEnv, TempCost);
    OutputCost.Add(TempCost);
  }
  // need to use a scan over this current relation
  CostFileScan(TableCard, AvgTupleSize, XXX_TRUE, DBEnv, TempCost);
  if (ToSortPlan1) {
    CostSort(TableCard, AvgTupleSize, TempCost, DBEnv, TempCost1);
    OutputCost.Add(TempCost1);
  }
  else
    OutputCost.Add(TempCost);

  // now the cost of the merge scan
  CostFileScan(TableCard, AvgTupleSize, XXX_TRUE, DBEnv, TempCost);
  OutputCost.Add(TempCost);
  CostFileScan(OuterLProps.Card, OuterLProps.TupleSize, XXX_TRUE, DBEnv, TempCost);
  OutputCost.Add(TempCost);

  // use 3 pages of buffer for merge
  OutputCost.FullUsage.Buf.PagesUsed += 3;
  OutputCost.FullUsage.Buf.PagesInUse += 3;
}

//Begin: Moustafa
void RelSPJOptimizer::
CostStreamScan(int TableCard, int AvgTupleSize,
	       XxxBool Sequential,
	       const XxxDBEnv& DBEnv,
	       RelOptEngine::Cost& OutputCost)
{
  OutputCost.SetMin();
  // TODO: must think in providing the correct cost
  OutputCost.FullUsage.Disk.SeqIO  = 0;
  OutputCost.FullUsage.Disk.RandIO = 0;
  OutputCost.FullUsage.Buf.PagesUsed = 0;
  OutputCost.FullUsage.Buf.PagesInUse = 0;
  OutputCost.FullUsage.Proc = 0;
}

//End: Moustafa

void RelSPJOptimizer::
CostFileScan(int TableCard, int AvgTupleSize,
	     XxxBool Sequential,
	     const XxxDBEnv& DBEnv,
	     RelOptEngine::Cost& OutputCost)
{
  OutputCost.SetMin();
  if (Sequential) {
    OutputCost.FullUsage.Disk.SeqIO = (int)
      (((double)TableCard * (AvgTupleSize + DBEnv.RecordOverhead)) / DBEnv.PageSize);
    if (OutputCost.FullUsage.Disk.SeqIO == 0) OutputCost.FullUsage.Disk.SeqIO = 1;
    // assign cost of one random I/O for initial latency
    OutputCost.FullUsage.Disk.RandIO = 1;
  }
  else {
    OutputCost.FullUsage.Disk.RandIO = (int)
      (((double)TableCard * (AvgTupleSize + DBEnv.RecordOverhead)) / DBEnv.PageSize);
    if (OutputCost.FullUsage.Disk.RandIO == 0) OutputCost.FullUsage.Disk.RandIO = 1;
  }
  // generous estimate of CPU cost per filescan
  OutputCost.FullUsage.Proc = TableCard * 500;
  // use 1 page of buffer
  OutputCost.FullUsage.Buf.PagesUsed = 1;
  OutputCost.FullUsage.Buf.PagesInUse = 1;
}

void RelSPJOptimizer::
CostIndexScan(int TableCard, int AvgTupleSize,
	      const RelImplInfo::IndexImplInfo* IndxInfo,
	      const XxxMatchInfo* MatchedInfo,
	      XxxBool IsEqualityPred,
	      double Selectivity,
	      const XxxDBEnv& DBEnv,
	      RelOptEngine::Cost& OutputCost)
{
  OutputCost.SetMin();

  int IndexFanout; // assume same fanout for leaf and non-leaf pages
  double IndexFillFactor;
  int IndexEntrySize;
  // should be computed from IndxInfo TODO:LOpt : for now, assume it
  IndexEntrySize = 24; // bytes
  IndexFillFactor= 0.67;
  IndexFanout = (int)((DBEnv.PageSize * IndexFillFactor)/ IndexEntrySize) ;
  int IndexHeight = (int)(ceil(log((double)TableCard)/log((double)IndexFanout)));

  int BufPagesAvailable = DBEnv.BufPoolSize/DBEnv.MPL;

  // buffer usage: try to keep most of the index leaf pages in buffer, but
  // dont exceed half of BufPagesAvailable (since we are pipelining)
  int NumIndexLeafPages = (int)(ceil(((double)TableCard)/((double)IndexFanout)));
  int BufPagesUsed;
  double IndexInMemRatio;
  if (NumIndexLeafPages < BufPagesAvailable/2) {
    BufPagesUsed = NumIndexLeafPages;
    IndexInMemRatio = 1;
  }
  else {
    BufPagesUsed = BufPagesAvailable/2;
    IndexInMemRatio = ((double)BufPagesUsed)/((double)NumIndexLeafPages);
  }

  OutputCost.FullUsage.Buf.PagesUsed += BufPagesUsed;
  OutputCost.FullUsage.Buf.PagesInUse += BufPagesUsed;

  OutputCost.FullUsage.Disk.RandIO = (int)ceil(IndexHeight*IndexInMemRatio);
  // arbitrary estimate of index traversal CPU cost
  OutputCost.FullUsage.Proc = IndexHeight * 1000;

  RelOptEngine::Cost TempCost;
  if (IndxInfo && IndxInfo->Unique && IsEqualityPred) {
    // unique index with equality pred
    TempCost.FullUsage.Disk.RandIO += 1;
    OutputCost.Add(TempCost);
  }
  else if (IndxInfo && IndxInfo->Clustered) {
    // sequential I/O on file
    CostFileScan(TableCard, AvgTupleSize, XXX_TRUE, DBEnv, TempCost);
    OutputCost.Add(TempCost, Selectivity, Selectivity);
  }
  else {
    // random I/O on file
    CostFileScan(TableCard, AvgTupleSize, XXX_FALSE, DBEnv, TempCost);
    OutputCost.Add(TempCost, Selectivity, Selectivity);
  }
}

//
// This assumes that all passes of the sort algorithm use virtual memory,
// not the database buffer pool
//
void RelSPJOptimizer::
CostSort(int TableCard, int AvgTupleSize,
	 const RelOptEngine::Cost& OuterCost,
	 const XxxDBEnv& DBEnv,
	 RelOptEngine::Cost& OutputCost)
{
  OutputCost.SetMin();
  RelOptEngine::Cost TempCost;

  int NumPasses;
  int PagesUsed;
  int NumPages = (int)(((double)TableCard * (AvgTupleSize + DBEnv.RecordOverhead)) / DBEnv.PageSize);
  int MemPagesAvailable = 
    DBEnv.MemorySize/DBEnv.MPL - OuterCost.FullUsage.Mem.PagesInUse;
  if (NumPages < MemPagesAvailable) {
    // 1-pass sort
    NumPasses = 1;
    PagesUsed = NumPages;
  }
  else {
    NumPasses = (int)(ceil(log((double)NumPages)/log((double)MemPagesAvailable)));
    PagesUsed = MemPagesAvailable;
  }

  // The first pass (generation of sorted runs) has a sequential
  // read and a sequential write of data. The read is passed in as
  // OuterCost
  // Add Outer cost: since sort does not pipeline, the memory and
  // buffer usage of its inputs will be thrown away.
  OutputCost.Add(OuterCost, 1.0, 1.0, 0.0, 0.0);
  // now add the output cost for the first pass
  CostFileScan(TableCard, AvgTupleSize, XXX_TRUE, DBEnv, TempCost);
  OutputCost.Add(TempCost, NumPasses, NumPasses);

  // Each subsequent merge pass involves a random read + a seq write of 
  // all the data.
  if (NumPasses > 1) {
    CostFileScan(TableCard, AvgTupleSize, XXX_TRUE, DBEnv, TempCost);
    OutputCost.Add(TempCost, NumPasses-1, NumPasses-1, 0.0, 0.0);
    CostFileScan(TableCard, AvgTupleSize, XXX_FALSE, DBEnv, TempCost);
    OutputCost.Add(TempCost, NumPasses-1, NumPasses-1, 0.0, 0.0);
  }
  
  // add CPU costs of each phase (sort-run gen + merge)
  // for now: some random guesses (not even well-founded, just placeholders)
  OutputCost.FullUsage.Proc += TableCard * 300;
  OutputCost.FullUsage.Proc += TableCard * 100; // merge cost per tuple

  // use 1 output page of buffer
  OutputCost.FullUsage.Buf.PagesUsed += 1;
  OutputCost.FullUsage.Buf.PagesInUse += 1;
  // total memory pages used in this operation
  OutputCost.FullUsage.Mem.PagesUsed = PagesUsed;
  OutputCost.FullUsage.Mem.PagesInUse = 0;
}




