#ifndef XXX_PLANSPJ_H
#define XXX_PLANSPJ_H

#include "relopt.h"
#include "misc.h"
#include "values.h"

/**
 * This is the root class of all SPJ optimizers
 */
class RelSPJOptimizer {
 public:
  XXX_EXTERN_CLASS_DECL(PlanOptions);
  /**
   *
   */
  class RelSPJBlockPlan : public RelBlockPlan { 
   public:
    RelSPJOptimizer* Opt; // this is the "optimizer" object
    virtual void Print(ostream& OStr) const;
    RelSPJBlockPlan(XxxErrCode& RC,
		    unsigned char BlockIdent,
		    RelOperator*, 
		    int NumChildren,
		    RelBlockPlan* Children[],
		    RelCatalogInfo* CatInfo,
		    RelOptEngine::OptQueryEnv* Env);
    virtual ~RelSPJBlockPlan();
    virtual XxxErrCode GeneratePlanOp(int PlanId,
				      RelPlanOp*& PlanOp) const;
    // This method replaces the dependencies of all correlation
    // variables (i.e XxxUnknownVars) within this plan.
    virtual XxxErrCode RedirectDependencies(const XxxValueExprPlan::VarRemapping& Remap,
					    int CorrelationHt = 0);
   };

  struct PredPlanStruct {
    XxxBooleanExpressionPlan* InitPlan;
    XxxBooleanExpressionPlan** SelPredPlans;
    XxxBooleanExpressionPlan** JoinPredPlans;
    XxxBooleanExpressionPlan** ResidPredPlans;
    PredPlanStruct() {
      InitPlan = NULL;
      SelPredPlans = NULL;
      JoinPredPlans = NULL;
      ResidPredPlans = NULL;
     }
    ~PredPlanStruct() {}
    void Reclaim() {
      if (SelPredPlans) delete [] SelPredPlans;
      if (JoinPredPlans) delete [] JoinPredPlans;
      if (ResidPredPlans) delete [] ResidPredPlans;
     }
   };
  //// dummy superclass that indicates glue information for a plan
  class PlanGlue {
   public:
    virtual ~PlanGlue() {}
   };
  //// Glue information for sort merge join
  class SMPlanGlue : public PlanGlue {
   public:
    int LeftRelIndex, RightRelIndex;
    XxxBool SortLeft, SortRight;
    XxxSortInfo LeftSInfo, RightSInfo;
    //// indicate whether the inputs need a temp to sort them
    XxxBool MaterializeLtInput, MaterializeRtInput;
    //// indicate max run sizes for sorting
    int LeftRunSize, RightRunSize;
    SMPlanGlue(int NumAttrs) 
      : LeftSInfo(NumAttrs), RightSInfo(NumAttrs) {}
    virtual ~SMPlanGlue() {}
   };

  XxxBool isValid;
 public:
  RelOptEngine::OptQueryEnv *Env;
  RelSPJBlockPlan* BlockPlan;
  XxxValExprPlanList* ExprPlanList; // list of target plans
  XxxBool Distinct;
  int NumPreds; // conjunctive boolean predicates
  XxxBooleanExpressionPlan** PredArray; // array of pred plans
  XxxBitMap* PredDependencyInfo; // array of predicate dependencies 
  double* PredSelectivity; // array of predicate selectivities

  const RelCatalogInfo* ChildCat(int i) const {
    if (i >= BlockPlan->Arity()) return NULL;
    else return BlockPlan->GetChild(i)->CatalogInfo;
   }
  
  RelSPJOptimizer(XxxErrCode& RC,
		  XxxValueExprList* TargetList,
		  XxxBooleanExpression* Condition,
		  XxxBool Unique,
		  RelSPJBlockPlan* SPJPlan,
		  RelOptEngine::OptQueryEnv* QEnv);
  virtual ~RelSPJOptimizer();

  const RelSpecParameters* OptSpecs() const { return Env->OptSpecs;}
    
  virtual XxxErrCode Optimize() = 0;
  // dump the contents of the RelSPJOptimizer
  virtual void PrintOptInfo(ostream& OStr) const;

  virtual XxxErrCode GeneratePlanOp(RelOptEngine::PlanObject* Plan,
				    RelPlanOp*& PlanOp) const = 0;

  XxxBool IsValid() { return isValid;}
  virtual void PrintPlans(const RelOptEngine::PlanObject* Plan,
			  ostream& OStr) const = 0;
  // Do the two sub-plans represented by LProps0 and LProps1 join?
  // What are the resulting JoinProps? What NewPreds are applied?
  virtual XxxErrCode 
    JoinsWith(const RelOptEngine::LogProperties& LProps0,
	      const RelOptEngine::LogProperties& LProps1,
	      XxxBool AvoidExpensivePreds, 
	      double MaxCheapPredCost, // definition of "expensive"
	      RelOptEngine::LogProperties& JoinProps,
	      XxxBitMap& NewPreds,
	      XxxBool& Joinable) const;
  // For a table scan, what are the resulting logical props?
  virtual XxxErrCode 
    SetRelLProps(unsigned char RelId,
		 XxxBool AvoidExpensivePreds, 
		 double MaxCheapPredCost, // definition of "expensive"
		 RelOptEngine::LogProperties& RelLProps) const;

  // various costing (and resource utilization estimation) functions
  static void CostPNLJoin(const RelOptEngine::LogProperties& InputLProps,
			  const RelOptEngine::PhysProperties& InputPProps,
			  const RelOptEngine::Cost& InputCost,
			  int TableCard, int AvgTupleSize,
			  const XxxDBEnv& DBEnv,
			  RelOptEngine::Cost& OutputCost,
			  int TuplesPerBlock);
  static void CostTNLJoin(const RelOptEngine::LogProperties& InputLProps,
			  const RelOptEngine::PhysProperties& InputPProps,
			  const RelOptEngine::Cost& InputCost,
			  int TableCard, int AvgTupleSize,
			  const XxxDBEnv& DBEnv,
			  RelOptEngine::Cost& OutputCost);
  static void CostINLJoin(const RelOptEngine::LogProperties& InputLProps,
			  const RelOptEngine::PhysProperties& InputPProps,
			  const RelOptEngine::Cost& InputCost,
			  int TableCard, int AvgTupleSize,
			  const RelImplInfo::IndexImplInfo* IndxInfo,
			  const XxxMatchInfo* MatchedInfo,
			  XxxBool IsEqualityPred,
			  double Selectivity,
			  const XxxDBEnv& DBEnv,
			  RelOptEngine::Cost& OutputCost);
  static void CostSMJoin(const RelOptEngine::LogProperties& InputLProps,
			 const RelOptEngine::PhysProperties& InputPProps,
			 const RelOptEngine::Cost& InputCost,
			 XxxBool ToSortPlan0,
			 int TableCard, int AvgTupleSize,
			 XxxBool ToSortPlan1,
			 const XxxDBEnv& DBEnv,
			 RelOptEngine::Cost& OutputCost,
			 int& SortRunLengthOuter,
			 int& SortRunLengthInner);
  //Begin: Moustafa
  static void CostStreamScan(int TableCard, int AvgTupleSize,
			     XxxBool Sequential,
			     const XxxDBEnv& DBEnv,
			     RelOptEngine::Cost& OutputCost);
  //End: Moustafa

  static void CostFileScan(int TableCard, int AvgTupleSize,
			   XxxBool Sequential,
			   const XxxDBEnv& DBEnv,
			   RelOptEngine::Cost& OutputCost);
  static void CostIndexScan(int TableCard, int AvgTupleSize,
			    const RelImplInfo::IndexImplInfo* IndxInfo,
			    const XxxMatchInfo* MatchedInfo,
			    XxxBool IsEqualityPred,
			    double Selectivity,
			    const XxxDBEnv& DBEnv,
			    RelOptEngine::Cost& OutputCost);
  static void CostSort(int TableCard, int AvgTupleSize,
		       const RelOptEngine::Cost& InputCost,
		       const XxxDBEnv& DBEnv,
		       RelOptEngine::Cost& OutputCost);
  
 };

#endif
