#include <iostream.h>

#include "relerror.h"
#include "relquery.h"
#include "planstop.h"
#include "relplanop.h"
#include "prd-sort.h"

XxxErrCode RelOptEngine::planStopBlock(RelQueryNode* Node,
				       unsigned char BlockIdent,
				       OptQueryEnv  *Env,
				       RelBlockPlan *& Plan)
{
  ASSERT((Node->Arity() == 1));
  
  RelStopAfterBlockPlan* STOPPlan = NULL;
  RelBlockPlan* ChildBlockPlan = NULL;
  
  RelQueryNode* ChildNode = (RelQueryNode*)(Node->GetChild(0));
  XxxBool ChildFound = XXX_FALSE;
  if ((findChildBlockPlan(ChildNode, Env, 
			  ChildFound, ChildBlockPlan) != XXX_OK) ||
      (ChildFound != XXX_TRUE))  {
    REL_NOTE_ERROR(REL_BADOPT, "Unable to find plan for child of STOP");
    return XXX_NOT_OK;
  }
  
  XxxErrCode rc = XXX_OK;
  STOPPlan = new RelStopAfterBlockPlan(rc, BlockIdent, Node->Op, 
				       ChildBlockPlan, Node->CatalogInfo, Env);
  Plan = STOPPlan;
  return rc;
}

RelStopAfterBlockPlan::RelStopAfterBlockPlan(XxxErrCode& RC,
					     unsigned char BlockIdent,
					     RelOperator* Op, 
					     RelBlockPlan* Child,
					     RelCatalogInfo* CatInfo,
					     RelOptEngine::OptQueryEnv* Env)
  : RelBlockPlan(RC, BlockIdent, 1, &Child, CatInfo, Env)
{
  
  K  = ((RelStopAfterOp *)Op)->K;
  
  // Step 1 : Create a PhysProp .
  RelOptEngine::PhysProperties DesiredProps, *ChildProps;
  XxxBool PlanFound = XXX_FALSE;
  int ChildPlanId;
  XxxGenericPlan::PlanUsage ChildUsage;
  
  if ((Child->GetBestPlan(DesiredProps, ChildProps, ChildPlanId,
			  ChildUsage, PlanFound) != XXX_OK) ||
      (PlanFound != XXX_TRUE)) {
    REL_NOTE_ERROR(REL_BADOPT, "Cannot get child plan");
    RC = XXX_NOT_OK;
    return;
  }
 
  //
  // Step 2:  Create a PlanObject using this child plan
  //
  StopPlan* ThisPlan = new StopPlan(HighestPlanId++,
				    BlockId,
				    ChildPlanId,
				    ChildUsage);
  // add the plan to this RelBlockPlan
  AddPlan(ThisPlan);
  RC = XXX_OK;
}

RelStopAfterBlockPlan::~RelStopAfterBlockPlan()
{
}

RelStopAfterBlockPlan::
StopPlan::StopPlan(int PlanIdent,
		   unsigned char BlockIdent,
		   int ChildPlan,
		   XxxGenericPlan::PlanUsage& ChildUsage)
  :RelOptEngine::PlanObject(PlanIdent, BlockIdent)
{
  ChildPlanId = ChildPlan;
  Usage.Add(ChildUsage);
}

void RelStopAfterBlockPlan::
StopPlan::Print(ostream& OStr) const
{
  OStr << "[STOP Plan]";
}

void RelStopAfterBlockPlan::Print(ostream& OStr) const
{
 OStr << "STOP AFTER PLAN";
 
 OStr << "["<<K<<"]" ;
 
}

XxxErrCode 
RelStopAfterBlockPlan::GeneratePlanOp(int PlanId,
				      RelPlanOp*& PlanOp) const
{
  ASSERT((PlanId >= 0));
  
  RelOptEngine::PlanObject* ChosenPlan;
  if (GetPlan(PlanId, ChosenPlan) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, "Unable to find chosen plan");
    return XXX_NOT_OK;
  }
  // currently, the StopPlan contains no info, so isn't used
  StopPlan* ThePlan = (StopPlan*)(ChosenPlan);
  
  RelExecParameters* Params = NULL; 
  RelPlanOp* ChildPlanOp;
  // first generate a planop for its child
  REL_CALL_FN(GetChild(0)->GeneratePlanOp(Params, ChildPlanOp),
	      REL_BADEXEC, "Cannot generate planop for child of Stop after");

  PlanOp = new RelStopAfterPlanOp(ChildPlanOp,K,ChildPlanOp->CatalogInfo);
 
  return XXX_OK;
}

XxxErrCode RelStopAfterBlockPlan::
RedirectDependencies(const XxxValueExprPlan::VarRemapping& Remap,
		     int CorrelationHt)
{
  REL_CALL_FN(redirectDependencies(Remap, CorrelationHt),
	      REL_BADOPT, "Unable to remap plan");
  
  return XXX_OK;
}
