//Ihab
// a new file for the semin naiive ooptimizer of a KRJ block (logical operator)

#include <iostream.h>

#include "relerror.h"
#include "relquery.h"
#include "relplanop.h"
#include "relenv.h"
#include "seminaiveKRJ.h"

//
// Methods of RelKRJSemiNaiveOpt
//

XxxErrCode RelKRJSemiNaiveOpt::
GeneratePlanOp(RelOptEngine::PlanObject* Plan,
	       RelPlanOp*& PlanOp) const
{
  KRJNaivePlan* KRJPlan = (KRJNaivePlan*)(Plan);
  int NumChildren = KRJPlan->ArraySize;
  int NumJoins = NumChildren-1;
  if (NumJoins == 0) 
    return RelKRJNaiveOptimizer::GeneratePlanOp(Plan, PlanOp);
  
  // this is entered only if there is at least one join
  int i;
  
  // Step 1: Generate plan ops for all the child query blocks.
  RelPlanOp** ChildPlanOps = new RelPlanOp*[NumChildren];
  // InputArity keeps track of the arity of each input relation
  // before join reordering. This could be needed for Remapping
  int *InputArity = new int[NumChildren];
  if (getChildPlans(KRJPlan, ChildPlanOps, InputArity) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, "Error getting child plans");
    return XXX_NOT_OK;
  }
  
  // Generating a pipeline of several KRJPlanOps,
  // one for each binary join.
  // The painful part is creating a cataloginfo
  // for each intermediate result.
  // However, the good news is that the recordschema
  // is easy to determine if projections are not
  // pushed down (it is the concatenation of the
  // input recordschema).
  
 // Step 2: construct catalog info, targetlist, predplans, residplan 
 // and initplan for each join
  RelKRJOptimizer::PredPlanStruct PredPlans;
  XxxValExprPlanList* PlanList = NULL;
  RelCatalogInfo* CatInfo;
  
 // Step 3: this remap is used to alter the various expressions
  XxxExprPlan::VarRemapping Remap(NumChildren);
  if (constructRemap(KRJPlan, Remap, 2, InputArity,
		     XXX_TRUE) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, "Error constructing remap");
    return XXX_NOT_OK;
  }
  
  PlanOp = ChildPlanOps[0];
  int CurrentArity = // keeps track of arity of partial result
    ChildPlanOps[0]->CatalogInfo->RecordSchema()->Arity();
  
  // step thru all the joins (remember that all joins are binary)
  for (i = 0; i < NumJoins; i++) {
    // construct the input planops
    RelPlanOp* TmpPlanOps[2];
    TmpPlanOps[0] = PlanOp; // outer relation is the temporary
    TmpPlanOps[1] = ChildPlanOps[i+1];
    
    // Step 4: get all the predicates to be applied
    PredPlans.InitPlan = NULL;
    // InitPlan makes sense only for the first join
    if (i == 0) {
      if (getInitPlan(KRJPlan, PredPlans.InitPlan) != XXX_OK) {
	REL_NOTE_ERROR(REL_BADOPT, "Error in getInitPlan");
	return XXX_NOT_OK;
      }
    } // if (i == 0)
    
    // determine the PredPlans to be applied
    if (getJoinPlans(KRJPlan, PredPlans, 2, i, XXX_TRUE) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADOPT, "Error in getJoinPlans");
      return XXX_NOT_OK;
    }
    
   // Step 5: modify the Remap to reflect the new temporary planop
   if (changeRemap(Remap, NumChildren, 
		   i+1 /* SourceToChange*/, 1 /* New source*/, 
		   0, InputArity) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error changing remap");
     return XXX_NOT_OK;
    }
   

   // step 5.5  Modify Ordering.. Need change.. as it dependes on the order 
   XxxValExprPlanList* tmpOrderPlanList = NULL;
   XxxValExprPlanList* OrderPos= OrderPlanList;
   if(OrderPlanList){
     for (int k=0;k<NumJoins-i-1;k++)
       OrderPos = OrderPos->Next;
     tmpOrderPlanList = OrderPos->Replicate();
     if(i>0) tmpOrderPlanList->Next = NULL;
   }
   
   // Step 6: modify all the expressions using the VarRemapping
   if (exprRemap(Remap, PredPlans, 2, NULL,tmpOrderPlanList) != XXX_OK) { // Ordering is as is
     REL_NOTE_ERROR(REL_BADOPT, "Error applying remap");
     return XXX_NOT_OK;
    }
       
   // Step 7: determine the targetlist and construct the catalog info
   XxxBool ReclaimCat = XXX_FALSE; // should the catinfo be reclaimed?
   if (makeOutputInfo(i, NumJoins, Remap, TmpPlanOps,
		      PlanList, CatInfo, ReclaimCat) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error making output info");
     return XXX_NOT_OK;
    }

   //Step 8: adjust for join preds and resid preds 
   if (checkKRJ(NumChildren,
	       PredPlans, Remap, 
	       KRJPlan->JoinMethods[i]) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error adjusting preds");
     return XXX_NOT_OK;
   }
   // Step 9: generate a plan operator for the KRJ.
   // Make sure "distinct" is handled only for the last op
   XxxBool ApplyDistinct = ((i < NumJoins-1)?XXX_FALSE:Distinct);
   // Top = 1 : top in more than 2 source
   int top=0;
   if (i==NumJoins-1) top =1;
   if(OrderPlanList==NULL) top =2;  // ordering by sequence..
   if (makePlanOp(TmpPlanOps, 2, PredPlans, PlanList,tmpOrderPlanList,
		  CatInfo, KRJPlan->JoinMethods[i], 
		  ApplyDistinct, PlanOp, top) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error making plan operator");
     return XXX_NOT_OK;
    }
   if (ReclaimCat) {
     CatInfo->Destroy();
     delete CatInfo;
    }
   // Step 10: modify the Remap to reflect the new temporary planop
   if (changeRemap(Remap, NumChildren, 
		   1 /* SourceToChange*/, 0 /* New source*/, 
		   CurrentArity, InputArity) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error changing remap");
     return XXX_NOT_OK;
   }
   
   // important: update this only at the end of the loop
   // becaue it is used to update the Remap
   // The temp schemas have been changed.. so add 2 for WG and BG
   CurrentArity +=  ChildPlanOps[i+1]->CatalogInfo->RecordSchema()->Arity() + 2; 

   // Do not reclaim the pred arrays, because the plan ops use them.
   // Change this eventually : TODO:L2
   // PredPlans.Reclaim();
  } // for NumJoins (main loop creating KRJPlanOps)

 delete [] InputArity;
 delete [] ChildPlanOps;
 
 return XXX_OK;
}
//---------------------------------------------------------------------------
XxxErrCode RelKRJSemiNaiveOpt::
makeOutputInfo(int CurrentJoinPos,
	       int NumJoins,
	       XxxExprPlan::VarRemapping& Remap,
	       RelPlanOp** TmpPlanOps,
	       XxxValExprPlanList*& TmpPlanList,
	       RelCatalogInfo*& TmpCatInfo,
	       XxxBool& ReclaimCat) const
{
 if (CurrentJoinPos == NumJoins-1) {
   // if this is the last join
   TmpCatInfo = BlockPlan->CatalogInfo;
   ReclaimCat = XXX_FALSE;
   TmpPlanList = ExprPlanList;
   // modify the mapping in the targetlist
   if (TmpPlanList &&
       TmpPlanList->RedirectDependencies(Remap) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error redirecting target");
     return XXX_NOT_OK;
    }
  }
 else {
   // if it is an intermediate join
   TmpCatInfo = new RelCatalogInfo();
   // give it a dummy implinfo: needed because SortOrders are attached to it
   TmpCatInfo->implInfo = new RelImplInfo();
   // make sure to reclaim this
   ReclaimCat = XXX_TRUE;

   // create the schema.. add two fields for the WG and BG 
   const XxxRecordSchema* TmpRecSchema[3];
   TmpRecSchema[0] = TmpPlanOps[0]->CatalogInfo->RecordSchema();
   TmpRecSchema[1] = TmpPlanOps[1]->CatalogInfo->RecordSchema();
   TmpRecSchema[2] = new XxxRecordSchema(2);
   (*TmpRecSchema[2])[0].Assign("WG" , XXX_DOUBLE,NULL); 
   (*TmpRecSchema[2])[1].Assign("BG" , XXX_DOUBLE,NULL);
   TmpCatInfo->recSchema = new XxxRecordSchema(3,TmpRecSchema);
   

   //  new targetlist is the concatenation of the record schemas
   TmpPlanList = NULL;
   // Two steps: first create an XxxValueExprList, then 
   // optimize it to get an XxxValExprPlanList
   XxxValueExprList *TmpExprList = NULL;
   for (int ci =0; ci < 2; ci++) {
     const XxxRecordSchema* ChildSchema = TmpRecSchema[ci];
     for (int j = 0; j < ChildSchema->Arity(); j++) {
       XxxValueExpression* NewExpr =  new XxxUnknownValue(j, 
							  ChildSchema->GetAttr(j)->Type,
							  ChildSchema->GetAttr(j)->TypeCatInfo);
       ((XxxUnknownValue *)NewExpr)->SrcChild = ci;
       if (!TmpExprList)	 TmpExprList = new XxxValueExprList(NewExpr);
       else 	                 TmpExprList = TmpExprList->AddExpr(NewExpr);
      } // for each attribute
    } // for both children

   // add the WG and BG to the target List as double constants

   XxxValueExpression* NewExpr =  new  XxxConstDoubleValue (0.0);
   TmpExprList = TmpExprList->AddExpr(NewExpr);
   NewExpr =  new  XxxConstDoubleValue (0.0);
   TmpExprList = TmpExprList->AddExpr(NewExpr);

   // now optimize TmpExprList to create plan list
   if (TmpExprList && 
       (TmpExprList->Optimize(TmpPlanList) != XXX_OK)) {
     REL_NOTE_ERROR(REL_BADOPT, "Error optimizing tmp expr list");
     return XXX_NOT_OK;
    }
  } // else

 
 return XXX_OK;
}

//--------------------------------------------------------------------------------
XxxErrCode RelKRJSemiNaiveOpt::checkKRJ(int NumChildren,
				      RelKRJOptimizer::PredPlanStruct& TmpPredPlan,
				      XxxExprPlan::VarRemapping& Remap,
				      RelOptEngine::JoinMethod& JoinMethod) const
{
  XxxBool KRJJoinable = XXX_FALSE;
  XxxRelBooleanPlan* TmpJoinPlan = NULL;
  if (isKRJJoinable(TmpPredPlan.JoinPredPlans[1], TmpJoinPlan,
		   TmpPredPlan.ResidPredPlans[1], KRJJoinable) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADOPT, "Error checking for KRJ");
    return XXX_NOT_OK;
  }
  
  if (KRJJoinable) {
    JoinMethod = RelOptEngine::RJOIN;
    TmpPredPlan.JoinPredPlans[1] = TmpJoinPlan;
  
  // need to further remap the SelPredPlans, because for 
  // Rank Join, these selections get applied independent of
  // a join context. So the "inner" should also be treated
  // as being input #0 (i.e. it looks a bit "bushy")
  // The original Remap should not be changed for this, because it
  // needs to be preserved to accurately remap later expressions
  // that reference the inner. Instead, use a dummy Remap
    
    XxxExprPlan::VarRemapping DummyRemap(Remap.OldArity);
    DummyRemap.NewArity = Remap.NewArity;
    for (int i = 0; i < Remap.OldArity; i++) {
      DummyRemap.SrcMappingArray[i].NewSrc = Remap.SrcMappingArray[i].NewSrc;
      // dont bother with AttrMappings, since they are not used here
    }
    if (changeRemap(DummyRemap, NumChildren,  1 , 0 ,   0, NULL) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADOPT, "Error changing remap");
      return XXX_NOT_OK;
    }
    // modify the selection expressions using the VarRemapping
    if (TmpPredPlan.SelPredPlans[1] &&
	TmpPredPlan.SelPredPlans[1]->RedirectDependencies(DummyRemap) != XXX_OK){
      REL_NOTE_ERROR(REL_BADOPT, "Error redirecting predicate");
      return XXX_NOT_OK;
    }
  }
  return XXX_OK;
}
//------------------------------------------------------------------
/**
 * Modify the Remap to reflect a new temporary planop.
 */
XxxErrCode RelKRJSemiNaiveOpt::
changeRemap(XxxExprPlan::VarRemapping& Remap,
	    int NumChildren,
	    int SourceToChange,
	    int NewSourceValue,
	    int AttrIncrement,
	    int* InputArity) const
{
 for (int k=0; k < NumChildren; k++) {
   int SrcNum = Remap.SrcMappingArray[k].NewSrc;
   if (SrcNum == SourceToChange) {
     Remap.SrcMappingArray[k].NewSrc = NewSourceValue;
     // if the change in the join order also affects
     // the positions of attrs for subsequent joins,
     // then modify the remap appropriately
     if (AttrIncrement > 0) {
       for (int j = 0; j < InputArity[k]; j++)
	 Remap.SrcMappingArray[k].NewAttrMapping[j] += AttrIncrement;
      }
    } // if SrcNum == SourceToChange
  } // for k < NumChildren
 return XXX_OK;
}
