#ifndef XXX_BIDX_H
#define XXX_BIDX_H

#include "XxxIndex.h"
/**
 * Definition of interface for B+ index
 */
class ShoreBTree : public XxxIDXClass {
public:
  //// BTrees define an index range type 
  class BTreeRange : public XxxIDXClass::IndexRange  {
  public:
    BTreeRange() { 
      Data = NULL; 
      Satisfiable = XXX_TRUE;
    }
    // Destructor avoids memory leak
    ~BTreeRange() { 
      if (Data) {
	delete[] Data; Data = NULL;
       }
    }
    void Copy(const XxxIDXClass::IndexRange& R);
    
    /// These are the SHORE structures that represent bounds
    scan_index_i::cmp_t    LowerCmp;
    cvec_t   LowerBound;
    scan_index_i::cmp_t    UpperCmp;
    cvec_t   UpperBound;

    //// Memory for the bounds
    char* Data;
    int DataLen;
    int LowerBoundOffset; // within Data
    int LowerBoundSize; 
    int UpperBoundOffset; // within Data
    int UpperBoundSize; 

    //// Can this range spec be satisfied?
    /// A value of XXX_TRUE means "maybe"
    XxxBool Satisfiable;

    void CheckSatisfiability();
   };
  //// Cursor over B+ tree
  class IScanCursorInfo : public XxxIDXClass::CursorInfo { 
   public:
    scan_index_i ScanIter;
    XxxShoreVolumeId VolId;
    const BTreeRange* RSpec;
    IScanCursorInfo(const XxxShoreLogicalOID& RelId,
		    const XxxShoreObjectId& IndexId,
		    XxxRecordSchema* Schema,
		    XxxBooleanExpressionPlan* Expr,
		    XxxBooleanExpressionPlan* RExpr, // residual
		    XxxBool Exact,
		    const BTreeRange* RangeSpec,
		    const RelExecParameters* Parameters = NULL,
		    const XxxRecord** SibRecs = NULL,
		    const XxxRecordSchema** SibSchema = NULL,
		    int CurrentJoinPos = 0);
    virtual ~IScanCursorInfo();
    virtual XxxErrCode NextRecord(XxxShoreObjectId*&, XxxRecord*&,
				  XxxBool& Success);
   };

  // ShoreBTree methods.

  // Constructors and Destructors
  ShoreBTree(const XxxIndexType id)
    : XxxIDXClass(id,"B","B+") { }
    
  ~ShoreBTree() {}

  virtual XxxErrCode Print(const XxxShoreVolumeId& VolumeId,
			   const XxxShoreObjectId& IndexId);
  
  virtual XxxErrCode Create(const XxxShoreVolumeId& VolumeId,
			    XxxShoreObjectId& Index,
			    XxxValueExpression *Expr,
			    const XxxBool Unique);

  virtual XxxErrCode Create(const XxxShoreVolumeId& VolumeId,
			    XxxShoreObjectId& Index,
			    XxxBasicType IndxColType,
			    const XxxBool Unique);
  
  virtual XxxErrCode Delete(const XxxShoreVolumeId& VolumeId,
			    XxxShoreObjectId& Index);

  virtual XxxErrCode InsertEntry(const XxxShoreVolumeId& VolumeId,
				 const XxxShoreObjectId& Index,
				 const XxxADTValue* RecValue,
				 const XxxShoreObjectId* RecId,
				 XxxValueExprPlan *ValuePlan);
  
  virtual XxxErrCode InsertEntry(const XxxShoreVolumeId& VolumeId,
				 const XxxShoreObjectId& Index,
				 const XxxADTValue* RecValue,
				 int RecSize,
				 XxxBasicType IndxColType,
				 const XxxShoreObjectId* RecId);

  virtual XxxErrCode DeleteAllEntries(const XxxShoreVolumeId& VolumeId,
				      const XxxShoreObjectId& Index,	
				      const XxxADTValue* RecValue,
				      int RecSize,
				      XxxBasicType IndxColType,
				      int& NumDeleted);
  
  virtual XxxErrCode DeleteOneEntry(const XxxShoreVolumeId& VolumeId,
				    const XxxShoreObjectId& Index,
				    const XxxADTValue* RecValue,
				    const XxxShoreObjectId* RecId,
				    XxxValueExprPlan *ValuePlan);
  
  virtual XxxErrCode DeleteOneEntry(const XxxShoreVolumeId& VolumeId,
				    const XxxShoreObjectId& Index,
				    const XxxADTValue* RecValue,
				    int RecSize,
				    XxxBasicType IndxColType,
				    const XxxShoreObjectId* RecId);
  
  virtual XxxErrCode GetEntry(const XxxShoreVolumeId& VolumeId,
			      const XxxShoreObjectId& Index,
			      const XxxADTValue* RecValue,
			      XxxValueExprPlan *ValuePlan,
			      void *DataBuf,
			      int DataBufSize,
			      bool& Found);
  
  virtual XxxErrCode GetEntry(const XxxShoreVolumeId& VolumeId,
			      const XxxShoreObjectId& Index,
			      const XxxADTValue* RecValue,
			      int RecSize,
			      XxxBasicType IndxColType,
			      void *DataBuf,
			      int DataBufSize,
			      bool& Found);

  virtual XxxIDXClass::IndexRange* CreateRange() {
    return new BTreeRange;
  }
  
  // Merge range R1 and R2 into R3
  virtual XxxErrCode MergeRanges(const XxxIDXClass::IndexRange* R1, 
				 const XxxIDXClass::IndexRange* R2,
				 XxxIDXClass::IndexRange* R3) const;

  // Now that we have a range, we need to open a cursor on the relation
  // using this index. The cursor is going to have to be index specific.
  // i.e each index should create it's own cursor class.

  virtual XxxIDXClass::CursorInfo* 
    GetCursor(const XxxShoreLogicalOID& RelId,
	      const XxxShoreObjectId& IndexId,
	      XxxRecordSchema* Schema,
	      XxxBooleanExpressionPlan* Expr,
	      XxxBooleanExpressionPlan* RExpr, // residual
	      XxxBool Exact,
	      const XxxIDXClass::IndexRange* RangeSpec,
	      const RelExecParameters* Parameters,
	      const XxxRecord** SibRecs,
	      const XxxRecordSchema** SibSchema,
	      int CurrentJoinPos);
  
protected:
  virtual XxxErrCode getRangeSpec(XxxExprPlan* LeftPlan,
				  XxxExprPlan* RightPlan,
				  XxxRelOpType OpType,
				  const XxxRecord** SibRecs,
				  const XxxRecordSchema** SibSchema,
				  const RelExecParameters* Parameters,
				  XxxIDXClass::IndexRange* RangeSpec,
				  const XxxIDXClass::MatchInfo* MatchedInfo) const;
  
};
  
#endif





