#ifndef XXX_RIDX_H
#define XXX_RIDX_H

#include "XxxIndex.h"
#include "BoundingBox.h"

/**
 * Definition of interface for R* index
 */
class ShoreRTree : public XxxIDXClass {
 public:
 //// Range type for R* tree
 class RTreeRange  : public XxxIDXClass::IndexRange  {
  public:
   nbox_t::sob_cmp_t  Cmp;
   XxxBoundingBox box;
   void Copy(const XxxIDXClass::IndexRange& R);
  };
 //// Cursor over R* tree 
 class RIScanCursorInfo : public XxxIDXClass::CursorInfo { 
  public:
   scan_rt_i ScanIter;
   XxxShoreVolumeId VolId;
   XxxBoundingBox   box;
   const RTreeRange* RSpec;    
   RIScanCursorInfo(const XxxShoreLogicalOID& RelId,
		    const XxxShoreObjectId& IndexId,
		    XxxRecordSchema* Schema,
		    XxxBooleanExpressionPlan* Expr,
		    XxxBooleanExpressionPlan* RExpr, // residual
		    XxxBool Exact,
		    const RTreeRange* RangeSpec,
		    const RelExecParameters* Parameters = NULL,
		    const XxxRecord** SibRecs = NULL,
		    const XxxRecordSchema** SibSchema = NULL,
		    int CurrentJoinPos = 0);
   virtual ~RIScanCursorInfo();
   virtual XxxErrCode NextRecord(XxxShoreObjectId*&, XxxRecord*&,
				 XxxBool& Success);
  };

 typedef enum { OWNER,
		ARG,
		NOWHERE } WhereMatched;
 
 //// Index match information for R* tree
 class RMatchInfo : public XxxIDXClass::RelMatchInfo  {
  public:
   XxxCompareOp Comp;
   WhereMatched Where;

   RMatchInfo(XxxCompareOp comp,WhereMatched where)
     : Comp(comp),Where(where) { Exact = XXX_FALSE;}
   ~RMatchInfo() {}
  };
 
 // ShoreRTree methods.

 //// Constructors and Destructors
 ShoreRTree(const XxxIndexType id)
   : XxxIDXClass(id,"R","R*") {}
 
 ~ShoreRTree() {}
 
 virtual XxxErrCode Print(const XxxShoreVolumeId& VolumeId,
			  const XxxShoreObjectId& IndexId);
 
 virtual XxxErrCode Create(const XxxShoreVolumeId& VolumeId,
			   XxxShoreObjectId& Index,
			   XxxValueExpression *Expr,
			   const XxxBool Unique);
 
 virtual XxxErrCode Create(const XxxShoreVolumeId& VolumeId,
			   XxxShoreObjectId& Index,
			   XxxBasicType IndxColType,
			   const XxxBool Unique);
 
 virtual XxxErrCode Delete(const XxxShoreVolumeId& VolumeId,
			   XxxShoreObjectId& Index);

 virtual XxxErrCode InsertEntry(const XxxShoreVolumeId& VolumeId,
				const XxxShoreObjectId& Index,
				const XxxADTValue* RecValue,
				const XxxShoreObjectId* RecId,
				XxxValueExprPlan *ValuePlan);

 virtual XxxErrCode InsertEntry(const XxxShoreVolumeId& VolumeId,
				const XxxShoreObjectId& Index,
				const XxxADTValue* RecValue,
				int RecSize,
				XxxBasicType IndxColType,
				const XxxShoreObjectId* RecId);
 
 //using Index expression and schema, evaluate record, get key from
 // evaluation and insert <key,RecId>

 virtual XxxErrCode DeleteOneEntry(const XxxShoreVolumeId& VolumeId,
				   const XxxShoreObjectId& Index,
				   const XxxADTValue* RecValue,
				   const XxxShoreObjectId* RecId,
				   XxxValueExprPlan *ValuePlan);
 
 virtual XxxErrCode DeleteOneEntry(const XxxShoreVolumeId& VolumeId,
				   const XxxShoreObjectId& Index,
				   const XxxADTValue* RecValue,
				   int RecSize,
				   XxxBasicType IndxColType,
				   const XxxShoreObjectId* RecId);
 
 virtual XxxErrCode GetEntry(const XxxShoreVolumeId& VolumeId,
			     const XxxShoreObjectId& Index,
			     const XxxADTValue* RecValue,
			     XxxValueExprPlan *ValuePlan,
			     void *DataBuf,
			     int DataBufSize,
			     bool& Found);

 virtual XxxErrCode GetEntry(const XxxShoreVolumeId& VolumeId,
			     const XxxShoreObjectId& Index,
			     const XxxADTValue* RecValue,
			     int RecSize,
			     XxxBasicType IndxColType,
			     void *DataBuf,
			     int DataBufSize,
			     bool& Found);
 
 
 // Index specific match function called when the indexexpr
 // does not exactly match the expr.
 virtual XxxErrCode Match(const XxxValueExpression* IndexExpr,
			  const XxxValueExpression* Expr,
			  XxxBool& Success,
			  RelMatchInfo*& MatchedInfo) const;
 
 // Given an expression, we need to get a range to search the index on.
 // We might want to make this a private method and have it called
 // from GetCursor.

 virtual XxxIDXClass::IndexRange* CreateRange() {
   return new RTreeRange;
  }

 // Now that we have a range, we need to open a cursor on the relation
 // using this index. The cursor is going to have to be index specific.
 // i.e each index should create it's own cursor class.

 virtual XxxIDXClass::CursorInfo* 
   GetCursor(const XxxShoreLogicalOID& RelId,
	     const XxxShoreObjectId& IndexId,
	     XxxRecordSchema* Schema,
	     XxxBooleanExpressionPlan* Expr,
	     XxxBooleanExpressionPlan* RExpr, // residual
	     XxxBool Exact,
	     const XxxIDXClass::IndexRange* RangeSpec,
	     const RelExecParameters* Parameters = NULL,
	     const XxxRecord** SibRecs = NULL,
	     const XxxRecordSchema** SibSchema = NULL,
	     int CurrentJoinPos = 0);

protected:

 virtual XxxErrCode getRangeSpec(XxxExprPlan* LeftPlan,
				 XxxExprPlan* RightPlan,
				 XxxRelOpType OpType,
				 const XxxRecord** SibRecs,
				 const XxxRecordSchema** SibSchema,
				 const RelExecParameters* Parameters,
				 XxxIDXClass::IndexRange* RangeSpec,
				 const XxxIDXClass::MatchInfo* MatchedInfo) const;
 
};

#endif
