#include "XxxIndex.h"
#include "ShoreBTree.h"
#include "ShoreRTree.h"
#include "releng.h"
#include "relerror.h"

// Global variables
XxxIDXClass* XxxIDXArray[XXX_IDX_MAX];
int XxxNumIDXs;

// Should get called by InitThread to set up all Indexes
void XxxInitializeIDXs()
{
 for (int i=0; i< XXX_IDX_MAX; XxxIDXArray[i++]=NULL);

 // XXX_BTREE is hardcoded as 0
 new ShoreBTree(XXX_BTREE);
 XxxNumIDXs = 1;
 new ShoreRTree(XxxNumIDXs++);
 // new GiSTRTree(XxxNumIDXs++);
}

// Also called by InitThread to close IDXs
void XxxTerminateIDXs()
{
 for (int i = 0; i < XxxNumIDXs; i++)
   delete XxxIDXArray[i];
}

XxxIDXClass::XxxIDXClass(const XxxIndexType id, const char* Name,
			 const char* oldName)
: TypeId(id)
{
 // Each new index adds itself to the index array.
 XxxIDXArray[id] = this;
 strcpy(TypeName, Name);
 strcpy(OldName, oldName);
}

XxxIDXClass::~XxxIDXClass() 
{}

//AUTHOR: Fabian Camargo DATE: 4/9/97
//PURPOSE: Extract a range from an expression plan that can be used
//to open an index scan on a relation.

XxxErrCode
XxxIDXClass::GetRangeSpec(const XxxBooleanExpressionPlan* ExprPlan,
			  XxxValueExpression* IndexExpr,
			  const MatchInfo* MatchedInfo,
			  const XxxRecord** SibRecs,
			  const XxxRecordSchema** SibSchema,
			  const RelExecParameters* Params,
			  int SibArity,
			  int CurrRelPos,
			  IndexRange* RangeSpec) const
{
 if (ExprPlan->BoolType == XXX_BOOL_BOOL_EXPR) 
   return boolRangeSpec((XxxBoolBooleanPlan *)ExprPlan,
			IndexExpr, MatchedInfo, SibRecs, SibSchema,
			Params, SibArity, CurrRelPos, RangeSpec);
 else 
   return relRangeSpec((XxxRelBooleanPlan *)ExprPlan,
		       IndexExpr, MatchedInfo, SibRecs, SibSchema,
		       Params, SibArity, CurrRelPos, RangeSpec);
}

XxxErrCode
XxxIDXClass::boolRangeSpec(const XxxBoolBooleanPlan* BoolExprPlan,
			   XxxValueExpression* IndexExpr,
			   const MatchInfo* MatchedInfo,
			   const XxxRecord** SibRecs,
			   const XxxRecordSchema** SibSchema,
			   const RelExecParameters* Params,
			   int SibArity,
			   int CurrRelPos,
			   IndexRange* RangeSpec) const
{
 IndexRange *RangeSpec1 = XxxIDXArray[TypeId]->CreateRange();
 XxxBool MatchedConjunct = XXX_FALSE;
 XxxBool Success = XXX_FALSE;
 
 if (!MatchedInfo) {
   REL_NOTE_ERROR(REL_BADINDEX, "Empty match info");
   return XXX_NOT_OK;
  }
 if (MatchedInfo->Type != XxxIDXClass::MatchInfo::BOOL) {
   REL_NOTE_ERROR(REL_BADINDEX, "Bad type for match info");
   return XXX_NOT_OK;
  }
 const XxxIDXClass::BoolMatchInfo* BMInfo = 
   (const XxxIDXClass::BoolMatchInfo*)MatchedInfo;

 int i;
 switch (BoolExprPlan->OpType) {
  case XXX_AND_OP:
   for (i = 0; i < BMInfo->Arity; i++) {
     if (BMInfo->MatchInfoArray[i]) {
       if (GetRangeSpec((XxxBooleanExpressionPlan *)(BoolExprPlan->SubPlans[i]),
			IndexExpr,
			BMInfo->MatchInfoArray[i],
			SibRecs, SibSchema, Params,
			SibArity, CurrRelPos, RangeSpec1) != XXX_OK) {
	 REL_NOTE_ERROR(REL_BADINDEX, "Error getting range spec for conjunct");
	 return XXX_NOT_OK;
	}
       if (!Success) { // if this is the first match
	 RangeSpec->Copy(*RangeSpec1); 
	}
       else {
	 // Take the conjunction of the ranges. This is so that a query 
	 // such that  x>10 and x<100 will use a range of 10 to 100 
	 if (MergeRanges(RangeSpec,RangeSpec1,RangeSpec) != XXX_OK) {
	   REL_NOTE_ERROR(REL_BADINDEX, "Error merging ranges");
	   return XXX_NOT_OK;
	  }
	}
       Success = XXX_TRUE;
      } // if there is a MatchInfoArray[i]
    } // for each subexpression
   
   break;
  case XXX_NOT_OP:
   // Have to invert the range!! Not implemented yet 
   break;
  case XXX_OR_OP:
   // has not been implemented yet : TODO:L2
  default:
   break;
  }
 // Clean up
 delete RangeSpec1;

 return (Success ? XXX_OK : XXX_NOT_OK);
}

XxxErrCode
XxxIDXClass::relRangeSpec(const XxxRelBooleanPlan* RelExprPlan,
			  XxxValueExpression* IndexExpr,
			  const MatchInfo* MatchedInfo,
			  const XxxRecord** SibRecs,
			  const XxxRecordSchema** SibSchema,
			  const RelExecParameters* Params,
			  int SibArity,
			  int CurrRelPos,
			  IndexRange* RangeSpec) const
{
 XxxBool Success = XXX_FALSE;

 if (!MatchedInfo) {
   REL_NOTE_ERROR(REL_BADINDEX, "Empty match info");
   return XXX_NOT_OK;
  }
 if (MatchedInfo->Type != XxxIDXClass::MatchInfo::REL) {
   REL_NOTE_ERROR(REL_BADINDEX, "Bad type for match info");
   return XXX_NOT_OK;
  }
 const XxxIDXClass::RelMatchInfo* RMInfo = 
   (const XxxIDXClass::RelMatchInfo*)MatchedInfo;

 if (RelExprPlan->Arity () != 2) {
   XXX_NOTE_ERROR (XXX_INTERNAL, "Arity of RelExpr not 2 !!");
   return XXX_NOT_OK;
  }
 XxxValueExprPlan *Child0, *Child1;
 
 Child0 = (XxxValueExprPlan *) (RelExprPlan->SubPlans[0]);
 Child1 = (XxxValueExprPlan *) (RelExprPlan->SubPlans[1]);
 
 const XxxValueExpression *Expr = NULL;

 // Get expr from non-constant part of plan. 
 // This is the expr that needs to be matched with the index expression.
 if (RMInfo->LeftConstant) 
   Expr = Child1->SourceExpr();
 else 
   Expr = Child0->SourceExpr();

 XxxRelOpType OpType = RelExprPlan->OperType;
 if (RelExprPlan->Inverted) {
   switch (OpType) {
   case XXX_GT_OP:
     OpType = XXX_LEQ_OP;
     break;
   case XXX_GEQ_OP:
     OpType = XXX_LT_OP;
     break;
   case XXX_EQ_OP:
     OpType = XXX_NEQ_OP;
     break;
   }
 } // if (Inverted)
 
 // fill in RangeSpec using a getRangeSpec function defined
 // by the index type. We need to pass both the subplans 
 // because only the type knows which to get the range from.
 if (getRangeSpec(RelExprPlan->SubPlans[0],
		  RelExprPlan->SubPlans[1],
		  OpType,
		  SibRecs, SibSchema, Params,
		  RangeSpec,
		  MatchedInfo) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_BADEXPR, "Error geting range spec for index");
   return XXX_NOT_OK;
  }
 Success = XXX_TRUE;

 return (Success ? XXX_OK : XXX_NOT_OK);
}

XxxBoolMatchInfo::XxxBoolMatchInfo(int arity) 
: XxxMatchInfo(BOOL)
{
 Arity = arity;
 MatchInfoArray = new XxxMatchInfo*[Arity];
 for (int i = 0; i < Arity; i++) {
   MatchInfoArray[i] = NULL;
  }
}

XxxBoolMatchInfo::~XxxBoolMatchInfo() 
{
 for (int i = 0; i < Arity; i++) {
   if (MatchInfoArray[i]) delete MatchInfoArray[i];
  }
 if (MatchInfoArray) delete [] MatchInfoArray;
}
