#ifndef XXX_RELIDX_H
#define XXX_RELIDX_H

#include "indexmatch.h" 
#include "xxx_thread.h"
#include "shoresm.h" 
#include "relation.h" 
#include "shorerel.h"

// Forward declarations
XXX_EXTERN_CLASS_DECL(RelExecParameters);
XXX_EXTERN_CLASS_DECL(XxxIDXClass);

extern XxxIDXClass* XxxIDXArray[];
extern int XxxNumIDXs;

/**
 * An XxxIDXClass provides a description of the indexing capabilities
 * of one particular kind of index. Each actual physical index is
 * _NOT_ an instance of this class -- rather, this class presents
 * an interface that provides the behavior of each index.
 * This is similar to the use of XxxADTClass to describe a data type.
 */
class XxxIDXClass {
public:  
  typedef XxxMatchInfo MatchInfo;
  typedef XxxBoolMatchInfo BoolMatchInfo;
  typedef XxxRelMatchInfo RelMatchInfo;

  /**
   * The IndexRange is used to specify the range of values desired
   * for some lookup.
   */
  class IndexRange {
  public:
    IndexRange() {}
    virtual ~IndexRange() {}
    virtual void Copy(const IndexRange& R) {}; 
  };

  /**
   *
   */
  typedef XxxShoreRelation::CursorInfo CursorInfo;
  /********** Cannot define a different CursorInfo here!
              When scanning a file, it is assumed to be
              a XxxShoreRelation::CursorInfo
  class CursorInfo : public XxxRelation::CommonCursorInfo { 
  public:
    CursorInfo(XxxRecordSchema* Schema,
	       XxxBooleanExpressionPlan* Expr,
	       XxxBool Exact,
	       const RelExecParameters* Parameters = NULL,
	       const XxxRecord** SibRecs = NULL,
	       const XxxRecordSchema** SibSchema = NULL,
	       int CurrentJoinPos = 0)
      : XxxRelation::CommonCursorInfo(Schema, Expr, Exact, Parameters, 
				      SibRecs, SibSchema, CurrentJoinPos) 
	{}
    virtual ~CursorInfo() {}
  };
  *****/

  const XxxIndexType TypeId; 
  char TypeName[XXX_IDX_NAME_MAX];
  char OldName[XXX_IDX_NAME_MAX]; // Name used for backward compatability
  
  // Constructors and Destructors
  XxxIDXClass(const XxxIndexType id, const char* Name, 
	      const char* oldName = "");
  
  virtual ~XxxIDXClass();
  
  virtual XxxErrCode Print(const XxxShoreVolumeId& VolumeId,
			   const XxxShoreObjectId& IndexId) 
			     { return XXX_NOT_OK; }
  
  virtual XxxErrCode Create(const XxxShoreVolumeId& VolumeId,
			    XxxShoreObjectId& Index,
			    XxxValueExpression *Expr,
			    const XxxBool Unique)
    { return XXX_NOT_OK; }

  virtual XxxErrCode Create(const XxxShoreVolumeId& VolumeId,
			    XxxShoreObjectId& Index,
			    XxxBasicType IndxColType,
			    const XxxBool Unique)
    { return XXX_NOT_OK; }

  virtual XxxErrCode Delete(const XxxShoreVolumeId& VolumeId,
			    XxxShoreObjectId& Index)
    { return XXX_NOT_OK; }


  // Functions to update the index. Some functions have more than
  // one version to allow for flexible usage.
  virtual XxxErrCode InsertEntry(const XxxShoreVolumeId& VolumeId,
				 const XxxShoreObjectId& Index,
				 const XxxADTValue* RecValue,
				 const XxxShoreObjectId* RecId,
				 XxxValueExprPlan *ValuePlan)
    { return XXX_NOT_OK; }

  virtual XxxErrCode InsertEntry(const XxxShoreVolumeId& VolumeId,
				 const XxxShoreObjectId& Index,
				 const XxxADTValue* RecValue,
				 int RecSize,
				 XxxBasicType IndxColType,
				 const XxxShoreObjectId* RecId)
    { return XXX_NOT_OK; }

  virtual XxxErrCode DeleteAllEntries(const XxxShoreVolumeId& VolumeId,
				      const XxxShoreObjectId& Index,	
				      const XxxADTValue* RecValue,
				      int RecSize,
				      XxxBasicType IndxColType,
				      int& NumDeleted)
    { return XXX_NOT_OK; }

  virtual XxxErrCode DeleteOneEntry(const XxxShoreVolumeId& VolumeId,
				    const XxxShoreObjectId& Index,
				    const XxxADTValue* RecValue,
				    const XxxShoreObjectId* RecId,
				    XxxValueExprPlan *ValuePlan)
    { return XXX_NOT_OK; }

  virtual XxxErrCode DeleteOneEntry(const XxxShoreVolumeId& VolumeId,
				    const XxxShoreObjectId& Index,
				    const XxxADTValue* RecValue,
				    int RecSize,
				    XxxBasicType IndxColType,
				    const XxxShoreObjectId* RecId)
    { return XXX_NOT_OK; }
  
  virtual XxxErrCode GetEntry(const XxxShoreVolumeId& VolumeId,
			      const XxxShoreObjectId& Index,
			      const XxxADTValue* RecValue,
			      XxxValueExprPlan *ValuePlan,
			      void *DataBuf,
			      int DataBufSize,
			      bool& Found)
    { return XXX_NOT_OK; }

  virtual XxxErrCode GetEntry(const XxxShoreVolumeId& VolumeId,
			      const XxxShoreObjectId& Index,
			      const XxxADTValue* RecValue,
			      int RecSize,
			      XxxBasicType IndxColType,
			      void *DataBuf,
			      int DataBufSize,
			      bool& Found)
    { return XXX_NOT_OK; }

  
  // Index specific match function called when the indexexpr
  // does not exactly match the expr.
  virtual XxxErrCode Match(const XxxValueExpression* IndexExpr,
			   const XxxValueExpression* Expr,
			   XxxBool& Success,
			   RelMatchInfo*& MatchedInfo) const {
    Success = XXX_FALSE;
    return XXX_OK;
   }


  // Given an expression, we need to get a range to search the index on.
  // We might want to make this a private method and have it called
  // from GetCursor.

  virtual IndexRange* CreateRange() 
    { return NULL; }

  virtual XxxErrCode MergeRanges(const IndexRange* R1, 
				 const IndexRange* R2,
				 IndexRange* R3) const 
				  { return XXX_NOT_OK; }

  virtual XxxErrCode GetRangeSpec(const XxxBooleanExpressionPlan* ExprPlan,
				  XxxValueExpression* IndexExpr,
				  const MatchInfo* MatchedInfo,
				  const XxxRecord** SibRecs,
				  const XxxRecordSchema** SibSchema,
				  const RelExecParameters* Parameters,
				  int SibArity,
				  int CurrentJoinPos,
				  IndexRange* RangeSpec) const;


  virtual CursorInfo* GetCursor(const XxxShoreLogicalOID& RelId,
				const XxxShoreObjectId& IndexId,
				XxxRecordSchema* Schema,
				XxxBooleanExpressionPlan* Expr,
				XxxBooleanExpressionPlan* RExpr, // residual
				XxxBool Exact,
				const IndexRange* RangeSpec,
				const RelExecParameters* Parameters = NULL,
				const XxxRecord** SibRecs = NULL,
				const XxxRecordSchema** SibSchema = NULL,
				int CurrentJoinPos = 0)
    { return NULL; }

protected:
  virtual XxxErrCode getRangeSpec(XxxExprPlan* LeftPlan,
				  XxxExprPlan* RightPlan,
				  XxxRelOpType OpType,
				  const XxxRecord** SibRecs,
				  const XxxRecordSchema** SibSchema,
				  const RelExecParameters* Parameters,
				  IndexRange* RangeSpec,
				  const MatchInfo* MatchedInfo) const = 0;

  virtual XxxErrCode boolRangeSpec(const XxxBoolBooleanPlan* BoolExprPlan,
				   XxxValueExpression* IndexExpr,
				   const MatchInfo* MatchedInfo,
				   const XxxRecord** SibRecs,
				   const XxxRecordSchema** SibSchema,
				   const RelExecParameters* Parameters,
				   int SibArity,
				   int CurrentJoinPos,
				   IndexRange* RangeSpec) const;
  virtual XxxErrCode relRangeSpec(const XxxRelBooleanPlan* BoolExprPlan,
				  XxxValueExpression* IndexExpr,
				  const MatchInfo* MatchedInfo,
				  const XxxRecord** SibRecs,
				  const XxxRecordSchema** SibSchema,
				  const RelExecParameters* Parameters,
				  int SibArity,
				  int CurrentJoinPos,
				  IndexRange* RangeSpec) const;

  // the subclasses of XxxIndexClass use this
  ss_m* getSSM(XxxStorageManager* ShoreSM) 
    { return ShoreSM->storageManager;}
};

#endif





