#include <stdio.h>
#include <stdlib.h>
#include <fstream.h>

#ifdef __GNUG__
#pragma implementation
#endif

#include "xxx_thread.h"
#include "reladt.h"
#include "relcat.h"
#include "reldata.h"
#include "releng.h"
#include "env.h"
#include "relenv.h"
#include "relerror.h"
#include "relation.h"
#include "values.h"
#include "funcexpr.h"
#include "castexpr.h"
#include "XxxIndex.h"

const int RelStatsInfo::DefaultTableCard = 1000;
const int RelStatsInfo::DefaultTupleSize = 100;

RelStatsInfo::RelStatsInfo(int NumFields)
{
 init();
 Arity = NumFields;
 AttrStats = new XxxADTClass::TypeStats*[Arity];
 for (int i = 0; i < Arity; i++) {
   AttrStats[i] = NULL;
  }
}

RelStatsInfo* RelStatsInfo::
Replicate(const XxxRecordSchema* RecSchema) const
{
 RelStatsInfo* NewSI = new RelStatsInfo();
 NewSI->Cardinality = Cardinality;
 NewSI->AvgTupleSize = AvgTupleSize;
 NewSI->Arity = Arity;
 NewSI->AttrStats = new XxxADTClass::TypeStats*[Arity];
 for (int i = 0; i < Arity; i++) {
   NewSI->AttrStats[i] = NULL;
   if (AttrStats[i] && RecSchema) {
     XxxADTMetaInfo *FieldMetaInfo = RecSchema->GetAttr(i)->TypeCatInfo;
     if (AttrStats[i]->Replicate(FieldMetaInfo, 
				 NewSI->AttrStats[i]) != XXX_OK) {
       REL_NOTE_ERROR(REL_BADCAT, "Error replicating statsinfo");
      }
    }
  } // for each attr
 return NewSI;
}

void RelStatsInfo::reclaim()
{
 if (AttrStats) {
   for (int i = 0; i < Arity; i++) {
     if (AttrStats[i]) {
       delete AttrStats[i];
      } // if
    } // for
   delete [] AttrStats;
  } // if
}

XxxErrCode RelStatsInfo::ResetStatsInfo(const XxxRecordSchema* RecSchema)
{
 // first delete the old stats info, if it exists
 reclaim();
 init();
 Cardinality = 0;
 Arity = RecSchema->Arity();
 AttrStats = new XxxADTClass::TypeStats*[Arity];
 for (int i = 0; i < Arity; i++) {
   XxxBasicType FieldType = RecSchema->GetAttr(i)->Type;
   XxxADTMetaInfo *FieldMetaInfo = RecSchema->GetAttr(i)->TypeCatInfo;
   if (XxxADTArray[FieldType]->CreateStatsInfo(FieldMetaInfo,
					       AttrStats[i]) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADCAT, "Cannot create stats info");
     return XXX_NOT_OK;
    }
  } // for 
 return XXX_OK;
}
XxxErrCode RelStatsInfo::AccumulateStats(XxxRecord* Rec,
					 const XxxRecordSchema* RecSchema)
{
 for (int i=0; i < Arity; i++) { 
   XxxBasicType FieldType = RecSchema->GetAttr(i)->Type;
   XxxADTMetaInfo *FieldMetaInfo = RecSchema->GetAttr(i)->TypeCatInfo;
    //Bjarke, sep. 14. 2001: added const qualifier to variables
   const char *FldiPtr;
   Rec->GetField(i, RecSchema, FldiPtr);
   if (AttrStats[i] &&
       (AttrStats[i]->Accum((XxxADTValue*)FldiPtr, 
			    FieldMetaInfo) != XXX_OK)) {
     REL_NOTE_ERROR(REL_BADCAT, "Cannot accumulate stats info");
     return XXX_NOT_OK;
    }
  } // for each field
 AvgTupleSize = ((Cardinality * AvgTupleSize) + 
		 Rec->Size())/(Cardinality+1);
 Cardinality++;

 return XXX_OK;
}

XxxErrCode RelStatsInfo::TerminateStats(const XxxRecordSchema* RecSchema)
{
 // first terminate the aggregation of statistics collection
 for (int i=0; i < Arity; i++) { 
   XxxADTMetaInfo *FieldMetaInfo = RecSchema->GetAttr(i)->TypeCatInfo;
   if (AttrStats[i] && 
       (AttrStats[i]->Term(FieldMetaInfo) != XXX_OK)) {
     REL_NOTE_ERROR(REL_BADCAT, "Cannot terminate stats collection");
     return XXX_NOT_OK;
    }
  } // for each field

 if (XXX_THISRELENV()->DebugStats) {
   Print(cerr, RecSchema);
 }

 return XXX_OK;
}

void RelStatsInfo::Print(ostream& OStr,
			 const XxxRecordSchema* RecSchema) const
{
 OStr << "Card = " << Cardinality << ", AvgTupSize = " << AvgTupleSize << endl;
 if (Arity && AttrStats && RecSchema) {
   for (int i = 0; i < Arity; i++) {
     XxxADTMetaInfo *FieldMetaInfo = RecSchema->GetAttr(i)->TypeCatInfo;
     if (AttrStats[i]) {
       AttrStats[i]->Print(OStr, FieldMetaInfo);
      } // AttrStats[i]
    } // for each attr i
  } // if attrs stats exist
}

// Added Indextype to index info. Fabian 11/30
RelImplInfo::
IndexImplInfo::IndexImplInfo(const char *IndxStr, 
			     const char *IndxName,
			     const XxxIndexType type,
			     XxxBool IsUnique,
			     XxxBool IsClustered,
			     const XxxRecordSchema *RecordSchema)
  : IndexType(type), Clustered(IsClustered), Unique(IsUnique)
{
 // Set the IndexPlan to NULL initially.
 // If it is still NULL when this constructor returns, it indicates
 // that there was an error here.
 IndexPlan = NULL;
 IndexExpr = NULL;

 IndexString = new char[strlen (IndxStr) + 1];
 strcpy (IndexString, IndxStr);
 IndexName = new char[strlen (IndxName) + 1];
 strcpy (IndexName, IndxName);
 
 XxxRelationEngine *RelEng =
   (XxxRelationEngine *) XxxDBEngineArray[RelationEngineId];
 if (RelEng->ParseValueExpression(IndexString, IndexExpr) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADINDEX, "Error parsing index expression");
   return;
  }
 
 if (IndexExpr->ResolveVars(RecordSchema) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADINDEX, "Error resolving index expression");
   return;
  }
 if (IndexExpr->TypeCheck() != XXX_OK) {
   REL_NOTE_ERROR(REL_BADINDEX, "Error type checking index expression");
   return;
  }
 
 // Obtain plan for expr. This avoids extra work later on 
 // when the index is used.
  if (IndexExpr->Optimize(IndexPlan) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADINDEX, "Error optimizing index expression");
   IndexPlan = NULL;
   return;
  }
}

void 
RelImplInfo::Print(ostream& OStr) const 
{
 if (NumIndexes) {
   OStr << " Indexes : ";
   IndexList* Tmp = Indexes;
   while (Tmp->Next) {
     Tmp->Next->IndxInfo->Print(OStr);
     OStr << " ";
     Tmp = Tmp->Next;
    }
  }
}

RelImplInfo::RelImplInfo(const RelImplInfo& IInfo, XxxBool CopyIndexes)
: RelType(IInfo.RelType), SourceIndex(IInfo.SourceIndex)
{
 if (IInfo.Indexes && CopyIndexes) {
   NumIndexes = IInfo.NumIndexes;
   Indexes = IInfo.Indexes->Replicate();
  }
 else {
   NumIndexes = 0;
   Indexes = new IndexList(NULL);
  }
}

RelImplInfo::~RelImplInfo()
{
 if (Indexes) {
   Indexes->Destroy();
   delete Indexes;
   Indexes = NULL;
  }
}

void RelImplInfo::IndexList::Destroy()
{
 while (Next) {
   IndexList* Tmp = Next->Next;
   if (Next->IndxInfo) delete Next->IndxInfo;
   delete Next;
   Next = Tmp;
  }
 if (IndxInfo) delete IndxInfo;
}

RelImplInfo::IndexList* RelImplInfo::IndexList::Replicate() const
{
 IndexList* ReturnList = new IndexList(NULL);
 IndexList* Tmp = Next;
 while (Tmp) {
   if (Tmp->IndxInfo) {
     ReturnList->Next = new IndexList(Tmp->IndxInfo->Replicate(),
				      ReturnList->Next);
    }
   Tmp = Tmp->Next;
  }
 return ReturnList;
}

XxxErrCode RelImplInfo::
IndexList::Matches(XxxBooleanExpressionPlan* ExprPlan,
		   int SibArity,
		   int CurrRelPos,
		   XxxBool& Success,
		   XxxMatchInfo*& MatchedInfo,
		   IndexImplInfo*& IndxInfo) const
{
 Success = XXX_FALSE;
 const RelImplInfo::IndexList *IndexList = this;
   
 while (IndexList->Next) {
   // Matches() is really a compile-time check based on 
   // the pattern of the expression and the query
   if (IndexList->Next->IndxInfo) {
     IndxInfo = IndexList->Next->IndxInfo;
     if (IndxInfo->Matches(ExprPlan, SibArity, CurrRelPos, 
			   Success, MatchedInfo) != XXX_OK) {
       REL_NOTE_ERROR(REL_BADINDEX, "Error matching with indx info");
       return XXX_NOT_OK;
      } // if Matches
     if (Success) return XXX_OK; //Index matched
    } // if there is an IndxInfo
   IndexList = IndexList->Next;
  }     // while
 // No index matched
 return XXX_OK;
}

RelImplInfo::
IndexImplInfo::IndexImplInfo(const IndexImplInfo& IInfo)
{
 Clustered = IInfo.Clustered;
 Unique = IInfo.Unique;
 // copy IndexString
 IndexString = NULL;
 IndexName = NULL;
 IndexExpr = NULL;
 IndexPlan = NULL;

 if (IInfo.IndexString) {
   IndexString = new char[strlen(IInfo.IndexString)+1];
   strcpy(IndexString, IInfo.IndexString);
  }
 if (IInfo.IndexName) {
   IndexName = new char[strlen(IInfo.IndexName)+1];
   strcpy(IndexName, IInfo.IndexName);
  }
 IndexType = IInfo.IndexType;
 if (IInfo.IndexExpr) {
   IndexExpr = (XxxValueExpression*)(IInfo.IndexExpr->Replicate());
   // TODO:L1 it would be nice to copy this instead of re-optimizing it
   if (IndexExpr->Optimize(IndexPlan) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADINDEX, "Error copying index impl info");
    }
  }
}

XxxErrCode RelImplInfo::
IndexImplInfo::Matches(XxxBooleanExpressionPlan* ExprPlan,
		       int SibArity,
		       int CurrRelPos,
		       XxxBool& Success,
		       XxxMatchInfo*& MatchedInfo) const
{
 // check that the index string "matches" the expression
 Success = XXX_FALSE;

 // here are the kinds of matches allowed:
 // 1) if the expression is an AND , recursively check the boolean 
 // factors. These will be inexact matches.
 //
 // 2) if the expression is an OR , recursively check the boolean
 // factors. If any matches, it is an exact match.
 //
 // 3) if the expression is =,<,>, and if one sub-expression
 // matches the index, while the other is constant, then  the index matches
 // exactly.
 //
 // 4) otherwise, the index does not match the predicate
 // 
 if (ExprPlan->BoolType == XXX_BOOL_BOOL_EXPR) {
   return matches((XxxBoolBooleanPlan*)ExprPlan,
		  SibArity, CurrRelPos, Success, MatchedInfo);
  } // if XXX_BOOL_BOOL_EXPR
 else if (ExprPlan->BoolType == XXX_REL_BOOL_EXPR) { 
   return matches((XxxRelBooleanPlan*)ExprPlan,
		  SibArity, CurrRelPos, Success, MatchedInfo);
  } // if it is a REL_BOOLEAN_EXPR

 else // TODO: should we match XXX_VAL_BOOL_EXPR ?
   return XXX_OK;
}

XxxErrCode RelImplInfo::
IndexImplInfo::matches(XxxBoolBooleanPlan* BExprPlan,
		       int SibArity,
		       int CurrRelPos,
		       XxxBool& Success,
		       XxxMatchInfo*& MatchedInfo) const
{
 XxxBoolMatchInfo* BMInfo = NULL;
 XxxBool PartialSuccess;
 int i;
 Success = XXX_FALSE;
 
 switch (BExprPlan->OpType) {
  case XXX_AND_OP: 
   BMInfo = new XxxBoolMatchInfo(BExprPlan->Arity());
   BMInfo->Exact = XXX_TRUE;
   for (i = 0; i < BExprPlan->Arity(); i++) {
     if (Matches((XxxBooleanExpressionPlan *) BExprPlan->SubPlans[i],
		 SibArity, CurrRelPos,
		 PartialSuccess, BMInfo->MatchInfoArray[i]) != XXX_OK) {
       REL_NOTE_ERROR(REL_BADINDEX, "Error matching bool expression");
       return XXX_NOT_OK;
      }
     if (PartialSuccess) 
       Success = XXX_TRUE;
     else
       BMInfo->Exact = XXX_FALSE;
    } // for each term of the conjunction
   break;
  case XXX_OR_OP:
  case XXX_NOT_OP:
  default:
   // // This is incomplete and wont work ...
   break;
  }
 MatchedInfo = BMInfo;
 return XXX_OK;
}

XxxErrCode RelImplInfo::
IndexImplInfo::matches(XxxRelBooleanPlan* RExprPlan,
		       int SibArity,
		       int CurrRelPos,
		       XxxBool& Success,
		       XxxMatchInfo*& MatchedInfo) const
{
 Success = XXX_FALSE;
 XxxRelMatchInfo* RMInfo = NULL;

 // for things like ANY, ALL, IN, etc
 if (RExprPlan->Arity () != 2) {
   return XXX_OK;
  }

 // only certain kinds of expressions can be satisfied with an 
 // index
 switch (RExprPlan->OperType) {
  case XXX_GT_OP:
  case XXX_GEQ_OP:
  case XXX_LT_OP:
  case XXX_LEQ_OP:
  case XXX_EQ_OP:
   break;
  case XXX_LIKE_OP:
  case XXX_NEQ_OP:
   return XXX_OK;
  default:
   REL_NOTE_ERROR(REL_BADEXEC, "Invalid operator type");
   return XXX_NOT_OK;
  }

 XxxValueExprPlan *Child0, *Child1;

 Child0 = (XxxValueExprPlan *) (RExprPlan->SubPlans[0]);
 Child1 = (XxxValueExprPlan *) (RExprPlan->SubPlans[1]);
 
 XxxBitMap Depends0(SibArity + 1);
 XxxBitMap Depends1(SibArity + 1);
 
 if (Child0->GetDependencies(Depends0) != XXX_OK) {
   XXX_NOTE_ERROR (XXX_BADEXPR, 
		   "Unable to get child expr 0 dependencies");
   return XXX_NOT_OK;
  }
 if (Child1->GetDependencies(Depends1) != XXX_OK) {
   XXX_NOTE_ERROR (XXX_BADEXPR, 
		   "Unable to get child expr 1 dependencies");
   return XXX_NOT_OK;
  }
 
 const XxxValueExpression *Expr = NULL;
 XxxBool LeftConstant;
 // see if Child0 is constant and Child1 matches the IndexString
 // or vice-versa. Check this: it should be non-constant!
 if (!Depends0[CurrRelPos+1] && Depends1[CurrRelPos+1]) {
   LeftConstant = XXX_TRUE;
   Expr = Child1->SourceExpr();
  }
 else if (!Depends1[CurrRelPos+1] && Depends0[CurrRelPos+1]) {
   LeftConstant = XXX_FALSE;
   Expr = Child0->SourceExpr();
  }
 else 
   return XXX_OK;
 
 // Now check to see if Expr matches the index expression
 // The second argument is used to dynamically bind vars to relation names
 char* Env = new char[XXX_NODE_NAME_MAX];
 strcpy(Env,"");
 
 Success = Expr->Match(IndexExpr,Env);

 if (Success) {
   RMInfo = new XxxRelMatchInfo;
   RMInfo->Exact = XXX_TRUE;
  }
 else {
   // If there's no exact expr match then see if the ADT that the 
   // index is built on can help us match the expression
   if (XxxIDXArray[IndexType]->Match(IndexExpr, Expr, Success,
				     RMInfo) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADINDEX, "Error matching expr");
     return XXX_NOT_OK;
    }
  }

 if (Success && RMInfo) { // RMInfo should exist if Success is true!
   RMInfo->LeftConstant = LeftConstant;
  }
 MatchedInfo = RMInfo;

 // reclaim any memory allocated 
 delete [] Env;
 return XXX_OK;
}

XxxBool RelImplInfo::
SortOrders::Matches(const XxxSortInfo& SortInfo) const
{
 unsigned int i;
 for (i = 0; i < NumSortOrders; i++) {
   if (SortInfoArray[i].Matches(SortInfo)) return XXX_TRUE;
  }
 return XXX_FALSE;
}

//
// Methods of RelCatalogRelInfo
//
XxxErrCode RelCatalogInfo::SerializeCopySize(int &Size) const
{
  return XXX_NOT_OK;
}

XxxErrCode RelCatalogInfo::SerializeCopy(char *Data) const
{
  return XXX_NOT_OK;
}

RelCatalogInfo* RelCatalogInfo::Replicate1(XxxBool CopyIndexes) const
{
 RelCatalogInfo* CatInfo = new RelCatalogInfo(RelName);
 CatInfo->Named = Named;
 CatInfo->recSchema = new XxxRecordSchema(*(RecordSchema()));

 if (StatsInfo())
   CatInfo->statsInfo = StatsInfo()->Replicate(RecordSchema());
 else 
   CatInfo->statsInfo = NULL;

 if (ImplInfo())
   CatInfo->implInfo = ImplInfo()->Replicate(CopyIndexes);
 else
   CatInfo->implInfo = NULL;

 return CatInfo;
}

void RelCatalogInfo::PrintSchema(ostream& OStr) const
{
  if (recSchema) {
    OStr << RelName << " : ";
    OStr << " Schema: ";
    recSchema->Print(OStr);
    OStr << endl;
  }
}

void RelCatalogInfo::Print(ostream& OStr) const
{
 OStr << RelName << (Named?"(*):":":") << endl;
 if (recSchema) {
   OStr << " Schema: ";
   recSchema->Print(OStr);
   OStr << endl;
  }
 if (implInfo) {
   OStr << " Impl: ";
   implInfo->Print(OStr);
   OStr << endl;
  }
 if (statsInfo) {
   OStr << " Stats: ";
   statsInfo->Print(OStr);
   OStr << endl;
  }
}

//
// methods of RelCatalog
//
static int printCatEntry(RelCatalogInfo& CatEntry, void *arg)
{
 CatEntry.Print(*((ostream *)arg));
 return 1;
}
void RelCatalog::PrintCats(ostream& OStr)
{
 // should print some identifier for the catalog src here
 ApplyAll(printCatEntry, (void *)(&OStr));
}

static int printCatSchema(RelCatalogInfo& CatEntry, void *arg)
{
 CatEntry.PrintSchema(*((ostream *)arg));
 return 1;
}
void RelCatalog::PrintSchemas(ostream& OStr)
{
 ApplyAll(printCatSchema, (void *)(&OStr));
}
