#include <stdio.h>
#include <stdlib.h>
#include <fstream.h>
#include <dirent.h>
#include <sys/stat.h>

#ifdef __GNUG__
#pragma implementation
#endif

#include "xxx_thread.h"
#include "env.h"
#include "reladt.h"
#include "relcat.h"
#include "relexec.h"
#include "relenv.h"
#include "relerror.h"
#include "relation.h"
#include "relplanop.h"
#include "values.h"
#include "shorerel.h"
#include "XxxIndex.h"
//Begin:Moustafa
// Stream Scan operator needs access to StrmMgr (Mar. 2003)
#include "genadt.h"
#include <math.h>
extern StreamManagerThread* StrmMgr;
//End: Moustafa

//
// Methods of XxxShoreFileRelation
//
XxxShoreFileRelation::XxxShoreFileRelation(RelCatalogInfo* Cat, 
					   const XxxShoreLogicalOID& RelId,
					   RelShoreDataSrc::RelShoreDevice* Dev)
:XxxShoreRelation(Cat, RelId, Dev)
{
}

XxxShoreFileRelation::~XxxShoreFileRelation()
{
}

//PURPOSE: Returns true if an index on relation matches Expr
//NOTES: Used to be part of InitCursor
XxxErrCode
XxxShoreFileRelation::matchIndex(XxxBooleanExpressionPlan* Expr,
				 int SibArity,
				 int CurrJoinPos,
				 XxxBool& Success,
				 XxxMatchInfo*& MatchedInfo,
				 RelShoreDataSrc::RelShoreImplInfo::
				 ShoreIndexImplInfo*& ShoreIndxInfo) const
{
 Success = XXX_FALSE;

 // step thru the indexes and see if any of them matches the 
 // desired filter predicate (if there is a filter predicate)
 if  (!(Expr && CatalogInfo->ImplInfo()))
   return XXX_OK;
 RelImplInfo::IndexList *IndexList = CatalogInfo->ImplInfo()->Indexes;
 if (!IndexList) 
   return XXX_OK;
 RelImplInfo::IndexImplInfo* IndxInfo;

 if (IndexList->Matches(Expr, SibArity, CurrJoinPos,
			Success, MatchedInfo, IndxInfo) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADINDEX, "Error matching with indexes");
   return XXX_NOT_OK;
  }
 ShoreIndxInfo =
   (RelShoreDataSrc::RelShoreImplInfo::ShoreIndexImplInfo *) IndxInfo;
 
 return XXX_OK;
}

XxxErrCode 
XxxShoreFileRelation::InitCursor1(XxxSetCursor & SCsr,
				  XxxBooleanExpressionPlan * Expr,
				  const XxxRecord * Rec[],
				  const XxxRecordSchema * Schema[],
				  const RelExecParameters * Params,
				  int SibArity,
				  int CurrJoinPos) const
{
 CursorInfo *CInfo = NULL;

 if (AccMode == UNDECIDED) {
   // have to choose an access path at run time
   if (chooseAccessPath(Expr, Rec, Schema, Params, 
			SibArity, CurrJoinPos, CInfo) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADEXEC, "Error choosing access path");
     return XXX_NOT_OK;
    }
  }
 else if (AccMode == FILESCAN) {
   if (fscanAccessPath(Expr, Rec, Schema, Params, 
		       SibArity, CurrJoinPos, CInfo) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADEXEC, "Error creating filescan path");
     return XXX_NOT_OK;
    }
  }
 else if (AccMode == INDEXED) {
   if (iscanAccessPath(Expr, Rec, Schema, Params, 
		       SibArity, CurrJoinPos, CInfo) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADEXEC, "Error creating indexscan path");
     return XXX_NOT_OK;
    }
 }
 //Begin:Moustafa
 // Choose SSCAN operator (Mar. 2003)
 else if (AccMode == SSCAN) {
   if (sscanAccessPath(Expr, Rec, Schema, Params, 
		       SibArity, CurrJoinPos, CInfo) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADEXEC, "Error creating streamscan path");
     return XXX_NOT_OK;
   }
 }
 //End: Moustafa
 else {
   XXX_NOTE_ERROR (XXX_INTERNAL, "Relation scan with unknown access path");
   return XXX_NOT_OK;
  }
 
 SCsr.State =  CInfo;
 SCsr.AtEnd = XXX_FALSE;
 SCsr.NumberRead = 0;
 return XXX_OK;
}

XxxErrCode 
XxxShoreFileRelation::chooseAccessPath(XxxBooleanExpressionPlan * Expr,
				       const XxxRecord * Rec[],
				       const XxxRecordSchema * Schema[],
				       const RelExecParameters * Parameters,
				       int SibArity,
				       int CurrJoinPos,
				       CursorInfo*& CInfo) const
{
 RelShoreDataSrc::RelShoreImplInfo::ShoreIndexImplInfo *ShoreIndxInfo;

 XxxBool Success = XXX_FALSE;
 XxxMatchInfo* MatchedInfo = NULL;

 if (matchIndex(Expr, SibArity, CurrJoinPos, Success,
		MatchedInfo, ShoreIndxInfo) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADINDEX, "Error matching index");
   return XXX_NOT_OK;
  }

 if (Success) {
   // use index scan
   if (indexedAccessPath(Expr, NULL, Rec, Schema, Parameters, SibArity, CurrJoinPos, 
			 MatchedInfo, ShoreIndxInfo, CInfo) != XXX_OK){
     REL_NOTE_ERROR(REL_BADEXEC, "Error creating indexscan path");
     return XXX_NOT_OK;
    }
  } // if there is a matching index
  else { // use a file scan
    if (fscanAccessPath(Expr, Rec, Schema, Parameters, 
			SibArity, CurrJoinPos, CInfo) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADEXEC, "Error creating filescan path");
      return XXX_NOT_OK;
     }
   }

 if (MatchedInfo) delete MatchedInfo;
 return XXX_OK;
}

XxxErrCode 
XxxShoreFileRelation::fscanAccessPath(XxxBooleanExpressionPlan* Expr,
				      const XxxRecord * Rec[],
				      const XxxRecordSchema * Schema[],
				      const RelExecParameters * Parameters,
				      int SibArity,
				      int CurrJoinPos,
				      CursorInfo*& CInfo) const
{
 // Use File scan.
 if (AccMode != UNDECIDED) {
   // In this case, the desired selection condition is already in the 
   // RelFScanPlanOp sitting in AccInfo. It was pre-detemrined by
   // the optimizer -- ignore the Expr passed in.
   RelFScanPlanOp* PlanOp = (RelFScanPlanOp*)AccInfo;
   if (!PlanOp) {
     REL_NOTE_ERROR(REL_BADEXEC, "PlanOp null for Filescan");
     return XXX_NOT_OK;
    }
   if (Expr) {
     // i dont think this should ever be non-NULL: just checking
     cerr << "Curious:FScan why does it ever come here" << endl;
   }
   Expr = PlanOp->Selection;
   /****
   if (XxxBooleanExpressionPlan::MakeConjunction
       (Expr, PlanOp->Selection) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADOPT, "Error making conjunction");
     return XXX_NOT_OK;
    }
    ***/
  }

 CInfo = new FScanCursorInfo (RelationId, CatalogInfo->RecordSchema(),
			      Expr, Parameters, Rec, Schema, CurrJoinPos);
 
 if (XXX_THISRELENV() && XXX_THISRELENV()->DebugStorage) {
   cerr << "Using FScan ";
   cerr << " of (SHORE file) relation " << CatalogInfo->RelName << endl;
 }
 return XXX_OK;
}
//Begin:Moustafa
// This function associates the Stream Scan to an operator (Mar. 2003)
XxxErrCode 
XxxShoreFileRelation::sscanAccessPath(XxxBooleanExpressionPlan* Expr,
				      const XxxRecord * Rec[],
				      const XxxRecordSchema * Schema[],
				      const RelExecParameters * Parameters,
				      int SibArity,
				      int CurrJoinPos,
				      CursorInfo*& CInfo) const
{
  // Use stream scan.

  CursorInfo *RCInfo = NULL;

  // have to choose an access path for the original table at run time
  if (chooseAccessPath(Expr, Rec, Schema, Parameters, 
		       SibArity, CurrJoinPos, RCInfo) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADEXEC, "Error choosing access path");
    return XXX_NOT_OK;
  }
  
  // assign CInfo to a field in a  special stream cursor
  if (RCInfo!=NULL)
    CInfo = new SScanCursorInfo (RCInfo);

  return XXX_OK;
}
//End: Moustafa

XxxErrCode 
XxxShoreFileRelation::iscanAccessPath(XxxBooleanExpressionPlan* IndxExpr,
				      const XxxRecord * Rec[],
				      const XxxRecordSchema * Schema[],
				      const RelExecParameters * Parameters,
				      int SibArity,
				      int CurrJoinPos,
				      CursorInfo*& CInfo) const
{
 // Use Index scan.
 if (AccMode != INDEXED) {
   REL_NOTE_ERROR(REL_BADEXEC, "Shouldnt enter here");
   return XXX_NOT_OK;
  }
 // The desired selection cond and residual cond are already in the 
 // RelIScanPlanOp sitting in AccInfo. It was pre-detemrined by
 // the optimizer -- ignore the IndxExr passed in.
 RelIScanPlanOp* PlanOp = (RelIScanPlanOp*)AccInfo;
 if (!PlanOp) {
   REL_NOTE_ERROR(REL_BADEXEC, "PlanOp null for indexscan");
   return XXX_NOT_OK;
  }
 if (IndxExpr) {
   // i dont think this should ever be non-NULL: just checking
   cerr << "Curious:IScan:IndxExpr why does it ever come here" << endl;
 }
 IndxExpr = PlanOp->IndxPlan;
 XxxBooleanExpressionPlan* ResidExpr = PlanOp->ResidPlan;

 // Step1: match the index based on the known Name
 RelShoreDataSrc::RelShoreImplInfo::
 ShoreIndexImplInfo* ShoreIndxInfo = NULL;
 RelImplInfo::IndexList* IndexList;

 // step thru the indexes and see if any of them matches the 
 // specified index name
 if (CatalogInfo->ImplInfo() && 
     (IndexList = CatalogInfo->ImplInfo()->Indexes)) {
   while (IndexList->Next) {
     if (IndexList->Next->IndxInfo &&
	 !strcmp(IndexList->Next->IndxInfo->IndexName,
		 PlanOp->IndxName)) {
       ShoreIndxInfo = (RelShoreDataSrc::RelShoreImplInfo::
			ShoreIndexImplInfo *) IndexList->Next->IndxInfo;
       break;
      } 
     IndexList = IndexList->Next;
    }     // while
  }
 if (!ShoreIndxInfo) {
   REL_NOTE_ERROR(REL_BADEXEC, "Unable to find matching index");
   return XXX_NOT_OK;
  }

 // Step2 : call indexedAccessPath ...
 if (indexedAccessPath(IndxExpr, ResidExpr, Rec, Schema, Parameters, 
		       SibArity, CurrJoinPos, 
		       PlanOp->MatchedInfo, 
		       ShoreIndxInfo, CInfo) != XXX_OK){
   REL_NOTE_ERROR(REL_BADEXEC, "Error creating indexscan path");
   return XXX_NOT_OK;
  }

 return XXX_OK;
}

XxxErrCode 
XxxShoreFileRelation::indexedAccessPath(XxxBooleanExpressionPlan * IndxExpr,
					XxxBooleanExpressionPlan* ResidExpr,
					const XxxRecord * Rec[],
					const XxxRecordSchema * Schema[],
					const RelExecParameters * Parameters,
					int SibArity,
					int CurrJoinPos,
					const XxxMatchInfo* MatchedInfo,
					RelShoreDataSrc::RelShoreImplInfo::
					ShoreIndexImplInfo *ShoreIndxInfo,
					CursorInfo*& CInfo) const
{
 // Get appropriate range object from index
 XxxIDXClass::IndexRange *RangeSpec = 
   XxxIDXArray[ShoreIndxInfo->IndexType]->CreateRange();
 
 if (RangeSpec == NULL) {
   XXX_NOTE_ERROR (XXX_INTERNAL, 
		   "Don't know how to get range for this type of index!");
   return XXX_NOT_OK;
  }
 
 if (XxxIDXArray[ShoreIndxInfo->IndexType]->
     GetRangeSpec(IndxExpr,
		  ShoreIndxInfo->IndexExpr,
		  MatchedInfo,
		  Rec,Schema,Parameters,
		  SibArity, CurrJoinPos,
		  RangeSpec) != XXX_OK) {
   XXX_NOTE_ERROR (XXX_INTERNAL, 
		   "Indx matches, but no range spec!");
   return XXX_NOT_OK;
  }
 
 CInfo = XxxIDXArray[ShoreIndxInfo->IndexType]->
   GetCursor(RelationId,
	     ShoreIndxInfo->IndexId,
	     CatalogInfo->RecordSchema(),
	     IndxExpr, ResidExpr,
	     MatchedInfo->Exact,
	     RangeSpec,
	     Parameters,
	     Rec,
	     Schema,
	     CurrJoinPos);
 
 if (CInfo == NULL) {
   XXX_NOTE_ERROR (XXX_INTERNAL, 
		   "Could not create cursor for index scan.");
   return XXX_NOT_OK;
  }
 
 if (XXX_THISRELENV() && XXX_THISRELENV()->DebugStorage) {
   cerr << "Using " << XxxIDXArray[ShoreIndxInfo->IndexType]->TypeName
        << "index scan on expression ";
   ShoreIndxInfo->IndexExpr->Print(cerr);
   cerr << " of (SHORE file) relation " 
	<< CatalogInfo->RelName << endl;
 }

 // delete RangeSpec;
 // Praveen: too early to clean this up, because the
 // range spec is used in the index cursor...
 // Putting in this delete leads to a nasty memory bug
 // which is really tough to detect, so i'm leaving this
 // comment in as a warning. Instead, the RangeSpec is
 // maintained along with the cursor and deleted when
 // the cursor is deleted.
 return XXX_OK;
}

XxxErrCode 
XxxShoreFileRelation::CloseCursor(XxxSetCursor& SCsr) const
{
 delete ((CursorInfo *)(SCsr.State));
 SCsr.State = 0;
 return XXX_OK;
}

// Note: This function allocates the new object id and record, and it is
//       responsibility of the calling function to destroy them.
XxxErrCode 
XxxShoreFileRelation::NextItem1(XxxSetCursor& SCsr,
				XxxShoreObjectId*& ObjId,
				XxxRecord*& ReturnRecord) const
{
 XXX_SM_CHECK;
 
 CursorInfo* CInfo = (CursorInfo*) (SCsr.State);
 while (1) {
   XxxBool Success = XXX_FALSE;
   if (CInfo->NextRecord(ObjId, ReturnRecord, Success) != XXX_OK) {
     XXX_NOTE_ERROR(XXX_BADEXPR, "Error in cursor NextRecord");
     return XXX_NOT_OK;
    }
   // check if there was a new record
   if (!Success) break;

   XxxCoreADTClass::BoolValue CondResult(XXX_TRUE);
   // if it is an exact match cursor, return success
   XxxBool MainPredDone = XXX_FALSE;
   XxxBool ResidPredDone = XXX_FALSE;
   
   // maybe these compile-time checks can be moved out of here
   if (CInfo->ExactMatch || !CInfo->Filter) MainPredDone = XXX_TRUE;
   if (!CInfo->RFilter) ResidPredDone = XXX_TRUE;
   if (MainPredDone && ResidPredDone) return XXX_OK;

   // TODO: ISSUE: should the resid preds be applied before the
   // non-exact index preds? i think the answer is yes, given
   // the kinds of predicates that non-exact index preds
   // correspond to (like polygon overlaps)
   // otherwise the condition needs to be checked again
   if (!CInfo->SiblingRecs || !CInfo->SiblingSchema) {
     // if there is no join environment
     if (!ResidPredDone && 
	 CInfo->RFilter->Evaluate(ReturnRecord, 
				  CatalogInfo->RecordSchema(),
				  &CondResult, 
				  (CInfo->Params?CInfo->Params->Params:0)) 
	 != XXX_OK) {
       XXX_NOTE_ERROR(XXX_BADEXPR, "Cannot evaluate residual filter expression");
       return XXX_NOT_OK;
      }
     if (!MainPredDone && CondResult.Value &&
	 CInfo->Filter->Evaluate(ReturnRecord, 
				 CatalogInfo->RecordSchema(),
				 &CondResult, 
				 (CInfo->Params?CInfo->Params->Params:0)) 
	 != XXX_OK) {
       XXX_NOTE_ERROR(XXX_BADEXPR, "Cannot evaluate filter expression");
       return XXX_NOT_OK;
      }
    } // if there is no join environment
    else {
      // if there is a join environment
     if (!ResidPredDone && 
	  CInfo->RFilter->Evaluate((const XxxRecord **)CInfo->SiblingRecs,
				   (const XxxRecordSchema **)CInfo->SiblingSchema,
				   &CondResult,
				   (CInfo->Params?CInfo->Params->Params:0))
	 != XXX_OK) {
       XXX_NOTE_ERROR(XXX_BADEXPR, "Cannot evaluate residual filter expression");
       return XXX_NOT_OK;
      }
      if (!MainPredDone && CondResult.Value &&
	  CInfo->Filter->Evaluate((const XxxRecord **)CInfo->SiblingRecs,
				  (const XxxRecordSchema **)CInfo->SiblingSchema,
				  &CondResult,
				  (CInfo->Params?CInfo->Params->Params:0))
	  != XXX_OK) {
	XXX_NOTE_ERROR(XXX_BADEXPR, "Cannot evaluate filter expression");
	return XXX_NOT_OK;
       }
     } // else if there is a join environment

   if (CondResult.Value == XXX_TRUE) {
     return XXX_OK;
    }
   else {
     delete ReturnRecord;
     // reset the ReturnRecord: this is important!!
     ReturnRecord = 0;
     delete ObjId;
     ObjId = 0;
    }
  } // while loop
 
 SCsr.AtEnd = XXX_TRUE;
 return XXX_OK;
}

XxxErrCode
XxxShoreFileRelation::DeleteRecords(XxxBooleanExpressionPlan *FilterPlan,
				    int &NumDeleted) 
{
 XXX_SM_CHECK;
 NumDeleted = 0;

 // Open FScan on relation. We need to force an FScan 
 // because index scans can't handle deleting records from a table while 
 // scanning.

 XxxSetCursor Cursor;

 CursorInfo* CInfo = new FScanCursorInfo (RelationId, 
					  CatalogInfo->RecordSchema(),
					  FilterPlan, NULL, 
					  NULL, NULL, 
					  0);

 Cursor.State = CInfo;
 Cursor.AtEnd = XXX_FALSE;
 Cursor.NumberRead = 0;

 XxxRecord* Rec =NULL;
 XxxShoreObjectId* RecId =NULL;

 while ((NextItem1(Cursor, RecId, Rec) == XXX_OK)
	&& (Cursor.IsValid())) {

   if (delFromAllIndexes(Rec, RecId) != XXX_OK) {
     REL_NOTE_ERROR (REL_BADTABLE, "Unable to delete record from indexes");
     return XXX_NOT_OK;
   }
   // Now delete the actual tuple from the table.
   if (XxxShoreSM->DeleteRecord(RelationId.VolId, *RecId) != XXX_OK) {
     REL_NOTE_ERROR (REL_BADTABLE, "Unable to delete record");
     return XXX_NOT_OK;
    }
   NumDeleted++;
   
   // Clean up
   delete Rec;
   delete RecId; 
  } // while loop

 if (CloseCursor(Cursor) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADSHOREREL, "Unable to close scan of relation");
   return XXX_NOT_OK;
  }

 return XXX_OK;
}

XxxErrCode 
XxxShoreFileRelation::UpdateRecords(XxxBooleanExpressionPlan *FilterPlan,
				    XxxValExprPlanList* TargetPlanList,
				    int &NumUpdated)
{
 XXX_SM_CHECK;
 NumUpdated = 0;
 XxxRecordSchema* Schema = CatalogInfo->RecordSchema();

 // Open FScan on relation. We need to force an FScan 
 // because index scans can't handle deleting/changing records from 
 // a table while scanning.
 XxxSetCursor Cursor;
 CursorInfo* CInfo = new FScanCursorInfo (RelationId, 
					  Schema,
					  FilterPlan, NULL, 
					  NULL, NULL, 
					  0);
 Cursor.State = CInfo;
 Cursor.AtEnd = XXX_FALSE;
 Cursor.NumberRead = 0;

 XxxRecord* Rec =NULL;
 XxxShoreObjectId* RecId =NULL;

 XxxValuesArray ValArray(TargetPlanList);
 

 while ((NextItem1(Cursor, RecId, Rec) == XXX_OK)
	&& (Cursor.IsValid())) {
   // Every time we get into this loop, we have found a record that
   // needs to be updated. The strategy used is:
   //    a) remove the record from all indexes
   //    b) modify the record
   //      b.1) create a new record
   //      b.2) put the new values into the new record
   //      b.3) replace the old record with the new record
   //    c) insert it into all indexes
   if (delFromAllIndexes(Rec, RecId) != XXX_OK) {
     REL_NOTE_ERROR (REL_BADTABLE, "Unable to delete record from indexes");
     return XXX_NOT_OK;
   }
   ASSERT((Schema->Arity() == TargetPlanList->Count())); 
   // get the record values into the ValArray
   REL_CALL_FN(TargetPlanList->Evaluate((const XxxRecord**)&Rec, 
					(const XxxRecordSchema**)&Schema,
					ValArray.Values),
	       REL_BADEXEC, "Error evaluating value list");

   // now make a new record out of the ValArray
   XxxRecord* NewRec = new XxxRecord(ValArray, Schema);

   // Now update the actual tuple 
   REL_CALL_FN(XxxShoreSM->ReplaceRecord(RelationId.VolId, *RecId, NewRec),
	       REL_BADEXEC, "Unable to update record");

   // put back the new record into the indexes
   REL_CALL_FN(addToAllIndexes(NewRec, RecId),
	       REL_BADTABLE, "Unable to delete record from indexes");

   NumUpdated++;
   // Clean up
   if (Rec) delete Rec;
   if (NewRec) delete NewRec;
   if (RecId) delete RecId; 
  } // while loop

 REL_CALL_FN(CloseCursor(Cursor),
	     REL_BADSHOREREL, "Unable to close scan of relation");
 return XXX_OK;
}


XxxErrCode XxxShoreFileRelation::RemoveItem(XxxRecord* const& DelRecord)
{
 XXX_NOTE_ERROR(XXX_INCOMPL, "RemoveItem from relation");
 return XXX_NOT_OK;
}

// this is used to change logging properties, set up bulk loads, etc
XxxErrCode XxxShoreFileRelation::PrepareForLoad()
{
 REL_CALL_FN(XxxShoreSM->SetStoreProperty(RelationId.VolId, RelationId.LocalId,
					  XxxStorageManager::XXX_SHORE_APPEND),
	     REL_BADTABLE, "Unable to set store property");

 // Go thru each of the indexes and set the store property
 // Shore disallows setting the store property on an index
 // though :-(.
 /*******
   RelShoreDataSrc::RelShoreImplInfo::ShoreIndexImplInfo *IndexInfo;
   RelImplInfo::IndexList   *IndexList;
   RelImplInfo              *ImplInfo;
   
   ImplInfo = CatalogInfo->ImplInfo();
   XxxIndexType IndexType;
   
   if (ImplInfo && (IndexList = ImplInfo->Indexes)) {
   while (IndexList->Next) {
   IndexInfo = 
   (RelShoreDataSrc::RelShoreImplInfo::ShoreIndexImplInfo *)
   (IndexList->Next->IndxInfo);
   IndexType = IndexInfo->IndexType;
   if (XxxShoreSM->SetStoreProperty(RelationId.VolId, 
   IndexInfo->IndexId,
   XxxStorageManager::XXX_SHORE_APPEND) != XXX_OK) {
   REL_NOTE_ERROR (REL_BADTABLE, "Unable to set index store prop.");
   return XXX_NOT_OK;
   }
   IndexList = IndexList->Next;
   } // while there are indexes
   } // if there is an ImplInfo
   *******/

 // set the store property on the "Large ADT" file, if there is a
 // large ADT being loaded
 XxxRecordSchema* Schema = CatalogInfo->RecordSchema();
 XxxBool TouchLargeFile = XXX_FALSE;
 for (int i = 0; i < Schema->Arity(); i++) {
   if (!XxxADTArray[Schema->GetAttr(i)->Type]->
       IsSmall(Schema->GetAttr(i)->TypeCatInfo)) {
     TouchLargeFile = XXX_TRUE;
     break;
    }
  }
 if (TouchLargeFile) {
   REL_CALL_FN(XxxShoreSM->SetStoreProperty(XxxShoreSM->DefaultVol(), 
					    XxxShoreSM->PersistentFileId(),
					    XxxStorageManager::XXX_SHORE_APPEND),
		REL_BADTABLE, "Unable to set store property");
 }
 return XXX_OK;
}

// Add the record to any indexes 
XxxErrCode 
XxxShoreFileRelation::addToAllIndexes(const XxxRecord* RecordToAdd,
				      const XxxShoreObjectId* RecordId)
{
 // Step thru the indexes in the ImplInfo
 RelShoreDataSrc::RelShoreImplInfo::ShoreIndexImplInfo *IndexInfo;
 RelImplInfo::IndexList   *IndexList;
 RelImplInfo              *ImplInfo;
 XxxValueExprPlan         *ExprPlan;
 
 ImplInfo = CatalogInfo->ImplInfo ();
 XxxIndexType IndexType;
 
 if (ImplInfo && (IndexList = ImplInfo->Indexes)) {
   while (IndexList->Next) {
     IndexInfo = 
       (RelShoreDataSrc::RelShoreImplInfo::ShoreIndexImplInfo *)
	 (IndexList->Next->IndxInfo);
     
     ExprPlan = (XxxValueExprPlan *)IndexInfo->IndexPlan;
     IndexType = IndexInfo->IndexType;
     
     int KeySize = XxxADTArray[ExprPlan->BasicType()]->
       MaxObjectSize(ExprPlan->GenerateCatInfo());
     char *KeyPtr = new char[KeySize];
     XxxADTValue* KeyValuePtr = (XxxADTValue*)KeyPtr;
     KeyValuePtr->Init();
     ExprPlan->Evaluate(RecordToAdd, 
			CatalogInfo->RecordSchema(), 
			KeyValuePtr); 

     if (XxxIDXArray[IndexType]->InsertEntry(RelationId.VolId, 
					     IndexInfo->IndexId,
					     KeyValuePtr, 
					     RecordId,
					     ExprPlan) != XXX_OK) {
       REL_NOTE_ERROR (REL_BADDEV, 
		       "Unable to insert record into index");
       delete [] KeyPtr;
       return XXX_NOT_OK;
      }
     
     // Clean up
     delete [] KeyPtr;
     
     IndexList = IndexList->Next;
    } // while there are indexes
  } // if there is an ImplInfo
 return XXX_OK;
}

XxxErrCode 
XxxShoreFileRelation::delFromAllIndexes(const XxxRecord * RecordToDel,
					const XxxShoreObjectId* RecordId)
{
 // Step thru the indexes in the ImplInfo
 RelShoreDataSrc::RelShoreImplInfo::ShoreIndexImplInfo *IndexInfo;
 RelImplInfo::IndexList   *IndexList;
 RelImplInfo              *ImplInfo;
 XxxValueExprPlan         *ExprPlan;

 ImplInfo = CatalogInfo->ImplInfo ();
 XxxIndexType IndexType;

 // Now step through all indexes and remove any pointers to the tuple;
 // then remove the actual tuple from the relation
 if (ImplInfo && (IndexList = ImplInfo->Indexes)) {
   while (IndexList->Next) {
     // remove the tuple from the index
     IndexInfo = 
       (RelShoreDataSrc::RelShoreImplInfo::ShoreIndexImplInfo *)
	 (IndexList->Next->IndxInfo);
     
     ExprPlan = (XxxValueExprPlan *)IndexInfo->IndexPlan;
     IndexType = IndexInfo->IndexType;

     // Create the record's key for the index
     int KeySize = XxxADTArray[ExprPlan->BasicType()]->
       MaxObjectSize(ExprPlan->GenerateCatInfo());
     char *KeyPtr = new char[KeySize];
     XxxADTValue* KeyValuePtr = (XxxADTValue*)KeyPtr;
     KeyValuePtr->Init();
     ExprPlan->Evaluate(RecordToDel, 
			CatalogInfo->RecordSchema(), 
			KeyValuePtr); 

     if (XxxIDXArray[IndexType]->DeleteOneEntry(RelationId.VolId, 
						IndexInfo->IndexId,
						KeyValuePtr, 
						RecordId,
						ExprPlan) != XXX_OK) {
       REL_NOTE_ERROR (REL_BADDEV, 
		       "Unable to delete single record from index");
       delete [] KeyPtr;
       return XXX_NOT_OK;
     }
     
     delete [] KeyPtr;

     IndexList = IndexList->Next;
    } // while there are indexes
 } // if there is an IndxInfo
 return XXX_OK;
}

XxxErrCode 
XxxShoreFileRelation::addItem(XxxRecord * const &RecordToAdd)
{
 XXX_SM_CHECK;
 
 XxxShoreObjectId RecordId;
 
 // TODO:L3 if record has any in-memory representations
 // convert them to a disk representation using TypeStoreCopy :
 
 // now add the record to the relation
 if (XxxShoreSM->InsertRecord(RelationId.VolId, RelationId.LocalId,
			      RecordToAdd, RecordId) != XXX_OK) {
   REL_NOTE_ERROR (REL_BADTABLE, "Unable to insert record");
   return XXX_NOT_OK;
  }
 if (addToAllIndexes(RecordToAdd, &RecordId) != XXX_OK) {
   REL_NOTE_ERROR (REL_BADTABLE, "Unable to add record to indexes");
   return XXX_NOT_OK;
 }
 return XXX_OK;
}

XxxErrCode 
XxxShoreFileRelation::AddIndex(const char *IndexName, 
			       XxxValueExpression *Expr,
			       XxxIndexType IndxType,
			       XxxBool Unique, 
			       const void *IndxColMetaInfo)
{
 XXX_SM_CHECK;
 
 // create the index
 XxxShoreObjectId NewIndex;
 
 REL_CALL_FN(XxxIDXArray[IndxType]->Create(RelationId.VolId, NewIndex,
					   Expr, Unique),
	     REL_BADDEV, "Unable to create new index");
 
 // insert it into the Root Catalog
 int StrValSize = strlen(IndexName) + 1 + XxxCoreADTClass::ValueOffset(XXX_STRING);
 char *TmpADTVal = new char[StrValSize];
 XxxCoreADTClass::StrValue* StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
 strcpy(StrVal->Value, IndexName);
 REL_CALL_FN(XxxIDXArray[XXX_BTREE]->InsertEntry(RelationId.VolId,
						 Device->RootIndex,
						 StrVal, StrValSize,
						 XXX_STRING,
						 &NewIndex),
	     REL_BADDEV, "Unable to insert index into REL root");
 delete [] TmpADTVal;

 // Get Plan from Expr
 XxxExprPlan *Plan = NULL;
 REL_CALL_FN(Expr->Optimize(Plan),
	     REL_BADDEV, "Unable to optimize index expression");
 XxxValueExprPlan *ValuePlan = (XxxValueExprPlan *)Plan;

 // Get type of column that index is built on. Expr must be typed checked
 // for this to work. Where should this be done???
 XxxBasicType IndxColType;
 IndxColType = Expr->BasicType();

 // Result holds evaluated records
 int KeySize = XxxADTArray[IndxColType]->
   MaxObjectSize(ValuePlan->GenerateCatInfo());
 XxxADTValue *Result = (XxxADTValue*)(new char[KeySize]);
 Result->Init();
 
 // Open FScan on relation
 XxxSetCursor Cursor;
 if (InitCursor(Cursor)!= XXX_OK) {
   XXX_NOTE_ERROR(XXX_BADSCAN, "Unable to open scan of relation");
   return XXX_NOT_OK;
  }
 
 XxxRecord* Rec = NULL;
 XxxShoreObjectId* RecId = NULL;
 
 while ((NextItem1(Cursor, RecId, Rec) == XXX_OK)
	&& (Cursor.IsValid())) {
   // don't forget to delete RecId and Rec when done with them
   
   // Evaluate record to extract key
   Plan->Evaluate(Rec,CatalogInfo->RecordSchema(), 
		  Result, NULL);
   
   if (XxxIDXArray[IndxType]->InsertEntry(RelationId.VolId, NewIndex,
					  Result, RecId,
					  ValuePlan) != XXX_OK) {
     REL_NOTE_ERROR (REL_BADDEV, "Unable to insert record into index");
     delete Plan;
     delete ((char*)Result);
     delete Rec;
     delete RecId;
     return XXX_NOT_OK;
    }
   // Need to clean thes up because they get allocated in NextItem()
   delete RecId;
   delete Rec;
   RecId = NULL;
   Rec = NULL;
  } // Next item
 
 // Clean up

 delete Plan;
 delete ((char*)Result);

 if (CloseCursor(Cursor) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADSHOREREL, "Unable to close scan of relation");
   return XXX_NOT_OK;
  }
 
 // no need to add the index to the RelImplInfo index list. This
 // relation object is going to be destroyed right after this.
 
 return XXX_OK;
}

// open a scan on the SHORE file.
// if there's a filter expr, the scan is not exact. otherwise, it is
XxxShoreFileRelation::
FScanCursorInfo::FScanCursorInfo(const XxxShoreLogicalOID& RelId,
				 XxxRecordSchema* Schema,
				 XxxBooleanExpressionPlan* Expr,
				 const RelExecParameters* Parameters,
				 const XxxRecord** SibRecs,
				 const XxxRecordSchema** SibSchema,
				 int CurrJoinPos)
  : CursorInfo(Schema, Expr, NULL, // no residual expression
	     (Expr?XXX_FALSE:XXX_TRUE), 
	     Parameters, SibRecs, SibSchema, CurrJoinPos),
	     ScanIter(RelId.VolId, RelId.LocalId)
{
 RelationId = RelId;
}

XxxShoreFileRelation::FScanCursorInfo::~FScanCursorInfo()
{
 // close the scan on the SHORE file
}

XxxErrCode XxxShoreFileRelation::
FScanCursorInfo::NextRecord(XxxShoreObjectId*& ObjId, 
			    XxxRecord*& ReturnRec, XxxBool& Success)
{
 Success = XXX_FALSE;
 XxxErrCode rc = XXX_OK;

 pin_i* Pin = 0;
 bool AtEnd;
 w_rc_t RetVal;
 XXX_CALL((XXX_SHORE_RETFN(ScanIter.next(Pin, 0, AtEnd), RetVal)),
	  XXX_SHOREERR, "Couldnt scan file", rc);
 XXX_BADRC_RCRETURN(rc);

 if (AtEnd == true) {
   return XXX_OK;
  }
 if (!XxxShoreSM->PinValid(Pin, SH)) {
   XXX_NOTE_ERROR(XXX_SHOREERR, " Invalid pin ");
   return XXX_NOT_OK;
  }

 int TotalSize = Pin->body_size();
 ReturnRec = new XxxRecord(TotalSize);
 if (XxxShoreSM->GetObjectPreAlloc(Pin, (char *)(ReturnRec->DataArea),
				   TotalSize, 0, TotalSize) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_SHOREERR, " Error getting pinned record");
   Pin->unpin();
   return XXX_NOT_OK;
  }
 ObjId = new XxxShoreObjectId(XxxShoreSM->PinObjectId(Pin));
 Pin->unpin();
 Success = XXX_TRUE;
 return XXX_OK;
}

//Begin:Moustafa
// Open a scan on the stream.
// This open completely traverses the original table 
// (that has stream type) and registers each stream tuple
// in the stream manager. It applies any Filtering condiditon 
// on the non-stream type at open cursor step (Mar. 2003)

XxxShoreFileRelation::
SScanCursorInfo::SScanCursorInfo(CursorInfo*& CInfo)
  : CursorInfo(CInfo->RecordSchema, CInfo->Filter, CInfo->RFilter, 
	       CInfo->ExactMatch, CInfo->Params, 
	       CInfo->SiblingRecs, CInfo->SiblingSchema, CInfo->CurrJoinPos)
{
  RelCInfo = CInfo;

  // obtain all the streaming tuples from the Rel

  XxxBool Success = XXX_FALSE;
  XxxRecord* Rec =NULL;
  XxxShoreObjectId* RecId =NULL;

  // create a result queue to hold the results of streaming
  resQueue = new ResultQueue();
  XxxRecordSchema* Schema = RelCInfo->RecordSchema;  
  int StrmFldPos = -1;
  // get stream field position
  XxxBool IsStrm = XXX_FALSE;
  for(int i=0; i< Schema->RecArity; i++){
    XxxADTClass* attrADT = XxxADTArray[(Schema->Attrs+i)->Type];
    if (!XxxCoreADTClass::IsCoreADT((Schema->Attrs+i)->Type)){
      if(attrADT->IsSmall(NULL))
	IsStrm = ((XxxSmallADT*)attrADT)->IsStream();
      else
	IsStrm = ((XxxLargeADT*)attrADT)->IsStream();
      if (IsStrm){
	StrmFldPos = i;
	break;
      }
    }
  }
  ASSERT(StrmFldPos>=0);

  //

  while(1){
    if (RelCInfo->NextRecord(RecId, Rec, Success) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADEXPR, "Error in cursor NextRecord");
      break;
    }
    // check if there was a new record
    if (!Success) break;
    // TODO: A check should be made here on the returned tuples 
    //       as in NextItem1; for the time delay that

    // for each new tuple, create a stream request    
    StrmMgr->addNewStream(RecId, Rec, resQueue, Schema, StrmFldPos);
  }

}

XxxShoreFileRelation::
SScanCursorInfo::~SScanCursorInfo()
{
  
  // just delete all stream request that are attached to the resultqueue, and clean the queue
  // this should be done by the streaming manager
  if(resQueue)
    resQueue->Enough();
  // TODO: This part must be completed
  //  StrmMgr->RemoveStreams(resQueue);
  // now delete all the records in the resQueue
  //  resQueue->flush(RelCInfo->RecordSchema);
  //  delete resQueue;
  // close the scan on the stream
  delete ((CursorInfo *)(RelCInfo));
}

XxxErrCode XxxShoreFileRelation::
SScanCursorInfo::NextRecord(XxxShoreObjectId*& ObjId, 
			    XxxRecord*& ReturnRec, XxxBool& Success)
{
 Success = XXX_FALSE;
 XxxErrCode rc = XXX_OK;

 Tuple *tmpTuple = NULL;

 while(1){
   // get a record from the stream buffer ResultQueue)
   resQueue->dequeue(tmpTuple);  
   if(tmpTuple!=NULL){
     //TODO: apply any filtering on the data
     ObjId = tmpTuple->RecordId;
     ReturnRec =  tmpTuple->Record;
     delete tmpTuple;
     //tmpTuple = NULL;
     break;
   }
   else if (resQueue->IsDone()){
     // the stream is finished
     delete resQueue;
     resQueue = NULL;
     ReturnRec = NULL;
     return XXX_OK;
   }
   else {
     ReturnRec = NULL;
     Success = XXX_TRUE;
     me()->yield();     // the calling thread will wait in this case
     return XXX_OK;
   }
 };
 Success = XXX_TRUE;
 return XXX_OK;
}
//End: Moustafa
