#include <stdio.h>
#include <stdlib.h>
#include <fstream.h>
#include <dirent.h>
#include <sys/stat.h>

#ifdef __GNUG__
#pragma implementation
#endif

#include "xxx_thread.h"
#include "env.h"
#include "reladt.h"
#include "relcat.h"
#include "relexec.h"
#include "relerror.h"
#include "relation.h"
#include "values.h"
#include "shorerel.h"
#include "engines.h"  // XxxDBEngineArray
#include "releng.h"  // XxxDBEngineArray
#include "XxxIndex.h"
#include "funcexpr.h"
#include "relopt.h"
#include "alias.h"

//
// Methods for RelShoreCatalog
//
RelShoreDataSrc::
RelShoreCatalog::RelShoreCatalog(RelShoreDataSrc::RelShoreDevice* Dev,
				 const char *DevPath)
: RelCatalog(), device(Dev)
{
 TablesRelation = AttrsRelation = IndxsRelation = StatsRelation = 
	 ViewsRelation = 0;
 AttrStatsRelation = NULL;

 XxxBool Bootstrap = XXX_TRUE;


 // this creates all the catalog tables
 if ((getCatalogRel(RelShoreDataSrc::RelShoreDevice::RELTABLESCAT,
		    Bootstrap, TablesRelation) != XXX_OK) 
     || (getCatalogRel(RelShoreDataSrc::RelShoreDevice::RELATTRSCAT,
		       Bootstrap, AttrsRelation) != XXX_OK) 
     || (getCatalogRel(RelShoreDataSrc::RelShoreDevice::RELINDXSCAT,
		       Bootstrap, IndxsRelation) != XXX_OK) 
     || (getCatalogRel(RelShoreDataSrc::RelShoreDevice::RELSTATSCAT,
		       Bootstrap, StatsRelation) != XXX_OK) 
     || (getCatalogRel(RelShoreDataSrc::RelShoreDevice::RELATTRSTATSCAT,
		       Bootstrap, AttrStatsRelation) != XXX_OK) 
     || (getCatalogRel(RelShoreDataSrc::RelShoreDevice::RELVIEWSCAT,
		       Bootstrap, ViewsRelation) != XXX_OK) 
     ) {
   REL_NOTE_ERROR(REL_MSSNGCAT, "Error in catalog creation");
   return;
  }

 // now, for each of them, throw away the CatalogInfo and create a new
 // catalogInfo. The reason is that an accurate cataloginfo can only be
 // created after all of these relations are created.
 Bootstrap = XXX_FALSE;
 if ((getCatalogRel(RelShoreDataSrc::RelShoreDevice::RELTABLESCAT,
		    Bootstrap, TablesRelation) != XXX_OK) 
     || (getCatalogRel(RelShoreDataSrc::RelShoreDevice::RELATTRSCAT,
		       Bootstrap, AttrsRelation) != XXX_OK) 
     || (getCatalogRel(RelShoreDataSrc::RelShoreDevice::RELINDXSCAT,
		       Bootstrap, IndxsRelation) != XXX_OK) 
     || (getCatalogRel(RelShoreDataSrc::RelShoreDevice::RELSTATSCAT,
		       Bootstrap, StatsRelation) != XXX_OK) 
     || (getCatalogRel(RelShoreDataSrc::RelShoreDevice::RELATTRSTATSCAT,
		       Bootstrap, AttrStatsRelation) != XXX_OK) 
     || (getCatalogRel(RelShoreDataSrc::RelShoreDevice::RELVIEWSCAT,
		       Bootstrap, ViewsRelation) != XXX_OK) 
     ) {
   REL_NOTE_ERROR(REL_MSSNGCAT, "Error in catalog recreation");
   return;
  }
}

void RelShoreDataSrc::
RelShoreCatalog::destroyCatRelation(XxxRelation* Rel)
{
 if (Rel) {
   RelCatalogInfo* TmpCatInfo = NULL;
   if (Rel->CatalogInfo) 
     TmpCatInfo = (RelCatalogInfo*)Rel->CatalogInfo;
   delete Rel;
   if (TmpCatInfo)  {
     TmpCatInfo->Destroy();
     delete TmpCatInfo;
   }
 }
}
RelShoreDataSrc::RelShoreCatalog::~RelShoreCatalog()
{
 destroyCatRelation(TablesRelation);
 destroyCatRelation(AttrsRelation);
 destroyCatRelation(IndxsRelation);
 destroyCatRelation(StatsRelation);
 destroyCatRelation(ViewsRelation);
 destroyCatRelation(AttrStatsRelation);
}

XxxErrCode RelShoreDataSrc::
RelShoreCatalog::ModifyStatsInfo(const RelCatalogInfo* CatInfo,
				 const RelStatsInfo* StatsInfo)
{
 if (!StatsRelation || !AttrStatsRelation) {
   REL_NOTE_ERROR(REL_BADDEV, "No stats or attrstats relation");
   return XXX_NOT_OK;
  }
 XxxBooleanExpressionPlan *FilterPlan;
 int NumDel = 0;
 // Step 1) delete any old stats tuple for this relation
 if (makeFilterPlan("relname", 
		    RelShoreDataSrc::RelShoreDevice::RELSTATSCAT, 
		    CatInfo->RelationName(),FilterPlan) != XXX_OK) {
   REL_NOTE_ERROR (REL_BADDEV, "Unable to make filter expr");
   return XXX_NOT_OK;
  }
 if (StatsRelation->DeleteRecords(FilterPlan, NumDel) != XXX_OK) {
   REL_NOTE_ERROR (REL_BADSHOREREL, "Error deleting records");
   return XXX_NOT_OK;
  }
 delete FilterPlan;

 // Step 2) delete any old AttrStats tuples for this relation
 if (makeFilterPlan("relname", 
		    RelShoreDataSrc::RelShoreDevice::RELATTRSTATSCAT, 
		    CatInfo->RelationName(),FilterPlan) != XXX_OK) {
   REL_NOTE_ERROR (REL_BADDEV, "Unable to make filter expr");
   return XXX_NOT_OK;
  }
 if (AttrStatsRelation->DeleteRecords(FilterPlan, NumDel) != XXX_OK) {
   REL_NOTE_ERROR (REL_BADSHOREREL, "Error deleting records");
   return XXX_NOT_OK;
  }
 delete FilterPlan;

 // Step 3) insert the new stats tuple
 // create a stats record and add to the StatsRelation
 XxxRecordSchema* RecSchema = StatsRelation->CatalogInfo->RecordSchema();
 XxxValuesArray ValArray(RecSchema);
 RelShoreDevice::statsInfo *SInfo
   = RelShoreDevice::newStatsInfo(CatInfo->RelationName(), 
				  StatsInfo->Cardinality, 
				  StatsInfo->AvgTupleSize);
 XxxRecord* NewRec;
 if (RelShoreDevice::makeStatsRec(RecSchema,
				  ValArray, *SInfo, NewRec) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create Stats record");
   return XXX_NOT_OK;   
  }
 delete SInfo;

 if (StatsRelation->AddItem(NewRec) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADCAT, "Adding record to Stats relation");
   return XXX_NOT_OK;
  }
 NewRec->Reclaim(RecSchema); // reclaim the record

 // Step 4) Store any stats info on the individual attributes
 XxxRecordSchema* RecSchema1 = AttrStatsRelation->CatalogInfo->RecordSchema();
 XxxValuesArray ValArray1(RecSchema1);
 XxxRecord* NewRec1;
 ASSERT((StatsInfo->Arity == CatInfo->RecordSchema()->Arity()));
 for (int i = 0; i < StatsInfo->Arity; i++) {
   XxxAttribute* CurrentAttr = CatInfo->RecordSchema()->GetAttr(i);
   RelShoreDevice::astatsInfo *ASInfo
     = RelShoreDevice::newAStatsInfo(CatInfo->RelationName(),
				     i, 
				     CurrentAttr->Type,
				     CurrentAttr->TypeCatInfo,
				     StatsInfo->AttrStats[i]);
   if (RelShoreDevice::makeAStatsRec(RecSchema1,
				     ValArray1, *ASInfo, NewRec1) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADDEV, "Unable to create AttrStats record");
     return XXX_NOT_OK;
    }
   delete ASInfo;

   if (AttrStatsRelation->AddItem(NewRec1) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADCAT, "Adding record to AttrStats");
     return XXX_NOT_OK;
    }
   NewRec1->Reclaim(RecSchema1); // reclaim the record
  } // for each attribute
 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreCatalog::AddCatInfo(const char *RelName, 
			    const RelDataSrc::AttrTypeList* Attrs,
			    const RelDataSrc::KeyList* Keys,
			    const RelShoreImplType ImplType)
{
 const RelDataSrc::AttrTypeList* Tmp = Attrs;
 int AttrCount= 0;
 while (Tmp) {
   AttrCount++;
   Tmp = Tmp->Next;
  }
 // add the entry to the TablesRelation
 XxxRecordSchema* RecSchema = TablesRelation->CatalogInfo->RecordSchema();
 XxxValuesArray ValArray(RecSchema);
 RelShoreDevice::tablesInfo *TablesInfo
   = RelShoreDevice::newTablesInfo(RelName, AttrCount, ImplType);
 XxxRecord* NewRec;

 if (RelShoreDevice::makeTablesRec(RecSchema,
				   ValArray, *TablesInfo, NewRec) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create Tables record");
   return XXX_NOT_OK;   
  }
 delete TablesInfo;

 if (TablesRelation->AddItem(NewRec) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADCAT, "Unable to add record for Tables relation");
   return XXX_NOT_OK;
  }
 NewRec->Reclaim(RecSchema); // reclaim the record

 RecSchema = AttrsRelation->CatalogInfo->RecordSchema();
 // make the entries in the AttrsRelation
 XxxValuesArray ValArray1(RecSchema);
 AttrCount = 0;
 while (Attrs) {
   int KeyVal = 0;
   const RelDataSrc::KeyList * KeyList = Keys;
   while (KeyList) {
     if (!strcmp(KeyList->AttrName, Attrs->Name)) {
       KeyVal = 1;
       break;
      }
     KeyList = KeyList->Next;
    }

   RelShoreDevice::attrsInfo *AttrsInfo
     = RelShoreDevice::newAttrsInfo(RelName, Attrs->Name, KeyVal,
				    Attrs->Type, AttrCount++, 
				    &(Attrs->MIHndl));

   if (RelShoreDevice::makeAttrsRec(RecSchema,
				    ValArray1, *AttrsInfo, NewRec) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADDEV, "Unable to create Attrs record");
     return XXX_NOT_OK;   
    }
   delete AttrsInfo;

   if (AttrsRelation->AddItem(NewRec) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADCAT, "Unable to add record to Attrs relation");
     return XXX_NOT_OK;
    }
   NewRec->Reclaim(RecSchema); // reclaim the record
   Attrs = Attrs->Next;
  } // while

 // finally add an empty record to the Stats relation: TODO:L2

 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreCatalog::deleteFromRelation(XxxRelation* Rel,
				    const char* AttrName,
				    const char* AttrVal,
				    int& NumDeleted)
{
 XxxBooleanExpressionPlan *FilterPlan;
 // Remove the record in the TablesRelation corresponding to the relation
 if (makeFilterPlan(AttrName,
		    Rel->CatalogInfo->RelationName(),
		    AttrVal, FilterPlan) != XXX_OK) {
   REL_NOTE_ERROR (REL_BADSHOREREL, "Unable to make filter expr");
   return XXX_NOT_OK;
  }
  // delete the tables relation's record corresponding to the relation
  NumDeleted = 0;
  if (Rel->DeleteRecords(FilterPlan, NumDeleted) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADSHOREREL, "Error deleting record");
    return XXX_NOT_OK;
  }
 delete FilterPlan;
 return XXX_OK;
}

//-> AUTHOR: Mark Paskin  DATE: 1/3/97
XxxErrCode RelShoreDataSrc::
RelShoreCatalog::DeleteCatInfo(const char *RelName) 
{
 int NumDeleted;
 if (deleteFromRelation(TablesRelation,
			"name", RelName, NumDeleted) ||
     (NumDeleted < 1)) {
    REL_NOTE_ERROR(REL_BADCAT, "deleting from table catalog");
    return XXX_NOT_OK;
  }

  // delete the attrs relation's records corresponding to the relation
 if (deleteFromRelation(AttrsRelation,
			"relname", RelName, NumDeleted) ||
     (NumDeleted < 1)) {
    REL_NOTE_ERROR(REL_BADCAT, "deleting from attr catalog");
    return XXX_NOT_OK;
  }

  // if the stats relation is non-empty, delete the record associated with
  // the relation
  if (StatsRelation) {
    if (deleteFromRelation(StatsRelation,
			   "relname", RelName, NumDeleted)) {
      REL_NOTE_ERROR(REL_BADCAT, "deleting from stats catalog");
      return XXX_NOT_OK;
     }
   }
  if (AttrStatsRelation) {
    if (deleteFromRelation(AttrStatsRelation,
			   "relname", RelName, NumDeleted)) {
      REL_NOTE_ERROR(REL_BADCAT, "deleting from attrstats catalog");
      return XXX_NOT_OK;
     }
   }

  // if the indxs relation is non-empty, delete the record associated with
  // the relation
  if (IndxsRelation) {
    if (deleteFromRelation(IndxsRelation,
			   "relname", RelName, NumDeleted)) {
      REL_NOTE_ERROR(REL_BADCAT, "deleting from indx catalog");
      return XXX_NOT_OK;
     }
  }

  return XXX_OK;
}

// make a simple filter plan
XxxErrCode RelShoreDataSrc::
RelShoreCatalog::makeFilterPlan(const char* AttrName,
				const char* CatTableName,
				const char* MatchString,
				XxxBooleanExpressionPlan*& FilterPlan)
{
 XxxExpression *Exprs[2];

 int AttrTypeId = XXX_UNKNOWN;
 int AttrIndex = 0;

 // find the attribute from the attrsInfos array
 for (int i = 0; i < RelShoreDevice::numAttrsInfos; i++) {
   if ((!strcmp(RelShoreDevice::AttrsInfo[i].TableName, CatTableName)) 
      && (!strcmp(RelShoreDevice::AttrsInfo[i].AttrName, AttrName))){
     AttrTypeId = RelShoreDevice::AttrsInfo[i].TypeId;
     AttrIndex = RelShoreDevice::AttrsInfo[i].AttrIndex;
     break;
    }
  } // for
 if (AttrTypeId == XXX_UNKNOWN)
  {
   REL_NOTE_ERROR(REL_BADSHOREREL, "Unable to find attr");
   return XXX_NOT_OK;
  }
 
 // NOTE: these two have to be dynamically allocated, because the
 // destructor for the statically allocated Filter expression will
 // delete its subexpressions
 Exprs[0] = new XxxUnknownValue (AttrName, CatTableName, 
				 AttrTypeId, AttrIndex, 0);
 Exprs[1] = new XxxConstStringValue (MatchString);
 XxxRelBooleanExpression TableFilter (XXX_EQ_OP, 2, Exprs);

 if (TableFilter.TypeCheck () != XXX_OK)
  {
   REL_NOTE_ERROR(REL_BADSHOREREL, 
		  "Unable to type check filter expression");
   return XXX_NOT_OK;
  }

 XxxExprPlan* TmpFilterPlan;
 if (TableFilter.Optimize(TmpFilterPlan) != XXX_OK)
  {
   REL_NOTE_ERROR(REL_BADSHOREREL, 
		  "Unable to optimize the filter expression");
   return XXX_NOT_OK;
  }
 FilterPlan = (XxxBooleanExpressionPlan*)(TmpFilterPlan);
 return XXX_OK;
}
XxxErrCode RelShoreDataSrc::
RelShoreCatalog::getAttrInfo(const char *RelName, 
			     XxxRecordSchema* RecSchema) const
{
 // set up a cursor over the _SATTRS relation to extract all the 
 // attributes of this particular relation
 XxxBooleanExpressionPlan *FilterPlan;
 XxxSetCursor AttrCsr;
 XxxRecord *AttrRec;
 if (makeFilterPlan("relname", 
		    RelShoreDataSrc::RelShoreDevice::RELATTRSCAT, 
		    RelName, FilterPlan) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADSHOREREL, "Unable to make filter expr");
   return XXX_NOT_OK;
  }
 if (AttrsRelation->InitCursor1(AttrCsr, FilterPlan) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADSHOREREL, 
		   "Unable to open scan of _SATTRS relation");
   return XXX_NOT_OK;
  }
 int Counter = 0;

 while ((AttrsRelation->NextItem (AttrCsr, AttrRec) == XXX_OK) &&
	(AttrCsr.IsValid ()))
  {
   Counter++;
   // Collect the name, iskey, type, and index from the attr 
   // record
   //Bjarke, sep. 14. 2001: added const qualifier to variable
   const char *FldPtr;
   AttrRec->GetField (1, AttrsRelation->CatalogInfo->RecordSchema(), FldPtr);
   char *AttrName = ((XxxCoreADTClass::StrValue*)FldPtr)->Value;

   int IsKey;
   AttrRec->GetField (2, AttrsRelation->CatalogInfo->RecordSchema(), FldPtr);
   memcpy (&IsKey, &(((XxxCoreADTClass::IntValue*)FldPtr)->Value), sizeof (int));

   XxxBasicType AttrType;
   AttrRec->GetField (3, AttrsRelation->CatalogInfo->RecordSchema(), FldPtr);
   char *BasicTypeName = ((XxxCoreADTClass::StrValue*)FldPtr)->Value;

   {
     XxxUseMutex AutoMutex(*XxxADTClass::Latch);
     for (AttrType = 0; AttrType < XxxADTClass::NumADTs; AttrType++)
       if (!strcasecmp (XxxADTClass::ADTArray[AttrType]->TypeName, 
			BasicTypeName))
	 break;
     if (AttrType == XxxADTClass::NumADTs) {
       char buf[100];
       sprintf (buf, "Unknown type %s for attribute %s", BasicTypeName, 
		AttrName);
       XXX_NOTE_ERROR (XXX_BADTYPE, buf);
       AttrType = XXX_UNKNOWN;
     }
   }

   int AttrIndex;
   AttrRec->GetField(4, AttrsRelation->CatalogInfo->RecordSchema(), FldPtr);
   memcpy (&AttrIndex, &(((XxxCoreADTClass::IntValue*)FldPtr)->Value), sizeof (int));

   // Assign collected info to attribute
   (*RecSchema)[AttrIndex].Assign(AttrName, AttrType, 0);
   RecSchema->KeyVector[AttrIndex] = IsKey;

   // now fill in any catalog info about nested complex
   // attributes
   if (XxxADTArray[AttrType]->MaintainsMetaInfo()) {
     // retrieve the MIHndl
     XxxADTClass::MetaInfoHandle MIHndl;
     AttrRec->GetField(5, AttrsRelation->CatalogInfo->RecordSchema(), FldPtr);
     XxxADTArray[XXX_MIBLOB]->TypeCopy((XxxADTValue*)FldPtr,
				       (XxxADTValue *)&MIHndl, 0);

     char buf[REL_NAME_MAX + XXX_ATTR_NAME_MAX + 1];

     sprintf (buf, "%s.%s", RelName, AttrName);
     XxxBool MIFound = XXX_FALSE;

     if ((XxxADTArray[AttrType]->
	  GetMetaInfo (buf,
		       MIHndl,
		       (*RecSchema)[AttrIndex].TypeCatInfo,
		       MIFound) != XXX_OK) 
	 || (MIFound != XXX_TRUE)) {
       XXX_NOTE_ERROR (XXX_BADADT, buf);
       return XXX_NOT_OK;
      }
    }
   else {
     (*RecSchema)[AttrIndex].TypeCatInfo = 0;
    }

   // done with AttrRec
   AttrRec->Reclaim(AttrsRelation->CatalogInfo->RecordSchema());
  }

 if (AttrsRelation->CloseCursor (AttrCsr) != XXX_OK)
  {
   REL_NOTE_ERROR (REL_BADSHOREREL, "Unable to close scan of attrs relation");
   return XXX_NOT_OK;
  }
 delete FilterPlan;

 if (Counter != RecSchema->Arity())
  {
   char buf[100];
   sprintf (buf, "Disparity in attr count of %s, %d instead of %d",
	    RelName, Counter, RecSchema->Arity ());
   REL_NOTE_ERROR (REL_BADCAT, buf);
   return XXX_NOT_OK;
  }

 RecSchema->ComputeStats();

 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreCatalog::getStatInfo(const char *RelName, 
			     const XxxRecordSchema* RecSchema,
			     RelStatsInfo* RelStats) const
{
 //Bjarke, sep. 14. 2001: added const qualifier to variable
 const char *FldPtr;
 if (StatsRelation) {
   // set up a scan over RELSTATSCAT
   XxxSetCursor StatsCsr;
   XxxRecord *StatsRec;
   XxxBooleanExpressionPlan *FilterPlan;

   if (makeFilterPlan("relname", 
		      RelShoreDataSrc::RelShoreDevice::RELSTATSCAT,
		      RelName, FilterPlan) != XXX_OK) {
     REL_NOTE_ERROR (REL_BADSHOREREL, "Unable to make filter expr");
     return XXX_NOT_OK;
    }
   if (StatsRelation->InitCursor1(StatsCsr, FilterPlan) != XXX_OK) {
     REL_NOTE_ERROR (REL_BADSHOREREL, 
		     "Unable to open scan of stats relation");
     return XXX_NOT_OK;
    }

   // while() loop not needed -- relname is a key
   if ((StatsRelation->NextItem(StatsCsr, StatsRec) == XXX_OK) &&
	  (StatsCsr.IsValid())) {
     XxxRecordSchema* StatsSchema = 
       StatsRelation->CatalogInfo->RecordSchema();
     StatsRec->GetField(1, StatsSchema, FldPtr);
     memcpy(&(RelStats->Cardinality), &(((XxxCoreADTClass::IntValue*)FldPtr)->Value), sizeof(int));
     StatsRec->GetField(2, StatsSchema, FldPtr);
     memcpy(&(RelStats->AvgTupleSize), &(((XxxCoreADTClass::IntValue*)FldPtr)->Value), sizeof(int));
     StatsRec->Reclaim(StatsSchema);
    }

   if (StatsRelation->CloseCursor(StatsCsr) != XXX_OK)
    {
     REL_NOTE_ERROR (REL_BADSHOREREL, "close scan of stats relation");
     return XXX_NOT_OK;
    }
   delete FilterPlan;
  } // if StatsRelation
 else {
   RelStats->Cardinality = RelStatsInfo::DefaultTableCard;
   RelStats->AvgTupleSize = RelStatsInfo::DefaultTupleSize;
  }

 if (AttrStatsRelation) {
   // read in the _SASTATS relation
   XxxBooleanExpressionPlan *FilterPlan;
   XxxSetCursor AStatCsr;
   XxxRecord *AStatRec;
   XxxRecordSchema* AStatSchema = 
     AttrStatsRelation->CatalogInfo->RecordSchema();
   if (makeFilterPlan("relname", 
		      RelShoreDataSrc::RelShoreDevice::RELATTRSTATSCAT, 
		      RelName, FilterPlan) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADSHOREREL, "Unable to make filter expr");
     return XXX_NOT_OK;
    }
   if (AttrStatsRelation->InitCursor1(AStatCsr, FilterPlan) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADSHOREREL, 
		    "Unable to open scan of _SASTATS relation");
     return XXX_NOT_OK;
    }
   while ((AttrStatsRelation->NextItem (AStatCsr, AStatRec) == XXX_OK) &&
	  (AStatCsr.IsValid ()))
    {
     int AttrIndex;
     AStatRec->GetField(1, AStatSchema, FldPtr);
     memcpy (&AttrIndex, &(((XxxCoreADTClass::IntValue*)FldPtr)->Value), sizeof(int));

     // retrieve the MIHndl
     XxxADTClass::MetaInfoHandle MIHndl;
     AStatRec->GetField(2, AStatSchema, FldPtr);
     XxxADTArray[XXX_MIBLOB]->TypeCopy((XxxADTValue*)FldPtr, 
				       (XxxADTValue *)&MIHndl, 0);

     XxxADTMetaInfo* AttrMetaInfo = (*RecSchema)[AttrIndex].TypeCatInfo;
     if (XxxADTArray[(*RecSchema)[AttrIndex].Type]->
	  RetrieveStatsInfo(AttrMetaInfo, MIHndl,
			    RelStats->AttrStats[AttrIndex]) != XXX_OK){
       XXX_NOTE_ERROR (XXX_BADADT, "Unable to recover attr stats");
       return XXX_NOT_OK;
      }
     // done with AStatRec
     AStatRec->Reclaim(AStatSchema);
     if (AttrMetaInfo &&
	 AttrMetaInfo->AbsorbStatsInfo(RelStats->AttrStats[AttrIndex]) != XXX_OK) {
       XXX_NOTE_ERROR (XXX_BADADT, "Unable to absorb attr stats");
       return XXX_NOT_OK;
      }
    }
   if (AttrStatsRelation->CloseCursor(AStatCsr) != XXX_OK) {
     REL_NOTE_ERROR (REL_BADSHOREREL, "close scan of _SASTATS relation");
     return XXX_NOT_OK;
    }
   delete FilterPlan;
  }
 return XXX_OK;
}

//AUTHOR: Fabian Camargo DATE: 4/8/97
//PURPOSE: Gets all the indexes for a relation and constructs the 
// RelShoreImplInfo for the relation.
//NOTES: none 


XxxErrCode RelShoreDataSrc::
RelShoreCatalog::getImplInfo(const char *RelName, 
			     const XxxRecordSchema* RecSchema,
			     RelShoreImplType ImplType,
			     RelShoreDataSrc::RelShoreImplInfo* NewImplInfo) const
{
 NewImplInfo->ImplType = ImplType;
 // ImplInfo also has a DeviceIndex field!! this is currently
 // not being stored. for now, use the default
 NewImplInfo->DeviceIndex = 0; 

 //Bjarke, sep. 14. 2001: added const qualifier to variable
 const char *FldPtr;

 if (IndxsRelation) {
   // set up a scan and retrieve all the indexes
   XxxSetCursor IndxCsr;
   XxxRecord *IndxRec;
   XxxBooleanExpressionPlan *FilterPlan;

   if (makeFilterPlan("relname", 
		      RelShoreDataSrc::RelShoreDevice::RELINDXSCAT,
		      RelName, FilterPlan) != XXX_OK) {
     REL_NOTE_ERROR (REL_BADSHOREREL, "Unable to make filter expr");
     return XXX_NOT_OK;
    }
   if (IndxsRelation->InitCursor1(IndxCsr, FilterPlan) != XXX_OK) {
     REL_NOTE_ERROR (REL_BADSHOREREL, 
		     "Unable to open scan of indxs relation");
     return XXX_NOT_OK;
    }

   while ((IndxsRelation->NextItem(IndxCsr, IndxRec) == XXX_OK) &&
	  (IndxCsr.IsValid())) {
     IndxRec->GetField(1, IndxsRelation->CatalogInfo->RecordSchema(), FldPtr);
     char *IndxName = ((XxxCoreADTClass::StrValue*)FldPtr)->Value;
     XxxADTValue *IndxNameValue = (XxxADTValue*)FldPtr;

     IndxRec->GetField(2, IndxsRelation->CatalogInfo->RecordSchema(), FldPtr);
     char *IndxString = ((XxxCoreADTClass::StrValue*)FldPtr)->Value;

     // Also get index type. 
     IndxRec->GetField(3, IndxsRelation->CatalogInfo->RecordSchema(), FldPtr);
     char *IndxTypeString = ((XxxCoreADTClass::StrValue*)FldPtr)->Value;

     XxxIndexType IndxType= XXX_UNKNOWN;
     
     for (int i=0;i<XxxNumIDXs;i++) {
       if (strcmp(IndxTypeString,XxxIDXArray[i]->TypeName)==0 ||
	   strcmp(IndxTypeString,XxxIDXArray[i]->OldName)==0) {
	 IndxType=i;
	 break;
       }
     }
     // check type.
     if (IndxType == XXX_UNKNOWN) {
       REL_NOTE_ERROR (REL_BADDEV, 
		       "Index of unknown type.");
       return XXX_NOT_OK;
     }
     int IsUnique;
     // is it an unique index
     IndxRec->GetField(4, IndxsRelation->CatalogInfo->RecordSchema(), FldPtr);
     memcpy (&IsUnique, &(((XxxCoreADTClass::IntValue*)FldPtr)->Value),
	     sizeof (int));

     // is it a clustered index
     int IsClustered;
     IndxRec->GetField(5, IndxsRelation->CatalogInfo->RecordSchema(), FldPtr);
     memcpy (&IsClustered, &(((XxxCoreADTClass::IntValue*)FldPtr)->Value),
	     sizeof (int));

     // look up the index in the Root Catalog
     XxxShoreObjectId IndxId;
     bool Found;

     if (XxxIDXArray[XXX_BTREE]->GetEntry(device->VolumeId, device->RootIndex,
					  IndxNameValue,
					  strlen(IndxName)+1+XxxCoreADTClass::ValueOffset(XXX_STRING),
					  XXX_STRING,
					  (void *) (&IndxId),
					  sizeof (XxxShoreObjectId), 
					  Found) != XXX_OK) {
       REL_NOTE_ERROR (REL_BADDEV, 
		       "Error getting index from Root Index");
       return XXX_NOT_OK;
     }
     
     if (!Found) {
       char buf[100];
       sprintf (buf, "Error getting index:%s from Root index", IndxName);
       REL_NOTE_ERROR (REL_BADINDEX, buf);
       return XXX_NOT_OK;
      }

     RelImplInfo::IndexImplInfo* IndxInfo =
       new RelShoreImplInfo::ShoreIndexImplInfo(IndxId, IndxString, 
						IndxName, IndxType,
						(XxxBool)IsUnique,
						(XxxBool)IsClustered,
						RecSchema);
     if (!IndxInfo->IndexPlan) {
       // there was a failure in the constructor!
       REL_NOTE_ERROR (REL_BADINDEX, "Error in IndexImplInfo constr.");
       delete IndxInfo;
       return XXX_NOT_OK;
      }
     NewImplInfo->AddIndex(IndxInfo);

     IndxRec->Reclaim(IndxsRelation->CatalogInfo->RecordSchema());
    }

   if (IndxsRelation->CloseCursor (IndxCsr) != XXX_OK) {
     REL_NOTE_ERROR (REL_BADSHOREREL, "Unable to close scan of indxs relation");
     return XXX_NOT_OK;
    }
   delete FilterPlan;
 } // if IndxsRelation

 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreCatalog::GetCatInfo(const char *RelName, 
			    RelCatalogInfo& CatEntry,
			    XxxBool& CatFound) const
{
 CatFound = XXX_FALSE;

 // find the desired catalog info
 XxxSetCursor TableCsr;
 XxxRecord *TableRec;
 XxxBooleanExpressionPlan *FilterPlan;
 //Bjarke, sep. 14. 2001: added const qualifier to variable
 const char *FldPtr;

 if (makeFilterPlan("name", 
		    RelShoreDataSrc::RelShoreDevice::RELTABLESCAT, 
		    RelName, FilterPlan) != XXX_OK)
  {
   REL_NOTE_ERROR (REL_BADSHOREREL, "Unable to make filter expr");
   return XXX_NOT_OK;
  }

 if (TablesRelation->InitCursor1(TableCsr, FilterPlan) != XXX_OK) {
   REL_NOTE_ERROR (REL_BADSHOREREL, "Unable to open scan of tables relation");
   return XXX_NOT_OK;
  }
 if (TablesRelation->NextItem (TableCsr, TableRec) != XXX_OK) {
   REL_NOTE_ERROR (REL_BADSHOREREL, "Unable to scan tables relation");
   return XXX_NOT_OK;
  }
 if (TableCsr.IsValid ())
  {
   // Construct a catalog entry for this table.

   // Initialized Relation Name
   strcpy (CatEntry.RelName, RelName);

   // Initialized Record Schema
   int RelArity;
   TableRec->GetField(1, TablesRelation->CatalogInfo->RecordSchema(), FldPtr);
   memcpy (&RelArity, &(((XxxCoreADTClass::IntValue*)FldPtr)->Value),
	   sizeof (int));

   CatEntry.recSchema = new XxxRecordSchema(RelArity);
   // also determine the impl type
   int ImplType;

   TableRec->GetField(2, TablesRelation->CatalogInfo->RecordSchema(), FldPtr);
   memcpy (&ImplType, &(((XxxCoreADTClass::IntValue*)FldPtr)->Value),
	   sizeof (int));

   // done with TableRec
   TableRec->Reclaim(TablesRelation->CatalogInfo->RecordSchema());
   delete FilterPlan;
   
   if (getAttrInfo(RelName, CatEntry.recSchema) != XXX_OK) {
     REL_NOTE_ERROR (REL_BADSHOREREL, "Error getting AttrInfo");
     return XXX_NOT_OK;
    }

   // Initialize Implementation Info
   RelShoreImplInfo *NewImplInfo = new RelShoreImplInfo;
   CatEntry.implInfo = NewImplInfo;
   if (getImplInfo(RelName, CatEntry.RecordSchema(), 
		   (RelShoreImplType) ImplType,
		   NewImplInfo) != XXX_OK) {
     REL_NOTE_ERROR (REL_BADSHOREREL, "Error getting ImplInfo");
     return XXX_NOT_OK;
    }

   // Initialize Stats Info
   CatEntry.statsInfo = new RelStatsInfo(CatEntry.RecordSchema()->Arity());
   if (getStatInfo(RelName, CatEntry.RecordSchema(),
		   CatEntry.statsInfo) != XXX_OK) {
     REL_NOTE_ERROR (REL_BADSHOREREL, "Error getting StatInfo");
     return XXX_NOT_OK;
    }

   CatFound = XXX_TRUE;
  }       // if TableCsr.IsValid()

 if (TablesRelation->CloseCursor (TableCsr) != XXX_OK)  {
   REL_NOTE_ERROR (REL_BADSHOREREL, "Unable to close scan of tables relation");
   return XXX_NOT_OK;
  }

 return XXX_OK;
}

//-> AUTHOR: Mark Paskin  DATE: 1/3/97
XxxBool  RelShoreDataSrc:: 
RelShoreCatalog::GetRelationOfIndex(const char *IndxName, 
				    char *&RelName) const 
{
 XxxBooleanExpressionPlan *FilterPlan;
 if (makeFilterPlan("indxname", 
		    RelShoreDataSrc::RelShoreDevice::RELINDXSCAT, 
		    IndxName, FilterPlan) != XXX_OK) {
   REL_NOTE_ERROR (REL_BADSHOREREL, "Unable to make filter expr");
   return XXX_NOT_OK;
  }

 XxxSetCursor IndxCsr;
 XxxRecord *IndxRec;
 
 if (IndxsRelation->InitCursor1(IndxCsr, FilterPlan) != XXX_OK) {
   REL_NOTE_ERROR (REL_BADSHOREREL, "Unable to open scan of indxs relation");
   return XXX_FALSE;
  }
 if (IndxsRelation->NextItem(IndxCsr, IndxRec) != XXX_OK) {
   REL_NOTE_ERROR (REL_BADSHOREREL, "Unable to scan indxs relation");
   return XXX_FALSE;
  }
 if (IndxCsr.IsValid()) {
   if (IndxsRelation->CloseCursor(IndxCsr) != XXX_OK) {
     REL_NOTE_ERROR (REL_BADSHOREREL,
		     "Unable to close scan of indxs relation");
     return XXX_FALSE;
    }
   //Bjarke, sep. 14. 2001: added const qualifier to variable
   const char *FldPtr;
   IndxRec->GetField(0, IndxsRelation->CatalogInfo->RecordSchema(), FldPtr);
   char *RName = ((XxxCoreADTClass::StrValue*)FldPtr)->Value;

   RelName = new char[strlen(RName) + 1];
   strcpy(RelName, RName);
   return XXX_TRUE;
  } 
 else {
   if (IndxsRelation->CloseCursor (IndxCsr) != XXX_OK) {
     REL_NOTE_ERROR (REL_BADSHOREREL,
		     "Unable to close scan of indxs relation");
     return XXX_FALSE;
    }
   return XXX_FALSE;
  }
}

//-> AUTHOR: Mark Paskin  DATE: 1/3/97
XxxErrCode RelShoreDataSrc::
RelShoreCatalog::DeleteIndex(const char *IndxName) 
{
 // This function removes all traces of an index from the catalog;
 // this includes both the IndxsRelation AND any and all indexes created
 // on it.  This _should_ not cause a problem when it is an index on 
 // the IndxsRelation; first the tuple and the key in all indexes are
 // deleted, and then the index itself is.

 XxxBooleanExpressionPlan *FilterPlan;
 if (makeFilterPlan("indxname", 
		    RelShoreDataSrc::RelShoreDevice::RELINDXSCAT, 
		    IndxName, FilterPlan) != XXX_OK) {
   REL_NOTE_ERROR (REL_BADSHOREREL, "Unable to make filter expr");
   return XXX_NOT_OK;
  }
 int NumDeleted = 0;
 if (IndxsRelation->DeleteRecords(FilterPlan, NumDeleted) != XXX_OK){
   REL_NOTE_ERROR (REL_BADCAT, 
		   "Unable to delete record of index from indxs relation");
   delete FilterPlan;
   return XXX_NOT_OK;
  }
 delete FilterPlan;
 if (NumDeleted != 1) {
   char Buf[100];
   sprintf (Buf, "Removed %d records of index %s from indxs relation!",
	    NumDeleted, IndxName);
   REL_NOTE_ERROR (REL_BADCAT, Buf);
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreCatalog::replaceCatInfo(const char* TableName,
				const RelCatalogInfo*& CInfo)
{
 RelCatalogInfo* CatInfo = new RelCatalogInfo(TableName);
 XxxBool CatFound = XXX_FALSE;
 if ((GetCatInfo(TableName, *CatInfo, CatFound) != XXX_OK) || 
     (CatFound != XXX_TRUE)) {
   CatInfo->Destroy(); 
   delete CatInfo;
   REL_NOTE_ERROR(REL_MSSNGCAT, "Unable to find catalog entry");
   return XXX_NOT_OK;
  }
 // need to cast it so that the compiler doesnt complain about
 // destroying the contents of a const pointer.
 RelCatalogInfo* TmpCatInfo = (RelCatalogInfo*)CInfo;
 if (TmpCatInfo) {
   TmpCatInfo->Destroy(); 
   delete TmpCatInfo;
  }
 CInfo = CatInfo;

 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreCatalog::getCatalogRel(const char* RelName,
			       XxxBool Bootstrap,
			       XxxRelation*& Rel)
{
 if (Bootstrap) {
   RelCatalogInfo* CatEntry = new RelCatalogInfo(RelName);
   // first add the catalog entry for RelName
   CatEntry->recSchema = RelShoreDevice::getInitSchema(RelName);
   CatEntry->statsInfo = new RelStatsInfo();
   // Need to create the ImplInfo: TODO:L2 Leave it as 0 for now

   RelHandle* Hndl;
   if (device->GetRelation(RelName, Hndl) != XXX_TRUE) {
     XXX_NOTE_ERROR(XXX_BADFILE, "Unable to access catalog table");
     return XXX_NOT_OK;
    }

   // then get the relation from the handle
   if (device->GetRelation((RelShoreHandle*)Hndl, Rel,
			   CatEntry->ImplInfo()) != XXX_OK) {
     XXX_NOTE_ERROR(XXX_INTERNAL, "Handle exists, but no relation");
     return XXX_NOT_OK;
    }

   Rel->CatalogInfo = CatEntry;
   delete Hndl;
  }
 else {
   if (replaceCatInfo(RelName, Rel->CatalogInfo) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADCAT, "Error replacing catalog info");
     return XXX_NOT_OK;
    }
  }
 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreCatalog::addItem(const RelCatalogInfo& CatInfo)
{
 if (!Valid()) {
   REL_NOTE_ERROR(REL_BADCAT, "Invalid catalog");
   return XXX_NOT_OK;
  }
 // construct a record for the Tables relation and insert it

 // construct a record for the Attrs relation and insert it

 // construct a record for the Stats relation and insert it
 
 XXX_NOTE_ERROR(XXX_INCOMPL, "addItem not done yet");
 return XXX_NOT_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreCatalog::RemoveItem(const RelCatalogInfo& CatInfo)
{
 if (!Valid()) {
   REL_NOTE_ERROR(REL_BADCAT, "Invalid catalog");
   return XXX_NOT_OK;
  }
 XXX_NOTE_ERROR(XXX_INCOMPL, "RemoveItem not done yet");
 return XXX_NOT_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreCatalog::InitCursor(XxxSetCursor& SCsr) const
{
 if (!Valid()) {
   REL_NOTE_ERROR(REL_BADCAT, "Invalid catalog");
   return XXX_NOT_OK;
  }
 return TablesRelation->InitCursor(SCsr);
}

XxxErrCode RelShoreDataSrc::
RelShoreCatalog::CloseCursor(XxxSetCursor& SCsr) const
{
 if (!Valid()) {
   REL_NOTE_ERROR(REL_BADCAT, "Invalid catalog");
   return XXX_NOT_OK;
  }
 return TablesRelation->CloseCursor(SCsr);
}

XxxErrCode RelShoreDataSrc::
RelShoreCatalog::NextItem(XxxSetCursor& SCsr, RelCatalogInfo& CatEntry) const
{
 XXX_SM_CHECK;

 XxxRecord* TableRec;
 if (TablesRelation->NextItem(SCsr, TableRec) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADCAT, "Error during NextItem");
   return XXX_NOT_OK;
  }
 if (!SCsr.IsValid()) {
   return XXX_OK;
  }
 char RelName[REL_RELNAME_MAX];
 // construct a catalog entry for this table
 //Bjarke, sep. 14. 2001: added const qualifier to variable
 const char *FldPtr;
 TableRec->GetField(0, TablesRelation->CatalogInfo->RecordSchema(), FldPtr);
 char *TabName = ((XxxCoreADTClass::StrValue*)FldPtr)->Value;
 strcpy(RelName, TabName);
	
 XxxBool CatFound = XXX_FALSE;
 if ((GetCatInfo(RelName, CatEntry, CatFound) != XXX_OK) ||
     !CatFound) {
   REL_NOTE_ERROR(REL_BADCAT, "Error getting catalog entry");
   return XXX_NOT_OK;   
 }

 return XXX_OK;
}

//
// Methods of RelShoreImplInfo
//
void RelShoreDataSrc::
RelShoreImplInfo::Print(ostream& OStr) const
{
 OStr << "Impl Type: ";
 switch (ImplType) {
  case REL_SHORE_FILE_IMPL: OStr << "File" << endl; break;
 default: OStr << "Unknown" << endl; break;
 }
 RelImplInfo::Print(OStr);
}

void RelShoreDataSrc::
RelShoreImplInfo::ShoreIndexImplInfo::Print(ostream& OStr) const
{
 OStr << IndexName << ':' << IndexString;
}

