#ifndef REL_SHORERELCAT_H
#define REL_SHORERELCAT_H

//
// This is included from within shorerel.h.
// RelShoreCatalog is a nested class within RelShoreDataSrc
//

/**
 *  A catalog maintains the meta-information for a relational device
 */
class RelShoreCatalog: public RelCatalog {
  friend RelShoreDevice;
 protected:
  RelShoreDevice* device;
  virtual XxxErrCode addItem(const RelCatalogInfo & item);
  // called from destructor
  void destroyCatRelation(XxxRelation* Rel);
  // called from within constructor
  // if Bootstrap is true, the catalogs are manually created
  XxxErrCode getCatalogRel(const char* RelName,
			   XxxBool Bootstrap,
			   XxxRelation*&Rel);
  // called from within getCatalogRel
  XxxErrCode replaceCatInfo(const char* TableName,
				   const RelCatalogInfo*& CInfo);
  // used to make filters for scans over catalog relations
  static XxxErrCode makeFilterPlan(const char* AttrName,
				   const char* CatTableName,
				   const char* MatchString,
				   XxxBooleanExpressionPlan*& FilterPlan);
  // called from within DeleteCatInfo
  static XxxErrCode deleteFromRelation(XxxRelation* Rel,
				       const char* AttrName,
				       const char* AttrVal,
				       int& NumDeleted);

  // the next methods are called from GetCatInfo
  XxxErrCode getAttrInfo(const char* RelName, 
			 XxxRecordSchema* RecSchema) const;
  XxxErrCode getStatInfo(const char* RelName, 
			 const XxxRecordSchema* RecSchema,
			 RelStatsInfo* RelStats) const;
  XxxErrCode getImplInfo(const char* RelName,
			 const XxxRecordSchema* RecSchema,
			 RelShoreImplType ImplType,
			 RelShoreImplInfo* NewImplInfo) const;

 public:
  XxxRelation* TablesRelation; // _relname, arity, impltype
  XxxRelation* AttrsRelation;  // _relname, _attrname, key, type, indx
  XxxRelation* IndxsRelation;  // _rel_name, _indx_name, indx_fields, type
  XxxRelation* StatsRelation;  // _rel_name, num_tuples, avg_tuple_size
  XxxRelation* AttrStatsRelation;  // _rel_name, attr_name, stats_info
  XxxRelation* ViewsRelation;  // viewname, viewinfo

  RelShoreCatalog(RelShoreDevice* Dev, const char *DevPath);
  virtual ~RelShoreCatalog();

  //-> AUTHOR: Mark Paskin DATE: 1/3/97
  virtual XxxErrCode DeleteCatInfo (const char* RelName);
  //-> AUTHOR: Mark Paskin DATE: 1/3/97
  XxxBool GetRelationOfIndex (const char *IndxName,
			      char *&RelName) const;
  //-> AUTHOR: Mark Paskin DATE: 1/3/97
  virtual XxxErrCode DeleteIndex (const char *IndxName);
  XxxErrCode AddCatInfo(const char* RelName,
			const RelDataSrc::AttrTypeList* Attrs,
			const RelDataSrc::KeyList* Keys,
			const RelShoreImplType ImplType);
  XxxErrCode ModifyStatsInfo(const RelCatalogInfo* CatInfo,
			     const RelStatsInfo* StatsInfo);
  virtual XxxErrCode GetCatInfo(const char* RelName, RelCatalogInfo &,
				XxxBool& Found) const;
  XxxBool Valid() const
   { return (TablesRelation && AttrsRelation && 
	     StatsRelation && IndxsRelation); }

  // to iterate over the RelCatalogInfo
  virtual XxxErrCode InitCursor(XxxSetCursor&) const;
  virtual XxxErrCode NextItem(XxxSetCursor&, RelCatalogInfo&) const;
  virtual XxxErrCode CloseCursor(XxxSetCursor&) const;
  virtual XxxErrCode RemoveItem(const RelCatalogInfo & item);
 };

#endif
