#include <stdio.h>
#include <stdlib.h>
#include <fstream.h>
#include <dirent.h>
#include <sys/stat.h>

#ifdef __GNUG__
#pragma implementation
#endif

#include "xxx_thread.h"
#include "env.h"
#include "reladt.h"
#include "relcat.h"
#include "relexec.h"
#include "relerror.h"
#include "relation.h"
#include "values.h"
#include "shorerel.h"
#include "XxxIndex.h"
#include "releng.h"

#define yyFlexLexer sqlFlexLexer
#include <FlexLexer.h>

const char *const RelShoreDataSrc::RelShoreDevice::RELROOTCAT = "_RELROOT";
const char *const RelShoreDataSrc::RelShoreDevice::RELTABLESCAT = "_STABLES";
const char *const RelShoreDataSrc::RelShoreDevice::RELTABINDXCAT = "_STABLES_TABLENAME";
const char *const RelShoreDataSrc::RelShoreDevice::RELATTRSCAT = "_SATTRS";
const char *const RelShoreDataSrc::RelShoreDevice::RELATTRSINDXCAT = "_SATTRS_ATTRNAME";
const char *const RelShoreDataSrc::RelShoreDevice::RELSTATSCAT = "_SSTATS";
const char *const RelShoreDataSrc::RelShoreDevice::RELSTATSINDXCAT = "_SSTATS_TABLENAME";
const char *const RelShoreDataSrc::RelShoreDevice::RELATTRSTATSCAT = "_SASTATS";
const char *const RelShoreDataSrc::RelShoreDevice::RELATTRSTATSINDXCAT = "_SASTATS_TABLENAME";
const char *const RelShoreDataSrc::RelShoreDevice::RELINDXSCAT = "_SINDXS";
const char *const RelShoreDataSrc::RelShoreDevice::RELINDXSINDXCAT1 = "_SINDXS_TABLENAME";
const char *const RelShoreDataSrc::RelShoreDevice::RELINDXSINDXCAT2 = "_SINDXS_INDXNAME";
const char *const RelShoreDataSrc::RelShoreDevice::RELVIEWSCAT = "_SVIEWS";
const char *const RelShoreDataSrc::RelShoreDevice::RELVIEWSINDXCAT = "_SVIEWS_VIEWNAME";
const char *const RelShoreDataSrc::RelShoreDevice::RELSUPERFILE = "_RELSUPERFILE";

//
// These following const definitions allow you to change the
// default catalogs. Be careful modifying this -- it is pretty
// tricky.
//
const int RelShoreDataSrc::RelShoreDevice::numTablesInfos = 6;
RelShoreDataSrc::RelShoreDevice::tablesInfo 
RelShoreDataSrc::RelShoreDevice::TablesInfo[RelShoreDataSrc::RelShoreDevice::numTablesInfos] = {
 {RELTABLESCAT, 3, REL_SHORE_FILE_IMPL},
 {RELATTRSCAT, 6, REL_SHORE_FILE_IMPL},
 {RELSTATSCAT, 3, REL_SHORE_FILE_IMPL},
 {RELATTRSTATSCAT, 3, REL_SHORE_FILE_IMPL},
 {RELINDXSCAT, 6, REL_SHORE_FILE_IMPL},
 {RELVIEWSCAT, 2, REL_SHORE_FILE_IMPL}
 };

const int RelShoreDataSrc::RelShoreDevice::numAttrsInfos = 23;
RelShoreDataSrc::RelShoreDevice::attrsInfo 
RelShoreDataSrc::RelShoreDevice::AttrsInfo[RelShoreDataSrc::RelShoreDevice::numAttrsInfos] = {
 {RELTABLESCAT, "name", 1, XXX_STRING, 0, &(XxxADTClass::NullMIHndl)},
 {RELTABLESCAT, "arity", 0, XXX_INT, 1, &(XxxADTClass::NullMIHndl)},
 {RELTABLESCAT, "impltype", 0, XXX_INT, 2, &(XxxADTClass::NullMIHndl)},
 {RELATTRSCAT, "relname", 1, XXX_STRING, 0, &(XxxADTClass::NullMIHndl)},
 {RELATTRSCAT, "attrname", 1, XXX_STRING, 1, &(XxxADTClass::NullMIHndl)},
 {RELATTRSCAT, "key", 0, XXX_INT, 2, &(XxxADTClass::NullMIHndl)},
 // important to put the name and not the id of the attr type, because of 
 // extensibility, and the need to have plug-in-plug-out modules
 {RELATTRSCAT, "attrtype", 0, XXX_STRING, 3, &(XxxADTClass::NullMIHndl)},
 {RELATTRSCAT, "attrindex", 0, XXX_INT, 4, &(XxxADTClass::NullMIHndl)},
 {RELATTRSCAT, "attr_mihndl", 0, XXX_MIBLOB, 5, &(XxxADTClass::NullMIHndl)},
 {RELSTATSCAT, "relname", 1, XXX_STRING, 0, &(XxxADTClass::NullMIHndl)},
 {RELSTATSCAT, "cardinality", 0, XXX_INT, 1, &(XxxADTClass::NullMIHndl)},
 {RELSTATSCAT, "tuplesize", 0, XXX_INT, 2, &(XxxADTClass::NullMIHndl)},
 {RELVIEWSCAT, "viewname", 1, XXX_STRING, 0, &(XxxADTClass::NullMIHndl)},
 {RELVIEWSCAT, "viewinfo", 0, XXX_STRING, 1, &(XxxADTClass::NullMIHndl)},
 {RELATTRSTATSCAT, "relname", 1, XXX_STRING, 0, &(XxxADTClass::NullMIHndl)},
 {RELATTRSTATSCAT, "attrindex", 1, XXX_INT, 1, &(XxxADTClass::NullMIHndl)},
 {RELATTRSTATSCAT, "stat_info", 0, XXX_MIBLOB, 2, &(XxxADTClass::NullMIHndl)},
 {RELINDXSCAT, "relname", 1, XXX_STRING, 0, &(XxxADTClass::NullMIHndl)},
 {RELINDXSCAT, "indxname", 1, XXX_STRING, 1, &(XxxADTClass::NullMIHndl)},
 {RELINDXSCAT, "indxinfo", 0, XXX_STRING, 2, &(XxxADTClass::NullMIHndl)},
 {RELINDXSCAT, "indxtype", 0, XXX_STRING, 3, &(XxxADTClass::NullMIHndl)},
 {RELINDXSCAT, "unique", 0, XXX_INT, 4, &(XxxADTClass::NullMIHndl)},
 {RELINDXSCAT, "clustered", 0, XXX_INT, 5, &(XxxADTClass::NullMIHndl)},
 };

const int RelShoreDataSrc::RelShoreDevice::numStatsInfos = 0;
RelShoreDataSrc::RelShoreDevice::statsInfo 
RelShoreDataSrc::RelShoreDevice::StatsInfo[RelShoreDataSrc::RelShoreDevice::numStatsInfos+1]; // VC++5.0 doesnt like 0 size array

const int RelShoreDataSrc::RelShoreDevice::numViewsInfos = 0;
RelShoreDataSrc::RelShoreDevice::viewsInfo 
RelShoreDataSrc::RelShoreDevice::ViewsInfo[RelShoreDataSrc::RelShoreDevice::numViewsInfos+1]; // VC++5.0 doesnt like 0 size array

// The IndexType field has been changed to an XxxIndexType for extensibility.
// See createIndxsEntry to see how it's used.
const int RelShoreDataSrc::RelShoreDevice::numIndxsInfos = 7;
RelShoreDataSrc::RelShoreDevice::indxsInfo 
RelShoreDataSrc::RelShoreDevice::IndxsInfo[RelShoreDataSrc::RelShoreDevice::numIndxsInfos] = {
 {RELTABLESCAT, RELTABINDXCAT, "_STABLES.name",XXX_BTREE, 1, 1},
 {RELATTRSCAT, RELATTRSINDXCAT, "_SATTRS.attrname",XXX_BTREE, 0, 0},
 {RELSTATSCAT, RELSTATSINDXCAT, "_SSTATS.relname",XXX_BTREE, 0, 0},
 {RELVIEWSCAT, RELVIEWSINDXCAT, "_SVIEWS.viewname",XXX_BTREE, 0, 0},
 {RELATTRSTATSCAT, RELATTRSTATSINDXCAT, "_SASTATS.relname",XXX_BTREE, 0, 0},
 {RELINDXSCAT, RELINDXSINDXCAT1, "_SINDXS.relname",XXX_BTREE, 0, 0},
 {RELINDXSCAT, RELINDXSINDXCAT2, "_SINDXS.indxname",XXX_BTREE, 0, 0}
 };

XxxRecordSchema* RelShoreDataSrc::
RelShoreDevice::getInitSchema(const char* CatName)
{
 int SchemaArity = getSchemaArity(CatName);

 XxxRecordSchema* RecSchema = new XxxRecordSchema(SchemaArity);

 for (int i = 0; i < numAttrsInfos; i++) {
   if (!strcmp(AttrsInfo[i].TableName, CatName)) {
     int AttrIndex = AttrsInfo[i].AttrIndex;
     (*RecSchema)[AttrIndex].Assign(AttrsInfo[i].AttrName, 
				    AttrsInfo[i].TypeId, 0);
     (*RecSchema).KeyVector[AttrIndex] = AttrsInfo[i].IsKey;
    }
  } // for
 (*RecSchema).ComputeStats();
 return RecSchema;
}

int RelShoreDataSrc::
RelShoreDevice::getSchemaArity(const char* CatName)
{
 int SchemaArity = 0;
 for (int i = 0; i < numAttrsInfos; i++) {
   if (!strcmp(AttrsInfo[i].TableName, CatName)) {
     SchemaArity++;
    }
  } // for
 return SchemaArity;
}

//
// Methods of RelShoreDevice
// 
RelShoreDataSrc::
RelShoreDevice::RelShoreDevice(const char* Name)
{
 // here we have to do the following:
 // 1) mount the shore device (if it isnt already mounted)
 // 2) get a handle on the RELROOT index, so that relations can be accessed
 // 3) get a handle on the SUPER file, so that orphan objects can be created
 // 4) generate the catalog for the Shore device

 Valid = XXX_FALSE;
 XXX_NULL_SM_CHECK;
 if (XxxShoreSM->MountDevice(Name, VolumeId) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, Name);
   return;   
  }

 if (XxxBeginTrans() != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to open transaction");
   return;   
  }
 // get a handle on the RELROOT index
 bool Found;
 XxxShoreKey CatalogKey((void *)(RELROOTCAT), strlen(RELROOTCAT));
 if (XxxShoreSM->GetFromRootIndex(VolumeId, CatalogKey, (void *)(&RootIndex), 
				  sizeof(XxxShoreObjectId), Found) 
     != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to get catalog from root");
   XxxAbortTrans();
   return ;
  }

 if (!Found) {
   REL_NOTE_ERROR(REL_BADDEV, "Cannot find root index");
   XxxAbortTrans();
   return;
  }

 XxxShoreKey SuperFileKey((void *)(RELSUPERFILE), strlen(RELSUPERFILE));
 if (XxxShoreSM->GetFromRootIndex(VolumeId, SuperFileKey, 
				  (void *)(&SuperFile), 
				  sizeof(XxxShoreObjectId), Found) 
     != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to get superfile from root");
   XxxAbortTrans();
   return ;
  }

 if (!Found) {
   REL_NOTE_ERROR(REL_BADDEV, "Cannot find super file");
   XxxAbortTrans();
   return;
  }

 // add a catalog from this device
 Catalog = new RelShoreDataSrc::RelShoreCatalog(this, Name);
 if (XXX_IS_ERROR()) {
   delete Catalog;
   Catalog = 0;
   REL_NOTE_ERROR(REL_BADDEV, "Error creating catalog");
   XxxAbortTrans();
   return;
  }
 if (XxxCommitTrans() != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to commit transaction");
   return;   
  }
 DevName = new char[strlen(Name)+1];
 strcpy(DevName, Name);
 Valid = XXX_TRUE;
}

RelShoreDataSrc::
RelShoreDevice::~RelShoreDevice()
{
 // unmount the device and delete the catalog
 if (Valid && Catalog) delete Catalog;
 if (XxxShoreSM && Valid && XxxShoreSM->UnMountDevice(DevName) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, DevName);
   return;
  }
 if (Valid) delete [] DevName;
 Valid = XXX_FALSE;
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::DestroyDev()
{
 if (Valid && Catalog) delete Catalog;
 if (XxxShoreSM && Valid && XxxShoreSM->DestroyDevice(DevName) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, DevName);
   return XXX_NOT_OK;
  }
 Valid = XXX_FALSE;
 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::FormatDevice(const XxxShoreVolumeId& VolId)
{
 XXX_SM_CHECK;
 XxxShoreObjectId CatalogIndex;
 if (XxxIDXArray[XXX_BTREE]->Create(VolId, CatalogIndex,XXX_STRING, XXX_TRUE) 
     != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create catalog index");
   return XXX_NOT_OK;
  }
 // create the various catalog files and enter them into the catalog index
 if (createCatalogs(VolId, CatalogIndex) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create catalog entries");
   return XXX_NOT_OK;
  }

 // add the catalog file to the volume root index
 XxxShoreKey CatalogKey((void *)(RELROOTCAT), strlen(RELROOTCAT));
 if (XxxShoreSM->InsertInRootIndex(VolId, CatalogKey, (void *)(&CatalogIndex), 
				   sizeof(XxxShoreObjectId)) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to insert catalog into root");
   return XXX_NOT_OK;
  }

 XxxShoreObjectId SuperFile;
 if (XxxShoreSM->CreateFile(VolId, SuperFile) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create super file");
   return XXX_NOT_OK;
  }

 // add the super file to the volume root index
 XxxShoreKey SuperFileKey((void *)(RELSUPERFILE), strlen(RELSUPERFILE));
 if (XxxShoreSM->InsertInRootIndex(VolId, SuperFileKey, 
				   (void *)(&SuperFile), 
				   sizeof(XxxShoreObjectId)) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to insert superfile into root");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

//
// Here's how the catalog initialization works:
//   1) create the tables, attrs, stats and indxs rels; add them to the Root
//   2) make entries for each of them in each of the rels.
//   3) create indexes on each of them and make entries in the indxs rel.
//
XxxErrCode RelShoreDataSrc::
RelShoreDevice::createCatalogs(const XxxShoreVolumeId VolId,
			       const XxxShoreObjectId& CatalogIndex)
{
 XXX_SM_CHECK;

 if (createTablesCat(VolId, CatalogIndex) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create Tables cat");
   return XXX_NOT_OK;
  }
 if (createAttrsCat(VolId, CatalogIndex) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create Attrs cat");
   return XXX_NOT_OK;
  }
 if (createStatsCat(VolId, CatalogIndex) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create Stats cat");
   return XXX_NOT_OK;
  }
 if (createAttrStatsCat(VolId, CatalogIndex) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create AttrStats cat");
   return XXX_NOT_OK;
  }
 if (createIndxsCat(VolId, CatalogIndex) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create Indxs cat");
   return XXX_NOT_OK;
  }
 if (createViewsCat(VolId, CatalogIndex) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create Views cat");
   return XXX_NOT_OK;
  }

 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::makeTablesRec(const XxxRecordSchema* RecSchema,
				XxxValuesArray& ValsTab,
				const struct tablesInfo& TablesInfo,
				XxxRecord*& ReturnRec)
{
 const char* TableName = TablesInfo.TableName;
 const int TableArity = TablesInfo.TableArity;
 const int ImplType = TablesInfo.ImplType;

 strcpy(((XxxCoreADTClass::StrValue *)(ValsTab.Values[0]))->Value, TableName);
 ((XxxCoreADTClass::IntValue *)(ValsTab.Values[1]))->Value = TableArity;
 ((XxxCoreADTClass::IntValue *)(ValsTab.Values[2]))->Value = ImplType;
 if (!(ReturnRec = new XxxRecord(ValsTab, RecSchema))) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create Tables record");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::createTablesEntry(const XxxShoreVolumeId VolId,
				  XxxShoreObjectId& TablesFile,
				  XxxShoreObjectId& TablesIndex,
				  const XxxRecordSchema* RecSchema,
				  XxxValuesArray& ValsTab,
				  struct tablesInfo& TablesInfo)
{
 XxxRecord* Rec;
 XxxShoreObjectId RecordId;

 if (makeTablesRec(RecSchema, ValsTab, TablesInfo, Rec) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create Tables record");
   return XXX_NOT_OK;   
  }
 // insert the record into the table and the index
 if (XxxShoreSM->InsertRecord(VolId, TablesFile, Rec, RecordId) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to insert Tables record");
   return XXX_NOT_OK;   
  }
 int StrValSize = strlen(TablesInfo.TableName)+1+XxxCoreADTClass::ValueOffset(XXX_STRING);
 char *TmpADTVal = new char[StrValSize];
 XxxCoreADTClass::StrValue* StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
 strcpy(StrVal->Value, TablesInfo.TableName);

 if (XxxIDXArray[XXX_BTREE]->InsertEntry(VolId,TablesIndex,
					 StrVal, StrValSize,
					 XXX_STRING,
					 &RecordId) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to index Tables record");
   return XXX_NOT_OK;
  }
 delete [] TmpADTVal;

 Rec->Reclaim(RecSchema);
 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::createTablesCat(const XxxShoreVolumeId VolId,
				const XxxShoreObjectId& CatalogIndex)
{
 XXX_SM_CHECK;
 XxxShoreObjectId TablesFile, TablesIndex;

 // create a file and an index
 if (XxxShoreSM->CreateFile(VolId, TablesFile) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create Tables relation");
   return XXX_NOT_OK;
  } 
 if (XxxIDXArray[XXX_BTREE]->Create(VolId, TablesIndex,
				    XXX_STRING, XXX_TRUE) != XXX_OK){
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create Tables index");
   return XXX_NOT_OK;
  } 
 // insert both into the global root index
 int StrValSize = strlen(RELTABLESCAT)+1+XxxCoreADTClass::ValueOffset(XXX_STRING);
 char *TmpADTVal = new char[StrValSize];
 XxxCoreADTClass::StrValue* StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
 strcpy(StrVal->Value, RELTABLESCAT);
 if (XxxIDXArray[XXX_BTREE]->InsertEntry(VolId, CatalogIndex,
					 StrVal, StrValSize,
					 XXX_STRING,
					 &TablesFile) != XXX_OK) { 
   REL_NOTE_ERROR(REL_BADDEV, "Unable to catalog Tables relation");
   return XXX_NOT_OK;
  }
 delete [] TmpADTVal;

 StrValSize = strlen(RELTABINDXCAT)+1+XxxCoreADTClass::ValueOffset(XXX_STRING);
 TmpADTVal = new char[StrValSize];
 StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
 strcpy(StrVal->Value, RELTABINDXCAT);
 if (XxxIDXArray[XXX_BTREE]->InsertEntry(VolId,CatalogIndex,
					 StrVal, StrValSize,
					 XXX_STRING,
					 &TablesIndex) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to catalog Tables index");
   return XXX_NOT_OK;
  }
 delete [] TmpADTVal;

 // first get a catalog entry for the _TABLES relation
 XxxRecordSchema* RecSchema = getInitSchema(RELTABLESCAT);
 XxxValuesArray ValsTab(RecSchema);
 // insert records corresponding to the default tables
 // insert records corresponding to the default indexes
 for (int i = 0; i < numTablesInfos; i++) {
   if (createTablesEntry(VolId, TablesFile, TablesIndex,
			 RecSchema, ValsTab, TablesInfo[i]) 
       != XXX_OK) {
     REL_NOTE_ERROR (REL_BADDEV, "Unable to insert Tables record");
     return XXX_NOT_OK;
    }
  } // for 

 delete RecSchema;
 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::makeAttrsRec(const XxxRecordSchema* RecSchema,
			     XxxValuesArray& ValsTab,
			     const struct attrsInfo& AttrsInfo,
			     XxxRecord*& NewRec)
{
 const char* TableName = AttrsInfo.TableName;
 const char* AttrName = AttrsInfo.AttrName;
 const int IsKey = AttrsInfo.IsKey;
 const char* TypeName = XxxADTArray[AttrsInfo.TypeId]->TypeName;;
 const int AttrIndex = AttrsInfo.AttrIndex;
 const XxxADTClass::MetaInfoHandle* AttrMIHndl = AttrsInfo.AttrMIHndl;

 strcpy(((XxxCoreADTClass::StrValue*)(ValsTab.Values[0]))->Value, TableName);
 strcpy(((XxxCoreADTClass::StrValue*)(ValsTab.Values[1]))->Value, AttrName);
 ((XxxCoreADTClass::IntValue *)(ValsTab.Values[2]))->Value = IsKey; 
 strcpy(((XxxCoreADTClass::StrValue*)(ValsTab.Values[3]))->Value, TypeName);
 ((XxxCoreADTClass::IntValue *)(ValsTab.Values[4]))->Value = AttrIndex; 
 memcpy(ValsTab.Values[5], (char *)AttrMIHndl,
	AttrMIHndl->CopySize());

 if (!(NewRec = new XxxRecord(ValsTab, RecSchema))) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create Attr record");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::createAttrsEntry(const XxxShoreVolumeId VolId,
				 XxxShoreObjectId& AttrsFile,
				 XxxShoreObjectId& AttrsIndex,
				 const XxxRecordSchema* RecSchema,
				 XxxValuesArray& ValsTab,
				 struct attrsInfo& AttrsInfo)
{
 XxxRecord* Rec;
 XxxShoreObjectId RecordId;

 if (makeAttrsRec(RecSchema, ValsTab, AttrsInfo, Rec) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create Attrs record");
   return XXX_NOT_OK;   
  }

 // insert the record into the table and the index
 if (XxxShoreSM->InsertRecord(VolId, AttrsFile, Rec, RecordId) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to insert Attr record");
   return XXX_NOT_OK;   
  }
 int StrValSize = strlen(AttrsInfo.AttrName)+1+XxxCoreADTClass::ValueOffset(XXX_STRING);
 char *TmpADTVal = new char[StrValSize];
 XxxCoreADTClass::StrValue* StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
 strcpy(StrVal->Value, AttrsInfo.AttrName);
 if (XxxIDXArray[XXX_BTREE]->InsertEntry(VolId, AttrsIndex,
					 StrVal, StrValSize,
					 XXX_STRING,
					 &RecordId) != XXX_OK) { 
   REL_NOTE_ERROR(REL_BADDEV, "Unable to index Attr record");
   return XXX_NOT_OK;
  }
 delete [] TmpADTVal;
 Rec->Reclaim(RecSchema);
 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::createAttrsCat(const XxxShoreVolumeId VolId,
			       const XxxShoreObjectId& CatalogIndex)
{
 XXX_SM_CHECK;
 XxxShoreObjectId AttrsFile, AttrsIndex;

 if (XxxShoreSM->CreateFile(VolId, AttrsFile) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create Attrs relation");
   return XXX_NOT_OK;
  } 
 int StrValSize = strlen(RELATTRSCAT)+1+XxxCoreADTClass::ValueOffset(XXX_STRING);
 char *TmpADTVal = new char[StrValSize];
 XxxCoreADTClass::StrValue* StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
 strcpy(StrVal->Value, RELATTRSCAT);
 if (XxxIDXArray[XXX_BTREE]->InsertEntry(VolId, CatalogIndex, 
					 StrVal, StrValSize,
					 XXX_STRING,
					 &AttrsFile) != XXX_OK) { 
   REL_NOTE_ERROR(REL_BADDEV, "Unable to catalog Attrs relation");
   return XXX_NOT_OK;
  }
 delete [] TmpADTVal;
 // non-unique index for Attrs
 if (XxxIDXArray[XXX_BTREE]->Create(VolId, AttrsIndex,
				    XXX_STRING, XXX_FALSE) != XXX_OK){
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create Attrs index");
   return XXX_NOT_OK;
  }
 StrValSize = strlen(RELATTRSINDXCAT)+1+XxxCoreADTClass::ValueOffset(XXX_STRING);
 TmpADTVal = new char[StrValSize];
 StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
 strcpy(StrVal->Value, RELATTRSINDXCAT);
 if (XxxIDXArray[XXX_BTREE]->InsertEntry(VolId, CatalogIndex,
					 StrVal, StrValSize,
					 XXX_STRING,
					 &AttrsIndex) != XXX_OK) { 
   REL_NOTE_ERROR(REL_BADDEV, "Unable to catalog Attrs index");
   return XXX_NOT_OK;
  }
 delete [] TmpADTVal;

 // first get a catalog entry for the _ATTRS relation
 XxxRecordSchema* RecSchema = getInitSchema(RELATTRSCAT);
 XxxValuesArray ValsTab(RecSchema);
 // insert records corresponding to the default tables
 for (int i = 0; i < numAttrsInfos; i++) {
   if (createAttrsEntry(VolId, AttrsFile, AttrsIndex,
			RecSchema, ValsTab, AttrsInfo[i]) 
       != XXX_OK) {
     REL_NOTE_ERROR (REL_BADDEV, "Unable to insert Attrs record");
     return XXX_NOT_OK;
    }
  } // for 

 delete RecSchema;
 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::makeStatsRec(const XxxRecordSchema* RecSchema,
			     XxxValuesArray& ValsTab,
			     const struct statsInfo& StatsInfo,
			     XxxRecord*& ReturnRec)
{
 strcpy(((XxxCoreADTClass::StrValue*)(ValsTab.Values[0]))->Value, 
	StatsInfo.TableName);
 ((XxxCoreADTClass::IntValue *)(ValsTab.Values[1]))->Value = StatsInfo.Card;
 ((XxxCoreADTClass::IntValue *)(ValsTab.Values[2]))->Value = StatsInfo.TupleSize;
 if (!(ReturnRec = new XxxRecord(ValsTab, RecSchema))) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create Stats record");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::createStatsEntry(const XxxShoreVolumeId VolId,
				 XxxShoreObjectId& StatsFile,
				 XxxShoreObjectId& StatsIndex,
				 const XxxRecordSchema* RecSchema,
				 XxxValuesArray& ValsTab,
				 struct statsInfo& StatsInfo)
{
 XxxRecord* Rec;
 XxxShoreObjectId RecordId;

 if (makeStatsRec(RecSchema, ValsTab, StatsInfo, Rec) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create Stats record");
   return XXX_NOT_OK;   
  }
 // insert the record into the table and the index
 if (XxxShoreSM->InsertRecord(VolId, StatsFile, Rec, RecordId) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to insert Stats record");
   return XXX_NOT_OK;   
  }
 int StrValSize = strlen(StatsInfo.TableName)+1+XxxCoreADTClass::ValueOffset(XXX_STRING);
 char *TmpADTVal = new char[StrValSize];
 XxxCoreADTClass::StrValue* StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
 strcpy(StrVal->Value, StatsInfo.TableName);
 if (XxxIDXArray[XXX_BTREE]->InsertEntry(VolId, StatsIndex, 
					 StrVal, StrValSize,
					 XXX_STRING,
					 &RecordId) != XXX_OK) { 
   REL_NOTE_ERROR(REL_BADDEV, "Unable to index Stats record");
   return XXX_NOT_OK;
  }
 delete [] TmpADTVal;
 Rec->Reclaim(RecSchema);
 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::createStatsCat(const XxxShoreVolumeId VolId,
			       const XxxShoreObjectId& CatalogIndex)
{
 XXX_SM_CHECK;
 XxxShoreObjectId StatsFile, StatsIndex;

 if (XxxShoreSM->CreateFile(VolId, StatsFile) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create Stats relation");
   return XXX_NOT_OK;
  } 
 int StrValSize = strlen(RELSTATSCAT)+1+XxxCoreADTClass::ValueOffset(XXX_STRING);
 char *TmpADTVal = new char[StrValSize];
 XxxCoreADTClass::StrValue* StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
 strcpy(StrVal->Value, RELSTATSCAT);
 if (XxxIDXArray[XXX_BTREE]->InsertEntry(VolId, CatalogIndex, 
					 StrVal, StrValSize,
					 XXX_STRING,
					 &StatsFile) != XXX_OK) { 
   REL_NOTE_ERROR(REL_BADDEV, "Unable to catalog Stats relation");
   return XXX_NOT_OK;
  }
 delete [] TmpADTVal;
 if (XxxIDXArray[XXX_BTREE]->Create(VolId, StatsIndex,
				    XXX_STRING, XXX_TRUE) != XXX_OK){
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create Stats index");
   return XXX_NOT_OK;
  }
 StrValSize = strlen(RELSTATSINDXCAT)+1+XxxCoreADTClass::ValueOffset(XXX_STRING);
 TmpADTVal = new char[StrValSize];
 StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
 strcpy(StrVal->Value, RELSTATSINDXCAT);
 if (XxxIDXArray[XXX_BTREE]->InsertEntry(VolId, CatalogIndex,
					 StrVal, StrValSize,
					 XXX_STRING,
					 &StatsIndex) != XXX_OK) { 
   REL_NOTE_ERROR(REL_BADDEV, "Unable to catalog Stats index");
   return XXX_NOT_OK;
  }
 delete [] TmpADTVal;

 // first get a catalog entry for the _STATS relation
 XxxRecordSchema* RecSchema = getInitSchema(RELSTATSCAT);
 XxxValuesArray ValsTab(RecSchema);
 XxxShoreObjectId RecordId;

 // insert records corresponding to the default tables
 for (int i = 0; i < numStatsInfos; i++) {
   if (createStatsEntry(VolId, StatsFile, StatsIndex,
			RecSchema, ValsTab, StatsInfo[i]) 
       != XXX_OK) {
     REL_NOTE_ERROR (REL_BADDEV, "Unable to insert Stats record");
     return XXX_NOT_OK;
    }
  } // for 

 delete RecSchema;
 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::makeViewsRec(const XxxRecordSchema* RecSchema,
			     XxxValuesArray& ValsTab,
			     const struct viewsInfo& ViewsInfo,
			     XxxRecord*& ReturnRec)
{
 strcpy(((XxxCoreADTClass::StrValue*)(ValsTab.Values[0]))->Value, 
	ViewsInfo.ViewName);
 strcpy(((XxxCoreADTClass::StrValue*)(ValsTab.Values[1]))->Value, 
	ViewsInfo.ViewInfo);
 if (!(ReturnRec = new XxxRecord(ValsTab, RecSchema))) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create Views record");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::createViewsEntry(const XxxShoreVolumeId VolId,
				 XxxShoreObjectId& ViewsFile,
				 XxxShoreObjectId& ViewsIndex,
				 const XxxRecordSchema* RecSchema,
				 XxxValuesArray& ValsTab,
				 struct viewsInfo& ViewsInfo)
{
 XxxRecord* Rec;
 XxxShoreObjectId RecordId;

 if (makeViewsRec(RecSchema, ValsTab, ViewsInfo, Rec) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create Views record");
   return XXX_NOT_OK;   
  }
 // insert the record into the table and the index
 if (XxxShoreSM->InsertRecord(VolId, ViewsFile, Rec, RecordId) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to insert Views record");
   return XXX_NOT_OK;   
  }
 int StrValSize = strlen(ViewsInfo.ViewName)+1+XxxCoreADTClass::ValueOffset(XXX_STRING);
 char *TmpADTVal = new char[StrValSize];
 XxxCoreADTClass::StrValue* StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
 strcpy(StrVal->Value, ViewsInfo.ViewName);
 if (XxxIDXArray[XXX_BTREE]->InsertEntry(VolId, ViewsIndex, 
					 StrVal, StrValSize,
					 XXX_STRING,
					 &RecordId) != XXX_OK) { 
   REL_NOTE_ERROR(REL_BADDEV, "Unable to index Views record");
   return XXX_NOT_OK;
  }
 delete [] TmpADTVal;
 Rec->Reclaim(RecSchema);
 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::deleteViewsEntry(const XxxShoreVolumeId VolId,
				 XxxShoreObjectId& ViewsFile,
				 XxxShoreObjectId& ViewsIndex,
				 const XxxRecordSchema* RecSchema,
				 XxxValuesArray& ValsTab,
				 struct viewsInfo& ViewsInfo)
{
 XxxShoreObjectId RecordId;
 bool Found;

 // Get recordid for the entry of this view
 int StrValSize = strlen(ViewsInfo.ViewName)+1+XxxCoreADTClass::ValueOffset(XXX_STRING);
 char *TmpADTVal = new char[StrValSize];
 XxxCoreADTClass::StrValue* StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
 strcpy(StrVal->Value, ViewsInfo.ViewName);
 if (XxxIDXArray[XXX_BTREE]->GetEntry(VolId, ViewsIndex,
					 StrVal, StrValSize, XXX_STRING,
					 &RecordId, sizeof(XxxShoreObjectId), Found) != XXX_OK) { 
   REL_NOTE_ERROR(REL_BADDEV, "Unable to get catalog info of Views index");
   return XXX_NOT_OK;
 }
 delete [] TmpADTVal;
 if(!Found){
   REL_NOTE_ERROR(REL_BADDEV, "Unable to find catalog info of Views index");
   return XXX_NOT_OK;
 }

 // Delete the record from the table and the index
 if (XxxShoreSM->DeleteRecord(VolId, RecordId) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to delete Views record");
   return XXX_NOT_OK;   
  }

 StrValSize = strlen(ViewsInfo.ViewName)+1+XxxCoreADTClass::ValueOffset(XXX_STRING);
 TmpADTVal = new char[StrValSize];
 StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
 strcpy(StrVal->Value, ViewsInfo.ViewName);
 if (XxxIDXArray[XXX_BTREE]->DeleteOneEntry(VolId, ViewsIndex, 
					 StrVal, StrValSize,
					 XXX_STRING,
					 &RecordId) != XXX_OK) { 
   REL_NOTE_ERROR(REL_BADDEV, "Unable to index Views record");
   return XXX_NOT_OK;
  }
 delete [] TmpADTVal;
 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::createViewsCat(const XxxShoreVolumeId VolId,
			       const XxxShoreObjectId& CatalogIndex)
{
 XXX_SM_CHECK;
 XxxShoreObjectId ViewsFile, ViewsIndex;

 if (XxxShoreSM->CreateFile(VolId, ViewsFile) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create Views relation");
   return XXX_NOT_OK;
  } 
 int StrValSize = strlen(RELVIEWSCAT)+1+XxxCoreADTClass::ValueOffset(XXX_STRING);
 char *TmpADTVal = new char[StrValSize];
 XxxCoreADTClass::StrValue* StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
 strcpy(StrVal->Value, RELVIEWSCAT);
 if (XxxIDXArray[XXX_BTREE]->InsertEntry(VolId, CatalogIndex, 
					 StrVal, StrValSize,
					 XXX_STRING,
					 &ViewsFile) != XXX_OK) { 
   REL_NOTE_ERROR(REL_BADDEV, "Unable to catalog Views relation");
   return XXX_NOT_OK;
  }
 delete [] TmpADTVal;
 if (XxxIDXArray[XXX_BTREE]->Create(VolId, ViewsIndex,
				    XXX_STRING, XXX_TRUE) != XXX_OK){
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create Views index");
   return XXX_NOT_OK;
  }
 StrValSize = strlen(RELVIEWSINDXCAT)+1+XxxCoreADTClass::ValueOffset(XXX_STRING);
 TmpADTVal = new char[StrValSize];
 StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
 strcpy(StrVal->Value, RELVIEWSINDXCAT);
 if (XxxIDXArray[XXX_BTREE]->InsertEntry(VolId, CatalogIndex,
					 StrVal, StrValSize,
					 XXX_STRING,
					 &ViewsIndex) != XXX_OK) { 
   REL_NOTE_ERROR(REL_BADDEV, "Unable to catalog Views index");
   return XXX_NOT_OK;
  }
 delete [] TmpADTVal;

 // first get a catalog entry for the _VIEWS relation
 XxxRecordSchema* RecSchema = getInitSchema(RELVIEWSCAT);
 XxxValuesArray ValsTab(RecSchema);
 XxxShoreObjectId RecordId;

 // insert records corresponding to the default tables
 for (int i = 0; i < numViewsInfos; i++) {
   if (createViewsEntry(VolId, ViewsFile, ViewsIndex,
			RecSchema, ValsTab, ViewsInfo[i]) 
       != XXX_OK) {
     REL_NOTE_ERROR (REL_BADDEV, "Unable to insert Views record");
     return XXX_NOT_OK;
    }
  } // for 

 delete RecSchema;
 return XXX_OK;
}


XxxErrCode RelShoreDataSrc::
RelShoreDevice::makeAStatsRec(const XxxRecordSchema* RecSchema,
			      XxxValuesArray& ValsTab,
			      const astatsInfo& AStatsInfo,
			      XxxRecord*& ReturnRec)
{
 XxxADTClass::MetaInfoHandle StatsMIHndl;
 XxxADTClass* ADTClass = XxxADTArray[AStatsInfo.TypeId];
 if (ADTClass->StoreStatsInfo(AStatsInfo.MetaInfo,
			      AStatsInfo.Data,
			      StatsMIHndl) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Error creating attrstats blob");
   return XXX_NOT_OK;
  }

 // TODO:L1 change all strcpy()s to strncpy()s
 strcpy(((XxxCoreADTClass::StrValue*)(ValsTab.Values[0]))->Value, AStatsInfo.TableName);
 ((XxxCoreADTClass::IntValue *)(ValsTab.Values[1]))->Value = AStatsInfo.AttrIndex;
 memcpy((char *)(ValsTab.Values[2]), (char *)&StatsMIHndl,
	StatsMIHndl.CopySize());

 if (!(ReturnRec = new XxxRecord(ValsTab, RecSchema))) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create AttrStats record");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}
XxxErrCode RelShoreDataSrc::
RelShoreDevice::createAttrStatsCat(const XxxShoreVolumeId VolId,
				   const XxxShoreObjectId& CatalogIndex)
{
 XXX_SM_CHECK;
 XxxShoreObjectId AStatsFile, AStatsIndex;

 if (XxxShoreSM->CreateFile(VolId, AStatsFile) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create AttrStats file");
   return XXX_NOT_OK;
  } 
 int StrValSize = strlen(RELATTRSTATSCAT)+1+XxxCoreADTClass::ValueOffset(XXX_STRING);
 char *TmpADTVal = new char[StrValSize];
 XxxCoreADTClass::StrValue* StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
 strcpy(StrVal->Value, RELATTRSTATSCAT);
 if (XxxIDXArray[XXX_BTREE]->InsertEntry(VolId, CatalogIndex, 
					 StrVal, StrValSize,
					 XXX_STRING,
					 &AStatsFile) != XXX_OK) { 
   REL_NOTE_ERROR(REL_BADDEV, "Unable to catalog AStats relation");
   return XXX_NOT_OK;
  }
 delete [] TmpADTVal;

 if (XxxIDXArray[XXX_BTREE]->Create(VolId, AStatsIndex,
				    XXX_STRING, XXX_FALSE) != XXX_OK){
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create AStats index");
   return XXX_NOT_OK;
  }
 StrValSize = strlen(RELATTRSTATSINDXCAT)+1+XxxCoreADTClass::ValueOffset(XXX_STRING);
 TmpADTVal = new char[StrValSize];
 StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
 strcpy(StrVal->Value, RELATTRSTATSINDXCAT);
 if (XxxIDXArray[XXX_BTREE]->InsertEntry(VolId, CatalogIndex, 
					 StrVal, StrValSize,
					 XXX_STRING,
					 &AStatsIndex) != XXX_OK) { 
   REL_NOTE_ERROR(REL_BADDEV, "Unable to catalog AStats index");
   return XXX_NOT_OK;
  }
 delete [] TmpADTVal;
 // no initial records
 return XXX_OK;
}
XxxErrCode RelShoreDataSrc::
RelShoreDevice::createIndxsEntry(const XxxShoreVolumeId VolId,
				 XxxShoreObjectId& IndxsFile,
				 XxxShoreObjectId& IndxsTableIndex,
				 XxxShoreObjectId& IndxsNameIndex,
				 const XxxRecordSchema* RecSchema,
				 XxxValuesArray& ValsTab,
				 struct indxsInfo& IndxInfo)
{
 const char* TableName = IndxInfo.TableName;
 const char* IndxName = IndxInfo.IndxName;
 const char* IndxString = IndxInfo.IndxString;
 const char* IndxType = XxxIDXArray[IndxInfo.IndxType]->TypeName;

 XxxRecord* Rec;
 XxxShoreObjectId RecordId;
 strcpy(((XxxCoreADTClass::StrValue*)(ValsTab.Values[0]))->Value, TableName);
 strcpy(((XxxCoreADTClass::StrValue*)(ValsTab.Values[1]))->Value, IndxName);
 strcpy(((XxxCoreADTClass::StrValue*)(ValsTab.Values[2]))->Value, IndxString);
 strcpy(((XxxCoreADTClass::StrValue*)(ValsTab.Values[3]))->Value, IndxType);
 ((XxxCoreADTClass::IntValue *)(ValsTab.Values[4]))->Value = IndxInfo.IsUnique;
 ((XxxCoreADTClass::IntValue *)(ValsTab.Values[5]))->Value = IndxInfo.IsClustered;

 if (!(Rec = new XxxRecord(ValsTab, RecSchema))) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to create Indxs record");
   return XXX_NOT_OK;
  }
 // insert the record into the table and the indexes
 if (XxxShoreSM->InsertRecord(VolId, IndxsFile, Rec, RecordId) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to insert Indxs record");
   return XXX_NOT_OK;   
  }
 if (XxxIDXArray[XXX_BTREE]->InsertEntry(VolId, IndxsTableIndex, 
					 ValsTab.Values[0],
					 (int)strlen(TableName) + 1 + XxxCoreADTClass::ValueOffset(XXX_STRING),
					 XXX_STRING,
					 &RecordId) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Unable to index Indxs record");
   return XXX_NOT_OK;
  }
 if (XxxIDXArray[XXX_BTREE]->InsertEntry(VolId, IndxsNameIndex, 
					 ValsTab.Values[1],
					 (int)strlen(IndxName) + 1 + XxxCoreADTClass::ValueOffset(XXX_STRING),
					 XXX_STRING,
					 &RecordId) != XXX_OK) { 
   REL_NOTE_ERROR(REL_BADDEV, "Unable to index Indxs record");
   return XXX_NOT_OK;
  }
 Rec->Reclaim(RecSchema);
 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::createIndxsCat(const XxxShoreVolumeId VolId, 
			       const XxxShoreObjectId& CatalogIndex)
{
 XXX_SM_CHECK;
 XxxShoreObjectId IndxsFile, IndxsIndex1, IndxsIndex2;

 if (XxxShoreSM->CreateFile (VolId, IndxsFile) != XXX_OK)
 {
  REL_NOTE_ERROR (REL_BADDEV, "Unable to create Indxs relation");
  return XXX_NOT_OK;
 }

 int StrValSize = strlen(RELINDXSCAT)+1+XxxCoreADTClass::ValueOffset(XXX_STRING);
 char *TmpADTVal = new char[StrValSize];
 XxxCoreADTClass::StrValue* StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
 strcpy(StrVal->Value, RELINDXSCAT);
 if (XxxIDXArray[XXX_BTREE]->InsertEntry(VolId, CatalogIndex,
					 StrVal, StrValSize,
					 XXX_STRING,
					 &IndxsFile) != XXX_OK) {
   REL_NOTE_ERROR (REL_BADDEV, "Unable to catalog Indxs relation");
   return XXX_NOT_OK;
 }
 delete [] TmpADTVal;

 // The index on table name is _not_ unique
 if (XxxIDXArray[XXX_BTREE]->Create(VolId, IndxsIndex1,
				    XXX_STRING, XXX_FALSE)
     != XXX_OK) {
   REL_NOTE_ERROR (REL_BADDEV, "Unable to create Indxs index 1");
  return XXX_NOT_OK;
 }

 StrValSize = strlen(RELINDXSINDXCAT1)+1+XxxCoreADTClass::ValueOffset(XXX_STRING);
 TmpADTVal = new char[StrValSize];
 StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
 strcpy(StrVal->Value, RELINDXSINDXCAT1);
 if (XxxIDXArray[XXX_BTREE]->InsertEntry(VolId, CatalogIndex,
					 StrVal, StrValSize,
					 XXX_STRING,
					 &IndxsIndex1) != XXX_OK) {
   REL_NOTE_ERROR (REL_BADDEV, "Unable to catalog Indxs index 1");
   return XXX_NOT_OK;
 }
 delete [] TmpADTVal;

 if (XxxIDXArray[XXX_BTREE]->Create(VolId, IndxsIndex2,
				    XXX_STRING, XXX_TRUE) != XXX_OK) {
   REL_NOTE_ERROR (REL_BADDEV, "Unable to create Indxs index 2");
   return XXX_NOT_OK;
 }

 StrValSize = strlen(RELINDXSINDXCAT2)+1+XxxCoreADTClass::ValueOffset(XXX_STRING);
 TmpADTVal = new char[StrValSize];
 StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
 strcpy(StrVal->Value, RELINDXSINDXCAT2);
 if (XxxIDXArray[XXX_BTREE]->InsertEntry(VolId, CatalogIndex,
					 StrVal, StrValSize,
					 XXX_STRING,
					 &IndxsIndex2) != XXX_OK) {
   REL_NOTE_ERROR (REL_BADDEV, "Unable to catalog Indxs index 2");
   return XXX_NOT_OK;
 }
 delete [] TmpADTVal;

 // first get a catalog entry for the _INDXS relation
 XxxRecordSchema* RecSchema = getInitSchema(RELINDXSCAT);
 XxxValuesArray ValsTab (RecSchema);

 // insert records corresponding to the default indexes
 for (int i = 0; i < numIndxsInfos; i++) {
   if (createIndxsEntry(VolId, IndxsFile, IndxsIndex1, IndxsIndex2,
			RecSchema, ValsTab, IndxsInfo[i]) 
       != XXX_OK) {
     REL_NOTE_ERROR (REL_BADDEV, "Unable to insert Indxs record");
     return XXX_NOT_OK;
    }
  } // for 

 delete RecSchema;
 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::CreateNewRelation(RelShoreHandle*& Handle,
				  RelShoreImplType ImplType,
				  const char* Name, 
				  RelDataSrc::AttrTypeList* Attrs,
				  RelDataSrc::KeyList* Keys,
				  XxxBool MakeCatEntry)
{
 XXX_SM_CHECK;

 // here determine the possible choices for ImplType
 XxxShoreObjectId NewRel;
 // ideally, we want a table of implementation types, not a switch stmt
 switch (ImplType) {
  case REL_SHORE_FILE_IMPL: {
    if (XxxShoreSM->CreateFile(VolumeId, NewRel) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADDEV, "Unable to create relation");
      return XXX_NOT_OK;
     }
    break;
   }
  default: 
   REL_NOTE_ERROR(REL_BADDEV, "Bad relation impl type");
   return XXX_NOT_OK;
  }

 Handle = new RelShoreHandle(VolumeId, NewRel);

 if (Name) {
   // named relation must be entered into the RELROOT index
   int StrValSize = strlen(Name)+1+XxxCoreADTClass::ValueOffset(XXX_STRING);
   char *TmpADTVal = new char[StrValSize];
   XxxCoreADTClass::StrValue* StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
   strcpy(StrVal->Value, Name);
   if (XxxIDXArray[XXX_BTREE]->InsertEntry(VolumeId, RootIndex, 
					   StrVal, StrValSize,
					   XXX_STRING,
					   &NewRel) != XXX_OK) { 
     REL_NOTE_ERROR(REL_BADDEV, "Unable to catalog new relation");
     return XXX_NOT_OK;
    }
   delete [] TmpADTVal;

   // create the catalog entries for the new relation
   if (MakeCatEntry &&
       Catalog->AddCatInfo(Name, Attrs, Keys, ImplType) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADDEV, "Unable to make catalog entries for new rel");
     return XXX_NOT_OK;
    }
  }   
 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::CreateNewView(const char* Name, const char* ViewInfo)
{
 XXX_SM_CHECK;
 XxxShoreObjectId ViewsFile, ViewsIndex;
 bool Found;

 if (Name) {
   // named view must be entered into the RELVIEW file and index
   // First, get the ViewsFile and ViewsIndex from CatalogIndex
   int StrValSize = strlen(RELVIEWSCAT)+1+XxxCoreADTClass::ValueOffset(XXX_STRING);
   char *TmpADTVal = new char[StrValSize];
   XxxCoreADTClass::StrValue* StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
   strcpy(StrVal->Value, RELVIEWSCAT);
   if (XxxIDXArray[XXX_BTREE]->GetEntry(VolumeId, RootIndex, 
					   StrVal, StrValSize, XXX_STRING,
					   &ViewsFile, sizeof(XxxShoreObjectId), Found) != XXX_OK) { 
     REL_NOTE_ERROR(REL_BADDEV, "Unable to get catalog info of Views relation");
     return XXX_NOT_OK;
   }
   delete [] TmpADTVal;
   if(!Found){
     REL_NOTE_ERROR(REL_BADDEV, "Unable to find catalog info of Views relation");
     return XXX_NOT_OK;
   }
   StrValSize = strlen(RELVIEWSINDXCAT)+1+XxxCoreADTClass::ValueOffset(XXX_STRING);
   TmpADTVal = new char[StrValSize];
   StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
   strcpy(StrVal->Value, RELVIEWSINDXCAT);
   if (XxxIDXArray[XXX_BTREE]->GetEntry(VolumeId, RootIndex,
					   StrVal, StrValSize, XXX_STRING,
					   &ViewsIndex, sizeof(XxxShoreObjectId), Found) != XXX_OK) { 
     REL_NOTE_ERROR(REL_BADDEV, "Unable to get catalog info of Views index");
     return XXX_NOT_OK;
   }
   delete [] TmpADTVal;
   if(!Found){
     REL_NOTE_ERROR(REL_BADDEV, "Unable to find catalog info of Views index");
     return XXX_NOT_OK;
   }

   // first get a catalog entry for the _VIEWS relation
   XxxRecordSchema* RecSchema = getInitSchema(RELVIEWSCAT);
   XxxValuesArray ValsTab(RecSchema);
   XxxShoreObjectId RecordId;

   viewsInfo *view = newViewsInfo(Name, ViewInfo);
   // create the view entries for the new view
   if (createViewsEntry(VolumeId, ViewsFile, ViewsIndex,
			  RecSchema, ValsTab, *view) 
         != XXX_OK) {
     REL_NOTE_ERROR (REL_BADDEV, "Unable to insert Views record");
     return XXX_NOT_OK;
   }
 }
 return XXX_OK;
}

// PURPOSE: Adds new index to shore device.
// NOTES: Using new index framework with XxxIDXArray to avoid switches.

XxxErrCode 
RelShoreDataSrc::
RelShoreDevice::CreateNewIndex( const char* RelName,
				const char* IndxName,
				XxxValueExpression *Expr,
				const XxxBool Unique,
				const XxxBool Clustered,
				XxxBool& Success,
				XxxIndexType IndxType)
{
 XxxRelation *Rel;
 RelHandle *Hndl;
 XxxBool CatFound = XXX_FALSE;

 XXX_SM_CHECK;

 if (!RelName || !IndxName) {
  REL_NOTE_ERROR (REL_BADDEV, "New index needs index and relation names");
  return XXX_NOT_OK;
 }
 
 // check if this relation is in this device
 RelCatalogInfo CatInfo(RelName);
 
 if ((GetCatInfo(RelName, CatInfo, CatFound) != XXX_OK) ||
     (CatFound != XXX_TRUE)) {
   REL_NOTE_ERROR (REL_BADDEV, "Error getting catalog info");
   return XXX_NOT_OK;
 }
 
 if (GetRelation(RelName, Hndl) != XXX_TRUE) {
   Success = XXX_FALSE;
   return XXX_OK;
 }
 
 if (GetRelation((RelShoreHandle *) Hndl, Rel, CatInfo.ImplInfo ()) 
     != XXX_OK) {
   REL_NOTE_ERROR (REL_BADDEV, "Error getting relation from handle");
   return XXX_NOT_OK;
 }
 Rel->CatalogInfo = &CatInfo;
 
 // Type check the expression.
 if (Expr->ResolveVars(CatInfo.RecordSchema()) != XXX_OK) {
   REL_NOTE_ERROR (REL_BADDEV, "Cannot resolve variables.");
   return XXX_NOT_OK;
 }
 
 if (Expr->TypeCheck() != XXX_OK) {
   REL_NOTE_ERROR (REL_BADDEV, "Type checks failed on expression.");
   return XXX_NOT_OK;
 }
 
 // Add the index to the relation. This will create the index, and
 // insert all the entries from the relation.

 if (Rel->AddIndex(IndxName, Expr, IndxType,
		   Unique, Expr->GenerateCatInfo()) != XXX_OK) {
   REL_NOTE_ERROR (REL_BADINDEX, "Cannot add index");
   delete Rel;
   
   return XXX_NOT_OK;
 }
 
 // add this to the _INDXS catalog
 XxxRecordSchema *RecSchema =
   Catalog->IndxsRelation->CatalogInfo->RecordSchema ();
 XxxValuesArray ValsTab(RecSchema);
 
 XxxRecord *NewRec;

 strcpy(((XxxCoreADTClass::StrValue*)(ValsTab.Values[0]))->Value, RelName);
 strcpy(((XxxCoreADTClass::StrValue*)(ValsTab.Values[1]))->Value, IndxName);
 
 // TODO:L2 bit of a hack. the expression could overflow the string!
 ostrstream str(((XxxCoreADTClass::StrValue*)(ValsTab.Values[2]))->Value, XXX_TEXTSTRING_MAX);
 Expr->Print(str);
 str << '\0';   // Must terminate the string !
 
 // Add type of index.
 strcpy(((XxxCoreADTClass::StrValue*)(ValsTab.Values[3]))->Value, XxxIDXArray[IndxType]->TypeName);
 
 // specify uniqueness
 ((XxxCoreADTClass::IntValue*)(ValsTab.Values[4]))->Value = (int)Unique;

 // specify clustering
 ((XxxCoreADTClass::IntValue *)(ValsTab.Values[5]))->Value = (int)Clustered;
 
 if (!(NewRec = new XxxRecord(ValsTab, RecSchema))) {
   REL_NOTE_ERROR (REL_BADDEV, "Unable to create new _INDXS record");
   return XXX_NOT_OK;
 }
 
 if (Catalog->IndxsRelation->AddItem(NewRec) != XXX_OK) {
     REL_NOTE_ERROR (REL_BADTABLE, "Unable to insert new _INDXS record");
     Success = XXX_FALSE;
     return XXX_NOT_OK;
 }
 
 NewRec->Reclaim(RecSchema);
 delete Rel;

 Success = XXX_TRUE;
 return XXX_OK;
}

XxxBool RelShoreDataSrc::
RelShoreDevice::GetRelationOfIndex(const char *IndxName, char *&RelName) const {
  return Catalog->GetRelationOfIndex(IndxName, RelName);
}

// PURPOSE: Delete index from system
//NOTES: Added if-statement for R* trees. Need to get rid of these ifs.

XxxErrCode RelShoreDataSrc::
RelShoreDevice::DeleteIndex(const char *IndxName, const char *RelName,
			    const RelShoreHandle *Handle) 
{
  // remove the index from the indexlist maintained by the relation's
  // catalog info
  int NumDeleted = 0;
  RelCatalogInfo CatInfo;
  XxxBool XFound;
  if (GetCatInfo(RelName, CatInfo, XFound) != XXX_OK) {
    char buf[100];
    sprintf (buf, "Cannot retrieve catalog information on relation %s",
	     RelName);
    REL_NOTE_ERROR(REL_BADDEV, buf);
    return XXX_NOT_OK;
  }
  if (!XFound) {
    char buf[100];
    sprintf (buf, "Cannot find catalog information on relation %s",
	     RelName);
    REL_NOTE_ERROR(REL_BADDEV, buf);
    return XXX_NOT_OK;
  }

  XxxIndexType IndexType; // stores the type of the index being deleted

  if (CatInfo.ImplInfo()->NumIndexes > 0) {
    if (CatInfo.ImplInfo()->Indexes) {
      RelShoreImplInfo::IndexList *Current = CatInfo.ImplInfo()->Indexes->Next;
      RelShoreImplInfo::IndexList *Last = 0;
      while (Current != 0) {
	if (!(strcmp((Current->IndxInfo)->IndexName, IndxName))) {
	  // Get index type so we know which delete function to call.
	  IndexType = Current->IndxInfo->IndexType;
	  RelShoreImplInfo::IndexList *Tmp = Current->Next;
	  delete Current;
	  NumDeleted++;
	  (CatInfo.ImplInfo()->NumIndexes)--;
	  if (Last == 0) // special case for first in list
	    CatInfo.ImplInfo()->Indexes = Tmp;
	  else
	    Last->Next = Tmp;
	  break; // get out of the loop
	} // if this is the index we want to delete
	Last = Current;
	Current = Current->Next;
      } // while there are indexes
    } // if there are any indexes
    if (NumDeleted != 1) {
      char Buf[100];
      sprintf (Buf, 
	       "Removed %d matching index list records from catalog info!", 
	       NumDeleted);
      REL_NOTE_ERROR (REL_BADCAT, Buf);
      // do not exit, lest index will exist without proper records
    }
  } else {
    REL_NOTE_ERROR(REL_BADCAT, 
		   "Catalog is unaware of any indexes on the relation");
    return XXX_NOT_OK;
  }

  // remove the record of the index from the catalog table _INDXS
  if (Catalog->DeleteIndex(IndxName) != XXX_OK) {
    REL_NOTE_ERROR (REL_BADCAT, "Cannot delete index from catalog");
    return XXX_NOT_OK;
  }
  
  // Delete the actual index
  // look up (and then delete) the index in the Root Index
  XxxShoreObjectId IndexId;
  bool Found;

  int StrValSize = strlen(IndxName)+1+XxxCoreADTClass::ValueOffset(XXX_STRING);
  char *TmpADTVal = new char[StrValSize];
  XxxCoreADTClass::StrValue* StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
  strcpy(StrVal->Value, IndxName);
  if (XxxIDXArray[XXX_BTREE]->GetEntry(VolumeId,RootIndex,
				       StrVal, StrValSize,
				       XXX_STRING,
				       &IndexId, 
				       sizeof(XxxShoreObjectId), 
				       Found) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADDEV, "Error getting from Root Index");
    return XXX_NOT_OK;
  }

  if (!Found) {
    REL_NOTE_ERROR(REL_BADDEV,
		   "No record of the index exists in the device's root index");
    return XXX_NOT_OK;
  }

  NumDeleted = 0;
  if (XxxIDXArray[XXX_BTREE]->DeleteAllEntries(VolumeId, RootIndex,
					       StrVal, StrValSize,
					       XXX_STRING,
					       NumDeleted) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADDEV, 
		   "Cannot delete record of index from device's root index");
    return XXX_NOT_OK;
  }
  delete [] TmpADTVal;

  if (NumDeleted != 1) {
    char buf[100];
    sprintf (buf, "Deleted %d entries of index %s from device's index!", 
	     NumDeleted, IndxName);
    REL_NOTE_ERROR(REL_BADDEV, buf);
    // note: do not return from here, lest the index exist with no record
    // of it anywhere
  }

  // should probably let this delete() remove the index from the root index.

  if (XxxIDXArray[IndexType]->Delete(VolumeId,IndexId) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADDEV, "Cannot delete index");
    return XXX_NOT_OK;
  }

  return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::DeleteRelation(const char *RelName,
			       const RelShoreHandle* Handle)
{
  XXX_SM_CHECK;
  // Get catalog information
  RelCatalogInfo CatInfo(RelName);
  XxxBool CatFound = XXX_FALSE;
  if ((GetCatInfo(RelName, CatInfo, CatFound) != XXX_OK) ||
      (CatFound != XXX_TRUE)) {
    REL_NOTE_ERROR(REL_BADDEV, "Error getting catalog info");
    return XXX_NOT_OK;
  }  
  // Pass information along to next level
  return (DeleteRelation(RelName, Handle, &CatInfo));
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::DeleteRelation(const char *RelName,
			       const RelShoreHandle* Handle,
			       const RelCatalogInfo* CatInfo)
{
  XXX_SM_CHECK;

  // delete all indexes on the relation
  int NumRemoved = 0;
  if (CatInfo->ImplInfo()->NumIndexes > 0) {
    RelShoreImplInfo::IndexList *IList = CatInfo->ImplInfo()->Indexes;
    while (IList->Next != 0) {
      if (DeleteIndex((IList->Next->IndxInfo)->IndexName, RelName, 
		      Handle) != XXX_OK) {
	REL_NOTE_ERROR (REL_BADDEV, 
			"Cannot delete an existing index on the relation");
	return XXX_NOT_OK;
      }
      IList = IList->Next;
    }
  }
  
  // Delete the record of the relation from the device's index
  int StrValSize = strlen(CatInfo->RelationName())+1+XxxCoreADTClass::ValueOffset(XXX_STRING);
  char *TmpADTVal = new char[StrValSize];
  XxxCoreADTClass::StrValue* StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
  strcpy(StrVal->Value, CatInfo->RelationName());
  if (XxxIDXArray[XXX_BTREE]->
      DeleteAllEntries(VolumeId, RootIndex, 
		       StrVal, StrValSize,
		       XXX_STRING,
		       NumRemoved) != XXX_OK){
    REL_NOTE_ERROR(REL_BADDEV, 
		   "Unable to delete relation from device's index");
    return XXX_NOT_OK;
  }
  delete [] TmpADTVal;
  
  if (NumRemoved != 1) {
    char buf[100];
    sprintf (buf, 
	     "There were %d records of the relation '%s' before deletion!", 
	     NumRemoved, CatInfo->RelationName());
    REL_NOTE_ERROR(REL_BADDEV, buf);
    return XXX_NOT_OK;
  }
  
  // delete the relation's SHORE representation
  switch (((RelShoreImplInfo *)(CatInfo->ImplInfo()))->ImplType) {
  case REL_SHORE_FILE_IMPL: 
    if (XxxShoreSM->DeleteFile(Handle->ShoreId) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADDEV, "Unable to delete relation file");
      return XXX_NOT_OK;
    }
      break;
  default: 
    REL_NOTE_ERROR (REL_BADSHOREREL,
		    "Attempt to delete a relation with bad impl type");
    return XXX_NOT_OK;
  }
  // Delete the information pertaining to this relation in the catalog
  if (Catalog->DeleteCatInfo(CatInfo->RelationName()) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADCAT, "Unable to remove catalog information");
    return XXX_NOT_OK;
  }
  
  return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::DeleteView(const char *Name)
{
 XxxShoreObjectId ViewsFile, ViewsIndex;
 bool Found;
 // First, get the ViewsFile and ViewsIndex from CatalogIndex

 int StrValSize = strlen(RELVIEWSCAT)+1+XxxCoreADTClass::ValueOffset(XXX_STRING);
 char *TmpADTVal = new char[StrValSize];
 XxxCoreADTClass::StrValue* StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
 strcpy(StrVal->Value, RELVIEWSCAT);
 if (XxxIDXArray[XXX_BTREE]->GetEntry(VolumeId, RootIndex, 
			   StrVal, StrValSize, XXX_STRING,
			   &ViewsFile, sizeof(XxxShoreObjectId), Found) != XXX_OK) { 
   REL_NOTE_ERROR(REL_BADDEV, "Unable to get catalog info of Views relation");
   return XXX_NOT_OK;
 }
 delete [] TmpADTVal;
 if(!Found){
   REL_NOTE_ERROR(REL_BADDEV, "Unable to find catalog info of Views relation");
   return XXX_NOT_OK;
 }
 StrValSize = strlen(RELVIEWSINDXCAT)+1+XxxCoreADTClass::ValueOffset(XXX_STRING);
 TmpADTVal = new char[StrValSize];
 StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
 strcpy(StrVal->Value, RELVIEWSINDXCAT);
 REL_CALL_FN(XxxIDXArray[XXX_BTREE]->
	     GetEntry(VolumeId, RootIndex, StrVal, StrValSize, 
		      XXX_STRING, &ViewsIndex, 
		      sizeof(XxxShoreObjectId), Found),
	     REL_BADDEV, "Unable to get catalog info of Views index");
 delete [] TmpADTVal;
 if (!Found){
   REL_NOTE_ERROR(REL_BADDEV, "Unable to find catalog info of Views index");
   return XXX_NOT_OK;
 }
 // first get a catalog entry for the _VIEWS relation
 XxxRecordSchema* RecSchema = getInitSchema(RELVIEWSCAT);
 XxxValuesArray ValsTab(RecSchema);
 XxxShoreObjectId RecordId;

 viewsInfo *view = newViewsInfo(Name, NULL);

 REL_CALL_FN(deleteViewsEntry(VolumeId, ViewsFile, ViewsIndex, 
			      RecSchema, ValsTab, *view),
	     REL_BADDEV, "Unable to delete view entry");

 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::RetrieveView(const char* ViewName, 
			     RelViewClass*& View,
			     XxxBool& Found) const
{
  Found = XXX_FALSE;
  // set up a cursor over the _SVIEWS relation
  XxxBooleanExpressionPlan *FilterPlan;
  XxxSetCursor ViewCsr;
  XxxRecord *ViewRec;
  REL_CALL_FN(Catalog->makeFilterPlan("viewname", 
				      RelShoreDataSrc::RelShoreDevice::RELVIEWSCAT, 
				      ViewName, FilterPlan),
	      REL_BADSHOREREL, "Unable to make filter expr");
  
  XxxRelation* ViewsRelation = Catalog->ViewsRelation;
  REL_CALL_FN(ViewsRelation->InitCursor1(ViewCsr, FilterPlan),
	      REL_BADSHOREREL, "Unable to open scan of _SVIEWS relation");

  char* ViewInfo;
  int flag = 0;
  XxxRecordSchema* recordSchema = ViewsRelation->CatalogInfo->RecordSchema();
  while ((ViewsRelation->NextItem (ViewCsr, ViewRec) == XXX_OK) &&
	 (ViewCsr.IsValid ())) {
    // Collect the viewinfo for view "ViewName"
    //Bjarke, sep. 14 2001: added const qualifier to variable
    const char *FldPtr;
    ViewRec->GetField (1, recordSchema, FldPtr);
    ViewInfo = ((XxxCoreADTClass::StrValue*)FldPtr)->Value;
    flag = 1;
  }
  
  // If the view doesn't exist
  if (flag == 0)
    return XXX_OK;
  
  Found = XXX_TRUE;
  int len=strlen(ViewInfo);
  
  // parse the view info string
  istrstream InStream ((char *)ViewInfo, len);
  XxxRelationEngine::parseStruct PDS (&InStream, NULL);
  
  XxxRelationEngine* relEngine = (XxxRelationEngine *)XxxDBEngineArray[RelationEngineId];
  int ParseReturn = relEngine->sqlparse(PDS);
  if (PDS.LexerPtr) {
    delete ((yyFlexLexer*)(PDS.LexerPtr));
    PDS.LexerPtr = NULL;
  }
  if (ParseReturn) {
    REL_NOTE_ERROR (REL_BADQUERY, "Error in view expression");
    return XXX_NOT_OK;
  }
  
  View = (RelViewClass *)PDS.OutArg;
  
  REL_CALL_FN(ViewsRelation->CloseCursor(ViewCsr),
	      REL_BADSHOREREL, "Unable to close scan of views relation");
  delete FilterPlan;
  return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::PrintViews(ostream& OStr) const
{
  // set up a cursor over the _SVIEWS relation
  XxxSetCursor ViewCsr;
  XxxRecord *ViewRec;
  
  XxxRelation* ViewsRelation = Catalog->ViewsRelation;
  REL_CALL_FN(ViewsRelation->InitCursor1(ViewCsr, NULL),
	      REL_BADSHOREREL, "Unable to open scan of _SVIEWS relation");

  char* ViewInfo, *ViewName;
  int flag = 0;
  XxxRecordSchema* RecordSchema = ViewsRelation->CatalogInfo->RecordSchema();
  while ((ViewsRelation->NextItem (ViewCsr, ViewRec) == XXX_OK) &&
	 (ViewCsr.IsValid ()))  {
    // Collect the name, iskey, type, and index from the attr 
    // record
    //Bjarke, sep. 14 2001: added const qualifier to variable
    const char *FldPtr;
    ViewRec->GetField (0, RecordSchema, FldPtr);
    ViewName = ((XxxCoreADTClass::StrValue*)FldPtr)->Value;
    OStr << "\"" << ViewName << "\"" ;
    
    ViewRec->GetField (1, RecordSchema, FldPtr);
    ViewInfo = ((XxxCoreADTClass::StrValue*)FldPtr)->Value;
    OStr << "\"" << ViewInfo << "\"" << endl;
  }
  return XXX_OK;
}

XxxBool RelShoreDataSrc::
RelShoreDevice::GetRelation(const char* Name, RelHandle*& Handle) const
{
 //XXX_SM_CHECK;
 if (!XxxShoreSM) {
   return XXX_FALSE;
  }
 // look up the relation in the Root Index
 XxxShoreObjectId NewFile;
 bool Found;
 int StrValSize = strlen(Name)+1+XxxCoreADTClass::ValueOffset(XXX_STRING);
 char *TmpADTVal = new char[StrValSize];
 XxxCoreADTClass::StrValue* StrVal = (XxxCoreADTClass::StrValue*)TmpADTVal;
 strcpy(StrVal->Value, Name);
 if (XxxIDXArray[XXX_BTREE]->GetEntry(VolumeId, RootIndex,
				      StrVal, StrValSize,
				      XXX_STRING,
				      &NewFile, 
				      sizeof(XxxShoreObjectId), 
				      Found) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Error getting from Root Index");
   return XXX_FALSE;
 }
 delete [] TmpADTVal;

 if (!Found) 
   return XXX_FALSE;
 else {
   Handle = new RelShoreHandle(VolumeId, NewFile);
   return XXX_TRUE;
  }
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::GetRelation(const RelShoreHandle* Handle,
			    XxxRelation*& AnswerRel,
			    const RelImplInfo* RelImplInfo)
{
 RelShoreImplInfo* ImplInfo = (RelShoreImplInfo*)RelImplInfo;
 // check the different types of ImplInfo here
 if (!ImplInfo ||
     ImplInfo->ImplType == REL_SHORE_FILE_IMPL) {
   AnswerRel = new XxxShoreFileRelation(0, Handle->ShoreId, this);
  }
 else {
   REL_NOTE_ERROR(REL_BADDEV, "Unimplemented or unknown relation format");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::addItem(const RelNamedHandle& Handle)
{
 XXX_NOTE_ERROR(XXX_INCOMPL, "addItem not done yet");
 return XXX_NOT_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::RemoveItem(const RelNamedHandle& Handle)
{
 XXX_NOTE_ERROR(XXX_INCOMPL, "RemoveItem not done yet");
 return XXX_NOT_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::InitCursor(XxxSetCursor& SCsr) const
{
 XXX_SM_CHECK;
 // step thru the RELTABLESCAT relation scanning the relation names
 if (Catalog && Catalog->TablesRelation) 
   return Catalog->TablesRelation->InitCursor(SCsr);
 else 
   return XXX_NOT_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::CloseCursor(XxxSetCursor& SCsr) const
{
 XXX_SM_CHECK;
 if (Catalog && Catalog->TablesRelation) 
   return Catalog->TablesRelation->CloseCursor(SCsr);
 else 
   return XXX_NOT_OK;
}

XxxErrCode RelShoreDataSrc::
RelShoreDevice::NextItem(XxxSetCursor& SCsr, RelNamedHandle& Handle) const
{
 XXX_SM_CHECK;
 XxxErrCode rc = XXX_OK;
 XxxRecord* TableRec;
 //Bjarke, sep. 14 2001: added const qualifier to variable
 const char *FldPtr;

 if (!Catalog || !Catalog->TablesRelation) 
   return XXX_NOT_OK;

 XxxRecordSchema* TablesSchema = 
   Catalog->TablesRelation->CatalogInfo->RecordSchema();
 while (1) {
   if (Catalog->TablesRelation->NextItem(SCsr, TableRec) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADTABLE, "Unable to get next item");
     return XXX_NOT_OK;
    }
   if (!SCsr.IsValid()) { // no more entries
     return XXX_OK;
    }

   // extract the table name
   TableRec->GetField(0, TablesSchema, FldPtr);
   char* TableName = ((XxxCoreADTClass::StrValue*)FldPtr)->Value;
   // get the object id from the Root Index
   if (GetRelation(TableName, Handle.Hndl) == XXX_TRUE) {
     strcpy(Handle.Name, TableName);
     TableRec->Reclaim(TablesSchema);
     return XXX_OK;
    }
   // otheriwse, it is an unnamed relation (used to signify a collection of rels)
   TableRec->Reclaim(TablesSchema);
  } // while (1)
}

