#include <stdio.h>
#include <stdlib.h>
#include <fstream.h>
#include <dirent.h>
#include <sys/stat.h>

#ifdef __GNUG__
#pragma implementation
#endif

#include "xxx_thread.h"
#include "env.h"
#include "reladt.h"
#include "relcat.h"
#include "relexec.h"
#include "relerror.h"
#include "relation.h"
#include "values.h"
#include "shorerel.h"

//
// Methods of RelShoreDataSrc
//
RelShoreDataSrc::RelShoreDataSrc(RelDataEngine* DataEng) :
RelDataSrc(DataEng, "shore", 10000), deviceArray(REL_SHOREDEV_MAX)
{
 for (int i=0; i < deviceArray.Cnt(); i++)
   deviceArray[i] = 0;

 SrcType = REL_SHORE_SRC;
 XXX_NULL_SM_CHECK;
}
RelShoreDataSrc::~RelShoreDataSrc()
{
 for (int i=0; i < deviceArray.Cnt(); i++)
   delete deviceArray[i];
}

XxxErrCode RelShoreDataSrc::CreateDevice(char*& DeviceName,
					 const int DeviceSize)
{
 XXX_SM_CHECK;
 if (!DeviceName) {
   // create a device with a default name
   DeviceName = new char[300];
   strcpy(DeviceName, "/tmp/");
   strcat(DeviceName, tmpnam(NULL));
  }

 XxxShoreVolumeId VolId;
 if (XxxShoreSM->CreateDevice(DeviceName, DeviceSize, VolId) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, DeviceName);
   return XXX_NOT_OK;   
  }

 if (XxxBeginTrans() != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Error opening transaction");
   return XXX_NOT_OK;   
  }
 // now make sure to initialize the volume with the appropriate catalogs
 if (RelShoreDevice::FormatDevice(VolId) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, DeviceName);
   if (XxxAbortTrans() != XXX_OK) {
     REL_NOTE_ERROR(REL_BADDEV, "Error aborting transaction");
    }
   return XXX_NOT_OK;
  }
 if (XxxCommitTrans() != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, "Error committing transaction");
   return XXX_NOT_OK;   
  }

 if (XxxShoreSM->UnMountDevice(DeviceName) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADDEV, DeviceName);
   return XXX_NOT_OK;   
  }

 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::MountDevice(const char* DeviceName)
{
 XXX_SM_CHECK;
 // search to see if the device exists already
 for (int i=0; i < deviceArray.Cnt(); i++)
   if (!strcmp(deviceArray[i]->DevName, DeviceName)) {
     char buf[100];
     sprintf(buf, "Device %s exists already!", DeviceName);
     XXX_NOTE_ERROR(XXX_BADFILE, buf);
     return XXX_NOT_OK;   
    }

 RelShoreDevice* NewDevice = new RelShoreDevice(DeviceName);
 if (!NewDevice->Valid) {
   delete NewDevice;
   REL_NOTE_ERROR(REL_BADDEV, DeviceName);
   return XXX_NOT_OK;
  }
 else {
   if (deviceArray.Count() == deviceArray.Size())
     deviceArray.Grow();
   deviceArray[deviceArray.Count()++] = NewDevice;
  }
 
 return XXX_OK;
}

XxxBool RelShoreDataSrc::IsMounted(const char* DeviceName) const
{
 // search to see if the device exists already
 for (int i=0; i < deviceArray.Cnt(); i++)
   if (!strcmp(deviceArray[i]->DevName, DeviceName)) {
     return XXX_TRUE;
    }
 return XXX_FALSE;
}

XxxErrCode RelShoreDataSrc::UnMountDevice(const char* DeviceName)
{
 XXX_SM_CHECK;
 // find the device and unmount it
 for (int i=0; i < deviceArray.Cnt(); i++) {
   if (!strcmp(deviceArray[i]->DevName, DeviceName)) {
     delete deviceArray[i]; // this performs the unmounting
     int j;
     for (j = i; j < deviceArray.Cnt(); j++)
       deviceArray[j] = deviceArray[j+1];
     deviceArray.Count() = j-1;
     return XXX_OK;
    }
  }
 REL_NOTE_ERROR(REL_BADDEV, "Unknown device to unmount");
 return XXX_NOT_OK;
}

XxxErrCode RelShoreDataSrc::DestroyDevice(const char* DeviceName)
{
 XXX_SM_CHECK;

 // find the device and destroy it
 for (int i=0; i < deviceArray.Cnt(); i++) {
   if (!strcmp(deviceArray[i]->DevName, DeviceName)) {
     deviceArray[i]->DestroyDev();  // is this Wrong? : PS
     delete deviceArray[i]; // this performs the destroy
     int j;
     for (j = i; j < deviceArray.Cnt(); j++)
       deviceArray[j] = deviceArray[j++];
     deviceArray.Count() = j-1;
     return XXX_OK;
    }
  }
 REL_NOTE_ERROR(REL_BADDEV, "Unknown device to unmount");
 return XXX_NOT_OK;
}

// methods that create/delete relations

XxxErrCode 
RelShoreDataSrc::CreateNewRelation(RelHandle*& Handle, 
				   const char* Name, 
				   const char* DevName,
				   const char* RelImplType,
				   AttrTypeList* Attrs,
				   KeyList* Keys,
				   XxxBool MakeCatEntry)
{
 RelShoreImplInfo ImplInfo;

 // find out the device being used
 if (!DevName || !strcmp(DevName, "default")) 
  { ImplInfo.DeviceIndex = DefaultDevIndex(); }
 else {
   int i;
   // search the device array to find the desired device
   for (i=0; i < deviceArray.Cnt(); i++) {
     if (!strcmp(deviceArray[i]->DevName, DevName)) {
       ImplInfo.DeviceIndex = i;
       break;
      } // if
    } // for
   if ((i >= deviceArray.Cnt()) || !deviceArray[i]) {
     char buf[200];
     sprintf(buf, "Unable to find relation device %s", DevName);
     REL_NOTE_ERROR(REL_BADDEV, buf);
     return XXX_NOT_OK;
    }
  }

 // find out the implementation type of the new relation
 ImplInfo.ImplType = DefaultImplType();
 if (RelImplType) {
   if (!strcmp(RelImplType, "file")) {
     ImplInfo.ImplType = REL_SHORE_FILE_IMPL;
    }
   else {
     char buf[200];
     sprintf(buf, "Unknown SHORE relation impl type %s; using default", 
	     RelImplType);
     REL_NOTE_ERROR(REL_BADDEV, buf);
    }
  }
 
 return CreateNewRelation(Handle, Name, &ImplInfo, Attrs, Keys, MakeCatEntry);
}

XxxErrCode 
RelShoreDataSrc::CreateNewRelation(RelHandle*& Handle, 
				   const char* Name, 
				   const RelImplInfo* RelImplInfo,
				   AttrTypeList* Attrs,
				   KeyList* Keys,
				   XxxBool MakeCatEntry)
{
 XXX_SM_CHECK;
 int DevIndex = DefaultDevIndex();
 RelShoreImplType ImplType = DefaultImplType();
 if (RelImplInfo && RelImplInfo->RelType == REL_STORED) {
   RelShoreImplInfo* ImplInfo = (RelShoreImplInfo*)RelImplInfo;
   DevIndex = ImplInfo->DeviceIndex;
   ImplType = ImplInfo->ImplType;
 }
 RelShoreDevice* Dev = deviceArray[DevIndex];
 RelShoreHandle* Tmp;

 if (Dev->CreateNewRelation(Tmp, ImplType, 
			    Name, Attrs, Keys, MakeCatEntry) == XXX_OK) {
   Handle = Tmp;
   return XXX_OK;
  }
 else return XXX_NOT_OK;
}

XxxErrCode RelShoreDataSrc::DeleteRelation(const char *RelName,
					   const RelHandle* Handle)	
{
 XXX_SM_CHECK;
 RelShoreHandle* ShoreHandle = (RelShoreHandle*) Handle;

 RelShoreDevice* Dev;
 int i;
 for (i=0; i < deviceArray.Cnt(); i++) {
   if (deviceArray[i]->VolumeId == ShoreHandle->ShoreId.VolId) {
     Dev = deviceArray[i];
     break;
    } // if
  } // for
 if ((i >= deviceArray.Cnt()) || !Dev) {
   char buf[200];
   sprintf(buf, "Unable to find device");
   REL_NOTE_ERROR(REL_BADDEV, buf);
   return XXX_NOT_OK;
  }

 return Dev->DeleteRelation(RelName, ShoreHandle);
}

/*
// OWNER: map  DATE: 1/31/97
// PURPOSE: sorts a relation
// NOTES: none
XxxErrCode RelShoreDataSrc::SortRelation (XxxRelation *InputRelation,
					  XxxRelation *&OutputRelation,
					  SortInfo *SInfo,
					  XxxRecordSchema *Schema,
					  RelHandle *&TempSortHandle)
{
  RelShoreDevice *ShoreDevice = deviceArray[DefaultDevIndex()];
  if (!ShoreDevice) {
    XXX_NOTE_ERROR(XXX_SHOREERR, "Couldn't find the SHORE default device");
    return XXX_NOT_OK;
  }
  XxxShoreLogicalOID *RelId = new XxxShoreLogicalOID();
  RelId->VolId = ShoreDevice->VolumeId;
  if (XxxShoreSM->CreateFile(RelId->VolId, RelId->LocalId) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_SHOREERR, "Unable to create temporary file");
    return XXX_NOT_OK;
  }
  XxxShoreObjectId RecordId; // Just a placeholder
  XxxRecord *Rec = 0;

  // open a cursor on the input relation
  XxxSetCursor Handle;
  if (InputRelation->InitCursor(Handle) != XXX_OK){
    XXX_NOTE_ERROR(XXX_BADCURSOR, 
		   "Unable to open input relation cursor");
    return XXX_NOT_OK; 
  }
  while (1) {
    if (Rec) {
      delete Rec;
      Rec = 0;
    }
    if (InputRelation->NextItem(Handle, Rec) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADEXEC, "Error getting NextItem from relation");
      return XXX_NOT_OK;
    }
    if (Handle.IsValid() == XXX_FALSE) {
      break;
    }
    if (XxxShoreSM->InsertRecord(RelId->VolId, RelId->LocalId, Rec,
				 RecordId) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_SHOREERR, 
		     "Could not insert record into temp SHORE file");
      return XXX_NOT_OK;
    }
  }
  // close cursor on the input relation
  if (InputRelation->CloseCursor(Handle) != XXX_OK){
    XXX_NOTE_ERROR(XXX_BADCURSOR, 
		   "Unable to close input relation cursor");
    return XXX_NOT_OK; 
  }
  // sort the SHORE file
  if (XxxShoreSM->SortFile(RelId->VolId,
			   RelId->LocalId,
			   Schema, *SInfo) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_SHOREERR, "Could not sort SHORE file");
    return XXX_NOT_OK;
  }
  
  // Create a SHORE file relation built on the sorted SHORE file
  RelCatalogInfo *CatInfo = new RelCatalogInfo();
  RelShoreDataSrc::RelShoreImplInfo *ImplInfo = 
    new RelShoreDataSrc::RelShoreImplInfo();
  RelStatsInfo *StatsInfo = new RelStatsInfo();
  strcpy (CatInfo->RelName, "Temporary SHORE Relation");
  CatInfo->recSchema = new XxxRecordSchema(*Schema);
  ImplInfo->ImplType = RelShoreDataSrc::REL_SHORE_FILE_IMPL;
  ImplInfo->DeviceIndex = DefaultDevIndex();
  CatInfo->implInfo = ImplInfo;
  CatInfo->statsInfo = StatsInfo;
  (XxxShoreFileRelation *)OutputRelation = 
    new XxxShoreFileRelation(CatInfo, (*RelId), ShoreDevice);

  // store the location in a temp handle
  TempSortHandle = new RelShoreHandle (RelId->VolId, RelId->LocalId);

  return XXX_OK;
}

XxxErrCode 
RelShoreDataSrc::ReclaimSortedRel (RelHandle *&SortHandle) {
  RelShoreHandle *Handle = (RelShoreHandle *)SortHandle;
  if (XxxShoreSM->DeleteFile(Handle->ShoreId.VolId,
			     Handle->ShoreId.LocalId) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot delete temporary SHORE file");
    return XXX_NOT_OK;
  }
  delete (RelShoreHandle *)SortHandle;
  return XXX_OK;
}
*/

XxxErrCode 
RelShoreDataSrc::DeleteTuples (const RelHandle *Handle,
			       XxxBooleanExpressionPlan *ExprPlan,
			       RelCatalogInfo &CatInfo) 
{
  XXX_SM_CHECK;

  XxxRelation *Rel;

  if (GetRelation(Handle, Rel, CatInfo.ImplInfo()) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADDEV, "Cannot retrieve relation from handle");
    return XXX_NOT_OK;
  }
  
  Rel->CatalogInfo = &CatInfo;

  int NumDeleted = 0;
  if (Rel->DeleteRecords(ExprPlan, NumDeleted) != XXX_OK) {
    REL_NOTE_ERROR(REL_BADDEV, "Cannot delete records from relation");
    return XXX_NOT_OK;
  }

  delete Rel;
  return XXX_OK;
}
	
// PURPOSE: Create a new index of specified type from expression
// passed in.
// NOTES: none
	 
XxxErrCode 
RelShoreDataSrc::CreateNewIndex(const char* RelName,
				const char* IndxName,
				XxxValueExpression *Expr,
				const XxxBool Unique,
				const XxxBool Clustered,
				XxxBool& Success,
				XxxIndexType IndxType)
{
 XXX_SM_CHECK;

 // search the device array to find the desired device

 for (int i = 0; i < deviceArray.Cnt (); i++) {
  if (deviceArray[i]->CreateNewIndex (RelName, IndxName, Expr, Unique, Clustered, 
    Success, IndxType) != XXX_OK) {
   REL_NOTE_ERROR (REL_BADDEV, "Unable to create new index");
   return XXX_NOT_OK;
  }

  if (Success)
   return XXX_OK;
 }

 Success = XXX_FALSE;
 return XXX_OK;
}
			 
XxxErrCode 
RelShoreDataSrc::CreateNewView(const char* Name, const char* ViewInfo)
{
 XXX_SM_CHECK;
 RelShoreDevice* Dev = GetDefaultDevice();
 if (Dev) {
   REL_CALL_FN(Dev->CreateNewView(Name, ViewInfo),
	       REL_BADDEV, "Error creating view");
 }
 else {
   REL_NOTE_ERROR (REL_BADDEV, "No default device");
   return XXX_NOT_OK;
 }
 return XXX_OK;
}
XxxErrCode 
RelShoreDataSrc::DeleteView(const char* ViewName)
{
 XXX_SM_CHECK;
 RelShoreDevice* Dev = GetDefaultDevice();
 if (Dev) {
   REL_CALL_FN(Dev->DeleteView(ViewName),
	       REL_BADDEV, "Error deleting view");
 }
 else {
   REL_NOTE_ERROR (REL_BADDEV, "No default device");
   return XXX_NOT_OK;
 }
 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::
RetrieveView(const char* ViewName, 
	     RelViewClass*& View, XxxBool& Found) const
{
 XXX_SM_CHECK;
 Found = XXX_FALSE;
 RelShoreDevice* Dev = GetDefaultDevice();
 if (Dev) {
   REL_CALL_FN(Dev->RetrieveView(ViewName, View, Found),
	       REL_BADDEV, "Error retrieving view");
 }
 else {
   REL_NOTE_ERROR (REL_BADDEV, "No default device");
   return XXX_NOT_OK;
 }
 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::PrintViews(ostream& OStr) const
{
 XXX_SM_CHECK;
 RelShoreDevice* Dev = GetDefaultDevice();
 if (Dev) {
   REL_CALL_FN(Dev->PrintViews(OStr),
	       REL_BADDEV, "Error printing view");
 }
 else {
   REL_NOTE_ERROR (REL_BADDEV, "No default device");
   return XXX_NOT_OK;
 }
 return XXX_OK;
}

XxxBool
RelShoreDataSrc::GetRelationOfIndex(const char *IndxName, char *&RelName) const 
{
  XXX_SM_CHECK;
  RelShoreDevice* Dev = 0;
  // search all devices
  for (int i=0; i < deviceArray.Cnt(); i++) 
    if (deviceArray[i]->GetRelationOfIndex(IndxName, RelName) == XXX_TRUE) {
      Dev = deviceArray[i];
      break;
    }
  if (Dev != 0)
    return XXX_TRUE;
  else
    return XXX_FALSE;
}

XxxErrCode
RelShoreDataSrc::DeleteIndex(const char *IndxName, const char *RelName) 
{
  XXX_SM_CHECK;

  RelHandle *Handle;
  GetRelation(RelName, Handle);
  RelShoreHandle* ShoreHandle = (RelShoreHandle*) Handle;
  RelShoreDevice* Dev;
  int i;
  for (i=0; i < deviceArray.Cnt(); i++) {
    if (deviceArray[i]->VolumeId == ShoreHandle->ShoreId.VolId) {
      Dev = deviceArray[i];
      break;
    } // if
  } // for
  if ((i >= deviceArray.Cnt()) || !Dev) {
    REL_NOTE_ERROR(REL_BADDEV, "Unable to find index's device");
    return XXX_NOT_OK;
  }

  return Dev->DeleteIndex(IndxName, RelName, ShoreHandle);
}

XxxBool RelShoreDataSrc::RelationIsProtected (const char *RelName) 
{
  return (!(strcmp(RelName, RelShoreDevice::RELTABLESCAT) &&
	    strcmp(RelName, RelShoreDevice::RELATTRSCAT) &&
	    strcmp(RelName, RelShoreDevice::RELSTATSCAT) &&
	    strcmp(RelName, RelShoreDevice::RELINDXSCAT)));
}

XxxBool RelShoreDataSrc::GetRelation(const char* RelName,
				     XxxRelation*& AnswerRel,
				     const RelImplInfo* ImplInfo)
{
 for (int i=0; i < deviceArray.Cnt(); i++) {
   RelHandle* Hndl;
   if (deviceArray[i]->GetRelation(RelName, Hndl) == XXX_TRUE) {
     if (deviceArray[i]->GetRelation((RelShoreHandle*)Hndl, 
				     AnswerRel, ImplInfo) 
	 != XXX_OK){
       XXX_NOTE_ERROR(XXX_INTERNAL, "Handle exists, but not the relation");
       if (Hndl) delete Hndl;
       return XXX_FALSE;
      }
     if (Hndl) delete Hndl;
     return XXX_TRUE;
    }
  }
 return XXX_FALSE;
}

XxxBool RelShoreDataSrc::GetRelation(const char* Name,
				     RelHandle*& Handle)
{
 for (int i=0; i < deviceArray.Cnt(); i++) {
   if (deviceArray[i]->GetRelation(Name, Handle) == XXX_TRUE)
     return XXX_TRUE;
  }
 return XXX_FALSE;
}

XxxErrCode RelShoreDataSrc::GetRelation(const RelHandle* Handle,
					XxxRelation*& AnswerRel,
					const RelImplInfo* ImplInfo)
{
 RelShoreHandle* ShoreHandle = (RelShoreHandle*) Handle;

 RelShoreDevice* Dev;
 int i;
 for (i=0; i < deviceArray.Cnt(); i++) {
   if (deviceArray[i]->VolumeId == ShoreHandle->ShoreId.VolId) {
     Dev = deviceArray[i];
     break;
    } // if
  } // for
 if ((i >= deviceArray.Cnt()) || !Dev) {
   char buf[200];
   sprintf(buf, "Unable to find device");
   REL_NOTE_ERROR(REL_BADDEV, buf);
   return XXX_NOT_OK;
  }

 return Dev->GetRelation(ShoreHandle, AnswerRel, ImplInfo);
}

XxxErrCode RelShoreDataSrc::InitCursor(XxxSetCursor& Csr) const {
  CursorHndl* Hndl = new CursorHndl;
  Hndl->ArrayIndex = 0;
  Csr.State = Hndl;
  Csr.AtEnd = XXX_FALSE;
  Csr.NumberRead = 0;
  return XXX_OK;
 }

XxxErrCode RelShoreDataSrc::NextItem(XxxSetCursor& Csr, 
				     RelNamedHandle& ReturnItem) const 
{
 CursorHndl* Hndl = (CursorHndl*)(Csr.State);
 for (;Hndl->ArrayIndex < deviceArray.Cnt(); Hndl->ArrayIndex++) {
   if (!Hndl->DevCsr.IsValid()) {
     if (deviceArray[Hndl->ArrayIndex]->InitCursor(Hndl->DevCsr) != XXX_OK) {
       XXX_NOTE_ERROR(REL_BADDEV, "Error opening cursor");
       return XXX_NOT_OK;
      }
    }
   if ((deviceArray[Hndl->ArrayIndex]->NextItem(Hndl->DevCsr, ReturnItem) 
	== XXX_OK) && Hndl->DevCsr.IsValid()) {
     return XXX_OK;
    }
   if (!Hndl->DevCsr.IsValid()) {
     if (deviceArray[Hndl->ArrayIndex]->CloseCursor(Hndl->DevCsr) != XXX_OK) {
       XXX_NOTE_ERROR(REL_BADDEV, "Error closing cursor");
       return XXX_NOT_OK;
      }
    }
   else {
     REL_NOTE_ERROR(REL_BADDEV, "Error reading from cursor");
     return XXX_NOT_OK;
    }
  } // for loop over all sources

 Csr.AtEnd = XXX_TRUE;
 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::CloseCursor(XxxSetCursor& Csr) const 
{
 CursorHndl* Hndl = (CursorHndl*)(Csr.State);
 delete Hndl;
 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::GetCatInfo(const char* Name,
				       RelCatalogInfo& ReturnCat,
				       XxxBool& Found)
{
 Found = XXX_FALSE;
 for (int i=0; i < deviceArray.Cnt(); i++) {
   if (deviceArray[i]->GetCatInfo(Name, ReturnCat, Found) != XXX_OK) {
     REL_NOTE_ERROR(REL_BADDEV, "Error getting catalog info");
     return XXX_NOT_OK;
    }
   if (Found) {
     RelShoreImplInfo* ImplInfo = (RelShoreImplInfo*)(ReturnCat.implInfo);
     if (ImplInfo) ImplInfo->DeviceIndex = i;
     break;
    }
  }
 
 return XXX_OK;
}

XxxErrCode RelShoreDataSrc::DeleteCatInfo(const char* RelName)
{
  XxxBool Found = XXX_FALSE;
  RelCatalogInfo CatInfo;
  
  for (int i=0; i < deviceArray.Cnt(); i++) {
    if (deviceArray[i]->GetCatInfo(RelName, CatInfo, Found) != XXX_OK) {
      REL_NOTE_ERROR(REL_BADDEV, "Error getting catalog info");
      return XXX_NOT_OK;
    }
    if (Found) {
      if (deviceArray[i]->DeleteCatInfo(RelName) != XXX_OK) {
	REL_NOTE_ERROR(REL_BADDEV, "Error deleting catalog info from device");
	return XXX_NOT_OK;
      }
      return XXX_OK;
    }
  }

  // Execution should never make it here, since RelDataEngine insures the 
  // catalog info exists--but just as a check for Catalog coherency...

  REL_NOTE_ERROR(REL_BADDEV, "Catalog info for specified relation not found on any device");
  return XXX_NOT_OK;
}

XxxErrCode RelShoreDataSrc::AddCatInfo(const char* RelName,
				       const char* DevName,
				       const char* RelImplType,
				       const RelDataSrc::AttrTypeList* Attrs,
				       const RelDataSrc::KeyList* Keys)
{
 XXX_SM_CHECK;
 // first find the device
 XxxErrCode rc = XXX_OK;
 RelShoreDevice* Dev;
 int DevIndex = -1;
 if (!DevName) { DevIndex = 0;}
 else {
   int i;
   // search the device array to find the desired device
   for (i=0; i < deviceArray.Cnt(); i++) {
     if (!strcmp(deviceArray[i]->DevName, DevName)) {
       DevIndex = i;
       break;
      } // if
    } // for
   if ((i >= deviceArray.Cnt()) || (DevIndex < 0)) {
     char buf[200];
     sprintf(buf, "Unable to find device %s", DevName);
     REL_NOTE_ERROR(REL_BADDEV, buf);
     return XXX_NOT_OK;
    }
  }
 Dev = deviceArray[DevIndex];
 // then find the implementation type
 RelShoreImplType ImplType = DefaultImplType();
 if (RelImplType) {
   if (!strcmp(RelImplType, "file")) {
     ImplType = REL_SHORE_FILE_IMPL;
    }
   else {
     char buf[200];
     sprintf(buf, "Unknown SHORE relation impl type %s; using default", 
	     RelImplType);
     REL_NOTE_ERROR(REL_BADDEV, buf);
    }
  }

 return Dev->AddCatInfo(RelName, Attrs, Keys, ImplType);
}

XxxErrCode RelShoreDataSrc::ModifyStatsInfo(const RelCatalogInfo* CatInfo,
					    const RelStatsInfo* StatsInfo)
{
 RelShoreImplInfo* ImplInfo = (RelShoreImplInfo*)CatInfo->ImplInfo();
 int DevIndex = ImplInfo->DeviceIndex;
 if (!deviceArray[DevIndex] ||
     deviceArray[DevIndex]->ModifyStatsInfo(CatInfo, 
					    StatsInfo) != XXX_OK) {
   REL_NOTE_ERROR(REL_BADCAT, "Error propagating stats to disk");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

void RelShoreDataSrc::PrintCats(ostream& OStr)
{
 for (int i=0; i < deviceArray.Cnt(); i++) {
   deviceArray[i]->Catalog->PrintCats(OStr);
  }
}

void RelShoreDataSrc::PrintSchemas(ostream& OStr)
{
 for (int i=0; i < deviceArray.Cnt(); i++) {
   deviceArray[i]->Catalog->PrintSchemas(OStr);
  }
}

void RelShoreDataSrc::PrintDevs(ostream& OStr)
{
 for (int i=0; i < deviceArray.Cnt(); i++) {
   OStr << deviceArray[i]->DevName << '(' << deviceArray[i]->Count() << ')' << endl;
  }
}

XxxErrCode RelShoreDataSrc::RemoveItem(const RelNamedHandle&) 
{
 XXX_NOTE_ERROR(XXX_INCOMPL, "RemoveItem");
 return XXX_NOT_OK;
}
XxxErrCode RelShoreDataSrc::addItem(const RelNamedHandle&) 
{
 XXX_NOTE_ERROR(XXX_INCOMPL, "addItem");
 return XXX_NOT_OK;
}


