#ifndef REL_SHOREREL_H
#define REL_SHOREREL_H
#include "relcat.h"
#include "reldata.h"
#include "relation.h"
#include "shoresm.h"
#include "prd-sort.h"
//Begin: Moustafa
#include "stream-mgr.h"
//End:  Moustafa

// Class defining a Shore Src.
// A Shore src uses a SHORE device to model a logical device.
// Each such device contains a number of relations (which are files).
// The device has an index which maintains a 

class RelShoreDataSrc : public RelDataSrc {
  friend XXX_EXTERN_CLASS_DECL(XxxShoreRelation);
  friend XXX_EXTERN_CLASS_DECL(XxxShoreFileRelation);

 public:
  typedef enum { REL_SHORE_ERROR_IMPL,
		 REL_SHORE_FILE_IMPL} RelShoreImplType;
  XXX_EXTERN_CLASS_DECL(RelShoreImplInfo); 

 protected:
  //// class that specifies a relation handle for shore relations
  class RelShoreHandle : public RelHandle  {
   public:
    XxxShoreLogicalOID ShoreId;
    // this is the ID of the object that "contains" the relation
    // this object contains the following information:
    //   1) what is the storage format of the relation
    //   2) any other info about the relation
    // The other option is to store this info
    // along with the handle. however this increases
    // the handle size unnecessarily.
    //
    RelShoreHandle(const XxxShoreVolumeId& VolumeId,
		   const XxxShoreObjectId& ObjId) : RelHandle(REL_SHORE_SRC)
     { ShoreId.VolId = VolumeId; ShoreId.LocalId = ObjId;}
   };

XXX_EXTERN_CLASS_DECL(RelShoreDevice);

public:
  #include "shorerel-cat.h"
  //@Include: shorerel-cat.h
//ivan@2.5.99 public: was moved 2 lines up because MSDEV 6.0 can't find the *.h file

  #include "shorerel-dev.h"
  //@Include: shorerel-dev.h
protected:

  // array of devices: device 0 is the default device
  XxxPtrArray<RelShoreDevice*> deviceArray;

  // allow new named relations to be created
  virtual XxxErrCode addItem(const RelNamedHandle & item);

  ///
  struct CursorHndl : public XxxSetCursor::CursorState {
    int ArrayIndex;
    XxxSetCursor DevCsr;
   };

 public:
  #include "shorerel-impl.h"
  //@Include: shorerel-impl.h

  RelShoreDataSrc(RelDataEngine* DataEng);
  virtual ~RelShoreDataSrc();

  virtual int HandleSize() const { return sizeof(RelShoreHandle); }
  virtual RelShoreImplType DefaultImplType() const 
    { return REL_SHORE_FILE_IMPL; }
  virtual int DefaultDevIndex() const 
    { return 0; }
  virtual RelShoreDevice *GetDefaultDevice() const {
    return deviceArray[DefaultDevIndex()];
  }

  // methods that manipulate devices
  virtual XxxErrCode CreateDevice(char*& DeviceName,
				  const int DeviceSize);
  virtual XxxErrCode MountDevice(const char* DeviceName);
  virtual XxxErrCode UnMountDevice(const char* DeviceName);
  virtual XxxErrCode DestroyDevice(const char* DeviceName);
  virtual XxxBool IsMounted(const char* DevName) const;

  virtual void PrintCats(ostream& OStr);
  virtual void PrintSchemas(ostream& OStr);
  virtual void PrintDevs(ostream& OStr);
  virtual XxxErrCode GetCatInfo(const char* Name, RelCatalogInfo &,
				XxxBool& Found);
  virtual XxxErrCode AddCatInfo(const char* RelName,
				const char* DevName,
				const char* ImplType,
				const RelDataSrc::AttrTypeList* Attrs,
				const RelDataSrc::KeyList* Keys);
  virtual XxxErrCode ModifyStatsInfo(const RelCatalogInfo* CatInfo,
				     const RelStatsInfo* StatsInfo);

  virtual XxxErrCode DeleteCatInfo(const char* RelName);

  // methods that create/delete relations
  virtual XxxErrCode CreateNewRelation(RelHandle*& Handle, const char* Name,
				       const char* DevName,
				       const char* RelImplType,
				       AttrTypeList* Attrs,
				       KeyList* Keys,
				       XxxBool MakeCatEntry);
  virtual XxxErrCode CreateNewRelation(RelHandle*& Handle, const char* Name,
				       const RelImplInfo* ImplInfo,
				       AttrTypeList* Attrs,
				       KeyList* Keys,
				       XxxBool MakeCatEntry);
  virtual XxxErrCode CreateNewIndex (const char *RelName,
				     const char *IndxName,
				     XxxValueExpression * Expr,
				     const XxxBool Unique,
				     const XxxBool Clustered,
				     XxxBool & Success,
				     XxxIndexType IndxType);

  XxxBool GetRelationOfIndex (const char *IndxName,
			      char *&RelName) const;
  XxxErrCode DeleteIndex (const char *IndxName, const char *RelName);
  virtual XxxErrCode DeleteTuples (const RelHandle *Handle,
				   XxxBooleanExpressionPlan *ExprPlan,
				   RelCatalogInfo &CatInfo);

  virtual XxxErrCode DeleteRelation(const char *RelName,
                                    const RelHandle* Handle);

  virtual XxxErrCode CreateNewView(const char* Name, 
				   const char* ViewInfo);
  virtual XxxErrCode DeleteView (const char *ViewName);
  virtual XxxErrCode RetrieveView(const char* ViewName, 
				  RelViewClass*& View,
				  XxxBool& Found) const;
  virtual XxxErrCode PrintViews(ostream& OStr) const;

  /*
  virtual 
  XxxErrCode SortRelation (XxxRelation *InputRelation,
			   XxxRelation *&OutputRelation,
			   SortInfo *SInfo,
			   XxxRecordSchema *Schema,		
			   RelHandle *&TempSortHandle); //-> paskin 1/24/97
  virtual 
  XxxErrCode ReclaimSortedRel (RelHandle *&SortHandle); //->paskin 2/1/97
  */ 
  // methods that retrieve relations and handles to them
  virtual XxxBool GetRelation(const char* Name,
			      XxxRelation*& AnswerRel, 
			      const RelImplInfo* ImplInfo);
  virtual XxxBool RelationIsProtected (const char *RelName);
  virtual XxxBool GetRelation(const char* Name, 
			      RelHandle*& Handle);
  virtual XxxErrCode GetRelation(const RelHandle* Handle,
				 XxxRelation*& AnswerRel, 
				 const RelImplInfo* ImplInfo);

  // to iterate over the RelNamedHandles
  virtual XxxErrCode InitCursor(XxxSetCursor&) const;
  virtual XxxErrCode NextItem(XxxSetCursor&, RelNamedHandle& ReturnItem) const;
  virtual XxxErrCode CloseCursor(XxxSetCursor&) const;
  virtual XxxErrCode RemoveItem(const RelNamedHandle & item);
};

/**
 * Implementation of a relation stored in SHORE.
 * This is a dummy superclass. There are various actual
 * implementations of a relation in SHORE. Each forms a 
 * subclass of this.
 */
class XxxShoreRelation : public XxxStoredRelation {
 protected:
  typedef XxxStack<XxxShoreObjectId *> ShoreOIDStack;
  typedef XxxStack<XxxRecord *> ShoreRecordStack;
  const XxxShoreLogicalOID RelationId;
  // the device that created this relation
  const RelShoreDataSrc::RelShoreDevice* Device;
 public:
  //// Cursor over a shore relation
  class CursorInfo : public XxxRelation::CommonCursorInfo { 
   public:
    XxxBooleanExpressionPlan* RFilter; // residual filter
    CursorInfo(XxxRecordSchema* Schema,
	       XxxBooleanExpressionPlan* Expr,
	       XxxBooleanExpressionPlan* RExpr, // residual
	       XxxBool Exact,
	       const RelExecParameters* Parameters = NULL,
	       const XxxRecord** SibRecs = NULL,
	       const XxxRecordSchema** SibSchema = NULL,
	       int CurrentJoinPos = 0)
      : CommonCursorInfo(Schema, Expr, Exact, Parameters, 
			 SibRecs, SibSchema, CurrentJoinPos) 
    { RFilter = RExpr; }
    virtual ~CursorInfo() {}
    virtual XxxErrCode NextRecord(XxxShoreObjectId*&, XxxRecord*&,
				  XxxBool& Success) = 0;
   };

  XxxShoreLogicalOID const ShoreOID() const { return RelationId; }
  XxxShoreRelation(RelCatalogInfo* Cat, const XxxShoreLogicalOID& RelId,
		   RelShoreDataSrc::RelShoreDevice* Dev)
    : XxxStoredRelation(Cat), RelationId(RelId), Device(Dev) {}
  virtual ~XxxShoreRelation() {}
 };

/**
// Implementation of a relation as a file. Most straightforward impl.
*/
class XxxShoreFileRelation : public XxxShoreRelation {
 protected:
  virtual XxxErrCode addItem(XxxRecord* const& AddRecord);
  XxxErrCode addToAllIndexes(const XxxRecord* RecordToAdd,
			     const XxxShoreObjectId* RecordId);
  XxxErrCode delFromAllIndexes(const XxxRecord* RecordToDel,
			       const XxxShoreObjectId* RecordId);

  //// Cursor for file scan
  class FScanCursorInfo : public XxxShoreRelation::CursorInfo { 
   public:
    scan_file_i ScanIter;
    XxxShoreLogicalOID RelationId;
    FScanCursorInfo(const XxxShoreLogicalOID& RelId,
		    XxxRecordSchema* Schema,
		    XxxBooleanExpressionPlan* Expr,
		    const RelExecParameters* Parameters = NULL,
		    const XxxRecord** SibRecs = NULL,
		    const XxxRecordSchema** SibSchema = NULL,
		    int CurrentJoinPos = 0);
    virtual ~FScanCursorInfo();
    virtual XxxErrCode NextRecord(XxxShoreObjectId*&, XxxRecord*&,
				  XxxBool& Success);
   };

  //Begin: Moustafa
  //// Cursor for stream scan
  class SScanCursorInfo : public XxxShoreRelation::CursorInfo { 
  private:
    ResultQueue *resQueue; // a queue to hold the streming tuples

  public:
    // A cursor on the original Shore relation
    CursorInfo *RelCInfo;
    SScanCursorInfo(CursorInfo*& CInfo);
    virtual ~SScanCursorInfo();
    virtual XxxErrCode NextRecord(XxxShoreObjectId*&, XxxRecord*&,
				  XxxBool& Success);
  };
  //End: Moustafa

  //Check if index on relation matches an expr
  virtual XxxErrCode matchIndex(XxxBooleanExpressionPlan * Expr,
				int SibArity,
				int CurrJoinPos,
				XxxBool& Success,
				XxxMatchInfo*& MatchedInfo,
				RelShoreDataSrc::RelShoreImplInfo::ShoreIndexImplInfo*&
				ShoreIndxInfo) const;
  XxxErrCode chooseAccessPath(XxxBooleanExpressionPlan*,
			      const XxxRecord* Rec[],
			      const XxxRecordSchema* Schema[],
			      const RelExecParameters* Parameters,
			      int SibArity, 
			      int CurrJoinPos,
			      CursorInfo*& CInfo) const;
  XxxErrCode fscanAccessPath(XxxBooleanExpressionPlan*,
			     const XxxRecord* Rec[],
			     const XxxRecordSchema* Schema[],
			     const RelExecParameters* Parameters,
			     int SibArity, 
			     int CurrJoinPos,
			     CursorInfo*& CInfo) const;
  XxxErrCode iscanAccessPath(XxxBooleanExpressionPlan* IndxExpr,
			     const XxxRecord* Rec[],
			     const XxxRecordSchema* Schema[],
			     const RelExecParameters* Parameters,
			     int SibArity, 
			     int CurrJoinPos,
			     CursorInfo*& CInfo) const;

  //Begin: Moustafa
  XxxErrCode sscanAccessPath(XxxBooleanExpressionPlan* Expr,
			     const XxxRecord * Rec[],
			     const XxxRecordSchema * Schema[],
			     const RelExecParameters * Parameters,
			     int SibArity,
			     int CurrJoinPos,
			     CursorInfo*& CInfo) const;
  //End: Moustafa

  XxxErrCode indexedAccessPath(XxxBooleanExpressionPlan* IndxExpr,
			       XxxBooleanExpressionPlan* ResidExpr,
			       const XxxRecord* Rec[],
			       const XxxRecordSchema* Schema[],
			       const RelExecParameters* Parameters,
			       int SibArity, 
			       int CurrJoinPos,
			       const XxxMatchInfo* MatchedInfo,
			       RelShoreDataSrc::RelShoreImplInfo::
			       ShoreIndexImplInfo *ShoreIndxInfo,
			       CursorInfo*& CInfo) const;
  
 public:
  XxxShoreFileRelation(RelCatalogInfo* Cat, const XxxShoreLogicalOID& RelId,
		       RelShoreDataSrc::RelShoreDevice* Dev);
  virtual ~XxxShoreFileRelation();

  // scan with filter predicate
  virtual XxxErrCode InitCursor1(XxxSetCursor&, 
				 XxxBooleanExpressionPlan*,
				 const XxxRecord* Rec[] = NULL,
				 const XxxRecordSchema* Schema[] = NULL,
				 const RelExecParameters* Parameters = NULL,
				 int SibArity = 1, 
				 int   CurrJoinPos = 0) const;
  virtual XxxErrCode InitCursor(XxxSetCursor& Csr) const {
    return InitCursor1(Csr, NULL);
   }

  virtual XxxErrCode NextItem1(XxxSetCursor&,
                              XxxShoreObjectId *&,
                              XxxRecord*&) const;
  virtual XxxErrCode NextItem(XxxSetCursor& SCrs,
                              XxxRecord*& Rec) const {
    XxxShoreObjectId *ObjId = 0;
    XxxErrCode Err = NextItem1(SCrs, ObjId, Rec);
    if (ObjId) delete ObjId;
    return Err;
  }
  virtual XxxErrCode CloseCursor(XxxSetCursor&) const;

  virtual XxxErrCode DeleteRecords(XxxBooleanExpressionPlan *FilterPlan,
				   int &NumDeleted);
  virtual XxxErrCode UpdateRecords(XxxBooleanExpressionPlan *FilterPlan,
				   XxxValExprPlanList* TargetPlanList,
				   int &NumUpdated);

  virtual XxxErrCode RemoveItem(XxxRecord* const& DelRecord);
  
  virtual XxxErrCode AddIndex(const char *IndexName, 
			      XxxValueExpression *Expr,
			      XxxIndexType IndxType,
			      XxxBool Unique, 
			      const void *IndxColMetaInfo);
  // this is used to change logging properties, set up bulk loads, etc
  virtual XxxErrCode PrepareForLoad();
 };

#endif



