/***************************************************************************
 Author: Moustafa Hammad
 Date: 10/31/2001

 The StrmMgr (Stream Manager is responsible for  servicing stream requests)
****************************************************************************/
#ifndef XXX_SM_H
#define XXX_SM_H

#define MAX_EXP_STREAMS 100      // Max number of expected streams cosidered for prefetching
#define MAX_SUPPORTED_STREAMS 10 // Max number of concurrent streams
#define MAX_TOTAL_STREAMS 50     // Max number of streams in all system queues
#define CONSIDERED_SEGMENT 100   // Max number of considered expected segments
#define MAX_REQUEST_QUEUE_SIZE 100 
#define MAX_RESULT_QUEUE_SIZE 2097152  // Max number of tuples waiting in the result queue
// return values from StreamRead function

#define STREAM_NEW_VALUE   0  // the function reads a new value 
#define STREAM_NO_VALUE    1  // currently there is no new value
#define STREAM_END         2  // the stream is totally scanned
#define STREAM_READ_ERROR -1  // a read error occured


#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include "xxx_thread.h"
#include "shoresm.h"

XXX_EXTERN_CLASS_DECL(StreamManagerThread);

//----------------------------------------------------------------------------
// the tuple information

class Tuple{

 public:
  XxxShoreObjectId* RecordId;
  XxxRecord* Record;

  Tuple(XxxShoreObjectId* RecId, XxxRecord* Rec);
  ~Tuple(){};

};

class ResultQueue; // forward declarations


//TODO: Latter on  you may need this????
// ===================================================
// This table is used to keep track of objects created
//   by the large streaming field in the da storage
//
/*
 class LargeObjectTable{
 public:
  class QueueEntry {
  public:
    struct timeval TimeStamp; 
    XxxShoreLogicalOID  LargeObjectId; // ID of the large object
    QueueEntry *NextEntry;    /// The next queue entry (NULL if none)
    QueueEntry *PrevEntry;    /// The previous queue entry (NULL if none)

    // Default Constructor

    QueueEntry (struct timeval TS, 
		XxxShoreLogicalOID LOID,
		QueueEntry *Next,
		QueueEntry *Prev);
    ~QueueEntry (){};
  };
  
    // iterator over the queue
 private:
  QueueEntry* currPos;
  XxxBool ValidScan;
 public:
  // Constructor.
  LargeObjectTable(){HeadTable=NULL; TailTable=NULL;};
  // Add to bottom of table
  XxxErrCode AddTuple(struct timeval TS, XxxShoreLogicalOID LOID); 
  // Remove tuple from table
  XxxErrCode RemoveTuple(struct timeval TS, XxxShoreLogicalOID LOID);
  
  bool isEmpty(){
    return (HeadTable==NULL);
  };
  void initQIter(XxxBool Head);
  void getNext(struct timeval *TS, XxxShoreLogicalOID *LOID);
  void getPrevious(struct timeval *TS, XxxShoreLogicalOID *LOID);
  XxxBool ScanIsValid() { return (ValidScan); };
  
  // Functions for time stamps assignment and comparison
  // TS1==TS2?
  XxxBool TimeStampsEqual(struct timeval TS1, struct timeval TS2){
    double diff = ((TS1.tv_sec-TS2.tv_sec)+
		   (TS1.tv_usec-TS2.tv_usec)/1000000.0);
    return ((diff==0?XXX_TRUE:XXX_FALSE));
  };
  // TS1>TS2?
  XxxBool TimeStampsGreater(struct timeval TS1, struct timeval TS2){
    double diff = ((TS1.tv_sec-TS2.tv_sec)+
		   (TS1.tv_usec-TS2.tv_usec)/1000000.0);
    return ((diff>0?XXX_TRUE:XXX_FALSE));
  };
  // TS1=TS2
  void TimeStampsAssign(struct timeval *TS1, struct timeval TS2){
    TS1->tv_sec = TS2.tv_sec;
    TS1->tv_usec = TS2.tv_usec;
  };
  // TS1-TS2
  double TimeStampsDiff(struct timeval TS1, struct timeval TS2){
    return ((TS1.tv_sec-TS2.tv_sec)+
	    (TS1.tv_usec-TS2.tv_usec)/1000000.0);
  };
  // State determination function
  unsigned char GetTupleState(struct timeval TS, double qw, double opw){
    struct timeval CTS; 
    struct timezone Temp; 
    gettimeofday(&CTS, &Temp); 
    
    double interval = TimeStampsDiff(CTS, TS);
     if (interval <0)
      ASSERT(0); // can't be negative
    if (interval < opw)
      return ACTIVE;
    else if (interval < qw)
      return ZOMBIE;
    else
      return EXPIRED;
  };
  
 protected:
  QueueEntry *HeadTable;  /// The Head of the Table
  QueueEntry *TailTable;  /// The Tail of the Table
 };
*/
//===================================================
class StreamRequest{
 public:

  Tuple *OrgTuple;      // the orignal tuple before streaming its contnets.

  ResultQueue* ResQ;    //output queue of the results
  XxxRecordSchema* RecordSchema;       // The scehma of the tuple
  int StrmFldPos;       // Position of the stream field in the tuple
  //  int  reqRank;         // a token given by SM to identify the request
  //  LargeObjectTable* LargeObjectT; // A pointer to the associated large object table if any
  bool newReq;          // to indicate first

  StreamRequest(Tuple *t, ResultQueue *q,  XxxRecordSchema *RecSchema, int SfPos);
  ~StreamRequest(){};

  StreamRequest& operator=(const StreamRequest& R);
};
//--------------------------------------------------------------------------------------------
/**
 * A queue of stream requests.
 */
class ReqQueue{
 public:
  class QueueEntry {
  public:
    StreamRequest* Req; 
    QueueEntry *NextEntry;  // The next queue entry (NULL if none) 
    QueueEntry *PrevEntry;  // The previous queue entry (NULL if none)
    /**
     * Default Constructor
     */
    QueueEntry (StreamRequest *Req, 
		QueueEntry *Next,
		QueueEntry *Prev);
    ~QueueEntry (){};
  };

  // iterator over the queue
 private:
  QueueEntry* currPos;
 public:
  void initQIter();
  QueueEntry* getNext();

  /**
   * Constructor.
   */
  ReqQueue(){Entries=NULL;};
  ~ReqQueue(){};
  XxxErrCode enqueue(StreamRequest *Req);
  XxxErrCode dequeue(StreamRequest *&s);

  bool isEmpty(){
    return (Entries==NULL);
  };
  XxxErrCode removeEntry(StreamRequest *Req);

 protected:
  QueueEntry *Entries;  /// The queue of items
  XxxBaseMutex Latch;  /// A latch to protect access to the queue
};

/**
 * This queue is used to hold the result stream records
 */

class ResultQueue{
 private:
  int AttacedStreams;
  int Stop;
 public:
  //(exp-only) must be removed
  XxxRecord* templateRec;
  struct timeval timeOfNextTuple;
  int remainInBurstCnt;
  //(exp-only)
  class QueueEntry {
  public:
    Tuple *ResultTuple;
    QueueEntry *NextEntry;    /// The next queue entry (NULL if none)
    QueueEntry *PrevEntry;    /// The previous queue entry (NULL if none)
    /**
     * Default Constructor
     */
    QueueEntry (Tuple *ResT,
		QueueEntry *Next,
		QueueEntry *Prev);
    ~QueueEntry (){};
  };

  // iterator over the queue
 private:
  QueueEntry* currPos;
  unsigned int CurrSize;
 public:

  /**
   * Constructor.
   */
  ResultQueue(){AttacedStreams=0; HeadEntries=NULL; TailEntries=NULL; Stop=0; CurrSize=0; templateRec=NULL; 
  struct timezone Temp; 
  timeOfNextTuple.tv_sec=0; timeOfNextTuple.tv_usec=0;};
  XxxErrCode enqueue(Tuple *ResT);
  XxxErrCode dequeue(Tuple *&ResT);
  XxxErrCode flush(XxxRecordSchema* Schema);
  bool isEmpty(){
    return (TailEntries==NULL);
  };
  bool isFull(){	
    return (CurrSize==MAX_RESULT_QUEUE_SIZE);
  };
  int QSize(){
	  return CurrSize;
  };
  bool IsDone(){
    return ((AttacedStreams==0)&&(isEmpty()));
  };
  void Enough(){
    Stop=1;
  };
  bool IsEnough(){
    return (Stop==1);
  };
  XxxErrCode AddStream(){
    AttacedStreams++;
    return XXX_OK;
  };
  XxxErrCode RemoveStream(){
    ASSERT(AttacedStreams>0);
    AttacedStreams--;
    return XXX_OK;
  };
 protected:
  QueueEntry *TailEntries;  /// The queue of items
  QueueEntry *HeadEntries;  /// The queue of items
  XxxBaseMutex Latch;  /// A latch to protect access to the queue
};

//---------------------------------------------------------------------------

class StreamManagerThread : public XxxBaseThread {

 protected:
  virtual void ExecCode();

  ReqQueue inputStreamQueue;   // holds the ready stream requests
  ReqQueue currStreamQueue;    // holds the current stream requests
  //  ResQueue zombieStreamQueue;  // hold the stream requests, finished but not yet consumed by the client

  int nMax;    // Maximum number of supported streams
  //  int period;  // Period length in milli seconds

  int cntStream;   // count of current number of streams
  int finish;    // at sever shutdown

  void finishCurrStream(StreamRequest *&s);
  XxxErrCode StreamManagerThread::InitNewRequest(StreamRequest *strm);
  XxxErrCode StreamManagerThread::GetNewTuple(StreamRequest *strm, int *retV);
  XxxErrCode StreamManagerThread::CloseRequest(StreamRequest *strm);

  void streamScheduler(); 

 public:  

  StreamManagerThread (int n); // constructor
  void addNewStream(XxxShoreObjectId* RecId, XxxRecord* Rec, ResultQueue *resQueue, XxxRecordSchema* Schema, int StrmFldPos);  // interface to the client
  // delete all stream request that are attached to the resultqueue
  void RemoveStreams(ResultQueue *resQueue);  
  void cleanup(){finish=1;};
  virtual ~StreamManagerThread(){};
};


#endif
