//mwm1 added line below
#include <stdarg.h>
#include <iostream.h>
#include "prd-config.h"
#include "types.h"
#include "adtfunc.h"
#include "expr.h"
#include "exprplan.h"
#include "funcexpr.h"
#include "suspcomp.h"
#include "smartfunc.h"

//
// XxxFuncParseInfo
//
// constructor used by QPEs
XxxFuncParseInfo::XxxFuncParseInfo(const char* Name,
				   XxxGenericParse* Exp,
				   int Index,
				   XxxBasicType RetType) 
{
  strncpy(FuncName, Name, XXX_FUNCNAME_MAX);
  FuncName[XXX_FUNCNAME_MAX-1] = 0;
  ExprParse = Exp;
  EngineIndex = Index;
  TypeId = XXX_UNKNOWN;
  ReturnType = RetType;
  IsTopLevel = XXX_TRUE;
  NumArgs=0;
  IsAggr = XXX_FALSE;
}

// constructor used by ADTs
XxxFuncParseInfo::XxxFuncParseInfo(const char* Name,
				   int MethIndex,
				   XxxGenericParse* ParseData, 
				   XxxBasicType ThisTypeId,
				   XxxBasicType RetType,
				   XxxADTMetaInfo* CatInfo)
{
  strncpy(FuncName, Name, XXX_FUNCNAME_MAX);
  FuncName[XXX_FUNCNAME_MAX-1] = 0;
  TypeId = ThisTypeId;
  ExprParse = ParseData;
  EngineIndex = 0;
  ReturnType = RetType;
  IsTopLevel = XXX_TRUE;
  NumArgs=0;
  MethodIndex = MethIndex;
  TypeCatInfo = CatInfo;
  IsAggr = XXX_FALSE;
}

//Bjarke, sep. 17 2001: explicitly initialized base class XxxGenericParse
XxxFuncParseInfo::XxxFuncParseInfo(const XxxFuncParseInfo& src) 
  : XxxGenericParse()
{
  NumArgs=src.NumArgs;
  ExprParse=NULL;

  if (src.ExprParse)
    ExprParse=src.ExprParse->Replicate();
  
  TypeId=src.TypeId;
  EngineIndex=src.EngineIndex;
  IsTopLevel=src.IsTopLevel;
  ReturnType=src.ReturnType;
  strncpy(FuncName, src.FuncName, XXX_FUNCNAME_MAX);
  FuncName[XXX_FUNCNAME_MAX-1] = 0;
  
  MethodIndex=src.MethodIndex;
  TypeCatInfo=NULL;
  if (src.TypeCatInfo)
    src.TypeCatInfo->Replicate(TypeCatInfo);
  IsAggr = src.IsAggr;
}



//
// XxxFuncPlanInfo
//
XxxFuncPlanInfo::XxxFuncPlanInfo(const XxxFuncPlanInfo &FPI)
  : XxxGenericPlan(FPI), NumArgs(FPI.NumArgs), 
  planUsage(FPI.planUsage), compressOutput(FPI.compressOutput), 
  decompressOwner(FPI.decompressOwner), 
  suspendComputation(FPI.suspendComputation), TypeId(FPI.TypeId), 
  EngineIndex(FPI.EngineIndex), ExprPlan(NULL)
{
  strncpy(FuncName, FPI.FuncName, XXX_FUNCNAME_MAX);
  if (FPI.ExprPlan) 
    ExprPlan = FPI.ExprPlan->Replicate();

  decompressInput = new XxxBool[NumArgs];
  for (int i = 0; i < NumArgs; i++)
    decompressInput[i] = FPI.decompressInput[i];
  ScratchPad = 0;
  ResultCache = 0;
}

XxxFuncPlanInfo::XxxFuncPlanInfo(const XxxFuncParseInfo* ParseInfo,
				 int ArgCount,
				 XxxGenericPlan* Plan)
{
  strncpy(FuncName, ParseInfo->FuncName, XXX_FUNCNAME_MAX);
  FuncName[XXX_FUNCNAME_MAX-1] = 0;
  TypeId = ParseInfo->TypeId;
  EngineIndex = ParseInfo->EngineIndex;
  ExprPlan = Plan;

  NumArgs = ArgCount;
  MethodIndex = ParseInfo->MethodIndex;
  compressOutput = XXX_TRUE;
  decompressInput = new XxxBool[ArgCount];
  for (int i =0; i < ArgCount; i++)
    decompressInput[i] = XXX_TRUE;
  decompressOwner = XXX_TRUE;
  suspendComputation = XXX_FALSE;
  ScratchPad = 0;
  ResultCache = 0;
}

XxxFuncPlanInfo::XxxFuncPlanInfo(const char* Name,
				 XxxBasicType Type,
				 int EngIndex,
				 XxxGenericPlan* Exp) 
{
  strncpy(FuncName, Name, XXX_FUNCNAME_MAX);
  FuncName[XXX_FUNCNAME_MAX-1] = 0;
  TypeId = Type;
  EngineIndex = EngIndex;
  ExprPlan = Exp;
  ScratchPad = 0;
  ResultCache = 0;
}

XxxFuncPlanInfo::~XxxFuncPlanInfo()
{
  if (ExprPlan) delete ExprPlan;
  if (decompressInput) delete [] decompressInput;
  if (ScratchPad) delete ScratchPad;
  if (ResultCache) delete ResultCache;
}

XxxErrCode 
XxxFuncPlanInfo::modifyArgCompression(XxxValueExprPlan* OwnerPlan,
				      XxxValExprPlanList* ArgPlans)
{
 XxxValExprPlanList* Tmp = ArgPlans;
 int InputCount = 0;
 while (Tmp) {
   if (Tmp->Plan->ValType() == XxxValueExpression::FUNC_EXPR) {
     // it is a function
     XxxFuncExpression::FuncPlan* ArgFuncPlan =
       (XxxFuncExpression::FuncPlan*)(Tmp->Plan);
     XxxFuncPlanInfo* ArgFuncPlanInfo = ArgFuncPlan->ExprPlan;
     if (ArgFuncPlanInfo && ArgFuncPlanInfo->TypeId == this->TypeId) {
       XxxFuncPlanInfo* ArgPlanInfo = (XxxFuncPlanInfo*)(ArgFuncPlanInfo);
       ArgPlanInfo->compressOutput = XXX_FALSE;
       decompressInput[InputCount] = XXX_FALSE;
      }
   } // if FUNC_EXPR
   Tmp = Tmp->Next; InputCount++;
  } // for each argument
 if (OwnerPlan) {
   if (OwnerPlan->ValType() == XxxValueExpression::FUNC_EXPR) {
     // it is a function
     XxxFuncExpression::FuncPlan* ArgFuncPlan =
       (XxxFuncExpression::FuncPlan*)(OwnerPlan);
     XxxFuncPlanInfo* ArgFuncPlanInfo = ArgFuncPlan->ExprPlan;
     if (ArgFuncPlanInfo && ArgFuncPlanInfo->TypeId == this->TypeId) {
       XxxFuncPlanInfo* ArgPlanInfo = (XxxFuncPlanInfo*)(ArgFuncPlanInfo);
       ArgPlanInfo->compressOutput = XXX_FALSE;
       decompressOwner = XXX_FALSE;
     }
   } // if FUNC_EXPR
 } // if Owner
 return XXX_OK;
}
//----------------------------------------------------------------------------
XxxErrCode XxxFuncPlanInfo::SuspendComputation()
{
  suspendComputation = XXX_TRUE;
  return XXX_OK;
}
//----------------------------------------------------------------------------
XxxErrCode XxxFuncPlanInfo::UnsuspendComputation()
{
  suspendComputation = XXX_FALSE;
  return XXX_OK;
}
//----------------------------------------------------------------------------
XxxErrCode 
XxxFuncPlanInfo::Evaluate(XxxValueEnv* Env,
			  XxxADTClass::FuncMethodInfo* ThisMethInfo,
			  XxxADTValue *ReturnValue) const
{ 
  if (!ThisMethInfo) {
    XXX_NOTE_ERROR(XXX_INTERNAL, "method info does not exist");
    return XXX_NOT_OK;
  }
  if (ThisMethInfo->ExecFunc == NULL) {
    XXX_NOTE_ERROR(XXX_INTERNAL, "method not defined");
    return XXX_NOT_OK;
  }
  ASSERT((Env->NumArgs == ThisMethInfo->NumArgs));


  XxxADTClass::ExecuteFunc MethodFunc = ThisMethInfo->ExecFunc;
  
  if (suspendComputation) {
    if (SuspendEvaluation(Env, ThisMethInfo, ReturnValue) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADFUNC, "Cannot suspend ADT method invocation");
      return XXX_NOT_OK;
    }
  } else {
    // if there is a function cache, look in it
    if (ResultCache &&
	ResultCache->Search(Env, ReturnValue)) {
      // found it in the cache
    }
    else { // otherwise, have to evaluate the method
      XXX_CALL_RET((this->*MethodFunc)(Env, ReturnValue),
		   XXX_BADFUNC, "error executing ADT method");
      // if there is a function cache, add ReturnValue to it
      if (ResultCache)
	ResultCache->Add(Env, ReturnValue);
    }
  }
  
  return XXX_OK;
}
//----------------------------------------------------------------------------
XxxErrCode 
XxxFuncPlanInfo::SuspendEvaluation(XxxValueEnv* Env, 
				   XxxADTClass::FuncMethodInfo* ThisMethInfo,
				   XxxADTValue *ReturnValue) const
{
  // Suspend the computation; this records the suspended computation in
  // this client's repository.
  XxxSuspCompLib::XxxSuspCompId ID;
  if (XXX_THISENV().SuspendedComputations->
      InsertSuspComp(new XxxSuspendedComp(this, Env, 
					  ThisMethInfo), ID) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot record suspended computation");
    return XXX_NOT_OK;
  }
  // Mark the return value as suspended
  ReturnValue->Set(XxxADTValue::SUSPENDED_COMPUTATION);
  // Record the new suspended computation
  ReturnValue->SuspensionId = ID;
  if (XXX_THISENV().DebugFlags.SuspComp)
    cerr << "Suspended computation " << ReturnValue->SuspensionId 
	 << " stored." << endl;
  return XXX_OK;
}

XxxErrCode XxxFuncPlanInfo::AggrInit(XxxAggrMethodInfo* MethInfo,
				     XxxAggrOpHandle *Handle) const
{
  ASSERT ((Handle && MethInfo));

  XxxADTClass::AggrInitFunc InitFunc = MethInfo->InitFunc;
  XXX_CALL_RET((this->*InitFunc)(*Handle),
	       XXX_BADFUNC, "error initializing ADT aggregate");
  return XXX_OK;
}

XxxErrCode XxxFuncPlanInfo::AggrTerm(XxxAggrMethodInfo* MethInfo,
				     XxxAggrOpHandle *Handle,
				     XxxADTValue* FinalVal) const
{
  ASSERT ((Handle && MethInfo));

  XxxADTClass::AggrTermFunc TermFunc = MethInfo->TermFunc;
  XXX_CALL_RET((this->*TermFunc)(*Handle, FinalVal),
	       XXX_BADFUNC, "error terminating ADT aggregate");
  return XXX_OK;
}

XxxErrCode XxxFuncPlanInfo::AggrAccum(XxxAggrMethodInfo* MethInfo,
				      XxxValueEnv* Env, 
				      XxxAggrOpHandle *Handle) const
{
  ASSERT ((Handle && MethInfo));
  ASSERT((Env->NumArgs == MethInfo->NumArgs));

  XxxADTClass::AggrAccumFunc AccumFunc = MethInfo->AccumFunc;
  XXX_CALL_RET((this->*AccumFunc)(Env, *Handle),
	       XXX_BADFUNC, "error accumulating ADT aggregate");
  return XXX_OK;
}

XxxErrCode XxxFuncPlanInfo::AggrDrop(XxxAggrMethodInfo* MethInfo,
				     XxxValueEnv* Env, 
				     XxxAggrOpHandle *Handle) const
{
  ASSERT ((Handle && MethInfo));
  ASSERT((Env->NumArgs == MethInfo->NumArgs));

  XxxADTClass::AggrDropFunc DropFunc = MethInfo->DropFunc;
  XXX_CALL_RET((this->*DropFunc)(Env, *Handle),
	       XXX_BADFUNC, "error dropping ADT aggregate");
  return XXX_OK;
}

//
// Methods of XxxMethodInfo
//
XxxMethodInfo::XxxMethodInfo(const XxxMethodInfo &MInfo) :
  NumArgs(MInfo.NumArgs), ReturnType(MInfo.ReturnType), 
  MIGen(MInfo.MIGen), OptFunc(MInfo.OptFunc), 
  SideEffectFree(MInfo.SideEffectFree)
{
  strncpy(Name, MInfo.Name, XXX_FUNCNAME_MAX);
}

//
// Methods of XxxFuncMethodInfo
//
XxxFuncMethodInfo::ArgInfo::ArgInfo(const ArgInfo& AInfo)
{
  Count = AInfo.Count;
  for (int i = 0; i < Count; i++)
    ArgTypes[i] = AInfo.ArgTypes[i];    
}
XxxFuncMethodInfo::ArgInfo::ArgInfo(int Arity, ...)
{
 Count = Arity;
 va_list ap;
 va_start(ap, Arity);
 int i = 0;
 while (i < Count) {
   ArgTypes[i] = va_arg(ap, XxxBasicType);
   i++;
  }
 va_end(ap);
}

XxxFuncMethodInfo::XxxFuncMethodInfo(const XxxFuncMethodInfo &MInfo) :
  XxxMethodInfo(MInfo), ExecFunc(MInfo.ExecFunc), 
  ArgumentTypes(MInfo.ArgumentTypes)
{
}
// maintained for backwards compatibility
XxxFuncMethodInfo::XxxFuncMethodInfo(const char* FName,
				     XxxBasicType RetType, 
				     XxxADTClass::MetaInfoGen MIG, 
				     XxxADTClass::OptimizeFunc OF,
				     XxxADTClass::ExecuteFunc EF, 
				     int Arity, ...) 
  : XxxMethodInfo(FName, Arity, RetType, MIG, OF), ExecFunc(EF)
{
 ArgumentTypes.Count = Arity;
 va_list ap;
 va_start(ap, Arity);
 int i = 0;
 while (i < ArgumentTypes.Count) {
   ArgumentTypes.ArgTypes[i] = va_arg(ap, XxxBasicType);
   i++;
  }
 va_end(ap);
}

XxxFuncMethodInfo::XxxFuncMethodInfo(const char* FName,
				     XxxBasicType RetType, 
				     ArgInfo& TypeSignature,
				     XxxADTClass::ExecuteFunc EF,  
				     XxxADTClass::MetaInfoGen MIG,
				     XxxADTClass::OptimizeFunc OF)
  : XxxMethodInfo(FName, TypeSignature.Count, RetType, MIG, OF), 
    ExecFunc(EF), ArgumentTypes(TypeSignature)
{
}

XxxFuncMethodInfo::XxxFuncMethodInfo(const char* FName,
				     XxxBasicType RetType, 
				     ArgInfo& TypeSignature,
				     XxxADTClass::ExecuteFunc EF,  
				     XxxADTClass::OptimizeFunc OF,
				     XxxADTClass::MetaInfoGen MIG)
  : XxxMethodInfo(FName, TypeSignature.Count, RetType, MIG, OF), 
    ExecFunc(EF), ArgumentTypes(TypeSignature)
{
}

XxxFuncMethodInfo::XxxFuncMethodInfo(const char* FName,
				     XxxBasicType RetType, 
				     ArgInfo& TypeSignature,
				     XxxADTClass::ExecuteFunc EF,  
				     const XxxGenericPlan::PlanUsage& PUsage,
				     XxxADTClass::MetaInfoGen MIG)
  : XxxMethodInfo(FName, TypeSignature.Count, RetType, MIG, 0), 
    ExecFunc(EF), ArgumentTypes(TypeSignature)
{
  Usage.Add(PUsage);
}

// TODO: haven't implemented MIGChoice yet ...
XxxFuncMethodInfo::XxxFuncMethodInfo(const char* FName, 
				     XxxBasicType RetType, 
				     ArgInfo& TypeSignature,
				     XxxADTClass::ExecuteFunc EF,  
				     const XxxGenericPlan::PlanUsage& PUsage,
				     MIGenOption MIGenChoice)
  : XxxMethodInfo(FName, TypeSignature.Count, RetType, 0, 0), 
    ExecFunc(EF), ArgumentTypes(TypeSignature)
{
  Usage.Add(PUsage);
}

XxxErrCode XxxFuncMethodInfo::Matches(const char*FuncName,
				      const XxxValueExpression* Owner,
				      const XxxValueExprList *ArgList,
				      XxxBasicType& RetType,
				      XxxADTMetaInfo*& ReturnMetaInfo,
				      XxxBool& Success) const
{
 int NArgs = (ArgList?ArgList->Count():0);
 Success = XXX_FALSE;
 if (!strncmp(Name, FuncName, XXX_FUNCNAME_MAX) &&
     (NumArgs == NArgs)) {
   // maybe it is a match -- check closer
   Success = XXX_TRUE;
   for (int i = 0; i < NumArgs; i++) {
     if (((*ArgList)[i])->BasicType() != ArgumentTypes.ArgTypes[i]) {
       Success = XXX_FALSE;
       break;
      }
    }
   if (Success) {
     RetType = ReturnType;
     if (MIGen) ReturnMetaInfo = MIGen(Owner, ArgList);
     else ReturnMetaInfo = NULL;
    }
  }
 return XXX_OK;
}

//
// Methods of XxxAggrMethodInfo
//
XxxAggrMethodInfo::
XxxAggrMethodInfo(const char* FName, 
		  XxxBasicType RetType,
		  ArgInfo& TypeSignature,
		  XxxADTClass::MetaInfoGen MIG, 
		  XxxADTClass::OptimizeFunc OF,
		  XxxADTClass::AggrInitFunc IFunc,
		  XxxADTClass::AggrAccumFunc AFunc, 
		  XxxADTClass::AggrTermFunc TFunc,
		  XxxADTClass::AggrDropFunc DFunc): 
  XxxFuncMethodInfo(FName, RetType, TypeSignature, 0, MIG, OF),
  InitFunc(IFunc), AccumFunc(AFunc), TermFunc(TFunc), DropFunc(DFunc) 
{ }
XxxAggrMethodInfo::XxxAggrMethodInfo(const XxxAggrMethodInfo &AMI) 
  : XxxFuncMethodInfo(AMI), InitFunc(AMI.InitFunc),
    AccumFunc(AMI.AccumFunc), TermFunc(AMI.TermFunc), 
    DropFunc(AMI.DropFunc) 
{ }



