#include <iostream.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/resource.h>
#include "prd-config.h"
#include "client.h"
#include "values.h"
#include "engines.h"
#include "defeng.h"
#include "shoresm.h"
#include "protocol.h"
#include "sockstream.h"
#include "httprequest.h"
#include "suspcomp.h"

#define MAX_REQUEST 2048
 
// These are the authenticating password/username pair
// OLD: const char *XXX_PASSWORD = getenv("XXXPASSWORD");
// OLD: const char *XXX_USERNAME = getenv("XXXUSERNAME");
// Password not checked yet!
const char *XXX_PASSWORD = NULL;
const char *XXX_USERNAME = NULL;

// static variable needs to be initialized
int XxxClientInfo::CurrentClientId = 0;
XxxBaseMutex XxxClientInfo::latch;
const unsigned int XxxRequestThread::DefaultRequestMaxSize = 5000;

ostream& operator<<(ostream& OStream, XxxClientInfo& ClientInfo)
{
 OStream << ClientInfo.UserName << " (" << ClientInfo.UserId 
	  << ")" << endl;
 return OStream;
}

XxxClientInfo::XxxClientInfo(XxxRequestThread* Thread) 
{
 XxxUseMutex AutoMutex(CBlatch);
 ThisThread = Thread;
 TotalClockTime = 0.0;
 NumCBThreads = 0;
 CallbacksPermitted = XXX_TRUE;
 UserName = NULL;
 Password = NULL;
 {
  XxxUseMutex AutoMutex(latch);
  UserId = CurrentClientId++;
 }
 Protocol = 0;
}
XxxClientInfo::~XxxClientInfo() 
{
 if (UserName) delete [] UserName;
 if (Password) delete [] Password;
 if (Protocol) delete Protocol;
}

XxxErrCode XxxClientInfo::AddCallback() 
{
  {
    XxxUseMutex AutoMutex(CBlatch);
    if (!CallbacksPermitted)
      return XXX_NOT_OK;
  }
  XxxBool Waiting = XXX_FALSE;
  while (XXX_TRUE) {
    {
      XxxUseMutex AutoMutex(CBlatch);
      if (NumCBThreads < smlevel_0::max_xct_thread) {
	NumCBThreads++;
	break;
      }
    }
    if (XXX_THISENV().DebugFlags.Callback) {
      cerr << "Maximum number of threads (" << smlevel_0::max_xct_thread 
	   << ") reached: callback waiting to be serviced." << endl;
    }
    Waiting = XXX_TRUE;
    CBThreadDone.Wait();
  }

  if (XXX_THISENV().DebugFlags.Callback) {
    if (Waiting)
      cerr << "Waiting callback permitted to execute." << endl;
  }
  return XXX_OK;
}

XxxErrCode XxxClientInfo::RemoveCallback() {
  {
    XxxUseMutex AutoMutex(CBlatch);
    NumCBThreads--;
  }
  NoCBThreads.Signal();
  CBThreadDone.Signal();
  if (XXX_THISENV().DebugFlags.Callback) {
    cerr << "A callback has completed." << endl;
  }
  return XXX_NOT_OK;
}

XxxErrCode XxxClientInfo::PermitCallbacks() {
  XxxUseMutex AutoMutex(CBlatch);
  CallbacksPermitted = XXX_TRUE;
  if (XXX_THISENV().DebugFlags.Callback) {
    cerr << "Callbacks are now permitted." << endl;
  }
  return XXX_OK;
}

XxxErrCode XxxClientInfo::WaitForCallbacks() {
  {
    XxxUseMutex AutoMutex(CBlatch);
    CallbacksPermitted = XXX_FALSE;
  }
  while (XXX_TRUE) {
    {
      XxxUseMutex AutoMutex(CBlatch);
      if (NumCBThreads == 0)
	break;
      if (XXX_THISENV().DebugFlags.Callback) {
	cerr << "Waiting for " << NumCBThreads 
	     << " callbacks to complete execution." << endl;
      }
    }
    NoCBThreads.Wait();
  }
  if (XXX_THISENV().DebugFlags.Callback) {
    cerr << "All callbacks have completed execution." << endl;
  }
  return XXX_OK;
}

XxxErrCode XxxClientInfo::StringSendReceive(const char* SendString,
					    char RecvString[],
					    int RecvStringLen)
{
 int Socket = ThisThread->SocketDesc();

 if (::send(Socket, SendString, 
	    strlen(SendString), 0) < (int)strlen(SendString)) {
   cerr << "Error sending data to client " << endl;
   return XXX_NOT_OK;
  }

 // if a reply is desired
 if (RecvStringLen > 0) {
   // read client characteristics
#if defined(XXX_WIN32) && defined(XXX_NONPREEMPTIVE)
   HANDLE handle = CreateEvent(0, TRUE, FALSE, 0);
   int rep = WSAEventSelect((unsigned int)Socket, handle, FD_READ);

   win32_compat_event_t* eventHandle = new win32_compat_event_t((HANDLE)handle);

   if (eventHandle->is_down() || 
       (bool)eventHandle->wait(::WAIT_FOREVER)) 
#else
   if (ThisThread->fileHandle->is_down() || 
       (bool)ThisThread->fileHandle->wait(::WAIT_FOREVER)) 
#endif
   {
     cerr << "Error reading from input socket" << endl;
     return XXX_NOT_OK;
    }
   
   int BytesRead;
   if ((BytesRead = ::recv(Socket, RecvString, RecvStringLen, 0)) < 1) {
     cerr << "Error getting client data" << endl;
     return XXX_NOT_OK;
    }
   RecvString[BytesRead] = 0; // null-terminate it
   if (XXX_THISENV().DebugFlags.Protocol) {
     cerr << "Got :" << RecvString << endl;
   }
  }
 return XXX_OK;
}

XxxErrCode XxxClientInfo::EstablishProtocol(int Socket)
{
 const int SETUPSTRINGMAX = 1000;
 char ClientBuf[SETUPSTRINGMAX];
 // ask for and get 
 //   a) name 
 //   b) password : currently not used
 //   c) protocol name
 if (StringSendReceive("Setup?\n", ClientBuf, SETUPSTRINGMAX) != XXX_OK) {
   cerr << "Error getting setup info" << endl;
   return XXX_NOT_OK;
  }
 
 // ask for setup info
 istrstream InStr(ClientBuf);
 const int ARGSTRINGMAX=300;
 char ArgBuf[ARGSTRINGMAX];

 InStr >> ArgBuf;
 if (!InStr.good()) {
   cerr << "Error getting client name" << endl;
   return XXX_NOT_OK;
  }
 UserName = new char[strlen(ArgBuf)+1];
 strcpy(UserName, ArgBuf);

 InStr >> ArgBuf;
 if (!InStr.good()) {
   cerr << "Error getting client password" << endl;
   return XXX_NOT_OK;
  }
 // do something with the password : nothing for now
 Password = new char[strlen(ArgBuf) + 1];
 strcpy (Password, ArgBuf);

 if((XXX_USERNAME != NULL) && (XXX_PASSWORD != NULL)) {
	if((strcmp(UserName,XXX_USERNAME) != 0) || (strcmp(Password,XXX_PASSWORD) != 0)){
		cerr << "Error authenticating user " << UserName << " with password " << Password << endl;
		return XXX_NOT_OK;
	}
 }

 InStr >> ArgBuf;
 if (InStr.bad()) {
   cerr << "Error getting protocol name" << endl;
   return XXX_NOT_OK;
  }
 char* ProtocolName = ArgBuf;

 XxxClientProtocol::ProtoDefn ProtocolDefn(ProtocolName);
 XxxClientProtocol::ProtoDefn* ProtoDefn= NULL;
 XxxDefaultEngine* DefEng = (XxxDefaultEngine*)XxxDBEngine::DefaultDBEngine();
 if ((DefEng->ProtocolTable.FindTableItem(&ProtocolDefn, 
					  ProtoDefn) 
      != XXX_TRUE) || (!ProtoDefn)) {
   cerr << "Invalid protocol name:" << ProtocolName << endl;
   return XXX_NOT_OK;
  }
 Protocol = ProtoDefn->GetProtocol(*this);

 if (!Protocol || (Protocol->OpenConnection() != XXX_OK)) {
   cerr << "Error opening connection using protocol" << endl;
   return XXX_NOT_OK;
  }

 return XXX_OK;
}

//
// methods for the server request thread 
//
XxxRequestThread::XxxRequestThread(XxxSocketConnection* Socket,
				   const XxxEnvironment& DefaultEnv,
				   unsigned int ReqBufSize)
: XxxBaseThread(DefaultEnv, "client_thread"), socket(Socket), 
  RequestBuf(NULL), ClientInfo(NULL)
{
 if (ReqBufSize <= 0) 
   RequestBufSize = DefaultRequestMaxSize;
 else
   RequestBufSize = ReqBufSize;
 valid = 1;
 RequestBuf = new char[RequestBufSize];

 // set defaults in environment
 // use a thread-safe output stream
 XxxEnv.OutputStr = new XxxSockStream(Socket->socketDesc);
 XxxEnv.ErrorStr = XxxEnv.OutputStr;

#if defined(XXX_WIN32) && defined(XXX_NONPREEMPTIVE)
  // eventHandle is initialized just before it is used - not clear why!
  handle = CreateEvent(0, TRUE, FALSE, 0);
  int rep = WSAEventSelect((unsigned int) Socket->socketDesc, handle, FD_READ);

  eventHandle = new win32_compat_event_t(handle);
#else
  fileHandle = new sfile_read_hdl_t(Socket->socketDesc);
#endif

  //Begin: Moustafa
  StopQuery = 0; // initially execute until query ends
  //End: Moustafa

 if (XXX_THISENV().DebugFlags.Connection) {
   cerr << "Created request thread." << endl;
 }
 W_COERCE(fork());
}

// The format of the input is <LANGNAME>:command. Find the engine
// with the LANGNAME
static XxxBool FindEngine(const char* RequestBuf,
			  int& ChosenEngine, int&ChosenLanguage) 
{
 ChosenEngine = -1;
 ChosenLanguage = -1;
 XxxUseMutex AutoMutex(*XxxDBEngine::Latch);
 for (int i = 0; i < XxxDBEngine::NumDBEngines; i++) {
   for (int j = 0; j < XxxDBEngine::DBEngineArray[i]->NumLanguages; j++) {
     if (!strncasecmp(XxxDBEngine::DBEngineArray[i]->LanguageNames[j], RequestBuf, 
		      strlen(XxxDBEngine::DBEngineArray[i]->LanguageNames[j]))) {
       ChosenEngine = i;
       ChosenLanguage = j;
       break;
      }
    } // for each language of engine i
   if (ChosenEngine >= 0) break;
  } // for each engine
 if (ChosenEngine >= 0) return XXX_TRUE;
 else return XXX_FALSE;
}


XxxErrCode XxxRequestThread::runClientSession()
{
  // set up protocol
  ClientInfo = new XxxClientInfo(this);
  if (ClientInfo->EstablishProtocol(socket->socketDesc) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot establish protocol");
    this->ShutDown();
    cleanup(XXX_FALSE,  // CleanEnvs
	    XXX_FALSE,  // CleanStorage
	    XXX_FALSE,  // CleanProtocol
	    XXX_TRUE);  // CleanSocket
    return XXX_NOT_OK;
  }
  // Initialize the storage
  if (XxxEnv.InitStorage() != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot initialize storage");
    this->ShutDown();
    cleanup(XXX_FALSE,  // CleanEnvs
	    XXX_FALSE,  // CleanStorage
	    XXX_TRUE,   // CleanProtocol
	    XXX_TRUE);  // CleanSocket
    return XXX_NOT_OK;
  }
  // create envs for all ADTs
  if (XxxEnv.CreateEnvs() != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot create ADT environments");
    this->ShutDown();
    cleanup(XXX_FALSE,  // CleanEnvs
	    XXX_TRUE,   // CleanStorage
	    XXX_TRUE,   // CleanProtocol
	    XXX_TRUE);  // CleanSocket
    return XXX_NOT_OK;
  }

  XxxDefaultEngine* DefEng = (XxxDefaultEngine*)XxxDBEngine::DefaultDBEngine();
  if (DefEng->AddClient(ClientInfo) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot add new client");
    this->ShutDown();
    cleanup(XXX_TRUE,   // CleanEnvs
	    XXX_TRUE,   // CleanStorage
	    XXX_TRUE,   // CleanProtocol
	    XXX_TRUE);  // CleanSocket
    return XXX_NOT_OK;
  }
  if (XXX_THISENV().DebugFlags.Connection) {
    cerr << "Starting client session for " << ClientInfo->UserName << endl;
  }

  // TODO: need to fix implementation of gettimeofday() in shorelib
#ifdef XXX_WIN32
  clock_t Time[4];
#else
  struct timeval Time[4];
  struct timezone TZ[4];
#endif

  unsigned int RequestCounter = 0;
  
  while (1) {
    // remove any error flags from previous queries
    XXX_THISENV().Clear();

    XxxBool Valid = XXX_FALSE;
    XxxBool ToBreak = XXX_TRUE;
    // get input from client
    getInput(Valid, ToBreak);
   
    if (ToBreak) {
      cleanup(XXX_TRUE,  // CleanEnvs
	      XXX_TRUE,  // CleanStorage
	      XXX_TRUE,  // CleanProtocol
	      XXX_TRUE); // CleanSocket
      break;
    }

    if (XXX_THISENV().DebugFlags.Connection) {
      cerr << "Received request from client " << ClientInfo->UserId 
	   << "." << endl;
    }
    RequestCounter++;
#ifndef XXX_WIN32    
    gettimeofday(&Time[0], &TZ[0]);
#else
    Time[0] = clock();
#endif
    if (Valid) {
      int ChosenEngine, ChosenLanguage;
      if (FindEngine(RequestBuf, ChosenEngine, ChosenLanguage)) {
	XxxBool Success = XXX_FALSE;
	// start a transaction if none exists already
	if (!XXX_THISTHREAD()->xct())
	  XXX_THISENV().MultiStmtXact = XXX_FALSE;
	else 
	  // this is probably not necessary, since it will stay set
	  XXX_THISENV().MultiStmtXact = XXX_TRUE;
	// Permit callbacks again
	if (ClientInfo->PermitCallbacks() != XXX_OK)
	  XXX_NOTE_ERROR(XXX_BADEXEC, "Unable to permit callbacks");
	if (!XXX_THISENV().MultiStmtXact && (XxxBeginTrans() != XXX_OK)) {
	  XXX_NOTE_ERROR(XXX_BADCMD, "Unable to begin transaction");
	  Success = XXX_FALSE;
	} else {
	  XxxBasicType DummyType; XxxGenericParse* DummyParse;
#ifndef XXX_WIN32    
	  gettimeofday(&Time[2], &TZ[2]);
#else
	  Time[2] = clock();
#endif
	  // Call the appropriate parser
	  if (XxxDBEngineArray[ChosenEngine]->
	      ExprParse(RequestBuf, NULL, DummyType, 
			DummyParse) != XXX_OK)
	    Success = XXX_FALSE;
	  else
	    Success = XXX_TRUE;
#ifndef XXX_WIN32    
	  gettimeofday(&Time[3], &TZ[3]);
#else
	  Time[3] = clock();
#endif
	} // else, there was no problem creating the Xact
	 
	// Print any errors to the client
	if (processErrors("Error processing user request") != XXX_OK) {
	  XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot process errors");
	  Success = Valid = XXX_FALSE;
	}
       	// to signal end of transmission
	if (ClientInfo->Protocol->CloseRequest() != XXX_OK) {
	  XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot close request");
	  Success = Valid = XXX_FALSE;
	}
	// now wait for any running callbacks (these are in the same
	// transaction, and must end before it can be committed.)
	if (ClientInfo->WaitForCallbacks() != XXX_OK)
	  XXX_NOTE_ERROR(XXX_BADEXEC, "Unable to remove callbacks");
	// If the Xact was not successful, abort
	if (!Success) {
	  if (XxxAbortTrans() != XXX_OK)
	    XXX_NOTE_ERROR(XXX_BADEXEC, "Unable to abort transaction");
	  // We may have been in the midst of a multi-stmt xact!
	  // All the same, its over now.
	  XXX_THISENV().MultiStmtXact = XXX_FALSE;
	} 
	else {
	  // Otherwise commit the transaction
	  if (!XXX_THISENV().MultiStmtXact && 
	      (XxxCommitTrans() != XXX_OK))
	    XXX_NOTE_ERROR(XXX_BADEXEC, "Unable to commit transaction");
	}
      } // if (ChosenEngine >= 0)
      else {
	XXX_NOTE_ERROR(XXX_BADEXEC, "Unknown engine name");
	// to signal end of transmission
	if (ClientInfo->Protocol->CloseRequest() != XXX_OK) {
	  XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot close request using protocol");
	  Valid = XXX_FALSE;
	}
      } // else if (ChosenEngine < 0)
    } // if (Valid)

    // Valid may have been initially false, or may have been set to false
    // because the client died
    if (!Valid) {
      if (XXX_THISENV().DebugFlags.Connection) {
	cerr << "client EOF" << endl;
      }
      cleanup(XXX_TRUE,  // CleanEnvs
	      XXX_TRUE,  // CleanStorage
	      XXX_TRUE,  // CleanProtocol
	      XXX_TRUE); // CleanSocket
      this->ShutDown(); 
      // continue;
      break;
    }

    if (XXX_THISENV().DebugFlags.Connection) {
      cerr << "Processed request.  " << endl;
    }
    // measure the time before flushing the output
    double TotalElapsedTime, ProcessingTime;
#ifndef XXX_WIN32    
    gettimeofday(&Time[1], &TZ[1]);
    TotalElapsedTime = (Time[1].tv_sec - Time[0].tv_sec) +
      (Time[1].tv_usec - Time[0].tv_usec)/1000000.0;
    ProcessingTime = (Time[3].tv_sec - Time[2].tv_sec) +
      (Time[3].tv_usec - Time[2].tv_usec)/1000000.0;
#else
    Time[1] = clock();
    TotalElapsedTime = ((double)(Time[1]-Time[0]))/CLOCKS_PER_SEC;
    ProcessingTime = ((double)(Time[3]-Time[2]))/CLOCKS_PER_SEC;;
#endif
    cerr << "Execution time: " << TotalElapsedTime 
	 << " seconds (Processing time: " << ProcessingTime << " seconds)."
	 << endl;
    ClientInfo->TotalClockTime += TotalElapsedTime;

    // Now make sure to reclaim the temp file storage if needed.
    // TODO: actually, this should check the size of the temp file
    // and make the decision based on that
    if (!(RequestCounter%10)) {
      if (XxxEnv.DumpStorage() != XXX_OK) {
	XXX_PRINT_ERROR(cerr, "Error dumping storage");
	XXX_CLEAR_ERROR();
      }
      if (XxxEnv.InitStorage() != XXX_OK) {
	XXX_PRINT_ERROR(cerr, "Request:Error dumping storage");
	XXX_CLEAR_ERROR();
      }
    } // if (!(RequestCounter%10))
  } // while (1)

  // This thread must get deleted, and ideally this should be
  // called by some scheduler thread that keeps track of which
  // request threads have completed. One way to do this is to associate
  // the thread with the ClientInfo structure, and to mark it as
  // "to be reclaimed", and put it on a free thread list
  if (DefEng->RemoveClient(ClientInfo) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADEXEC, "Error while removing the client!");
  }
  return XXX_OK;
}

static XxxBool IsCallbackInContext(int ClientId, 
				   XxxShoreXactId CallbackXactId) {
  XxxClientInfo* ClientInfo;
  if (((XxxDefaultEngine*)XxxDBEngine::DefaultDBEngine())->
      GetClient(ClientId, ClientInfo) != XXX_OK) {
    return XXX_FALSE;
  }
  if (!(ClientInfo->ThisThread->xct())) {
    return XXX_FALSE;
  }
  XxxShoreXactId ClientXactId;
  if (XxxShoreSM->XactToId(ClientInfo->ThisThread->xct(),
			   ClientXactId) != XXX_OK) {
    return XXX_FALSE;
  }
  if (ClientXactId != CallbackXactId) {
    return XXX_FALSE;
  }
  return XXX_TRUE;
}

static XxxBool IsValidCallback(const int ClientId, 
			       const XxxShoreXactId *XactId,
			       const int ADTTypeId,//Bjarke aug. 3: int added as type
			       const XxxADTValue *ObjectRef,
			       const XxxADTMetaInfo *MetaInfo,
			       const int MessageLength,
			       const char *Message)
{
  // The following code works correctly.  It was removed to permit callbacks
  // out of transactional context.  Ideally, the client protocol will determine
  // when such callbacks are permissible.
  /*
    if (!(IsCallbackInContext(ClientId, *XactId)))
      return XXX_FALSE;
  */
  if (!(XxxADTArray[ADTTypeId]->SupportsCallback())) {
    return XXX_FALSE;
  }
  return XXX_TRUE;
}

XxxErrCode XxxRequestThread::serviceHTTPRequest(const char *RequestLine)
{
  XxxHTTPRequest request(RequestLine);
  // If it is a callback, service it
  int ClientId;
  int TypeId;
  XxxADTValue *ObjectRef;
  XxxADTMetaInfo *MetaInfo;
  char *Message;
  int MessageSize;
  XxxShoreXactId *CallbackXactId, ClientXactId;
  if (!(request.IsCallback(ClientId, 
			   CallbackXactId,
			   TypeId,
			   ObjectRef,
			   MetaInfo,
			   MessageSize,
			   Message))) {
    // The request is not a callback, and is unsupported.
    if (request.WriteResponseHeader(*(XxxEnv.OutputStr),
				    XxxHTTPRequest::HTTP_BAD_REQUEST,
				    NULL, 0, NULL, NULL) != XXX_OK) {
      cerr << "Cannot write HTTP response header." << endl;
    }
    return XXX_NOT_OK;
  }
  if (XXX_THISENV().DebugFlags.Callback) {
    cerr << "Received Callback: [Client " << ClientId << "]" << endl;
  }

  if (!(IsValidCallback(ClientId, 
			CallbackXactId,
			TypeId,
			ObjectRef,
			MetaInfo,
			MessageSize,
			Message))) {
    // The callback is not valid
    if (request.WriteResponseHeader(*(XxxEnv.OutputStr),
				    XxxHTTPRequest::HTTP_BAD_REQUEST,
				    NULL, 0, NULL, NULL) != XXX_OK) {
      cerr << "Cannot write HTTP response header." << endl;
    }
    return XXX_NOT_OK;
  }
  XxxClientInfo *clientInfo;
  if (((XxxDefaultEngine*)XxxDBEngine::DefaultDBEngine())->
      GetClient(ClientId, clientInfo) != XXX_OK) {
    // The client with which this callback is associated is no longer on.
    // This happens for CGI script-type applications in which a single
    // connection is made per query.
    clientInfo = NULL;
  }
  const XxxBool 
    AttachedXact = (clientInfo && 
		            IsCallbackInContext(ClientId, *CallbackXactId));
  if (AttachedXact) {
    // This is a blocking call; if there are insufficient resources
    // to support this callback (SHORE limits the number of threads
    // which can attach to a Xact) it will block until it is safe to proceed.
    // If callbacks are no longer allowed, the function will return an
    // error.
    if (clientInfo && (clientInfo->AddCallback() != XXX_OK)) {
      XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot add callback to client info");
      cleanup(XXX_FALSE,  // CleanEnvs
	      XXX_TRUE,   // CleanStorage
	      XXX_FALSE,   // CleanProtocol
	      XXX_TRUE);  // CleanSocket
      return XXX_NOT_OK;
    }
    // Identify with the client's current transaction
    if (AttachXact(clientInfo->ThisThread->xct()) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot attach the client's transaction");
      return XXX_NOT_OK;
    }
    // Copy the client's environment to this thread's environment
    // Note: there is an operator= for XxxEnv
    XxxEnv = clientInfo->ThisThread->XxxEnv;
  } else {
    if (XxxBeginTrans() != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot begin a transaction");
      return XXX_NOT_OK;
    }
  }
  // Initialize the storage
  if (XxxEnv.InitStorage() != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot initialize storage");
    return XXX_NOT_OK;
  }
  // If this thread is unattached to a main client thread, initialize the
  // ADT environments
  if (!AttachedXact && (XxxEnv.CreateEnvs() != XXX_OK)) {
    XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot create ADT environments");
    cleanup(XXX_FALSE,  // CleanEnvs
	    XXX_TRUE,   // CleanStorage
	    XXX_FALSE,   // CleanProtocol
	    XXX_TRUE);  // CleanSocket
    return XXX_NOT_OK;
  }

  XxxADTClass* ADT = (XxxADTArray[TypeId]);
  // Get a handle on the result to be served
  while (XXX_TRUE) {
    // This is not a true loop; it's just convenient for error handling.
    // It is only executed once (note the break at the end).

    // If the object that was sent over is suspended, execute its
    // plan.
    if (ObjectRef->Is(XxxADTValue::SUSPENDED_COMPUTATION)) {
      XxxSuspendedComp *SuspComp = NULL;
      if (XXX_THISENV().SuspendedComputations->
	  GetSuspComp(ObjectRef->SuspensionId, SuspComp) != XXX_OK) {
	if (request.WriteResponseHeader(*(XxxEnv.OutputStr),
					XxxHTTPRequest::HTTP_NOT_FOUND,
					NULL, 0, NULL) != XXX_OK) {
	  cerr << "Cannot write HTTP error response." << endl;
	}
	break;
      }
      if (SuspComp->Evaluate(ObjectRef) != XXX_OK) {
	if (request.WriteResponseHeader(*(XxxEnv.OutputStr),
					XxxHTTPRequest::
					HTTP_INTERNAL_SERVER_ERROR,
					NULL, 0, NULL) != XXX_OK) {
	  cerr << "Cannot write HTTP error response." << endl;
	}
	break;
      }
    }

	int Size;
    // Determine the result's size
    if (ADT->GetResultSize(ObjectRef,
			   MetaInfo,
			   Message,
			   MessageSize, Size) != XXX_OK) {
      if (request.WriteResponseHeader(*(XxxEnv.OutputStr),
				      XxxHTTPRequest::
				      HTTP_INTERNAL_SERVER_ERROR,
				      NULL, 0, NULL) != XXX_OK) {
	cerr << "Cannot write HTTP error response." << endl;
      }
      break;
    }
    // Send over the HTTP header
    if (request.WriteResponseHeader(*(XxxEnv.OutputStr),
				    XxxHTTPRequest::HTTP_OK, NULL, Size, 
				    ADT->GetContentType(ObjectRef, MetaInfo, Message, MessageSize)) 
	!= XXX_OK) {
      cerr << "Cannot write HTTP response header." << endl;
      break;
    }
    // Hand this off to the correct ADT
    if (ADT->ServiceCallback(ObjectRef, MetaInfo, Message, MessageSize,
			     *(XxxEnv.OutputStr)) != XXX_OK) {
      XXX_CHECK_ERROR("Callback could not be serviced");
      break;
    }
	break;
  }
  cleanup(XXX_FALSE,  // CleanEnvs
	  XXX_TRUE,   // CleanStorage
	  XXX_FALSE,   // CleanProtocol
	  XXX_TRUE);  // CleanSocket
  // Delete the ADT environments, if they were created
  if (!AttachedXact && (XxxEnv.DeleteEnvs() != XXX_OK)) {
    XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot delete ADT environments");
    return XXX_NOT_OK;
  }

  if (AttachedXact) {
    // Detach from the client's transaction
    if (DetachXact(clientInfo->ThisThread->xct()) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADEXEC, "Callback thread cannot detach transaction");
    }
    if (clientInfo && (clientInfo->RemoveCallback() != XXX_OK)) {
      XXX_NOTE_ERROR(XXX_BADEXEC, "Callback thread cannot remove itself");
    }
  } else {
    if (XxxCommitTrans() != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot commit a transaction");
      return XXX_NOT_OK;
    }
  }
  return XXX_OK;
}

XxxErrCode XxxRequestThread::AttachXact(XxxShoreXact *Xact) {
  attach_xct(Xact);
  return XXX_OK;
}

XxxErrCode XxxRequestThread::DetachXact(XxxShoreXact *Xact) {
  detach_xct(Xact);
  return XXX_OK;
}

XxxErrCode XxxRequestThread::processErrors(const char *Message)
{
  if (XXX_IS_ERROR()) {
    if (ClientInfo->Protocol->WriteErrorStack(*(XXX_THISENV().ErrorStr),
					      XXX_THISERROR(), 
					      Message) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot write errors to client");
      return XXX_NOT_OK;
    }
  } else {
    if (ClientInfo->Protocol->WriteNoError(*(XXX_THISENV().ErrorStr)) 
	!= XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot write no errors to client");
      return XXX_NOT_OK;
    }
  }
  XXX_CLEAR_ERROR();
  return XXX_OK;
}

const char* XXX_NEW_SESSION_REQUEST = "NEW SESSION\n";

static XxxBool isNewSessionRequest(const char *request)
{
  if (!strcmp(request, XXX_NEW_SESSION_REQUEST))
    return XXX_TRUE;
  return XXX_FALSE;
}

void XxxRequestThread::ExecCode()
{

  /* mmaasha no java jvm
  // create the JVM
  if (XXX_THISENV().StartJVM() != XXX_OK) {
    cerr << "Error creating JVM" << endl;
  }
  */
  // get the current settings
  XXX_THISENV().DBEnv.GetCurrentValues();

  XxxDefaultEngine* DefEng = (XxxDefaultEngine*)XxxDBEngine::DefaultDBEngine();
  // Get a line to determine the connection type
  char requestLine[MAX_REQUEST];

#if defined(XXX_WIN32) && defined(XXX_NONPREEMPTIVE)
  if (eventHandle->is_down() || (bool)eventHandle->wait(::WAIT_FOREVER)) 
#else
  if (fileHandle->is_down() || (bool)fileHandle->wait(::WAIT_FOREVER)) 
#endif
  {
      XXX_NOTE_ERROR(XXX_BADEXEC, "Error waiting for connection request");
  }
  

  if (socket->ReadLine(requestLine, MAX_REQUEST) < 0) {
    cerr << "Error getting connection request." << endl;
    delete socket;
    socket = NULL;
    if (DefEng->MarkThreadAsFree(this) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADEXEC, "Error marking request thread as free");
    }
    goto exit_req;
  }
  // Now process the request
  if (isNewSessionRequest((const char *)requestLine)) {
    if (runClientSession() != XXX_OK) {
      XXX_PRINT_ERROR(cerr, "An error occurred during a client session");
      XXX_CLEAR_ERROR();
    }
  } else if (XxxHTTPRequest::IsHTTPRequest((const char *)requestLine)) {
    if (serviceHTTPRequest(requestLine) != XXX_OK) {
      XXX_PRINT_ERROR(cerr, 
		      "Received an HTTP request which cannot be serviced.");
      XXX_CLEAR_ERROR();
    }
  } else {
    cerr << "Connection failed: unrecognized protocol." << endl;
  }
  if (DefEng->MarkThreadAsFree(this) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADEXEC, "Error marking request thread as free");
  }

  /* mmaasha
  // close the JVM
  if (XXX_THISENV().ShutJVM() != XXX_OK) {
    cerr << "Error closing JVM" << endl;
  }
  */
exit_req:
  return;
}

XxxRequestThread::~XxxRequestThread()
{ 
  if (XXX_THISENV().DebugFlags.Connection) {
    cerr << "Deleting " << ((ClientInfo != NULL) ? "client" : "callback") 
	 << " thread...";
  }
  if (ClientInfo != NULL) {
    // This is a client thread
    delete ClientInfo;
    if (XxxEnv.OutputStr && (XxxEnv.OutputStr == XxxEnv.ErrorStr))
      delete XxxEnv.OutputStr;
    else {
      if (XxxEnv.OutputStr) delete XxxEnv.OutputStr;
      if (XxxEnv.ErrorStr) delete XxxEnv.ErrorStr;
    }
  } else {
    // This is a callback thread
    if (XxxEnv.OutputStr) delete XxxEnv.OutputStr;
  }
  if (socket) delete socket;
  if (RequestBuf) delete [] RequestBuf;
  if (XXX_THISENV().DebugFlags.Connection) {
    cerr << "done." << endl;
  }
}

void XxxRequestThread::cleanup(XxxBool CleanEnvs, 
			       XxxBool CleanStorage,
			       XxxBool CleanProtocol, 
			       XxxBool CleanSocket)
{
  if (XXX_THISENV().DebugFlags.Connection) {
    cerr << "Deleting connection handle... ";
  }

 // if there is still a multi-stmt xact running, abort it
 if (XXX_THISENV().MultiStmtXact)
   XxxAbortTrans();

 if (CleanEnvs && (XxxEnv.DeleteEnvs() != XXX_OK)) {
   XXX_PRINT_ERROR(cerr, "Request:Error deleting envs");
   XXX_CLEAR_ERROR();
 }

 if (CleanStorage && (XxxEnv.DumpStorage() != XXX_OK)) {
   XXX_PRINT_ERROR(cerr, "Request:Error dumping storage");
   XXX_CLEAR_ERROR();
 }

 if (CleanProtocol && 
     (ClientInfo->Protocol->CloseConnection() != XXX_OK)) {
   XXX_PRINT_ERROR(cerr, "Error closing connection using protocol");
   XXX_CLEAR_ERROR();
 }

 // close the socket to the client
 if (CleanSocket && socket) {
   delete socket;
   socket = NULL;
 }
 if (XXX_THISENV().DebugFlags.Connection) {
   cerr << "done" << endl;
 }
}

void XxxRequestThread::getInput(XxxBool& Valid, XxxBool& ToBreak)
{
 Valid = XXX_FALSE;
 ToBreak = XXX_FALSE;


#if defined(XXX_WIN32) && defined(XXX_NONPREEMPTIVE)
 HANDLE handle = CreateEvent(0, TRUE, FALSE, 0);
 int rep = WSAEventSelect((unsigned int) SocketDesc(), handle, FD_READ);

 eventHandle = new win32_compat_event_t(handle);

 if (eventHandle->is_down() || (bool)eventHandle->wait(::WAIT_FOREVER)) 
#else
 if (fileHandle->is_down() || (bool)fileHandle->wait(::WAIT_FOREVER)) 
#endif
 {
   ToBreak = XXX_TRUE;
   return;
  }

 unsigned long BytesToRead = 0;
 // First get the number of bytes to read. There could be an error.
 if (::recv(socket->socketDesc, (char *)&BytesToRead, 
	    sizeof(unsigned long), 0) < sizeof(unsigned long)) {
   Valid = XXX_FALSE;
   return;
  }
 BytesToRead = ntohl(BytesToRead);

 // the client could have sent EOF
 if (BytesToRead == 0) {
   if (XXX_THISENV().DebugFlags.Connection) {
     cerr << "Client EOF " << endl;
   }
   Valid = XXX_FALSE;
   return;
 }

 // check that BytesToRead < RequestBufSize
 if ((int)BytesToRead > RequestBufSize) {
   cerr << "Client command too long!" << endl;
   Valid = XXX_FALSE;
   return;
  }

 unsigned int BytesRead = 0;
 while (BytesRead < BytesToRead) {
   unsigned int NewBytesRead = 0;

#if defined(XXX_WIN32) && defined(XXX_NONPREEMPTIVE)
	handle = CreateEvent(0, TRUE, FALSE, 0);
	int rep = WSAEventSelect((unsigned int) SocketDesc(), handle, FD_READ);

	eventHandle = new win32_compat_event_t(handle);

	if (eventHandle->is_down() || (bool)eventHandle->wait(::WAIT_FOREVER)) 
#else
	if (fileHandle->is_down() || (bool)fileHandle->wait(::WAIT_FOREVER)) 
#endif
	{
		 ToBreak = XXX_TRUE;
		return;
	}

   NewBytesRead = ::recv(socket->socketDesc, RequestBuf+BytesRead, 
			 (BytesToRead - BytesRead), 0);
   // There could be an error, 
   if (NewBytesRead < 1) {
     cerr << "Error reading bytes" << endl;
     Valid = XXX_FALSE;
     return;
    }
   BytesRead += NewBytesRead;

  } // while

 RequestBuf[BytesRead] = 0; // null-terminate, in case
 
 Valid = XXX_TRUE;
}

//Begin: Moustafa
XxxBool XxxRequestThread::IsQueryRequestStop(){
  XxxBool Valid = XXX_FALSE;
  XxxBool ToBreak = XXX_TRUE;
  // get input from client
  if(fileHandle->is_down()|| ((bool)fileHandle->wait(WAIT_FOREVER)))
    return XXX_FALSE;

  getInput(Valid, ToBreak);
  if (strstr(RequestBuf, "stream_stop")!=NULL)
    return XXX_TRUE;
  else
    return XXX_FALSE;
  
}
//End: Moustafa
