#ifndef XXX_CLIENT_H
#define XXX_CLIENT_H

#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include "xxx_thread.h"
#include "engines.h"

XXX_EXTERN_CLASS_DECL(XxxRequestThread);
// all protocols ae defined in protocol.h
XXX_EXTERN_CLASS_DECL(XxxClientProtocol);
/**
 * one object of this class is maintained for each connected client
 */
class XxxClientInfo {
  friend ostream& operator<<(ostream&, XxxClientInfo&);
  friend class XxxClientProtocol;
 private:
  // mutex used to coordinate access to 
  static int CurrentClientId;
  static XxxBaseMutex latch;
 public:
  int UserId;
  char *UserName;
  char *Password;
  XxxRequestThread* ThisThread;
  double TotalClockTime;

  XxxClientProtocol* Protocol;

  // mutex used to coordinate access to the following
  XxxBaseMutex CBlatch;
  int NumCBThreads;
  XxxBool CallbacksPermitted;
  XxxCondVar NoCBThreads;
  XxxCondVar CBThreadDone;

  XxxClientInfo(XxxRequestThread* Thread);
  ~XxxClientInfo();
  XxxRequestThread* Thread() { return ThisThread;}
  XxxErrCode StringSendReceive(const char* SendString,
			       char RecvString[],
			       int RecvStringSize);
  // set up connection
  XxxErrCode EstablishProtocol(int Socket);
  // wait for read input on thread
  inline int WaitForInput();
  
  XxxErrCode AddCallback();
  XxxErrCode RemoveCallback();
  XxxErrCode WaitForCallbacks();
  XxxErrCode PermitCallbacks();
 };

/**
 * One such thread operates for each client
 */
class XxxRequestThread : public XxxBaseThread {
  friend class XxxClientInfo;
 protected:
  void cleanup(XxxBool CleanEnvs, XxxBool CleanStorage,
	       XxxBool CleanProtocol, XxxBool CleanSocket);
  void getInput(XxxBool& Valid, XxxBool& ToBreak);
  /**
   * If there is a pending server error, this function writes it out
   * on the client's error stream using the appropriate protocol.
   */
  XxxErrCode processErrors(const char *Message);
  XxxErrCode runClientSession();
  XxxErrCode serviceHTTPRequest(const char *RequestLine);

  virtual void ExecCode();
  XxxSocketConnection* socket;
 public:
  static const unsigned int DefaultRequestMaxSize;
  //Begin: Moustafa
  // to stop indefinite query (for stream)
  unsigned char StopQuery;
  XxxBool IsQueryRequestStop();
  //End: Moustafa

  char* RequestBuf;
  int RequestBufSize;

  XxxClientInfo* ClientInfo;
  int SocketDesc() const { return (socket?socket->socketDesc:0);}
  /**
   * Associates this thread with the given transaction.
   * @param Xact The transaction currently being processed by the 
   *             XxxRequestThread with which this thread is associated
   */
  XxxErrCode AttachXact (XxxShoreXact *Xact);
  /**
   * Deassociates this thread with the associated transaction.
   * @param Xact The transaction currently being processed by the 
   *             XxxRequestThread with which this thread is associated
   */
  XxxErrCode DetachXact (XxxShoreXact *Xact);
  XxxRequestThread(XxxSocketConnection*, 
		   const XxxEnvironment& DefaultEnv,
		   unsigned int ReqBufSize = 0);
  virtual ~XxxRequestThread();
 };

// Get the current client info (only valid for a XxxRequestThread)
inline XxxClientInfo* XXX_CLIENTINFO() \
  { return ((XxxRequestThread*)(XXX_THISTHREAD()))->ClientInfo;}

inline int XxxClientInfo::WaitForInput()
{
#if defined(XXX_WIN32) && defined(XXX_NONPREEMPTIVE)
 HANDLE handle = CreateEvent(0, TRUE, FALSE, 0);
 int rep = WSAEventSelect((unsigned int) ((XxxRequestThread*) XXX_THISTHREAD())->SocketDesc(), handle, FD_READ);

 win32_compat_event_t* eventHandle = new win32_compat_event_t(handle);
	
 return (eventHandle->is_down() || 
		(bool)eventHandle->wait(::WAIT_FOREVER));
#else
	return (ThisThread->fileHandle->is_down() || 
		(bool)ThisThread->fileHandle->wait(::WAIT_FOREVER));
#endif
}

#endif

