    #include <stdio.h>
    #include <string.h>
    #include <iostream.h>
    #include "templates.h"
    #include "values.h"
    #include "engines.h"
    #include "defeng.h"
    #include "shoresm.h"
    #include "error.h"
    #include "JavaUDF.h"
    
    /// error tables
    const int XxxDefaultEngine::NumErrorConds = 29;
    int XxxDefaultEngine::DefEngId = 0;
    const char *const XxxDefaultEngine::ErrorTable[] = {
      "No Error",              // 0 XXX_ALLWELL
      "Incomplete System Command", // 1 XXX_INCOMPL
      "XXX Internal Error",    // 2 XXX_INTERNAL
      "Bad Command Line Flag", // 3 XXX_BADFLAG
      "Bad File",              // 4 XXX_BADFILE
      "Bad Pipe",              // 5 XXX_BADPIPE
      "Bad Socket",            // 6 XXX_BADSOCK
      "Bad Registry",          // 7 XXX_BADREG
      "Table Overflow",        // 8 XXX_OVFLO
      "Duplicate Set Entry",   // 9 XXX_SETDUPL
      "Bad Server",            // 10 XXX_BADSERVER
      "Bad Engine",            // 11 XXX_BADENG
      "Error from SHORE",      // 12 XXX_SHOREERR
      "Error in Parsing",      // 13 XXX_BADPARSE
      "Error in Optimization", // 14 XXX_BADOPT
      "Error in Execution",    // 15 XXX_BADEXEC
      "Bad Command",           // 16 XXX_BADCMD
      "Bad Query",             // 17 XXX_BADQUERY
      "Bad Expression",        // 18 XXX_BADEXPR
      "Bad Value",             // 19 XXX_BADVALUE
      "Bad Function Call",     // 20 XXX_BADFUNC
      "Bad Record",            // 21 XXX_BADRECORD
      "Bad Type",              // 22 XXX_BADTYPE
      "Bad ADT",               // 23 XXX_BADADT
      "Invalid Cursor",        // 24 XXX_BADCURSOR
      "Bad Scan",              // 25 XXX_BADSCAN
      "Bad Handle",            // 26 XXX_BADHANDLE
      "Bad Device",            // 27 XXX_BADDEVICE
      "Error in JavaVM"        // 28 XXX_BADJVM
       };
    
    const char* XXX_DEF_SETADTENV = "set_adtenv";
    const char* XXX_DEF_LISTADTENV = "list_adtenv";
    
    //
    // Methods of XxxDefaultEngine
    //
    // XxxDefaultEngine constructor
    XxxDefaultEngine::XxxDefaultEngine(const int Id) : 
      XxxDBEngine(Id, "__default", 1),
      ProtocolTable(10) // max number of protocols : move to a constant
    {
      DefEngId=Id;
      strcpy(LanguageNames[0], "XXX");
      ProtocolTable.AddItem(new XxxJavaClientProtocol::JProtoDefn());
      ProtocolTable.AddItem(new XxxTxtClientProtocol::TProtoDefn());
      XxxJaguarExtendProtocolTable(ProtocolTable);
#ifdef XXX_COMPRESS
      // adding compression protocol
      ProtocolTable.AddItem(new XxxCompressProtocol::CProtoDefn());
#endif
     }
    
    static int deleteProtocol(XxxClientProtocol::ProtoDefn*& TableEntry, void *arg)
    {
      delete TableEntry;
      return 1;
    }
    // XxxDefaultEngine destructor
    XxxDefaultEngine::~XxxDefaultEngine()
    {
      ProtocolTable.ApplyAll(deleteProtocol, 0);
    }
     
    const char* XxxDefaultEngine::ErrorString(XxxErrCond Condition) const
    {
     if ((Condition >= NumErrorConds) || (Condition < 0))
       return "???";
     else return ErrorTable[Condition];
    }
    
    //
    // Commands for the default engine
    //
    
#ifdef PROF_PAPI
    const int XxxDefaultEngine::maxCommands = 23;
#else
    const int XxxDefaultEngine::maxCommands = 21;
#endif

    struct XxxDefaultEngine::commandInfo XxxDefaultEngine::commandArray[XxxDefaultEngine::maxCommands] = {
      {"adts", &XxxDefaultEngine::execListAdts, XXX_TRUE, XXX_TRUE},
      {"engines", &XxxDefaultEngine::execListEngines, XXX_TRUE, XXX_TRUE},
      {"stats_eng", &XxxDefaultEngine::execStatsEng, XXX_TRUE, XXX_FALSE},
      {"stats_adt", &XxxDefaultEngine::execStatsAdt, XXX_TRUE, XXX_FALSE},
      {"stats_gen", &XxxDefaultEngine::execStatsGen, XXX_TRUE, XXX_FALSE},
      {"stats_reset", &XxxDefaultEngine::execStatsReset, XXX_TRUE, XXX_FALSE},
      {"help", &XxxDefaultEngine::execHelp, XXX_TRUE, XXX_TRUE},
      {XXX_DEF_SETADTENV, &XxxDefaultEngine::execSetAdtEnv, XXX_TRUE, XXX_FALSE},
      {XXX_DEF_LISTADTENV, &XxxDefaultEngine::execListAdtEnv, XXX_TRUE, XXX_FALSE},
      {"quiet", &XxxDefaultEngine::execQuiet, XXX_FALSE, XXX_TRUE},
      {"verbose", &XxxDefaultEngine::execVerbose, XXX_FALSE, XXX_TRUE},
      {"user", &XxxDefaultEngine::execUser, XXX_FALSE, XXX_TRUE},
      {"clients", &XxxDefaultEngine::execListClients, XXX_FALSE, XXX_TRUE},
      {"sys_disk", &XxxDefaultEngine::execSysDisk, XXX_FALSE, XXX_TRUE},
      {"sys_shore", &XxxDefaultEngine::execSysShore, XXX_FALSE, XXX_TRUE},
      {"sys_stats", &XxxDefaultEngine::execSysStats, XXX_FALSE, XXX_TRUE},
      {"sys_reset", &XxxDefaultEngine::execSysReset, XXX_FALSE, XXX_TRUE},
      {"set", &XxxDefaultEngine::execSet, XXX_FALSE, XXX_TRUE},
      {"clear", &XxxDefaultEngine::execClear, XXX_FALSE, XXX_TRUE},
      {"extend", &XxxDefaultEngine::execExtend, XXX_TRUE, XXX_TRUE},
      //Begin: Moustafa
      // this command is to stop a client query.
      // useful for streaming input
      {"stream_stop", &XxxDefaultEngine::execStreamStop, XXX_TRUE, XXX_TRUE},
      //End: Moustafa
#ifdef PROF_PAPI
      {"papi_start", &XxxDefaultEngine::execPapiStart, XXX_FALSE, XXX_TRUE},
      {"papi_stop", &XxxDefaultEngine::execPapiStop, XXX_FALSE, XXX_TRUE},
#endif
      {NULL, NULL, XXX_FALSE, XXX_FALSE}
    };
    
    XxxErrCode XxxDefaultEngine::matchCommand(const char* Command,
    					  ExecFunc& EF,
    					  XxxBool& ClientAllowed,
    					  XxxBool& MonitorAllowed,
    					  const char* &CommandSuffix)
    {
      for (int i = 0; i < maxCommands; i++) {
        if (commandArray[i].Name) {
          if (!strncmp(Command, commandArray[i].Name, strlen(commandArray[i].Name))) {
    	EF = commandArray[i].Func;
    	ClientAllowed = commandArray[i].Client;
    	MonitorAllowed = commandArray[i].Monitor;
    	CommandSuffix = Command + strlen(commandArray[i].Name) + 1;
    	return XXX_OK;
          } // if
        } // if
      } // for
      return XXX_NOT_OK;
    }
    
    XxxErrCode XxxDefaultEngine::execListAdts(const char* Params)
    {
      XxxUseMutex AutoMutex(*XxxADTClass::Latch);
      *(XXX_THISENV().OutputStr) << "Available ADTs" << endl;
      for (int i = 0; i < XxxADTClass::NumADTs; i++) {
        *(XXX_THISENV().OutputStr) << "\t";
        XxxADTClass::ADTArray[i]->Print(*(XXX_THISENV().OutputStr), NULL);
        *(XXX_THISENV().OutputStr) << endl;
      }
      return XXX_OK;
    }
    
    XxxErrCode XxxDefaultEngine::execListEngines(const char* Params)
    {
      XxxUseMutex AutoMutex(*XxxDBEngine::Latch);
      *(XXX_THISENV().OutputStr) << "Available Engines" << endl;
      for (int i = 0; i < XxxDBEngine::NumDBEngines; i++)
        XxxDBEngine::DBEngineArray[i]->Print(*(XXX_THISENV().OutputStr));
      return XXX_OK;
    }
    
    XxxErrCode XxxDefaultEngine::execStatsEng(const char* Params)
    {
      XxxUseMutex AutoMutex(*XxxDBEngine::Latch);
      *(XXX_THISENV().OutputStr) << "Engine Statistics" << endl;
      for (int i = 0; i < XxxDBEngine::NumDBEngines; i++)
        XxxDBEngine::DBEngineArray[i]->PrintStats(*(XXX_THISENV().OutputStr));
      return XXX_OK;
    }
    
    XxxErrCode XxxDefaultEngine::execStatsAdt(const char* Params)
    {
      XxxUseMutex AutoMutex(*XxxADTClass::Latch);
      *(XXX_THISENV().OutputStr) << "ADT Statistics" << endl;
      for (int i = 0; i < XxxADTClass::NumADTs; i++)
        XxxADTClass::ADTArray[i]->PrintStats(*(XXX_THISENV().OutputStr));
      return XXX_OK;
    }
    
    XxxErrCode XxxDefaultEngine::execStatsGen(const char* Params)
    {
      XXX_THISENV().PrintStats(*(XXX_THISENV().OutputStr));
      return XXX_OK;
    }
    
    XxxErrCode XxxDefaultEngine::execStatsReset(const char* Params)
    {
      *(XXX_THISENV().OutputStr) << "Resetting all statistics ...";
      *(XXX_THISENV().OutputStr) << " general ...";
      XXX_THISENV().ResetStats();
      *(XXX_THISENV().OutputStr) << " adts ...";
      int i; 
      {
        XxxUseMutex AutoMutex(*XxxADTClass::Latch);
        for (i = 0; i < XxxADTClass::NumADTs; i++)
          XxxADTClass::ADTArray[i]->ResetStats();
      }
      {
        XxxUseMutex AutoMutex(*XxxDBEngine::Latch);
        for (i = 0; i < XxxDBEngine::NumDBEngines; i++)
          XxxDBEngine::DBEngineArray[i]->ResetStats();
      }
      *(XXX_THISENV().OutputStr) << " engines ...";
      *(XXX_THISENV().OutputStr) << " done" << endl;
      return XXX_OK;
    }
    
    XxxErrCode XxxDefaultEngine::execHelp(const char* Params)
    {
      *(XXX_THISENV().OutputStr) << "Common monitor/client commands are:" 
    			     << endl;
      *(XXX_THISENV().OutputStr) << "\tadts       : list available ADTs" << endl;
      *(XXX_THISENV().OutputStr) << "\tengines    : list available engines" 
    			     << endl;
      *(XXX_THISENV().OutputStr) << "\thelp       : help on commands" << endl;
      *(XXX_THISENV().OutputStr) << endl;
        
      *(XXX_THISENV().OutputStr) << "Exclusive client commands are:" << endl;
      *(XXX_THISENV().OutputStr) << "\tstats_gen  : client general stats" 
    			     << endl;
      *(XXX_THISENV().OutputStr) << "\tstats_adt  : client adt stats" << endl;
      *(XXX_THISENV().OutputStr) << "\tstats_eng  : client engine stats" << endl;
      *(XXX_THISENV().OutputStr) << "\tstats_reset: reset client stats" << endl;
      *(XXX_THISENV().OutputStr) << "set_adtenv [ADTName] [VarName] [NewValue]"
        " : set variable VarName of type ADTName\n\t\tto value NewValue" << endl;
      *(XXX_THISENV().OutputStr) << "list_adtenv [ADTName] : list all "
        "environment variable names and values for type\n\t\tADTName" << endl;
      *(XXX_THISENV().OutputStr) << endl;
        
      *(XXX_THISENV().OutputStr) << "Exclusive monitor commands are:" << endl;
      *(XXX_THISENV().OutputStr) << "\tclients    : list current clients" 
    			     << endl;
      *(XXX_THISENV().OutputStr) << "\textend  : add a new Java method :" << endl
         << "\t  extend methodname Java-name subclasser URL result owner arity arguments" << endl;
      *(XXX_THISENV().OutputStr) << "\tshutdown   : terminate server" << endl;
      *(XXX_THISENV().OutputStr) << "\tquiet      : make the server quiet" 
    			     << endl;
      *(XXX_THISENV().OutputStr) << "\tsys_stats  : print server statistics" 
    			     << endl;
      *(XXX_THISENV().OutputStr) << "\tsys_disk   : show disk volume stats" 
    			     << endl;
      *(XXX_THISENV().OutputStr) << "\tsys_shore  : show SHORE stats" << endl;
      *(XXX_THISENV().OutputStr) << "\tsys_reset  : reset all stats" << endl;
      *(XXX_THISENV().OutputStr) << "\tuser [a|d|l] : add|delete|list users" 
    			     << endl;
      *(XXX_THISENV().OutputStr) << "\tverbose  : make the server verbose" 
    			     << endl;
      *(XXX_THISENV().OutputStr) << endl;
      //Begin: Moustafa
      *(XXX_THISENV().OutputStr) << "\tstream_stop <client number> : stops query at client" 
				 << endl;
      //End: Moustafa

#ifdef PROF_PAPI
      *(XXX_THISENV().OutputStr) << "PAPI monitor commands are:" << endl;
      *(XXX_THISENV().OutputStr) << "\tpapi_start <number_evts> evt1 [evt2 evt3 ...] : start counting hardware events using papi" << endl;
      *(XXX_THISENV().OutputStr) << "\tpapi_stop  : stop counting hardware events" << endl;
#endif
      return XXX_OK;
    }
    
    XxxErrCode XxxDefaultEngine::execSetAdtEnv(const char* Params)
    {
      // find out which ADT's environment requires modification
      istrstream ReqStr((char*)Params);
      char AdtName[XXX_ADT_NAME_MAX];
      char AdtArg[XXX_ATTR_NAME_MAX];
      char NewVal[10];
      // set_adtenv <adtname> <variable-name> <value>
      ReqStr >> AdtName >> AdtArg >> NewVal;
      if (ReqStr.bad()) {
        XXX_NOTE_ERROR(XXX_BADCMD, "set_adtenv");
        return XXX_NOT_OK;
      }
      int ind=0;
      {
        XxxUseMutex AutoMutex(*XxxADTClass::Latch);
        for (; ind < XxxADTClass::NumADTs; ind++)
          if (!strcmp(AdtName, XxxADTClass::ADTArray[ind]->TypeName))
    	break;
        // ind contains the index of the desired ADT
        if (ind==XxxADTClass::NumADTs) {
          // adtname not found
          char buf[200];
          sprintf(buf,"Unknown ADT Type: %s in %s command.", AdtName, "set_adtenv");
          XXX_NOTE_ERROR(XXX_BADCMD, buf);
          return XXX_NOT_OK;
        }
      }
      return XXX_THISADTENV(ind)->SetValue(AdtArg,NewVal);
    }
    
    XxxErrCode XxxDefaultEngine::execListAdtEnv(const char* Params)
    {
      // find out which ADT's environment is to be viewed.
      int ind=0;
      const char *AdtName = Params;
      {
        XxxUseMutex AutoMutex(*XxxADTClass::Latch);
        for (; ind < XxxADTClass::NumADTs; ind++) {
          if (!strcmp(AdtName, XxxADTClass::ADTArray[ind]->TypeName))
    	break;
        }
        // ind contains the index of the desired ADT
        if (ind==XxxADTClass::NumADTs) {
          // adtname not found
          char buf[200];
          sprintf(buf,"Unknown ADT Type: %s in %s command",AdtName, "list_adtenv");
          XXX_NOTE_ERROR(XXX_BADCMD, buf);
          return XXX_NOT_OK;
        }
      }
      XXX_THISADTENV(ind)->Print(*(XXX_THISENV().OutputStr));
      return XXX_OK;
    }
    
    XxxErrCode XxxDefaultEngine::execQuiet(const char* Params)
    {
      servState.SetQuiet(XXX_TRUE);
      return XXX_OK;
    }
    
    XxxErrCode XxxDefaultEngine::execVerbose(const char* Params)
    {
      servState.SetQuiet(XXX_FALSE);
      return XXX_OK;
    }
    
    XxxErrCode XxxDefaultEngine::execUser(const char* Params)
    {
      return processUserService(Params);
    }
    
    XxxErrCode XxxDefaultEngine::execListClients(const char* Params)
    {
      return servState.ListClients(*(XXX_THISENV().OutputStr));
    }
//Begin: Moustafa
XxxErrCode XxxDefaultEngine::execStreamStop(const char* Params)
{
  // must first get the client information and set the stopquery falg to 1 
  istrstream ReqStr((char*)Params);
  int ClientId;
  ReqStr >> ClientId;
  XxxClientInfo* ClientInfo;
  servState.GetClient(ClientId, ClientInfo);
  if(ClientInfo->ThisThread!=NULL)
    ((XxxRequestThread* )(ClientInfo->ThisThread))->StopQuery = 1;
  return XXX_OK;
}
//End: Moustafa

#ifdef PROF_PAPI
    XxxErrCode XxxDefaultEngine::execPapiStart(const char* Params)
    {
      *(XXX_THISENV().OutputStr) << "\tStarting PAPI" << endl;
      PapiProfiler.PapiStart(Params);
      return XXX_OK;
    }

    XxxErrCode XxxDefaultEngine::execPapiStop(const char* Params)
    {
      *(XXX_THISENV().OutputStr) << "\tStopping PAPI" << endl;
      PapiProfiler.PapiStop();
      return XXX_OK;
    }
#endif

    XxxErrCode XxxDefaultEngine::execSysDisk(const char* Params)
    {
      /*******
        if (servState.NumClients() > 0) {
        // we have occasionally had problems with SHORE crashing
        *(XXX_THISENV().OutputStr) << "There must be no connected clients for this command to run" << endl;
        return XXX_OK;
        }
        ******/
      if (XxxShoreSM)
        return XxxShoreSM->DiskUsage(*(XXX_THISENV().OutputStr));
      else {
        *(XXX_THISENV().OutputStr) << "SHORE not enabled: no disk stats "
          "available\n" << endl;
        return XXX_OK;
      }     
    }
    
    XxxErrCode XxxDefaultEngine::execSysShore(const char* Params)
    {
      if (XxxShoreSM)
        return XxxShoreSM->PrintStats(*(XXX_THISENV().OutputStr));
      else {
        *(XXX_THISENV().OutputStr) << "SHORE not enabled: no SHORE stats "
          "available\n" << endl;
        return XXX_OK;
      }     
    }
    
    XxxErrCode XxxDefaultEngine::execSysStats(const char* Params)
    {
      *(XXX_THISENV().OutputStr) << "General server stats not available\n" 
    			     << endl;
      return XXX_OK;
    }
    
    XxxErrCode XxxDefaultEngine::execSysReset(const char* Params)
    {
      *(XXX_THISENV().OutputStr) << "Resetting all statistics ...";
      *(XXX_THISENV().OutputStr) << " General server stats ..." ;
      if (XxxShoreSM) {
        *(XXX_THISENV().OutputStr) << " SHORE ..." ;
        XxxShoreSM->ResetStats();
      }
      *(XXX_THISENV().OutputStr) << " done" << endl;
      return XXX_OK;
    }
    
    XxxErrCode XxxDefaultEngine::execSet(const char* Params)
    {
      istrstream ReqStr((char*)Params);
      char FieldName[XXX_ADT_NAME_MAX];
      // set_adtenv <adtname> <variable-name> <value>
      ReqStr >> FieldName;
      if (ReqStr.bad()) {
        XXX_NOTE_ERROR(XXX_BADCMD, "defeng::set");
        return XXX_NOT_OK;
      }
     if (XXX_THISENV().SetValue(FieldName, 1) != XXX_OK) {
       XXX_NOTE_ERROR(XXX_BADCMD, "Bad Set command");
       return XXX_NOT_OK;
     }
     return XXX_OK;
    }
    
    XxxErrCode XxxDefaultEngine::execClear(const char* Params)
    {
      istrstream ReqStr((char*)Params);
      char FieldName[XXX_ADT_NAME_MAX];
      // set_adtenv <adtname> <variable-name> <value>
      ReqStr >> FieldName;
      if (ReqStr.bad()) {
        XXX_NOTE_ERROR(XXX_BADCMD, "defeng::set");
        return XXX_NOT_OK;
      }
     if (XXX_THISENV().SetValue(FieldName, 0) != XXX_OK) {
       XXX_NOTE_ERROR(XXX_BADCMD, "Bad Set command");
       return XXX_NOT_OK;
     }
     return XXX_OK;
    }
    
    XxxErrCode XxxDefaultEngine::execExtend(const char* Params)
    {
      // find out which ADT's environment requires modification
      istrstream ReqStr((char*)Params);
      char OwnerADTName[XXX_ADT_NAME_MAX];
      XxxBasicType OwnerType;
      char SMethName[50];
      char SCName[50];
      char JMethName[50];
      char url[500];
      char ResultADTName[XXX_ADT_NAME_MAX];
      char ArgCountString[10];
      XxxBasicType ResultType;
      XxxFuncMethodInfo::ArgInfo Signature;
      ReqStr >> SMethName >> JMethName >> SCName >> url 
    	  >> ResultADTName >> OwnerADTName >> ArgCountString;
      if (ReqStr.bad()) {
        XXX_NOTE_ERROR(XXX_BADCMD, "extend");
        return XXX_NOT_OK;
      }
      Signature.Count = atoi(ArgCountString);
      char **ArgADTNames = new char*[Signature.Count];
      // convenient class to reclaim this memory
      class ArgADTNameReclaimer { 
        char **names;
        int count;
      public:
        ArgADTNameReclaimer(char **Names, int Count) 
        { names = Names; count = Count;}
        ~ArgADTNameReclaimer() { 
          if (names && count) {
    	for (int i=0; i<count; i++)
    	  if (names[i]) delete [] names[i];
    	delete [] names;
          }
        }
      };
      ArgADTNameReclaimer Foo(ArgADTNames, Signature.Count);
    
      for (int i=0; i<Signature.Count; i++) {
        ArgADTNames[i] = new char[XXX_ADT_NAME_MAX];
        ReqStr >> ArgADTNames[i];
      }
      if (ReqStr.bad()) {
        XXX_NOTE_ERROR(XXX_BADCMD, "extend");
        return XXX_NOT_OK;
      }
      {
        XxxUseMutex AutoMutex(*XxxADTClass::Latch);
	//Bjarke, aug. 3.. decl of ind moved outside for stmt.
	int ind;
        for (ind=0; ind < XxxADTClass::NumADTs; ind++) 
          if (!strcmp(OwnerADTName, XxxADTClass::ADTArray[ind]->TypeName))
	    break;
        if (ind != XxxADTClass::NumADTs)
          OwnerType = ind;
        else {
          XXX_NOTE_ERROR(XXX_BADCMD, "extend: bad ADT name");
          return XXX_NOT_OK;
        }
        for (ind=0; ind < XxxADTClass::NumADTs; ind++) 
          if (!strcmp(ResultADTName, XxxADTClass::ADTArray[ind]->TypeName))
	    break;
        if (ind != XxxADTClass::NumADTs)
          ResultType = ind;
        else {
          XXX_NOTE_ERROR(XXX_BADCMD, "extend: bad ADT name");
          return XXX_NOT_OK;
        }
        for (int j=0;j<Signature.Count;j++) {
          for (ind=0; ind < XxxADTClass::NumADTs; ind++) 
	    if (!strcmp(ArgADTNames[j], XxxADTClass::ADTArray[ind]->TypeName))
	      break;
          if(ind != XxxADTClass::NumADTs)
	    Signature.ArgTypes[j] = ind;
          else {
	    XXX_NOTE_ERROR(XXX_BADCMD, "extend: bad ADT name");
	    return XXX_NOT_OK;
          }
        } // for each argument
      }
      XxxJavaFuncMethodInfo *JFMI = 
        new XxxJavaFuncMethodInfo(SMethName,ResultType,0,
  	XxxJavaFuncMethodInfo::OptFunc,
    			      Signature,SCName, 
    			      JMethName, 
    			      new XxxJavaRemoteSourceSecurity(url));
      int Index;
      XXX_CALL_RET(XxxADTArray[OwnerType]->AddMethod(JFMI, Index),
    	       XXX_BADCMD, "extend");
    
      // *(XXX_THISENV().OutputStr) << "Added Method with index " << index << endl;
      return XXX_OK;
    }
    
    XxxErrCode 
    XxxDefaultEngine::ExprParse(char* ExprString,
    			    const XxxArgEnv*, XxxBasicType&, 
    			    XxxGenericParse*&) 
    { 
      if (strncasecmp(LanguageNames[0], ExprString, strlen(LanguageNames[0]))) { 
        XXX_NOTE_ERROR(XXX_BADCMD, "Unknown XXX Command"); 
        return XXX_NOT_OK; 
      } 
      char* Request = ExprString+strlen(LanguageNames[0])+1;
      // TODO: yikes! this needs to be changed
      // if the sender of the command is the server, there is a '\n' at the end 
      // of the command.  If not, then the command came from the client.
      XxxBool FromClient; 
      FromClient = (Request[strlen(Request)-1]=='\n') ? XXX_FALSE : XXX_TRUE;
      // remove ending \n if it's there 
      if (FromClient == XXX_FALSE) 
        Request[strlen(Request)-1] = 0; 
      // remove ending ';' if it's there 
      if (Request[strlen(Request)-1] == ';') 
        Request[strlen(Request)-1] = 0;
    
      XxxBool ClientAllowed;
      XxxBool MonitorAllowed;
      ExecFunc EF;
      //Bjarke, sep. 13 2001: added const qualifier to variable
      const char* ReqSuffix;
      if (matchCommand(Request, EF, 
    		   ClientAllowed, MonitorAllowed,
    		   ReqSuffix) != XXX_OK) {
        XXX_NOTE_ERROR(XXX_BADCMD, "Invalid command");
        return XXX_NOT_OK;
      }
      /*** TODO: commented out for now
      if (FromClient && !ClientAllowed) {
        XXX_NOTE_ERROR(XXX_BADCMD, "The command is restricted to the monitor");
        return XXX_NOT_OK;
      }
      ***/
      if (!FromClient && !MonitorAllowed) {
        XXX_NOTE_ERROR(XXX_BADCMD, "The command is restricted to the client");
        return XXX_NOT_OK;
      }
      return (this->*EF)(ReqSuffix);
    }
    

    XxxErrCode XxxDefaultEngine::processUserService(const char* Request)
    {
     XXX_NOTE_ERROR(XXX_INCOMPL, "User IDs");
     return XXX_NOT_OK;
    }
    
    //
    // Methods of XxxDefaultEngine::XxxServerState
    //
    
    XxxErrCode XxxDefaultEngine::
    ServerState::AddClient(XxxClientInfo* Client)
    {
      XxxUseMutex AutoMutex(latch);
    
      // delete any threads in free thread list
      int NumDeleted = 0;
      for (int i = free_thread_count-1; i >= 0; i--) {
        if (free_threads[i]) {
          delete free_threads[i];
          free_threads[i] = NULL;
          free_thread_count--;
          NumDeleted++;
        } else {
          XXX_NOTE_ERROR(XXX_BADSERVER, "Free thread appears to be NULL!\n");
          return XXX_NOT_OK;
        }
      }
      if (XXX_THISENV().DebugFlags.Server) {
        cerr << "Deleted " << NumDeleted << " request thread(s)." << endl;
      }
      if (count == XXX_CLIENTS_MAX-1) {
        XXX_NOTE_ERROR(XXX_BADSERVER, "Limit on number of clients allowed!\n");
        return XXX_NOT_OK;
      }
      clients[count++] = Client;
      return XXX_OK;
    }
    XxxErrCode XxxDefaultEngine::
    ServerState::MarkThreadAsFree(XxxRequestThread *Thread)
    {
      XxxUseMutex AutoMutex(latch);
      // this cannot exceed XXX_CLIENTS_MAX, so no need to check
      if (free_thread_count >= XXX_CLIENTS_MAX) {
        for (int i = 0; i < free_thread_count; i++) {
          delete free_threads[i];
          free_threads[i] = NULL;
        }
        free_thread_count = 0;
      }
      free_threads[free_thread_count++] = Thread;
      return XXX_OK;
    }
    
    XxxErrCode XxxDefaultEngine::
    ServerState::RemoveClient(XxxClientInfo* Client)
    {
     XxxUseMutex AutoMutex(latch);
    
     for (int i = 0; i < count; i++) {
       if (clients[i] == Client) {
         // the i'th entry must be removed
         for (int j = i; j < count; j++) {
           clients[j] = clients[j+1];
          }
         count--;
         return XXX_OK;
        }
      }
     return XXX_NOT_OK;
    }
    
    XxxErrCode XxxDefaultEngine::
    ServerState::ListClients(ostream& OStream)
    {
     XxxUseMutex AutoMutex(latch);
     OStream << "Active XXX Clients" << endl;
     for (int i = 0; i < count; i++)
       OStream << *(clients[i]);
     return XXX_OK;
    }
    
    XxxErrCode XxxDefaultEngine::
    ServerState::GetClient(int ClientId,
    		       XxxClientInfo*& ClientInfo)
    {
      XxxUseMutex AutoMutex(latch);
      for (int i = 0; i < count; i++)
        if (clients[i]->UserId == ClientId) {
          ClientInfo = clients[i];
          return XXX_OK;
        }
      XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot get client info for supplied client ID");
      return XXX_NOT_OK;
    }
    
    XxxDefaultEngine::ServerState::ServerState()
    {
     count = 0; 
     free_thread_count = 0; 
     quiet = XXX_TRUE;
    
     for (int i = 0; i < XXX_CLIENTS_MAX; i++) {
       clients[i] = NULL;
       free_threads[i] = NULL;
      }
    }
    
    XxxDefaultEngine::ServerState::~ServerState()
    { 
      // delete any free threads
      XxxUseMutex AutoMutex(latch);
      
      // delete any threads in free thread list
      int NumDeleted = 0;
      for (int i = free_thread_count-1; i >= 0; i--) {
        if (free_threads[i]) {
          delete free_threads[i];
          free_threads[i] = NULL;
          free_thread_count--;
          NumDeleted++;
        } else {
          XXX_NOTE_ERROR(XXX_BADSERVER, "Free thread appears to be NULL!\n");
        }
      }
      if (XXX_THISENV().DebugFlags.Server) {
        cerr << "Deleted " << NumDeleted << " request thread(s)." << endl;
      }
    }
    
