#ifndef XXX_DEFENG_H
#define XXX_DEFENG_H

#include "engines.h"
#include "client.h"
#include "protocol.h"
#include "xxx_thread.h"
#ifdef PROF_PAPI
#include "papi_prof.h"
#endif


// the following constants are used by the RelationEngine to hook into the
// set_adtenv and list_adtenv commands of the Default Engine.
extern const char* XXX_DEF_SETADTENV;
extern const char* XXX_DEF_LISTADTENV;

// Possible TODO:L2
// Structure XxxDefaultEngine's command parsing like RelationEngine.
// RelationEngine does much more type-checking, etc.

/**
 // Dummy DB engine: provides no query processing, but 
 //  1) handles generic error codes for the system. 
 //  2) has a simple command language 
 //  3) maintains the state of the system
 */
class XxxDefaultEngine : public XxxDBEngine {
  static const int NumErrorConds;
  static const char *const ErrorTable[];

  /**
  // An object of this class maintains the state of the XXX server!!
  // Every access to this had better be sandwiched by a Mutex acquire
  // and release. To ensure this, make all the public functions
  // acquire and release the mutex.
  */
  class ServerState {
    XxxBaseMutex latch;
    XxxClientInfo *clients[XXX_CLIENTS_MAX];
    XxxBaseThread *free_threads[XXX_CLIENTS_MAX];
    int count;
    int free_thread_count;
    XxxBool quiet;
   public:
    ServerState();
    ~ServerState();
    XxxErrCode AddClient(XxxClientInfo* Client);
    XxxErrCode RemoveClient(XxxClientInfo* Client);
    XxxErrCode MarkThreadAsFree(XxxRequestThread *Thread);
    XxxErrCode ListClients(ostream& OStream);
    XxxErrCode GetClient(int ClientId,
			 XxxClientInfo*& ClientInfo);
    int NumClients() { XxxUseMutex TmpM(latch); return count;}
    XxxBool BeQuiet() {XxxUseMutex TmpM(latch);  return quiet;}
    void SetQuiet(XxxBool Value) {
      XxxUseMutex TmpM(latch);
      quiet = Value;
     }
   };
  ServerState servState;
  XxxErrCode processUserService(const char* Request);

  typedef XxxErrCode (XxxDefaultEngine::*ExecFunc)(const char*);
  XxxErrCode execListAdts(const char*);
  XxxErrCode execListEngines(const char*);
  XxxErrCode execStatsEng(const char*);
  XxxErrCode execStatsAdt(const char*);
  XxxErrCode execStatsGen(const char*);
  XxxErrCode execStatsReset(const char*);
  XxxErrCode execHelp(const char*);
  XxxErrCode execSetAdtEnv(const char*);
  XxxErrCode execListAdtEnv(const char*);
  XxxErrCode execQuiet(const char*);
  XxxErrCode execVerbose(const char*);
  XxxErrCode execUser(const char*);
  XxxErrCode execListClients(const char*);
  XxxErrCode execSysDisk(const char*);
  XxxErrCode execSysShore(const char*);
  XxxErrCode execSysStats(const char*);
  XxxErrCode execSysReset(const char*);
  XxxErrCode execSet(const char*);
  XxxErrCode execClear(const char*);
  XxxErrCode execExtend(const char*);
  //Begin:Moustafa
  XxxErrCode execStreamStop(const char*);
  //End Moustafa
  XxxErrCode matchCommand(const char *Command, ExecFunc& EF,
			  XxxBool& ClientAllowed,
			  XxxBool& MonitorAllowed,
			  const char*& CommandSuffix);

#ifdef PROF_PAPI
  XxxErrCode execPapiStart(const char*);
  XxxErrCode execPapiStop(const char*);
  PapiProf<2> PapiProfiler;
#endif

  static const int maxCommands;
  struct commandInfo {
    const char* Name;
    ExecFunc Func;
    XxxBool Client;
    XxxBool Monitor;
  };
  static struct commandInfo commandArray[];

 public:
  static int DefEngId; // type id accessible without object.

  // table of known client protocols
  XxxPtrTable<XxxClientProtocol::ProtoDefn*> ProtocolTable;

  XxxDefaultEngine(const int Id);
  virtual ~XxxDefaultEngine();
  virtual const char* ErrorString(XxxErrCond Condition) const;
  
  virtual XxxErrCode ExprParse(char* ExprString,
			       //int NumParams,
			       //XxxValueExprList *ParamList,
			       const XxxArgEnv* ArgEnv,
			       XxxBasicType& ReturnType,
			       XxxGenericParse*& ParseStructure
			       );
  virtual XxxErrCode ExprOptimize(XxxGenericParse* ParseStructure,
				  XxxGenericPlan*& PlanStructure) 
   { return XXX_NOT_OK; }

  virtual XxxErrCode ExprEvaluate(XxxGenericPlan* PlanStructure,
				  const XxxValueEnv* ArgEnv,
				  XxxADTValue *ReturnValue) 
   { return XXX_NOT_OK; }


  XxxErrCode AddClient(XxxClientInfo* Client) {
    return servState.AddClient(Client);
   }
  XxxErrCode RemoveClient(XxxClientInfo* Client) {
    return servState.RemoveClient(Client);
   }
  XxxErrCode MarkThreadAsFree(XxxRequestThread *Thread) {
    return servState.MarkThreadAsFree(Thread);
  }
  int NumClients() {
    return servState.NumClients();
   }
  XxxErrCode GetClient(int ClientId,
		       XxxClientInfo*& ClientInfo) {
    return servState.GetClient(ClientId, ClientInfo);
  }
};

#endif
