#ifndef XXX_ENV_H
#define XXX_ENV_H

#include <fstream.h>
#include "error.h"
#include "templates.h"
#include "shoresm-defs.h"
#include "options.h"


XXX_EXTERN_CLASS_DECL(XxxADTEnv);
XXX_EXTERN_CLASS_DECL(XxxEngineEnv);
XXX_EXTERN_CLASS_DECL(XxxJavaVM); // in utils/JavaVM.h
XXX_EXTERN_CLASS_DECL(XxxRemoteExecEngine); // in utils/JavaPlans.h
XXX_EXTERN_CLASS_DECL(XxxSuspCompLib); // in suspcomp.h
XXX_EXTERN_CLASS_DECL(XxxProxyServerManager); // in utils/ProxyServerManager.h

/**
 * Defines the environment for each client.
 * This class encapsulates all the per-client "global" variables.
 */
class XxxEnvironment {
  void print(ostream& OStream) const;
  void enableExceptionHandlers(XxxSigHandlerType, XxxNewHandlerType);
  XxxErrInfo ErrorInfo;
  ////
  class GeneralStats {
    //// record-level activity for this client
    class RecordStats {
     public:
      // records
      int RecCreated;
      int RecDeleted;
      double AvgSizeRecCreated;
      double AvgSizeRecDeleted;
      int RecConcat;
      int RecGetField;
      int RecReclaim;
      int RecPrint;
      // attrs
      int AttrCreated;
      int AttrDeleted;
      // schemas
      int SchemaCompStats;
      int GetAttr0;
      int GetAttr1;
      int GetAttr2;
      int GetOffset;

      int Extra[5];

      void Reset() {
	RecCreated = RecDeleted = RecConcat = RecGetField = RecReclaim = 0;
	RecPrint = AttrCreated = AttrDeleted = SchemaCompStats = 0;
	GetAttr0 = GetAttr1 = GetAttr2 = GetOffset = 0;
	AvgSizeRecCreated = AvgSizeRecDeleted = 0.0;

	Extra[0] = Extra[1] = Extra[2] = Extra[3] = Extra[4] = 0;
       }
      void Print(ostream& OStr) const;
      RecordStats() { Reset(); }
     };
    //// SHORE activity for this client
    class ShoreStats {
     public:
      // fill this in
      void Reset() {}
      void Print(ostream& OStr) const {
	OStr << "     SHORE Stats : Not avaliable " << endl;
       }
      ShoreStats() { Reset(); }
     };
    //// expr-level activity for this client
    class ExprStats {
     public:
      void Reset() {}
      void Print(ostream& OStr) const {
	OStr << "     Expr Stats : Not avaliable " << endl;
       }
      ExprStats() { Reset(); }
     };
   public:
    RecordStats Record;
    ShoreStats Shore;
    ExprStats Expr;

    void Reset() {
      Record.Reset();
      Shore.Reset();
      Expr.Reset();
     }
    void Print(ostream& OStr) const;
    GeneralStats() { Reset(); }
  }; // class GeneralStats

 public:

  class DebugFlags {
  public:
    XxxBool Connection;
    XxxBool Socket;
    XxxBool Callback;
    XxxBool SuspComp;
    XxxBool Server;
    XxxBool Protocol;
    XxxBool ADTRules;
    XxxBool Memory;
    XxxBool Shore;
    XxxBool JavaVM;
    XxxBool Expression;
    XxxBool FutureUse0;
    XxxBool FutureUse1;
    XxxBool FutureUse2;
    XxxBool FutureUse3;
    XxxBool FutureUse4;
    DebugFlags();
    XxxErrCode SetValue(const char* ValName,
			char NewVal);
  };

  // the variables that define the cpu speed, etc
  XxxDBEnv DBEnv; // defined in generic-plan.h

  // variables used by InitThread and MonitorThread: common to all clients
  int  StorageManager;  
  char *HostName;
  int Port;

  // options (replace environment variables)
  options * Options;

#ifdef XXX_JAVA_VIRTUAL_MACHINE 
protected:
  /// This is the JavaVM wrapper object for this env (i.e. its thread)
  XxxJavaVM *JVM;
  // This is the remote execution engine wrapper object for this env.
  XxxRemoteExecEngine *RemoteEE;
public:

  // Returns the Java Virtual Machine used for this thread.
  XxxJavaVM *GetJVM()
  { return JVM;}
  // Returns the Remote Execution Engine used for this thread
  XxxRemoteExecEngine *GetRemoteExecEngine()
  {return RemoteEE;}
#endif

#ifdef XXX_DEF_COUGAR
protected:
	//ProxyServerManager
	XxxProxyServerManager *PSM;
public:
	XxxProxyServerManager* GetProxyServerManager()
	{ return PSM; }
	XxxErrCode StartProxyServers();
	XxxErrCode ShutProxyServers();

#endif


public:
  // Variables that can be set by each client
  DebugFlags DebugFlags;
  int  VerboseLevel;

  // Interrupt Occurred
  XxxBool Interrupt;
  // Reason for Interrupt
  int InterruptSignal;

  // Is there a multi-stmt xact in progress?
  XxxBool MultiStmtXact;

  ostream* ErrorStr;
  ostream* OutputStr;

  // The absolute path of the directory into which files can be
  // written such that they are accessible by the Java Client Program
  char *TmpDataDir;

  // The volume ID of a volume on which the following files reside
  XxxShoreVolumeId StorageVolume;
  // The OID of a persistent file which is globally accessible.
  XxxShoreObjectId StorageFile;
  // The OID of a temporary file which is globally accessible.
  XxxShoreObjectId TempFileId;
  // The nuber of milliseconds the temp file will stay around after 
  // disconnection (for out-of-context callbacks)
  long TempFileDeleteDelay;
  // The buffer which holds all of the suspended computations for this
  // client.
  XxxSuspCompLib *SuspendedComputations;

  // Array of environments, one for each engine.
  // This is needed to allow users to specify defaults for each
  // of the engines.
  XxxEngineEnv **EngineEnvs;
  int NumEngineEnvs;

  // Array of environments, one for each ADT
  // This is needed to allow each ADT to maintain stats per client
  XxxADTEnv **ADTEnvs;
  int NumADTEnvs;

  // Maintain a stack of output streams for diagnostics.
  // This is accessed, for instance, deep inside code in the
  // utils directory, and it is too painful to pass this in as a
  // parameter.
  XxxStack<ostream *> DiagnosticStream;

  // statistics maintained
  GeneralStats Stats;

  void PrintError(const XxxErrInfo*, const char *msg = "") const;
  void Display() { this->print(*(this->ErrorStr)); }

  XxxBool ProcessCommandLine(int argc, char **argv);
  void Initialize(XxxSigHandlerType, XxxNewHandlerType, int Port);
  // remove signal and error flags that might be present.
  // to be used at the start of each query.
  void Clear();
  XxxErrCode SetValue(const char* ValName,
		      char NewVal);

  /**
   * Initializes storage information for the environment.  This
   * requires that SHORE is up and running and should be run for
   * <I>every client's</I> environment, since it generates a 
   * temporary file.  However, it should not be run on callbacks'
   * environments, since the temp file must be shared.
   */
  XxxErrCode InitStorage();
  /**
   * Removes the temp file.
   */
  XxxErrCode DumpStorage();

#ifdef XXX_JAVA_VIRTUAL_MACHINE
  //// Initialize the JVM
  XxxErrCode StartJVM();
  //// Close the JVM
  XxxErrCode ShutJVM();
  // Initialize RemoteExecEngine 
  XxxErrCode StartRemoteExecEngine();
  // Close RemoteExecEngine 
  XxxErrCode ShutRemoteExecEngine();
#endif

  const XxxEnvironment &operator=(const XxxEnvironment &other);

  void PrintHelp(ostream& OStr) const;
  void PrintStats(ostream& OStr) const
   { Stats.Print(OStr); }
  void ResetStats()
   { Stats.Reset(); }


  XxxEnvironment();
  XxxEnvironment(const XxxEnvironment& Env);
  ~XxxEnvironment();

  // routines to set up and destroy engine and adt envs.
  // these are not part of the constr/destr because they
  // need to be invoked when each thread starts running,
  // not when each thread gets created.
  XxxErrCode CreateEnvs();
  XxxErrCode DeleteEnvs();
 };

//// Checks if the file exists and opens it
extern XxxErrCode CheckFileExists(ifstream& InFStream,
				  const char *FileName,
				  const char *DirPath = NULL, 
				  char WithComments = 0);

/**
 * Superclass for environments of E-ADTs.
 */
class XxxADTEnv {
 public:
  // apply topdown rules when doing function rewrites
  static const char* const APPLY_TOPDOWN;

  // apply bottom up rules when doing function rewrites
  static const char* const APPLY_BOTTOMUP;

  // pipe output from result of one operation into input of another
  static const char* const ADT_PIPE;

  // suspend the computation of methods
  static const char* const SUSPEND_COMPUTATION;

  // cache the result of methods
  static const char* const METHOD_CACHE;

  // use a method scratch area
  static const char* const METHOD_SCRATCH;

  // various flags that store the current value of some features.
  XxxBool ApplyTopDown;
  XxxBool ApplyBottomUp;
  XxxBool AdtPipe;
  XxxBool SuspendComputation;
  XxxBool MethodCache; 
  XxxBool MethodScratchArea;
  // these last two are just hints that individual methods can ignore

  XxxADTEnv();
  virtual ~XxxADTEnv();
  virtual void Print(ostream& OStr) const;
  virtual XxxErrCode SetValue(const char* ValName,
			      void *NewVal);
};

/**
 * Superclass for environments of engines
 */
class XxxEngineEnv {
 public:
  XxxEngineEnv();
  virtual ~XxxEngineEnv();
  virtual void Print(ostream& OStr) const = 0;
  virtual XxxErrCode SetValue(const char* ValName,
			      char NewVal) = 0;
};

#endif  /* XXX_ENV_H */
