#include <iostream.h>
#include "prd-config.h"
#include "types.h"
#include "expr.h"
#include "generic.h"
#include "shoresm.h"

void XxxGenericPlan::PlanUsage::Print(ostream& OStr) const
{
  OStr << "INFO ";
  OStr << " COST: " << FullUsage.Cost();
  OStr << endl << "INFO ";
  OStr << " SEQIO:" << FullUsage.Disk.SeqIO;
  OStr << endl << "INFO ";
  OStr << " RNDIO:" << FullUsage.Disk.RandIO;
  OStr << endl << "INFO ";
  OStr << " CPU:" << FullUsage.Proc;
  OStr << endl << "INFO ";
  OStr << " (Overlap:" << FullUsage.DiskProcOverlap;
  OStr << " MemUsed:" << FullUsage.Mem.PagesUsed;
  OStr << " MemInUse:" << FullUsage.Mem.PagesInUse;
  OStr << " BufUsed:" << FullUsage.Buf.PagesUsed;
  OStr << " BufInUse:" << FullUsage.Buf.PagesInUse;
  OStr << ")" << endl ;
}

//
// XxxDBEnv
//
XxxDBEnv::XxxDBEnv()
{
 // Can't get real values here because some envs may be created
 // before the ShoreSM is created
 MPL = 1;
 PageSize = 8192;  // bytes
 RecordOverhead = 50;  // bytes: my estimate of Shore overhead
 BufPoolSize = 1000;
 MemorySize = 1000;

 CPUSpeed = 100 * 1000 * 1000;    // 100M instr/sec

 DiskLatency = 8.0 / 1000.0 ; // 8 msec
 DiskXferRate = 10 * 1000 * 1000; // 10 MBytes/sec
 computeDerivedVals();
}

void XxxDBEnv::GetCurrentValues()
{
 PageSize = SM_PAGESIZE;
 if (XxxShoreSM) 
   BufPoolSize = XxxShoreSM->NumBuffers();
 // assume working mem is half the size of the buffer pool
 MemorySize = BufPoolSize/2;

 computeDerivedVals();
}

void XxxDBEnv::computeDerivedVals()
{
  SeqIOWeight = 1.0; // define other weights based on this
  double SeqIOTime = PageSize/DiskXferRate;

  RndIOWeight = (DiskLatency + SeqIOTime)/SeqIOTime;

  CPUInstrWeight = 1.0/(CPUSpeed * SeqIOTime);
}

void XxxDBEnv::Print(ostream& OStr) const
{
  OStr << "\t MPL = " << MPL << endl;
  OStr << "\t PageSize = " << PageSize << endl;
  OStr << "\t RecordOverhead = " << RecordOverhead << endl;
  OStr << "\t BufPoolSize = " << BufPoolSize << endl;
  OStr << "\t MemorySize = " << MemorySize << endl;
  OStr << "\t CPUSpeed = " << CPUSpeed << endl;
  OStr << "\t DiskLatency = " << DiskLatency << endl;
  OStr << "\t DiskXferRate = " << DiskXferRate << endl;
  OStr << "\t SeqIOWeight = " << SeqIOWeight << endl;
  OStr << "\t RndIOWeight = " << RndIOWeight << endl;
  OStr << "\t CPUInstrWeight = " << CPUInstrWeight << endl;
}

XxxErrCode XxxDBEnv::SetValue(const char* ValName, char NewVal) 
{
 if (!strcmp(ValName, "mpl")) {
   MPL = NewVal;
  } 
 else if (!strcmp(ValName, "bufsize")) {
   BufPoolSize = NewVal*1000;
  } 
 else if (!strcmp(ValName, "memsize")) {
   MemorySize = NewVal*1000;
  } 
 else if (!strcmp(ValName, "cpu")) {
   CPUSpeed = NewVal*100000;
  } 
 else if (!strcmp(ValName, "latency")) {
   DiskLatency = NewVal / 1000;
  } 
 else if (!strcmp(ValName, "xfer")) {
   DiskXferRate = NewVal*10000;
  } 
 else {
   return XXX_NOT_OK;
 }
 computeDerivedVals();
 return XXX_OK;
}

// put this here for now: TODO: move this into generic-plan.h
double XxxGenericPlan::UsageComponents::Cost() const
{
  if ((Disk.SeqIO == INT_MAX) ||
      (Disk.RandIO == INT_MAX) ||
      (Proc == DBL_MAX)) { return DBL_MAX;}

  double CurrIOCost = 0.0;
  CurrIOCost += Disk.SeqIO * XXX_THISENV().DBEnv.SeqIOWeight;
  CurrIOCost += Disk.RandIO * XXX_THISENV().DBEnv.RndIOWeight;
  double CurrCPUCost = 0.0;
  CurrCPUCost += Proc * XXX_THISENV().DBEnv.CPUInstrWeight;

  // for now, ignore the IO - CPU overlap
  return (CurrIOCost + CurrCPUCost);
}

