// Note: this file contains code that defines "generic" functions --
// those that can be used in an SQL query, but are not methods of any ADT.
// They are modeled as belonging to XxxGenFuncADTClass.
#include <iostream.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "prd-config.h"
#include "xxx_thread.h"
#include "error.h"
#include "genfunc.h"
#include "generic.h"
#include "values.h"
#include "engines.h"
#include "bytearray.h"
#include "xxx_shmem.h"
#include "smartfunc.h"
#include "strrules.h"

/********* TRIAL OF EXPENSIVE FUNCTIONS *********/
class ExpensiveFuncPlan : public XxxFuncPlanInfo {
  static XxxErrCode ExpensiveOptFunc(XxxFuncParseInfo* FParse,
				   XxxValueExprPlan* OwnerPlan,
				   XxxValExprPlanList* ArgPlans,
				   XxxFuncPlanInfo*& FuncPlan);
 public: 
  class ScratchSpace : public XxxFuncScratchPad {
  public:
    int RunningSum;
    ScratchSpace() { RunningSum = 0; }
  };
  /// execute expintplus()
  XxxErrCode execIntPlus(XxxValueEnv* Env, 
			 XxxADTValue *ReturnValue) const {
    ASSERT((ReturnValue && (Env->NumArgs == 2)));
    int Arg0 = ((XxxCoreADTClass::IntValue *)(Env->ArgValsPtrs[0]))->Value;
    int Arg1 = ((XxxCoreADTClass::IntValue *)(Env->ArgValsPtrs[1]))->Value;
    ((XxxCoreADTClass::IntValue *)ReturnValue)->Value = Arg1 + Arg0;
    return XXX_OK;
  }
  /// execute oddnum()
  XxxErrCode execOddNum(XxxValueEnv* Env, 
			 XxxADTValue *ReturnValue) const {
    ASSERT((ReturnValue && (Env->NumArgs == 1)));
    int Arg0 = ((XxxCoreADTClass::IntValue *)(Env->ArgValsPtrs[0]))->Value;
    ((XxxCoreADTClass::BoolValue *)ReturnValue)->Value = 
      ((Arg0/2 * 2 == Arg0) ? XXX_FALSE : XXX_TRUE);
    return XXX_OK;
  }
  /// execute evennum()
  XxxErrCode execEvenNum(XxxValueEnv* Env, 
			 XxxADTValue *ReturnValue) const {
    ASSERT((ReturnValue && (Env->NumArgs == 1)));
    int Arg0 = ((XxxCoreADTClass::IntValue *)(Env->ArgValsPtrs[0]))->Value;
    ((XxxCoreADTClass::BoolValue *)ReturnValue)->Value = 
      ((Arg0/2 * 2 == Arg0) ? XXX_TRUE : XXX_FALSE);
    return XXX_OK;
  }
  /// execute runningsum()
  XxxErrCode execRunningSum(XxxValueEnv* Env, 
			    XxxADTValue *ReturnValue) const {
    ASSERT((ReturnValue && (Env->NumArgs == 1)));
    int Arg0 = ((XxxCoreADTClass::IntValue *)(Env->ArgValsPtrs[0]))->Value;
    ScratchSpace* SS = (ScratchSpace*)ScratchPad;
    ((XxxCoreADTClass::IntValue *)ReturnValue)->Value = Arg0 + SS->RunningSum;
    SS->RunningSum += Arg0;
    return XXX_OK;
  }
  ExpensiveFuncPlan(const XxxFuncParseInfo* FuncParse, int argCount)
    : XxxFuncPlanInfo(FuncParse, argCount){
      planUsage.FullUsage.Disk.SeqIO = 100; // trial
      ScratchPad = new ScratchSpace();
  }
};
static XxxErrCode ExpensiveOptFunc(XxxFuncParseInfo* FParse,
				   XxxValueExprPlan* OwnerPlan,
				   XxxValExprPlanList* ArgPlans,
				   XxxFuncPlanInfo*& FuncPlan)
{
  ExpensiveFuncPlan* ExpFuncPlan = 
    new ExpensiveFuncPlan(FParse, (ArgPlans?ArgPlans->Count():0));
  // add in the costs of OwnerPlan and ArgPlans
  if (OwnerPlan)
    ExpFuncPlan->getPlanUsage().Add(OwnerPlan->GetPlanUsage());
 // step thru arg plans adding cost
 const XxxValExprPlanList* Tmp = ArgPlans;
 while (Tmp) {
   ExpFuncPlan->getPlanUsage().Add(Tmp->Plan->GetPlanUsage());
   Tmp = Tmp->Next;
 }
 FuncPlan = ExpFuncPlan;
 return XXX_OK;
}


//
// Methods of XxxGenFuncADTClass
//

XxxErrCode XxxGenFuncADTClass::FuncTypeCheck(const char*FuncName,
					     const XxxValueExpression* Owner,
					     const XxxValueExprList *ArgList,
					     XxxBool IsAggr,
					     XxxBasicType& ReturnType,
					     XxxFuncParseInfo*& FuncParse) const
{
 FuncParse = NULL;
 int MethodIndex;

 XxxADTMetaInfo *ReturnMetaInfo = NULL;
 XxxBool Success = XXX_FALSE;
 XxxGenericParse* ExprParse = NULL;

 if (IsAggr) {
   XxxFuncExpression::XxxAggrType AggType;
   XxxFuncExpression::findAggrType(FuncName, AggType);
   if (AggType != XxxFuncExpression::ADT) {
     if (Owner || !ArgList || ArgList->Count() != 1) {
       XXX_NOTE_ERROR(XXX_BADTYPE, "SQL-92 aggregate should have exactly one parameter");
       return XXX_NOT_OK;
     }
     MethodIndex = (int)AggType;
     switch (AggType) {
     case XxxFuncExpression::MIN:
     case XxxFuncExpression::MAX:
       ReturnType = (*ArgList)[0]->BasicType();
       if (!((ReturnType == XXX_INT) || (ReturnType == XXX_DOUBLE) ||
           (ReturnType == XXX_STRING))) {
	 XXX_NOTE_ERROR(XXX_BADTYPE, "Bad argument type for MIN or MAX");
	 return XXX_NOT_OK;
       }
       Success = XXX_TRUE;
       break;
     case XxxFuncExpression::SUM:
     case XxxFuncExpression::PROD:
     case XxxFuncExpression::AVG:
       ReturnType = (*ArgList)[0]->BasicType();
       if ((ReturnType != XXX_INT) && (ReturnType != XXX_DOUBLE)) {
	 XXX_NOTE_ERROR(XXX_BADTYPE, "Bad argument type for SUM, PROD, or AVG");
	 return XXX_NOT_OK;
       }
       Success = XXX_TRUE;
       break;
     case XxxFuncExpression::COUNT:
       ReturnType = XXX_INT;
       Success = XXX_TRUE;
       break;
     default:
       XXX_NOTE_ERROR(XXX_BADTYPE, "Invalid aggregate");
       return XXX_NOT_OK;
     } // switch
   } // if AggType != ADT
 } // if IsAggr

 if (!Success) {
   // check for valid method names
   if (matchMethodName(FuncName, Owner, ArgList, IsAggr,
		       MethodIndex, ReturnType,
		       ReturnMetaInfo, Success) != XXX_OK) {
     char buf[100];
     sprintf(buf, "Invalid function name %s", FuncName);
     XXX_NOTE_ERROR(XXX_BADEXPR, buf);
     return XXX_NOT_OK;
   }
 }

 if (!Success) {
   // Check if this is an "engine" function, like $SQL()
   // These are modeled as functions.
   // The first argument _must_ be a constant string
   if ((ArgList->Count() < 1) || !ArgList->Expr->IsConstant() ||
       (ArgList->Expr->BasicType() != XXX_STRING)) {
     char buf[100];
     sprintf(buf, "Invalid function/language name %s", FuncName);
     XXX_NOTE_ERROR(XXX_BADFUNC, buf);
     return XXX_NOT_OK;
   }
   XxxBool LangFound = XXX_FALSE;
   for (int i = 0; i < XxxNumDBEngines; i++) {
     for (int j = 0; j < XxxDBEngineArray[i]->NumLanguages; j++) {
       if (!strncasecmp(XxxDBEngineArray[i]->LanguageNames[j], FuncName, 
			strlen(XxxDBEngineArray[i]->LanguageNames[j]))) {
	 LangFound = XXX_TRUE;
	 char* ParseString = 
	   new char[strlen(((XxxConstStringValue *)(ArgList->Expr))->c) +
		   strlen(FuncName)+5];
	 sprintf(ParseString, "%s:%s",FuncName,
		 ((XxxConstStringValue *)(ArgList->Expr))->c);

	 XxxArgEnv* Env;
	 Env = (ArgList->Count()-1 ? new XxxArgEnv(ArgList->Count()-1, ArgList->Next) : 0);
	 if (XxxDBEngineArray[i]->ExprParse(ParseString, Env,
					    ReturnType, ExprParse) != XXX_OK) {
	   // reclaim alloated space
	   delete ParseString;
	   if (Env) delete Env; 
	   char buf[100];
	   sprintf(buf, "Invalid %s language expression", FuncName);
	   XXX_NOTE_ERROR(XXX_BADEXPR, buf);
	   return XXX_NOT_OK;
	 }
	 else {
	   delete ParseString;
	   if (Env) delete Env; 
	   // FuncParse = new XxxFuncParseInfo(FuncName, ExprParse, i, ReturnType);
	   // return XXX_OK;
	   break;
	 }
       } // if the languages match
     } // for each language for engine
     if (LangFound) {
       Success = XXX_TRUE;
       break;
     }
   } // for each engine

 } // if (!Success)

 if (Success) {
   // since it is a valid function, construct a parse data structure
   FuncParse = new XxxFuncParseInfo(FuncName, MethodIndex, ExprParse, XXX_GENFUNC,
			     ReturnType, ReturnMetaInfo);
   FuncParse->IsAggr = IsAggr;
   return XXX_OK;;
 }
 else {
   char buf[100];
   sprintf(buf, "Invalid function/language name %s", FuncName);
   XXX_NOTE_ERROR(XXX_BADEXPR, buf);
   return XXX_NOT_OK;
 }
}

// Optmizes a method
XxxErrCode XxxGenFuncADTClass::methOptimize(XxxFuncParseInfo* FuncParse,
					    XxxValueExprPlan* OwnerPlan,
					    XxxValExprPlanList* ArgPlans,
					    XxxFuncPlanInfo*& FuncPlan) const 
{
 FuncPlan = NULL;
 // if this corresponds to a language expression
 if (!FuncParse) {
   XXX_NOTE_ERROR(XXX_BADFUNC, "No FuncParse passed");
   return XXX_NOT_OK;
 }
 if (FuncParse->ExprParse) {
   XxxGenericPlan* ExprPlan;
   if (FuncParse->EngineIndex >= XxxNumDBEngines) {
     char buf[35];
     sprintf(buf, "Invalid engine index %d", FuncParse->EngineIndex);
     XXX_NOTE_ERROR(XXX_BADADT, buf);
     return XXX_NOT_OK;
    }
   if (XxxDBEngineArray[FuncParse->EngineIndex]->
       ExprOptimize(FuncParse->ExprParse, ExprPlan) != XXX_OK)
     return XXX_NOT_OK;
   else {
     FuncPlan = new XxxFuncPlanInfo(FuncParse, (ArgPlans?ArgPlans->Count():0), 
				    ExprPlan);
     if (ExprPlan) {
      XxxGenericPlan::PlanUsage Tmp;
      ExprPlan->UsageOfPlan(Tmp);
      FuncPlan->getPlanUsage().Add(Tmp);
     }
     return XXX_OK;
    }
  }
 else if (FuncParse->IsAggr && 
	  ((XxxFuncExpression::XxxAggrType)(FuncParse->MethodIndex) != XxxFuncExpression::ADT)) {
   // do nothing for SQL-92 aggregates
   return XXX_OK;
 }
 else {
   // it is not a language expression
   return XxxADTClass::methOptimize(FuncParse, OwnerPlan,
				    ArgPlans, FuncPlan);
 }
}

//
// define an "OptimizeFunc" for each function
//
static XxxErrCode XxxDefaultOptFunc(XxxFuncParseInfo* FParse,
				    XxxValueExprPlan* OwnerPlan,
				    XxxValExprPlanList* ArgPlans,
				    XxxFuncPlanInfo*& FuncPlan)
{
 FuncPlan = new XxxGenFuncADTClass::SimpleFuncPlanInfo(FParse, (ArgPlans?ArgPlans->Count():0));
 return XXX_OK;
}

static XxxErrCode XxxMemOptFunc(XxxFuncParseInfo* FuncParse,
				XxxValueExprPlan* OwnerPlan,
				XxxValExprPlanList* ArgPlans,
				XxxFuncPlanInfo*& FuncPlan)
{
  XxxGenFuncADTClass::MemFuncPlanInfo* MemFPlan;
  MemFPlan = new XxxGenFuncADTClass::MemFuncPlanInfo(FuncParse,
						     NULL, ArgPlans);
  if (!MemFPlan || (!MemFPlan->Initialized)) {
    if (MemFPlan) delete MemFPlan;
    XXX_NOTE_ERROR(XXX_BADOPT, "Unable to create mem func plan");
    return XXX_NOT_OK;
  }
  FuncPlan = MemFPlan;
  return XXX_OK;
}

/**
 *   intplus(int1, int2) : returns int1+int2
 *   matches(string1, string2) : returns 1 if string1 is a substr of string2
 */
XxxGenFuncADTClass::XxxGenFuncADTClass(XxxBasicType Id, const char* Name)
    : XxxCoreADTClass(Id,Name) 
{
  registerMethods();

  registerRules();
}

void XxxGenFuncADTClass::registerMethods()
{
  NumMethods = 21; // unconditional methods (not #ifdef'ed)
#ifdef XXX_REMOTE_FUNCS
  NumMethods += 2;
#endif
#ifdef XXX_JAVA_FUNCS
  NumMethods += 1;
#endif
  
  NumMethodSlots = NumMethods + XxxADTClass::NUM_SPARE_METHOD_SLOTS;
  MethArray = new FuncMethodInfo*[NumMethodSlots];

  int i;

  i = 0;
  MethArray[i++] = new FuncMethodInfo("intplus", XXX_INT, 0, XxxDefaultOptFunc,
				    (ExecuteFunc) &(SimpleFuncPlanInfo::execIntPlus),
				    2, XXX_INT, XXX_INT);
  MethArray[i++] = new FuncMethodInfo("matches", XXX_INT, 0, XxxDefaultOptFunc,
				    (ExecuteFunc) &(SimpleFuncPlanInfo::execMatches),
				    2, XXX_STRING, XXX_STRING);
  MethArray[i++] = new FuncMethodInfo("testfunc", XXX_INT, 0, XxxMemOptFunc,
				    (ExecuteFunc) &(MemFuncPlanInfo::execTestFunc),
				    3, XXX_INT, XXX_INT, XxxByteArrayADT::ByteArrayTypeId);
#ifdef XXX_REMOTE_FUNCS
  MethArray[i++] = new FuncMethodInfo("rintplus", XXX_INT, 0, RemoteOptFunc,
				    (ExecuteFunc) &(PlanInfo::execRIntPlus),
				    2, XXX_INT, XXX_INT);
  MethArray[i++] = new FuncMethodInfo("rtestfunc", XXX_INT, 0, RemoteOptFunc,
				    (ExecuteFunc) &(PlanInfo::execRTestFunc),
				    3, XXX_INT, XXX_INT, XxxByteArrayADT::ByteArrayTypeId);
#endif
#ifdef XXX_JAVA_FUNCS
  MethArray[i++] = new FuncMethodInfo("jtestfunc", XXX_INT, 0, JavaOptFunc,
				    (ExecuteFunc) &(PlanInfo::execJTestFunc),
				    3, XXX_INT, XXX_INT, XxxByteArrayADT::ByteArrayTypeId);
#endif
  // trial : an "expensive" intplus function
  MethArray[i++] = new FuncMethodInfo("expintplus", XXX_INT, 0, ExpensiveOptFunc,
				    (ExecuteFunc) &(ExpensiveFuncPlan::execIntPlus),
				    2, XXX_INT, XXX_INT);
  // trial : an "expensive" intplus function
  MethArray[i++] = new FuncMethodInfo("exprunningsum", XXX_INT, 0, ExpensiveOptFunc,
				    (ExecuteFunc) &(ExpensiveFuncPlan::execRunningSum),
				    1, XXX_INT);
  // trial : a boolean function
  XxxFuncMethodInfo::ArgInfo TypeSignature(1, XXX_INT);
  XxxGenericPlan::PlanUsage Usage;
  MethArray[i++] = new FuncMethodInfo("oddnum", XXX_BOOLEAN,
				    TypeSignature, 
				    (ExecuteFunc) &(ExpensiveFuncPlan::execOddNum),
				    Usage);

  MethArray[i++] = new FuncMethodInfo("evennum", XXX_BOOLEAN,
				     TypeSignature, 
				     (ExecuteFunc) &(ExpensiveFuncPlan::execEvenNum),
				     Usage);

  MethArray[i++] = new FuncMethodInfo("mod", XXX_INT, 0, XxxDefaultOptFunc,
				     (ExecuteFunc) &(SimpleFuncPlanInfo::execMod),
				     2, XXX_INT, XXX_INT);

  // Like
  MethArray[i++] = new FuncMethodInfo("strmatch", XXX_BOOLEAN, 0, XxxDefaultOptFunc,
				     (ExecuteFunc) &(SimpleFuncPlanInfo::execLike),
				     2, XXX_STRING, XXX_STRING);
  MethArray[i++] = new FuncMethodInfo("notstrmatch", XXX_BOOLEAN, 0, XxxDefaultOptFunc,
				     (ExecuteFunc) &(SimpleFuncPlanInfo::execNotLike),
				     2, XXX_STRING, XXX_STRING);
  // string functions
  MethArray[i++] = new FuncMethodInfo("concat", XXX_STRING, 0, XxxDefaultOptFunc,
				     (ExecuteFunc) &(SimpleFuncPlanInfo::execConcat),
				     2, XXX_STRING, XXX_STRING);
  MethArray[i++] = new FuncMethodInfo("ucase", XXX_STRING, 0, XxxDefaultOptFunc,
				     (ExecuteFunc) &(SimpleFuncPlanInfo::execUcase),
				     1, XXX_STRING);
  MethArray[i++] = new FuncMethodInfo("lcase", XXX_STRING, 0, XxxDefaultOptFunc,
				     (ExecuteFunc) &(SimpleFuncPlanInfo::execLcase),
				     1, XXX_STRING);
  MethArray[i++] = new FuncMethodInfo("trim", XXX_STRING, 0, XxxDefaultOptFunc,
				     (ExecuteFunc) &(SimpleFuncPlanInfo::execTrim),
				     3, XXX_STRING, XXX_INT, XXX_STRING);
  MethArray[i++] = new FuncMethodInfo("ltrim", XXX_STRING, 0, XxxDefaultOptFunc,
				     (ExecuteFunc) &(SimpleFuncPlanInfo::execLtrim),
				     1, XXX_STRING);
  MethArray[i++] = new FuncMethodInfo("rtrim", XXX_STRING, 0, XxxDefaultOptFunc,
				     (ExecuteFunc) &(SimpleFuncPlanInfo::execRtrim),
				     1, XXX_STRING);



  // double functions
  MethArray[i++] = new FuncMethodInfo("abs", XXX_DOUBLE, 0, XxxDefaultOptFunc,
				     (ExecuteFunc) &(SimpleFuncPlanInfo::execAbs),
				     1, XXX_DOUBLE);
  MethArray[i++] = new FuncMethodInfo("ceiling", XXX_DOUBLE, 0, XxxDefaultOptFunc,
				     (ExecuteFunc) &(SimpleFuncPlanInfo::execCeiling),
				     1, XXX_DOUBLE);
  MethArray[i++] = new FuncMethodInfo("floor", XXX_DOUBLE, 0, XxxDefaultOptFunc,
				     (ExecuteFunc) &(SimpleFuncPlanInfo::execFloor),
				     1, XXX_DOUBLE);
  MethArray[i++] = new FuncMethodInfo("power", XXX_DOUBLE, 0, XxxDefaultOptFunc,
				     (ExecuteFunc) &(SimpleFuncPlanInfo::execPower),
				     2, XXX_DOUBLE, XXX_DOUBLE);
  MethArray[i++] = new FuncMethodInfo("sign", XXX_DOUBLE, 0, XxxDefaultOptFunc,
				     (ExecuteFunc) &(SimpleFuncPlanInfo::execSign),
				     1, XXX_DOUBLE);
  ASSERT((NumMethods == i)); // was NumMethods set correctly?
}

void XxxGenFuncADTClass::registerRules()
{
  RuleEngine = new XxxOptRuleEngine(10, 10, 1 /* traversals */);

  RuleEngine->AddRule(new XxxStrRule("[[$oddnum(X) and $evennum(X)]]", "[[1 = 0]]",XXX_GENFUNC, "EvenOddCancel"));
  RuleEngine->AddRule(new XxxStrRule("[[X = Y and Y = Z]]", 
				     "[[X=Y and Y=Z and X=Z]]",
				     XXX_GENFUNC, 
				     "EqualityClosure", 
				     XxxOptRuleEngine::BOTTOM_UP));

  RuleEngine->AddRule(new XxxStrRule("[[A.X.boxequals(B.Y) and B.Y.height() < Z]]", 
				     "[[A.X.boxequals(B.Y) and B.Y.height() < Z and A.X.height() < Z]]",
				     XXX_GENFUNC,
				     "ImageRule",
				     XxxOptRuleEngine::BOTTOM_UP));
}

//
// methods of XxxGenFuncADTClass::SimpleFuncPlanInfo
//
XxxErrCode XxxGenFuncADTClass::
SimpleFuncPlanInfo::execIntPlus(XxxValueEnv* ValueEnv, 
				XxxADTValue* ReturnValue) const
{
  ASSERT((ReturnValue && (ValueEnv->NumArgs == 2)));
  int Arg0 = ((XxxCoreADTClass::IntValue *)(ValueEnv->ArgValsPtrs[0]))->Value;
  int Arg1 = ((XxxCoreADTClass::IntValue *)(ValueEnv->ArgValsPtrs[1]))->Value;
  ((XxxCoreADTClass::IntValue *)ReturnValue)->Value = Arg1 + Arg0;
  return XXX_OK;
}

XxxErrCode XxxGenFuncADTClass::
SimpleFuncPlanInfo::execMod(XxxValueEnv* ValueEnv, 
			    XxxADTValue* ReturnValue) const
{
  ASSERT((ReturnValue && (ValueEnv->NumArgs == 2)));
  int Arg0 = ((XxxCoreADTClass::IntValue *)(ValueEnv->ArgValsPtrs[0]))->Value;
  int Arg1 = ((XxxCoreADTClass::IntValue *)(ValueEnv->ArgValsPtrs[1]))->Value;
  ((XxxCoreADTClass::IntValue *)ReturnValue)->Value = Arg0 % Arg1;
  return XXX_OK;
}

/*
 * LIKE.
 * (S1 LIKE S2) is true, if S2 is contained in S1, subject to wildcards.
 * TODO: support wild cards.
 */
XxxErrCode XxxGenFuncADTClass::
SimpleFuncPlanInfo::execLike(XxxValueEnv* ValueEnv, 
			     XxxADTValue* ReturnValue) const
{
  ASSERT((ReturnValue && (ValueEnv->NumArgs == 2)));
  ((BoolValue *)ReturnValue)->Value = 
    !(strstr(((XxxCoreADTClass::StrValue *)ValueEnv->ArgValsPtrs[0])->Value,
	     ((XxxCoreADTClass::StrValue *)ValueEnv->ArgValsPtrs[1])->Value) == NULL);
  return XXX_OK;
}

/*
 * Not LIKE.
 */
XxxErrCode XxxGenFuncADTClass::
SimpleFuncPlanInfo::execNotLike(XxxValueEnv* ValueEnv, 
			     XxxADTValue* ReturnValue) const
{
  ASSERT((ReturnValue && (ValueEnv->NumArgs == 2)));
  ((BoolValue *)ReturnValue)->Value = 
    (strstr(((XxxCoreADTClass::StrValue *)ValueEnv->ArgValsPtrs[0])->Value,
	     ((XxxCoreADTClass::StrValue *)ValueEnv->ArgValsPtrs[1])->Value) == NULL);
  return XXX_OK;
}

/*
 * Concatenate two strings.
 */
XxxErrCode XxxGenFuncADTClass::
SimpleFuncPlanInfo::execConcat(XxxValueEnv* ValueEnv, 
	  		       XxxADTValue* ReturnValue) const
{
  char *to, *from;

  ASSERT((ReturnValue && (ValueEnv->NumArgs == 2)));
  to = ((StrValue *)(ReturnValue))->Value;
  to[0] = '\0';
  from = ((XxxCoreADTClass::StrValue *)(ValueEnv->ArgValsPtrs[0]))->Value;
  strcat(to, from);
  from = ((XxxCoreADTClass::StrValue *)(ValueEnv->ArgValsPtrs[1]))->Value;
  strcat(to, from);
  return XXX_OK;
}

/*
 * UPPER.
 */
XxxErrCode XxxGenFuncADTClass::
SimpleFuncPlanInfo::execUcase(XxxValueEnv* ValueEnv, 
			      XxxADTValue* ReturnValue) const
{
  ASSERT((ReturnValue && (ValueEnv->NumArgs == 1)));
  char *sp1 = ((XxxCoreADTClass::StrValue *)(ValueEnv->ArgValsPtrs[0]))->Value;
  char *sp2 = ((StrValue *)(ReturnValue))->Value;
  while (*sp1 != '\0') {
    if (islower(*sp1))
      *sp2++ = toupper(*sp1++);
    else
      *sp2++ = *sp1++;
  }
  *sp2 = '\0';
  return XXX_OK;
}

/*
 * LOWER.
 */
XxxErrCode XxxGenFuncADTClass::
SimpleFuncPlanInfo::execLcase(XxxValueEnv* ValueEnv, 
			      XxxADTValue* ReturnValue) const
{
  ASSERT((ReturnValue && (ValueEnv->NumArgs == 1)));
  char *sp1 = ((XxxCoreADTClass::StrValue *)(ValueEnv->ArgValsPtrs[0]))->Value;
  char *sp2 = ((StrValue *)(ReturnValue))->Value;
  while (*sp1 != '\0') {
    if (isupper(*sp1))
      *sp2++ = tolower(*sp1++);
    else
      *sp2++ = *sp1++;
  }
  *sp2 = '\0';
  return XXX_OK;
}

/*
 * Trim subroutine.
 */
static void trimcore(char* to, char *from, int loc, char c)
{
  int lastvalid;
  char *last, *sp1, *sp2;

  sp1 = from;
  sp2 = to;

  // Skip leading c's.
  if (loc == XXX_TRIM_LEADING || loc == XXX_TRIM_BOTH) {
    while (*sp1 == c && *sp1 != '\0')
      *sp1++;
  }

  // Copy the string remebering the start of trailing c's.
  lastvalid = 0;
  while (*sp1 != '\0') {
    if(*sp1 == c && !lastvalid) {
      lastvalid = 1;
      last = sp2;
    }
    else if(*sp1 != c && lastvalid) {

      lastvalid = 0;
    }
    *sp2++ = *sp1++;
  }
  if (lastvalid && (loc == XXX_TRIM_TRAILING || loc == XXX_TRIM_BOTH))
    *last = '\0';
  else
    *sp2 = '\0';
}

/*
 * Trim.
 */
XxxErrCode XxxGenFuncADTClass::
SimpleFuncPlanInfo::execTrim(XxxValueEnv* ValueEnv, 
			     XxxADTValue* ReturnValue) const
{
  ASSERT((ReturnValue && (ValueEnv->NumArgs == 3)));
  char *from = ((XxxCoreADTClass::StrValue *)(ValueEnv->ArgValsPtrs[0]))->Value;
  int loc = ((XxxCoreADTClass::IntValue *)(ValueEnv->ArgValsPtrs[1]))->Value;
  char *c = ((XxxCoreADTClass::StrValue *)(ValueEnv->ArgValsPtrs[2]))->Value;
  char *to = ((StrValue *)(ReturnValue))->Value;
  trimcore(to, from, loc, c[0]);
  return XXX_OK;
}

/*
 * Ltrim.
 */
XxxErrCode XxxGenFuncADTClass::
SimpleFuncPlanInfo::execLtrim(XxxValueEnv* ValueEnv, 
			      XxxADTValue* ReturnValue) const
{
  ASSERT((ReturnValue && (ValueEnv->NumArgs == 1)));
  char *from = ((XxxCoreADTClass::StrValue *)(ValueEnv->ArgValsPtrs[0]))->Value;
  char *to = ((StrValue *)(ReturnValue))->Value;
  trimcore(to, from, XXX_TRIM_LEADING, ' ');
  return XXX_OK;
}

/*
 * Rtrim.
 */
XxxErrCode XxxGenFuncADTClass::
SimpleFuncPlanInfo::execRtrim(XxxValueEnv* ValueEnv, 
			      XxxADTValue* ReturnValue) const
{
  ASSERT((ReturnValue && (ValueEnv->NumArgs == 1)));
  char *from = ((XxxCoreADTClass::StrValue *)(ValueEnv->ArgValsPtrs[0]))->Value;
  char *to = ((StrValue *)(ReturnValue))->Value;
  trimcore(to, from, XXX_TRIM_TRAILING, ' ');
  return XXX_OK;
}


/*
 * Absolute.
 */
XxxErrCode XxxGenFuncADTClass::
SimpleFuncPlanInfo::execAbs(XxxValueEnv* ValueEnv, 
			    XxxADTValue* ReturnValue) const
{
  ASSERT((ReturnValue && (ValueEnv->NumArgs == 1)));
  double Arg = ((XxxCoreADTClass::DblValue *)(ValueEnv->ArgValsPtrs[0]))->Value;
  ((XxxCoreADTClass::DblValue *)ReturnValue)->Value = fabs(Arg) ;
  return XXX_OK;
}

/*
 * Ceiling.  (Note: might want to return integer type.)
 */
XxxErrCode XxxGenFuncADTClass::
SimpleFuncPlanInfo::execCeiling(XxxValueEnv* ValueEnv, 
			        XxxADTValue* ReturnValue) const
{
  ASSERT((ReturnValue && (ValueEnv->NumArgs == 1)));
  double Arg = ((XxxCoreADTClass::DblValue *)(ValueEnv->ArgValsPtrs[0]))->Value;
  ((XxxCoreADTClass::DblValue *)ReturnValue)->Value = ceil(Arg) ;
  return XXX_OK;
}

/*
 * Floor.  (Note: might want to return integer type.
 */
XxxErrCode XxxGenFuncADTClass::
SimpleFuncPlanInfo::execFloor(XxxValueEnv* ValueEnv, 
			      XxxADTValue* ReturnValue) const
{
  ASSERT((ReturnValue && (ValueEnv->NumArgs == 1)));
  double Arg = ((XxxCoreADTClass::DblValue *)(ValueEnv->ArgValsPtrs[0]))->Value;
  ((XxxCoreADTClass::DblValue *)ReturnValue)->Value = floor(Arg) ;
  return XXX_OK;
}

/*
 * Power.
 */
XxxErrCode XxxGenFuncADTClass::
SimpleFuncPlanInfo::execPower(XxxValueEnv* ValueEnv, 
			      XxxADTValue* ReturnValue) const
{
  ASSERT((ReturnValue && (ValueEnv->NumArgs == 2)));
  double Arg0 = ((XxxCoreADTClass::DblValue *)(ValueEnv->ArgValsPtrs[0]))->Value;
  double Arg1 = ((XxxCoreADTClass::DblValue *)(ValueEnv->ArgValsPtrs[1]))->Value;
  ((XxxCoreADTClass::DblValue *)ReturnValue)->Value = pow(Arg0, Arg1) ;
  return XXX_OK;
}

/*
 * Sign.  (Note: might want to return integer type.
 */
XxxErrCode XxxGenFuncADTClass::
SimpleFuncPlanInfo::execSign(XxxValueEnv* ValueEnv, 
			     XxxADTValue* ReturnValue) const
{
  ASSERT((ReturnValue && (ValueEnv->NumArgs == 1)));
  double Arg = ((XxxCoreADTClass::DblValue *)(ValueEnv->ArgValsPtrs[0]))->Value;
  if (Arg < 0)
    ((XxxCoreADTClass::DblValue *)ReturnValue)->Value = -1.0;
  else if (Arg > 0)
    ((XxxCoreADTClass::DblValue *)ReturnValue)->Value = 1.0;
  else
    ((XxxCoreADTClass::DblValue *)ReturnValue)->Value = 0.0;
  return XXX_OK;
}



XxxErrCode XxxGenFuncADTClass::
SimpleFuncPlanInfo::execMatches(XxxValueEnv* ValueEnv, 
				XxxADTValue* ReturnValue) const
{
  ASSERT((ReturnValue && (ValueEnv->NumArgs == 2)));
  ((IntValue *)ReturnValue)->Value = 
    !(strstr(((XxxCoreADTClass::StrValue *)ValueEnv->ArgValsPtrs[0])->Value,
	     ((XxxCoreADTClass::StrValue *)ValueEnv->ArgValsPtrs[1])->Value) == NULL);
  return XXX_OK;
}

//
// methods of XxxGenFuncADTClass::MemFuncPlanInfo
//
XxxGenFuncADTClass::
MemFuncPlanInfo::MemFuncPlanInfo(const XxxFuncParseInfo* FuncParse,
				 XxxGenericPlan* Exp,
				 XxxValExprPlanList* ArgPlans)
  : XxxFuncPlanInfo(FuncParse, (ArgPlans?ArgPlans->Count():0), Exp) ,
    ArgCount((ArgPlans?ArgPlans->Count():0))
{
  Initialized = XXX_FALSE;
  // create data to hold any large objects in their entirety
  SerializedArgs = new struct ArgData[ArgCount];
  for (int i = 0; i < ArgCount; i++) {
    XxxValueExprPlan* ArgPlan = (*ArgPlans)[i];
    XxxADTClass* ArgADT = XxxADTArray[ArgPlan->BasicType()];
    XxxADTMetaInfo* MInfo = ArgPlan->GenerateCatInfo();
    if (ArgADT->IsSmall(MInfo)) {
      // nothing special: dont waste time copying it
      SerializedArgs[i].SetEmpty();
    }
    else {
      SerializedArgs[i].ArgMetaInfo = MInfo;
      SerializedArgs[i].TypeId = ArgPlan->BasicType();
      unsigned int MemSize;
      if (ArgADT->SerializeCopySize(SerializedArgs[i].ArgMetaInfo,
				    MemSize) != XXX_OK) {
	XXX_NOTE_ERROR(XXX_BADOPT, "Error getting copy size");
	return;
      }
      SerializedArgs[i].SetSize(MemSize); 
    } // it is not Small()
  } // for each arg

  Initialized = XXX_TRUE;
}

XxxGenFuncADTClass::
MemFuncPlanInfo::~MemFuncPlanInfo()
{
  if (SerializedArgs && ArgCount) {
    delete [] SerializedArgs;
  }
}

// some hardcode functions to execute the "testfunc" function
static int ComputeTestFunc(int CompDIndep,
			   int CompDDep,
			   char* Data,
			   int ArraySize)
{
  if (*((int *)Data) > ArraySize) {
    cerr << "strange data sizes: " << ArraySize << " and " << *((int *)Data) << endl;
  }
  // ArraySize passed in is the whole data area size.
  // The actual size of the array is passed in the first 4 bytes
  ArraySize = *((int *)Data);
  char *DataArray = Data + sizeof(int);
  int Sum = 0;
  int i, j;
  for (i = 0; i < CompDDep; i++) {
    for (j = 0; j < ArraySize; j++) {
      Sum = Sum + DataArray[j];
    } // for each array element
  } // as many times as CompDDep
  for (i = 0; i < CompDIndep; i++) {
    Sum = Sum + 5;
  } // as many times as CompDIndep
  return Sum;
}

XxxErrCode XxxGenFuncADTClass::
MemFuncPlanInfo::execTestFunc(XxxValueEnv* ValueEnv, 
			      XxxADTValue* ReturnValue) const
{
  // hardcoded for now
  int Arg0 = ((XxxCoreADTClass::IntValue *)(ValueEnv->ArgValsPtrs[0]))->Value;
  int Arg1 = ((XxxCoreADTClass::IntValue *)(ValueEnv->ArgValsPtrs[1]))->Value;
  XxxADTClass* Arg2ADT = XxxADTArray[SerializedArgs[2].TypeId];
  if (Arg2ADT->SerializeCopy(ValueEnv->ArgValsPtrs[2],
			     SerializedArgs[2].ArgMetaInfo,
			     SerializedArgs[2].MemArea) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADOPT, "Error in serialize copy");
    return XXX_NOT_OK;
  }
  ((IntValue *)ReturnValue)->Value = 
    ComputeTestFunc(Arg0, Arg1, SerializedArgs[2].MemArea, SerializedArgs[2].MemSize);
  return XXX_OK;
}

