#ifndef XXX_GENFUNC_H
#define XXX_GENFUNC_H
#include "types.h"
#include "generic.h"
#include "basictypes.h"
#include "adtfunc.h"

/*
 * for TRIM fuction.
 */
//Bjarke, aug. 3: int added as type for constants
const int XXX_TRIM_LEADING	= 0x10;
const int XXX_TRIM_TRAILING	= 0x01;
const int XXX_TRIM_BOTH	        = 0x11;

/**
 // This dummy class is the owner of all function calls that are not 
 // methods of some specific class. It is also the owner of all
 // expressions that span classes.
 */
class XxxGenFuncADTClass : public XxxCoreADTClass {
 public:
  class SimpleFuncPlanInfo : public XxxFuncPlanInfo {
    friend class XxxGenFuncADTClass;
  protected:
    /// execute matches
    XxxErrCode execMatches(XxxValueEnv* Env, 
			   XxxADTValue *ReturnValue) const;
    /// execute jmatches (implemented in Java)
    XxxErrCode execJMatches(XxxValueEnv* Env, 
			    XxxADTValue *ReturnValue) const;
    XxxErrCode execIntPlus(XxxValueEnv* Env, 
			   XxxADTValue *ReturnValue) const;
    XxxErrCode execMod(XxxValueEnv* Env, 
		       XxxADTValue *ReturnValue) const;

    // string functions
    XxxErrCode execLike(XxxValueEnv* Env, 
		        XxxADTValue *ReturnValue) const;
    XxxErrCode execNotLike(XxxValueEnv* Env, 
		        XxxADTValue *ReturnValue) const;
    XxxErrCode execConcat(XxxValueEnv* Env, 
		          XxxADTValue *ReturnValue) const;
    XxxErrCode execUcase(XxxValueEnv* Env, 
		         XxxADTValue *ReturnValue) const;
    XxxErrCode execLcase(XxxValueEnv* Env, 
		         XxxADTValue *ReturnValue) const;
    XxxErrCode execTrim(XxxValueEnv* Env, 
		        XxxADTValue *ReturnValue) const;
    XxxErrCode execLtrim(XxxValueEnv* Env, 
		         XxxADTValue *ReturnValue) const;
    XxxErrCode execRtrim(XxxValueEnv* Env, 
		        XxxADTValue *ReturnValue) const;

    // double functions
    XxxErrCode execAbs(XxxValueEnv* Env, 
		       XxxADTValue *ReturnValue) const;
    XxxErrCode execCeiling(XxxValueEnv* Env, 
		           XxxADTValue *ReturnValue) const;
    XxxErrCode execFloor(XxxValueEnv* Env, 
		         XxxADTValue *ReturnValue) const;
    XxxErrCode execPower(XxxValueEnv* Env, 
		         XxxADTValue *ReturnValue) const;
    XxxErrCode execSign(XxxValueEnv* Env, 
		        XxxADTValue *ReturnValue) const;

  public:
    SimpleFuncPlanInfo(const XxxFuncParseInfo* FuncParse, int argCount)
      : XxxFuncPlanInfo(FuncParse, argCount){}
  };

  // func plan with extra memory allocated
  // Used as a plan for the "testfunc" function
  class MemFuncPlanInfo : public XxxFuncPlanInfo {
    friend class XxxGenFuncADTClass;
  protected:
    struct ArgData {
      int TypeId;
      unsigned int MemSize;
      char* MemArea;
      XxxADTMetaInfo* ArgMetaInfo;
      ArgData() { SetEmpty();}
      ~ArgData() { 
	if (MemArea && MemSize) 
	  delete [] MemArea;
	MemArea = NULL;
      }
      void SetEmpty() { MemSize = 0; MemArea = 0; ArgMetaInfo = 0;}
      void SetSize(int size) { 
	MemSize = size; 
	MemArea = new char[MemSize];
      }
    };
    int ArgCount;
    struct ArgData *SerializedArgs;

    /// execute testfunc()
    XxxErrCode execTestFunc(XxxValueEnv* Env, 
			    XxxADTValue *ReturnValue) const;
  public:
    XxxBool Initialized;
  
    MemFuncPlanInfo(const XxxFuncParseInfo* FuncParse,
		    XxxGenericPlan* Exp,
		    XxxValExprPlanList* ArgPlans);
    virtual ~MemFuncPlanInfo();
  };

 protected:
  virtual XxxErrCode methOptimize(XxxFuncParseInfo* FuncParse,
				  XxxValueExprPlan* OwnerPlan,
				  XxxValExprPlanList* ArgPlans,
				  XxxFuncPlanInfo*& FPlan) const;
  //// helper method used by constructor
  void registerMethods();
  //// helper method used by constructor
  void registerRules();

 public:
  XxxGenFuncADTClass(XxxBasicType Id, const char* Name);
  virtual XxxErrCode FuncTypeCheck(const char*FuncName,
				   const XxxValueExpression* Owner,
				   const XxxValueExprList *ArgList,
				   XxxBool IsAggr,
				   XxxBasicType& ReturnType,
				   XxxFuncParseInfo*& FuncParse) const;
 };

#endif
