#ifndef XXX_ERROR_H
#define XXX_ERROR_H
#include <string.h>
#include <unistd.h>
#include <iostream.h>
#include "error_def.h"
#include "prd-config.h"

//   The Error mechanism works as follows:
//     Error codes returned by functions are always XXX_OK or XXX_NOT_OK.
//     As a special case, XXX_QUIT shuts down the system.
//
//     When an error code is returned, a global error condition is set, which
//     can be any of a number of error flags. The most recent error condition,
//     and the place where it was caused can be accessed at any time.
//     Each ADT defines its own error codes. Therefore each error code must
//     be stored along with the type causing the error. Global error conditions
//     belong to the GENFUNC pseudo-ADT. Look in its definition for the
//     global error cond text

const XxxErrCond XXX_BADJVM = 28;
const XxxErrCond XXX_BADDEV = 27;
const XxxErrCond XXX_BADHANDLE = 26;
const XxxErrCond XXX_BADSCAN = 25;
const XxxErrCond XXX_BADCURSOR = 24;
const XxxErrCond XXX_BADADT = 23;
const XxxErrCond XXX_BADTYPE = 22;
const XxxErrCond XXX_BADRECORD = 21;
const XxxErrCond XXX_BADFUNC = 20;
const XxxErrCond XXX_BADVALUE = 19;
const XxxErrCond XXX_BADEXPR = 18;
const XxxErrCond XXX_BADQUERY = 17;
const XxxErrCond XXX_BADCMD = 16;
const XxxErrCond XXX_BADEXEC = 15;
const XxxErrCond XXX_BADOPT = 14;
const XxxErrCond XXX_BADPARSE = 13;
const XxxErrCond XXX_SHOREERR = 12;
const XxxErrCond XXX_BADENG = 11;
const XxxErrCond XXX_BADSERVER = 10;
const XxxErrCond XXX_SETDUPL = 9;
const XxxErrCond XXX_OVFLO = 8;
const XxxErrCond XXX_BADREG = 7;
const XxxErrCond XXX_BADSOCK = 6;
const XxxErrCond XXX_BADPIPE = 5 ;
const XxxErrCond XXX_BADFILE = 4;
const XxxErrCond XXX_BADFLAG = 3;
const XxxErrCond XXX_INTERNAL = 2;
const XxxErrCond XXX_INCOMPL = 1;
const XxxErrCond XXX_ALLWELL = 0;

// Each user environment maintains a stack of XxxErrInfos that holds the
// current error stack
class XxxErrInfo {
 private:
  void printThis(ostream& OStr) const;
 public:
  int SourceEngine;
  XxxErrCond Condition;
  char Message[XXX_ERRMESSAGE_MAX];
  char FileName[XXX_ERRMESSAGE_MAX];
  int LineNumber;
  XxxErrInfo* Next;

  XxxErrInfo(XxxErrCond Condn, const char *Msg, int Engine,
	     const char *FileNm, int LnNum){
    Condition = Condn;
    strcpy(Message, Msg);
    SourceEngine = Engine;
    strcpy(FileName, FileNm);
    LineNumber = LnNum;
    Next = NULL;
   }
  XxxErrInfo() { Condition = XXX_ALLWELL; Next = NULL;}
  XxxErrCode Get(XxxErrInfo* Error);
  XxxErrCode Note(XxxErrCond Cond, const char *Message, int Engine,
		  const char* FileName, int LineNumber);
  void Clear();
  void Print(ostream& OStr, const char* Msg) const;
  XxxBool IsError() const { return (Next?XXX_TRUE:XXX_FALSE);}
 };

#endif  /* XXX_ERROR_H */
