#ifndef XXX_TYPES_H
#define XXX_TYPES_H

#include <iostream.h>
#include <strstream.h>
#include <string.h>
#include "error_def.h"
#include "prd-config.h"



/// The domain of integer identifiers given to system types (including ADTs)
typedef int XxxBasicType;

XXX_EXTERN_CLASS_DECL(smutex_t);
XXX_EXTERN_CLASS_DECL(XxxBoundingBox);     // look in BoundingBox.h 
XXX_EXTERN_CLASS_DECL(ValuesArray);        // look in record.h
XXX_EXTERN_CLASS_DECL(XxxValueExprList);
XXX_EXTERN_CLASS_DECL(XxxValueExpression);
XXX_EXTERN_CLASS_DECL(XxxValueEnv);
XXX_EXTERN_CLASS_DECL(XxxValueExprPlan);
XXX_EXTERN_CLASS_DECL(XxxValExprPlanList);
XXX_EXTERN_CLASS_DECL(XxxFuncParseInfo);   // a generic function parse structure
XXX_EXTERN_CLASS_DECL(XxxFuncPlanInfo);    // adtfunc.h
XXX_EXTERN_CLASS_DECL(XxxFuncMethodInfo);  // adtfunc.h
XXX_EXTERN_CLASS_DECL(XxxAggrMethodInfo);  // adtfunc.h

XXX_EXTERN_CLASS_DECL(XxxFuncExpression);  // funcexpr.h
XXX_EXTERN_CLASS_DECL(XxxAggrExpression);  // aggrexpr.h
XXX_EXTERN_CLASS_DECL(XxxAggrOpSpec);      // aggrexpr.h
XXX_EXTERN_CLASS_DECL(XxxAggrOpHandle);    // aggrexpr.h
XXX_EXTERN_CLASS_DECL(XxxOptRuleEngine);   // optrules.h
XXX_EXTERN_CLASS_DECL(XxxADTEnv);          // env.h
XXX_EXTERN_CLASS_DECL(XxxSetCursor);

extern XxxErrCode XxxFunctionReportAddMethod(XxxBasicType Type, XxxFuncMethodInfo *FMInfo);

// can find no better place to put this :-(
typedef int XxxIndexType;
// TODO:L2 This hardcoding should be removed
const XxxIndexType XXX_BTREE = 0;
//// Comparison categories for indexing
typedef enum { XXX_IDX_EQUALS,
	       XXX_IDX_GT,
	       XXX_IDX_GTE,
	       XXX_IDX_LT,
	       XXX_IDX_LTE,
	       XXX_IDX_MEETS,
	       XXX_IDX_OVERLAPS,
	       XXX_IDX_INSIDE,
	       XXX_IDX_COVERS,
	       XXX_IDX_COVERED_BY,
	       XXX_IDX_CONTAINS,
	       XXX_IDX_DISJOINT,
	       XXX_IDX_INVALID} XxxCompareOp;

#include "generic.h"

//// Wrapper class for type plus meta-info
typedef struct XxxFullType {
  XxxBasicType  TypeId; 
  XxxADTMetaInfo* MInfo;
} XxxFullType;

/**
 * Root class defining a value of an ADT. This is to be distinguished
 * from the XxxADTClass itself, which defines the behavior of the type.
 * Actual values in records are byte-copies of instances of the XxxADTValue
 * class or its subclasses.
 * ABSOLUTELY NO VIRTUAL METHODS IN THIS CLASS!! Instances of this class
 * are going to move from memory to disk WITHOUT swizzling.
 * This parent class is exactly 4 bytes large.
 * Most of the information (8 bits worth) is in the StatusFlag.
 */
class XxxADTValue {
public:
  /**
   * Set on NULL values.
   */
  static const char NULL_VALUE; 
  /**
   * Set on values which are incomplete due to suspended computation.
   */
  static const char SUSPENDED_COMPUTATION; 
  /**
   * Set on values which are in big-endian format.
   */

  static const char PRED_BIG_ENDIAN; 

  /**
   * Set on compressed values.
   */
  static const char COMPRESSED; 
public:
  char StatusFlag;
  char ExtraFlag;               // I think we could store the type id here
  unsigned short SuspensionId;  // 2 bytes for suspension id

  inline XxxBool Is(const char Flag) const { return (StatusFlag & Flag); }

  inline void Set(const char Flag) { StatusFlag |= Flag; }

  inline void Unset(const char Flag) { StatusFlag &= ~Flag; }

  XxxADTValue() {
    Init();
  }

  void Init();

  ~XxxADTValue() {} 
};

//// Not using this yet: Praveen
typedef struct XxxFullValue {
  XxxADTValue* Value; 
  XxxADTMetaInfo* MInfo;
} XxxFullValue;

//  typedefs useful for easy method definition
typedef XxxADTMetaInfo* (*XxxMetaInfoGen)(const XxxValueExpression* Owner,
					  const XxxValueExprList *ArgList);
typedef XxxErrCode (*XxxOptimizeFunc)(XxxFuncParseInfo* FParse,
				      XxxValueExprPlan* OwnerPlan,
				      XxxValExprPlanList* ArgPlans,
				      XxxFuncPlanInfo*& FuncPlan);
typedef XxxErrCode (XxxFuncPlanInfo::*XxxExecuteFunc)(XxxValueEnv* Env,
						      XxxADTValue *ReturnValue) const;
// typedefs needed for aggregates
typedef XxxErrCode (XxxFuncPlanInfo::*XxxAggrInitFunc)(XxxAggrOpHandle& Hndl) const;
typedef XxxErrCode (XxxFuncPlanInfo::*XxxAggrAccumFunc)(XxxValueEnv* Env,
						       XxxAggrOpHandle& Hndl) const;
typedef XxxErrCode (XxxFuncPlanInfo::*XxxAggrDropFunc)(XxxValueEnv* Env,
						      XxxAggrOpHandle& Hndl) const;
typedef XxxErrCode (XxxFuncPlanInfo::*XxxAggrTermFunc)(XxxAggrOpHandle& Hndl, 
						       XxxADTValue* FinalVal) const;
/**
 * The XxxADTClass represents a data type. Actual data of this type
 * will be manipulated solely via methods of this class.  This
 * gives a totally extensible type system.  We have special-cased
 * integers, doubles and strings. All other types will be ADTs.
 * Methods of XxxADTClass are all virtual.
 *
 * Every ADT is either a collection ADT or an atomic ADT.
 * The collection ADTs provide methods to obtain the
 * various component data items.
 * Both types of ADTs can register functions that can be
 * invoked. The semantics of this is procedural.
 * Both types of ADTs can also provide an expression language
 * in which declarative expressions can be specified. To make
 * this possible, the ADT must supply an expression parser,
 * an optimizer and an evaluation engine.
 * Each ADT also provides a catalog manager (to manage catalogs
 * for objects that belong to the ADT), and a data manager to
 * manage the objects.
 * Each ADT can optionally be a top-level data type, and can
 * maintain named objects that a user can directly manipulate.
 */
class XxxADTClass {
public:
  //// Latch is used to synchronize access to the type array
  static smutex_t* Latch;
  //// This is PREDATOR's type array; each ADT inserts itself into this
  static XxxADTClass* ADTArray[];
  //// The number of ADTs defined in the system
  static int NumADTs;

#define XxxADTArray XxxADTClass::ADTArray

  typedef XxxMetaInfoGen MetaInfoGen;
  typedef XxxOptimizeFunc OptimizeFunc;
  typedef XxxExecuteFunc ExecuteFunc;
  typedef XxxAggrInitFunc AggrInitFunc;
  typedef XxxAggrAccumFunc AggrAccumFunc;
  typedef XxxAggrDropFunc AggrDropFunc;
  typedef XxxAggrTermFunc AggrTermFunc;

  typedef XxxFuncMethodInfo FuncMethodInfo;
  typedef XxxAggrMethodInfo AggrMethodInfo;

  /// This serves as an index into the global table of ADTs
  const XxxBasicType TypeId;
  /// The name of the ADT
  char TypeName[XXX_ADT_NAME_MAX];
  /// An optional handle on the rule engine
  XxxOptRuleEngine *RuleEngine;

protected:
  /// The maximum size of an object (in record)
  const int maxObjectSize;
  // Method registry
  FuncMethodInfo** MethArray; 
  int NumMethods;
  int NumMethodSlots;
  // the following is allocated in MethArray additionally to NumMethods
  // for dynamic extensions (TODO:TOBIAS: per user/per ADT dynamic tables)
  const int NUM_SPARE_METHOD_SLOTS;
  AggrMethodInfo** AggrMethArray; 
  int NumAggrMethods;

public:
    //// Add a methodinfo to the end of the array, does not replicate 
  virtual XxxErrCode XxxADTClass::AddMethod(FuncMethodInfo* NewMethod, int &ItsIndex)
  {
    ASSERT((NumMethods <= NumMethodSlots))
    // don't do it if method slots are exhausted already 
    if(NumMethods == NumMethodSlots)
      return XXX_NOT_OK;
    MethArray[NumMethods] = NewMethod;
    ItsIndex = NumMethods++;
    XxxFunctionReportAddMethod(TypeId, NewMethod);
    return XXX_OK;
  }

  //// Get the methodinfo for a method by its index in meth array
  virtual FuncMethodInfo* GetMethByIndex(int index) const { 
    if (index <0 || index >= NumMethods)
      return NULL;
    else 
      return MethArray[index];
  }
  //// Get the number of methods
  virtual int GetNumMethods() const { return NumMethods; }

  // Get aggrmethinfo for an aggregate method by its index in
  // aggr meth array
  virtual AggrMethodInfo* GetAggrMethByIndex(int index) const { 
    if (index <0 || index >= NumAggrMethods) // was NumMethods, cngd by Tobias
      return NULL;
    else 
      return AggrMethArray[index];
  }

  /**
   * Maximum "size" of any object of this type. This size is the number of
   * bytes needed to store it in a record. If the objects are really large,
   * they will not be stored in the record. Instead, they will be stored in
   * separately, and an ID for them will be stored in the records. The object
   * size in this case is the size of the id.
   */
  virtual int MaxObjectSize(const XxxADTMetaInfo*) const 
   { return maxObjectSize;}
protected:
  /// XXX_TRUE if the type requires meta information
  XxxBool maintainsMetaInfo;
public:
  /**
   * Check whether this type maintains metainfo. Used by 
   * RelShoreDataSrc::RelShoreCatalog::getAttrInfo to determine whether it
   * should retrieve metainfo for this ADT.
   * Alias ADT will overwrite it, and return the maintainsMetaInfo field
   * of the ADT it aliases to.
   */
  virtual XxxBool MaintainsMetaInfo()
  { return maintainsMetaInfo; }

  /**
   * Check whether the meta info has been stored for this ADT. 
   * When an alias is created, and if the type it aliases to maintains metainfo,
   * there're two options, 
   *    1: when the alias is created, meta string is specified.
   *    2: meta string is specified when a table is created, and 
   *       this alias is the type of one of its attributes. 
   * "SpecifiedMetaInfo" is used to support case (1), because when a 
   * table is created, with this alias as the type of one attribute, 
   * metainfo for the type to which the alias aliases
   * @return XXX_TRUE if the ADT needs cats info, and that's not 
   * stored yet. return XXX_FALSE if ADT doen't need meta info, or, 
   * it needs meta info, but has already stored it.
   */
  virtual XxxBool SpecifiedMetaInfo()
  { return maintainsMetaInfo;}

  /**
   * Used to check whether the defined type is an alias.
   */
  virtual XxxBool IsAlias(){ return XXX_FALSE;}
  /**
   * Should a type object be stored in its record field, or should it be placed
   * at the end of the record with a pointer to it instead?
   */
  XxxBool StoreInField;      
  
  /**
   * Constructor.  
   */
  XxxADTClass(XxxBasicType Id, const char* Name, int ObjSize,
	      XxxBool InField, XxxBool MaintainsCat);

  /**
   * Destructor.
   */
  virtual ~XxxADTClass();

  /**
   * Check whether the new data type with "Name" exists in the system ADT array
   */
  static XxxErrCode CheckName(XxxBasicType typeId, const char *Name);

  /**
   * Register a record about this new type into "TypeListFile" in the defdisk
   */
  virtual XxxErrCode Register(){return XXX_OK;}
  /**
   * Determines if two ADTClasses are identical.
   */
  int operator==(const XxxADTClass& Other) { 
    return (!strcmp(Other.TypeName, TypeName));
   }
  /**
   * Are the values "small" i.e., stored totally within the record?
   */
  virtual XxxBool IsSmall(const XxxADTMetaInfo*) const 
   { return XXX_TRUE; }

  /**
   * Prints the name of the ADT.
   */
  virtual void Print(ostream& OStr, const XxxADTMetaInfo* MetaInfo) const;

  /**
   * Prints any statistics maintained by the ADT.
   */
  virtual void PrintStats(ostream& OStr) const {
    OStr << TypeName << ":  No Stats Maintained" << endl;
  }

  /**
   * Resets any statistics maintained by the ADT.
   */
  virtual void ResetStats() {}

  /**
   * Returns the size of a particular object of this type when stored 
   * in a tuple.
   */
  virtual int TypeSize(const XxxADTValue* /* Object */, 
		       const XxxADTMetaInfo* MetaInfo) const
   { return MaxObjectSize(MetaInfo);}

  /**
   * Extracts an object of this type from the data string into the
   * preallocated result object. increments ref count, if maintained
   * @return The size of the resultant object
   */
  virtual int TypeCopy(const XxxADTValue* SrcObject, 
		       XxxADTValue *ResultObject,
		       const XxxADTMetaInfo* MetaInfo) const 
  { 
    int Size = TypeSize(SrcObject, MetaInfo);
    memcpy((char*)ResultObject, (char *)SrcObject, Size);
    return Size;
  }
  
  /**
   * A method to determine if two objects of this type are identical.
   * TODO: this method and lt and gt need metainfos
   */
  virtual XxxBool  Equals(const XxxADTValue* Obj1, const XxxADTMetaInfo*, 
			  const XxxADTValue* Obj2, const XxxADTMetaInfo*)  const 
  { return XXX_FALSE;}

  /**
   * All types which are subsets of ordered domains will return XXX_TRUE for
   * this method.
   */
  virtual XxxBool OrderedType(const XxxADTMetaInfo*) const 
   { return XXX_FALSE;}

  /**
   * Compares two objects of this type.
   * @return XXX_TRUE if the first argument is "less than" the second
   * note: only valid for ordered types
   */
  virtual XxxBool  LessThan(const XxxADTValue* Val1, 
			    const XxxADTMetaInfo*,
			    const XxxADTValue* Val2,
			    const XxxADTMetaInfo*) const 
   { return XXX_FALSE;}
  /**
   * Compares two objects of this type.
   * @return XXX_TRUE if the first argument is "greater than" the second
   * note: only valid for ordered types
   */
  virtual XxxBool  GreaterThan(const XxxADTValue* Val1, 
			       const XxxADTMetaInfo*,
			       const XxxADTValue* Val2,
			       const XxxADTMetaInfo*) const
   { return XXX_FALSE;}

  /**
   * Makes a copy that can be stored (in memory representations need to
   * be converted to disk representations). Not curently used.
   */
  virtual int TypeStoreCopy(const XxxADTValue* SrcObject, 
			    XxxADTValue *ResultObject,
			    const XxxADTMetaInfo* MetaInfo) const 
    { return TypeCopy(SrcObject, ResultObject, MetaInfo); }

  /**
   * Makes a copy that can be shipped over a network (i.e. big-endian)
   * Each ADT must implement this -- default returns error.
   * The core types wont implement it because the function never gets
   * called for them (it is special-cased for them)
   */
  virtual XxxErrCode CopyHToN(const XxxADTValue* SrcObject, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const = 0;
  virtual XxxErrCode CopyNToH(const XxxADTValue* SrcObject, 
			      XxxADTValue *ResultObject,
			      const XxxADTMetaInfo* MetaInfo) const = 0;

  /**
   *
   */
  virtual char* TypeCreate(const XxxADTValue* SrcObject, 
			   const XxxADTMetaInfo* MetaInfo) const {
    char* Tmp = new char[TypeSize(SrcObject, MetaInfo)];
    TypeCopy(SrcObject, (XxxADTValue *)Tmp, MetaInfo);
    return Tmp;
  }

  /**
   * Destroys an object of the type when stored in a tuple.
   * This could involve decrementing ref counts, etc.
   */
  virtual void TypeDestroy(XxxADTValue *Object, 
			   const XxxADTMetaInfo* MetaInfo) const {};

  /**
   * A collection of methods for dispatching objects to external
   * processes by serializing them
   */
  /**
   * This assumes that an upperbound on the size can be determined.
   * Otherwise, return an error.
   */
  virtual XxxErrCode SerializeCopySize(const XxxADTMetaInfo* MetaInfo,
				       unsigned int& Size) const;
  /**
   * Make a serialized copy
   */
  virtual XxxErrCode SerializeCopy(const XxxADTValue* Object,
				   const XxxADTMetaInfo* MetaInfo,
				   char* MemLocation) const;
  /**
   * Copy a serialized copy into a normal object
   */
  virtual XxxErrCode UnSerializeCopy(const char* MemLocation,
				     const XxxADTMetaInfo* MetaInfo,
				     XxxADTValue* Object) const;

  /**
   * Assigns a "default" value, if any exists.
   */
  virtual XxxErrCode DefaultVal(XxxADTValue* ObjectRef, 
				const XxxADTMetaInfo* MetaInfo) const;

  /**
   * Makes a readable text string version of the object from the tuple
   * representation.
   */
  virtual XxxErrCode WriteText(ostream& OutStream, 
			       const XxxADTValue* Object,
			       const XxxADTMetaInfo* MetaInfo) const = 0;
  /**
   * Reads a text version of the object into the tuple representation of it,
   * after memory has been allocated for the tuple representation.
   */
  virtual XxxErrCode ReadText(istream& InStream, 
			      XxxADTValue* ObjectRef,
			      const XxxADTMetaInfo* MetaInfo) const = 0;

  /**
   * This structure is used to store meta-information for complex types.
   * For example, the audio ADT stores information on the format of the 
   * audio source, the sampling rate, etc.
   */
  struct MetaInfoHandle {
    // Size MUST be the first field, and Data MUST be the last!
    // Do not interchange! 
    // Each ADT can define a stats or meta-info structure that will
    // be stored in the Data area. Since this structure needs to be
    // double word-aligned, add padding integer field
    unsigned int Size;
    unsigned int Padding;
    char Data[XXX_MIHNDL_MAX];
    MetaInfoHandle() { Size = 0; }
    MetaInfoHandle(const MetaInfoHandle& MIH) {
      Size = MIH.Size;
      if (Size > 0) memcpy(Data, MIH.Data, MIH.Size);
     }
    int CopySize() const { return sizeof(Size) + sizeof(Padding) + Size;}
   };

  static const struct MetaInfoHandle NullMIHndl;

  virtual XxxErrCode AddMetaInfo(const char* Name, 
				 const XxxBool Named,
				 const char* MetaInfoString,
				 MetaInfoHandle& MIHandle);
//Bjarke, aug. 8: const added to char* parameter
  virtual XxxErrCode GetMetaInfo(const char* ObjName,
				 const MetaInfoHandle& MIHandle,
				 XxxADTMetaInfo* &MetaInfo,
				 XxxBool& Found) const;

  /**
   * Must return XXX_TRUE for ADTs which wish to implement callbacks.
   */
  virtual XxxBool SupportsCallback() const
  { return XXX_FALSE; }
  /**
   * This function will only be called on ADTs which support callbacks.
   * Given a handle on the result, this function must
   * determine the size (in bytes) of the response it will send in a 
   * call to the ServiceCallback() function.
   * @param ObjectRef The object for which a callback is being made
   * @param MetaInfo The meta info for that object
   * @param Message A binary message that was sent with the callback
   *                (or NULL if none was sent)
   * @param MessageSize The size of the Message array
   * @param ResultSize (out) The size of the object to be sent
   */
  virtual XxxErrCode GetResultSize(const XxxADTValue *ObjectRef,
				   const XxxADTMetaInfo *MetaInfo,
				   const char *Message,
				   const int MessageSize,
				   int &ResultSize) const
  { return XXX_NOT_OK; }
  /**
   * This function will only be called on ADTs which support callbacks.
   * Any data type which would like to take advantage of browser capability
   * to handle some data specially, for instance, displaying images, must
   * return a valid Content Type string as described in RFC 1590.
   */
  virtual const char * GetContentType(const XxxADTValue *ObjectRef,
				      const XxxADTMetaInfo *MetaInfo,
				      const char *Message,
				      const int MessageSize) const
  { return "application/octet-stream"; }
  /**
   * This function will only be called on ADTs which support callbacks.
   * This is the sister-function to the Callback function in the Java
   * Client Library.  It communicates with the JCL and transfers the 
   * "full value" of a data object over the stream provided.
   * @param RequestBuffer The request message sent from the JCL
   * @param BufferSize The buffer's size
   * @param OStr The output stream to the JCL data type
   */
  virtual XxxErrCode ServiceCallback(const XxxADTValue *ObjectRef,
				     const XxxADTMetaInfo *MetaInfo,
				     const char *Message,
				     const int MessageSize,
				     ostream &OStr) const 
  { return XXX_NOT_OK; }

  /**
   * The class which defines statistics for each ADT.
   */
  class TypeStats {
   public:
    TypeStats () {}
    virtual ~TypeStats() {}
    virtual int Size(const XxxADTMetaInfo* MetaInfo) const = 0;
    virtual XxxErrCode Accum(const XxxADTValue* Object, 
			     const XxxADTMetaInfo* MetaInfo) = 0;
    virtual XxxErrCode Term(const XxxADTMetaInfo* MetaInfo) = 0;
    //// allocate and copy a replica
    virtual XxxErrCode Replicate(const XxxADTMetaInfo* MetaInfo, 
				 TypeStats*& NewSI) const=0;
    //// copy a replica into preallocated space
    virtual XxxErrCode Copy(const XxxADTMetaInfo* MetaInfo, 
			    TypeStats* NewSI) const=0;
    virtual void Print(ostream& OStr, 
		       const XxxADTMetaInfo* MetaInfo) const = 0;  
   };

  /**
   * The class which defines implementation for each ADT.
   */
  class TypeImpl {
   public:
    TypeImpl () {}
    virtual ~TypeImpl() {}
    //// allocate and copy a replica
    virtual XxxErrCode Replicate(const XxxADTMetaInfo* MetaInfo, 
				 TypeImpl*& NewImpl) const=0;
    //// copy a replica into preallocated space
    virtual XxxErrCode Copy(const XxxADTMetaInfo* MetaInfo, 
			    TypeImpl* NewImpl) const=0;
    virtual void Print(ostream& OStr, 
		       const XxxADTMetaInfo* MetaInfo) const = 0;  
   };

  /**
   * Creates a new StatsInfo object for this ADT. Redefine this
   * in a subclass that wants stats (along with makeStatsInfo)

   */
  virtual XxxErrCode CreateStatsInfo(const XxxADTMetaInfo *MetaInfo,
				     TypeStats*& StatsInfo) const
   { StatsInfo = NULL; return XXX_OK;}

  /**
   * Stores the in-memory StatsInfo into a MIHndl MIBLOB
   */
  virtual XxxErrCode StoreStatsInfo(const XxxADTMetaInfo *MetaInfo,
				    const TypeStats* StatsInfo,
				    MetaInfoHandle& MIHandle);
  /**
   * Retrieves the StatsInfo from a meta-info handle on a BLOB ADT
   */
  virtual XxxErrCode RetrieveStatsInfo(const XxxADTMetaInfo *MetaInfo,
				       const MetaInfoHandle& MIHandle,
				       TypeStats*& StatsInfo);

  /**
   * Function that provides type checking for allowed methods
   * of the ADT. During type checking, the return parameter is
   * ReturnType.
   *
   * To allow for variable methods (methods with the same name
   * but either with different type signatures), the actual
   * expression list must also be passed. For such cases, it is
   * usually sufficient in PL to pass the list of argument _types_.
   * However, in our case, we must also support function invocations
   * in which the first argument is an expression string, and the
   * function type signature depends on what is in the string.
   * Therefore, the actual list of argument expressions needs to
   * be passed in.
   */
  virtual XxxErrCode FuncTypeCheck(const char*FuncName,
				   const XxxValueExpression* Owner,
				   const XxxValueExprList *ArgList,
				   XxxBool IsAggr,
				   XxxBasicType& ReturnType,
				   XxxFuncParseInfo*& FuncParse) const;

  /**
   * Function that does any reorganization of a function expression
   * usually by traversing its expression tree and applying rules to
   * the funcexpressions
   */
  virtual XxxErrCode FuncReOrganize(XxxValueExpression*& FuncExpression) const;
  
  /**
   * Function that optimizes function expressions into execution plans
   */
  virtual XxxErrCode FuncOptimize(XxxFuncParseInfo* FuncParse,
				  XxxValueExprPlan* OwnerPlan,
				  XxxValExprPlanList* ArgPlans,
				  XxxFuncPlanInfo*& FuncPlan) const;

  /**
   * Determines if this ADT can cast an object of another type into
   * its own type.
   * @param SourceType The type to cast from
   * @param SourceMetaInfo The meta info of the object to be casted
   * @param TargetMetaInfo The meta info of the new object to be created
   */
  virtual XxxBool CanCastFrom(XxxBasicType SourceType, 
			      const XxxADTMetaInfo* SourceMetaInfo,
			      const XxxADTMetaInfo* TargetMetaInfo) 
  { return XXX_FALSE;}
  /**
   * Executes a cast from another type into this type.
   * @param CastType The type id of the old object
   * @param SourceMetaInfo The meta info of the object to be casted
   * @param TargetMetaInfo The meta info of the new object to be created
   * @param OldObject The object to be casted
   * @param NewObject The new object to be created
   */
  virtual XxxErrCode CastFrom(XxxBasicType SourceType, 
			      const XxxADTMetaInfo* SourceMetaInfo,
			      const XxxADTMetaInfo* TargetMetaInfo,
			      XxxADTValue* OldObject, 
			      XxxADTValue* NewObject) 
  { return XXX_NOT_OK;}
  /**
   * Determines if this ADT can cast an object of its type into
   * another type.
   * @param TargetType The type to cast to
   * @param SourceMetaInfo The meta info of the object to be casted
   * @param TargetMetaInfo The meta info of the new object to be created
   */
  virtual XxxBool CanCastTo(XxxBasicType TargetType, 
			    const XxxADTMetaInfo* SourceMetaInfo,
			    const XxxADTMetaInfo* TargetMetaInfo) 
  { return XXX_FALSE;}
  /**
   * Executes a cast from this type into another type.
   * @param CastType The type id of the new object
   * @param SourceMetaInfo The meta info of the object to be casted
   * @param TargetMetaInfo The meta info of the new object to be created
   * @param OldObject The object to be casted
   * @param NewObject The new object to be created
   */
  virtual XxxErrCode CastTo(XxxBasicType TargetType, 
			    const XxxADTMetaInfo* SourceMetaInfo,
			    const XxxADTMetaInfo* TargetMetaInfo,
			    XxxADTValue* OldObject, 
			    XxxADTValue* NewObject) 
  { return XXX_NOT_OK;}

  // Each ADT may maintain stats on a per user basis.
  // The env encapsulates all such stats
  /**
   * Creates an ADT environment.
   * Each ADT maintains statistics on a per-user basis.  The environment
   * for that ADT maintains such statistics.
   */
  virtual XxxErrCode CreateEnv( XxxADTEnv* &Env ) const;

  //// Reclaims an ADT environment.
  virtual XxxErrCode DeleteEnv( XxxADTEnv* Env ) const;

  /**
   * Gets a bounding box for an object of this type.
   * @return XXX_FALSE if this is not possible
   */
  virtual XxxBool GetBoundingBox(const XxxADTValue* &Object, 
				 XxxBoundingBox& Box) 
    { return XXX_FALSE;}

  /**
   * Gets a comparison operator for this type
   */
  virtual XxxErrCode GetCompOp(const char* FuncName, 
			       XxxCompareOp& CompOp, XxxBool& Success) const
  { Success = XXX_FALSE; return XXX_OK;}

 protected:
    // any subclass can override this
  virtual XxxErrCode methOptimize(XxxFuncParseInfo* FuncParse,
				  XxxValueExprPlan* OwnerPlan,
				  XxxValExprPlanList* ArgPlans,
				  XxxFuncPlanInfo*& FPlan) const ;
  // just match the method name, arg types and arity
  XxxErrCode matchMethodName(const char* FuncName,
			     const XxxValueExpression* Owner,
			     const XxxValueExprList *ArgList,
			     XxxBool IsAggr,
			     int& MethodIndex, // index in array of methods
			     XxxBasicType& ReturnType,
			     XxxADTMetaInfo*& ReturnMetaInfo,
			     XxxBool& Success) const;

  virtual XxxErrCode parseMetaInfo(const char* MetaInfoString,
				   MetaInfoHandle& MIHandle) const;
  virtual XxxErrCode makeMetaInfo(const MetaInfoHandle& MIHandle,
				  XxxADTMetaInfo*& MI) const;
  // to maintain stats, define CreateStatsInfo and this method
  /**
   * Create a stats info out of MIHandle.
   */
  virtual XxxErrCode makeStatsInfo(const XxxADTMetaInfo *MetaInfo,
				   const MetaInfoHandle& MIHandle,
				   TypeStats*& StatsInfo) const {
    return XXX_NOT_OK;
   }
 };

/**
 * This is the superclass for ADT meta-information.
 */
class XxxADTMetaInfo { 
 public:
  // classes to help parse meta-info
  class Itemizer { 
   public:
    const char* const Separator;
    char* CurrentItem;
    char* CurrentString;
    Itemizer(const char* Sep) : Separator(Sep) {}
    void InitScan(XxxSetCursor& SCsr, const char* InputString);
    void NextItem(XxxSetCursor& SCsr, char*& ItemString);
    void CloseScan(XxxSetCursor& SCsr);
   };
  class StringTokenizer { 
   public:
    char ItemString[50]; // bound each token to 50 chars
    istrstream* InStr;
    void InitScan(XxxSetCursor& SCsr, const char* InputString);
    void NextItem(XxxSetCursor& SCsr, char*& ItemString);
    void CloseScan(XxxSetCursor& SCsr);
   };
    
  /// Constructor
  XxxADTMetaInfo() {}
  /// Destructor
  virtual ~XxxADTMetaInfo() {}
  /**
   * Makes a copy of this meta-info object
   */
  virtual XxxErrCode Replicate(XxxADTMetaInfo*& NewMI) const = 0;
  /**
   * Prints this meta-info object to the output stream
   */
  virtual void Print(ostream& OStr) const = 0;
  /**
   * Checks if two meta-infos are compatible (needed in order
   *  to process type-cast constants being inserted into a column
   *  of an existing table)
   */
  virtual XxxBool IsCompatible(const XxxADTMetaInfo* OtherMI) const = 0;
  /**
   * Returns the number of bytes necessary to hold the serialized version
   * of this meta-info object.
   * @param Size (out) The required number of bytes
   * @return XXX_NOT_OK if the object cannot be serialized, XXX_OK otherwise
   */
  virtual XxxErrCode SerializeCopySize(int &Size) const = 0;
  /**
   * Creates a binary version of this meta-info object.
   * @param Data An array of bytes of at least the size specified by the
   *             SerializeCopySize() function.
   * @return XXX_NOT_OK if the object cannot be serialized, XXX_OK otherwise
   */
  virtual XxxErrCode SerializeCopy(char *Data) const = 0;
  
  /**
   * Now four virtual methods to be used by fancy ADTs
   */
  //// add in the statistics info
  virtual XxxErrCode AbsorbStatsInfo(XxxADTClass::TypeStats* Stats) 
  { return XXX_OK; }
  //// add in the implementation info
  virtual XxxErrCode AbsorbImplInfo(XxxADTClass::TypeImpl* Impl)
  { return XXX_OK; }
  //// get the statistics info
  virtual XxxADTClass::TypeStats* GetStatsInfo() const
  { return NULL; }
  //// get the implementation info
  virtual XxxADTClass::TypeImpl* GetImplInfo() const
  { return NULL; }
};

//
// some useful functions
//
#include <sys/types.h>
#include <netinet/in.h>

inline void XxxNetworkInt(int* NValue, int Value)
{
  unsigned long LTemp = htonl(Value);
  int IVal = LTemp;
  memcpy((char *)NValue, (char *)&IVal, sizeof(IVal));
}
inline void XxxNetworkLong(long* NValue, long Value)
{
  unsigned long LTemp = htonl(Value);
  memcpy((char *)NValue, (char *)&LTemp, sizeof(LTemp));
}
inline void XxxNetworkShort(short* HValue, short Value)
{
  unsigned short STemp = htons(Value);
  memcpy((char *)HValue, (char *)&STemp, sizeof(STemp));
}
inline void XxxNetworkDouble(double* NValue, double Value)
{
  // TODO: This assumes all architectures use IEEE floating point
  //       standard.
#ifdef XXX_LITTLEENDIAN
  // Reverse all 8 bytes
  for (int i = 0; i < sizeof(double); i++)
    ((char *)NValue)[i] = ((char *)(&Value))[sizeof(double) - 1 - i];
#else
  memcpy((char *)NValue, (char*)&Value, sizeof(double));
#endif
}
inline void XxxHostInt(int* HValue, int Value)
{
  unsigned long LTemp = ntohl(Value);
  int IVal = LTemp;
  memcpy((char *)HValue, (char *)&IVal, sizeof(IVal));
}
inline void XxxHostLong(long* HValue, long Value)
{
  unsigned long LTemp = ntohl(Value);
  memcpy((char *)HValue, (char *)&LTemp, sizeof(LTemp));
}
inline void XxxHostShort(short* HValue, short Value)
{
  unsigned short STemp = ntohs(Value);
  memcpy((char *)HValue, (char *)&STemp, sizeof(STemp));
}
inline void XxxHostDouble(double* HValue, double Value)
{
  // TODO: This assumes all architectures use IEEE floating point
  //       standard.
#ifdef XXX_LITTLEENDIAN
  // Reverse all 8 bytes
  for (int i = 0; i < sizeof(double); i++)
    ((char *)HValue)[i] = ((char *)(&Value))[sizeof(double) - 1 - i];
#else
  memcpy((char *)HValue, (char*)&Value, sizeof(double));
#endif
}

#include "basictypes.h"

#endif  /* XXX_TYPES_H */
