    #include <iostream.h>
    #include <stdlib.h>
    #include <string.h>
    
    #include "protocol.h"
    #include "shoresm.h"
    #include "socket.h"
    #include "sockstream.h"
    
    // 
    // Methods of XxxClientProtocol
    //
    XxxErrCode XxxClientProtocol::initTokenScan(XxxSetCursor& Csr, 
    					    const char* TokenString)
    {
     tokenScanState* TSS = new tokenScanState(TokenString);
     Csr.State = TSS;
     Csr.AtEnd = XXX_FALSE;
     Csr.NumberRead = 0;
     return XXX_OK;
    }
    XxxErrCode XxxClientProtocol::endTokenScan(XxxSetCursor& Csr)
    {
     tokenScanState* TSS = (tokenScanState *)(Csr.State);
     delete TSS;
     return XXX_OK;
    }
    XxxErrCode XxxClientProtocol::getNextToken(XxxSetCursor& Csr, 
    					   char& FlagChar, char*& Token)
    {
     tokenScanState* TSS = (tokenScanState *)(Csr.State);
     TSS->InStr >> TSS->FlagBuf;
     if (TSS->InStr.good()) {
       if (TSS->FlagBuf[0] != '-') {
         cerr << "Bad client setup flag " << TSS->FlagBuf << endl;
         return XXX_NOT_OK;
        }
       FlagChar = TSS->FlagBuf[1];
       TSS->InStr >> TSS->Buf;
       if (!TSS->InStr.bad()) {
         Token = TSS->Buf; 
         Csr.NumberRead++;
         // warning : we are not "copying" out the value.
         // so it is usable only till the next getNextToken() call
         return XXX_OK;
        }
       else return XXX_NOT_OK;
      }
     else if (TSS->InStr.eof()) {
       Csr.AtEnd = XXX_TRUE;
       return XXX_OK;
      }
     else {
       cerr << "Something strange: bad token string" << endl;
       return XXX_NOT_OK;
      }
    }
    
    XxxErrCode XxxClientProtocol::OpenConnection()
    {
     return XXX_OK;
    }
    XxxErrCode XxxClientProtocol::CloseConnection()
    {
     // to signal clean shutdown to client
     if (XXX_THISENV().OutputStr->good()) {
       *(XXX_THISENV().OutputStr) << "\n" << (char)0 << flush;
      }
    
     if (XXX_THISENV().OutputStr->bad())
       return XXX_NOT_OK;
     else
       return XXX_OK;
    }
    XxxErrCode XxxClientProtocol::CloseRequest()
    {
     // terminate with a newline and a null character
     *(XXX_THISENV().OutputStr) << "\n" << (char)0;
     *(XXX_THISENV().ErrorStr) << flush;
     *(XXX_THISENV().OutputStr) << flush;
    
     if (!XXX_THISENV().OutputStr->good()) {
       return XXX_NOT_OK;
      }
     else
       return XXX_OK;
    }
    XxxErrCode XxxClientProtocol::PrefixResults(ostream* OutputStr,
    					    XxxRecordSchema* AnswerSchema)
    {
     return XXX_OK;
    }
    XxxErrCode XxxClientProtocol::WriteRecord(ostream* OutputStr,
    					  const XxxRecordSchema* Schema,
    					  const XxxRecord* Record)
    {
      Record->Print(*(OutputStr), (XxxRecordSchema*)Schema,
                    BinaryPrint);
    
     return XXX_OK;
    }
    
    XxxErrCode XxxClientProtocol::WriteNoError(ostream &OutputStr)
    {
      return XXX_OK;
    }
    
    XxxErrCode XxxClientProtocol::WriteErrorStack(ostream &OutputStr,
    					      const XxxErrInfo &Top,
    					      const char *ErrorMessage)
    {
      Top.Print(OutputStr, ErrorMessage);
      OutputStr << flush;
      if (!OutputStr.good()) {
        XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot write error message to client");
        return XXX_NOT_OK;
      }
      return XXX_OK;
    }
    
    // 
    // Methods of XxxJavaClientProtocol
    //
    
    XxxJavaClientProtocol::XxxJavaClientProtocol(XxxClientInfo& CInfo) :
      XxxClientProtocol(CInfo), 
      MCConnection(NULL), MCPortNumber(0)
    {
#if defined(XXX_WIN32) && defined(XXX_NONPREEMPTIVE)
	  MCEventHandle = NULL;
#else
	  MCFileHandle = NULL;
#endif
      BinaryPrint = XXX_TRUE;
      SchemaPrint = XXX_TRUE;
    }
    
    XxxJavaClientProtocol::~XxxJavaClientProtocol() {
      if (MCConnection)
        delete MCConnection;
#if defined(XXX_WIN32) && defined(XXX_NONPREEMPTIVE)
	  if (MCEventHandle)
        delete MCEventHandle;
#else
      if (MCFileHandle)
        delete MCFileHandle;
#endif

    }
    
    XxxErrCode XxxJavaClientProtocol::OpenConnection()
    {
      ostream& OutputStr = *(ClientInfo.ThisThread->XxxEnv.OutputStr);
      int JavaADTConsSize;
      char *JavaADTCons;
      
      // first write the client ID number
      unsigned long ClientId = htonl(XXX_CLIENTINFO()->UserId);
      OutputStr.write((char *) (&ClientId), sizeof(unsigned long));
      if (XXX_THISENV().DebugFlags.Protocol) {
        cerr << "Telling Java client its ID: " << XXX_CLIENTINFO()->UserId << endl;
      }
      {
        // then write the number of ADT's
        unsigned long NumADTs = htonl(XxxADTClass::NumADTs);
        OutputStr.write((char *) (&NumADTs), sizeof(unsigned long));
        int i;
    
        XxxUseMutex AutoMutex(*XxxADTClass::Latch);
        for (i = 0; i < XxxADTClass::NumADTs; i++) {
          // write the Type Id
          unsigned long Id = htonl(i);
          OutputStr.write((char *) (&Id), sizeof(unsigned long));    
          JavaADTCons = (XxxADTClass::ADTArray[i])->TypeName;
          // write the Type Name length
          JavaADTConsSize = strlen(JavaADTCons);
          unsigned long LTemp = htonl(JavaADTConsSize);
          OutputStr.write((char *) (&LTemp), sizeof(unsigned long));
          // now write the Type Name
          OutputStr.write(JavaADTCons, JavaADTConsSize);
        }
        OutputStr << flush;
      }
    
      // Open a socket for meta-communication on a random port
      while (1) {
        MCPortNumber = (rand() % 5000) + 5000;
        MCConnection = new XxxSocketConnection(MCPortNumber);
        if (MCConnection->Valid()) {
          break;
        } else {
          delete MCConnection;
        }
      }
      if (XXX_THISENV().DebugFlags.Protocol) {
        cerr << "Java Client Protocol: Bound port " << MCPortNumber
    	 << " for meta-communication." << endl << "Waiting for connection...";
      }
    
      // send the port number to the client
      unsigned long PortNum = htonl(MCPortNumber);
      OutputStr.write((char *) (&PortNum), sizeof(unsigned long));
      OutputStr.flush();
    
      // now wait (thread-safe) for the client to connect to it
#if defined(XXX_WIN32) && defined(XXX_NONPREEMPTIVE)
	  // SHORE V2
	  handle = CreateEvent(NULL, TRUE, FALSE, NULL);
	  if (WSAEventSelect(MCConnection->socketDesc, handle, FD_ACCEPT|FD_READ) != 0)
	  {
		cerr << "Bad association between socket and Win32 event";
		return XXX_NOT_OK;
	  }
	  MCEventHandle = new win32_compat_event_t(handle);
#else
      MCFileHandle = new sfile_read_hdl_t(MCConnection->socketDesc);
#endif

#if defined(XXX_WIN32) && defined(XXX_NONPREEMPTIVE)
	  if (MCEventHandle->is_down() || (bool)MCEventHandle->wait(::WAIT_FOREVER)) 
#else
	  if (MCFileHandle->is_down()  || (bool)MCFileHandle->wait(::WAIT_FOREVER)) 
#endif
	  {
        XXX_NOTE_ERROR(XXX_BADEXEC, "Error waiting on secondary connection");
        return XXX_NOT_OK;
       }

      XxxSocketConnection *SecondaryConn = MCConnection->AcceptConnection();
      if (!SecondaryConn) {
        XXX_NOTE_ERROR(XXX_BADEXEC, "Client's secondary connection was bad");
        return XXX_NOT_OK;
      }
      if (XXX_THISENV().DebugFlags.Protocol) {
        cerr << "done." << endl;
      }
      // Now close the original socket to free it up
      delete MCConnection;
      // reclaim the file handle
#if defined(XXX_WIN32) && defined(XXX_NONPREEMPTIVE)
	  if (MCEventHandle) {
        delete MCEventHandle;
        MCEventHandle = 0;
      }
#else
      if (MCFileHandle) {
        delete MCFileHandle;
        MCFileHandle = 0;
      }
#endif
    

      MCConnection = SecondaryConn;

#if defined(XXX_WIN32) && defined(XXX_NONPREEMPTIVE)
	   // MCEventHandle is initialized just before it is used - not clear why! 
#else	  
      MCFileHandle = new sfile_read_hdl_t(MCConnection->socketDesc);
#endif
	  
      // Make the client's error stream go to the secondary connection
      // Use a thread-safe output stream
      ClientInfo.ThisThread->XxxEnv.ErrorStr = 
        new XxxSockStream(MCConnection->socketDesc);
    
      return XXX_OK;
    }
    
    XxxErrCode XxxJavaClientProtocol::CloseConnection()
    {
      delete MCConnection;
#if defined(XXX_WIN32) && defined(XXX_NONPREEMPTIVE)
	  if (MCEventHandle) delete MCEventHandle;
#else
      if (MCFileHandle) delete MCFileHandle;
#endif
      MCConnection  = NULL;

#if defined(XXX_WIN32) && defined(XXX_NONPREEMPTIVE)
	  MCEventHandle = NULL;
#else
      MCFileHandle  = NULL;
#endif
      return XxxClientProtocol::CloseConnection();
    }
    
    XxxErrCode XxxJavaClientProtocol::CloseRequest()
    {
      if (XxxClientProtocol::CloseRequest() != XXX_OK) {
        XXX_NOTE_ERROR(XXX_BADCMD, "Error sending close request");
        return XXX_NOT_OK;
      }
      if (XXX_THISENV().DebugFlags.Protocol) {
        cerr << "Waiting for client to end transaction." << endl << flush;
      }
      
      while (1) {
        if (ClientInfo.WaitForInput()) {
          XXX_NOTE_ERROR(XXX_BADCMD, "Error waiting for termination");
          return XXX_NOT_OK;
        }
        char RecvBuf[3];
        if (::recv(ClientInfo.ThisThread->SocketDesc(), 
    	       RecvBuf, 3, 0) != 3) {
          XXX_NOTE_ERROR(XXX_BADCMD, "Error getting client communication");
          return XXX_NOT_OK;
        }
        if (!strncmp(RecvBuf, "END", 3)) {
          *(XXX_THISENV().OutputStr) << "OK" << flush;
          if (XXX_THISENV().DebugFlags.Protocol) {
    	cerr << "Client has ended transaction." << endl;
          }
          break;
        } else {
          cerr << "Client has sent something other than an \"END\" message." 
    	   << endl;
        }
      }
      return XXX_OK;
    }
    
    XxxErrCode XxxJavaClientProtocol::PrefixResults(ostream* OutputStr,
    						XxxRecordSchema* AnswerSchema)
    {
      unsigned long LTemp; // used for htonl
      // Prefix results with a sentinel byte (integer value 27)
      *(OutputStr) << (unsigned char)27;
	  *(OutputStr) << flush;
      // Send over the Xact Id, length first
      int XactIdLength = sizeof(XxxShoreXactId);
      XxxShoreXactId ClientXactId;
      if (XxxShoreSM->XactToId(XXX_THISTHREAD()->xct(), ClientXactId) != XXX_OK) {
        XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot determine Xact's ID");
        return XXX_NOT_OK;
      }
      LTemp = htonl(XactIdLength);
      OutputStr->write((char *)(&LTemp), sizeof(LTemp));
      OutputStr->write((char *)&(ClientXactId), XactIdLength);
	  *(OutputStr) << flush;
      // print the schema if necessary
      if (SchemaPrint) {
        XxxRecordSchema &Schema = *AnswerSchema;
        int Arity = Schema.Arity();
        int j;
        int AnswerTypeId, AttrNameLen, AttrOffset, MetaInfoSize;
        XxxADTMetaInfo *MetaInfo = NULL;
        char *buf;
        // send the schema over (arity first)
        LTemp = htonl(Arity);
        OutputStr->write((char *) (&LTemp), sizeof(LTemp));
        for (j = 0; j < Arity; j++) {
          // write the attr type id
          AnswerTypeId = Schema[j].Type;
          LTemp = htonl(AnswerTypeId);
          OutputStr->write((char *)(&LTemp), sizeof(LTemp));
          // write the attr name
          AttrNameLen = strlen(Schema[j].Name);
          LTemp = htonl(AttrNameLen);
          OutputStr->write((char *)(&LTemp), sizeof(LTemp));
          OutputStr->write(Schema[j].Name, AttrNameLen);
          // write the attr offset
          AttrOffset = Schema.GetOffset(j);
          LTemp = htonl(AttrOffset);
          OutputStr->write((char *)(&LTemp), sizeof(LTemp));
          // write out the meta-info, size first
          MetaInfo = Schema[j].TypeCatInfo;
          if ((MetaInfo != NULL) && 
    	  (MetaInfo->SerializeCopySize(MetaInfoSize) != XXX_NOT_OK)) {
            buf = new char[MetaInfoSize];
            if (MetaInfo->SerializeCopy(buf) != XXX_OK) {
              XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot serialize meta-info");
              return XXX_NOT_OK;
            }
            LTemp = htonl(MetaInfoSize);
            OutputStr->write((char *)(&LTemp), sizeof(LTemp));
            OutputStr->write(buf, MetaInfoSize);
            delete [] buf;
          } else {
            // if there is no meta info to write, write a 0
            MetaInfoSize = 0;
            LTemp = htonl(MetaInfoSize);
            OutputStr->write((char *)(&LTemp), sizeof(LTemp));
          }
        } // for
		*(OutputStr) << flush;
      } // if (SchemaPrint)
      return XXX_OK;
    }
    XxxErrCode XxxJavaClientProtocol::WriteRecord(ostream* OutputStr,
    					      const XxxRecordSchema* Schema,
    					      const XxxRecord* Record)
    {
      return XxxClientProtocol::WriteRecord(OutputStr, Schema, Record);
    }
    
    XxxErrCode XxxJavaClientProtocol::WriteNoError(ostream &OutputStr)
    {
      // Write over a zero to signal no error
      unsigned long LTemp = htonl(0);
      OutputStr.write((char *)&(LTemp), sizeof(LTemp));
      OutputStr << flush;
      if (!OutputStr.good()) {
        XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot write error message to client");
        return XXX_NOT_OK;
      }
      // Now wait for the client to send over a receipt code
      while (1) {
#if defined(XXX_WIN32) && defined(XXX_NONPREEMPTIVE)
	    handle = CreateEvent(NULL, TRUE, FALSE, NULL);
	    if (WSAEventSelect(MCConnection->socketDesc, handle, FD_ACCEPT|FD_READ) != 0)
		{
			cerr << "Bad association between socket and Win32 event";
			return XXX_NOT_OK;
		}
	    MCEventHandle = new win32_compat_event_t(handle);

		if (MCEventHandle->is_down() ||  (bool)MCEventHandle->wait(::WAIT_FOREVER)) 
#else
		if (MCFileHandle->is_down() ||  (bool)MCFileHandle->wait(::WAIT_FOREVER)) 
#endif
		{
          XXX_NOTE_ERROR(XXX_BADCMD, "Error waiting for termination");
          return XXX_NOT_OK;
        }

        char RecvBuf[3];
        if (::recv(MCConnection->socketDesc, RecvBuf, 3, 0) != 3) {
          XXX_NOTE_ERROR(XXX_BADCMD, "Error getting client communication");
          return XXX_NOT_OK;
        }
        if (strncmp(RecvBuf, "END", 3)) {
          cerr << "Client has incorrectly terminated error communication." 
    	   << endl;
        } else break;
      }
    
      return XXX_OK;
    }
    
    XxxErrCode XxxJavaClientProtocol::WriteErrorStack(ostream &OutputStr,
    						  const XxxErrInfo &Top,
    						  const char *ErrorMessage)
    {
      unsigned long LTemp; // used for htonl
    
      // First send the size of the error message and then the message
      int size = strlen(ErrorMessage);
      LTemp = htonl(size);
      OutputStr.write((char *)&(LTemp), sizeof(LTemp));
      OutputStr.write(ErrorMessage, size);
      // Now write out the error stack items one at a time
      const XxxErrInfo *Err = Top.Next;
      while (Err) {
        // Write out the filename, size first, and then line number
        size = strlen(Err->FileName);
        LTemp = htonl(size);
        OutputStr.write((char *)(&LTemp), sizeof(LTemp));
        OutputStr.write(Err->FileName, size);
        LTemp = htonl(Err->LineNumber);
        OutputStr.write((char *)(&LTemp), sizeof(LTemp));
        // Write out the error string, size first
        const char *ErrorString = NULL;
        if ((Err->SourceEngine >= 0) && (Err->SourceEngine < XxxDBEngine::NumDBEngines))
          ErrorString = 
    	XxxDBEngine::DBEngineArray[Err->SourceEngine]->ErrorString(Err->Condition);
        if (ErrorString) {
          size = strlen(ErrorString);
          LTemp = htonl(size);
          OutputStr.write((char *)(&LTemp), sizeof(LTemp));
          OutputStr.write(ErrorString, size);
        } else {
          size = 0;
          LTemp = htonl(size);
          OutputStr.write((char *)(&LTemp), sizeof(LTemp));
        }
        // Write out the message
        size = strlen(Err->Message);
        LTemp = htonl(size);
        OutputStr.write((char *)(&LTemp), sizeof(LTemp));
        OutputStr.write(Err->Message, size);
        Err = Err->Next;
      }
      // Terminate with a zero-length filename
      size = 0;
      LTemp = htonl(size);
      OutputStr.write((char *)(&LTemp), sizeof(LTemp));
      OutputStr << flush;
      if (!OutputStr.good()) {
        XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot write error message to client");
        return XXX_NOT_OK;
      }
      // Now wait for the client to send over a receipt code
      while (1) {
#if defined(XXX_WIN32) && defined(XXX_NONPREEMPTIVE)
		handle = CreateEvent(NULL, TRUE, FALSE, NULL);
		if (WSAEventSelect(MCConnection->socketDesc, handle, FD_ACCEPT|FD_READ) != 0)
		{
			cerr << "Bad association between socket and Win32 event";
			return XXX_NOT_OK;
		}
		MCEventHandle = new win32_compat_event_t(handle);

		if (MCEventHandle->is_down() ||  (bool)MCEventHandle->wait(::WAIT_FOREVER)) 
#else
		if (MCFileHandle->is_down()  ||  (bool)MCFileHandle->wait(::WAIT_FOREVER)) 
#endif
		{
          XXX_NOTE_ERROR(XXX_BADCMD, "Error waiting for termination");
          return XXX_NOT_OK;
        }

        char RecvBuf[3];
        if (::recv(MCConnection->socketDesc, RecvBuf, 3, 0) != 3) {
          XXX_NOTE_ERROR(XXX_BADCMD, "Error getting client communication");
          return XXX_NOT_OK;
        }
        if (strncmp(RecvBuf, "END", 3)) {
          cerr << "Client has incorrectly terminated error communication." 
    	   << endl;
        } else break;
      }
    
      return XXX_OK;
    }
    
    // 
    // Methods of XxxTxtClientProtocol
    //
    XxxErrCode XxxTxtClientProtocol::OpenConnection()
    {
     const int OPTIONSMAX = 1000;
     char OptionsBuf[OPTIONSMAX];
     // ask for and get options string
     if (ClientInfo.StringSendReceive("Options?\n", OptionsBuf, OPTIONSMAX) != XXX_OK) {
       cerr << "Error getting options info" << endl;
       return XXX_NOT_OK;
      }
    
     XxxSetCursor Csr;
     if (initTokenScan(Csr, OptionsBuf) != XXX_OK) {
       cerr << "Problems opening options scan" << endl;
       ClientInfo.StringSendReceive("bad", NULL, 0);
       return XXX_NOT_OK;
      }
     while (1) {
       char* Token; char FlagLetter;
       if (getNextToken(Csr, FlagLetter, Token) != XXX_OK) {
         cerr << "Problems getting next option" << endl;
         ClientInfo.StringSendReceive("bad", NULL, 0);
         return XXX_NOT_OK;
        }
       if (!Csr.IsValid()) break;
       // handle the token appropriately:
       // these options are not very useful currently,
       // but this mechanism can be used for more complex clients
       switch (FlagLetter) {
         // binary mode?
        case 'b' :
         BinaryPrint = atoi(Token);
         break;
         // should schema be printed?
        case 's' :
         SchemaPrint = atoi(Token);
         break;
        }
      } // while (1)
    
     if (endTokenScan(Csr) != XXX_OK) {
       cerr << "Problems closing options scan" << endl;
       ClientInfo.StringSendReceive("bad", NULL, 0);
       return XXX_NOT_OK;
      }
    
     if (ClientInfo.StringSendReceive("good", NULL, 0) != XXX_OK) {
       cerr << "Error negotiating options info" << endl;
       return XXX_NOT_OK;
      }
     return XXX_OK;
    }
    
    XxxErrCode XxxTxtClientProtocol::WriteErrorStack(ostream &OutputStr,
    						 const XxxErrInfo &Top,
    						 const char *ErrorMessage)
    {
      return XxxClientProtocol::WriteErrorStack(OutputStr, Top, ErrorMessage);
    }
    
    XxxErrCode XxxTxtClientProtocol::CloseConnection()
    {
     return XxxClientProtocol::CloseConnection();
    }
    XxxErrCode XxxTxtClientProtocol::CloseRequest()
    {
     return XxxClientProtocol::CloseRequest();
    }
    XxxErrCode XxxTxtClientProtocol::PrefixResults(ostream* OutputStr,
    					       XxxRecordSchema* AnswerSchema)
    {
     // print the schema if necessary
     if (SchemaPrint) {
       AnswerSchema->Print(*(OutputStr));
       *(OutputStr) << endl ;
     } 
     return XXX_OK;
   }

XxxErrCode XxxTxtClientProtocol::WriteRecord(ostream* OutputStr,
					     const XxxRecordSchema* Schema,
					     const XxxRecord* Record)
{
 return XxxClientProtocol::WriteRecord(OutputStr, Schema, Record);
}

// Methods of Compress protocol
#ifdef XXX_COMPRESS

   XxxCompressProtocol::~XxxCompressProtocol() {
   	if (compressEngine)
   		delete compressEngine;
   //	if (os)
   //		delete os;
   }
   // 
   // Methods of XxxCompressProtocol methods
   //
   // Open the connection
   XxxErrCode XxxCompressProtocol::OpenConnection()
   {
    const int OPTIONSMAX = 1000;
    char OptionsBuf[OPTIONSMAX];
    // ask for and get options string
    if (ClientInfo.StringSendReceive("Options?\n", OptionsBuf, OPTIONSMAX) != XXX_OK) {
      cerr << "Error getting options info" << endl;
      return XXX_NOT_OK;
     }
   
    XxxSetCursor Csr;
    if (initTokenScan(Csr, OptionsBuf) != XXX_OK) {
      cerr << "Problems opening options scan" << endl;
      ClientInfo.StringSendReceive("bad", NULL, 0);
      return XXX_NOT_OK;
     }
    while (1) {
      char* Token; char FlagLetter;
      if (getNextToken(Csr, FlagLetter, Token) != XXX_OK) {
        cerr << "Problems getting next option" << endl;
        ClientInfo.StringSendReceive("bad", NULL, 0);
        return XXX_NOT_OK;
       }
      if (!Csr.IsValid()) break;
      // handle the token appropriately:
      // these options are not very useful currently,
      // but this mechanism can be used for more complex clients
      switch (FlagLetter) {
        // binary mode?
       case 'b' :
        BinaryPrint = atoi(Token);
        break;
        // should schema be printed?
       case 's' :
        SchemaPrint = atoi(Token);
        break;
       }
     } // while (1)
   
    if (endTokenScan(Csr) != XXX_OK) {
      cerr << "Problems closing options scan" << endl;
      ClientInfo.StringSendReceive("bad", NULL, 0);
      return XXX_NOT_OK;
     }
   
    if (ClientInfo.StringSendReceive("good", NULL, 0) != XXX_OK) {
      cerr << "Error negotiating options info" << endl;
      return XXX_NOT_OK;
     }
    return XXX_OK;
   }
   
   XxxErrCode XxxCompressProtocol::WriteErrorStack(ostream &OutputStr,
   						 const XxxErrInfo &Top,
   						 const char *ErrorMessage)
   {
     return XxxClientProtocol::WriteErrorStack(OutputStr, Top, ErrorMessage);
   }
   
   XxxErrCode XxxCompressProtocol::CloseConnection()
   {
    return XxxClientProtocol::CloseConnection();
   }
   
   XxxErrCode XxxCompressProtocol::CloseRequest()
   {
   	if (compressEngine && compressEngine->compPlan) {
   		compressEngine->compPlan->Flush(*os);
   		compressEngine->compPlan->CompressCleanUp();
   		}
    
    return XxxClientProtocol::CloseRequest();
   }
   
   // prefix results, output the schema and compression plan to client
   XxxErrCode XxxCompressProtocol::PrefixResults(ostream* OutputStr,
   					       XxxRecordSchema* AnswerSchema)
   {
    // print the schema if necessary
   /* if (SchemaPrint) {
      AnswerSchema->Print(*(OutputStr));
      *(OutputStr) << endl ;
     } 
   */ // output compress plan: 
    compressEngine = new CompressEngine();
    ifstream tmpIs("optflag.txt");
    short optFlag;
    tmpIs >> optFlag;
    if (optFlag)
   	compressEngine->optimizer->Optimize(AnswerSchema,compressEngine->compPlan,tmpIs);
    else
   	 compressEngine->optimizer->Optimize(AnswerSchema,compressEngine->compPlan);
    compressEngine->compPlan->CompressInit((*OutputStr));
    os = OutputStr;
   // os = OutputStr;
    // here we store it in another place
   // os = new ofstream("test.bin",ios::out | ios::binary);
    return XXX_OK;
   }
   XxxErrCode XxxCompressProtocol::WriteRecord(ostream* OutputStr,
   					     const XxxRecordSchema* Schema,
   					     const XxxRecord* Record)
   {
    // first compress the data, then write the record to client
   // for( short i = 0; i < 10; i++) 
   //	compressEngine->compPlan->CompressNext(Record,(*OutputStr));
   	compressEngine->compPlan->CompressNext(Record,(*os));
    return XXX_OK;
   // else
   //	 return XXX_NOT_OK;
   // return XxxClientProtocol::WriteRecord(OutputStr, Schema, Record);
   }
   
#endif
