    #ifndef XXX_CLIENTPROTOCOL_H
    #define XXX_CLIENTPROTOCOL_H
    
    #include <iostream.h>
    #include <stdlib.h>
    #include <string.h>
    #include "client.h"
    
    /**
     * Root class for client protocols
     */
    class XxxClientProtocol 
    {
     protected:
      //// stores the state of a scan
      class tokenScanState : public XxxSetCursor::CursorState { 
       public:
        istrstream InStr;
        char FlagBuf[10];
        char Buf[100];
        tokenScanState(const char* TokenString) : InStr((char *)TokenString) {}
       };
    
      static XxxErrCode initTokenScan(XxxSetCursor& Csr, const char* TokenString);
      static XxxErrCode endTokenScan(XxxSetCursor& Csr);
      static XxxErrCode getNextToken(XxxSetCursor& Csr, char& FlagChar, char*& Token);
    
     public:
      //// there will be table of protocol definitions
      class ProtoDefn { 
       public:
        char Name[XXX_PROTOCOL_NAME_MAX];
        // needed by the Table class
        int Hash() const { return Name[0] + Name[1];}
        int operator==(const ProtoDefn& Proto1) const
         { return !strcmp(Name, Proto1.Name); }
        ProtoDefn(const char* ProtoName) {
          strncpy(Name, ProtoName, XXX_PROTOCOL_NAME_MAX);
          Name[XXX_PROTOCOL_NAME_MAX-1] = 0;
         }
        virtual ~ProtoDefn() {}
        virtual XxxClientProtocol* GetProtocol(XxxClientInfo& CInfo) 
          const { return NULL;}
       };
      
#if defined(XXX_WIN32) && defined(XXX_NONPREEMPTIVE)
	  HANDLE handle;
#endif

      XxxClientInfo& ClientInfo;
      XxxBool BinaryPrint; // should answers be printed in binary?
      XxxBool SchemaPrint; // should schemas be provided first?
    
      XxxClientProtocol(XxxClientInfo& CInfo)
        : ClientInfo(CInfo) {}
      virtual ~XxxClientProtocol() {}
      // start of connection
      virtual XxxErrCode OpenConnection();
      virtual XxxErrCode CloseConnection();
      virtual XxxErrCode CloseRequest();
      virtual XxxErrCode PrefixResults(ostream* OutputStr,
    				   XxxRecordSchema* AnswerSchema);
      virtual XxxErrCode WriteRecord(ostream* OutputStr,
    				 const XxxRecordSchema* Schema,
    				 const XxxRecord* Record);
      virtual XxxErrCode WriteNoError(ostream &OutputStr);
      virtual XxxErrCode WriteErrorStack(ostream &OutputStr,
    				     const XxxErrInfo &Top,
    				     const char *ErrorMessage);
    };
    
    /**
     * Java client
     */
    class XxxJavaClientProtocol : public XxxClientProtocol 
    {
    protected:
      XxxSocketConnection *MCConnection;
      int MCPortNumber; 
#if defined(XXX_WIN32) && defined(XXX_NONPREEMPTIVE)
	  win32_compat_event_t* MCEventHandle;
#else
      sfile_read_hdl_t* MCFileHandle;
#endif
    public:
      //// definition of the protocol
      class JProtoDefn : public XxxClientProtocol::ProtoDefn { 
       public:
        JProtoDefn() : ProtoDefn("javaclient") { }
        virtual XxxClientProtocol* GetProtocol(XxxClientInfo& CInfo) const {
          return new XxxJavaClientProtocol(CInfo);
         }
       };
    
      XxxJavaClientProtocol(XxxClientInfo& CInfo);
      ~XxxJavaClientProtocol();
      // start of connection
      virtual XxxErrCode OpenConnection();
      virtual XxxErrCode CloseConnection();
      virtual XxxErrCode CloseRequest();
      virtual XxxErrCode PrefixResults(ostream* OutputStr,
    				   XxxRecordSchema* AnswerSchema);
      virtual XxxErrCode WriteRecord(ostream* OutputStr,
    				 const XxxRecordSchema* Schema,
    				 const XxxRecord* Record);
      virtual XxxErrCode WriteNoError(ostream &OutputStr);
      virtual XxxErrCode WriteErrorStack(ostream &OutputStr,
    				     const XxxErrInfo &Top,
    				     const char *ErrorMessage);
     };
    
    /**
     * Ascii client
     */
    class XxxTxtClientProtocol : public XxxClientProtocol 
    {
     public:
      //// definition of the protocol
      class TProtoDefn : public XxxClientProtocol::ProtoDefn { 
       public:
        TProtoDefn() : ProtoDefn("txtclient") { }
        virtual XxxClientProtocol* GetProtocol(XxxClientInfo& CInfo) const {
          return new XxxTxtClientProtocol(CInfo);
         }
       };
    
      XxxTxtClientProtocol(XxxClientInfo& CInfo)
        : XxxClientProtocol(CInfo) {
          BinaryPrint = XXX_FALSE;
          SchemaPrint = XXX_TRUE;
         }
      // start of connection
      virtual XxxErrCode OpenConnection();
      virtual XxxErrCode CloseConnection();
      virtual XxxErrCode CloseRequest();
      virtual XxxErrCode PrefixResults(ostream* OutputStr,
    				   XxxRecordSchema* AnswerSchema);
      virtual XxxErrCode WriteRecord(ostream* OutputStr,
    				 const XxxRecordSchema* Schema,
    				 const XxxRecord* Record);
      virtual XxxErrCode WriteErrorStack(ostream &OutputStr,
    				     const XxxErrInfo &Top,
    				     const char *ErrorMessage);
     };

#ifdef XXX_COMPRESS
    
 /**
  *  compression protocol
  */
 #include "compressapi.h"
 class XxxCompressProtocol : public XxxClientProtocol 
 {
  public:
   	ostream *os; // keep it in order to flush
   	CompressEngine *compressEngine;
     //// definition of the protocol
     class CProtoDefn : public XxxClientProtocol::ProtoDefn { 
      public:
       CProtoDefn() : ProtoDefn("compressclient") { }
       virtual XxxClientProtocol* GetProtocol(XxxClientInfo& CInfo) const {
         return new XxxCompressProtocol(CInfo);
        }
      };
   
     XxxCompressProtocol(XxxClientInfo & CInfo) : XxxClientProtocol(CInfo) {
         BinaryPrint = XXX_TRUE;
         SchemaPrint = XXX_TRUE;
   			compressEngine = new CompressEngine();
        }
   	~XxxCompressProtocol();
     // start of connection
     virtual XxxErrCode OpenConnection();
     virtual XxxErrCode CloseConnection();
     virtual XxxErrCode CloseRequest();
     virtual XxxErrCode PrefixResults(ostream* OutputStr,
   				   XxxRecordSchema* AnswerSchema);
     virtual XxxErrCode WriteRecord(ostream* OutputStr,
   				 const XxxRecordSchema* Schema,
   				 const XxxRecord* Record);
     virtual XxxErrCode WriteErrorStack(ostream &OutputStr,
   				     const XxxErrInfo &Top,
   				     const char *ErrorMessage);
    };

#endif // XXX_COMPRESS
#endif







