#ifndef XXX_SMARTFUNC_H
#define XXX_SMARTFUNC_H

/**
 * A fixed size scratch area to be used by functions to maintain
 * any kind of global state (for instance, a count of the invocations)
 */
class XxxFuncScratchPad {
public:
  char* DataArea;
  int BufferSize;
  XxxBool Valid;
  XxxFuncScratchPad(int SizeInBytes = 0) { 
    BufferSize = SizeInBytes;
    if (BufferSize)
      DataArea = new char[BufferSize];
    else
      DataArea = NULL;
    Valid = XXX_FALSE;
  }
  virtual ~XxxFuncScratchPad() {
    if (DataArea) delete [] DataArea;
  }
};

/**
 * A cache for function results, with a fixed number of entries.
 * Each entry is an XxxADTValue. Entries are copied into the cache and
 * discarded when replaced from the cache.
 * All entries in the cache have the same type.
 */
class XxxFuncResultCache {
public:
  int NumEntries;
  XxxBasicType TypeId;
  XxxADTMetaInfo* MetaInfo;
  int ValueSize;

  class CacheEntry {
  public:
    XxxADTValue* Value;
  };
  XxxADTValue* *Entries;

  // mapping from identifiers to entries : TODO

  // replacement policy : TODO

  XxxFuncResultCache(int NumberOfEntries = 0) {
    NumEntries = NumberOfEntries;
    if (NumEntries) {
      Entries = new XxxADTValue*[NumEntries];
      for (int i = 0; i < NumEntries; i++)
	Entries[i] = NULL;
    }
    else
      Entries = NULL;
  }
  virtual ~XxxFuncResultCache() {
    if (Entries && NumEntries) {
      for (int i = 0; i < NumEntries; i++)
	if (Entries[i]) delete Entries[i];
      delete [] Entries;
    }
  }
  //// add a value to the cache
  XxxErrCode Add(const XxxValueEnv* Env, const XxxADTValue* Value);
  //// search for a specific value in the cache
  XxxBool Search(const XxxValueEnv* Env, XxxADTValue* Value);
  // TODO: these two functions need to be implemented
};

#endif
