#ifndef XXX_STRINGTYPES_H
#define XXX_STRINGTYPES_H

#include "basictypes.h"

/**
 * ADT defining variable length strings (subject to a maximum size)
 */
class XxxStringADTClass : public XxxCoreADTClass {
 public:
  XxxStringADTClass()
    : XxxCoreADTClass(XXX_STRING, "string", sizeof(StrValue)+XXX_TEXTSTRING_MAX+1, XXX_FALSE) {}
  static int ValueSize(const XxxADTValue* Object,
		       const XxxADTMetaInfo* MetaInfo)
   { return XxxCoreADTClass::ValueOffset(XXX_STRING) + 
            strlen(((StrValue*)Object)->Value) + 1; }
 };


/**
 * ADT defining variable length strings (subject to a maximum size)
 */
class XxxFixedStringADT : public XxxCoreADTClass {
 public:
  class FStringMetaInfo : public XxxADTMetaInfo { 
  public:
    unsigned int Length;
    FStringMetaInfo() { Length = 0;}
    virtual ~FStringMetaInfo() {}
    virtual XxxErrCode Replicate(XxxADTMetaInfo*& NewMI) const {
      FStringMetaInfo* NewMetaInfo = new FStringMetaInfo(*this);
      NewMI = NewMetaInfo;
      return XXX_OK;
    }
    virtual void Print(ostream& OStr) const {
      OStr << "[" << Length << "]" ;
    }
    virtual XxxBool IsCompatible(const XxxADTMetaInfo* OtherMI) const {
      if (Length <= ((FStringMetaInfo*)OtherMI)->Length)
	return XXX_TRUE;
      else return XXX_FALSE;
    }
    //// number of bytes necessary to hold the serialized version
    virtual XxxErrCode SerializeCopySize(int &Size) const {
      Size = sizeof(Length);
      return XXX_OK;
    }
    //// Creates a serialized version of this meta-info object.
    virtual XxxErrCode SerializeCopy(char *Data) const {
      memcpy(Data, (char *)(&Length), sizeof(Length));
      return XXX_OK;
    }
  }; // classFStringMetaInfo
  
  XxxFixedStringADT()
    : XxxCoreADTClass(XXX_FIXEDSTRING, "char", sizeof(StrValue)+XXX_TEXTSTRING_MAX+1, XXX_TRUE, XXX_TRUE) {}
  static int ValueSize(const XxxADTMetaInfo* MetaInfo){ 
    return XxxCoreADTClass::ValueOffset(XXX_FIXEDSTRING) + 
      ((FStringMetaInfo*)(MetaInfo))->Length + 1; 
  }
protected:
  // these to methods are needed for ADTs with metainfo
  virtual XxxErrCode parseMetaInfo(const char* MetaInfoString,
				   MetaInfoHandle& MIHandle) const {
    MIHandle.Size = sizeof(FStringMetaInfo);
    ASSERT((MIHandle.Size <= (unsigned int)XXX_MIHNDL_MAX)); //added cast to remove warning
    FStringMetaInfo* FMI = (FStringMetaInfo*) (MIHandle.Data);
    istrstream InStr((char *)MetaInfoString);
    InStr >> FMI->Length;
    if (InStr.fail()) return XXX_NOT_OK;
    return XXX_OK;
  }
  virtual XxxErrCode makeMetaInfo(const MetaInfoHandle& MIHandle,
				  XxxADTMetaInfo*& MI) const {
    FStringMetaInfo* HandleInfo = (FStringMetaInfo*)(MIHandle.Data);
    MI = new FStringMetaInfo(*HandleInfo);
    return XXX_OK;
  }
 };

#endif
