#ifndef XXX_TEMPLATES_H
#define XXX_TEMPLATES_H
#include <stdio.h>
#include <stdlib.h>
#include "types.h"


//Bjarke, aug. 6: #pragma interface/implementation not needed any longer
//#ifdef __GNUG__ 
//#pragma interface 
//#endif 
 

// This file defines templates for a Stack, a Table and a Set

// define a Template Stack class used to implement a Trace Stack

template <class T> class XxxStackFrame;

/**
 * Template for generic stack
 */
template <class Type>
class XxxStack {
 private:
  XxxStackFrame<Type> *top;
  int height;
 public:
  XxxBool IsEmpty() const { return (XxxBool)(top == NULL);}
  int Height() const { return height; }
  void PopStack(Type& TopVal);
  Type& TopStack() const { ASSERT((!(IsEmpty()))); return top->item;}
  void PushStack(const Type&);
  void ApplyToDepth(int depth, void (*)(Type&, void *), void *);
  XxxStack() {top = NULL; height = 0;}
 };

/**
 * Template for generic stack frame
 */
template <class Type>
class XxxStackFrame{
 friend class XxxStack<Type>;
 private:
  XxxStackFrame(const Type &t) : item(t) { next = NULL;}
  Type item;
  XxxStackFrame *next;
 };

/**
 * Simple set cursor. 
 * The responsibility of maintaining the state is
 * left to the Set over which the cursor is used
 */
class XxxSetCursor {
 public:
  //// State for the cursor
  class CursorState { 
   public:
    CursorState() {}
    virtual ~CursorState() {}
  };
  XxxBool AtEnd;
  CursorState *State;
  int NumberRead;
  XxxSetCursor() { State = NULL; AtEnd = XXX_TRUE; NumberRead = 0;}
  XxxBool IsValid() const { return ((AtEnd== XXX_TRUE)?XXX_FALSE:XXX_TRUE);}
 };

/**
 // Template MultiSet. Subclasses can be defined by specifying
 // addItem, RemoveItem, InitCursor, NextItem and CloseCursor
 */
template <class Type>
class XxxMultiSet {
 protected:
  int setCount;
  XxxBool checkDupl;
  int (*EqualFunc)(const Type&, void*);
  virtual XxxErrCode addItem(const Type& item) = 0;
 public:
  XxxMultiSet() { setCount = 0;checkDupl = XXX_FALSE; EqualFunc = NULL;}
  virtual ~XxxMultiSet() {}

  virtual XxxErrCode InitCursor(XxxSetCursor&) const = 0;
  virtual XxxErrCode NextItem(XxxSetCursor&, Type& ReturnItem) const = 0;
  virtual XxxErrCode CloseCursor(XxxSetCursor&) const = 0;

  void ApplyAll(int (*)(Type&, void *), void *);
  int ApplyFirst(int (*)(Type&, void *), void *);
  XxxErrCode AddItem(const Type& item);
  virtual XxxErrCode RemoveItem(const Type& ) = 0;
  virtual XxxBool FindItem(int (*Func)(const Type&, void*),
			   void* Arg, Type& ReturnItem) const;
  XxxBool IsMember(const Type& Item) const {
    Type ReturnItem;
    if (EqualFunc)
      return FindItem(EqualFunc, (void *)(&Item), ReturnItem);
    else
      return XXX_FALSE;
   }
  int Count() const { return setCount;}
 };


/**
 // Template Set : Multi-Set without duplicates
 */
template <class Type>
class XxxSet : public virtual XxxMultiSet<Type> {
  static int XxxSetEqualFunc(const Type& Item, void* Arg)
    {
      return (Item == *((Type *)Arg));
    }
 public:
  XxxSet() { checkDupl = XXX_TRUE; EqualFunc = XxxSetEqualFunc;}
 };


template <class Type> class XxxRootTable;
/**
 * Table template class of <Type>
 */
template <class Type>
class XxxTableEntry{
 friend class XxxRootTable<Type>;
 // HACK:	Philippe changed from private to public.
 public:
  XxxTableEntry(const Type &t) : item(t) {}
  XxxTableEntry() {}
  void fill(const Type& t) { item = t; }
  Type item;
 };

/**
 * The root Table class template which is an indexed Set
 */
template <class Type>
class XxxRootTable : virtual public XxxSet<Type> {
 // HACK : Philippe changed from private to public
 public:
  XxxTableEntry<Type> *regItems;
  XxxBool* valid;
  int regCount;
  int regMax;
  void Init();
 protected:
  virtual XxxErrCode addItem(const Type& item);
  virtual int Hash(const Type& Item) const = 0;
  virtual int Equal(const Type& Item1, const Type& Item2) const = 0;
 public:
  XxxRootTable(int size) {
    // create and initialize to NULL
    regItems = new XxxTableEntry<Type>[size];
    valid = new XxxBool[size];
    regMax = size;
    regCount = 0;
    Init();
   }
  ~XxxRootTable();

  virtual XxxErrCode InitCursor(XxxSetCursor& SCsr) const;
  virtual XxxErrCode NextItem(XxxSetCursor&, Type& ReturnItem) const;
  virtual XxxErrCode CloseCursor(XxxSetCursor& SCsr) const;

  // just changed this to XxxBool!!
  virtual XxxBool FindTableItem(const Type& PatternItem, Type& ReturnItem) const;
  virtual XxxErrCode RemoveItem(const Type& );
 };

/**
 * The simple Table class template
 */
template <class Type>
class XxxTable : public XxxRootTable<Type> {
 protected:
  virtual int Hash(const Type& Item) const { return Item.Hash(); }
  virtual int Equal(const Type& Item1, const Type& Item2) const
   { return (Item1 == Item2); }
 public:
  XxxTable(int Size):XxxRootTable<Type>(Size){}
 };

/**
 * The ptr Table class template
 */
template <class Type>
class XxxPtrTable : public XxxRootTable<Type> {
 protected:
  virtual int Hash(const Type& Item) const { return Item->Hash(); }
  virtual int Equal(const Type& Item1, const Type& Item2) const
   { return *(Item1) == *(Item2); }
 public:
	 XxxPtrTable(int Size):XxxRootTable<Type>(Size){}
 };


/**
 * Philippe: Another ptr Table class template
 */
template <class Type>
class XxxPtrTable1 : public XxxRootTable<Type> {
 public:
  int *key;
  XxxBool *dupl;
  int *lastpos;
 protected:
  virtual int Hash(const Type& Item) const { return Item->Hash(); }
  virtual int Equal(const Type& Item1, const Type& Item2) const
   { return (Item1->Equal(Item2)); }

  virtual XxxErrCode addItem(const Type& item);

 public:
  XxxPtrTable1(int Size);
  ~XxxPtrTable1();

  virtual XxxErrCode RemoveItem(const Type& );
  virtual XxxBool    FindTableItem(const Type& PatternItem, Type& ReturnItem) const; 
  virtual XxxBool    FindTableItem1(const Type& PatternItem, Type& ReturnItem) const; 

 };


/**
 * The Array class  template
 */
template <class Type>
class XxxRootArray {
 protected:
  Type *memberArray;
  int arraySize;
  int arrayEnd;
  int incrRatio;
  virtual int Equal(const Type& Entry1, const Type& Entry2) const = 0;
 public:
  XxxRootArray(int Size, int IncrRatio = 2) {
    arraySize = Size;
    arrayEnd = 0;
    memberArray = new Type[Size];
   }
  virtual ~XxxRootArray() {
    delete [] memberArray;
   }
  int& Count() { return arrayEnd; }
  int Cnt() const { return arrayEnd; }
  int Size() const { return arraySize; }

  Type& operator [] (int index) const {
    return memberArray[index];
   }
  void Grow();
  virtual int Find(const Type& Match) const;
 };

/**
 * The Array class template
 */
template <class Type>
class XxxArray : public XxxRootArray<Type> {
 protected:
  virtual int Equal(const Type& Entry1, const Type& Entry2) const
   { return (Entry1 == Entry2); }
 public:
  XxxArray(int Size, int IncrRatio = 2):XxxRootArray<Type>(Size, IncrRatio){}
 };

/**
 * Array of Ptrs
 */
template <class Type>
class XxxPtrArray : public XxxRootArray<Type> {
 protected:
  virtual int Equal(const Type& Entry1, const Type& Entry2) const;

 public:
  XxxPtrArray(int Size, int IncrRatio = 2):XxxRootArray<Type>(Size, IncrRatio){}
 };

// June 26 - Win32 style inlining extended to all platforms
//#ifdef XXX_WIN32
// the template definitions need to be inlined
#include "templates.C"
//#endif


#endif
