#include <stdio.h>
#include <string.h>
#include "types.h"
#include "basictypes.h"
#include "genfunc.h"
#include "xxx_thread.h"
#include "counter.h"
#include "date.h"
#include "document.h"
#include "complex.h"
#include "bytearray.h"
#include "funcexpr.h"
#include "suspcomp.h"
#include "typelist.h"
#include "alias.h"
#include "enum.h"
#include "struct.h"
// vcheng added
#include "gentimeseries.h"

#ifdef XXX_DEF_COUGAR
#include "door.h"
#include "javaring.h"
#include "tempsensor.h"
#endif
#ifdef XXX_DEF_RELADT
#include "reladt.h"
#endif
#ifdef XXX_DEF_IMGADT
#include "image.h"
#endif
#ifdef XXX_DEF_SEQADT
#include "seqadt.h"
#endif
#ifdef XXX_DEF_AUDADT
#include "audio.h"
#endif

#ifdef XXX_DEF_GIS
#include "box.h"
#include "circle.h"
#include "point.h"
#include "polygon.h"
#include "raster.h"
#include "geometry.h"
#endif

#ifdef XXX_SEMISTRUCT
#include "union.h"
#include "unionvar.h"
#include "FlexFixedRecord.h"
#endif

//Begin: Moustafa
#include "STransX.h"
//End: Moustafa


// static variable needs to be initialized
const struct XxxADTClass::MetaInfoHandle XxxADTClass::NullMIHndl;

const char XxxADTValue::NULL_VALUE = 0x01; 
const char XxxADTValue::SUSPENDED_COMPUTATION = 0x02;
const char XxxADTValue::PRED_BIG_ENDIAN = 0x04;
const char XxxADTValue::COMPRESSED = 0x08;

XxxBaseMutex* XxxADTClass::Latch = new XxxBaseMutex();
// Only global variables in main code!!
XxxADTClass* XxxADTClass::ADTArray[XXX_GENFUNC+XXX_ADT_MAX];
int XxxADTClass::NumADTs;

// This function gets invoked by the InitThread to set up all the ADTs
// To add a new E-ADT : do the following
//    a) include the appropriate header file
//    b) assign the type id by incrementing NumADTs
//    c) create a new type object (the constructor will automatically
//        insert itself into the ADT array)
// Note that all the type ids are assigned before all the
// types are created. This is because some of the types
// need to access each other. This will be rectified in
// the next release.
void XxxInitializeADTs()
{
 // Latch is used to synchronize access to the ADT array
 // In this file, may not be necessary, because during the 
 // startup of the server, only one thread will access the array
 XxxUseMutex AutoMutex(*XxxADTClass::Latch);
 
 for (int i = 0; i < XXX_GENFUNC+XXX_ADT_MAX; 
      XxxADTClass::ADTArray[i++] = NULL);
 // initialize NumADTs to 0
 XxxADTClass::NumADTs = 0;
 // register the core ADTs
 XxxCoreADTClass::RegisterCoreADTs();

 XxxADTClass::NumADTs = XXX_GENFUNC+1;

 // now the "user-defined" ADTs begin

 // stage 1 : acquire type idsa

#ifdef XXX_DEF_SEQADT
 // the constructor for XxxSeqADT may itself create a few more ADTs
 // like OrderDomains, etc and insert them into the XxxADTArray
 XxxSeqADT::SeqTypeId = XxxADTClass::NumADTs++;
#endif
 
//Begin: Moustafa
 XxxSTransXADT::STransXTypeId = XxxADTClass::NumADTs++;	
//End: Moustafa

#ifdef XXX_DEF_RELADT
 RelationTypeId = XxxADTClass::NumADTs++;
#endif

 XxxCounterADT::CounterTypeId = XxxADTClass::NumADTs++;

 XxxComplexADT::ComplexTypeId = XxxADTClass::NumADTs++;

#ifdef XXX_DEF_COUGAR
 XxxDoorADT::DoorTypeId = XxxADTClass::NumADTs++;

 XxxJavaRingADT::JavaRingTypeId = XxxADTClass::NumADTs++;

 XxxTempSensorADT::TempSensorTypeId = XxxADTClass::NumADTs++;
#endif


 JulianMDYCalADT::MDYOrdDomId = XxxADTClass::NumADTs++;
 
#ifdef XXX_DEF_IMGADT
 XxxImageADT::ImageTypeId = XxxADTClass::NumADTs++;
#endif

#ifdef XXX_DEF_AUDADT
 XxxAudioADT::AudioTypeId = XxxADTClass::NumADTs++;
#endif

#ifdef XXX_DEF_GIS
 XxxPointADT::PointTypeId = XxxADTClass::NumADTs++;

 XxxBoxADT::BoxTypeId = XxxADTClass::NumADTs++;

 XxxCircleADT::CircleTypeId = XxxADTClass::NumADTs++;

 XxxPolygonADT::PolygonTypeId = XxxADTClass::NumADTs++;

 // Philippe: June 25
 // XxxRasterADT::RasterTypeId = XxxADTClass::NumADTs++;

#endif

 XxxByteArrayADT::ByteArrayTypeId = XxxADTClass::NumADTs++;
 XxxDocumentADT::DocumentTypeId = XxxADTClass::NumADTs++;
#ifdef XXX_SEMISTRUCT
  // The following ADTs apply to semi-structured data types
  XxxUnionADT::UnionTypeId = XxxADTClass::NumADTs++;
  XxxUnionVarADT::UnionVarTypeId = XxxADTClass::NumADTs++;
  XxxFlexFixedRecordADT::FlexFixedRecordTypeId = XxxADTClass::NumADTs++;
#endif

#ifdef XXX_STRUCT
 XxxStructADT::StructTypeId = XxxADTClass::NumADTs++;
#endif
 
 // vcheng added
XxxGenTimeSeriesADT::GenTimeSeriesTypeId = XxxADTClass::NumADTs++;

 // stage 2 : create type objects using type ids

#ifdef XXX_DEF_SEQADT
 // 
 // the constructor for XxxSeqADT may itself create a few more ADTs
 // like OrderDomains, etc and insert them into the XxxADTArray
 //
 new XxxSeqADT(XxxSeqADT::SeqTypeId);
#endif
 
#ifdef XXX_DEF_RELADT
 new XxxRelADT(RelationTypeId);
#endif

 // Here's where all new types should be added. Need to find a nicer
 // mechanism to do this.
 XxxADTClass* NewADT;

 //Begin: Moustafa
 NewADT = new XxxSTransXADT(XxxSTransXADT::STransXTypeId);
 //End: Moustafa

 NewADT = new XxxCounterADT(XxxCounterADT::CounterTypeId);
 // cerr << "ADT " << NewADT->TypeName << " size: " << NewADT->MaxObjectSize(0) << endl;

 NewADT = new XxxComplexADT(XxxComplexADT::ComplexTypeId);

#ifdef XXX_DEF_COUGAR
 NewADT = new XxxDoorADT(XxxDoorADT::DoorTypeId);
 NewADT = new XxxJavaRingADT(XxxJavaRingADT::JavaRingTypeId);
 NewADT = new XxxTempSensorADT(XxxTempSensorADT::TempSensorTypeId);
#endif

 NewADT = new JulianMDYCalADT(JulianMDYCalADT::MDYOrdDomId);

#ifdef XXX_DEF_IMGADT
 NewADT = new XxxImageADT(XxxImageADT::ImageTypeId);
#endif

#ifdef XXX_DEF_AUDADT
 NewADT = new XxxAudioADT(XxxAudioADT::AudioTypeId);
#endif

#ifdef XXX_DEF_GIS
 NewADT = new XxxPointADT(XxxPointADT::PointTypeId);
 NewADT = new XxxBoxADT(XxxBoxADT::BoxTypeId);
 NewADT = new XxxCircleADT(XxxCircleADT::CircleTypeId);
 NewADT = new XxxPolygonADT(XxxPolygonADT::PolygonTypeId);
 // Philippe, June 25.
 // NewADT = new XxxRasterADT(XxxRasterADT::RasterTypeId);
#endif

 NewADT = new XxxByteArrayADT(XxxByteArrayADT::ByteArrayTypeId);

 NewADT = new XxxDocumentADT(XxxDocumentADT::DocumentTypeId);

#ifdef XXX_SEMISTRUCT
  NewADT = new XxxUnionADT(XxxUnionADT::UnionTypeId);
  NewADT = new XxxUnionVarADT(XxxUnionVarADT::UnionVarTypeId);
  NewADT = new XxxFlexFixedRecordADT(XxxFlexFixedRecordADT::FlexFixedRecordTypeId);
#endif

#ifdef XXX_STRUCT
 NewADT = new XxxStructADT(XxxStructADT::StructTypeId);
#endif

 // vcheng added
 NewADT = new XxxGenTimeSeriesADT(XxxGenTimeSeriesADT::GenTimeSeriesTypeId);

 // put this at the end, because the functions registered
 // in GenFunc may use the other types
 new XxxGenFuncADTClass(XXX_GENFUNC, "_genfunc");

 if (XxxShoreSM->BeginTrans() != XXX_OK) {
   XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot start transaction");
   return;
 }
 
 XxxErrCode rc = XXX_OK;
 
 // Create an File scan iterator to read all records in the type list file
 scan_file_i ScanIter(XxxShoreSM->defaultVolId, XxxShoreSM->typeFileId);
 
 pin_i* Pin = 0;
 bool AtEnd=false;
 w_rc_t RetVal;
 
 XXX_CALL((XXX_SHORE_RETFN(ScanIter.next(Pin, 0, AtEnd), RetVal)),
	  XXX_SHOREERR, "Couldnt scan file", rc);
 
 while (rc == XXX_OK && AtEnd != true) {
   
   if (!XxxShoreSM->PinValid(Pin, SH)) {
     XXX_NOTE_ERROR(XXX_SHOREERR, " Invalid pin ");
     return;
   }
   int TotalSize = Pin->body_size();
   char* typeRecord = new char[TotalSize];
   if (XxxShoreSM->GetObjectPreAlloc(Pin, typeRecord, TotalSize, 0, TotalSize) != XXX_OK) {
     XXX_NOTE_ERROR(XXX_SHOREERR, " Error getting pinned record");
     Pin->unpin();
     return;
   }
   
   TypeInfo typeInfo(typeRecord);
   
   if (!strcmp(typeInfo.type, "alias")) {
     XxxBasicType Id;
     char* aliasToName, *metaString;
     if ((XxxAliasADT::UnSerialize(typeInfo.tInfo, Id, aliasToName, metaString)) != XXX_OK) {
       XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot unserialize the alias data from defdisk");
       if (XxxShoreSM->AbortTrans() != XXX_OK)
	 XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot abort transaction");
       return;
     }
     XxxErrCode status;
     new XxxAliasADT(Id, typeInfo.name, aliasToName, metaString, status);
     if (status != XXX_OK){
       XXX_NOTE_ERROR(XXX_BADADT, "Cannot create alias type");
       return;
     }
   } else if(strcmp(typeInfo.type, "enum") == 0){
     int numEnum;
     char **enumNames;
     if((XxxEnumADT::UnSerialize(typeInfo.tInfo, numEnum, enumNames)) != XXX_OK) {
       XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot unserialize the alias data from defdisk");
       if (XxxShoreSM->AbortTrans() != XXX_OK)
	 XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot abort transaction");
       return;
     }
     
     new XxxEnumADT(typeInfo.typeId, typeInfo.name, numEnum, enumNames, TotalSize);
     
   } else {
     XXX_NOTE_ERROR(XXX_SHOREERR, "Right now, only info about alias, enum, struct is stored in the TypeList file");
     if (XxxShoreSM->AbortTrans() != XXX_OK)
       XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot abort transaction");
     return;
   }
   
   XxxADTClass::NumADTs++;
   
   XXX_CALL((XXX_SHORE_RETFN(ScanIter.next(Pin, 0, AtEnd), RetVal)),
	    XXX_SHOREERR, "Couldnt scan file", rc);
   
 }
 
 if (XxxShoreSM->CommitTrans() != XXX_OK) {
   XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot commit transaction");
   if (XxxShoreSM->AbortTrans() != XXX_OK)
     XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot abort transaction");
   return;
 }
 
}

XxxErrCode XxxListAlias(ostream& Ostr)
{
  // Latch is used to synchronize access to the ADT array
  XxxUseMutex AutoMutex(*XxxADTClass::Latch);
  
  XxxErrCode rc = XXX_OK;
  
  // Create an File scan iterator to read all records in the type list file
  scan_file_i ScanIter(XxxShoreSM->defaultVolId, XxxShoreSM->typeFileId);
  
  pin_i* Pin = 0;
  bool AtEnd = false;
  w_rc_t RetVal;
  
  XXX_CALL((XXX_SHORE_RETFN(ScanIter.next(Pin, 0, AtEnd), RetVal)),
	   XXX_SHOREERR, "Couldnt scan file", rc);
  
  while (AtEnd != true) {
    
    if (!XxxShoreSM->PinValid(Pin, SH)) {
      XXX_NOTE_ERROR(XXX_SHOREERR, " Invalid pin ");
      return XXX_NOT_OK;
    }
    
    int TotalSize = Pin->body_size();
    
    char* typeRecord = new char[TotalSize];
    
    if (XxxShoreSM->GetObjectPreAlloc(Pin, typeRecord, TotalSize, 0, TotalSize) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_SHOREERR, " Error getting pinned record");
      Pin->unpin();
      return XXX_NOT_OK;
    }
    
    TypeInfo typeInfo(typeRecord);
    
    Ostr << " Types are " << endl;
    if(strcmp(typeInfo.type, "alias") == 0){
      XxxBasicType Id;
      char* aliasToName, *metaString;
      //Bjarke:TOFIX problem with deallocation of metastring Also,
      //error message willl never be displayed, since UnSerialize
      //always returns XXX_OK
      XXX_CALL_RET(XxxAliasADT::UnSerialize(typeInfo.tInfo, Id, 
					    aliasToName, metaString),
		   XXX_SHOREERR, 
		   "Cannot unserialize the alias data from defdisk");
      
      Ostr << " alias " << typeInfo.name << " ==> " << aliasToName << " , meta string (" <<
	metaString << " )" << endl;
      
    } else if (strcmp(typeInfo.type, "enum") == 0) {
      int numEnum;
      char **enumNames;
      XXX_CALL_RET (XxxEnumADT::UnSerialize(typeInfo.tInfo, numEnum, enumNames),
		    XXX_SHOREERR, "Cannot unserialize the alias data from defdisk");
      
      Ostr << " enum " << typeInfo.name << " (";
      for (int j=0; j<numEnum; j++) {
	Ostr << enumNames[j];
	if (j!=(numEnum-1)) {
	  Ostr << " , ";
	}
      }
      Ostr << " )" << endl;
    } 
    else {
      XXX_NOTE_ERROR(XXX_SHOREERR, "Right now, only info about alias, enum, struct is stored in the TypeList file");
      return XXX_NOT_OK;
    }
    
    XXX_CALL((XXX_SHORE_RETFN(ScanIter.next(Pin, 0, AtEnd), RetVal)),
	     XXX_SHOREERR, "Couldnt scan file", rc);
    
  }
  
  return XXX_OK;
}

// This function gets invoked by the InitThread to close all the ADTs
void XxxTerminateADTs()
{
 XxxUseMutex AutoMutex(*XxxADTClass::Latch);
 for (int i = 0; i < XxxADTClass::NumADTs; i++)
   delete XxxADTClass::ADTArray[i];
}

//
// Methods of XxxADTClass
//
XxxADTClass::XxxADTClass(XxxBasicType Id, const char* Name, 
			 int ObjSize, XxxBool InField, 
			 XxxBool MaintainsCat) 
: TypeId(Id), maxObjectSize((ObjSize > 0)? ObjSize : 1), NUM_SPARE_METHOD_SLOTS(20)
{
 if (ObjSize <= 0) 
   cerr << "Object Size of Type " << Name << "must be at least 1!" << endl;
 XxxADTArray[Id] = this;
 strncpy(TypeName, Name, XXX_ADT_NAME_MAX);
 TypeName[XXX_ADT_NAME_MAX-1] = 0;
 maintainsMetaInfo = MaintainsCat;
 StoreInField = InField;

 // initialize the arrays to NULL
 MethArray = NULL;
 NumMethods = 0;
 NumMethodSlots = 0;
 AggrMethArray = NULL;
 NumAggrMethods = 0;
 
 // If you need to allocate a rule engine for optimization rules,
 // do this in the ADT subclass constructor.
 // RuleEngine = new XxxOptRuleEngine;
 RuleEngine = NULL;
}

XxxADTClass::~XxxADTClass()
{
 if (RuleEngine) {
   delete RuleEngine;
   RuleEngine=NULL;
  }
 // reclaim the method arrays
 int i;
 if (MethArray) {
   for (i = 0; i < NumMethods; i++) 
     delete MethArray[i];
   delete [] MethArray;
   MethArray = NULL;
  }
 if (AggrMethArray) {
   for (i = 0; i < NumAggrMethods; i++) 
     delete AggrMethArray[i];
   delete [] AggrMethArray;
   AggrMethArray = NULL;
  }
}

XxxErrCode XxxADTClass::CheckName(XxxBasicType typeId, const char *Name)
{
  XxxUseMutex AutoMutex(*XxxADTClass::Latch);
  int Id;
  for (Id = 0; Id < XxxADTClass::NumADTs; Id++){
    if (!strcasecmp(XxxADTClass::ADTArray[Id]->TypeName, Name)) {
      break;
    }
  }
  if (Id != XxxADTClass::NumADTs) {
    XXX_NOTE_ERROR(XXX_BADTYPE, "The type with this name already exists in system");
    return XXX_NOT_OK;
  }
  return XXX_OK;
}

void XxxADTClass::Print(ostream& OStr, const XxxADTMetaInfo *MetaInfo) const
{
 OStr << TypeName ;
 if (MetaInfo && maintainsMetaInfo) {
   MetaInfo->Print(OStr);
  }
}

XxxErrCode XxxADTClass::AddMetaInfo(const char* Name, 
				    const XxxBool Named,
				    const char* MetaInfoString,
				    MetaInfoHandle& MIHandle)
{
 // Assumes that Name and Named are not of consequence.
 // TODO:L2 fix this eventually.
 if (parseMetaInfo(MetaInfoString, MIHandle) != XXX_OK){
   XXX_NOTE_ERROR(XXX_BADADT, "Error parsing ADT meta info");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

//Bjarke, aug. 8: const added to char* parameter
XxxErrCode XxxADTClass::GetMetaInfo(const char* ObjName,
				    const MetaInfoHandle& MIHandle,
				    XxxADTMetaInfo* &MetaInfo,
				    XxxBool& Found) const
{
 XxxADTMetaInfo* ThisMetaInfo;
 if (makeMetaInfo(MIHandle, ThisMetaInfo) != XXX_OK)
   return XXX_NOT_OK;
 MetaInfo = ThisMetaInfo;
 Found = XXX_TRUE;
 return XXX_OK;
}

XxxErrCode XxxADTClass::
StoreStatsInfo(const XxxADTMetaInfo *MetaInfo,
	       const TypeStats* StatsInfo,
	       MetaInfoHandle& MIHandle) 
{
 if (!StatsInfo) {
   MIHandle.Size = 0;
   return XXX_OK;
  }
 MIHandle.Size = StatsInfo->Size(MetaInfo);
 ASSERT((MIHandle.Size <= XXX_MIHNDL_MAX));
 TypeStats* StoreSI = (TypeStats*)(MIHandle.Data);
 return StatsInfo->Copy(MetaInfo, StoreSI);
}

XxxErrCode XxxADTClass::
RetrieveStatsInfo(const XxxADTMetaInfo *MetaInfo,
		  const MetaInfoHandle& MIHandle,
		  TypeStats*& StatsInfo)
{
 if (CreateStatsInfo(MetaInfo, StatsInfo) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_BADADT, "Error creating stats info");
   return XXX_NOT_OK;
  }

 // if no stats are maintained
 if (!StatsInfo) return XXX_OK;

 if (makeStatsInfo(MetaInfo, MIHandle, StatsInfo) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_BADADT, "Error making stats info");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

XxxErrCode XxxADTClass::CreateEnv(XxxADTEnv* &Env) const
{
 Env = new XxxADTEnv;
 return XXX_OK;
}

XxxErrCode XxxADTClass::DeleteEnv( XxxADTEnv* Env ) const
{ 
 if (Env) 
   delete Env;
 return XXX_OK;
}

// Function that does any reorganization of a function expression
// by traversing its expression tree and applying rules to
// the funcexpressions. This method is virtual, so e-adts can over-ride it.
// But in general, it will work well as a default.
// This is called from FuncExpression::Optimize()
XxxErrCode XxxADTClass::FuncReOrganize(XxxValueExpression*& VExpr) const
{
 if (RuleEngine) {
   XxxADTEnv *ThisEnv = XXX_THISADTENV(TypeId);
   if (RuleEngine->FireAllRules(VExpr, ThisEnv->ApplyTopDown,
				ThisEnv->ApplyBottomUp) != XXX_OK) {
     XXX_NOTE_ERROR(XXX_BADOPT, "Error firing rule engine");
     return XXX_NOT_OK;
    }
  }
 return XXX_OK;
}

//
// Methods of XxxADTValue
//
void XxxADTValue::Init()
{
  StatusFlag = 0x00; // clear all bits
#ifdef XXX_BIGENDIAN
  Set(PRED_BIG_ENDIAN);
#else
  Unset(PRED_BIG_ENDIAN);
#endif
  SuspensionId = XxxSuspCompLib::INVALID_ID;
  ExtraFlag = 0;
}

// any subclass can override this
XxxErrCode XxxADTClass::methOptimize(XxxFuncParseInfo* FParse,
				     XxxValueExprPlan* OwnerPlan,
				     XxxValExprPlanList* ArgPlans,
				     XxxFuncPlanInfo*& FPlan) const 
{
  FuncMethodInfo* ThisMethInfo;
  if (FParse->IsAggr)
    ThisMethInfo = AggrMethArray[FParse->MethodIndex];
  else
    ThisMethInfo = MethArray[FParse->MethodIndex];

  if (!ThisMethInfo) {
    XXX_NOTE_ERROR(XXX_INTERNAL, "Method info doesnt exist");
    return XXX_NOT_OK;
  }
  OptimizeFunc OptFunc = ThisMethInfo->OptFunc;
  if (!OptFunc) {
    FPlan = new XxxFuncPlanInfo(FParse, (ArgPlans?ArgPlans->Count():0));
    // add in the costs of OwnerPlan and ArgPlans
    if (OwnerPlan)
      FPlan->getPlanUsage().Add(OwnerPlan->GetPlanUsage());
    // step thru arg plans adding cost
    const XxxValExprPlanList* Tmp = ArgPlans;
    while (Tmp) {
      FPlan->getPlanUsage().Add(Tmp->Plan->GetPlanUsage());
      Tmp = Tmp->Next;
    }
    // add in actual cost of function
    FPlan->getPlanUsage().Add(ThisMethInfo->Usage);
  }
  else {
   if ((*OptFunc)(FParse, OwnerPlan, ArgPlans, FPlan) != XXX_OK) {
     XXX_NOTE_ERROR(XXX_BADFUNC, "Error optimizing function");
     return XXX_NOT_OK;
   }
  }
 return XXX_OK;
}

XxxErrCode 
XxxADTClass::matchMethodName(const char*FuncName,
			     const XxxValueExpression* Owner,
			     const XxxValueExprList *ArgList,
			     XxxBool IsAggr,
			     int& MethodIndex,
			     XxxBasicType& ReturnType,
			     XxxADTMetaInfo*& ReturnMetaInfo,
			     XxxBool& Success) const
{
 MethodIndex = -1;
 ReturnType = XXX_UNKNOWN;
 ReturnMetaInfo = NULL;
 Success = XXX_FALSE;

 XxxFuncMethodInfo** MethodArray;
 int MethodCount;
 if (IsAggr) {
   MethodArray = (XxxFuncMethodInfo**)AggrMethArray;
   MethodCount = NumAggrMethods;
 }
 else {
   MethodArray = MethArray;
   MethodCount = NumMethods;
 }

 for (int i = 0; i < MethodCount; i++) {
   if (!MethodArray[i]) continue;
   if (MethodArray[i]->Matches(FuncName, Owner, ArgList, 
			       ReturnType, ReturnMetaInfo, Success) != XXX_OK) {
     XXX_NOTE_ERROR(XXX_BADTYPE, "Error matching method");
     return XXX_NOT_OK;
    }
   if (Success) {
     MethodIndex = i;
     break;
    }
  }
 return XXX_OK;
}

XxxErrCode 
XxxADTClass::parseMetaInfo(const char* MetaInfoString,
			   MetaInfoHandle& MIHandle) const
{
 if (!maintainsMetaInfo) return XXX_OK;

 XXX_NOTE_ERROR(XXX_BADADT, "Should redefine parseMetaInfo!");
 return XXX_NOT_OK;
}

XxxErrCode 
XxxADTClass::makeMetaInfo(const MetaInfoHandle& MIHandle,
			  XxxADTMetaInfo*& MI) const
{
 if (!maintainsMetaInfo) return XXX_OK;

 XXX_NOTE_ERROR(XXX_BADADT, "Should redefine makeMetaInfo!");
 return XXX_NOT_OK;
}

XxxErrCode XxxADTClass::FuncTypeCheck(const char*FuncName,
				      const XxxValueExpression* Owner,
				      const XxxValueExprList *ArgList,
				      XxxBool IsAggr,
				      XxxBasicType& ReturnType,
				      XxxFuncParseInfo*& FuncParse) const
{
 FuncParse = NULL;
 int MethodIndex;

 XxxADTMetaInfo *ReturnMetaInfo;
 XxxBool Success = XXX_FALSE;
 // check for valid method names
 if ((matchMethodName(FuncName, Owner, ArgList, IsAggr,
		      MethodIndex, ReturnType,
		      ReturnMetaInfo, Success) != XXX_OK) || !Success) {
   char buf[100];
   sprintf(buf, "Invalid method name %s for type %s", FuncName, TypeName);
   XXX_NOTE_ERROR(XXX_BADEXPR, buf);
   return XXX_NOT_OK;
  }
 // since it is a valid function, construct a parse data structure
 FuncParse = new XxxFuncParseInfo(FuncName, MethodIndex, NULL, TypeId,
				  ReturnType, ReturnMetaInfo);
 FuncParse->IsAggr = IsAggr;
 return XXX_OK;
}

XxxErrCode XxxADTClass::FuncOptimize(XxxFuncParseInfo* FParse,
				     XxxValueExprPlan* OwnerPlan,
				     XxxValExprPlanList* ArgPlans,
				     XxxFuncPlanInfo*& FuncPlan) const
{
 FuncPlan = NULL;
 if (!FParse) {
   char buf[100];
   sprintf(buf, "Optimizing method of %s with empty parse", TypeName);
   XXX_NOTE_ERROR(XXX_BADADT, buf);
   return XXX_NOT_OK;
  }
 
 XxxFuncPlanInfo* ThisPlanInfo;
 if (methOptimize(FParse, OwnerPlan, ArgPlans, ThisPlanInfo) 
     != XXX_OK) {
   XXX_NOTE_ERROR(XXX_BADADT, "Error optimizing method");
   return XXX_NOT_OK;
  }

 XxxADTEnv *Env = (XxxADTEnv *) XXX_THISADTENV(TypeId);
 if (Env->AdtPipe && ThisPlanInfo) {
   // put in optimizations to avoid materializing on the boundaries
   if (ThisPlanInfo->modifyArgCompression(OwnerPlan, ArgPlans) != XXX_OK) {
     XXX_NOTE_ERROR(XXX_BADADT, "Error modifying compression of functions");
     return XXX_NOT_OK;
    }
  }

 // If the adt has set the suspend computation flag and the type supports
 // callbacks, suspend the plan's execution

 // TODO: For now, we just suspend the first plan.  Ideally, ADTs can decide
 // which portions of a plan should be suspended, and which should be
 // immediately computed.

 if (ThisPlanInfo && Env->SuspendComputation && XxxADTArray[TypeId]->SupportsCallback())
   ThisPlanInfo->SuspendComputation();

 FuncPlan = ThisPlanInfo;
 return XXX_OK;
}

//
// Default implementation assumes no special serialization needed
//
XxxErrCode XxxADTClass::
SerializeCopySize(const XxxADTMetaInfo* MetaInfo,
		  unsigned int& Size) const
{
  Size = MaxObjectSize(MetaInfo);
  if (Size < 1) {
    XXX_NOTE_ERROR(XXX_BADADT, "Variable size : not serializable");
    return XXX_NOT_OK;
  }
  return XXX_OK;
}

XxxErrCode XxxADTClass::
SerializeCopy(const XxxADTValue* Object,
	      const XxxADTMetaInfo* MetaInfo,
	      char* MemLocation) const
{
 TypeCopy(Object, (XxxADTValue *)MemLocation, MetaInfo);
 return XXX_OK;
}
XxxErrCode XxxADTClass::
UnSerializeCopy(const char* MemLocation,
		const XxxADTMetaInfo* MetaInfo,
		XxxADTValue* Object) const
{
 TypeCopy((XxxADTValue *)MemLocation, Object, MetaInfo);
 return XXX_OK;
}

XxxErrCode XxxADTClass::
DefaultVal(XxxADTValue* ObjectRef, 
	   const XxxADTMetaInfo* MetaInfo) const
{
  ObjectRef->Init();
  ObjectRef->Set(XxxADTValue::NULL_VALUE);
  return XXX_OK;
}

//
// Methods of XxxADTMetaInfo::Itemizer
// 
void XxxADTMetaInfo::
Itemizer::InitScan(XxxSetCursor& SCsr, const char* InputString)
{
 CurrentItem = new char[strlen(InputString)+1];
 strcpy(CurrentItem, InputString);
 CurrentString = CurrentItem;
 SCsr.AtEnd = XXX_FALSE;
}

void XxxADTMetaInfo::
Itemizer::NextItem(XxxSetCursor& SCsr, char*& ItemString)
{
 ItemString = strtok(CurrentString, Separator);
 CurrentString = NULL; // needed for subsequent calls to strtok
 if (!ItemString) {
   SCsr.AtEnd = XXX_TRUE;
   return;
  }
}

void XxxADTMetaInfo::
Itemizer::CloseScan(XxxSetCursor& SCsr)
{
 delete [] CurrentItem;
}

//
// Methods of XxxADTMetaInfo::StringTokenizer
// 
void XxxADTMetaInfo::
StringTokenizer::InitScan(XxxSetCursor& SCsr, const char* InputString)
{
 InStr = new istrstream((char *)InputString);
 SCsr.AtEnd = XXX_FALSE;
}

void XxxADTMetaInfo::
StringTokenizer::NextItem(XxxSetCursor& SCsr, char*& ReturnItem)
{
 (*InStr) >> ItemString;
 if ((*InStr).bad() || !ItemString[0]) {
   SCsr.AtEnd = XXX_TRUE;
  }
 ReturnItem = ItemString;
 return;
}

void XxxADTMetaInfo::
StringTokenizer::CloseScan(XxxSetCursor& SCsr)
{
 delete InStr;
}
