#ifndef XXX_SHMEM_H
#define XXX_SHMEM_H

// need to define this for WIN32
#ifndef XXX_WIN32

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <sys/wait.h>

inline XxxErrCode XxxSemDown(int SemId, int SemNum) {
  struct sembuf SemOp;
  SemOp.sem_num = SemNum;
  SemOp.sem_op = -1;
  SemOp.sem_flg = 0;
  while (semop(SemId, &SemOp, 1) < 0) {
    if (errno == EINTR) {
      cerr << "Recvd signal while in semop!!" << endl;
      continue;
    }
    XXX_NOTE_ERROR(XXX_BADEXEC, strerror(errno));
    return XXX_NOT_OK;
  }
  return XXX_OK;
}
inline XxxErrCode XxxSemUp(int SemId, int SemNum) {
  struct sembuf SemOp;
  SemOp.sem_num = SemNum;
  SemOp.sem_op = 1;
  SemOp.sem_flg = 0;
  while (semop(SemId, &SemOp, 1) < 0) {
    if (errno == EINTR) {
      cerr << "Recvd signal while in semop!!" << endl;
      continue;
    }
    XXX_NOTE_ERROR(XXX_BADEXEC, strerror(errno));
    return XXX_NOT_OK;
  }
  return XXX_OK;
}

/** Func plan with shared memory for external function
 *  Here is the layout for the data passed in shared memory:
 *     1) A "valid" flag ::  int
 *     2) Number of args ::  int
 *     3) Owner offset   :: int
 *     4) Arg offsets   :: array of ints
 *     5) Result offset :: int
 *     6) Rest of the data (owner data, args, result)
 *
 *  Each field is aligned to a 8 byte word boundary, to be on the 
 *  safe side.
 */
class ShmFuncPlanInfo : public XxxFuncPlanInfo {
 protected:
  // identifier for shared memory segment
  int ShmId;
  // identifier for semaphores
  int SemId;
  // id for child process
  pid_t ChildPid;

  XxxErrCode createShMem(int ShmSize) {
    key_t ShmKey = IPC_PRIVATE;
    int ShmFlag = IPC_CREAT | 0644;
    if ((ShmId = shmget(ShmKey, ShmSize, ShmFlag)) < 0) {
      XXX_NOTE_ERROR(XXX_BADEXEC, strerror(errno));
      return XXX_NOT_OK;
    }
    return XXX_OK;
  }
  XxxErrCode createSem(int NumSems) {
    key_t SemKey = IPC_PRIVATE;
    int SemFlag = IPC_CREAT | 0644;
    if ((SemId = semget(SemKey, NumSems, SemFlag)) < 0) {
      XXX_NOTE_ERROR(XXX_BADEXEC, strerror(errno));
      return XXX_NOT_OK;
    }
    return XXX_OK;
  }
  XxxErrCode semDown(int SemNum) {
    return XxxSemDown(SemId, SemNum);
  }
  XxxErrCode semUp(int SemNum) {
    return XxxSemUp(SemId, SemNum);
  }
  
public:
  // is this structure properly initialized
  XxxBool Initialized;

  // pointer to start of shared memory segment
  char* ShmAddr;
  // offsets for the various data items that will have
  // to be passed thru shared memory
  int OwnerOffset;
  int *ArgOffsets;
  int ResultOffset;

  XxxValueExprPlan* OwnerPlanPtr;
  XxxValExprPlanList* ArgPlansPtr;
  XxxBasicType RetType;
  XxxADTMetaInfo* RetMInfo;  

  ShmFuncPlanInfo(const char* Name,
		  XxxBasicType Type,
		  int EngIndex,
		  XxxValueExprPlan* OwnerPlan,
		  XxxValExprPlanList* ArgPlans,
		  XxxBasicType ReturnType,
		  XxxADTMetaInfo* ReturnMInfo,
		  XxxGenericPlan* Exp);
  virtual ~ShmFuncPlanInfo();
  XxxErrCode Execute();
};


#endif // ifndef XXX_WIN32

#endif
