#ifndef XXX_THREAD_H
#define XXX_THREAD_H

#include <iostream.h>
#include "shoresm-defs.h"
#include "sfile_handler.h"
// MODIF RELEASE
#if defined(XXX_WIN32) && defined(XXX_NONPREEMPTIVE)
#include "win32_events.h"
#endif
#include "prd-config.h"
#include "error.h"
#include "env.h"
#include "socket.h"


typedef smutex_t XxxBaseMutex;
/**
 // Simplified version of SHORE conditional variable.
 // Instead of a separate cond var and a mutex, this combines them
 */
class XxxCondVar { public:
  XxxBaseMutex Mutex;
  scond_t      Var;

  inline XxxErrCode   Wait() { 
    W_COERCE(Mutex.acquire(0));
    W_COERCE(Var.wait(Mutex));
    Mutex.release();
    return XXX_OK;
   }

  inline XxxErrCode  Signal() { 
    W_COERCE(Mutex.acquire(0));
    Var.signal();
    Mutex.release();
    return XXX_OK;
   }

  inline XxxErrCode BroadCast() {
    W_COERCE(Mutex.acquire(0));
    Var.broadcast();
    Mutex.release();
    return XXX_OK;
   }
  XxxCondVar (const char* Name = 0) : Var(Name) {}
 };

/**
 * The root class for all Predator threads
 */
class XxxBaseThread : public smthread_t {
 protected:
  char valid;
  // all Xxx threads will wait for an event on a file handle
#if defined(XXX_WIN32) && defined(XXX_NONPREEMPTIVE)
  // SHORE V2.0 - New I/O
  win32_compat_event_t* eventHandle;
  HANDLE handle;
#else
  sfile_read_hdl_t*   fileHandle;
#endif
  // the code to be executed for this thread
  virtual void ExecCode() {
    cout << "ERROR: should not invoke virtual ExecCode()!!!" << endl;
  }
 public:
  // a copy of the XXXEnv
  XxxEnvironment XxxEnv;
  // error stack for this thread
  XxxErrInfo Error;
  XxxBaseThread(const char* Name = "default_thread") 
    :
#ifdef SHORE_INTERIM_0
    smthread_t(smthread_t::t_regular, // priority
	       false, // block immediate
	       false, // auto delete off
	       Name, // name
	       WAIT_IMMEDIATE), // lockto
#else
    smthread_t(smthread_t::t_regular, // priority
	       Name, // name
	       WAIT_IMMEDIATE), // lockto
#endif    
    valid(0)
  
  {
#if defined(XXX_WIN32) && defined(XXX_NONPREEMPTIVE)
	  eventHandle = NULL;
#else
	  fileHandle  = NULL;
#endif
  }
  XxxBaseThread(const XxxEnvironment& DefaultEnv,
		const char* Name = "default_thread") 
    :
#ifdef SHORE_INTERIM_0
    smthread_t(smthread_t::t_regular, // priority
	       false, // block immediate
	       false, // auto delete off
	       Name, // name
	       WAIT_IMMEDIATE), // lockto
#else
    smthread_t(smthread_t::t_regular, // priority
	       Name, // name
	       WAIT_IMMEDIATE), // lockto
#endif
    valid(0), XxxEnv(DefaultEnv)
  {
#if defined(XXX_WIN32) && defined(XXX_NONPREEMPTIVE)
	  eventHandle = NULL;
#else
	  fileHandle  = NULL;
#endif
  }

  virtual void ShutDown() {
#if defined(XXX_WIN32) && defined(XXX_NONPREEMPTIVE)
     if (eventHandle) {
      eventHandle->shutdown();
     }
#else
     if (fileHandle) {
      fileHandle->shutdown();
     }
#endif
   }
  virtual ~XxxBaseThread() { 
#if defined(XXX_WIN32) && defined(XXX_NONPREEMPTIVE)
     if (eventHandle) {
      delete eventHandle;
     }
#else
   if (fileHandle)
      delete fileHandle; 
#endif
   }
  virtual void	run() { ExecCode(); }
  char Valid() { return valid;}
  
  // this is how to access the current thread from anywhere in the code
  static XxxBaseThread* me() { return (XxxBaseThread*) smthread_t::me(); }
 };

/**
 * This thread is used to initialize Predator
 */
class XxxInitThread : public XxxBaseThread {
 protected:
  int SockNumber;
  virtual void ExecCode();
  XxxCondVar QuitCond;  // condition used by MonitorThread to tell XXX to shutdown
 public:
  XxxCondVar Shut;  // condition that signals end of XXX run

  XxxInitThread(int argc, char** argv);
  virtual ~XxxInitThread();
 };

/**
 * This thread reads socket input and services it
 */
class XxxServerThread : public XxxBaseThread {
 protected:
  virtual void ExecCode();
  XxxSocketConnection Connection;
  int ServerPortNumber; 
 public:
  XxxCondVar Shut;  // condition that signals end of server run
  XxxServerThread(int PortNumber, const XxxEnvironment& DefaultEnv);
  virtual ~XxxServerThread();
 };

/**
 * This thread is used to provide a front end for the server
 */
class XxxMonitorThread : public XxxBaseThread {
 protected:
  virtual void ExecCode();
  char prompt[XXX_TEXTSTRING_MAX];
  XxxCondVar& QuitCond; // the condition to signal when done
 public:
  XxxCondVar Shut;
  XxxMonitorThread(const char* Prompt, const XxxEnvironment& DefaultEnv,
		   XxxCondVar& QCond);
  virtual ~XxxMonitorThread();
 };

/**
 * This class is used to cleanly acquire and release mutex'es 
 */
class XxxUseMutex {
  XxxBaseMutex& mutex;
 public:
  XxxUseMutex(XxxBaseMutex& Mutex) : mutex(Mutex) {
    W_COERCE(mutex.acquire(0));
   }
  ~XxxUseMutex() {
    mutex.release();
   }
 };

// Now, some very important global inline functions

// Get the current thread of execution
inline XxxBaseThread* XXX_THISTHREAD() { return XxxBaseThread::me();}

// Get the current environment
inline XxxEnvironment& XXX_THISENV() { return XXX_THISTHREAD()->XxxEnv;}

// Get the current ADT environment
inline XxxADTEnv* XXX_THISADTENV(int TypeId)
{ return (XXX_THISENV().ADTEnvs ? XXX_THISENV().ADTEnvs[TypeId] : 0); }

// Get the current engine environment
inline XxxEngineEnv* XXX_THISENGENV(int EngineId)
{ return (XXX_THISENV().EngineEnvs ? XXX_THISENV().EngineEnvs[EngineId] : 0); }

//
// To manipulate the error stack
//
// Get the current error stack (used for later inline function)
inline XxxErrInfo& XXX_THISERROR() { return XXX_THISTHREAD()->Error;}
// Notify an error (used for later inline function)
inline XxxErrCode XXX_ERROR(const char* FileName, int LineNum, 
			    XxxErrCond X, const char* Msg, 
			    int Engine = 0)
{ return XXX_THISERROR().Note(X,Msg,Engine,FileName,LineNum);}

// Notify an error (has to be a macro because of the use of the FILE
// and LINE macros withn it)
#define XXX_NOTE_ERROR(X, Msg) (XXX_ERROR(__FILE__, __LINE__, X, Msg))

// Retrieve the current error
inline XxxErrCode XXX_GET_ERROR(XxxErrInfo* X) { return XXX_THISERROR().Get(X);}
// Print the current error stack
inline void XXX_PRINT_ERROR(ostream& OStr, const char* Msg) 
{ XXX_THISERROR().Print(OStr, Msg); }
// Clear the error stack
inline void XXX_CLEAR_ERROR() { XXX_THISERROR().Clear();}
// Check if there is anything on the error stack
inline XxxBool XXX_IS_ERROR() { return XXX_THISERROR().IsError();}

// check, print and clear the error stack. return 1 if there was an error.
inline int XXX_CHECK_ERROR(const char* Msg) {
  int RetVal = 0;
  if (XXX_IS_ERROR()) {
    XXX_PRINT_ERROR(*(XXX_THISENV().ErrorStr), Msg);
    XXX_CLEAR_ERROR();
    RetVal = 1;
   }
  return RetVal;
 }

#define XXX_CALL(FnCall, ErrCode, ErrMessage, RC) \
 { \
  if ((RC=FnCall) != XXX_OK) { \
    XXX_NOTE_ERROR(ErrCode, ErrMessage); \
   } \
 }

#define XXX_CALL_RET(FnCall, ErrCode, ErrMessage) \
 { \
  if (FnCall != XXX_OK) { \
    XXX_NOTE_ERROR(ErrCode, ErrMessage); \
    return XXX_NOT_OK; \
   } \
 }

#define XXX_BADRC_RCRETURN(RC) \
 { \
  if (RC != XXX_OK) \
    return RC; \
 }

#define XXX_BADRC_VOIDRETURN(RC) \
 { \
  if (RC != XXX_OK) \
    return; \
 }

#endif
