/*****************
This file contains defaults for the compilation of PREDATOR.
*****************/
#ifndef XXX_CONFIG_H
#define XXX_CONFIG_H

#include <limits.h>
#include <float.h>

// defaults for ADTs : these may be changed on occasion
const int XXX_ADT_NAME_MAX = 15;   // maximum length of an ADT name
const int XXX_ADT_MAX = 50;        // maximum number of ADTs
const int XXX_FUNCNAME_MAX = 100;  // maximum length of an ADT function name
const int XXX_LANGNAME_MAX = 50;   // maximum length of an ADT language name
const int XXX_ARGCNT_MAX = 10;     // maximum number of args to an ADT function
const int XXX_MIHNDL_MAX = 100;    // maximum size of meta-info handle
const int XXX_TEXTSTRING_MAX = 5000; // maximum size of XXX_STRING

// the rest are unlikely to be changed

// constants for textual I/O
const int XXX_TEXTLINE_MAX = 3000;     // maximum length of an input line
const int XXX_DIRNAME_MAX = 300;       // maximum length og a directory name
const int XXX_ERRMESSAGE_MAX = 200;    // maximum length of an error message

// general defaults used in utils
const int XXX_ATTR_NAME_MAX  = 15; // name of an attribute
const int XXX_NODE_NAME_MAX  = 15; // name of a node of a query graph
// general defaults used in server
const int XXX_CLIENTS_MAX = 100;   // maximum number of simultaneous clients
const int XXX_PROTOCOL_NAME_MAX = 10; // maximum number of client protocols

// defaults for DB engines
const int XXX_ENGINE_NAME_MAX = 15;   // maximum length of a DB engine name
const int XXX_ENGINE_MAX = 10;        // maximum number of DB engines
const int XXX_SRCNAME_MAX = 25;       // maximum length of a data src name
const int XXX_SUNDRYNAME_MAX = 20;    // maximum length for generic names
const int XXX_REGISTRY_MAX = 49;      // defunct 

#ifdef XXX_DEBUG
#define ASSERT(x) \
	if(!(x)) { \
		cerr << "Assertion ( x ) failed at line " << __LINE__ << " of file \"" << __FILE__ << "\"." << endl; \
      		abort() ;\
	}
#else
#define ASSERT(x) ;
#endif

// Since we also run tools like doc++ over
// our source code, we can use this to
// "comment" out any code that causes the tools
// to crash.
#ifdef XXX_COMPILE
#define XXX_EXTERN_CLASS_DECL(foobar) class foobar
#endif

// Now for Architecture dependent stuff
// for word alignment on disk (should be arch dependent)
const int XXX_ALIGN_SIZE = sizeof(double);
#define XXX_WORD_ALIGN(Value) \
    if ((Value % XXX_ALIGN_SIZE) != 0) \
      Value += XXX_ALIGN_SIZE - (Value % XXX_ALIGN_SIZE);

/** Macro definitions are hpux, mips, __cplusplus, __GNUC__, _WIN32  ***/

// WIN_CHG
#ifdef XXX_WIN32
#include <io.h>
#ifndef _WINSOCKAPI_
#define _WINSOCKAPI_   // Prevent inclusion of winsock.h in windows.h 
#endif
#include <windows.h>
#include <winsock2.h>
#define strcasecmp _stricmp
#define strncasecmp _strnicmp
#define XXX_WRITE(A,B,C) _write(A,B,C)
#define XXX_READ(A,B,C) _read(A,B,C)
// Bison requires the definition of MSDOS
#ifndef MSDOS
#define MSDOS
#endif

#else

#define XXX_WRITE(A,B,C) write(A,B,C)
#define XXX_READ(A,B,C) read(A,B,C)

#endif

// Need to define XXX_BIGENDIAN or XXX_LITTLEENDIAN
#ifdef XXX_WIN32
#define XXX_LITTLEENDIAN
#endif

#ifdef Sparc
#define XXX_BIGENDIAN
#define USE_MONITOR
#endif

#endif /* XXX_CONFIG_H */
