#ifndef _BOUNDING_BOX_H_
#define _BOUNDING_BOX_H_

#include <iostream.h>
#include <string.h>
#include <nbox.h>
#include <types.h>

/**
  * This is a bounding  box class to be used with R*trees
  * Conceptually, it represents the smallest bounding multidimensional
  * box for some object. Theoretically the dimensions can be of any type.
  * For example, you could have a 4 dimensional box, with 2 integer dimensions
  * a floating point, and maybe a complex dimension.
  * However, right now, only bounding boxes with same type dimensions are 
  * supported. 
  **/
class XxxBoundingBox {
private:
  XxxBasicType dimension_type;
  nbox_t *box;
public:
  // constructors, destructors.
  XxxBoundingBox();
  /**
   * This is the default constructor. The parameters are the # of dimensions,
   * a string describing the type (e.g "Int")
   * and a pointer to a bunch of values of that type.
   * More specifically, for a box of type T, and dimensionality D,
   * _box would be an array of 2*D values of type D where the values
   * are represent 2 points a and b, and the values 
   * are a0, a1, .... aD-1, b0, b1,.. bD-1
   */
  XxxBoundingBox(const int dim,const XxxBasicType type,void *_box);
  XxxBoundingBox(const XxxBoundingBox& B);
  XxxBoundingBox(const nbox_t& _box);
  ~XxxBoundingBox();
  
  // operators
  friend ostream& operator<<(ostream&,const XxxBoundingBox&);
  XxxBoundingBox& operator=(const XxxBoundingBox& B);

  // extractors
  int Dimension() const;
  XxxBasicType Type() const { return dimension_type;}

  /// This function returns a SHORE nbox_t object from a XxxBoundingBox object
  nbox_t ShoreBox() const { return *box; }
  const void *Data() { return box->kval(); }
};
  
inline int 
XxxBoundingBox::Dimension() const
{
  if (box!=NULL)
    return box->dimension();
  else
    return -1;
}

#endif









