#ifdef XXX_JAVA_VIRTUAL_MACHINE

#include "..\jaguar\ClientSiteOperator.h"

#else

#ifndef CLIENTSITEOPERATOR_H
#define CLIENTSITEOPERATOR_H

#include "relplanop.h"
#include "planspj.h"
#include "socket.h"

//////////////////////////////////////////////////////////////////////
/// RelClientSitePlanOp //////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

class XxxClientSiteSenderThread;
class BatchCounter;

class RelClientSitePlanOp : public RelDerivedPlanOp {
  friend class XxxClientSiteSenderThread;
protected:
  
  XxxBool initOK;

  /* In the following we identify among all targets the client site ones.
  These are the targets which are not just a column reference, but an 
  actual expression. Simple column references, also in the case of the 
  predicate expression, are not sent to the client. */

  // the predicate expression used to select on the client
  XxxValueExprPlan* PredicateExpressionPlan;
  // all target expressions in the projection of this operator
  XxxValExprPlanList* TargetExpressionPlans;
  // the target expressions which are produced on the client
  XxxValExprPlanList* ClientSiteTargetPlans;
  // 
  XxxValueExprList *InputUnknowns;
  XxxValExprPlanList *ReferencedUnknownPlans;
  // overall number of targets (c.s. or not)
  int NumberOfTargets;
  // overall number of unknowns in the predicate and the targets.
  int NumberOfUnknowns;  
  int NumberOfClientSiteTargets;
  int *ClientSiteTargets;
  // indices of targets that are just unknowns, and as such not c.s.
  // otherwise -1 (for c.s. targets)
  int *ServerSiteTargetsIndices; 
  // the number of referenced unknowns in the predicate 
  int UnknownsPerPredicate;
  // unknowns, that are needed on the c.s., per target
  int *UnknownsPerTarget;
  // is the predicate evaluated on the c.s.?
  XxxBool ClientSitePredicate;

  // used as size for the record buffer and for the BatchCounter
  int BatchSize;

  /* Buffer of records that were sent but whose results are not yet received.
   * The buffer is initialized with the size, which determines the upper limit
   * for records that are concurrently in the sender - client - receiver 
   * pipeline. The actual synchronization for this 'concurrency factor'
   * is done by a BatchCounter, whose initial value should be less or equal
   * to the buffer size. */
  class Buffer {    
  private:
	// points to a XxxRecord pointer array of size Size
    XxxRecord **RecBuffer;
    int Size;
	/* pointers into the array. Write points to the next free slot.
	 Read points to the next readable slot. If Read equals Write
	 then the buffer is either empty and cannot be read or it is full,
	 and can be read but not written (then and only then full is true). */
    int Read, Write;
    XxxBool full;
  public:
    Buffer ( int Size );
    ~Buffer ();
    void Add ( XxxRecord* Record );
    void Fetch ( XxxRecord* &ReturnRec );
    XxxBool Empty () const
    { return ( ( Read == Write ) && !full ); }
    XxxBool Full () const
    { return ( full ); }
  };
  /* Handle has to encapsulate all dynamic (changing with the processed 
	tuples) state. This due to the architecture or Predator's operators.
	*/
  class Handle {
  public:
    // Semaphore used to synchronize before sending and receiving. 
    BatchCounter* BCounter;

    XxxBool initOK;

	/* TODO : port numbers for the socket server for relational operators
	are now always 6020 which certainly won't do for multiple operators */
    static XxxSocketConnection* ServerConnection;

	/* The actual connection which is used for the data. */
    XxxSocketConnection* Connection;
	/* stream related to Connection */
    ostream *OutStream;

	/* The thread that sends the arguments to the client and enqueues them
	on the buffer. If Sender is null, this is done within the same thread. */
    XxxClientSiteSenderThread* Sender;

    XxxValExprPlanList* ClientSiteOperations;
    int NumberOfClientSiteOperations;
    
    // just for testing
    FILE *inFile, *outFile;

	/* the buffer used for the records whose argument columns 
	are in the pipeline */
    Buffer *ChildRecBuffer;

    XxxBool BatchingDone;  // nothing more to be processed

    Handle ( XxxValExprPlanList* ClientSiteOperations,
	     int NumberOfClientSiteTargets,
	     int BatchSize, 
	     XxxDerivedRelation::CursorInfo* Handle,
	     XxxRelation* *RelationArray,
	     const RelClientSitePlanOp *PlanOp );
    ~Handle ();

    XxxBool wasInitOK()
    { return XXX_TRUE; }

    // helper function for the constructors  
    XxxErrCode getClientTaskConnection ();
    
    XxxErrCode initializeClient ();

  };

  XxxBool BatchFull(Handle *OpHandle) const;
  XxxBool BatchEmpty(Handle *OpHandle) const;

  XxxErrCode NextOfBatch(Handle *OpHandle,
			 XxxRecord* &ReturnRec) const;

  XxxErrCode SendToClient(Handle *OpHandle,
			  XxxRecord *ChildRecords[],
			  XxxRecordSchema *ChildSchemas[]) const;

  XxxBool BatchDone (Handle *OpHandle) const;

  void EndBatch (Handle *OpHandle) const;

  XxxErrCode GetNext ( XxxSetCursor& SCsr,
		       XxxRelation* *RelationArray,
		       XxxRecord*& ReturnRec ) const;

  XxxErrCode GetStoredResults ( XxxDerivedRelation::CursorInfo* Handle,
				RelClientSitePlanOp::Handle *OpHandle,
				XxxRecord *Record ,
				XxxValuesArray *&Results ,
				XxxADTValue *&PredicateResult ) const
  {     Results = 0; return XXX_OK; }

  XxxErrCode GetClientResults ( XxxDerivedRelation::CursorInfo* Handle,
				RelClientSitePlanOp::Handle *OpHandle,
				const XxxRecordSchema* Schema,
				XxxRecord *Record ,
				XxxValuesArray *&Results ,
				XxxADTValue *&PredicateResult,
				bool &DropRecord ) const;
   
  XxxErrCode StoreResults ( Handle *OpHandle,
			    XxxRecord *Record ,
			    XxxValuesArray *Results ) const
  { return XXX_OK; }  

  XxxErrCode Join ( XxxValuesArray *Results ,
		    XxxADTValue *&PredicateResult ,
		    XxxRecord *InputRec,
		    XxxRecord* &ReturnRec,
		    const XxxRecordSchema* Schema,
		    const XxxRecordSchema* ChildSchema,
		    bool &DropRecord ) const;

  // helper function for the constructors
  XxxErrCode determineInputUnknownsAndClientSiteTargets ();
  /* mmaasha
  XxxErrCode BlockingReceive ( XxxSocket s,
			       char* buffer,
			       int length,
			       int flags ) const;
			       */

public:

  RelClientSitePlanOp ( RelPlanOp** ChildPlanOps,
			int NumChildren,
			RelSPJOptimizer::PredPlanStruct& PredPlans,
			XxxValExprPlanList* ExprPlanList,
			RelCatalogInfo* CatInfo,
      int BatchSize = 1) : 
  //ivan@2.5.99 added dummy
  RelDerivedPlanOp ( 1, ChildPlanOps, CatInfo )
  { }

  virtual ~RelClientSitePlanOp()
  //ivan@2.5.99 added dummy
  { }

  XxxBool wasInitOK() const
  { return initOK; }
  
  // These methods are called from XxxDerivedRelation::InitCursor 
  // and XxxDerivedRelation::CloseCursor
  virtual XxxErrCode InitHandle(XxxDerivedRelation::CursorInfo* Handle,
				XxxRelation* *RelationArray) const
  //ivan@2.5.99 added dummy
  { return XXX_NOT_OK; }
  virtual XxxErrCode CloseHandle(XxxDerivedRelation::CursorInfo* Handle,
				 XxxRelation* *RelationArray) const
  //ivan@2.5.99 added dummy
  { return XXX_NOT_OK; }
  virtual XxxErrCode GetNextRecord(XxxSetCursor& SCsr,
				   XxxRelation* *RelationArray,
				   const XxxRecordSchema* Schema,
           XxxRecord*& ReturnRec) const 
  //ivan@2.5.99 added dummy
  { return XXX_NOT_OK; }

  virtual void NodePrint (ostream &OStr, int curBlockId, int Level, 
			  int numChildren, int *childrenIds) const
  //ivan@2.5.99 added dummy
  { }

  virtual const char *PrintString()  const 
  { return "Client Side UDF" ;}
};


#endif // CLIENTSITEOPERATOR_H

#endif // XXX_JAVA_VIRTUAL_MACHINE





