#include "Report.h"
#include "adtfunc.h"
#include "funcexpr.h"

bool XxxFunctionReport::everCalled = false;
bool XxxFunctionReport::newReport = true;
char *XxxFunctionReport::serverSiteFunctions = 0;
char *XxxFunctionReport::endOfServerSiteFunctions = 0;
char *XxxFunctionReport::clientSiteFunctions = 0;
char *XxxFunctionReport::endOfClientSiteFunctions = 0;
int XxxFunctionReport::numClassLoaders=0;
char **XxxFunctionReport::classLoaders=0;
char **XxxFunctionReport::endOfClassLoaders=0;
int XxxFunctionReport::NumADTs = 1;
XxxBasicType XxxFunctionReport::ADTs[] = { XXX_GENFUNC }; 
const int XxxFunctionReport::MAX_LENGTH = 1000;
const int XxxFunctionReport::MAX_NUM_CL = 25;

void XxxFunctionReport::checkInitialization()
{
    if(!everCalled) {
        everCalled = true;
        newReport = true;
        endOfServerSiteFunctions = serverSiteFunctions = new char[MAX_LENGTH];
        endOfClientSiteFunctions = clientSiteFunctions = new char[MAX_LENGTH];
        serverSiteFunctions[0] = 0;
        clientSiteFunctions[0] = 0;
        numClassLoaders = 0;
        endOfClassLoaders = classLoaders = new char*[MAX_NUM_CL];
    }
}

void XxxFunctionReport::deleteInformation()
{
    // the initialization check, should be in every function
    checkInitialization();

    // reset everything
    serverSiteFunctions[0] = 0;
    endOfServerSiteFunctions = serverSiteFunctions;
    clientSiteFunctions[0] = 0;
    endOfClientSiteFunctions = clientSiteFunctions;
    numClassLoaders = 0;
}

void XxxFunctionReport::recollectInformation()
{
    // the initialization check, should be in every function
    checkInitialization();

    // delete all old information
    deleteInformation();

    // now go through the method tables
    for(int i=0; i<NumADTs; i++) {
        int NumMethods = XxxADTArray[ADTs[i]]->GetNumMethods();
        for(int j=0; j<NumMethods; j++) {
            XxxFuncMethodInfo *FMInfo =
                XxxADTArray[ADTs[i]]->GetMethByIndex(j);
            AddMethod(ADTs[i], FMInfo);
            }
        }

    // for the moment, forget about classloaders (TOBIAS:TODO)

}

XxxErrCode XxxFunctionReport::AddMethod(XxxBasicType Type, XxxFuncMethodInfo *FMInfo)
{
    // the initialization check, should be in every function
    checkInitialization();

    bool found = false;
    for(int i=0; i<NumADTs; i++) 
        if( ADTs[i] == Type ) {
            found = true;
            break;
        }

    if(!found)
        return XXX_OK;
    
    if(FMInfo->IsClientSiteExecutable()) {

        strcpy(endOfClientSiteFunctions, XxxADTArray[Type]->TypeName);
        endOfClientSiteFunctions += strlen(XxxADTArray[Type]->TypeName);
        endOfClientSiteFunctions++[0] = ' ';
        strcpy(endOfClientSiteFunctions, FMInfo->Name);
        endOfClientSiteFunctions += strlen(FMInfo->Name);
        endOfClientSiteFunctions++[0] = '\n';
        endOfClientSiteFunctions[0] = 0;


    } else {

        strcpy(endOfServerSiteFunctions, XxxADTArray[Type]->TypeName);
        endOfServerSiteFunctions += strlen(XxxADTArray[Type]->TypeName);
        endOfServerSiteFunctions++[0] = ' ';
        strcpy(endOfServerSiteFunctions, FMInfo->Name);
        endOfServerSiteFunctions += strlen(FMInfo->Name);
        endOfServerSiteFunctions++[0] = '\n';
        endOfServerSiteFunctions[0] = 0;

    }

    return XXX_OK;
}

XxxErrCode XxxFunctionReportAddMethod(XxxBasicType Type, XxxFuncMethodInfo *FMInfo)
{
#ifdef NO_FUNCTION_REPORTS
    return XXX_OK;
#else
    return XxxFunctionReport::AddMethod(Type,FMInfo);
#endif
}

        
XxxErrCode XxxFunctionReport::PrintReport(ostream *ostr)
{
    // the initialization check, should be in every function
    checkInitialization();

    newReport = true;

    return PrintIncrementalReport(ostr);
}


XxxErrCode XxxFunctionReport::PrintIncrementalReport(ostream *ostr)
{
    // the initialization check, should be in every function
    checkInitialization();

    // a new report is done by reinitializing everything
    if(newReport) {
        newReport = false;
        // collect all information
        recollectInformation();
    }

    // just output the data collected since the last report
    *ostr << serverSiteFunctions;
    *ostr << '\n';
    *ostr << clientSiteFunctions;
    *ostr << '\n';
    for(int i=0; i<numClassLoaders; i++) {
        *ostr << classLoaders[i];
        *ostr << '\n';
    }
    *ostr << '\n';

    // printed information should be deleted
    deleteInformation();

    return XXX_OK;
}

