#include <stdio.h>
#include <stdlib.h>
#include <string.h>

//Bjarke, aug. 6: #pragma interface/implementation not needed any longer
//#ifdef __GNUG__
//#pragma implementation
//#endif

#include "prd-config.h"
#include "error.h"
#include "xxx_thread.h"
#include "absquery.h"
#include "record.h"
#include "values.h"

// static members
const XxxQueryDag::TraversalType XxxQueryDag::DEPTH_FIRST_POSTFIX = 0;
const XxxQueryDag::TraversalType XxxQueryDag::DEPTH_FIRST_PREFIX = 1;

//
// Methods of XxxQueryNode
//
// OWNER: as  DATE: 02/24/97
// PURPOSE: Constructor
// NOTES: The parent and childNumber fields have been added to facilitate query
// rewrite
XxxQueryNode::XxxQueryNode(int ChildArity, XxxQueryNode** Children,
			   int NumNames, char** NodeNames, XxxQueryNode *Par,
			   int ChildNum)
{
 int i;
 if (ChildArity>0)
   children = new XxxQueryNode*[ChildArity];
 else
   children=NULL;
 arity = ChildArity;
 parent=Par;
 childNumber=ChildNum;
 blockId = 0;
 for (i = 0; i < arity; i++) {
   children[i] = Children[i];
   children[i]->SetChildNumber(i);
   children[i]->refCnt++;
  }
 refCnt = 0;
 nameInQuery = new XxxQueryNameStruct();

 for (i = 0; i < NumNames; i++) {
   if (i == 0) {
     strncpy(nameInQuery->Name, NodeNames[i], XXX_NODE_NAME_MAX);
     nameInQuery->Name[XXX_NODE_NAME_MAX-1] = 0;
    }
   else {
     AddName(NodeNames[i]);
    }
  }
}

XxxQueryNode::~XxxQueryNode() 
{
 XxxQueryNameStruct* Tmp;
 while (nameInQuery) {
   Tmp = nameInQuery->Next;
   delete nameInQuery;
   nameInQuery = Tmp;
  }

 // destroy the children
 for (int i = 0; i < Arity(); i++)
   children[i]->Destroy();
 if (children) 
   delete [] children;
}

// OWNER: as  DATE: 02/24/97
// PURPOSE: Replace child # 'Index' with 'NewChild', returns 'OldChild'
// NOTES: added to facilitate query rewrite
XxxErrCode XxxQueryNode::ReplaceChild(int Index, XxxQueryNode *NewChild, 
				      XxxQueryNode *&OldChild) {
  if (Index>=arity) {
    char buf[100];
    sprintf(buf,"XxxQueryNode::ReplaceChild: Tried to reset child querynode "
	    "%d when there are only %d child nodes!",Index,arity);
    XXX_NOTE_ERROR(XXX_BADEXPR, buf);
    return XXX_NOT_OK;
  }    

  OldChild = children[Index];
  children[Index]=NewChild;
  children[Index]->SetChildNumber(Index);
  children[Index]->SetParent(this);
  return XXX_OK;
}

// OWNER: as  DATE: 02/24/97
// PURPOSE: Sets child # 'Index' to 'NewChild'
// NOTES: added to facilitate query rewrite
XxxErrCode XxxQueryNode::SetChild(int Index, XxxQueryNode *NewNode) {
  XxxQueryNode *OldChild;
  ReplaceChild(Index, NewNode, OldChild);

  if(OldChild != NULL)
    OldChild->Destroy();

  return XXX_OK;
}

// OWNER: as  DATE: 02/24/97
// PURPOSE: add new children to the end of the children list
// NOTES: added to facilitate query rewrite
XxxErrCode XxxQueryNode::AddChildren(XxxQueryNode **NodeList, int Quantity) {
  if (Quantity==0)
    return XXX_OK;
  XxxQueryNode **NewChildList=new XxxQueryNode*[arity+Quantity];
  if (NewChildList==NULL) {
    char buf[100];
    sprintf(buf,"XxxQueryNode::AddChildren: Couldn't allocate more memory to" 
	    " add children nodes!");
    XXX_NOTE_ERROR(XXX_BADEXPR, buf);
    return XXX_NOT_OK;
  }    
  for (int i=0;i<arity;i++) 
    NewChildList[i]=children[i];
  for (int j=0;j<Quantity;j++) {
    NewChildList[arity+j]=NodeList[j];
    NewChildList[arity+j]->SetChildNumber(arity+j);
    NewChildList[arity+j]->SetParent(this);
  }
  arity+=Quantity;
  if (children!=NULL)
    delete[] children;
  children=NewChildList;
  return XXX_OK;
}

void XxxQueryNode::Destroy()
{
 if (--refCnt <= 0)
   delete this;
}

void XxxQueryNode::Print(ostream& OStr) const 
{
 PrintThis(OStr);
 if (Arity()) {
   OStr << "(";
   for (int i = 0; i < Arity(); i++) {
     GetChild(i)->Print(OStr);
     if (i < Arity() - 1)
       OStr << ",";
    }
   OStr << ")" ;
  }
}

XxxQueryNode* XxxQueryNode::RemoveDuplicates(XxxQueryNode *Root)
{
 // if base node, 
 if (Arity() == 0)
   return Root->CheckNode(this);

 for (int i = 0; i < Arity(); i++) 
   children[i] = children[i]->RemoveDuplicates(Root);

 return this;
}

XxxQueryNode* XxxQueryNode::CheckNode(XxxQueryNode* Node)
{
 // Node must be a leaf node
 ASSERT((Node && !Node->Arity()));

 if ((void *)Node == (void *)this) return Node;

 //if ((!strcmp(nameInQuery->Name, Node->nameInQuery->Name)) ||
 //(!strcmp(nameInQuery->Name, Node->name))) {
 if (MatchNodeName(Node->nameInQuery)) {
   // this is the duplicate that we have been looking for
   // note:: replace all uses of Node by this
   XxxQueryNameStruct* TmpName = new XxxQueryNameStruct();
   TmpName->Next = nameInQuery->Next;
   nameInQuery->Next = TmpName;
   strcpy(nameInQuery->Next->Name, Node->nameInQuery->Name);
   this->refCnt += Node->refCnt;
   delete Node;
   return this;
  }

 for (int i = 0; i < Arity(); i++) {
   XxxQueryNode* AnswerNode = children[i]->CheckNode(Node);
   if (AnswerNode != Node) return AnswerNode;
  }
 return Node;
}

int XxxQueryNode::FindValueAttrIndex(XxxUnknownValue *Value) const
{
 XxxRecordSchema* Sch = Schema();

 if (!Sch) {
   XXX_NOTE_ERROR(XXX_INTERNAL, "Calling FindValueAttrIndex without schema!");
   return -1;
  }

 XxxBool ConsiderThis = XXX_TRUE;
 if (strcmp(Value->XxxName, ""))
   ConsiderThis = nameInQuery->MatchNodeName(Value->XxxName);
 // else it is not defined : this happens for temporary attributes
 // that are automatically assigned by the cool code in relquery.C
 // written by the amazing Paskin

 // ConsiderThis is a complicated scheme for pursuing names
 // across several query blocks! I wrote it, and i dont understand
 // it, let alone somebody else having to understand it.
 // Instead, use the easier scheme of forcing the names to match at 
 // each query block! Eventually, remove the need to pass ConsiderThis
 // to FindValueAttr
 // if (!ConsiderThis) return -1;

 for (int i = 0; i < Sch->Arity(); i++) {
   // Value is either of the form A.#1 or A.colname
   if (Value->SrcAttrIndex >= 0) { // A.#1
     if (i == Value->SrcAttrIndex) {
       Value->BaseType = (*Sch)[i].Type;
       return i;
      }
    }
   else if ((*Sch)[i].FindValueAttr(this, Value, ConsiderThis)) {
     Value->BaseType = (*Sch)[i].Type;
     return i;
    }
  }

 return -1;
}

XxxBool XxxQueryNode::MatchNodeName(XxxQueryNameStruct* MatchNames) const
{
 while (MatchNames) {
   if (nameInQuery->MatchNodeName(MatchNames->Name))
     return XXX_TRUE;
   MatchNames = MatchNames->Next;
  }
 return XXX_FALSE;
}

//
// Methods of XxxQueryDag
//
static void NodeIsEqual(XxxQueryNode*& Node, void *TmpStructPtr)
{
 if (Node == ((XxxQueryDag::NodeStackSearch *)TmpStructPtr)->TestNode) 
   ((XxxQueryDag::NodeStackSearch *)TmpStructPtr)->Found = XXX_TRUE;
}
XxxQueryDag::
NodeStackSearch::NodeStackSearch(XxxQueryNode* Node,
				 XxxQueryDag::NodeStack* NStack)
{
 TestNode = Node;
 Found = XXX_FALSE;

 NStack->ApplyToDepth(NStack->Height(), NodeIsEqual, this);
 if (Found == XXX_TRUE) {
   XxxQueryDag::NodeStack* TmpStack = new XxxStack<XxxQueryNode*>;
   // pop from NStack and push onto TmpStack
   XxxQueryNode* TmpNode;
   while (1) {
     if (NStack->IsEmpty() == XXX_TRUE) {
       XXX_NOTE_ERROR(XXX_BADPARSE, "Error in stack of query blocks");
       delete TmpStack;
       return;
      }
     NStack->PopStack(TmpNode);
     if (TmpNode == Node) break;  // break when the duplicate is found
     // otherwise push onto the TmpStack
     TmpStack->PushStack(TmpNode);
    } // while
   // now push back everything from TmpStack to NStack
   while (TmpStack->IsEmpty() == XXX_FALSE) {
     TmpStack->PopStack(TmpNode);
     NStack->PushStack(TmpNode);
    }
   delete TmpStack;
  }
}

XxxErrCode XxxQueryDag::prefixStack(NodeStack* NStack, 
				    XxxQueryNode* Node)
{
 // Check if this node is already on the stack
 // This probably need only be done if RefCnt() > 1.
 // Checking is done within constructor!
 NodeStackSearch Checker(Node, NStack); 

 NStack->PushStack(Node);
 for (int i = 0; i < Node->Arity(); i++) {
   prefixStack(NStack, Node->GetChild(i));
  }
 return XXX_OK;
}

XxxErrCode XxxQueryDag::postfixStack(NodeStack* NStack, 
				     XxxQueryNode* Node)
{
 // Check if this node is already on the stack.
 // This probably need only be done if RefCnt() > 1
 // Checking is done within constructor!
 NodeStackSearch Checker(Node, NStack);

 for (int i = 0; i < Node->Arity(); i++) {
   postfixStack(NStack, Node->GetChild(i));
  }
 NStack->PushStack(Node);
 return XXX_OK;
}


//
// We need to establish a traversal of the query graph in 
// a a bottom-up manner. One way to do this is to first
// push all the nodes onto a stack in a depth-first prefix
// manner. This means that children nodes appear higher in 
// the stack than the parents. Popping the stack gives the
// desired bottom-up traversal. The presence of CSEs makes
// life more difficult. A CSE should occur only once in
// the traversal. There are two ways of ensuring this:
//   1) If a node is already present on the stack, it is not 
//      pushed on it again. However, to make sure that
//      all bottom-up dependencies are met, the lower
//      stack entry must be deleted, and the node should
//      be added at the top.
//   2) Push duplicate copies of the node. When reading from
//      the stack, check if it is a duplicate, and if so
//      delete the copies from within the stack.
//
// In both these, the deletion from within the stack is the painful
// part. So, choose (1) which at least builds a correct stack.
//
XxxErrCode XxxQueryDag::InitTraversal(void *&Cursor, 
				      TraversalType TType,
				      XxxQueryNode* RootNode) const
{
 NodeStack* NStack = new XxxStack<XxxQueryNode*>;
 // traverse the query and push the nodes onto the stack
 if (TType == DEPTH_FIRST_POSTFIX) {
   prefixStack(NStack, RootNode);
  }
 else { // DEPTH_FIRST_PREFIX
   postfixStack(NStack, RootNode);
  }
 Cursor = (void *)NStack;
 return XXX_OK;
}
XxxErrCode XxxQueryDag::TraverseNext(void *Cursor, 
				     XxxQueryNode*& Node,
				     XxxBool& Success) const
{
 NodeStack* NStack = (NodeStack*)Cursor;
 if (NStack->IsEmpty()) {
   Success = XXX_FALSE;
   Node = NULL;
  }
 else {
   Success = XXX_TRUE;
   NStack->PopStack(Node);
  }
 return XXX_OK;
}
XxxErrCode XxxQueryDag::EndTraversal(void *Cursor) const
{
 NodeStack* NStack = (NodeStack*)Cursor;
 delete NStack;
 return XXX_OK;
}

