//
//  This is a provided as a library to help build queries for each data type
//
#ifndef XXX_ABSQUERY_H
#define XXX_ABSQUERY_H
#include <string.h>
#include "prd-config.h"
#include "types.h"
#include "generic.h"

//These are the classes defined here for abstract query graphs
XXX_EXTERN_CLASS_DECL(XxxViewClass);
XXX_EXTERN_CLASS_DECL(XxxViewList);

XXX_EXTERN_CLASS_DECL(XxxQueryNode);
XXX_EXTERN_CLASS_DECL(XxxQueryNameStruct);
XXX_EXTERN_CLASS_DECL(XxxRecordSchema);
XXX_EXTERN_CLASS_DECL(XxxUnknownValue);
XXX_EXTERN_CLASS_DECL(XxxArgEnv);

/**
// Encapsulation of a view. Portions of the view may
// have further XxxViewClass objects embedded inside it.
*/
class XxxViewClass {
 public:
  char ViewName[XXX_NODE_NAME_MAX];
  XxxViewClass() {}
  XxxViewClass(const char *HeadName) {
    strcpy(ViewName, HeadName);
    }
  virtual ~XxxViewClass() { }
  void Destroy() { delete this; }
  virtual void Print(ostream& OStr) const = 0;
  // methods needed by the XxxTable template
  virtual int Hash() const { return (ViewName[0] + ViewName[1]);}
  virtual int operator==(const XxxViewClass& View1) const
   { return !strcmp(ViewName,View1.ViewName);}  

 };

/**
  */
class XxxViewList {
 public:
  XxxViewClass* View;
  XxxViewList* Next;
  XxxViewList(XxxViewClass* V, XxxViewList* N) { View = V; Next = N;}
 };

/**
 *
 */
class XxxQueryNameStruct {
 public:
  char Name[XXX_NODE_NAME_MAX];
  XxxQueryNameStruct* Next;
  XxxQueryNameStruct(const char* NodeName = "__dummy") { 
    Next = NULL;
    strncpy(Name, NodeName, XXX_NODE_NAME_MAX);
    Name[XXX_NODE_NAME_MAX-1] = 0;
    }
  ~XxxQueryNameStruct() {}
  XxxBool MatchNodeName(const char *MatchName) const  {
    if (!strcmp(Name, MatchName)) return XXX_TRUE;
    else 
      if (Next) return Next->MatchNodeName(MatchName);
      else
	return XXX_FALSE;
   }
  void Print(ostream& OStr) const {
    // if this is a dummy name, skip it
    if (Name && strcmp(Name, "__dummy"))
      OStr << Name << ": ";
    if (Next)
      Next->Print(OStr);
  }
};

/**
// Abstract class for nodes of a query graph. Each node is one
// unit of optimization. Each algebra will special case this 
// with its operators.
*/
class XxxQueryNode {
 protected:
  int refCnt; // needed because of sharing of nodes
  XxxQueryNode **children;
  XxxQueryNode *parent;
  int childNumber; // my index in my parent's children array
  int arity;
  char blockId;

  // this is a list of names for this node.
  // if the node has a real name (as against an alias), it is
  // at the head of the list. Use this for looking up catalogs, etc.
  XxxQueryNameStruct* nameInQuery;
  // used to see if this node has the same name as Node. Used to remove
  // duplicates
  virtual XxxQueryNode* CheckNode(XxxQueryNode* Node);
 public:
  int Arity() const { return arity;}
  virtual void PrintThis(ostream& OStr) const = 0;
  virtual void Print(ostream& OStr) const;
  virtual void TreePrint(ostream& OStr, int &curBlockId, 
			 int Level = 0) const = 0;
  virtual void NodePrint(ostream &OStr, int curBlockId, int Level, 
			 int numChildren, int *childrenIds) const = 0;
  virtual XxxQueryNode* GetChild(int i) const { 
    if (i >= arity) return NULL;
    else return children[i];
   }
  XxxQueryNode() { 
    children = NULL; arity = 0; refCnt = 0;
    childNumber = -1;
    nameInQuery = new XxxQueryNameStruct();
    nameInQuery = NULL;
    parent = NULL;
    blockId = 0;
  } // --> Anil Sachdeva 2/24/96
  XxxQueryNode(int ChildArity, XxxQueryNode **Children,
	       int NumNames, char **NodeNames, XxxQueryNode *Par=NULL,
	       int ChildNum=-1); // --> Anil Sachdeva 2/24/96

  // The following methods are used in setting and getting blockId's for the
  // query node tree (used in printing block info for GUI of query node tree.)
  char GetBlockID() {return blockId;}
  void SetBlockIDs(); //--> sandman 02/26/97

  // All of these members were added to facilitate query rewrite -- Anil
  void SetChildNumber(int NewChildNum) { childNumber=NewChildNum; }
  int WhichChildAmI() const { return childNumber; } 
  void SetParent(XxxQueryNode *Par) { parent=Par; }
  XxxQueryNode *GetParent() const { return parent; }  
  XxxErrCode AddChildren(XxxQueryNode **NodeList, int Quantity);
  XxxErrCode SetChild(int Index, XxxQueryNode *NewNode);
  XxxErrCode ReplaceChild(int Index, XxxQueryNode *NewChild, 
			  XxxQueryNode *&OldChild);

  virtual ~XxxQueryNode() ; // inline this later
  virtual void Destroy() ; // inline this later

  // go thru the dag and unite base nodes that are duplicates
  virtual XxxQueryNode* RemoveDuplicates(XxxQueryNode *Root);

  virtual XxxErrCode PropagateMetaInfoUp(const XxxArgEnv* Env = NULL) = 0;
  virtual XxxErrCode PropagateMetaInfoDown() = 0;

  void AddName(const char* NewName) {
    XxxQueryNameStruct* Tmp = new XxxQueryNameStruct(NewName);
    Tmp->Next = nameInQuery->Next;
    nameInQuery->Next = Tmp;
   }
  const char* BasicName() { return nameInQuery->Name; }
  XxxBool MatchNodeName(XxxQueryNameStruct* MatchNames) const;
  XxxBool MatchNodeName(const char* MatchName) const {
    return nameInQuery->MatchNodeName(MatchName);
   }

  void IncrRef() { ++refCnt; }
  void DecrRef() { --refCnt; }
  virtual XxxRecordSchema* Schema() const { return NULL;}
  int FindValueAttrIndex(XxxUnknownValue *Value) const;
 };

/**
// This is used to create a dag of query operators.
// There is some redundancy between this and the methods of XxxQueryNode,
// which also do graph traversals. Resolve this at some point.
*/
class XxxQueryDag : public XxxGenericParse {
 public:
  typedef int TraversalType;
  static const TraversalType DEPTH_FIRST_POSTFIX;
  static const TraversalType DEPTH_FIRST_PREFIX;

  typedef XxxStack<XxxQueryNode*> NodeStack;

  struct NodeStackSearch {
    XxxQueryNode* TestNode;
    XxxBool Found;
    NodeStackSearch(XxxQueryNode* Node,
		    NodeStack* NStack);
   };


  XxxQueryDag() : XxxGenericParse() { }
  virtual ~XxxQueryDag() {}

  // methods for a traversal of the query graph
  virtual XxxErrCode InitTraversal(void *&Cursor,
				   TraversalType TType,
				   XxxQueryNode* RootNode) const;
  virtual XxxErrCode TraverseNext(void *Cursor, 
				  XxxQueryNode*& Node,
				  XxxBool& Success) const;
  virtual XxxErrCode EndTraversal(void *Cursor) const;
 private:
  static XxxErrCode prefixStack(NodeStack* NStack, XxxQueryNode *Node);
  static XxxErrCode postfixStack(NodeStack* NStack, XxxQueryNode *Node);
 };

#endif
