#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <iostream.h>
#include "aggrexpr.h"
#include "funcexpr.h"
#include "prd-config.h"
#include "strrules.h"
#include "adtfunc.h"


// Methods of AggrPlan
XxxFuncExpression::
AggrPlan::AggrPlan(const AggrPlan &Plan) : 
  XxxFuncExpression::FuncPlan(Plan), AggType(Plan.AggType)
{
}

XxxErrCode XxxFuncExpression::
AggrPlan::MakeOpHandle(XxxAggrOpHandle*& Handle) const 
{
  Handle = new XxxAggrOpHandle(BasicType(), GenerateCatInfo());
  return XXX_OK;
}

XxxErrCode XxxFuncExpression::
AggrPlan::Init(XxxAggrOpHandle *Handle) const 
{
  Handle->Init();
  if (AggType < ADT) {
    return XxxFuncExpression::AggrOpSpecs[AggType]->Init(this, *Handle);
  }
  else {
    XxxADTClass* AttrClass = XxxADTArray[ownerType];
    XxxAggrMethodInfo* ThisMethInfo = AttrClass->GetAggrMethByIndex(ExprPlan->MethodIndex);
    XXX_CALL_RET(ExprPlan->AggrInit(ThisMethInfo, Handle),
		 XXX_BADFUNC, "Error initializing adt aggregate method");
    return XXX_OK;
  }
}
XxxErrCode XxxFuncExpression::
AggrPlan::Terminate(XxxAggrOpHandle *Handle,
		    XxxADTValue *FinalVal) const 
{
  if (AggType < ADT) {
    return XxxFuncExpression::AggrOpSpecs[AggType]->Terminate(this, *Handle, FinalVal);
  }
  else {
    XxxADTClass* AttrClass = XxxADTArray[ownerType];
    XxxAggrMethodInfo* ThisMethInfo = AttrClass->GetAggrMethByIndex(ExprPlan->MethodIndex);
    XXX_CALL_RET(ExprPlan->AggrTerm(ThisMethInfo, Handle, FinalVal),
		 XXX_BADFUNC, "Error terminating adt aggregate method");
    return XXX_OK;
  }
}

XxxErrCode XxxFuncExpression::
AggrPlan::Accumulate(const XxxRecord* Rec[],
		     const XxxRecordSchema* Schema[],
		     XxxAggrOpHandle *Handle,
		     const XxxValueEnv* Env) const
{
 ASSERT((Handle));
 ASSERT ((AggType <= XxxFuncExpression::ADT));

 if (AggType == XxxFuncExpression::COUNT) {
   // for count, there is no need to pass in a value
   ASSERT((AggrOpSpecs[AggType]));
   if (AggrOpSpecs[AggType]->Accum(this, *Handle) != XXX_OK) {
     XXX_NOTE_ERROR(XXX_BADEXEC, "Unable to accumulate aggr op");
     return XXX_NOT_OK;
    }
   Handle->Count++;
   return XXX_OK;
  }
 // if the owner exists, and should be evaluated, do so
 XXX_CALL_RET(evalOwner(Rec, Schema, ValueEnv->OwnerValPtr, Env),
	      XXX_BADEXEC, "Error evaluating owner");
 // compute the arguments
 XXX_CALL_RET(evalArgs(Rec, Schema, ValueEnv->ArgValsPtrs, Env),
	      XXX_BADEXEC, "Error evaluating func args");

 if (AggType < ADT) {
   XXX_CALL_RET(AggrOpSpecs[AggType]->Accum(this, *Handle),
		XXX_BADEXEC, "Unable to accumulate aggr op");
 }
 else {
    XxxADTClass* AttrClass = XxxADTArray[ownerType];
    XxxAggrMethodInfo* ThisMethInfo = AttrClass->GetAggrMethByIndex(ExprPlan->MethodIndex);
    XXX_CALL_RET(ExprPlan->AggrAccum(ThisMethInfo, ValueEnv, Handle),
		 XXX_BADFUNC, "Error accumulating adt aggregate method");
 }
 Handle->Count++;
 return XXX_OK;
}

XxxErrCode XxxFuncExpression::
AggrPlan::Drop(const XxxRecord* Rec[],
	       const XxxRecordSchema* Schema[],
	       XxxAggrOpHandle *Handle,
	       const XxxValueEnv* Env) const
{
 ASSERT((Handle));
 ASSERT ((AggType <= XxxFuncExpression::ADT));

 if (AggType == XxxFuncExpression::COUNT) {
   ASSERT((AggrOpSpecs[AggType]));
   // for count, there is no need to pass in a value
   if (AggrOpSpecs[AggType]->Drop(this, *Handle) != XXX_OK) {
     XXX_NOTE_ERROR(XXX_BADEXEC, "Unable to drop aggr op");
     return XXX_NOT_OK;
   }
   Handle->Count--;
   return XXX_OK;
  }
 // if the owner exists, and should be evaluated, do so
 XXX_CALL_RET(evalOwner(Rec, Schema, ValueEnv->OwnerValPtr, Env),
	      XXX_BADEXEC, "Error evaluating owner");
 // compute the arguments
 XXX_CALL_RET(evalArgs(Rec, Schema, ValueEnv->ArgValsPtrs, Env),
	      XXX_BADEXEC, "Error evaluating func args");
 
 if (AggType < ADT) {
   XXX_CALL_RET(AggrOpSpecs[AggType]->Drop(this, *Handle),
		XXX_BADEXEC, "Unable to drop aggr op");
 }
 else {
    XxxADTClass* AttrClass = XxxADTArray[ownerType];
    XxxAggrMethodInfo* ThisMethInfo = AttrClass->GetAggrMethByIndex(ExprPlan->MethodIndex);
    XXX_CALL_RET(ExprPlan->AggrDrop(ThisMethInfo, ValueEnv, Handle),
		 XXX_BADFUNC, "Error accumulating adt aggregate method");
 }
 Handle->Count--;
 return XXX_OK;
}

