#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <iostream.h>
#include "error.h"
#include "expr.h"
#include "record.h"
#include "exprplan.h"
#include "values.h"
#include "basictypes.h"
#include "funcexpr.h"
#include "sel-defaults.h"
#include "valbool.h"

//
// Methods of XxxBooleanExpression
//
XxxErrCode 
XxxBooleanExpression::ConvertToCNF(XxxBooleanExpression* BExpr)
{
 // Cop out! TODO:L2
 // Actually need to change the contents of BExpr
 return XXX_OK;
}

XxxErrCode
XxxBooleanExpression::GenerateBooleanFactors(XxxBooleanExpression* Cond,
					     int& NumFactors,
					     XxxBooleanExpression**& Factors,
					     XxxBool ApplyXforms) 
{
  if (!Cond) {
    XXX_NOTE_ERROR(XXX_BADEXPR, "Error generating boolean factors");
    return XXX_NOT_OK;
  }
#if 0
  // TODO: need to fix this
  if (ApplyXforms) {
    // try to reorganize the expression here!
    XxxBasicType OwnerType = Cond->ChildExpr(0)->BasicType();
    XxxBool TryRewrite = XXX_TRUE;
    for (int j = 1; j < Cond->Arity(); j++) {
      if (Cond->ChildExpr(j)->BasicType() != OwnerType) {
	TryRewrite = XXX_FALSE;
	break;
      }
    } // for
    // TODO: the owner type is screwed up here (always XXX_BOOLEAN)
    // For now, hardcode it
    OwnerType = XXX_GENFUNC;
    if (TryRewrite) {
      XxxValueExpression* NewCond;
      NewCond = Cond->Replicate();
      if (XxxADTArray[OwnerType]->FuncReOrganize(NewCond) != XXX_OK) {
	XXX_NOTE_ERROR(XXX_BADFUNC, "Error reorganizing boolean expression");
	return XXX_NOT_OK;
      }
      if (!NewCond || (NewCond->ValType() != BOOL_EXPR)) {
	XXX_NOTE_ERROR(XXX_BADFUNC, "Error reorganizing boolean expression");
	return XXX_NOT_OK;
      }
      // assign Cond to the NewCond (TODO: who reclaims the memory?)
      Cond = (XxxBooleanExpression*)NewCond;
    } // if (TryRewrite)
  } // if (ApplyXforms)
#endif

 if (Cond->BoolType == XXX_BOOL_BOOL_EXPR) {
   XxxBoolBooleanExpression* BoolExpr = 
     (XxxBoolBooleanExpression*)Cond;
   switch (BoolExpr->OpType) {
    case XXX_OR_OP:
    case XXX_NOT_OP:
    case XXX_FALSE_OP:
    case XXX_TRUE_OP:
     // we are too primitive to try pushing OR and NOT selections. 
     // give up here and treat it just like a REL_BOOL_EXPR
     NumFactors = 1;
     Factors = new XxxBooleanExpression*[1];
     Factors[0] = Cond;
     break;
    case XXX_AND_OP: {
      int NumChild0Factors, NumChild1Factors;
      XxxBooleanExpression** Child0Factors, ** Child1Factors;
      // here, lets try to break up the selections
      if ((GenerateBooleanFactors((XxxBooleanExpression*)
				  (BoolExpr->ChildExpr(0)),
				  NumChild0Factors, 
				  Child0Factors, XXX_FALSE) != XXX_OK) ||
	  (GenerateBooleanFactors((XxxBooleanExpression*)
				  (BoolExpr->ChildExpr(1)),
				  NumChild1Factors, 
				  Child1Factors, XXX_FALSE) != XXX_OK)) {
	XXX_NOTE_ERROR(XXX_BADEXPR, "unable to generate bool factors");
	return XXX_NOT_OK;
       }
      // merge the two sets of boolean factors
      NumFactors = NumChild0Factors + NumChild1Factors;
      Factors = new XxxBooleanExpression*[NumFactors];
      int i;
      for (i = 0; i <NumChild0Factors; i++) {
	Factors[i] = Child0Factors[i];
       }
      delete [] Child0Factors;
      for (i = NumChild0Factors; 
	   i < NumChild0Factors+NumChild1Factors; i++) {
	Factors[i] = Child1Factors[i-NumChild0Factors];
       }
      delete [] Child1Factors;
     }
    } // switch
  } // if
  else {
    NumFactors = 1;
    Factors = new XxxBooleanExpression*[1];
    Factors[0] = Cond;
   }
 return XXX_OK;
}

// default, for any predicate that doesnt have a nice selectivity formula
double XxxBooleanExpression::
Selectivity(SelectivityStatsInfo* StatsArray, int ArraySize) const
{
 return XXX_DEFAULT_EXPR_SEL;
}


//
// Methods of BoolBooleanExpression
//
XxxBoolBooleanExpression::XxxBoolBooleanExpression(XxxBoolOpType Type,
						   int NumChildren,
						   XxxExpression* Children[]) :
						   XxxBooleanExpression(NumChildren, Children)
{
 BoolType = XXX_BOOL_BOOL_EXPR;
 OpType = Type;
 // Allow an AND or OR with many sub-clauses

 if (arity == 0) {
   if ((Type != XXX_TRUE_OP) && (Type != XXX_FALSE_OP))
     XXX_NOTE_ERROR(XXX_BADEXPR, "Incompatible boolean expression");
  }
 else if (arity == 1) {
   if (Type != XXX_NOT_OP)
     XXX_NOTE_ERROR(XXX_BADEXPR, "Incompatible boolean expression");
  }
 else if (arity >= 2) {
   if ((Type != XXX_AND_OP) && (Type != XXX_OR_OP))
     XXX_NOTE_ERROR(XXX_BADEXPR, "Incompatible boolean expression");
  }
 else {
   XXX_NOTE_ERROR(XXX_BADEXPR, "Incompatible boolean expression");
  }
}

XxxBool XxxBoolBooleanExpression::Match(const XxxExpression* OtherExpr, 
					char *Env) const
{
  // Initital type check
  if (OtherExpr->ValType() != BOOL_EXPR) {
    return XXX_FALSE;
  }
  if (((XxxBooleanExpression *)OtherExpr)->BoolType != XXX_BOOL_BOOL_EXPR) {
    return XXX_FALSE;
  }
  // Check operator and children
  XxxBoolBooleanExpression *Expr = (XxxBoolBooleanExpression *)OtherExpr;
  if (Expr->OpType != OpType)
    return XXX_FALSE;
  if (arity != Expr->arity)
    return XXX_FALSE;
  for (int i = 0; i < arity; i++)
    if (!(ChildExpr(i)->Match(Expr->ChildExpr(i))))
      return XXX_FALSE;
  return XXX_TRUE;
}

XxxErrCode XxxBoolBooleanExpression::TypeCheck()
{
 XxxErrCode rc;

 if ((rc = XxxExpression::TypeCheck()) != XXX_OK) {
   return rc;
  }
 
 for (int i = 0; i < arity; i++) {
   if (ChildExpr(i)->ValType() != BOOL_EXPR) {
     XXX_NOTE_ERROR(XXX_BADTYPE, "Bad operand of BoolBoolean Op!");
     return XXX_NOT_OK;
    }
  }

 return XXX_OK;
}

void XxxBoolBooleanExpression::Print(ostream& OStr) const 
{
 if (arity == 0) {
   if (OpType == XXX_TRUE_OP)
     OStr << "TRUE";
   else
     OStr << "FALSE";
  }
 else if (arity == 1) {
   if (OpType == XXX_NOT_OP) {
     OStr << " NOT ";
     ChildExpr(0)->Print(OStr);
    }
   else {
     XXX_NOTE_ERROR(XXX_BADEXPR, "Bad Boolean Op!");
    }     
  }
 else {
   for (int i = 0; i < arity; i++) {
     ChildExpr(i)->Print(OStr);
     if (i < arity - 1) {
       switch (OpType) {
	case XXX_AND_OP:
	 OStr << " AND ";
	 break;
	case XXX_OR_OP:
	 OStr << " OR ";
	 break;
	default:
	 XXX_NOTE_ERROR(XXX_BADEXPR, "Bad Boolean Op!");
	}; // switch
      } // if not the last sub-clause
    } // for each sub-clause
  } // if the arity >= 2
}

XxxErrCode XxxBoolBooleanExpression::Optimize(XxxValueExprPlan*& ExecPlan) 
{
 /**********
 // TODO: try to reorganize the expression here!
 XxxBasicType OwnerType = ChildExpr(0)->BasicType();
 XxxBool TryRewrite = XXX_TRUE;
 for (int j = 1; j < Arity(); j++) {
   if (ChildExpr(j)->BasicType() != OwnerType) {
     TryRewrite = XXX_FALSE;
     break;
   }
 }
 // UGLY HACK: see below
 if (TryRewrite &&  !((int)(MetaInfo))) {
   XxxValueExpression *ThisCopy;
   ThisCopy = new XxxBoolBooleanExpression(*this);
   if (XxxADTArray[OwnerType]->FuncReOrganize(ThisCopy) != XXX_OK) {
     XXX_NOTE_ERROR(XXX_BADFUNC, "Error reorganizing boolean expression");
     if (ThisCopy)
       delete ThisCopy;
     return XXX_NOT_OK;
   }
   // how do i prevent it from recursively doing the same thing?
   // UGLY HACK!! : TODO
   // For now, use the MetaInfo, which should normally be NULL
   ThisCopy->MetaInfo = (XxxADTMetaInfo*)1;
   if (ThisCopy->Optimize(ExecPlan) != XXX_OK) {
     delete ThisCopy;
     XXX_NOTE_ERROR(XXX_BADOPT, "Error optimizing function expression");
     return XXX_NOT_OK;
   }
   ThisCopy->MetaInfo = 0;
   delete ThisCopy;
   return XXX_OK;
 }
 ***********/

 XxxValueExprPlan** TmpPlans = new XxxValueExprPlan*[Arity()];
 for (int i = 0; i < Arity(); i++) {
   XXX_CALL_RET(ChildExpr(i)->Optimize(TmpPlans[i]),
		XXX_BADOPT, "Error optimizing subexp");
  }
 ExecPlan = new XxxBoolBooleanPlan(this, 
				   (XxxBooleanExpressionPlan**)TmpPlans);
 delete [] TmpPlans;
 return XXX_OK;
}

double XxxBoolBooleanExpression::
Selectivity(SelectivityStatsInfo* StatsArray, int ArraySize) const 
{
 double Sel = 1.0;
 switch(OpType) {
  case XXX_TRUE_OP:
   Sel = 1.0;
   break;
  case XXX_FALSE_OP:
   Sel = 0.0;
   break;
  case XXX_NOT_OP:
   Sel = 1.0 - ((XxxBooleanExpression*)ChildExpr(0))->Selectivity(StatsArray, ArraySize);
   break;
  case XXX_AND_OP:
   Sel = ((XxxBooleanExpression*)ChildExpr(0))->Selectivity(StatsArray, ArraySize) *
     ((XxxBooleanExpression*)ChildExpr(1))->Selectivity(StatsArray, ArraySize);
   break;
  case XXX_OR_OP: {
    double Child0Sel = 
      ((XxxBooleanExpression*)ChildExpr(0))->Selectivity(StatsArray, ArraySize);
    double Child1Sel = 
      ((XxxBooleanExpression*)ChildExpr(1))->Selectivity(StatsArray, ArraySize);
    Sel = Child0Sel + Child1Sel - (Child0Sel*Child1Sel);
    break;
   }
  }
 return Sel;
}


//
// Methods of RelBooleanExpression
//

XxxRelBooleanExpression::XxxRelBooleanExpression(XxxRelOpType Type,
						 int NumChildren,
						 XxxExpression* Children[]) 
: XxxBooleanExpression(NumChildren, Children)
{
 BoolType = XXX_REL_BOOL_EXPR;
 OpType = Type;
 if ((OpType != XXX_OTHER_OP) && (NumChildren != 2)) {
   XXX_NOTE_ERROR(XXX_BADEXPR, "Incompatible rel boolean expression");
  }
}

XxxBool XxxRelBooleanExpression::Match(const XxxExpression* OtherExpr, 
				       char *Env) const
{
  // Initital type check
  if (OtherExpr->ValType() != BOOL_EXPR) {
    return XXX_FALSE;
  }
  if (((XxxBooleanExpression *)OtherExpr)->BoolType != XXX_REL_BOOL_EXPR) {
    return XXX_FALSE;
  }
  // Check operator and children
  XxxBoolBooleanExpression *Expr = (XxxBoolBooleanExpression *)OtherExpr;
  if (Expr->OpType != OpType)
    return XXX_FALSE;
  if (Arity() != Expr->Arity())
    return XXX_FALSE;
  for (int i = 0; i < arity; i++)
    if (!(ChildExpr(i)->Match(Expr->ChildExpr(i))))
      return XXX_FALSE;
  return XXX_TRUE;
}

XxxErrCode XxxRelBooleanExpression::TypeCheck()
{
 XxxErrCode rc;

 if ((rc = XxxExpression::TypeCheck()) != XXX_OK) {
   return rc;
  }

 // arity must be 2. these are old checks!
 if ((arity > 0) && ChildExpr(0)->ValType() == BOOL_EXPR) {
   XXX_NOTE_ERROR(XXX_BADTYPE, "Bad left operand of RelBoolean Op!");
   return XXX_NOT_OK;
  }
 if ((arity > 1) && ChildExpr(1)->ValType() == BOOL_EXPR) {
   XXX_NOTE_ERROR(XXX_BADTYPE, "Bad right operand of RelBoolean Op!");
   return XXX_NOT_OK;
  }
 XxxBasicType LType, RType;
 XxxADTMetaInfo *LMInfo, *RMInfo;
 LType = (arity>0)?
   ((XxxValueExpression*)ChildExpr(0))->BasicType(): XXX_UNKNOWN;
 LMInfo = (arity>0)?
   ((XxxValueExpression*)ChildExpr(0))->GenerateCatInfo(): NULL;
 RType = (arity>1)?
   ((XxxValueExpression*)ChildExpr(1))->BasicType(): XXX_UNKNOWN;
 RMInfo = (arity>0)?
   ((XxxValueExpression*)ChildExpr(1))->GenerateCatInfo(): NULL;

 // TODO: put in automatic type coercion here
 switch (OpType) {
  case XXX_EQ_OP:
  case XXX_NEQ_OP: 
   if (LType == RType)
     break;
   // otherwise, fall thru to other cases
  case XXX_GT_OP:
  case XXX_LT_OP:
  case XXX_LEQ_OP:
  case XXX_GEQ_OP:
   // Changed this, so that it does not allow int and doubles to be mixed
   // because there are places in the code where this causes problems,
   // specifically when getting a range for an indexed search
   // Fabian 2/18/97
   if ((LType == XXX_DOUBLE && RType == XXX_INT) ||
       (LType == XXX_INT && RType == XXX_DOUBLE))
     break;
   if ((RType == LType) &&
       XxxADTArray[RType]->OrderedType(RMInfo))
     break;
   XXX_NOTE_ERROR(XXX_BADTYPE, "Bad operands of RelBoolean Op!");
   return XXX_NOT_OK;

  case XXX_LIKE_OP:
  default: {
    XXX_NOTE_ERROR(XXX_BADTYPE, "Bad operator of RelBoolean Op!");
    return XXX_NOT_OK;
   }
  } // switch

 return XXX_OK;
}

void XxxRelBooleanExpression::Print(ostream& OStr) const 
{
 if (arity == 2) {
   ChildExpr(0)->Print(OStr);
   switch (OpType) {
    case XXX_GT_OP:
     OStr << " > ";
     break;
    case XXX_LT_OP:
     OStr << " < ";
     break;
    case XXX_EQ_OP:
     OStr << " = ";
     break;
    case XXX_GEQ_OP:
     OStr << " >= ";
     break;
    case XXX_LEQ_OP:
     OStr << " <= ";
     break;
    case XXX_NEQ_OP:
     OStr << " <> ";
     break;
    default:
     XXX_NOTE_ERROR(XXX_BADEXPR, "Bad Rel Op!");
    };
   ChildExpr(1)->Print(OStr);
  }
}

XxxErrCode XxxRelBooleanExpression::Optimize(XxxValueExprPlan*& ExecPlan)
{
 XxxValueExprPlan** TmpPlans = new XxxValueExprPlan*[Arity()];
 for (int i = 0; i < Arity(); i++) {
   XXX_CALL_RET(ChildExpr(i)->Optimize(TmpPlans[i]),
		XXX_BADOPT, "Error optimizing subexp");
  }
 ExecPlan = new XxxRelBooleanPlan(this, TmpPlans);
 delete [] TmpPlans;
 return XXX_OK;
}

// gets the stats for the column represented by Expr
XxxErrCode 
XxxRelBooleanExpression::getStats(SelectivityStatsInfo *StatsArray, 
				  int ArraySize, XxxValueExpression *VExpr, 
				  double &High, double &Low, double &Avg,
				  int &Count)
{
 High = Low = Avg = 0.0;
 Count = 0;
 
 if(VExpr->ValType() != XxxValueExpression::COLUMN_VAR) {
   XXX_NOTE_ERROR(XXX_BADOPT, "XxxRelBooleanExpression::getStats: "
		  "Received bad expression");
   return XXX_NOT_OK;
  }

 XxxUnknownValue *Expr = (XxxUnknownValue *)VExpr;
 
 int SrcChild = Expr->SrcChild;
 if(SrcChild >= ArraySize) {
   XXX_NOTE_ERROR(XXX_BADOPT, "XxxRelBooleanExpression::getStats: "
		  "Expression and recSchema don't match");
   return XXX_NOT_OK;
  }
 
 int SrcAttrIndex = Expr->SrcAttrIndex;
 if(SrcAttrIndex >= StatsArray[SrcChild].RecSchema->Arity()) {
   XXX_NOTE_ERROR(XXX_BADOPT, "XxxRelBooleanExpression::getStats: "
		  "Expression and recSchema don't match");
   return XXX_NOT_OK;
  }
 
 XxxADTClass::TypeStats *Stats = StatsArray[SrcChild].AttrStats[SrcAttrIndex];
 if(!Stats) {
   if (XXX_THISENV().DebugFlags.Expression) {
     cerr << "XxxRelBooleanExpression::getStats: Cannot retrieve stats."
	  << endl;
   }
   return XXX_NOT_OK;
  }
 
 XxxBasicType Type = Expr->BasicType();
 switch(Type) {
  case XXX_INT: {
    XxxCoreADTClass::IntStats *IntSt = (XxxCoreADTClass::IntStats *)Stats;
    High = IntSt->MaxVal;
    Low = IntSt->MinVal;
    Avg = IntSt->Avg;
    Count = IntSt->Count;
    break;
   }
  case XXX_DOUBLE: {
    XxxCoreADTClass::DblStats *DblSt 
     = (XxxCoreADTClass::DblStats *)Stats;
    High = DblSt->MaxVal;
    Low = DblSt->MinVal;
    Avg = DblSt->Avg;
    Count = DblSt->Count;
    break;
   }
  default:
  if (XXX_THISENV().DebugFlags.Expression) {
    cerr << "XxxRelBooleanExpression::getStats: Type not supported." << endl;
  }
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

double XxxRelBooleanExpression::
eqSelectivity(SelectivityStatsInfo* StatsArray, int ArraySize,
	      XxxValueExpression* LExpr, 
	      XxxValueExpression* RExpr)
{
 double Sel = XXX_COL_EQ_CONST_SEL;

 // either expr passed in may be NULL 
 // (because the original expressions were deemed too complex)
 if (!LExpr || !RExpr) return Sel;

 XxxBasicType LType = LExpr->BasicType();
 XxxBasicType RType = RExpr->BasicType();

 // We only handle ints and doubles right now
 if (((LType != XXX_INT) && (LType != XXX_DOUBLE)) ||
     ((RType != XXX_INT) && (RType != XXX_DOUBLE)))
   return Sel;

 XxxBool LeftIsConst = LExpr->IsConstant();
 if ((LExpr->ValType() == COLUMN_VAR) &&
     (((XxxUnknownValue*)LExpr)->CorrelationHeight > 0))
   LeftIsConst = XXX_TRUE;
 XxxBool RightIsConst = RExpr->IsConstant();
 if ((RExpr->ValType() == COLUMN_VAR) &&
     (((XxxUnknownValue*)RExpr)->CorrelationHeight > 0))
   RightIsConst = XXX_TRUE;
 
 if (LeftIsConst && RightIsConst) {
   // condition of type [constant = constant]
   double LeftVal, RightVal;
   // What if there is a big subquery or function that produces a double?
   if ((LExpr->ValType() != XxxValueExpression::CONST_EXPR) ||
       (RExpr->ValType() != XxxValueExpression::CONST_EXPR)) {
     // So we are unwilling to compute this, but what is the likelihood
     // of two consts being the same? No different from that of a column
     // value being equal to a constant
     return Sel;
    }
   if (LType == XXX_INT) 
     LeftVal = ((XxxConstIntValue*)LExpr)->GetDblValue();
   else 
     LeftVal = ((XxxConstDoubleValue*)LExpr)->GetDblValue();
   if (RType == XXX_INT) 
     RightVal = ((XxxConstIntValue*)RExpr)->GetDblValue();
   else 
     RightVal = ((XxxConstDoubleValue*)RExpr)->GetDblValue();

   Sel = (LeftVal == RightVal ? 1.0 : 0.0);
  } 
 else if (LeftIsConst || RightIsConst) {
   // condition of [constant = column] or [column = constant]
   // Sel = 1/(distinct column values)
   XxxValueExpression *NonConstExpr = (LeftIsConst ? RExpr: LExpr);
   double Hi, Lo, Avg;
   int Card;
   if (getStats(StatsArray, ArraySize, NonConstExpr, Hi, Lo, Avg, 
		Card) != XXX_OK)
     return Sel;
   Sel = 1.0/(double)Card;
  } 
 else {
   // neither is constant, condition of type [column = column]
   // Sel = 1/max(distinct values of equality column)
   double LeftHi, LeftLo, LeftAvg, RightHi, RightLo, RightAvg;
   int LeftCard, RightCard;
   if ((getStats(StatsArray, ArraySize, LExpr, LeftHi, LeftLo, LeftAvg, 
		 LeftCard) != XXX_OK) ||
       (getStats(StatsArray, ArraySize, RExpr, RightHi, RightLo, RightAvg,
		 RightCard) != XXX_OK))
     return Sel;
   int MaxCard = (LeftCard > RightCard? LeftCard : RightCard);
   Sel = 1.0/(double)MaxCard;
  }
 return Sel;
}

// LExpr > RExpr
double XxxRelBooleanExpression::
gtSelectivity(SelectivityStatsInfo* StatsArray, int ArraySize,
	      XxxValueExpression* LExpr, 
	      XxxValueExpression* RExpr)
{
 double Sel = XXX_COL_INEQ_CONST_SEL;

 // either expr passed in may be NULL 
 // (because the original expressions were deemed too complex)
 if (!LExpr || !RExpr) return Sel;

 XxxBasicType LType = LExpr->BasicType();
 XxxBasicType RType = RExpr->BasicType();

 // we only handle ints and doubles right now
 if(((LType != XXX_INT) && (LType != XXX_DOUBLE)) ||
    ((RType != XXX_INT) && (RType != XXX_DOUBLE)))
   return Sel;

 XxxBool LeftIsConst = LExpr->IsConstant();
 XxxBool RightIsConst = RExpr->IsConstant();
 double LeftVal, RightVal;
 if (LeftIsConst && (LExpr->ValType() == XxxValueExpression::CONST_EXPR)) {
   if (LType == XXX_INT) 
     LeftVal = ((XxxConstIntValue*)LExpr)->GetDblValue();
   else 
     LeftVal = ((XxxConstDoubleValue*)LExpr)->GetDblValue();
  }
 if (RightIsConst && (RExpr->ValType() == XxxValueExpression::CONST_EXPR)) {
   if (RType == XXX_INT) 
     RightVal = ((XxxConstIntValue*)RExpr)->GetDblValue();
   else 
     RightVal = ((XxxConstDoubleValue*)RExpr)->GetDblValue();
  }
 
 if (LeftIsConst && RightIsConst) {
   // condition of type [constant > constant]
   // What if there is a big subquery or function that produces a double?
   if ((LExpr->ValType() != XxxValueExpression::CONST_EXPR) ||
       (RExpr->ValType() != XxxValueExpression::CONST_EXPR)) {
     // we are unwilling to compute this
     return Sel;
    }
   Sel = (LeftVal > RightVal? 1.0:0.0);
  } 
 else if (LeftIsConst || RightIsConst) {
   // condition of [constant = column] or [column = constant]
   // Sel = (high value - value)/(high value - low value)
   XxxValueExpression *NonConstExpr = (LeftIsConst ? RExpr: LExpr);
   double Hi, Lo, Avg;
   double ConstVal = (LeftIsConst? LeftVal: RightVal);
   int Card;
   if (getStats(StatsArray, ArraySize, NonConstExpr, Hi, Lo, Avg, 
		Card) != XXX_OK)
     return Sel;

   // if it is of the form 5 > X, i.e. X < 5
   if (LeftIsConst)
     Sel = (ConstVal - Lo)/(Hi - Lo);
   else 
     // if it is of the form X > 5
     Sel = (Hi - ConstVal)/(Hi - Lo);
  } 
 else {
   // neither is constant, condition of type [column = column]
   // Sel = (high value 1 - avg value 2)/(high value 2 - low value 2)
   double LeftHi, LeftLo, LeftAvg, RightHi, RightLo, RightAvg;
   int LeftCard, RightCard;
   if ((getStats(StatsArray, ArraySize, LExpr, LeftHi, LeftLo, LeftAvg, 
		 LeftCard) != XXX_OK) ||
       (getStats(StatsArray, ArraySize, RExpr, RightHi, RightLo, RightAvg,
		 RightCard) != XXX_OK))
     return Sel;

   Sel = (LeftHi - RightAvg)/(LeftHi - LeftLo);
  }      

 // check for roundoff errors
 if (Sel > 1.0) 
   Sel = 1.0;
 else if (Sel < 0.0)
   Sel = 0.0;
 return Sel;
}

/*******
static XxxValueExpression* getUnknownExpr(XxxValueExpression* ChildExpr)
{
 XxxValueExprList *Unknowns = ChildExpr->ValExprsReferenced();
 if (Unknowns) {
   Unknowns = Unknowns->TypeFilter(XxxValueExpression::COLUMN_VAR);
 XxxValueExpression* UnknownExpr;

 if (Unknowns) {
   UnknownExpr = (XxxValueExpression *)(Unknowns->Expr->Replicate());
   Unknowns->Destroy();
  } 
 else {
   UnknownExpr = (XxxValueExpression *)ChildExpr;
  }
 return UnknownExpr;
}
***********/

double XxxRelBooleanExpression::
Selectivity(SelectivityStatsInfo* StatsArray, int ArraySize) const 
{
 double Sel = XXX_DEFAULT_EXPR_SEL;
 
 if (ChildExpr(0)->ValType() == BOOL_EXPR) {
   XXX_NOTE_ERROR(XXX_BADTYPE, "XxxRelBooleanExpression::Selectivity: Bad "
		  "left operand of RelBoolean Op!");
   return Sel;
  }
 if (ChildExpr(1)->ValType() == BOOL_EXPR) {
   XXX_NOTE_ERROR(XXX_BADTYPE, "XxxRelBooleanExpression::Selectivity: Bad "
		  "right operand of RelBoolean Op!");
   return Sel;
  }

 XxxValueExpression *LExpr = (XxxValueExpression *)ChildExpr(0);
 XxxValueExpression *RExpr = (XxxValueExpression *)ChildExpr(1);
 // We do not handle expressions of the type f(E.id, C.foo) = 1, or
 // (E.id +C.id) = 1 correctly : TODO:L2
 if ((LExpr->ValType() != XxxValueExpression::COLUMN_VAR) &&
     ((LExpr->ValType() != XxxValueExpression::CONST_EXPR)))
   LExpr = NULL;
 if ((RExpr->ValType() != XxxValueExpression::COLUMN_VAR) &&
     ((RExpr->ValType() != XxxValueExpression::CONST_EXPR)))
   RExpr = NULL;

 /**
 LExpr = getUnknownExpr((XxxValueExpression*)(ChildExpr(0)));
 RExpr = getUnknownExpr((XxxValueExpression*)(ChildExpr(1)));
 **/
 
 XxxBool Invert = XXX_FALSE;
 switch(OpType) {
  case XXX_NEQ_OP:  // same as 1 - (equal) : reverse this at end of func
   Invert = XXX_TRUE;
  case XXX_EQ_OP:
   Sel = eqSelectivity(StatsArray, ArraySize, LExpr, RExpr);
   break;
  case XXX_LEQ_OP: // same as 1 - (gt) : reverse this at end of func
   Invert = XXX_TRUE;
  case XXX_GT_OP:
   Sel = gtSelectivity(StatsArray, ArraySize, LExpr, RExpr);
   break;
  case XXX_GEQ_OP: // same as 1 - (lt) : reverse this at end of func
   Invert = XXX_TRUE;
  case XXX_LT_OP:  // similar to (greater than) : reverse this at end of func
   Sel = gtSelectivity(StatsArray, ArraySize, RExpr, LExpr);
   break;
  case XXX_LIKE_OP:
  default: {
    XXX_NOTE_ERROR(XXX_BADTYPE, "Bad operator of RelBoolean Op!");
    return XXX_NOT_OK;
   }
  } // switch

 if (Invert)
   Sel = 1.0 - Sel;
 
 return Sel;
}

//
// Methods of ValBooleanExpression
//
XxxValBooleanExpression::
XxxValBooleanExpression(XxxExpression* Child) 
  : XxxBooleanExpression(1, &Child)
{
	BoolType = XXX_VAL_BOOL_EXPR;
}
XxxValBooleanExpression::
XxxValBooleanExpression(const XxxValBooleanExpression& src) 
  : XxxBooleanExpression(src)
{
	BoolType = XXX_VAL_BOOL_EXPR;
}

XxxErrCode XxxValBooleanExpression::TypeCheck()
{
 XxxErrCode rc;
 if ((rc = XxxExpression::TypeCheck()) != XXX_OK) {
   return rc;
  }
 if (ChildExpr(0)->BasicType() != XXX_BOOLEAN) {
   XXX_NOTE_ERROR(XXX_BADTYPE, "Bad operand of ValBoolean expression!");
   return XXX_NOT_OK;
 }
 return XXX_OK;
}

void XxxValBooleanExpression::Print(ostream& OStr) const 
{
  ChildExpr(0)->Print(OStr);
}

XxxErrCode XxxValBooleanExpression::Optimize(XxxValueExprPlan*& ExecPlan) 
{
 XxxValueExprPlan* TmpPlan;
 XXX_CALL_RET(ChildExpr(0)->Optimize(TmpPlan),
	      XXX_BADOPT, "Error optimizing subexp");
 ExecPlan = new XxxValBooleanPlan(this, TmpPlan);
 return XXX_OK;
}

XxxBool XxxValBooleanExpression::Match(const XxxExpression* OtherExpr, 
				       char *Env) const
{
  // Initital type check
  if ((OtherExpr->ValType() != ValType()) ||
      (OtherExpr->BasicType() != BasicType())) {
    return XXX_FALSE;
  }
  /*
  if (((XxxBooleanExpression *)OtherExpr)->BoolType != XXX_VAL_BOOL_EXPR) {
    return XXX_FALSE;
  }
  */
  // Check operator and children
  XxxValBooleanExpression *Expr = (XxxValBooleanExpression *)OtherExpr;
  if (!(ChildExpr(0)->Match(Expr->ChildExpr(0))))
    return XXX_FALSE;
  return XXX_TRUE;
}

double XxxValBooleanExpression::
Selectivity(SelectivityStatsInfo* StatsArray, int ArraySize) const 
{
  // TODO: do a better job of selectivity
  return 0.5;
}


