#ifndef XXX_CASEEXPR_H
#define XXX_CASEEXPR_H
#include <iostream.h>
#include "expr.h"
#include "values.h"
#include "exprplan.h"

/**
 * This class currently models only one type of case expression:
 *   CASE WHEN cond THEN expr ... ELSE expr END
 * The children argment of the constructor should contain a series of
 * condition and corresponding expression (when and then), and
 * default expression (else). 
 *
 * TODO: implement the other type of case expression:
 *     CASE expr WHEN expr THEN expr ... ELSE expr END
 */
class XxxCaseExpression : public XxxValueExpression {
  friend class CasePlan;
 protected:
 public:
  //// Plan for case expressions
  class CasePlan : public XxxValueExprPlan { public:
    XxxADTValue *ConditionVal;
    XxxBasicType ChildType;

    CasePlan(XxxCaseExpression* CaseExpr,
	     XxxValueExprPlan* ChildrenPlan[]);
    CasePlan(const CasePlan &Plan);
    virtual ~CasePlan() { delete ConditionVal; };

    virtual XxxGenericPlan * Replicate() const {
      return (XxxGenericPlan *)(new CasePlan(*this));
    }    
    virtual XxxErrCode Evaluate(const XxxRecord* Rec[],
				const XxxRecordSchema* Schema[],
				XxxADTValue *Result,
				const XxxValueEnv* Env = NULL) const;
   };

  XxxCaseExpression(int NChildren,
		    XxxExpression* Children[]);
  XxxCaseExpression(const XxxCaseExpression& src);

  // Provides a convenient way to copy arbitrary expressions
  virtual XxxExpression *Replicate() const 
    { return new XxxCaseExpression(*this); }

  virtual ValExprType ValType() const {return CASE_EXPR;}
  virtual void Print(ostream&) const;
  virtual XxxErrCode Optimize(XxxExprPlan*& ExecPlan);
  virtual XxxErrCode TypeCheck();

  // Check to see if matches other expression
  virtual XxxBool Match(const XxxExpression* OtherExpr, 
			char *Env = NULL) const;
 };

#endif
