#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include "castexpr.h"
#include "generic.h"

//
// Methods of XxxCastValueExpression
//
XxxCastValueExpression::
XxxCastValueExpression(XxxBasicType Type,
		       XxxADTMetaInfo* MInfo,
		       XxxExpression *Child) :
XxxValueExpression(1, &Child)
{
 CastType = Type;
 MetaInfo = MInfo;
}

XxxCastValueExpression::
XxxCastValueExpression(const XxxCastValueExpression& src):
  XxxValueExpression(src), CastType(src.CastType) { }

void XxxCastValueExpression::Print(ostream& OStr) const 
{
 OStr << "(";
 OStr << '[';
 OStr << XxxADTArray[CastType]->TypeName ;
 if (MetaInfo) {
   OStr << "(";
   MetaInfo->Print(OStr);
   OStr << ")";
  }
 OStr << "]";
 OStr << "(";
 ChildExpr(0)->Print(OStr);
 OStr << "))";
}

XxxErrCode XxxCastValueExpression::TypeCheck()
{
  XxxErrCode rc;
  // first type check the child
  if ((rc = XxxExpression::TypeCheck()) != XXX_OK) {
    return rc;
  }
  if (ChildExpr(0)->ValType() == BOOL_EXPR) {
    XXX_NOTE_ERROR(XXX_BADTYPE, "Bad operand for Cast Op!");
    return XXX_NOT_OK;
  }
  XxxBasicType InType;
  InType = ((XxxValueExpression*)ChildExpr(0))->BasicType();
  
  XxxADTMetaInfo *ChildCatInfo = 
    ((XxxValueExpression*)ChildExpr(0))->GenerateCatInfo();
  // Check to see if the ADT of the input can cast to the current type
  if ((InType != XXX_UNKNOWN) && XxxADTArray[InType] && 
      (XxxADTArray[InType]->CanCastTo(CastType, ChildCatInfo,
				      MetaInfo) == XXX_TRUE)) {
    BaseType = CastType;
    return XXX_OK;
  } else if ((CastType != XXX_UNKNOWN) && XxxADTArray[CastType] && 
      (XxxADTArray[CastType]->CanCastFrom(InType, MetaInfo, 
					  ChildCatInfo) == XXX_TRUE)) {
    BaseType = CastType;
    return XXX_OK;
  } else {
    BaseType = XXX_UNKNOWN;
    XXX_NOTE_ERROR(XXX_BADTYPE, "Bad operand for cast");
    return XXX_NOT_OK;
  }
}

XxxErrCode XxxCastValueExpression::Optimize(XxxValueExprPlan*& ExprPlan)
{
 XxxValueExprPlan* TmpPlan;
 XXX_CALL_RET(((XxxValueExpression *)(ChildExpr(0)))->Optimize(TmpPlan),
	      XXX_BADEXEC, "Unable to optimize cast expression");

 CastPlan *CEPlan = new CastPlan(this, TmpPlan);

 // This will be fine for constants like 1 or 2, but what about
 // host variables? either those must be made available here
 // (this is out of the question), or something else must be done.
 if (ChildExpr(0)->IsConstant()) {
   XXX_CALL_RET(CEPlan->PreCompute(), XXX_BADEXEC, 
		"Unable to precompute constant cast expression");
  }
 ExprPlan = CEPlan;
 return XXX_OK;
} 

XxxCastValueExpression::
CastPlan::CastPlan(XxxCastValueExpression* CExpr,
		   XxxValueExprPlan* ChildPlan)
: XxxValueExprPlan(CExpr, &ChildPlan) 
{ 
 StoredConstVal = NULL;

 InType = ChildPlan->BasicType();

 if ((InType == XXX_UNKNOWN) || !XxxADTArray[InType]) {
   XXX_NOTE_ERROR(XXX_BADTYPE, "Strange type for child of cast");
   return;
  }
 // who reclaims this ??
 ChildCatInfo = ChildPlan->GenerateCatInfo();

 ChildAnswer = (XxxADTValue*)(new char[XxxADTArray[InType]->MaxObjectSize(ChildCatInfo)]);
 ChildAnswer->Init();

 getPlanUsage().Add(ChildPlan->GetPlanUsage());
}

XxxCastValueExpression::
CastPlan::CastPlan(const CastPlan &Plan) : 
  XxxValueExprPlan(Plan), StoredConstVal(NULL), InType(Plan.InType)
{
  if (Plan.ChildCatInfo && 
      (Plan.ChildCatInfo->Replicate(ChildCatInfo) != XXX_OK)) {
    ChildCatInfo = NULL;
  }
  ChildAnswer = (XxxADTValue*)(new char[XxxADTArray[InType]->MaxObjectSize(ChildCatInfo)]);
  ChildAnswer->Init();
  memcpy(ChildAnswer, Plan.ChildAnswer, 
	 XxxADTArray[InType]->MaxObjectSize(ChildCatInfo));
}

/***Compiler doesnt like this!!
virtual XxxCastValueExpression::CastPlan::~CastPlan()
{
 delete ChildAnswer;
 if (StoredConstVal) delete [] StoredConstVal;
}
****/


XxxErrCode XxxCastValueExpression::
CastPlan::PreCompute()
{
  // check if it is a constant. if so, copy answer into StoredConstVal
  if (IsConstant()) {
    if (SubPlans[0]->Evaluate((XxxRecord*)NULL, NULL,
			       ChildAnswer, NULL) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADEXEC, "Error precomputing child of cast");
      return XXX_NOT_OK;
    }
    XxxADTMetaInfo* AnswerCatInfo = GenerateCatInfo();
    StoredConstVal = (XxxADTValue*)
      (new char[XxxADTArray[BasicType()]->MaxObjectSize(AnswerCatInfo)]);
    StoredConstVal->Init();
    if ((InType != XXX_UNKNOWN) && XxxADTArray[InType] && 
	(XxxADTArray[InType]->CanCastTo(BasicType(),
					ChildCatInfo,
					AnswerCatInfo) == XXX_TRUE)) {
      if (XxxADTArray[InType]->CastTo(BasicType(),
				      ChildCatInfo,
				      AnswerCatInfo,
				      ChildAnswer,
				      StoredConstVal) != XXX_OK) {
	XXX_NOTE_ERROR(XXX_BADEXEC, "Error evaluating type cast");
	return XXX_NOT_OK;
      }
    } else if ((BasicType() != XXX_UNKNOWN) && XxxADTArray[BasicType()] && 
	       (XxxADTArray[BasicType()]->
		CanCastFrom(InType, ChildCatInfo,
			    AnswerCatInfo) == XXX_TRUE)) {
      if (XxxADTArray[InType]->CastFrom(BasicType(),
					ChildCatInfo,
					AnswerCatInfo,
					ChildAnswer, 
					StoredConstVal) != XXX_OK) {
	XXX_NOTE_ERROR(XXX_BADEXEC, "Error evaluating type cast");
	return XXX_NOT_OK;
      }
    }
  } else {
    // precompute should not be called on a non-constant
    XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot precompute non-constant");
    return XXX_NOT_OK;
  }
  return XXX_OK;
}

XxxErrCode XxxCastValueExpression::
CastPlan::Evaluate(const XxxRecord* InputRecords[],
		   const XxxRecordSchema* InputSchemas[],
		   XxxADTValue *Result,
		   const XxxValueEnv* Env) const 
{
 XxxADTMetaInfo* AnswerCatInfo = GenerateCatInfo();
 if (StoredConstVal) {
   // the value has already been computed!!
   // Make sure to add MetaInfo!!
   XxxADTArray[BasicType()]->TypeCopy(StoredConstVal, Result,
				      AnswerCatInfo);
   return XXX_OK;
  }
 
 if (SubPlans[0]->Evaluate(InputRecords, InputSchemas,
			   ChildAnswer, Env) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_BADEXPR, "Error evaluating child of cast");
   return XXX_NOT_OK;
  }

 if ((InType != XXX_UNKNOWN) && XxxADTArray[InType] && 
     (XxxADTArray[InType]->
      CanCastTo(BasicType(), ChildCatInfo, AnswerCatInfo) == XXX_TRUE)) {
   if (XxxADTArray[InType]->CastTo(BasicType(),
				   ChildCatInfo,
				   AnswerCatInfo,
				   ChildAnswer,
				   Result) != XXX_OK) {
     XXX_NOTE_ERROR(XXX_BADEXEC, "Error evaluating type cast");
     return XXX_NOT_OK;
   }
 } else if ((BasicType() != XXX_UNKNOWN) && XxxADTArray[BasicType()] && 
	    (XxxADTArray[BasicType()]->
	     CanCastFrom(InType, ChildCatInfo, AnswerCatInfo) == XXX_TRUE)) {
   if (XxxADTArray[BasicType()]->CastFrom(InType,
					  ChildCatInfo,
					  AnswerCatInfo,
					  ChildAnswer,
					  Result) != XXX_OK) {
     XXX_NOTE_ERROR(XXX_BADEXEC, "Error evaluating type cast");
     return XXX_NOT_OK;
   }
 }

 /*******
   Compiler doesnt like this, because this is supposed to be a const function!!
   Despite problems (see below), move this to plan constructor
 
 // check if it is a constant. if so, copy answer into StoredConstVal
 if (ConstantExpr) {
   // yikes: this is supposed to be a constant member function,
   // but i have to cache this result!!
   // Down the road, move this to the plan constructor.
   // This will be fine for constants like 1 or 2, but what about
   // host variables? either those must be made available to the
   // plan constructor (this is out of the question), or
   // something else will have to be done.
   StoredConstVal = (XxxADTValue*)
     new char[XxxADTArray[CastExpr->BasicType()]->MaxObjectSize()];
   XxxADTArray[CastExpr->BasicType()]->TypeCopy(Result, StoredConstVal);
  }
 ****/

 return XXX_OK;
}

XxxBool
XxxCastValueExpression::Match(const XxxExpression *OtherExpr, 
			      char* Env) const
{
 if (!XxxValueExpression::Match(OtherExpr, Env))
   return XXX_FALSE;

  XxxCastValueExpression *CastExpr = (XxxCastValueExpression *)OtherExpr;
  
  if (CastType != CastExpr->CastType)
    return XXX_FALSE;
  
  return XXX_TRUE; 
}
