#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <iostream.h>
#include "error.h"
#include "expr.h"
#include "record.h"
#include "exprplan.h"
#include "values.h"
#include "basictypes.h"
#include "funcexpr.h"
#include "strrules.h"

//
// Methods of XxxValueExpression::
//
XxxValueExpression::XxxValueExpression(int NumChildren, XxxValueExpression* Children[])
{
 ExprName = NULL;
 arity = NumChildren;
 if (arity) {
   childExpr = new XxxValueExpression*[arity];
  }
 else
   childExpr = NULL;

 for (int i = 0; i < arity; i++)
   childExpr[i] = Children[i];

 // stuff for value expressions
 BaseType = XXX_UNKNOWN; 
 MetaInfo = NULL;
}

XxxValueExpression::XxxValueExpression(const XxxValueExpression& src)
{
 arity = src.arity;
 ExprName = NULL;
 childExpr = NULL;
 if (src.ExprName) {
   ExprName = new char[XXX_ATTR_NAME_MAX];
   strncpy(ExprName, src.ExprName, XXX_ATTR_NAME_MAX);
   ExprName[XXX_ATTR_NAME_MAX-1] = 0;
  }
 if (arity)
   childExpr = new XxxValueExpression*[arity];

 // create the childExpr array
  for (int i = 0; i < arity; i++) {
    childExpr[i] = src.childExpr[i]->Replicate();
  }
 BaseType = src.BaseType;
 MetaInfo = NULL;
 if (src.MetaInfo)
   // TODO:L2 check the return code
   src.MetaInfo->Replicate(MetaInfo);
}

XxxValueExpression::~XxxValueExpression()
{
  for (int i = 0; i < arity; i++)
   delete childExpr[i];

 if (childExpr) delete [] childExpr;
}

XxxErrCode 
XxxValueExpression::ResolveVars(const XxxRecordSchema *Schema)
{
 // resolve all variables in children and type check
 // them
 for (int i = 0; i < arity; i++)
   if (ChildExpr(i)->ResolveVars(Schema) != XXX_OK)
     return XXX_NOT_OK;
 return XXX_OK;
}

// try to resolve the vars in this expression using the schema of Node
// or its children
XxxErrCode XxxValueExpression::ResolveVars(XxxQueryNode* Node, 
					   char Source)
{
 // resolve all variables in children and type check them
 for (int i = 0; i < arity; i++)
   if (ChildExpr(i)->ResolveVars(Node, Source) != XXX_OK) {
     XXX_NOTE_ERROR(XXX_BADEXPR, "Error resolving variable");
     return XXX_NOT_OK;
    }
 
 return XXX_OK;
}

XxxBool XxxValueExpression::Match(const XxxValueExpression* OtherExpr, 
			     char *Env) const
{
  if (Arity() != OtherExpr->Arity())
    return XXX_FALSE;
  if (ValType() != ((XxxValueExpression *)OtherExpr)->ValType())
    return XXX_FALSE;
  if (BasicType() != ((XxxValueExpression *)OtherExpr)->BasicType())
    return XXX_FALSE;
  for (int i = 0; i < Arity(); i++)
    if (!(ChildExpr(i)->Match(OtherExpr->ChildExpr(i), Env)))
      return XXX_FALSE;
  return XXX_TRUE;
}

XxxBool XxxValueExpression::Match(const XxxValueExpression* RuleExpr, 
				  XxxBindingEnv *BindEnv) const
{
  if (((XxxValueExpression *)RuleExpr)->ValType() == COLUMN_VAR) {
    // Note: If the corresponding value for "RuleExpr" has
	// already been entered. Then that value will be matched
	// against "this" to make sure they are the same. otherwise,
	// Match will return fail
    if (BindEnv->SetElement((XxxUnknownValue *)RuleExpr, 
			    (XxxValueExpression*)this) != XXX_OK)
      return XXX_FALSE;
    return XXX_TRUE;
  }

  if (Arity() != RuleExpr->Arity())
    return XXX_FALSE;
  for (int i = 0; i < Arity(); i++)
    if (!(ChildExpr(i)->Match(RuleExpr->ChildExpr(i), BindEnv)))
      return XXX_FALSE;
  if (ValType() != ((XxxValueExpression *)RuleExpr)->ValType())
    return XXX_FALSE;
  return XXX_TRUE;
}

// try to resolve the host vars in this expression using the arg environment
XxxErrCode XxxValueExpression::ResolveVars(const XxxArgEnv* Env)
{
 // try to resolve variables in children
 for (int i = 0; i < arity; i++)
   if (ChildExpr(i)->ResolveVars(Env) != XXX_OK) 
     return XXX_NOT_OK;
 return XXX_OK;
}

XxxErrCode XxxValueExpression::TypeCheck()
{
 XxxErrCode rc = XXX_OK;
 for (int i = 0; i < arity; i++) {
   if ((rc = childExpr[i]->TypeCheck()) != XXX_OK) return rc;
  }
 return rc;
}

// PURPOSE: Set child # 'Index' to 'NewChild', Kill the OldChild
// NOTES: added to facilitate query rewrite
XxxErrCode XxxValueExpression::SetChild(int Index, XxxValueExpression *NewExp) {
  if (Index>=arity) {
    char buf[100];
    sprintf(buf,"XxxValueExpression::SetChild: Tried to reset child expression %d, "
	    "when there are only %d child expressions!",Index,arity);
    XXX_NOTE_ERROR(XXX_BADEXPR, buf);
    return XXX_NOT_OK;
   }    
  if (childExpr[Index] != NULL)
    delete childExpr[Index];
  childExpr[Index] = NewExp;
  return XXX_OK;
 }

XxxBool XxxValueExpression::IsConstant() const
{
 for (int i = 0; i < arity; i++) {
   if (!childExpr[i]->IsConstant()) return XXX_FALSE;
  }
 return XXX_TRUE;
}

XxxBool XxxValueExpression::ReferencesAggregate() const
{
  for (int i = 0; i < arity; i++) {
    if (childExpr[i]->ReferencesAggregate()) return XXX_TRUE;
  }
  return XXX_FALSE;
}

// TODO: find out where this is used!
XxxErrCode XxxValueExpression::Replace (const XxxValueExpression *Target,
				   const XxxValueExpression *Replacement)
{
  for (int i = 0; i < arity; i++) {
    if (childExpr[i]->Match(Target)) {
      delete childExpr[i];
      childExpr[i] = Replacement->Replicate();
    } 
  }
  return XXX_OK;
}

XxxErrCode XxxValueExpression::
Replace (const XxxUnknownValue *Target,
	 const XxxValueExpression *Replacement)
{
  for (int i = 0; i < arity; i++) {
    if (childExpr[i]->ValType() == XxxValueExpression::COLUMN_VAR && 
	((XxxUnknownValue *)childExpr[i])->Equals(Target)) {
      XxxUnknownValue *expr = (XxxUnknownValue *)childExpr[i];
      delete expr;
      childExpr[i] = Replacement->Replicate();
    } 
    else 
      childExpr[i]->Replace(Target, Replacement);
  }
  return XXX_OK;
}

XxxErrCode XxxValueExpression::UpdateUnknowns(int ChildNumber,
					      int OldAttrIndex,
					      int NewAttrIndex,
					      XxxBool &Appears)
{
  for (int i = 0; i < arity; i++) 
    if(childExpr[i]->UpdateUnknowns(ChildNumber,
				    OldAttrIndex,
				    NewAttrIndex,
				    Appears) != XXX_OK)
      return XXX_NOT_OK;
  return XXX_OK;
}

 
XxxErrCode 
XxxValueExpression::GetDependencies(XxxBitMap& Dependency,
				    int CorrelationHt) const
{
 if (arity > 0) {
   for (int i = 0; i < arity; i++) {
     XXX_CALL_RET(childExpr[i]->GetDependencies(Dependency, CorrelationHt),
		  XXX_BADEXPR, "Error getting dependencies of child");
   }
   return XXX_OK;
 }
 else if (IsConstant()) {
   return XXX_OK;
  }
 else {
   XXX_NOTE_ERROR(XXX_BADEXPR, "GetDependencies should be defined for subclass");
   return XXX_NOT_OK;
  }
}

XxxValueExprList * XxxValueExpression::DirectSubexpressions()
{
  XxxValueExprList *Result = NULL;
  XxxValueExprList **TmpResult = &Result;
  for(int i=0; i < Arity(); i++) {
    *TmpResult = new XxxValueExprList(ChildExpr(i)->Replicate());
    while(*TmpResult)
      TmpResult = &(*TmpResult)->Next; }
  return Result;
}
  
XxxValueExprList * XxxValueExpression::ValExprsReferenced() const
{
 XxxValueExprList *ReturnList = NULL;
 for (int i = 0; i < arity; i++) {
   XxxValueExprList *ChildVals = childExpr[i]->ValExprsReferenced();
   if (ChildVals) {
     if (ReturnList) {	
       ReturnList->Append(ChildVals);
      } 
     else {
       ReturnList = ChildVals;
      }
    }
  }
 // base case
 if (ValType() != BOOL_EXPR) {
   XxxValueExprList *Me = new XxxValueExprList(Replicate());
   if (ReturnList) {
     ReturnList->Append(Me);
    } 
   else {
     ReturnList = Me;
    }
  }
 return ReturnList;
}

XxxValueExprList * XxxValueExpression::SubExprsReferenced() const
{
 XxxValueExprList *ReturnList = NULL;
 for (int i = 0; i < arity; i++) {
   XxxValueExprList *ChildVals = childExpr[i]->SubExprsReferenced();
   if (ChildVals) {
     if (ReturnList) {	
       ReturnList->Append(ChildVals);
      } 
     else {
       ReturnList = ChildVals;
      }
    }
  }
 // base case
 XxxValueExprList *Me = new XxxValueExprList(Replicate());
 if (ReturnList) {
   ReturnList->Append(Me);
 } 
 else {
   ReturnList = Me;
 }
 return ReturnList;
}

//
// Methods of XxxValueExprList::
//

XxxValueExprList::XxxValueExprList(const XxxValueExprList& src) 
{
 Expr = (XxxValueExpression*)src.Expr->Replicate();
 if (src.Next)
   Next = new XxxValueExprList(*src.Next);
 else 
   Next = NULL;
}

XxxErrCode 
XxxValueExprList::ResolveVars(const XxxRecordSchema* Schema)
{
 XxxValueExprList *Tmp = this;

 while (Tmp) {
   if (Tmp->Expr->ResolveVars (Schema) != XXX_OK)
     return XXX_NOT_OK;
   Tmp = Tmp->Next;
  }
 return XXX_OK;
}

// this is usually called from INSERT INTO
XxxErrCode 
XxxValueExprList::TypeCheck(const XxxRecordSchema *RecSchema)
{
 if (Count() != RecSchema->Arity()) {
   XXX_NOTE_ERROR(XXX_BADEXPR, "Wrong arity!");
   return XXX_NOT_OK;
  }

 XxxValueExprList *Tmp = this;
 int i = 0;
 while (Tmp)
  {
   // In the case of cast expressions, the meta-info may be provided.
   // Make sure that the provided meta-info is compatible with the
   // corresponding meta-info in the schema.
   if (Tmp->Expr->MetaInfo && (*RecSchema)[i].TypeCatInfo) {
     if (!Tmp->Expr->MetaInfo->IsCompatible((*RecSchema)[i].TypeCatInfo)){
       XXX_NOTE_ERROR(XXX_BADEXPR, "Meta-info of cast expression not compatible");
       return XXX_NOT_OK;
      }
    }
   // If necessary, copy the meta-info over
   else if (!Tmp->Expr->MetaInfo && (*RecSchema)[i].TypeCatInfo) {
     // make sure to copy the meta-info
     ((*RecSchema)[i].TypeCatInfo)->Replicate(Tmp->Expr->MetaInfo);
     // TODO:L1 this needs to be reclaimed
    }
   Tmp = Tmp->Next;
   i++;
  }

 // TypeCheck can be executed only after the metaInfo assoicated with this
 // expression list has been copied over from table catalog.
 XXX_CALL_RET(TypeCheck(), XXX_BADEXPR, "Typecheck failed!");

 // reset Tmp
 Tmp = this;
 i=0;
 while (Tmp) {
   XxxBasicType ExprType = Tmp->Expr->BasicType();
   if ((*RecSchema)[i].Type != ExprType) {
     XXX_NOTE_ERROR(XXX_BADEXPR, "Exprlist doesnt match schema");
     return XXX_NOT_OK;
    }
   Tmp = Tmp->Next;
   i++;
  }

 return XXX_OK;
}
void XxxValueExprList::Print(ostream& OStr) const 
{
 const XxxValueExprList* Tmp = this;
 while (Tmp) {
   Tmp->printThis(OStr);
   if(Tmp->Expr) {
     if (Tmp->Expr->ExprName)
       OStr << " AS " << Tmp->Expr->ExprName;
    }     
   Tmp = Tmp->Next;
   if (Tmp) 
     OStr << ", ";
  }
}

void XxxValueExprList::DestroyListOnly()
{
 while (Next) {
   XxxValueExprList* Tmp = Next->Next;
   delete Next;
   Next = Tmp;
  }
 delete this;
}

void XxxValueExprList::Destroy()
{
 while (Next) {
   XxxValueExprList* Tmp = Next->Next;
   delete Next->Expr;
   delete Next;
   Next = Tmp;
  }
 delete Expr;
 delete this;
}

// PURPOSE: Sets the ind'th element of the list to Exp.  
// NOTES: If the list has n elements (0..n-1), and ind>=n, as many new
//        elements as necessary are created.  Unused elements have NULL
//        Expr members.  If the desired slot is already occupied, the old
//        expresssion is "deleted" before inserting the new exp.
XxxErrCode XxxValueExprList::SetElement(const int ind, 
					XxxValueExpression *Exp) 
{
 int i=0;
 XxxValueExprList *Iter=this;
 // move to the ind'th expression, or as close to it as we can get
 while (i<ind && Iter->Next) {
   Iter=Iter->Next;
   i++;
  }
 // if we're not there, add new elements until we are there.
 for (;i<ind;i++) {
   Iter->Next=new XxxValueExprList(NULL);
   Iter=Iter->Next;
  }
 // Iter now points to the ind'th element.  Set its Expr field to Exp
 if (Iter->Expr)
   delete Iter->Expr;
 Iter->Expr=Exp;
 return XXX_OK;
}

// PURPOSE: Gets the ind'th element of the list and returns it.
// NOTES: If there are < ind elements in the list, return NULL.  Also,
//        the caller can change the contents of the expression and this will
//        take effect in the list (i.e. we're not returning a copy of the
//        desired expresssion, but the actual pointer to it.
XxxErrCode
XxxValueExprList::GetElement(const int ind, XxxValueExpression*& Exp) const
{
 int i=0;
 const XxxValueExprList *Iter = this;
 while (i<ind && Iter->Next) {
   Iter=Iter->Next;
   i++;
  }
 if (i<ind) {
   // illegal index given
   Exp=NULL;
   XXX_NOTE_ERROR(XXX_OVFLO, "Index out of range while attempting to get "
		  "expression from XxxValueExprList");
   return XXX_NOT_OK;
  }
 Exp=Iter->Expr;
 return XXX_OK;
}

XxxValueExprList*
XxxValueExprList::Append(XxxValueExprList *ToAdd) 
{
 XxxValueExprList *Tmp=this;
 while (Tmp->Next!=NULL)
   Tmp=Tmp->Next;
 Tmp->Next=ToAdd;
 return this;
}
XxxValueExprList *
XxxValueExprList::TypeFilter(const XxxValueExpression::ValExprType Type) {
  if (Next)
    Next = Next->TypeFilter(Type);
  if (Expr->ValType() != Type) {
    XxxValueExprList *NextNode = Next;
    Next = NULL;
    Destroy();
    return NextNode;
  }
  return this;
}

XxxBool XxxValueExprList::Contains(XxxValueExpression *Expression)
{
 if (Expr->Match(Expression))
   return XXX_TRUE;
 else if (Next)
   return Next->Contains(Expression);
 else
   return XXX_FALSE;
}

XxxBool XxxValueExprList::IsSubsetOf(XxxValueExprList *SuperSet)
{
 XxxValueExprList *Tmp = SuperSet;
 while (Tmp) {
   if (Expr->Match(Tmp->Expr)) {
     if (!Next)
       return XXX_TRUE;
     else 
       return Next->IsSubsetOf(SuperSet);
    }
   Tmp = Tmp->Next;
  }
 return XXX_FALSE;
}

XxxBool XxxValueExprList::ReferencesAggregate() const
{
 const XxxValueExprList* Tmp = this;
 while (Tmp) {
   if (Tmp->Expr->ReferencesAggregate())
     return XXX_TRUE;
   Tmp = Tmp->Next;
  }
 return XXX_FALSE;
}

XxxBool XxxValueExprList::IsConstant() const
{
 const XxxValueExprList* Tmp = this;
 while (Tmp) {
   if (Tmp->Expr->IsConstant() != XXX_TRUE)
     return XXX_FALSE;
   Tmp = Tmp->Next;
  }
 return XXX_TRUE;
}

// try to resolve the vars in this expression using the schema of Node
// or its children
XxxErrCode XxxValueExprList::ResolveVars(XxxQueryNode* Node, char Source)
{
 XxxValueExprList* Tmp = this;
 while (Tmp) {
   if (Tmp->Expr->ResolveVars(Node, Source) != XXX_OK)
     return XXX_NOT_OK;
   Tmp = Tmp->Next;
  }
 return XXX_OK;
}
// try to resolve the vars in this expression using the host env
XxxErrCode XxxValueExprList::ResolveVars(const XxxArgEnv* Env)
{
 XxxValueExprList* Tmp = this;
 while (Tmp) {
   if (Tmp->Expr->ResolveVars(Env) != XXX_OK)
     return XXX_NOT_OK;
   Tmp = Tmp->Next;
  }
 return XXX_OK;
}

XxxErrCode XxxValueExprList::TypeCheck()
{
 XxxValueExprList* Tmp = this;
 while (Tmp) {
   if (Tmp->Expr->TypeCheck()!= XXX_OK)
     return XXX_NOT_OK;
   Tmp = Tmp->Next;
  }
 return XXX_OK;
}

XxxErrCode 
XxxValueExprList::GetDependencies(XxxBitMap& Dependency,
 				  int CorrelationHt) const
{
 const XxxValueExprList* Tmp = this;
 while (Tmp) {
   XXX_CALL_RET(Tmp->Expr->GetDependencies(Dependency, CorrelationHt),
		XXX_BADEXPR, "Error getting expr dependencies");
   Tmp = Tmp->Next;
  }
 return XXX_OK;
}

XxxErrCode XxxValueExprList::Optimize(XxxValExprPlanList*& PlanList)
{
 PlanList = NULL;
 XxxValueExprList* Tmp = this;
 XxxValueExprPlan* TmpPlan;
 while (Tmp) {
   XXX_CALL_RET(Tmp->Expr->Optimize(TmpPlan),
		XXX_BADOPT, "Error optimizing expr list");
   // add the plan to the list
   if (!PlanList)
     PlanList = new XxxValExprPlanList(TmpPlan, NULL);
   else
     PlanList = PlanList->AddExprPlan(TmpPlan);
   Tmp = Tmp->Next;
  }
 return XXX_OK;
}

int XxxValueExprList::Count() const
{
  int i = 1;
  XxxValueExprList* Tmp = Next;
  while (Tmp) {
    i++;
    Tmp = Tmp->Next;
  }
  return i;
}

XxxValueExpression* XxxValueExprList::operator[](int index) const
{
 const XxxValueExprList* Tmp = this;
 while (Tmp && index-- > 0)  Tmp = Tmp->Next;
 return (Tmp?Tmp->Expr:0);
}

XxxErrCode XxxValueExprList::
Replace (const XxxValueExpression *Target,
	 const XxxValueExpression *Replacement)
{
  XxxValueExprList* Tmp = this;
  while (Tmp) {
    if (Tmp->Expr->Match(Target)) {
      delete Tmp->Expr;
      Tmp->Expr = (XxxValueExpression *)Replacement->Replicate();
    } 
    else 
      XXX_CALL_RET(Tmp->Expr->Replace(Target, Replacement),
		   XXX_BADEXEC, "Cannot replace expression in list");
    Tmp = Tmp->Next;
  }
  return XXX_OK;
}

XxxErrCode XxxValueExprList::Replace (const XxxUnknownValue *Target,
				      const XxxValueExpression *Replacement)
{
  XxxValueExprList* Tmp = this;
  while (Tmp) {
    XxxValueExpression * CastExpr = (XxxValueExpression *)(Tmp->Expr);
    if (CastExpr->ValType() == XxxValueExpression::COLUMN_VAR && ((XxxUnknownValue *)CastExpr)->Equals(Target)) {
      delete Tmp->Expr;
      Tmp->Expr = (XxxValueExpression *)Replacement->Replicate();
    } 
    else {
      XXX_CALL_RET(Tmp->Expr->Replace(Target, Replacement),
		   XXX_BADEXEC, "Cannot replace expression in list");
    }
    Tmp = Tmp->Next;
  }
  return XXX_OK;
}

XxxErrCode XxxValueExprList::UpdateUnknowns(int ChildNumber,
					    int OldAttrIndex,
					    int NewAttrIndex,
					    XxxBool &Appears)
{
  XxxValueExprList* Tmp = this;
  while (Tmp) {
    XXX_CALL_RET(Tmp->Expr->UpdateUnknowns(ChildNumber,
					   OldAttrIndex,
					   NewAttrIndex,
					   Appears),
		 XXX_BADEXPR, "Error updating unknowns in subexpr");
    Tmp = Tmp->Next;
  }
  return XXX_OK;
}

// PURPOSE: Adds an expression to the end of the list
XxxValueExprList* XxxValueExprList::AddExpr(XxxValueExpression *NextExpr)
{  
 XxxValueExprList *ReturnValue = NULL;
 return AddExpr(NextExpr, ReturnValue);
}

// PURPOSE: Adds an expression to the end of the list.  Returns newly created
// element in second arg
XxxValueExprList *XxxValueExprList::AddExpr(XxxValueExpression *NextExpr, 
					    XxxValueExprList *&NewElt) 
{
 XxxValueExprList* Tmp = this;
 while (Tmp->Next)  
   Tmp = Tmp->Next;
 Tmp->Next = new XxxValueExprList(NextExpr);
 NewElt = Tmp->Next;
 return this;
}

