#include "exprplan.h"
#include "error.h"

//
// Methods of XxxValueExprPlan::
//

XxxValueExprPlan::XxxValueExprPlan(XxxValueExpression * Exp,
				   XxxValueExprPlan* Children[])
  : dependency(1) 
// initialy, the only dependency is for possible env args
{
 arity = Exp->Arity();
 isConstant = Exp->IsConstant();
 ostrstream Strm(printBuf, 100);
 Exp->Print(Strm);
 Strm << '\0';

 Exp->GetDependencies(dependency, 0);

 basicType = ((XxxValueExpression*)Exp)->BasicType();
 valType = ((XxxValueExpression*)Exp)->ValType();
 catInfo = ((XxxValueExpression*)Exp)->GenerateCatInfo();
 // Praveen : added this because the original expression is needed
 // at a couple of places at run time (mostly in index matching code).
 // Ideally, the index matching code should be a method of ExprPlan.
 // For now, the easiest thing is to replicate the expression. TODO:L2
 sourceExpr = (XxxValueExpression*)Exp->Replicate();

 SubPlans = NULL;
 if (arity) {
   SubPlans = new XxxValueExprPlan*[arity];
   for (int i = 0; i < arity; i++) {
     // note: the subplans are not replicated, because the expectation
     // is that they are created explicitly in order to create this plan
     SubPlans[i] = Children[i];
     // do not add costs here: costs need to be determined by the
     // specific kind of plan (note: some plans may not always
     // evaluate their subplans, so additive costs may not be right)
   }
 }
}

XxxValueExprPlan::XxxValueExprPlan(
					XxxValueExpression::ValExprType ExprType,
					int Arity,
					XxxBool Constant,
					XxxValueExprPlan* Children[])
: dependency(1)
{ 
 valType	= ExprType;
 arity		= Arity;
 isConstant = Constant;
 sourceExpr = NULL;
 catInfo	= NULL;

 SubPlans = NULL;
 if (arity) {
   SubPlans = new XxxValueExprPlan*[arity];
   for (int i = 0; i < arity; i++) {
     // note: the subplans are not replicated, because the expectation
     // is that they are created explicitly in order to create this plan
     SubPlans[i] = Children[i];
   }
 }
}

//Bjarke, sep. 14: explicitly instantiated base class XxxGenericPlan
XxxValueExprPlan::XxxValueExprPlan (const XxxValueExprPlan &Plan) : 
  XxxGenericPlan(),
  basicType(Plan.basicType), valType(Plan.valType), 
  planUsage(Plan.planUsage), isConstant(Plan.isConstant),
  arity(Plan.arity), dependency(Plan.dependency)
{
  strncpy(printBuf, Plan.printBuf, 100);
  /*
  if (Plan.catInfo && (Plan.catInfo->Replicate(catInfo) != XXX_OK)) {
    catInfo = NULL;
  } substituted by the next lines (Tobias): */
  if (!Plan.catInfo)
    catInfo = NULL;
  else if (Plan.catInfo && 
	   ((Plan.catInfo)->Replicate(catInfo) != XXX_OK)) {
    catInfo = NULL;
  }
  if (Plan.sourceExpr)
    sourceExpr = (XxxValueExpression*)Plan.sourceExpr->Replicate();
  else
    sourceExpr = NULL;

 SubPlans = NULL;
 if (arity) {
   SubPlans = new XxxValueExprPlan*[arity];
   for (int i = 0; i < arity; i++) {
     SubPlans[i] = (XxxValueExprPlan*)((Plan.SubPlans[i] ? Plan.SubPlans[i]->Replicate() : 0));
   }
 }
}

XxxValueExprPlan::~XxxValueExprPlan() 
{
  if (sourceExpr) delete sourceExpr; 
  if (arity && SubPlans) {
    for (int i = 0; i < arity; i++) {
      if (SubPlans[i]) delete SubPlans[i];
    }
    delete [] SubPlans;
  }
}

XxxErrCode 
XxxValueExprPlan::GetDependencies(XxxBitMap& Dependency) const
{
 if (Dependency.ArraySize < dependency.ArraySize) {
   char buf[50];
   sprintf(buf, "Difference between expr input cards: %d and %d",
 	   Dependency.ArraySize, dependency.ArraySize);
   XXX_NOTE_ERROR(XXX_BADEXPR, buf);
   return XXX_NOT_OK;
  }
 Dependency.CopyBits(dependency);
 return XXX_OK;
}

XxxErrCode XxxValueExprPlan::
RedirectDependencies(const VarRemapping& Remap,
		     int CorrelationHt)
{
  for (int i = 0; i < arity; i++) {
    if (SubPlans[i]) {
      XXX_CALL_RET(SubPlans[i]->RedirectDependencies(Remap, 
						     CorrelationHt),
		   XXX_BADOPT, "Error changing depends mapping");
    }
  } // for

  if (CorrelationHt == 0) {
    XXX_CALL_RET(changeDependency(Remap),
		 XXX_BADOPT, "Error changing depends mapping");
  }
  return XXX_OK;
}

// all subclasses should actually change the SrcIndex of any
// XxxUnknownValue inside them, and then call this superclass
// method to change the stored dependency array.
XxxErrCode XxxValueExprPlan::changeDependency(const VarRemapping& Remap)
{
 // first create a new depends array and initialize to XXX_FALSE
 XxxBitMap NewDep(Remap.NewArity+1);

 // now copy new dependencies from mapping
 // for (int i = 0; i < Remap.NewArity; i++) {
 for (int i = 0; i < Remap.OldArity; i++) {
   if (dependency[i] == XXX_TRUE) {
     // be cautious!
     if (Remap.SrcMappingArray[i].NewSrc < 0) { // invalid remapping
       NewDep.SetBit(i); // copy the old dependency
       continue; 
     }
     if (Remap.SrcMappingArray[i].NewSrc >= Remap.NewArity) {
       XXX_NOTE_ERROR(XXX_BADEXPR, "Error in redirect mapping");
       return XXX_NOT_OK;
      }
     NewDep.SetBit(Remap.SrcMappingArray[i].NewSrc);
    }
  }

 // copy the dependency on host variables
 if (dependency[Remap.NewArity] == XXX_TRUE)
   NewDep.SetBit(Remap.NewArity);

 // change to new array
 dependency = NewDep;

 return XXX_OK;
}

XxxErrCode XxxValueExprPlan::MakeOpHandle(XxxAggrOpHandle *&Handle) const
{
  /*******
  // TODO: do this recursively
  Handle = new XxxAggrOpHandle(BasicType(), GenerateCatInfo());
  return XXX_OK;
  ****/
  return XXX_NOT_OK;
}

XxxErrCode XxxValueExprPlan::Init(XxxAggrOpHandle *Handle) const
{
  if (!Handle) return XXX_OK;
  /******
  for (int i = 0; i < arity; i++) {
    if (SubPlans[i]) {
      XXX_CALL_RET(SubPlans[i]->Init(Handle->GetSubHandle(i)),
		   XXX_BADOPT, "Error initializing aggregate");
    }
  } // for
  return XXX_OK;
  ******/
  return XXX_NOT_OK;
}
XxxErrCode XxxValueExprPlan::Accumulate(const XxxRecord* Rec[],
					const XxxRecordSchema* Schema[],
					XxxAggrOpHandle *Handle,
					const XxxValueEnv* Env) const
{
  if (!Handle) return XXX_OK;
  /******
  for (int i = 0; i < arity; i++) {
    if (SubPlans[i]) {
      XXX_CALL_RET(SubPlans[i]->Accumulate(Rec, Schema,
					   Handle->GetSubHandle(i), Env),
		   XXX_BADOPT, "Error accumulating aggregate");
    }
  } // for
  return XXX_OK;
  *******/
  return XXX_NOT_OK;
}
XxxErrCode XxxValueExprPlan::Terminate(XxxAggrOpHandle *Handle,
				       const XxxRecord* Rec[],
				       const XxxRecordSchema* Schema[],
				       XxxADTValue *FinalVal,
				       const XxxValueEnv* Env) const
{
  if (!Handle) return XXX_OK;
  /******* TODO: this needs to be thought through a little more to
        get aggregate expressions to work within other expressions
  for (int i = 0; i < arity; i++) {
    if (SubPlans[i]) {
      XXX_CALL_RET(SubPlans[i]->Terminate(Handle->GetSubHandle(i)),
		   XXX_BADOPT, "Error initializing aggregate");
    }
  } // for
  return XXX_OK;
  *******/
  return XXX_NOT_OK;
}

//
// Methods of XxxValueExprPlan::VarRemapping
//

XxxValueExprPlan::VarRemapping::VarRemapping(int Size)
{
 OldArity = Size;
 SrcMappingArray = new struct SrcMapping[Size];
 for (int i = 0; i < Size; i++) {
   SrcMappingArray[i].NewSrc = -1;
   SrcMappingArray[i].NewAttrMapping = NULL;
  }
}

XxxValueExprPlan::VarRemapping::~VarRemapping()
{
 for (int i = 0; i < OldArity; i++) {
   if (SrcMappingArray[i].NewAttrMapping)
     delete [] SrcMappingArray[i].NewAttrMapping;
  }
 delete [] SrcMappingArray;
}

//
// XxxValExprPlanList::
//

XxxValExprPlanList * XxxValExprPlanList::Replicate() const
{
  XxxValExprPlanList *Item = 
    new XxxValExprPlanList((XxxValueExprPlan *)Plan->Replicate());
  Item->Next = ((Next != NULL) ? Next->Replicate() : NULL);
  return Item;
}

void XxxValExprPlanList::Print(ostream& OStr) const 
{
 const XxxValExprPlanList* Tmp = this;
 while (Tmp) {
   Tmp->printThis(OStr);
   Tmp = Tmp->Next;
   if (Tmp) 
     OStr << ", ";
  }
}

XxxErrCode XxxValExprPlanList::
RedirectDependencies(const XxxValueExprPlan::VarRemapping& Remap,
		     int CorrelationHt)
{
 const XxxValExprPlanList *Tmp = this;
 while (Tmp) {
   if (Tmp->Plan->RedirectDependencies(Remap, CorrelationHt) != XXX_OK)
     return XXX_NOT_OK;
   Tmp = Tmp->Next;
  }
 return XXX_OK;
}

XxxErrCode
XxxValExprPlanList::Evaluate(const XxxRecord* InputRecords[],
			     const XxxRecordSchema* InputSchemas[],
			     XxxADTValue *Result[],
			     const XxxValueEnv* Env) const 
{
 const XxxValExprPlanList *Tmp = this;
 int i = 0;
 while (Tmp) {
   if (Tmp->Plan->Evaluate(InputRecords, InputSchemas, 
			   Result[i], Env) != XXX_OK) {
     char buf[100];
     sprintf(buf, "Error evaluating list expression %d", i);
     XXX_NOTE_ERROR(XXX_BADEXEC, buf);
     return XXX_NOT_OK;
    }
   i++;
   Tmp = Tmp->Next;
  }
 return XXX_OK;
}

// PURPOSE: destroys only outer shell
void XxxValExprPlanList::DestroyListOnly() 
{
  while(Next) {
    XxxValExprPlanList *Tmp = Next->Next;
    delete Next;
    Next = Tmp;
  }
  delete this;
}

// PURPOSE: destroys the entire list
void XxxValExprPlanList::Destroy()
{
  while (Next) {
    XxxValExprPlanList* Tmp = Next->Next;
    if(Next->Plan)
      delete Next->Plan;
    delete Next;
    Next = Tmp;
  }
  delete Plan;
  delete this;
}

XxxValExprPlanList::~XxxValExprPlanList()
{
}

int XxxValExprPlanList::Count()
{
 int i = 1;
 XxxValExprPlanList* Tmp = Next;
 while (Tmp) {
   i++;
   Tmp = Tmp->Next;
  }
 return i;
}

XxxValueExprPlan* XxxValExprPlanList::operator[](int index) const
{
 const XxxValExprPlanList* Tmp = this;
 while (Tmp && index-- > 0)  Tmp = Tmp->Next;
 return (Tmp?Tmp->Plan:0);
}

XxxValExprPlanList* 
XxxValExprPlanList::AddExprPlan(XxxValueExprPlan *NextExprPlan)
{
 XxxValExprPlanList* Tmp = this;
 while (Tmp->Next)  Tmp = Tmp->Next;
 Tmp->Next = new XxxValExprPlanList(NextExprPlan);
 return this;
}





