#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include "funcexpr.h"
#include "aggrexpr.h"
#include "generic.h"
#include "prd-config.h"
#include "strrules.h"
#include "adtfunc.h"
#include "engines.h"

// defined in optrules.C
extern void XxxAssignTopLevel(XxxValueExpression* ChildExpr,
			      int ParentTypeId);
const int XxxFuncExpression::aggrTypeCount = 10;
const XxxAggrOpSpec* const XxxFuncExpression::AggrOpSpecs[10] = 
{
 new XxxMinOpSpec,
 new XxxMaxOpSpec,
 new XxxCountOpSpec,
 new XxxSumOpSpec,
 new XxxAvgOpSpec,
 new XxxProdOpSpec,
 NULL,     // Any : not implemented
 NULL,     // All : not implemented
 NULL,     // Most : not implemented
 NULL      // ADT
  };

XxxBool XxxFuncExpression::stdAggr(const char* AggrName)
{
  XxxAggrType AggType = ADT;
  findAggrType(AggrName, AggType);
  if (AggType == ADT) return XXX_FALSE;
  else return XXX_TRUE;
}

void XxxFuncExpression::findAggrType(const char* AggrName, 
				     XxxAggrType& AggType)
{
  AggType = ADT;
  int i;
  for (i = 0; i < aggrTypeCount; i++) {
    if (AggrOpSpecs[i] && 
	!strcasecmp(AggrName, AggrOpSpecs[i]->AggrName())) {
      AggType = (XxxAggrType)i;
      break;
    }
  }
}

//Bjarke, aug. 8: const added to char* parameter
XxxFuncExpression::XxxFuncExpression(const char* FuncName, 
				     XxxValueExprList* ArgL,
				     XxxValueExpression* OwnerObject)
{
  //Bjarke, sep. 17 2001: added cast
  ASSERT((strlen(FuncName) < (unsigned int)XXX_FUNCNAME_MAX));
  // skip dummy $ for generic functions
  if(FuncName && FuncName[0] == '$')
    strcpy(Name, FuncName + 1);
  else
    strcpy(Name, FuncName);
  ArgList = ArgL;
  Owner = OwnerObject;
  OwnerTypeCatInfo = NULL;
  FuncParse = NULL;
  isAggr = XXX_FALSE;
}

XxxFuncExpression::XxxFuncExpression(const XxxFuncExpression& src):
  XxxValueExpression(src)
{
  OwnerType = src.OwnerType;
  FuncParse = NULL;
  ArgList = NULL;
  Owner = NULL;
  OwnerTypeCatInfo = NULL;
  if (src.OwnerTypeCatInfo)
    src.OwnerTypeCatInfo->Replicate(OwnerTypeCatInfo);
	
  strncpy(Name, src.Name, XXX_FUNCNAME_MAX);
  Name[XXX_FUNCNAME_MAX-1] = 0;
	
  if(src.FuncParse)
    FuncParse = (XxxFuncParseInfo*)src.FuncParse->Replicate();
  if (src.ArgList)
    ArgList = new XxxValueExprList(*(src.ArgList));
  if (src.Owner)
    Owner = (XxxValueExpression*)src.Owner->Replicate();
  AggType = src.AggType;
  isAggr = src.isAggr;
}

XxxFuncExpression::~XxxFuncExpression()
{
  if (ArgList) 
    ArgList->Destroy();
  if (Owner) delete Owner;
  if (OwnerTypeCatInfo) 
    delete OwnerTypeCatInfo;
  // reclaim FuncParse
  if (FuncParse) {
    delete FuncParse;
  }
}

XxxValueExprList *XxxFuncExpression::DirectSubexpressions()
{
  XxxValueExprList *Result = XxxValueExpression::DirectSubexpressions();
  XxxValueExprList **TmpResult = &Result;
  while(*TmpResult) 
    TmpResult = &(*TmpResult)->Next;
  (*TmpResult) = new XxxValueExprList(Owner);
  TmpResult = &(*TmpResult)->Next;
  if(ArgList)
    (*TmpResult) = ArgList->Replicate();
  return Result;
}
  
XxxValueExprList * XxxFuncExpression::ValExprsReferenced() const 
{
  XxxValueExprList *ReturnList = XxxValueExpression::ValExprsReferenced();
  XxxValueExprList *Tmp = NULL;
  if (Owner && (Tmp = Owner->ValExprsReferenced())) {
    if (ReturnList)
      ReturnList->Append(Tmp);
    else
      ReturnList = Tmp;
  }

  if (ArgList) {
    /*****
    Tmp = ArgList->Replicate();
    if (ReturnList)
      // TODO:IMM shouldnt this be ArgList->ValExprReferenced()??
      ReturnList->Append(Tmp);
    else
      ReturnList = Tmp;
      *****/
    // step thru the ArgList 
    XxxValueExprList* TmpList = ArgList;
    while (TmpList) {
      if (Tmp = TmpList->Expr->ValExprsReferenced()) {
	if (ReturnList) ReturnList->Append(Tmp);
	else ReturnList = Tmp;
      }
      TmpList = TmpList->Next;
    } // while (TmpList)
  } // if (ArgList)

  return ReturnList;
}

XxxValueExprList * XxxFuncExpression::SubExprsReferenced() const 
{
  XxxValueExprList *ReturnList = XxxValueExpression::SubExprsReferenced();
  XxxValueExprList *Tmp = NULL;
  if (Owner && (Tmp = Owner->SubExprsReferenced())) {
    if (ReturnList)
      ReturnList->Append(Tmp);
    else
      ReturnList = Tmp;
  }

  if (ArgList) {
    XxxValueExprList* TmpList = ArgList;
    while (TmpList) {
      if (Tmp = TmpList->Expr->SubExprsReferenced()) {
	if (ReturnList) ReturnList->Append(Tmp);
	else ReturnList = Tmp;
      }
      TmpList = TmpList->Next;
    } // while (TmpList)
  } // if (ArgList)

  return ReturnList;
}

void XxxFuncExpression::Print(ostream& OStr) const
{
  if (Owner) {
    Owner->Print(OStr);
    OStr << '.';
  }
  // generic function name should be preceeded by $
  if(!isAggr && Name && Name[0] != '$')
    OStr << "$" ;
  OStr << Name << "(" ;
  if (ArgList)
    ArgList->Print(OStr);
  OStr << ")";
}

XxxBool XxxFuncExpression::ReferencesAggregate() const
{
  // trivially true if this is an aggregate
  if (IsAggr()) return XXX_TRUE;

  if (ArgList) {
    if (ArgList->ReferencesAggregate()) {
      return XXX_TRUE;
    }
  }
  if (Owner && Owner->ReferencesAggregate())
    return XXX_TRUE;
  return XXX_FALSE;
}

XxxBool XxxFuncExpression::IsConstant() const
{
  // An aggregate expression is never a "constant".
  // It is always tied to the value/cardinality of the data
  if (IsAggr()) return XXX_FALSE;

  if (ArgList) {
    if (ArgList->IsConstant() != XXX_TRUE) {
      return XXX_FALSE;
    }
  }
  if (Owner) {
    if (Owner->IsConstant() != XXX_TRUE) {
      return XXX_FALSE;
    }
  }
  return XXX_TRUE;
}

XxxErrCode XxxFuncExpression::Replace (const XxxExpression *Target,
				       const XxxExpression *Replacement)
{
  if (XxxExpression::Replace(Target, Replacement) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADEXEC, "Attempt to replace failed for superclass");
    return XXX_NOT_OK;
  }
  if (ArgList) {
    if (ArgList->Replace(Target, Replacement) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot replace expression in arguments");
      return XXX_NOT_OK;
    }
  }
  if (Owner) {
    if (Owner->Match(Target)) {
      delete Owner;
      Owner = (XxxValueExpression *)Replacement->Replicate();
    }
  }
  return XXX_OK;
}

XxxErrCode XxxFuncExpression::Replace (const XxxUnknownValue *Target,
				       const XxxExpression *Replacement)
{
  if (XxxExpression::Replace(Target, Replacement) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADEXEC, "Attempt to replace failed for superclass");
    return XXX_NOT_OK;
  }
  if (ArgList) {
    if (ArgList->Replace(Target, Replacement) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot replace expression in arguments");
      return XXX_NOT_OK;
    }
  }
  if (Owner) {
    if (Owner->ValType() == COLUMN_VAR && ((XxxUnknownValue* )Owner)->Equals(Target)) {
      delete Owner;
      Owner = (XxxValueExpression *)Replacement->Replicate();
    } else {
      if (Owner->Replace(Target, Replacement) != XXX_OK) {
	XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot replace expression in owner");
      }
    }
  }
  return XXX_OK;
}

XxxErrCode XxxFuncExpression::UpdateUnknowns(int ChildNumber,
					      int OldAttrIndex,
					      int NewAttrIndex,
					      XxxBool &Appears)
{
  if (ArgList)
    if (ArgList->UpdateUnknowns(ChildNumber,
				 OldAttrIndex,
				 NewAttrIndex,
				 Appears) != XXX_OK) 
      return XXX_NOT_OK;
  if (Owner)
    if (Owner->UpdateUnknowns( ChildNumber,
				 OldAttrIndex,
				 NewAttrIndex,
				 Appears) != XXX_OK) 
      return XXX_NOT_OK;
  return XXX_OK;
}

XxxErrCode
XxxFuncExpression::ResolveVars(const XxxRecordSchema *Schema)
{
  XxxErrCode rc = XXX_OK;
  if (ArgList) {
    if ((rc = ArgList->ResolveVars(Schema)) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADFUNC, "Unable to resolve function arguments");
      return rc;
    }
  }
  if (Owner) {
    if ((rc = Owner->ResolveVars(Schema)) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADFUNC, "Unable to resolve expression");
      return rc;
    }
  }
  return rc;
}

XxxErrCode XxxFuncExpression::ResolveVars(XxxQueryNode* Node, char Source)
{
  XxxErrCode rc = XXX_OK;
  if (ArgList) {
    if ((rc = ArgList->ResolveVars(Node, Source)) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADFUNC, "Unable to resolve function arguments");
      return rc;
    }
  }
  if (Owner) {
    if ((rc = Owner->ResolveVars(Node, Source)) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADFUNC, "Unable to resolve expression");
      return rc;
    }
  }
  return rc;
}

XxxErrCode 
XxxFuncExpression::GetDependencies(XxxBitMap& Dependency,
 				   int CorrelationHt) const
{
  if (ArgList) {
    XXX_CALL_RET(ArgList->GetDependencies(Dependency, CorrelationHt),
		 XXX_BADFUNC, "Unable to get deps of function arguments");
  }
  if (Owner) {
    XXX_CALL_RET(Owner->GetDependencies(Dependency, CorrelationHt),
		 XXX_BADFUNC, "Unable to resolve expression");
  }
  return XXX_OK;
}

XxxErrCode XxxFuncExpression::TypeCheck()
{
  XxxErrCode rc = XXX_OK;
	
  int NumArgs = (ArgList?ArgList->Count():0);
  if (NumArgs >= XXX_ARGCNT_MAX) {
    XXX_NOTE_ERROR(XXX_OVFLO, "Too many args for function");
    return XXX_NOT_OK;
  }
  if (IsAggr()) {
    if (ArgList && ArgList->ReferencesAggregate()) {
      XXX_NOTE_ERROR(XXX_BADTYPE, "Nested aggregates not allowed");
      return XXX_NOT_OK;
    }
    findAggrType(Name, AggType);
  }

  // type check the arguments
  if (ArgList && (rc = ArgList->TypeCheck()) != XXX_OK) {
    char buf[100];
    sprintf(buf, "Type check failed for arg of func %s", Name);
    XXX_NOTE_ERROR(XXX_BADTYPE, buf);
    return rc;
  }
  if (Owner) {
    if (Owner->TypeCheck() != XXX_OK) {
      char buf[100];
      sprintf(buf, "Type check failed for owner of func %s", Name);
      XXX_NOTE_ERROR(XXX_BADTYPE, buf);
      return XXX_NOT_OK;
    }
    XxxADTMetaInfo *OwnerCatInfo = Owner->GenerateCatInfo();
    if (OwnerCatInfo)
      OwnerCatInfo->Replicate(OwnerTypeCatInfo);
    else 
      OwnerTypeCatInfo = NULL;
    OwnerType = Owner->BasicType();
  } 
  else {
    // this case should never arise, i think
    OwnerType = XXX_GENFUNC;
    OwnerTypeCatInfo = NULL;
  }
	
  XxxADTClass* AttrClass = XxxADTArray[OwnerType];
  // type check the function itself
  XXX_CALL_RET(AttrClass->FuncTypeCheck(Name, Owner, ArgList, IsAggr(), 
					BaseType, FuncParse),
	       XXX_BADFUNC, Name);
	
  // Reset IsTopLevel flag in ArgList expressions' FuncParse to XXX_FALSE 
  // because *this is at a higher level (if the types are compatible).
  XxxValueExprList *Iter=ArgList;
  while (Iter) {
    XxxAssignTopLevel(Iter->Expr, FuncParse->TypeId);
    Iter = Iter->Next;
  }
	
  // type check the Owner, and assign OwnerType
  if (Owner) {
    // Reset IsTopLevel flag in Owner->FuncParse to XXX_FALSE because
    // *this is at a higher level (if they are type compatible).
    XxxAssignTopLevel(Owner, FuncParse->TypeId);
  }

  return XXX_OK;
}

XxxADTMetaInfo *XxxFuncExpression::GenerateCatInfo() const
{
  if (!FuncParse)
    return XxxValueExpression::GenerateCatInfo();
  else {
    // need to get catalog information from Parse Tree:
    return FuncParse->GenerateCatInfo();
  }
}

XxxErrCode XxxFuncExpression::Optimize(XxxValueExprPlan*& ExprPlan)
{
  // The most simple optimizations involve simply reordering
  // functions, and collapsing several functions into one.
  // Do this here before the rest of optimization. 
  // Note: FuncReOrganize can totally change the contents of "this"
  XxxValueExpression *ThisCopy;
  if (FuncParse && FuncParse->IsTopLevel == XXX_TRUE) {
    ThisCopy = new XxxFuncExpression(*this);
    if (XxxADTArray[FuncParse->TypeId]->FuncReOrganize(ThisCopy) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADFUNC, Name);
      if (ThisCopy)
	delete ThisCopy;
      return XXX_NOT_OK;
    }
    // now that it has been reorganized, set "top-level" to false
    // and reinvoke Optimize() --- this is a bit clumsy
    if (ThisCopy->ValType() == XxxValueExpression::FUNC_EXPR) {
      ((XxxFuncExpression*)ThisCopy)->FuncParse->IsTopLevel = XXX_FALSE;
    } 
		
    if (ThisCopy->Optimize(ExprPlan) != XXX_OK) {
      delete ThisCopy;
      XXX_NOTE_ERROR(XXX_BADOPT, "Error optimizing function expression");
      return XXX_NOT_OK;
    }
	// Philippe: 08/99 the following line corrupts ExprPlan (OwnerPlan.catinfo)
	// Philippe: the destructor for XxxValueExprPlan deletes all the source expression.
    //delete ThisCopy;
    return XXX_OK;
  }

  // if we get this far, we're not dealing with a top-level funcexpr.
  FuncPlan* Plan;
  if (IsAggr())
    Plan = new AggrPlan(this);
  else 
    Plan = new FuncPlan(this);
	
  // optimize the owner and the value expr list
  int OwnerSize = 0;
  if (Owner) {
    OwnerTypeCatInfo = Owner->GenerateCatInfo();
    OwnerSize = XxxADTArray[(int)(Owner->BasicType())]
      ->MaxObjectSize(OwnerTypeCatInfo);
    Plan->OwnerPlan = NULL;
		
    //optimize the owner
    XXX_CALL_RET(Owner->Optimize(Plan->OwnerPlan),
		 XXX_BADFUNC, "Error optimizing function owner");
  }
  Plan->ArgPlans = NULL;
  XxxValueExprList* Tmp = ArgList;
  int argCnt = 0;
  while (Tmp) {
    XxxValueExprPlan* ArgPlan = NULL;
    if (Tmp->Expr->Optimize(ArgPlan)!= XXX_OK) {
      char buf[100];
      sprintf(buf, "Error optimizing function argument %d", argCnt-1);
      XXX_NOTE_ERROR(XXX_BADFUNC, buf);
      return XXX_NOT_OK;
    }
    argCnt++;
		
    // add the plan to the list
    if (!Plan->ArgPlans)
      Plan->ArgPlans = new XxxValExprPlanList((XxxValueExprPlan*)ArgPlan, 
					      NULL);
    else
      Plan->ArgPlans = Plan->ArgPlans->AddExprPlan((XxxValueExprPlan*)ArgPlan);
    // move down the list
    Tmp = Tmp->Next;
  }
	
	
  if (FuncParse) {
    // invoke the optimizer
    XxxADTClass* AttrClass = XxxADTArray[OwnerType];
    if (AttrClass->FuncOptimize(FuncParse, 
				Plan->OwnerPlan,
				Plan->ArgPlans,
				Plan->ExprPlan) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADFUNC, "Error optimizing function expression");
      return XXX_NOT_OK;
    }
    // add the cost of the ExprPlan (which holds the costs of OwnerPlan
    // and ArgPlans within it
    if (Plan->ExprPlan) {
      Plan->getPlanUsage().Add(Plan->ExprPlan->GetPlanUsage());
    }
  }
  else { // TODO: check if this case ever gets called
    // add in the costs of OwnerPlan and ArgPlans
    if (Plan->OwnerPlan)
      Plan->getPlanUsage().Add(Plan->OwnerPlan->GetPlanUsage());
    // step thru arg plans adding cost
    const XxxValExprPlanList* Tmp = Plan->ArgPlans;
    while (Tmp) {
      Plan->getPlanUsage().Add(Tmp->Plan->GetPlanUsage());
      Tmp = Tmp->Next;
    }
  }

  // have to copy the data lest it get reclaimed
  XxxADTMetaInfo *OCatInfo=NULL;
  if(OwnerTypeCatInfo) {
    OwnerTypeCatInfo->Replicate(OCatInfo);
  }
  // allocate space for run-time arguments
  Plan->ValueEnv = new XxxValueEnv(ArgList, OwnerSize, OCatInfo);
	
  ExprPlan = Plan;
  return XXX_OK;
}

XxxFuncExpression::
FuncPlan::FuncPlan(const FuncPlan &Plan) : 
  XxxValueExprPlan(Plan), ownerType(Plan.ownerType), 
  returnType(Plan.returnType)
{
  strncpy(Name, Plan.Name, XXX_FUNCNAME_MAX);
  if (Plan.ownerTypeCatInfo && 
      (Plan.ownerTypeCatInfo->Replicate(ownerTypeCatInfo) != XXX_OK))
    ownerTypeCatInfo = NULL;
  if (Plan.returnCatInfo && 
      (Plan.returnCatInfo->Replicate(returnCatInfo) != XXX_OK))
    returnCatInfo = NULL;
  if (Plan.ExprPlan)
    ExprPlan = (XxxFuncPlanInfo *)Plan.ExprPlan->Replicate();
  else ExprPlan = NULL;
  if (Plan.ValueEnv)
    ValueEnv = Plan.ValueEnv->Replicate();
  else ValueEnv = NULL;
  if (Plan.OwnerPlan)
    OwnerPlan = (XxxValueExprPlan *)Plan.OwnerPlan->Replicate();
  else OwnerPlan = NULL;
  if (Plan.ArgPlans)
    ArgPlans = Plan.ArgPlans->Replicate();
  else ArgPlans = NULL;
}

XxxFuncExpression::
FuncPlan::FuncPlan(XxxFuncExpression* FuncExpr)
  : XxxValueExprPlan(FuncExpr, 0) 
{
  strncpy(Name, FuncExpr->Name, XXX_FUNCNAME_MAX);
  Name[XXX_FUNCNAME_MAX-1] = 0;
  OwnerPlan = NULL;
  ArgPlans = NULL;
  ValueEnv = NULL;
  ExprPlan = NULL;
  ownerType = FuncExpr->OwnerType;
  ownerTypeCatInfo = FuncExpr->OwnerTypeCatInfo;
  returnType = FuncExpr->BasicType();
  returnCatInfo = FuncExpr->GenerateCatInfo();
}

XxxFuncExpression::
FuncPlan::~FuncPlan()
{
  if (ExprPlan) delete ExprPlan;
  if (ValueEnv) delete ValueEnv;
  if (OwnerPlan) delete OwnerPlan;
  if (ArgPlans) delete ArgPlans;
}

XxxErrCode XxxFuncExpression::
FuncPlan::RedirectDependencies(const VarRemapping& Remap,
 			       int CorrelationHt)
{
  if (OwnerPlan) {
    XXX_CALL_RET(OwnerPlan->RedirectDependencies(Remap, CorrelationHt),
		 XXX_BADOPT, "Error changing owner plan depends mapping");
  }
  if (ArgPlans) {
    XXX_CALL_RET(ArgPlans->RedirectDependencies(Remap, CorrelationHt),
		 XXX_BADOPT, "Error changing arg plans depends mapping");
  }
  if (CorrelationHt == 0) {
    XXX_CALL_RET(changeDependency(Remap),
 		 XXX_BADOPT, "Error changing depends mapping");
  }
  return XXX_OK;
}

// PURPOSE: AggrList and GroupList are subsequences of TargetList
// To evaluate the function, we must first evaluate all the arguments to
// the function. We then call the function with the evaluated arguments
XxxErrCode XxxFuncExpression::
FuncPlan::Evaluate(const XxxRecord* Rec[],
		   const XxxRecordSchema* Schema[],
		   XxxADTValue *Result,
		   const XxxValueEnv* Env) const
{
  XxxErrCode rc = XXX_OK;

  // if the owner exists, and should be evaluated, do so
  XXX_CALL_RET(evalOwner(Rec, Schema, ValueEnv->OwnerValPtr, Env),
	       XXX_BADEXEC, "Error evaluating owner");
  return Evaluate1(Rec, Schema, NULL, Result, Env);
}

// OWNER: as  DATE: 2/20/97
// PURPOSE: AggrList and GroupList are subsequences of TargetList
// To evaluate the function, we must first evaluate all the arguments to
// the function. We then call the function with the evaluated arguments
// If Precomputed Owner is NULL, the Owner Value in ValueEnv is used
// NOTES: 
// I am overloading this method, to compute a func-expression with a 
// pre-computed owner
XxxErrCode XxxFuncExpression::
FuncPlan::Evaluate1(const XxxRecord* Rec[], 
		    const XxxRecordSchema* Schema[],
		    XxxADTValue *PreComputedOwner, 
		    XxxADTValue *Result, 
		    const XxxValueEnv* Env) const 
{
  if (PreComputedOwner) {
    // copy this value into the the ValueEnv->OwnerValPtr
    XxxADTArray[ownerType]->TypeCopy(PreComputedOwner, 
				     ValueEnv->OwnerValPtr,
				     ownerTypeCatInfo);
  }
	
  // compute the arguments
  XXX_CALL_RET(evalArgs(Rec, Schema, ValueEnv->ArgValsPtrs, Env),
	       XXX_BADEXEC, "Error evaluating func args");
  
  // compute the function
  ASSERT((ExprPlan));

  if (ExprPlan->ExprPlan) {
    // this is an embedded language expression (like $SQL)
    XXX_CALL_RET(XxxDBEngineArray[ExprPlan->EngineIndex]->
		 ExprEvaluate(ExprPlan->ExprPlan, ValueEnv, Result),
		 XXX_BADEXEC, "Error in language expression");
  }
  else {
    XxxADTClass* AttrClass = XxxADTArray[ownerType];
    XxxFuncMethodInfo* ThisMethInfo = 
		AttrClass->GetMethByIndex(ExprPlan->MethodIndex);
    if (!ThisMethInfo) {
      XXX_NOTE_ERROR(XXX_INTERNAL, "Method info doesnt exist");
      return XXX_NOT_OK;
    }
    XXX_CALL_RET(ExprPlan->Evaluate(ValueEnv, 
				    ThisMethInfo, Result),
		 XXX_BADFUNC, "Error evaluating adt method");
  }

  return XXX_OK;
}

XxxErrCode XxxFuncExpression::
FuncPlan::evalOwner(const XxxRecord* Rec[],
		    const XxxRecordSchema* Schema[],
		    XxxADTValue *OwnerVal,
		    const XxxValueEnv* Env) const
{
  if (OwnerPlan) {
    XXX_CALL_RET(OwnerPlan->Evaluate(Rec, Schema, OwnerVal, Env),
		 XXX_BADFUNC, "Error in evaluating func owner");
  }
  return XXX_OK;
}

XxxErrCode XxxFuncExpression::
FuncPlan::evalArgs(const XxxRecord* Rec[],
		   const XxxRecordSchema* Schema[],
		   XxxADTValue **ArgVals,
		   const XxxValueEnv* Env) const
{
  int i;
  XxxValExprPlanList* EList;
  for (i=0,EList = ArgPlans; EList; i++) {
    XXX_CALL_RET(EList->Plan->Evaluate(Rec, Schema, 
				       ArgVals[i], Env),
		 XXX_BADFUNC, "Error in evaluating func args");
    EList = EList->Next;
  } // for
  return XXX_OK;
}

XxxBool 
XxxFuncExpression::Match(const XxxExpression* OtherExpr, 
			 char* Env) const
{
  if (!XxxValueExpression::Match(OtherExpr, Env))
    return XXX_FALSE;
  XxxFuncExpression *CastExpr = (XxxFuncExpression *)OtherExpr;
  // Compare their names
  if (strcmp(Name,CastExpr->Name))
    return XXX_FALSE;
  // Check if Owners match
  if ((Owner && !CastExpr->Owner) ||
      (CastExpr->Owner && !Owner)) 
    return XXX_FALSE;
  if (Owner && !Owner->Match(CastExpr->Owner, Env))
    return XXX_FALSE;
  // Check to see if all arguments match
  if (ArgList) {
    if (!CastExpr->ArgList)
      return XXX_FALSE;
    if (ArgList->Count() != CastExpr->ArgList->Count()) {
      return XXX_FALSE;
    }
    else {
      int i;
      for (i=0;i<ArgList->Count();i++) {
	if (! ((*ArgList)[i]->Match((*CastExpr->ArgList)[i], Env)))
	  return XXX_FALSE;
      }
    }
  }
 if ((IsAggr() && !CastExpr->IsAggr()) ||
     (!IsAggr() && CastExpr->IsAggr()))
   return XXX_FALSE;
  return XXX_TRUE;
}

// TODO: this is wrong: need to back out any partial bindings
// in case of error
XxxBool 
XxxFuncExpression::Match(const XxxExpression* RuleExpr, 
			 XxxBindingEnv *BindEnv) const
{
  if (!XxxValueExpression::Match(RuleExpr, BindEnv))
    return XXX_FALSE;
  XxxFuncExpression *CastExpr = (XxxFuncExpression *)RuleExpr;
  // Compare their names
  //if (strcmp(Name,CastExpr->Name))
  if (BindEnv->Match(CastExpr->Name, Name) != XXX_TRUE)
    return XXX_FALSE;

  if ((Owner && !CastExpr->Owner) ||
      (CastExpr->Owner && !Owner)) 
    return XXX_FALSE;

  // If The owner of CastExpr is an UnknownValue, Use the match value
  // to update it.
  if (CastExpr->Owner &&
      (CastExpr->Owner->ValType() == COLUMN_VAR)) {
    XxxExpression* BoundExpr 
      = BindEnv->GetElement((XxxUnknownValue *)CastExpr->Owner);
    if (BoundExpr) { 
      // check if it matches Owner
      if (!BoundExpr->Match(Owner))
	return XXX_FALSE;
    }
    else {
      // add new binding to the binding env
      if ((BindEnv->SetElement((XxxUnknownValue *)CastExpr->Owner, 
			       Owner)) != XXX_OK) {
	XXX_NOTE_ERROR(XXX_BADOPT, "Cannot set the binding environment");
	return XXX_FALSE;
      }
    } // else
  } // if the owner is a COLUMN_VAR
  else { 
    // Check if Owners match
    if (Owner && !Owner->Match(CastExpr->Owner, BindEnv))
      return XXX_FALSE;
  }
  // Check to see if all arguments match
  if (ArgList) {
    if (!CastExpr->ArgList) {
      return XXX_FALSE;
    }

    if (ArgList->Count() != CastExpr->ArgList->Count()) {
      return XXX_FALSE;
    } 
    else {
      int i;
      for (i=0;i<ArgList->Count();i++) {
	if (! ((*ArgList)[i]->Match((*CastExpr->ArgList)[i],BindEnv)) ){
	  return XXX_FALSE;
	} // if
      } // for each arg
    } // else
  } // if (ArgList)

 if ((IsAggr() && !CastExpr->IsAggr()) ||
     (!IsAggr() && CastExpr->IsAggr()))
   return XXX_FALSE;

 // TODO: commit bindings only at this stage
 return XXX_TRUE;
}


