#ifndef XXX_FUNCEXPR_H
#define XXX_FUNCEXPR_H
#include <iostream.h>
#include "expr.h"
#include "values.h"
#include "exprplan.h"
#include "record.h"

XXX_EXTERN_CLASS_DECL(XxxAggrOpSpec); // defined in aggrexpr.h

/**
 * Models all functions (and class methods)
 */
class XxxFuncExpression: public XxxValueExpression {
  friend class XxxGenFuncADTClass;// eventually remove the need for this
 protected:
  XxxBool isAggr;
  typedef enum {MIN = 0, 
		MAX = 1, 
		COUNT = 2, 
		SUM =3, 
		AVG = 4, 
		PROD = 5, 
		ANY = 6, 
		ALL = 7, 
		MOST = 8,
	        ADT = 9} XxxAggrType;
  static const int aggrTypeCount;
  static const XxxAggrOpSpec* const AggrOpSpecs[];

  static XxxBool stdAggr(const char* AggrName);
  static void findAggrType(const char* AggrName, XxxAggrType& AggType);

 public:
  //// Plan for XxxFuncExpression
  class FuncPlan : public XxxValueExprPlan { 
  protected:
    friend class XxxFuncExpression;

    XxxBasicType ownerType;
    XxxADTMetaInfo *ownerTypeCatInfo; // any additional information about the owner
    XxxBasicType returnType;
    XxxADTMetaInfo *returnCatInfo;

    XxxErrCode evalOwner(const XxxRecord* Rec[],
			 const XxxRecordSchema* Schema[],
			 XxxADTValue *OwnerVal,
			 const XxxValueEnv* Env = NULL) const;
    XxxErrCode evalArgs(const XxxRecord* Rec[],
			const XxxRecordSchema* Schema[],
			XxxADTValue** ArgVals,
			 const XxxValueEnv* Env = NULL) const;
    
   public:
    XxxFuncPlanInfo* ExprPlan; // for language expressions
    XxxValueExprPlan* OwnerPlan;   // plan for the owner
    XxxValExprPlanList* ArgPlans; // plan for the arguments
    char  Name[XXX_FUNCNAME_MAX];

    // TODO: i dont think i like this. it would be nice to keep the
    // plans independent of run-time structures specific to each 
    // evaluation (this stuff should be in a separate handle)
    XxxValueEnv* ValueEnv; // to store values of arguments at run time

    FuncPlan(XxxFuncExpression* FuncExpr);
    FuncPlan(const FuncPlan &Plan);
    virtual XxxGenericPlan * Replicate() const {
      return (XxxGenericPlan *)(new FuncPlan(*this));
    }    
    virtual ~FuncPlan();
    XxxBasicType ReturnType() { return returnType; }
    XxxADTMetaInfo *ReturnCatInfo() { return returnCatInfo; }
    virtual XxxErrCode Evaluate(const XxxRecord* Rec[],
				const XxxRecordSchema* Schema[],
				XxxADTValue *Result,
				const XxxValueEnv* Env = NULL) const;
    virtual XxxErrCode Evaluate1(const XxxRecord* Rec[],
				 const XxxRecordSchema* Schema[],
				 XxxADTValue *PreComputedOwner, 
				 XxxADTValue *Result,
				 const XxxValueEnv* Env = NULL) const;
    virtual XxxErrCode RedirectDependencies(const VarRemapping& Remap,
     					    int CorrelationHt);
    virtual XxxBool IsAggr() const { return XXX_FALSE;}
  }; // end of Class FuncPlan
  friend class FuncPlan;

  class AggrPlan : public FuncPlan { 
   public:
    XxxAggrType AggType;

	/**
	 * Used in gbyplanop.C to determine what to do with empty aggregates
	 */
	XxxBool IsCountAggr() { return AggType == COUNT; }

    AggrPlan(XxxFuncExpression* AExpr) : 
      XxxFuncExpression::FuncPlan(AExpr) { 
      AggType = AExpr->AggType;
     }
    AggrPlan(const AggrPlan &Plan);
    virtual ~AggrPlan() {}

    virtual XxxGenericPlan * Replicate() const {
      return (XxxGenericPlan *)(new AggrPlan(*this));
    } 
    /// Evaluate() is not meaningful for an aggregate expression.
    virtual XxxErrCode Evaluate(const XxxRecord* Rec[],
				const XxxRecordSchema* Schema[],
				XxxADTValue *Result,
				const XxxValueEnv* Env = NULL) const {
      return XXX_NOT_OK;
    }

    virtual XxxErrCode MakeOpHandle(XxxAggrOpHandle*& Handle) const;
    virtual XxxErrCode Init(XxxAggrOpHandle *Handle) const;
    virtual XxxErrCode Accumulate(const XxxRecord* Rec[],
				  const XxxRecordSchema* Schema[],
				  XxxAggrOpHandle *Handle,
				  const XxxValueEnv* Env = NULL) const;
    //// This drops the input from the AggrOpHandle.
    virtual XxxErrCode Drop(const XxxRecord* Rec[],
			    const XxxRecordSchema* Schema[],
			    XxxAggrOpHandle *Handle,
			    const XxxValueEnv* Env = NULL) const;

    virtual XxxErrCode Terminate(XxxAggrOpHandle *Handle,
				 XxxADTValue *FinalVal) const;
    virtual XxxBool IsAggr() const { return XXX_TRUE;}
  }; // end of class AggrPlan
  friend class AggrPlan;

 public:
  XxxBasicType OwnerType;
  XxxAggrType AggType; // not meaningful unless its an aggregate
  XxxADTMetaInfo *OwnerTypeCatInfo;  // any additional information about the owner type
  // structure to store the parse tree for the function
  XxxFuncParseInfo *FuncParse;

  char  Name[XXX_FUNCNAME_MAX];
  XxxValueExprList* ArgList;
  XxxValueExpression* Owner;
//Bjarke, aug. 8: const added to char* parameter
  XxxFuncExpression(const char* FuncName, XxxValueExprList* ArgList, 
		    XxxValueExpression* OwnerObject = NULL);
  XxxFuncExpression(const XxxFuncExpression& src);
  /**
   * Returns a list of the immediate subexpressions of the expression.
   * The list consists of the children or the owner and arguments of an 
   * expression. Used in the AggregateFunctionPushDownRule. (mayr@cs)
   */
  virtual XxxValueExprList *DirectSubexpressions();  
  /**
   * Returns a list of replications of all value expressions referenced in
   * the expression.  It is the caller's responsibility to destroy the
   * list.
   */
  virtual XxxValueExprList *ValExprsReferenced() const;
  virtual XxxValueExprList *SubExprsReferenced() const; 
  /// Provides a convenient way to copy arbitrary expressions
  virtual XxxExpression * Replicate() const
    { return new XxxFuncExpression(*this); }
  virtual ~XxxFuncExpression();

  virtual ValExprType ValType() const {return FUNC_EXPR;}
  virtual void Print(ostream&) const;
  /**
   * Replaces all references to the target expression with references
   * to the replacement expression.  Uses the match function to determine
   * if the target expression is referenced.
   */
  virtual XxxErrCode Replace (const XxxExpression *Target,
			      const XxxExpression *Replacement);
  /**
   * Replaces all references to the target expression with references
   * to the replacement expression.  Used in XxxStrRule (Binquan)
   */
  virtual XxxErrCode Replace (const XxxUnknownValue *Target,
			      const XxxExpression *Replacement);
  /**
   * Substitutes NewAttrIndex for OldAttrIndex in the SrcAttrIndex field 
   * of XxxUnknownValues subexpressions with SrcChild == ChildNumber
   */
  virtual XxxErrCode UpdateUnknowns(int ChildNumber,
				    int OldAttrIndex,
				    int NewAttrIndex,
				    XxxBool &Appears);
  /// Is this an aggregate function?
  XxxBool IsAggr() const { return isAggr;}
  /// Is this expression based in any way on an aggregate expression?
  virtual XxxBool ReferencesAggregate() const;
  virtual XxxBool IsConstant() const;
  virtual XxxADTMetaInfo *GenerateCatInfo() const;
  XxxErrCode ResolveVars (const XxxRecordSchema * Schema);
  XxxErrCode ResolveVars(XxxQueryNode* Node, char Source);
  virtual XxxErrCode TypeCheck();
  // Gets input stream dependencies (that were resolved in ResolveVars)
  // Sets flags in the input argument DependsArray. The
  // array must be initialized to be all false. The size of
  // the array is 1+ the actual number of input streams.
  // The extra entry stands for the run-time Env
  virtual XxxErrCode GetDependencies(XxxBitMap& Dependendency,
 				     int CorrelationHt) const;
  virtual XxxErrCode Optimize(XxxExprPlan*& Plan);

  // Check to see if matches other expression
  virtual XxxBool Match(const XxxExpression* OtherExpr, 
			char* Env = NULL) const;
   /**
   *  Checks if the expression matches another rule input expression 
   *  If match is successful, return true, and assign BindTerm array
   *  inside XxxBindingEnv class.
   */
  virtual XxxBool Match(const XxxExpression* RuleExpr, 
			XxxBindingEnv *BindEnv) const;
};

#endif
