#ifndef XXX_GENERICSRC_H
#define XXX_GENERICSRC_H

class XxxGenDataSrc {
 protected:
  int defaultDevSize;
 public:
  // class used during parsing a Create Table/Seq stmt
  class AttrTypeList { public:
    char Name[XXX_ATTR_NAME_MAX];
    XxxBasicType Type;
    XxxBool CanBeNull;
    XxxADTClass::MetaInfoHandle MIHndl;
    AttrTypeList* Next;
    AttrTypeList(const char* AttrName, XxxBasicType AttrType, 
		 XxxBool NullOK, const XxxADTClass::MetaInfoHandle& MIH)
      : MIHndl(MIH) {
     CanBeNull = NullOK;
     Type = AttrType;
     strncpy(Name, AttrName, XXX_ATTR_NAME_MAX-1);
     Name[XXX_ATTR_NAME_MAX-1]=0;
     if(strlen(AttrName)>XXX_ATTR_NAME_MAX-1)   //--> emc 2/25/97
       cerr << "Warning: " << AttrName << " will be truncated to " << Name 
	    << "\n";
     Next = NULL;
    }
    XxxBool Contains(const char *AttrName)
    {
      if (!strcmp(Name, AttrName))
	return XXX_TRUE;
      if (Next && Next->Contains(AttrName))
	return XXX_TRUE;
      return XXX_FALSE;
    }
    AttrTypeList* Add(const char* AttrName, XxxBasicType AttrType, 
		      XxxBool NullOK, const XxxADTClass::MetaInfoHandle& MIH) {
      AttrTypeList* Tmp = new AttrTypeList(AttrName, AttrType, NullOK, MIH);
      Tmp->Next = this;
      return Tmp;
     }
   };
  // class used to store key list while parsing Create Table stmt
  class KeyList { public:
    char AttrName[XXX_ATTR_NAME_MAX];
    KeyList* Next;
    KeyList(const char* Name) {
      strncpy(AttrName, Name, XXX_ATTR_NAME_MAX);
      AttrName[XXX_ATTR_NAME_MAX-1] = 0;
      Next = NULL;
     }
    KeyList* Add(const char* Name) {
      KeyList* Tmp = new KeyList(Name);
      Tmp->Next = this;
      return Tmp;
     }
   };

  char SrcName[XXX_SRCNAME_MAX];

  XxxGenDataSrc(const char* Name, int defDevSize) {
   defaultDevSize = defDevSize;
   strncpy(SrcName, Name, XXX_SRCNAME_MAX);
   SrcName[XXX_SRCNAME_MAX-1] = 0;
  }
  virtual ~XxxGenDataSrc() {};

  // this is important! each data src must specify the size of its
  // handle, so that the handle can be safely stored in a record
  virtual int HandleSize() const = 0;

  virtual void Print(ostream& OStr) {
    OStr << SrcName;
   }
  virtual void PrintSchemas(ostream& OStr) = 0;
  virtual void PrintCats(ostream& OStr) = 0;
  virtual void PrintDevs(ostream& OStr) = 0;

  // If the device name is provided, a device of that name is created
  // If the provided name is NULL, a new device is created and its name
  // is returned via the DeviceName argument.
  virtual XxxErrCode CreateDevice(char*& DeviceName, 
				  const int DeviceSize)=0;
  // Define one version of MountDevice in terms of the other!!
  virtual XxxErrCode MountDevice(const char* DevicePath) {
    return MountDevice(DevicePath, NULL);
   }
  virtual XxxErrCode MountDevice(const char* DevicePath,
				 const char* DeviceName) {
    // MountDevice(DevicePath);
    return XXX_NOT_OK;
   }
  virtual XxxErrCode UnMountDevice(const char* DeviceName) = 0;
  virtual XxxErrCode DestroyDevice(const char* DeviceName) = 0;  
  virtual XxxBool IsMounted(const char* DevName) const=0;

  int DefaultDevSize() const { return defaultDevSize; }
 };

#endif
