#include <iostream.h>
#include <iomanip.h>
#include <time.h>
#include "httprequest.h"
#include "xxx_thread.h"
#include "shoresm-defs.h"

#define URI_MAX_LENGTH 2048

//
// Methods of XxxHTTPRequest
//
XxxHTTPRequest::XxxHTTPRequest(const char *request) : 
  Major(0), Minor(9), RequestType(HTTP_GET), URI(NULL), 
  RequestBuffer(NULL), BufferLength(0)
{
  istrstream IStr((char *)request);
  // Read the request type
  char Method[10];
  IStr >> Method;
  if (!strcmp(Method, "GET"))
    RequestType = HTTP_GET;
  else if (!strcmp(Method, "POST"))
    RequestType = HTTP_POST;
  else if (!strcmp(Method, "HEAD"))
    RequestType = HTTP_HEAD;
  // Read the URI
  char URIarray[URI_MAX_LENGTH];
  IStr >> URIarray;
  URI = new char[strlen(URIarray) + 1];
  strcpy(URI, URIarray);
  // Read the version
  IStr >> setw(5) >> URIarray >> Major >> URIarray[0] >> Minor;
}

XxxHTTPRequest::~XxxHTTPRequest()
{
  if (URI != NULL)
    delete [] URI;
  if (RequestBuffer != NULL)
    delete [] RequestBuffer;
}

static void PrintCurrentDate(ostream &OStr)
{

#ifdef XXX_WIN32
  clock_t Time = clock();
  OStr << Time;
  OStr << flush;
#else
  struct timeval Time;
  struct timezone TZ;
  gettimeofday(&Time, &TZ);
  long seconds_since_UCT = Time.tv_sec;
  // PRD: CHG2
  asctime(gmtime(&seconds_since_UCT));
  char buf[26];
  istrstream IStr(buf);
  char weekday[4], month[4], colon;
  int hours, minutes, seconds, day, year;
  IStr >> weekday >> month >> day 
       >> hours >> colon >> minutes >> colon >> seconds
       >> year;
  OStr << weekday << ", " 
       << day << " " << month << " " << year << " "
       << hours << ":" << minutes << ":" << seconds << " GMT";
  OStr << flush;

#endif

}

static void WriteByteToHex(char byte, ostream& OStr) {
  char c = (byte & 0xf0) >> 4;

  if (c < 10)
    OStr << (char)('0' + c);
  else
    OStr << (char)('A' + c - 10);
  c = (byte & 0x0f);
  if (c < 10)
    OStr << (char)('0' + c);
  else
    OStr << (char)('A' + c - 10);
}

XxxErrCode XxxHTTPRequest::WriteCallbackURL(const int ClientId, 
					    const XxxShoreXactId &XactId,
					    const int ADTTypeId, //Bjarke aug. 3: int added as type
					    const XxxADTValue *ObjectRef,
					    const XxxADTMetaInfo *MetaInfo,
					    const int MessageLength,
					    const char *Message,
					    ostream &OStr)
{
  /*
   * Parse the URI:
   *   (a.) (b.)    (c.) (d.) (e.)  (f.) (g.)  (h.) (i.)
   *  |----|--...--|----|----|-...-|----|-...-|----|-...-|
   *    a. Client ID (integer)
   *    b. Transaction Id (a binary sequence of known length)
   *    c. Type ID (integer)
   *    d. Size of the object (integer)
   *    e. Object (sequence of (d.) bytes)
   *    f. Size of the object's meta-info (integer)
   *    g. Meta-info (sequence of (f.) bytes)
   *    h. Size of the additional message (integer)
   *    i. Additional message (sequence of (h.) bytes)
   */
  OStr << "http://" 
       << XXX_THISENV().HostName << ':' << XXX_THISENV().Port << '/';
  char c = 0;
  int i;
  int TmpClientId;
  XxxNetworkInt(&TmpClientId, ClientId);
  // Write out the client ID
  for (i = 0; i < (int)sizeof(int); i++)
    WriteByteToHex(*((char *)(&TmpClientId) + i), OStr);
  // Write out the Transaction ID
  for (i = 0; i < (int)sizeof(XxxShoreXactId); i++)
    WriteByteToHex(*((char *)(&XactId) + i), OStr);
  int TmpTypeId;
  XxxNetworkInt(&TmpTypeId, ADTTypeId);
  // Write out the Type ID
  for (i = 0; i < (int)sizeof(int); i++)
    WriteByteToHex(*((char *)(&TmpTypeId) + i), OStr);
  // Write out the size of the object
  int TmpLength;
  int Size = XxxADTArray[ADTTypeId]->TypeSize(ObjectRef, MetaInfo);
  XxxNetworkInt(&TmpLength, Size);
  for (i = 0; i < (int)sizeof(int); i++)
    WriteByteToHex(*((char *)(&TmpLength) + i), OStr);
  // Write out the object
  char *NetworkObject = new char[Size];
  if (XxxCoreADTClass::IsCoreADT(ADTTypeId)) {
    if (XxxCoreADTClass::SCopyHToN(ADTTypeId, 
				   ObjectRef,
				   (XxxADTValue *)NetworkObject, 
				   MetaInfo)!= XXX_OK){
      XXX_NOTE_ERROR(XXX_BADADT, "Error creating big-endian version");
      return XXX_NOT_OK;
    }
  } else {
    if (XxxADTArray[ADTTypeId]->CopyHToN(ObjectRef, 
					 (XxxADTValue *)NetworkObject,	
					 MetaInfo) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_BADADT, "Error creating big-endian version");
      return XXX_NOT_OK;
    }
  }
  for (i = 0; i < Size; i++)
    WriteByteToHex(NetworkObject[i], OStr);
  delete [] NetworkObject;
  // Write out the size of the metainfo
  if (MetaInfo->SerializeCopySize(Size) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADADT, "Cannot determine size of meta info");
    return XXX_NOT_OK;
  }
  XxxNetworkInt(&TmpLength, Size);
  for (i = 0; i < (int)sizeof(int); i++)
    WriteByteToHex(*((char *)(&TmpLength) + i), OStr);
  // Write out the metainfo
  char *NetworkMI = new char[Size];
  if (MetaInfo->SerializeCopy(NetworkMI) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADADT, "Cannot serialize meta info");
    return XXX_NOT_OK;
  }
  for (i = 0; i < Size; i++)
    WriteByteToHex(NetworkMI[i], OStr);
  // Write out the size of the message
  XxxNetworkInt(&TmpLength, MessageLength);
  for (i = 0; i < (int)sizeof(int); i++)
    WriteByteToHex(*((char *)(&TmpLength) + i), OStr);
  // Write out the message
  for (i = 0; i < MessageLength; i++)
    WriteByteToHex(Message[i], OStr);
  OStr << flush;

  if (!OStr.good()) {
    XXX_NOTE_ERROR(XXX_BADEXEC, "URL buffer not large enough");
    return XXX_NOT_OK;
  }
  return XXX_OK;
}

// TODO: Yikes! i dont like this (though its not clear whats better)
static const char *HTTP_NOT_FOUND_REPONSE = "<HTML><HEAD><TITLE>PREDATOR Response: Item Not Found</TITLE></HEAD><BODY BGCOLOR=\"#FFFFFF\" TEXT=\"#000000\" LINK=\"#CC1111\" VLINK=\"#CC0000\" ALINK=\"#CC0000\"><CENTER><IMG SRC=\"http://www.cs.cornell.edu/Info/Projects/PREDATOR/predator.gif\"></CENTER><HR><H3>Not Found</H3>The item you requested was not found.  If your request was a callback, the item was either a temporary result or has been deleted by another transaction.<HR><A HREF=\"http://www.cs.cornell.edu/Info/Projects/PREDATOR/predator.html\"><FONT COLOR=\"#CC0000\"><FONT SIZE=+1>P</FONT>REDATOR</FONT> Home Page</A></BODY></HTML>";

static const char *HTTP_BAD_REQUEST_REPONSE = "<HTML><HEAD><TITLE>PREDATOR Response: Bad Request</TITLE></HEAD><BODY BGCOLOR=\"#FFFFFF\" TEXT=\"#000000\" LINK=\"#CC1111\" VLINK=\"#CC0000\" ALINK=\"#CC0000\"><CENTER><IMG SRC=\"http://www.cs.cornell.edu/Info/Projects/PREDATOR/predator.gif\"></CENTER><HR><H3>Bad Request</H3>Your request was not a properly formatted callback.<HR><A HREF=\"http://www.cs.cornell.edu/Info/Projects/PREDATOR/predator.html\"><FONT COLOR=\"#CC0000\"><FONT SIZE=+1>P</FONT>REDATOR</FONT> Home Page</A></BODY></HTML>";

static const char *HTTP_SERVER_ERROR_REPONSE = "<HTML><HEAD><TITLE>PREDATOR Response: Server Error</TITLE></HEAD><BODY BGCOLOR=\"#FFFFFF\" TEXT=\"#000000\" LINK=\"#CC1111\" VLINK=\"#CC0000\" ALINK=\"#CC0000\"><CENTER><IMG SRC=\"http://www.cs.cornell.edu/Info/Projects/PREDATOR/predator.gif\"></CENTER><HR><H3>Server Error</H3>An error occurred while processing your request.<HR><A HREF=\"http://www.cs.cornell.edu/Info/Projects/PREDATOR/predator.html\"><FONT COLOR=\"#CC0000\"><FONT SIZE=+1>P</FONT>REDATOR</FONT> Home Page</A></BODY></HTML>";

XxxErrCode 
XxxHTTPRequest::WriteResponseHeader(ostream &OStr, 
				    XxxHTTPReturnCode ReturnCode,
				    const char *Message,
				    int ContentLength,
				    const char *ContentType,
				    const char *ContentEncoding) const
{
  if (XXX_THISENV().DebugFlags.Callback) {
    cerr << "Response with Content-Type:" << ContentType << " and Length " << ContentLength << endl;
  }
  OStr << "HTTP/1.0 " << (int)ReturnCode << " ";
  if (Message)
    OStr << Message;
  else
    OStr << "(No message supplied)";
  OStr << endl;
  OStr << "Server: PREDATOR" << endl;
  OStr << "Date: ";
  PrintCurrentDate(OStr);
  OStr << endl;
  if (((int)ReturnCode / 100) == 2) {
    // The response includes data; write out the rest of the header
    OStr << "Content-Type: " << ContentType << endl;
    OStr << "Content-Length: " << ContentLength << endl;
    if (ContentEncoding)
      OStr << "Content-encoding: " << ContentEncoding << endl;
    OStr << "Expires: ";
    PrintCurrentDate(OStr);
    OStr << endl;
    OStr << "Last Modified: ";
    PrintCurrentDate(OStr);
    OStr << endl;
    OStr << endl << flush;
  } else if (((int)ReturnCode / 100) == 3) {
    // A response with no data following
    OStr << endl << flush;
  } else if (ReturnCode == HTTP_NOT_FOUND) {
    // The request could not be satisfied because the resource wasn't found
    OStr << "Content-Type: text/html" << endl;
    OStr << "Content-Length: " << (int)strlen(HTTP_NOT_FOUND_REPONSE) 
	 << endl;
    OStr << endl << HTTP_NOT_FOUND_REPONSE << flush;
  } else if (ReturnCode == HTTP_BAD_REQUEST) {
    // The request could not be satisfied because the request was 
    // not a callback
    OStr << "Content-Type: text/html" << endl;
    OStr << "Content-Length: " << (int)strlen(HTTP_BAD_REQUEST_REPONSE) 
	 << endl;
    OStr << "Last Modified: ";
    PrintCurrentDate(OStr);
    OStr << endl;
    OStr << endl << HTTP_BAD_REQUEST_REPONSE << flush;
  } else if (((int)ReturnCode / 100) == 5) {
    // The request could not be satisfied because the server was incapable
    OStr << "Content-Type: text/html" << endl;
    OStr << "Content-Length: " << (int)strlen(HTTP_SERVER_ERROR_REPONSE) 
	 << endl;
    OStr << "Last Modified: ";
    PrintCurrentDate(OStr);
    OStr << endl;
    OStr << endl << HTTP_SERVER_ERROR_REPONSE << flush;
  }
  return XXX_OK;
}

static char HexToByte(char c) {
  if ((c >= '0') && (c <= '9'))
    return (c - '0');
  else
    return (c - 'A' + 10);
}

XxxBool XxxHTTPRequest::IsCallback(int &ClientId, 
				   XxxShoreXactId *&XactId,
				   int &ADTTypeId,
				   XxxADTValue *&ObjectRef,
				   XxxADTMetaInfo *&MetaInfo,
				   int &MessageLength,
				   char *&Message)
{
  if (RequestType != HTTP_GET)
    return XXX_FALSE;
  // The URI must be prefixed by a slash ('/')
  if (URI[0] != '/')
    return XXX_FALSE;
  // Override the global definition for ease
  char *URI = &(this->URI[1]);
  // Attempt converting the URI from hexadecimal to binary
  if (RequestBuffer != NULL)
    delete [] RequestBuffer;
  BufferLength = (strlen(URI) / 2);
  RequestBuffer = new char[BufferLength];
  int i;
  for (i = 0; i < (int)strlen(URI); i += 2) {
    if ((((URI[i] >= 'A') && (URI[i] <= 'F')) ||
	 ((URI[i] >= '0') && (URI[i] <= '9'))) &&
	(((URI[i + 1] >= 'A') && (URI[i + 1] <= 'F')) ||
	 ((URI[i + 1] >= '0') && (URI[i + 1] <= '9')))) {
      RequestBuffer[i / 2] = HexToByte(URI[i]) << 4 | HexToByte(URI[i + 1]);
    } else {
      delete [] RequestBuffer;
      RequestBuffer = NULL;
      return XXX_FALSE;
    }
  }
  /*
   * Parse the URI:
   *   (a.) (b.)    (c.) (d.) (e.)  (f.) (g.)  (h.) (i.)
   *  |----|--...--|----|----|-...-|----|-...-|----|-...-|
   *    a. Client ID (integer)
   *    b. Transaction Id (a binary sequence of known length)
   *    c. Type ID (integer)
   *    d. Size of the object (integer)
   *    e. Object (sequence of (d.) bytes)
   *    f. Size of the object's meta-info (integer)
   *    g. Meta-info (sequence of (f.) bytes)
   *    h. Size of the additional message (integer)
   *    i. Additional message (sequence of (h.) bytes)
   */
  while (XXX_TRUE) {
    int curPos = 0;
    XxxHostInt(&ClientId, *((int *)(RequestBuffer + curPos)));
    curPos += sizeof(int);
    if (curPos >= BufferLength)
      break;
    XactId = (XxxShoreXactId *)(RequestBuffer + curPos);
    curPos += sizeof(XxxShoreXactId);
    if (curPos >= BufferLength)
      break;
    XxxHostInt(&ADTTypeId, *((int *)(RequestBuffer + curPos)));
    curPos += sizeof(int);
    if (curPos >= BufferLength)
      break;
    int Size;
    XxxHostInt(&Size, *((int *)(RequestBuffer + curPos)));
    curPos += sizeof(int);
    if (curPos >= BufferLength)
      break;
    ObjectRef = (XxxADTValue *)(RequestBuffer + curPos);
    curPos += Size;
    if (curPos >= BufferLength)
      break;
    XxxHostInt(&Size, *((int *)(RequestBuffer + curPos)));
    curPos += sizeof(int);
    if (curPos >= BufferLength)
      break;
    MetaInfo = (XxxADTMetaInfo *)(RequestBuffer + curPos);
    curPos += Size;
    if (curPos >= BufferLength)
      break;
    XxxHostInt(&MessageLength, *((int *)(RequestBuffer + curPos)));
    curPos += sizeof(int);
    if ((curPos >= BufferLength) && (MessageLength > 0))
      break;
    Message = (RequestBuffer + curPos);

    if ((curPos + MessageLength) > BufferLength)
      break;

    return XXX_TRUE;
  }
  
  delete [] RequestBuffer;
  RequestBuffer = NULL;

  return XXX_FALSE;
}
