#ifndef XXX_HTTPREQUEST_H
#define XXX_HTTPREQUEST_H

#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include "types.h"
#include "shoresm-defs.h"

/**
 * A HyperText Transfer Protocol request.
 * @author Mark A. Paskin, paskin@cs.cornell.edu
 * @version September 5, 1997
 */
class XxxHTTPRequest {
public:
  /**
   * The type of the request.
   */
  enum XxxHTTPRequestTypeEnum {
    HTTP_GET = 0,
    HTTP_POST,
    HTTP_HEAD
  };
  /// A typedef for easy reference
  typedef enum XxxHTTPRequestTypeEnum XxxHTTPRequestType;
  /**
   * Supported return code types.
   */
  enum XxxHTTPReturnCodeEnum {
    HTTP_OK                     = 200,
    HTTP_CREATED                = 201,
    HTTP_ACCEPTED               = 202,
    HTTP_NO_CONTENT             = 204,
    HTTP_MOVED_PERMANENTLY      = 301,
    HTTP_MOVED_TEMPORARILY      = 302,
    HTTP_NOT_MODIFIED           = 303,
    HTTP_BAD_REQUEST            = 400,
    HTTP_UNAUTHORIZED           = 401,
    HTTP_FORBIDDEN              = 403,
    HTTP_NOT_FOUND              = 404,
    HTTP_INTERNAL_SERVER_ERROR  = 500,
    HTTP_NOT_IMPLEMENTED        = 501,
    HTTP_BAD_GATEWAY            = 502,
    HTTP_SERVICE_UNAVAILABLE    = 503
  };
  /// A typedef for easy reference
  typedef enum XxxHTTPReturnCodeEnum XxxHTTPReturnCode;
  /// The major version of this HTTP request
  int Major;
  /// The minor version of this HTTP request
  int Minor;
  /// The request type of this request
  XxxHTTPRequestType RequestType;
  /// The URI (Uniform resource identifier)
  char *URI;
  /// If this request is a callback, the request message
  char *RequestBuffer;
  /// The length of that buffer
  int BufferLength;
  /**
   * Constructor.
   */
  XxxHTTPRequest(const char *request);
  /**
   * Destructor.
   */
  ~XxxHTTPRequest();
  /**
   * Determines if the given string (terminated with a newline) is a 
   * valid HTTP request.
   */
  static XxxBool IsHTTPRequest(const char *request);
  /**
   * Constructs a URL for use in callbacks, writing it to the supplied stream.
   */
  static XxxErrCode WriteCallbackURL(const int ClientId, 
				     const XxxShoreXactId &XactId,
				     const int ADTTypeId,//Bjarke aug. 3: int added as type
				     const XxxADTValue *ObjectRef,
				     const XxxADTMetaInfo *MetaInfo,
				     const int MessageLength,
				     const char *Message,
				     ostream &OStr);

  /**
   * Determines if this HTTP request is a callback from a Java ADT.  If
   * so, it returns XXX_TRUE and sets the parameters to their proper values.
   * All of the pointers herein modified point to data internal to this
   * object.  Consequently, it should not be deleted until they are no
   * longer going to be referenced.
   */
  XxxBool IsCallback(int &ClientId, 
		     XxxShoreXactId *&XactId,
		     int &ADTTypeId,
		     XxxADTValue *&ObjectRef,
		     XxxADTMetaInfo *&MetaInfo,
		     int &MessageLength,
		     char *&Message);
  /**
   * Writes the response header for this request on the supplied output
   * stream.
   */
  XxxErrCode WriteResponseHeader(ostream &OStr, 
				 XxxHTTPReturnCode ReturnCode,
				 const char *Message,
				 int ContentLength,
				 const char *ContentType,
				 const char *ContentEncoding = NULL) const;
};

inline XxxBool XxxHTTPRequest::IsHTTPRequest(const char *request)
{
  istrstream IStr((char *)request);
  char Method[10];
  IStr >> Method;
  if (strcmp(Method, "GET") && strcmp(Method, "POST") &&
      strcmp(Method, "HEAD"))
    return XXX_FALSE;
  return XXX_TRUE;
}

#endif
