#include "inlistexpr.h"

XxxInListExpression::
XxxInListExpression(int NChildren,
		  XxxExpression *Children[])
  : XxxBooleanExpression(NChildren, Children)
{
}

XxxInListExpression::
XxxInListExpression(const XxxInListExpression& src)
  : XxxBooleanExpression(src)
{
}

void XxxInListExpression::Print(ostream& OStr) const 
{
  OStr << "IN: (";
  for (int i = 1; i < arity; i++) {
    ChildExpr(i)->Print(OStr);
    OStr << " ";
  }
  OStr << ")";
}

/**
 * TODO: check for metainfo?
 */
XxxErrCode XxxInListExpression::TypeCheck()
{
  XxxErrCode rc;

  // First type check the child.
  if ((rc = XxxExpression::TypeCheck()) != XXX_OK) {
    return rc;
  }

  if (arity < 2) {
      XXX_NOTE_ERROR(XXX_BADTYPE, "Missing argments for InList Op!");
      return XXX_NOT_OK;
  }

  // Check that every expression has the same type.
  BaseType = ChildExpr(0)->BasicType();
  for (int i = 1; i < arity; i++) {
    if (ChildExpr(i)->BasicType() != BaseType) {
      XXX_NOTE_ERROR(XXX_BADTYPE, "All expressions must have the same type!");
      return XXX_NOT_OK;
    }
  }
  return XXX_OK;
}

XxxErrCode XxxInListExpression::Optimize(XxxValueExprPlan*& ExprPlan)
{

 XxxValueExprPlan** TmpPlans = new XxxValueExprPlan*[Arity()];
 for (int i = 0; i < arity; i++)
   XXX_CALL_RET(((XxxValueExpression *)(ChildExpr(i)))->Optimize(TmpPlans[i]),
	      XXX_BADEXEC, "Unable to optimize case expression");

 InListPlan *ILPlan = new InListPlan(this, TmpPlans);
 ExprPlan = ILPlan;
 return XXX_OK;
} 

XxxInListExpression::
InListPlan::InListPlan(XxxInListExpression* CExpr,
		   XxxValueExprPlan* ChildrenPlan[])
  : XxxBooleanExpressionPlan(CExpr, ChildrenPlan) 
{ 
  TargetType = SubPlans[0]->BasicType();
  TargetMetaInfo = SubPlans[0]->GenerateCatInfo();
  TargetVal = (XxxADTValue *)(new char[XxxADTArray[TargetType]->MaxObjectSize(TargetMetaInfo)]);
  ChoiceVal = (XxxADTValue *)(new char[XxxADTArray[TargetType]->MaxObjectSize(TargetMetaInfo)]);

  for (int i = 0; i < arity; i++) {
    getPlanUsage().Add(SubPlans[i]->GetPlanUsage());
  }
}

XxxInListExpression::
InListPlan::InListPlan(const InListPlan &Plan)
  : XxxBooleanExpressionPlan(Plan)
{
  TargetType = Plan.TargetType;
  if (!Plan.TargetMetaInfo ||
      (Plan.TargetMetaInfo->Replicate((XxxADTMetaInfo *&)TargetMetaInfo) != XXX_OK
)) {
    TargetMetaInfo = NULL;
  }

  TargetVal = (XxxADTValue *)(new char[XxxADTArray[TargetType]->MaxObjectSize(TargetMetaInfo)]);
  ChoiceVal = (XxxADTValue *)(new char[XxxADTArray[TargetType]->MaxObjectSize(TargetMetaInfo)]);
}


//Bjarke aug. 3: XxxErrCode added as return type
static XxxErrCode ExprEval(const XxxADTValue* ArgL,
                     XxxBasicType TypeL,
                     const XxxADTMetaInfo* MetaInfoL,
                     const XxxADTValue* ArgR,
                     XxxBasicType TypeR,
                     const XxxADTMetaInfo* MetaInfoR,
                     XxxBool& ExprResult)
{
  // check if either input is a NULL
  if (ArgL->Is(XxxADTValue::NULL_VALUE) ||
      ArgR->Is(XxxADTValue::NULL_VALUE)) {
    ExprResult = XXX_FALSE;
    return XXX_OK;
  }

  XxxCoreADTClass::IntValue *IntArg[2];
  XxxCoreADTClass::DblValue *DoubleArg[2];
  XxxCoreADTClass::StrValue *StringArg[2];
  IntArg[0] = (XxxCoreADTClass::IntValue *)ArgL;
  IntArg[1] = (XxxCoreADTClass::IntValue *)ArgR;
  DoubleArg[0] = (XxxCoreADTClass::DblValue *)ArgL;
  DoubleArg[1] = (XxxCoreADTClass::DblValue *)ArgR;
  StringArg[0] = (XxxCoreADTClass::StrValue *)ArgL;
  StringArg[1] = (XxxCoreADTClass::StrValue *)ArgR;

  if (XxxCoreADTClass::IsCoreADT(TypeL)) {
    ExprResult = XxxCoreADTClass::
      SEquals(TypeL, ArgL, MetaInfoL, ArgR, MetaInfoR);
  }
  else {
    ExprResult = XxxADTArray[TypeL]->
      Equals(ArgL, MetaInfoL, ArgR, MetaInfoR);
  }

  return XXX_OK;
}


XxxErrCode XxxInListExpression::
InListPlan::Evaluate(const XxxRecord* Records[],
		   const XxxRecordSchema* Schemas[],
		   XxxADTValue *Result,
		   const XxxValueEnv* Env) const 
{
  XxxCoreADTClass::BoolValue* Answer = (XxxCoreADTClass::BoolValue*)Result;
  TargetVal->Init();
  XXX_CALL_RET(SubPlans[0]->Evaluate(Records, Schemas, TargetVal, Env),
            XXX_BADEXPR, "Error evaluating in list subexpression");

  // Step through list expressions until one same as the target
  // expression found.
  for (int i = 1; i < arity; i++) {
    // Reset the null bit for each choice expression.
    ChoiceVal->Init();
    XXX_CALL_RET(SubPlans[i]->Evaluate(Records, Schemas, ChoiceVal, Env),
              XXX_BADEXPR, "Error evaluating in list subexpression");

    // check if choice expression equal tartet expression.
    XXX_CALL_RET(ExprEval(TargetVal, TargetType, 0,
                          ChoiceVal, TargetType, 0,
                          Answer->Value),
            XXX_BADEXPR, "Error evaluating in list subexpression");
    if (Answer->Value == XXX_TRUE)
      return XXX_OK;
  }

  // No matching expression.  Return false.
  Answer->Value = XXX_FALSE;
  return XXX_OK;
}

XxxBool XxxInListExpression::
Match(const XxxExpression *OtherExpr, char* Env) const
{
  return XXX_FALSE;
}
