/* Memory Mapping

 * Han Wang

 */

#include "memmap.h"
#ifdef PRED_LINUX
#include <fcntl.h>
#endif

XxxMemoryMap::XxxMemoryMap(const char* FileName)

{
  valid = XXX_FALSE;
  memHandle = NULL;

#ifdef XXX_WIN32

  // 1) create a file handle
  fileHandle = CreateFile(FileName, GENERIC_READ, FILE_SHARE_READ, 
			  NULL, OPEN_EXISTING, 0, NULL);
  if (fileHandle == INVALID_HANDLE_VALUE) {
    XXX_NOTE_ERROR(XXX_BADFILE, FileName);
    return;
  }

  fileSize = (int)GetFileSize(fileHandle, NULL);

  // 2) CreateFileMapping of that handle
  mapHandle = CreateFileMapping(fileHandle, NULL, PAGE_READONLY,
				0, 0, NULL);
  if ((mapHandle == INVALID_HANDLE_VALUE) || !mapHandle) {
    XXX_NOTE_ERROR(XXX_BADFILE, "Error mapping file");
    return;
  }

  if (map(mapHandle, memHandle) != XXX_OK) {
     return;
   }

#else

  fileDes = open(FileName, O_RDONLY);
  if (fileDes < 0) {
    XXX_NOTE_ERROR(XXX_BADFILE, FileName);
    return;
  }

  fileSize = lseek(fileDes, 0, SEEK_END);
  lseek(fileDes, 0, SEEK_SET);
  if (map(fileDes, fileSize, memHandle) != XXX_OK) {
     return;
   }

#endif

  valid = XXX_TRUE;
}


XxxMemoryMap::~XxxMemoryMap()
{
	if (valid) {
		if (close() != XXX_OK) {
			return;
		}
	}
	return;
}

// Kills the memory map - the old destructor
XxxErrCode XxxMemoryMap::close()
{
  if (!valid) return XXX_NOT_OK;

#ifdef XXX_WIN32
  if (unmap(memHandle) != XXX_OK) {
    return XXX_NOT_OK;
  }

  if (!CloseHandle(mapHandle)) {
    XXX_NOTE_ERROR(XXX_BADFILE, "Error closing map handle");
    return XXX_NOT_OK;
  }

  if (!CloseHandle(fileHandle)) {
    XXX_NOTE_ERROR(XXX_BADFILE, "Error closing file handle");
    return XXX_NOT_OK;
  }

#else

  if (unmap(memHandle, fileSize) != XXX_OK) {
    return XXX_NOT_OK;
  }

  if (::close(fileDes)) {
     XXX_NOTE_ERROR(XXX_BADFILE, "Error closing file handle");
     return XXX_NOT_OK;
     }

#endif

  valid = XXX_FALSE;
  return XXX_OK;
}

#ifdef _WIN32
/* WINDOWS NT Implementation */



XxxErrCode XxxMemoryMap::map(HANDLE FileDes, 
			     void *& MemHandle)
{
  MemHandle = (void *)MapViewOfFile(FileDes, FILE_MAP_READ,0,0,0);
  if (!MemHandle) {
    LPVOID lpMsgBuf;
    FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | 
                  FORMAT_MESSAGE_FROM_SYSTEM |
		  FORMAT_MESSAGE_MAX_WIDTH_MASK,
                  NULL,
		  GetLastError(),
		  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
		  (LPTSTR) &lpMsgBuf,
		  0,
		  NULL);
    XXX_NOTE_ERROR(XXX_BADFILE, (char *)lpMsgBuf);

    // Free the buffer.
    LocalFree( lpMsgBuf );
    return XXX_NOT_OK;
   }

  return XXX_OK;
}


XxxErrCode XxxMemoryMap::unmap(void *MemHandle)
{
  if (!UnmapViewOfFile((LPCVOID)MemHandle)) {
   XXX_NOTE_ERROR(XXX_BADFILE, "Error unmapping file");
    return XXX_NOT_OK;
   }

  return XXX_OK;
}


#else

/* UNIX Implemantation */

XxxErrCode XxxMemoryMap::map(int FileDes,
			     int FileSize, 
			     void *& MemHandle)
{
  MemHandle =(void *)mmap((caddr_t)0, FileSize, PROT_READ|PROT_EXEC,
                       MAP_PRIVATE|MAP_NORESERVE, FileDes, 0); 
  if (MemHandle == MAP_FAILED) {
    XXX_NOTE_ERROR(XXX_BADFILE, "Unable to mmap");
    return XXX_NOT_OK;
  }

  return XXX_OK;
}


XxxErrCode XxxMemoryMap::unmap(void *MemHandle, 
			       int FileSize)
{
  if (munmap((caddr_t)MemHandle, FileSize)) {
    XXX_NOTE_ERROR(XXX_BADFILE, "Unable to munmap");
    return XXX_NOT_OK;
   }

  return XXX_OK;
}

#endif
