#include "nullexpr.h"

//
// Methods of RelIsNullExpression
//

XxxIsNullExpression::XxxIsNullExpression(XxxValueExpression* Expr)
  : XxxRelBooleanExpression(XXX_OTHER_OP, 1, (XxxExpression**)(&Expr))
{
}

XxxErrCode XxxIsNullExpression::TypeCheck()
{
 XxxErrCode rc;

 if ((rc = XxxExpression::TypeCheck()) != XXX_OK) {
   return rc;
  }
 return XXX_OK;
}

XxxErrCode XxxIsNullExpression::Optimize(XxxExprPlan*& ExecPlan)
{
  XxxExprPlan* Tmp;
  if (ChildExpr(0)->Optimize(Tmp) != XXX_OK) {
    delete Tmp;
    XXX_NOTE_ERROR(XXX_BADOPT, "Error optimizing subexp");
    return XXX_NOT_OK;
  }

  IsNullPlan* RPlan = new IsNullPlan(this,
				     (XxxValueExprPlan*)Tmp);
  ExecPlan = RPlan;

  return XXX_OK;
}

void XxxIsNullExpression::Print(ostream& OStr) const 
{
  ChildExpr(0)->Print(OStr);
  OStr << " IS NULL ";
}

XxxIsNullExpression::
IsNullPlan::IsNullPlan(XxxIsNullExpression* Expr,
		       XxxValueExprPlan* ExprPlan)
  : XxxBooleanExpressionPlan(Expr, &ExprPlan) 
{
  ChildVal = NULL;
  ChildType = SubPlans[0]->BasicType();
  ChildMetaInfo = SubPlans[0]->GenerateCatInfo();
  ChildVal = (XxxADTValue *)(new char[XxxADTArray[ChildType]->MaxObjectSize(ChildMetaInfo)]);
  ChildVal->Init();
  if (SubPlans[0]) getPlanUsage().Add(SubPlans[0]->GetPlanUsage());
}

XxxIsNullExpression::
IsNullPlan::IsNullPlan(const IsNullPlan& Plan)
  : XxxBooleanExpressionPlan(Plan)
{
  ChildType = Plan.ChildType;
  if (!Plan.ChildMetaInfo ||
      (Plan.ChildMetaInfo->Replicate((XxxADTMetaInfo *&)ChildMetaInfo) != XXX_OK)) {
    ChildMetaInfo = NULL;
  }
  ChildVal = (XxxADTValue *)(new char[XxxADTArray[ChildType]->MaxObjectSize(ChildMetaInfo)]);
  ChildVal->Init();
}

XxxIsNullExpression::IsNullPlan::~IsNullPlan()
{
  if (ChildVal) delete [] ((char*)ChildVal);
}

XxxErrCode XxxIsNullExpression::
IsNullPlan::Evaluate(const XxxRecord* Rec[],
		     const XxxRecordSchema* Schema[],
		     XxxADTValue *Result,
		     const XxxValueEnv* Env) const
{
 // interpret result as a ptr to a XxxBool
 XxxCoreADTClass::BoolValue* Answer = (XxxCoreADTClass::BoolValue *)Result;

 Answer->Value = XXX_FALSE;

 XXX_CALL_RET(SubPlans[0]->Evaluate(Rec, Schema, ChildVal, Env),
	      XXX_BADEXPR, "Error evaluating subexpression");
 
 if (ChildVal->Is(XxxADTValue::NULL_VALUE))
   Answer->Value = XXX_TRUE;

 return XXX_OK;
}

