#ifndef XXX_NULLEXPR_H
#define XXX_NULLEXPR_H

#include "expr.h"
#include "exprplan.h"

/**
 * This models an IS NULL expression
 */
class XxxIsNullExpression : public XxxRelBooleanExpression {
protected:
public:
  /// Plan for exists subqueries
  class IsNullPlan : public XxxBooleanExpressionPlan { 
    friend class XxxIsNullExpression;
  protected:
    XxxADTValue *ChildVal;
    XxxBasicType ChildType;
    XxxADTMetaInfo* ChildMetaInfo;
  public:
    IsNullPlan(XxxIsNullExpression* Expr,
	       XxxValueExprPlan* ExprPlan);
    IsNullPlan(const IsNullPlan& Plan); 
    virtual ~IsNullPlan();
    virtual XxxGenericPlan * Replicate() const 
    { return new IsNullPlan(*this);}
    virtual XxxErrCode Evaluate(const XxxRecord* Rec[],
				const XxxRecordSchema* Schema[],
				XxxADTValue *Result,
				const XxxValueEnv* Env = NULL) const;
  };
  friend class IsNullPlan;

  XxxIsNullExpression(XxxValueExpression* Expr);
  /**
   * Copy constructor.  Makes a complete copy (recursing if necessary) of
   * the supplied expression.
   */
  XxxIsNullExpression(const XxxIsNullExpression& src) : 
    XxxRelBooleanExpression(src)  { }
  virtual ~XxxIsNullExpression() { }
  virtual XxxErrCode Optimize(XxxExprPlan*& ExecPlan);
  virtual XxxErrCode TypeCheck();
  /// Provides a convenient way to copy arbitrary expressions
  virtual XxxExpression * Replicate() const
    { return new XxxIsNullExpression(*this); }
  virtual void Print(ostream&) const;
  virtual XxxBool Match(const XxxExpression* Expr, 
			char *Env = NULL) const
  { return XXX_FALSE;} // never matches an index
  // TODO: actually keep stats on NULL values too!
  virtual double Selectivity(SelectivityStatsInfo* StatsArray, 
			     int ArraySize) const 
  { return 0.3;} // random choice of selectivity
};

#endif
