#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <iostream.h>
#include <strstream.h>
#include <fstream.h>

//Bjarke, aug. 6: #pragma interface/implementation not needed any longer
//#ifdef __GNUG__
//#pragma implementation "record.h"
//#endif

#include "xxx_thread.h"
#include "error.h"
#include "record.h"
#include "values.h"
#include "absquery.h"
#include "basictypes.h"

static XxxErrCode CopyNetworkOrder(XxxRecord* ToRec, const XxxRecord* FromRec,
				   const XxxRecordSchema* RecSchema)
{
 int i, RecordPos;
 int IntVal;

 // Get the record values
 for (i=0; i < RecSchema->Arity(); i++) {
   RecordPos = RecSchema->RecordOffsets[i];
   XxxADTClass* AttrClass = XxxADTArray[(*RecSchema)[i].Type];
   const XxxADTValue* FromAttrPtr = (const XxxADTValue*)(FromRec->DataArea+RecordPos);
   XxxADTValue* ToAttrPtr = (XxxADTValue*)(ToRec->DataArea+RecordPos);
   if (!AttrClass->StoreInField) {
     // convert the in-field information (offset)
     memcpy((char *)(&IntVal), (char *)(FromRec->DataArea+RecordPos), 
	    sizeof(int));
     XxxNetworkInt((int *)(ToRec->DataArea+RecordPos), IntVal);
     FromAttrPtr = (XxxADTValue*)(FromRec->DataArea+IntVal);
     ToAttrPtr = (XxxADTValue*)(ToRec->DataArea+IntVal);
   }
   *ToAttrPtr = *FromAttrPtr; // IMPORTANT: Copies header
   // now convert the actual field
   if (XxxCoreADTClass::IsCoreADT((*RecSchema)[i].Type)) {
     if (XxxCoreADTClass::SCopyHToN((*RecSchema)[i].Type, 
				    FromAttrPtr,
				    ToAttrPtr, 
				    (*RecSchema)[i].TypeCatInfo)!= XXX_OK){
        XXX_NOTE_ERROR(XXX_BADADT, "Error creating big-endian version");
        return XXX_NOT_OK;
     }
   }
   else {
     if (AttrClass->CopyHToN(FromAttrPtr, ToAttrPtr,
			     (*RecSchema)[i].TypeCatInfo) != XXX_OK) {
       XXX_NOTE_ERROR(XXX_BADADT, "Error creating big-endian version");
       return XXX_NOT_OK;
     }
   }
  } // for each field
 return XXX_OK;
}

//
// Methods of XxxAttribute
//

void XxxAttribute::Print(ostream& OStr) const
{
 XxxADTClass* AttrClass = XxxADTArray[Type];
 if (!AttrClass) {
   XXX_NOTE_ERROR(XXX_BADTYPE, "");
   return;
  }

 OStr << Name << ":" ;

 AttrClass->Print(OStr, TypeCatInfo);
}

// TODO: need to change this eventually
XxxBool XxxAttribute::FindValueAttr(const XxxQueryNode* Node, 
				    XxxUnknownValue* Value, 
				    XxxBool ConsiderThis) const
{
 if (ConsiderThis == XXX_TRUE) {
   if (Value->Name[0] != 0) { // A.colname
     if (!strcmp(Name, Value->Name))
       return XXX_TRUE;
     else
       return XXX_FALSE;
    }
   else {
     // shouldnt ever enter here!
     return XXX_FALSE;
    }
  }

 // The AttrSource check shouldbe sufficient.
 // But due to an error in the SHORE sequence catalogs (which has been fixed)
 // sometimes the AttrSource isnt correct, so we need the other check!
 if ((AttrSource < 0) ||
     (Node->Arity() <= AttrSource))
   return XXX_FALSE;
 else {
   ConsiderThis = Node->GetChild(AttrSource)->MatchNodeName(Value->XxxName);
   return (*(Node->GetChild(AttrSource)->Schema()))[AttrIndex].
     FindValueAttr(Node->GetChild(AttrSource), Value, ConsiderThis);
  }
}

//
// methods of XxxRecordSchema
//

XxxRecordSchema::XxxRecordSchema(const XxxRecordSchema& Schema) 
{
 int i;

 RecArity = Schema.Arity();
 RecordSize = Schema.RecordSize;
 RecordOffsets = new int[Arity()];
 for (i=0; i < Arity(); i++)
   RecordOffsets[i] = Schema.RecordOffsets[i];
 ExtraAreaOffset = Schema.ExtraAreaOffset;
 Attrs = new XxxAttribute[Arity()];
 for (i = 0; i < Arity(); i++) {
   Attrs[i].Assign(Schema[i]);
   Attrs[i].AttrSource = Schema[i].AttrSource;
   Attrs[i].AttrIndex = Schema[i].AttrIndex;
  }
 KeyVector = new char[Arity()];
 memcpy(KeyVector, Schema.KeyVector, Arity());
}

XxxRecordSchema::XxxRecordSchema(int NumSchemas,
				 const XxxRecordSchema* Schemas[]) 
{
 RecArity = 0;
 RecordSize = -1;
 int i;
 for (i=0; i < NumSchemas; i++) {
   RecArity += Schemas[i]->Arity();
  }

 KeyVector = new char[Arity()];
 Attrs = new XxxAttribute[Arity()];
 i = 0;
 for (int j = 0; j < NumSchemas; j++) {
   for (int k = 0; k < Schemas[j]->Arity(); k++) {
     // assign Attrs[i] = (*Schemas[j])[k]
     Attrs[i].Assign((*Schemas[j])[k]);
     Attrs[i].AttrSource = (*Schemas[j])[k].AttrSource;
     Attrs[i].AttrIndex = (*Schemas[j])[k].AttrIndex;
     i++;
    } // for each attribute
  } // for each schema

 ComputeStats();
}

// returns an attribute on success, or a NULL on failure
XxxAttribute* XxxRecordSchema::GetAttr(char *AttrName) const
{
 XxxAttribute *SearchAttr;

 for (int i = 0; i < Arity(); i++) {
   if (!strcmp((SearchAttr = GetAttr(i))->Name, AttrName)) {
     return SearchAttr;
    }
  }
 return NULL;
}

void XxxRecordSchema::Print(ostream& OStr) const
{
 OStr << "<";
 for (int i = 0; i < Arity(); i++) {
   Attrs[i].Print(OStr);
   if (i < Arity() - 1)
     OStr << ", ";
  }
 OStr << ">";
}

void XxxRecordSchema::ComputeStats()
{
 int i;
 RecordSize = 0;

 RecordOffsets = new int[Arity()];
 
 int AttrSize;

 // fill in the offsets
 for (i=0; i < Arity(); i++) {
   XxxADTClass* AttrClass = XxxADTArray[Attrs[i].Type];
   if (AttrClass->StoreInField) {
    if (XxxCoreADTClass::IsCoreADT(Attrs[i].Type)) {
      AttrSize = XxxCoreADTClass::SMaxObjectSize(Attrs[i].Type,
						 Attrs[i].TypeCatInfo);
    }
    else AttrSize = AttrClass->MaxObjectSize(Attrs[i].TypeCatInfo);
   }
   else {
     AttrSize = sizeof(XxxADTValue)+sizeof(int);
   }
   RecordOffsets[i] = RecordSize;
   RecordSize += AttrSize;
   // do padded word-alignment for every field!!
   XXX_WORD_ALIGN(RecordSize);
  } // for

 ExtraAreaOffset = RecordSize;
}

//SMostafa
// methods of class TheTimeStampVector
TheTimeStampVector::TheTimeStampVector(TheTimeStampVector* v){

  minTS.tv_sec = v->TheVec[0].tv_sec; 
  minTS.tv_usec = v->TheVec[0].tv_usec; 
  
  maxTS.tv_sec = v->TheVec[0].tv_sec; 
  maxTS.tv_usec = v->TheVec[0].tv_usec; 
  
  VecSize = v->VecSize;
  TheVec = new  struct timeval[VecSize];
  for(int i=0; i < VecSize;i++){
    TheVec[i].tv_sec = v->TheVec[i].tv_sec; 
    TheVec[i].tv_usec = v->TheVec[i].tv_usec; 
    
    if(TimeStampsGreater(TheVec[i], maxTS)){
      maxTS.tv_sec = TheVec[i].tv_sec; 
      maxTS.tv_usec = TheVec[i].tv_usec; 
    }
    if(TimeStampsGreater(minTS, TheVec[i])){
      minTS.tv_sec = TheVec[i].tv_sec; 
      minTS.tv_usec = TheVec[i].tv_usec; 
    }
  }
}

TheTimeStampVector::TheTimeStampVector(struct timeval V){
  VecSize = 1;
  TheVec = new struct timeval[VecSize];
  TheVec[0].tv_sec = V.tv_sec; 
  TheVec[0].tv_usec = V.tv_usec; 
  
  minTS.tv_sec = TheVec[0].tv_sec; 
  minTS.tv_usec = TheVec[0].tv_usec; 
  
  maxTS.tv_sec = TheVec[0].tv_sec; 
  maxTS.tv_usec = TheVec[0].tv_usec; 
  
}

TheTimeStampVector::TheTimeStampVector(TheTimeStampVector* v, struct timeval ts){

  minTS.tv_sec = v->TheVec[0].tv_sec; 
  minTS.tv_usec = v->TheVec[0].tv_usec; 
  
  maxTS.tv_sec = v->TheVec[0].tv_sec; 
  maxTS.tv_usec = v->TheVec[0].tv_usec; 
  
  VecSize = v->VecSize;
  TheVec = new  struct timeval[VecSize+1];
  for(int i=0; i < VecSize;i++){
    TheVec[i].tv_sec = v->TheVec[i].tv_sec; 
    TheVec[i].tv_usec = v->TheVec[i].tv_usec; 
    
    if(TimeStampsGreater(TheVec[i], maxTS)){
      maxTS.tv_sec = TheVec[i].tv_sec; 
      maxTS.tv_usec = TheVec[i].tv_usec; 
    }
    if(TimeStampsGreater(minTS, TheVec[i])){
      minTS.tv_sec = TheVec[i].tv_sec; 
      minTS.tv_usec = TheVec[i].tv_usec; 
    }
  }
  
  TheVec[VecSize].tv_sec = ts.tv_sec; 
  TheVec[VecSize].tv_usec = ts.tv_usec; 
  
  if(TimeStampsGreater(ts, maxTS)){
    maxTS.tv_sec = ts.tv_sec; 
    maxTS.tv_usec = ts.tv_usec; 
  }
  if(TimeStampsGreater(minTS, ts)){
    minTS.tv_sec = ts.tv_sec; 
    minTS.tv_usec = ts.tv_usec; 
  }
  
  VecSize ++;
  
}
  
TheTimeStampVector::TheTimeStampVector(int size){
    
  VecSize = size;
  TheVec= new struct timeval[VecSize];
}

TheTimeStampVector::~TheTimeStampVector(){
  if(TheVec)
    delete [] TheVec;
}

void TheTimeStampVector::SetMaximum(TheTimeStampVector* v){

 
  int change = 0;

  for(int i=0; i< v->VecSize;i++){
    if(TimeStampsGreater(v->TheVec[i], TheVec[i])){
      TheVec[i].tv_sec = v->TheVec[i].tv_sec; 
      TheVec[i].tv_usec = v->TheVec[i].tv_usec; 
      change  = 1;
    }
  }

  if(change){
    // update the max and min
    minTS.tv_sec = TheVec[0].tv_sec; 
    minTS.tv_usec = TheVec[0].tv_usec; 
    
    maxTS.tv_sec = TheVec[0].tv_sec; 
    maxTS.tv_usec = TheVec[0].tv_usec; 
    for(int i=1; i< VecSize;i++){
      if(TimeStampsGreater(TheVec[i], maxTS)){
	maxTS.tv_sec = TheVec[i].tv_sec; 
	maxTS.tv_usec = TheVec[i].tv_usec; 
      }
      if(TimeStampsGreater(minTS, TheVec[i])){
	minTS.tv_sec = TheVec[i].tv_sec; 
	minTS.tv_usec = TheVec[i].tv_usec; 
      }
    }
  }
}

XxxBool TheTimeStampVector::IsToLeftByWindowOf(TheTimeStampVector* v, int w){

  double interval;
  if(v==NULL)
	return XXX_FALSE;
  interval = TimeStampsDiff(v->minTS, maxTS);

  if(interval <= w) return XXX_FALSE;
  else
    return XXX_TRUE;
}

//EMostafa

//
// Methods of XxxRecord
//
       // This is one PRIMARY place where TypeCopy gets called.
       // (The other is in values.C where a Unknown value or Host variable
       //  gets copied into the XxxValuesArray).
       // There should be matching calls to TypeDestroy when the record
       // gets deleted, and when the XxxValuesArray entry gets overwritten.
XxxRecord::XxxRecord(const XxxValuesArray& ValArray,
		     const XxxRecordSchema* RecSchema)
{
 borrowedData = XXX_FALSE;
 size = RecSchema->GetRecordSize(ValArray);
 DataArea = new char[size];
 
 int i;
 int RecordPos = 0;
 int ExtraPos = RecSchema->Size();

 // fill the record values: special case INT, DOUBLE and STRING
 for (i=0; i < RecSchema->Arity(); i++) {
   RecordPos = RecSchema->RecordOffsets[i];
   XxxADTClass* AttrClass = XxxADTArray[(*RecSchema)[i].Type];
   if (!AttrClass || AttrClass->TypeId == XXX_UNKNOWN) {
     XXX_NOTE_ERROR(XXX_BADTYPE, "");
     return;
    }
   XxxADTValue* ToAttrPtr = (XxxADTValue*)(DataArea+RecordPos);
   int FieldLen;
   // now copy the actual value
   if (!AttrClass->StoreInField) {
     memcpy(DataArea+RecordPos, (char *)(&ExtraPos), sizeof(int));
     ToAttrPtr = (XxxADTValue*)(DataArea+ExtraPos);
   }
   // note: old code used to remove quoteson strings here
   if (XxxCoreADTClass::IsCoreADT((*RecSchema)[i].Type)) {
     FieldLen = XxxCoreADTClass::STypeCopy((*RecSchema)[i].Type,
					   ValArray.Values[i],
					   ToAttrPtr,
					   (*RecSchema)[i].TypeCatInfo);
   }
   else {
     FieldLen = AttrClass->TypeCopy(ValArray.Values[i],
				    ToAttrPtr,
				    (*RecSchema)[i].TypeCatInfo);
   }
   // now adjust the extra pos
   if (!AttrClass->StoreInField) {
     ExtraPos += FieldLen;
     XXX_WORD_ALIGN(ExtraPos);
   }
  } // for each field
 //SMostafa
 TSVec = NULL;
 //EMostafa
}

XxxBool 
XxxRecord::Equals(const XxxRecord *Rec, const XxxRecordSchema* Schema) const
{
  int i;
  for (i = 0; i < Schema->Arity(); i++) {
    // get the two ADT values to compare
    XxxADTValue *Val0, *Val1;
    //Bjarke, sep. 13. 2001: added const qualifier to variables
    const char *Data0, *Data1;
    GetField(i, Schema, Data0);
    Rec->GetField(i, Schema, Data1);
    Val0 = (XxxADTValue*)Data0;
    Val1 = (XxxADTValue*)Data1;
    if (XxxCoreADTClass::IsCoreADT((*Schema)[i].Type)) {
      if (!(XxxCoreADTClass::SEquals((*Schema)[i].Type,
				     Val0, (*Schema)[i].TypeCatInfo,
				     Val1, (*Schema)[i].TypeCatInfo)))
	return XXX_FALSE;
    }
    else {
      if (!(XxxADTArray[(*Schema)[i].Type]->
	    Equals(Val0, (*Schema)[i].TypeCatInfo,
		   Val1, (*Schema)[i].TypeCatInfo)))
	return XXX_FALSE;
    }
  } // for each attribute
  return XXX_TRUE;
}

XxxRecord* XxxRecord::Concat(const RecSchemaList& RSList, int TotalSize)
{
 // first allocate the new record
 XxxRecord* NewRec = new XxxRecord(TotalSize);
 //Bjarke, sep. 13. 2001: added const qualifier to variables
 const XxxRecord* Rec;
 const XxxRecordSchema* RecSchema;

 void *Cursor;
 // first copy the regular fields
 RSList.StartScan(Cursor);
 int RecordPos = 0;
 while (RSList.GetNext(Cursor, Rec, RecSchema) == XXX_TRUE) {
   memcpy(NewRec->DataArea+RecordPos, Rec->DataArea, RecSchema->Size());
   RecordPos += RecSchema->Size();
  }
 RSList.EndScan(Cursor);

 int TotalSchemaSize = RecordPos;

 // now copy the extra area for variable-length fields
 RSList.StartScan(Cursor);
 while (RSList.GetNext(Cursor, Rec, RecSchema) == XXX_TRUE) {
   memcpy(NewRec->DataArea+RecordPos, Rec->DataArea + RecSchema->Size(),
	  Rec->Size() - RecSchema->Size());
   RecordPos += Rec->Size() - RecSchema->Size();
  }
 RSList.EndScan(Cursor);

 int TotalRecordSize = RecordPos;

 ASSERT((TotalRecordSize == TotalSize));

 int ExtraRecSizeSoFar = 0;
 int SchemaSizeSoFar = 0;
 // now fix the offsets to variable length fields
 RecordPos = 0;
 RSList.StartScan(Cursor);
 while (RSList.GetNext(Cursor, Rec, RecSchema) == XXX_TRUE) {
   int i; int ADTOffset;
   for (i=0; i < RecSchema->Arity(); i++) {
     RecordPos = SchemaSizeSoFar + RecSchema->RecordOffsets[i];
     XxxADTClass* AttrClass = XxxADTArray[(*RecSchema)[i].Type];
     if (!AttrClass->StoreInField) {
       ADTOffset = *((int *)(NewRec->DataArea+RecordPos)) 
	 + TotalSchemaSize - RecSchema->Size()
	 + ExtraRecSizeSoFar;
       memcpy(NewRec->DataArea+RecordPos, (char *)(&ADTOffset), sizeof(int));
     }
    } // for each attribute
   ExtraRecSizeSoFar += Rec->Size() - RecSchema->Size();
   SchemaSizeSoFar += RecSchema->Size();
  } // while there are records to be concatenated

 RSList.EndScan(Cursor);

 return NewRec;
}

void XxxRecord::printBinary(ostream& OStr, const XxxRecordSchema* RecSchema) const
{
 int RecSize = Size();
 int NRecSize;
 XxxNetworkInt(&NRecSize, RecSize);
 OStr.write((char *)&NRecSize, sizeof(int));

#ifdef XXX_BIGENDIAN 
 OStr.write(DataArea, RecSize);
#else
 // otherwise, make a copy of the record with the appropriate bytes flipped.
 XxxRecord* NewRec = new XxxRecord(RecSize);
 CopyNetworkOrder(NewRec, this, RecSchema);
 OStr.write(NewRec->DataArea, RecSize);
 delete NewRec;
#endif
}

void XxxRecord::printAscii(ostream& OStr, const XxxRecordSchema* RecSchema) const
{
 int i, RecordPos;

 // Get the record values
 for (i=0; i < RecSchema->Arity(); i++) {
   XxxADTClass* AttrClass = XxxADTArray[(*RecSchema)[i].Type];
   if (!AttrClass || AttrClass->TypeId == XXX_UNKNOWN) {
     XXX_NOTE_ERROR(XXX_BADTYPE, "");
     return;
    }
   RecordPos = RecSchema->RecordOffsets[i];
   XxxADTValue* Object = (XxxADTValue*)(DataArea+RecordPos);
   if (!AttrClass->StoreInField) {
     Object = (XxxADTValue*)(DataArea + *((int *)(DataArea+RecordPos)));
   }
   if (Object->Is(XxxADTValue::NULL_VALUE)) {
     OStr << "[";
     XxxADTArray[(*RecSchema)[i].Type]->Print(OStr,
					      (*RecSchema)[i].TypeCatInfo);
     OStr << "]NULL";
   }
   else {
     if (XxxCoreADTClass::IsCoreADT((*RecSchema)[i].Type)) {
       XxxCoreADTClass::SWriteText((*RecSchema)[i].Type, OStr,
				   Object, (*RecSchema)[i].TypeCatInfo); 
     }
     else {
       AttrClass->WriteText(OStr, Object, (*RecSchema)[i].TypeCatInfo);
     }
   } // if not a NULL field
   
   OStr << "\t"; // TODO: change \t to be a settable field separator
  } // for each field
 OStr << endl;
 OStr << flush;
}

XxxErrCode XxxRecord::GetField(int i, const XxxRecordSchema *Schema,
			       const char*& FieldPtr) const
{
 int Offset;
 FieldPtr = NULL;
 if ((Offset = Schema->GetOffset(i)) < 0 ) {
   XXX_NOTE_ERROR(XXX_BADRECORD, "Bad field index");
   return XXX_NOT_OK;
  }
 XxxADTClass* AttrClass = XxxADTArray[(*Schema)[i].Type];
 if (AttrClass->StoreInField) {
   FieldPtr = DataArea+Offset;
 }
 else {
   int ActualOffset;
   memcpy(&ActualOffset, DataArea+Offset, sizeof(int));
   FieldPtr = DataArea+ActualOffset;
 }
 return XXX_OK;
}

void XxxRecord::Reclaim(const XxxRecordSchema* RecSchema)
{
 int i, RecordPos;

 // Get the record values
 for (i=0; i < RecSchema->Arity(); i++) {
   XxxADTClass* AttrClass = XxxADTArray[(*RecSchema)[i].Type];
   if (!AttrClass || AttrClass->TypeId == XXX_UNKNOWN) {
     XXX_NOTE_ERROR(XXX_BADTYPE, "");
     return;
    }
   RecordPos = RecSchema->RecordOffsets[i];
   if (XxxCoreADTClass::IsCoreADT((*RecSchema)[i].Type)) {
     // do nothing: no reclaim method implemented so far: TODO
   }
   else {
     if (AttrClass->StoreInField) {
       AttrClass->TypeDestroy((XxxADTValue*)(DataArea+RecordPos), 
			      (*RecSchema)[i].TypeCatInfo);
     }
     else {
       AttrClass->TypeDestroy((XxxADTValue*)(DataArea + *((int *)(DataArea+RecordPos))),
			      (*RecSchema)[i].TypeCatInfo);
     }
   }
  } // for each attribute

 // finally, delete the record
 delete this;
}

//
// Return the size needed to create the record.
// (Word alignment is taken into consideratino)
//
unsigned int XxxRecordSchema::GetRecordSize(const XxxValuesArray& ValArray) const
{
 int RecSize = 0;
 int ExtraRecSize = 0;

 for (int i = 0; i < Arity() ; i++) {
  XxxADTClass *AttrClass = XxxADTArray[Attrs[i].Type];
  if (AttrClass->StoreInField) {
    if (XxxCoreADTClass::IsCoreADT(Attrs[i].Type)) {
      RecSize += XxxCoreADTClass::SMaxObjectSize(Attrs[i].Type,
						 Attrs[i].TypeCatInfo);
    }
    else RecSize += AttrClass->MaxObjectSize(Attrs[i].TypeCatInfo);
  }
  else {
    RecSize += sizeof(int);
    if (XxxCoreADTClass::IsCoreADT(Attrs[i].Type)) {
      ExtraRecSize += XxxCoreADTClass::STypeSize(Attrs[i].Type,
						 ValArray.Values[i],
						 Attrs[i].TypeCatInfo);
    }
    else ExtraRecSize += 
	   AttrClass->TypeSize(ValArray.Values[i], Attrs[i].TypeCatInfo);
  }

  // do padded word-alignment for every field!!
  XXX_WORD_ALIGN(RecSize);
  XXX_WORD_ALIGN(ExtraRecSize);
 }
 return RecSize+ExtraRecSize;
}

// update col to attribute column 
XxxErrCode 
XxxRecordSchema::GetAttrCol(char *AttrName, int &col) const
{
 XxxAttribute *SearchAttr;

 for (int i = 0; i < Arity(); i++)
 {
  if (!strcmp ((SearchAttr = GetAttr(i))->Name, AttrName))
  {
   col = i;
   return XXX_OK;
  }
 }
 return XXX_NOT_OK;
}

XxxErrCode 
XxxRecordSchema::GetAttrType(char *AttrName, XxxBasicType &Type) const
{
 XxxAttribute *SearchAttr;

 for (int i = 0; i < Arity(); i++) {
  if (!strcmp ((SearchAttr = GetAttr(i))->Name, AttrName)) {
   Type = SearchAttr->Type;
   return XXX_OK;
  }
 }
 return XXX_NOT_OK;
}

XxxErrCode 
XxxRecordSchema::Matches(const XxxRecordSchema& Schema, XxxBool& Match) const
{
 Match = XXX_FALSE;
 if (Arity() == Schema.Arity()) {
   Match = XXX_TRUE;
   for (int i = 0; i < Arity(); i++) {
     if (GetAttr(i)->Type != Schema.GetAttr(i)->Type) {
       Match = XXX_FALSE;
       break;
     }
     // now check if the meta-infos match, if they exist
     XxxADTMetaInfo* MInfo1, *MInfo2;
     MInfo1 = GetAttr(i)->TypeCatInfo;
     MInfo2 = Schema.GetAttr(i)->TypeCatInfo;
     if ((!MInfo1 && MInfo2) || (MInfo1 && !MInfo2) ||
	 (MInfo1 && !MInfo1->IsCompatible(MInfo2))) {
       Match = XXX_FALSE;
       break;
     }
   } // for
 } // if arities match
 return XXX_OK;
}

//
// Methods of XxxValuesArray
//
XxxValuesArray::XxxValuesArray(const XxxValuesArray &ValArray) :
  NumValues(ValArray.NumValues)
{
  Values = new XxxADTValue*[NumValues];
  Types = new XxxBasicType[NumValues];
  ValSize = new int[NumValues];
  for (int i = 0; i < NumValues; i++) {
    Types[i] = ValArray.Types[i];
    ValSize[i] = ValArray.ValSize[i];
    Values[i] = (XxxADTValue*)(new char[ValSize[i]]);
    memcpy(Values[i], ValArray.Values[i], ValSize[i]);
  }
}

XxxValuesArray::XxxValuesArray(XxxValueExprList* ExprList)
{
 if (!ExprList) {
   NumValues = 0;
  }
 else {
   NumValues = ExprList->Count();
   Types = new XxxBasicType[NumValues];
   Values = new XxxADTValue*[NumValues];
   ValSize = new int[NumValues];

   int i;
   for (i=0; (i < NumValues) && ExprList; i++, ExprList = ExprList->Next) {
     Types[i] = ExprList->Expr->BasicType();
     if (XxxCoreADTClass::IsCoreADT(Types[i])) {
       ValSize[i] = XxxCoreADTClass::
	 SMaxObjectSize(Types[i], ExprList->Expr->GenerateCatInfo());
     }
     else ValSize[i] = 
	    XxxADTArray[Types[i]]->MaxObjectSize(ExprList->Expr->GenerateCatInfo());
     Values[i] = (XxxADTValue*)(new char[ValSize[i]]);
     Values[i]->Init();
    }
  } // else
}

XxxValuesArray::XxxValuesArray(XxxValExprPlanList* PlanList)
{
 if (!PlanList) {
   NumValues = 0;
  }
 else {
   NumValues = PlanList->Count();
   Types = new XxxBasicType[NumValues];
   Values = new XxxADTValue*[NumValues];
   ValSize = new int[NumValues];

   int i;
   for (i=0; (i < NumValues) && PlanList; i++, PlanList = PlanList->Next) {
     Types[i] = PlanList->Plan->BasicType();
     if (XxxCoreADTClass::IsCoreADT(Types[i])) {
       ValSize[i] = XxxCoreADTClass::
	 SMaxObjectSize(Types[i], PlanList->Plan->GenerateCatInfo());
     }
     else ValSize[i] = 
	    XxxADTArray[Types[i]]->MaxObjectSize(PlanList->Plan->GenerateCatInfo());
     Values[i] = (XxxADTValue*)(new char[ValSize[i]]);
     Values[i]->Init();
   }
 } // else
}

XxxValuesArray::XxxValuesArray(const XxxRecordSchema* RecSchema)
{
  if (!RecSchema) {
    NumValues  = 0;
  }
  else {
    NumValues = RecSchema->Arity();
    Types = new XxxBasicType[NumValues];
    Values = new XxxADTValue*[NumValues];
    ValSize = new int[NumValues];
    
    for (int i=0; i < NumValues; i++) {
      ASSERT(((*RecSchema)[i].Type >= 0));
      Types[i] = (*RecSchema)[i].Type;
     if (XxxCoreADTClass::IsCoreADT(Types[i])) {
       ValSize[i] = XxxCoreADTClass::
	 SMaxObjectSize(Types[i], (*RecSchema)[i].TypeCatInfo);
     }
     else ValSize[i] = 
	XxxADTArray[Types[i]]->MaxObjectSize((*RecSchema)[i].TypeCatInfo);
      Values[i] = (XxxADTValue*)(new char[ValSize[i]]);
      Values[i]->Init();
    }
  }
}

XxxValuesArray::~XxxValuesArray()
{
 int i;
 if (NumValues) {
   for (i=0; i < NumValues; i++) {
     delete [] ((char*)Values[i]);
    }
   
   // TODO:L1 : is there a memory leak here?
   delete [] Values;
   delete [] Types;
   delete [] ValSize;
  }
}


// PURPOSE: Reads record from InStream into relation
// Note: The XxxValuesArray should come pre-allocated (i.e.) it 
// should be have been created using the XxxValuesArray(RecSchema)
// constructor, so that the input arrays have been defined.
// COMMENT: This whole piece of code is ugly, and I'd like 
//  to change it eventually.
XxxErrCode 
XxxValuesArray::GetTextRecordValues(istream& InStream,
				    XxxRecordSchema* RecSchema,
				    unsigned int& RecSize, 
				    XxxBool Delimited,
				    char Delimiter)
{
 int i;

 // Get the record values
 for (i=0; i < RecSchema->Arity(); i++) {
   XxxADTClass* AttrClass = XxxADTArray[(*RecSchema)[i].Type];
   Values[i]->Init();
   // sanity check
   if (!AttrClass || AttrClass->TypeId == XXX_UNKNOWN) {
     XXX_NOTE_ERROR(XXX_BADTYPE, "");
     return XXX_NOT_OK;
    }

   // If Delimited: Determine if field contains data
   // Non-delimited mode does not allow empty fields

   char FChar;
   InStream.get(FChar);
   // Failure usually only occurs at EOF
   if (InStream.fail())
     return XXX_NOT_OK;
   if(Delimited && FChar==Delimiter) {
     // Field contains no data
     if (handleMissingField(InStream, (*RecSchema)[i], i, 
			    Delimited, Delimiter) != XXX_OK) {
       return XXX_NOT_OK;
      }
     continue; // move on to the next field
    }

   // else, there is some data in this field
   InStream.putback(FChar);

   if (XxxCoreADTClass::IsCoreADT((*RecSchema)[i].Type)) {
     ((XxxCoreADTClass*)XxxADTArray[(*RecSchema)[i].Type])
       ->SReadText((*RecSchema)[i].Type,
		   InStream, 
		   Values[i], 
		   (*RecSchema)[i].TypeCatInfo,
		   Delimited,
		   Delimiter);
   }
   else {
     // reaches here if it is an ADT
     if (AttrClass->ReadText(InStream, Values[i], 
			     (*RecSchema)[i].TypeCatInfo) != XXX_OK) {
       // Clear the error stack in case an error was noted
       XXX_CLEAR_ERROR();
       return XXX_NOT_OK;
     }
     if (Delimited)
       XxxCoreADTClass::skipRecNextField(InStream, Delimiter);
   } // switch
 } // for each attribute
 RecSize = RecSchema->GetRecordSize(*this);
 return XXX_OK;
}

XxxErrCode 
XxxValuesArray::handleMissingField(istream& InStream,
				   const XxxAttribute& Attr,
				   int i,
				   XxxBool Delimited,
				   char Delimiter)
{
 XxxADTClass* AttrClass = XxxADTArray[Attr.Type];
 if (AttrClass->DefaultVal(Values[i],
			   Attr.TypeCatInfo) != XXX_OK) {
   XXX_NOTE_ERROR(XXX_BADADT, "Couldnt assign default value");
   return XXX_NOT_OK;
  }
 return XXX_OK;
}
