#include <sys/time.h>
#include "xxx_thread.h"
#include "shoresm.h"
#include "expr.h"
#include "exprplan.h"
#include "BoundingBox.h"
#include "filestream.h"
#include "typelist.h"
#include "alias.h"
#include "syscmd.h"
#include "prd-sort.h"

#ifndef XXX_SHORE_V2
// HACK: this is a stop-gap measure for Shore 1.1
// This function should be implemented in the Shore library
// If you get a link-time conflict due to this function being
// doubly-defined
// comment this function out.
vid_t ss_m::get_vid(const lvid_t& lvid)
{
 return (io->get_vid(lvid));
}
#endif

//Bjarke, aug. 14 01: 
//get the physical volume id 
XxxShorePhysVolumeId
XxxStorageManager::GetPhysVolumeId(const XxxShoreVolumeId& VolumeId)
{
  XxxShorePhysVolumeId PhysVolumeId;
  storageManager->lvid_to_vid(VolumeId, PhysVolumeId);
  return PhysVolumeId;
}

XxxShorePhysVolumeId
XxxStorageManager::GetTempVolume(const XxxShoreVolumeId& VolumeId)
{
  //this is used as an indirection to figure out where to put
  //temporary files.  For now it just gives the Physical Volume Id of
  //the VolumeId passed.  It could, however, do some more intelligent
  //examination of available space etc.
  return GetPhysVolumeId(VolumeId);
}
//end, Bjarke's additions


// define static members
const int XxxStorageManager::MIN_NUM_BUFFERS = 100;
int FilesQueued = 0;

XxxStorageManager* XxxShoreSM = NULL;

XxxStorageManager::XxxStorageManager(int BufCount) : numBuffers(BufCount),
  DeletionQueue()
{
  XxxErrCode RC = XXX_OK;
  storageManager = NULL;  // set this as the default at the beginning
  // if storagemanager is NULL, it means that initialization is incorrect!!
  smOptions = NULL;

  // Shore expects a BufSize parameter in KB
  if (BufCount < MIN_NUM_BUFFERS)
    BufCount = MIN_NUM_BUFFERS;
  int BufSize = BufCount * (SM_PAGESIZE / 1024);
  cerr << "Using a " << BufSize << " KB buffer pool of " << BufCount
       << " pages." << endl;

  smOptions = new option_group_t(2);
  w_rc_t RetVal;

  XXX_CALL((XXX_SHORE_RETFN(smOptions->add_class_level("SHORE"), RetVal)), 
	   XXX_SHOREERR, "Failed to add class level SHORE", RC);
  XXX_BADRC_VOIDRETURN(RC);

  XXX_CALL((XXX_SHORE_RETFN(smOptions->add_class_level("XXX"), RetVal)),
	   XXX_SHOREERR, "Failed to add class level XXX", RC);
  XXX_BADRC_VOIDRETURN(RC);

  XXX_CALL((XXX_SHORE_RETFN(ss_m::setup_options(smOptions), RetVal)),
	   XXX_SHOREERR, "Failed to add SHORE options", RC);
  XXX_BADRC_VOIDRETURN(RC);

  option_t* Option;
  ostrstream err_stream;

  XXX_CALL((XXX_SHORE_RETFN(smOptions->lookup("sm_bufpoolsize", true, Option),
			    RetVal)),
	   XXX_SHOREERR, "Failed to lookup buffer size", RC);
  XXX_BADRC_VOIDRETURN(RC);
  char buf[150];
  sprintf(buf, "%d", BufSize);
  XXX_CALL((XXX_SHORE_RETFN(Option->set_value(buf, true, &err_stream), RetVal)),
	   XXX_SHOREERR, "Failed to set buffer size", RC);
  XXX_BADRC_VOIDRETURN(RC);

  XXX_CALL((XXX_SHORE_RETFN(smOptions->lookup("sm_logdir", true, Option), 
			    RetVal)),
	   XXX_SHOREERR, "Failed to lookup sm_logdir", RC);
  XXX_BADRC_VOIDRETURN(RC);	  
  // OLD: char *LogDir = getenv("XXXSHORELOGDIR");
  char *LogDir = XXX_THISENV().Options->getopt("XXXSHORELOGDIR");
  // Test to see if the log directory actually exists--if it doesn't, SHORE crashes


  if (!LogDir) {
    char tmp[100];
    sprintf(tmp, "XXXSHORELOGDIR not specified: check environment");
    // Necessary since the init thread doesn't print an error stack
    cerr << tmp << endl;
    XXX_NOTE_ERROR(XXX_SHOREERR, tmp);
    RC = XXX_NOT_OK;
    return;
  }

  if (XxxSystemCmd::TestExistDir(LogDir)) {
    cerr << endl << "Problem detected with XXXSHORELOGDIR" << LogDir << endl;
    char tmp[200];
    sprintf(tmp, "XXXSHORELOGDIR (%s) is not a valid directory.", LogDir);
    // Necessary since the init thread doesn't print an error stack
    cerr << endl << tmp << endl;
    XXX_NOTE_ERROR(XXX_SHOREERR, tmp);
    RC = XXX_NOT_OK;
    exit(1);
  }

  XXX_CALL((XXX_SHORE_RETFN(Option->set_value(LogDir, true, &err_stream),
			    RetVal)),
	   XXX_SHOREERR, "Failed to set sm_logdir", RC);
  XXX_BADRC_VOIDRETURN(RC);
  cerr << "Using log directory = " << LogDir << endl;

  XXX_CALL((XXX_SHORE_RETFN(smOptions->lookup("sm_diskrw", true, Option),
			    RetVal)),
	   XXX_SHOREERR, "Failed to lookup sm_diskrw", RC);
  XXX_BADRC_VOIDRETURN(RC);
  // OLD: char *Diskrw = getenv("XXXSHOREDISKRW");
  char *Diskrw = XXX_THISENV().Options->getopt("XXXSHOREDISKRW");

  if (!Diskrw) {
    char tmp[100];
    sprintf(tmp, "XXXSHOREDISKRW not specified: check environment");
    // Necessary since the init thread doesn't print an error stack
    cerr << tmp << endl;
    XXX_NOTE_ERROR(XXX_SHOREERR, tmp);
    RC = XXX_NOT_OK;
    return;
  }
  cerr << "Using Diskrw = " << Diskrw << endl;

  XXX_CALL((XXX_SHORE_RETFN(Option->set_value(Diskrw, true, &err_stream),
			    RetVal)),
	   XXX_SHOREERR, "Failed to set sm_diskrw", RC);
  XXX_BADRC_VOIDRETURN(RC);

  XXX_CALL((XXX_SHORE_RETFN(smOptions->lookup("sm_logging", true, Option),
			    RetVal)),
	   XXX_SHOREERR, "Failed to lookup sm_logging flag", RC);
  XXX_BADRC_VOIDRETURN(RC);
  XXX_CALL((XXX_SHORE_RETFN(Option->set_value("yes", 
					      true, &err_stream), RetVal)),
	   XXX_SHOREERR, "Failed to enable logging", RC);
  XXX_BADRC_VOIDRETURN(RC);

  XXX_CALL((XXX_SHORE_RETFN(smOptions->lookup("sm_logsize", true, Option),
			    RetVal)),
	   XXX_SHOREERR, "Failed to lookup sm_logsize flag", RC);
  XXX_BADRC_VOIDRETURN(RC);

  //Bjarke, oct. 11: reading log size from environment
  char* MaxLogSize = XXX_THISENV().Options->getopt("XXXSHOREMAXLOGSIZE");
  if (!MaxLogSize) {
    XXX_PRINT_ERROR(cerr, 
		    "XXXSHOREMAXLOGSIZE not specified: check environment");
    storageManager = NULL;
    return;
  }
  sprintf(buf, "%s", MaxLogSize); // max log size of MaxLogSize bytes
  XXX_CALL((XXX_SHORE_RETFN(Option->set_value(buf, true, &err_stream), 
			    RetVal)),
	   XXX_SHOREERR, "Failed to set max log size", RC);
  XXX_BADRC_VOIDRETURN(RC);

  /***
  XXX_CALL((XXX_SHORE_RETFN(smOptions->lookup("sm_prefetch", true, Option),
			    RetVal)),
	   XXX_SHOREERR, "Failed to lookup sm_prefetch flag", RC);
  XXX_BADRC_VOIDRETURN(RC);
  XXX_CALL((XXX_SHORE_RETFN(Option->set_value("yes", 
					      true, &err_stream), RetVal)),
	   XXX_SHOREERR, "Failed to enable prefetch", RC);
  XXX_BADRC_VOIDRETURN(RC);
  ****/

  if (smOptions->check_required(&err_stream)) {
    char tmp[500];
    sprintf (tmp, "Options not set: %s", err_stream.str());
    // Necessary since the init thread doesn't print an error stack
    cerr << tmp << endl;
    XXX_NOTE_ERROR(XXX_SHOREERR, tmp);
    storageManager = NULL;
    return;
  }
  else
    storageManager = new ss_m();

  // create a default volume
  // OLD: char* DefaultDevPath = getenv("XXXSHOREDEFDEVPATH");
  char* DefaultDevPath = XXX_THISENV().Options->getopt("XXXSHOREDEFDEVPATH");
  if (!DefaultDevPath) {
    XXX_NOTE_ERROR(XXX_SHOREERR, 
		   "XXXSHOREDEFDEVPATH not specified: check environment");
    storageManager = NULL;
    return;
  }

  // first check if the def volume exists already!!
  XxxBool Found = XXX_FALSE;
  /*****
    // No point calling DeviceExists, because it checks the mounted devices.
    // At this stage, there are _no_ mounted devices.
    // So instead, check if the named _file_ exists
    if (DeviceExists(DefaultDevPath, Found) != XXX_OK) {
    cerr << "Could not search for default device!" << endl;
    storageManager = NULL;
    return;
    }
    *****/
  ifstream InFStream;
  InFStream.open(DefaultDevPath, ios::nocreate);
  Found = (InFStream.good() ? XXX_TRUE : XXX_FALSE);
  InFStream.close();

  if (!Found) {
    // OLD: char* DefaultDevSize = getenv("XXXSHOREDEFDEVSIZE");
	char* DefaultDevSize = XXX_THISENV().Options->getopt("XXXSHOREDEFDEVSIZE");
    if (!DefaultDevSize) {
      XXX_PRINT_ERROR(cerr, 
		      "XXXSHOREDEFDEVSIZE not specified: check environment");
      storageManager = NULL;
      return;
    }
    // creating default device!
    cerr << "creating default device " << DefaultDevPath << endl;
    if (CreateDefDevice(DefaultDevPath, 
			atoi(DefaultDevSize)) != XXX_OK) {
      XXX_PRINT_ERROR(cerr, "Could not create default device");
      XXX_NOTE_ERROR(XXX_SHOREERR, 
		     "Could not create default device");
      storageManager = NULL;
      return;
    }
  } else {
    if (MountDefDevice(DefaultDevPath) != XXX_OK) {
      XXX_PRINT_ERROR(cerr, "Could not mount default device");
      XXX_NOTE_ERROR(XXX_SHOREERR, 
		     "Could not mount default device");
      storageManager = NULL;
      return;
    }
  }
}

XxxStorageManager::~XxxStorageManager()
{
  if (DeletionQueue.ForceFlush() != XXX_OK) {
    cerr << "Cannot flush deletion queue!" << endl;
  }
  if (storageManager) {
    XxxErrCode rc;
    w_rc_t RetVal;
    // dismount all devices
    XXX_CALL((XXX_SHORE_RETFN(storageManager->dismount_all(), RetVal)),
	     XXX_SHOREERR, "Cannot dismount all volumes", rc);
    
    if (smOptions) delete smOptions; // reclaim space for options
    delete storageManager;
  }
}

XxxErrCode XxxStorageManager::BeginTrans()
{
 XxxErrCode rc = XXX_OK;
 // begin a xact
 w_rc_t RetVal;
 XXX_CALL((XXX_SHORE_RETFN(storageManager->begin_xct(), RetVal)),
	  XXX_SHOREERR, "Could not start transaction", rc);
 XXX_BADRC_RCRETURN(rc);
 return rc;
}

XxxErrCode XxxStorageManager::CommitTrans()
{
 XxxErrCode rc = XXX_OK;
 // commit the transaction
 w_rc_t RetVal;
 XXX_CALL((XXX_SHORE_RETFN(storageManager->commit_xct(), RetVal)),
	  XXX_SHOREERR, "Unable to commit transaction", rc);
 XXX_BADRC_RCRETURN(rc);
 return rc;
}

XxxErrCode XxxStorageManager::AbortTrans()
{
 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;
 XXX_CALL((XXX_SHORE_RETFN(storageManager->abort_xct(), RetVal)),
	  XXX_SHOREERR, "Could not abort transaction", rc);
 XXX_BADRC_RCRETURN(rc);
 return rc;
}

XxxErrCode XxxStorageManager::CreateDevice(const char* FPath, 
					   const int DeviceSize,
					   XxxShoreVolumeId& VolId)
{
 XxxErrCode rc = XXX_OK;
 u_int VolCnt = 0;

 // create a Shore device
 w_rc_t RetVal;
 XXX_CALL((XXX_SHORE_RETFN(storageManager->format_dev(FPath, DeviceSize, true),
			   RetVal)),
	  XXX_SHOREERR, "Could not format device", rc);
 if (rc == XXX_NOT_OK) cerr << "Could not format device!" << endl;
 XXX_BADRC_RCRETURN(rc);

 devid_t TmpDevId; // device id's are not used by shore: just a dummy
 XXX_CALL((XXX_SHORE_RETFN(storageManager->mount_dev(FPath, VolCnt, TmpDevId),
			   RetVal)),
	  XXX_SHOREERR, "Could not mount device", rc);
 if (rc == XXX_NOT_OK) cerr << "Could not mount device!" << endl;
 XXX_BADRC_RCRETURN(rc);

 // create a Shore volume
 XXX_CALL((XXX_SHORE_RETFN(storageManager->generate_new_lvid(VolId), RetVal)),
	  XXX_SHOREERR, "Could not generate new logical id", rc);
 if (rc == XXX_NOT_OK) cerr << "Could not generate new logical id" << endl;
 XXX_BADRC_RCRETURN(rc);

 // added a parameter that skips zeroing of the new volume. this saves on
 // time during volume creation. only for development mode.
 XXX_CALL((XXX_SHORE_RETFN(storageManager->create_vol(FPath,VolId,DeviceSize, false),
			   RetVal)),
	  XXX_SHOREERR, "Could not generate new volume", rc);
 if (rc == XXX_NOT_OK) cerr << "Could not generate new volume" << endl;
 XXX_BADRC_RCRETURN(rc);
 
 // add logical id index
 XXX_CALL((XXX_SHORE_RETFN(storageManager->add_logical_id_index(VolId, 0, 0),
			   RetVal)),
	  XXX_SHOREERR, "Could not add logical index", rc);
 if (rc == XXX_NOT_OK) cerr << "Could not add logical index" << endl;
 XXX_BADRC_RCRETURN(rc);
	  
 // dismount the device
 //XXX_CALL((XXX_SHORE_FN(storageManager->dismount_dev(FPath))),
 //XXX_SHOREERR, "Could not dismount device", rc);
 //XXX_BADRC_RCRETURN(rc);

 return rc;
}

XxxErrCode XxxStorageManager::CreateDefDevice(const char* FPath, 
					      const int DeviceSize)
{
  if (CreateDevice(FPath, DeviceSize, defaultVolId) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot create the default device");
    return XXX_NOT_OK;
  }

  if (BeginTrans() != XXX_OK) {
    XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot start transaction");
    return XXX_NOT_OK;
  }

  // Create the default persistent file
  if (CreateFile(defaultVolId, persistentFileId, XXX_SHORE_APPEND) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot create persistent file");
    if (AbortTrans() != XXX_OK)
      XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot abort transaction");
    return XXX_NOT_OK;
  }
  // Insert a record of the file in the index
  XxxShoreKey PFKey((void *)("PFKey"), strlen("PFKey"));
  if (InsertInRootIndex(defaultVolId, PFKey, (void *)&persistentFileId,
			sizeof(XxxShoreObjectId)) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_SHOREERR, 
		   "Cannot enter record of persistent file in root index");
    if (AbortTrans() != XXX_OK)
      XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot abort transaction");
    return XXX_NOT_OK;
  }
  // Create a persistent file for type list
  if (CreateFile(defaultVolId, typeFileId, XXX_SHORE_APPEND) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot create persistent file for type list");
    if (AbortTrans() != XXX_OK)
      XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot abort transaction");
    return XXX_NOT_OK;
  }
  // Insert a record of the file in the index
  XxxShoreKey TPFKey((void *)("TPFKey"), strlen("TPFKey"));
  if (InsertInRootIndex(defaultVolId, TPFKey, (void *)&typeFileId,
			sizeof(XxxShoreObjectId)) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_SHOREERR, 
		   "Cannot enter record of persistent file for type list in root index");
    if (AbortTrans() != XXX_OK)
      XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot abort transaction");
    return XXX_NOT_OK;
  }
  if (CommitTrans() != XXX_OK) {
    XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot commit the transaction.");
    if (AbortTrans() != XXX_OK)
      XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot abort transaction");
    return XXX_NOT_OK;
  }
  return XXX_OK;
}

XxxErrCode XxxStorageManager::MountDevice(const char* FPath,
					  XxxShoreVolumeId& VolId)
{
 XxxErrCode rc = XXX_OK;
 char ErrBuf[1000];
 w_rc_t RetVal;

 u_int VolCnt;
 devid_t TmpDevId; // device id's are not used by shore: just a dummy
 XXX_CALL((XXX_SHORE_RETFN(storageManager->mount_dev(FPath, VolCnt, TmpDevId),
			   RetVal)),
	   XXX_SHOREERR, " Bad device mount", rc);
 XXX_BADRC_RCRETURN(rc);
 if (VolCnt != 1) {
   sprintf(ErrBuf, "Wierd number of volumes : %d", VolCnt);
   XXX_NOTE_ERROR(XXX_SHOREERR, ErrBuf);
   return XXX_NOT_OK;
  }

 // determine the logical vol id of this device
 XxxShoreVolumeId* lvid_list = NULL;
 u_int lvid_cnt;
 XXX_CALL((XXX_SHORE_RETFN(storageManager->list_volumes(FPath,lvid_list,
							lvid_cnt),
			   RetVal)),
	  XXX_SHOREERR, "Unable to list volumes", rc);
 XXX_BADRC_RCRETURN(rc);

 if (lvid_cnt != 1) {
   sprintf(ErrBuf, "Wierd number of volumes : %d", VolCnt);
   XXX_NOTE_ERROR(XXX_SHOREERR, ErrBuf);
   return XXX_NOT_OK;
  }
 VolId = lvid_list[0];
 delete [] lvid_list;

 return rc;
}

XxxErrCode XxxStorageManager::MountDefDevice(const char* FilePath) 
{
  if (MountDevice(FilePath, defaultVolId) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot mount default device");
    return XXX_NOT_OK;
  }
  if (BeginTrans() != XXX_OK) {
    XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot start transaction");
    return XXX_NOT_OK;
  }
  // Get a handle on the default persistent file in the device
  // Insert a record of the file in the index
  XxxShoreKey PFKey((void *)("PFKey"), strlen("PFKey"));
  bool Found = false;
  if (GetFromRootIndex(defaultVolId, PFKey, (void *)&persistentFileId,
		       sizeof(XxxShoreObjectId), Found) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_SHOREERR, 
		   "Cannot get record of persistent file in root index");
    if (AbortTrans() != XXX_OK)
      XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot abort transaction");
   return XXX_NOT_OK;
  }

  // Get a handle on the default persistent file(about data type list) in the device
  XxxShoreKey TPFKey((void *)("TPFKey"), strlen("TPFKey"));
  bool typeFound = false;
  if (GetFromRootIndex(defaultVolId, TPFKey, (void *)&typeFileId,
		       sizeof(XxxShoreObjectId), typeFound) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_SHOREERR, 
		   "Cannot get record of persistent file(type list) in root index");
    if (AbortTrans() != XXX_OK)
      XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot abort transaction");
   return XXX_NOT_OK;
  }
  if (CommitTrans() != XXX_OK) {
    XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot commit transaction");
    if (AbortTrans() != XXX_OK)
      XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot abort transaction");
    return XXX_NOT_OK;
  }
  if (!Found || !typeFound) {
    XXX_NOTE_ERROR(XXX_SHOREERR, 
		   "Cannot find record of persistent file, type file, view file in root index");
    return XXX_NOT_OK;
  }
  return XXX_OK;
}

XxxErrCode XxxStorageManager::UnMountDevice(const char* FPath)
{
 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;
 // dismount the device
 XXX_CALL((XXX_SHORE_RETFN(storageManager->dismount_dev(FPath), RetVal)),
	  XXX_SHOREERR, "Could not dismount device", rc);
 XXX_BADRC_RCRETURN(rc);

 return rc;
}

XxxErrCode XxxStorageManager::DestroyDevice(const char* FPath)
{
 XXX_NOTE_ERROR(XXX_INCOMPL, "Unable to destroy a SHORE device");
 return XXX_NOT_OK;
}

XxxErrCode XxxStorageManager::DeviceExists(const char* FPath, 
					   XxxBool& Found) const 
{
 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;

 // step thru all the devices
 const char**  DevNameList = NULL;
 devid_t*      DevIdList = NULL;
 u_int         DevCnt;

 XXX_CALL((XXX_SHORE_RETFN(storageManager->list_devices(DevNameList,
							DevIdList,
							DevCnt),
			   RetVal)),
	  XXX_SHOREERR, "Unable to list devices", rc);
 XXX_BADRC_RCRETURN(rc);

 Found = XXX_FALSE;
 // for each device
 for (unsigned int i = 0; i < DevCnt; i++) {
   if (!strcmp(DevNameList[i], FPath)) {
     Found = XXX_TRUE;
     break;
    }
  }
 if (DevIdList) delete [] DevIdList;
 if (DevNameList) delete [] DevNameList;
 return XXX_OK;
}

XxxErrCode 
XxxStorageManager::CreateFile(const XxxShoreVolumeId& VolumeId,
			      XxxShoreObjectId& File,
			      XxxStoreProperty Property)
{
 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;

 ss_m::store_property_t ShoreProp = ss_m::t_regular;
 switch (Property) {
 case XXX_SHORE_REGULAR:
   ShoreProp = ss_m::t_regular;
   break;
 case XXX_SHORE_APPEND:
   ShoreProp = ss_m::t_insert_file;
   break;
 case XXX_SHORE_LOAD:
   ShoreProp = ss_m::t_load_file;
   break;
 case XXX_SHORE_TEMP:
   XXX_NOTE_ERROR(XXX_BADEXEC, "SHORE does not permit logical temp files");
   // SHORE doesnt allow temp files with logical ids.
   // and all our files have logical ids.
   // So create it with t_loadfile.
   return XXX_NOT_OK;
   break;
 default:
   XXX_NOTE_ERROR(XXX_SHOREERR, "Bad store property");
   return XXX_NOT_OK;
 }

 XXX_CALL((XXX_SHORE_RETFN(storageManager->create_file(VolumeId, File,
						       ShoreProp),
			   RetVal)),
	  XXX_SHOREERR, "Couldnt create temp file", rc);
 XXX_BADRC_RCRETURN(rc);
 return rc;
}

XxxErrCode  
XxxStorageManager::LogicalToPhysical(const XxxShoreLogicalOID &OID,
				     XxxShorePhysicalId &PhysId)
{
  XxxErrCode rc = XXX_OK;
  w_rc_t RetVal;
  
  XXX_CALL((XXX_SHORE_RETFN(storageManager->
			    serial_to_stid(OID.VolId, OID.LocalId, PhysId),
			    RetVal)),
	   XXX_SHOREERR, "Couldn't convert logical to physical id", rc);
  XXX_BADRC_RCRETURN(rc);
  return rc;  
}

XxxErrCode 
XxxStorageManager::CreateTempFile(const XxxShoreVolumeId& VolumeId,
				  XxxShorePhysicalId& FileId)
{
 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;
 XXX_CALL((XXX_SHORE_RETFN(storageManager->
			   create_file(// Bjarke, obsolete: storageManager->get_vid(VolumeId), use GetTempVolume instead
				       GetTempVolume(VolumeId),
				       FileId,
				       ss_m::t_temporary),
			   RetVal)),
	  XXX_SHOREERR, "Couldn't create temporary file", rc);
 XXX_BADRC_RCRETURN(rc);

 return rc;
}

XxxErrCode 
XxxStorageManager::DeleteTempFile(const XxxShorePhysicalId& FileId)
{
 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;

 XXX_CALL((XXX_SHORE_RETFN(storageManager->destroy_file(FileId),
			   RetVal)),
	  XXX_SHOREERR, "Couldn't delete temporary file", rc);
 XXX_BADRC_RCRETURN(rc);
 return rc;
}

XxxErrCode
XxxStorageManager::InsertTempRecord(const XxxShorePhysicalId& FileId,
				    const XxxRecord* Record) 
{
  XxxErrCode rc = XXX_OK;
  w_rc_t RetVal;
  XxxShoreKey RecHdr;
  XxxShoreKey RecData((void *)(Record->DataArea), Record->Size());
  rid_t RecordId;
  XXX_CALL((XXX_SHORE_RETFN(storageManager->create_rec(FileId, RecHdr,
						       Record->Size(), 
						       RecData,
						       RecordId),
			    RetVal)),
	   XXX_SHOREERR, "Cannot enter record into temp file", rc);
  XXX_BADRC_RCRETURN(rc);
  
  return rc;
}

XxxErrCode 
XxxStorageManager::SortPhysicalFile (const XxxShorePhysicalId &InFileId,
				     const XxxShoreVolumeId &VolumeId,
				     XxxShorePhysicalId &OutFileId,
				     XxxComparisonFunction Func,
				     const XxxComparisonData *Data,
				     XxxBool Distinct,
				     int RunSize)
{
  XxxErrCode rc = XXX_OK;
  w_rc_t RetVal; 

#ifdef XXX_SHORE_V2
//  OLDSORT_COMPATIBILITY determines the version of sort_file used in sm.h
#undef OLDSORT_COMPATIBILITY

	XxxShoreSortKeyList	*ShoreSortKeyList;

	if (constructSortKey(Data, ShoreSortKeyList) != XXX_OK)
	{
		XXX_NOTE_ERROR (XXX_SHOREERR, "Cannot construct SHORE sort key");
		return XXX_NOT_OK;
	}
	
#ifndef XXX_WIN32
  struct timeval Time[2];
  struct timezone TZ[2];
  if (XXX_THISENV().DebugFlags.Shore) {
    // Time the sort
    gettimeofday(&Time[0], &TZ[0]);
  }
#endif

  XxxShorePhysVolumeId vidArray[1];
//Bjarke, obsolete: storageManager->get_vid(VolumeId); Use GetPhysVolumeId instead
  vidArray[0] = GetPhysVolumeId(VolumeId);

  XXX_CALL((XXX_SHORE_RETFN(storageManager->
			    sort_file(InFileId, OutFileId,
					1 /*1 volume*/, vidArray, 
					*ShoreSortKeyList,
					4 /* minimum record size: Philippe??*/,
				    RunSize /* nb of pages for a run*/, 
					RunSize*ss_m::page_sz /* nb of pages for tmp space (scratch)*/), RetVal)),
	   XXX_SHOREERR, "Couldn't sort temporary file", rc);
  XXX_BADRC_RCRETURN(rc);

#ifndef XXX_WIN32
  if (XXX_THISENV().DebugFlags.Shore) {
    gettimeofday(&Time[1], &TZ[1]);
    double ElapsedTime = (Time[1].tv_sec - Time[0].tv_sec) +
      (Time[1].tv_usec - Time[0].tv_usec)/1000000.0;
    cerr << "Sorted temporary file in " << ElapsedTime 
	 << " seconds." << endl;
  }
#endif
#else
//  OLDSORT_COMPATIBILITY determines the version of sort_file used in sm.h
#define OLDSORT_COMPATIBILITY
  XxxShoreKeyInfo ShoreKey;
  if (constructSortKeyInfo(Func, Data, ShoreKey) != XXX_OK) {
    XXX_NOTE_ERROR (XXX_SHOREERR, "Cannot construct SHORE key info");
    return XXX_NOT_OK;
  }

  XxxShorePhysicalId SortedFile;

#ifndef XXX_WIN32
  struct timeval Time[2];
  struct timezone TZ[2];
  if (XXX_THISENV().DebugFlags.Shore) {
    // Time the sort
    gettimeofday(&Time[0], &TZ[0]);
  }
#endif

  cout << "*** sort 2" << endl;

  XXX_CALL((XXX_SHORE_RETFN(storageManager->
			    sort_file(InFileId, 
				      //Bjarke, obsolete: storagemanager->get_vid(VolumeId), use GetPhysVolumeId instead
				      GetPhysVolumeId(VolumeId), 
				      SortedFile,
				      ss_m::t_temporary, ShoreKey,
				      RunSize, 
				      (Distinct ? true : false), 
				      (InFileId == OutFileId) ? true : false),
			    RetVal)),
	   XXX_SHOREERR, "Couldn't sort temporary file", rc);
  XXX_BADRC_RCRETURN(rc);
 
#ifndef XXX_WIN32
  if (XXX_THISENV().DebugFlags.Shore) {
    gettimeofday(&Time[1], &TZ[1]);
    double ElapsedTime = (Time[1].tv_sec - Time[0].tv_sec) +
      (Time[1].tv_usec - Time[0].tv_usec)/1000000.0;
    cerr << "Sorted temporary file in " << ElapsedTime 
	 << " seconds." << endl;
  }
#endif

  OutFileId = SortedFile;

#endif


  return rc;
}

XxxErrCode 
XxxStorageManager::SetStoreProperty(const XxxShoreVolumeId& VolumeId,
				    const XxxShoreObjectId& FileId,
				    XxxStoreProperty Property)
{
 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;
 ss_m::store_property_t ShoreProp = ss_m::t_regular;
 switch (Property) {
 case XXX_SHORE_APPEND:
   ShoreProp = ss_m::t_insert_file;
   break;
 case XXX_SHORE_LOAD:
   ShoreProp = ss_m::t_load_file;
   break;
 case XXX_SHORE_REGULAR:
   ShoreProp = ss_m::t_regular;
   break;
 case XXX_SHORE_TEMP:
   ShoreProp = ss_m::t_temporary;
   break;
 default:
   XXX_NOTE_ERROR(XXX_SHOREERR, "Bad store property");
   return XXX_NOT_OK;
 }

 XXX_CALL((XXX_SHORE_RETFN(storageManager->
			   set_store_property(VolumeId, FileId,
					      ShoreProp),
			   RetVal)),
	  XXX_SHOREERR, "Couldn't change store property", rc);
 XXX_BADRC_RCRETURN(rc);
 return rc;
}

XxxErrCode 
XxxStorageManager::GetStoreProperty(const XxxShoreVolumeId& VolumeId,
				    const XxxShoreObjectId& FileId,
				    XxxStoreProperty& Property)
{
 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;
 ss_m::store_property_t ShoreProp = ss_m::t_regular;
 XXX_CALL((XXX_SHORE_RETFN(storageManager->
			   get_store_property(VolumeId, FileId,
					      ShoreProp),
			   RetVal)),
	  XXX_SHOREERR, "Couldn't get store property", rc);
 XXX_BADRC_RCRETURN(rc);
 switch (ShoreProp) {
 case ss_m::t_insert_file:
   Property = XXX_SHORE_APPEND;
   break;
 case ss_m::t_regular:
   Property = XXX_SHORE_REGULAR;
   break;
 case ss_m::t_temporary:
   Property = XXX_SHORE_TEMP;
   break;
 case ss_m::t_load_file:
   Property = XXX_SHORE_LOAD;
   break;
 default:
   XXX_NOTE_ERROR(XXX_SHOREERR, "Bad store property");
   return XXX_NOT_OK;
 }
 return XXX_OK;
}


XxxErrCode 
XxxStorageManager::constructSortKeyInfo (XxxComparisonFunction Func,
					 const XxxComparisonData *Data,
					 XxxShoreKeyInfo &ShoreKey)
{
#ifndef XXX_SHORE_V2
  ShoreKey.type = sortorder::kt_custom;
  ShoreKey.len = 0;
  ShoreKey.offset = 0;
  ShoreKey.cmp = Func;
  ShoreKey.comp_data = (void *)Data;
#endif
  return XXX_OK;
}


#ifdef XXX_SHORE_V2
XxxErrCode 
XxxStorageManager::constructSortKey (const XxxComparisonData *Data,
					 XxxShoreSortKeyList *&ShoreSortKeyList)
{
	// Get nbKeys from SortInfo
	int nbKeys = Data->SortInfo->NumOrderedAttrs();

	// create ShoreSortKeyList
	ShoreSortKeyList = new sort_keys_t(nbKeys);

	// initialize ShoreSortKey - that contains nbKeys
	for (int i = 0; i < nbKeys; i++)
	{
		// initialize cookie for this key
		XxxKeyCompData *cookie = new XxxKeyCompData(Data->Schema, Data->SortInfo, i);

		// initialize Comparison Function for this key
		XxxBasicType TypeId = (*(Data->Schema))[(*(Data->SortInfo))[i].AttrIndex].Type;


		// Philippe:
		// With the new version of Shore we do not support sorting on complex ADT values 
		// (i.e. non core ADT types that involve MetaData)
		// attributes). 
		// TODO: A fix would involve extending the interface of Comparison Functions in Shore 
		// (sm/s-sort.h)to include meta data.
		
		XxxKeyComparisonFunction cf;
		XxxBool IsLexico;

		switch (TypeId)
		{
			case XXX_INT:
				cf = XxxShoreSortKeyList::int4_cmp;
				IsLexico = XXX_FALSE;
				break;
			case XXX_DOUBLE:
				cf = XxxShoreSortKeyList::f8_cmp;
				IsLexico = XXX_FALSE;
				break;
			case XXX_STRING:
				cf = XxxShoreSortKeyList::string_cmp;
				IsLexico = XXX_TRUE;
				break;
			default:
				cf = &XxxCompareFuncADT;
				IsLexico = XXX_TRUE;
				break;
		}



		ShoreSortKeyList->set_sortkey_derived(i, &XxxKeyExtractionFunction, (void*) cookie,
					'\0'/*in header?no*/, '\1'/*aligned?yes*/, IsLexico/*on several pages?no*/, cf);
		ShoreSortKeyList->set_for_file(/*deepcopy?yes*/'\1', /*keeporig?yes*/'\1', /*carry_obj?no*/'\0');

		if ((*(Data->SortInfo))[i].Asc)
		{
			ShoreSortKeyList->set_ascending('\1');
		}
		else
		{
			ShoreSortKeyList->set_ascending('\0');
		}

	}
		
  return XXX_OK;
}

#endif

XxxErrCode XxxStorageManager::SortFile (const XxxShoreVolumeId &InVolumeId,
					const XxxShoreObjectId &InFileId,
					XxxBool DestroyInputFile,
					XxxShoreVolumeId &OutVolumeId,
					XxxShoreObjectId &OutFileId,
					XxxStoreProperty NewFileProp,
					XxxComparisonFunction Func,
					const XxxComparisonData *Data,
					XxxBool Distinct,
					int RunSize)
{
  XxxErrCode rc = XXX_OK;

#ifndef XXX_SHORE_V2

  w_rc_t RetVal; 

  XxxShoreKeyInfo ShoreKey;
  if (constructSortKeyInfo(Func, Data, ShoreKey) != XXX_OK) {
    XXX_NOTE_ERROR (XXX_SHOREERR, "Cannot construct SHORE key info");
    return XXX_NOT_OK;
  }
  
  ss_m::store_property_t ShoreProp;
  switch (NewFileProp) {
  case XXX_SHORE_APPEND:
    ShoreProp = ss_m::t_insert_file;
    break;
  case XXX_SHORE_LOAD:
    ShoreProp = ss_m::t_load_file;
    break;
  case XXX_SHORE_REGULAR:
    ShoreProp = ss_m::t_regular;
    break;
  case XXX_SHORE_TEMP:
    ShoreProp = ss_m::t_temporary;
    break;
  default:
    XXX_NOTE_ERROR(XXX_SHOREERR, "Bad store property");
    return XXX_NOT_OK;
  }

  struct timeval Time[2];
  struct timezone TZ[2];
  if (XXX_THISENV().DebugFlags.Shore) {
    // Time the sort
    gettimeofday(&Time[0], &TZ[0]);
  }

  cout << "*** sort 3" << endl;
  XXX_CALL((XXX_SHORE_RETFN(storageManager->
			    sort_file(InVolumeId, InFileId,
				      OutVolumeId, OutFileId, 
				      ShoreProp, ShoreKey,
				      RunSize, 
				      (Distinct ? true : false), 
				      (DestroyInputFile ? true : false)),
			    RetVal)),
	   XXX_SHOREERR, "Couldn't sort file", rc);
  XXX_BADRC_RCRETURN(rc);


  if (XXX_THISENV().DebugFlags.Shore) {
    gettimeofday(&Time[1], &TZ[1]);
    double ElapsedTime = (Time[1].tv_sec - Time[0].tv_sec) +
      (Time[1].tv_usec - Time[0].tv_usec)/1000000.0;
    cerr << "Sorted file in " << ElapsedTime << " seconds." << endl;
  }

#endif
  return rc;
}

XxxErrCode 
XxxStorageManager::DeleteFile(const XxxShoreVolumeId& VolumeId,
			      const XxxShoreObjectId& File)
{
 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;

 XXX_CALL((XXX_SHORE_RETFN(storageManager->destroy_file(VolumeId, File),
			   RetVal)),
	  XXX_SHOREERR, "Couldnt destroy file", rc);
 XXX_BADRC_RCRETURN(rc);
 return rc;
}

// functions for root index

XxxErrCode 
XxxStorageManager::InsertInRootIndex(const XxxShoreVolumeId& VolumeId,
				     const XxxShoreKey& Key,
				     void *Datum,
				     int DatumSize)
{
 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;

 XxxShoreObjectId LogicalIndexId;
 XXX_CALL((XXX_SHORE_RETFN(storageManager->vol_root_index(VolumeId,
							  LogicalIndexId),
			   RetVal)),
	  XXX_SHOREERR, "Cannot locate root index", rc);
 XXX_BADRC_RCRETURN(rc);

 // enter it into the root b-tree index
 XxxShoreKey DataElem(Datum, DatumSize);
 XXX_CALL((XXX_SHORE_RETFN(storageManager->create_assoc(VolumeId, 
							LogicalIndexId, 
							Key, DataElem),
			   RetVal)),
	  XXX_SHOREERR, "Cannot enter file into index", rc);
 XXX_BADRC_RCRETURN(rc);

 return rc;
}

XxxErrCode
XxxStorageManager::GetFromRootIndex(const XxxShoreVolumeId& VolumeId,
				    const XxxShoreKey& Key,
				    void *DataBuf,
				    int DataBufSize,
				    bool& Found)
{
 XxxErrCode rc  = XXX_OK;
 w_rc_t RetVal;
 XxxShoreObjectId LogicalIndexId;
 XXX_CALL((XXX_SHORE_RETFN(storageManager->vol_root_index(VolumeId, 
							  LogicalIndexId),
			   RetVal)),
	  XXX_SHOREERR, "Cannot locate root index", rc);
 XXX_BADRC_RCRETURN(rc);

 smsize_t DBufSize = DataBufSize;
 XXX_CALL((XXX_SHORE_RETFN(storageManager->find_assoc(VolumeId,
						      LogicalIndexId, Key,
						      DataBuf, DBufSize,
						      Found), RetVal)),
	  XXX_SHOREERR, "Cannot find catalog file in root index", rc);
 XXX_BADRC_RCRETURN(rc);
 DataBufSize = DBufSize;

 // is a memcpy needed here for alignment bugs??

 return rc;
}

XxxErrCode
XxxStorageManager::DeleteFromRootIndex(const XxxShoreVolumeId& VolumeId,
				       const XxxShoreKey& Key,
				       int& NumRemoved)
{
 XxxErrCode rc  = XXX_OK;
 w_rc_t RetVal;
 XxxShoreObjectId LogicalIndexId;
 XXX_CALL((XXX_SHORE_RETFN(storageManager->vol_root_index(VolumeId, 
							  LogicalIndexId),
			   RetVal)),
	  XXX_SHOREERR, "Cannot locate root index", rc);
 XXX_BADRC_RCRETURN(rc);

 XXX_CALL((XXX_SHORE_RETFN(storageManager->destroy_all_assoc(VolumeId,
							     LogicalIndexId, 
							     Key,
							     NumRemoved), 
			   RetVal)),
	  XXX_SHOREERR, "Cannot delete entry in root index", rc);
 XXX_BADRC_RCRETURN(rc);

 return rc;
}

// functions for records.

XxxErrCode XxxStorageManager::GetObjectSize(const XxxShoreVolumeId& VolumeId,
					    const XxxShoreObjectId& ObjectId,
					    int &Size)
{
  XxxErrCode rc = XXX_OK;
  w_rc_t RetVal;
  pin_i Pin;
  XXX_CALL((XXX_SHORE_RETFN(Pin.pin(VolumeId, ObjectId, 0,SH), 
			    RetVal)), 
	   XXX_SHOREERR, "Couldn't pin object", rc);
  XXX_BADRC_RCRETURN (rc);
  if (!PinValid(&Pin, SH)) {
    XXX_NOTE_ERROR(XXX_SHOREERR, "Unable to pin requested object");
    return XXX_NOT_OK;
  }
  Size = Pin.body_size();
  Pin.unpin();

  return XXX_OK;
}

XxxErrCode
XxxStorageManager::InsertRecord(const XxxShoreVolumeId& VolumeId,
				const XxxShoreObjectId& FileId,
				const XxxRecord* Record,
				XxxShoreObjectId& RecordId)
{
 return InsertRecord(VolumeId, FileId, 
		     (void *)(Record->DataArea),
		     Record->Size(), RecordId);
}

XxxErrCode
XxxStorageManager::InsertRecord(const XxxShoreVolumeId& VolumeId,
				const XxxShoreObjectId& FileId,
				const void* RecordData,
				size_t RecordSize,
				XxxShoreObjectId& RecordId)
{
 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;
 XxxShoreKey  RecHdr;
 XxxShoreKey RecData((void *)(RecordData), RecordSize);

 XXX_CALL((XXX_SHORE_RETFN(storageManager->create_rec(VolumeId, FileId, RecHdr,
						      RecordSize, RecData,
						      RecordId),
			   RetVal)),
	  XXX_SHOREERR, "Cannot enter record into file", rc);
 XXX_BADRC_RCRETURN(rc);

 return rc;
}

XxxErrCode
XxxStorageManager::InsertRecordWithId(const XxxShoreVolumeId& VolumeId,
				      const XxxShoreObjectId& FileId,
				      const void* RecordData,
				      size_t RecordSize,
				      const XxxShoreObjectId& RecordId)
{
  XxxErrCode rc = XXX_OK;
  w_rc_t RetVal;
  XxxShoreKey  RecHdr;
  XxxShoreKey RecData((void *)(RecordData), RecordSize);
  
  XXX_CALL((XXX_SHORE_RETFN(storageManager->create_rec_id(VolumeId, 
							  FileId, RecHdr,
							  RecordSize, RecData,
							  RecordId),
			    RetVal)),
	   XXX_SHOREERR, "Cannot enter record into file", rc);
  XXX_BADRC_RCRETURN(rc);
  
  return rc;
}

XxxErrCode 
XxxStorageManager::DeleteRecord(const XxxShoreVolumeId& VolumeId,
				const XxxShoreObjectId& RecordId)
{
 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;

 XXX_CALL((XXX_SHORE_RETFN(storageManager->destroy_rec(VolumeId, RecordId),
			   RetVal)),
	  XXX_SHOREERR, "Cannot delete record from file", rc);
 XXX_BADRC_RCRETURN(rc);

 return rc;
}

XxxErrCode 
XxxStorageManager::GetObjectPreAlloc(const XxxShoreVolumeId& VolumeId,
				     const XxxShoreObjectId& ObjectId,
				     char* MemoryBuffer,
				     unsigned int BufferSize,
				     unsigned int StartByte,
				     unsigned int Length)
{
 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;

 pin_i Pin;

 XXX_CALL((XXX_SHORE_RETFN(Pin.pin(VolumeId, ObjectId, StartByte ,SH), 
			   RetVal)), 
	  XXX_SHOREERR, "Couldnt pin object", rc);
 XXX_BADRC_RCRETURN (rc);

 if (!PinValid(&Pin, SH)) {
   XXX_NOTE_ERROR(XXX_SHOREERR, "Unable to pin requested object");
   return XXX_NOT_OK;
  }

 rc = GetObjectPreAlloc(&Pin, MemoryBuffer, BufferSize, StartByte, Length);

 Pin.unpin();
 return rc;
}

XxxErrCode 
XxxStorageManager::GetAllocateObject(const XxxShoreVolumeId& VolumeId,
				     const XxxShoreObjectId& ObjectId,
				     char*& MemoryBuffer,
				     unsigned int& BufferSize,
				     unsigned int StartByte,
				     unsigned int Length)
{
 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;

 pin_i Pin;

 XXX_CALL((XXX_SHORE_RETFN(Pin.pin(VolumeId, ObjectId, StartByte ,SH), 
			   RetVal)), 
	  XXX_SHOREERR, "Couldnt pin object", rc);
 XXX_BADRC_RCRETURN (rc);

 if (!PinValid(&Pin, SH)) {
   XXX_NOTE_ERROR(XXX_SHOREERR, "Unable to pin requested object");
   return XXX_NOT_OK;
  }

 rc = GetAllocateObject(&Pin, MemoryBuffer, BufferSize, StartByte, Length);

 Pin.unpin();
 return rc;
}

XxxErrCode 
XxxStorageManager::GetAllocateRecord(const XxxShoreVolumeId& VolumeId,
				     const XxxShoreObjectId& ObjectId,
				     XxxRecord*& Record)
{
 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;

 pin_i Pin;

 XXX_CALL((XXX_SHORE_RETFN(Pin.pin(VolumeId, ObjectId, 0 ,SH), 
			   RetVal)), 
	  XXX_SHOREERR, "Couldnt pin object", rc);
 XXX_BADRC_RCRETURN (rc);

 if (!PinValid(&Pin, SH)) {
   XXX_NOTE_ERROR(XXX_SHOREERR, "Unable to pin requested object");
   return XXX_NOT_OK;
  }

 rc = GetAllocateRecord(&Pin, Record);

 Pin.unpin();
 return rc;
}


XxxErrCode 
XxxStorageManager::GetObjectPreAlloc(pin_i* Pin,
				     char* MemoryBuffer,
				     unsigned int BufferSize,
				     unsigned int StartByte,
				     unsigned int Length)
{  
 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;
 bool AtEnd;

 // check the validity of the Pin
 if (!PinValid(Pin, SH)) {
   char buf[100];
   PinPrint(Pin, buf);
   XXX_NOTE_ERROR(XXX_SHOREERR, buf);
   return XXX_NOT_OK;
  }
     
 if (!Length) {
   Length = Pin->body_size()-StartByte;
  }

 if (Length > BufferSize) {
   XXX_NOTE_ERROR(XXX_SHOREERR, 
		  "Bad parameters to GetObject: Length > BufSize");
   return XXX_NOT_OK;
  }

 if (StartByte+Length > Pin->body_size()) {
   XXX_NOTE_ERROR(XXX_SHOREERR, 
		  "Bad parameters to GetObject:Length too long");
   return XXX_NOT_OK;
  }

 // adjust the StartByte to reflect Pin->start_byte()
 StartByte -= Pin->start_byte();

 unsigned int BytesRead = 0;
 unsigned int CurrentByte = 0;
 while (BytesRead < Length) {
   int PinLength = Pin->length();
   if ((CurrentByte + PinLength > StartByte) &&
       (PinLength > 0)) {
     int CopyStartByte = 0;
     int CopyLength = PinLength;
     if (CurrentByte < StartByte) {
       CopyStartByte = StartByte - CurrentByte;
       CopyLength = PinLength - CopyStartByte;
      }
     if (BytesRead+CopyLength > Length) 
       CopyLength = Length-BytesRead;

     char* DataToCopy = ((char *)Pin->body()) + CopyStartByte;
     char* CopyInto = MemoryBuffer+BytesRead;
     memcpy (CopyInto, DataToCopy, CopyLength);
     
     BytesRead += CopyLength;
    } 
   CurrentByte += PinLength;
   // if we are done reading the object, quit
   if (CurrentByte >= StartByte + Length) break;

   XXX_CALL((XXX_SHORE_RETFN(Pin->next_bytes(AtEnd), RetVal)),
	    XXX_SHOREERR, "Couldnt pin next bytes", rc);
   if (rc != XXX_OK)
     return XXX_NOT_OK;
   if (AtEnd) {
     XXX_NOTE_ERROR(XXX_SHOREERR, "Pin ended too early");
     return XXX_NOT_OK;
    }
  } // while

 return rc;
}

XxxErrCode 
XxxStorageManager::GetAllocateObject(pin_i* Pin,
				     char*& MemoryBuffer,
				     unsigned int& BufferSize,
				     unsigned int StartByte,
				     unsigned int Length)
{
 BufferSize = Pin->body_size()-StartByte;
 MemoryBuffer = new char[BufferSize];

 return GetObjectPreAlloc(Pin, MemoryBuffer, 
			  BufferSize, StartByte, Length);
}

XxxErrCode 
XxxStorageManager::GetAllocateRecord(pin_i* Pin,
				     XxxRecord*& Record)
{
 Record = new XxxRecord(Pin->body_size());

 return GetObjectPreAlloc(Pin, Record->DataArea, 
			  Pin->body_size(), 0, Pin->body_size());
}

// Assumption is that the update does not change the length of the record
XxxErrCode XxxStorageManager::UpdateRecord(const XxxShoreVolumeId& VolumeId,
					   const XxxShoreObjectId& RecordId,
					   const int StartByte, 
					   const int Length,
					   const char *NewData)
{
  XxxErrCode rc = XXX_OK;
  w_rc_t RetVal;
  XxxShoreKey Replacement(NewData, Length);
  XXX_CALL((XXX_SHORE_RETFN(storageManager->update_rec(VolumeId, 
						       RecordId, 
						       StartByte,
						       Replacement),
			    RetVal)),
	   XXX_SHOREERR, "Cannot upate record", rc);
  XXX_BADRC_RCRETURN(rc);
  return XXX_OK;
}

XxxErrCode XxxStorageManager::
ReplaceRecord(const XxxShoreVolumeId& VolumeId,
	      const XxxShoreObjectId& RecordId,
	      const XxxRecord* NewRec)
{
  XxxErrCode rc = XXX_OK;
  w_rc_t RetVal;
  XxxShoreKey Replacement((void *)(NewRec->DataArea), NewRec->Size());

  pin_i RecPin;
  XXX_CALL((XXX_SHORE_RETFN(RecPin.pin(VolumeId, RecordId, 0, EX), 
			    RetVal)), 
	   XXX_SHOREERR, "Couldn't pin object", rc);
  XXX_BADRC_RCRETURN (rc);
  if (!PinValid(&RecPin, EX)) {
    XXX_NOTE_ERROR(XXX_SHOREERR, "Unable to pin requested object");
    return XXX_NOT_OK;
  }
  unsigned int Size = RecPin.body_size();
  if (Size > NewRec->Size()) {
    // needs to be truncated
    XXX_CALL((XXX_SHORE_RETFN(RecPin.truncate_rec(NewRec->Size()), RetVal)), 
	     XXX_SHOREERR, "Couldn't truncate object", rc);
    XXX_BADRC_RCRETURN (rc);
  }
  else if (Size < NewRec->Size()) {
    // needs to be expanded
    // make a dummy vec_t
    XxxShoreKey DummyData((void *)(NewRec->DataArea), NewRec->Size()-Size);
    XXX_CALL((XXX_SHORE_RETFN(RecPin.append_rec(DummyData), RetVal)), 
	     XXX_SHOREERR, "Couldn't expand object", rc);
    XXX_BADRC_RCRETURN (rc);
  }
  // make sure to repin the object since the size has changed
  XXX_CALL((XXX_SHORE_RETFN(RecPin.repin(EX), RetVal)), 
	   XXX_SHOREERR, "Couldn't repin object", rc);
  XXX_BADRC_RCRETURN (rc);

  // now update the record
  XXX_CALL((XXX_SHORE_RETFN(RecPin.update_rec(0, Replacement), RetVal)),
	   XXX_SHOREERR, "Cannot upate record", rc);
  XXX_BADRC_RCRETURN(rc);
  return XXX_OK;
}

XxxErrCode 
XxxStorageManager::WriteToOSFile(const XxxShoreVolumeId& VolumeId,
				 const XxxShoreObjectId& RecordId,
				 const char* FileName)
{
  // Open a file stream (use the sock-stream for non-blocking)
  FILE *file = fopen(FileName, "w");
  if (file == NULL) {
    XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot open stream to file");
    return XXX_NOT_OK;
  }
  XxxFileStream OStr(fileno(file));
  // Write the object out
  if (WriteToStream(VolumeId, RecordId, OStr) != XXX_OK) {
    XXX_NOTE_ERROR(XXX_BADEXEC, "Cannot write object to stream");
    return XXX_NOT_OK;
  }
  // Close the stream
  OStr.flush();
  if (!OStr.good()) {
    XXX_NOTE_ERROR(XXX_BADEXEC, "Could not close the file stream");
    return XXX_NOT_OK;
  }
  fclose(file);
  return XXX_OK;
}

XxxErrCode XxxStorageManager::WriteToStream(const XxxShoreVolumeId& VolumeId,
					    const XxxShoreObjectId& RecordId,
					    ostream &OStr)
{
  // Check the stream
  if (!OStr.good()) {
    XXX_NOTE_ERROR(XXX_BADEXEC, "The stream is not valid");
    return XXX_NOT_OK;
  }

  XxxErrCode rc = XXX_OK;
  w_rc_t RetVal;
  bool AtEnd;

  // pin the particular object
  pin_i Pin;
  XXX_CALL((XXX_SHORE_RETFN(Pin.pin(VolumeId, 
				    RecordId, 0, SH), 
			    RetVal)),
	   XXX_SHOREERR, "Couldn't pin object", rc);
  XXX_BADRC_RCRETURN(rc);
  // check the validity of the Pin
  if (!(PinValid(&Pin, SH) && (Pin.start_byte() == 0))) {
    char buf[100];
    PinPrint(&Pin, buf);
    XXX_NOTE_ERROR(XXX_SHOREERR, buf);
    return XXX_NOT_OK;
  }
  int BytesWritten = 0;
  // repeatedly read bytes of the object and write them to the stream
  while (XXX_TRUE) {
    // Check the stream
    if (!(OStr.good())) {
      XXX_NOTE_ERROR(XXX_BADEXEC, "The stream is not valid");
      return XXX_NOT_OK;
    }
    OStr.write(Pin.body(), Pin.length());
    BytesWritten += Pin.length();
    // pin next extent of file
    XXX_CALL((XXX_SHORE_RETFN(Pin.next_bytes(AtEnd), RetVal)),
	     XXX_SHOREERR, "Couldn't pin next bytes", rc);
    XXX_BADRC_RCRETURN(rc);
    if (AtEnd) {
      break;
    }
  }
  // unpin the object
  Pin.unpin();
  // flush the stream
  OStr << flush;
  if (!(OStr.good())) {
    XXX_NOTE_ERROR(XXX_BADEXEC, "The stream is not valid");
    return XXX_NOT_OK;
  }

  // cerr << " Bytes of large object written to stream : " << BytesWritten << endl;
  return XXX_OK;
}

XxxErrCode
XxxStorageManager::DiskUsage(ostream& OStr)
{
 XxxErrCode rc = XXX_OK;
 w_rc_t     RetVal;

 // step thru all the devices
 const char**  DevNameList = NULL;
 devid_t*      DevIdList = NULL;
 u_int         DevCnt;
 XXX_CALL((XXX_SHORE_RETFN(storageManager->list_devices(DevNameList,
							DevIdList,
							DevCnt),
			   RetVal)),
	  XXX_SHOREERR, "Unable to list devices", rc);
 XXX_BADRC_RCRETURN(rc);

 // for each device
 for (unsigned int i = 0; i < DevCnt; i++) {
   // print device info
#ifdef XXX_SHORE_V2
   smlevel_0::smksize_t QuotaKB, UsedKB;
#else
   smksize_t QuotaKB, UsedKB;
#endif

   XXX_CALL((XXX_SHORE_RETFN(storageManager->get_device_quota(DevNameList[i],
							      QuotaKB, UsedKB),
			     RetVal)),
	    XXX_SHOREERR, "Unable to get device quotas", rc);
   XXX_BADRC_RCRETURN(rc);
   OStr << "Device #" << DevIdList[i] << " : " << DevNameList[i];
   OStr << ", quota(KB) " << QuotaKB << ", used(KB) " << UsedKB << endl;
   

   // step thru all the volumes
   XxxShoreVolumeId* LVidList = NULL;
   u_int LVidCnt;
   XXX_CALL((XXX_SHORE_RETFN(storageManager->list_volumes(DevNameList[i],
							  LVidList,
							  LVidCnt),
			     RetVal)),
	    XXX_SHOREERR, "Unable to list volumes in device", rc);
   XXX_BADRC_RCRETURN(rc);
   // for each volume, print vol info
   for (unsigned int j = 0; j < LVidCnt; j++) {
     XXX_CALL((XXX_SHORE_RETFN(storageManager->get_volume_quota(LVidList[j],
								QuotaKB, 
								UsedKB),
			       RetVal)),
	    XXX_SHOREERR, "Unable to get volume quotas", rc);
     XXX_BADRC_RCRETURN(rc);
     OStr << "\tVolume Id" << LVidList[i] << ", quota(KB) " << QuotaKB << ", used(KB) " << UsedKB << endl;

     // Needs a transaction running!
     XxxBool TransBeganHere = XXX_FALSE;
     if (XxxShoreSM && !XXX_THISTHREAD()->xct()) {
       XxxShoreSM->BeginTrans();
       TransBeganHere = XXX_TRUE;
      }
     
     // now get the du stats for the volume
     sm_du_stats_t DUStats;
     XXX_CALL((XXX_SHORE_RETFN(storageManager->get_du_statistics(LVidList[j], 
								 DUStats,
								false),
			       RetVal)),
	    XXX_SHOREERR, "Unable to get volume du stats", rc);
     if (rc) {
       if (XxxShoreSM && TransBeganHere) {
	 XxxShoreSM->AbortTrans();
	 return XXX_NOT_OK;
	}
      }
     // print out the du stats
     OStr << "    DU Stats : " << DUStats;

     if (XxxShoreSM && TransBeganHere) {
       XxxShoreSM->CommitTrans();
      }
    } // for each volume

  } // for each device

 if (DevIdList) delete [] DevIdList;
 if (DevNameList) delete [] DevNameList;
 return rc;
}

XxxErrCode
XxxStorageManager::PrintStats(ostream& OStr) 
{
 sm_stats_info_t Stats;
 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;
 XXX_CALL((XXX_SHORE_RETFN(storageManager->gather_stats(Stats), 
			   RetVal)),
	  XXX_SHOREERR, "Couldn't gather stats", rc);
 if (rc != XXX_OK) {
   return XXX_NOT_OK;
  }
 // actually display the stats
 //w_statistics_t WStats << Stats;
 //OStr << WStats;
 OStr << Stats;

 return XXX_OK;
}

XxxErrCode
XxxStorageManager::ResetStats() 
{
 sm_stats_info_t Stats;
 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;
 // passing "true" for the second parameter resets stats
 XXX_CALL((XXX_SHORE_RETFN(storageManager->gather_stats(Stats, true),
			   RetVal)),
	  XXX_SHOREERR, "Couldn't reset stats", rc);
 if (rc != XXX_OK) {
   return XXX_NOT_OK;
  }
 return XXX_OK;
}

//
// Methods of XxxStorageManager::ObjectIterator
//
XxxStorageManager::
ObjectIterator::ObjectIterator(const XxxShoreVolumeId& VolumeId,
			       const XxxShoreObjectId& RecordId)
{
 Valid = XXX_FALSE;
 Pin = NULL;
 firstAccess = XXX_TRUE;

 XxxErrCode rc = XXX_OK;
 w_rc_t RetVal;
 
 // pin the particular object
 Pin = new pin_i;
 XXX_CALL((XXX_SHORE_RETFN(Pin->pin(VolumeId, RecordId, 0, SH), 
			   RetVal)),
	  XXX_SHOREERR, "Couldnt pin object", rc);
 if (rc != XXX_OK) {
   delete Pin;
   Pin = NULL;
   return;
  }
 
 // check the validity of the Pin
 if (!(PinValid(Pin, SH) && (Pin->start_byte() == 0))) {
   char buf[100];
#if (SHORE_INTERIM < 2)    
    const char* format = "Pin() failed to pin!: %d %d %d %d %ld %ld";
#else
    const char* format = "Pin() failed to pin!: %d %d %d %d %d %d";
#endif
    sprintf(buf, format,
	    Pin->pinned(), Pin->up_to_date(), Pin->pinned_all(),
	    Pin->is_small(), Pin->start_byte(), Pin->length());
    XXX_NOTE_ERROR(XXX_SHOREERR, buf);
    delete Pin;
    Pin = NULL;
    return;
 }

 // open the scan 
 Valid = XXX_TRUE;
}

XxxErrCode XxxStorageManager::
ObjectIterator::GetNextBytes(const char*& NextBytes, unsigned int& Length)
{
  XxxErrCode rc = XXX_OK;
  w_rc_t RetVal;
  bool AtEnd;

 if (!Valid) {
   XXX_NOTE_ERROR(XXX_SHOREERR, "Invalid object iterator");
   return XXX_NOT_OK;
  }
 if (!firstAccess) {
   // get the next bytes
   XXX_CALL((XXX_SHORE_RETFN(Pin->next_bytes(AtEnd), RetVal)),
	    XXX_SHOREERR, "Couldnt pin next bytes", rc);
   XXX_BADRC_RCRETURN(rc);
   if (AtEnd) {
     Valid = XXX_FALSE;
     return XXX_OK;
    }
  } // if !firstAccess

 firstAccess = XXX_FALSE;
 NextBytes = Pin->body();
 Length = Pin->length();
 return XXX_OK;
}

XxxStorageManager::
ObjectIterator::~ObjectIterator()
{
 if (Pin) {
   // close the scan
   Pin->unpin();
   delete Pin;
  }
}
 
XxxErrCode XxxStorageManager::XactToId (const XxxShoreXact *Xact,
					XxxShoreXactId &XactId) const
{
  XactId = storageManager->xct_to_tid(Xact);
  return XXX_OK;
}

XxxErrCode XxxStorageManager::IdToXact (const XxxShoreXactId &XactId,
					XxxShoreXact *&Xact) const
{
  Xact = storageManager->tid_to_xct(XactId);
  return XXX_OK;
}

XxxErrCode XxxBeginTrans()
{
 if (XxxShoreSM)
   if (XxxShoreSM->BeginTrans() != XXX_OK) {
     XXX_NOTE_ERROR(XXX_BADCMD, "Unable to begin transaction");
     return XXX_NOT_OK;
    }
 return XXX_OK;
}

XxxErrCode XxxCommitTrans()
{
 if (XxxShoreSM)
   if (XxxShoreSM->CommitTrans() != XXX_OK) {
     XXX_NOTE_ERROR(XXX_BADCMD, "Unable to commit transaction");
     return XXX_NOT_OK;
    }
 return XXX_OK;
}

XxxErrCode XxxAbortTrans()
{
 if (XxxShoreSM)
   if (XxxShoreSM->AbortTrans() != XXX_OK) {
     XXX_NOTE_ERROR(XXX_BADCMD, "Unable to abort transaction");
     return XXX_NOT_OK;
    }
 return XXX_OK;
}



/**************************************************************************/
// Methods of file deletion queue

ShoreFileDeletionQueue::QueueEntry::QueueEntry (XxxShoreVolumeId VolumeId, 
						XxxShoreObjectId LocalId,
						long DeletionTime,
						QueueEntry *Next,
						QueueEntry *Prev) :
  DeleteTime(DeletionTime), VolId(VolumeId), FileId(LocalId), 
  NextEntry(Next), PrevEntry(Prev)
{
  if (NextEntry != NULL)
    NextEntry->PrevEntry = this;
  if (PrevEntry != NULL)
    PrevEntry->NextEntry = this;
}

ShoreFileDeletionQueue::ShoreFileDeletionQueue()
{
  XxxUseMutex AutoMutex(Latch);
  Entries = NULL;
}

long ShoreFileDeletionQueue::GetTimeFromNow(long Delay)
{
#ifdef XXX_WIN32
  clock_t CurrentClockTicks = clock();
  return (long)(((double)CurrentClockTicks / CLOCKS_PER_SEC)*1000 + Delay);
#else
  struct timeval tv;
  struct timezone tz;
  gettimeofday(&tv, &tz);
  return ((tv.tv_sec) + ((tv.tv_usec) + Delay)/1000);
#endif
}

XxxErrCode ShoreFileDeletionQueue::EnqueueFile(XxxShoreVolumeId VolId, 
					       XxxShoreObjectId FileId,
					       long DeletionTime)
{
  XxxUseMutex AutoMutex(Latch);
  Entries = new QueueEntry(VolId, FileId, DeletionTime, Entries, NULL);
  // cerr << "Enqueuing file:" << ++FilesQueued << ": d.time " << DeletionTime << endl;
  return XXX_OK;
}

XxxErrCode ShoreFileDeletionQueue::GarbageCollect()
{  
  XxxUseMutex AutoMutex(Latch);

  if (Entries == NULL)
    return XXX_OK;

  XxxBool XctStartedHere = XXX_FALSE;
  if (!XXX_THISTHREAD()->xct())
    XctStartedHere = XXX_TRUE;
  if (XctStartedHere && (XxxBeginTrans() != XXX_OK)) {
    XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot begin transaction");
    return XXX_NOT_OK;
  }
  XxxBool ToAbort = XXX_FALSE;
  XxxErrCode rc = XXX_OK;

  QueueEntry *Tmp = Entries;
  while (Tmp != NULL) {
    if (Tmp->DeleteTime <= GetTimeFromNow(0)) {
      // cerr << "Dequeuing file:" << --FilesQueued << endl;
      if (XxxShoreSM->DeleteFile(Tmp->VolId, Tmp->FileId) != XXX_OK) {
	XXX_NOTE_ERROR(XXX_SHOREERR, "Unable to delete queued file");
	ToAbort = XXX_TRUE;
	rc = XXX_NOT_OK;
      }
      if (XXX_THISENV().DebugFlags.Shore) {
	cerr << "Temporary results in [" << Tmp->VolId 
	     << ", " << Tmp->FileId << "] have expired." << endl;
      }

      if (Tmp->PrevEntry == NULL)
	Entries = Tmp->NextEntry;
      else
	Tmp->PrevEntry->NextEntry = Tmp->NextEntry;
      if (Tmp->NextEntry != NULL)
	Tmp->NextEntry->PrevEntry = Tmp->PrevEntry;
      delete Tmp;
      Tmp = Entries;
    } else
      Tmp = Tmp->NextEntry;
  }
  if (!ToAbort && XctStartedHere && (XxxCommitTrans() != XXX_OK)) {
    XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot commit transaction");
    ToAbort = XXX_TRUE;
    rc = XXX_NOT_OK;
  }
  if (XctStartedHere && ToAbort && (XxxAbortTrans() != XXX_OK)) {
    XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot abort transaction");
  }
  return rc;
}

XxxErrCode ShoreFileDeletionQueue::ForceFlush()
{
  XxxUseMutex AutoMutex(Latch);

  XxxBool XctStartedHere = XXX_FALSE;
  if (!XXX_THISTHREAD()->xct())
    XctStartedHere = XXX_TRUE;
  if (XctStartedHere && (XxxBeginTrans() != XXX_OK)) {
    XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot begin transaction");
    return XXX_NOT_OK;
  }
  XxxBool ToAbort = XXX_FALSE;
  XxxErrCode rc = XXX_OK;

  while (Entries != NULL) {
    QueueEntry *Tmp;
    // cerr << "Dequeuing file:" << --FilesQueued << endl;
    if (XxxShoreSM->DeleteFile(Entries->VolId, Entries->FileId) != XXX_OK) {
      XXX_NOTE_ERROR(XXX_SHOREERR, "Unable to delete queued file");
      ToAbort = XXX_TRUE;
      rc = XXX_NOT_OK;
    }

    if (XXX_THISENV().DebugFlags.Shore) {
      cerr << "Temporary results in [" << Entries->VolId 
	   << ", " << Entries->FileId << "] have expired." << endl;
    }

    Tmp = Entries;
    Entries = Entries->NextEntry;
    delete Tmp;
  }
  if (!ToAbort && XctStartedHere && (XxxCommitTrans() != XXX_OK)) {
    XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot commit transaction");
    ToAbort = XXX_TRUE;
    rc = XXX_NOT_OK;
  }
  if (XctStartedHere && ToAbort && (XxxAbortTrans() != XXX_OK)) {
    XXX_NOTE_ERROR(XXX_SHOREERR, "Cannot abort transaction");
  }
  return rc;
}
