
#include "strrules.h"
#include "releng.h"

extern int RelationEngineId;

XxxBindingEnv::XxxBindingEnv(XxxValueExprList* ExprList)
{
	int i, j;
	// first determine the number of unique variables in ExprList
	XxxValueExprList* TmpList = ExprList;
    XxxUnknownValue* *ValArray = new XxxUnknownValue*[TmpList->Count()];
	int DistinctCount = 0;
	int Counter = 0;
	while (TmpList) {
		XxxUnknownValue* ThisUVal = (XxxUnknownValue*)(TmpList->Expr);
		ValArray[Counter] = NULL;
		for (i = 0; i < Counter; i++) {
			if (ValArray[i] && ThisUVal->Equals(ValArray[i]))
				break;
		}
		if (i == Counter) {
			ValArray[Counter] = ThisUVal;
			DistinctCount++;
		}
		TmpList = TmpList->Next;
		Counter++;
	} // while TmpList
	
	
	BindArray = new BindTerm[DistinctCount];
	Len = DistinctCount;
	
	for (i=0,j=0; i < Counter; i++) {
		if (ValArray[i]) {
			BindArray[j].UnknownVar = (XxxUnknownValue*)ValArray[i]->Replicate();
			BindArray[j].BindValue = NULL;
			j++;
		}
	}
	// now j should be == DistinctCount
	if (j != Len) {
		cerr << "Something needs to be checked out here" << endl;
	}
}

XxxBindingEnv::XxxBindingEnv(XxxBindingEnv& Old)
{
	int i;
	Len = Old.Len;
	BindArray = new BindTerm[Len];
	
	for(i=0; i<Len; i++) {
		BindArray[i].UnknownVar = (XxxUnknownValue *)Old.BindArray[i].UnknownVar->Replicate();
		BindArray[i].BindValue = Old.BindArray[i].BindValue->Replicate();
	}
}

XxxBindingEnv::~XxxBindingEnv()
{
	int i;
	
	for(i=0; i<Len; i++) {
		delete BindArray[i].UnknownVar;
		
		if(BindArray[i].BindValue){
			delete BindArray[i].BindValue;
		}
	}
	if(BindArray) 
		delete [] BindArray;
}

XxxBool XxxBindingEnv::Match(const char* foo, const char* funcName)
{
	if(strcmp(foo, funcName)){
		return XXX_FALSE;
	} else {
		return XXX_TRUE;
	}
}

//
XxxErrCode XxxBindingEnv::SetElement(XxxUnknownValue* Var, XxxExpression* Value)
{
	for (int i=0; i<Len; i++) {
		if (BindArray[i].UnknownVar->Equals(Var)) {
			if (BindArray[i].BindValue == NULL){
				BindArray[i].BindValue = Value->Replicate();
				return XXX_OK;
			} 
			else {
				// Check whether the old value of "Var", BindArray[i].BindValue,
				// is equal to expression "Value"
				if(BindArray[i].BindValue->Match(Value) == XXX_TRUE)
					return XXX_OK;
				else
					return XXX_NOT_OK;
			}
		}
	} // for
	// Should not reach here
	return XXX_NOT_OK;
}

XxxExpression* XxxBindingEnv::GetElement(XxxUnknownValue* Var)
{
	for (int i=0; i < Len; i++) {
		if (BindArray[i].UnknownVar->Equals(Var)) {
			return BindArray[i].BindValue;
		} 
	} // for
	
	return NULL;
}


XxxStrRule::XxxStrRule(const char* InString, const char* OutString, 
					   XxxBasicType type, const char* Name, 
					   const XxxOptRuleEngine::TravDir TD, const char* CondString)
					   : XxxRule(NULL, NULL, Name, TD)
{
	Type = type;
	InStr = new char[strlen(InString)+1];
	OutStr = new char[strlen(OutString)+1];
	
	if(CondString != NULL){
		CondStr = new char[strlen(CondString)+1];
		strcpy(CondStr, CondString);
	} else {
		CondStr = NULL;
	}
	
	strcpy(InStr, InString);
	strcpy(OutStr, OutString);
	
}

XxxStrRule::~XxxStrRule() 
{
	// Remeber to clean the things up.
	delete [] InStr;
	delete [] OutStr;
	if(CondStr != NULL){
		delete [] CondStr;
	}
} 

XxxErrCode XxxStrRule::PreCondition(XxxValueExpression* VExpr,
									XxxBool& Success,
									void*& RuleData) const
{
	Success = XXX_FALSE;
	
	XxxValueExpression *InExpr, *OutExpr;
	XxxBooleanExpression *CondExpr;
	
	// TODO: parsing should be done once and that's it -- not in here
	// First parse the three strings, better way is put it into the constructor, but
	// then, the engine array hasn't been initialized, should change later
	XxxRelationEngine * RelEngine = (XxxRelationEngine *)XxxDBEngineArray[RelationEngineId];
	
	if ((RelEngine->ParseValueExpression(InStr, InExpr) != XXX_OK) 
		|| (!InExpr)) {
		XXX_NOTE_ERROR(XXX_BADEXEC, "The format for InString is wrong");
		return XXX_NOT_OK;
	}
	
	if ((RelEngine->ParseValueExpression(OutStr, OutExpr) != XXX_OK)
		|| (!OutExpr)) {
		XXX_NOTE_ERROR(XXX_BADEXEC, "The format for OutString is wrong");
		return XXX_NOT_OK;
	}
	
	if (CondStr != NULL){
		if ((RelEngine->ParseBoolExpression(CondStr, CondExpr) != XXX_OK)
			|| (!CondExpr)) {
			XXX_NOTE_ERROR(XXX_BADEXEC, "The format for CondString is wrong");
			return XXX_NOT_OK;
		}
	}
	/* Debug
	cerr << "In Str \n";
	InExpr->Print(cerr);
	cerr << "Out Str \n";
	OutExpr->Print(cerr);
	cerr << endl;
	*/
	// Create a list of value expression referred in the InExpr
	XxxValueExprList *UnknownList, *TempList = InExpr->ValExprsReferenced();
	
	// Filter out only those expressions that refer to the unknown types
	UnknownList = TempList->TypeFilter(XxxValueExpression::COLUMN_VAR);
	
	// Create a binding environment class
	XxxBindingEnv *BindEnv = new XxxBindingEnv(UnknownList);
	
	// Check the VExpr for the match, and replace the unknown variable in the binding
	// environment by the actual value
	Success = VExpr->Match(InExpr, BindEnv);
	
	if (Success == XXX_TRUE && (CondStr != NULL)){
		// Determine whether the VExpr satisfies the condition expression
		int i;
		
		if (XXX_THISENV().DebugFlags.ADTRules) {
			cerr << "Initial Condition expression" << endl;
			CondExpr->Print(cerr);
			cerr << endl;
		}
		
		// Replace the unknown variables in the condition expression
		for (i=0; i<BindEnv->Len; i++) {
			XxxUnknownValue *Target = BindEnv->BindArray[i].UnknownVar;
			XxxExpression *Replacement = BindEnv->BindArray[i].BindValue;
			
			// Replace the unknown expression inside the OExpr
			if (CondExpr->Replace(Target, Replacement) != XXX_OK) {
				XXX_NOTE_ERROR(XXX_BADOPT, "Error replacing variables in Condition Expression");
				return XXX_NOT_OK;
			}
			
		}
		
		if (XXX_THISENV().DebugFlags.ADTRules) {
			cerr << "The condition expression with actual value" << endl;
			CondExpr->Print(cerr);
			cerr << endl;
		}
		// Type check, optimize and evaluate the condition expression(should be boolean expression
		
		if(CondExpr->TypeCheck() != XXX_OK){
			XXX_NOTE_ERROR(XXX_BADOPT, "Error Type Check for Condition Expression");
			return XXX_NOT_OK;
		}
		
		XxxExprPlan* plan;
		
		if(CondExpr->Optimize(plan) != XXX_OK){
			XXX_NOTE_ERROR(XXX_BADOPT, "Error Optimize for Condition Expression");
			return XXX_NOT_OK;
		}
		
		const XxxRecord** InputRecords=NULL;
		const XxxRecordSchema** InputSchemas = NULL;
		
		XxxCoreADTClass::BoolValue SuccessValue(XXX_FALSE);
		if(((XxxBooleanExpressionPlan *)plan)->Evaluate(InputRecords, InputSchemas, &SuccessValue, NULL) != XXX_OK){
			XXX_NOTE_ERROR(XXX_BADOPT, "Error Evaluate for Out Expression");
			return XXX_NOT_OK;
		}
		Success = SuccessValue.Value;
	}
	
	if(Success == XXX_TRUE){
		RuleDataStruct* ruleData = new RuleDataStruct(BindEnv, OutExpr);
		
		RuleData = (void *)ruleData;
	} else {
		// Clean up the binding values
		delete BindEnv;
	}
	
	return XXX_OK;  
}




XxxErrCode XxxStrRule::Action(XxxValueExpression* InVExpr, void* RuleData, 
							  XxxValueExpression*& OutVExpr) const
{
	
	int i;
	RuleDataStruct* ruleData = (RuleDataStruct* )RuleData;
	
	XxxValueExpression *OutExpr = ruleData->OutExpr;
	
	OutVExpr = (XxxValueExpression *)OutExpr->Replicate();
	XxxBindingEnv* BindEnv = ruleData->BindEnv;
	
	for(i=0; i<BindEnv->Len; i++) {
		XxxUnknownValue *Target = BindEnv->BindArray[i].UnknownVar;
		XxxExpression *Replacement = BindEnv->BindArray[i].BindValue;
		
		// Replace the unknown expression inside the OExpr
		if (OutVExpr->Replace(Target, Replacement) != XXX_OK) {
			XXX_NOTE_ERROR(XXX_BADOPT, "Error replacing variables in Out Expression");
			return XXX_NOT_OK;
		}
	}
	
	// Type check for OutVExpr
	if (OutVExpr->TypeCheck() != XXX_OK) {
		XXX_NOTE_ERROR(XXX_BADOPT, "Error Type Check for Out Expression");
		return XXX_NOT_OK;
	}
	
	if (XXX_THISENV().DebugFlags.ADTRules) {
		cerr << "\nIn Str \n";
		InVExpr->Print(cerr);
		cerr << "\nOut Str \n";
		OutVExpr->Print(cerr);
		cerr << endl;
	}
	delete ruleData->BindEnv;
	delete ruleData;
	return XXX_OK;
}
