#include "structcastexpr.h"

//---------------------------------------------------------------
XxxStructCastExpression::
XxxStructCastExpression(XxxBasicType Type, 
			XxxStructADT::StructMetaInfo *metaInfo, 
			XxxValueExprList* ExprList) : 
  XxxCastValueExpression(Type, (XxxADTMetaInfo*)metaInfo, ExprList->Expr)//(XxxExpression*)NULL)
{ 
  ChildList = new XxxValueExprList(*ExprList);
}

//---------------------------------------------------------------
XxxStructCastExpression::
XxxStructCastExpression(const  XxxStructCastExpression& src) : 
  XxxCastValueExpression(src)
{ 
  ChildList = new XxxValueExprList(*src.ChildList);
}

//---------------------------------------------------------------
void XxxStructCastExpression::Print(ostream& OStr) const
{
  OStr << '[';
  OStr << XxxADTArray[CastType]->TypeName;
  OStr << ']';
  OStr << '(';
  XxxValueExprList *Tmp = ChildList;
  while (Tmp) {
    Tmp->Expr->Print(OStr);
    if (!(Tmp=Tmp->Next))
      OStr << ',';
  }
  OStr << ')';
}
 
//---------------------------------------------------------------
XxxErrCode XxxStructCastExpression::TypeCheck()
{
  XxxErrCode rc;

  if(MetaInfo == NULL) {
    XXX_NOTE_ERROR(XXX_BADEXEC, 
		   "The Meta Info for struct expression should not be null in TypeCheck");
    return XXX_NOT_OK;
  }

  XxxRecordSchema *StructSchema = ((XxxStructADT::StructMetaInfo *)MetaInfo)->StructSchema;

  if ((rc=ChildList->TypeCheck(StructSchema))!=XXX_OK)
    return rc;
  BaseType=CastType;
  return XXX_OK;
}

//---------------------------------------------------------------
XxxErrCode XxxStructCastExpression::Optimize(XxxExprPlan*& ExecPlan)
{
  if(MetaInfo == NULL) {
    XXX_NOTE_ERROR(XXX_BADEXEC, 
		   "The Meta Info for struct expression should not be null in TypeCheck");
    return XXX_NOT_OK;
  }

  XxxRecordSchema *StructSchema = ((XxxStructADT::StructMetaInfo *)MetaInfo)->StructSchema;

  StructCastPlan *Plan = new StructCastPlan(this, StructSchema);
  XXX_CALL_RET(ChildList->Optimize(Plan->StructPlanList), 
	       XXX_BADEXEC, "Unable to optimize struct cast expression");
  ExecPlan = Plan;
  return XXX_OK;
}

//---------------------------------------------------------------
XxxBool XxxStructCastExpression::Match(const XxxExpression* OtherExpr, char *Env) const
{
  XxxStructCastExpression *CastExpr = (XxxStructCastExpression*)OtherExpr;
  if (CastType != CastExpr->CastType)
    return XXX_FALSE;
  XxxValueExprList *Tmp1 = ChildList, *Tmp2 = CastExpr->ChildList;
  while (Tmp1 && Tmp2){
    if (!Tmp1->Expr->Match(Tmp2->Expr,Env))
      return XXX_FALSE;
  }
  if (Tmp1||Tmp2)
    return XXX_FALSE;
  return XXX_TRUE;
}

//---------------------------------------------------------------
XxxBool 
XxxStructCastExpression::IsConstant() const
{
  return (ChildList->IsConstant());
}


//---------------------------------------------------------------
XxxErrCode 
XxxStructCastExpression::StructCastPlan::PreCompute()
{
  return XXX_OK;
}

//---------------------------------------------------------------
XxxErrCode 
XxxStructCastExpression::StructCastPlan::Evaluate(const XxxRecord* Rec[],
						  const XxxRecordSchema* Schema[],
						  XxxADTValue *Result,
						  const XxxValueEnv* Env) const
{
  XxxValuesArray ValArray(StructSchema);
  if (StructPlanList->Evaluate((const XxxRecord**)NULL,
			       (const XxxRecordSchema**)NULL,
			       ValArray.Values) != XXX_OK)
    return XXX_NOT_OK;
  XxxRecord *Recrd = new XxxRecord(ValArray,StructSchema);
  unsigned int size = Recrd->Size();
  memcpy(((char *)Result),(char*)(&size),sizeof(unsigned int));
  memcpy(((char *)Result)+sizeof(unsigned int),Recrd->DataArea,Recrd->Size());

  return XXX_OK;
}

//---------------------------------------------------------------
XxxErrCode XxxStructCastExpression::
StructCastPlan::RedirectDependencies(const VarRemapping& Remap,
				     int CorrelationHt)
{
  return (StructPlanList->RedirectDependencies(Remap, CorrelationHt));
}
