#ifndef XXX_STRUCTCASTEXPR_H
#define XXX_STRUCTCASTEXPR_H

#include "castexpr.h"
#include "struct.h"

/**
 *  After parsed by sql.y, expression such as [struct](...) will be parsed
 *  into XxxStructCastExpression.
 */
class XxxStructCastExpression : public XxxCastValueExpression
{
  friend class StructCastPlan;
public:
  class StructCastPlan : public XxxValueExprPlan
  {
  public:
    /// The attribute list contained in this struct object
    XxxValExprPlanList *StructPlanList;
    // The schema for this struct object
    XxxRecordSchema *StructSchema;
    // The cast expression
    XxxValueExpression * Exp;
    StructCastPlan(XxxStructCastExpression* CastExpr, XxxRecordSchema *Schema) 
		: XxxValueExprPlan(XxxValueExpression::CAST_EXPR, 0, XXX_TRUE, NULL) 
    { StructSchema = new XxxRecordSchema(*Schema);
    Exp = CastExpr; }
    StructCastPlan(const StructCastPlan &Plan) : XxxValueExprPlan(*this)
    { 
      StructPlanList = Plan.StructPlanList;
      StructSchema = new XxxRecordSchema(*Plan.StructSchema);
      Exp = Plan.Exp;
    }
    virtual ~StructCastPlan() {
      if(StructSchema) delete StructSchema;
    }
    virtual XxxErrCode PreCompute();
    virtual XxxErrCode Evaluate(const XxxRecord* Rec[],
				const XxxRecordSchema* Schema[],
				XxxADTValue *Result,
				const XxxValueEnv* Env = NULL) const;
    virtual XxxErrCode RedirectDependencies(const VarRemapping& Remap,
					    int CorrelationHt = 0);
    virtual XxxGenericPlan *Replicate() const
    { return (XxxGenericPlan *)(new StructCastPlan(*this)); }
  };
  // The attribute list contained in this struct cast expression
  XxxValueExprList* ChildList;
  XxxStructCastExpression(XxxBasicType Type, XxxStructADT::StructMetaInfo *metaInfo, XxxValueExprList* ExprList);
  XxxStructCastExpression(const XxxStructCastExpression& src);
  virtual ~XxxStructCastExpression() {
    if(ChildList) delete ChildList;
  }
  virtual void Print(ostream&) const;
  virtual XxxExpression *Replicate() const 
  { return new XxxStructCastExpression(*this); }

  virtual XxxErrCode Optimize(XxxExprPlan*& ExecPlan);
  virtual XxxErrCode TypeCheck();
  virtual XxxBool Match(const XxxExpression* OtherExpr, char *Env=NULL) const;
  virtual XxxBool IsConstant() const;
};

#endif
