#ifndef XXX_VALBOOL_H

#define XXX_VALBOOL_H

/**
 * A boolean expression wrapper for a boolean value
 */
class XxxValBooleanExpression : public XxxBooleanExpression {
  friend class XxxValBooleanPlan;
 public:
  XxxValBooleanExpression(XxxExpression* Child);
  XxxValBooleanExpression(const XxxValBooleanExpression& src);
  /**
   * Provides a convenient way to copy arbitrary expressions.  It is unsafe
   * to use copy constructors for expressions, since a derived class's
   * copy constructor may not be called for a parent class.  Thus every
   * expression has a replicate method, which provides a common way to
   * duplicate an expression.
   */
  virtual XxxExpression *Replicate() const
    { return new XxxValBooleanExpression(*this); }
  virtual void Print(ostream&) const;
  virtual XxxErrCode Optimize(XxxValueExprPlan*& ExecPlan);
  virtual XxxErrCode TypeCheck();
  virtual XxxBool Match(const XxxExpression* Expr, 
			char *Env = NULL) const;
  virtual double Selectivity(SelectivityStatsInfo* StatsArray, 
			     int ArraySize) const;
 };

/**
 * Plan class for all bool boolean expressions
 */
class XxxValBooleanPlan : public XxxBooleanExpressionPlan { 
  friend class XxxValBooleanExpression;
 public:
  XxxValBooleanPlan(XxxValBooleanExpression* BExpr,
		    XxxValueExprPlan* Child)
    : XxxBooleanExpressionPlan(BExpr, &Child) {
      // add in the subplan usage
      getPlanUsage().Add(Child->GetPlanUsage());
  }
  XxxValBooleanPlan(const XxxValBooleanPlan &Plan)
    : XxxBooleanExpressionPlan(Plan) {}
  virtual XxxGenericPlan * Replicate() const 
  { return new XxxValBooleanPlan(*this); }
  virtual ~XxxValBooleanPlan() {}
  virtual XxxErrCode Evaluate(const XxxRecord* Rec[],
			      const XxxRecordSchema* Schema[],
			      XxxADTValue *Result,
			      const XxxValueEnv* Env = NULL) const;
 };


#endif
