#ifndef XXX_VALUES_H
#define XXX_VALUES_H

#include "expr.h"
#include "exprplan.h"
#include "record.h"
#include <string.h>
#include <iostream.h>
#include <strstream.h>

XXX_EXTERN_CLASS_DECL(XxxConstValue);
XXX_EXTERN_CLASS_DECL(XxxConstIntValue);
XXX_EXTERN_CLASS_DECL(XxxConstDoubleValue);
XXX_EXTERN_CLASS_DECL(XxxConstStringValue);

typedef enum {
  XXX_PLUS_OP,
  XXX_MINUS_OP,
  XXX_TIMES_OP,
  XXX_DIV_OP,
  XXX_UMINUS_OP
   } XxxArithOpType;

/**
 * Simple arithmetic expressions.
 * @author Praveen
 */
class XxxArithExpression : public XxxValueExpression {
  friend class XxxArithPlan;
 protected:
  //// Plan for arithmetic expressions
  class ArithPlan : public XxxValueExprPlan { 
  private:
    void init(); // shared initialization for constructors
  protected:
    inline XxxErrCode PreEvaluate(const XxxRecord* Rec[],
				  const XxxRecordSchema* Schema[],
				  XxxADTValue *Result,
				  const XxxValueEnv* Env) const {
      XXX_CALL_RET(SubPlans[0]->Evaluate(Rec, Schema, ArgL, Env),
		   XXX_BADEXPR, "Error evaluating left subexpression");
      if (Arity() == 2) {
	XXX_CALL_RET(SubPlans[1]->Evaluate(Rec, Schema, ArgR, Env),
		     XXX_BADEXPR, "Error evaluating right subexpression");
      }
      return XXX_OK;
    }
  public:
    XxxArithOpType OpType;

    // TODO:DESIGN i dont think i like this. it would be nice to keep the
    // plans independent of run-time structures specific to each 
    // evaluation (this stuff should be in a separate handle)
    XxxADTValue *ArgL, *ArgR;
    XxxBasicType TypeL, TypeR;
    XxxCoreADTClass::IntValue IntArg[2];
    XxxCoreADTClass::DblValue DoubleArg[2];

    ArithPlan(XxxArithExpression* AExpr, XxxValueExprPlan* Children[]);
    ArithPlan(const ArithPlan &Plan) :
      XxxValueExprPlan(Plan), OpType(Plan.OpType)
    { init(); }
    virtual XxxGenericPlan * Replicate() const 
    { return (XxxGenericPlan *)(new ArithPlan(*this)); }    
    virtual ~ArithPlan() {}
    virtual XxxErrCode Evaluate(const XxxRecord* Rec[],
				const XxxRecordSchema* Schema[],
				XxxADTValue *Result,
				const XxxValueEnv* Env = NULL) const;
  }; // class ArithPlan
  class PlusPlan : public ArithPlan { 
  public:
    PlusPlan(XxxArithExpression* AExpr, XxxValueExprPlan* Children[])
      :ArithPlan(AExpr, Children) {}
    PlusPlan(const PlusPlan &Plan): ArithPlan(Plan) {}
    virtual XxxGenericPlan * Replicate() const 
    { return new PlusPlan(*this);}    
    virtual XxxErrCode Evaluate(const XxxRecord* Rec[],
				const XxxRecordSchema* Schema[],
				XxxADTValue *Result,
				const XxxValueEnv* Env) const;
  }; // class PlusPlan
  class MinusPlan : public ArithPlan { 
  public:
    MinusPlan(XxxArithExpression* AExpr,XxxValueExprPlan* Children[])
      :ArithPlan(AExpr, Children) {}
    MinusPlan(const MinusPlan &Plan): ArithPlan(Plan) {}
    virtual XxxGenericPlan * Replicate() const 
    { return new MinusPlan(*this);}    
    virtual XxxErrCode Evaluate(const XxxRecord* Rec[],
				const XxxRecordSchema* Schema[],
				XxxADTValue *Result,
				const XxxValueEnv* Env) const;
  }; // class MinusPlan
  class TimesPlan : public ArithPlan { 
  public:
    TimesPlan(XxxArithExpression* AExpr, XxxValueExprPlan* Children[])
      :ArithPlan(AExpr, Children) {}
    TimesPlan(const TimesPlan &Plan): ArithPlan(Plan) {}
    virtual XxxGenericPlan * Replicate() const 
    { return new TimesPlan(*this);}    
    virtual XxxErrCode Evaluate(const XxxRecord* Rec[],
				const XxxRecordSchema* Schema[],
				XxxADTValue *Result,
				const XxxValueEnv* Env) const;
  }; // class TimesPlan
  class DivPlan : public ArithPlan { 
  public:
    DivPlan(XxxArithExpression* AExpr, XxxValueExprPlan* Children[])
      :ArithPlan(AExpr, Children) {}
    DivPlan(const DivPlan &Plan): ArithPlan(Plan) {}
    virtual XxxGenericPlan * Replicate() const 
    { return new DivPlan(*this);}    
    virtual XxxErrCode Evaluate(const XxxRecord* Rec[],
				const XxxRecordSchema* Schema[],
				XxxADTValue *Result,
				const XxxValueEnv* Env) const;
  }; // class DivPlan
  class UMinusPlan : public ArithPlan { 
  public:
    UMinusPlan(XxxArithExpression* AExpr, XxxValueExprPlan* Children[])
      :ArithPlan(AExpr, Children) {}
    UMinusPlan(const UMinusPlan &Plan): ArithPlan(Plan) {}
    virtual XxxGenericPlan * Replicate() const 
    { return new UMinusPlan(*this);}    
    virtual XxxErrCode Evaluate(const XxxRecord* Rec[],
				const XxxRecordSchema* Schema[],
				XxxADTValue *Result,
				const XxxValueEnv* Env) const;
  }; // class UMinusPlan

 public:
  XxxArithOpType OpType;
  XxxArithExpression(XxxArithOpType OperatorType, 
		     int NumChildren,
		     XxxExpression* Children[]);
  /// Copy constructor.
  XxxArithExpression(const XxxArithExpression &src);
  /// Provides a convenient way to copy arbitrary expressions
  virtual XxxExpression * Replicate() const
    { return new XxxArithExpression(*this); }
  virtual ValExprType ValType() const {return ARITH_EXPR;}
  virtual void Print(ostream&) const;
  virtual XxxErrCode Optimize(XxxExprPlan*& ExecPlan);
  virtual XxxErrCode TypeCheck();
  /// Checks to see if the expression matches another expression
  virtual XxxBool Match(const XxxExpression* OtherExpr, 
			char *Env = NULL) const;
  /**
   *  Checks if the expression matches another rule input expression 
   *  If match is successful, return true, and assign BindTerm array
   *  inside XxxBindingEnv class.
   */
  virtual XxxBool Match(const XxxExpression* RuleExpr, 
			XxxBindingEnv *BindEnv) const;
};

/**
 * Represents column value expressions (like Emp.salary)
 * @author Praveen
 */
class XxxUnknownValue : public XxxValueExpression {
 protected:
  XxxADTMetaInfo *typeCatInfo;
 public:
  /**
   * Execution plan for column expressions
   */
  class UnknownValuePlan : public XxxValueExprPlan { public:
    int SrcChild;
    int SrcAttrIndex;
    int CorrelationHeight;
    UnknownValuePlan(XxxUnknownValue* UVal) 
      : XxxValueExprPlan(UVal, 0) 
    {
      SrcChild = UVal->SrcChild;
      SrcAttrIndex = UVal->SrcAttrIndex;
      CorrelationHeight = UVal->CorrelationHeight;
    }
    UnknownValuePlan(const UnknownValuePlan &Plan) :
      XxxValueExprPlan(Plan), SrcChild(Plan.SrcChild), 
      SrcAttrIndex(Plan.SrcAttrIndex), 
      CorrelationHeight(Plan.CorrelationHeight) { }
    virtual XxxGenericPlan * Replicate() const {
      return (XxxGenericPlan *)(new UnknownValuePlan(*this));
    }    
    virtual ~UnknownValuePlan() {}
    virtual XxxErrCode Evaluate(const XxxRecord* Rec[],
				const XxxRecordSchema* Schema[],
				XxxADTValue *Result,
				const XxxValueEnv* Env = NULL) const;
    virtual XxxErrCode RedirectDependencies(const VarRemapping& Remap,
					    int CorrelationHt = 0);
  };
  friend XXX_EXTERN_CLASS_DECL(UnknownValuePlan);
  /// The name of the attribute
  char Name[XXX_ATTR_NAME_MAX];
  /// The name of the query node in which it is defined
  char XxxName[XXX_NODE_NAME_MAX];
  /// The input index of the child in which this attribute is defined
  char SrcChild;
  /// The index of this attribute in the schema of that node
  int SrcAttrIndex;
  /// How many level above this did the correlation come from?
  /// 0 indicates uncorrelated
  int CorrelationHeight;

  /**
   * This constructor is called for variable names inside queries.
   * the constructor leaves the variable with incomplete info
   * The info is filled in later in the context of the whole query.
   */
  XxxUnknownValue (const char *VarName, const char *NodeName = NULL,
		   const XxxBasicType Type = XXX_UNKNOWN, 
		   const int AttrIndex = -1, const int Child = -1);
  /**
   * This constructor is called for variables specified by positions 
   * inside queries. The constructor leaves the variable with incomplete 
   * info (on the SrcChild). The info is filled in later in the context 
   * of the whole query.
   */
  XxxUnknownValue(int Index, const char *NodeName);
  /** 
   * This constructor is called when a complete variable needs to be created
   * in order to create an expression (possibly to set up a scan)
   */
  XxxUnknownValue(int Index, XxxBasicType Type, 
		  XxxADTMetaInfo *CatInfo = NULL);
  /**
   * Copy constructor.
   */
  XxxUnknownValue(const XxxUnknownValue& src);
  /// Provides a convenient way to copy arbitrary expressions
  virtual XxxExpression * Replicate() const
    { return new XxxUnknownValue(*this); }

  /// Check whether these two  refer to the same unknown expr. Binquan
  virtual XxxBool Equals(const XxxUnknownValue *Var);

  virtual ValExprType ValType() const {return COLUMN_VAR;}
  virtual void Print(ostream&) const;
  /**
   * An unknown value is <B>NOT</B> a constant
   */
  virtual XxxBool IsConstant() const { return XXX_FALSE; }

  /**
   * Substitutes NewAttrIndex for OldAttrIndex in the SrcAttrIndex field 
   * of XxxUnknownValues subexpressions with SrcChild == ChildNumber
   */
  virtual XxxErrCode UpdateUnknowns(int ChildNumber,
 				    int OldAttrIndex,
				    int NewAttrIndex,
				    XxxBool &Appears);

  virtual XxxErrCode ResolveVars (const XxxRecordSchema *Schema);

  virtual XxxErrCode ResolveVars(XxxQueryNode* Node, 
				 char Source);
  /// to resolve correlation variables
  virtual XxxErrCode ResolveVars(const XxxArgEnv* Env);

  virtual XxxErrCode TypeCheck()
   { return ((BaseType != XXX_UNKNOWN)? XXX_OK : XXX_NOT_OK);}

  virtual XxxErrCode Optimize(XxxExprPlan*& ExecPlan);
  // needed for variables that correspond to complex data types
  virtual XxxADTMetaInfo *GenerateCatInfo() const { return typeCatInfo;}
  /// Checks to see if the expression matches another expression
  virtual XxxBool Match(const XxxExpression* OtherExpr, 
			char *Env = NULL) const;
  /**
   * Gets input stream dependencies (that were resolved in ResolveVars)
   * Sets flags in the input argument DependsArray. The
   * array must be initialized to be all false. The size of
   * the array is 1+ the actual number of input streams.
   * The extra entry stands for the run-time Env
   */
  virtual XxxErrCode GetDependencies(XxxBitMap& Dependendency,
				     int CorrelationHt) const;
 };

/**
 * The XxxArgEnv supports two kinds of arguments to a (sub)query 
 *  1) "Host" variables that are bound as run-time arguments 
 *     They appear in a query as $1, $2, etc.
 *  2) Correlation values
 * During type checking, there must be a phase in which a list of all
 * such values is extracted and maintained. 
 * At run time, they will get instantiated from the execution environment
 */
class XxxArgEnv {
 public:
  /// Class used as an environment for correlations
  class NestingParseEnv { 
  public:
    int NumChildren;
    XxxQueryNode** Children;
    const XxxArgEnv* Env;

    NestingParseEnv() { Init(0,NULL,NULL);}
    void Init(int NChildren, XxxQueryNode** ChildNodes,
	      const XxxArgEnv* ArgEnv) {
      NumChildren = NChildren;
      Children = ChildNodes;
      Env = ArgEnv;
    }
  };

  // list of all "host" variables
  XxxValueExprList* HostExprList;
  int ExprCount;

  // Nesting environment (for correlations)
  NestingParseEnv NPEnv;
  
  XxxValueExpression* operator[](int index) const {
    if (HostExprList)
      return (*HostExprList)[index];
    else return NULL;
   }
  XxxArgEnv(int Count, XxxValueExprList* HostArgs) {
    ExprCount = Count;
    HostExprList = HostArgs;
   }
 };

/**
 * A structure which is used store values. Typically used in
 * the evaluatuon of functions (to store arguments) and to
 * pass host variables.
 */
class XxxValueEnv {
 public:
  /// Class used as an environment for correlations
  class NestingEvalEnv {  
  public:
    const XxxRecord** InputRecords;
    const XxxRecordSchema** InputSchemas;
    const XxxValueEnv* Env;
    XxxBool Valid;

    NestingEvalEnv() { Init(NULL,NULL,NULL); Valid = XXX_FALSE;}
    void Init(const XxxRecord* IRecords[],
	      const XxxRecordSchema* ISchemas[],
	      const XxxValueEnv* IEnv) {
      InputRecords = IRecords;
      InputSchemas = ISchemas;
      Env = IEnv;
      Valid= XXX_TRUE;
    }
   };
  XxxValuesArray ArgVals;
  XxxADTValue **ArgValsPtrs;
  int OwnerSize;
  XxxADTValue *OwnerValPtr;
  int NumArgs;
  XxxADTMetaInfo *OwnerCatInfo;
  NestingEvalEnv NEEnv;

  XxxADTValue* operator[](int index) const {
    if (ArgValsPtrs && index < NumArgs)
      return ArgValsPtrs[index];
    else return NULL;
   }
  XxxValueEnv(XxxValueExprList* HostExprs, int OwnerSize, 
	      XxxADTMetaInfo *OwnerTypeCatInfo);
  XxxValueEnv(const XxxValueEnv &Env);
  virtual XxxValueEnv * Replicate() const
    { return new XxxValueEnv(*this); }
  virtual ~XxxValueEnv();

 };

/**
 */
class XxxArgValue : public XxxValueExpression {
  friend class ArgValuePlan;
 protected:
  ////
  class ArgValuePlan : public XxxValueExprPlan { public:
    int ArgIndex;
    XxxADTMetaInfo *ArgCatInfo;
    ArgValuePlan(XxxArgValue* AExpr, XxxValueExprPlan* SubPlan) 
      : XxxValueExprPlan(AExpr, &SubPlan)
     { ArgIndex = AExpr->ArgIndex; }
    ArgValuePlan(const ArgValuePlan &Plan) :
      XxxValueExprPlan(Plan), ArgIndex(Plan.ArgIndex), ArgCatInfo(NULL)
    { 
      if (Plan.ArgCatInfo && (Plan.ArgCatInfo->Replicate(ArgCatInfo))) {
	ArgCatInfo = NULL;
      }
    }
    virtual XxxGenericPlan * Replicate() const {
      return (XxxGenericPlan *)(new ArgValuePlan(*this));
    }    
    virtual ~ArgValuePlan() {}
    virtual XxxErrCode Evaluate(const XxxRecord* Rec[],
				const XxxRecordSchema* Schema[],
				XxxADTValue *Result,
				const XxxValueEnv* Env = NULL) const;
    virtual XxxErrCode RedirectDependencies(const VarRemapping& Remap,
					    int CorrelationHt = 0)
     { return ((CorrelationHt==0)?changeDependency(Remap):XXX_OK);}

   };

 public:
  int ArgIndex;
  XxxValueExpression* Expr;
  XxxArgValue(int i) 
   { ArgIndex  = i; BaseType = XXX_UNKNOWN; }
  XxxArgValue(const XxxArgValue& src);
  /// Provides a convenient way to copy arbitrary expressions
  virtual XxxExpression * Replicate() const
    { return new XxxArgValue(*this); }
  virtual ValExprType ValType() const {return HOST_VAR;}
  virtual void Print(ostream&) const;
  // ArgValues are treated as constants
  virtual XxxBool IsConstant() const { return XXX_TRUE; }
  virtual XxxErrCode ResolveVars(const XxxArgEnv* Env);
  virtual XxxErrCode TypeCheck()
   { return ((Expr && (BaseType != XXX_UNKNOWN))? XXX_OK : XXX_NOT_OK);}
  virtual XxxADTMetaInfo *GenerateCatInfo() const
   { return (Expr?Expr->GenerateCatInfo():0);}
  // set the dependency on the ValueEnv
  virtual XxxErrCode GetDependencies(XxxBitMap& Dependency,
				     int CorrelationHt) const {
    Dependency.SetBit(0);
    return XXX_OK;
   }
  virtual XxxErrCode Optimize(XxxExprPlan*& ExecPlan);
  /// Checks to see if the expression matches another expression
  virtual XxxBool Match(const XxxExpression* OtherExpr, 
			char *Env = NULL) const;
 };

/**
 * Parent class of all constant expressions.
 * @author Praveen
 */
class XxxConstValue : public XxxValueExpression {
 public:
  ////
  class ConstPlan : public XxxValueExprPlan { 
   public:
    ConstPlan(XxxConstValue* CV) 
      : XxxValueExprPlan(CV, 0) {}
    ConstPlan(const ConstPlan &Plan) : XxxValueExprPlan(Plan) {}
    virtual XxxGenericPlan * Replicate() const = 0;
    virtual ~ConstPlan() {}
    // default here for constant values
    virtual XxxErrCode RedirectDependencies(const VarRemapping& Remap,
					    int CorrelationHt = 0)
     { return ((CorrelationHt==0)?changeDependency(Remap):XXX_OK);}
   };

  XxxConstValue() : XxxValueExpression() { }
  XxxConstValue(const XxxConstValue &src) : XxxValueExpression(src) { }
  /// This should never be called
  virtual XxxExpression * Replicate() const { ASSERT(1 == 0); return NULL;}
  virtual ValExprType ValType() const {return CONST_EXPR;}
  virtual XxxBool IsConstant() const { return XXX_TRUE; }
  virtual XxxErrCode TypeCheck() { return XXX_OK; }
  /// Checks to see if the expression matches another expression
  virtual XxxBool Match(const XxxExpression* OtherExpr, 
			char *Env = NULL) const;
  /**
   *  Checks if the expression matches another rule input expression 
   *  If match is successful, return true, and assign BindTerm array
   *  inside XxxBindingEnv class.
   */
  virtual XxxBool Match(const XxxExpression* RuleExpr, 
			XxxBindingEnv *BindEnv) const;
};

/**
 */
class XxxConstIntValue : public XxxConstValue {
 protected:  
  ////
  class ConstIntPlan : public XxxConstValue::ConstPlan { public:
    int Val;
    ConstIntPlan(XxxConstIntValue* CInt) : XxxConstValue::ConstPlan(CInt) 
     {Val = CInt->i;}
    ConstIntPlan(const ConstIntPlan &Plan) : XxxConstValue::ConstPlan(Plan), 
       Val(Plan.Val) {}
    virtual XxxGenericPlan * Replicate() const {
      return (XxxGenericPlan *)(new ConstIntPlan(*this));
    }    
    virtual ~ConstIntPlan() {}
    virtual XxxErrCode Evaluate(const XxxRecord* Rec[],
				const XxxRecordSchema* Schema[],
				XxxADTValue *Result,
				const XxxValueEnv* Env = NULL) const
				 {
				  ((XxxCoreADTClass::IntValue *)Result)->Value = Val;
				  return XXX_OK;
				 }
  };
 public:
  int i;
  XxxConstIntValue(int val) : XxxConstValue()
    { i = val; BaseType = XXX_INT;}
  XxxConstIntValue (const XxxConstIntValue &src) : XxxConstValue(src)
    { i = src.i; }
  /// Provides a convenient way to copy arbitrary expressions
  virtual XxxExpression * Replicate() const
    { return new XxxConstIntValue(*this); }
  virtual void Print(ostream&) const;
  virtual XxxErrCode Optimize(XxxExprPlan*& ExecPlan)
   { ExecPlan = new ConstIntPlan(this); return XXX_OK;}
  /// Checks to see if the expression matches another expression
  virtual XxxBool Match(const XxxExpression* OtherExpr, 
			char *Env = NULL) const;
  double GetDblValue() const { return (double)i;}
 };

/**
 */
class XxxConstDoubleValue : public XxxConstValue {
 protected:  
  ////
  class ConstDoublePlan : public XxxConstValue::ConstPlan { public:
    double Val;
    ConstDoublePlan(XxxConstDoubleValue* CDoub) : XxxConstValue::ConstPlan(CDoub) 
     {Val = CDoub->d;}
    ConstDoublePlan(const ConstDoublePlan &Plan) : XxxConstValue::ConstPlan(Plan), 
       Val(Plan.Val) {}
    virtual XxxGenericPlan * Replicate() const {
      return (XxxGenericPlan *)(new ConstDoublePlan(*this));
    }    
    virtual ~ConstDoublePlan() {}
    virtual XxxErrCode Evaluate(const XxxRecord* Rec[],
				const XxxRecordSchema* Schema[],
				XxxADTValue *Result,
				const XxxValueEnv* Env = NULL) const
				 {
				  ((XxxCoreADTClass::DblValue *)Result)->Value = Val;
				  return XXX_OK;
				 }
   };
 public:
  double d;
  XxxConstDoubleValue(double val)  : XxxConstValue()
   { d = val; BaseType = XXX_DOUBLE;}
  XxxConstDoubleValue (const XxxConstDoubleValue &src) : XxxConstValue(src)
    { d = src.d; }
  /// Provides a convenient way to copy arbitrary expressions
  virtual XxxExpression * Replicate() const
    { return new XxxConstDoubleValue(*this); }
  virtual void Print(ostream&) const;
  virtual XxxErrCode Optimize(XxxExprPlan*& ExecPlan)
   { ExecPlan = new ConstDoublePlan(this); return XXX_OK;}

  /// Checks to see if the expression matches another expression
  virtual XxxBool Match(const XxxExpression* OtherExpr, 
			char *Env = NULL) const;
  double GetDblValue() const { return d;}
 };

/**
 */
class XxxConstStringValue : public XxxConstValue {
 protected:  
  ////
  class ConstStringPlan : public XxxConstValue::ConstPlan { public:
    char Val[XXX_TEXTSTRING_MAX+1];
    ConstStringPlan(XxxConstStringValue* CStr) : XxxConstValue::ConstPlan(CStr) { 
      strncpy(Val, CStr->c, XXX_TEXTSTRING_MAX);
      Val[XXX_TEXTSTRING_MAX] = 0;
     }
    ConstStringPlan(const ConstStringPlan &Plan) : XxxConstValue::ConstPlan(Plan) {
      strncpy(Val, Plan.Val, XXX_TEXTSTRING_MAX);
      Val[XXX_TEXTSTRING_MAX] = 0;
    }
    virtual XxxGenericPlan * Replicate() const {
      return (XxxGenericPlan *)(new ConstStringPlan(*this));
    }    
    virtual ~ConstStringPlan() {}
    virtual XxxErrCode Evaluate(const XxxRecord* Rec[],
				const XxxRecordSchema* Schema[],
				XxxADTValue *Result,
				const XxxValueEnv* Env = NULL) const
				 {
				  strncpy(((XxxCoreADTClass::StrValue *)Result)->Value, 
					  Val, 
					  XXX_TEXTSTRING_MAX);
				  return XXX_OK;
				 }
   };
 public:
  char c[XXX_TEXTSTRING_MAX+1];
  XxxConstStringValue(const char *val) : XxxConstValue() 
  { 
    strncpy(c,val, XXX_TEXTSTRING_MAX); 
    c[XXX_TEXTSTRING_MAX] = 0;
    BaseType = XXX_STRING;
  }
  XxxConstStringValue(const XxxConstStringValue &src) : XxxConstValue(src) { 
    strncpy(c, src.c, XXX_TEXTSTRING_MAX); 
    c[XXX_TEXTSTRING_MAX] = 0;
   } 
  /// Provides a convenient way to copy arbitrary expressions
  virtual XxxExpression * Replicate() const
    { return new XxxConstStringValue(*this); }
  virtual void Print(ostream&) const;
  virtual XxxErrCode Optimize(XxxExprPlan*& ExecPlan)
   { ExecPlan = new ConstStringPlan(this); return XXX_OK;}
  /// Checks to see if the expression matches another expression
  virtual XxxBool Match(const XxxExpression* OtherExpr, 
			char *Env = NULL) const;
 };

#endif
