#include "record.h"
#include "shoresm.h"
#include "prd-sort.h"
#include "exprplan.h"
#include "values.h"

//Bjarke feb 1 2002: commented out ifdef, since we need in in windows as well
//#ifdef PRED_LINUX
#include <new.h>
//#endif

//
// Methods of XxxSortInfo
//
void XxxSortInfo::Init(const XxxSortInfo& SInfo)
{
  if (SInfo.numOrderedAttrs <= 0) {
    numOrderedAttrs = 0;
    orderedAttrs = NULL;
  } else {
    numOrderedAttrs = SInfo.numOrderedAttrs;
    orderedAttrs = new OrderedAttrInfo[numOrderedAttrs];
    unsigned int i;
    for (i = 0; i < numOrderedAttrs; i++) {
      orderedAttrs[i] = SInfo.orderedAttrs[i];
    }
  }
}

void XxxSortInfo::Init(unsigned int NumAttrs)
{
 numOrderedAttrs = NumAttrs;
 if (NumAttrs <= 0)
   orderedAttrs = NULL;
 else 
   orderedAttrs = new OrderedAttrInfo[NumAttrs];
}

XxxBool 
XxxSortInfo::Matches(const XxxSortInfo& SInfo) const
{
  XxxBool Matches = XXX_TRUE;
  
  if (numOrderedAttrs != SInfo.numOrderedAttrs) {
    return XXX_FALSE;
  }
  
  for (unsigned int i = 0; i < numOrderedAttrs; i++) {
    if ((orderedAttrs[i].AttrIndex != SInfo.orderedAttrs[i].AttrIndex) ||
	(orderedAttrs[i].Asc != SInfo.orderedAttrs[i].Asc)) {
      return XXX_FALSE;
    }
  }
  return XXX_TRUE;
}

XxxBool 
XxxSortInfo::interestingMatches(const XxxSortInfo& SInfo) const
{
  XxxBool Matches = XXX_TRUE;
  unsigned int matchAttr = 0;

  if (numOrderedAttrs < SInfo.numOrderedAttrs ) {
    return XXX_FALSE;
  }
  
  for (unsigned int i = 0; i < numOrderedAttrs && i < SInfo.numOrderedAttrs; i++) {
	  if ((orderedAttrs[i].AttrIndex != SInfo.orderedAttrs[i].AttrIndex) ||
		  (orderedAttrs[i].Asc != SInfo.orderedAttrs[i].Asc)) {
		  break;
	  }
	  else
		matchAttr++;
  }

  if ( matchAttr == SInfo.numOrderedAttrs ) 
	return XXX_TRUE;
  else
	return XXX_FALSE;

}

XxxSortInfo::OrderedAttrInfo & XxxSortInfo::operator[](int index) const
{
  return orderedAttrs[index];
}

/**
 * Returns a count of the number of attributes in the prefix of the
 * XxxSortInfo that are also found as XxxUnknownValues in the TargetList.
 * This is used to detemine sort orders for relational operations.
 * The order of the attrs in the grouping list matters, since that
 * corresponds to the sort order.
 */
int XxxSortInfo::CommonAttrs(const XxxValExprPlanList* TargetList,
			     int InputSrc) const
{
 int OutputSortedAttrs = 0;
 unsigned int i;
 for (i=0; i < NumOrderedAttrs(); i++) {
   const XxxValExprPlanList *TList = TargetList;
   while (TList) {
     if (TList->Plan->ValType() == XxxValueExpression::COLUMN_VAR) {
       XxxUnknownValue::UnknownValuePlan *UPlan =
	 (XxxUnknownValue::UnknownValuePlan *)(TList->Plan);
       if ((UPlan->SrcAttrIndex == orderedAttrs[i].AttrIndex) &&
	   (UPlan->SrcChild == InputSrc)) {
	 OutputSortedAttrs++;
	 break;
	}
      }
     TList = TList->Next;
    } // while TList
   if (!TList) break;
  } // for each SInfo attr
 return OutputSortedAttrs;
}

XxxErrCode XxxSortInfo::FillTarget(const XxxValExprPlanList* TargetList,
				   XxxSortInfo& OutSInfo,
				   int InputSrc) const
{
 if (NumOrderedAttrs() < OutSInfo.NumOrderedAttrs()) {
   XXX_NOTE_ERROR(XXX_INTERNAL, "Bad call to XxxSortInfo::FillTarget");
   return XXX_NOT_OK;
  }
 unsigned int i;
 for (i=0; i < OutSInfo.NumOrderedAttrs(); i++) {
   const XxxValExprPlanList *TList = TargetList;
   int j = 0;
   while (TList) {
     if (TList->Plan->ValType() == XxxValueExpression::COLUMN_VAR) {
       XxxUnknownValue::UnknownValuePlan *UPlan =
	 (XxxUnknownValue::UnknownValuePlan *)(TList->Plan);
       if ((UPlan->SrcAttrIndex == orderedAttrs[i].AttrIndex) &&
	   (UPlan->SrcChild == InputSrc)) {
	 OutSInfo[i].AttrIndex = j;
	 OutSInfo[i].Asc = orderedAttrs[i].Asc;
	 break;
	}
      }
     TList = TList->Next;
     j++;
    } // while TList
   if (!TList) {
     XXX_NOTE_ERROR(XXX_INTERNAL, "Sort attr expected but not found");
     return XXX_NOT_OK;
    }
  } // for each OutSInfo attr
 return XXX_OK;
}

int XxxCompareFunc(size_type size1,
		   const void *Rec1,
		   size_type size2,
		   const void *Rec2, 
		   const void *ExtraData)
{
  XxxRecord R1(size1, (char *)Rec1);
  XxxRecord R2(size2, (char *)Rec2);
  const XxxRecordSchema *Schema = 
    ((const XxxCompData *)ExtraData)->Schema;
  const XxxSortInfo *SortInfo = 
    ((const XxxCompData *)ExtraData)->SortInfo;
  //Bjarke, sep. 13 2001: added const qualifier to variables
  const char *Fld1, *Fld2;
  XxxADTValue *A1, *A2;
  unsigned int i;
  for (i = 0; i < SortInfo->NumOrderedAttrs(); i++) {
    // Get the ordering attributes
    R1.GetField((*SortInfo)[i].AttrIndex, Schema, Fld1);
    R2.GetField((*SortInfo)[i].AttrIndex, Schema, Fld2);
    A1 = (XxxADTValue*)Fld1;
    A2 = (XxxADTValue*)Fld2;
    XxxBasicType TypeId = (*Schema)[(*SortInfo)[i].AttrIndex].Type;
    XxxADTMetaInfo* MInfo = (*Schema)[(*SortInfo)[i].AttrIndex].TypeCatInfo;
    if (XxxCoreADTClass::IsCoreADT(TypeId)) {
     if (XxxCoreADTClass::
         SGreaterThan(TypeId, A1, MInfo, A2, MInfo))
	return ((*SortInfo)[i].Asc ? 1 : -1);
     else if (XxxCoreADTClass::
         SLessThan(TypeId, A1, MInfo, A2, MInfo))
	return ((*SortInfo)[i].Asc ? -1 : 1);
     else continue;
     }
    else {
      XxxADTClass *ADT = XxxADTArray[TypeId];
      if (ADT->OrderedType(MInfo)) {
	if (ADT->GreaterThan(A1, MInfo, A2, MInfo))
	  return ((*SortInfo)[i].Asc ? 1 : -1);
	else if (ADT->LessThan(A1, MInfo, A2, MInfo))
	  return ((*SortInfo)[i].Asc ? -1 : 1);
	else continue;
      } else {
	// NOTE: This code should never be executed.
	// We need to make sure we get equality right for distinct sorts
	if (ADT->Equals(A1, MInfo, A2, MInfo))
	  continue;
	else
	  // This is to insure repeatable arbitrary ordering
	  return strncmp((char*)A1,(char*) A2, ((size1 < size2) ? size1 : size2));
      }
    }
  }
  return 0;
}


#ifdef XXX_SHORE_V2
int XxxCompareFuncADT(size_type size1,
		   const void *Rec1,
		   size_type size2,
		   const void *Rec2)
{
	// obtain type id
	int TypeId1;
	int TypeId2;

	memcpy((void*)&TypeId1, Rec1, sizeof(int));
	memcpy((void*)&TypeId2, Rec2, sizeof(int));

	ASSERT((TypeId1 == TypeId2));

	// obtain ADT values
	XxxADTValue* A1 = (XxxADTValue*)((char*)Rec1+sizeof(int));
	XxxADTValue* A2 = (XxxADTValue*)((char*)Rec2+sizeof(int));

	// compare
	XxxADTClass *ADT = XxxADTArray[TypeId1];
    if (ADT->OrderedType(NULL)) 
	{
		if (ADT->GreaterThan(A1, NULL, A2, NULL))
		{
			return 1;
		}
		else if (ADT->LessThan(A1, NULL, A2, NULL))
		{
			return -1;
		}
		else
		{
			return 0;
		}
    } else {
		if (ADT->Equals(A1, NULL, A2, NULL))
		{
			return 0;
		}
		else
		{
		  // This is to insure repeatable arbitrary ordering
		 return strncmp((char*)A1,(char*) A2, ((size1 < size2) ? size1 : size2));
		}
	}

	return 0;
}


XxxShoreRetVal XxxKeyExtractionFunction(const XxxShoreRecId&   rid,
		const XxxShoreSortObject&	in_obj,
		XxxShoreSortKeyData	cookie,  
		XxxShoreFactory&	factory,
		XxxShoreSortKey*	skey)
{
	// get record from in_obj
	XxxRecord R(in_obj.contig_body_size(), (char *) in_obj.body(0/*offset*/));

	// get key info from cookie
	const XxxRecordSchema *Schema = 
		((XxxCompData)((XxxKeyCompData*)cookie)->keyInfo).Schema;
	const XxxSortInfo *SortInfo = 
		((XxxCompData)((XxxKeyCompData*)cookie)->keyInfo).SortInfo;
	int i = ((XxxKeyCompData*)cookie)->keyIndex;
	// get key value
	//Bjarke, sep. 13: added const qualifier to variable
	const char *Fld;
	XxxADTValue *Val;
	 
    R.GetField((*SortInfo)[i].AttrIndex, Schema, Fld);
    Val	= (XxxADTValue*) Fld;

	// create shore key
	XxxBasicType TypeId = (*Schema)[(*SortInfo)[i].AttrIndex].Type;
 
	
    if (XxxCoreADTClass::IsCoreADT(TypeId)) {

      new (skey) skey_t((void*) Val, /* offset*/ XxxCoreADTClass::ValueOffset(TypeId), /* length of the key*/ XxxCoreADTClass::STypeSize(TypeId, Val, NULL) - XxxCoreADTClass::ValueOffset(TypeId), factory);
	

	}
	else
	{
		// skey is composed of typeid (int), ADT value
		// TODO: mechanism to pass Meta Info! Preferably, with a new signature of the
		// SHORE comparison function to include MetaInfo
		// For the moment meta info is assumed to be null here and in XxxComparisonFuncADT (see above)
		int BufSize = sizeof(int) + XxxADTArray[TypeId]->TypeSize(Val, NULL);
		char* Buf = new char[BufSize];
		char* TmpBuf;

		TmpBuf = Buf;
		memcpy(TmpBuf, (char*)&TypeId, sizeof(int));
		TmpBuf += sizeof(int);
		XxxADTArray[TypeId]->TypeCopy(Val, (XxxADTValue*)TmpBuf, NULL);


		new (skey) skey_t((void*) Buf, /* offset*/ 0, /* length of the key*/ BufSize, factory);

	}

	return RCOK;
}

#endif
