#include "predconnect.h"

//
// Methods of PredConnection
//
const char *PredConnection::initString = "NEW SESSION\n";
PredConnection::PredConnection(const char* HostName, int PortNumber,
			       const char* UserName, const char* Password,
			       bool Binary, bool GetSchemas)
{
 init();
 binaryMode = Binary;
 getSchemas = GetSchemas;
 portNumber = PortNumber;
 hostName = new char[strlen(HostName)+1];
 strcpy(hostName, HostName);
 userName = new char[strlen(UserName)+1];
 strcpy(userName, UserName);
 password = new char[strlen(Password)+1];
 strcpy(password, Password);

 socket = new XxxSocketConnection(hostName, portNumber);
 if (!socket->Valid()) {
   cerr << "Did not make connection to server" << endl;
   return;
  }

 // set up protocol
 if (!setupProtocol()) {
   cerr << "Could not set up protocol" << endl;
   return;
  }
 valid = true;
}

PredConnection::~PredConnection()
{
 // expect the server to close out the protocol with a \0
 if (valid) {
   if (!terminateProtocol()) {
     cerr << "No clean shutdown from server" << endl;
    }
  }
 reclaim();
}

int PredConnection::setupProtocol()
{
 if (!socket) {
   cerr << "Invalid connection" << endl;
   return 0;
  }
 int SocketDesc = socket->socketDesc;
 // Send the string which requests a new client session
 if (::send(SocketDesc, initString, strlen(initString),0) != strlen(initString)) {
   cerr << "Bad send" << endl;
   return 0;
 }

 char Buf[500];
 int BytesRead;

 if ((BytesRead = ::recv(SocketDesc, Buf, 500, 0)) <= 0) {
   cerr << "Bad Read" << endl;
   return 0;
  }
 if (strncmp(Buf, "Setup?\n", BytesRead)) {
   cerr << "Unexpected protocol string from server:" << Buf << endl;
   cerr << "Bytes Read: " << BytesRead << endl;
   return 0;
  }
 // TODO: We want to modify this to NOT use 'txtclient' method of interfacing
 // this is because is is textual and we want binary for enhanced control over
 // the data returned by the server.

 // now send back a string with initial info
 sprintf (Buf, "%s %s %s ", userName, password, // passwd
	  "txtclient"); // protocolname
 if (::send(SocketDesc, Buf, strlen(Buf)+1, 0) != strlen(Buf)+1) {
   cerr << "Bad send" << endl;
   return 0;
 }

 // now handshake on the options
 if ((BytesRead = ::recv(SocketDesc, Buf, 500, 0)) <= 0) {
   cerr << "Bad Read" << endl;
   return 0;
  }
 if (strncmp(Buf, "Options?\n", BytesRead)) {
   cerr << "Unexpected protocol string from server:" << Buf << endl;
   cerr << "Bytes Read: " << BytesRead << endl;
   return 0;
  }
 sprintf (Buf, "-s %d -b %d", (getSchemas?1:0), (binaryMode?1:0));
 if (::send(SocketDesc, Buf, strlen(Buf), 0) != strlen(Buf)) {
   cerr << "Bad send" << endl;
   return 0;
 }

 // now get back a "good" or a "bad"
 if ((BytesRead = ::recv(SocketDesc, Buf, 500, 0)) <= 0) {
   cerr << "Bad Read" << endl;
   return 0;
  }
 if (!strncmp(Buf, "good", BytesRead))
   return 1;
 else {
   cerr << "Setup failed at server" << endl;
   return 0;
  }
}

bool PredConnection::requestTerminated(const char*Buf, int BytesRead) const
{
 if (!valid) {
   cerr << "Invalid connection" << endl;
   return true;
  }
 // answers terminated by \0
 if (!Buf[BytesRead-1]) { 
   return true;
  }
 return false;
}

int PredConnection::terminateProtocol()
{
 if (!valid) {
   cerr << "Invalid connection" << endl;
   return 0;
  }
 int SocketDesc = socket->socketDesc;
 char Buf[100];
 int BytesToRead = 0; // signal EOF
 if (::send(SocketDesc, (char*)&BytesToRead, sizeof(BytesToRead), 0) !=
     sizeof(BytesToRead)) {
   cerr << "Bad send" << endl;
   return 0;
 }

 if (::recv(SocketDesc, Buf, 100, 0) <= 0) {
   cerr << "Bad recv" << endl;
   return 0;
  }
 return 1;
}

int PredConnection::SendRequest(const char* Buf)
{
 if (!valid) {
   cerr << "Invalid connection" << endl;
   return 0;
  }
 int SocketDesc = socket->socketDesc;
 // first send the length
 int BufLength = strlen(Buf)+1;
 int NetworkBufLength = htonl(BufLength);
 if (::send(SocketDesc, (char*)&NetworkBufLength, sizeof(int), 0) != sizeof(int)) {
   cerr << "Error writing data length" << endl;
   valid = false;
   return 0;
  }
 // then send the data
 if (::send(SocketDesc, Buf, strlen(Buf)+1, 0) != (int)(strlen(Buf)+1)) {
   cerr << "Error writing data" << endl;
   valid = false;
   return 0;
  }
 return 1;
}

int PredConnection::PrintReply(ostream& OStr)
{
 if (!valid) {
   cerr << "Invalid connection" << endl;
   return 0;
  }
 int SocketDesc = socket->socketDesc;
 char Buf[RESULT_BUFFER_MAX+1];
 int BytesRead;

 while (1) {
   BytesRead = ::recv(SocketDesc, Buf, RESULT_BUFFER_MAX, 0);
   if (BytesRead <= 0) {
     perror("Server error:");
     valid = 0;
     return 0;
    }
   Buf[BytesRead] = 0;
   OStr << Buf;
   OStr << flush;

   if (requestTerminated(Buf, BytesRead))
     break;
  } // while (1)
 return 1;
}

ReplyScan* PredConnection::CreateReplyScan()
{
 if (!valid) {
   cerr << "Invalid connection" << endl;
   return NULL;
  }
 int SocketDesc = socket->socketDesc;
 char Buf[RESULT_BUFFER_MAX+1];
 int BytesRead;
 ReplyScan* replyScan = new ReplyScan();

 while (1) {
   BytesRead = ::recv(SocketDesc, Buf, RESULT_BUFFER_MAX, 0);
   if (BytesRead <= 0) {
     perror("Server error:");
     valid = 0;
     return NULL;
    }
   Buf[BytesRead] = 0;
   replyScan->AnsOStr << Buf;

   if (requestTerminated(Buf, BytesRead))
     break;
  } // while (1)
 replyScan->init();
 return replyScan;
}


bool PredConnection::CreateBufferScan(char* Buffer, bool &TimeOut, bool &AtEnd)
{
 if (!valid) {
   cerr << "Invalid connection" << endl;
   return false;
  }
 int SocketDesc = socket->socketDesc;
 char Buf[RESULT_BUFFER_MAX+1];
 int BytesRead;

 struct timeval Timeout = {0,50}; // 1 seconds time-out

 fd_set ReadSockets;
 FD_ZERO(&ReadSockets);
 FD_SET(SocketDesc, &ReadSockets);

 // wait until a proxy server sends data
 int selval = select(1, &ReadSockets, NULL, NULL, &Timeout);
		
 if (selval < 0) {
	 return false;
 } else if (selval == 0) {
	 // time out
	 strcpy(Buffer, "<I></I>");
	 TimeOut = true;
	 AtEnd = false;
	 return true;

 } else {
	TimeOut = false;
	BytesRead = ::recv(SocketDesc, Buf, RESULT_BUFFER_MAX, 0);
	if (BytesRead <= 0) {
		perror("Server error:");
		valid = 0;
		return false;
	}
	Buf[BytesRead] = 0;

	if (requestTerminated(Buf, BytesRead)) {
		AtEnd = true;
	} else {
		AtEnd = false;
	}
	 
	strcpy(Buffer, Buf);
	return true;
 }

}





ReplyScan::ReplyScan()
{
  ptr = NULL;
  atEnd = true;
  PrintedAlready = false;
}

ReplyScan::~ReplyScan()
{
  if (ptr) delete [] ptr;
}

void ReplyScan::init()
{
  ptr = AnsOStr.str();
  ptr[AnsOStr.pcount()] = 0;
  if (!(*ptr)) {
    atEnd = true;
  }else{
    atEnd = false;
  }
}

char ReplyScan::GetNextChar()
{
  return(*ptr++);
  PrintedAlready = true;
}

bool ReplyScan::IsError()
{
  char *where = strstr(ptr, "Error: Error processing user request");
  if(!where)   return false;
  return true;
}

bool ReplyScan::AnyResults()
{
  char *where = strchr(ptr, '\t');
  if(!where)   return false;
  return true;
}

void ReplyScan::PrintSchema(ostream& OStr)
{
  char *where1 = strchr(ptr, '<');
  char *where2 = strchr(ptr, '>');
  char *where3 = strchr(ptr, ':');
  if (!PrintedAlready && where1 && where2 && where3) {
    PrintedAlready = true;
    while ((*ptr != '<') && (*ptr)) ptr++; // gets you to the first char safely
    if (!(*ptr)) {
      OStr << "<td><center><b>SOMETHING WRONG</B></CENTER></TD>"<< endl;
      return;
    }
    OStr << "<td><center><font size+=2 color = \"red\"><b>";
    ptr++;
    bool InParen = false;
    while (*ptr) { // loop through the stream, printing out the schema
      if (*ptr == '>') {
	OStr << "</b></font></center></td>";
	ptr++;
	break;
      }
      if (*ptr == '(') {
	InParen = true;
      }
      if (*ptr == ')') {
	InParen = false;
      }
      if (*ptr == ',' && !InParen) {
	OStr << "</b></font></center></td><td><center><font size+=2 color = \"red\"><b>";
	ptr++;
      } 
      else {
	OStr << *ptr++;
      }
    } // while
  } // for
}
// Answers come back separated by tabs, so the functions following GetNextItem
// show how messy it is to use a textual interface to PREDATOR.  Make it binary
// and this code cleans up very, very quickly.
int ReplyScan::GetNextItem(char* first)
{
  PrintedAlready = true;
  char *where = strchr(ptr, '\t');
  if(!where)   return 0;
  strncpy(first,ptr,where-ptr);
  first[where-ptr] = 0;
  ptr += (where-ptr+1);
  if (!(*ptr)) {
    atEnd = true;
  }else{
    atEnd = false;
  }
  return 1;
}

void ReplyScan::GetNextLine(ostream& OStr)
{
  PrintedAlready = true;
  if(!(*ptr)){
    atEnd = true;
  }else{
    bool hitQuote = false;
    while (1) {
      if(!(*ptr)){
	atEnd = true;
	OStr << 0;
	break;
      }
      if((*ptr == '\n') && (!hitQuote)) {
	OStr << 0;
	break;
      }
      if((*ptr == '\t') && (!hitQuote)){
      }else{
	if(*ptr == '\"'){
	  hitQuote = !hitQuote;
	}else{
	  OStr.put(*ptr);
	}
      }
      ptr++;
    }
    ptr++;
  }
}

void ReplyScan::GetNextLineFormatted(ostream& OStr)
{
  PrintedAlready = true;
  if (!(*ptr)) {
    atEnd = true;
  } 
  else {
    bool hitQuote = false;
    bool firstTime = true;
    bool printTD = false;
    while (1) {
      if (!(*ptr)) {
	atEnd = true;
	break;
      }
      if ((*ptr == '\n') && !hitQuote) {
	break;
      }
      if ((*ptr == '\t') && !hitQuote){
	printTD = true;
      } 
      else {
	if ((printTD) && !firstTime) {
	  OStr << "</CENTER><TD><CENTER>";
	  printTD = false;
	}
	if (firstTime) {
	  OStr << "<TD><CENTER>";
	  firstTime = false;
	}
	if ((*ptr == 'h') && (*(ptr+1) == 't') && (*(ptr+2) == 't' ) && (*(ptr+3) == 'p') && (*(ptr+4) == ':') && (*(ptr+5) == '/') && (*(ptr+6) == '/')) {
	  OStr << "<A TARGET=\"Document\" HREF=\"";
	  while((*ptr != '\t') && (*ptr != '\n') && (*ptr)) {
	    OStr.put(*ptr);
	    ptr++;
	  }
	  OStr << "\">View</A></CENTER><TD><CENTER>";
	} // if http value
	else{
	  if (*ptr == '\"') {
	    hitQuote = !hitQuote;
	  }
	  else{
	    OStr.put(*ptr);
	  }
	} // else if not http value
      } // else if !((*ptr == '\t') && (!hitQuote))
      ptr++;
      firstTime = false;
    }// while(1)
    ptr++;
    firstTime = false;
  }
}
